# ME-NeuralFMU from the Modelica Conference 2021
Tutorial by Johannes Stoljar, Tobias Thummerer

## License
Copyright (c) 2021 Tobias Thummerer, Lars Mikelsons, Johannes Stoljar

Licensed under the MIT license. See [LICENSE](https://github.com/thummeto/FMIFlux.jl/blob/main/LICENSE) file in the project root for details.

## Motivation
The Julia Package *FMIFlux.jl* is motivated by the application of hybrid modeling. This package enables the user to integrate his simulation model between neural networks (NeuralFMU). For this, the simulation model must be exported as FMU (functional mock-up unit), which corresponds to a widely used standard. The big advantage of hybrid modeling with artificial neural networks is, that the effects that are difficult to model (because they might be unknown) can be easily learned by the neural networks. For this purpose, the NeuralFMU is trained with measurement data containing the unmodeled physical effect. The final product is a simulation model including the orignially unmodeled effects. Another big advantage of the NeuralFMU is that it works with little data, because the FMU already contains the characterisitic functionality of the simulation and only the missing effects are added.

NeuralFMUs do not need to be as easy as in this example. Basically a NeuralFMU can combine different ANN topologies that manipulate any FMU-input (system state, system inputs, time) and any FMU-output (system state derivative, system outputs, other system variables). However, for this example a NeuralFMU topology as shown in the following picture is used.

![NeuralFMU.svg](https://github.com/thummeto/FMIFlux.jl/blob/main/docs/src/examples/pics/NeuralFMU.svg?raw=true)

*NeuralFMU (ME) from* [[1]](#Source).

## Introduction to the example
In this example, simplified modeling of a one-dimensional spring pendulum (without friction) is compared to a model of the same system that includes a nonlinear friction model. The FMU with the simplified model will be named *simpleFMU* in the following and the model with the friction will be named *realFMU*. At the beginning, the actual state of both simulations is shown, whereby clear deviations can be seen in the graphs. In addition, the initial states are changed for both models and these graphs are also contrasted, and the differences can again be clearly seen. The *realFMU* serves as a reference graph. The *simpleFMU* is then integrated into a NeuralFMU architecture and a training of the entire network is performed. After the training the final state is compared again to the *realFMU*. It can be clearly seen that by using the NeuralFMU, learning of the friction process has taken place.  


## Target group
The example is primarily intended for users who work in the field of first principle and/or hybrid modeling and are further interested in hybrid model building. The example wants to show how simple it is to combine FMUs with machine learning and to illustrate the advantages of this approach.


## Other formats
Besides this [Jupyter Notebook](https://github.com/thummeto/FMIFlux.jl/blob/main/example/modelica_conference_2021.ipynb) there is also a [Julia file](https://github.com/thummeto/FMIFlux.jl/blob/main/example/modelica_conference_2021.jl) with the same name, which contains only the code cells. For the documentation there is a [Markdown file](https://github.com/thummeto/FMIFlux.jl/blob/main/docs/src/examples/modelica_conference_2021.md) corresponding to the notebook.  


## Getting started

### Installation prerequisites
|    | Description                       | Command     |  Alternative  |   
|:--- |:---                              |:---         |:---|
|1.  | Enter Package Manager via         |     ]       |     |
|2.  | Install FMI via                   |   add FMI   | add " https://github.com/ThummeTo/FMI.jl "   |
|3.  | Install FMIFlux via               | add FMIFlux | add " https://github.com/ThummeTo/FMIFlux.jl " |
|4.  | Install FMIZoo via                | add FMIZoo  | add " https://github.com/ThummeTo/FMIZoo.jl " |
|5.  | Install Flux via                  |  add Flux   |     |
|6.  | Install DifferentialEquations via | add DifferentialEquations |  |
|7.  | Install Plots via                 | add Plots   |     |

## Code section

To run the example, the previously installed packages must be included. 


```julia
# imports
using FMI
using FMIFlux
using FMIZoo
using Flux
using DifferentialEquations: Tsit5
import Plots
```

    WARNING: Method definition (::Type{DiffEqFlux.DeterministicCNF{M, P, RE, D, T, A, K} where K where A where T where D where RE where P where M})(M, P, RE, D, T, A, K) where {M, P, RE, D, T, A, K} in module DiffEqFlux at deprecated.jl:70 overwritten at C:\Users\thummeto\.julia\packages\DiffEqFlux\A5156\src\ffjord.jl:41.
      ** incremental compilation may be fatally broken for this module **
    


After importing the packages, the path to the *Functional Mock-up Units* (FMUs) is set. The exported FMU is a model meeting the *Functional Mock-up Interface* (FMI) Standard. The FMI is a free standard ([fmi-standard.org](http://fmi-standard.org/)) that defines a container and an interface to exchange dynamic models using a combination of XML files, binaries and C code zipped into a single file. 

The object-orientated structure of the *SpringPendulum1D* (*simpleFMU*) can be seen in the following graphic and corresponds to a simple modeling.

![svg](https://github.com/thummeto/FMIFlux.jl/blob/main/docs/src/examples/pics/SpringPendulum1D.svg?raw=true)

In contrast, the model *SpringFrictionPendulum1D* (*realFMU*) is somewhat more accurate, because it includes a friction component. 

![svg](https://github.com/thummeto/FMIFlux.jl/blob/main/docs/src/examples/pics/SpringFrictionPendulum1D.svg?raw=true)

Next, the start time and end time of the simulation are set. Finally, a step size is specified to store the results of the simulation at these time steps.


```julia
tStart = 0.0
tStep = 0.01
tStop = 4.0
tSave = collect(tStart:tStep:tStop)
```


    401-element Vector{Float64}:
     0.0
     0.01
     0.02
     0.03
     0.04
     0.05
     0.06
     0.07
     0.08
     0.09
     ⋮
     3.92
     3.93
     3.94
     3.95
     3.96
     3.97
     3.98
     3.99
     4.0


### RealFMU

In the next lines of code the FMU of the *realFMU* model from *FMIZoo.jl* is loaded and instantiated.  


```julia
realFMU = fmiLoad("SpringFrictionPendulum1D", "Dymola", "2022x")
fmiInstantiate!(realFMU; loggingOn=false)
fmiInfo(realFMU)
```

    ┌ Info: fmi2Unzip(...): Successfully unzipped 29 files at `C:\Users\thummeto\AppData\Local\Temp\fmijl_Qpq0sv\SpringFrictionPendulum1D`.
    └ @ FMIImport C:\Users\thummeto\Documents\FMIImport.jl\src\FMI2_ext.jl:77
    ┌ Info: fmi2Load(...): FMU resources location is `file:///C:/Users/thummeto/AppData/Local/Temp/fmijl_Qpq0sv/SpringFrictionPendulum1D/resources`
    └ @ FMIImport C:\Users\thummeto\Documents\FMIImport.jl\src\FMI2_ext.jl:192
    ┌ Info: fmi2Load(...): FMU supports both CS and ME, using CS as default if nothing specified.
    └ @ FMIImport C:\Users\thummeto\Documents\FMIImport.jl\src\FMI2_ext.jl:195


    #################### Begin information for FMU ####################
    	Model name:			SpringFrictionPendulum1D
    	FMI-Version:			2.0
    	GUID:				{df491d8d-0598-4495-913e-5b025e54d7f2}
    	Generation tool:		Dymola Version 2022x (64-bit), 2021-10-08
    	Generation time:		2022-03-03T15:09:18Z
    	Var. naming conv.:		structured
    	Event indicators:		24
    	Inputs:				0
    	Outputs:			0
    	States:				2
    		33554432 ["mass.s"]
    		33554433 ["mass.v", "mass.v_relfric"]
    	Supports Co-Simulation:		true
    		Model identifier:	SpringFrictionPendulum1D
    		Get/Set State:		true
    		Serialize State:	true
    		Dir. Derivatives:	true
    		Var. com. steps:	true
    		Input interpol.:	true
    		Max order out. der.:	1
    	Supports Model-Exchange:	true
    		Model identifier:	SpringFrictionPendulum1D
    		Get/Set State:		true
    		Serialize State:	true
    		Dir. Derivatives:	true
    ##################### End information for FMU #####################


In the following two subsections, the *realFMU* is simulated twice with different initial states to show what effect the choice of initial states has.

#### Default initial states

The start and end time are set via the `fmiSetupExperiment()` function. In the next steps the initial states are set. The first state is the position of the mass, which is initilized with $0.5m$, the second state is the velocity, which is initialized with $0\frac{m}{s}$.   


```julia
fmiSetupExperiment(realFMU, tStart, tStop)
states = ["s0", "v0"]
x₀ = [0.5, 0.0]

fmiSetReal(realFMU, states, x₀)
fmiEnterInitializationMode(realFMU)
fmiExitInitializationMode(realFMU);
```

In the following code block the *realFMU* is simulated, still specifying which variables are included. After the simulation is finished the result of the *realFMU* can be plotted. This plot also serves as a reference for the other model (*simpleFMU*). The extracted data will still be needed later on.


```julia
vrs = ["mass.s", "mass.v", "mass.a", "mass.f"]
success, realSimData = fmiSimulate(realFMU, tStart, tStop; recordValues=vrs, saveat=tSave, setup=false, reset=false)
posReal = collect(data[1] for data in realSimData.saveval)
velReal = collect(data[2] for data in realSimData.saveval)
fmiPlot(realFMU, vrs, realSimData)
```


<?xml version="1.0" encoding="utf-8"?>
<svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" width="600" height="400" viewBox="0 0 2400 1600">
<defs>
  <clipPath id="clip940">
    <rect x="0" y="0" width="2400" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip940)" d="
M0 1600 L2400 1600 L2400 0 L0 0  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip941">
    <rect x="480" y="0" width="1681" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip940)" d="
M149.191 1423.18 L2352.76 1423.18 L2352.76 47.2441 L149.191 47.2441  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip942">
    <rect x="149" y="47" width="2205" height="1377"/>
  </clipPath>
</defs>
<polyline clip-path="url(#clip942)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  211.556,1423.18 211.556,47.2441 
  "/>
<polyline clip-path="url(#clip942)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  731.265,1423.18 731.265,47.2441 
  "/>
<polyline clip-path="url(#clip942)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1250.97,1423.18 1250.97,47.2441 
  "/>
<polyline clip-path="url(#clip942)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1770.68,1423.18 1770.68,47.2441 
  "/>
<polyline clip-path="url(#clip942)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  2290.39,1423.18 2290.39,47.2441 
  "/>
<polyline clip-path="url(#clip940)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  149.191,1423.18 2352.76,1423.18 
  "/>
<polyline clip-path="url(#clip940)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  211.556,1423.18 211.556,1404.28 
  "/>
<polyline clip-path="url(#clip940)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  731.265,1423.18 731.265,1404.28 
  "/>
<polyline clip-path="url(#clip940)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1250.97,1423.18 1250.97,1404.28 
  "/>
<polyline clip-path="url(#clip940)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1770.68,1423.18 1770.68,1404.28 
  "/>
<polyline clip-path="url(#clip940)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  2290.39,1423.18 2290.39,1404.28 
  "/>
<path clip-path="url(#clip940)" d="M211.556 1454.1 Q207.945 1454.1 206.116 1457.66 Q204.311 1461.2 204.311 1468.33 Q204.311 1475.44 206.116 1479.01 Q207.945 1482.55 211.556 1482.55 Q215.19 1482.55 216.996 1479.01 Q218.824 1475.44 218.824 1468.33 Q218.824 1461.2 216.996 1457.66 Q215.19 1454.1 211.556 1454.1 M211.556 1450.39 Q217.366 1450.39 220.422 1455 Q223.5 1459.58 223.5 1468.33 Q223.5 1477.06 220.422 1481.67 Q217.366 1486.25 211.556 1486.25 Q205.746 1486.25 202.667 1481.67 Q199.612 1477.06 199.612 1468.33 Q199.612 1459.58 202.667 1455 Q205.746 1450.39 211.556 1450.39 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M721.647 1481.64 L729.285 1481.64 L729.285 1455.28 L720.975 1456.95 L720.975 1452.69 L729.239 1451.02 L733.915 1451.02 L733.915 1481.64 L741.554 1481.64 L741.554 1485.58 L721.647 1485.58 L721.647 1481.64 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1245.63 1481.64 L1261.95 1481.64 L1261.95 1485.58 L1240 1485.58 L1240 1481.64 Q1242.66 1478.89 1247.25 1474.26 Q1251.85 1469.61 1253.03 1468.27 Q1255.28 1465.74 1256.16 1464.01 Q1257.06 1462.25 1257.06 1460.56 Q1257.06 1457.8 1255.12 1456.07 Q1253.2 1454.33 1250.09 1454.33 Q1247.89 1454.33 1245.44 1455.09 Q1243.01 1455.86 1240.23 1457.41 L1240.23 1452.69 Q1243.06 1451.55 1245.51 1450.97 Q1247.96 1450.39 1250 1450.39 Q1255.37 1450.39 1258.57 1453.08 Q1261.76 1455.77 1261.76 1460.26 Q1261.76 1462.39 1260.95 1464.31 Q1260.16 1466.2 1258.06 1468.8 Q1257.48 1469.47 1254.38 1472.69 Q1251.27 1475.88 1245.63 1481.64 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1774.93 1466.95 Q1778.29 1467.66 1780.16 1469.93 Q1782.06 1472.2 1782.06 1475.53 Q1782.06 1480.65 1778.54 1483.45 Q1775.02 1486.25 1768.54 1486.25 Q1766.37 1486.25 1764.05 1485.81 Q1761.76 1485.39 1759.3 1484.54 L1759.3 1480.02 Q1761.25 1481.16 1763.56 1481.74 Q1765.88 1482.32 1768.4 1482.32 Q1772.8 1482.32 1775.09 1480.58 Q1777.41 1478.84 1777.41 1475.53 Q1777.41 1472.48 1775.25 1470.77 Q1773.12 1469.03 1769.3 1469.03 L1765.28 1469.03 L1765.28 1465.19 L1769.49 1465.19 Q1772.94 1465.19 1774.77 1463.82 Q1776.6 1462.43 1776.6 1459.84 Q1776.6 1457.18 1774.7 1455.77 Q1772.82 1454.33 1769.3 1454.33 Q1767.38 1454.33 1765.18 1454.75 Q1762.99 1455.16 1760.35 1456.04 L1760.35 1451.88 Q1763.01 1451.14 1765.32 1450.77 Q1767.66 1450.39 1769.72 1450.39 Q1775.05 1450.39 1778.15 1452.83 Q1781.25 1455.23 1781.25 1459.35 Q1781.25 1462.22 1779.61 1464.21 Q1777.96 1466.18 1774.93 1466.95 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M2293.4 1455.09 L2281.59 1473.54 L2293.4 1473.54 L2293.4 1455.09 M2292.17 1451.02 L2298.05 1451.02 L2298.05 1473.54 L2302.98 1473.54 L2302.98 1477.43 L2298.05 1477.43 L2298.05 1485.58 L2293.4 1485.58 L2293.4 1477.43 L2277.8 1477.43 L2277.8 1472.92 L2292.17 1451.02 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1199.28 1522.27 L1199.28 1532.4 L1211.35 1532.4 L1211.35 1536.95 L1199.28 1536.95 L1199.28 1556.3 Q1199.28 1560.66 1200.46 1561.9 Q1201.67 1563.14 1205.33 1563.14 L1211.35 1563.14 L1211.35 1568.04 L1205.33 1568.04 Q1198.55 1568.04 1195.97 1565.53 Q1193.4 1562.98 1193.4 1556.3 L1193.4 1536.95 L1189.1 1536.95 L1189.1 1532.4 L1193.4 1532.4 L1193.4 1522.27 L1199.28 1522.27 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1239.23 1518.52 L1252.72 1518.52 L1252.72 1523.07 L1245.09 1523.07 L1245.09 1572.09 L1252.72 1572.09 L1252.72 1576.64 L1239.23 1576.64 L1239.23 1518.52 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1287.93 1533.45 L1287.93 1538.98 Q1285.44 1537.71 1282.77 1537.07 Q1280.1 1536.44 1277.23 1536.44 Q1272.87 1536.44 1270.68 1537.77 Q1268.51 1539.11 1268.51 1541.79 Q1268.51 1543.82 1270.07 1545 Q1271.63 1546.15 1276.34 1547.2 L1278.35 1547.64 Q1284.58 1548.98 1287.19 1551.43 Q1289.84 1553.85 1289.84 1558.21 Q1289.84 1563.17 1285.89 1566.07 Q1281.97 1568.97 1275.1 1568.97 Q1272.23 1568.97 1269.12 1568.39 Q1266.03 1567.85 1262.59 1566.74 L1262.59 1560.69 Q1265.84 1562.38 1268.99 1563.24 Q1272.14 1564.07 1275.23 1564.07 Q1279.36 1564.07 1281.59 1562.66 Q1283.82 1561.23 1283.82 1558.65 Q1283.82 1556.27 1282.2 1554.99 Q1280.61 1553.72 1275.16 1552.54 L1273.13 1552.07 Q1267.68 1550.92 1265.26 1548.56 Q1262.85 1546.18 1262.85 1542.04 Q1262.85 1537.01 1266.41 1534.27 Q1269.98 1531.54 1276.53 1531.54 Q1279.78 1531.54 1282.64 1532.01 Q1285.51 1532.49 1287.93 1533.45 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1312.85 1518.52 L1312.85 1576.64 L1299.35 1576.64 L1299.35 1572.09 L1306.96 1572.09 L1306.96 1523.07 L1299.35 1523.07 L1299.35 1518.52 L1312.85 1518.52 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip942)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  149.191,1247.37 2352.76,1247.37 
  "/>
<polyline clip-path="url(#clip942)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  149.191,969.637 2352.76,969.637 
  "/>
<polyline clip-path="url(#clip942)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  149.191,691.907 2352.76,691.907 
  "/>
<polyline clip-path="url(#clip942)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  149.191,414.176 2352.76,414.176 
  "/>
<polyline clip-path="url(#clip942)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  149.191,136.446 2352.76,136.446 
  "/>
<polyline clip-path="url(#clip940)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  149.191,1423.18 149.191,47.2441 
  "/>
<polyline clip-path="url(#clip940)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  149.191,1247.37 168.089,1247.37 
  "/>
<polyline clip-path="url(#clip940)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  149.191,969.637 168.089,969.637 
  "/>
<polyline clip-path="url(#clip940)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  149.191,691.907 168.089,691.907 
  "/>
<polyline clip-path="url(#clip940)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  149.191,414.176 168.089,414.176 
  "/>
<polyline clip-path="url(#clip940)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  149.191,136.446 168.089,136.446 
  "/>
<path clip-path="url(#clip940)" d="M50.9921 1247.82 L80.6679 1247.82 L80.6679 1251.75 L50.9921 1251.75 L50.9921 1247.82 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M103.608 1234.16 L91.8021 1252.61 L103.608 1252.61 L103.608 1234.16 M102.381 1230.09 L108.26 1230.09 L108.26 1252.61 L113.191 1252.61 L113.191 1256.5 L108.26 1256.5 L108.26 1264.65 L103.608 1264.65 L103.608 1256.5 L88.0058 1256.5 L88.0058 1251.99 L102.381 1230.09 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M53.0754 970.088 L82.7512 970.088 L82.7512 974.023 L53.0754 974.023 L53.0754 970.088 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M96.8715 982.982 L113.191 982.982 L113.191 986.917 L91.2465 986.917 L91.2465 982.982 Q93.9086 980.227 98.4919 975.597 Q103.098 970.945 104.279 969.602 Q106.524 967.079 107.404 965.343 Q108.307 963.584 108.307 961.894 Q108.307 959.139 106.362 957.403 Q104.441 955.667 101.339 955.667 Q99.14 955.667 96.6863 956.431 Q94.2558 957.195 91.478 958.746 L91.478 954.024 Q94.3021 952.889 96.7558 952.311 Q99.2095 951.732 101.246 951.732 Q106.617 951.732 109.811 954.417 Q113.006 957.102 113.006 961.593 Q113.006 963.723 112.196 965.644 Q111.408 967.542 109.302 970.135 Q108.723 970.806 105.621 974.023 Q102.52 977.218 96.8715 982.982 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M101.246 677.705 Q97.6354 677.705 95.8067 681.27 Q94.0012 684.812 94.0012 691.941 Q94.0012 699.048 95.8067 702.613 Q97.6354 706.154 101.246 706.154 Q104.881 706.154 106.686 702.613 Q108.515 699.048 108.515 691.941 Q108.515 684.812 106.686 681.27 Q104.881 677.705 101.246 677.705 M101.246 674.002 Q107.057 674.002 110.112 678.608 Q113.191 683.191 113.191 691.941 Q113.191 700.668 110.112 705.275 Q107.057 709.858 101.246 709.858 Q95.4363 709.858 92.3576 705.275 Q89.3021 700.668 89.3021 691.941 Q89.3021 683.191 92.3576 678.608 Q95.4363 674.002 101.246 674.002 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M96.8715 427.521 L113.191 427.521 L113.191 431.456 L91.2465 431.456 L91.2465 427.521 Q93.9086 424.766 98.4919 420.137 Q103.098 415.484 104.279 414.142 Q106.524 411.618 107.404 409.882 Q108.307 408.123 108.307 406.433 Q108.307 403.679 106.362 401.943 Q104.441 400.206 101.339 400.206 Q99.14 400.206 96.6863 400.97 Q94.2558 401.734 91.478 403.285 L91.478 398.563 Q94.3021 397.429 96.7558 396.85 Q99.2095 396.271 101.246 396.271 Q106.617 396.271 109.811 398.956 Q113.006 401.642 113.006 406.132 Q113.006 408.262 112.196 410.183 Q111.408 412.081 109.302 414.674 Q108.723 415.345 105.621 418.563 Q102.52 421.757 96.8715 427.521 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M103.608 123.24 L91.8021 141.689 L103.608 141.689 L103.608 123.24 M102.381 119.166 L108.26 119.166 L108.26 141.689 L113.191 141.689 L113.191 145.578 L108.26 145.578 L108.26 153.726 L103.608 153.726 L103.608 145.578 L88.0058 145.578 L88.0058 141.064 L102.381 119.166 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip942)" style="stroke:#009af9; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  211.556,622.474 216.753,622.444 221.95,622.355 227.147,622.206 232.344,621.997 237.541,621.727 242.738,621.397 247.936,621.006 253.133,620.555 258.33,620.043 
  263.527,619.472 268.724,618.841 273.921,618.15 279.118,617.401 284.315,616.593 289.512,615.728 294.709,614.806 299.906,613.828 305.103,612.795 310.301,611.708 
  315.498,610.567 320.695,609.375 325.892,608.131 331.089,606.837 336.286,605.495 341.483,604.105 346.68,602.669 351.877,601.189 357.074,599.665 362.271,598.1 
  367.469,596.494 372.666,594.85 377.863,593.17 383.06,591.453 388.257,589.703 393.454,587.921 398.651,586.109 403.848,584.269 409.045,582.403 414.242,580.512 
  419.439,578.598 424.636,576.663 429.834,574.71 435.031,572.739 440.228,570.754 445.425,568.756 450.622,566.748 455.819,564.731 461.016,562.707 466.213,560.678 
  471.41,558.647 476.607,556.615 481.804,554.585 487.002,552.558 492.199,550.536 497.396,548.522 502.593,546.517 507.79,544.523 512.987,542.543 518.184,540.578 
  523.381,538.63 528.578,536.702 533.775,534.794 538.972,532.91 544.17,531.051 549.367,529.219 554.564,527.416 559.761,525.643 564.958,523.902 570.155,522.195 
  575.352,520.525 580.549,518.891 585.746,517.297 590.943,515.743 596.14,514.232 601.337,512.765 606.535,511.343 611.732,509.968 616.929,508.641 622.126,507.363 
  627.323,506.137 632.52,504.963 637.717,503.842 642.914,502.776 648.111,501.766 653.308,500.812 658.505,499.917 663.703,499.081 668.9,498.305 674.097,497.59 
  679.294,496.936 684.491,496.346 689.688,495.819 694.885,495.356 700.082,494.959 705.279,494.627 710.476,494.362 715.673,494.164 720.87,494.034 726.068,493.973 
  731.265,493.977 736.462,494.032 741.659,494.136 746.856,494.289 752.053,494.492 757.25,494.745 762.447,495.048 767.644,495.401 772.841,495.804 778.038,496.257 
  783.236,496.759 788.433,497.31 793.63,497.911 798.827,498.56 804.024,499.257 809.221,500.001 814.418,500.792 819.615,501.63 824.812,502.513 830.009,503.44 
  835.206,504.412 840.403,505.427 845.601,506.484 850.798,507.581 855.995,508.719 861.192,509.896 866.389,511.111 871.586,512.362 876.783,513.648 881.98,514.969 
  887.177,516.323 892.374,517.709 897.571,519.125 902.769,520.57 907.966,522.042 913.163,523.54 918.36,525.063 923.557,526.609 928.754,528.176 933.951,529.763 
  939.148,531.369 944.345,532.991 949.542,534.628 954.739,536.279 959.936,537.941 965.134,539.614 970.331,541.295 975.528,542.983 980.725,544.676 985.922,546.373 
  991.119,548.071 996.316,549.77 1001.51,551.467 1006.71,553.161 1011.91,554.849 1017.1,556.531 1022.3,558.204 1027.5,559.868 1032.7,561.519 1037.89,563.157 
  1043.09,564.78 1048.29,566.387 1053.48,567.975 1058.68,569.543 1063.88,571.09 1069.08,572.613 1074.27,574.112 1079.47,575.585 1084.67,577.03 1089.86,578.447 
  1095.06,579.832 1100.26,581.186 1105.45,582.507 1110.65,583.792 1115.85,585.042 1121.05,586.255 1126.24,587.429 1131.44,588.563 1136.64,589.656 1141.83,590.707 
  1147.03,591.715 1152.23,592.679 1157.43,593.597 1162.62,594.469 1167.82,595.293 1173.02,596.069 1178.21,596.796 1183.41,597.473 1188.61,598.098 1193.81,598.672 
  1199,599.193 1204.2,599.661 1209.4,600.075 1214.59,600.435 1219.79,600.739 1224.99,600.987 1230.19,601.178 1235.38,601.313 1240.58,601.389 1245.78,601.407 
  1250.97,601.381 1256.17,601.316 1261.37,601.213 1266.56,601.071 1271.76,600.89 1276.96,600.67 1282.16,600.411 1287.35,600.113 1292.55,599.776 1297.75,599.4 
  1302.94,598.985 1308.14,598.532 1313.34,598.04 1318.54,597.511 1323.73,596.944 1328.93,596.339 1334.13,595.699 1339.32,595.022 1344.52,594.31 1349.72,593.562 
  1354.92,592.781 1360.11,591.965 1365.31,591.116 1370.51,590.236 1375.7,589.323 1380.9,588.38 1386.1,587.408 1391.29,586.406 1396.49,585.377 1401.69,584.321 
  1406.89,583.24 1412.08,582.133 1417.28,581.003 1422.48,579.85 1427.67,578.676 1432.87,577.482 1438.07,576.268 1443.27,575.036 1448.46,573.788 1453.66,572.524 
  1458.86,571.246 1464.05,569.955 1469.25,568.652 1474.45,567.338 1479.65,566.015 1484.84,564.684 1490.04,563.347 1495.24,562.005 1500.43,560.659 1505.63,559.311 
  1510.83,557.961 1516.02,556.611 1521.22,555.263 1526.42,553.918 1531.62,552.578 1536.81,551.243 1542.01,549.915 1547.21,548.595 1552.4,547.285 1557.6,545.986 
  1562.8,544.699 1568,543.426 1573.19,542.168 1578.39,540.926 1583.59,539.701 1588.78,538.495 1593.98,537.309 1599.18,536.144 1604.38,535.002 1609.57,533.883 
  1614.77,532.788 1619.97,531.719 1625.16,530.677 1630.36,529.663 1635.56,528.678 1640.75,527.723 1645.95,526.799 1651.15,525.907 1656.35,525.048 1661.54,524.223 
  1666.74,523.433 1671.94,522.678 1677.13,521.96 1682.33,521.28 1687.53,520.638 1692.73,520.035 1697.92,519.471 1703.12,518.948 1708.32,518.467 1713.51,518.027 
  1718.71,517.63 1723.91,517.276 1729.11,516.965 1734.3,516.699 1739.5,516.478 1744.7,516.302 1749.89,516.172 1755.09,516.088 1760.29,516.051 1765.49,516.056 
  1770.68,516.089 1775.88,516.148 1781.08,516.235 1786.27,516.349 1791.47,516.49 1796.67,516.658 1801.86,516.855 1807.06,517.078 1812.26,517.329 1817.46,517.607 
  1822.65,517.913 1827.85,518.246 1833.05,518.605 1838.24,518.991 1843.44,519.404 1848.64,519.843 1853.84,520.307 1859.03,520.798 1864.23,521.313 1869.43,521.853 
  1874.62,522.418 1879.82,523.006 1885.02,523.617 1890.22,524.251 1895.41,524.908 1900.61,525.586 1905.81,526.286 1911,527.005 1916.2,527.745 1921.4,528.504 
  1926.59,529.282 1931.79,530.077 1936.99,530.889 1942.19,531.717 1947.38,532.56 1952.58,533.418 1957.78,534.29 1962.97,535.175 1968.17,536.071 1973.37,536.979 
  1978.57,537.897 1983.76,538.825 1988.96,539.761 1994.16,540.704 1999.35,541.654 2004.55,542.61 2009.75,543.571 2014.95,544.535 2020.14,545.502 2025.34,546.471 
  2030.54,547.441 2035.73,548.411 2040.93,549.38 2046.13,550.346 2051.32,551.31 2056.52,552.27 2061.72,553.225 2066.92,554.174 2072.11,555.116 2077.31,556.05 
  2082.51,556.976 2087.7,557.892 2092.9,558.797 2098.1,559.69 2103.3,560.571 2108.49,561.439 2113.69,562.292 2118.89,563.13 2124.08,563.952 2129.28,564.757 
  2134.48,565.545 2139.68,566.313 2144.87,567.062 2150.07,567.791 2155.27,568.499 2160.46,569.185 2165.66,569.849 2170.86,570.489 2176.05,571.105 2181.25,571.696 
  2186.45,572.262 2191.65,572.802 2196.84,573.315 2202.04,573.801 2207.24,574.259 2212.43,574.688 2217.63,575.088 2222.83,575.458 2228.03,575.798 2233.22,576.108 
  2238.42,576.386 2243.62,576.632 2248.81,576.846 2254.01,577.028 2259.21,577.177 2264.41,577.292 2269.6,577.373 2274.8,577.42 2280,577.433 2285.19,577.424 
  2290.39,577.401 
  "/>
<polyline clip-path="url(#clip942)" style="stroke:#e26f46; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  211.556,691.907 216.753,685.979 221.95,680.007 227.147,673.999 232.344,667.966 237.541,661.916 242.738,655.86 247.936,649.806 253.133,643.764 258.33,637.743 
  263.527,631.751 268.724,625.797 273.921,619.889 279.118,614.035 284.315,608.242 289.512,602.519 294.709,596.872 299.906,591.309 305.103,585.836 310.301,580.46 
  315.498,575.187 320.695,570.025 325.892,564.977 331.089,560.05 336.286,555.251 341.483,550.585 346.68,546.056 351.877,541.669 357.074,537.43 362.271,533.344 
  367.469,529.413 372.666,525.644 377.863,522.04 383.06,518.604 388.257,515.34 393.454,512.253 398.651,509.344 403.848,506.617 409.045,504.075 414.242,501.72 
  419.439,499.556 424.636,497.583 429.834,495.804 435.031,494.222 440.228,492.837 445.425,491.652 450.622,490.667 455.819,489.883 461.016,489.301 466.213,488.921 
  471.41,488.745 476.607,488.772 481.804,489.002 487.002,489.435 492.199,490.071 497.396,490.908 502.593,491.947 507.79,493.185 512.987,494.623 518.184,496.258 
  523.381,498.088 528.578,500.113 533.775,502.33 538.972,504.737 544.17,507.331 549.367,510.111 554.564,513.073 559.761,516.215 564.958,519.533 570.155,523.025 
  575.352,526.687 580.549,530.517 585.746,534.509 590.943,538.662 596.14,542.969 601.337,547.429 606.535,552.038 611.732,556.79 616.929,561.681 622.126,566.708 
  627.323,571.866 632.52,577.151 637.717,582.558 642.914,588.082 648.111,593.721 653.308,599.469 658.505,605.322 663.703,611.277 668.9,617.327 674.097,623.469 
  679.294,629.7 684.491,636.015 689.688,642.411 694.885,648.885 700.082,655.433 705.279,662.053 710.476,668.744 715.673,675.503 720.87,682.328 726.068,689.219 
  731.265,694.895 736.462,699.806 741.659,704.751 746.856,709.72 752.053,714.709 757.25,719.707 762.447,724.71 767.644,729.709 772.841,734.698 778.038,739.668 
  783.236,744.613 788.433,749.526 793.63,754.4 798.827,759.23 804.024,764.008 809.221,768.728 814.418,773.385 819.615,777.972 824.812,782.485 830.009,786.916 
  835.206,791.262 840.403,795.516 845.601,799.674 850.798,803.731 855.995,807.683 861.192,811.524 866.389,815.251 871.586,818.86 876.783,822.345 881.98,825.702 
  887.177,828.93 892.374,832.022 897.571,834.977 902.769,837.791 907.966,840.461 913.163,842.983 918.36,845.357 923.557,847.578 928.754,849.645 933.951,851.555 
  939.148,853.306 944.345,854.897 949.542,856.325 954.739,857.59 959.936,858.69 965.134,859.623 970.331,860.389 975.528,860.987 980.725,861.416 985.922,861.675 
  991.119,861.765 996.316,861.685 1001.51,861.435 1006.71,861.016 1011.91,860.427 1017.1,859.669 1022.3,858.743 1027.5,857.65 1032.7,856.391 1037.89,854.966 
  1043.09,853.378 1048.29,851.627 1053.48,849.716 1058.68,847.645 1063.88,845.418 1069.08,843.036 1074.27,840.501 1079.47,837.815 1084.67,834.982 1089.86,832.003 
  1095.06,828.881 1100.26,825.62 1105.45,822.222 1110.65,818.689 1115.85,815.026 1121.05,811.236 1126.24,807.321 1131.44,803.285 1136.64,799.131 1141.83,794.862 
  1147.03,790.484 1152.23,785.998 1157.43,781.409 1162.62,776.719 1167.82,771.932 1173.02,767.053 1178.21,762.085 1183.41,757.032 1188.61,751.895 1193.81,746.68 
  1199,741.389 1204.2,736.025 1209.4,730.591 1214.59,725.09 1219.79,719.524 1224.99,713.895 1230.19,708.206 1235.38,702.458 1240.58,696.654 1245.78,691.191 
  1250.97,687.373 1256.17,683.525 1261.37,679.654 1266.56,675.765 1271.76,671.864 1276.96,667.954 1282.16,664.042 1287.35,660.134 1292.55,656.234 1297.75,652.349 
  1302.94,648.484 1308.14,644.643 1313.34,640.831 1318.54,637.055 1323.73,633.318 1328.93,629.625 1334.13,625.982 1339.32,622.391 1344.52,618.859 1349.72,615.389 
  1354.92,611.985 1360.11,608.654 1365.31,605.397 1370.51,602.219 1375.7,599.124 1380.9,596.115 1386.1,593.196 1391.29,590.371 1396.49,587.642 1401.69,585.013 
  1406.89,582.486 1412.08,580.065 1417.28,577.753 1422.48,575.551 1427.67,573.462 1432.87,571.489 1438.07,569.635 1443.27,567.9 1448.46,566.287 1453.66,564.798 
  1458.86,563.435 1464.05,562.198 1469.25,561.09 1474.45,560.112 1479.65,559.265 1484.84,558.55 1490.04,557.967 1495.24,557.518 1500.43,557.203 1505.63,557.023 
  1510.83,556.977 1516.02,557.067 1521.22,557.291 1526.42,557.651 1531.62,558.145 1536.81,558.774 1542.01,559.536 1547.21,560.432 1552.4,561.46 1557.6,562.62 
  1562.8,563.911 1568,565.331 1573.19,566.879 1578.39,568.555 1583.59,570.356 1588.78,572.281 1593.98,574.327 1599.18,576.495 1604.38,578.78 1609.57,581.182 
  1614.77,583.699 1619.97,586.329 1625.16,589.068 1630.36,591.915 1635.56,594.867 1640.75,597.922 1645.95,601.078 1651.15,604.332 1656.35,607.681 1661.54,611.122 
  1666.74,614.653 1671.94,618.272 1677.13,621.975 1682.33,625.76 1687.53,629.624 1692.73,633.564 1697.92,637.578 1703.12,641.662 1708.32,645.814 1713.51,650.033 
  1718.71,654.314 1723.91,658.657 1729.11,663.057 1734.3,667.513 1739.5,672.022 1744.7,676.583 1749.89,681.193 1755.09,685.852 1760.29,690.557 1765.49,693.813 
  1770.68,696.498 1775.88,699.203 1781.08,701.923 1786.27,704.655 1791.47,707.395 1796.67,710.141 1801.86,712.888 1807.06,715.634 1812.26,718.375 1817.46,721.106 
  1822.65,723.825 1827.85,726.528 1833.05,729.211 1838.24,731.871 1843.44,734.505 1848.64,737.109 1853.84,739.681 1859.03,742.216 1864.23,744.712 1869.43,747.165 
  1874.62,749.573 1879.82,751.933 1885.02,754.242 1890.22,756.497 1895.41,758.695 1900.61,760.832 1905.81,762.908 1911,764.919 1916.2,766.863 1921.4,768.737 
  1926.59,770.539 1931.79,772.268 1936.99,773.92 1942.19,775.495 1947.38,776.99 1952.58,778.404 1957.78,779.734 1962.97,780.979 1968.17,782.138 1973.37,783.209 
  1978.57,784.192 1983.76,785.084 1988.96,785.885 1994.16,786.593 1999.35,787.208 2004.55,787.729 2009.75,788.155 2014.95,788.485 2020.14,788.72 2025.34,788.858 
  2030.54,788.899 2035.73,788.843 2040.93,788.69 2046.13,788.44 2051.32,788.092 2056.52,787.648 2061.72,787.107 2066.92,786.469 2072.11,785.736 2077.31,784.907 
  2082.51,783.983 2087.7,782.965 2092.9,781.853 2098.1,780.649 2103.3,779.353 2108.49,777.967 2113.69,776.492 2118.89,774.928 2124.08,773.277 2129.28,771.54 
  2134.48,769.718 2139.68,767.814 2144.87,765.829 2150.07,763.763 2155.27,761.619 2160.46,759.399 2165.66,757.104 2170.86,754.735 2176.05,752.296 2181.25,749.786 
  2186.45,747.209 2191.65,744.567 2196.84,741.86 2202.04,739.092 2207.24,736.263 2212.43,733.377 2217.63,730.434 2222.83,727.438 2228.03,724.389 2233.22,721.29 
  2238.42,718.143 2243.62,714.949 2248.81,711.711 2254.01,708.43 2259.21,705.108 2264.41,701.747 2269.6,698.349 2274.8,694.915 2280,691.722 2285.19,690.316 
  2290.39,688.899 
  "/>
<polyline clip-path="url(#clip942)" style="stroke:#3da44d; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  211.556,101.73 216.753,96.6426 221.95,92.579 227.147,89.5214 232.344,87.451 237.541,86.3465 242.738,86.1857 247.936,86.946 253.133,88.6044 258.33,91.1371 
  263.527,94.5199 268.724,98.7307 273.921,103.744 279.118,109.541 284.315,116.094 289.512,123.386 294.709,131.391 299.906,140.089 305.103,149.458 310.301,159.476 
  315.498,170.125 320.695,181.386 325.892,193.234 331.089,205.652 336.286,218.621 341.483,232.123 346.68,246.134 351.877,260.637 357.074,275.613 362.271,291.041 
  367.469,306.903 372.666,323.178 377.863,339.846 383.06,356.891 388.257,374.292 393.454,392.029 398.651,410.079 403.848,428.424 409.045,447.041 414.242,465.909 
  419.439,485.021 424.636,504.347 429.834,523.867 435.031,543.558 440.228,563.4 445.425,583.369 450.622,603.443 455.819,623.612 461.016,643.849 466.213,664.134 
  471.41,684.445 476.607,704.763 481.804,725.069 487.002,745.343 492.199,765.564 497.396,785.712 502.593,805.768 507.79,825.713 512.987,845.527 518.184,865.191 
  523.381,884.685 528.578,903.994 533.775,923.098 538.972,941.978 544.17,960.617 549.367,978.998 554.564,997.106 559.761,1014.92 564.958,1032.44 570.155,1049.63 
  575.352,1066.48 580.549,1082.99 585.746,1099.13 590.943,1114.9 596.14,1130.29 601.337,1145.28 606.535,1159.86 611.732,1174.03 616.929,1187.78 622.126,1201.1 
  627.323,1213.99 632.52,1226.45 637.717,1238.46 642.914,1250.03 648.111,1261.17 653.308,1271.87 658.505,1282.15 663.703,1292 668.9,1301.43 674.097,1310.47 
  679.294,1319.13 684.491,1327.42 689.688,1335.36 694.885,1342.99 700.082,1350.33 705.279,1357.43 710.476,1364.31 715.673,1371.04 720.87,1377.66 726.068,1384.24 
  731.265,1181.18 736.462,1184.93 741.659,1187.9 746.856,1190.1 752.053,1191.55 757.25,1192.24 762.447,1192.19 767.644,1191.42 772.841,1189.94 778.038,1187.75 
  783.236,1184.88 788.433,1181.34 793.63,1177.14 798.827,1172.31 804.024,1166.85 809.221,1160.79 814.418,1154.13 819.615,1146.9 824.812,1139.11 830.009,1130.78 
  835.206,1121.92 840.403,1112.54 845.601,1102.67 850.798,1092.33 855.995,1081.52 861.192,1070.26 866.389,1058.58 871.586,1046.49 876.783,1033.99 881.98,1021.11 
  887.177,1007.87 892.374,994.282 897.571,980.359 902.769,966.12 907.966,951.584 913.163,936.766 918.36,921.684 923.557,906.354 928.754,890.795 933.951,875.022 
  939.148,859.054 944.345,842.907 949.542,826.598 954.739,810.146 959.936,793.566 965.134,776.877 970.331,760.094 975.528,743.236 980.725,726.32 985.922,709.364 
  991.119,692.383 996.316,675.395 1001.51,658.416 1006.71,641.464 1011.91,624.557 1017.1,607.711 1022.3,590.94 1027.5,574.26 1032.7,557.69 1037.89,541.245 
  1043.09,524.942 1048.29,508.793 1053.48,492.813 1058.68,477.017 1063.88,461.422 1069.08,446.04 1074.27,430.885 1079.47,415.968 1084.67,401.302 1089.86,386.901 
  1095.06,372.776 1100.26,358.938 1105.45,345.396 1110.65,332.159 1115.85,319.238 1121.05,306.641 1126.24,294.376 1131.44,282.446 1136.64,270.86 1141.83,259.622 
  1147.03,248.736 1152.23,238.205 1157.43,228.029 1162.62,218.209 1167.82,208.743 1173.02,199.631 1178.21,190.867 1183.41,182.445 1188.61,174.359 1193.81,166.598 
  1199,159.151 1204.2,152.005 1209.4,145.142 1214.59,138.544 1219.79,132.188 1224.99,126.051 1230.19,120.101 1235.38,114.308 1240.58,108.632 1245.78,311.718 
  1250.97,308.497 1256.17,305.829 1261.37,303.718 1266.56,302.162 1271.76,301.159 1276.96,300.706 1282.16,300.798 1287.35,301.431 1292.55,302.599 1297.75,304.293 
  1302.94,306.507 1308.14,309.234 1313.34,312.462 1318.54,316.184 1323.73,320.39 1328.93,325.07 1334.13,330.214 1339.32,335.808 1344.52,341.841 1349.72,348.305 
  1354.92,355.189 1360.11,362.483 1365.31,370.172 1370.51,378.245 1375.7,386.691 1380.9,395.497 1386.1,404.65 1391.29,414.136 1396.49,423.944 1401.69,434.061 
  1406.89,444.473 1412.08,455.166 1417.28,466.129 1422.48,477.347 1427.67,488.808 1432.87,500.498 1438.07,512.405 1443.27,524.513 1448.46,536.81 1453.66,549.283 
  1458.86,561.915 1464.05,574.697 1469.25,587.614 1474.45,600.652 1479.65,613.796 1484.84,627.033 1490.04,640.347 1495.24,653.725 1500.43,667.155 1505.63,680.626 
  1510.83,694.121 1516.02,707.626 1521.22,721.126 1526.42,734.609 1531.62,748.061 1536.81,761.471 1542.01,774.827 1547.21,788.113 1552.4,801.318 1557.6,814.429 
  1562.8,827.434 1568,840.32 1573.19,853.078 1578.39,865.694 1583.59,878.158 1588.78,890.46 1593.98,902.59 1599.18,914.536 1604.38,926.29 1609.57,937.841 
  1614.77,949.181 1619.97,960.302 1625.16,971.195 1630.36,981.855 1635.56,992.273 1640.75,1002.44 1645.95,1012.36 1651.15,1022.02 1656.35,1031.41 1661.54,1040.54 
  1666.74,1049.4 1671.94,1057.98 1677.13,1066.29 1682.33,1074.33 1687.53,1082.09 1692.73,1089.57 1697.92,1096.79 1703.12,1103.74 1708.32,1110.42 1713.51,1116.85 
  1718.71,1123.02 1723.91,1128.95 1729.11,1134.65 1734.3,1140.13 1739.5,1145.39 1744.7,1150.47 1749.89,1155.36 1755.09,1160.1 1760.29,1164.69 1765.49,959.396 
  1770.68,961.498 1775.88,963.251 1781.08,964.65 1786.27,965.694 1791.47,966.381 1796.67,966.712 1801.86,966.685 1807.06,966.303 1812.26,965.566 1817.46,964.475 
  1822.65,963.035 1827.85,961.247 1833.05,959.117 1838.24,956.647 1843.44,953.842 1848.64,950.708 1853.84,947.249 1859.03,943.471 1864.23,939.381 1869.43,934.984 
  1874.62,930.286 1879.82,925.299 1885.02,920.027 1890.22,914.476 1895.41,908.654 1900.61,902.567 1905.81,896.227 1911,889.64 1916.2,882.814 1921.4,875.759 
  1926.59,868.483 1931.79,860.996 1936.99,853.307 1942.19,845.424 1947.38,837.358 1952.58,829.118 1957.78,820.713 1962.97,812.153 1968.17,803.447 1973.37,794.606 
  1978.57,785.638 1983.76,776.553 1988.96,767.362 1994.16,758.074 1999.35,748.699 2004.55,739.247 2009.75,729.727 2014.95,720.149 2020.14,710.524 2025.34,700.861 
  2030.54,691.17 2035.73,681.461 2040.93,671.742 2046.13,662.024 2051.32,652.317 2056.52,642.629 2061.72,632.971 2066.92,623.35 2072.11,613.777 2077.31,604.26 
  2082.51,594.807 2087.7,585.428 2092.9,576.13 2098.1,566.923 2103.3,557.813 2108.49,548.809 2113.69,539.919 2118.89,531.148 2124.08,522.505 2129.28,513.997 
  2134.48,505.629 2139.68,497.409 2144.87,489.341 2150.07,481.432 2155.27,473.687 2160.46,466.111 2165.66,458.708 2170.86,451.483 2176.05,444.439 2181.25,437.58 
  2186.45,430.908 2191.65,424.427 2196.84,418.139 2202.04,412.045 2207.24,406.146 2212.43,400.442 2217.63,394.935 2222.83,389.624 2228.03,384.507 2233.22,379.584 
  2238.42,374.852 2243.62,370.308 2248.81,365.949 2254.01,361.771 2259.21,357.769 2264.41,353.938 2269.6,350.271 2274.8,346.762 2280,551.962 2285.19,550.739 
  2290.39,549.67 
  "/>
<polyline clip-path="url(#clip942)" style="stroke:#c271d2; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  211.556,587.758 216.753,593.143 221.95,598.099 227.147,602.647 232.344,606.81 237.541,610.611 242.738,614.074 247.936,617.221 253.133,620.075 258.33,622.657 
  263.527,624.988 268.724,627.09 273.921,628.98 279.118,630.677 284.315,632.198 289.512,633.558 294.709,634.773 299.906,635.855 305.103,636.817 310.301,637.671 
  315.498,638.427 320.695,639.096 325.892,639.685 331.089,640.203 336.286,640.657 341.483,641.054 346.68,641.4 351.877,641.701 357.074,641.962 362.271,642.186 
  367.469,642.379 372.666,642.544 377.863,642.684 383.06,642.803 388.257,642.902 393.454,642.985 398.651,643.054 403.848,643.111 409.045,643.157 414.242,643.194 
  419.439,643.223 424.636,643.247 429.834,643.265 435.031,643.278 440.228,643.289 445.425,643.296 450.622,643.302 455.819,643.305 461.016,643.308 466.213,643.309 
  471.41,643.31 476.607,643.31 481.804,643.309 487.002,643.307 492.199,643.304 497.396,643.3 502.593,643.294 507.79,643.286 512.987,643.275 518.184,643.26 
  523.381,643.241 528.578,643.216 533.775,643.185 538.972,643.145 544.17,643.097 549.367,643.037 554.564,642.964 559.761,642.877 564.958,642.772 570.155,642.648 
  575.352,642.5 580.549,642.327 585.746,642.125 590.943,641.889 596.14,641.616 601.337,641.299 606.535,640.935 611.732,640.517 616.929,640.037 622.126,639.49 
  627.323,638.865 632.52,638.155 637.717,637.348 642.914,636.434 648.111,635.399 653.308,634.229 658.505,632.909 663.703,631.42 668.9,629.744 674.097,627.857 
  679.294,625.736 684.491,623.353 689.688,620.677 694.885,617.674 700.082,614.306 705.279,610.529 710.476,606.295 715.673,601.547 720.87,596.226 726.068,590.26 
  731.265,793.28 736.462,788.984 741.659,784.972 746.856,781.234 752.053,777.759 757.25,774.536 762.447,771.552 767.644,768.795 772.841,766.252 778.038,763.911 
  783.236,761.759 788.433,759.784 793.63,757.975 798.827,756.32 804.024,754.808 809.221,753.428 814.418,752.171 819.615,751.028 824.812,749.988 830.009,749.045 
  835.206,748.191 840.403,747.417 845.601,746.717 850.798,746.085 855.995,745.515 861.192,745.002 866.389,744.541 871.586,744.126 876.783,743.754 881.98,743.421 
  887.177,743.123 892.374,742.858 897.571,742.621 902.769,742.41 907.966,742.223 913.163,742.058 918.36,741.912 923.557,741.783 928.754,741.67 933.951,741.571 
  939.148,741.485 944.345,741.411 949.542,741.347 954.739,741.293 959.936,741.248 965.134,741.21 970.331,741.181 975.528,741.158 980.725,741.142 985.922,741.132 
  991.119,741.129 996.316,741.132 1001.51,741.141 1006.71,741.157 1011.91,741.179 1017.1,741.209 1022.3,741.245 1027.5,741.29 1032.7,741.344 1037.89,741.408 
  1043.09,741.482 1048.29,741.568 1053.48,741.666 1058.68,741.779 1063.88,741.908 1069.08,742.055 1074.27,742.221 1079.47,742.408 1084.67,742.62 1089.86,742.859 
  1095.06,743.128 1100.26,743.429 1105.45,743.767 1110.65,744.145 1115.85,744.567 1121.05,745.039 1126.24,745.566 1131.44,746.152 1136.64,746.805 1141.83,747.532 
  1147.03,748.339 1152.23,749.235 1157.43,750.229 1162.62,751.331 1167.82,752.553 1173.02,753.905 1178.21,755.402 1183.41,757.057 1188.61,758.887 1193.81,760.909 
  1199,763.142 1204.2,765.608 1209.4,768.331 1214.59,771.335 1219.79,774.65 1224.99,778.308 1230.19,782.343 1235.38,786.794 1240.58,791.704 1245.78,588.434 
  1250.97,591.92 1256.17,595.234 1261.37,598.377 1266.56,601.353 1271.76,604.165 1276.96,606.818 1282.16,609.316 1287.35,611.665 1292.55,613.869 1297.75,615.936 
  1302.94,617.869 1308.14,619.677 1313.34,621.364 1318.54,622.936 1323.73,624.4 1328.93,625.762 1334.13,627.028 1339.32,628.202 1344.52,629.291 1349.72,630.3 
  1354.92,631.233 1360.11,632.095 1365.31,632.891 1370.51,633.626 1375.7,634.303 1380.9,634.927 1386.1,635.5 1391.29,636.027 1396.49,636.51 1401.69,636.954 
  1406.89,637.359 1412.08,637.73 1417.28,638.069 1422.48,638.378 1427.67,638.658 1432.87,638.913 1438.07,639.143 1443.27,639.351 1448.46,639.538 1453.66,639.704 
  1458.86,639.853 1464.05,639.983 1469.25,640.098 1474.45,640.197 1479.65,640.28 1484.84,640.35 1490.04,640.406 1495.24,640.448 1500.43,640.478 1505.63,640.495 
  1510.83,640.499 1516.02,640.491 1521.22,640.47 1526.42,640.436 1531.62,640.389 1536.81,640.328 1542.01,640.254 1547.21,640.164 1552.4,640.06 1557.6,639.939 
  1562.8,639.801 1568,639.645 1573.19,639.47 1578.39,639.273 1583.59,639.055 1588.78,638.812 1593.98,638.544 1599.18,638.247 1604.38,637.921 1609.57,637.561 
  1614.77,637.167 1619.97,636.734 1625.16,636.261 1630.36,635.742 1635.56,635.175 1640.75,634.556 1645.95,633.88 1651.15,633.142 1656.35,632.338 1661.54,631.461 
  1666.74,630.506 1671.94,629.466 1677.13,628.334 1682.33,627.102 1687.53,625.763 1692.73,624.307 1697.92,622.724 1703.12,621.004 1708.32,619.137 1713.51,617.109 
  1718.71,614.906 1723.91,612.516 1729.11,609.922 1734.3,607.107 1739.5,604.054 1744.7,600.742 1749.89,597.15 1755.09,593.253 1760.29,589.027 1765.49,794.27 
  1770.68,791.842 1775.88,789.494 1781.08,787.229 1786.27,785.047 1791.47,782.947 1796.67,780.931 1801.86,778.996 1807.06,777.144 1812.26,775.371 1817.46,773.678 
  1822.65,772.063 1827.85,770.525 1833.05,769.06 1838.24,767.668 1843.44,766.346 1848.64,765.093 1853.84,763.905 1859.03,762.78 1864.23,761.717 1869.43,760.713 
  1874.62,759.766 1879.82,758.873 1885.02,758.031 1890.22,757.24 1895.41,756.497 1900.61,755.799 1905.81,755.145 1911,754.533 1916.2,753.96 1921.4,753.426 
  1926.59,752.927 1931.79,752.464 1936.99,752.033 1942.19,751.634 1947.38,751.265 1952.58,750.925 1957.78,750.612 1962.97,750.326 1968.17,750.065 1973.37,749.829 
  1978.57,749.616 1983.76,749.426 1988.96,749.258 1994.16,749.112 1999.35,748.986 2004.55,748.88 2009.75,748.795 2014.95,748.729 2020.14,748.682 2025.34,748.655 
  2030.54,748.647 2035.73,748.658 2040.93,748.688 2046.13,748.738 2051.32,748.807 2056.52,748.897 2061.72,749.006 2066.92,749.137 2072.11,749.289 2077.31,749.464 
  2082.51,749.661 2087.7,749.883 2092.9,750.129 2098.1,750.401 2103.3,750.701 2108.49,751.029 2113.69,751.387 2118.89,751.777 2124.08,752.199 2129.28,752.657 
  2134.48,753.152 2139.68,753.687 2144.87,754.262 2150.07,754.882 2155.27,755.548 2160.46,756.264 2165.66,757.032 2170.86,757.856 2176.05,758.738 2181.25,759.684 
  2186.45,760.696 2191.65,761.778 2196.84,762.935 2202.04,764.173 2207.24,765.494 2212.43,766.906 2217.63,768.413 2222.83,770.021 2228.03,771.737 2233.22,773.567 
  2238.42,775.518 2243.62,777.598 2248.81,779.815 2254.01,782.177 2259.21,784.693 2264.41,787.372 2269.6,790.225 2274.8,793.262 2280,587.933 2285.19,589.251 
  2290.39,590.551 
  "/>
<path clip-path="url(#clip940)" d="
M1390.67 352.309 L2279.3 352.309 L2279.3 93.1086 L1390.67 93.1086  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<polyline clip-path="url(#clip940)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1390.67,352.309 2279.3,352.309 2279.3,93.1086 1390.67,93.1086 1390.67,352.309 
  "/>
<polyline clip-path="url(#clip940)" style="stroke:#009af9; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1415.15,144.949 1562.06,144.949 
  "/>
<path clip-path="url(#clip940)" d="M1606.89 141.28 Q1608.49 138.409 1610.71 137.044 Q1612.93 135.678 1615.94 135.678 Q1619.99 135.678 1622.19 138.525 Q1624.39 141.349 1624.39 146.581 L1624.39 162.229 L1620.11 162.229 L1620.11 146.719 Q1620.11 142.993 1618.79 141.187 Q1617.47 139.382 1614.76 139.382 Q1611.45 139.382 1609.53 141.581 Q1607.61 143.78 1607.61 147.576 L1607.61 162.229 L1603.33 162.229 L1603.33 146.719 Q1603.33 142.969 1602.01 141.187 Q1600.69 139.382 1597.93 139.382 Q1594.67 139.382 1592.75 141.604 Q1590.83 143.803 1590.83 147.576 L1590.83 162.229 L1586.54 162.229 L1586.54 136.303 L1590.83 136.303 L1590.83 140.331 Q1592.28 137.946 1594.32 136.812 Q1596.36 135.678 1599.16 135.678 Q1601.98 135.678 1603.95 137.113 Q1605.94 138.548 1606.89 141.28 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1644.67 149.196 Q1639.51 149.196 1637.51 150.377 Q1635.52 151.557 1635.52 154.405 Q1635.52 156.673 1637.01 158.016 Q1638.51 159.335 1641.08 159.335 Q1644.62 159.335 1646.75 156.835 Q1648.9 154.312 1648.9 150.145 L1648.9 149.196 L1644.67 149.196 M1653.16 147.437 L1653.16 162.229 L1648.9 162.229 L1648.9 158.293 Q1647.45 160.655 1645.27 161.789 Q1643.09 162.9 1639.95 162.9 Q1635.96 162.9 1633.6 160.678 Q1631.26 158.432 1631.26 154.682 Q1631.26 150.307 1634.18 148.085 Q1637.12 145.863 1642.93 145.863 L1648.9 145.863 L1648.9 145.446 Q1648.9 142.507 1646.96 140.909 Q1645.04 139.289 1641.54 139.289 Q1639.32 139.289 1637.21 139.821 Q1635.11 140.354 1633.16 141.419 L1633.16 137.483 Q1635.5 136.581 1637.7 136.141 Q1639.9 135.678 1641.98 135.678 Q1647.61 135.678 1650.39 138.594 Q1653.16 141.511 1653.16 147.437 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1678.46 137.067 L1678.46 141.094 Q1676.66 140.169 1674.71 139.706 Q1672.77 139.243 1670.69 139.243 Q1667.51 139.243 1665.92 140.215 Q1664.34 141.187 1664.34 143.131 Q1664.34 144.613 1665.48 145.469 Q1666.61 146.303 1670.04 147.067 L1671.5 147.391 Q1676.03 148.363 1677.93 150.145 Q1679.85 151.905 1679.85 155.076 Q1679.85 158.687 1676.98 160.793 Q1674.13 162.9 1669.13 162.9 Q1667.05 162.9 1664.78 162.483 Q1662.54 162.09 1660.04 161.28 L1660.04 156.881 Q1662.4 158.108 1664.69 158.733 Q1666.98 159.335 1669.23 159.335 Q1672.24 159.335 1673.86 158.317 Q1675.48 157.275 1675.48 155.4 Q1675.48 153.664 1674.3 152.738 Q1673.14 151.812 1669.18 150.956 L1667.7 150.608 Q1663.74 149.775 1661.98 148.062 Q1660.22 146.326 1660.22 143.317 Q1660.22 139.659 1662.82 137.669 Q1665.41 135.678 1670.18 135.678 Q1672.54 135.678 1674.62 136.025 Q1676.7 136.372 1678.46 137.067 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1703.16 137.067 L1703.16 141.094 Q1701.36 140.169 1699.41 139.706 Q1697.47 139.243 1695.38 139.243 Q1692.21 139.243 1690.62 140.215 Q1689.04 141.187 1689.04 143.131 Q1689.04 144.613 1690.18 145.469 Q1691.31 146.303 1694.74 147.067 L1696.19 147.391 Q1700.73 148.363 1702.63 150.145 Q1704.55 151.905 1704.55 155.076 Q1704.55 158.687 1701.68 160.793 Q1698.83 162.9 1693.83 162.9 Q1691.75 162.9 1689.48 162.483 Q1687.24 162.09 1684.74 161.28 L1684.74 156.881 Q1687.1 158.108 1689.39 158.733 Q1691.68 159.335 1693.93 159.335 Q1696.94 159.335 1698.56 158.317 Q1700.18 157.275 1700.18 155.4 Q1700.18 153.664 1699 152.738 Q1697.84 151.812 1693.88 150.956 L1692.4 150.608 Q1688.44 149.775 1686.68 148.062 Q1684.92 146.326 1684.92 143.317 Q1684.92 139.659 1687.51 137.669 Q1690.11 135.678 1694.88 135.678 Q1697.24 135.678 1699.32 136.025 Q1701.4 136.372 1703.16 137.067 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1711.94 156.349 L1716.82 156.349 L1716.82 162.229 L1711.94 162.229 L1711.94 156.349 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1742.93 137.067 L1742.93 141.094 Q1741.13 140.169 1739.18 139.706 Q1737.24 139.243 1735.15 139.243 Q1731.98 139.243 1730.38 140.215 Q1728.81 141.187 1728.81 143.131 Q1728.81 144.613 1729.94 145.469 Q1731.08 146.303 1734.5 147.067 L1735.96 147.391 Q1740.5 148.363 1742.4 150.145 Q1744.32 151.905 1744.32 155.076 Q1744.32 158.687 1741.45 160.793 Q1738.6 162.9 1733.6 162.9 Q1731.52 162.9 1729.25 162.483 Q1727 162.09 1724.51 161.28 L1724.51 156.881 Q1726.87 158.108 1729.16 158.733 Q1731.45 159.335 1733.69 159.335 Q1736.7 159.335 1738.32 158.317 Q1739.94 157.275 1739.94 155.4 Q1739.94 153.664 1738.76 152.738 Q1737.61 151.812 1733.65 150.956 L1732.17 150.608 Q1728.21 149.775 1726.45 148.062 Q1724.69 146.326 1724.69 143.317 Q1724.69 139.659 1727.28 137.669 Q1729.88 135.678 1734.64 135.678 Q1737 135.678 1739.09 136.025 Q1741.17 136.372 1742.93 137.067 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1776.4 126.257 Q1773.3 131.581 1771.8 136.789 Q1770.29 141.997 1770.29 147.344 Q1770.29 152.692 1771.8 157.946 Q1773.32 163.178 1776.4 168.479 L1772.7 168.479 Q1769.23 163.039 1767.49 157.784 Q1765.78 152.53 1765.78 147.344 Q1765.78 142.182 1767.49 136.951 Q1769.2 131.72 1772.7 126.257 L1776.4 126.257 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1799.44 143.594 Q1802.79 144.312 1804.67 146.581 Q1806.56 148.849 1806.56 152.182 Q1806.56 157.298 1803.05 160.099 Q1799.53 162.9 1793.05 162.9 Q1790.87 162.9 1788.56 162.46 Q1786.26 162.043 1783.81 161.187 L1783.81 156.673 Q1785.75 157.807 1788.07 158.386 Q1790.38 158.965 1792.91 158.965 Q1797.31 158.965 1799.6 157.229 Q1801.91 155.493 1801.91 152.182 Q1801.91 149.127 1799.76 147.414 Q1797.63 145.678 1793.81 145.678 L1789.78 145.678 L1789.78 141.835 L1794 141.835 Q1797.44 141.835 1799.27 140.469 Q1801.1 139.081 1801.1 136.488 Q1801.1 133.826 1799.2 132.414 Q1797.33 130.979 1793.81 130.979 Q1791.89 130.979 1789.69 131.395 Q1787.49 131.812 1784.85 132.692 L1784.85 128.525 Q1787.51 127.784 1789.83 127.414 Q1792.17 127.044 1794.23 127.044 Q1799.55 127.044 1802.65 129.474 Q1805.75 131.882 1805.75 136.002 Q1805.75 138.872 1804.11 140.863 Q1802.47 142.831 1799.44 143.594 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1829.6 143.594 Q1832.95 144.312 1834.83 146.581 Q1836.73 148.849 1836.73 152.182 Q1836.73 157.298 1833.21 160.099 Q1829.69 162.9 1823.21 162.9 Q1821.03 162.9 1818.72 162.46 Q1816.43 162.043 1813.97 161.187 L1813.97 156.673 Q1815.92 157.807 1818.23 158.386 Q1820.55 158.965 1823.07 158.965 Q1827.47 158.965 1829.76 157.229 Q1832.07 155.493 1832.07 152.182 Q1832.07 149.127 1829.92 147.414 Q1827.79 145.678 1823.97 145.678 L1819.94 145.678 L1819.94 141.835 L1824.16 141.835 Q1827.61 141.835 1829.43 140.469 Q1831.26 139.081 1831.26 136.488 Q1831.26 133.826 1829.37 132.414 Q1827.49 130.979 1823.97 130.979 Q1822.05 130.979 1819.85 131.395 Q1817.65 131.812 1815.01 132.692 L1815.01 128.525 Q1817.68 127.784 1819.99 127.414 Q1822.33 127.044 1824.39 127.044 Q1829.71 127.044 1832.81 129.474 Q1835.92 131.882 1835.92 136.002 Q1835.92 138.872 1834.27 140.863 Q1832.63 142.831 1829.6 143.594 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1845.64 127.669 L1863.99 127.669 L1863.99 131.604 L1849.92 131.604 L1849.92 140.076 Q1850.94 139.729 1851.96 139.567 Q1852.98 139.382 1854 139.382 Q1859.78 139.382 1863.16 142.553 Q1866.54 145.724 1866.54 151.141 Q1866.54 156.719 1863.07 159.821 Q1859.6 162.9 1853.28 162.9 Q1851.1 162.9 1848.83 162.53 Q1846.59 162.159 1844.18 161.418 L1844.18 156.719 Q1846.26 157.854 1848.49 158.409 Q1850.71 158.965 1853.18 158.965 Q1857.19 158.965 1859.53 156.858 Q1861.87 154.752 1861.87 151.141 Q1861.87 147.53 1859.53 145.423 Q1857.19 143.317 1853.18 143.317 Q1851.31 143.317 1849.43 143.733 Q1847.58 144.15 1845.64 145.03 L1845.64 127.669 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1875.8 127.669 L1894.16 127.669 L1894.16 131.604 L1880.08 131.604 L1880.08 140.076 Q1881.1 139.729 1882.12 139.567 Q1883.14 139.382 1884.16 139.382 Q1889.94 139.382 1893.32 142.553 Q1896.7 145.724 1896.7 151.141 Q1896.7 156.719 1893.23 159.821 Q1889.76 162.9 1883.44 162.9 Q1881.26 162.9 1878.99 162.53 Q1876.75 162.159 1874.34 161.418 L1874.34 156.719 Q1876.43 157.854 1878.65 158.409 Q1880.87 158.965 1883.35 158.965 Q1887.35 158.965 1889.69 156.858 Q1892.03 154.752 1892.03 151.141 Q1892.03 147.53 1889.69 145.423 Q1887.35 143.317 1883.35 143.317 Q1881.47 143.317 1879.6 143.733 Q1877.74 144.15 1875.8 145.03 L1875.8 127.669 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1918.76 131.743 L1906.96 150.192 L1918.76 150.192 L1918.76 131.743 M1917.54 127.669 L1923.42 127.669 L1923.42 150.192 L1928.35 150.192 L1928.35 154.081 L1923.42 154.081 L1923.42 162.229 L1918.76 162.229 L1918.76 154.081 L1903.16 154.081 L1903.16 149.567 L1917.54 127.669 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1948.93 131.743 L1937.12 150.192 L1948.93 150.192 L1948.93 131.743 M1947.7 127.669 L1953.58 127.669 L1953.58 150.192 L1958.51 150.192 L1958.51 154.081 L1953.58 154.081 L1953.58 162.229 L1948.93 162.229 L1948.93 154.081 L1933.32 154.081 L1933.32 149.567 L1947.7 127.669 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1980.41 143.594 Q1983.76 144.312 1985.64 146.581 Q1987.54 148.849 1987.54 152.182 Q1987.54 157.298 1984.02 160.099 Q1980.5 162.9 1974.02 162.9 Q1971.84 162.9 1969.53 162.46 Q1967.24 162.043 1964.78 161.187 L1964.78 156.673 Q1966.73 157.807 1969.04 158.386 Q1971.36 158.965 1973.88 158.965 Q1978.28 158.965 1980.57 157.229 Q1982.88 155.493 1982.88 152.182 Q1982.88 149.127 1980.73 147.414 Q1978.6 145.678 1974.78 145.678 L1970.75 145.678 L1970.75 141.835 L1974.97 141.835 Q1978.42 141.835 1980.24 140.469 Q1982.07 139.081 1982.07 136.488 Q1982.07 133.826 1980.17 132.414 Q1978.3 130.979 1974.78 130.979 Q1972.86 130.979 1970.66 131.395 Q1968.46 131.812 1965.82 132.692 L1965.82 128.525 Q1968.49 127.784 1970.8 127.414 Q1973.14 127.044 1975.2 127.044 Q1980.52 127.044 1983.62 129.474 Q1986.73 131.882 1986.73 136.002 Q1986.73 138.872 1985.08 140.863 Q1983.44 142.831 1980.41 143.594 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M2000.43 158.293 L2016.75 158.293 L2016.75 162.229 L1994.8 162.229 L1994.8 158.293 Q1997.47 155.539 2002.05 150.909 Q2006.66 146.256 2007.84 144.914 Q2010.08 142.391 2010.96 140.655 Q2011.86 138.895 2011.86 137.206 Q2011.86 134.451 2009.92 132.715 Q2008 130.979 2004.9 130.979 Q2002.7 130.979 2000.24 131.743 Q1997.81 132.507 1995.04 134.057 L1995.04 129.335 Q1997.86 128.201 2000.31 127.622 Q2002.77 127.044 2004.8 127.044 Q2010.17 127.044 2013.37 129.729 Q2016.56 132.414 2016.56 136.905 Q2016.56 139.034 2015.75 140.956 Q2014.97 142.854 2012.86 145.446 Q2012.28 146.118 2009.18 149.335 Q2006.08 152.53 2000.43 158.293 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M2025.29 126.257 L2028.99 126.257 Q2032.47 131.72 2034.18 136.951 Q2035.92 142.182 2035.92 147.344 Q2035.92 152.53 2034.18 157.784 Q2032.47 163.039 2028.99 168.479 L2025.29 168.479 Q2028.37 163.178 2029.87 157.946 Q2031.4 152.692 2031.4 147.344 Q2031.4 141.997 2029.87 136.789 Q2028.37 131.581 2025.29 126.257 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip940)" style="stroke:#e26f46; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1415.15,196.789 1562.06,196.789 
  "/>
<path clip-path="url(#clip940)" d="M1606.89 193.12 Q1608.49 190.249 1610.71 188.884 Q1612.93 187.518 1615.94 187.518 Q1619.99 187.518 1622.19 190.365 Q1624.39 193.189 1624.39 198.421 L1624.39 214.069 L1620.11 214.069 L1620.11 198.559 Q1620.11 194.833 1618.79 193.027 Q1617.47 191.222 1614.76 191.222 Q1611.45 191.222 1609.53 193.421 Q1607.61 195.62 1607.61 199.416 L1607.61 214.069 L1603.33 214.069 L1603.33 198.559 Q1603.33 194.809 1602.01 193.027 Q1600.69 191.222 1597.93 191.222 Q1594.67 191.222 1592.75 193.444 Q1590.83 195.643 1590.83 199.416 L1590.83 214.069 L1586.54 214.069 L1586.54 188.143 L1590.83 188.143 L1590.83 192.171 Q1592.28 189.786 1594.32 188.652 Q1596.36 187.518 1599.16 187.518 Q1601.98 187.518 1603.95 188.953 Q1605.94 190.388 1606.89 193.12 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1644.67 201.036 Q1639.51 201.036 1637.51 202.217 Q1635.52 203.397 1635.52 206.245 Q1635.52 208.513 1637.01 209.856 Q1638.51 211.175 1641.08 211.175 Q1644.62 211.175 1646.75 208.675 Q1648.9 206.152 1648.9 201.985 L1648.9 201.036 L1644.67 201.036 M1653.16 199.277 L1653.16 214.069 L1648.9 214.069 L1648.9 210.133 Q1647.45 212.495 1645.27 213.629 Q1643.09 214.74 1639.95 214.74 Q1635.96 214.74 1633.6 212.518 Q1631.26 210.272 1631.26 206.522 Q1631.26 202.147 1634.18 199.925 Q1637.12 197.703 1642.93 197.703 L1648.9 197.703 L1648.9 197.286 Q1648.9 194.347 1646.96 192.749 Q1645.04 191.129 1641.54 191.129 Q1639.32 191.129 1637.21 191.661 Q1635.11 192.194 1633.16 193.259 L1633.16 189.323 Q1635.5 188.421 1637.7 187.981 Q1639.9 187.518 1641.98 187.518 Q1647.61 187.518 1650.39 190.434 Q1653.16 193.351 1653.16 199.277 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1678.46 188.907 L1678.46 192.934 Q1676.66 192.009 1674.71 191.546 Q1672.77 191.083 1670.69 191.083 Q1667.51 191.083 1665.92 192.055 Q1664.34 193.027 1664.34 194.971 Q1664.34 196.453 1665.48 197.309 Q1666.61 198.143 1670.04 198.907 L1671.5 199.231 Q1676.03 200.203 1677.93 201.985 Q1679.85 203.745 1679.85 206.916 Q1679.85 210.527 1676.98 212.633 Q1674.13 214.74 1669.13 214.74 Q1667.05 214.74 1664.78 214.323 Q1662.54 213.93 1660.04 213.12 L1660.04 208.721 Q1662.4 209.948 1664.69 210.573 Q1666.98 211.175 1669.23 211.175 Q1672.24 211.175 1673.86 210.157 Q1675.48 209.115 1675.48 207.24 Q1675.48 205.504 1674.3 204.578 Q1673.14 203.652 1669.18 202.796 L1667.7 202.448 Q1663.74 201.615 1661.98 199.902 Q1660.22 198.166 1660.22 195.157 Q1660.22 191.499 1662.82 189.509 Q1665.41 187.518 1670.18 187.518 Q1672.54 187.518 1674.62 187.865 Q1676.7 188.212 1678.46 188.907 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1703.16 188.907 L1703.16 192.934 Q1701.36 192.009 1699.41 191.546 Q1697.47 191.083 1695.38 191.083 Q1692.21 191.083 1690.62 192.055 Q1689.04 193.027 1689.04 194.971 Q1689.04 196.453 1690.18 197.309 Q1691.31 198.143 1694.74 198.907 L1696.19 199.231 Q1700.73 200.203 1702.63 201.985 Q1704.55 203.745 1704.55 206.916 Q1704.55 210.527 1701.68 212.633 Q1698.83 214.74 1693.83 214.74 Q1691.75 214.74 1689.48 214.323 Q1687.24 213.93 1684.74 213.12 L1684.74 208.721 Q1687.1 209.948 1689.39 210.573 Q1691.68 211.175 1693.93 211.175 Q1696.94 211.175 1698.56 210.157 Q1700.18 209.115 1700.18 207.24 Q1700.18 205.504 1699 204.578 Q1697.84 203.652 1693.88 202.796 L1692.4 202.448 Q1688.44 201.615 1686.68 199.902 Q1684.92 198.166 1684.92 195.157 Q1684.92 191.499 1687.51 189.509 Q1690.11 187.518 1694.88 187.518 Q1697.24 187.518 1699.32 187.865 Q1701.4 188.212 1703.16 188.907 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1711.94 208.189 L1716.82 208.189 L1716.82 214.069 L1711.94 214.069 L1711.94 208.189 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1723.35 188.143 L1727.86 188.143 L1735.96 209.902 L1744.07 188.143 L1748.58 188.143 L1738.86 214.069 L1733.07 214.069 L1723.35 188.143 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1779.76 178.097 Q1776.66 183.421 1775.15 188.629 Q1773.65 193.837 1773.65 199.184 Q1773.65 204.532 1775.15 209.786 Q1776.68 215.018 1779.76 220.319 L1776.06 220.319 Q1772.58 214.879 1770.85 209.624 Q1769.13 204.37 1769.13 199.184 Q1769.13 194.022 1770.85 188.791 Q1772.56 183.56 1776.06 178.097 L1779.76 178.097 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1802.79 195.434 Q1806.15 196.152 1808.02 198.421 Q1809.92 200.689 1809.92 204.022 Q1809.92 209.138 1806.4 211.939 Q1802.88 214.74 1796.4 214.74 Q1794.23 214.74 1791.91 214.3 Q1789.62 213.883 1787.17 213.027 L1787.17 208.513 Q1789.11 209.647 1791.43 210.226 Q1793.74 210.805 1796.26 210.805 Q1800.66 210.805 1802.95 209.069 Q1805.27 207.333 1805.27 204.022 Q1805.27 200.967 1803.12 199.254 Q1800.99 197.518 1797.17 197.518 L1793.14 197.518 L1793.14 193.675 L1797.35 193.675 Q1800.8 193.675 1802.63 192.309 Q1804.46 190.921 1804.46 188.328 Q1804.46 185.666 1802.56 184.254 Q1800.69 182.819 1797.17 182.819 Q1795.25 182.819 1793.05 183.235 Q1790.85 183.652 1788.21 184.532 L1788.21 180.365 Q1790.87 179.624 1793.19 179.254 Q1795.52 178.884 1797.58 178.884 Q1802.91 178.884 1806.01 181.314 Q1809.11 183.722 1809.11 187.842 Q1809.11 190.712 1807.47 192.703 Q1805.82 194.671 1802.79 195.434 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1832.95 195.434 Q1836.31 196.152 1838.18 198.421 Q1840.08 200.689 1840.08 204.022 Q1840.08 209.138 1836.56 211.939 Q1833.05 214.74 1826.56 214.74 Q1824.39 214.74 1822.07 214.3 Q1819.78 213.883 1817.33 213.027 L1817.33 208.513 Q1819.27 209.647 1821.59 210.226 Q1823.9 210.805 1826.43 210.805 Q1830.82 210.805 1833.12 209.069 Q1835.43 207.333 1835.43 204.022 Q1835.43 200.967 1833.28 199.254 Q1831.15 197.518 1827.33 197.518 L1823.3 197.518 L1823.3 193.675 L1827.51 193.675 Q1830.96 193.675 1832.79 192.309 Q1834.62 190.921 1834.62 188.328 Q1834.62 185.666 1832.72 184.254 Q1830.85 182.819 1827.33 182.819 Q1825.41 182.819 1823.21 183.235 Q1821.01 183.652 1818.37 184.532 L1818.37 180.365 Q1821.03 179.624 1823.35 179.254 Q1825.69 178.884 1827.75 178.884 Q1833.07 178.884 1836.17 181.314 Q1839.27 183.722 1839.27 187.842 Q1839.27 190.712 1837.63 192.703 Q1835.99 194.671 1832.95 195.434 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1849 179.509 L1867.35 179.509 L1867.35 183.444 L1853.28 183.444 L1853.28 191.916 Q1854.3 191.569 1855.31 191.407 Q1856.33 191.222 1857.35 191.222 Q1863.14 191.222 1866.52 194.393 Q1869.9 197.564 1869.9 202.981 Q1869.9 208.559 1866.43 211.661 Q1862.95 214.74 1856.63 214.74 Q1854.46 214.74 1852.19 214.37 Q1849.94 213.999 1847.54 213.258 L1847.54 208.559 Q1849.62 209.694 1851.84 210.249 Q1854.06 210.805 1856.54 210.805 Q1860.55 210.805 1862.88 208.698 Q1865.22 206.592 1865.22 202.981 Q1865.22 199.37 1862.88 197.263 Q1860.55 195.157 1856.54 195.157 Q1854.67 195.157 1852.79 195.573 Q1850.94 195.99 1849 196.87 L1849 179.509 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1879.16 179.509 L1897.51 179.509 L1897.51 183.444 L1883.44 183.444 L1883.44 191.916 Q1884.46 191.569 1885.48 191.407 Q1886.49 191.222 1887.51 191.222 Q1893.3 191.222 1896.68 194.393 Q1900.06 197.564 1900.06 202.981 Q1900.06 208.559 1896.59 211.661 Q1893.12 214.74 1886.8 214.74 Q1884.62 214.74 1882.35 214.37 Q1880.11 213.999 1877.7 213.258 L1877.7 208.559 Q1879.78 209.694 1882 210.249 Q1884.23 210.805 1886.7 210.805 Q1890.71 210.805 1893.05 208.698 Q1895.38 206.592 1895.38 202.981 Q1895.38 199.37 1893.05 197.263 Q1890.71 195.157 1886.7 195.157 Q1884.83 195.157 1882.95 195.573 Q1881.1 195.99 1879.16 196.87 L1879.16 179.509 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1922.12 183.583 L1910.31 202.032 L1922.12 202.032 L1922.12 183.583 M1920.89 179.509 L1926.77 179.509 L1926.77 202.032 L1931.7 202.032 L1931.7 205.921 L1926.77 205.921 L1926.77 214.069 L1922.12 214.069 L1922.12 205.921 L1906.52 205.921 L1906.52 201.407 L1920.89 179.509 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1952.28 183.583 L1940.48 202.032 L1952.28 202.032 L1952.28 183.583 M1951.05 179.509 L1956.93 179.509 L1956.93 202.032 L1961.86 202.032 L1961.86 205.921 L1956.93 205.921 L1956.93 214.069 L1952.28 214.069 L1952.28 205.921 L1936.68 205.921 L1936.68 201.407 L1951.05 179.509 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1983.76 195.434 Q1987.12 196.152 1988.99 198.421 Q1990.89 200.689 1990.89 204.022 Q1990.89 209.138 1987.37 211.939 Q1983.86 214.74 1977.37 214.74 Q1975.2 214.74 1972.88 214.3 Q1970.59 213.883 1968.14 213.027 L1968.14 208.513 Q1970.08 209.647 1972.4 210.226 Q1974.71 210.805 1977.24 210.805 Q1981.63 210.805 1983.92 209.069 Q1986.24 207.333 1986.24 204.022 Q1986.24 200.967 1984.09 199.254 Q1981.96 197.518 1978.14 197.518 L1974.11 197.518 L1974.11 193.675 L1978.32 193.675 Q1981.77 193.675 1983.6 192.309 Q1985.43 190.921 1985.43 188.328 Q1985.43 185.666 1983.53 184.254 Q1981.66 182.819 1978.14 182.819 Q1976.22 182.819 1974.02 183.235 Q1971.82 183.652 1969.18 184.532 L1969.18 180.365 Q1971.84 179.624 1974.16 179.254 Q1976.49 178.884 1978.55 178.884 Q1983.88 178.884 1986.98 181.314 Q1990.08 183.722 1990.08 187.842 Q1990.08 190.712 1988.44 192.703 Q1986.8 194.671 1983.76 195.434 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M2013.92 195.434 Q2017.28 196.152 2019.16 198.421 Q2021.05 200.689 2021.05 204.022 Q2021.05 209.138 2017.54 211.939 Q2014.02 214.74 2007.54 214.74 Q2005.36 214.74 2003.05 214.3 Q2000.75 213.883 1998.3 213.027 L1998.3 208.513 Q2000.24 209.647 2002.56 210.226 Q2004.87 210.805 2007.4 210.805 Q2011.8 210.805 2014.09 209.069 Q2016.4 207.333 2016.4 204.022 Q2016.4 200.967 2014.25 199.254 Q2012.12 197.518 2008.3 197.518 L2004.27 197.518 L2004.27 193.675 L2008.48 193.675 Q2011.93 193.675 2013.76 192.309 Q2015.59 190.921 2015.59 188.328 Q2015.59 185.666 2013.69 184.254 Q2011.82 182.819 2008.3 182.819 Q2006.38 182.819 2004.18 183.235 Q2001.98 183.652 1999.34 184.532 L1999.34 180.365 Q2002 179.624 2004.32 179.254 Q2006.66 178.884 2008.72 178.884 Q2014.04 178.884 2017.14 181.314 Q2020.24 183.722 2020.24 187.842 Q2020.24 190.712 2018.6 192.703 Q2016.96 194.671 2013.92 195.434 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M2028.65 178.097 L2032.35 178.097 Q2035.82 183.56 2037.54 188.791 Q2039.27 194.022 2039.27 199.184 Q2039.27 204.37 2037.54 209.624 Q2035.82 214.879 2032.35 220.319 L2028.65 220.319 Q2031.73 215.018 2033.23 209.786 Q2034.76 204.532 2034.76 199.184 Q2034.76 193.837 2033.23 188.629 Q2031.73 183.421 2028.65 178.097 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip940)" style="stroke:#3da44d; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1415.15,248.629 1562.06,248.629 
  "/>
<path clip-path="url(#clip940)" d="M1606.89 244.96 Q1608.49 242.089 1610.71 240.724 Q1612.93 239.358 1615.94 239.358 Q1619.99 239.358 1622.19 242.205 Q1624.39 245.029 1624.39 250.261 L1624.39 265.909 L1620.11 265.909 L1620.11 250.399 Q1620.11 246.673 1618.79 244.867 Q1617.47 243.062 1614.76 243.062 Q1611.45 243.062 1609.53 245.261 Q1607.61 247.46 1607.61 251.256 L1607.61 265.909 L1603.33 265.909 L1603.33 250.399 Q1603.33 246.649 1602.01 244.867 Q1600.69 243.062 1597.93 243.062 Q1594.67 243.062 1592.75 245.284 Q1590.83 247.483 1590.83 251.256 L1590.83 265.909 L1586.54 265.909 L1586.54 239.983 L1590.83 239.983 L1590.83 244.011 Q1592.28 241.626 1594.32 240.492 Q1596.36 239.358 1599.16 239.358 Q1601.98 239.358 1603.95 240.793 Q1605.94 242.228 1606.89 244.96 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1644.67 252.876 Q1639.51 252.876 1637.51 254.057 Q1635.52 255.237 1635.52 258.085 Q1635.52 260.353 1637.01 261.696 Q1638.51 263.015 1641.08 263.015 Q1644.62 263.015 1646.75 260.515 Q1648.9 257.992 1648.9 253.825 L1648.9 252.876 L1644.67 252.876 M1653.16 251.117 L1653.16 265.909 L1648.9 265.909 L1648.9 261.973 Q1647.45 264.335 1645.27 265.469 Q1643.09 266.58 1639.95 266.58 Q1635.96 266.58 1633.6 264.358 Q1631.26 262.112 1631.26 258.362 Q1631.26 253.987 1634.18 251.765 Q1637.12 249.543 1642.93 249.543 L1648.9 249.543 L1648.9 249.126 Q1648.9 246.187 1646.96 244.589 Q1645.04 242.969 1641.54 242.969 Q1639.32 242.969 1637.21 243.501 Q1635.11 244.034 1633.16 245.099 L1633.16 241.163 Q1635.5 240.261 1637.7 239.821 Q1639.9 239.358 1641.98 239.358 Q1647.61 239.358 1650.39 242.274 Q1653.16 245.191 1653.16 251.117 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1678.46 240.747 L1678.46 244.774 Q1676.66 243.849 1674.71 243.386 Q1672.77 242.923 1670.69 242.923 Q1667.51 242.923 1665.92 243.895 Q1664.34 244.867 1664.34 246.811 Q1664.34 248.293 1665.48 249.149 Q1666.61 249.983 1670.04 250.747 L1671.5 251.071 Q1676.03 252.043 1677.93 253.825 Q1679.85 255.585 1679.85 258.756 Q1679.85 262.367 1676.98 264.473 Q1674.13 266.58 1669.13 266.58 Q1667.05 266.58 1664.78 266.163 Q1662.54 265.77 1660.04 264.96 L1660.04 260.561 Q1662.4 261.788 1664.69 262.413 Q1666.98 263.015 1669.23 263.015 Q1672.24 263.015 1673.86 261.997 Q1675.48 260.955 1675.48 259.08 Q1675.48 257.344 1674.3 256.418 Q1673.14 255.492 1669.18 254.636 L1667.7 254.288 Q1663.74 253.455 1661.98 251.742 Q1660.22 250.006 1660.22 246.997 Q1660.22 243.339 1662.82 241.349 Q1665.41 239.358 1670.18 239.358 Q1672.54 239.358 1674.62 239.705 Q1676.7 240.052 1678.46 240.747 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1703.16 240.747 L1703.16 244.774 Q1701.36 243.849 1699.41 243.386 Q1697.47 242.923 1695.38 242.923 Q1692.21 242.923 1690.62 243.895 Q1689.04 244.867 1689.04 246.811 Q1689.04 248.293 1690.18 249.149 Q1691.31 249.983 1694.74 250.747 L1696.19 251.071 Q1700.73 252.043 1702.63 253.825 Q1704.55 255.585 1704.55 258.756 Q1704.55 262.367 1701.68 264.473 Q1698.83 266.58 1693.83 266.58 Q1691.75 266.58 1689.48 266.163 Q1687.24 265.77 1684.74 264.96 L1684.74 260.561 Q1687.1 261.788 1689.39 262.413 Q1691.68 263.015 1693.93 263.015 Q1696.94 263.015 1698.56 261.997 Q1700.18 260.955 1700.18 259.08 Q1700.18 257.344 1699 256.418 Q1697.84 255.492 1693.88 254.636 L1692.4 254.288 Q1688.44 253.455 1686.68 251.742 Q1684.92 250.006 1684.92 246.997 Q1684.92 243.339 1687.51 241.349 Q1690.11 239.358 1694.88 239.358 Q1697.24 239.358 1699.32 239.705 Q1701.4 240.052 1703.16 240.747 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1711.94 260.029 L1716.82 260.029 L1716.82 265.909 L1711.94 265.909 L1711.94 260.029 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1738.19 252.876 Q1733.02 252.876 1731.03 254.057 Q1729.04 255.237 1729.04 258.085 Q1729.04 260.353 1730.52 261.696 Q1732.03 263.015 1734.6 263.015 Q1738.14 263.015 1740.27 260.515 Q1742.42 257.992 1742.42 253.825 L1742.42 252.876 L1738.19 252.876 M1746.68 251.117 L1746.68 265.909 L1742.42 265.909 L1742.42 261.973 Q1740.96 264.335 1738.79 265.469 Q1736.61 266.58 1733.46 266.58 Q1729.48 266.58 1727.12 264.358 Q1724.78 262.112 1724.78 258.362 Q1724.78 253.987 1727.7 251.765 Q1730.64 249.543 1736.45 249.543 L1742.42 249.543 L1742.42 249.126 Q1742.42 246.187 1740.48 244.589 Q1738.56 242.969 1735.06 242.969 Q1732.84 242.969 1730.73 243.501 Q1728.63 244.034 1726.68 245.099 L1726.68 241.163 Q1729.02 240.261 1731.22 239.821 Q1733.42 239.358 1735.5 239.358 Q1741.13 239.358 1743.9 242.274 Q1746.68 245.191 1746.68 251.117 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1775.15 273.779 L1775.15 277.089 L1750.52 277.089 L1750.52 273.779 L1775.15 273.779 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1794.18 243.964 Q1793.46 243.548 1792.61 243.362 Q1791.77 243.154 1790.75 243.154 Q1787.14 243.154 1785.2 245.515 Q1783.28 247.853 1783.28 252.251 L1783.28 265.909 L1779 265.909 L1779 239.983 L1783.28 239.983 L1783.28 244.011 Q1784.62 241.649 1786.77 240.515 Q1788.93 239.358 1792 239.358 Q1792.44 239.358 1792.98 239.427 Q1793.51 239.474 1794.16 239.589 L1794.18 243.964 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1819.78 251.881 L1819.78 253.964 L1800.2 253.964 Q1800.48 258.362 1802.84 260.677 Q1805.22 262.969 1809.46 262.969 Q1811.91 262.969 1814.2 262.367 Q1816.52 261.765 1818.79 260.561 L1818.79 264.589 Q1816.5 265.561 1814.09 266.071 Q1811.68 266.58 1809.2 266.58 Q1803 266.58 1799.37 262.969 Q1795.75 259.358 1795.75 253.2 Q1795.75 246.835 1799.18 243.108 Q1802.63 239.358 1808.46 239.358 Q1813.69 239.358 1816.73 242.737 Q1819.78 246.094 1819.78 251.881 M1815.52 250.631 Q1815.48 247.136 1813.56 245.052 Q1811.66 242.969 1808.51 242.969 Q1804.94 242.969 1802.79 244.983 Q1800.66 246.997 1800.34 250.654 L1815.52 250.631 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1826.77 229.89 L1831.03 229.89 L1831.03 265.909 L1826.77 265.909 L1826.77 229.89 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1853.07 229.89 L1853.07 233.432 L1849 233.432 Q1846.7 233.432 1845.8 234.358 Q1844.92 235.284 1844.92 237.691 L1844.92 239.983 L1851.93 239.983 L1851.93 243.293 L1844.92 243.293 L1844.92 265.909 L1840.64 265.909 L1840.64 243.293 L1836.56 243.293 L1836.56 239.983 L1840.64 239.983 L1840.64 238.177 Q1840.64 233.849 1842.65 231.881 Q1844.67 229.89 1849.04 229.89 L1853.07 229.89 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1871.66 243.964 Q1870.94 243.548 1870.08 243.362 Q1869.25 243.154 1868.23 243.154 Q1864.62 243.154 1862.68 245.515 Q1860.75 247.853 1860.75 252.251 L1860.75 265.909 L1856.47 265.909 L1856.47 239.983 L1860.75 239.983 L1860.75 244.011 Q1862.1 241.649 1864.25 240.515 Q1866.4 239.358 1869.48 239.358 Q1869.92 239.358 1870.45 239.427 Q1870.99 239.474 1871.63 239.589 L1871.66 243.964 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1876.12 239.983 L1880.38 239.983 L1880.38 265.909 L1876.12 265.909 L1876.12 239.983 M1876.12 229.89 L1880.38 229.89 L1880.38 235.284 L1876.12 235.284 L1876.12 229.89 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1907.95 240.978 L1907.95 244.96 Q1906.15 243.964 1904.32 243.478 Q1902.51 242.969 1900.66 242.969 Q1896.52 242.969 1894.23 245.608 Q1891.93 248.224 1891.93 252.969 Q1891.93 257.714 1894.23 260.353 Q1896.52 262.969 1900.66 262.969 Q1902.51 262.969 1904.32 262.483 Q1906.15 261.973 1907.95 260.978 L1907.95 264.913 Q1906.17 265.747 1904.25 266.163 Q1902.35 266.58 1900.2 266.58 Q1894.34 266.58 1890.89 262.899 Q1887.44 259.219 1887.44 252.969 Q1887.44 246.626 1890.92 242.992 Q1894.41 239.358 1900.48 239.358 Q1902.44 239.358 1904.32 239.775 Q1906.19 240.168 1907.95 240.978 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1940.66 229.937 Q1937.56 235.261 1936.05 240.469 Q1934.55 245.677 1934.55 251.024 Q1934.55 256.372 1936.05 261.626 Q1937.58 266.858 1940.66 272.159 L1936.96 272.159 Q1933.49 266.719 1931.75 261.464 Q1930.04 256.21 1930.04 251.024 Q1930.04 245.862 1931.75 240.631 Q1933.46 235.4 1936.96 229.937 L1940.66 229.937 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1949.57 231.349 L1967.93 231.349 L1967.93 235.284 L1953.86 235.284 L1953.86 243.756 Q1954.87 243.409 1955.89 243.247 Q1956.91 243.062 1957.93 243.062 Q1963.72 243.062 1967.1 246.233 Q1970.48 249.404 1970.48 254.821 Q1970.48 260.399 1967 263.501 Q1963.53 266.58 1957.21 266.58 Q1955.04 266.58 1952.77 266.21 Q1950.52 265.839 1948.11 265.098 L1948.11 260.399 Q1950.2 261.534 1952.42 262.089 Q1954.64 262.645 1957.12 262.645 Q1961.12 262.645 1963.46 260.538 Q1965.8 258.432 1965.8 254.821 Q1965.8 251.21 1963.46 249.103 Q1961.12 246.997 1957.12 246.997 Q1955.24 246.997 1953.37 247.413 Q1951.52 247.83 1949.57 248.71 L1949.57 231.349 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1989.69 249.497 Q1986.36 249.497 1984.43 251.279 Q1982.54 253.061 1982.54 256.186 Q1982.54 259.311 1984.43 261.094 Q1986.36 262.876 1989.69 262.876 Q1993.02 262.876 1994.94 261.094 Q1996.86 259.288 1996.86 256.186 Q1996.86 253.061 1994.94 251.279 Q1993.05 249.497 1989.69 249.497 M1985.01 247.506 Q1982 246.765 1980.31 244.705 Q1978.65 242.645 1978.65 239.682 Q1978.65 235.538 1981.59 233.131 Q1984.55 230.724 1989.69 230.724 Q1994.85 230.724 1997.79 233.131 Q2000.73 235.538 2000.73 239.682 Q2000.73 242.645 1999.04 244.705 Q1997.37 246.765 1994.39 247.506 Q1997.77 248.293 1999.64 250.585 Q2001.54 252.876 2001.54 256.186 Q2001.54 261.21 1998.46 263.895 Q1995.41 266.58 1989.69 266.58 Q1983.97 266.58 1980.89 263.895 Q1977.84 261.21 1977.84 256.186 Q1977.84 252.876 1979.74 250.585 Q1981.63 248.293 1985.01 247.506 M1983.3 240.122 Q1983.3 242.807 1984.97 244.312 Q1986.66 245.816 1989.69 245.816 Q1992.7 245.816 1994.39 244.312 Q1996.1 242.807 1996.1 240.122 Q1996.1 237.437 1994.39 235.932 Q1992.7 234.427 1989.69 234.427 Q1986.66 234.427 1984.97 235.932 Q1983.3 237.437 1983.3 240.122 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M2008.67 231.349 L2030.89 231.349 L2030.89 233.339 L2018.35 265.909 L2013.46 265.909 L2025.27 235.284 L2008.67 235.284 L2008.67 231.349 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M2044.04 261.973 L2060.36 261.973 L2060.36 265.909 L2038.42 265.909 L2038.42 261.973 Q2041.08 259.219 2045.66 254.589 Q2050.27 249.936 2051.45 248.594 Q2053.69 246.071 2054.57 244.335 Q2055.48 242.575 2055.48 240.886 Q2055.48 238.131 2053.53 236.395 Q2051.61 234.659 2048.51 234.659 Q2046.31 234.659 2043.86 235.423 Q2041.42 236.187 2038.65 237.737 L2038.65 233.015 Q2041.47 231.881 2043.92 231.302 Q2046.38 230.724 2048.42 230.724 Q2053.79 230.724 2056.98 233.409 Q2060.17 236.094 2060.17 240.585 Q2060.17 242.714 2059.36 244.636 Q2058.58 246.534 2056.47 249.126 Q2055.89 249.798 2052.79 253.015 Q2049.69 256.21 2044.04 261.973 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M2080.17 234.427 Q2076.56 234.427 2074.73 237.992 Q2072.93 241.534 2072.93 248.663 Q2072.93 255.77 2074.73 259.335 Q2076.56 262.876 2080.17 262.876 Q2083.81 262.876 2085.61 259.335 Q2087.44 255.77 2087.44 248.663 Q2087.44 241.534 2085.61 237.992 Q2083.81 234.427 2080.17 234.427 M2080.17 230.724 Q2085.98 230.724 2089.04 235.33 Q2092.12 239.913 2092.12 248.663 Q2092.12 257.39 2089.04 261.997 Q2085.98 266.58 2080.17 266.58 Q2074.36 266.58 2071.29 261.997 Q2068.23 257.39 2068.23 248.663 Q2068.23 239.913 2071.29 235.33 Q2074.36 230.724 2080.17 230.724 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M2104.36 261.973 L2120.68 261.973 L2120.68 265.909 L2098.74 265.909 L2098.74 261.973 Q2101.4 259.219 2105.98 254.589 Q2110.59 249.936 2111.77 248.594 Q2114.02 246.071 2114.9 244.335 Q2115.8 242.575 2115.8 240.886 Q2115.8 238.131 2113.85 236.395 Q2111.93 234.659 2108.83 234.659 Q2106.63 234.659 2104.18 235.423 Q2101.75 236.187 2098.97 237.737 L2098.97 233.015 Q2101.79 231.881 2104.25 231.302 Q2106.7 230.724 2108.74 230.724 Q2114.11 230.724 2117.3 233.409 Q2120.5 236.094 2120.5 240.585 Q2120.5 242.714 2119.69 244.636 Q2118.9 246.534 2116.79 249.126 Q2116.22 249.798 2113.11 253.015 Q2110.01 256.21 2104.36 261.973 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M2130.54 231.349 L2148.9 231.349 L2148.9 235.284 L2134.83 235.284 L2134.83 243.756 Q2135.85 243.409 2136.86 243.247 Q2137.88 243.062 2138.9 243.062 Q2144.69 243.062 2148.07 246.233 Q2151.45 249.404 2151.45 254.821 Q2151.45 260.399 2147.97 263.501 Q2144.5 266.58 2138.18 266.58 Q2136.01 266.58 2133.74 266.21 Q2131.49 265.839 2129.09 265.098 L2129.09 260.399 Q2131.17 261.534 2133.39 262.089 Q2135.61 262.645 2138.09 262.645 Q2142.1 262.645 2144.43 260.538 Q2146.77 258.432 2146.77 254.821 Q2146.77 251.21 2144.43 249.103 Q2142.1 246.997 2138.09 246.997 Q2136.22 246.997 2134.34 247.413 Q2132.49 247.83 2130.54 248.71 L2130.54 231.349 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M2171.24 246.765 Q2168.09 246.765 2166.24 248.918 Q2164.41 251.071 2164.41 254.821 Q2164.41 258.548 2166.24 260.723 Q2168.09 262.876 2171.24 262.876 Q2174.39 262.876 2176.22 260.723 Q2178.07 258.548 2178.07 254.821 Q2178.07 251.071 2176.22 248.918 Q2174.39 246.765 2171.24 246.765 M2180.52 232.113 L2180.52 236.372 Q2178.76 235.538 2176.96 235.099 Q2175.17 234.659 2173.41 234.659 Q2168.79 234.659 2166.33 237.784 Q2163.9 240.909 2163.55 247.228 Q2164.92 245.214 2166.98 244.149 Q2169.04 243.062 2171.52 243.062 Q2176.72 243.062 2179.73 246.233 Q2182.77 249.381 2182.77 254.821 Q2182.77 260.145 2179.62 263.362 Q2176.47 266.58 2171.24 266.58 Q2165.24 266.58 2162.07 261.997 Q2158.9 257.39 2158.9 248.663 Q2158.9 240.469 2162.79 235.608 Q2166.68 230.724 2173.23 230.724 Q2174.99 230.724 2176.77 231.071 Q2178.58 231.418 2180.52 232.113 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M2191.63 261.973 L2199.27 261.973 L2199.27 235.608 L2190.96 237.275 L2190.96 233.015 L2199.22 231.349 L2203.9 231.349 L2203.9 261.973 L2211.54 261.973 L2211.54 265.909 L2191.63 265.909 L2191.63 261.973 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M2219.71 229.937 L2223.41 229.937 Q2226.89 235.4 2228.6 240.631 Q2230.34 245.862 2230.34 251.024 Q2230.34 256.21 2228.6 261.464 Q2226.89 266.719 2223.41 272.159 L2219.71 272.159 Q2222.79 266.858 2224.29 261.626 Q2225.82 256.372 2225.82 251.024 Q2225.82 245.677 2224.29 240.469 Q2222.79 235.261 2219.71 229.937 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip940)" style="stroke:#c271d2; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1415.15,300.469 1562.06,300.469 
  "/>
<path clip-path="url(#clip940)" d="M1606.89 296.8 Q1608.49 293.929 1610.71 292.564 Q1612.93 291.198 1615.94 291.198 Q1619.99 291.198 1622.19 294.045 Q1624.39 296.869 1624.39 302.101 L1624.39 317.749 L1620.11 317.749 L1620.11 302.239 Q1620.11 298.513 1618.79 296.707 Q1617.47 294.902 1614.76 294.902 Q1611.45 294.902 1609.53 297.101 Q1607.61 299.3 1607.61 303.096 L1607.61 317.749 L1603.33 317.749 L1603.33 302.239 Q1603.33 298.489 1602.01 296.707 Q1600.69 294.902 1597.93 294.902 Q1594.67 294.902 1592.75 297.124 Q1590.83 299.323 1590.83 303.096 L1590.83 317.749 L1586.54 317.749 L1586.54 291.823 L1590.83 291.823 L1590.83 295.851 Q1592.28 293.466 1594.32 292.332 Q1596.36 291.198 1599.16 291.198 Q1601.98 291.198 1603.95 292.633 Q1605.94 294.068 1606.89 296.8 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1644.67 304.716 Q1639.51 304.716 1637.51 305.897 Q1635.52 307.077 1635.52 309.925 Q1635.52 312.193 1637.01 313.536 Q1638.51 314.855 1641.08 314.855 Q1644.62 314.855 1646.75 312.355 Q1648.9 309.832 1648.9 305.665 L1648.9 304.716 L1644.67 304.716 M1653.16 302.957 L1653.16 317.749 L1648.9 317.749 L1648.9 313.813 Q1647.45 316.175 1645.27 317.309 Q1643.09 318.42 1639.95 318.42 Q1635.96 318.42 1633.6 316.198 Q1631.26 313.952 1631.26 310.202 Q1631.26 305.827 1634.18 303.605 Q1637.12 301.383 1642.93 301.383 L1648.9 301.383 L1648.9 300.966 Q1648.9 298.027 1646.96 296.429 Q1645.04 294.809 1641.54 294.809 Q1639.32 294.809 1637.21 295.341 Q1635.11 295.874 1633.16 296.939 L1633.16 293.003 Q1635.5 292.101 1637.7 291.661 Q1639.9 291.198 1641.98 291.198 Q1647.61 291.198 1650.39 294.114 Q1653.16 297.031 1653.16 302.957 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1678.46 292.587 L1678.46 296.614 Q1676.66 295.689 1674.71 295.226 Q1672.77 294.763 1670.69 294.763 Q1667.51 294.763 1665.92 295.735 Q1664.34 296.707 1664.34 298.651 Q1664.34 300.133 1665.48 300.989 Q1666.61 301.823 1670.04 302.587 L1671.5 302.911 Q1676.03 303.883 1677.93 305.665 Q1679.85 307.425 1679.85 310.596 Q1679.85 314.207 1676.98 316.313 Q1674.13 318.42 1669.13 318.42 Q1667.05 318.42 1664.78 318.003 Q1662.54 317.61 1660.04 316.8 L1660.04 312.401 Q1662.4 313.628 1664.69 314.253 Q1666.98 314.855 1669.23 314.855 Q1672.24 314.855 1673.86 313.837 Q1675.48 312.795 1675.48 310.92 Q1675.48 309.184 1674.3 308.258 Q1673.14 307.332 1669.18 306.476 L1667.7 306.128 Q1663.74 305.295 1661.98 303.582 Q1660.22 301.846 1660.22 298.837 Q1660.22 295.179 1662.82 293.189 Q1665.41 291.198 1670.18 291.198 Q1672.54 291.198 1674.62 291.545 Q1676.7 291.892 1678.46 292.587 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1703.16 292.587 L1703.16 296.614 Q1701.36 295.689 1699.41 295.226 Q1697.47 294.763 1695.38 294.763 Q1692.21 294.763 1690.62 295.735 Q1689.04 296.707 1689.04 298.651 Q1689.04 300.133 1690.18 300.989 Q1691.31 301.823 1694.74 302.587 L1696.19 302.911 Q1700.73 303.883 1702.63 305.665 Q1704.55 307.425 1704.55 310.596 Q1704.55 314.207 1701.68 316.313 Q1698.83 318.42 1693.83 318.42 Q1691.75 318.42 1689.48 318.003 Q1687.24 317.61 1684.74 316.8 L1684.74 312.401 Q1687.1 313.628 1689.39 314.253 Q1691.68 314.855 1693.93 314.855 Q1696.94 314.855 1698.56 313.837 Q1700.18 312.795 1700.18 310.92 Q1700.18 309.184 1699 308.258 Q1697.84 307.332 1693.88 306.476 L1692.4 306.128 Q1688.44 305.295 1686.68 303.582 Q1684.92 301.846 1684.92 298.837 Q1684.92 295.179 1687.51 293.189 Q1690.11 291.198 1694.88 291.198 Q1697.24 291.198 1699.32 291.545 Q1701.4 291.892 1703.16 292.587 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1711.94 311.869 L1716.82 311.869 L1716.82 317.749 L1711.94 317.749 L1711.94 311.869 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1739.53 281.73 L1739.53 285.272 L1735.45 285.272 Q1733.16 285.272 1732.26 286.198 Q1731.38 287.124 1731.38 289.531 L1731.38 291.823 L1738.39 291.823 L1738.39 295.133 L1731.38 295.133 L1731.38 317.749 L1727.1 317.749 L1727.1 295.133 L1723.02 295.133 L1723.02 291.823 L1727.1 291.823 L1727.1 290.017 Q1727.1 285.689 1729.11 283.721 Q1731.13 281.73 1735.5 281.73 L1739.53 281.73 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1768.39 281.777 Q1765.29 287.101 1763.79 292.309 Q1762.28 297.517 1762.28 302.864 Q1762.28 308.212 1763.79 313.466 Q1765.31 318.698 1768.39 323.999 L1764.69 323.999 Q1761.22 318.559 1759.48 313.304 Q1757.77 308.05 1757.77 302.864 Q1757.77 297.702 1759.48 292.471 Q1761.19 287.24 1764.69 281.777 L1768.39 281.777 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1787.84 298.605 Q1784.69 298.605 1782.84 300.758 Q1781.01 302.911 1781.01 306.661 Q1781.01 310.388 1782.84 312.563 Q1784.69 314.716 1787.84 314.716 Q1790.99 314.716 1792.81 312.563 Q1794.67 310.388 1794.67 306.661 Q1794.67 302.911 1792.81 300.758 Q1790.99 298.605 1787.84 298.605 M1797.12 283.953 L1797.12 288.212 Q1795.36 287.378 1793.56 286.939 Q1791.77 286.499 1790.01 286.499 Q1785.38 286.499 1782.93 289.624 Q1780.5 292.749 1780.15 299.068 Q1781.52 297.054 1783.58 295.989 Q1785.64 294.902 1788.12 294.902 Q1793.32 294.902 1796.33 298.073 Q1799.37 301.221 1799.37 306.661 Q1799.37 311.985 1796.22 315.202 Q1793.07 318.42 1787.84 318.42 Q1781.84 318.42 1778.67 313.837 Q1775.5 309.23 1775.5 300.503 Q1775.5 292.309 1779.39 287.448 Q1783.28 282.564 1789.83 282.564 Q1791.59 282.564 1793.37 282.911 Q1795.18 283.258 1797.12 283.953 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1821.59 299.114 Q1824.94 299.832 1826.82 302.101 Q1828.72 304.369 1828.72 307.702 Q1828.72 312.818 1825.2 315.619 Q1821.68 318.42 1815.2 318.42 Q1813.02 318.42 1810.71 317.98 Q1808.42 317.563 1805.96 316.707 L1805.96 312.193 Q1807.91 313.327 1810.22 313.906 Q1812.54 314.485 1815.06 314.485 Q1819.46 314.485 1821.75 312.749 Q1824.06 311.013 1824.06 307.702 Q1824.06 304.647 1821.91 302.934 Q1819.78 301.198 1815.96 301.198 L1811.94 301.198 L1811.94 297.355 L1816.15 297.355 Q1819.6 297.355 1821.43 295.989 Q1823.25 294.601 1823.25 292.008 Q1823.25 289.346 1821.36 287.934 Q1819.48 286.499 1815.96 286.499 Q1814.04 286.499 1811.84 286.915 Q1809.64 287.332 1807 288.212 L1807 284.045 Q1809.67 283.304 1811.98 282.934 Q1814.32 282.564 1816.38 282.564 Q1821.7 282.564 1824.81 284.994 Q1827.91 287.402 1827.91 291.522 Q1827.91 294.392 1826.26 296.383 Q1824.62 298.351 1821.59 299.114 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1836.4 283.189 L1858.62 283.189 L1858.62 285.179 L1846.08 317.749 L1841.19 317.749 L1853 287.124 L1836.4 287.124 L1836.4 283.189 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1867.79 283.189 L1886.15 283.189 L1886.15 287.124 L1872.07 287.124 L1872.07 295.596 Q1873.09 295.249 1874.11 295.087 Q1875.13 294.902 1876.15 294.902 Q1881.93 294.902 1885.31 298.073 Q1888.69 301.244 1888.69 306.661 Q1888.69 312.239 1885.22 315.341 Q1881.75 318.42 1875.43 318.42 Q1873.25 318.42 1870.99 318.05 Q1868.74 317.679 1866.33 316.938 L1866.33 312.239 Q1868.42 313.374 1870.64 313.929 Q1872.86 314.485 1875.34 314.485 Q1879.34 314.485 1881.68 312.378 Q1884.02 310.272 1884.02 306.661 Q1884.02 303.05 1881.68 300.943 Q1879.34 298.837 1875.34 298.837 Q1873.46 298.837 1871.59 299.253 Q1869.74 299.67 1867.79 300.55 L1867.79 283.189 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1912.07 299.114 Q1915.43 299.832 1917.3 302.101 Q1919.2 304.369 1919.2 307.702 Q1919.2 312.818 1915.68 315.619 Q1912.17 318.42 1905.68 318.42 Q1903.51 318.42 1901.19 317.98 Q1898.9 317.563 1896.45 316.707 L1896.45 312.193 Q1898.39 313.327 1900.71 313.906 Q1903.02 314.485 1905.55 314.485 Q1909.94 314.485 1912.24 312.749 Q1914.55 311.013 1914.55 307.702 Q1914.55 304.647 1912.4 302.934 Q1910.27 301.198 1906.45 301.198 L1902.42 301.198 L1902.42 297.355 L1906.63 297.355 Q1910.08 297.355 1911.91 295.989 Q1913.74 294.601 1913.74 292.008 Q1913.74 289.346 1911.84 287.934 Q1909.97 286.499 1906.45 286.499 Q1904.53 286.499 1902.33 286.915 Q1900.13 287.332 1897.49 288.212 L1897.49 284.045 Q1900.15 283.304 1902.47 282.934 Q1904.8 282.564 1906.87 282.564 Q1912.19 282.564 1915.29 284.994 Q1918.39 287.402 1918.39 291.522 Q1918.39 294.392 1916.75 296.383 Q1915.11 298.351 1912.07 299.114 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1940.92 287.263 L1929.11 305.712 L1940.92 305.712 L1940.92 287.263 M1939.69 283.189 L1945.57 283.189 L1945.57 305.712 L1950.5 305.712 L1950.5 309.601 L1945.57 309.601 L1945.57 317.749 L1940.92 317.749 L1940.92 309.601 L1925.31 309.601 L1925.31 305.087 L1939.69 283.189 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1962.26 313.813 L1978.58 313.813 L1978.58 317.749 L1956.63 317.749 L1956.63 313.813 Q1959.3 311.059 1963.88 306.429 Q1968.49 301.776 1969.67 300.434 Q1971.91 297.911 1972.79 296.175 Q1973.69 294.415 1973.69 292.726 Q1973.69 289.971 1971.75 288.235 Q1969.83 286.499 1966.73 286.499 Q1964.53 286.499 1962.07 287.263 Q1959.64 288.027 1956.86 289.577 L1956.86 284.855 Q1959.69 283.721 1962.14 283.142 Q1964.6 282.564 1966.63 282.564 Q1972 282.564 1975.2 285.249 Q1978.39 287.934 1978.39 292.425 Q1978.39 294.554 1977.58 296.476 Q1976.8 298.374 1974.69 300.966 Q1974.11 301.638 1971.01 304.855 Q1967.91 308.05 1962.26 313.813 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1989.2 313.813 L1996.84 313.813 L1996.84 287.448 L1988.53 289.115 L1988.53 284.855 L1996.8 283.189 L2001.47 283.189 L2001.47 313.813 L2009.11 313.813 L2009.11 317.749 L1989.2 317.749 L1989.2 313.813 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M2019.36 313.813 L2027 313.813 L2027 287.448 L2018.69 289.115 L2018.69 284.855 L2026.96 283.189 L2031.63 283.189 L2031.63 313.813 L2039.27 313.813 L2039.27 317.749 L2019.36 317.749 L2019.36 313.813 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M2047.44 281.777 L2051.15 281.777 Q2054.62 287.24 2056.33 292.471 Q2058.07 297.702 2058.07 302.864 Q2058.07 308.05 2056.33 313.304 Q2054.62 318.559 2051.15 323.999 L2047.44 323.999 Q2050.52 318.698 2052.03 313.466 Q2053.55 308.212 2053.55 302.864 Q2053.55 297.517 2052.03 292.309 Q2050.52 287.101 2047.44 281.777 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /></svg>



#### Define functions

The structure of the previous code section is used more often in the further sections, so for clarity the previously explained code sections for resetting, initializing and simulating are combined into one function `simulate()`.


```julia
function simulate(FMU, states, x₀, variables, tStart, tStop, tSave)
    fmiReset(FMU)
    fmiSetupExperiment(FMU, tStart, tStop)

    fmiSetReal(FMU, states, x₀)
    fmiEnterInitializationMode(FMU)
    fmiExitInitializationMode(FMU)


    success, simData = fmiSimulate(FMU, tStart, tStop; recordValues=variables, saveat=tSave, setup=false, reset=false)
    return simData
end
```


    simulate (generic function with 1 method)


Also, a function to extract the position and velocity from the simulation data is created.


```julia
function extractPosVel(simData)
    posData = collect(data[1] for data in simData)
    velData = collect(data[2] for data in simData)
    return posData, velData
end
```


    extractPosVel (generic function with 1 method)


#### Modified initial states

In contrast to the previous section, other initial states are selected. The position of the mass is initilized with $1.0m$ and the velocity is initialized with $-1.5\frac{m}{s}$. With the modified initial states the *realFMU* is simulated and a graph is generated.


```julia
xMod₀ = [1.0, -1.5]
realSimDataMod = simulate(realFMU, states, xMod₀, vrs, tStart, tStop, tSave)
fmiPlot(realFMU, vrs, realSimDataMod)
```

    ┌ Warning: fmi2Reset(_): Should be only called in FMU state `modelSetableFMUstate`.
    └ @ FMIImport C:\Users\thummeto\Documents\FMIImport.jl\src\FMI2_c.jl:219



<?xml version="1.0" encoding="utf-8"?>
<svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" width="600" height="400" viewBox="0 0 2400 1600">
<defs>
  <clipPath id="clip030">
    <rect x="0" y="0" width="2400" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip030)" d="
M0 1600 L2400 1600 L2400 0 L0 0  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip031">
    <rect x="480" y="0" width="1681" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip030)" d="
M149.191 1423.18 L2352.76 1423.18 L2352.76 47.2441 L149.191 47.2441  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip032">
    <rect x="149" y="47" width="2205" height="1377"/>
  </clipPath>
</defs>
<polyline clip-path="url(#clip032)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  211.556,1423.18 211.556,47.2441 
  "/>
<polyline clip-path="url(#clip032)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  731.265,1423.18 731.265,47.2441 
  "/>
<polyline clip-path="url(#clip032)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1250.97,1423.18 1250.97,47.2441 
  "/>
<polyline clip-path="url(#clip032)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1770.68,1423.18 1770.68,47.2441 
  "/>
<polyline clip-path="url(#clip032)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  2290.39,1423.18 2290.39,47.2441 
  "/>
<polyline clip-path="url(#clip030)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  149.191,1423.18 2352.76,1423.18 
  "/>
<polyline clip-path="url(#clip030)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  211.556,1423.18 211.556,1404.28 
  "/>
<polyline clip-path="url(#clip030)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  731.265,1423.18 731.265,1404.28 
  "/>
<polyline clip-path="url(#clip030)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1250.97,1423.18 1250.97,1404.28 
  "/>
<polyline clip-path="url(#clip030)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1770.68,1423.18 1770.68,1404.28 
  "/>
<polyline clip-path="url(#clip030)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  2290.39,1423.18 2290.39,1404.28 
  "/>
<path clip-path="url(#clip030)" d="M211.556 1454.1 Q207.945 1454.1 206.116 1457.66 Q204.311 1461.2 204.311 1468.33 Q204.311 1475.44 206.116 1479.01 Q207.945 1482.55 211.556 1482.55 Q215.19 1482.55 216.996 1479.01 Q218.824 1475.44 218.824 1468.33 Q218.824 1461.2 216.996 1457.66 Q215.19 1454.1 211.556 1454.1 M211.556 1450.39 Q217.366 1450.39 220.422 1455 Q223.5 1459.58 223.5 1468.33 Q223.5 1477.06 220.422 1481.67 Q217.366 1486.25 211.556 1486.25 Q205.746 1486.25 202.667 1481.67 Q199.612 1477.06 199.612 1468.33 Q199.612 1459.58 202.667 1455 Q205.746 1450.39 211.556 1450.39 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M721.647 1481.64 L729.285 1481.64 L729.285 1455.28 L720.975 1456.95 L720.975 1452.69 L729.239 1451.02 L733.915 1451.02 L733.915 1481.64 L741.554 1481.64 L741.554 1485.58 L721.647 1485.58 L721.647 1481.64 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1245.63 1481.64 L1261.95 1481.64 L1261.95 1485.58 L1240 1485.58 L1240 1481.64 Q1242.66 1478.89 1247.25 1474.26 Q1251.85 1469.61 1253.03 1468.27 Q1255.28 1465.74 1256.16 1464.01 Q1257.06 1462.25 1257.06 1460.56 Q1257.06 1457.8 1255.12 1456.07 Q1253.2 1454.33 1250.09 1454.33 Q1247.89 1454.33 1245.44 1455.09 Q1243.01 1455.86 1240.23 1457.41 L1240.23 1452.69 Q1243.06 1451.55 1245.51 1450.97 Q1247.96 1450.39 1250 1450.39 Q1255.37 1450.39 1258.57 1453.08 Q1261.76 1455.77 1261.76 1460.26 Q1261.76 1462.39 1260.95 1464.31 Q1260.16 1466.2 1258.06 1468.8 Q1257.48 1469.47 1254.38 1472.69 Q1251.27 1475.88 1245.63 1481.64 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1774.93 1466.95 Q1778.29 1467.66 1780.16 1469.93 Q1782.06 1472.2 1782.06 1475.53 Q1782.06 1480.65 1778.54 1483.45 Q1775.02 1486.25 1768.54 1486.25 Q1766.37 1486.25 1764.05 1485.81 Q1761.76 1485.39 1759.3 1484.54 L1759.3 1480.02 Q1761.25 1481.16 1763.56 1481.74 Q1765.88 1482.32 1768.4 1482.32 Q1772.8 1482.32 1775.09 1480.58 Q1777.41 1478.84 1777.41 1475.53 Q1777.41 1472.48 1775.25 1470.77 Q1773.12 1469.03 1769.3 1469.03 L1765.28 1469.03 L1765.28 1465.19 L1769.49 1465.19 Q1772.94 1465.19 1774.77 1463.82 Q1776.6 1462.43 1776.6 1459.84 Q1776.6 1457.18 1774.7 1455.77 Q1772.82 1454.33 1769.3 1454.33 Q1767.38 1454.33 1765.18 1454.75 Q1762.99 1455.16 1760.35 1456.04 L1760.35 1451.88 Q1763.01 1451.14 1765.32 1450.77 Q1767.66 1450.39 1769.72 1450.39 Q1775.05 1450.39 1778.15 1452.83 Q1781.25 1455.23 1781.25 1459.35 Q1781.25 1462.22 1779.61 1464.21 Q1777.96 1466.18 1774.93 1466.95 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M2293.4 1455.09 L2281.59 1473.54 L2293.4 1473.54 L2293.4 1455.09 M2292.17 1451.02 L2298.05 1451.02 L2298.05 1473.54 L2302.98 1473.54 L2302.98 1477.43 L2298.05 1477.43 L2298.05 1485.58 L2293.4 1485.58 L2293.4 1477.43 L2277.8 1477.43 L2277.8 1472.92 L2292.17 1451.02 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1199.28 1522.27 L1199.28 1532.4 L1211.35 1532.4 L1211.35 1536.95 L1199.28 1536.95 L1199.28 1556.3 Q1199.28 1560.66 1200.46 1561.9 Q1201.67 1563.14 1205.33 1563.14 L1211.35 1563.14 L1211.35 1568.04 L1205.33 1568.04 Q1198.55 1568.04 1195.97 1565.53 Q1193.4 1562.98 1193.4 1556.3 L1193.4 1536.95 L1189.1 1536.95 L1189.1 1532.4 L1193.4 1532.4 L1193.4 1522.27 L1199.28 1522.27 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1239.23 1518.52 L1252.72 1518.52 L1252.72 1523.07 L1245.09 1523.07 L1245.09 1572.09 L1252.72 1572.09 L1252.72 1576.64 L1239.23 1576.64 L1239.23 1518.52 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1287.93 1533.45 L1287.93 1538.98 Q1285.44 1537.71 1282.77 1537.07 Q1280.1 1536.44 1277.23 1536.44 Q1272.87 1536.44 1270.68 1537.77 Q1268.51 1539.11 1268.51 1541.79 Q1268.51 1543.82 1270.07 1545 Q1271.63 1546.15 1276.34 1547.2 L1278.35 1547.64 Q1284.58 1548.98 1287.19 1551.43 Q1289.84 1553.85 1289.84 1558.21 Q1289.84 1563.17 1285.89 1566.07 Q1281.97 1568.97 1275.1 1568.97 Q1272.23 1568.97 1269.12 1568.39 Q1266.03 1567.85 1262.59 1566.74 L1262.59 1560.69 Q1265.84 1562.38 1268.99 1563.24 Q1272.14 1564.07 1275.23 1564.07 Q1279.36 1564.07 1281.59 1562.66 Q1283.82 1561.23 1283.82 1558.65 Q1283.82 1556.27 1282.2 1554.99 Q1280.61 1553.72 1275.16 1552.54 L1273.13 1552.07 Q1267.68 1550.92 1265.26 1548.56 Q1262.85 1546.18 1262.85 1542.04 Q1262.85 1537.01 1266.41 1534.27 Q1269.98 1531.54 1276.53 1531.54 Q1279.78 1531.54 1282.64 1532.01 Q1285.51 1532.49 1287.93 1533.45 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1312.85 1518.52 L1312.85 1576.64 L1299.35 1576.64 L1299.35 1572.09 L1306.96 1572.09 L1306.96 1523.07 L1299.35 1523.07 L1299.35 1518.52 L1312.85 1518.52 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip032)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  149.191,1334.39 2352.76,1334.39 
  "/>
<polyline clip-path="url(#clip032)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  149.191,1060.05 2352.76,1060.05 
  "/>
<polyline clip-path="url(#clip032)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  149.191,785.706 2352.76,785.706 
  "/>
<polyline clip-path="url(#clip032)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  149.191,511.362 2352.76,511.362 
  "/>
<polyline clip-path="url(#clip032)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  149.191,237.019 2352.76,237.019 
  "/>
<polyline clip-path="url(#clip030)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  149.191,1423.18 149.191,47.2441 
  "/>
<polyline clip-path="url(#clip030)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  149.191,1334.39 168.089,1334.39 
  "/>
<polyline clip-path="url(#clip030)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  149.191,1060.05 168.089,1060.05 
  "/>
<polyline clip-path="url(#clip030)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  149.191,785.706 168.089,785.706 
  "/>
<polyline clip-path="url(#clip030)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  149.191,511.362 168.089,511.362 
  "/>
<polyline clip-path="url(#clip030)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  149.191,237.019 168.089,237.019 
  "/>
<path clip-path="url(#clip030)" d="M50.9921 1334.85 L80.6679 1334.85 L80.6679 1338.78 L50.9921 1338.78 L50.9921 1334.85 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M103.608 1321.19 L91.8021 1339.64 L103.608 1339.64 L103.608 1321.19 M102.381 1317.11 L108.26 1317.11 L108.26 1339.64 L113.191 1339.64 L113.191 1343.53 L108.26 1343.53 L108.26 1351.67 L103.608 1351.67 L103.608 1343.53 L88.0058 1343.53 L88.0058 1339.01 L102.381 1317.11 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M53.0754 1060.5 L82.7512 1060.5 L82.7512 1064.44 L53.0754 1064.44 L53.0754 1060.5 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M96.8715 1073.39 L113.191 1073.39 L113.191 1077.33 L91.2465 1077.33 L91.2465 1073.39 Q93.9086 1070.64 98.4919 1066.01 Q103.098 1061.36 104.279 1060.02 Q106.524 1057.49 107.404 1055.76 Q108.307 1054 108.307 1052.31 Q108.307 1049.55 106.362 1047.82 Q104.441 1046.08 101.339 1046.08 Q99.14 1046.08 96.6863 1046.84 Q94.2558 1047.61 91.478 1049.16 L91.478 1044.44 Q94.3021 1043.3 96.7558 1042.72 Q99.2095 1042.15 101.246 1042.15 Q106.617 1042.15 109.811 1044.83 Q113.006 1047.52 113.006 1052.01 Q113.006 1054.14 112.196 1056.06 Q111.408 1057.96 109.302 1060.55 Q108.723 1061.22 105.621 1064.44 Q102.52 1067.63 96.8715 1073.39 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M101.246 771.505 Q97.6354 771.505 95.8067 775.07 Q94.0012 778.611 94.0012 785.741 Q94.0012 792.847 95.8067 796.412 Q97.6354 799.954 101.246 799.954 Q104.881 799.954 106.686 796.412 Q108.515 792.847 108.515 785.741 Q108.515 778.611 106.686 775.07 Q104.881 771.505 101.246 771.505 M101.246 767.801 Q107.057 767.801 110.112 772.408 Q113.191 776.991 113.191 785.741 Q113.191 794.468 110.112 799.074 Q107.057 803.657 101.246 803.657 Q95.4363 803.657 92.3576 799.074 Q89.3021 794.468 89.3021 785.741 Q89.3021 776.991 92.3576 772.408 Q95.4363 767.801 101.246 767.801 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M96.8715 524.707 L113.191 524.707 L113.191 528.642 L91.2465 528.642 L91.2465 524.707 Q93.9086 521.953 98.4919 517.323 Q103.098 512.67 104.279 511.328 Q106.524 508.805 107.404 507.068 Q108.307 505.309 108.307 503.619 Q108.307 500.865 106.362 499.129 Q104.441 497.393 101.339 497.393 Q99.14 497.393 96.6863 498.156 Q94.2558 498.92 91.478 500.471 L91.478 495.749 Q94.3021 494.615 96.7558 494.036 Q99.2095 493.457 101.246 493.457 Q106.617 493.457 109.811 496.143 Q113.006 498.828 113.006 503.318 Q113.006 505.448 112.196 507.369 Q111.408 509.267 109.302 511.86 Q108.723 512.531 105.621 515.749 Q102.52 518.943 96.8715 524.707 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M103.608 223.813 L91.8021 242.262 L103.608 242.262 L103.608 223.813 M102.381 219.739 L108.26 219.739 L108.26 242.262 L113.191 242.262 L113.191 246.15 L108.26 246.15 L108.26 254.299 L103.608 254.299 L103.608 246.15 L88.0058 246.15 L88.0058 241.637 L102.381 219.739 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip032)" style="stroke:#009af9; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  211.556,648.534 216.753,650.589 221.95,652.637 227.147,654.675 232.344,656.702 237.541,658.717 242.738,660.716 247.936,662.698 253.133,664.662 258.33,666.604 
  263.527,668.524 268.724,670.42 273.921,672.289 279.118,674.129 284.315,675.939 289.512,677.717 294.709,679.461 299.906,681.169 305.103,682.839 310.301,684.471 
  315.498,686.062 320.695,687.61 325.892,689.114 331.089,690.573 336.286,691.985 341.483,693.348 346.68,694.661 351.877,695.923 357.074,697.133 362.271,698.288 
  367.469,699.389 372.666,700.433 377.863,701.42 383.06,702.348 388.257,703.217 393.454,704.025 398.651,704.772 403.848,705.456 409.045,706.077 414.242,706.633 
  419.439,707.125 424.636,707.551 429.834,707.911 435.031,708.203 440.228,708.427 445.425,708.583 450.622,708.669 455.819,708.686 461.016,708.647 466.213,708.559 
  471.41,708.419 476.607,708.229 481.804,707.987 487.002,707.694 492.199,707.351 497.396,706.955 502.593,706.509 507.79,706.012 512.987,705.465 518.184,704.867 
  523.381,704.219 528.578,703.523 533.775,702.777 538.972,701.983 544.17,701.142 549.367,700.253 554.564,699.319 559.761,698.34 564.958,697.316 570.155,696.249 
  575.352,695.14 580.549,693.99 585.746,692.799 590.943,691.569 596.14,690.301 601.337,688.996 606.535,687.656 611.732,686.282 616.929,684.875 622.126,683.437 
  627.323,681.969 632.52,680.472 637.717,678.948 642.914,677.399 648.111,675.826 653.308,674.231 658.505,672.614 663.703,670.979 668.9,669.326 674.097,667.657 
  679.294,665.974 684.491,664.278 689.688,662.571 694.885,660.855 700.082,659.132 705.279,657.403 710.476,655.67 715.673,653.934 720.87,652.198 726.068,650.464 
  731.265,648.732 736.462,647.004 741.659,645.283 746.856,643.57 752.053,641.867 757.25,640.176 762.447,638.497 767.644,636.834 772.841,635.186 778.038,633.557 
  783.236,631.948 788.433,630.36 793.63,628.796 798.827,627.256 804.024,625.743 809.221,624.257 814.418,622.8 819.615,621.374 824.812,619.98 830.009,618.62 
  835.206,617.296 840.403,616.007 845.601,614.756 850.798,613.544 855.995,612.373 861.192,611.243 866.389,610.156 871.586,609.112 876.783,608.114 881.98,607.162 
  887.177,606.256 892.374,605.399 897.571,604.591 902.769,603.832 907.966,603.125 913.163,602.469 918.36,601.866 923.557,601.316 928.754,600.82 933.951,600.379 
  939.148,599.994 944.345,599.665 949.542,599.392 954.739,599.177 959.936,599.021 965.134,598.922 970.331,598.884 975.528,598.898 980.725,598.951 985.922,599.044 
  991.119,599.178 996.316,599.352 1001.51,599.566 1006.71,599.821 1011.91,600.117 1017.1,600.452 1022.3,600.829 1027.5,601.245 1032.7,601.702 1037.89,602.197 
  1043.09,602.733 1048.29,603.307 1053.48,603.92 1058.68,604.571 1063.88,605.26 1069.08,605.986 1074.27,606.748 1079.47,607.547 1084.67,608.38 1089.86,609.248 
  1095.06,610.15 1100.26,611.084 1105.45,612.05 1110.65,613.046 1115.85,614.073 1121.05,615.129 1126.24,616.213 1131.44,617.323 1136.64,618.46 1141.83,619.621 
  1147.03,620.807 1152.23,622.014 1157.43,623.243 1162.62,624.492 1167.82,625.76 1173.02,627.046 1178.21,628.348 1183.41,629.665 1188.61,630.996 1193.81,632.339 
  1199,633.694 1204.2,635.058 1209.4,636.431 1214.59,637.81 1219.79,639.196 1224.99,640.585 1230.19,641.978 1235.38,643.372 1240.58,644.766 1245.78,646.159 
  1250.97,647.549 1256.17,648.935 1261.37,650.316 1266.56,651.69 1271.76,653.056 1276.96,654.412 1282.16,655.757 1287.35,657.09 1292.55,658.409 1297.75,659.714 
  1302.94,661.002 1308.14,662.272 1313.34,663.524 1318.54,664.755 1323.73,665.965 1328.93,667.153 1334.13,668.316 1339.32,669.454 1344.52,670.567 1349.72,671.651 
  1354.92,672.708 1360.11,673.734 1365.31,674.73 1370.51,675.694 1375.7,676.626 1380.9,677.523 1386.1,678.386 1391.29,679.213 1396.49,680.004 1401.69,680.757 
  1406.89,681.471 1412.08,682.147 1417.28,682.782 1422.48,683.377 1427.67,683.93 1432.87,684.441 1438.07,684.908 1443.27,685.333 1448.46,685.713 1453.66,686.048 
  1458.86,686.338 1464.05,686.581 1469.25,686.778 1474.45,686.929 1479.65,687.031 1484.84,687.085 1490.04,687.092 1495.24,687.067 1500.43,687.014 1505.63,686.931 
  1510.83,686.82 1516.02,686.68 1521.22,686.51 1526.42,686.312 1531.62,686.084 1536.81,685.827 1542.01,685.54 1547.21,685.225 1552.4,684.881 1557.6,684.509 
  1562.8,684.108 1568,683.679 1573.19,683.222 1578.39,682.737 1583.59,682.225 1588.78,681.687 1593.98,681.122 1599.18,680.531 1604.38,679.914 1609.57,679.273 
  1614.77,678.607 1619.97,677.918 1625.16,677.205 1630.36,676.47 1635.56,675.713 1640.75,674.935 1645.95,674.137 1651.15,673.318 1656.35,672.481 1661.54,671.626 
  1666.74,670.753 1671.94,669.864 1677.13,668.96 1682.33,668.04 1687.53,667.107 1692.73,666.161 1697.92,665.203 1703.12,664.234 1708.32,663.255 1713.51,662.266 
  1718.71,661.27 1723.91,660.266 1729.11,659.256 1734.3,658.24 1739.5,657.221 1744.7,656.198 1749.89,655.173 1755.09,654.147 1760.29,653.121 1765.49,652.095 
  1770.68,651.072 1775.88,650.051 1781.08,649.034 1786.27,648.022 1791.47,647.017 1796.67,646.018 1801.86,645.028 1807.06,644.046 1812.26,643.075 1817.46,642.114 
  1822.65,641.166 1827.85,640.23 1833.05,639.309 1838.24,638.403 1843.44,637.512 1848.64,636.639 1853.84,635.783 1859.03,634.945 1864.23,634.128 1869.43,633.33 
  1874.62,632.554 1879.82,631.8 1885.02,631.069 1890.22,630.361 1895.41,629.678 1900.61,629.02 1905.81,628.388 1911,627.783 1916.2,627.205 1921.4,626.656 
  1926.59,626.134 1931.79,625.643 1936.99,625.181 1942.19,624.75 1947.38,624.35 1952.58,623.981 1957.78,623.645 1962.97,623.342 1968.17,623.072 1973.37,622.836 
  1978.57,622.634 1983.76,622.467 1988.96,622.334 1994.16,622.237 1999.35,622.176 2004.55,622.152 2009.75,622.157 2014.95,622.179 2020.14,622.217 2025.34,622.271 
  2030.54,622.342 2035.73,622.429 2040.93,622.533 2046.13,622.654 2051.32,622.791 2056.52,622.945 2061.72,623.116 2066.92,623.303 2072.11,623.507 2077.31,623.727 
  2082.51,623.964 2087.7,624.217 2092.9,624.485 2098.1,624.77 2103.3,625.07 2108.49,625.386 2113.69,625.717 2118.89,626.064 2124.08,626.425 2129.28,626.8 
  2134.48,627.189 2139.68,627.593 2144.87,628.01 2150.07,628.44 2155.27,628.883 2160.46,629.339 2165.66,629.806 2170.86,630.285 2176.05,630.776 2181.25,631.277 
  2186.45,631.789 2191.65,632.31 2196.84,632.841 2202.04,633.38 2207.24,633.928 2212.43,634.484 2217.63,635.047 2222.83,635.617 2228.03,636.193 2233.22,636.774 
  2238.42,637.361 2243.62,637.953 2248.81,638.548 2254.01,639.147 2259.21,639.749 2264.41,640.353 2269.6,640.958 2274.8,641.565 2280,642.172 2285.19,642.779 
  2290.39,643.385 
  "/>
<polyline clip-path="url(#clip032)" style="stroke:#e26f46; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  211.556,991.464 216.753,990.877 221.95,990.089 227.147,989.097 232.344,987.901 237.541,986.504 242.738,984.906 247.936,983.11 253.133,981.115 258.33,978.926 
  263.527,976.542 268.724,973.968 273.921,971.205 279.118,968.256 284.315,965.123 289.512,961.811 294.709,958.321 299.906,954.658 305.103,950.825 310.301,946.825 
  315.498,942.662 320.695,938.34 325.892,933.862 331.089,929.234 336.286,924.459 341.483,919.54 346.68,914.483 351.877,909.292 357.074,903.971 362.271,898.525 
  367.469,892.958 372.666,887.275 377.863,881.48 383.06,875.577 388.257,869.572 393.454,863.468 398.651,857.269 403.848,850.979 409.045,844.604 414.242,838.145 
  419.439,831.606 424.636,824.992 429.834,818.304 435.031,811.544 440.228,804.715 445.425,797.819 450.622,790.857 455.819,784.38 461.016,779.36 466.213,774.304 
  471.41,769.22 476.607,764.114 481.804,758.995 487.002,753.87 492.199,748.747 497.396,743.633 502.593,738.536 507.79,733.463 512.987,728.42 518.184,723.416 
  523.381,718.456 528.578,713.547 533.775,708.696 538.972,703.909 544.17,699.191 549.367,694.549 554.564,689.988 559.761,685.513 564.958,681.13 570.155,676.846 
  575.352,672.663 580.549,668.587 585.746,664.623 590.943,660.776 596.14,657.049 601.337,653.447 606.535,649.974 611.732,646.634 616.929,643.431 622.126,640.367 
  627.323,637.447 632.52,634.673 637.717,632.049 642.914,629.577 648.111,627.26 653.308,625.1 658.505,623.1 663.703,621.261 668.9,619.586 674.097,618.077 
  679.294,616.735 684.491,615.561 689.688,614.557 694.885,613.724 700.082,613.063 705.279,612.574 710.476,612.258 715.673,612.116 720.87,612.147 726.068,612.351 
  731.265,612.73 736.462,613.281 741.659,614.005 746.856,614.901 752.053,615.968 757.25,617.205 762.447,618.611 767.644,620.185 772.841,621.926 778.038,623.831 
  783.236,625.899 788.433,628.127 793.63,630.515 798.827,633.058 804.024,635.757 809.221,638.607 814.418,641.607 819.615,644.753 824.812,648.042 830.009,651.472 
  835.206,655.039 840.403,658.741 845.601,662.574 850.798,666.535 855.995,670.62 861.192,674.826 866.389,679.148 871.586,683.585 876.783,688.132 881.98,692.785 
  887.177,697.541 892.374,702.396 897.571,707.346 902.769,712.388 907.966,717.517 913.163,722.732 918.36,728.028 923.557,733.401 928.754,738.849 933.951,744.369 
  939.148,749.958 944.345,755.613 949.542,761.332 954.739,767.113 959.936,772.954 965.134,778.853 970.331,784.802 975.528,789.059 980.725,793.039 985.922,797.046 
  991.119,801.073 996.316,805.116 1001.51,809.167 1006.71,813.223 1011.91,817.276 1017.1,821.321 1022.3,825.353 1027.5,829.366 1032.7,833.354 1037.89,837.313 
  1043.09,841.238 1048.29,845.123 1053.48,848.962 1058.68,852.752 1063.88,856.486 1069.08,860.161 1074.27,863.771 1079.47,867.313 1084.67,870.782 1089.86,874.175 
  1095.06,877.486 1100.26,880.713 1105.45,883.851 1110.65,886.896 1115.85,889.846 1121.05,892.697 1126.24,895.446 1131.44,898.088 1136.64,900.622 1141.83,903.045 
  1147.03,905.353 1152.23,907.545 1157.43,909.617 1162.62,911.567 1167.82,913.394 1173.02,915.095 1178.21,916.668 1183.41,918.112 1188.61,919.424 1193.81,920.604 
  1199,921.65 1204.2,922.561 1209.4,923.336 1214.59,923.974 1219.79,924.474 1224.99,924.835 1230.19,925.058 1235.38,925.141 1240.58,925.085 1245.78,924.889 
  1250.97,924.555 1256.17,924.081 1261.37,923.468 1266.56,922.717 1271.76,921.828 1276.96,920.802 1282.16,919.64 1287.35,918.343 1292.55,916.912 1297.75,915.349 
  1302.94,913.654 1308.14,911.83 1313.34,909.877 1318.54,907.798 1323.73,905.594 1328.93,903.267 1334.13,900.82 1339.32,898.254 1344.52,895.572 1349.72,892.776 
  1354.92,889.869 1360.11,886.852 1365.31,883.728 1370.51,880.501 1375.7,877.172 1380.9,873.744 1386.1,870.221 1391.29,866.604 1396.49,862.897 1401.69,859.103 
  1406.89,855.224 1412.08,851.263 1417.28,847.223 1422.48,843.107 1427.67,838.918 1432.87,834.658 1438.07,830.33 1443.27,825.937 1448.46,821.481 1453.66,816.966 
  1458.86,812.392 1464.05,807.763 1469.25,803.08 1474.45,798.346 1479.65,793.562 1484.84,788.73 1490.04,784.636 1495.24,781.787 1500.43,778.917 1505.63,776.029 
  1510.83,773.127 1516.02,770.215 1521.22,767.297 1526.42,764.376 1531.62,761.456 1536.81,758.541 1542.01,755.635 1547.21,752.742 1552.4,749.865 1557.6,747.008 
  1562.8,744.175 1568,741.37 1573.19,738.596 1578.39,735.856 1583.59,733.154 1588.78,730.494 1593.98,727.879 1599.18,725.311 1604.38,722.794 1609.57,720.332 
  1614.77,717.927 1619.97,715.582 1625.16,713.299 1630.36,711.083 1635.56,708.935 1640.75,706.858 1645.95,704.854 1651.15,702.927 1656.35,701.077 1661.54,699.308 
  1666.74,697.621 1671.94,696.018 1677.13,694.502 1682.33,693.073 1687.53,691.735 1692.73,690.487 1697.92,689.333 1703.12,688.273 1708.32,687.308 1713.51,686.44 
  1718.71,685.67 1723.91,684.998 1729.11,684.426 1734.3,683.954 1739.5,683.584 1744.7,683.315 1749.89,683.148 1755.09,683.083 1760.29,683.121 1765.49,683.261 
  1770.68,683.505 1775.88,683.851 1781.08,684.3 1786.27,684.851 1791.47,685.504 1796.67,686.259 1801.86,687.114 1807.06,688.07 1812.26,689.126 1817.46,690.28 
  1822.65,691.532 1827.85,692.881 1833.05,694.326 1838.24,695.865 1843.44,697.498 1848.64,699.223 1853.84,701.039 1859.03,702.944 1864.23,704.936 1869.43,707.014 
  1874.62,709.177 1879.82,711.422 1885.02,713.749 1890.22,716.154 1895.41,718.636 1900.61,721.194 1905.81,723.825 1911,726.527 1916.2,729.299 1921.4,732.138 
  1926.59,735.043 1931.79,738.011 1936.99,741.039 1942.19,744.127 1947.38,747.272 1952.58,750.472 1957.78,753.724 1962.97,757.028 1968.17,760.381 1973.37,763.78 
  1978.57,767.224 1983.76,770.712 1988.96,774.241 1994.16,777.809 1999.35,781.416 2004.55,785.056 2009.75,787.035 2014.95,788.663 2020.14,790.303 2025.34,791.952 
  2030.54,793.609 2035.73,795.272 2040.93,796.939 2046.13,798.609 2051.32,800.278 2056.52,801.946 2061.72,803.61 2066.92,805.269 2072.11,806.919 2077.31,808.56 
  2082.51,810.188 2087.7,811.803 2092.9,813.402 2098.1,814.983 2103.3,816.544 2108.49,818.084 2113.69,819.6 2118.89,821.091 2124.08,822.554 2129.28,823.988 
  2134.48,825.392 2139.68,826.763 2144.87,828.1 2150.07,829.401 2155.27,830.664 2160.46,831.888 2165.66,833.072 2170.86,834.214 2176.05,835.312 2181.25,836.365 
  2186.45,837.372 2191.65,838.332 2196.84,839.242 2202.04,840.103 2207.24,840.914 2212.43,841.671 2217.63,842.376 2222.83,843.027 2228.03,843.623 2233.22,844.163 
  2238.42,844.647 2243.62,845.074 2248.81,845.442 2254.01,845.752 2259.21,846.004 2264.41,846.196 2269.6,846.327 2274.8,846.399 2280,846.41 2285.19,846.361 
  2290.39,846.25 
  "/>
<polyline clip-path="url(#clip032)" style="stroke:#3da44d; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  211.556,737.711 216.753,717.164 221.95,696.687 227.147,676.301 232.344,656.025 237.541,635.876 242.738,615.875 247.936,596.046 253.133,576.397 258.33,556.953 
  263.527,537.726 268.724,518.738 273.921,500.006 279.118,481.548 284.315,463.38 289.512,445.519 294.709,427.981 299.906,410.78 305.103,393.93 310.301,377.444 
  315.498,361.336 320.695,345.618 325.892,330.299 331.089,315.391 336.286,300.902 341.483,286.842 346.68,273.216 351.877,260.03 357.074,247.29 362.271,234.995 
  367.469,223.149 372.666,211.751 377.863,200.797 383.06,190.284 388.257,180.203 393.454,170.547 398.651,161.302 403.848,152.454 409.045,143.984 414.242,135.871 
  419.439,128.087 424.636,120.602 429.834,113.38 435.031,106.38 440.228,99.5524 445.425,92.8426 450.622,86.1857 455.819,285.821 461.016,281.709 466.213,278.407 
  471.41,275.906 476.607,274.197 481.804,273.268 487.002,273.106 492.199,273.698 497.396,275.031 502.593,277.087 507.79,279.852 512.987,283.31 518.184,287.444 
  523.381,292.237 528.578,297.675 533.775,303.738 538.972,310.412 544.17,317.678 549.367,325.521 554.564,333.924 559.761,342.867 564.958,352.336 570.155,362.317 
  575.352,372.789 580.549,383.737 585.746,395.145 590.943,406.998 596.14,419.276 601.337,431.964 606.535,445.046 611.732,458.504 616.929,472.321 622.126,486.48 
  627.323,500.964 632.52,515.756 637.717,530.84 642.914,546.197 648.111,561.81 653.308,577.662 658.505,593.735 663.703,610.012 668.9,626.476 674.097,643.108 
  679.294,659.892 684.491,676.81 689.688,693.843 694.885,710.975 700.082,728.188 705.279,745.464 710.476,762.785 715.673,780.135 720.87,797.496 726.068,814.85 
  731.265,832.18 736.462,849.471 741.659,866.704 746.856,883.863 752.053,900.931 757.25,917.89 762.447,934.724 767.644,951.422 772.841,967.965 778.038,984.338 
  783.236,1000.53 788.433,1016.51 793.63,1032.28 798.827,1047.82 804.024,1063.12 809.221,1078.16 814.418,1092.94 819.615,1107.44 824.812,1121.64 830.009,1135.54 
  835.206,1149.13 840.403,1162.39 845.601,1175.33 850.798,1187.92 855.995,1200.18 861.192,1212.07 866.389,1223.62 871.586,1234.8 876.783,1245.62 881.98,1256.07 
  887.177,1266.15 892.374,1275.87 897.571,1285.23 902.769,1294.23 907.966,1302.88 913.163,1311.18 918.36,1319.15 923.557,1326.79 928.754,1334.12 933.951,1341.16 
  939.148,1347.92 944.345,1354.43 949.542,1360.71 954.739,1366.8 959.936,1372.72 965.134,1378.52 970.331,1384.24 975.528,1182.3 980.725,1185.25 985.922,1187.61 
  991.119,1189.39 996.316,1190.59 1001.51,1191.2 1006.71,1191.25 1011.91,1190.73 1017.1,1189.65 1022.3,1188.01 1027.5,1185.83 1032.7,1183.12 1037.89,1179.88 
  1043.09,1176.13 1048.29,1171.88 1053.48,1167.13 1058.68,1161.89 1063.88,1156.18 1069.08,1150.01 1074.27,1143.4 1079.47,1136.34 1084.67,1128.86 1089.86,1120.97 
  1095.06,1112.68 1100.26,1104 1105.45,1094.96 1110.65,1085.55 1115.85,1075.79 1121.05,1065.71 1126.24,1055.3 1131.44,1044.58 1136.64,1033.57 1141.83,1022.28 
  1147.03,1010.72 1152.23,998.915 1157.43,986.867 1162.62,974.595 1167.82,962.111 1173.02,949.43 1178.21,936.566 1183.41,923.535 1188.61,910.35 1193.81,897.024 
  1199,883.572 1204.2,870.009 1209.4,856.349 1214.59,842.606 1219.79,828.794 1224.99,814.927 1230.19,801.02 1235.38,787.087 1240.58,773.14 1245.78,759.196 
  1250.97,745.266 1256.17,731.365 1261.37,717.506 1266.56,703.703 1271.76,689.968 1276.96,676.316 1282.16,662.758 1287.35,649.307 1292.55,635.977 1297.75,622.778 
  1302.94,609.723 1308.14,596.823 1313.34,584.09 1318.54,571.534 1323.73,559.167 1328.93,546.997 1334.13,535.036 1339.32,523.293 1344.52,511.775 1349.72,500.492 
  1354.92,489.452 1360.11,478.661 1365.31,468.127 1370.51,457.855 1375.7,447.852 1380.9,438.121 1386.1,428.668 1391.29,419.494 1396.49,410.604 1401.69,401.998 
  1406.89,393.677 1412.08,385.641 1417.28,377.889 1422.48,370.419 1427.67,363.227 1432.87,356.309 1438.07,349.658 1443.27,343.268 1448.46,337.13 1453.66,331.233 
  1458.86,325.566 1464.05,320.116 1469.25,314.865 1474.45,309.796 1479.65,304.89 1484.84,300.123 1490.04,501.995 1495.24,499.641 1500.43,497.666 1505.63,496.072 
  1510.83,494.863 1516.02,494.039 1521.22,493.601 1526.42,493.547 1531.62,493.877 1536.81,494.59 1542.01,495.682 1547.21,497.151 1552.4,498.991 1557.6,501.2 
  1562.8,503.774 1568,506.705 1573.19,509.989 1578.39,513.622 1583.59,517.594 1588.78,521.901 1593.98,526.535 1599.18,531.487 1604.38,536.752 1609.57,542.322 
  1614.77,548.188 1619.97,554.341 1625.16,560.773 1630.36,567.477 1635.56,574.441 1640.75,581.659 1645.95,589.119 1651.15,596.813 1656.35,604.731 1661.54,612.864 
  1666.74,621.201 1671.94,629.732 1677.13,638.448 1682.33,647.339 1687.53,656.394 1692.73,665.603 1697.92,674.956 1703.12,684.443 1708.32,694.052 1713.51,703.774 
  1718.71,713.599 1723.91,723.516 1729.11,733.514 1734.3,743.583 1739.5,753.713 1744.7,763.894 1749.89,774.115 1755.09,784.366 1760.29,794.635 1765.49,804.913 
  1770.68,815.191 1775.88,825.459 1781.08,835.706 1786.27,845.922 1791.47,856.097 1796.67,866.222 1801.86,876.285 1807.06,886.281 1812.26,896.2 1817.46,906.032 
  1822.65,915.768 1827.85,925.399 1833.05,934.917 1838.24,944.313 1843.44,953.583 1848.64,962.716 1853.84,971.706 1859.03,980.545 1864.23,989.228 1869.43,997.748 
  1874.62,1006.1 1879.82,1014.28 1885.02,1022.27 1890.22,1030.08 1895.41,1037.71 1900.61,1045.14 1905.81,1052.37 1911,1059.41 1916.2,1066.24 1921.4,1072.87 
  1926.59,1079.3 1931.79,1085.51 1936.99,1091.53 1942.19,1097.33 1947.38,1102.93 1952.58,1108.32 1957.78,1113.51 1962.97,1118.5 1968.17,1123.3 1973.37,1127.9 
  1978.57,1132.31 1983.76,1136.54 1988.96,1140.6 1994.16,1144.49 1999.35,1148.22 2004.55,1151.8 2009.75,947.848 2014.95,949.129 2020.14,950.22 2025.34,951.119 
  2030.54,951.824 2035.73,952.331 2040.93,952.64 2046.13,952.75 2051.32,952.659 2056.52,952.368 2061.72,951.875 2066.92,951.181 2072.11,950.286 2077.31,949.191 
  2082.51,947.897 2087.7,946.406 2092.9,944.718 2098.1,942.835 2103.3,940.761 2108.49,938.497 2113.69,936.045 2118.89,933.408 2124.08,930.59 2129.28,927.594 
  2134.48,924.423 2139.68,921.079 2144.87,917.571 2150.07,913.899 2155.27,910.066 2160.46,906.079 2165.66,901.944 2170.86,897.663 2176.05,893.241 2181.25,888.683 
  2186.45,883.997 2191.65,879.184 2196.84,874.251 2202.04,869.205 2207.24,864.051 2212.43,858.792 2217.63,853.436 2222.83,847.989 2228.03,842.456 2233.22,836.842 
  2238.42,831.154 2243.62,825.398 2248.81,819.58 2254.01,813.705 2259.21,807.78 2264.41,801.812 2269.6,795.804 2274.8,789.764 2280,783.699 2285.19,777.613 
  2290.39,771.513 
  "/>
<polyline clip-path="url(#clip032)" style="stroke:#c271d2; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  211.556,833.702 216.753,833.702 221.95,833.702 227.147,833.703 232.344,833.706 237.541,833.71 242.738,833.716 247.936,833.726 253.133,833.74 258.33,833.76 
  263.527,833.786 268.724,833.819 273.921,833.862 279.118,833.916 284.315,833.984 289.512,834.066 294.709,834.166 299.906,834.287 305.103,834.431 310.301,834.602 
  315.498,834.803 320.695,835.04 325.892,835.315 331.089,835.636 336.286,836.007 341.483,836.436 346.68,836.929 351.877,837.494 357.074,838.141 362.271,838.879 
  367.469,839.72 372.666,840.676 377.863,841.762 383.06,842.992 388.257,844.385 393.454,845.959 398.651,847.737 403.848,849.743 409.045,852.004 414.242,854.55 
  419.439,857.417 424.636,860.642 429.834,864.27 435.031,868.348 440.228,872.933 445.425,878.087 450.622,883.88 455.819,684.075 461.016,688.571 466.213,692.762 
  471.41,696.658 476.607,700.271 481.804,703.616 487.002,706.705 492.199,709.552 497.396,712.171 502.593,714.576 507.79,716.781 512.987,718.798 518.184,720.642 
  523.381,722.323 528.578,723.855 533.775,725.248 538.972,726.513 544.17,727.66 549.367,728.7 554.564,729.64 559.761,730.489 564.958,731.255 570.155,731.945 
  575.352,732.565 580.549,733.123 585.746,733.623 590.943,734.071 596.14,734.471 601.337,734.829 606.535,735.148 611.732,735.431 616.929,735.683 622.126,735.906 
  627.323,736.104 632.52,736.278 637.717,736.431 642.914,736.566 648.111,736.684 653.308,736.787 658.505,736.877 663.703,736.954 668.9,737.021 674.097,737.078 
  679.294,737.126 684.491,737.167 689.688,737.2 694.885,737.227 700.082,737.247 705.279,737.262 710.476,737.271 715.673,737.276 720.87,737.275 726.068,737.269 
  731.265,737.257 736.462,737.24 741.659,737.218 746.856,737.189 752.053,737.153 757.25,737.11 762.447,737.058 767.644,736.998 772.841,736.927 778.038,736.845 
  783.236,736.75 788.433,736.641 793.63,736.516 798.827,736.374 804.024,736.211 809.221,736.027 814.418,735.818 819.615,735.581 824.812,735.314 830.009,735.013 
  835.206,734.674 840.403,734.293 845.601,733.866 850.798,733.387 855.995,732.85 861.192,732.251 866.389,731.581 871.586,730.834 876.783,730 881.98,729.072 
  887.177,728.04 892.374,726.891 897.571,725.615 902.769,724.198 907.966,722.626 913.163,720.882 918.36,718.949 923.557,716.807 928.754,714.434 933.951,711.807 
  939.148,708.9 944.345,705.682 949.542,702.122 954.739,698.184 959.936,693.827 965.134,689.006 970.331,683.681 975.528,885.481 980.725,881.997 985.922,878.7 
  991.119,875.586 996.316,872.651 1001.51,869.889 1006.71,867.295 1011.91,864.862 1017.1,862.585 1022.3,860.457 1027.5,858.471 1032.7,856.621 1037.89,854.898 
  1043.09,853.297 1048.29,851.81 1053.48,850.432 1058.68,849.156 1063.88,847.975 1069.08,846.884 1074.27,845.877 1079.47,844.948 1084.67,844.092 1089.86,843.304 
  1095.06,842.58 1100.26,841.914 1105.45,841.304 1110.65,840.744 1115.85,840.231 1121.05,839.762 1126.24,839.333 1131.44,838.942 1136.64,838.585 1141.83,838.261 
  1147.03,837.966 1152.23,837.699 1157.43,837.457 1162.62,837.239 1167.82,837.042 1173.02,836.866 1178.21,836.709 1183.41,836.57 1188.61,836.447 1193.81,836.339 
  1199,836.246 1204.2,836.167 1209.4,836.101 1214.59,836.048 1219.79,836.006 1224.99,835.977 1230.19,835.959 1235.38,835.952 1240.58,835.956 1245.78,835.972 
  1250.97,836 1256.17,836.039 1261.37,836.09 1266.56,836.154 1271.76,836.231 1276.96,836.321 1282.16,836.427 1287.35,836.548 1292.55,836.685 1297.75,836.84 
  1302.94,837.015 1308.14,837.21 1313.34,837.427 1318.54,837.669 1323.73,837.936 1328.93,838.232 1334.13,838.558 1339.32,838.918 1344.52,839.314 1349.72,839.749 
  1354.92,840.227 1360.11,840.752 1365.31,841.327 1370.51,841.957 1375.7,842.647 1380.9,843.402 1386.1,844.227 1391.29,845.129 1396.49,846.115 1401.69,847.191 
  1406.89,848.366 1412.08,849.648 1417.28,851.046 1422.48,852.57 1427.67,854.231 1432.87,856.041 1438.07,858.013 1443.27,860.16 1448.46,862.498 1453.66,865.043 
  1458.86,867.812 1464.05,870.826 1469.25,874.106 1474.45,877.674 1479.65,881.555 1484.84,885.778 1490.04,683.836 1495.24,686.441 1500.43,688.952 1505.63,691.369 
  1510.83,693.691 1516.02,695.919 1521.22,698.052 1526.42,700.093 1531.62,702.042 1536.81,703.9 1542.01,705.669 1547.21,707.352 1552.4,708.95 1557.6,710.467 
  1562.8,711.903 1568,713.263 1573.19,714.549 1578.39,715.763 1583.59,716.909 1588.78,717.989 1593.98,719.005 1599.18,719.962 1604.38,720.86 1609.57,721.704 
  1614.77,722.494 1619.97,723.235 1625.16,723.929 1630.36,724.577 1635.56,725.182 1640.75,725.746 1645.95,726.271 1651.15,726.76 1656.35,727.213 1661.54,727.633 
  1666.74,728.022 1671.94,728.38 1677.13,728.71 1682.33,729.012 1687.53,729.289 1692.73,729.54 1697.92,729.768 1703.12,729.973 1708.32,730.157 1713.51,730.319 
  1718.71,730.46 1723.91,730.582 1729.11,730.685 1734.3,730.768 1739.5,730.834 1744.7,730.881 1749.89,730.91 1755.09,730.921 1760.29,730.915 1765.49,730.89 
  1770.68,730.848 1775.88,730.787 1781.08,730.707 1786.27,730.609 1791.47,730.49 1796.67,730.352 1801.86,730.193 1807.06,730.012 1812.26,729.809 1817.46,729.582 
  1822.65,729.33 1827.85,729.052 1833.05,728.747 1838.24,728.414 1843.44,728.049 1848.64,727.653 1853.84,727.222 1859.03,726.755 1864.23,726.25 1869.43,725.704 
  1874.62,725.115 1879.82,724.479 1885.02,723.794 1890.22,723.057 1895.41,722.264 1900.61,721.412 1905.81,720.497 1911,719.514 1916.2,718.459 1921.4,717.327 
  1926.59,716.114 1931.79,714.813 1936.99,713.42 1942.19,711.928 1947.38,710.33 1952.58,708.619 1957.78,706.789 1962.97,704.831 1968.17,702.737 1973.37,700.498 
  1978.57,698.104 1983.76,695.546 1988.96,692.812 1994.16,689.892 1999.35,686.772 2004.55,683.442 2009.75,887.335 2014.95,885.839 2020.14,884.369 2025.34,882.928 
  2030.54,881.516 2035.73,880.135 2040.93,878.785 2046.13,877.469 2051.32,876.186 2056.52,874.937 2061.72,873.722 2066.92,872.543 2072.11,871.4 2077.31,870.292 
  2082.51,869.22 2087.7,868.184 2092.9,867.184 2098.1,866.219 2103.3,865.29 2108.49,864.396 2113.69,863.536 2118.89,862.711 2124.08,861.92 2129.28,861.162 
  2134.48,860.437 2139.68,859.745 2144.87,859.084 2150.07,858.455 2155.27,857.856 2160.46,857.287 2165.66,856.748 2170.86,856.237 2176.05,855.755 2181.25,855.301 
  2186.45,854.873 2191.65,854.473 2196.84,854.098 2202.04,853.749 2207.24,853.425 2212.43,853.126 2217.63,852.852 2222.83,852.601 2228.03,852.373 2233.22,852.169 
  2238.42,851.987 2243.62,851.829 2248.81,851.692 2254.01,851.578 2259.21,851.486 2264.41,851.417 2269.6,851.369 2274.8,851.343 2280,851.339 2285.19,851.357 
  2290.39,851.397 
  "/>
<path clip-path="url(#clip030)" d="
M1390.67 352.309 L2279.3 352.309 L2279.3 93.1086 L1390.67 93.1086  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<polyline clip-path="url(#clip030)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1390.67,352.309 2279.3,352.309 2279.3,93.1086 1390.67,93.1086 1390.67,352.309 
  "/>
<polyline clip-path="url(#clip030)" style="stroke:#009af9; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1415.15,144.949 1562.06,144.949 
  "/>
<path clip-path="url(#clip030)" d="M1606.89 141.28 Q1608.49 138.409 1610.71 137.044 Q1612.93 135.678 1615.94 135.678 Q1619.99 135.678 1622.19 138.525 Q1624.39 141.349 1624.39 146.581 L1624.39 162.229 L1620.11 162.229 L1620.11 146.719 Q1620.11 142.993 1618.79 141.187 Q1617.47 139.382 1614.76 139.382 Q1611.45 139.382 1609.53 141.581 Q1607.61 143.78 1607.61 147.576 L1607.61 162.229 L1603.33 162.229 L1603.33 146.719 Q1603.33 142.969 1602.01 141.187 Q1600.69 139.382 1597.93 139.382 Q1594.67 139.382 1592.75 141.604 Q1590.83 143.803 1590.83 147.576 L1590.83 162.229 L1586.54 162.229 L1586.54 136.303 L1590.83 136.303 L1590.83 140.331 Q1592.28 137.946 1594.32 136.812 Q1596.36 135.678 1599.16 135.678 Q1601.98 135.678 1603.95 137.113 Q1605.94 138.548 1606.89 141.28 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1644.67 149.196 Q1639.51 149.196 1637.51 150.377 Q1635.52 151.557 1635.52 154.405 Q1635.52 156.673 1637.01 158.016 Q1638.51 159.335 1641.08 159.335 Q1644.62 159.335 1646.75 156.835 Q1648.9 154.312 1648.9 150.145 L1648.9 149.196 L1644.67 149.196 M1653.16 147.437 L1653.16 162.229 L1648.9 162.229 L1648.9 158.293 Q1647.45 160.655 1645.27 161.789 Q1643.09 162.9 1639.95 162.9 Q1635.96 162.9 1633.6 160.678 Q1631.26 158.432 1631.26 154.682 Q1631.26 150.307 1634.18 148.085 Q1637.12 145.863 1642.93 145.863 L1648.9 145.863 L1648.9 145.446 Q1648.9 142.507 1646.96 140.909 Q1645.04 139.289 1641.54 139.289 Q1639.32 139.289 1637.21 139.821 Q1635.11 140.354 1633.16 141.419 L1633.16 137.483 Q1635.5 136.581 1637.7 136.141 Q1639.9 135.678 1641.98 135.678 Q1647.61 135.678 1650.39 138.594 Q1653.16 141.511 1653.16 147.437 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1678.46 137.067 L1678.46 141.094 Q1676.66 140.169 1674.71 139.706 Q1672.77 139.243 1670.69 139.243 Q1667.51 139.243 1665.92 140.215 Q1664.34 141.187 1664.34 143.131 Q1664.34 144.613 1665.48 145.469 Q1666.61 146.303 1670.04 147.067 L1671.5 147.391 Q1676.03 148.363 1677.93 150.145 Q1679.85 151.905 1679.85 155.076 Q1679.85 158.687 1676.98 160.793 Q1674.13 162.9 1669.13 162.9 Q1667.05 162.9 1664.78 162.483 Q1662.54 162.09 1660.04 161.28 L1660.04 156.881 Q1662.4 158.108 1664.69 158.733 Q1666.98 159.335 1669.23 159.335 Q1672.24 159.335 1673.86 158.317 Q1675.48 157.275 1675.48 155.4 Q1675.48 153.664 1674.3 152.738 Q1673.14 151.812 1669.18 150.956 L1667.7 150.608 Q1663.74 149.775 1661.98 148.062 Q1660.22 146.326 1660.22 143.317 Q1660.22 139.659 1662.82 137.669 Q1665.41 135.678 1670.18 135.678 Q1672.54 135.678 1674.62 136.025 Q1676.7 136.372 1678.46 137.067 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1703.16 137.067 L1703.16 141.094 Q1701.36 140.169 1699.41 139.706 Q1697.47 139.243 1695.38 139.243 Q1692.21 139.243 1690.62 140.215 Q1689.04 141.187 1689.04 143.131 Q1689.04 144.613 1690.18 145.469 Q1691.31 146.303 1694.74 147.067 L1696.19 147.391 Q1700.73 148.363 1702.63 150.145 Q1704.55 151.905 1704.55 155.076 Q1704.55 158.687 1701.68 160.793 Q1698.83 162.9 1693.83 162.9 Q1691.75 162.9 1689.48 162.483 Q1687.24 162.09 1684.74 161.28 L1684.74 156.881 Q1687.1 158.108 1689.39 158.733 Q1691.68 159.335 1693.93 159.335 Q1696.94 159.335 1698.56 158.317 Q1700.18 157.275 1700.18 155.4 Q1700.18 153.664 1699 152.738 Q1697.84 151.812 1693.88 150.956 L1692.4 150.608 Q1688.44 149.775 1686.68 148.062 Q1684.92 146.326 1684.92 143.317 Q1684.92 139.659 1687.51 137.669 Q1690.11 135.678 1694.88 135.678 Q1697.24 135.678 1699.32 136.025 Q1701.4 136.372 1703.16 137.067 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1711.94 156.349 L1716.82 156.349 L1716.82 162.229 L1711.94 162.229 L1711.94 156.349 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1742.93 137.067 L1742.93 141.094 Q1741.13 140.169 1739.18 139.706 Q1737.24 139.243 1735.15 139.243 Q1731.98 139.243 1730.38 140.215 Q1728.81 141.187 1728.81 143.131 Q1728.81 144.613 1729.94 145.469 Q1731.08 146.303 1734.5 147.067 L1735.96 147.391 Q1740.5 148.363 1742.4 150.145 Q1744.32 151.905 1744.32 155.076 Q1744.32 158.687 1741.45 160.793 Q1738.6 162.9 1733.6 162.9 Q1731.52 162.9 1729.25 162.483 Q1727 162.09 1724.51 161.28 L1724.51 156.881 Q1726.87 158.108 1729.16 158.733 Q1731.45 159.335 1733.69 159.335 Q1736.7 159.335 1738.32 158.317 Q1739.94 157.275 1739.94 155.4 Q1739.94 153.664 1738.76 152.738 Q1737.61 151.812 1733.65 150.956 L1732.17 150.608 Q1728.21 149.775 1726.45 148.062 Q1724.69 146.326 1724.69 143.317 Q1724.69 139.659 1727.28 137.669 Q1729.88 135.678 1734.64 135.678 Q1737 135.678 1739.09 136.025 Q1741.17 136.372 1742.93 137.067 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1776.4 126.257 Q1773.3 131.581 1771.8 136.789 Q1770.29 141.997 1770.29 147.344 Q1770.29 152.692 1771.8 157.946 Q1773.32 163.178 1776.4 168.479 L1772.7 168.479 Q1769.23 163.039 1767.49 157.784 Q1765.78 152.53 1765.78 147.344 Q1765.78 142.182 1767.49 136.951 Q1769.2 131.72 1772.7 126.257 L1776.4 126.257 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1799.44 143.594 Q1802.79 144.312 1804.67 146.581 Q1806.56 148.849 1806.56 152.182 Q1806.56 157.298 1803.05 160.099 Q1799.53 162.9 1793.05 162.9 Q1790.87 162.9 1788.56 162.46 Q1786.26 162.043 1783.81 161.187 L1783.81 156.673 Q1785.75 157.807 1788.07 158.386 Q1790.38 158.965 1792.91 158.965 Q1797.31 158.965 1799.6 157.229 Q1801.91 155.493 1801.91 152.182 Q1801.91 149.127 1799.76 147.414 Q1797.63 145.678 1793.81 145.678 L1789.78 145.678 L1789.78 141.835 L1794 141.835 Q1797.44 141.835 1799.27 140.469 Q1801.1 139.081 1801.1 136.488 Q1801.1 133.826 1799.2 132.414 Q1797.33 130.979 1793.81 130.979 Q1791.89 130.979 1789.69 131.395 Q1787.49 131.812 1784.85 132.692 L1784.85 128.525 Q1787.51 127.784 1789.83 127.414 Q1792.17 127.044 1794.23 127.044 Q1799.55 127.044 1802.65 129.474 Q1805.75 131.882 1805.75 136.002 Q1805.75 138.872 1804.11 140.863 Q1802.47 142.831 1799.44 143.594 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1829.6 143.594 Q1832.95 144.312 1834.83 146.581 Q1836.73 148.849 1836.73 152.182 Q1836.73 157.298 1833.21 160.099 Q1829.69 162.9 1823.21 162.9 Q1821.03 162.9 1818.72 162.46 Q1816.43 162.043 1813.97 161.187 L1813.97 156.673 Q1815.92 157.807 1818.23 158.386 Q1820.55 158.965 1823.07 158.965 Q1827.47 158.965 1829.76 157.229 Q1832.07 155.493 1832.07 152.182 Q1832.07 149.127 1829.92 147.414 Q1827.79 145.678 1823.97 145.678 L1819.94 145.678 L1819.94 141.835 L1824.16 141.835 Q1827.61 141.835 1829.43 140.469 Q1831.26 139.081 1831.26 136.488 Q1831.26 133.826 1829.37 132.414 Q1827.49 130.979 1823.97 130.979 Q1822.05 130.979 1819.85 131.395 Q1817.65 131.812 1815.01 132.692 L1815.01 128.525 Q1817.68 127.784 1819.99 127.414 Q1822.33 127.044 1824.39 127.044 Q1829.71 127.044 1832.81 129.474 Q1835.92 131.882 1835.92 136.002 Q1835.92 138.872 1834.27 140.863 Q1832.63 142.831 1829.6 143.594 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1845.64 127.669 L1863.99 127.669 L1863.99 131.604 L1849.92 131.604 L1849.92 140.076 Q1850.94 139.729 1851.96 139.567 Q1852.98 139.382 1854 139.382 Q1859.78 139.382 1863.16 142.553 Q1866.54 145.724 1866.54 151.141 Q1866.54 156.719 1863.07 159.821 Q1859.6 162.9 1853.28 162.9 Q1851.1 162.9 1848.83 162.53 Q1846.59 162.159 1844.18 161.418 L1844.18 156.719 Q1846.26 157.854 1848.49 158.409 Q1850.71 158.965 1853.18 158.965 Q1857.19 158.965 1859.53 156.858 Q1861.87 154.752 1861.87 151.141 Q1861.87 147.53 1859.53 145.423 Q1857.19 143.317 1853.18 143.317 Q1851.31 143.317 1849.43 143.733 Q1847.58 144.15 1845.64 145.03 L1845.64 127.669 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1875.8 127.669 L1894.16 127.669 L1894.16 131.604 L1880.08 131.604 L1880.08 140.076 Q1881.1 139.729 1882.12 139.567 Q1883.14 139.382 1884.16 139.382 Q1889.94 139.382 1893.32 142.553 Q1896.7 145.724 1896.7 151.141 Q1896.7 156.719 1893.23 159.821 Q1889.76 162.9 1883.44 162.9 Q1881.26 162.9 1878.99 162.53 Q1876.75 162.159 1874.34 161.418 L1874.34 156.719 Q1876.43 157.854 1878.65 158.409 Q1880.87 158.965 1883.35 158.965 Q1887.35 158.965 1889.69 156.858 Q1892.03 154.752 1892.03 151.141 Q1892.03 147.53 1889.69 145.423 Q1887.35 143.317 1883.35 143.317 Q1881.47 143.317 1879.6 143.733 Q1877.74 144.15 1875.8 145.03 L1875.8 127.669 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1918.76 131.743 L1906.96 150.192 L1918.76 150.192 L1918.76 131.743 M1917.54 127.669 L1923.42 127.669 L1923.42 150.192 L1928.35 150.192 L1928.35 154.081 L1923.42 154.081 L1923.42 162.229 L1918.76 162.229 L1918.76 154.081 L1903.16 154.081 L1903.16 149.567 L1917.54 127.669 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1948.93 131.743 L1937.12 150.192 L1948.93 150.192 L1948.93 131.743 M1947.7 127.669 L1953.58 127.669 L1953.58 150.192 L1958.51 150.192 L1958.51 154.081 L1953.58 154.081 L1953.58 162.229 L1948.93 162.229 L1948.93 154.081 L1933.32 154.081 L1933.32 149.567 L1947.7 127.669 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1980.41 143.594 Q1983.76 144.312 1985.64 146.581 Q1987.54 148.849 1987.54 152.182 Q1987.54 157.298 1984.02 160.099 Q1980.5 162.9 1974.02 162.9 Q1971.84 162.9 1969.53 162.46 Q1967.24 162.043 1964.78 161.187 L1964.78 156.673 Q1966.73 157.807 1969.04 158.386 Q1971.36 158.965 1973.88 158.965 Q1978.28 158.965 1980.57 157.229 Q1982.88 155.493 1982.88 152.182 Q1982.88 149.127 1980.73 147.414 Q1978.6 145.678 1974.78 145.678 L1970.75 145.678 L1970.75 141.835 L1974.97 141.835 Q1978.42 141.835 1980.24 140.469 Q1982.07 139.081 1982.07 136.488 Q1982.07 133.826 1980.17 132.414 Q1978.3 130.979 1974.78 130.979 Q1972.86 130.979 1970.66 131.395 Q1968.46 131.812 1965.82 132.692 L1965.82 128.525 Q1968.49 127.784 1970.8 127.414 Q1973.14 127.044 1975.2 127.044 Q1980.52 127.044 1983.62 129.474 Q1986.73 131.882 1986.73 136.002 Q1986.73 138.872 1985.08 140.863 Q1983.44 142.831 1980.41 143.594 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M2000.43 158.293 L2016.75 158.293 L2016.75 162.229 L1994.8 162.229 L1994.8 158.293 Q1997.47 155.539 2002.05 150.909 Q2006.66 146.256 2007.84 144.914 Q2010.08 142.391 2010.96 140.655 Q2011.86 138.895 2011.86 137.206 Q2011.86 134.451 2009.92 132.715 Q2008 130.979 2004.9 130.979 Q2002.7 130.979 2000.24 131.743 Q1997.81 132.507 1995.04 134.057 L1995.04 129.335 Q1997.86 128.201 2000.31 127.622 Q2002.77 127.044 2004.8 127.044 Q2010.17 127.044 2013.37 129.729 Q2016.56 132.414 2016.56 136.905 Q2016.56 139.034 2015.75 140.956 Q2014.97 142.854 2012.86 145.446 Q2012.28 146.118 2009.18 149.335 Q2006.08 152.53 2000.43 158.293 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M2025.29 126.257 L2028.99 126.257 Q2032.47 131.72 2034.18 136.951 Q2035.92 142.182 2035.92 147.344 Q2035.92 152.53 2034.18 157.784 Q2032.47 163.039 2028.99 168.479 L2025.29 168.479 Q2028.37 163.178 2029.87 157.946 Q2031.4 152.692 2031.4 147.344 Q2031.4 141.997 2029.87 136.789 Q2028.37 131.581 2025.29 126.257 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip030)" style="stroke:#e26f46; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1415.15,196.789 1562.06,196.789 
  "/>
<path clip-path="url(#clip030)" d="M1606.89 193.12 Q1608.49 190.249 1610.71 188.884 Q1612.93 187.518 1615.94 187.518 Q1619.99 187.518 1622.19 190.365 Q1624.39 193.189 1624.39 198.421 L1624.39 214.069 L1620.11 214.069 L1620.11 198.559 Q1620.11 194.833 1618.79 193.027 Q1617.47 191.222 1614.76 191.222 Q1611.45 191.222 1609.53 193.421 Q1607.61 195.62 1607.61 199.416 L1607.61 214.069 L1603.33 214.069 L1603.33 198.559 Q1603.33 194.809 1602.01 193.027 Q1600.69 191.222 1597.93 191.222 Q1594.67 191.222 1592.75 193.444 Q1590.83 195.643 1590.83 199.416 L1590.83 214.069 L1586.54 214.069 L1586.54 188.143 L1590.83 188.143 L1590.83 192.171 Q1592.28 189.786 1594.32 188.652 Q1596.36 187.518 1599.16 187.518 Q1601.98 187.518 1603.95 188.953 Q1605.94 190.388 1606.89 193.12 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1644.67 201.036 Q1639.51 201.036 1637.51 202.217 Q1635.52 203.397 1635.52 206.245 Q1635.52 208.513 1637.01 209.856 Q1638.51 211.175 1641.08 211.175 Q1644.62 211.175 1646.75 208.675 Q1648.9 206.152 1648.9 201.985 L1648.9 201.036 L1644.67 201.036 M1653.16 199.277 L1653.16 214.069 L1648.9 214.069 L1648.9 210.133 Q1647.45 212.495 1645.27 213.629 Q1643.09 214.74 1639.95 214.74 Q1635.96 214.74 1633.6 212.518 Q1631.26 210.272 1631.26 206.522 Q1631.26 202.147 1634.18 199.925 Q1637.12 197.703 1642.93 197.703 L1648.9 197.703 L1648.9 197.286 Q1648.9 194.347 1646.96 192.749 Q1645.04 191.129 1641.54 191.129 Q1639.32 191.129 1637.21 191.661 Q1635.11 192.194 1633.16 193.259 L1633.16 189.323 Q1635.5 188.421 1637.7 187.981 Q1639.9 187.518 1641.98 187.518 Q1647.61 187.518 1650.39 190.434 Q1653.16 193.351 1653.16 199.277 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1678.46 188.907 L1678.46 192.934 Q1676.66 192.009 1674.71 191.546 Q1672.77 191.083 1670.69 191.083 Q1667.51 191.083 1665.92 192.055 Q1664.34 193.027 1664.34 194.971 Q1664.34 196.453 1665.48 197.309 Q1666.61 198.143 1670.04 198.907 L1671.5 199.231 Q1676.03 200.203 1677.93 201.985 Q1679.85 203.745 1679.85 206.916 Q1679.85 210.527 1676.98 212.633 Q1674.13 214.74 1669.13 214.74 Q1667.05 214.74 1664.78 214.323 Q1662.54 213.93 1660.04 213.12 L1660.04 208.721 Q1662.4 209.948 1664.69 210.573 Q1666.98 211.175 1669.23 211.175 Q1672.24 211.175 1673.86 210.157 Q1675.48 209.115 1675.48 207.24 Q1675.48 205.504 1674.3 204.578 Q1673.14 203.652 1669.18 202.796 L1667.7 202.448 Q1663.74 201.615 1661.98 199.902 Q1660.22 198.166 1660.22 195.157 Q1660.22 191.499 1662.82 189.509 Q1665.41 187.518 1670.18 187.518 Q1672.54 187.518 1674.62 187.865 Q1676.7 188.212 1678.46 188.907 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1703.16 188.907 L1703.16 192.934 Q1701.36 192.009 1699.41 191.546 Q1697.47 191.083 1695.38 191.083 Q1692.21 191.083 1690.62 192.055 Q1689.04 193.027 1689.04 194.971 Q1689.04 196.453 1690.18 197.309 Q1691.31 198.143 1694.74 198.907 L1696.19 199.231 Q1700.73 200.203 1702.63 201.985 Q1704.55 203.745 1704.55 206.916 Q1704.55 210.527 1701.68 212.633 Q1698.83 214.74 1693.83 214.74 Q1691.75 214.74 1689.48 214.323 Q1687.24 213.93 1684.74 213.12 L1684.74 208.721 Q1687.1 209.948 1689.39 210.573 Q1691.68 211.175 1693.93 211.175 Q1696.94 211.175 1698.56 210.157 Q1700.18 209.115 1700.18 207.24 Q1700.18 205.504 1699 204.578 Q1697.84 203.652 1693.88 202.796 L1692.4 202.448 Q1688.44 201.615 1686.68 199.902 Q1684.92 198.166 1684.92 195.157 Q1684.92 191.499 1687.51 189.509 Q1690.11 187.518 1694.88 187.518 Q1697.24 187.518 1699.32 187.865 Q1701.4 188.212 1703.16 188.907 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1711.94 208.189 L1716.82 208.189 L1716.82 214.069 L1711.94 214.069 L1711.94 208.189 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1723.35 188.143 L1727.86 188.143 L1735.96 209.902 L1744.07 188.143 L1748.58 188.143 L1738.86 214.069 L1733.07 214.069 L1723.35 188.143 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1779.76 178.097 Q1776.66 183.421 1775.15 188.629 Q1773.65 193.837 1773.65 199.184 Q1773.65 204.532 1775.15 209.786 Q1776.68 215.018 1779.76 220.319 L1776.06 220.319 Q1772.58 214.879 1770.85 209.624 Q1769.13 204.37 1769.13 199.184 Q1769.13 194.022 1770.85 188.791 Q1772.56 183.56 1776.06 178.097 L1779.76 178.097 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1802.79 195.434 Q1806.15 196.152 1808.02 198.421 Q1809.92 200.689 1809.92 204.022 Q1809.92 209.138 1806.4 211.939 Q1802.88 214.74 1796.4 214.74 Q1794.23 214.74 1791.91 214.3 Q1789.62 213.883 1787.17 213.027 L1787.17 208.513 Q1789.11 209.647 1791.43 210.226 Q1793.74 210.805 1796.26 210.805 Q1800.66 210.805 1802.95 209.069 Q1805.27 207.333 1805.27 204.022 Q1805.27 200.967 1803.12 199.254 Q1800.99 197.518 1797.17 197.518 L1793.14 197.518 L1793.14 193.675 L1797.35 193.675 Q1800.8 193.675 1802.63 192.309 Q1804.46 190.921 1804.46 188.328 Q1804.46 185.666 1802.56 184.254 Q1800.69 182.819 1797.17 182.819 Q1795.25 182.819 1793.05 183.235 Q1790.85 183.652 1788.21 184.532 L1788.21 180.365 Q1790.87 179.624 1793.19 179.254 Q1795.52 178.884 1797.58 178.884 Q1802.91 178.884 1806.01 181.314 Q1809.11 183.722 1809.11 187.842 Q1809.11 190.712 1807.47 192.703 Q1805.82 194.671 1802.79 195.434 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1832.95 195.434 Q1836.31 196.152 1838.18 198.421 Q1840.08 200.689 1840.08 204.022 Q1840.08 209.138 1836.56 211.939 Q1833.05 214.74 1826.56 214.74 Q1824.39 214.74 1822.07 214.3 Q1819.78 213.883 1817.33 213.027 L1817.33 208.513 Q1819.27 209.647 1821.59 210.226 Q1823.9 210.805 1826.43 210.805 Q1830.82 210.805 1833.12 209.069 Q1835.43 207.333 1835.43 204.022 Q1835.43 200.967 1833.28 199.254 Q1831.15 197.518 1827.33 197.518 L1823.3 197.518 L1823.3 193.675 L1827.51 193.675 Q1830.96 193.675 1832.79 192.309 Q1834.62 190.921 1834.62 188.328 Q1834.62 185.666 1832.72 184.254 Q1830.85 182.819 1827.33 182.819 Q1825.41 182.819 1823.21 183.235 Q1821.01 183.652 1818.37 184.532 L1818.37 180.365 Q1821.03 179.624 1823.35 179.254 Q1825.69 178.884 1827.75 178.884 Q1833.07 178.884 1836.17 181.314 Q1839.27 183.722 1839.27 187.842 Q1839.27 190.712 1837.63 192.703 Q1835.99 194.671 1832.95 195.434 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1849 179.509 L1867.35 179.509 L1867.35 183.444 L1853.28 183.444 L1853.28 191.916 Q1854.3 191.569 1855.31 191.407 Q1856.33 191.222 1857.35 191.222 Q1863.14 191.222 1866.52 194.393 Q1869.9 197.564 1869.9 202.981 Q1869.9 208.559 1866.43 211.661 Q1862.95 214.74 1856.63 214.74 Q1854.46 214.74 1852.19 214.37 Q1849.94 213.999 1847.54 213.258 L1847.54 208.559 Q1849.62 209.694 1851.84 210.249 Q1854.06 210.805 1856.54 210.805 Q1860.55 210.805 1862.88 208.698 Q1865.22 206.592 1865.22 202.981 Q1865.22 199.37 1862.88 197.263 Q1860.55 195.157 1856.54 195.157 Q1854.67 195.157 1852.79 195.573 Q1850.94 195.99 1849 196.87 L1849 179.509 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1879.16 179.509 L1897.51 179.509 L1897.51 183.444 L1883.44 183.444 L1883.44 191.916 Q1884.46 191.569 1885.48 191.407 Q1886.49 191.222 1887.51 191.222 Q1893.3 191.222 1896.68 194.393 Q1900.06 197.564 1900.06 202.981 Q1900.06 208.559 1896.59 211.661 Q1893.12 214.74 1886.8 214.74 Q1884.62 214.74 1882.35 214.37 Q1880.11 213.999 1877.7 213.258 L1877.7 208.559 Q1879.78 209.694 1882 210.249 Q1884.23 210.805 1886.7 210.805 Q1890.71 210.805 1893.05 208.698 Q1895.38 206.592 1895.38 202.981 Q1895.38 199.37 1893.05 197.263 Q1890.71 195.157 1886.7 195.157 Q1884.83 195.157 1882.95 195.573 Q1881.1 195.99 1879.16 196.87 L1879.16 179.509 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1922.12 183.583 L1910.31 202.032 L1922.12 202.032 L1922.12 183.583 M1920.89 179.509 L1926.77 179.509 L1926.77 202.032 L1931.7 202.032 L1931.7 205.921 L1926.77 205.921 L1926.77 214.069 L1922.12 214.069 L1922.12 205.921 L1906.52 205.921 L1906.52 201.407 L1920.89 179.509 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1952.28 183.583 L1940.48 202.032 L1952.28 202.032 L1952.28 183.583 M1951.05 179.509 L1956.93 179.509 L1956.93 202.032 L1961.86 202.032 L1961.86 205.921 L1956.93 205.921 L1956.93 214.069 L1952.28 214.069 L1952.28 205.921 L1936.68 205.921 L1936.68 201.407 L1951.05 179.509 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1983.76 195.434 Q1987.12 196.152 1988.99 198.421 Q1990.89 200.689 1990.89 204.022 Q1990.89 209.138 1987.37 211.939 Q1983.86 214.74 1977.37 214.74 Q1975.2 214.74 1972.88 214.3 Q1970.59 213.883 1968.14 213.027 L1968.14 208.513 Q1970.08 209.647 1972.4 210.226 Q1974.71 210.805 1977.24 210.805 Q1981.63 210.805 1983.92 209.069 Q1986.24 207.333 1986.24 204.022 Q1986.24 200.967 1984.09 199.254 Q1981.96 197.518 1978.14 197.518 L1974.11 197.518 L1974.11 193.675 L1978.32 193.675 Q1981.77 193.675 1983.6 192.309 Q1985.43 190.921 1985.43 188.328 Q1985.43 185.666 1983.53 184.254 Q1981.66 182.819 1978.14 182.819 Q1976.22 182.819 1974.02 183.235 Q1971.82 183.652 1969.18 184.532 L1969.18 180.365 Q1971.84 179.624 1974.16 179.254 Q1976.49 178.884 1978.55 178.884 Q1983.88 178.884 1986.98 181.314 Q1990.08 183.722 1990.08 187.842 Q1990.08 190.712 1988.44 192.703 Q1986.8 194.671 1983.76 195.434 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M2013.92 195.434 Q2017.28 196.152 2019.16 198.421 Q2021.05 200.689 2021.05 204.022 Q2021.05 209.138 2017.54 211.939 Q2014.02 214.74 2007.54 214.74 Q2005.36 214.74 2003.05 214.3 Q2000.75 213.883 1998.3 213.027 L1998.3 208.513 Q2000.24 209.647 2002.56 210.226 Q2004.87 210.805 2007.4 210.805 Q2011.8 210.805 2014.09 209.069 Q2016.4 207.333 2016.4 204.022 Q2016.4 200.967 2014.25 199.254 Q2012.12 197.518 2008.3 197.518 L2004.27 197.518 L2004.27 193.675 L2008.48 193.675 Q2011.93 193.675 2013.76 192.309 Q2015.59 190.921 2015.59 188.328 Q2015.59 185.666 2013.69 184.254 Q2011.82 182.819 2008.3 182.819 Q2006.38 182.819 2004.18 183.235 Q2001.98 183.652 1999.34 184.532 L1999.34 180.365 Q2002 179.624 2004.32 179.254 Q2006.66 178.884 2008.72 178.884 Q2014.04 178.884 2017.14 181.314 Q2020.24 183.722 2020.24 187.842 Q2020.24 190.712 2018.6 192.703 Q2016.96 194.671 2013.92 195.434 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M2028.65 178.097 L2032.35 178.097 Q2035.82 183.56 2037.54 188.791 Q2039.27 194.022 2039.27 199.184 Q2039.27 204.37 2037.54 209.624 Q2035.82 214.879 2032.35 220.319 L2028.65 220.319 Q2031.73 215.018 2033.23 209.786 Q2034.76 204.532 2034.76 199.184 Q2034.76 193.837 2033.23 188.629 Q2031.73 183.421 2028.65 178.097 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip030)" style="stroke:#3da44d; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1415.15,248.629 1562.06,248.629 
  "/>
<path clip-path="url(#clip030)" d="M1606.89 244.96 Q1608.49 242.089 1610.71 240.724 Q1612.93 239.358 1615.94 239.358 Q1619.99 239.358 1622.19 242.205 Q1624.39 245.029 1624.39 250.261 L1624.39 265.909 L1620.11 265.909 L1620.11 250.399 Q1620.11 246.673 1618.79 244.867 Q1617.47 243.062 1614.76 243.062 Q1611.45 243.062 1609.53 245.261 Q1607.61 247.46 1607.61 251.256 L1607.61 265.909 L1603.33 265.909 L1603.33 250.399 Q1603.33 246.649 1602.01 244.867 Q1600.69 243.062 1597.93 243.062 Q1594.67 243.062 1592.75 245.284 Q1590.83 247.483 1590.83 251.256 L1590.83 265.909 L1586.54 265.909 L1586.54 239.983 L1590.83 239.983 L1590.83 244.011 Q1592.28 241.626 1594.32 240.492 Q1596.36 239.358 1599.16 239.358 Q1601.98 239.358 1603.95 240.793 Q1605.94 242.228 1606.89 244.96 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1644.67 252.876 Q1639.51 252.876 1637.51 254.057 Q1635.52 255.237 1635.52 258.085 Q1635.52 260.353 1637.01 261.696 Q1638.51 263.015 1641.08 263.015 Q1644.62 263.015 1646.75 260.515 Q1648.9 257.992 1648.9 253.825 L1648.9 252.876 L1644.67 252.876 M1653.16 251.117 L1653.16 265.909 L1648.9 265.909 L1648.9 261.973 Q1647.45 264.335 1645.27 265.469 Q1643.09 266.58 1639.95 266.58 Q1635.96 266.58 1633.6 264.358 Q1631.26 262.112 1631.26 258.362 Q1631.26 253.987 1634.18 251.765 Q1637.12 249.543 1642.93 249.543 L1648.9 249.543 L1648.9 249.126 Q1648.9 246.187 1646.96 244.589 Q1645.04 242.969 1641.54 242.969 Q1639.32 242.969 1637.21 243.501 Q1635.11 244.034 1633.16 245.099 L1633.16 241.163 Q1635.5 240.261 1637.7 239.821 Q1639.9 239.358 1641.98 239.358 Q1647.61 239.358 1650.39 242.274 Q1653.16 245.191 1653.16 251.117 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1678.46 240.747 L1678.46 244.774 Q1676.66 243.849 1674.71 243.386 Q1672.77 242.923 1670.69 242.923 Q1667.51 242.923 1665.92 243.895 Q1664.34 244.867 1664.34 246.811 Q1664.34 248.293 1665.48 249.149 Q1666.61 249.983 1670.04 250.747 L1671.5 251.071 Q1676.03 252.043 1677.93 253.825 Q1679.85 255.585 1679.85 258.756 Q1679.85 262.367 1676.98 264.473 Q1674.13 266.58 1669.13 266.58 Q1667.05 266.58 1664.78 266.163 Q1662.54 265.77 1660.04 264.96 L1660.04 260.561 Q1662.4 261.788 1664.69 262.413 Q1666.98 263.015 1669.23 263.015 Q1672.24 263.015 1673.86 261.997 Q1675.48 260.955 1675.48 259.08 Q1675.48 257.344 1674.3 256.418 Q1673.14 255.492 1669.18 254.636 L1667.7 254.288 Q1663.74 253.455 1661.98 251.742 Q1660.22 250.006 1660.22 246.997 Q1660.22 243.339 1662.82 241.349 Q1665.41 239.358 1670.18 239.358 Q1672.54 239.358 1674.62 239.705 Q1676.7 240.052 1678.46 240.747 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1703.16 240.747 L1703.16 244.774 Q1701.36 243.849 1699.41 243.386 Q1697.47 242.923 1695.38 242.923 Q1692.21 242.923 1690.62 243.895 Q1689.04 244.867 1689.04 246.811 Q1689.04 248.293 1690.18 249.149 Q1691.31 249.983 1694.74 250.747 L1696.19 251.071 Q1700.73 252.043 1702.63 253.825 Q1704.55 255.585 1704.55 258.756 Q1704.55 262.367 1701.68 264.473 Q1698.83 266.58 1693.83 266.58 Q1691.75 266.58 1689.48 266.163 Q1687.24 265.77 1684.74 264.96 L1684.74 260.561 Q1687.1 261.788 1689.39 262.413 Q1691.68 263.015 1693.93 263.015 Q1696.94 263.015 1698.56 261.997 Q1700.18 260.955 1700.18 259.08 Q1700.18 257.344 1699 256.418 Q1697.84 255.492 1693.88 254.636 L1692.4 254.288 Q1688.44 253.455 1686.68 251.742 Q1684.92 250.006 1684.92 246.997 Q1684.92 243.339 1687.51 241.349 Q1690.11 239.358 1694.88 239.358 Q1697.24 239.358 1699.32 239.705 Q1701.4 240.052 1703.16 240.747 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1711.94 260.029 L1716.82 260.029 L1716.82 265.909 L1711.94 265.909 L1711.94 260.029 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1738.19 252.876 Q1733.02 252.876 1731.03 254.057 Q1729.04 255.237 1729.04 258.085 Q1729.04 260.353 1730.52 261.696 Q1732.03 263.015 1734.6 263.015 Q1738.14 263.015 1740.27 260.515 Q1742.42 257.992 1742.42 253.825 L1742.42 252.876 L1738.19 252.876 M1746.68 251.117 L1746.68 265.909 L1742.42 265.909 L1742.42 261.973 Q1740.96 264.335 1738.79 265.469 Q1736.61 266.58 1733.46 266.58 Q1729.48 266.58 1727.12 264.358 Q1724.78 262.112 1724.78 258.362 Q1724.78 253.987 1727.7 251.765 Q1730.64 249.543 1736.45 249.543 L1742.42 249.543 L1742.42 249.126 Q1742.42 246.187 1740.48 244.589 Q1738.56 242.969 1735.06 242.969 Q1732.84 242.969 1730.73 243.501 Q1728.63 244.034 1726.68 245.099 L1726.68 241.163 Q1729.02 240.261 1731.22 239.821 Q1733.42 239.358 1735.5 239.358 Q1741.13 239.358 1743.9 242.274 Q1746.68 245.191 1746.68 251.117 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1775.15 273.779 L1775.15 277.089 L1750.52 277.089 L1750.52 273.779 L1775.15 273.779 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1794.18 243.964 Q1793.46 243.548 1792.61 243.362 Q1791.77 243.154 1790.75 243.154 Q1787.14 243.154 1785.2 245.515 Q1783.28 247.853 1783.28 252.251 L1783.28 265.909 L1779 265.909 L1779 239.983 L1783.28 239.983 L1783.28 244.011 Q1784.62 241.649 1786.77 240.515 Q1788.93 239.358 1792 239.358 Q1792.44 239.358 1792.98 239.427 Q1793.51 239.474 1794.16 239.589 L1794.18 243.964 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1819.78 251.881 L1819.78 253.964 L1800.2 253.964 Q1800.48 258.362 1802.84 260.677 Q1805.22 262.969 1809.46 262.969 Q1811.91 262.969 1814.2 262.367 Q1816.52 261.765 1818.79 260.561 L1818.79 264.589 Q1816.5 265.561 1814.09 266.071 Q1811.68 266.58 1809.2 266.58 Q1803 266.58 1799.37 262.969 Q1795.75 259.358 1795.75 253.2 Q1795.75 246.835 1799.18 243.108 Q1802.63 239.358 1808.46 239.358 Q1813.69 239.358 1816.73 242.737 Q1819.78 246.094 1819.78 251.881 M1815.52 250.631 Q1815.48 247.136 1813.56 245.052 Q1811.66 242.969 1808.51 242.969 Q1804.94 242.969 1802.79 244.983 Q1800.66 246.997 1800.34 250.654 L1815.52 250.631 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1826.77 229.89 L1831.03 229.89 L1831.03 265.909 L1826.77 265.909 L1826.77 229.89 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1853.07 229.89 L1853.07 233.432 L1849 233.432 Q1846.7 233.432 1845.8 234.358 Q1844.92 235.284 1844.92 237.691 L1844.92 239.983 L1851.93 239.983 L1851.93 243.293 L1844.92 243.293 L1844.92 265.909 L1840.64 265.909 L1840.64 243.293 L1836.56 243.293 L1836.56 239.983 L1840.64 239.983 L1840.64 238.177 Q1840.64 233.849 1842.65 231.881 Q1844.67 229.89 1849.04 229.89 L1853.07 229.89 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1871.66 243.964 Q1870.94 243.548 1870.08 243.362 Q1869.25 243.154 1868.23 243.154 Q1864.62 243.154 1862.68 245.515 Q1860.75 247.853 1860.75 252.251 L1860.75 265.909 L1856.47 265.909 L1856.47 239.983 L1860.75 239.983 L1860.75 244.011 Q1862.1 241.649 1864.25 240.515 Q1866.4 239.358 1869.48 239.358 Q1869.92 239.358 1870.45 239.427 Q1870.99 239.474 1871.63 239.589 L1871.66 243.964 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1876.12 239.983 L1880.38 239.983 L1880.38 265.909 L1876.12 265.909 L1876.12 239.983 M1876.12 229.89 L1880.38 229.89 L1880.38 235.284 L1876.12 235.284 L1876.12 229.89 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1907.95 240.978 L1907.95 244.96 Q1906.15 243.964 1904.32 243.478 Q1902.51 242.969 1900.66 242.969 Q1896.52 242.969 1894.23 245.608 Q1891.93 248.224 1891.93 252.969 Q1891.93 257.714 1894.23 260.353 Q1896.52 262.969 1900.66 262.969 Q1902.51 262.969 1904.32 262.483 Q1906.15 261.973 1907.95 260.978 L1907.95 264.913 Q1906.17 265.747 1904.25 266.163 Q1902.35 266.58 1900.2 266.58 Q1894.34 266.58 1890.89 262.899 Q1887.44 259.219 1887.44 252.969 Q1887.44 246.626 1890.92 242.992 Q1894.41 239.358 1900.48 239.358 Q1902.44 239.358 1904.32 239.775 Q1906.19 240.168 1907.95 240.978 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1940.66 229.937 Q1937.56 235.261 1936.05 240.469 Q1934.55 245.677 1934.55 251.024 Q1934.55 256.372 1936.05 261.626 Q1937.58 266.858 1940.66 272.159 L1936.96 272.159 Q1933.49 266.719 1931.75 261.464 Q1930.04 256.21 1930.04 251.024 Q1930.04 245.862 1931.75 240.631 Q1933.46 235.4 1936.96 229.937 L1940.66 229.937 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1949.57 231.349 L1967.93 231.349 L1967.93 235.284 L1953.86 235.284 L1953.86 243.756 Q1954.87 243.409 1955.89 243.247 Q1956.91 243.062 1957.93 243.062 Q1963.72 243.062 1967.1 246.233 Q1970.48 249.404 1970.48 254.821 Q1970.48 260.399 1967 263.501 Q1963.53 266.58 1957.21 266.58 Q1955.04 266.58 1952.77 266.21 Q1950.52 265.839 1948.11 265.098 L1948.11 260.399 Q1950.2 261.534 1952.42 262.089 Q1954.64 262.645 1957.12 262.645 Q1961.12 262.645 1963.46 260.538 Q1965.8 258.432 1965.8 254.821 Q1965.8 251.21 1963.46 249.103 Q1961.12 246.997 1957.12 246.997 Q1955.24 246.997 1953.37 247.413 Q1951.52 247.83 1949.57 248.71 L1949.57 231.349 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1989.69 249.497 Q1986.36 249.497 1984.43 251.279 Q1982.54 253.061 1982.54 256.186 Q1982.54 259.311 1984.43 261.094 Q1986.36 262.876 1989.69 262.876 Q1993.02 262.876 1994.94 261.094 Q1996.86 259.288 1996.86 256.186 Q1996.86 253.061 1994.94 251.279 Q1993.05 249.497 1989.69 249.497 M1985.01 247.506 Q1982 246.765 1980.31 244.705 Q1978.65 242.645 1978.65 239.682 Q1978.65 235.538 1981.59 233.131 Q1984.55 230.724 1989.69 230.724 Q1994.85 230.724 1997.79 233.131 Q2000.73 235.538 2000.73 239.682 Q2000.73 242.645 1999.04 244.705 Q1997.37 246.765 1994.39 247.506 Q1997.77 248.293 1999.64 250.585 Q2001.54 252.876 2001.54 256.186 Q2001.54 261.21 1998.46 263.895 Q1995.41 266.58 1989.69 266.58 Q1983.97 266.58 1980.89 263.895 Q1977.84 261.21 1977.84 256.186 Q1977.84 252.876 1979.74 250.585 Q1981.63 248.293 1985.01 247.506 M1983.3 240.122 Q1983.3 242.807 1984.97 244.312 Q1986.66 245.816 1989.69 245.816 Q1992.7 245.816 1994.39 244.312 Q1996.1 242.807 1996.1 240.122 Q1996.1 237.437 1994.39 235.932 Q1992.7 234.427 1989.69 234.427 Q1986.66 234.427 1984.97 235.932 Q1983.3 237.437 1983.3 240.122 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M2008.67 231.349 L2030.89 231.349 L2030.89 233.339 L2018.35 265.909 L2013.46 265.909 L2025.27 235.284 L2008.67 235.284 L2008.67 231.349 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M2044.04 261.973 L2060.36 261.973 L2060.36 265.909 L2038.42 265.909 L2038.42 261.973 Q2041.08 259.219 2045.66 254.589 Q2050.27 249.936 2051.45 248.594 Q2053.69 246.071 2054.57 244.335 Q2055.48 242.575 2055.48 240.886 Q2055.48 238.131 2053.53 236.395 Q2051.61 234.659 2048.51 234.659 Q2046.31 234.659 2043.86 235.423 Q2041.42 236.187 2038.65 237.737 L2038.65 233.015 Q2041.47 231.881 2043.92 231.302 Q2046.38 230.724 2048.42 230.724 Q2053.79 230.724 2056.98 233.409 Q2060.17 236.094 2060.17 240.585 Q2060.17 242.714 2059.36 244.636 Q2058.58 246.534 2056.47 249.126 Q2055.89 249.798 2052.79 253.015 Q2049.69 256.21 2044.04 261.973 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M2080.17 234.427 Q2076.56 234.427 2074.73 237.992 Q2072.93 241.534 2072.93 248.663 Q2072.93 255.77 2074.73 259.335 Q2076.56 262.876 2080.17 262.876 Q2083.81 262.876 2085.61 259.335 Q2087.44 255.77 2087.44 248.663 Q2087.44 241.534 2085.61 237.992 Q2083.81 234.427 2080.17 234.427 M2080.17 230.724 Q2085.98 230.724 2089.04 235.33 Q2092.12 239.913 2092.12 248.663 Q2092.12 257.39 2089.04 261.997 Q2085.98 266.58 2080.17 266.58 Q2074.36 266.58 2071.29 261.997 Q2068.23 257.39 2068.23 248.663 Q2068.23 239.913 2071.29 235.33 Q2074.36 230.724 2080.17 230.724 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M2104.36 261.973 L2120.68 261.973 L2120.68 265.909 L2098.74 265.909 L2098.74 261.973 Q2101.4 259.219 2105.98 254.589 Q2110.59 249.936 2111.77 248.594 Q2114.02 246.071 2114.9 244.335 Q2115.8 242.575 2115.8 240.886 Q2115.8 238.131 2113.85 236.395 Q2111.93 234.659 2108.83 234.659 Q2106.63 234.659 2104.18 235.423 Q2101.75 236.187 2098.97 237.737 L2098.97 233.015 Q2101.79 231.881 2104.25 231.302 Q2106.7 230.724 2108.74 230.724 Q2114.11 230.724 2117.3 233.409 Q2120.5 236.094 2120.5 240.585 Q2120.5 242.714 2119.69 244.636 Q2118.9 246.534 2116.79 249.126 Q2116.22 249.798 2113.11 253.015 Q2110.01 256.21 2104.36 261.973 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M2130.54 231.349 L2148.9 231.349 L2148.9 235.284 L2134.83 235.284 L2134.83 243.756 Q2135.85 243.409 2136.86 243.247 Q2137.88 243.062 2138.9 243.062 Q2144.69 243.062 2148.07 246.233 Q2151.45 249.404 2151.45 254.821 Q2151.45 260.399 2147.97 263.501 Q2144.5 266.58 2138.18 266.58 Q2136.01 266.58 2133.74 266.21 Q2131.49 265.839 2129.09 265.098 L2129.09 260.399 Q2131.17 261.534 2133.39 262.089 Q2135.61 262.645 2138.09 262.645 Q2142.1 262.645 2144.43 260.538 Q2146.77 258.432 2146.77 254.821 Q2146.77 251.21 2144.43 249.103 Q2142.1 246.997 2138.09 246.997 Q2136.22 246.997 2134.34 247.413 Q2132.49 247.83 2130.54 248.71 L2130.54 231.349 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M2171.24 246.765 Q2168.09 246.765 2166.24 248.918 Q2164.41 251.071 2164.41 254.821 Q2164.41 258.548 2166.24 260.723 Q2168.09 262.876 2171.24 262.876 Q2174.39 262.876 2176.22 260.723 Q2178.07 258.548 2178.07 254.821 Q2178.07 251.071 2176.22 248.918 Q2174.39 246.765 2171.24 246.765 M2180.52 232.113 L2180.52 236.372 Q2178.76 235.538 2176.96 235.099 Q2175.17 234.659 2173.41 234.659 Q2168.79 234.659 2166.33 237.784 Q2163.9 240.909 2163.55 247.228 Q2164.92 245.214 2166.98 244.149 Q2169.04 243.062 2171.52 243.062 Q2176.72 243.062 2179.73 246.233 Q2182.77 249.381 2182.77 254.821 Q2182.77 260.145 2179.62 263.362 Q2176.47 266.58 2171.24 266.58 Q2165.24 266.58 2162.07 261.997 Q2158.9 257.39 2158.9 248.663 Q2158.9 240.469 2162.79 235.608 Q2166.68 230.724 2173.23 230.724 Q2174.99 230.724 2176.77 231.071 Q2178.58 231.418 2180.52 232.113 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M2191.63 261.973 L2199.27 261.973 L2199.27 235.608 L2190.96 237.275 L2190.96 233.015 L2199.22 231.349 L2203.9 231.349 L2203.9 261.973 L2211.54 261.973 L2211.54 265.909 L2191.63 265.909 L2191.63 261.973 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M2219.71 229.937 L2223.41 229.937 Q2226.89 235.4 2228.6 240.631 Q2230.34 245.862 2230.34 251.024 Q2230.34 256.21 2228.6 261.464 Q2226.89 266.719 2223.41 272.159 L2219.71 272.159 Q2222.79 266.858 2224.29 261.626 Q2225.82 256.372 2225.82 251.024 Q2225.82 245.677 2224.29 240.469 Q2222.79 235.261 2219.71 229.937 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip030)" style="stroke:#c271d2; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1415.15,300.469 1562.06,300.469 
  "/>
<path clip-path="url(#clip030)" d="M1606.89 296.8 Q1608.49 293.929 1610.71 292.564 Q1612.93 291.198 1615.94 291.198 Q1619.99 291.198 1622.19 294.045 Q1624.39 296.869 1624.39 302.101 L1624.39 317.749 L1620.11 317.749 L1620.11 302.239 Q1620.11 298.513 1618.79 296.707 Q1617.47 294.902 1614.76 294.902 Q1611.45 294.902 1609.53 297.101 Q1607.61 299.3 1607.61 303.096 L1607.61 317.749 L1603.33 317.749 L1603.33 302.239 Q1603.33 298.489 1602.01 296.707 Q1600.69 294.902 1597.93 294.902 Q1594.67 294.902 1592.75 297.124 Q1590.83 299.323 1590.83 303.096 L1590.83 317.749 L1586.54 317.749 L1586.54 291.823 L1590.83 291.823 L1590.83 295.851 Q1592.28 293.466 1594.32 292.332 Q1596.36 291.198 1599.16 291.198 Q1601.98 291.198 1603.95 292.633 Q1605.94 294.068 1606.89 296.8 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1644.67 304.716 Q1639.51 304.716 1637.51 305.897 Q1635.52 307.077 1635.52 309.925 Q1635.52 312.193 1637.01 313.536 Q1638.51 314.855 1641.08 314.855 Q1644.62 314.855 1646.75 312.355 Q1648.9 309.832 1648.9 305.665 L1648.9 304.716 L1644.67 304.716 M1653.16 302.957 L1653.16 317.749 L1648.9 317.749 L1648.9 313.813 Q1647.45 316.175 1645.27 317.309 Q1643.09 318.42 1639.95 318.42 Q1635.96 318.42 1633.6 316.198 Q1631.26 313.952 1631.26 310.202 Q1631.26 305.827 1634.18 303.605 Q1637.12 301.383 1642.93 301.383 L1648.9 301.383 L1648.9 300.966 Q1648.9 298.027 1646.96 296.429 Q1645.04 294.809 1641.54 294.809 Q1639.32 294.809 1637.21 295.341 Q1635.11 295.874 1633.16 296.939 L1633.16 293.003 Q1635.5 292.101 1637.7 291.661 Q1639.9 291.198 1641.98 291.198 Q1647.61 291.198 1650.39 294.114 Q1653.16 297.031 1653.16 302.957 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1678.46 292.587 L1678.46 296.614 Q1676.66 295.689 1674.71 295.226 Q1672.77 294.763 1670.69 294.763 Q1667.51 294.763 1665.92 295.735 Q1664.34 296.707 1664.34 298.651 Q1664.34 300.133 1665.48 300.989 Q1666.61 301.823 1670.04 302.587 L1671.5 302.911 Q1676.03 303.883 1677.93 305.665 Q1679.85 307.425 1679.85 310.596 Q1679.85 314.207 1676.98 316.313 Q1674.13 318.42 1669.13 318.42 Q1667.05 318.42 1664.78 318.003 Q1662.54 317.61 1660.04 316.8 L1660.04 312.401 Q1662.4 313.628 1664.69 314.253 Q1666.98 314.855 1669.23 314.855 Q1672.24 314.855 1673.86 313.837 Q1675.48 312.795 1675.48 310.92 Q1675.48 309.184 1674.3 308.258 Q1673.14 307.332 1669.18 306.476 L1667.7 306.128 Q1663.74 305.295 1661.98 303.582 Q1660.22 301.846 1660.22 298.837 Q1660.22 295.179 1662.82 293.189 Q1665.41 291.198 1670.18 291.198 Q1672.54 291.198 1674.62 291.545 Q1676.7 291.892 1678.46 292.587 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1703.16 292.587 L1703.16 296.614 Q1701.36 295.689 1699.41 295.226 Q1697.47 294.763 1695.38 294.763 Q1692.21 294.763 1690.62 295.735 Q1689.04 296.707 1689.04 298.651 Q1689.04 300.133 1690.18 300.989 Q1691.31 301.823 1694.74 302.587 L1696.19 302.911 Q1700.73 303.883 1702.63 305.665 Q1704.55 307.425 1704.55 310.596 Q1704.55 314.207 1701.68 316.313 Q1698.83 318.42 1693.83 318.42 Q1691.75 318.42 1689.48 318.003 Q1687.24 317.61 1684.74 316.8 L1684.74 312.401 Q1687.1 313.628 1689.39 314.253 Q1691.68 314.855 1693.93 314.855 Q1696.94 314.855 1698.56 313.837 Q1700.18 312.795 1700.18 310.92 Q1700.18 309.184 1699 308.258 Q1697.84 307.332 1693.88 306.476 L1692.4 306.128 Q1688.44 305.295 1686.68 303.582 Q1684.92 301.846 1684.92 298.837 Q1684.92 295.179 1687.51 293.189 Q1690.11 291.198 1694.88 291.198 Q1697.24 291.198 1699.32 291.545 Q1701.4 291.892 1703.16 292.587 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1711.94 311.869 L1716.82 311.869 L1716.82 317.749 L1711.94 317.749 L1711.94 311.869 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1739.53 281.73 L1739.53 285.272 L1735.45 285.272 Q1733.16 285.272 1732.26 286.198 Q1731.38 287.124 1731.38 289.531 L1731.38 291.823 L1738.39 291.823 L1738.39 295.133 L1731.38 295.133 L1731.38 317.749 L1727.1 317.749 L1727.1 295.133 L1723.02 295.133 L1723.02 291.823 L1727.1 291.823 L1727.1 290.017 Q1727.1 285.689 1729.11 283.721 Q1731.13 281.73 1735.5 281.73 L1739.53 281.73 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1768.39 281.777 Q1765.29 287.101 1763.79 292.309 Q1762.28 297.517 1762.28 302.864 Q1762.28 308.212 1763.79 313.466 Q1765.31 318.698 1768.39 323.999 L1764.69 323.999 Q1761.22 318.559 1759.48 313.304 Q1757.77 308.05 1757.77 302.864 Q1757.77 297.702 1759.48 292.471 Q1761.19 287.24 1764.69 281.777 L1768.39 281.777 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1787.84 298.605 Q1784.69 298.605 1782.84 300.758 Q1781.01 302.911 1781.01 306.661 Q1781.01 310.388 1782.84 312.563 Q1784.69 314.716 1787.84 314.716 Q1790.99 314.716 1792.81 312.563 Q1794.67 310.388 1794.67 306.661 Q1794.67 302.911 1792.81 300.758 Q1790.99 298.605 1787.84 298.605 M1797.12 283.953 L1797.12 288.212 Q1795.36 287.378 1793.56 286.939 Q1791.77 286.499 1790.01 286.499 Q1785.38 286.499 1782.93 289.624 Q1780.5 292.749 1780.15 299.068 Q1781.52 297.054 1783.58 295.989 Q1785.64 294.902 1788.12 294.902 Q1793.32 294.902 1796.33 298.073 Q1799.37 301.221 1799.37 306.661 Q1799.37 311.985 1796.22 315.202 Q1793.07 318.42 1787.84 318.42 Q1781.84 318.42 1778.67 313.837 Q1775.5 309.23 1775.5 300.503 Q1775.5 292.309 1779.39 287.448 Q1783.28 282.564 1789.83 282.564 Q1791.59 282.564 1793.37 282.911 Q1795.18 283.258 1797.12 283.953 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1821.59 299.114 Q1824.94 299.832 1826.82 302.101 Q1828.72 304.369 1828.72 307.702 Q1828.72 312.818 1825.2 315.619 Q1821.68 318.42 1815.2 318.42 Q1813.02 318.42 1810.71 317.98 Q1808.42 317.563 1805.96 316.707 L1805.96 312.193 Q1807.91 313.327 1810.22 313.906 Q1812.54 314.485 1815.06 314.485 Q1819.46 314.485 1821.75 312.749 Q1824.06 311.013 1824.06 307.702 Q1824.06 304.647 1821.91 302.934 Q1819.78 301.198 1815.96 301.198 L1811.94 301.198 L1811.94 297.355 L1816.15 297.355 Q1819.6 297.355 1821.43 295.989 Q1823.25 294.601 1823.25 292.008 Q1823.25 289.346 1821.36 287.934 Q1819.48 286.499 1815.96 286.499 Q1814.04 286.499 1811.84 286.915 Q1809.64 287.332 1807 288.212 L1807 284.045 Q1809.67 283.304 1811.98 282.934 Q1814.32 282.564 1816.38 282.564 Q1821.7 282.564 1824.81 284.994 Q1827.91 287.402 1827.91 291.522 Q1827.91 294.392 1826.26 296.383 Q1824.62 298.351 1821.59 299.114 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1836.4 283.189 L1858.62 283.189 L1858.62 285.179 L1846.08 317.749 L1841.19 317.749 L1853 287.124 L1836.4 287.124 L1836.4 283.189 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1867.79 283.189 L1886.15 283.189 L1886.15 287.124 L1872.07 287.124 L1872.07 295.596 Q1873.09 295.249 1874.11 295.087 Q1875.13 294.902 1876.15 294.902 Q1881.93 294.902 1885.31 298.073 Q1888.69 301.244 1888.69 306.661 Q1888.69 312.239 1885.22 315.341 Q1881.75 318.42 1875.43 318.42 Q1873.25 318.42 1870.99 318.05 Q1868.74 317.679 1866.33 316.938 L1866.33 312.239 Q1868.42 313.374 1870.64 313.929 Q1872.86 314.485 1875.34 314.485 Q1879.34 314.485 1881.68 312.378 Q1884.02 310.272 1884.02 306.661 Q1884.02 303.05 1881.68 300.943 Q1879.34 298.837 1875.34 298.837 Q1873.46 298.837 1871.59 299.253 Q1869.74 299.67 1867.79 300.55 L1867.79 283.189 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1912.07 299.114 Q1915.43 299.832 1917.3 302.101 Q1919.2 304.369 1919.2 307.702 Q1919.2 312.818 1915.68 315.619 Q1912.17 318.42 1905.68 318.42 Q1903.51 318.42 1901.19 317.98 Q1898.9 317.563 1896.45 316.707 L1896.45 312.193 Q1898.39 313.327 1900.71 313.906 Q1903.02 314.485 1905.55 314.485 Q1909.94 314.485 1912.24 312.749 Q1914.55 311.013 1914.55 307.702 Q1914.55 304.647 1912.4 302.934 Q1910.27 301.198 1906.45 301.198 L1902.42 301.198 L1902.42 297.355 L1906.63 297.355 Q1910.08 297.355 1911.91 295.989 Q1913.74 294.601 1913.74 292.008 Q1913.74 289.346 1911.84 287.934 Q1909.97 286.499 1906.45 286.499 Q1904.53 286.499 1902.33 286.915 Q1900.13 287.332 1897.49 288.212 L1897.49 284.045 Q1900.15 283.304 1902.47 282.934 Q1904.8 282.564 1906.87 282.564 Q1912.19 282.564 1915.29 284.994 Q1918.39 287.402 1918.39 291.522 Q1918.39 294.392 1916.75 296.383 Q1915.11 298.351 1912.07 299.114 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1940.92 287.263 L1929.11 305.712 L1940.92 305.712 L1940.92 287.263 M1939.69 283.189 L1945.57 283.189 L1945.57 305.712 L1950.5 305.712 L1950.5 309.601 L1945.57 309.601 L1945.57 317.749 L1940.92 317.749 L1940.92 309.601 L1925.31 309.601 L1925.31 305.087 L1939.69 283.189 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1962.26 313.813 L1978.58 313.813 L1978.58 317.749 L1956.63 317.749 L1956.63 313.813 Q1959.3 311.059 1963.88 306.429 Q1968.49 301.776 1969.67 300.434 Q1971.91 297.911 1972.79 296.175 Q1973.69 294.415 1973.69 292.726 Q1973.69 289.971 1971.75 288.235 Q1969.83 286.499 1966.73 286.499 Q1964.53 286.499 1962.07 287.263 Q1959.64 288.027 1956.86 289.577 L1956.86 284.855 Q1959.69 283.721 1962.14 283.142 Q1964.6 282.564 1966.63 282.564 Q1972 282.564 1975.2 285.249 Q1978.39 287.934 1978.39 292.425 Q1978.39 294.554 1977.58 296.476 Q1976.8 298.374 1974.69 300.966 Q1974.11 301.638 1971.01 304.855 Q1967.91 308.05 1962.26 313.813 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1989.2 313.813 L1996.84 313.813 L1996.84 287.448 L1988.53 289.115 L1988.53 284.855 L1996.8 283.189 L2001.47 283.189 L2001.47 313.813 L2009.11 313.813 L2009.11 317.749 L1989.2 317.749 L1989.2 313.813 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M2019.36 313.813 L2027 313.813 L2027 287.448 L2018.69 289.115 L2018.69 284.855 L2026.96 283.189 L2031.63 283.189 L2031.63 313.813 L2039.27 313.813 L2039.27 317.749 L2019.36 317.749 L2019.36 313.813 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M2047.44 281.777 L2051.15 281.777 Q2054.62 287.24 2056.33 292.471 Q2058.07 297.702 2058.07 302.864 Q2058.07 308.05 2056.33 313.304 Q2054.62 318.559 2051.15 323.999 L2047.44 323.999 Q2050.52 318.698 2052.03 313.466 Q2053.55 308.212 2053.55 302.864 Q2053.55 297.517 2052.03 292.309 Q2050.52 287.101 2047.44 281.777 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /></svg>



 After the plots are created, the FMU is unloaded.


```julia
fmiUnload(realFMU)
```

### SimpleFMU

The following lines load and instantiate the *simpleFMU* from *FMIZoo.jl*. 


```julia
simpleFMU = fmiLoad("SpringPendulum1D", "Dymola", "2022x")
fmiInstantiate!(simpleFMU; loggingOn=false)
fmiInfo(simpleFMU)
```

    ┌ Info: fmi2Unzip(...): Successfully unzipped 29 files at `C:\Users\thummeto\AppData\Local\Temp\fmijl_Fe7xyu\SpringPendulum1D`.
    └ @ FMIImport C:\Users\thummeto\Documents\FMIImport.jl\src\FMI2_ext.jl:77


    #################### Begin information for FMU ####################
    	Model name:			SpringPendulum1D
    	FMI-Version:			2.0
    	GUID:				{a3e886fa-675c-4308-8e91-3490e598ba11}
    	Generation tool:		Dymola Version 2022x (64-bit), 2021-10-08
    	Generation time:		2022-03-03T15:08:57Z
    	Var. naming conv.:		structured
    	Event indicators:		0
    	Inputs:				0
    	Outputs:			0
    	States:				2
    		33554432 ["mass.s"]
    		33554433 ["mass.v"]
    	Supports Co-Simulation:		true
    		Model identifier:	SpringPendulum1D
    		Get/Set State:		true
    		Serialize State:	true
    		Dir. Derivatives:	true
    		Var. com. steps:	true
    		Input interpol.:	true
    		Max order out. der.:	1
    	Supports Model-Exchange:	true
    		Model identifier:	SpringPendulum1D
    		Get/Set State:		true
    		Serialize State:	true
    		Dir. Derivatives:	true
    ##################### End information for FMU #####################


    ┌ Info: fmi2Load(...): FMU resources location is `file:///C:/Users/thummeto/AppData/Local/Temp/fmijl_Fe7xyu/SpringPendulum1D/resources`
    └ @ FMIImport C:\Users\thummeto\Documents\FMIImport.jl\src\FMI2_ext.jl:192
    ┌ Info: fmi2Load(...): FMU supports both CS and ME, using CS as default if nothing specified.
    └ @ FMIImport C:\Users\thummeto\Documents\FMIImport.jl\src\FMI2_ext.jl:195


The differences between both systems can be clearly seen from the plots in the subchapters. In the plot for the *realFMU* it can be seen that the oscillation continues to decrease due to the effect of the friction. If you would simulate long enough, the oscillation would come to a standstill in a certain time. The oscillation in the *simpleFMU* behaves differently, since the friction was not taken into account here. The oscillation in this model would continue to infinity with the same oscillation amplitude. From this observation the desire of an improvement of this model arises.     


In the following two subsections, the *simpleFMU* is simulated twice with different initial states to show what effect the choice of initial states has.

#### Default initial states

Similar to the simulation of the *realFMU*, the *simpleFMU* is also simulated with the default values for the position and velocity of the mass and then plotted. There is one difference, however, as another state representing a fixed displacement is set. In addition, the last variable is also removed from the varibals to be plotted.


```julia
states = ["mass_s0", "mass_v0", "fixed.s0"]
displacement = 0.1
xSimple₀ = vcat(x₀, displacement)
vrs = vrs[1:end-1]

simpleSimData = simulate(simpleFMU, states, xSimple₀, vrs, tStart, tStop, tSave)
fmiPlot(simpleFMU, vrs, simpleSimData)
```

    ┌ Warning: fmi2Reset(_): Should be only called in FMU state `modelSetableFMUstate`.
    └ @ FMIImport C:\Users\thummeto\Documents\FMIImport.jl\src\FMI2_c.jl:219



<?xml version="1.0" encoding="utf-8"?>
<svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" width="600" height="400" viewBox="0 0 2400 1600">
<defs>
  <clipPath id="clip120">
    <rect x="0" y="0" width="2400" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip120)" d="
M0 1600 L2400 1600 L2400 0 L0 0  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip121">
    <rect x="480" y="0" width="1681" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip120)" d="
M193.936 1423.18 L2352.76 1423.18 L2352.76 47.2441 L193.936 47.2441  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip122">
    <rect x="193" y="47" width="2160" height="1377"/>
  </clipPath>
</defs>
<polyline clip-path="url(#clip122)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  255.035,1423.18 255.035,47.2441 
  "/>
<polyline clip-path="url(#clip122)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  764.19,1423.18 764.19,47.2441 
  "/>
<polyline clip-path="url(#clip122)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1273.35,1423.18 1273.35,47.2441 
  "/>
<polyline clip-path="url(#clip122)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1782.5,1423.18 1782.5,47.2441 
  "/>
<polyline clip-path="url(#clip122)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  2291.66,1423.18 2291.66,47.2441 
  "/>
<polyline clip-path="url(#clip120)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  193.936,1423.18 2352.76,1423.18 
  "/>
<polyline clip-path="url(#clip120)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  255.035,1423.18 255.035,1404.28 
  "/>
<polyline clip-path="url(#clip120)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  764.19,1423.18 764.19,1404.28 
  "/>
<polyline clip-path="url(#clip120)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1273.35,1423.18 1273.35,1404.28 
  "/>
<polyline clip-path="url(#clip120)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1782.5,1423.18 1782.5,1404.28 
  "/>
<polyline clip-path="url(#clip120)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  2291.66,1423.18 2291.66,1404.28 
  "/>
<path clip-path="url(#clip120)" d="M255.035 1454.1 Q251.424 1454.1 249.595 1457.66 Q247.789 1461.2 247.789 1468.33 Q247.789 1475.44 249.595 1479.01 Q251.424 1482.55 255.035 1482.55 Q258.669 1482.55 260.474 1479.01 Q262.303 1475.44 262.303 1468.33 Q262.303 1461.2 260.474 1457.66 Q258.669 1454.1 255.035 1454.1 M255.035 1450.39 Q260.845 1450.39 263.9 1455 Q266.979 1459.58 266.979 1468.33 Q266.979 1477.06 263.9 1481.67 Q260.845 1486.25 255.035 1486.25 Q249.225 1486.25 246.146 1481.67 Q243.09 1477.06 243.09 1468.33 Q243.09 1459.58 246.146 1455 Q249.225 1450.39 255.035 1450.39 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M754.572 1481.64 L762.211 1481.64 L762.211 1455.28 L753.901 1456.95 L753.901 1452.69 L762.165 1451.02 L766.841 1451.02 L766.841 1481.64 L774.48 1481.64 L774.48 1485.58 L754.572 1485.58 L754.572 1481.64 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M1268 1481.64 L1284.32 1481.64 L1284.32 1485.58 L1262.37 1485.58 L1262.37 1481.64 Q1265.04 1478.89 1269.62 1474.26 Q1274.23 1469.61 1275.41 1468.27 Q1277.65 1465.74 1278.53 1464.01 Q1279.43 1462.25 1279.43 1460.56 Q1279.43 1457.8 1277.49 1456.07 Q1275.57 1454.33 1272.47 1454.33 Q1270.27 1454.33 1267.81 1455.09 Q1265.38 1455.86 1262.61 1457.41 L1262.61 1452.69 Q1265.43 1451.55 1267.88 1450.97 Q1270.34 1450.39 1272.37 1450.39 Q1277.74 1450.39 1280.94 1453.08 Q1284.13 1455.77 1284.13 1460.26 Q1284.13 1462.39 1283.32 1464.31 Q1282.54 1466.2 1280.43 1468.8 Q1279.85 1469.47 1276.75 1472.69 Q1273.65 1475.88 1268 1481.64 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M1786.75 1466.95 Q1790.11 1467.66 1791.98 1469.93 Q1793.88 1472.2 1793.88 1475.53 Q1793.88 1480.65 1790.36 1483.45 Q1786.84 1486.25 1780.36 1486.25 Q1778.18 1486.25 1775.87 1485.81 Q1773.58 1485.39 1771.12 1484.54 L1771.12 1480.02 Q1773.07 1481.16 1775.38 1481.74 Q1777.7 1482.32 1780.22 1482.32 Q1784.62 1482.32 1786.91 1480.58 Q1789.23 1478.84 1789.23 1475.53 Q1789.23 1472.48 1787.07 1470.77 Q1784.94 1469.03 1781.12 1469.03 L1777.1 1469.03 L1777.1 1465.19 L1781.31 1465.19 Q1784.76 1465.19 1786.59 1463.82 Q1788.42 1462.43 1788.42 1459.84 Q1788.42 1457.18 1786.52 1455.77 Q1784.64 1454.33 1781.12 1454.33 Q1779.2 1454.33 1777 1454.75 Q1774.8 1455.16 1772.17 1456.04 L1772.17 1451.88 Q1774.83 1451.14 1777.14 1450.77 Q1779.48 1450.39 1781.54 1450.39 Q1786.86 1450.39 1789.97 1452.83 Q1793.07 1455.23 1793.07 1459.35 Q1793.07 1462.22 1791.43 1464.21 Q1789.78 1466.18 1786.75 1466.95 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M2294.67 1455.09 L2282.86 1473.54 L2294.67 1473.54 L2294.67 1455.09 M2293.44 1451.02 L2299.32 1451.02 L2299.32 1473.54 L2304.25 1473.54 L2304.25 1477.43 L2299.32 1477.43 L2299.32 1485.58 L2294.67 1485.58 L2294.67 1477.43 L2279.06 1477.43 L2279.06 1472.92 L2293.44 1451.02 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M1221.66 1522.27 L1221.66 1532.4 L1233.72 1532.4 L1233.72 1536.95 L1221.66 1536.95 L1221.66 1556.3 Q1221.66 1560.66 1222.83 1561.9 Q1224.04 1563.14 1227.7 1563.14 L1233.72 1563.14 L1233.72 1568.04 L1227.7 1568.04 Q1220.92 1568.04 1218.35 1565.53 Q1215.77 1562.98 1215.77 1556.3 L1215.77 1536.95 L1211.47 1536.95 L1211.47 1532.4 L1215.77 1532.4 L1215.77 1522.27 L1221.66 1522.27 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M1261.6 1518.52 L1275.1 1518.52 L1275.1 1523.07 L1267.46 1523.07 L1267.46 1572.09 L1275.1 1572.09 L1275.1 1576.64 L1261.6 1576.64 L1261.6 1518.52 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M1310.3 1533.45 L1310.3 1538.98 Q1307.82 1537.71 1305.14 1537.07 Q1302.47 1536.44 1299.6 1536.44 Q1295.24 1536.44 1293.05 1537.77 Q1290.88 1539.11 1290.88 1541.79 Q1290.88 1543.82 1292.44 1545 Q1294 1546.15 1298.71 1547.2 L1300.72 1547.64 Q1306.96 1548.98 1309.57 1551.43 Q1312.21 1553.85 1312.21 1558.21 Q1312.21 1563.17 1308.26 1566.07 Q1304.35 1568.97 1297.47 1568.97 Q1294.61 1568.97 1291.49 1568.39 Q1288.4 1567.85 1284.96 1566.74 L1284.96 1560.69 Q1288.21 1562.38 1291.36 1563.24 Q1294.51 1564.07 1297.6 1564.07 Q1301.74 1564.07 1303.97 1562.66 Q1306.19 1561.23 1306.19 1558.65 Q1306.19 1556.27 1304.57 1554.99 Q1302.98 1553.72 1297.54 1552.54 L1295.5 1552.07 Q1290.06 1550.92 1287.64 1548.56 Q1285.22 1546.18 1285.22 1542.04 Q1285.22 1537.01 1288.78 1534.27 Q1292.35 1531.54 1298.9 1531.54 Q1302.15 1531.54 1305.02 1532.01 Q1307.88 1532.49 1310.3 1533.45 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M1335.22 1518.52 L1335.22 1576.64 L1321.73 1576.64 L1321.73 1572.09 L1329.33 1572.09 L1329.33 1523.07 L1321.73 1523.07 L1321.73 1518.52 L1335.22 1518.52 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip122)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  193.936,1276.04 2352.76,1276.04 
  "/>
<polyline clip-path="url(#clip122)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  193.936,1005.61 2352.76,1005.61 
  "/>
<polyline clip-path="url(#clip122)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  193.936,735.195 2352.76,735.195 
  "/>
<polyline clip-path="url(#clip122)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  193.936,464.774 2352.76,464.774 
  "/>
<polyline clip-path="url(#clip122)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  193.936,194.354 2352.76,194.354 
  "/>
<polyline clip-path="url(#clip120)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  193.936,1423.18 193.936,47.2441 
  "/>
<polyline clip-path="url(#clip120)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  193.936,1276.04 212.834,1276.04 
  "/>
<polyline clip-path="url(#clip120)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  193.936,1005.61 212.834,1005.61 
  "/>
<polyline clip-path="url(#clip120)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  193.936,735.195 212.834,735.195 
  "/>
<polyline clip-path="url(#clip120)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  193.936,464.774 212.834,464.774 
  "/>
<polyline clip-path="url(#clip120)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  193.936,194.354 212.834,194.354 
  "/>
<path clip-path="url(#clip120)" d="M50.9921 1276.49 L80.6679 1276.49 L80.6679 1280.42 L50.9921 1280.42 L50.9921 1276.49 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M90.8067 1258.76 L109.163 1258.76 L109.163 1262.69 L95.0891 1262.69 L95.0891 1271.16 Q96.1076 1270.82 97.1261 1270.65 Q98.1447 1270.47 99.1632 1270.47 Q104.95 1270.47 108.33 1273.64 Q111.709 1276.81 111.709 1282.23 Q111.709 1287.81 108.237 1290.91 Q104.765 1293.99 98.4456 1293.99 Q96.2697 1293.99 94.0012 1293.62 Q91.7558 1293.25 89.3484 1292.51 L89.3484 1287.81 Q91.4317 1288.94 93.6539 1289.5 Q95.8761 1290.05 98.353 1290.05 Q102.358 1290.05 104.696 1287.94 Q107.033 1285.84 107.033 1282.23 Q107.033 1278.62 104.696 1276.51 Q102.358 1274.4 98.353 1274.4 Q96.478 1274.4 94.603 1274.82 Q92.7512 1275.24 90.8067 1276.12 L90.8067 1258.76 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M120.922 1287.44 L125.807 1287.44 L125.807 1293.32 L120.922 1293.32 L120.922 1287.44 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M145.992 1261.83 Q142.381 1261.83 140.552 1265.4 Q138.746 1268.94 138.746 1276.07 Q138.746 1283.18 140.552 1286.74 Q142.381 1290.28 145.992 1290.28 Q149.626 1290.28 151.431 1286.74 Q153.26 1283.18 153.26 1276.07 Q153.26 1268.94 151.431 1265.4 Q149.626 1261.83 145.992 1261.83 M145.992 1258.13 Q151.802 1258.13 154.857 1262.74 Q157.936 1267.32 157.936 1276.07 Q157.936 1284.8 154.857 1289.4 Q151.802 1293.99 145.992 1293.99 Q140.181 1293.99 137.103 1289.4 Q134.047 1284.8 134.047 1276.07 Q134.047 1267.32 137.103 1262.74 Q140.181 1258.13 145.992 1258.13 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M51.9875 1006.07 L81.6633 1006.07 L81.6633 1010 L51.9875 1010 L51.9875 1006.07 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M95.7836 1018.96 L112.103 1018.96 L112.103 1022.89 L90.1586 1022.89 L90.1586 1018.96 Q92.8206 1016.21 97.4039 1011.58 Q102.01 1006.92 103.191 1005.58 Q105.436 1003.06 106.316 1001.32 Q107.219 999.562 107.219 997.872 Q107.219 995.117 105.274 993.381 Q103.353 991.645 100.251 991.645 Q98.0521 991.645 95.5984 992.409 Q93.1678 993.173 90.3901 994.724 L90.3901 990.002 Q93.2141 988.867 95.6678 988.289 Q98.1215 987.71 100.159 987.71 Q105.529 987.71 108.723 990.395 Q111.918 993.08 111.918 997.571 Q111.918 999.701 111.108 1001.62 Q110.321 1003.52 108.214 1006.11 Q107.635 1006.78 104.534 1010 Q101.432 1013.2 95.7836 1018.96 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M121.918 1017.02 L126.802 1017.02 L126.802 1022.89 L121.918 1022.89 L121.918 1017.02 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M137.033 988.335 L155.39 988.335 L155.39 992.27 L141.316 992.27 L141.316 1000.74 Q142.334 1000.4 143.353 1000.23 Q144.371 1000.05 145.39 1000.05 Q151.177 1000.05 154.556 1003.22 Q157.936 1006.39 157.936 1011.81 Q157.936 1017.39 154.464 1020.49 Q150.992 1023.57 144.672 1023.57 Q142.496 1023.57 140.228 1023.2 Q137.982 1022.83 135.575 1022.08 L135.575 1017.39 Q137.658 1018.52 139.881 1019.08 Q142.103 1019.63 144.58 1019.63 Q148.584 1019.63 150.922 1017.52 Q153.26 1015.42 153.26 1011.81 Q153.26 1008.2 150.922 1006.09 Q148.584 1003.98 144.58 1003.98 Q142.705 1003.98 140.83 1004.4 Q138.978 1004.82 137.033 1005.7 L137.033 988.335 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M100.76 720.993 Q97.1493 720.993 95.3206 724.558 Q93.515 728.1 93.515 735.229 Q93.515 742.336 95.3206 745.9 Q97.1493 749.442 100.76 749.442 Q104.395 749.442 106.2 745.9 Q108.029 742.336 108.029 735.229 Q108.029 728.1 106.2 724.558 Q104.395 720.993 100.76 720.993 M100.76 717.29 Q106.571 717.29 109.626 721.896 Q112.705 726.479 112.705 735.229 Q112.705 743.956 109.626 748.563 Q106.571 753.146 100.76 753.146 Q94.9502 753.146 91.8715 748.563 Q88.816 743.956 88.816 735.229 Q88.816 726.479 91.8715 721.896 Q94.9502 717.29 100.76 717.29 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M120.922 746.595 L125.807 746.595 L125.807 752.475 L120.922 752.475 L120.922 746.595 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M145.992 720.993 Q142.381 720.993 140.552 724.558 Q138.746 728.1 138.746 735.229 Q138.746 742.336 140.552 745.9 Q142.381 749.442 145.992 749.442 Q149.626 749.442 151.431 745.9 Q153.26 742.336 153.26 735.229 Q153.26 728.1 151.431 724.558 Q149.626 720.993 145.992 720.993 M145.992 717.29 Q151.802 717.29 154.857 721.896 Q157.936 726.479 157.936 735.229 Q157.936 743.956 154.857 748.563 Q151.802 753.146 145.992 753.146 Q140.181 753.146 137.103 748.563 Q134.047 743.956 134.047 735.229 Q134.047 726.479 137.103 721.896 Q140.181 717.29 145.992 717.29 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M95.7836 478.119 L112.103 478.119 L112.103 482.054 L90.1586 482.054 L90.1586 478.119 Q92.8206 475.364 97.4039 470.735 Q102.01 466.082 103.191 464.739 Q105.436 462.216 106.316 460.48 Q107.219 458.721 107.219 457.031 Q107.219 454.277 105.274 452.54 Q103.353 450.804 100.251 450.804 Q98.0521 450.804 95.5984 451.568 Q93.1678 452.332 90.3901 453.883 L90.3901 449.161 Q93.2141 448.027 95.6678 447.448 Q98.1215 446.869 100.159 446.869 Q105.529 446.869 108.723 449.554 Q111.918 452.24 111.918 456.73 Q111.918 458.86 111.108 460.781 Q110.321 462.679 108.214 465.272 Q107.635 465.943 104.534 469.161 Q101.432 472.355 95.7836 478.119 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M121.918 476.175 L126.802 476.175 L126.802 482.054 L121.918 482.054 L121.918 476.175 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M137.033 447.494 L155.39 447.494 L155.39 451.429 L141.316 451.429 L141.316 459.902 Q142.334 459.554 143.353 459.392 Q144.371 459.207 145.39 459.207 Q151.177 459.207 154.556 462.378 Q157.936 465.55 157.936 470.966 Q157.936 476.545 154.464 479.647 Q150.992 482.725 144.672 482.725 Q142.496 482.725 140.228 482.355 Q137.982 481.985 135.575 481.244 L135.575 476.545 Q137.658 477.679 139.881 478.235 Q142.103 478.79 144.58 478.79 Q148.584 478.79 150.922 476.684 Q153.26 474.577 153.26 470.966 Q153.26 467.355 150.922 465.249 Q148.584 463.142 144.58 463.142 Q142.705 463.142 140.83 463.559 Q138.978 463.976 137.033 464.855 L137.033 447.494 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M90.8067 177.074 L109.163 177.074 L109.163 181.009 L95.0891 181.009 L95.0891 189.481 Q96.1076 189.134 97.1261 188.972 Q98.1447 188.787 99.1632 188.787 Q104.95 188.787 108.33 191.958 Q111.709 195.129 111.709 200.546 Q111.709 206.125 108.237 209.226 Q104.765 212.305 98.4456 212.305 Q96.2697 212.305 94.0012 211.935 Q91.7558 211.564 89.3484 210.824 L89.3484 206.125 Q91.4317 207.259 93.6539 207.814 Q95.8761 208.37 98.353 208.37 Q102.358 208.37 104.696 206.263 Q107.033 204.157 107.033 200.546 Q107.033 196.935 104.696 194.828 Q102.358 192.722 98.353 192.722 Q96.478 192.722 94.603 193.139 Q92.7512 193.555 90.8067 194.435 L90.8067 177.074 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M120.922 205.754 L125.807 205.754 L125.807 211.634 L120.922 211.634 L120.922 205.754 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M145.992 180.153 Q142.381 180.153 140.552 183.717 Q138.746 187.259 138.746 194.389 Q138.746 201.495 140.552 205.06 Q142.381 208.601 145.992 208.601 Q149.626 208.601 151.431 205.06 Q153.26 201.495 153.26 194.389 Q153.26 187.259 151.431 183.717 Q149.626 180.153 145.992 180.153 M145.992 176.449 Q151.802 176.449 154.857 181.055 Q157.936 185.639 157.936 194.389 Q157.936 203.115 154.857 207.722 Q151.802 212.305 145.992 212.305 Q140.181 212.305 137.103 207.722 Q134.047 203.115 134.047 194.389 Q134.047 185.639 137.103 181.055 Q140.181 176.449 145.992 176.449 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip122)" style="stroke:#009af9; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  255.035,681.11 260.126,681.078 265.218,680.981 270.309,680.818 275.401,680.592 280.492,680.301 285.584,679.946 290.676,679.527 295.767,679.045 300.859,678.5 
  305.95,677.893 311.042,677.224 316.133,676.494 321.225,675.704 326.316,674.855 331.408,673.947 336.5,672.981 341.591,671.958 346.683,670.88 351.774,669.748 
  356.866,668.561 361.957,667.322 367.049,666.032 372.14,664.692 377.232,663.304 382.324,661.868 387.415,660.387 392.507,658.861 397.598,657.293 402.69,655.683 
  407.781,654.034 412.873,652.348 417.964,650.625 423.056,648.868 428.148,647.078 433.239,645.257 438.331,643.407 443.422,641.53 448.514,639.628 453.605,637.702 
  458.697,635.755 463.788,633.788 468.88,631.803 473.972,629.803 479.063,627.79 484.155,625.765 489.246,623.73 494.338,621.688 499.429,619.64 504.521,617.589 
  509.612,615.536 514.704,613.485 519.796,611.435 524.887,609.391 529.979,607.354 535.07,605.325 540.162,603.307 545.253,601.302 550.345,599.312 555.437,597.339 
  560.528,595.385 565.62,593.451 570.711,591.541 575.803,589.655 580.894,587.795 585.986,585.964 591.077,584.164 596.169,582.395 601.261,580.66 606.352,578.961 
  611.444,577.298 616.535,575.675 621.627,574.093 626.718,572.552 631.81,571.055 636.901,569.603 641.993,568.198 647.085,566.841 652.176,565.533 657.268,564.276 
  662.359,563.071 667.451,561.919 672.542,560.821 677.634,559.779 682.725,558.793 687.817,557.865 692.909,556.995 698,556.184 703.092,555.433 708.183,554.743 
  713.275,554.114 718.366,553.547 723.458,553.044 728.549,552.603 733.641,552.226 738.733,551.913 743.824,551.664 748.916,551.479 754.007,551.36 759.099,551.305 
  764.19,551.315 769.282,551.39 774.373,551.53 779.465,551.735 784.557,552.004 789.648,552.337 794.74,552.734 799.831,553.195 804.923,553.719 810.014,554.305 
  815.106,554.953 820.197,555.662 825.289,556.432 830.381,557.262 835.472,558.15 840.564,559.097 845.655,560.101 850.747,561.161 855.838,562.276 860.93,563.445 
  866.021,564.666 871.113,565.94 876.205,567.263 881.296,568.635 886.388,570.055 891.479,571.522 896.571,573.032 901.662,574.586 906.754,576.182 911.845,577.817 
  916.937,579.491 922.029,581.202 927.12,582.948 932.212,584.727 937.303,586.537 942.395,588.377 947.486,590.245 952.578,592.139 957.669,594.057 962.761,595.997 
  967.853,597.957 972.944,599.936 978.036,601.931 983.127,603.94 988.219,605.962 993.31,607.993 998.402,610.033 1003.49,612.079 1008.59,614.13 1013.68,616.182 
  1018.77,618.234 1023.86,620.285 1028.95,622.331 1034.04,624.371 1039.13,626.403 1044.23,628.425 1049.32,630.435 1054.41,632.43 1059.5,634.409 1064.59,636.37 
  1069.68,638.311 1074.78,640.23 1079.87,642.124 1084.96,643.993 1090.05,645.834 1095.14,647.645 1100.23,649.425 1105.32,651.171 1110.42,652.883 1115.51,654.558 
  1120.6,656.194 1125.69,657.791 1130.78,659.345 1135.87,660.857 1140.97,662.324 1146.06,663.745 1151.15,665.118 1156.24,666.443 1161.33,667.717 1166.42,668.939 
  1171.51,670.109 1176.61,671.225 1181.7,672.286 1186.79,673.291 1191.88,674.239 1196.97,675.128 1202.06,675.959 1207.16,676.73 1212.25,677.441 1217.34,678.09 
  1222.43,678.677 1227.52,679.202 1232.61,679.664 1237.71,680.062 1242.8,680.397 1247.89,680.668 1252.98,680.874 1258.07,681.015 1263.16,681.091 1268.25,681.103 
  1273.35,681.049 1278.44,680.931 1283.53,680.748 1288.62,680.501 1293.71,680.19 1298.8,679.814 1303.9,679.374 1308.99,678.872 1314.08,678.307 1319.17,677.679 
  1324.26,676.991 1329.35,676.242 1334.44,675.432 1339.54,674.563 1344.63,673.636 1349.72,672.651 1354.81,671.611 1359.9,670.514 1364.99,669.364 1370.09,668.16 
  1375.18,666.904 1380.27,665.598 1385.36,664.242 1390.45,662.838 1395.54,661.387 1400.63,659.892 1405.73,658.353 1410.82,656.771 1415.91,655.149 1421,653.488 
  1426.09,651.79 1431.18,650.056 1436.28,648.289 1441.37,646.489 1446.46,644.659 1451.55,642.801 1456.64,640.916 1461.73,639.007 1466.83,637.074 1471.92,635.121 
  1477.01,633.149 1482.1,631.16 1487.19,629.156 1492.28,627.139 1497.37,625.111 1502.47,623.074 1507.56,621.031 1512.65,618.982 1517.74,616.931 1522.83,614.879 
  1527.92,612.829 1533.02,610.782 1538.11,608.74 1543.2,606.706 1548.29,604.681 1553.38,602.668 1558.47,600.668 1563.56,598.684 1568.66,596.718 1573.75,594.771 
  1578.84,592.845 1583.93,590.943 1589.02,589.067 1594.11,587.217 1599.21,585.396 1604.3,583.606 1609.39,581.849 1614.48,580.126 1619.57,578.44 1624.66,576.791 
  1629.75,575.181 1634.85,573.613 1639.94,572.087 1645.03,570.605 1650.12,569.169 1655.21,567.78 1660.3,566.439 1665.4,565.148 1670.49,563.909 1675.58,562.721 
  1680.67,561.587 1685.76,560.508 1690.85,559.484 1695.95,558.518 1701.04,557.608 1706.13,556.758 1711.22,555.967 1716.31,555.236 1721.4,554.566 1726.49,553.958 
  1731.59,553.412 1736.68,552.929 1741.77,552.509 1746.86,552.153 1751.95,551.861 1757.04,551.633 1762.14,551.47 1767.23,551.371 1772.32,551.338 1777.41,551.369 
  1782.5,551.466 1787.59,551.626 1792.68,551.852 1797.78,552.142 1802.87,552.495 1807.96,552.913 1813.05,553.394 1818.14,553.938 1823.23,554.544 1828.33,555.212 
  1833.42,555.94 1838.51,556.729 1843.6,557.578 1848.69,558.485 1853.78,559.45 1858.87,560.471 1863.97,561.549 1869.06,562.68 1874.15,563.866 1879.24,565.103 
  1884.33,566.392 1889.42,567.731 1894.52,569.118 1899.61,570.553 1904.7,572.033 1909.79,573.557 1914.88,575.123 1919.97,576.731 1925.07,578.378 1930.16,580.063 
  1935.25,581.785 1940.34,583.54 1945.43,585.329 1950.52,587.148 1955.61,588.996 1960.71,590.871 1965.8,592.772 1970.89,594.696 1975.98,596.642 1981.07,598.607 
  1986.16,600.59 1991.26,602.588 1996.35,604.6 2001.44,606.624 2006.53,608.657 2011.62,610.698 2016.71,612.744 2021.8,614.794 2026.9,616.845 2031.99,618.895 
  2037.08,620.943 2042.17,622.986 2047.26,625.022 2052.35,627.049 2057.45,629.066 2062.54,631.069 2067.63,633.058 2072.72,635.03 2077.81,636.983 2082.9,638.915 
  2087.99,640.824 2093.09,642.709 2098.18,644.568 2103.27,646.398 2108.36,648.198 2113.45,649.965 2118.54,651.699 2123.64,653.398 2128.73,655.059 2133.82,656.682 
  2138.91,658.263 2144,659.803 2149.09,661.3 2154.19,662.751 2159.28,664.155 2164.37,665.512 2169.46,666.819 2174.55,668.076 2179.64,669.281 2184.73,670.433 
  2189.83,671.53 2194.92,672.573 2200.01,673.558 2205.1,674.487 2210.19,675.357 2215.28,676.168 2220.38,676.919 2225.47,677.609 2230.56,678.238 2235.65,678.805 
  2240.74,679.309 2245.83,679.75 2250.92,680.128 2256.02,680.442 2261.11,680.691 2266.2,680.876 2271.29,680.997 2276.38,681.052 2281.47,681.043 2286.57,680.969 
  2291.66,680.83 
  "/>
<polyline clip-path="url(#clip122)" style="stroke:#e26f46; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  255.035,735.195 260.126,728.706 265.218,722.224 270.309,715.754 275.401,709.304 280.492,702.88 285.584,696.488 290.676,690.135 295.767,683.826 300.859,677.569 
  305.95,671.37 311.042,665.234 316.133,659.169 321.225,653.179 326.316,647.271 331.408,641.451 336.5,635.725 341.591,630.098 346.683,624.576 351.774,619.164 
  356.866,613.868 361.957,608.694 367.049,603.646 372.14,598.73 377.232,593.95 382.324,589.312 387.415,584.819 392.507,580.477 397.598,576.29 402.69,572.261 
  407.781,568.396 412.873,564.697 417.964,561.169 423.056,557.815 428.148,554.638 433.239,551.642 438.331,548.829 443.422,546.203 448.514,543.766 453.605,541.52 
  458.697,539.468 463.788,537.611 468.88,535.952 473.972,534.493 479.063,533.234 484.155,532.177 489.246,531.323 494.338,530.673 499.429,530.227 504.521,529.986 
  509.612,529.951 514.704,530.121 519.796,530.496 524.887,531.076 529.979,531.859 535.07,532.846 540.162,534.036 545.253,535.426 550.345,537.017 555.437,538.805 
  560.528,540.79 565.62,542.97 570.711,545.341 575.803,547.903 580.894,550.652 585.986,553.585 591.077,556.7 596.169,559.993 601.261,563.462 606.352,567.102 
  611.444,570.911 616.535,574.883 621.627,579.016 626.718,583.306 631.81,587.747 636.901,592.335 641.993,597.067 647.085,601.936 652.176,606.939 657.268,612.07 
  662.359,617.325 667.451,622.697 672.542,628.181 677.634,633.773 682.725,639.466 687.817,645.255 692.909,651.134 698,657.096 703.092,663.137 708.183,669.25 
  713.275,675.429 718.366,681.668 723.458,687.96 728.549,694.299 733.641,700.68 738.733,707.095 743.824,713.537 748.916,720.002 754.007,726.482 759.099,732.971 
  764.19,739.461 769.282,745.948 774.373,752.424 779.465,758.882 784.557,765.317 789.648,771.722 794.74,778.09 799.831,784.415 804.923,790.691 810.014,796.912 
  815.106,803.071 820.197,809.162 825.289,815.179 830.381,821.116 835.472,826.968 840.564,832.728 845.655,838.391 850.747,843.95 855.838,849.4 860.93,854.736 
  866.021,859.953 871.113,865.045 876.205,870.007 881.296,874.833 886.388,879.52 891.479,884.062 896.571,888.456 901.662,892.696 906.754,896.779 911.845,900.7 
  916.937,904.455 922.029,908.042 927.12,911.455 932.212,914.693 937.303,917.75 942.395,920.626 947.486,923.316 952.578,925.818 957.669,928.129 962.761,930.247 
  967.853,932.171 972.944,933.897 978.036,935.425 983.127,936.753 988.219,937.879 993.31,938.802 998.402,939.522 1003.49,940.037 1008.59,940.348 1013.68,940.453 
  1018.77,940.353 1023.86,940.047 1028.95,939.537 1034.04,938.822 1039.13,937.904 1044.23,936.783 1049.32,935.46 1054.41,933.938 1059.5,932.216 1064.59,930.296 
  1069.68,928.182 1074.78,925.875 1079.87,923.376 1084.96,920.69 1090.05,917.819 1095.14,914.765 1100.23,911.531 1105.32,908.121 1110.42,904.537 1115.51,900.785 
  1120.6,896.866 1125.69,892.787 1130.78,888.55 1135.87,884.16 1140.97,879.62 1146.06,874.936 1151.15,870.112 1156.24,865.154 1161.33,860.065 1166.42,854.852 
  1171.51,849.52 1176.61,844.073 1181.7,838.517 1186.79,832.858 1191.88,827.101 1196.97,821.252 1202.06,815.318 1207.16,809.303 1212.25,803.215 1217.34,797.058 
  1222.43,790.84 1227.52,784.565 1232.61,778.242 1237.71,771.875 1242.8,765.472 1247.89,759.039 1252.98,752.582 1258.07,746.107 1263.16,739.622 1268.25,733.132 
  1273.35,726.644 1278.44,720.165 1283.53,713.701 1288.62,707.259 1293.71,700.845 1298.8,694.465 1303.9,688.125 1308.99,681.833 1314.08,675.594 1319.17,669.415 
  1324.26,663.302 1329.35,657.261 1334.44,651.297 1339.54,645.418 1344.63,639.628 1349.72,633.934 1354.81,628.342 1359.9,622.856 1364.99,617.483 1370.09,612.227 
  1375.18,607.095 1380.27,602.09 1385.36,597.219 1390.45,592.486 1395.54,587.895 1400.63,583.452 1405.73,579.161 1410.82,575.025 1415.91,571.05 1421,567.24 
  1426.09,563.597 1431.18,560.126 1436.28,556.83 1441.37,553.712 1446.46,550.775 1451.55,548.024 1456.64,545.459 1461.73,543.085 1466.83,540.902 1471.92,538.914 
  1477.01,537.122 1482.1,535.528 1487.19,534.133 1492.28,532.94 1497.37,531.949 1502.47,531.162 1507.56,530.578 1512.65,530.199 1517.74,530.025 1522.83,530.056 
  1527.92,530.292 1533.02,530.734 1538.11,531.38 1543.2,532.23 1548.29,533.282 1553.38,534.537 1558.47,535.992 1563.56,537.646 1568.66,539.497 1573.75,541.545 
  1578.84,543.787 1583.93,546.22 1589.02,548.841 1594.11,551.649 1599.21,554.64 1604.3,557.812 1609.39,561.161 1614.48,564.685 1619.57,568.379 1624.66,572.24 
  1629.75,576.264 1634.85,580.446 1639.94,584.783 1645.03,589.27 1650.12,593.904 1655.21,598.679 1660.3,603.59 1665.4,608.634 1670.49,613.803 1675.58,619.094 
  1680.67,624.5 1685.76,630.017 1690.85,635.64 1695.95,641.362 1701.04,647.178 1706.13,653.082 1711.22,659.068 1716.31,665.13 1721.4,671.262 1726.49,677.458 
  1731.59,683.712 1736.68,690.017 1741.77,696.367 1746.86,702.756 1751.95,709.178 1757.04,715.625 1762.14,722.091 1767.23,728.571 1772.32,735.058 1777.41,741.545 
  1782.5,748.025 1787.59,754.492 1792.68,760.94 1797.78,767.362 1802.87,773.752 1807.96,780.104 1813.05,786.41 1818.14,792.665 1823.23,798.863 1828.33,804.997 
  1833.42,811.061 1838.51,817.049 1843.6,822.955 1848.69,828.773 1853.78,834.498 1858.87,840.123 1863.97,845.644 1869.06,851.054 1874.15,856.348 1879.24,861.521 
  1884.33,866.567 1889.42,871.482 1894.52,876.26 1899.61,880.898 1904.7,885.39 1909.79,889.731 1914.88,893.919 1919.97,897.947 1925.07,901.812 1930.16,905.511 
  1935.25,909.039 1940.34,912.394 1945.43,915.571 1950.52,918.568 1955.61,921.382 1960.71,924.009 1965.8,926.448 1970.89,928.695 1975.98,930.748 1981.07,932.606 
  1986.16,934.267 1991.26,935.728 1996.35,936.99 2001.44,938.049 2006.53,938.905 2011.62,939.557 2016.71,940.005 2021.8,940.248 2026.9,940.287 2031.99,940.12 
  2037.08,939.749 2042.17,939.172 2047.26,938.392 2052.35,937.408 2057.45,936.222 2062.54,934.835 2067.63,933.249 2072.72,931.465 2077.81,929.485 2082.9,927.309 
  2087.99,924.942 2093.09,922.385 2098.18,919.641 2103.27,916.712 2108.36,913.602 2113.45,910.314 2118.54,906.851 2123.64,903.216 2128.73,899.412 2133.82,895.445 
  2138.91,891.317 2144,887.034 2149.09,882.598 2154.19,878.016 2159.28,873.29 2164.37,868.427 2169.46,863.429 2174.55,858.304 2179.64,853.056 2184.73,847.69 
  2189.83,842.212 2194.92,836.627 2200.01,830.94 2205.1,825.157 2210.19,819.284 2215.28,813.328 2220.38,807.293 2225.47,801.186 2230.56,795.014 2235.65,788.782 
  2240.74,782.496 2245.83,776.163 2250.92,769.788 2256.02,763.38 2261.11,756.943 2266.2,750.484 2271.29,744.011 2276.38,737.529 2281.47,731.044 2286.57,724.564 
  2291.66,718.094 
  "/>
<polyline clip-path="url(#clip122)" style="stroke:#3da44d; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  255.035,86.1857 260.126,86.5121 265.218,87.4852 270.309,89.1057 275.401,91.3722 280.492,94.2819 285.584,97.832 290.676,102.019 295.767,106.839 300.859,112.288 
  305.95,118.359 311.042,125.048 316.133,132.345 321.225,140.247 326.316,148.741 331.408,157.821 336.5,167.478 341.591,177.705 346.683,188.486 351.774,199.814 
  356.866,211.679 361.957,224.07 367.049,236.97 372.14,250.369 377.232,264.253 382.324,278.61 387.415,293.424 392.507,308.68 397.598,324.363 402.69,340.457 
  407.781,356.946 412.873,373.813 417.964,391.042 423.056,408.614 428.148,426.514 433.239,444.722 438.331,463.221 443.422,481.991 448.514,501.015 453.605,520.272 
  458.697,539.745 463.788,559.413 468.88,579.256 473.972,599.256 479.063,619.392 484.155,639.643 489.246,659.99 494.338,680.413 499.429,700.89 504.521,721.401 
  509.612,741.926 514.704,762.445 519.796,782.936 524.887,803.38 529.979,823.755 535.07,844.042 540.162,864.22 545.253,884.269 550.345,904.169 555.437,923.9 
  560.528,943.442 565.62,962.776 570.711,981.883 575.803,1000.74 580.894,1019.34 585.986,1037.65 591.077,1055.65 596.169,1073.34 601.261,1090.69 606.352,1107.68 
  611.444,1124.31 616.535,1140.54 621.627,1156.36 626.718,1171.77 631.81,1186.74 636.901,1201.26 641.993,1215.31 647.085,1228.88 652.176,1241.96 657.268,1254.53 
  662.359,1266.58 667.451,1278.1 672.542,1289.08 677.634,1299.5 682.725,1309.36 687.817,1318.64 692.909,1327.34 698,1335.45 703.092,1342.96 708.183,1349.86 
  713.275,1356.15 718.366,1361.82 723.458,1366.85 728.549,1371.26 733.641,1375.03 738.733,1378.16 743.824,1380.65 748.916,1382.5 754.007,1383.69 759.099,1384.24 
  764.19,1384.14 769.282,1383.39 774.373,1381.99 779.465,1379.94 784.557,1377.25 789.648,1373.92 794.74,1369.95 799.831,1365.34 804.923,1360.1 810.014,1354.24 
  815.106,1347.76 820.197,1340.67 825.289,1332.97 830.381,1324.67 835.472,1315.79 840.564,1306.32 845.655,1296.28 850.747,1285.68 855.838,1274.53 860.93,1262.84 
  866.021,1250.63 871.113,1237.89 876.205,1224.66 881.296,1210.94 886.388,1196.74 891.479,1182.08 896.571,1166.97 901.662,1151.43 906.754,1135.47 911.845,1119.12 
  916.937,1102.38 922.029,1085.27 927.12,1067.81 932.212,1050.02 937.303,1031.92 942.395,1013.52 947.486,994.838 952.578,975.9 957.669,956.72 962.761,937.319 
  967.853,917.715 972.944,897.93 978.036,877.981 983.127,857.89 988.219,837.675 993.31,817.358 998.402,796.958 1003.49,776.497 1008.59,755.995 1013.68,735.472 
  1018.77,714.949 1023.86,694.445 1028.95,673.981 1034.04,653.579 1039.13,633.258 1044.23,613.04 1049.32,592.944 1054.41,572.991 1059.5,553.199 1064.59,533.589 
  1069.68,514.181 1074.78,494.994 1079.87,476.048 1084.96,457.361 1090.05,438.952 1095.14,420.841 1100.23,403.043 1105.32,385.578 1110.42,368.462 1115.51,351.714 
  1120.6,335.349 1125.69,319.385 1130.78,303.836 1135.87,288.719 1140.97,274.049 1146.06,259.841 1151.15,246.108 1156.24,232.865 1161.33,220.124 1166.42,207.897 
  1171.51,196.198 1176.61,185.038 1181.7,174.429 1186.79,164.381 1191.88,154.905 1196.97,146.008 1202.06,137.7 1207.16,129.989 1212.25,122.883 1217.34,116.39 
  1222.43,110.517 1227.52,105.269 1232.61,100.65 1237.71,96.6656 1242.8,93.3193 1247.89,90.6143 1252.98,88.5535 1258.07,87.1413 1263.16,86.3772 1268.25,86.2619 
  1273.35,86.7954 1278.44,87.9771 1283.53,89.8054 1288.62,92.2782 1293.71,95.3945 1298.8,99.1522 1303.9,103.546 1308.99,108.571 1314.08,114.223 1319.17,120.495 
  1324.26,127.382 1329.35,134.875 1334.44,142.97 1339.54,151.658 1344.63,160.93 1349.72,170.775 1354.81,181.185 1359.9,192.149 1364.99,203.654 1370.09,215.69 
  1375.18,228.248 1380.27,241.313 1385.36,254.872 1390.45,268.911 1395.54,283.416 1400.63,298.371 1405.73,313.763 1410.82,329.576 1415.91,345.796 1421,362.406 
  1426.09,379.388 1431.18,396.726 1436.28,414.402 1441.37,432.398 1446.46,450.696 1451.55,469.279 1456.64,488.128 1461.73,507.225 1466.83,526.55 1471.92,546.082 
  1477.01,565.804 1482.1,585.694 1487.19,605.733 1492.28,625.902 1497.37,646.181 1502.47,666.549 1507.56,686.985 1512.65,707.469 1517.74,727.98 1522.83,748.498 
  1527.92,769.003 1533.02,789.474 1538.11,809.891 1543.2,830.233 1548.29,850.48 1553.38,870.611 1558.47,890.607 1563.56,910.446 1568.66,930.111 1573.75,949.58 
  1578.84,968.835 1583.93,987.857 1589.02,1006.62 1594.11,1025.12 1599.21,1043.33 1604.3,1061.23 1609.39,1078.8 1614.48,1096.03 1619.57,1112.89 1624.66,1129.38 
  1629.75,1145.48 1634.85,1161.16 1639.94,1176.42 1645.03,1191.24 1650.12,1205.6 1655.21,1219.49 1660.3,1232.9 1665.4,1245.81 1670.49,1258.2 1675.58,1270.08 
  1680.67,1281.42 1685.76,1292.21 1690.85,1302.45 1695.95,1312.12 1701.04,1321.21 1706.13,1329.71 1711.22,1337.62 1716.31,1344.93 1721.4,1351.63 1726.49,1357.71 
  1731.59,1363.17 1736.68,1368 1741.77,1372.2 1746.86,1375.76 1751.95,1378.68 1757.04,1380.96 1762.14,1382.59 1767.23,1383.58 1772.32,1383.91 1777.41,1383.6 
  1782.5,1382.64 1787.59,1381.03 1792.68,1378.77 1797.78,1375.87 1802.87,1372.34 1807.96,1368.16 1813.05,1363.35 1818.14,1357.91 1823.23,1351.85 1828.33,1345.17 
  1833.42,1337.89 1838.51,1330 1843.6,1321.51 1848.69,1312.44 1853.78,1302.79 1858.87,1292.58 1863.97,1281.8 1869.06,1270.49 1874.15,1258.63 1879.24,1246.26 
  1884.33,1233.37 1889.42,1219.98 1894.52,1206.11 1899.61,1191.76 1904.7,1176.96 1909.79,1161.72 1914.88,1146.06 1919.97,1129.98 1925.07,1113.51 1930.16,1096.66 
  1935.25,1079.44 1940.34,1061.89 1945.43,1044 1950.52,1025.81 1955.61,1007.33 1960.71,988.579 1965.8,969.571 1970.89,950.329 1975.98,930.872 1981.07,911.22 
  1986.16,891.392 1991.26,871.408 1996.35,851.29 2001.44,831.054 2006.53,810.722 2011.62,790.314 2016.71,769.852 2021.8,749.355 2026.9,728.845 2031.99,708.341 
  2037.08,687.865 2042.17,667.435 2047.26,647.073 2052.35,626.8 2057.45,606.635 2062.54,586.599 2067.63,566.712 2072.72,546.993 2077.81,527.463 2082.9,508.141 
  2087.99,489.046 2093.09,470.197 2098.18,451.613 2103.27,433.313 2108.36,415.315 2113.45,397.637 2118.54,380.297 2123.64,363.312 2128.73,346.699 2133.82,330.475 
  2138.91,314.656 2144,299.257 2149.09,284.294 2154.19,269.782 2159.28,255.736 2164.37,242.169 2169.46,229.096 2174.55,216.529 2179.64,204.481 2184.73,192.963 
  2189.83,181.988 2194.92,171.565 2200.01,161.707 2205.1,152.423 2210.19,143.722 2215.28,135.612 2220.38,128.102 2225.47,121.198 2230.56,114.908 2235.65,109.239 
  2240.74,104.196 2245.83,99.7854 2250.92,96.0098 2256.02,92.8732 2261.11,90.3787 2266.2,88.5284 2271.29,87.3239 2276.38,86.7676 2281.47,86.8599 2286.57,87.6003 
  2291.66,88.9879 
  "/>
<path clip-path="url(#clip120)" d="
M1557.04 300.469 L2280.8 300.469 L2280.8 93.1086 L1557.04 93.1086  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<polyline clip-path="url(#clip120)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1557.04,300.469 2280.8,300.469 2280.8,93.1086 1557.04,93.1086 1557.04,300.469 
  "/>
<polyline clip-path="url(#clip120)" style="stroke:#009af9; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1581.03,144.949 1724.95,144.949 
  "/>
<path clip-path="url(#clip120)" d="M1769.28 141.28 Q1770.88 138.409 1773.1 137.044 Q1775.32 135.678 1778.33 135.678 Q1782.38 135.678 1784.58 138.525 Q1786.78 141.349 1786.78 146.581 L1786.78 162.229 L1782.5 162.229 L1782.5 146.719 Q1782.5 142.993 1781.18 141.187 Q1779.86 139.382 1777.15 139.382 Q1773.84 139.382 1771.92 141.581 Q1770 143.78 1770 147.576 L1770 162.229 L1765.72 162.229 L1765.72 146.719 Q1765.72 142.969 1764.4 141.187 Q1763.08 139.382 1760.32 139.382 Q1757.06 139.382 1755.14 141.604 Q1753.22 143.803 1753.22 147.576 L1753.22 162.229 L1748.94 162.229 L1748.94 136.303 L1753.22 136.303 L1753.22 140.331 Q1754.68 137.946 1756.71 136.812 Q1758.75 135.678 1761.55 135.678 Q1764.37 135.678 1766.34 137.113 Q1768.33 138.548 1769.28 141.28 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M1807.06 149.196 Q1801.9 149.196 1799.91 150.377 Q1797.92 151.557 1797.92 154.405 Q1797.92 156.673 1799.4 158.016 Q1800.9 159.335 1803.47 159.335 Q1807.01 159.335 1809.14 156.835 Q1811.3 154.312 1811.3 150.145 L1811.3 149.196 L1807.06 149.196 M1815.56 147.437 L1815.56 162.229 L1811.3 162.229 L1811.3 158.293 Q1809.84 160.655 1807.66 161.789 Q1805.49 162.9 1802.34 162.9 Q1798.36 162.9 1796 160.678 Q1793.66 158.432 1793.66 154.682 Q1793.66 150.307 1796.57 148.085 Q1799.51 145.863 1805.32 145.863 L1811.3 145.863 L1811.3 145.446 Q1811.3 142.507 1809.35 140.909 Q1807.43 139.289 1803.93 139.289 Q1801.71 139.289 1799.61 139.821 Q1797.5 140.354 1795.56 141.419 L1795.56 137.483 Q1797.89 136.581 1800.09 136.141 Q1802.29 135.678 1804.37 135.678 Q1810 135.678 1812.78 138.594 Q1815.56 141.511 1815.56 147.437 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M1840.86 137.067 L1840.86 141.094 Q1839.05 140.169 1837.11 139.706 Q1835.16 139.243 1833.08 139.243 Q1829.91 139.243 1828.31 140.215 Q1826.74 141.187 1826.74 143.131 Q1826.74 144.613 1827.87 145.469 Q1829 146.303 1832.43 147.067 L1833.89 147.391 Q1838.43 148.363 1840.32 150.145 Q1842.24 151.905 1842.24 155.076 Q1842.24 158.687 1839.37 160.793 Q1836.53 162.9 1831.53 162.9 Q1829.44 162.9 1827.18 162.483 Q1824.93 162.09 1822.43 161.28 L1822.43 156.881 Q1824.79 158.108 1827.08 158.733 Q1829.37 159.335 1831.62 159.335 Q1834.63 159.335 1836.25 158.317 Q1837.87 157.275 1837.87 155.4 Q1837.87 153.664 1836.69 152.738 Q1835.53 151.812 1831.57 150.956 L1830.09 150.608 Q1826.13 149.775 1824.37 148.062 Q1822.62 146.326 1822.62 143.317 Q1822.62 139.659 1825.21 137.669 Q1827.8 135.678 1832.57 135.678 Q1834.93 135.678 1837.01 136.025 Q1839.1 136.372 1840.86 137.067 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M1865.55 137.067 L1865.55 141.094 Q1863.75 140.169 1861.8 139.706 Q1859.86 139.243 1857.78 139.243 Q1854.61 139.243 1853.01 140.215 Q1851.43 141.187 1851.43 143.131 Q1851.43 144.613 1852.57 145.469 Q1853.7 146.303 1857.13 147.067 L1858.59 147.391 Q1863.12 148.363 1865.02 150.145 Q1866.94 151.905 1866.94 155.076 Q1866.94 158.687 1864.07 160.793 Q1861.23 162.9 1856.23 162.9 Q1854.14 162.9 1851.87 162.483 Q1849.63 162.09 1847.13 161.28 L1847.13 156.881 Q1849.49 158.108 1851.78 158.733 Q1854.07 159.335 1856.32 159.335 Q1859.33 159.335 1860.95 158.317 Q1862.57 157.275 1862.57 155.4 Q1862.57 153.664 1861.39 152.738 Q1860.23 151.812 1856.27 150.956 L1854.79 150.608 Q1850.83 149.775 1849.07 148.062 Q1847.31 146.326 1847.31 143.317 Q1847.31 139.659 1849.91 137.669 Q1852.5 135.678 1857.27 135.678 Q1859.63 135.678 1861.71 136.025 Q1863.8 136.372 1865.55 137.067 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M1874.33 156.349 L1879.21 156.349 L1879.21 162.229 L1874.33 162.229 L1874.33 156.349 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M1905.32 137.067 L1905.32 141.094 Q1903.52 140.169 1901.57 139.706 Q1899.63 139.243 1897.55 139.243 Q1894.37 139.243 1892.78 140.215 Q1891.2 141.187 1891.2 143.131 Q1891.2 144.613 1892.34 145.469 Q1893.47 146.303 1896.9 147.067 L1898.36 147.391 Q1902.89 148.363 1904.79 150.145 Q1906.71 151.905 1906.71 155.076 Q1906.71 158.687 1903.84 160.793 Q1900.99 162.9 1895.99 162.9 Q1893.91 162.9 1891.64 162.483 Q1889.4 162.09 1886.9 161.28 L1886.9 156.881 Q1889.26 158.108 1891.55 158.733 Q1893.84 159.335 1896.09 159.335 Q1899.1 159.335 1900.72 158.317 Q1902.34 157.275 1902.34 155.4 Q1902.34 153.664 1901.16 152.738 Q1900 151.812 1896.04 150.956 L1894.56 150.608 Q1890.6 149.775 1888.84 148.062 Q1887.08 146.326 1887.08 143.317 Q1887.08 139.659 1889.68 137.669 Q1892.27 135.678 1897.04 135.678 Q1899.4 135.678 1901.48 136.025 Q1903.56 136.372 1905.32 137.067 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M1938.8 126.257 Q1935.69 131.581 1934.19 136.789 Q1932.68 141.997 1932.68 147.344 Q1932.68 152.692 1934.19 157.946 Q1935.72 163.178 1938.8 168.479 L1935.09 168.479 Q1931.62 163.039 1929.88 157.784 Q1928.17 152.53 1928.17 147.344 Q1928.17 142.182 1929.88 136.951 Q1931.6 131.72 1935.09 126.257 L1938.8 126.257 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M1961.83 143.594 Q1965.18 144.312 1967.06 146.581 Q1968.96 148.849 1968.96 152.182 Q1968.96 157.298 1965.44 160.099 Q1961.92 162.9 1955.44 162.9 Q1953.26 162.9 1950.95 162.46 Q1948.66 162.043 1946.2 161.187 L1946.2 156.673 Q1948.15 157.807 1950.46 158.386 Q1952.78 158.965 1955.3 158.965 Q1959.7 158.965 1961.99 157.229 Q1964.3 155.493 1964.3 152.182 Q1964.3 149.127 1962.15 147.414 Q1960.02 145.678 1956.2 145.678 L1952.17 145.678 L1952.17 141.835 L1956.39 141.835 Q1959.84 141.835 1961.67 140.469 Q1963.49 139.081 1963.49 136.488 Q1963.49 133.826 1961.6 132.414 Q1959.72 130.979 1956.2 130.979 Q1954.28 130.979 1952.08 131.395 Q1949.88 131.812 1947.24 132.692 L1947.24 128.525 Q1949.91 127.784 1952.22 127.414 Q1954.56 127.044 1956.62 127.044 Q1961.94 127.044 1965.05 129.474 Q1968.15 131.882 1968.15 136.002 Q1968.15 138.872 1966.5 140.863 Q1964.86 142.831 1961.83 143.594 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M1991.99 143.594 Q1995.35 144.312 1997.22 146.581 Q1999.12 148.849 1999.12 152.182 Q1999.12 157.298 1995.6 160.099 Q1992.08 162.9 1985.6 162.9 Q1983.42 162.9 1981.11 162.46 Q1978.82 162.043 1976.36 161.187 L1976.36 156.673 Q1978.31 157.807 1980.62 158.386 Q1982.94 158.965 1985.46 158.965 Q1989.86 158.965 1992.15 157.229 Q1994.47 155.493 1994.47 152.182 Q1994.47 149.127 1992.31 147.414 Q1990.18 145.678 1986.36 145.678 L1982.34 145.678 L1982.34 141.835 L1986.55 141.835 Q1990 141.835 1991.83 140.469 Q1993.66 139.081 1993.66 136.488 Q1993.66 133.826 1991.76 132.414 Q1989.88 130.979 1986.36 130.979 Q1984.44 130.979 1982.24 131.395 Q1980.05 131.812 1977.41 132.692 L1977.41 128.525 Q1980.07 127.784 1982.38 127.414 Q1984.72 127.044 1986.78 127.044 Q1992.11 127.044 1995.21 129.474 Q1998.31 131.882 1998.31 136.002 Q1998.31 138.872 1996.67 140.863 Q1995.02 142.831 1991.99 143.594 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M2008.03 127.669 L2026.39 127.669 L2026.39 131.604 L2012.31 131.604 L2012.31 140.076 Q2013.33 139.729 2014.35 139.567 Q2015.37 139.382 2016.39 139.382 Q2022.17 139.382 2025.55 142.553 Q2028.93 145.724 2028.93 151.141 Q2028.93 156.719 2025.46 159.821 Q2021.99 162.9 2015.67 162.9 Q2013.49 162.9 2011.23 162.53 Q2008.98 162.159 2006.57 161.418 L2006.57 156.719 Q2008.66 157.854 2010.88 158.409 Q2013.1 158.965 2015.58 158.965 Q2019.58 158.965 2021.92 156.858 Q2024.26 154.752 2024.26 151.141 Q2024.26 147.53 2021.92 145.423 Q2019.58 143.317 2015.58 143.317 Q2013.7 143.317 2011.83 143.733 Q2009.98 144.15 2008.03 145.03 L2008.03 127.669 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M2038.19 127.669 L2056.55 127.669 L2056.55 131.604 L2042.48 131.604 L2042.48 140.076 Q2043.49 139.729 2044.51 139.567 Q2045.53 139.382 2046.55 139.382 Q2052.34 139.382 2055.72 142.553 Q2059.1 145.724 2059.1 151.141 Q2059.1 156.719 2055.62 159.821 Q2052.15 162.9 2045.83 162.9 Q2043.66 162.9 2041.39 162.53 Q2039.14 162.159 2036.73 161.418 L2036.73 156.719 Q2038.82 157.854 2041.04 158.409 Q2043.26 158.965 2045.74 158.965 Q2049.74 158.965 2052.08 156.858 Q2054.42 154.752 2054.42 151.141 Q2054.42 147.53 2052.08 145.423 Q2049.74 143.317 2045.74 143.317 Q2043.86 143.317 2041.99 143.733 Q2040.14 144.15 2038.19 145.03 L2038.19 127.669 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M2081.16 131.743 L2069.35 150.192 L2081.16 150.192 L2081.16 131.743 M2079.93 127.669 L2085.81 127.669 L2085.81 150.192 L2090.74 150.192 L2090.74 154.081 L2085.81 154.081 L2085.81 162.229 L2081.16 162.229 L2081.16 154.081 L2065.55 154.081 L2065.55 149.567 L2079.93 127.669 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M2111.32 131.743 L2099.51 150.192 L2111.32 150.192 L2111.32 131.743 M2110.09 127.669 L2115.97 127.669 L2115.97 150.192 L2120.9 150.192 L2120.9 154.081 L2115.97 154.081 L2115.97 162.229 L2111.32 162.229 L2111.32 154.081 L2095.72 154.081 L2095.72 149.567 L2110.09 127.669 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M2142.8 143.594 Q2146.16 144.312 2148.03 146.581 Q2149.93 148.849 2149.93 152.182 Q2149.93 157.298 2146.41 160.099 Q2142.89 162.9 2136.41 162.9 Q2134.23 162.9 2131.92 162.46 Q2129.63 162.043 2127.17 161.187 L2127.17 156.673 Q2129.12 157.807 2131.43 158.386 Q2133.75 158.965 2136.27 158.965 Q2140.67 158.965 2142.96 157.229 Q2145.28 155.493 2145.28 152.182 Q2145.28 149.127 2143.12 147.414 Q2140.99 145.678 2137.17 145.678 L2133.15 145.678 L2133.15 141.835 L2137.36 141.835 Q2140.81 141.835 2142.64 140.469 Q2144.47 139.081 2144.47 136.488 Q2144.47 133.826 2142.57 132.414 Q2140.69 130.979 2137.17 130.979 Q2135.25 130.979 2133.05 131.395 Q2130.85 131.812 2128.22 132.692 L2128.22 128.525 Q2130.88 127.784 2133.19 127.414 Q2135.53 127.044 2137.59 127.044 Q2142.91 127.044 2146.02 129.474 Q2149.12 131.882 2149.12 136.002 Q2149.12 138.872 2147.47 140.863 Q2145.83 142.831 2142.8 143.594 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M2162.82 158.293 L2179.14 158.293 L2179.14 162.229 L2157.2 162.229 L2157.2 158.293 Q2159.86 155.539 2164.44 150.909 Q2169.05 146.256 2170.23 144.914 Q2172.47 142.391 2173.35 140.655 Q2174.26 138.895 2174.26 137.206 Q2174.26 134.451 2172.31 132.715 Q2170.39 130.979 2167.29 130.979 Q2165.09 130.979 2162.64 131.743 Q2160.21 132.507 2157.43 134.057 L2157.43 129.335 Q2160.25 128.201 2162.71 127.622 Q2165.16 127.044 2167.2 127.044 Q2172.57 127.044 2175.76 129.729 Q2178.96 132.414 2178.96 136.905 Q2178.96 139.034 2178.15 140.956 Q2177.36 142.854 2175.25 145.446 Q2174.67 146.118 2171.57 149.335 Q2168.47 152.53 2162.82 158.293 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M2187.68 126.257 L2191.39 126.257 Q2194.86 131.72 2196.57 136.951 Q2198.31 142.182 2198.31 147.344 Q2198.31 152.53 2196.57 157.784 Q2194.86 163.039 2191.39 168.479 L2187.68 168.479 Q2190.76 163.178 2192.27 157.946 Q2193.79 152.692 2193.79 147.344 Q2193.79 141.997 2192.27 136.789 Q2190.76 131.581 2187.68 126.257 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip120)" style="stroke:#e26f46; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1581.03,196.789 1724.95,196.789 
  "/>
<path clip-path="url(#clip120)" d="M1769.28 193.12 Q1770.88 190.249 1773.1 188.884 Q1775.32 187.518 1778.33 187.518 Q1782.38 187.518 1784.58 190.365 Q1786.78 193.189 1786.78 198.421 L1786.78 214.069 L1782.5 214.069 L1782.5 198.559 Q1782.5 194.833 1781.18 193.027 Q1779.86 191.222 1777.15 191.222 Q1773.84 191.222 1771.92 193.421 Q1770 195.62 1770 199.416 L1770 214.069 L1765.72 214.069 L1765.72 198.559 Q1765.72 194.809 1764.4 193.027 Q1763.08 191.222 1760.32 191.222 Q1757.06 191.222 1755.14 193.444 Q1753.22 195.643 1753.22 199.416 L1753.22 214.069 L1748.94 214.069 L1748.94 188.143 L1753.22 188.143 L1753.22 192.171 Q1754.68 189.786 1756.71 188.652 Q1758.75 187.518 1761.55 187.518 Q1764.37 187.518 1766.34 188.953 Q1768.33 190.388 1769.28 193.12 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M1807.06 201.036 Q1801.9 201.036 1799.91 202.217 Q1797.92 203.397 1797.92 206.245 Q1797.92 208.513 1799.4 209.856 Q1800.9 211.175 1803.47 211.175 Q1807.01 211.175 1809.14 208.675 Q1811.3 206.152 1811.3 201.985 L1811.3 201.036 L1807.06 201.036 M1815.56 199.277 L1815.56 214.069 L1811.3 214.069 L1811.3 210.133 Q1809.84 212.495 1807.66 213.629 Q1805.49 214.74 1802.34 214.74 Q1798.36 214.74 1796 212.518 Q1793.66 210.272 1793.66 206.522 Q1793.66 202.147 1796.57 199.925 Q1799.51 197.703 1805.32 197.703 L1811.3 197.703 L1811.3 197.286 Q1811.3 194.347 1809.35 192.749 Q1807.43 191.129 1803.93 191.129 Q1801.71 191.129 1799.61 191.661 Q1797.5 192.194 1795.56 193.259 L1795.56 189.323 Q1797.89 188.421 1800.09 187.981 Q1802.29 187.518 1804.37 187.518 Q1810 187.518 1812.78 190.434 Q1815.56 193.351 1815.56 199.277 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M1840.86 188.907 L1840.86 192.934 Q1839.05 192.009 1837.11 191.546 Q1835.16 191.083 1833.08 191.083 Q1829.91 191.083 1828.31 192.055 Q1826.74 193.027 1826.74 194.971 Q1826.74 196.453 1827.87 197.309 Q1829 198.143 1832.43 198.907 L1833.89 199.231 Q1838.43 200.203 1840.32 201.985 Q1842.24 203.745 1842.24 206.916 Q1842.24 210.527 1839.37 212.633 Q1836.53 214.74 1831.53 214.74 Q1829.44 214.74 1827.18 214.323 Q1824.93 213.93 1822.43 213.12 L1822.43 208.721 Q1824.79 209.948 1827.08 210.573 Q1829.37 211.175 1831.62 211.175 Q1834.63 211.175 1836.25 210.157 Q1837.87 209.115 1837.87 207.24 Q1837.87 205.504 1836.69 204.578 Q1835.53 203.652 1831.57 202.796 L1830.09 202.448 Q1826.13 201.615 1824.37 199.902 Q1822.62 198.166 1822.62 195.157 Q1822.62 191.499 1825.21 189.509 Q1827.8 187.518 1832.57 187.518 Q1834.93 187.518 1837.01 187.865 Q1839.1 188.212 1840.86 188.907 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M1865.55 188.907 L1865.55 192.934 Q1863.75 192.009 1861.8 191.546 Q1859.86 191.083 1857.78 191.083 Q1854.61 191.083 1853.01 192.055 Q1851.43 193.027 1851.43 194.971 Q1851.43 196.453 1852.57 197.309 Q1853.7 198.143 1857.13 198.907 L1858.59 199.231 Q1863.12 200.203 1865.02 201.985 Q1866.94 203.745 1866.94 206.916 Q1866.94 210.527 1864.07 212.633 Q1861.23 214.74 1856.23 214.74 Q1854.14 214.74 1851.87 214.323 Q1849.63 213.93 1847.13 213.12 L1847.13 208.721 Q1849.49 209.948 1851.78 210.573 Q1854.07 211.175 1856.32 211.175 Q1859.33 211.175 1860.95 210.157 Q1862.57 209.115 1862.57 207.24 Q1862.57 205.504 1861.39 204.578 Q1860.23 203.652 1856.27 202.796 L1854.79 202.448 Q1850.83 201.615 1849.07 199.902 Q1847.31 198.166 1847.31 195.157 Q1847.31 191.499 1849.91 189.509 Q1852.5 187.518 1857.27 187.518 Q1859.63 187.518 1861.71 187.865 Q1863.8 188.212 1865.55 188.907 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M1874.33 208.189 L1879.21 208.189 L1879.21 214.069 L1874.33 214.069 L1874.33 208.189 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M1885.74 188.143 L1890.25 188.143 L1898.36 209.902 L1906.46 188.143 L1910.97 188.143 L1901.25 214.069 L1895.46 214.069 L1885.74 188.143 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M1942.15 178.097 Q1939.05 183.421 1937.55 188.629 Q1936.04 193.837 1936.04 199.184 Q1936.04 204.532 1937.55 209.786 Q1939.07 215.018 1942.15 220.319 L1938.45 220.319 Q1934.98 214.879 1933.24 209.624 Q1931.53 204.37 1931.53 199.184 Q1931.53 194.022 1933.24 188.791 Q1934.95 183.56 1938.45 178.097 L1942.15 178.097 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M1965.18 195.434 Q1968.54 196.152 1970.42 198.421 Q1972.31 200.689 1972.31 204.022 Q1972.31 209.138 1968.8 211.939 Q1965.28 214.74 1958.8 214.74 Q1956.62 214.74 1954.3 214.3 Q1952.01 213.883 1949.56 213.027 L1949.56 208.513 Q1951.5 209.647 1953.82 210.226 Q1956.13 210.805 1958.66 210.805 Q1963.05 210.805 1965.35 209.069 Q1967.66 207.333 1967.66 204.022 Q1967.66 200.967 1965.51 199.254 Q1963.38 197.518 1959.56 197.518 L1955.53 197.518 L1955.53 193.675 L1959.74 193.675 Q1963.19 193.675 1965.02 192.309 Q1966.85 190.921 1966.85 188.328 Q1966.85 185.666 1964.95 184.254 Q1963.08 182.819 1959.56 182.819 Q1957.64 182.819 1955.44 183.235 Q1953.24 183.652 1950.6 184.532 L1950.6 180.365 Q1953.26 179.624 1955.58 179.254 Q1957.92 178.884 1959.98 178.884 Q1965.3 178.884 1968.4 181.314 Q1971.5 183.722 1971.5 187.842 Q1971.5 190.712 1969.86 192.703 Q1968.22 194.671 1965.18 195.434 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M1995.35 195.434 Q1998.7 196.152 2000.58 198.421 Q2002.48 200.689 2002.48 204.022 Q2002.48 209.138 1998.96 211.939 Q1995.44 214.74 1988.96 214.74 Q1986.78 214.74 1984.47 214.3 Q1982.17 213.883 1979.72 213.027 L1979.72 208.513 Q1981.67 209.647 1983.98 210.226 Q1986.3 210.805 1988.82 210.805 Q1993.22 210.805 1995.51 209.069 Q1997.82 207.333 1997.82 204.022 Q1997.82 200.967 1995.67 199.254 Q1993.54 197.518 1989.72 197.518 L1985.69 197.518 L1985.69 193.675 L1989.91 193.675 Q1993.36 193.675 1995.18 192.309 Q1997.01 190.921 1997.01 188.328 Q1997.01 185.666 1995.11 184.254 Q1993.24 182.819 1989.72 182.819 Q1987.8 182.819 1985.6 183.235 Q1983.4 183.652 1980.76 184.532 L1980.76 180.365 Q1983.42 179.624 1985.74 179.254 Q1988.08 178.884 1990.14 178.884 Q1995.46 178.884 1998.56 181.314 Q2001.67 183.722 2001.67 187.842 Q2001.67 190.712 2000.02 192.703 Q1998.38 194.671 1995.35 195.434 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M2011.39 179.509 L2029.74 179.509 L2029.74 183.444 L2015.67 183.444 L2015.67 191.916 Q2016.69 191.569 2017.71 191.407 Q2018.73 191.222 2019.74 191.222 Q2025.53 191.222 2028.91 194.393 Q2032.29 197.564 2032.29 202.981 Q2032.29 208.559 2028.82 211.661 Q2025.35 214.74 2019.03 214.74 Q2016.85 214.74 2014.58 214.37 Q2012.34 213.999 2009.93 213.258 L2009.93 208.559 Q2012.01 209.694 2014.23 210.249 Q2016.46 210.805 2018.93 210.805 Q2022.94 210.805 2025.28 208.698 Q2027.61 206.592 2027.61 202.981 Q2027.61 199.37 2025.28 197.263 Q2022.94 195.157 2018.93 195.157 Q2017.06 195.157 2015.18 195.573 Q2013.33 195.99 2011.39 196.87 L2011.39 179.509 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M2041.55 179.509 L2059.91 179.509 L2059.91 183.444 L2045.83 183.444 L2045.83 191.916 Q2046.85 191.569 2047.87 191.407 Q2048.89 191.222 2049.91 191.222 Q2055.69 191.222 2059.07 194.393 Q2062.45 197.564 2062.45 202.981 Q2062.45 208.559 2058.98 211.661 Q2055.51 214.74 2049.19 214.74 Q2047.01 214.74 2044.74 214.37 Q2042.5 213.999 2040.09 213.258 L2040.09 208.559 Q2042.17 209.694 2044.4 210.249 Q2046.62 210.805 2049.1 210.805 Q2053.1 210.805 2055.44 208.698 Q2057.78 206.592 2057.78 202.981 Q2057.78 199.37 2055.44 197.263 Q2053.1 195.157 2049.1 195.157 Q2047.22 195.157 2045.35 195.573 Q2043.49 195.99 2041.55 196.87 L2041.55 179.509 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M2084.51 183.583 L2072.71 202.032 L2084.51 202.032 L2084.51 183.583 M2083.29 179.509 L2089.16 179.509 L2089.16 202.032 L2094.1 202.032 L2094.1 205.921 L2089.16 205.921 L2089.16 214.069 L2084.51 214.069 L2084.51 205.921 L2068.91 205.921 L2068.91 201.407 L2083.29 179.509 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M2114.67 183.583 L2102.87 202.032 L2114.67 202.032 L2114.67 183.583 M2113.45 179.509 L2119.33 179.509 L2119.33 202.032 L2124.26 202.032 L2124.26 205.921 L2119.33 205.921 L2119.33 214.069 L2114.67 214.069 L2114.67 205.921 L2099.07 205.921 L2099.07 201.407 L2113.45 179.509 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M2146.16 195.434 Q2149.51 196.152 2151.39 198.421 Q2153.28 200.689 2153.28 204.022 Q2153.28 209.138 2149.77 211.939 Q2146.25 214.74 2139.77 214.74 Q2137.59 214.74 2135.28 214.3 Q2132.98 213.883 2130.53 213.027 L2130.53 208.513 Q2132.47 209.647 2134.79 210.226 Q2137.1 210.805 2139.63 210.805 Q2144.03 210.805 2146.32 209.069 Q2148.63 207.333 2148.63 204.022 Q2148.63 200.967 2146.48 199.254 Q2144.35 197.518 2140.53 197.518 L2136.5 197.518 L2136.5 193.675 L2140.72 193.675 Q2144.16 193.675 2145.99 192.309 Q2147.82 190.921 2147.82 188.328 Q2147.82 185.666 2145.92 184.254 Q2144.05 182.819 2140.53 182.819 Q2138.61 182.819 2136.41 183.235 Q2134.21 183.652 2131.57 184.532 L2131.57 180.365 Q2134.23 179.624 2136.55 179.254 Q2138.89 178.884 2140.95 178.884 Q2146.27 178.884 2149.37 181.314 Q2152.47 183.722 2152.47 187.842 Q2152.47 190.712 2150.83 192.703 Q2149.19 194.671 2146.16 195.434 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M2176.32 195.434 Q2179.67 196.152 2181.55 198.421 Q2183.45 200.689 2183.45 204.022 Q2183.45 209.138 2179.93 211.939 Q2176.41 214.74 2169.93 214.74 Q2167.75 214.74 2165.44 214.3 Q2163.15 213.883 2160.69 213.027 L2160.69 208.513 Q2162.64 209.647 2164.95 210.226 Q2167.27 210.805 2169.79 210.805 Q2174.19 210.805 2176.48 209.069 Q2178.79 207.333 2178.79 204.022 Q2178.79 200.967 2176.64 199.254 Q2174.51 197.518 2170.69 197.518 L2166.66 197.518 L2166.66 193.675 L2170.88 193.675 Q2174.33 193.675 2176.16 192.309 Q2177.98 190.921 2177.98 188.328 Q2177.98 185.666 2176.09 184.254 Q2174.21 182.819 2170.69 182.819 Q2168.77 182.819 2166.57 183.235 Q2164.37 183.652 2161.73 184.532 L2161.73 180.365 Q2164.4 179.624 2166.71 179.254 Q2169.05 178.884 2171.11 178.884 Q2176.43 178.884 2179.53 181.314 Q2182.64 183.722 2182.64 187.842 Q2182.64 190.712 2180.99 192.703 Q2179.35 194.671 2176.32 195.434 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M2191.04 178.097 L2194.74 178.097 Q2198.22 183.56 2199.93 188.791 Q2201.66 194.022 2201.66 199.184 Q2201.66 204.37 2199.93 209.624 Q2198.22 214.879 2194.74 220.319 L2191.04 220.319 Q2194.12 215.018 2195.62 209.786 Q2197.15 204.532 2197.15 199.184 Q2197.15 193.837 2195.62 188.629 Q2194.12 183.421 2191.04 178.097 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip120)" style="stroke:#3da44d; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1581.03,248.629 1724.95,248.629 
  "/>
<path clip-path="url(#clip120)" d="M1769.28 244.96 Q1770.88 242.089 1773.1 240.724 Q1775.32 239.358 1778.33 239.358 Q1782.38 239.358 1784.58 242.205 Q1786.78 245.029 1786.78 250.261 L1786.78 265.909 L1782.5 265.909 L1782.5 250.399 Q1782.5 246.673 1781.18 244.867 Q1779.86 243.062 1777.15 243.062 Q1773.84 243.062 1771.92 245.261 Q1770 247.46 1770 251.256 L1770 265.909 L1765.72 265.909 L1765.72 250.399 Q1765.72 246.649 1764.4 244.867 Q1763.08 243.062 1760.32 243.062 Q1757.06 243.062 1755.14 245.284 Q1753.22 247.483 1753.22 251.256 L1753.22 265.909 L1748.94 265.909 L1748.94 239.983 L1753.22 239.983 L1753.22 244.011 Q1754.68 241.626 1756.71 240.492 Q1758.75 239.358 1761.55 239.358 Q1764.37 239.358 1766.34 240.793 Q1768.33 242.228 1769.28 244.96 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M1807.06 252.876 Q1801.9 252.876 1799.91 254.057 Q1797.92 255.237 1797.92 258.085 Q1797.92 260.353 1799.4 261.696 Q1800.9 263.015 1803.47 263.015 Q1807.01 263.015 1809.14 260.515 Q1811.3 257.992 1811.3 253.825 L1811.3 252.876 L1807.06 252.876 M1815.56 251.117 L1815.56 265.909 L1811.3 265.909 L1811.3 261.973 Q1809.84 264.335 1807.66 265.469 Q1805.49 266.58 1802.34 266.58 Q1798.36 266.58 1796 264.358 Q1793.66 262.112 1793.66 258.362 Q1793.66 253.987 1796.57 251.765 Q1799.51 249.543 1805.32 249.543 L1811.3 249.543 L1811.3 249.126 Q1811.3 246.187 1809.35 244.589 Q1807.43 242.969 1803.93 242.969 Q1801.71 242.969 1799.61 243.501 Q1797.5 244.034 1795.56 245.099 L1795.56 241.163 Q1797.89 240.261 1800.09 239.821 Q1802.29 239.358 1804.37 239.358 Q1810 239.358 1812.78 242.274 Q1815.56 245.191 1815.56 251.117 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M1840.86 240.747 L1840.86 244.774 Q1839.05 243.849 1837.11 243.386 Q1835.16 242.923 1833.08 242.923 Q1829.91 242.923 1828.31 243.895 Q1826.74 244.867 1826.74 246.811 Q1826.74 248.293 1827.87 249.149 Q1829 249.983 1832.43 250.747 L1833.89 251.071 Q1838.43 252.043 1840.32 253.825 Q1842.24 255.585 1842.24 258.756 Q1842.24 262.367 1839.37 264.473 Q1836.53 266.58 1831.53 266.58 Q1829.44 266.58 1827.18 266.163 Q1824.93 265.77 1822.43 264.96 L1822.43 260.561 Q1824.79 261.788 1827.08 262.413 Q1829.37 263.015 1831.62 263.015 Q1834.63 263.015 1836.25 261.997 Q1837.87 260.955 1837.87 259.08 Q1837.87 257.344 1836.69 256.418 Q1835.53 255.492 1831.57 254.636 L1830.09 254.288 Q1826.13 253.455 1824.37 251.742 Q1822.62 250.006 1822.62 246.997 Q1822.62 243.339 1825.21 241.349 Q1827.8 239.358 1832.57 239.358 Q1834.93 239.358 1837.01 239.705 Q1839.1 240.052 1840.86 240.747 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M1865.55 240.747 L1865.55 244.774 Q1863.75 243.849 1861.8 243.386 Q1859.86 242.923 1857.78 242.923 Q1854.61 242.923 1853.01 243.895 Q1851.43 244.867 1851.43 246.811 Q1851.43 248.293 1852.57 249.149 Q1853.7 249.983 1857.13 250.747 L1858.59 251.071 Q1863.12 252.043 1865.02 253.825 Q1866.94 255.585 1866.94 258.756 Q1866.94 262.367 1864.07 264.473 Q1861.23 266.58 1856.23 266.58 Q1854.14 266.58 1851.87 266.163 Q1849.63 265.77 1847.13 264.96 L1847.13 260.561 Q1849.49 261.788 1851.78 262.413 Q1854.07 263.015 1856.32 263.015 Q1859.33 263.015 1860.95 261.997 Q1862.57 260.955 1862.57 259.08 Q1862.57 257.344 1861.39 256.418 Q1860.23 255.492 1856.27 254.636 L1854.79 254.288 Q1850.83 253.455 1849.07 251.742 Q1847.31 250.006 1847.31 246.997 Q1847.31 243.339 1849.91 241.349 Q1852.5 239.358 1857.27 239.358 Q1859.63 239.358 1861.71 239.705 Q1863.8 240.052 1865.55 240.747 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M1874.33 260.029 L1879.21 260.029 L1879.21 265.909 L1874.33 265.909 L1874.33 260.029 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M1900.58 252.876 Q1895.42 252.876 1893.43 254.057 Q1891.43 255.237 1891.43 258.085 Q1891.43 260.353 1892.92 261.696 Q1894.42 263.015 1896.99 263.015 Q1900.53 263.015 1902.66 260.515 Q1904.81 257.992 1904.81 253.825 L1904.81 252.876 L1900.58 252.876 M1909.07 251.117 L1909.07 265.909 L1904.81 265.909 L1904.81 261.973 Q1903.36 264.335 1901.18 265.469 Q1899 266.58 1895.86 266.58 Q1891.87 266.58 1889.51 264.358 Q1887.18 262.112 1887.18 258.362 Q1887.18 253.987 1890.09 251.765 Q1893.03 249.543 1898.84 249.543 L1904.81 249.543 L1904.81 249.126 Q1904.81 246.187 1902.87 244.589 Q1900.95 242.969 1897.45 242.969 Q1895.23 242.969 1893.12 243.501 Q1891.02 244.034 1889.07 245.099 L1889.07 241.163 Q1891.41 240.261 1893.61 239.821 Q1895.81 239.358 1897.89 239.358 Q1903.52 239.358 1906.3 242.274 Q1909.07 245.191 1909.07 251.117 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M1943.15 229.937 Q1940.05 235.261 1938.54 240.469 Q1937.04 245.677 1937.04 251.024 Q1937.04 256.372 1938.54 261.626 Q1940.07 266.858 1943.15 272.159 L1939.44 272.159 Q1935.97 266.719 1934.24 261.464 Q1932.52 256.21 1932.52 251.024 Q1932.52 245.862 1934.24 240.631 Q1935.95 235.4 1939.44 229.937 L1943.15 229.937 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M1952.06 231.349 L1970.42 231.349 L1970.42 235.284 L1956.34 235.284 L1956.34 243.756 Q1957.36 243.409 1958.38 243.247 Q1959.4 243.062 1960.42 243.062 Q1966.2 243.062 1969.58 246.233 Q1972.96 249.404 1972.96 254.821 Q1972.96 260.399 1969.49 263.501 Q1966.02 266.58 1959.7 266.58 Q1957.52 266.58 1955.25 266.21 Q1953.01 265.839 1950.6 265.098 L1950.6 260.399 Q1952.68 261.534 1954.91 262.089 Q1957.13 262.645 1959.61 262.645 Q1963.61 262.645 1965.95 260.538 Q1968.29 258.432 1968.29 254.821 Q1968.29 251.21 1965.95 249.103 Q1963.61 246.997 1959.61 246.997 Q1957.73 246.997 1955.86 247.413 Q1954 247.83 1952.06 248.71 L1952.06 231.349 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M1992.17 249.497 Q1988.84 249.497 1986.92 251.279 Q1985.02 253.061 1985.02 256.186 Q1985.02 259.311 1986.92 261.094 Q1988.84 262.876 1992.17 262.876 Q1995.51 262.876 1997.43 261.094 Q1999.35 259.288 1999.35 256.186 Q1999.35 253.061 1997.43 251.279 Q1995.53 249.497 1992.17 249.497 M1987.5 247.506 Q1984.49 246.765 1982.8 244.705 Q1981.13 242.645 1981.13 239.682 Q1981.13 235.538 1984.07 233.131 Q1987.04 230.724 1992.17 230.724 Q1997.34 230.724 2000.28 233.131 Q2003.22 235.538 2003.22 239.682 Q2003.22 242.645 2001.53 244.705 Q1999.86 246.765 1996.87 247.506 Q2000.25 248.293 2002.13 250.585 Q2004.03 252.876 2004.03 256.186 Q2004.03 261.21 2000.95 263.895 Q1997.89 266.58 1992.17 266.58 Q1986.46 266.58 1983.38 263.895 Q1980.32 261.21 1980.32 256.186 Q1980.32 252.876 1982.22 250.585 Q1984.12 248.293 1987.5 247.506 M1985.79 240.122 Q1985.79 242.807 1987.45 244.312 Q1989.14 245.816 1992.17 245.816 Q1995.18 245.816 1996.87 244.312 Q1998.59 242.807 1998.59 240.122 Q1998.59 237.437 1996.87 235.932 Q1995.18 234.427 1992.17 234.427 Q1989.14 234.427 1987.45 235.932 Q1985.79 237.437 1985.79 240.122 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M2011.16 231.349 L2033.38 231.349 L2033.38 233.339 L2020.83 265.909 L2015.95 265.909 L2027.75 235.284 L2011.16 235.284 L2011.16 231.349 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M2046.53 261.973 L2062.85 261.973 L2062.85 265.909 L2040.9 265.909 L2040.9 261.973 Q2043.56 259.219 2048.15 254.589 Q2052.75 249.936 2053.93 248.594 Q2056.18 246.071 2057.06 244.335 Q2057.96 242.575 2057.96 240.886 Q2057.96 238.131 2056.02 236.395 Q2054.1 234.659 2050.99 234.659 Q2048.79 234.659 2046.34 235.423 Q2043.91 236.187 2041.13 237.737 L2041.13 233.015 Q2043.96 231.881 2046.41 231.302 Q2048.86 230.724 2050.9 230.724 Q2056.27 230.724 2059.47 233.409 Q2062.66 236.094 2062.66 240.585 Q2062.66 242.714 2061.85 244.636 Q2061.06 246.534 2058.96 249.126 Q2058.38 249.798 2055.28 253.015 Q2052.17 256.21 2046.53 261.973 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M2082.66 234.427 Q2079.05 234.427 2077.22 237.992 Q2075.41 241.534 2075.41 248.663 Q2075.41 255.77 2077.22 259.335 Q2079.05 262.876 2082.66 262.876 Q2086.29 262.876 2088.1 259.335 Q2089.93 255.77 2089.93 248.663 Q2089.93 241.534 2088.1 237.992 Q2086.29 234.427 2082.66 234.427 M2082.66 230.724 Q2088.47 230.724 2091.53 235.33 Q2094.6 239.913 2094.6 248.663 Q2094.6 257.39 2091.53 261.997 Q2088.47 266.58 2082.66 266.58 Q2076.85 266.58 2073.77 261.997 Q2070.72 257.39 2070.72 248.663 Q2070.72 239.913 2073.77 235.33 Q2076.85 230.724 2082.66 230.724 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M2106.85 261.973 L2123.17 261.973 L2123.17 265.909 L2101.22 265.909 L2101.22 261.973 Q2103.89 259.219 2108.47 254.589 Q2113.08 249.936 2114.26 248.594 Q2116.5 246.071 2117.38 244.335 Q2118.29 242.575 2118.29 240.886 Q2118.29 238.131 2116.34 236.395 Q2114.42 234.659 2111.32 234.659 Q2109.12 234.659 2106.66 235.423 Q2104.23 236.187 2101.46 237.737 L2101.46 233.015 Q2104.28 231.881 2106.73 231.302 Q2109.19 230.724 2111.22 230.724 Q2116.6 230.724 2119.79 233.409 Q2122.98 236.094 2122.98 240.585 Q2122.98 242.714 2122.17 244.636 Q2121.39 246.534 2119.28 249.126 Q2118.7 249.798 2115.6 253.015 Q2112.5 256.21 2106.85 261.973 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M2133.03 231.349 L2151.39 231.349 L2151.39 235.284 L2137.31 235.284 L2137.31 243.756 Q2138.33 243.409 2139.35 243.247 Q2140.37 243.062 2141.39 243.062 Q2147.17 243.062 2150.55 246.233 Q2153.93 249.404 2153.93 254.821 Q2153.93 260.399 2150.46 263.501 Q2146.99 266.58 2140.67 266.58 Q2138.49 266.58 2136.22 266.21 Q2133.98 265.839 2131.57 265.098 L2131.57 260.399 Q2133.66 261.534 2135.88 262.089 Q2138.1 262.645 2140.58 262.645 Q2144.58 262.645 2146.92 260.538 Q2149.26 258.432 2149.26 254.821 Q2149.26 251.21 2146.92 249.103 Q2144.58 246.997 2140.58 246.997 Q2138.7 246.997 2136.83 247.413 Q2134.97 247.83 2133.03 248.71 L2133.03 231.349 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M2173.72 246.765 Q2170.58 246.765 2168.72 248.918 Q2166.9 251.071 2166.9 254.821 Q2166.9 258.548 2168.72 260.723 Q2170.58 262.876 2173.72 262.876 Q2176.87 262.876 2178.7 260.723 Q2180.55 258.548 2180.55 254.821 Q2180.55 251.071 2178.7 248.918 Q2176.87 246.765 2173.72 246.765 M2183.01 232.113 L2183.01 236.372 Q2181.25 235.538 2179.44 235.099 Q2177.66 234.659 2175.9 234.659 Q2171.27 234.659 2168.82 237.784 Q2166.39 240.909 2166.04 247.228 Q2167.41 245.214 2169.47 244.149 Q2171.53 243.062 2174 243.062 Q2179.21 243.062 2182.22 246.233 Q2185.25 249.381 2185.25 254.821 Q2185.25 260.145 2182.1 263.362 Q2178.96 266.58 2173.72 266.58 Q2167.73 266.58 2164.56 261.997 Q2161.39 257.39 2161.39 248.663 Q2161.39 240.469 2165.28 235.608 Q2169.16 230.724 2175.72 230.724 Q2177.47 230.724 2179.26 231.071 Q2181.06 231.418 2183.01 232.113 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M2194.12 261.973 L2201.76 261.973 L2201.76 235.608 L2193.45 237.275 L2193.45 233.015 L2201.71 231.349 L2206.39 231.349 L2206.39 261.973 L2214.03 261.973 L2214.03 265.909 L2194.12 265.909 L2194.12 261.973 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M2222.2 229.937 L2225.9 229.937 Q2229.37 235.4 2231.09 240.631 Q2232.82 245.862 2232.82 251.024 Q2232.82 256.21 2231.09 261.464 Q2229.37 266.719 2225.9 272.159 L2222.2 272.159 Q2225.28 266.858 2226.78 261.626 Q2228.31 256.372 2228.31 251.024 Q2228.31 245.677 2226.78 240.469 Q2225.28 235.261 2222.2 229.937 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /></svg>



#### Modified initial states

The same values for the initial states are used for this simulation as for the simulation from the *realFMU* with the modified initial states.


```julia
xSimpleMod₀ = vcat(xMod₀, displacement)

simpleSimDataMod = simulate(simpleFMU, states, xSimpleMod₀, vrs, tStart, tStop, tSave)
fmiPlot(simpleFMU, vrs, simpleSimDataMod)
```

    ┌ Warning: fmi2Reset(_): Should be only called in FMU state `modelSetableFMUstate`.
    └ @ FMIImport C:\Users\thummeto\Documents\FMIImport.jl\src\FMI2_c.jl:219



<?xml version="1.0" encoding="utf-8"?>
<svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" width="600" height="400" viewBox="0 0 2400 1600">
<defs>
  <clipPath id="clip210">
    <rect x="0" y="0" width="2400" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip210)" d="
M0 1600 L2400 1600 L2400 0 L0 0  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip211">
    <rect x="480" y="0" width="1681" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip210)" d="
M149.191 1423.18 L2352.76 1423.18 L2352.76 47.2441 L149.191 47.2441  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip212">
    <rect x="149" y="47" width="2205" height="1377"/>
  </clipPath>
</defs>
<polyline clip-path="url(#clip212)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  211.556,1423.18 211.556,47.2441 
  "/>
<polyline clip-path="url(#clip212)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  731.265,1423.18 731.265,47.2441 
  "/>
<polyline clip-path="url(#clip212)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1250.97,1423.18 1250.97,47.2441 
  "/>
<polyline clip-path="url(#clip212)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1770.68,1423.18 1770.68,47.2441 
  "/>
<polyline clip-path="url(#clip212)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  2290.39,1423.18 2290.39,47.2441 
  "/>
<polyline clip-path="url(#clip210)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  149.191,1423.18 2352.76,1423.18 
  "/>
<polyline clip-path="url(#clip210)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  211.556,1423.18 211.556,1404.28 
  "/>
<polyline clip-path="url(#clip210)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  731.265,1423.18 731.265,1404.28 
  "/>
<polyline clip-path="url(#clip210)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1250.97,1423.18 1250.97,1404.28 
  "/>
<polyline clip-path="url(#clip210)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1770.68,1423.18 1770.68,1404.28 
  "/>
<polyline clip-path="url(#clip210)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  2290.39,1423.18 2290.39,1404.28 
  "/>
<path clip-path="url(#clip210)" d="M211.556 1454.1 Q207.945 1454.1 206.116 1457.66 Q204.311 1461.2 204.311 1468.33 Q204.311 1475.44 206.116 1479.01 Q207.945 1482.55 211.556 1482.55 Q215.19 1482.55 216.996 1479.01 Q218.824 1475.44 218.824 1468.33 Q218.824 1461.2 216.996 1457.66 Q215.19 1454.1 211.556 1454.1 M211.556 1450.39 Q217.366 1450.39 220.422 1455 Q223.5 1459.58 223.5 1468.33 Q223.5 1477.06 220.422 1481.67 Q217.366 1486.25 211.556 1486.25 Q205.746 1486.25 202.667 1481.67 Q199.612 1477.06 199.612 1468.33 Q199.612 1459.58 202.667 1455 Q205.746 1450.39 211.556 1450.39 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M721.647 1481.64 L729.285 1481.64 L729.285 1455.28 L720.975 1456.95 L720.975 1452.69 L729.239 1451.02 L733.915 1451.02 L733.915 1481.64 L741.554 1481.64 L741.554 1485.58 L721.647 1485.58 L721.647 1481.64 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1245.63 1481.64 L1261.95 1481.64 L1261.95 1485.58 L1240 1485.58 L1240 1481.64 Q1242.66 1478.89 1247.25 1474.26 Q1251.85 1469.61 1253.03 1468.27 Q1255.28 1465.74 1256.16 1464.01 Q1257.06 1462.25 1257.06 1460.56 Q1257.06 1457.8 1255.12 1456.07 Q1253.2 1454.33 1250.09 1454.33 Q1247.89 1454.33 1245.44 1455.09 Q1243.01 1455.86 1240.23 1457.41 L1240.23 1452.69 Q1243.06 1451.55 1245.51 1450.97 Q1247.96 1450.39 1250 1450.39 Q1255.37 1450.39 1258.57 1453.08 Q1261.76 1455.77 1261.76 1460.26 Q1261.76 1462.39 1260.95 1464.31 Q1260.16 1466.2 1258.06 1468.8 Q1257.48 1469.47 1254.38 1472.69 Q1251.27 1475.88 1245.63 1481.64 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1774.93 1466.95 Q1778.29 1467.66 1780.16 1469.93 Q1782.06 1472.2 1782.06 1475.53 Q1782.06 1480.65 1778.54 1483.45 Q1775.02 1486.25 1768.54 1486.25 Q1766.37 1486.25 1764.05 1485.81 Q1761.76 1485.39 1759.3 1484.54 L1759.3 1480.02 Q1761.25 1481.16 1763.56 1481.74 Q1765.88 1482.32 1768.4 1482.32 Q1772.8 1482.32 1775.09 1480.58 Q1777.41 1478.84 1777.41 1475.53 Q1777.41 1472.48 1775.25 1470.77 Q1773.12 1469.03 1769.3 1469.03 L1765.28 1469.03 L1765.28 1465.19 L1769.49 1465.19 Q1772.94 1465.19 1774.77 1463.82 Q1776.6 1462.43 1776.6 1459.84 Q1776.6 1457.18 1774.7 1455.77 Q1772.82 1454.33 1769.3 1454.33 Q1767.38 1454.33 1765.18 1454.75 Q1762.99 1455.16 1760.35 1456.04 L1760.35 1451.88 Q1763.01 1451.14 1765.32 1450.77 Q1767.66 1450.39 1769.72 1450.39 Q1775.05 1450.39 1778.15 1452.83 Q1781.25 1455.23 1781.25 1459.35 Q1781.25 1462.22 1779.61 1464.21 Q1777.96 1466.18 1774.93 1466.95 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M2293.4 1455.09 L2281.59 1473.54 L2293.4 1473.54 L2293.4 1455.09 M2292.17 1451.02 L2298.05 1451.02 L2298.05 1473.54 L2302.98 1473.54 L2302.98 1477.43 L2298.05 1477.43 L2298.05 1485.58 L2293.4 1485.58 L2293.4 1477.43 L2277.8 1477.43 L2277.8 1472.92 L2292.17 1451.02 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1199.28 1522.27 L1199.28 1532.4 L1211.35 1532.4 L1211.35 1536.95 L1199.28 1536.95 L1199.28 1556.3 Q1199.28 1560.66 1200.46 1561.9 Q1201.67 1563.14 1205.33 1563.14 L1211.35 1563.14 L1211.35 1568.04 L1205.33 1568.04 Q1198.55 1568.04 1195.97 1565.53 Q1193.4 1562.98 1193.4 1556.3 L1193.4 1536.95 L1189.1 1536.95 L1189.1 1532.4 L1193.4 1532.4 L1193.4 1522.27 L1199.28 1522.27 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1239.23 1518.52 L1252.72 1518.52 L1252.72 1523.07 L1245.09 1523.07 L1245.09 1572.09 L1252.72 1572.09 L1252.72 1576.64 L1239.23 1576.64 L1239.23 1518.52 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1287.93 1533.45 L1287.93 1538.98 Q1285.44 1537.71 1282.77 1537.07 Q1280.1 1536.44 1277.23 1536.44 Q1272.87 1536.44 1270.68 1537.77 Q1268.51 1539.11 1268.51 1541.79 Q1268.51 1543.82 1270.07 1545 Q1271.63 1546.15 1276.34 1547.2 L1278.35 1547.64 Q1284.58 1548.98 1287.19 1551.43 Q1289.84 1553.85 1289.84 1558.21 Q1289.84 1563.17 1285.89 1566.07 Q1281.97 1568.97 1275.1 1568.97 Q1272.23 1568.97 1269.12 1568.39 Q1266.03 1567.85 1262.59 1566.74 L1262.59 1560.69 Q1265.84 1562.38 1268.99 1563.24 Q1272.14 1564.07 1275.23 1564.07 Q1279.36 1564.07 1281.59 1562.66 Q1283.82 1561.23 1283.82 1558.65 Q1283.82 1556.27 1282.2 1554.99 Q1280.61 1553.72 1275.16 1552.54 L1273.13 1552.07 Q1267.68 1550.92 1265.26 1548.56 Q1262.85 1546.18 1262.85 1542.04 Q1262.85 1537.01 1266.41 1534.27 Q1269.98 1531.54 1276.53 1531.54 Q1279.78 1531.54 1282.64 1532.01 Q1285.51 1532.49 1287.93 1533.45 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1312.85 1518.52 L1312.85 1576.64 L1299.35 1576.64 L1299.35 1572.09 L1306.96 1572.09 L1306.96 1523.07 L1299.35 1523.07 L1299.35 1518.52 L1312.85 1518.52 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip212)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  149.191,1270.87 2352.76,1270.87 
  "/>
<polyline clip-path="url(#clip212)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  149.191,1003.08 2352.76,1003.08 
  "/>
<polyline clip-path="url(#clip212)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  149.191,735.289 2352.76,735.289 
  "/>
<polyline clip-path="url(#clip212)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  149.191,467.498 2352.76,467.498 
  "/>
<polyline clip-path="url(#clip212)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  149.191,199.707 2352.76,199.707 
  "/>
<polyline clip-path="url(#clip210)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  149.191,1423.18 149.191,47.2441 
  "/>
<polyline clip-path="url(#clip210)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  149.191,1270.87 168.089,1270.87 
  "/>
<polyline clip-path="url(#clip210)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  149.191,1003.08 168.089,1003.08 
  "/>
<polyline clip-path="url(#clip210)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  149.191,735.289 168.089,735.289 
  "/>
<polyline clip-path="url(#clip210)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  149.191,467.498 168.089,467.498 
  "/>
<polyline clip-path="url(#clip210)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  149.191,199.707 168.089,199.707 
  "/>
<path clip-path="url(#clip210)" d="M50.9921 1271.32 L80.6679 1271.32 L80.6679 1275.26 L50.9921 1275.26 L50.9921 1271.32 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M103.608 1257.67 L91.8021 1276.11 L103.608 1276.11 L103.608 1257.67 M102.381 1253.59 L108.26 1253.59 L108.26 1276.11 L113.191 1276.11 L113.191 1280 L108.26 1280 L108.26 1288.15 L103.608 1288.15 L103.608 1280 L88.0058 1280 L88.0058 1275.49 L102.381 1253.59 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M53.0754 1003.53 L82.7512 1003.53 L82.7512 1007.47 L53.0754 1007.47 L53.0754 1003.53 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M96.8715 1016.43 L113.191 1016.43 L113.191 1020.36 L91.2465 1020.36 L91.2465 1016.43 Q93.9086 1013.67 98.4919 1009.04 Q103.098 1004.39 104.279 1003.05 Q106.524 1000.52 107.404 998.786 Q108.307 997.027 108.307 995.337 Q108.307 992.583 106.362 990.847 Q104.441 989.111 101.339 989.111 Q99.14 989.111 96.6863 989.874 Q94.2558 990.638 91.478 992.189 L91.478 987.467 Q94.3021 986.333 96.7558 985.754 Q99.2095 985.175 101.246 985.175 Q106.617 985.175 109.811 987.861 Q113.006 990.546 113.006 995.036 Q113.006 997.166 112.196 999.087 Q111.408 1000.99 109.302 1003.58 Q108.723 1004.25 105.621 1007.47 Q102.52 1010.66 96.8715 1016.43 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M101.246 721.088 Q97.6354 721.088 95.8067 724.653 Q94.0012 728.194 94.0012 735.324 Q94.0012 742.431 95.8067 745.995 Q97.6354 749.537 101.246 749.537 Q104.881 749.537 106.686 745.995 Q108.515 742.431 108.515 735.324 Q108.515 728.194 106.686 724.653 Q104.881 721.088 101.246 721.088 M101.246 717.384 Q107.057 717.384 110.112 721.991 Q113.191 726.574 113.191 735.324 Q113.191 744.051 110.112 748.657 Q107.057 753.241 101.246 753.241 Q95.4363 753.241 92.3576 748.657 Q89.3021 744.051 89.3021 735.324 Q89.3021 726.574 92.3576 721.991 Q95.4363 717.384 101.246 717.384 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M96.8715 480.843 L113.191 480.843 L113.191 484.778 L91.2465 484.778 L91.2465 480.843 Q93.9086 478.089 98.4919 473.459 Q103.098 468.806 104.279 467.464 Q106.524 464.94 107.404 463.204 Q108.307 461.445 108.307 459.755 Q108.307 457.001 106.362 455.265 Q104.441 453.528 101.339 453.528 Q99.14 453.528 96.6863 454.292 Q94.2558 455.056 91.478 456.607 L91.478 451.885 Q94.3021 450.751 96.7558 450.172 Q99.2095 449.593 101.246 449.593 Q106.617 449.593 109.811 452.278 Q113.006 454.964 113.006 459.454 Q113.006 461.584 112.196 463.505 Q111.408 465.403 109.302 467.996 Q108.723 468.667 105.621 471.885 Q102.52 475.079 96.8715 480.843 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M103.608 186.501 L91.8021 204.95 L103.608 204.95 L103.608 186.501 M102.381 182.427 L108.26 182.427 L108.26 204.95 L113.191 204.95 L113.191 208.839 L108.26 208.839 L108.26 216.987 L103.608 216.987 L103.608 208.839 L88.0058 208.839 L88.0058 204.325 L102.381 182.427 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip212)" style="stroke:#009af9; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  211.556,601.394 216.753,603.395 221.95,605.381 227.147,607.349 232.344,609.297 237.541,611.224 242.738,613.128 247.936,615.007 253.133,616.859 258.33,618.682 
  263.527,620.475 268.724,622.235 273.921,623.962 279.118,625.652 284.315,627.305 289.512,628.919 294.709,630.492 299.906,632.022 305.103,633.508 310.301,634.949 
  315.498,636.343 320.695,637.689 325.892,638.985 331.089,640.23 336.286,641.423 341.483,642.562 346.68,643.647 351.877,644.676 357.074,645.649 362.271,646.563 
  367.469,647.42 372.666,648.216 377.863,648.953 383.06,649.629 388.257,650.243 393.454,650.794 398.651,651.283 403.848,651.709 409.045,652.071 414.242,652.369 
  419.439,652.602 424.636,652.771 429.834,652.875 435.031,652.915 440.228,652.889 445.425,652.799 450.622,652.643 455.819,652.423 461.016,652.139 466.213,651.791 
  471.41,651.378 476.607,650.903 481.804,650.364 487.002,649.763 492.199,649.101 497.396,648.377 502.593,647.593 507.79,646.749 512.987,645.846 518.184,644.886 
  523.381,643.869 528.578,642.796 533.775,641.668 538.972,640.486 544.17,639.252 549.367,637.967 554.564,636.632 559.761,635.248 564.958,633.817 570.155,632.34 
  575.352,630.819 580.549,629.255 585.746,627.65 590.943,626.005 596.14,624.322 601.337,622.603 606.535,620.849 611.732,619.063 616.929,617.245 622.126,615.398 
  627.323,613.524 632.52,611.624 637.717,609.7 642.914,607.755 648.111,605.79 653.308,603.807 658.505,601.809 663.703,599.796 668.9,597.772 674.097,595.738 
  679.294,593.696 684.491,591.649 689.688,589.598 694.885,587.546 700.082,585.493 705.279,583.444 710.476,581.398 715.673,579.36 720.87,577.33 726.068,575.311 
  731.265,573.305 736.462,571.313 741.659,569.338 746.856,567.382 752.053,565.446 757.25,563.533 762.447,561.644 767.644,559.782 772.841,557.948 778.038,556.144 
  783.236,554.372 788.433,552.634 793.63,550.931 798.827,549.265 804.024,547.638 809.221,546.051 814.418,544.506 819.615,543.005 824.812,541.549 830.009,540.139 
  835.206,538.777 840.403,537.464 845.601,536.202 850.798,534.992 855.995,533.834 861.192,532.731 866.389,531.683 871.586,530.692 876.783,529.757 881.98,528.881 
  887.177,528.064 892.374,527.307 897.571,526.611 902.769,525.976 907.966,525.403 913.163,524.893 918.36,524.446 923.557,524.062 928.754,523.743 933.951,523.488 
  939.148,523.297 944.345,523.171 949.542,523.109 954.739,523.113 959.936,523.181 965.134,523.315 970.331,523.513 975.528,523.775 980.725,524.102 985.922,524.493 
  991.119,524.947 996.316,525.464 1001.51,526.044 1006.71,526.686 1011.91,527.389 1017.1,528.152 1022.3,528.976 1027.5,529.858 1032.7,530.799 1037.89,531.797 
  1043.09,532.851 1048.29,533.96 1053.48,535.123 1058.68,536.34 1063.88,537.607 1069.08,538.925 1074.27,540.293 1079.47,541.708 1084.67,543.169 1089.86,544.675 
  1095.06,546.224 1100.26,547.815 1105.45,549.447 1110.65,551.117 1115.85,552.823 1121.05,554.565 1126.24,556.341 1131.44,558.148 1136.64,559.984 1141.83,561.849 
  1147.03,563.74 1152.23,565.656 1157.43,567.593 1162.62,569.551 1167.82,571.528 1173.02,573.521 1178.21,575.528 1183.41,577.548 1188.61,579.578 1193.81,581.617 
  1199,583.662 1204.2,585.711 1209.4,587.763 1214.59,589.815 1219.79,591.865 1224.99,593.911 1230.19,595.952 1235.38,597.984 1240.58,600.006 1245.78,602.017 
  1250.97,604.013 1256.17,605.993 1261.37,607.956 1266.56,609.898 1271.76,611.819 1276.96,613.715 1282.16,615.586 1287.35,617.429 1292.55,619.243 1297.75,621.026 
  1302.94,622.775 1308.14,624.49 1313.34,626.169 1318.54,627.809 1323.73,629.409 1328.93,630.968 1334.13,632.484 1339.32,633.955 1344.52,635.381 1349.72,636.759 
  1354.92,638.088 1360.11,639.368 1365.31,640.596 1370.51,641.771 1375.7,642.892 1380.9,643.959 1386.1,644.97 1391.29,645.923 1396.49,646.819 1401.69,647.656 
  1406.89,648.433 1412.08,649.15 1417.28,649.806 1422.48,650.4 1427.67,650.931 1432.87,651.4 1438.07,651.805 1443.27,652.146 1448.46,652.424 1453.66,652.636 
  1458.86,652.785 1464.05,652.868 1469.25,652.886 1474.45,652.84 1479.65,652.729 1484.84,652.553 1490.04,652.313 1495.24,652.008 1500.43,651.639 1505.63,651.207 
  1510.83,650.711 1516.02,650.153 1521.22,649.532 1526.42,648.85 1531.62,648.107 1536.81,647.304 1542.01,646.442 1547.21,645.522 1552.4,644.544 1557.6,643.509 
  1562.8,642.419 1568,641.274 1573.19,640.076 1578.39,638.826 1583.59,637.525 1588.78,636.175 1593.98,634.777 1599.18,633.331 1604.38,631.841 1609.57,630.307 
  1614.77,628.73 1619.97,627.113 1625.16,625.456 1630.36,623.762 1635.56,622.033 1640.75,620.269 1645.95,618.473 1651.15,616.647 1656.35,614.792 1661.54,612.91 
  1666.74,611.003 1671.94,609.074 1677.13,607.123 1682.33,605.153 1687.53,603.166 1692.73,601.164 1697.92,599.149 1703.12,597.122 1708.32,595.087 1713.51,593.044 
  1718.71,590.996 1723.91,588.946 1729.11,586.894 1734.3,584.844 1739.5,582.797 1744.7,580.755 1749.89,578.72 1755.09,576.694 1760.29,574.68 1765.49,572.679 
  1770.68,570.694 1775.88,568.726 1781.08,566.777 1786.27,564.849 1791.47,562.945 1796.67,561.065 1801.86,559.213 1807.06,557.389 1812.26,555.596 1817.46,553.836 
  1822.65,552.109 1827.85,550.419 1833.05,548.766 1838.24,547.152 1843.44,545.579 1848.64,544.049 1853.84,542.563 1859.03,541.122 1864.23,539.728 1869.43,538.382 
  1874.62,537.086 1879.82,535.84 1885.02,534.647 1890.22,533.508 1895.41,532.423 1900.61,531.393 1905.81,530.42 1911,529.505 1916.2,528.648 1921.4,527.85 
  1926.59,527.113 1931.79,526.436 1936.99,525.821 1942.19,525.269 1947.38,524.779 1952.58,524.352 1957.78,523.989 1962.97,523.69 1968.17,523.455 1973.37,523.285 
  1978.57,523.179 1983.76,523.139 1988.96,523.163 1994.16,523.252 1999.35,523.406 2004.55,523.624 2009.75,523.907 2014.95,524.254 2020.14,524.664 2025.34,525.138 
  2030.54,525.675 2035.73,526.274 2040.93,526.935 2046.13,527.657 2051.32,528.439 2056.52,529.281 2061.72,530.182 2066.92,531.14 2072.11,532.155 2077.31,533.226 
  2082.51,534.352 2087.7,535.532 2092.9,536.764 2098.1,538.047 2103.3,539.38 2108.49,540.762 2113.69,542.191 2118.89,543.666 2124.08,545.185 2129.28,546.747 
  2134.48,548.35 2139.68,549.993 2144.87,551.674 2150.07,553.391 2155.27,555.143 2160.46,556.927 2165.66,558.743 2170.86,560.588 2176.05,562.46 2181.25,564.358 
  2186.45,566.28 2191.65,568.223 2196.84,570.186 2202.04,572.167 2207.24,574.163 2212.43,576.174 2217.63,578.196 2222.83,580.228 2228.03,582.268 2233.22,584.314 
  2238.42,586.363 2243.62,588.414 2248.81,590.464 2254.01,592.512 2259.21,594.556 2264.41,596.593 2269.6,598.621 2274.8,600.639 2280,602.644 2285.19,604.634 
  2290.39,606.608 
  "/>
<polyline clip-path="url(#clip212)" style="stroke:#e26f46; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  211.556,936.133 216.753,934.689 221.95,933.051 227.147,931.215 232.344,929.183 237.541,926.958 242.738,924.542 247.936,921.937 253.133,919.145 258.33,916.169 
  263.527,913.012 268.724,909.677 273.921,906.168 279.118,902.488 284.315,898.64 289.512,894.629 294.709,890.459 299.906,886.134 305.103,881.658 310.301,877.036 
  315.498,872.272 320.695,867.371 325.892,862.338 331.089,857.178 336.286,851.896 341.483,846.497 346.68,840.987 351.877,835.372 357.074,829.655 362.271,823.845 
  367.469,817.946 372.666,811.964 377.863,805.905 383.06,799.775 388.257,793.581 393.454,787.329 398.651,781.025 403.848,774.675 409.045,768.286 414.242,761.864 
  419.439,755.415 424.636,748.946 429.834,742.463 435.031,735.973 440.228,729.483 445.425,722.998 450.622,716.526 455.819,710.072 461.016,703.644 466.213,697.247 
  471.41,690.888 476.607,684.573 481.804,678.309 487.002,672.102 492.199,665.958 497.396,659.883 502.593,653.884 507.79,647.966 512.987,642.136 518.184,636.398 
  523.381,630.76 528.578,625.227 533.775,619.804 538.972,614.496 544.17,609.309 549.367,604.248 554.564,599.318 559.761,594.525 564.958,589.872 570.155,585.364 
  575.352,581.006 580.549,576.803 585.746,572.758 590.943,568.875 596.14,565.159 601.337,561.613 606.535,558.241 611.732,555.045 616.929,552.03 622.126,549.198 
  627.323,546.552 632.52,544.095 637.717,541.829 642.914,539.756 648.111,537.879 653.308,536.2 658.505,534.719 663.703,533.439 668.9,532.361 674.097,531.486 
  679.294,530.815 684.491,530.349 689.688,530.087 694.885,530.03 700.082,530.179 705.279,530.534 710.476,531.093 715.673,531.856 720.87,532.823 726.068,533.992 
  731.265,535.362 736.462,536.932 741.659,538.701 746.856,540.667 752.053,542.827 757.25,545.179 762.447,547.721 767.644,550.451 772.841,553.366 778.038,556.463 
  783.236,559.739 788.433,563.19 793.63,566.813 798.827,570.605 804.024,574.56 809.221,578.677 814.418,582.951 819.615,587.377 824.812,591.951 830.009,596.668 
  835.206,601.523 840.403,606.512 845.601,611.63 850.798,616.872 855.995,622.232 861.192,627.706 866.389,633.286 871.586,638.969 876.783,644.748 881.98,650.617 
  887.177,656.571 892.374,662.604 897.571,668.71 902.769,674.882 907.966,681.114 913.163,687.401 918.36,693.735 923.557,700.111 928.754,706.522 933.951,712.962 
  939.148,719.424 944.345,725.902 949.542,732.389 954.739,738.879 959.936,745.366 965.134,751.842 970.331,758.302 975.528,764.739 980.725,771.146 985.922,777.518 
  991.119,783.847 996.316,790.127 1001.51,796.353 1006.71,802.518 1011.91,808.615 1017.1,814.639 1022.3,820.583 1027.5,826.443 1032.7,832.211 1037.89,837.882 
  1043.09,843.45 1048.29,848.91 1053.48,854.257 1058.68,859.485 1063.88,864.588 1069.08,869.562 1074.27,874.402 1079.47,879.103 1084.67,883.659 1089.86,888.068 
  1095.06,892.323 1100.26,896.422 1105.45,900.359 1110.65,904.132 1115.85,907.735 1121.05,911.166 1126.24,914.421 1131.44,917.497 1136.64,920.391 1141.83,923.099 
  1147.03,925.62 1152.23,927.95 1157.43,930.088 1162.62,932.031 1167.82,933.777 1173.02,935.325 1178.21,936.673 1183.41,937.819 1188.61,938.763 1193.81,939.503 
  1199,940.038 1204.2,940.37 1209.4,940.496 1214.59,940.417 1219.79,940.133 1224.99,939.644 1230.19,938.95 1235.38,938.053 1240.58,936.953 1245.78,935.652 
  1250.97,934.151 1256.17,932.45 1261.37,930.552 1266.56,928.459 1271.76,926.172 1276.96,923.695 1282.16,921.03 1287.35,918.179 1292.55,915.145 1297.75,911.931 
  1302.94,908.54 1308.14,904.976 1313.34,901.243 1318.54,897.344 1323.73,893.282 1328.93,889.063 1334.13,884.69 1339.32,880.167 1344.52,875.5 1349.72,870.692 
  1354.92,865.749 1360.11,860.676 1365.31,855.478 1370.51,850.158 1375.7,844.725 1380.9,839.181 1386.1,833.534 1391.29,827.789 1396.49,821.952 1401.69,816.027 
  1406.89,810.022 1412.08,803.943 1417.28,797.794 1422.48,791.583 1427.67,785.316 1432.87,778.999 1438.07,772.639 1443.27,766.241 1448.46,759.811 1453.66,753.358 
  1458.86,746.886 1464.05,740.403 1469.25,733.915 1474.45,727.429 1479.65,720.95 1484.84,714.485 1490.04,708.042 1495.24,701.625 1500.43,695.243 1505.63,688.9 
  1510.83,682.604 1516.02,676.361 1521.22,670.176 1526.42,664.057 1531.62,658.009 1536.81,652.038 1542.01,646.151 1547.21,640.352 1552.4,634.649 1557.6,629.046 
  1562.8,623.55 1568,618.166 1573.19,612.898 1578.39,607.753 1583.59,602.736 1588.78,597.851 1593.98,593.103 1599.18,588.498 1604.38,584.04 1609.57,579.733 
  1614.77,575.581 1619.97,571.589 1625.16,567.761 1630.36,564.101 1635.56,560.612 1640.75,557.297 1645.95,554.161 1651.15,551.205 1656.35,548.434 1661.54,545.849 
  1666.74,543.454 1671.94,541.251 1677.13,539.242 1682.33,537.429 1687.53,535.814 1692.73,534.398 1697.92,533.183 1703.12,532.17 1708.32,531.361 1713.51,530.755 
  1718.71,530.354 1723.91,530.158 1729.11,530.167 1734.3,530.381 1739.5,530.8 1744.7,531.424 1749.89,532.251 1755.09,533.282 1760.29,534.514 1765.49,535.947 
  1770.68,537.579 1775.88,539.409 1781.08,541.436 1786.27,543.656 1791.47,546.067 1796.67,548.668 1801.86,551.456 1807.06,554.426 1812.26,557.578 1817.46,560.908 
  1822.65,564.412 1827.85,568.088 1833.05,571.93 1838.24,575.935 1843.44,580.1 1848.64,584.419 1853.84,588.89 1859.03,593.508 1864.23,598.267 1869.43,603.163 
  1874.62,608.191 1879.82,613.346 1885.02,618.623 1890.22,624.016 1895.41,629.521 1900.61,635.132 1905.81,640.843 1911,646.648 1916.2,652.542 1921.4,658.518 
  1926.59,664.572 1931.79,670.696 1936.99,676.884 1942.19,683.131 1947.38,689.43 1952.58,695.775 1957.78,702.159 1962.97,708.576 1968.17,715.021 1973.37,721.485 
  1978.57,727.963 1983.76,734.449 1988.96,740.935 1994.16,747.415 1999.35,753.883 2004.55,760.333 2009.75,766.758 2014.95,773.151 2020.14,779.506 2025.34,785.817 
  2030.54,792.078 2035.73,798.281 2040.93,804.422 2046.13,810.493 2051.32,816.489 2056.52,822.404 2061.72,828.232 2066.92,833.967 2072.11,839.603 2077.31,845.135 
  2082.51,850.557 2087.7,855.863 2092.9,861.049 2098.1,866.109 2103.3,871.039 2108.49,875.832 2113.69,880.485 2118.89,884.993 2124.08,889.351 2129.28,893.555 
  2134.48,897.601 2139.68,901.484 2144.87,905.201 2150.07,908.749 2155.27,912.122 2160.46,915.319 2165.66,918.336 2170.86,921.169 2176.05,923.817 2181.25,926.277 
  2186.45,928.545 2191.65,930.62 2196.84,932.5 2202.04,934.182 2207.24,935.666 2212.43,936.949 2217.63,938.031 2222.83,938.91 2228.03,939.584 2233.22,940.055 
  2238.42,940.321 2243.62,940.382 2248.81,940.238 2254.01,939.89 2259.21,939.336 2264.41,938.578 2269.6,937.617 2274.8,936.454 2280,935.089 2285.19,933.525 
  2290.39,931.763 
  "/>
<polyline clip-path="url(#clip212)" style="stroke:#3da44d; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  211.556,601.394 216.753,581.384 221.95,561.527 227.147,541.847 232.344,522.361 237.541,503.089 242.738,484.05 247.936,465.263 253.133,446.744 258.33,428.512 
  263.527,410.584 268.724,392.978 273.921,375.714 279.118,358.808 284.315,342.279 289.512,326.142 294.709,310.415 299.906,295.112 305.103,280.249 310.301,265.839 
  315.498,251.899 320.695,238.444 325.892,225.483 331.089,213.032 336.286,201.104 341.483,189.71 346.68,178.862 351.877,168.57 357.074,158.845 362.271,149.697 
  367.469,141.135 372.666,133.167 377.863,125.802 383.06,119.046 388.257,112.906 393.454,107.389 398.651,102.5 403.848,98.2435 409.045,94.624 414.242,91.6451 
  419.439,89.3098 424.636,87.6204 429.834,86.5787 435.031,86.1857 440.228,86.4417 445.425,87.3465 450.622,88.8993 455.819,91.0984 461.016,93.9418 466.213,97.4264 
  471.41,101.549 476.607,106.305 481.804,111.691 487.002,117.7 492.199,124.327 497.396,131.564 502.593,139.406 507.79,147.843 512.987,156.868 518.184,166.472 
  523.381,176.644 528.578,187.374 533.775,198.653 538.972,210.468 544.17,222.808 549.367,235.66 554.564,249.012 559.761,262.85 564.958,277.161 570.155,291.929 
  575.352,307.141 580.549,322.781 585.746,338.833 590.943,355.282 596.14,372.111 601.337,389.303 606.535,406.841 611.732,424.707 616.929,442.884 622.126,461.353 
  627.323,480.097 632.52,499.095 637.717,518.33 642.914,537.782 648.111,557.432 653.308,577.259 658.505,597.244 663.703,617.368 668.9,637.61 674.097,657.95 
  679.294,678.368 684.491,698.842 689.688,719.352 694.885,739.877 700.082,760.399 705.279,780.896 710.476,801.347 715.673,821.732 720.87,842.03 726.068,862.221 
  731.265,882.285 736.462,902.202 741.659,921.952 746.856,941.515 752.053,960.872 757.25,980.003 762.447,998.888 767.644,1017.51 772.841,1035.85 778.038,1053.89 
  783.236,1071.61 788.433,1088.99 793.63,1106.02 798.827,1122.68 804.024,1138.95 809.221,1154.82 814.418,1170.27 819.615,1185.28 824.812,1199.84 830.009,1213.94 
  835.206,1227.56 840.403,1240.69 845.601,1253.31 850.798,1265.42 855.995,1276.99 861.192,1288.02 866.389,1298.5 871.586,1308.42 876.783,1317.76 881.98,1326.52 
  887.177,1334.69 892.374,1342.26 897.571,1349.22 902.769,1355.57 907.966,1361.3 913.163,1366.4 918.36,1370.87 923.557,1374.71 928.754,1377.9 933.951,1380.46 
  939.148,1382.36 944.345,1383.63 949.542,1384.24 954.739,1384.2 959.936,1383.52 965.134,1382.18 970.331,1380.2 975.528,1377.58 980.725,1374.31 985.922,1370.4 
  991.119,1365.86 996.316,1360.69 1001.51,1354.89 1006.71,1348.48 1011.91,1341.44 1017.1,1333.81 1022.3,1325.57 1027.5,1316.75 1032.7,1307.34 1037.89,1297.36 
  1043.09,1286.82 1048.29,1275.73 1053.48,1264.1 1058.68,1251.94 1063.88,1239.26 1069.08,1226.08 1074.27,1212.41 1079.47,1198.26 1084.67,1183.64 1089.86,1168.58 
  1095.06,1153.09 1100.26,1137.18 1105.45,1120.87 1110.65,1104.17 1115.85,1087.1 1121.05,1069.68 1126.24,1051.93 1131.44,1033.86 1136.64,1015.49 1141.83,996.84 
  1147.03,977.929 1152.23,958.776 1157.43,939.4 1162.62,919.819 1167.82,900.055 1173.02,880.126 1178.21,860.053 1183.41,839.854 1188.61,819.551 1193.81,799.163 
  1199,778.712 1204.2,758.218 1209.4,737.701 1214.59,717.182 1219.79,696.681 1224.99,676.219 1230.19,655.815 1235.38,635.491 1240.58,615.268 1245.78,595.164 
  1250.97,575.201 1256.17,555.398 1261.37,535.775 1266.56,516.351 1271.76,497.147 1276.96,478.18 1282.16,459.471 1287.35,441.038 1292.55,422.899 1297.75,405.073 
  1302.94,387.577 1308.14,370.429 1313.34,353.646 1318.54,337.245 1323.73,321.241 1328.93,305.652 1334.13,290.492 1339.32,275.778 1344.52,261.523 1349.72,247.742 
  1354.92,234.449 1360.11,221.656 1365.31,209.376 1370.51,197.624 1375.7,186.409 1380.9,175.743 1386.1,165.636 1391.29,156.099 1396.49,147.141 1401.69,138.771 
  1406.89,130.999 1412.08,123.83 1417.28,117.274 1422.48,111.335 1427.67,106.02 1432.87,101.334 1438.07,97.2809 1443.27,93.8676 1448.46,91.0958 1453.66,88.9682 
  1458.86,87.4868 1464.05,86.6529 1469.25,86.4671 1474.45,86.9294 1479.65,88.0418 1484.84,89.8014 1490.04,92.2066 1495.24,95.2547 1500.43,98.9425 1505.63,103.266 
  1510.83,108.221 1516.02,113.804 1521.22,120.009 1526.42,126.83 1531.62,134.258 1536.81,142.288 1542.01,150.91 1547.21,160.115 1552.4,169.896 1557.6,180.244 
  1562.8,191.147 1568,202.593 1573.19,214.572 1578.39,227.071 1583.59,240.078 1588.78,253.58 1593.98,267.565 1599.18,282.017 1604.38,296.923 1609.57,312.266 
  1614.77,328.033 1619.97,344.205 1625.16,360.769 1630.36,377.709 1635.56,395.006 1640.75,412.643 1645.95,430.602 1651.15,448.866 1656.35,467.415 1661.54,486.233 
  1666.74,505.3 1671.94,524.597 1677.13,544.105 1682.33,563.803 1687.53,583.673 1692.73,603.694 1697.92,623.846 1703.12,644.11 1708.32,664.466 1713.51,684.892 
  1718.71,705.368 1723.91,725.875 1729.11,746.39 1734.3,766.893 1739.5,787.366 1744.7,807.786 1749.89,828.134 1755.09,848.389 1760.29,868.53 1765.49,888.538 
  1770.68,908.393 1775.88,928.074 1781.08,947.563 1786.27,966.839 1791.47,985.884 1796.67,1004.68 1801.86,1023.2 1807.06,1041.44 1812.26,1059.37 1817.46,1076.97 
  1822.65,1094.24 1827.85,1111.14 1833.05,1127.67 1838.24,1143.81 1843.44,1159.54 1848.64,1174.84 1853.84,1189.71 1859.03,1204.11 1864.23,1218.06 1869.43,1231.51 
  1874.62,1244.47 1879.82,1256.93 1885.02,1268.86 1890.22,1280.26 1895.41,1291.11 1900.61,1301.4 1905.81,1311.13 1911,1320.29 1916.2,1328.86 1921.4,1336.83 
  1926.59,1344.21 1931.79,1350.97 1936.99,1357.12 1942.19,1362.65 1947.38,1367.55 1952.58,1371.81 1957.78,1375.45 1962.97,1378.44 1968.17,1380.78 1973.37,1382.48 
  1978.57,1383.54 1983.76,1383.95 1988.96,1383.7 1994.16,1382.81 1999.35,1381.28 2004.55,1379.09 2009.75,1376.26 2014.95,1372.79 2020.14,1368.69 2025.34,1363.95 
  2030.54,1358.58 2035.73,1352.59 2040.93,1345.98 2046.13,1338.76 2051.32,1330.94 2056.52,1322.52 2061.72,1313.51 2066.92,1303.93 2072.11,1293.78 2077.31,1283.07 
  2082.51,1271.81 2087.7,1260.01 2092.9,1247.69 2098.1,1234.86 2103.3,1221.53 2108.49,1207.71 2113.69,1193.42 2118.89,1178.68 2124.08,1163.48 2129.28,1147.86 
  2134.48,1131.83 2139.68,1115.4 2144.87,1098.6 2150.07,1081.42 2155.27,1063.91 2160.46,1046.06 2165.66,1027.9 2170.86,1009.45 2176.05,990.73 2181.25,971.752 
  2186.45,952.536 2191.65,933.103 2196.84,913.473 2202.04,893.664 2207.24,873.698 2212.43,853.594 2217.63,833.371 2222.83,813.05 2228.03,792.651 2233.22,772.195 
  2238.42,751.703 2243.62,731.194 2248.81,710.689 2254.01,690.21 2259.21,669.775 2264.41,649.405 2269.6,629.122 2274.8,608.945 2280,588.895 2285.19,568.991 
  2290.39,549.253 
  "/>
<path clip-path="url(#clip210)" d="
M1550.58 300.469 L2279.3 300.469 L2279.3 93.1086 L1550.58 93.1086  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<polyline clip-path="url(#clip210)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1550.58,300.469 2279.3,300.469 2279.3,93.1086 1550.58,93.1086 1550.58,300.469 
  "/>
<polyline clip-path="url(#clip210)" style="stroke:#009af9; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1575.06,144.949 1721.97,144.949 
  "/>
<path clip-path="url(#clip210)" d="M1766.8 141.28 Q1768.39 138.409 1770.62 137.044 Q1772.84 135.678 1775.85 135.678 Q1779.9 135.678 1782.1 138.525 Q1784.3 141.349 1784.3 146.581 L1784.3 162.229 L1780.01 162.229 L1780.01 146.719 Q1780.01 142.993 1778.69 141.187 Q1777.38 139.382 1774.67 139.382 Q1771.36 139.382 1769.44 141.581 Q1767.51 143.78 1767.51 147.576 L1767.51 162.229 L1763.23 162.229 L1763.23 146.719 Q1763.23 142.969 1761.91 141.187 Q1760.59 139.382 1757.84 139.382 Q1754.57 139.382 1752.65 141.604 Q1750.73 143.803 1750.73 147.576 L1750.73 162.229 L1746.45 162.229 L1746.45 136.303 L1750.73 136.303 L1750.73 140.331 Q1752.19 137.946 1754.23 136.812 Q1756.26 135.678 1759.07 135.678 Q1761.89 135.678 1763.86 137.113 Q1765.85 138.548 1766.8 141.28 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1804.57 149.196 Q1799.41 149.196 1797.42 150.377 Q1795.43 151.557 1795.43 154.405 Q1795.43 156.673 1796.91 158.016 Q1798.42 159.335 1800.99 159.335 Q1804.53 159.335 1806.66 156.835 Q1808.81 154.312 1808.81 150.145 L1808.81 149.196 L1804.57 149.196 M1813.07 147.437 L1813.07 162.229 L1808.81 162.229 L1808.81 158.293 Q1807.35 160.655 1805.18 161.789 Q1803 162.9 1799.85 162.9 Q1795.87 162.9 1793.51 160.678 Q1791.17 158.432 1791.17 154.682 Q1791.17 150.307 1794.09 148.085 Q1797.03 145.863 1802.84 145.863 L1808.81 145.863 L1808.81 145.446 Q1808.81 142.507 1806.87 140.909 Q1804.94 139.289 1801.45 139.289 Q1799.23 139.289 1797.12 139.821 Q1795.01 140.354 1793.07 141.419 L1793.07 137.483 Q1795.41 136.581 1797.61 136.141 Q1799.81 135.678 1801.89 135.678 Q1807.51 135.678 1810.29 138.594 Q1813.07 141.511 1813.07 147.437 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1838.37 137.067 L1838.37 141.094 Q1836.56 140.169 1834.62 139.706 Q1832.68 139.243 1830.59 139.243 Q1827.42 139.243 1825.82 140.215 Q1824.25 141.187 1824.25 143.131 Q1824.25 144.613 1825.38 145.469 Q1826.52 146.303 1829.94 147.067 L1831.4 147.391 Q1835.94 148.363 1837.84 150.145 Q1839.76 151.905 1839.76 155.076 Q1839.76 158.687 1836.89 160.793 Q1834.04 162.9 1829.04 162.9 Q1826.96 162.9 1824.69 162.483 Q1822.44 162.09 1819.94 161.28 L1819.94 156.881 Q1822.31 158.108 1824.6 158.733 Q1826.89 159.335 1829.13 159.335 Q1832.14 159.335 1833.76 158.317 Q1835.38 157.275 1835.38 155.4 Q1835.38 153.664 1834.2 152.738 Q1833.05 151.812 1829.09 150.956 L1827.61 150.608 Q1823.65 149.775 1821.89 148.062 Q1820.13 146.326 1820.13 143.317 Q1820.13 139.659 1822.72 137.669 Q1825.31 135.678 1830.08 135.678 Q1832.44 135.678 1834.53 136.025 Q1836.61 136.372 1838.37 137.067 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1863.07 137.067 L1863.07 141.094 Q1861.26 140.169 1859.32 139.706 Q1857.37 139.243 1855.29 139.243 Q1852.12 139.243 1850.52 140.215 Q1848.95 141.187 1848.95 143.131 Q1848.95 144.613 1850.08 145.469 Q1851.22 146.303 1854.64 147.067 L1856.1 147.391 Q1860.64 148.363 1862.54 150.145 Q1864.46 151.905 1864.46 155.076 Q1864.46 158.687 1861.59 160.793 Q1858.74 162.9 1853.74 162.9 Q1851.66 162.9 1849.39 162.483 Q1847.14 162.09 1844.64 161.28 L1844.64 156.881 Q1847 158.108 1849.3 158.733 Q1851.59 159.335 1853.83 159.335 Q1856.84 159.335 1858.46 158.317 Q1860.08 157.275 1860.08 155.4 Q1860.08 153.664 1858.9 152.738 Q1857.75 151.812 1853.79 150.956 L1852.31 150.608 Q1848.35 149.775 1846.59 148.062 Q1844.83 146.326 1844.83 143.317 Q1844.83 139.659 1847.42 137.669 Q1850.01 135.678 1854.78 135.678 Q1857.14 135.678 1859.23 136.025 Q1861.31 136.372 1863.07 137.067 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1871.84 156.349 L1876.73 156.349 L1876.73 162.229 L1871.84 162.229 L1871.84 156.349 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1902.84 137.067 L1902.84 141.094 Q1901.03 140.169 1899.09 139.706 Q1897.14 139.243 1895.06 139.243 Q1891.89 139.243 1890.29 140.215 Q1888.72 141.187 1888.72 143.131 Q1888.72 144.613 1889.85 145.469 Q1890.99 146.303 1894.41 147.067 L1895.87 147.391 Q1900.41 148.363 1902.3 150.145 Q1904.23 151.905 1904.23 155.076 Q1904.23 158.687 1901.36 160.793 Q1898.51 162.9 1893.51 162.9 Q1891.43 162.9 1889.16 162.483 Q1886.91 162.09 1884.41 161.28 L1884.41 156.881 Q1886.77 158.108 1889.06 158.733 Q1891.36 159.335 1893.6 159.335 Q1896.61 159.335 1898.23 158.317 Q1899.85 157.275 1899.85 155.4 Q1899.85 153.664 1898.67 152.738 Q1897.51 151.812 1893.56 150.956 L1892.07 150.608 Q1888.12 149.775 1886.36 148.062 Q1884.6 146.326 1884.6 143.317 Q1884.6 139.659 1887.19 137.669 Q1889.78 135.678 1894.55 135.678 Q1896.91 135.678 1898.99 136.025 Q1901.08 136.372 1902.84 137.067 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1936.31 126.257 Q1933.21 131.581 1931.7 136.789 Q1930.2 141.997 1930.2 147.344 Q1930.2 152.692 1931.7 157.946 Q1933.23 163.178 1936.31 168.479 L1932.61 168.479 Q1929.13 163.039 1927.4 157.784 Q1925.68 152.53 1925.68 147.344 Q1925.68 142.182 1927.4 136.951 Q1929.11 131.72 1932.61 126.257 L1936.31 126.257 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1959.34 143.594 Q1962.7 144.312 1964.57 146.581 Q1966.47 148.849 1966.47 152.182 Q1966.47 157.298 1962.95 160.099 Q1959.43 162.9 1952.95 162.9 Q1950.78 162.9 1948.46 162.46 Q1946.17 162.043 1943.72 161.187 L1943.72 156.673 Q1945.66 157.807 1947.98 158.386 Q1950.29 158.965 1952.81 158.965 Q1957.21 158.965 1959.5 157.229 Q1961.82 155.493 1961.82 152.182 Q1961.82 149.127 1959.67 147.414 Q1957.54 145.678 1953.72 145.678 L1949.69 145.678 L1949.69 141.835 L1953.9 141.835 Q1957.35 141.835 1959.18 140.469 Q1961.01 139.081 1961.01 136.488 Q1961.01 133.826 1959.11 132.414 Q1957.24 130.979 1953.72 130.979 Q1951.8 130.979 1949.6 131.395 Q1947.4 131.812 1944.76 132.692 L1944.76 128.525 Q1947.42 127.784 1949.74 127.414 Q1952.07 127.044 1954.13 127.044 Q1959.46 127.044 1962.56 129.474 Q1965.66 131.882 1965.66 136.002 Q1965.66 138.872 1964.02 140.863 Q1962.37 142.831 1959.34 143.594 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1989.5 143.594 Q1992.86 144.312 1994.74 146.581 Q1996.63 148.849 1996.63 152.182 Q1996.63 157.298 1993.11 160.099 Q1989.6 162.9 1983.11 162.9 Q1980.94 162.9 1978.62 162.46 Q1976.33 162.043 1973.88 161.187 L1973.88 156.673 Q1975.82 157.807 1978.14 158.386 Q1980.45 158.965 1982.98 158.965 Q1987.37 158.965 1989.67 157.229 Q1991.98 155.493 1991.98 152.182 Q1991.98 149.127 1989.83 147.414 Q1987.7 145.678 1983.88 145.678 L1979.85 145.678 L1979.85 141.835 L1984.06 141.835 Q1987.51 141.835 1989.34 140.469 Q1991.17 139.081 1991.17 136.488 Q1991.17 133.826 1989.27 132.414 Q1987.4 130.979 1983.88 130.979 Q1981.96 130.979 1979.76 131.395 Q1977.56 131.812 1974.92 132.692 L1974.92 128.525 Q1977.58 127.784 1979.9 127.414 Q1982.24 127.044 1984.3 127.044 Q1989.62 127.044 1992.72 129.474 Q1995.82 131.882 1995.82 136.002 Q1995.82 138.872 1994.18 140.863 Q1992.54 142.831 1989.5 143.594 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M2005.55 127.669 L2023.9 127.669 L2023.9 131.604 L2009.83 131.604 L2009.83 140.076 Q2010.85 139.729 2011.86 139.567 Q2012.88 139.382 2013.9 139.382 Q2019.69 139.382 2023.07 142.553 Q2026.45 145.724 2026.45 151.141 Q2026.45 156.719 2022.98 159.821 Q2019.5 162.9 2013.18 162.9 Q2011.01 162.9 2008.74 162.53 Q2006.49 162.159 2004.09 161.418 L2004.09 156.719 Q2006.17 157.854 2008.39 158.409 Q2010.61 158.965 2013.09 158.965 Q2017.1 158.965 2019.43 156.858 Q2021.77 154.752 2021.77 151.141 Q2021.77 147.53 2019.43 145.423 Q2017.1 143.317 2013.09 143.317 Q2011.22 143.317 2009.34 143.733 Q2007.49 144.15 2005.55 145.03 L2005.55 127.669 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M2035.71 127.669 L2054.06 127.669 L2054.06 131.604 L2039.99 131.604 L2039.99 140.076 Q2041.01 139.729 2042.03 139.567 Q2043.04 139.382 2044.06 139.382 Q2049.85 139.382 2053.23 142.553 Q2056.61 145.724 2056.61 151.141 Q2056.61 156.719 2053.14 159.821 Q2049.67 162.9 2043.35 162.9 Q2041.17 162.9 2038.9 162.53 Q2036.66 162.159 2034.25 161.418 L2034.25 156.719 Q2036.33 157.854 2038.55 158.409 Q2040.78 158.965 2043.25 158.965 Q2047.26 158.965 2049.6 156.858 Q2051.93 154.752 2051.93 151.141 Q2051.93 147.53 2049.6 145.423 Q2047.26 143.317 2043.25 143.317 Q2041.38 143.317 2039.5 143.733 Q2037.65 144.15 2035.71 145.03 L2035.71 127.669 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M2078.67 131.743 L2066.86 150.192 L2078.67 150.192 L2078.67 131.743 M2077.44 127.669 L2083.32 127.669 L2083.32 150.192 L2088.25 150.192 L2088.25 154.081 L2083.32 154.081 L2083.32 162.229 L2078.67 162.229 L2078.67 154.081 L2063.07 154.081 L2063.07 149.567 L2077.44 127.669 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M2108.83 131.743 L2097.03 150.192 L2108.83 150.192 L2108.83 131.743 M2107.6 127.669 L2113.48 127.669 L2113.48 150.192 L2118.41 150.192 L2118.41 154.081 L2113.48 154.081 L2113.48 162.229 L2108.83 162.229 L2108.83 154.081 L2093.23 154.081 L2093.23 149.567 L2107.6 127.669 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M2140.31 143.594 Q2143.67 144.312 2145.54 146.581 Q2147.44 148.849 2147.44 152.182 Q2147.44 157.298 2143.92 160.099 Q2140.41 162.9 2133.92 162.9 Q2131.75 162.9 2129.43 162.46 Q2127.14 162.043 2124.69 161.187 L2124.69 156.673 Q2126.63 157.807 2128.95 158.386 Q2131.26 158.965 2133.79 158.965 Q2138.18 158.965 2140.47 157.229 Q2142.79 155.493 2142.79 152.182 Q2142.79 149.127 2140.64 147.414 Q2138.51 145.678 2134.69 145.678 L2130.66 145.678 L2130.66 141.835 L2134.87 141.835 Q2138.32 141.835 2140.15 140.469 Q2141.98 139.081 2141.98 136.488 Q2141.98 133.826 2140.08 132.414 Q2138.21 130.979 2134.69 130.979 Q2132.77 130.979 2130.57 131.395 Q2128.37 131.812 2125.73 132.692 L2125.73 128.525 Q2128.39 127.784 2130.71 127.414 Q2133.04 127.044 2135.1 127.044 Q2140.43 127.044 2143.53 129.474 Q2146.63 131.882 2146.63 136.002 Q2146.63 138.872 2144.99 140.863 Q2143.35 142.831 2140.31 143.594 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M2160.34 158.293 L2176.66 158.293 L2176.66 162.229 L2154.71 162.229 L2154.71 158.293 Q2157.37 155.539 2161.96 150.909 Q2166.56 146.256 2167.74 144.914 Q2169.99 142.391 2170.87 140.655 Q2171.77 138.895 2171.77 137.206 Q2171.77 134.451 2169.83 132.715 Q2167.91 130.979 2164.8 130.979 Q2162.6 130.979 2160.15 131.743 Q2157.72 132.507 2154.94 134.057 L2154.94 129.335 Q2157.77 128.201 2160.22 127.622 Q2162.67 127.044 2164.71 127.044 Q2170.08 127.044 2173.28 129.729 Q2176.47 132.414 2176.47 136.905 Q2176.47 139.034 2175.66 140.956 Q2174.87 142.854 2172.77 145.446 Q2172.19 146.118 2169.09 149.335 Q2165.98 152.53 2160.34 158.293 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M2185.2 126.257 L2188.9 126.257 Q2192.37 131.72 2194.09 136.951 Q2195.82 142.182 2195.82 147.344 Q2195.82 152.53 2194.09 157.784 Q2192.37 163.039 2188.9 168.479 L2185.2 168.479 Q2188.28 163.178 2189.78 157.946 Q2191.31 152.692 2191.31 147.344 Q2191.31 141.997 2189.78 136.789 Q2188.28 131.581 2185.2 126.257 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip210)" style="stroke:#e26f46; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1575.06,196.789 1721.97,196.789 
  "/>
<path clip-path="url(#clip210)" d="M1766.8 193.12 Q1768.39 190.249 1770.62 188.884 Q1772.84 187.518 1775.85 187.518 Q1779.9 187.518 1782.1 190.365 Q1784.3 193.189 1784.3 198.421 L1784.3 214.069 L1780.01 214.069 L1780.01 198.559 Q1780.01 194.833 1778.69 193.027 Q1777.38 191.222 1774.67 191.222 Q1771.36 191.222 1769.44 193.421 Q1767.51 195.62 1767.51 199.416 L1767.51 214.069 L1763.23 214.069 L1763.23 198.559 Q1763.23 194.809 1761.91 193.027 Q1760.59 191.222 1757.84 191.222 Q1754.57 191.222 1752.65 193.444 Q1750.73 195.643 1750.73 199.416 L1750.73 214.069 L1746.45 214.069 L1746.45 188.143 L1750.73 188.143 L1750.73 192.171 Q1752.19 189.786 1754.23 188.652 Q1756.26 187.518 1759.07 187.518 Q1761.89 187.518 1763.86 188.953 Q1765.85 190.388 1766.8 193.12 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1804.57 201.036 Q1799.41 201.036 1797.42 202.217 Q1795.43 203.397 1795.43 206.245 Q1795.43 208.513 1796.91 209.856 Q1798.42 211.175 1800.99 211.175 Q1804.53 211.175 1806.66 208.675 Q1808.81 206.152 1808.81 201.985 L1808.81 201.036 L1804.57 201.036 M1813.07 199.277 L1813.07 214.069 L1808.81 214.069 L1808.81 210.133 Q1807.35 212.495 1805.18 213.629 Q1803 214.74 1799.85 214.74 Q1795.87 214.74 1793.51 212.518 Q1791.17 210.272 1791.17 206.522 Q1791.17 202.147 1794.09 199.925 Q1797.03 197.703 1802.84 197.703 L1808.81 197.703 L1808.81 197.286 Q1808.81 194.347 1806.87 192.749 Q1804.94 191.129 1801.45 191.129 Q1799.23 191.129 1797.12 191.661 Q1795.01 192.194 1793.07 193.259 L1793.07 189.323 Q1795.41 188.421 1797.61 187.981 Q1799.81 187.518 1801.89 187.518 Q1807.51 187.518 1810.29 190.434 Q1813.07 193.351 1813.07 199.277 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1838.37 188.907 L1838.37 192.934 Q1836.56 192.009 1834.62 191.546 Q1832.68 191.083 1830.59 191.083 Q1827.42 191.083 1825.82 192.055 Q1824.25 193.027 1824.25 194.971 Q1824.25 196.453 1825.38 197.309 Q1826.52 198.143 1829.94 198.907 L1831.4 199.231 Q1835.94 200.203 1837.84 201.985 Q1839.76 203.745 1839.76 206.916 Q1839.76 210.527 1836.89 212.633 Q1834.04 214.74 1829.04 214.74 Q1826.96 214.74 1824.69 214.323 Q1822.44 213.93 1819.94 213.12 L1819.94 208.721 Q1822.31 209.948 1824.6 210.573 Q1826.89 211.175 1829.13 211.175 Q1832.14 211.175 1833.76 210.157 Q1835.38 209.115 1835.38 207.24 Q1835.38 205.504 1834.2 204.578 Q1833.05 203.652 1829.09 202.796 L1827.61 202.448 Q1823.65 201.615 1821.89 199.902 Q1820.13 198.166 1820.13 195.157 Q1820.13 191.499 1822.72 189.509 Q1825.31 187.518 1830.08 187.518 Q1832.44 187.518 1834.53 187.865 Q1836.61 188.212 1838.37 188.907 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1863.07 188.907 L1863.07 192.934 Q1861.26 192.009 1859.32 191.546 Q1857.37 191.083 1855.29 191.083 Q1852.12 191.083 1850.52 192.055 Q1848.95 193.027 1848.95 194.971 Q1848.95 196.453 1850.08 197.309 Q1851.22 198.143 1854.64 198.907 L1856.1 199.231 Q1860.64 200.203 1862.54 201.985 Q1864.46 203.745 1864.46 206.916 Q1864.46 210.527 1861.59 212.633 Q1858.74 214.74 1853.74 214.74 Q1851.66 214.74 1849.39 214.323 Q1847.14 213.93 1844.64 213.12 L1844.64 208.721 Q1847 209.948 1849.3 210.573 Q1851.59 211.175 1853.83 211.175 Q1856.84 211.175 1858.46 210.157 Q1860.08 209.115 1860.08 207.24 Q1860.08 205.504 1858.9 204.578 Q1857.75 203.652 1853.79 202.796 L1852.31 202.448 Q1848.35 201.615 1846.59 199.902 Q1844.83 198.166 1844.83 195.157 Q1844.83 191.499 1847.42 189.509 Q1850.01 187.518 1854.78 187.518 Q1857.14 187.518 1859.23 187.865 Q1861.31 188.212 1863.07 188.907 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1871.84 208.189 L1876.73 208.189 L1876.73 214.069 L1871.84 214.069 L1871.84 208.189 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1883.25 188.143 L1887.77 188.143 L1895.87 209.902 L1903.97 188.143 L1908.49 188.143 L1898.76 214.069 L1892.98 214.069 L1883.25 188.143 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1939.67 178.097 Q1936.56 183.421 1935.06 188.629 Q1933.55 193.837 1933.55 199.184 Q1933.55 204.532 1935.06 209.786 Q1936.59 215.018 1939.67 220.319 L1935.96 220.319 Q1932.49 214.879 1930.75 209.624 Q1929.04 204.37 1929.04 199.184 Q1929.04 194.022 1930.75 188.791 Q1932.47 183.56 1935.96 178.097 L1939.67 178.097 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1962.7 195.434 Q1966.05 196.152 1967.93 198.421 Q1969.83 200.689 1969.83 204.022 Q1969.83 209.138 1966.31 211.939 Q1962.79 214.74 1956.31 214.74 Q1954.13 214.74 1951.82 214.3 Q1949.53 213.883 1947.07 213.027 L1947.07 208.513 Q1949.02 209.647 1951.33 210.226 Q1953.65 210.805 1956.17 210.805 Q1960.57 210.805 1962.86 209.069 Q1965.18 207.333 1965.18 204.022 Q1965.18 200.967 1963.02 199.254 Q1960.89 197.518 1957.07 197.518 L1953.05 197.518 L1953.05 193.675 L1957.26 193.675 Q1960.71 193.675 1962.54 192.309 Q1964.36 190.921 1964.36 188.328 Q1964.36 185.666 1962.47 184.254 Q1960.59 182.819 1957.07 182.819 Q1955.15 182.819 1952.95 183.235 Q1950.75 183.652 1948.11 184.532 L1948.11 180.365 Q1950.78 179.624 1953.09 179.254 Q1955.43 178.884 1957.49 178.884 Q1962.81 178.884 1965.92 181.314 Q1969.02 183.722 1969.02 187.842 Q1969.02 190.712 1967.37 192.703 Q1965.73 194.671 1962.7 195.434 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1992.86 195.434 Q1996.22 196.152 1998.09 198.421 Q1999.99 200.689 1999.99 204.022 Q1999.99 209.138 1996.47 211.939 Q1992.95 214.74 1986.47 214.74 Q1984.3 214.74 1981.98 214.3 Q1979.69 213.883 1977.24 213.027 L1977.24 208.513 Q1979.18 209.647 1981.49 210.226 Q1983.81 210.805 1986.33 210.805 Q1990.73 210.805 1993.02 209.069 Q1995.34 207.333 1995.34 204.022 Q1995.34 200.967 1993.18 199.254 Q1991.05 197.518 1987.24 197.518 L1983.21 197.518 L1983.21 193.675 L1987.42 193.675 Q1990.87 193.675 1992.7 192.309 Q1994.53 190.921 1994.53 188.328 Q1994.53 185.666 1992.63 184.254 Q1990.75 182.819 1987.24 182.819 Q1985.31 182.819 1983.11 183.235 Q1980.92 183.652 1978.28 184.532 L1978.28 180.365 Q1980.94 179.624 1983.25 179.254 Q1985.59 178.884 1987.65 178.884 Q1992.98 178.884 1996.08 181.314 Q1999.18 183.722 1999.18 187.842 Q1999.18 190.712 1997.54 192.703 Q1995.89 194.671 1992.86 195.434 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M2008.9 179.509 L2027.26 179.509 L2027.26 183.444 L2013.18 183.444 L2013.18 191.916 Q2014.2 191.569 2015.22 191.407 Q2016.24 191.222 2017.26 191.222 Q2023.05 191.222 2026.42 194.393 Q2029.8 197.564 2029.8 202.981 Q2029.8 208.559 2026.33 211.661 Q2022.86 214.74 2016.54 214.74 Q2014.36 214.74 2012.1 214.37 Q2009.85 213.999 2007.44 213.258 L2007.44 208.559 Q2009.53 209.694 2011.75 210.249 Q2013.97 210.805 2016.45 210.805 Q2020.45 210.805 2022.79 208.698 Q2025.13 206.592 2025.13 202.981 Q2025.13 199.37 2022.79 197.263 Q2020.45 195.157 2016.45 195.157 Q2014.57 195.157 2012.7 195.573 Q2010.85 195.99 2008.9 196.87 L2008.9 179.509 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M2039.06 179.509 L2057.42 179.509 L2057.42 183.444 L2043.35 183.444 L2043.35 191.916 Q2044.36 191.569 2045.38 191.407 Q2046.4 191.222 2047.42 191.222 Q2053.21 191.222 2056.59 194.393 Q2059.97 197.564 2059.97 202.981 Q2059.97 208.559 2056.49 211.661 Q2053.02 214.74 2046.7 214.74 Q2044.53 214.74 2042.26 214.37 Q2040.01 213.999 2037.61 213.258 L2037.61 208.559 Q2039.69 209.694 2041.91 210.249 Q2044.13 210.805 2046.61 210.805 Q2050.61 210.805 2052.95 208.698 Q2055.29 206.592 2055.29 202.981 Q2055.29 199.37 2052.95 197.263 Q2050.61 195.157 2046.61 195.157 Q2044.73 195.157 2042.86 195.573 Q2041.01 195.99 2039.06 196.87 L2039.06 179.509 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M2082.03 183.583 L2070.22 202.032 L2082.03 202.032 L2082.03 183.583 M2080.8 179.509 L2086.68 179.509 L2086.68 202.032 L2091.61 202.032 L2091.61 205.921 L2086.68 205.921 L2086.68 214.069 L2082.03 214.069 L2082.03 205.921 L2066.42 205.921 L2066.42 201.407 L2080.8 179.509 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M2112.19 183.583 L2100.38 202.032 L2112.19 202.032 L2112.19 183.583 M2110.96 179.509 L2116.84 179.509 L2116.84 202.032 L2121.77 202.032 L2121.77 205.921 L2116.84 205.921 L2116.84 214.069 L2112.19 214.069 L2112.19 205.921 L2096.59 205.921 L2096.59 201.407 L2110.96 179.509 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M2143.67 195.434 Q2147.03 196.152 2148.9 198.421 Q2150.8 200.689 2150.8 204.022 Q2150.8 209.138 2147.28 211.939 Q2143.76 214.74 2137.28 214.74 Q2135.1 214.74 2132.79 214.3 Q2130.5 213.883 2128.04 213.027 L2128.04 208.513 Q2129.99 209.647 2132.3 210.226 Q2134.62 210.805 2137.14 210.805 Q2141.54 210.805 2143.83 209.069 Q2146.15 207.333 2146.15 204.022 Q2146.15 200.967 2143.99 199.254 Q2141.86 197.518 2138.04 197.518 L2134.02 197.518 L2134.02 193.675 L2138.23 193.675 Q2141.68 193.675 2143.51 192.309 Q2145.34 190.921 2145.34 188.328 Q2145.34 185.666 2143.44 184.254 Q2141.56 182.819 2138.04 182.819 Q2136.12 182.819 2133.92 183.235 Q2131.73 183.652 2129.09 184.532 L2129.09 180.365 Q2131.75 179.624 2134.06 179.254 Q2136.4 178.884 2138.46 178.884 Q2143.79 178.884 2146.89 181.314 Q2149.99 183.722 2149.99 187.842 Q2149.99 190.712 2148.35 192.703 Q2146.7 194.671 2143.67 195.434 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M2173.83 195.434 Q2177.19 196.152 2179.06 198.421 Q2180.96 200.689 2180.96 204.022 Q2180.96 209.138 2177.44 211.939 Q2173.92 214.74 2167.44 214.74 Q2165.27 214.74 2162.95 214.3 Q2160.66 213.883 2158.21 213.027 L2158.21 208.513 Q2160.15 209.647 2162.47 210.226 Q2164.78 210.805 2167.3 210.805 Q2171.7 210.805 2173.99 209.069 Q2176.31 207.333 2176.31 204.022 Q2176.31 200.967 2174.16 199.254 Q2172.03 197.518 2168.21 197.518 L2164.18 197.518 L2164.18 193.675 L2168.39 193.675 Q2171.84 193.675 2173.67 192.309 Q2175.5 190.921 2175.5 188.328 Q2175.5 185.666 2173.6 184.254 Q2171.72 182.819 2168.21 182.819 Q2166.29 182.819 2164.09 183.235 Q2161.89 183.652 2159.25 184.532 L2159.25 180.365 Q2161.91 179.624 2164.22 179.254 Q2166.56 178.884 2168.62 178.884 Q2173.95 178.884 2177.05 181.314 Q2180.15 183.722 2180.15 187.842 Q2180.15 190.712 2178.51 192.703 Q2176.86 194.671 2173.83 195.434 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M2188.55 178.097 L2192.26 178.097 Q2195.73 183.56 2197.44 188.791 Q2199.18 194.022 2199.18 199.184 Q2199.18 204.37 2197.44 209.624 Q2195.73 214.879 2192.26 220.319 L2188.55 220.319 Q2191.63 215.018 2193.14 209.786 Q2194.66 204.532 2194.66 199.184 Q2194.66 193.837 2193.14 188.629 Q2191.63 183.421 2188.55 178.097 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip210)" style="stroke:#3da44d; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1575.06,248.629 1721.97,248.629 
  "/>
<path clip-path="url(#clip210)" d="M1766.8 244.96 Q1768.39 242.089 1770.62 240.724 Q1772.84 239.358 1775.85 239.358 Q1779.9 239.358 1782.1 242.205 Q1784.3 245.029 1784.3 250.261 L1784.3 265.909 L1780.01 265.909 L1780.01 250.399 Q1780.01 246.673 1778.69 244.867 Q1777.38 243.062 1774.67 243.062 Q1771.36 243.062 1769.44 245.261 Q1767.51 247.46 1767.51 251.256 L1767.51 265.909 L1763.23 265.909 L1763.23 250.399 Q1763.23 246.649 1761.91 244.867 Q1760.59 243.062 1757.84 243.062 Q1754.57 243.062 1752.65 245.284 Q1750.73 247.483 1750.73 251.256 L1750.73 265.909 L1746.45 265.909 L1746.45 239.983 L1750.73 239.983 L1750.73 244.011 Q1752.19 241.626 1754.23 240.492 Q1756.26 239.358 1759.07 239.358 Q1761.89 239.358 1763.86 240.793 Q1765.85 242.228 1766.8 244.96 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1804.57 252.876 Q1799.41 252.876 1797.42 254.057 Q1795.43 255.237 1795.43 258.085 Q1795.43 260.353 1796.91 261.696 Q1798.42 263.015 1800.99 263.015 Q1804.53 263.015 1806.66 260.515 Q1808.81 257.992 1808.81 253.825 L1808.81 252.876 L1804.57 252.876 M1813.07 251.117 L1813.07 265.909 L1808.81 265.909 L1808.81 261.973 Q1807.35 264.335 1805.18 265.469 Q1803 266.58 1799.85 266.58 Q1795.87 266.58 1793.51 264.358 Q1791.17 262.112 1791.17 258.362 Q1791.17 253.987 1794.09 251.765 Q1797.03 249.543 1802.84 249.543 L1808.81 249.543 L1808.81 249.126 Q1808.81 246.187 1806.87 244.589 Q1804.94 242.969 1801.45 242.969 Q1799.23 242.969 1797.12 243.501 Q1795.01 244.034 1793.07 245.099 L1793.07 241.163 Q1795.41 240.261 1797.61 239.821 Q1799.81 239.358 1801.89 239.358 Q1807.51 239.358 1810.29 242.274 Q1813.07 245.191 1813.07 251.117 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1838.37 240.747 L1838.37 244.774 Q1836.56 243.849 1834.62 243.386 Q1832.68 242.923 1830.59 242.923 Q1827.42 242.923 1825.82 243.895 Q1824.25 244.867 1824.25 246.811 Q1824.25 248.293 1825.38 249.149 Q1826.52 249.983 1829.94 250.747 L1831.4 251.071 Q1835.94 252.043 1837.84 253.825 Q1839.76 255.585 1839.76 258.756 Q1839.76 262.367 1836.89 264.473 Q1834.04 266.58 1829.04 266.58 Q1826.96 266.58 1824.69 266.163 Q1822.44 265.77 1819.94 264.96 L1819.94 260.561 Q1822.31 261.788 1824.6 262.413 Q1826.89 263.015 1829.13 263.015 Q1832.14 263.015 1833.76 261.997 Q1835.38 260.955 1835.38 259.08 Q1835.38 257.344 1834.2 256.418 Q1833.05 255.492 1829.09 254.636 L1827.61 254.288 Q1823.65 253.455 1821.89 251.742 Q1820.13 250.006 1820.13 246.997 Q1820.13 243.339 1822.72 241.349 Q1825.31 239.358 1830.08 239.358 Q1832.44 239.358 1834.53 239.705 Q1836.61 240.052 1838.37 240.747 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1863.07 240.747 L1863.07 244.774 Q1861.26 243.849 1859.32 243.386 Q1857.37 242.923 1855.29 242.923 Q1852.12 242.923 1850.52 243.895 Q1848.95 244.867 1848.95 246.811 Q1848.95 248.293 1850.08 249.149 Q1851.22 249.983 1854.64 250.747 L1856.1 251.071 Q1860.64 252.043 1862.54 253.825 Q1864.46 255.585 1864.46 258.756 Q1864.46 262.367 1861.59 264.473 Q1858.74 266.58 1853.74 266.58 Q1851.66 266.58 1849.39 266.163 Q1847.14 265.77 1844.64 264.96 L1844.64 260.561 Q1847 261.788 1849.3 262.413 Q1851.59 263.015 1853.83 263.015 Q1856.84 263.015 1858.46 261.997 Q1860.08 260.955 1860.08 259.08 Q1860.08 257.344 1858.9 256.418 Q1857.75 255.492 1853.79 254.636 L1852.31 254.288 Q1848.35 253.455 1846.59 251.742 Q1844.83 250.006 1844.83 246.997 Q1844.83 243.339 1847.42 241.349 Q1850.01 239.358 1854.78 239.358 Q1857.14 239.358 1859.23 239.705 Q1861.31 240.052 1863.07 240.747 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1871.84 260.029 L1876.73 260.029 L1876.73 265.909 L1871.84 265.909 L1871.84 260.029 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1898.09 252.876 Q1892.93 252.876 1890.94 254.057 Q1888.95 255.237 1888.95 258.085 Q1888.95 260.353 1890.43 261.696 Q1891.93 263.015 1894.5 263.015 Q1898.05 263.015 1900.18 260.515 Q1902.33 257.992 1902.33 253.825 L1902.33 252.876 L1898.09 252.876 M1906.59 251.117 L1906.59 265.909 L1902.33 265.909 L1902.33 261.973 Q1900.87 264.335 1898.69 265.469 Q1896.52 266.58 1893.37 266.58 Q1889.39 266.58 1887.03 264.358 Q1884.69 262.112 1884.69 258.362 Q1884.69 253.987 1887.61 251.765 Q1890.55 249.543 1896.36 249.543 L1902.33 249.543 L1902.33 249.126 Q1902.33 246.187 1900.38 244.589 Q1898.46 242.969 1894.97 242.969 Q1892.74 242.969 1890.64 243.501 Q1888.53 244.034 1886.59 245.099 L1886.59 241.163 Q1888.93 240.261 1891.12 239.821 Q1893.32 239.358 1895.41 239.358 Q1901.03 239.358 1903.81 242.274 Q1906.59 245.191 1906.59 251.117 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1940.66 229.937 Q1937.56 235.261 1936.05 240.469 Q1934.55 245.677 1934.55 251.024 Q1934.55 256.372 1936.05 261.626 Q1937.58 266.858 1940.66 272.159 L1936.96 272.159 Q1933.49 266.719 1931.75 261.464 Q1930.04 256.21 1930.04 251.024 Q1930.04 245.862 1931.75 240.631 Q1933.46 235.4 1936.96 229.937 L1940.66 229.937 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1949.57 231.349 L1967.93 231.349 L1967.93 235.284 L1953.86 235.284 L1953.86 243.756 Q1954.87 243.409 1955.89 243.247 Q1956.91 243.062 1957.93 243.062 Q1963.72 243.062 1967.1 246.233 Q1970.48 249.404 1970.48 254.821 Q1970.48 260.399 1967 263.501 Q1963.53 266.58 1957.21 266.58 Q1955.04 266.58 1952.77 266.21 Q1950.52 265.839 1948.11 265.098 L1948.11 260.399 Q1950.2 261.534 1952.42 262.089 Q1954.64 262.645 1957.12 262.645 Q1961.12 262.645 1963.46 260.538 Q1965.8 258.432 1965.8 254.821 Q1965.8 251.21 1963.46 249.103 Q1961.12 246.997 1957.12 246.997 Q1955.24 246.997 1953.37 247.413 Q1951.52 247.83 1949.57 248.71 L1949.57 231.349 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1989.69 249.497 Q1986.36 249.497 1984.43 251.279 Q1982.54 253.061 1982.54 256.186 Q1982.54 259.311 1984.43 261.094 Q1986.36 262.876 1989.69 262.876 Q1993.02 262.876 1994.94 261.094 Q1996.86 259.288 1996.86 256.186 Q1996.86 253.061 1994.94 251.279 Q1993.05 249.497 1989.69 249.497 M1985.01 247.506 Q1982 246.765 1980.31 244.705 Q1978.65 242.645 1978.65 239.682 Q1978.65 235.538 1981.59 233.131 Q1984.55 230.724 1989.69 230.724 Q1994.85 230.724 1997.79 233.131 Q2000.73 235.538 2000.73 239.682 Q2000.73 242.645 1999.04 244.705 Q1997.37 246.765 1994.39 247.506 Q1997.77 248.293 1999.64 250.585 Q2001.54 252.876 2001.54 256.186 Q2001.54 261.21 1998.46 263.895 Q1995.41 266.58 1989.69 266.58 Q1983.97 266.58 1980.89 263.895 Q1977.84 261.21 1977.84 256.186 Q1977.84 252.876 1979.74 250.585 Q1981.63 248.293 1985.01 247.506 M1983.3 240.122 Q1983.3 242.807 1984.97 244.312 Q1986.66 245.816 1989.69 245.816 Q1992.7 245.816 1994.39 244.312 Q1996.1 242.807 1996.1 240.122 Q1996.1 237.437 1994.39 235.932 Q1992.7 234.427 1989.69 234.427 Q1986.66 234.427 1984.97 235.932 Q1983.3 237.437 1983.3 240.122 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M2008.67 231.349 L2030.89 231.349 L2030.89 233.339 L2018.35 265.909 L2013.46 265.909 L2025.27 235.284 L2008.67 235.284 L2008.67 231.349 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M2044.04 261.973 L2060.36 261.973 L2060.36 265.909 L2038.42 265.909 L2038.42 261.973 Q2041.08 259.219 2045.66 254.589 Q2050.27 249.936 2051.45 248.594 Q2053.69 246.071 2054.57 244.335 Q2055.48 242.575 2055.48 240.886 Q2055.48 238.131 2053.53 236.395 Q2051.61 234.659 2048.51 234.659 Q2046.31 234.659 2043.86 235.423 Q2041.42 236.187 2038.65 237.737 L2038.65 233.015 Q2041.47 231.881 2043.92 231.302 Q2046.38 230.724 2048.42 230.724 Q2053.79 230.724 2056.98 233.409 Q2060.17 236.094 2060.17 240.585 Q2060.17 242.714 2059.36 244.636 Q2058.58 246.534 2056.47 249.126 Q2055.89 249.798 2052.79 253.015 Q2049.69 256.21 2044.04 261.973 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M2080.17 234.427 Q2076.56 234.427 2074.73 237.992 Q2072.93 241.534 2072.93 248.663 Q2072.93 255.77 2074.73 259.335 Q2076.56 262.876 2080.17 262.876 Q2083.81 262.876 2085.61 259.335 Q2087.44 255.77 2087.44 248.663 Q2087.44 241.534 2085.61 237.992 Q2083.81 234.427 2080.17 234.427 M2080.17 230.724 Q2085.98 230.724 2089.04 235.33 Q2092.12 239.913 2092.12 248.663 Q2092.12 257.39 2089.04 261.997 Q2085.98 266.58 2080.17 266.58 Q2074.36 266.58 2071.29 261.997 Q2068.23 257.39 2068.23 248.663 Q2068.23 239.913 2071.29 235.33 Q2074.36 230.724 2080.17 230.724 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M2104.36 261.973 L2120.68 261.973 L2120.68 265.909 L2098.74 265.909 L2098.74 261.973 Q2101.4 259.219 2105.98 254.589 Q2110.59 249.936 2111.77 248.594 Q2114.02 246.071 2114.9 244.335 Q2115.8 242.575 2115.8 240.886 Q2115.8 238.131 2113.85 236.395 Q2111.93 234.659 2108.83 234.659 Q2106.63 234.659 2104.18 235.423 Q2101.75 236.187 2098.97 237.737 L2098.97 233.015 Q2101.79 231.881 2104.25 231.302 Q2106.7 230.724 2108.74 230.724 Q2114.11 230.724 2117.3 233.409 Q2120.5 236.094 2120.5 240.585 Q2120.5 242.714 2119.69 244.636 Q2118.9 246.534 2116.79 249.126 Q2116.22 249.798 2113.11 253.015 Q2110.01 256.21 2104.36 261.973 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M2130.54 231.349 L2148.9 231.349 L2148.9 235.284 L2134.83 235.284 L2134.83 243.756 Q2135.85 243.409 2136.86 243.247 Q2137.88 243.062 2138.9 243.062 Q2144.69 243.062 2148.07 246.233 Q2151.45 249.404 2151.45 254.821 Q2151.45 260.399 2147.97 263.501 Q2144.5 266.58 2138.18 266.58 Q2136.01 266.58 2133.74 266.21 Q2131.49 265.839 2129.09 265.098 L2129.09 260.399 Q2131.17 261.534 2133.39 262.089 Q2135.61 262.645 2138.09 262.645 Q2142.1 262.645 2144.43 260.538 Q2146.77 258.432 2146.77 254.821 Q2146.77 251.21 2144.43 249.103 Q2142.1 246.997 2138.09 246.997 Q2136.22 246.997 2134.34 247.413 Q2132.49 247.83 2130.54 248.71 L2130.54 231.349 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M2171.24 246.765 Q2168.09 246.765 2166.24 248.918 Q2164.41 251.071 2164.41 254.821 Q2164.41 258.548 2166.24 260.723 Q2168.09 262.876 2171.24 262.876 Q2174.39 262.876 2176.22 260.723 Q2178.07 258.548 2178.07 254.821 Q2178.07 251.071 2176.22 248.918 Q2174.39 246.765 2171.24 246.765 M2180.52 232.113 L2180.52 236.372 Q2178.76 235.538 2176.96 235.099 Q2175.17 234.659 2173.41 234.659 Q2168.79 234.659 2166.33 237.784 Q2163.9 240.909 2163.55 247.228 Q2164.92 245.214 2166.98 244.149 Q2169.04 243.062 2171.52 243.062 Q2176.72 243.062 2179.73 246.233 Q2182.77 249.381 2182.77 254.821 Q2182.77 260.145 2179.62 263.362 Q2176.47 266.58 2171.24 266.58 Q2165.24 266.58 2162.07 261.997 Q2158.9 257.39 2158.9 248.663 Q2158.9 240.469 2162.79 235.608 Q2166.68 230.724 2173.23 230.724 Q2174.99 230.724 2176.77 231.071 Q2178.58 231.418 2180.52 232.113 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M2191.63 261.973 L2199.27 261.973 L2199.27 235.608 L2190.96 237.275 L2190.96 233.015 L2199.22 231.349 L2203.9 231.349 L2203.9 261.973 L2211.54 261.973 L2211.54 265.909 L2191.63 265.909 L2191.63 261.973 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M2219.71 229.937 L2223.41 229.937 Q2226.89 235.4 2228.6 240.631 Q2230.34 245.862 2230.34 251.024 Q2230.34 256.21 2228.6 261.464 Q2226.89 266.719 2223.41 272.159 L2219.71 272.159 Q2222.79 266.858 2224.29 261.626 Q2225.82 256.372 2225.82 251.024 Q2225.82 245.677 2224.29 240.469 Q2222.79 235.261 2219.71 229.937 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /></svg>



## NeuralFMU

#### Loss function

In order to train our model, a loss function must be implemented. The solver of the NeuralFMU can calculate the gradient of the loss function. The gradient descent is needed to adjust the weights in the neural network so that the sum of the error is reduced and the model becomes more accurate.

The error function in this implementation consists of the mean of the mean squared erros. The first part of the addition is the deviation of the position and the second part is the deviation of the velocity. The mean squared error (mse) for the position consists from the real position of the *realFMU* simulation (posReal) and the position data of the network (posNet). The mean squared error for the velocity consists of the real velocity of the *realFMU* simualtion (velReal) and the velocity data of the network (velNet).
$$ loss = \frac{1}{2} \Bigl[ \frac{1}{n} \sum\limits_{i=0}^n (posReal[i] - posNet[i])^2 + \frac{1}{n} \sum\limits_{i=0}^n (velReal[i] - velNet[i])^2 \Bigr]$$


```julia
# loss function for training
function lossSum()
    global x₀
    solution = neuralFMU(x₀)

    posNet, velNet = extractPosVel(solution.u)

    (Flux.Losses.mse(posReal, posNet) + Flux.Losses.mse(velReal, velNet)) / 2.0
end
```


    lossSum (generic function with 1 method)


#### Callback

To output the loss in certain time intervals, a callback is implemented as a function in the following. Here a counter is incremented, every fiftieth pass the loss function is called and the average error is printed out. Also the parmaters for the velocity in the first layer are kept to a fixed value.


```julia
# callback function for training
global counter = 0
function callb()
    global counter, paramsNet
    counter += 1

    # freeze first layer parameters (2,4,6) for velocity -> (static) direct feed trough for velocity
    # parameters for position (1,3,5) are learned
    paramsNet[1][2] = 0.0
    paramsNet[1][4] = 1.0
    paramsNet[1][6] = 0.0

    if counter % 50 == 1
        avgLoss = lossSum()
        @info "  Loss [$counter]: $(round(avgLoss, digits=5))
        Avg displacement in data: $(round(sqrt(avgLoss), digits=5))
        Weight/Scale: $(paramsNet[1][1])   Bias/Offset: $(paramsNet[1][5])"
    end
end
```


    callb (generic function with 1 method)


#### Functions for plotting

In this section some important functions for plotting are defined. The function `generate_figure()` creates a new figure object and sets some attributes.


```julia
function generate_figure(title, xLabel, yLabel, xlim="auto")
    Plots.plot(
        title=title, xlabel=xLabel, ylabel=yLabel, linewidth=2,
        xtickfontsize=12, ytickfontsize=12, xguidefontsize=12, yguidefontsize=12,
        legendfontsize=12, legend=:topright, xlim=xlim)
end
```


    generate_figure (generic function with 2 methods)


In the following function, the data of the *realFMU*, *simpleFMU* and *neuralFMU* are summarized and displayed in a graph.


```julia
function plot_results(title, xLabel, yLabel, interval, realData, simpleData, neuralData)
    linestyles = [:dot, :solid]
    
    fig = generate_figure(title, xLabel, yLabel)
    Plots.plot!(fig, interval, simpleData, label="SimpleFMU", linewidth=2)
    Plots.plot!(fig, interval, realData, label="Reference", linewidth=2)
    for i in 1:length(neuralData)
        Plots.plot!(fig, neuralData[i][1], neuralData[i][2], label="NeuralFMU ($(i*2500))", 
                    linewidth=2, linestyle=linestyles[i], linecolor=:green)
    end
    Plots.display(fig)
end
```


    plot_results (generic function with 1 method)


This is the superordinate function, which at the beginning extracts the position and velocity from the simulation data (`realSimData`, `realSimDataMod`, `simpleSimData`,..., `solutionAfterMod`). Four graphs are then generated, each comparing the corresponding data from the *realFMU*, *simpleFMU*, and *neuralFMU*. The comparison is made with the simulation data from the simulation with the default and modified initial states. According to the data, the designation of the title and the naming of the axes is adapted.


```julia
function plot_all_results(realSimData, realSimDataMod, simpleSimData, 
        simpleSimDataMod, solutionAfter, solutionAfterMod)    
    # collect all data
    posReal, velReal = extractPosVel(realSimData.saveval)
    posRealMod, velRealMod = extractPosVel(realSimDataMod.saveval)
    posSimple, velSimple = extractPosVel(simpleSimData.saveval)
    posSimpleMod, velSimpleMod = extractPosVel(simpleSimDataMod.saveval)
    
    run = length(solutionAfter)
    
    posNeural, velNeural = [], []
    posNeuralMod, velNeuralMod = [], []
    for i in 1:run
        dataNeural = extractPosVel(solutionAfter[i].u)
        push!(posNeural, (solutionAfter[i].t, dataNeural[1]))
        push!(velNeural, (solutionAfter[i].t, dataNeural[2]))
        
        dataNeuralMod = extractPosVel(solutionAfterMod[i].u)
        push!(posNeuralMod, (solutionAfterMod[i].t, dataNeuralMod[1]))
        push!(velNeuralMod, (solutionAfterMod[i].t, dataNeuralMod[2]))
    end
         
    # plot results s (default initial states)
    xLabel="t [s]"
    yLabel="mass position [m]"
    title = "Default: Mass position after Run: $(run)"
    plot_results(title, xLabel, yLabel, tSave, posReal, posSimple, posNeural)

    # plot results s (modified initial states)
    title = "Modified: Mass position after Run: $(run)"
    plot_results(title, xLabel, yLabel, tSave, posRealMod, posSimpleMod, posNeuralMod)

    # plot results v (default initial states)
    yLabel="mass velocity [m/s]"
    title = "Default: Mass velocity after Run: $(run)"
    plot_results(title, xLabel, yLabel, tSave, velReal, velSimple, velNeural)

    # plot results v (modified initial states)    
    title = "Modified: Mass velocity after Run: $(run)"
    plot_results(title, xLabel, yLabel, tSave, velRealMod, velSimpleMod, velNeuralMod)
end
```


    plot_all_results (generic function with 1 method)


The function `plot_friction_model()` compares the friction model of the *realFMU*, *simpleFMU* and *neuralFMU*. For this, the velocity and force from the simulation data of the *realFMU* is needed. The force data is calculated with the extracted last layer of the *neuralFMU* to the real velocity in line 9 by iterating over the vector `velReal`. In the next rows, the velocity and force data (if available) for each of the three FMUs are combined into a matrix. The first row of the matrix corresponds to the later x-axis and here the velocity is plotted. The second row corresponds to the y-axis and here the force is plotted. This matrix is sorted and plotted by the first entries (velocity) with the function `sortperm()`. The graph with at least three graphs is plotted in line 33. As output this function has the forces of the *neuralFMU*.


```julia
function plot_friction_model(realSimData, netBottom, forces)    
    linestyles = [:dot, :solid]
    
    velReal = collect(data[2] for data in realSimData.saveval)
    forceReal = collect(data[4] for data in realSimData.saveval)

    push!(forces, zeros(length(velReal)))
    for i in 1:length(velReal)
        forces[end][i] = -netBottom([velReal[i], 0.0])[2]
    end

    run = length(forces) 
    
    fig = generate_figure("Friction model $(run)", "v [m/s]", "friction force [N]", (-1.25, 1.25))

    fricSimple = hcat(velReal, zeros(length(velReal)))
    fricSimple[sortperm(fricSimple[:, 1]), :]
    Plots.plot!(fig, fricSimple[:,1], fricSimple[:,2], label="SimpleFMU", linewidth=2)

    fricReal = hcat(velReal, forceReal)
    fricReal[sortperm(fricReal[:, 1]), :]
    Plots.plot!(fig, fricReal[:,1], fricReal[:,2], label="reference", linewidth=2)

    for i in 1:run
        fricNeural = hcat(velReal, forces[i])
        fricNeural[sortperm(fricNeural[:, 1]), :]
        Plots.plot!(fig, fricNeural[:,1], fricNeural[:,2], label="NeuralFMU ($(i*2500))", 
                    linewidth=2, linestyle=linestyles[i], linecolor=:green)
        @info "Friction model $i mse: $(Flux.Losses.mse(fricNeural[:,2], fricReal[:,2]))"
    end
    flush(stderr)

    Plots.display(fig)
    
    return forces   
end
```


    plot_friction_model (generic function with 1 method)


The following function is used to display the different displacement modells of the *realFMU*, *simpleFMU* and *neuralFMU*. The displacement of the *realFMU* and *simpleFMU* is very trivial and is only a constant. The position data of the *realFMU* is needed to calculate the displacement. The displacement for the *neuralFMU* is calculated using the first extracted layer of the neural network, subtracting the real position and the displacement of the *simpleFMU*. Also in this function, the graphs of the three FMUs are compared in a plot.


```julia
function plot_displacement_model(realSimData, netTop, displacements, tSave, displacement)
    linestyles = [:dot, :solid]
    
    posReal = collect(data[1] for data in realSimData.saveval)
    
    push!(displacements, zeros(length(posReal)))
    for i in 1:length(posReal)
        displacements[end][i] = netTop([posReal[i], 0.0])[1] - posReal[i] - displacement
    end

    run = length(displacements)
    fig = generate_figure("Displacement model $(run)", "t [s]", "displacement [m]")
    Plots.plot!(fig, [tSave[1], tSave[end]], [displacement, displacement], label="simpleFMU", linewidth=2)
    Plots.plot!(fig, [tSave[1], tSave[end]], [0.0, 0.0], label="reference", linewidth=2)
    for i in 1:run
        Plots.plot!(fig, tSave, displacements[i], label="NeuralFMU ($(i*2500))", 
                    linewidth=2, linestyle=linestyles[i], linecolor=:green)
    end

    Plots.display(fig)
    
    return displacements
end
```


    plot_displacement_model (generic function with 1 method)


#### Structure of the NeuralFMU

In the following, the topology of the NeuralFMU is constructed. It consists of a dense layer that has exactly as many inputs and outputs as the model has states `numStates` (and therefore state derivatives). It also sets the initial weights and offsets for the first dense layer, as well as the activation function, which consists of the identity. An input layer follows, which then leads into the *simpleFMU* model. The ME-FMU computes the state derivatives for a given system state. Following the *simpleFMU* is a dense layer that has `numStates` states. The output of this layer consists of 8 output nodes and a *identity* activation function. The next layer has 8 input and output nodes with a *tanh* activation function. The last layer is again a dense layer with 8 input nodes and the number of states as outputs. Here, it is important that no *tanh*-activation function follows, because otherwise the pendulums state values would be limited to the interval $[-1;1]$.


```julia
# NeuralFMU setup
numStates = fmiGetNumberOfStates(simpleFMU)

net = Chain(Dense(numStates, numStates, identity; 
                  initW = (out, in) -> [[1.0, 0.0] [0.0, 1.0]], 
                  initb = out -> zeros(out)),
            inputs -> fmi2EvaluateME(simpleFMU, inputs),
            Dense(numStates, 8, identity),
            Dense(8, 8, tanh),
            Dense(8, numStates))
```


    Chain(
      Dense(2, 2),                          [90m# 6 parameters[39m
      var"#19#22"(),
      Dense(2, 8),                          [90m# 24 parameters[39m
      Dense(8, 8, tanh),                    [90m# 72 parameters[39m
      Dense(8, 2),                          [90m# 18 parameters[39m
    )[90m                   # Total: 8 arrays, [39m120 parameters, 1016 bytes.


#### Definition of the NeuralFMU

The instantiation of the ME-NeuralFMU is done as a one-liner. The FMU (*simpleFMU*), the structure of the network `net`, start `tStart` and end time `tStop`, the numerical solver `Tsit5()` and the time steps `tSave` for saving are specified.


```julia
neuralFMU = ME_NeuralFMU(simpleFMU, net, (tStart, tStop), Tsit5(); saveat=tSave);
```

#### Plot before training

Here the state trajactory of the *simpleFMU* is recorded. Doesn't really look like a pendulum yet, but the system is random initialized by default. In the later plots, the effect of learning can be seen.


```julia
solutionBefore = neuralFMU(x₀)
fmiPlot(simpleFMU, solutionBefore)
```


<?xml version="1.0" encoding="utf-8"?>
<svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" width="600" height="400" viewBox="0 0 2400 1600">
<defs>
  <clipPath id="clip300">
    <rect x="0" y="0" width="2400" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip300)" d="
M0 1600 L2400 1600 L2400 0 L0 0  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip301">
    <rect x="480" y="0" width="1681" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip300)" d="
M149.191 1423.18 L2352.76 1423.18 L2352.76 47.2441 L149.191 47.2441  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip302">
    <rect x="149" y="47" width="2205" height="1377"/>
  </clipPath>
</defs>
<polyline clip-path="url(#clip302)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  211.556,1423.18 211.556,47.2441 
  "/>
<polyline clip-path="url(#clip302)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  731.265,1423.18 731.265,47.2441 
  "/>
<polyline clip-path="url(#clip302)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1250.97,1423.18 1250.97,47.2441 
  "/>
<polyline clip-path="url(#clip302)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1770.68,1423.18 1770.68,47.2441 
  "/>
<polyline clip-path="url(#clip302)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  2290.39,1423.18 2290.39,47.2441 
  "/>
<polyline clip-path="url(#clip300)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  149.191,1423.18 2352.76,1423.18 
  "/>
<polyline clip-path="url(#clip300)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  211.556,1423.18 211.556,1404.28 
  "/>
<polyline clip-path="url(#clip300)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  731.265,1423.18 731.265,1404.28 
  "/>
<polyline clip-path="url(#clip300)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1250.97,1423.18 1250.97,1404.28 
  "/>
<polyline clip-path="url(#clip300)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1770.68,1423.18 1770.68,1404.28 
  "/>
<polyline clip-path="url(#clip300)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  2290.39,1423.18 2290.39,1404.28 
  "/>
<path clip-path="url(#clip300)" d="M211.556 1454.1 Q207.945 1454.1 206.116 1457.66 Q204.311 1461.2 204.311 1468.33 Q204.311 1475.44 206.116 1479.01 Q207.945 1482.55 211.556 1482.55 Q215.19 1482.55 216.996 1479.01 Q218.824 1475.44 218.824 1468.33 Q218.824 1461.2 216.996 1457.66 Q215.19 1454.1 211.556 1454.1 M211.556 1450.39 Q217.366 1450.39 220.422 1455 Q223.5 1459.58 223.5 1468.33 Q223.5 1477.06 220.422 1481.67 Q217.366 1486.25 211.556 1486.25 Q205.746 1486.25 202.667 1481.67 Q199.612 1477.06 199.612 1468.33 Q199.612 1459.58 202.667 1455 Q205.746 1450.39 211.556 1450.39 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M721.647 1481.64 L729.285 1481.64 L729.285 1455.28 L720.975 1456.95 L720.975 1452.69 L729.239 1451.02 L733.915 1451.02 L733.915 1481.64 L741.554 1481.64 L741.554 1485.58 L721.647 1485.58 L721.647 1481.64 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1245.63 1481.64 L1261.95 1481.64 L1261.95 1485.58 L1240 1485.58 L1240 1481.64 Q1242.66 1478.89 1247.25 1474.26 Q1251.85 1469.61 1253.03 1468.27 Q1255.28 1465.74 1256.16 1464.01 Q1257.06 1462.25 1257.06 1460.56 Q1257.06 1457.8 1255.12 1456.07 Q1253.2 1454.33 1250.09 1454.33 Q1247.89 1454.33 1245.44 1455.09 Q1243.01 1455.86 1240.23 1457.41 L1240.23 1452.69 Q1243.06 1451.55 1245.51 1450.97 Q1247.96 1450.39 1250 1450.39 Q1255.37 1450.39 1258.57 1453.08 Q1261.76 1455.77 1261.76 1460.26 Q1261.76 1462.39 1260.95 1464.31 Q1260.16 1466.2 1258.06 1468.8 Q1257.48 1469.47 1254.38 1472.69 Q1251.27 1475.88 1245.63 1481.64 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1774.93 1466.95 Q1778.29 1467.66 1780.16 1469.93 Q1782.06 1472.2 1782.06 1475.53 Q1782.06 1480.65 1778.54 1483.45 Q1775.02 1486.25 1768.54 1486.25 Q1766.37 1486.25 1764.05 1485.81 Q1761.76 1485.39 1759.3 1484.54 L1759.3 1480.02 Q1761.25 1481.16 1763.56 1481.74 Q1765.88 1482.32 1768.4 1482.32 Q1772.8 1482.32 1775.09 1480.58 Q1777.41 1478.84 1777.41 1475.53 Q1777.41 1472.48 1775.25 1470.77 Q1773.12 1469.03 1769.3 1469.03 L1765.28 1469.03 L1765.28 1465.19 L1769.49 1465.19 Q1772.94 1465.19 1774.77 1463.82 Q1776.6 1462.43 1776.6 1459.84 Q1776.6 1457.18 1774.7 1455.77 Q1772.82 1454.33 1769.3 1454.33 Q1767.38 1454.33 1765.18 1454.75 Q1762.99 1455.16 1760.35 1456.04 L1760.35 1451.88 Q1763.01 1451.14 1765.32 1450.77 Q1767.66 1450.39 1769.72 1450.39 Q1775.05 1450.39 1778.15 1452.83 Q1781.25 1455.23 1781.25 1459.35 Q1781.25 1462.22 1779.61 1464.21 Q1777.96 1466.18 1774.93 1466.95 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M2293.4 1455.09 L2281.59 1473.54 L2293.4 1473.54 L2293.4 1455.09 M2292.17 1451.02 L2298.05 1451.02 L2298.05 1473.54 L2302.98 1473.54 L2302.98 1477.43 L2298.05 1477.43 L2298.05 1485.58 L2293.4 1485.58 L2293.4 1477.43 L2277.8 1477.43 L2277.8 1472.92 L2292.17 1451.02 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1199.28 1522.27 L1199.28 1532.4 L1211.35 1532.4 L1211.35 1536.95 L1199.28 1536.95 L1199.28 1556.3 Q1199.28 1560.66 1200.46 1561.9 Q1201.67 1563.14 1205.33 1563.14 L1211.35 1563.14 L1211.35 1568.04 L1205.33 1568.04 Q1198.55 1568.04 1195.97 1565.53 Q1193.4 1562.98 1193.4 1556.3 L1193.4 1536.95 L1189.1 1536.95 L1189.1 1532.4 L1193.4 1532.4 L1193.4 1522.27 L1199.28 1522.27 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1239.23 1518.52 L1252.72 1518.52 L1252.72 1523.07 L1245.09 1523.07 L1245.09 1572.09 L1252.72 1572.09 L1252.72 1576.64 L1239.23 1576.64 L1239.23 1518.52 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1287.93 1533.45 L1287.93 1538.98 Q1285.44 1537.71 1282.77 1537.07 Q1280.1 1536.44 1277.23 1536.44 Q1272.87 1536.44 1270.68 1537.77 Q1268.51 1539.11 1268.51 1541.79 Q1268.51 1543.82 1270.07 1545 Q1271.63 1546.15 1276.34 1547.2 L1278.35 1547.64 Q1284.58 1548.98 1287.19 1551.43 Q1289.84 1553.85 1289.84 1558.21 Q1289.84 1563.17 1285.89 1566.07 Q1281.97 1568.97 1275.1 1568.97 Q1272.23 1568.97 1269.12 1568.39 Q1266.03 1567.85 1262.59 1566.74 L1262.59 1560.69 Q1265.84 1562.38 1268.99 1563.24 Q1272.14 1564.07 1275.23 1564.07 Q1279.36 1564.07 1281.59 1562.66 Q1283.82 1561.23 1283.82 1558.65 Q1283.82 1556.27 1282.2 1554.99 Q1280.61 1553.72 1275.16 1552.54 L1273.13 1552.07 Q1267.68 1550.92 1265.26 1548.56 Q1262.85 1546.18 1262.85 1542.04 Q1262.85 1537.01 1266.41 1534.27 Q1269.98 1531.54 1276.53 1531.54 Q1279.78 1531.54 1282.64 1532.01 Q1285.51 1532.49 1287.93 1533.45 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1312.85 1518.52 L1312.85 1576.64 L1299.35 1576.64 L1299.35 1572.09 L1306.96 1572.09 L1306.96 1523.07 L1299.35 1523.07 L1299.35 1518.52 L1312.85 1518.52 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip302)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  149.191,1194.85 2352.76,1194.85 
  "/>
<polyline clip-path="url(#clip302)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  149.191,853.724 2352.76,853.724 
  "/>
<polyline clip-path="url(#clip302)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  149.191,512.596 2352.76,512.596 
  "/>
<polyline clip-path="url(#clip302)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  149.191,171.468 2352.76,171.468 
  "/>
<polyline clip-path="url(#clip300)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  149.191,1423.18 149.191,47.2441 
  "/>
<polyline clip-path="url(#clip300)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  149.191,1194.85 168.089,1194.85 
  "/>
<polyline clip-path="url(#clip300)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  149.191,853.724 168.089,853.724 
  "/>
<polyline clip-path="url(#clip300)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  149.191,512.596 168.089,512.596 
  "/>
<polyline clip-path="url(#clip300)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  149.191,171.468 168.089,171.468 
  "/>
<path clip-path="url(#clip300)" d="M51.3162 1195.3 L80.992 1195.3 L80.992 1199.24 L51.3162 1199.24 L51.3162 1195.3 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M101.663 1192.99 Q98.515 1192.99 96.6632 1195.14 Q94.8345 1197.29 94.8345 1201.04 Q94.8345 1204.77 96.6632 1206.95 Q98.515 1209.1 101.663 1209.1 Q104.811 1209.1 106.64 1206.95 Q108.492 1204.77 108.492 1201.04 Q108.492 1197.29 106.64 1195.14 Q104.811 1192.99 101.663 1192.99 M110.946 1178.34 L110.946 1182.6 Q109.186 1181.76 107.381 1181.32 Q105.598 1180.88 103.839 1180.88 Q99.2095 1180.88 96.7558 1184.01 Q94.3252 1187.13 93.978 1193.45 Q95.3437 1191.44 97.4039 1190.37 Q99.4641 1189.29 101.941 1189.29 Q107.149 1189.29 110.158 1192.46 Q113.191 1195.61 113.191 1201.04 Q113.191 1206.37 110.043 1209.59 Q106.895 1212.8 101.663 1212.8 Q95.6678 1212.8 92.4965 1208.22 Q89.3253 1203.61 89.3253 1194.89 Q89.3253 1186.69 93.2141 1181.83 Q97.103 1176.95 103.654 1176.95 Q105.413 1176.95 107.196 1177.29 Q109.001 1177.64 110.946 1178.34 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M50.9921 854.176 L80.6679 854.176 L80.6679 858.111 L50.9921 858.111 L50.9921 854.176 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M103.608 840.518 L91.8021 858.967 L103.608 858.967 L103.608 840.518 M102.381 836.444 L108.26 836.444 L108.26 858.967 L113.191 858.967 L113.191 862.856 L108.26 862.856 L108.26 871.004 L103.608 871.004 L103.608 862.856 L88.0058 862.856 L88.0058 858.342 L102.381 836.444 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M53.0754 513.047 L82.7512 513.047 L82.7512 516.983 L53.0754 516.983 L53.0754 513.047 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M96.8715 525.941 L113.191 525.941 L113.191 529.876 L91.2465 529.876 L91.2465 525.941 Q93.9086 523.186 98.4919 518.557 Q103.098 513.904 104.279 512.561 Q106.524 510.038 107.404 508.302 Q108.307 506.543 108.307 504.853 Q108.307 502.098 106.362 500.362 Q104.441 498.626 101.339 498.626 Q99.14 498.626 96.6863 499.39 Q94.2558 500.154 91.478 501.705 L91.478 496.983 Q94.3021 495.848 96.7558 495.27 Q99.2095 494.691 101.246 494.691 Q106.617 494.691 109.811 497.376 Q113.006 500.061 113.006 504.552 Q113.006 506.682 112.196 508.603 Q111.408 510.501 109.302 513.094 Q108.723 513.765 105.621 516.983 Q102.52 520.177 96.8715 525.941 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M101.246 157.266 Q97.6354 157.266 95.8067 160.831 Q94.0012 164.373 94.0012 171.502 Q94.0012 178.609 95.8067 182.174 Q97.6354 185.715 101.246 185.715 Q104.881 185.715 106.686 182.174 Q108.515 178.609 108.515 171.502 Q108.515 164.373 106.686 160.831 Q104.881 157.266 101.246 157.266 M101.246 153.563 Q107.057 153.563 110.112 158.169 Q113.191 162.753 113.191 171.502 Q113.191 180.229 110.112 184.836 Q107.057 189.419 101.246 189.419 Q95.4363 189.419 92.3576 184.836 Q89.3021 180.229 89.3021 171.502 Q89.3021 162.753 92.3576 158.169 Q95.4363 153.563 101.246 153.563 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip302)" style="stroke:#009af9; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  211.556,86.1857 216.753,89.2423 221.95,92.3212 227.147,95.4212 232.344,98.5406 237.541,101.678 242.738,104.833 247.936,108.003 253.133,111.189 258.33,114.387 
  263.527,117.598 268.724,120.82 273.921,124.053 279.118,127.296 284.315,130.547 289.512,133.806 294.709,137.072 299.906,140.345 305.103,143.624 310.301,146.908 
  315.498,150.197 320.695,153.49 325.892,156.787 331.089,160.087 336.286,163.39 341.483,166.695 346.68,170.003 351.877,173.312 357.074,176.623 362.271,179.936 
  367.469,183.249 372.666,186.563 377.863,189.877 383.06,193.192 388.257,196.508 393.454,199.823 398.651,203.138 403.848,206.453 409.045,209.767 414.242,213.082 
  419.439,216.395 424.636,219.708 429.834,223.02 435.031,226.332 440.228,229.643 445.425,232.952 450.622,236.261 455.819,239.569 461.016,242.875 466.213,246.181 
  471.41,249.486 476.607,252.789 481.804,256.092 487.002,259.393 492.199,262.693 497.396,265.992 502.593,269.29 507.79,272.587 512.987,275.883 518.184,279.177 
  523.381,282.47 528.578,285.763 533.775,289.054 538.972,292.344 544.17,295.632 549.367,298.92 554.564,302.207 559.761,305.492 564.958,308.777 570.155,312.06 
  575.352,315.342 580.549,318.624 585.746,321.904 590.943,325.183 596.14,328.461 601.337,331.739 606.535,335.015 611.732,338.291 616.929,341.565 622.126,344.839 
  627.323,348.111 632.52,351.383 637.717,354.654 642.914,357.924 648.111,361.193 653.308,364.462 658.505,367.73 663.703,370.996 668.9,374.263 674.097,377.528 
  679.294,380.793 684.491,384.056 689.688,387.32 694.885,390.582 700.082,393.844 705.279,397.105 710.476,400.366 715.673,403.626 720.87,406.885 726.068,410.143 
  731.265,413.401 736.462,416.659 741.659,419.916 746.856,423.172 752.053,426.428 757.25,429.683 762.447,432.938 767.644,436.192 772.841,439.446 778.038,442.699 
  783.236,445.952 788.433,449.204 793.63,452.456 798.827,455.707 804.024,458.958 809.221,462.209 814.418,465.459 819.615,468.709 824.812,471.958 830.009,475.207 
  835.206,478.455 840.403,481.703 845.601,484.951 850.798,488.199 855.995,491.446 861.192,494.693 866.389,497.939 871.586,501.185 876.783,504.431 881.98,507.676 
  887.177,510.922 892.374,514.166 897.571,517.411 902.769,520.655 907.966,523.899 913.163,527.143 918.36,530.387 923.557,533.63 928.754,536.873 933.951,540.116 
  939.148,543.358 944.345,546.6 949.542,549.842 954.739,553.084 959.936,556.326 965.134,559.567 970.331,562.808 975.528,566.049 980.725,569.29 985.922,572.53 
  991.119,575.771 996.316,579.011 1001.51,582.251 1006.71,585.491 1011.91,588.73 1017.1,591.97 1022.3,595.209 1027.5,598.448 1032.7,601.687 1037.89,604.926 
  1043.09,608.165 1048.29,611.403 1053.48,614.642 1058.68,617.88 1063.88,621.118 1069.08,624.356 1074.27,627.594 1079.47,630.832 1084.67,634.07 1089.86,637.307 
  1095.06,640.545 1100.26,643.782 1105.45,647.019 1110.65,650.256 1115.85,653.493 1121.05,656.73 1126.24,659.967 1131.44,663.203 1136.64,666.44 1141.83,669.676 
  1147.03,672.913 1152.23,676.149 1157.43,679.385 1162.62,682.622 1167.82,685.858 1173.02,689.094 1178.21,692.33 1183.41,695.565 1188.61,698.801 1193.81,702.037 
  1199,705.272 1204.2,708.508 1209.4,711.743 1214.59,714.979 1219.79,718.214 1224.99,721.449 1230.19,724.684 1235.38,727.92 1240.58,731.155 1245.78,734.39 
  1250.97,737.625 1256.17,740.859 1261.37,744.094 1266.56,747.329 1271.76,750.564 1276.96,753.799 1282.16,757.033 1287.35,760.268 1292.55,763.502 1297.75,766.737 
  1302.94,769.971 1308.14,773.206 1313.34,776.44 1318.54,779.674 1323.73,782.909 1328.93,786.143 1334.13,789.377 1339.32,792.611 1344.52,795.845 1349.72,799.079 
  1354.92,802.314 1360.11,805.548 1365.31,808.782 1370.51,812.015 1375.7,815.249 1380.9,818.483 1386.1,821.717 1391.29,824.951 1396.49,828.185 1401.69,831.418 
  1406.89,834.652 1412.08,837.886 1417.28,841.12 1422.48,844.353 1427.67,847.587 1432.87,850.82 1438.07,854.054 1443.27,857.287 1448.46,860.521 1453.66,863.754 
  1458.86,866.988 1464.05,870.221 1469.25,873.455 1474.45,876.688 1479.65,879.922 1484.84,883.155 1490.04,886.388 1495.24,889.622 1500.43,892.855 1505.63,896.088 
  1510.83,899.322 1516.02,902.555 1521.22,905.788 1526.42,909.021 1531.62,912.255 1536.81,915.488 1542.01,918.721 1547.21,921.954 1552.4,925.187 1557.6,928.42 
  1562.8,931.654 1568,934.887 1573.19,938.12 1578.39,941.353 1583.59,944.586 1588.78,947.819 1593.98,951.052 1599.18,954.285 1604.38,957.518 1609.57,960.751 
  1614.77,963.984 1619.97,967.217 1625.16,970.45 1630.36,973.683 1635.56,976.916 1640.75,980.149 1645.95,983.382 1651.15,986.615 1656.35,989.848 1661.54,993.081 
  1666.74,996.314 1671.94,999.547 1677.13,1002.78 1682.33,1006.01 1687.53,1009.25 1692.73,1012.48 1697.92,1015.71 1703.12,1018.94 1708.32,1022.18 1713.51,1025.41 
  1718.71,1028.64 1723.91,1031.88 1729.11,1035.11 1734.3,1038.34 1739.5,1041.57 1744.7,1044.81 1749.89,1048.04 1755.09,1051.27 1760.29,1054.51 1765.49,1057.74 
  1770.68,1060.97 1775.88,1064.2 1781.08,1067.44 1786.27,1070.67 1791.47,1073.9 1796.67,1077.14 1801.86,1080.37 1807.06,1083.6 1812.26,1086.83 1817.46,1090.07 
  1822.65,1093.3 1827.85,1096.53 1833.05,1099.76 1838.24,1103 1843.44,1106.23 1848.64,1109.46 1853.84,1112.7 1859.03,1115.93 1864.23,1119.16 1869.43,1122.39 
  1874.62,1125.63 1879.82,1128.86 1885.02,1132.09 1890.22,1135.32 1895.41,1138.56 1900.61,1141.79 1905.81,1145.02 1911,1148.26 1916.2,1151.49 1921.4,1154.72 
  1926.59,1157.95 1931.79,1161.19 1936.99,1164.42 1942.19,1167.65 1947.38,1170.88 1952.58,1174.12 1957.78,1177.35 1962.97,1180.58 1968.17,1183.82 1973.37,1187.05 
  1978.57,1190.28 1983.76,1193.51 1988.96,1196.75 1994.16,1199.98 1999.35,1203.21 2004.55,1206.44 2009.75,1209.68 2014.95,1212.91 2020.14,1216.14 2025.34,1219.37 
  2030.54,1222.61 2035.73,1225.84 2040.93,1229.07 2046.13,1232.3 2051.32,1235.54 2056.52,1238.77 2061.72,1242 2066.92,1245.24 2072.11,1248.47 2077.31,1251.7 
  2082.51,1254.93 2087.7,1258.17 2092.9,1261.4 2098.1,1264.63 2103.3,1267.86 2108.49,1271.1 2113.69,1274.33 2118.89,1277.56 2124.08,1280.79 2129.28,1284.03 
  2134.48,1287.26 2139.68,1290.49 2144.87,1293.72 2150.07,1296.96 2155.27,1300.19 2160.46,1303.42 2165.66,1306.66 2170.86,1309.89 2176.05,1313.12 2181.25,1316.35 
  2186.45,1319.59 2191.65,1322.82 2196.84,1326.05 2202.04,1329.28 2207.24,1332.52 2212.43,1335.75 2217.63,1338.98 2222.83,1342.21 2228.03,1345.45 2233.22,1348.68 
  2238.42,1351.91 2243.62,1355.14 2248.81,1358.38 2254.01,1361.61 2259.21,1364.84 2264.41,1368.08 2269.6,1371.31 2274.8,1374.54 2280,1377.77 2285.19,1381.01 
  2290.39,1384.24 
  "/>
<polyline clip-path="url(#clip302)" style="stroke:#e26f46; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  211.556,171.468 216.753,173.367 221.95,175.26 227.147,177.146 232.344,179.026 237.541,180.901 242.738,182.771 247.936,184.636 253.133,186.498 258.33,188.356 
  263.527,190.211 268.724,192.064 273.921,193.914 279.118,195.764 284.315,197.612 289.512,199.459 294.709,201.305 299.906,203.152 305.103,204.998 310.301,206.845 
  315.498,208.692 320.695,210.541 325.892,212.39 331.089,214.24 336.286,216.092 341.483,217.946 346.68,219.801 351.877,221.658 357.074,223.518 362.271,225.379 
  367.469,227.242 372.666,229.108 377.863,230.975 383.06,232.846 388.257,234.718 393.454,236.593 398.651,238.471 403.848,240.351 409.045,242.233 414.242,244.118 
  419.439,246.006 424.636,247.896 429.834,249.789 435.031,251.684 440.228,253.582 445.425,255.483 450.622,257.386 455.819,259.291 461.016,261.199 466.213,263.109 
  471.41,265.022 476.607,266.937 481.804,268.855 487.002,270.775 492.199,272.697 497.396,274.621 502.593,276.548 507.79,278.477 512.987,280.408 518.184,282.342 
  523.381,284.277 528.578,286.215 533.775,288.154 538.972,290.096 544.17,292.04 549.367,293.985 554.564,295.933 559.761,297.882 564.958,299.833 570.155,301.786 
  575.352,303.741 580.549,305.698 585.746,307.656 590.943,309.616 596.14,311.578 601.337,313.541 606.535,315.506 611.732,317.472 616.929,319.44 622.126,321.409 
  627.323,323.38 632.52,325.352 637.717,327.326 642.914,329.301 648.111,331.277 653.308,333.254 658.505,335.233 663.703,337.213 668.9,339.195 674.097,341.177 
  679.294,343.161 684.491,345.146 689.688,347.132 694.885,349.119 700.082,351.107 705.279,353.096 710.476,355.087 715.673,357.078 720.87,359.07 726.068,361.063 
  731.265,363.057 736.462,365.052 741.659,367.048 746.856,369.045 752.053,371.043 757.25,373.041 762.447,375.04 767.644,377.04 772.841,379.041 778.038,381.043 
  783.236,383.045 788.433,385.048 793.63,387.052 798.827,389.057 804.024,391.062 809.221,393.068 814.418,395.074 819.615,397.081 824.812,399.089 830.009,401.097 
  835.206,403.106 840.403,405.116 845.601,407.126 850.798,409.136 855.995,411.147 861.192,413.159 866.389,415.171 871.586,417.184 876.783,419.197 881.98,421.211 
  887.177,423.225 892.374,425.239 897.571,427.254 902.769,429.27 907.966,431.285 913.163,433.302 918.36,435.318 923.557,437.335 928.754,439.353 933.951,441.371 
  939.148,443.389 944.345,445.408 949.542,447.427 954.739,449.446 959.936,451.466 965.134,453.485 970.331,455.506 975.528,457.526 980.725,459.547 985.922,461.569 
  991.119,463.59 996.316,465.612 1001.51,467.634 1006.71,469.656 1011.91,471.679 1017.1,473.702 1022.3,475.725 1027.5,477.748 1032.7,479.772 1037.89,481.796 
  1043.09,483.82 1048.29,485.844 1053.48,487.869 1058.68,489.894 1063.88,491.918 1069.08,493.944 1074.27,495.969 1079.47,497.994 1084.67,500.02 1089.86,502.046 
  1095.06,504.072 1100.26,506.098 1105.45,508.125 1110.65,510.151 1115.85,512.178 1121.05,514.205 1126.24,516.232 1131.44,518.259 1136.64,520.287 1141.83,522.314 
  1147.03,524.342 1152.23,526.369 1157.43,528.397 1162.62,530.425 1167.82,532.453 1173.02,534.482 1178.21,536.51 1183.41,538.539 1188.61,540.567 1193.81,542.596 
  1199,544.625 1204.2,546.654 1209.4,548.683 1214.59,550.712 1219.79,552.741 1224.99,554.771 1230.19,556.8 1235.38,558.83 1240.58,560.86 1245.78,562.889 
  1250.97,564.919 1256.17,566.949 1261.37,568.979 1266.56,571.009 1271.76,573.039 1276.96,575.07 1282.16,577.1 1287.35,579.13 1292.55,581.161 1297.75,583.191 
  1302.94,585.222 1308.14,587.253 1313.34,589.284 1318.54,591.314 1323.73,593.345 1328.93,595.376 1334.13,597.407 1339.32,599.438 1344.52,601.469 1349.72,603.501 
  1354.92,605.532 1360.11,607.563 1365.31,609.594 1370.51,611.626 1375.7,613.657 1380.9,615.689 1386.1,617.72 1391.29,619.752 1396.49,621.784 1401.69,623.815 
  1406.89,625.847 1412.08,627.879 1417.28,629.911 1422.48,631.943 1427.67,633.974 1432.87,636.006 1438.07,638.038 1443.27,640.07 1448.46,642.103 1453.66,644.135 
  1458.86,646.167 1464.05,648.199 1469.25,650.231 1474.45,652.263 1479.65,654.296 1484.84,656.328 1490.04,658.36 1495.24,660.392 1500.43,662.425 1505.63,664.457 
  1510.83,666.49 1516.02,668.522 1521.22,670.555 1526.42,672.587 1531.62,674.62 1536.81,676.652 1542.01,678.685 1547.21,680.717 1552.4,682.75 1557.6,684.783 
  1562.8,686.815 1568,688.848 1573.19,690.881 1578.39,692.913 1583.59,694.946 1588.78,696.979 1593.98,699.011 1599.18,701.044 1604.38,703.077 1609.57,705.11 
  1614.77,707.143 1619.97,709.175 1625.16,711.208 1630.36,713.241 1635.56,715.274 1640.75,717.307 1645.95,719.34 1651.15,721.373 1656.35,723.406 1661.54,725.439 
  1666.74,727.471 1671.94,729.504 1677.13,731.537 1682.33,733.57 1687.53,735.603 1692.73,737.636 1697.92,739.669 1703.12,741.702 1708.32,743.735 1713.51,745.769 
  1718.71,747.802 1723.91,749.835 1729.11,751.868 1734.3,753.901 1739.5,755.934 1744.7,757.967 1749.89,760 1755.09,762.033 1760.29,764.067 1765.49,766.1 
  1770.68,768.133 1775.88,770.166 1781.08,772.199 1786.27,774.232 1791.47,776.266 1796.67,778.299 1801.86,780.332 1807.06,782.365 1812.26,784.398 1817.46,786.432 
  1822.65,788.465 1827.85,790.498 1833.05,792.531 1838.24,794.565 1843.44,796.598 1848.64,798.631 1853.84,800.664 1859.03,802.698 1864.23,804.731 1869.43,806.764 
  1874.62,808.797 1879.82,810.831 1885.02,812.864 1890.22,814.897 1895.41,816.931 1900.61,818.964 1905.81,820.997 1911,823.031 1916.2,825.064 1921.4,827.097 
  1926.59,829.131 1931.79,831.164 1936.99,833.197 1942.19,835.231 1947.38,837.264 1952.58,839.297 1957.78,841.331 1962.97,843.364 1968.17,845.397 1973.37,847.431 
  1978.57,849.464 1983.76,851.497 1988.96,853.531 1994.16,855.564 1999.35,857.598 2004.55,859.631 2009.75,861.664 2014.95,863.698 2020.14,865.731 2025.34,867.764 
  2030.54,869.798 2035.73,871.831 2040.93,873.865 2046.13,875.898 2051.32,877.931 2056.52,879.965 2061.72,881.998 2066.92,884.032 2072.11,886.065 2077.31,888.098 
  2082.51,890.132 2087.7,892.165 2092.9,894.199 2098.1,896.232 2103.3,898.266 2108.49,900.299 2113.69,902.332 2118.89,904.366 2124.08,906.399 2129.28,908.433 
  2134.48,910.466 2139.68,912.5 2144.87,914.533 2150.07,916.566 2155.27,918.6 2160.46,920.633 2165.66,922.667 2170.86,924.7 2176.05,926.734 2181.25,928.767 
  2186.45,930.8 2191.65,932.834 2196.84,934.867 2202.04,936.901 2207.24,938.934 2212.43,940.968 2217.63,943.001 2222.83,945.035 2228.03,947.068 2233.22,949.101 
  2238.42,951.135 2243.62,953.168 2248.81,955.202 2254.01,957.235 2259.21,959.269 2264.41,961.302 2269.6,963.336 2274.8,965.369 2280,967.403 2285.19,969.436 
  2290.39,971.469 
  "/>
<path clip-path="url(#clip300)" d="
M1581.73 248.629 L2279.3 248.629 L2279.3 93.1086 L1581.73 93.1086  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<polyline clip-path="url(#clip300)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1581.73,248.629 2279.3,248.629 2279.3,93.1086 1581.73,93.1086 1581.73,248.629 
  "/>
<polyline clip-path="url(#clip300)" style="stroke:#009af9; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1606.22,144.949 1753.12,144.949 
  "/>
<path clip-path="url(#clip300)" d="M1797.95 141.28 Q1799.55 138.409 1801.77 137.044 Q1804 135.678 1807 135.678 Q1811.06 135.678 1813.25 138.525 Q1815.45 141.349 1815.45 146.581 L1815.45 162.229 L1811.17 162.229 L1811.17 146.719 Q1811.17 142.993 1809.85 141.187 Q1808.53 139.382 1805.82 139.382 Q1802.51 139.382 1800.59 141.581 Q1798.67 143.78 1798.67 147.576 L1798.67 162.229 L1794.39 162.229 L1794.39 146.719 Q1794.39 142.969 1793.07 141.187 Q1791.75 139.382 1789 139.382 Q1785.73 139.382 1783.81 141.604 Q1781.89 143.803 1781.89 147.576 L1781.89 162.229 L1777.61 162.229 L1777.61 136.303 L1781.89 136.303 L1781.89 140.331 Q1783.35 137.946 1785.38 136.812 Q1787.42 135.678 1790.22 135.678 Q1793.05 135.678 1795.01 137.113 Q1797 138.548 1797.95 141.28 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1835.73 149.196 Q1830.57 149.196 1828.58 150.377 Q1826.59 151.557 1826.59 154.405 Q1826.59 156.673 1828.07 158.016 Q1829.57 159.335 1832.14 159.335 Q1835.68 159.335 1837.81 156.835 Q1839.97 154.312 1839.97 150.145 L1839.97 149.196 L1835.73 149.196 M1844.23 147.437 L1844.23 162.229 L1839.97 162.229 L1839.97 158.293 Q1838.51 160.655 1836.33 161.789 Q1834.16 162.9 1831.01 162.9 Q1827.03 162.9 1824.67 160.678 Q1822.33 158.432 1822.33 154.682 Q1822.33 150.307 1825.25 148.085 Q1828.19 145.863 1834 145.863 L1839.97 145.863 L1839.97 145.446 Q1839.97 142.507 1838.02 140.909 Q1836.1 139.289 1832.61 139.289 Q1830.38 139.289 1828.28 139.821 Q1826.17 140.354 1824.23 141.419 L1824.23 137.483 Q1826.56 136.581 1828.76 136.141 Q1830.96 135.678 1833.05 135.678 Q1838.67 135.678 1841.45 138.594 Q1844.23 141.511 1844.23 147.437 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1869.53 137.067 L1869.53 141.094 Q1867.72 140.169 1865.78 139.706 Q1863.83 139.243 1861.75 139.243 Q1858.58 139.243 1856.98 140.215 Q1855.41 141.187 1855.41 143.131 Q1855.41 144.613 1856.54 145.469 Q1857.68 146.303 1861.1 147.067 L1862.56 147.391 Q1867.1 148.363 1868.99 150.145 Q1870.92 151.905 1870.92 155.076 Q1870.92 158.687 1868.05 160.793 Q1865.2 162.9 1860.2 162.9 Q1858.12 162.9 1855.85 162.483 Q1853.6 162.09 1851.1 161.28 L1851.1 156.881 Q1853.46 158.108 1855.75 158.733 Q1858.05 159.335 1860.29 159.335 Q1863.3 159.335 1864.92 158.317 Q1866.54 157.275 1866.54 155.4 Q1866.54 153.664 1865.36 152.738 Q1864.2 151.812 1860.25 150.956 L1858.76 150.608 Q1854.81 149.775 1853.05 148.062 Q1851.29 146.326 1851.29 143.317 Q1851.29 139.659 1853.88 137.669 Q1856.47 135.678 1861.24 135.678 Q1863.6 135.678 1865.68 136.025 Q1867.77 136.372 1869.53 137.067 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1894.23 137.067 L1894.23 141.094 Q1892.42 140.169 1890.48 139.706 Q1888.53 139.243 1886.45 139.243 Q1883.28 139.243 1881.68 140.215 Q1880.11 141.187 1880.11 143.131 Q1880.11 144.613 1881.24 145.469 Q1882.37 146.303 1885.8 147.067 L1887.26 147.391 Q1891.8 148.363 1893.69 150.145 Q1895.62 151.905 1895.62 155.076 Q1895.62 158.687 1892.74 160.793 Q1889.9 162.9 1884.9 162.9 Q1882.81 162.9 1880.55 162.483 Q1878.3 162.09 1875.8 161.28 L1875.8 156.881 Q1878.16 158.108 1880.45 158.733 Q1882.74 159.335 1884.99 159.335 Q1888 159.335 1889.62 158.317 Q1891.24 157.275 1891.24 155.4 Q1891.24 153.664 1890.06 152.738 Q1888.9 151.812 1884.94 150.956 L1883.46 150.608 Q1879.5 149.775 1877.74 148.062 Q1875.99 146.326 1875.99 143.317 Q1875.99 139.659 1878.58 137.669 Q1881.17 135.678 1885.94 135.678 Q1888.3 135.678 1890.38 136.025 Q1892.47 136.372 1894.23 137.067 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1903 156.349 L1907.88 156.349 L1907.88 162.229 L1903 162.229 L1903 156.349 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1933.99 137.067 L1933.99 141.094 Q1932.19 140.169 1930.24 139.706 Q1928.3 139.243 1926.22 139.243 Q1923.05 139.243 1921.45 140.215 Q1919.87 141.187 1919.87 143.131 Q1919.87 144.613 1921.01 145.469 Q1922.14 146.303 1925.57 147.067 L1927.03 147.391 Q1931.56 148.363 1933.46 150.145 Q1935.38 151.905 1935.38 155.076 Q1935.38 158.687 1932.51 160.793 Q1929.67 162.9 1924.67 162.9 Q1922.58 162.9 1920.31 162.483 Q1918.07 162.09 1915.57 161.28 L1915.57 156.881 Q1917.93 158.108 1920.22 158.733 Q1922.51 159.335 1924.76 159.335 Q1927.77 159.335 1929.39 158.317 Q1931.01 157.275 1931.01 155.4 Q1931.01 153.664 1929.83 152.738 Q1928.67 151.812 1924.71 150.956 L1923.23 150.608 Q1919.27 149.775 1917.51 148.062 Q1915.75 146.326 1915.75 143.317 Q1915.75 139.659 1918.35 137.669 Q1920.94 135.678 1925.71 135.678 Q1928.07 135.678 1930.15 136.025 Q1932.24 136.372 1933.99 137.067 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1967.47 126.257 Q1964.36 131.581 1962.86 136.789 Q1961.36 141.997 1961.36 147.344 Q1961.36 152.692 1962.86 157.946 Q1964.39 163.178 1967.47 168.479 L1963.76 168.479 Q1960.29 163.039 1958.55 157.784 Q1956.84 152.53 1956.84 147.344 Q1956.84 142.182 1958.55 136.951 Q1960.27 131.72 1963.76 126.257 L1967.47 126.257 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1990.5 143.594 Q1993.86 144.312 1995.73 146.581 Q1997.63 148.849 1997.63 152.182 Q1997.63 157.298 1994.11 160.099 Q1990.59 162.9 1984.11 162.9 Q1981.93 162.9 1979.62 162.46 Q1977.33 162.043 1974.87 161.187 L1974.87 156.673 Q1976.82 157.807 1979.13 158.386 Q1981.45 158.965 1983.97 158.965 Q1988.37 158.965 1990.66 157.229 Q1992.98 155.493 1992.98 152.182 Q1992.98 149.127 1990.82 147.414 Q1988.69 145.678 1984.87 145.678 L1980.85 145.678 L1980.85 141.835 L1985.06 141.835 Q1988.51 141.835 1990.34 140.469 Q1992.17 139.081 1992.17 136.488 Q1992.17 133.826 1990.27 132.414 Q1988.39 130.979 1984.87 130.979 Q1982.95 130.979 1980.75 131.395 Q1978.55 131.812 1975.92 132.692 L1975.92 128.525 Q1978.58 127.784 1980.89 127.414 Q1983.23 127.044 1985.29 127.044 Q1990.61 127.044 1993.72 129.474 Q1996.82 131.882 1996.82 136.002 Q1996.82 138.872 1995.17 140.863 Q1993.53 142.831 1990.5 143.594 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M2020.66 143.594 Q2024.02 144.312 2025.89 146.581 Q2027.79 148.849 2027.79 152.182 Q2027.79 157.298 2024.27 160.099 Q2020.75 162.9 2014.27 162.9 Q2012.1 162.9 2009.78 162.46 Q2007.49 162.043 2005.04 161.187 L2005.04 156.673 Q2006.98 157.807 2009.3 158.386 Q2011.61 158.965 2014.13 158.965 Q2018.53 158.965 2020.82 157.229 Q2023.14 155.493 2023.14 152.182 Q2023.14 149.127 2020.98 147.414 Q2018.86 145.678 2015.04 145.678 L2011.01 145.678 L2011.01 141.835 L2015.22 141.835 Q2018.67 141.835 2020.5 140.469 Q2022.33 139.081 2022.33 136.488 Q2022.33 133.826 2020.43 132.414 Q2018.55 130.979 2015.04 130.979 Q2013.11 130.979 2010.92 131.395 Q2008.72 131.812 2006.08 132.692 L2006.08 128.525 Q2008.74 127.784 2011.05 127.414 Q2013.39 127.044 2015.45 127.044 Q2020.78 127.044 2023.88 129.474 Q2026.98 131.882 2026.98 136.002 Q2026.98 138.872 2025.34 140.863 Q2023.69 142.831 2020.66 143.594 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M2036.7 127.669 L2055.06 127.669 L2055.06 131.604 L2040.98 131.604 L2040.98 140.076 Q2042 139.729 2043.02 139.567 Q2044.04 139.382 2045.06 139.382 Q2050.85 139.382 2054.23 142.553 Q2057.61 145.724 2057.61 151.141 Q2057.61 156.719 2054.13 159.821 Q2050.66 162.9 2044.34 162.9 Q2042.17 162.9 2039.9 162.53 Q2037.65 162.159 2035.24 161.418 L2035.24 156.719 Q2037.33 157.854 2039.55 158.409 Q2041.77 158.965 2044.25 158.965 Q2048.25 158.965 2050.59 156.858 Q2052.93 154.752 2052.93 151.141 Q2052.93 147.53 2050.59 145.423 Q2048.25 143.317 2044.25 143.317 Q2042.37 143.317 2040.5 143.733 Q2038.65 144.15 2036.7 145.03 L2036.7 127.669 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M2066.86 127.669 L2085.22 127.669 L2085.22 131.604 L2071.15 131.604 L2071.15 140.076 Q2072.17 139.729 2073.18 139.567 Q2074.2 139.382 2075.22 139.382 Q2081.01 139.382 2084.39 142.553 Q2087.77 145.724 2087.77 151.141 Q2087.77 156.719 2084.29 159.821 Q2080.82 162.9 2074.5 162.9 Q2072.33 162.9 2070.06 162.53 Q2067.81 162.159 2065.41 161.418 L2065.41 156.719 Q2067.49 157.854 2069.71 158.409 Q2071.93 158.965 2074.41 158.965 Q2078.42 158.965 2080.75 156.858 Q2083.09 154.752 2083.09 151.141 Q2083.09 147.53 2080.75 145.423 Q2078.42 143.317 2074.41 143.317 Q2072.54 143.317 2070.66 143.733 Q2068.81 144.15 2066.86 145.03 L2066.86 127.669 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M2109.83 131.743 L2098.02 150.192 L2109.83 150.192 L2109.83 131.743 M2108.6 127.669 L2114.48 127.669 L2114.48 150.192 L2119.41 150.192 L2119.41 154.081 L2114.48 154.081 L2114.48 162.229 L2109.83 162.229 L2109.83 154.081 L2094.23 154.081 L2094.23 149.567 L2108.6 127.669 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M2139.99 131.743 L2128.18 150.192 L2139.99 150.192 L2139.99 131.743 M2138.76 127.669 L2144.64 127.669 L2144.64 150.192 L2149.57 150.192 L2149.57 154.081 L2144.64 154.081 L2144.64 162.229 L2139.99 162.229 L2139.99 154.081 L2124.39 154.081 L2124.39 149.567 L2138.76 127.669 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M2171.47 143.594 Q2174.83 144.312 2176.7 146.581 Q2178.6 148.849 2178.6 152.182 Q2178.6 157.298 2175.08 160.099 Q2171.56 162.9 2165.08 162.9 Q2162.91 162.9 2160.59 162.46 Q2158.3 162.043 2155.85 161.187 L2155.85 156.673 Q2157.79 157.807 2160.1 158.386 Q2162.42 158.965 2164.94 158.965 Q2169.34 158.965 2171.63 157.229 Q2173.95 155.493 2173.95 152.182 Q2173.95 149.127 2171.79 147.414 Q2169.66 145.678 2165.85 145.678 L2161.82 145.678 L2161.82 141.835 L2166.03 141.835 Q2169.48 141.835 2171.31 140.469 Q2173.14 139.081 2173.14 136.488 Q2173.14 133.826 2171.24 132.414 Q2169.36 130.979 2165.85 130.979 Q2163.92 130.979 2161.72 131.395 Q2159.53 131.812 2156.89 132.692 L2156.89 128.525 Q2159.55 127.784 2161.86 127.414 Q2164.2 127.044 2166.26 127.044 Q2171.59 127.044 2174.69 129.474 Q2177.79 131.882 2177.79 136.002 Q2177.79 138.872 2176.15 140.863 Q2174.5 142.831 2171.47 143.594 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M2191.49 158.293 L2207.81 158.293 L2207.81 162.229 L2185.87 162.229 L2185.87 158.293 Q2188.53 155.539 2193.11 150.909 Q2197.72 146.256 2198.9 144.914 Q2201.15 142.391 2202.03 140.655 Q2202.93 138.895 2202.93 137.206 Q2202.93 134.451 2200.98 132.715 Q2199.06 130.979 2195.96 130.979 Q2193.76 130.979 2191.31 131.743 Q2188.88 132.507 2186.1 134.057 L2186.1 129.335 Q2188.92 128.201 2191.38 127.622 Q2193.83 127.044 2195.87 127.044 Q2201.24 127.044 2204.43 129.729 Q2207.63 132.414 2207.63 136.905 Q2207.63 139.034 2206.82 140.956 Q2206.03 142.854 2203.92 145.446 Q2203.35 146.118 2200.24 149.335 Q2197.14 152.53 2191.49 158.293 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M2216.35 126.257 L2220.06 126.257 Q2223.53 131.72 2225.24 136.951 Q2226.98 142.182 2226.98 147.344 Q2226.98 152.53 2225.24 157.784 Q2223.53 163.039 2220.06 168.479 L2216.35 168.479 Q2219.43 163.178 2220.94 157.946 Q2222.47 152.692 2222.47 147.344 Q2222.47 141.997 2220.94 136.789 Q2219.43 131.581 2216.35 126.257 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip300)" style="stroke:#e26f46; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1606.22,196.789 1753.12,196.789 
  "/>
<path clip-path="url(#clip300)" d="M1797.95 193.12 Q1799.55 190.249 1801.77 188.884 Q1804 187.518 1807 187.518 Q1811.06 187.518 1813.25 190.365 Q1815.45 193.189 1815.45 198.421 L1815.45 214.069 L1811.17 214.069 L1811.17 198.559 Q1811.17 194.833 1809.85 193.027 Q1808.53 191.222 1805.82 191.222 Q1802.51 191.222 1800.59 193.421 Q1798.67 195.62 1798.67 199.416 L1798.67 214.069 L1794.39 214.069 L1794.39 198.559 Q1794.39 194.809 1793.07 193.027 Q1791.75 191.222 1789 191.222 Q1785.73 191.222 1783.81 193.444 Q1781.89 195.643 1781.89 199.416 L1781.89 214.069 L1777.61 214.069 L1777.61 188.143 L1781.89 188.143 L1781.89 192.171 Q1783.35 189.786 1785.38 188.652 Q1787.42 187.518 1790.22 187.518 Q1793.05 187.518 1795.01 188.953 Q1797 190.388 1797.95 193.12 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1835.73 201.036 Q1830.57 201.036 1828.58 202.217 Q1826.59 203.397 1826.59 206.245 Q1826.59 208.513 1828.07 209.856 Q1829.57 211.175 1832.14 211.175 Q1835.68 211.175 1837.81 208.675 Q1839.97 206.152 1839.97 201.985 L1839.97 201.036 L1835.73 201.036 M1844.23 199.277 L1844.23 214.069 L1839.97 214.069 L1839.97 210.133 Q1838.51 212.495 1836.33 213.629 Q1834.16 214.74 1831.01 214.74 Q1827.03 214.74 1824.67 212.518 Q1822.33 210.272 1822.33 206.522 Q1822.33 202.147 1825.25 199.925 Q1828.19 197.703 1834 197.703 L1839.97 197.703 L1839.97 197.286 Q1839.97 194.347 1838.02 192.749 Q1836.1 191.129 1832.61 191.129 Q1830.38 191.129 1828.28 191.661 Q1826.17 192.194 1824.23 193.259 L1824.23 189.323 Q1826.56 188.421 1828.76 187.981 Q1830.96 187.518 1833.05 187.518 Q1838.67 187.518 1841.45 190.434 Q1844.23 193.351 1844.23 199.277 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1869.53 188.907 L1869.53 192.934 Q1867.72 192.009 1865.78 191.546 Q1863.83 191.083 1861.75 191.083 Q1858.58 191.083 1856.98 192.055 Q1855.41 193.027 1855.41 194.971 Q1855.41 196.453 1856.54 197.309 Q1857.68 198.143 1861.1 198.907 L1862.56 199.231 Q1867.1 200.203 1868.99 201.985 Q1870.92 203.745 1870.92 206.916 Q1870.92 210.527 1868.05 212.633 Q1865.2 214.74 1860.2 214.74 Q1858.12 214.74 1855.85 214.323 Q1853.6 213.93 1851.1 213.12 L1851.1 208.721 Q1853.46 209.948 1855.75 210.573 Q1858.05 211.175 1860.29 211.175 Q1863.3 211.175 1864.92 210.157 Q1866.54 209.115 1866.54 207.24 Q1866.54 205.504 1865.36 204.578 Q1864.2 203.652 1860.25 202.796 L1858.76 202.448 Q1854.81 201.615 1853.05 199.902 Q1851.29 198.166 1851.29 195.157 Q1851.29 191.499 1853.88 189.509 Q1856.47 187.518 1861.24 187.518 Q1863.6 187.518 1865.68 187.865 Q1867.77 188.212 1869.53 188.907 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1894.23 188.907 L1894.23 192.934 Q1892.42 192.009 1890.48 191.546 Q1888.53 191.083 1886.45 191.083 Q1883.28 191.083 1881.68 192.055 Q1880.11 193.027 1880.11 194.971 Q1880.11 196.453 1881.24 197.309 Q1882.37 198.143 1885.8 198.907 L1887.26 199.231 Q1891.8 200.203 1893.69 201.985 Q1895.62 203.745 1895.62 206.916 Q1895.62 210.527 1892.74 212.633 Q1889.9 214.74 1884.9 214.74 Q1882.81 214.74 1880.55 214.323 Q1878.3 213.93 1875.8 213.12 L1875.8 208.721 Q1878.16 209.948 1880.45 210.573 Q1882.74 211.175 1884.99 211.175 Q1888 211.175 1889.62 210.157 Q1891.24 209.115 1891.24 207.24 Q1891.24 205.504 1890.06 204.578 Q1888.9 203.652 1884.94 202.796 L1883.46 202.448 Q1879.5 201.615 1877.74 199.902 Q1875.99 198.166 1875.99 195.157 Q1875.99 191.499 1878.58 189.509 Q1881.17 187.518 1885.94 187.518 Q1888.3 187.518 1890.38 187.865 Q1892.47 188.212 1894.23 188.907 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1903 208.189 L1907.88 208.189 L1907.88 214.069 L1903 214.069 L1903 208.189 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1914.41 188.143 L1918.93 188.143 L1927.03 209.902 L1935.13 188.143 L1939.64 188.143 L1929.92 214.069 L1924.13 214.069 L1914.41 188.143 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1970.82 178.097 Q1967.72 183.421 1966.22 188.629 Q1964.71 193.837 1964.71 199.184 Q1964.71 204.532 1966.22 209.786 Q1967.74 215.018 1970.82 220.319 L1967.12 220.319 Q1963.65 214.879 1961.91 209.624 Q1960.2 204.37 1960.2 199.184 Q1960.2 194.022 1961.91 188.791 Q1963.62 183.56 1967.12 178.097 L1970.82 178.097 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1993.86 195.434 Q1997.21 196.152 1999.09 198.421 Q2000.99 200.689 2000.99 204.022 Q2000.99 209.138 1997.47 211.939 Q1993.95 214.74 1987.47 214.74 Q1985.29 214.74 1982.98 214.3 Q1980.68 213.883 1978.23 213.027 L1978.23 208.513 Q1980.17 209.647 1982.49 210.226 Q1984.8 210.805 1987.33 210.805 Q1991.73 210.805 1994.02 209.069 Q1996.33 207.333 1996.33 204.022 Q1996.33 200.967 1994.18 199.254 Q1992.05 197.518 1988.23 197.518 L1984.2 197.518 L1984.2 193.675 L1988.42 193.675 Q1991.86 193.675 1993.69 192.309 Q1995.52 190.921 1995.52 188.328 Q1995.52 185.666 1993.62 184.254 Q1991.75 182.819 1988.23 182.819 Q1986.31 182.819 1984.11 183.235 Q1981.91 183.652 1979.27 184.532 L1979.27 180.365 Q1981.93 179.624 1984.25 179.254 Q1986.59 178.884 1988.65 178.884 Q1993.97 178.884 1997.07 181.314 Q2000.17 183.722 2000.17 187.842 Q2000.17 190.712 1998.53 192.703 Q1996.89 194.671 1993.86 195.434 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M2024.02 195.434 Q2027.37 196.152 2029.25 198.421 Q2031.15 200.689 2031.15 204.022 Q2031.15 209.138 2027.63 211.939 Q2024.11 214.74 2017.63 214.74 Q2015.45 214.74 2013.14 214.3 Q2010.85 213.883 2008.39 213.027 L2008.39 208.513 Q2010.34 209.647 2012.65 210.226 Q2014.97 210.805 2017.49 210.805 Q2021.89 210.805 2024.18 209.069 Q2026.49 207.333 2026.49 204.022 Q2026.49 200.967 2024.34 199.254 Q2022.21 197.518 2018.39 197.518 L2014.36 197.518 L2014.36 193.675 L2018.58 193.675 Q2022.03 193.675 2023.86 192.309 Q2025.68 190.921 2025.68 188.328 Q2025.68 185.666 2023.79 184.254 Q2021.91 182.819 2018.39 182.819 Q2016.47 182.819 2014.27 183.235 Q2012.07 183.652 2009.43 184.532 L2009.43 180.365 Q2012.1 179.624 2014.41 179.254 Q2016.75 178.884 2018.81 178.884 Q2024.13 178.884 2027.23 181.314 Q2030.34 183.722 2030.34 187.842 Q2030.34 190.712 2028.69 192.703 Q2027.05 194.671 2024.02 195.434 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M2040.06 179.509 L2058.42 179.509 L2058.42 183.444 L2044.34 183.444 L2044.34 191.916 Q2045.36 191.569 2046.38 191.407 Q2047.4 191.222 2048.42 191.222 Q2054.2 191.222 2057.58 194.393 Q2060.96 197.564 2060.96 202.981 Q2060.96 208.559 2057.49 211.661 Q2054.02 214.74 2047.7 214.74 Q2045.52 214.74 2043.25 214.37 Q2041.01 213.999 2038.6 213.258 L2038.6 208.559 Q2040.68 209.694 2042.91 210.249 Q2045.13 210.805 2047.61 210.805 Q2051.61 210.805 2053.95 208.698 Q2056.29 206.592 2056.29 202.981 Q2056.29 199.37 2053.95 197.263 Q2051.61 195.157 2047.61 195.157 Q2045.73 195.157 2043.86 195.573 Q2042 195.99 2040.06 196.87 L2040.06 179.509 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M2070.22 179.509 L2088.58 179.509 L2088.58 183.444 L2074.5 183.444 L2074.5 191.916 Q2075.52 191.569 2076.54 191.407 Q2077.56 191.222 2078.58 191.222 Q2084.36 191.222 2087.74 194.393 Q2091.12 197.564 2091.12 202.981 Q2091.12 208.559 2087.65 211.661 Q2084.18 214.74 2077.86 214.74 Q2075.68 214.74 2073.42 214.37 Q2071.17 213.999 2068.76 213.258 L2068.76 208.559 Q2070.85 209.694 2073.07 210.249 Q2075.29 210.805 2077.77 210.805 Q2081.77 210.805 2084.11 208.698 Q2086.45 206.592 2086.45 202.981 Q2086.45 199.37 2084.11 197.263 Q2081.77 195.157 2077.77 195.157 Q2075.89 195.157 2074.02 195.573 Q2072.17 195.99 2070.22 196.87 L2070.22 179.509 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M2113.18 183.583 L2101.38 202.032 L2113.18 202.032 L2113.18 183.583 M2111.96 179.509 L2117.84 179.509 L2117.84 202.032 L2122.77 202.032 L2122.77 205.921 L2117.84 205.921 L2117.84 214.069 L2113.18 214.069 L2113.18 205.921 L2097.58 205.921 L2097.58 201.407 L2111.96 179.509 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M2143.35 183.583 L2131.54 202.032 L2143.35 202.032 L2143.35 183.583 M2142.12 179.509 L2148 179.509 L2148 202.032 L2152.93 202.032 L2152.93 205.921 L2148 205.921 L2148 214.069 L2143.35 214.069 L2143.35 205.921 L2127.74 205.921 L2127.74 201.407 L2142.12 179.509 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M2174.83 195.434 Q2178.18 196.152 2180.06 198.421 Q2181.96 200.689 2181.96 204.022 Q2181.96 209.138 2178.44 211.939 Q2174.92 214.74 2168.44 214.74 Q2166.26 214.74 2163.95 214.3 Q2161.66 213.883 2159.2 213.027 L2159.2 208.513 Q2161.15 209.647 2163.46 210.226 Q2165.78 210.805 2168.3 210.805 Q2172.7 210.805 2174.99 209.069 Q2177.3 207.333 2177.3 204.022 Q2177.3 200.967 2175.15 199.254 Q2173.02 197.518 2169.2 197.518 L2165.17 197.518 L2165.17 193.675 L2169.39 193.675 Q2172.84 193.675 2174.66 192.309 Q2176.49 190.921 2176.49 188.328 Q2176.49 185.666 2174.6 184.254 Q2172.72 182.819 2169.2 182.819 Q2167.28 182.819 2165.08 183.235 Q2162.88 183.652 2160.24 184.532 L2160.24 180.365 Q2162.91 179.624 2165.22 179.254 Q2167.56 178.884 2169.62 178.884 Q2174.94 178.884 2178.04 181.314 Q2181.15 183.722 2181.15 187.842 Q2181.15 190.712 2179.5 192.703 Q2177.86 194.671 2174.83 195.434 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M2204.99 195.434 Q2208.35 196.152 2210.22 198.421 Q2212.12 200.689 2212.12 204.022 Q2212.12 209.138 2208.6 211.939 Q2205.08 214.74 2198.6 214.74 Q2196.42 214.74 2194.11 214.3 Q2191.82 213.883 2189.36 213.027 L2189.36 208.513 Q2191.31 209.647 2193.62 210.226 Q2195.94 210.805 2198.46 210.805 Q2202.86 210.805 2205.15 209.069 Q2207.47 207.333 2207.47 204.022 Q2207.47 200.967 2205.31 199.254 Q2203.18 197.518 2199.36 197.518 L2195.34 197.518 L2195.34 193.675 L2199.55 193.675 Q2203 193.675 2204.83 192.309 Q2206.66 190.921 2206.66 188.328 Q2206.66 185.666 2204.76 184.254 Q2202.88 182.819 2199.36 182.819 Q2197.44 182.819 2195.24 183.235 Q2193.04 183.652 2190.41 184.532 L2190.41 180.365 Q2193.07 179.624 2195.38 179.254 Q2197.72 178.884 2199.78 178.884 Q2205.1 178.884 2208.21 181.314 Q2211.31 183.722 2211.31 187.842 Q2211.31 190.712 2209.66 192.703 Q2208.02 194.671 2204.99 195.434 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M2219.71 178.097 L2223.41 178.097 Q2226.89 183.56 2228.6 188.791 Q2230.34 194.022 2230.34 199.184 Q2230.34 204.37 2228.6 209.624 Q2226.89 214.879 2223.41 220.319 L2219.71 220.319 Q2222.79 215.018 2224.29 209.786 Q2225.82 204.532 2225.82 199.184 Q2225.82 193.837 2224.29 188.629 Q2222.79 183.421 2219.71 178.097 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /></svg>



#### Training of the NeuralFMU

For the training of the NeuralFMU the parameters are extracted. All parameters of the first layer are set to the absolute value.


```julia
# train
paramsNet = Flux.params(neuralFMU)

for i in 1:length(paramsNet[1])
    if paramsNet[1][i] < 0.0 
        paramsNet[1][i] = -paramsNet[1][i]
    end
end
```

The well-known ADAM optimizer for minimizing the gradient descent is used as further passing parameters. Additionally, the previously defined loss and callback function as well as a one for the number of epochs are passed. Only one epoch is trained so that the NeuralFMU is precompiled.


```julia
optim = ADAM()
Flux.train!(lossSum, paramsNet, Iterators.repeated((), 1), optim; cb=callb) 
```

    ┌ Info:   Loss [1]: 0.37281
    │         Avg displacement in data: 0.61058
    │         Weight/Scale: 1.0009999999708   Bias/Offset: 0.0009999999670082423
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15


Some vectors for collecting data are initialized and the number of runs, epochs and iterations are set.


```julia
solutionAfter = []
solutionAfterMod = []
forces = []
displacements = []

numRuns = 2
numEpochs= 5
numIterations = 500;
```

#### Training loop

The code section shown here represents the training loop. The loop is structured so that it has `numRuns` runs, where each run has `numEpochs` epochs, and the training is performed at each epoch with `numIterations` iterations. In lines 9 and 10, the data for the *neuralFMU* for the default and modified initial states are appended to the corresponding vectors. The plots for the opposition of position and velocity is done in line 13 by calling the function `plot_all_results`. In the following lines the last layers are extracted from the *neuralFMU* and formed into an independent network `netBottom`. The parmaters for the `netBottom` network come from the original architecture and are shared. In line 20, the new network is used to represent the friction model in a graph. An analogous construction of the next part of the training loop, where here the first layer is taken from the *neuralFMU* and converted to its own network `netTop`. This network is used to record the displacement model. The different graphs are generated for each run and can thus be compared. 


```julia
for run in 1:numRuns    
    @time for epoch in 1:numEpochs
        @info "Run: $(run)/$(numRuns)  Epoch: $(epoch)/$(numEpochs)"
        Flux.train!(lossSum, paramsNet, Iterators.repeated((), numIterations), optim; cb=callb)
    end
    flush(stderr)
    flush(stdout)
    
    push!(solutionAfter, neuralFMU(x₀))
    push!(solutionAfterMod, neuralFMU(xMod₀))

    # generate all plots for the position and velocity
    plot_all_results(realSimData, realSimDataMod, simpleSimData, simpleSimDataMod, solutionAfter, solutionAfterMod)
    
    # friction model extraction
    layersBottom = neuralFMU.neuralODE.model.layers[3:5]
    netBottom = Chain(layersBottom...)
    transferParams!(netBottom, paramsNet, 7)
    
    forces = plot_friction_model(realSimData, netBottom, forces) 
    
    # displacement model extraction
    layersTop = neuralFMU.neuralODE.model.layers[1:1]
    netTop = Chain(layersTop...)
    transferParams!(netTop, paramsNet, 1)

    displacements = plot_displacement_model(realSimData, netTop, displacements, tSave, displacement)
end
```

    ┌ Info: Run: 1/2  Epoch: 1/5
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:3
    ┌ Info:   Loss [51]: 0.29271
    │         Avg displacement in data: 0.54103
    │         Weight/Scale: 1.0261700470827426   Bias/Offset: 0.026781892116802514
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [101]: 0.27598
    │         Avg displacement in data: 0.52534
    │         Weight/Scale: 1.0166941766096746   Bias/Offset: 0.018766602263892727
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [151]: 0.14655
    │         Avg displacement in data: 0.38282
    │         Weight/Scale: 0.9881678087519986   Bias/Offset: 0.009921884875380626
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [201]: 0.03493
    │         Avg displacement in data: 0.18689
    │         Weight/Scale: 1.0409497258014302   Bias/Offset: 0.07412065482413917
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [251]: 0.02909
    │         Avg displacement in data: 0.17055
    │         Weight/Scale: 1.0445126475609419   Bias/Offset: 0.07482981104487602
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [301]: 0.025
    │         Avg displacement in data: 0.15812
    │         Weight/Scale: 1.0450995524918851   Bias/Offset: 0.07274866450816542
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [351]: 0.02164
    │         Avg displacement in data: 0.1471
    │         Weight/Scale: 1.0449575243976053   Bias/Offset: 0.07100912607928997
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [401]: 0.01917
    │         Avg displacement in data: 0.13847
    │         Weight/Scale: 1.0438239812810226   Bias/Offset: 0.06913720933273998
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [451]: 0.01715
    │         Avg displacement in data: 0.13097
    │         Weight/Scale: 1.0419133983493345   Bias/Offset: 0.06715393880817992
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [501]: 0.01545
    │         Avg displacement in data: 0.12429
    │         Weight/Scale: 1.0394782537736347   Bias/Offset: 0.06503844578767762
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info: Run: 1/2  Epoch: 2/5
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:3
    ┌ Info:   Loss [551]: 0.01401
    │         Avg displacement in data: 0.11837
    │         Weight/Scale: 1.0368577211126218   Bias/Offset: 0.06293846743637753
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [601]: 0.0128
    │         Avg displacement in data: 0.11315
    │         Weight/Scale: 1.0343833826971995   Bias/Offset: 0.06109646559257188
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [651]: 0.01178
    │         Avg displacement in data: 0.10852
    │         Weight/Scale: 1.0322364161236024   Bias/Offset: 0.05961904466855851
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [701]: 0.01089
    │         Avg displacement in data: 0.10434
    │         Weight/Scale: 1.0304486687608865   Bias/Offset: 0.05845025707582767
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [751]: 0.01012
    │         Avg displacement in data: 0.10061
    │         Weight/Scale: 1.029023239504709   Bias/Offset: 0.05757655372371887
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [801]: 0.00948
    │         Avg displacement in data: 0.09734
    │         Weight/Scale: 1.0279051016975647   Bias/Offset: 0.056966787973587195
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [851]: 0.00892
    │         Avg displacement in data: 0.09443
    │         Weight/Scale: 1.0270105715576456   Bias/Offset: 0.056576648049736125
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [901]: 0.00842
    │         Avg displacement in data: 0.09173
    │         Weight/Scale: 1.026261773580901   Bias/Offset: 0.05635249617185529
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [951]: 0.00795
    │         Avg displacement in data: 0.08918
    │         Weight/Scale: 1.025605096167402   Bias/Offset: 0.056249304827566717
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [1001]: 0.00752
    │         Avg displacement in data: 0.08669
    │         Weight/Scale: 1.025008248961602   Bias/Offset: 0.056239420583352906
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info: Run: 1/2  Epoch: 3/5
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:3
    ┌ Info:   Loss [1051]: 0.0071
    │         Avg displacement in data: 0.08428
    │         Weight/Scale: 1.0244534147669253   Bias/Offset: 0.05631034261565172
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [1101]: 0.0067
    │         Avg displacement in data: 0.08186
    │         Weight/Scale: 1.023916884544763   Bias/Offset: 0.05644765533830105
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [1151]: 0.00632
    │         Avg displacement in data: 0.07949
    │         Weight/Scale: 1.0233950200628592   Bias/Offset: 0.056653478181622564
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [1201]: 0.00593
    │         Avg displacement in data: 0.07699
    │         Weight/Scale: 1.022856901275902   Bias/Offset: 0.05689341577806724
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [1251]: 0.00556
    │         Avg displacement in data: 0.07458
    │         Weight/Scale: 1.0223196582908227   Bias/Offset: 0.05720147506252516
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [1301]: 0.00521
    │         Avg displacement in data: 0.07217
    │         Weight/Scale: 1.0217296189187148   Bias/Offset: 0.057504150886319806
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [1351]: 0.00484
    │         Avg displacement in data: 0.06959
    │         Weight/Scale: 1.0210934548931594   Bias/Offset: 0.05777431429993874
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [1401]: 0.00447
    │         Avg displacement in data: 0.06687
    │         Weight/Scale: 1.0204569192191018   Bias/Offset: 0.05802013715785489
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [1451]: 0.00413
    │         Avg displacement in data: 0.06425
    │         Weight/Scale: 1.0198279264957875   Bias/Offset: 0.058185806588078315
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [1501]: 0.00383
    │         Avg displacement in data: 0.06186
    │         Weight/Scale: 1.0192215104569753   Bias/Offset: 0.05821758923925437
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info: Run: 1/2  Epoch: 4/5
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:3
    ┌ Info:   Loss [1551]: 0.00357
    │         Avg displacement in data: 0.05976
    │         Weight/Scale: 1.0186619096260496   Bias/Offset: 0.058117573979332815
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [1601]: 0.00336
    │         Avg displacement in data: 0.05793
    │         Weight/Scale: 1.0181667928533207   Bias/Offset: 0.05793415005320181
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [1651]: 0.00317
    │         Avg displacement in data: 0.05627
    │         Weight/Scale: 1.0177384616962248   Bias/Offset: 0.05772042637626792
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [1701]: 0.00301
    │         Avg displacement in data: 0.05483
    │         Weight/Scale: 1.0173319838860602   Bias/Offset: 0.05752543564342985
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [1751]: 0.00285
    │         Avg displacement in data: 0.05343
    │         Weight/Scale: 1.0169584704776444   Bias/Offset: 0.05731313291343434
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [1801]: 0.00272
    │         Avg displacement in data: 0.05211
    │         Weight/Scale: 1.01662361152564   Bias/Offset: 0.05711736572660747
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [1851]: 0.00259
    │         Avg displacement in data: 0.05088
    │         Weight/Scale: 1.016316559157972   Bias/Offset: 0.056933744164988385
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [1901]: 0.00247
    │         Avg displacement in data: 0.04973
    │         Weight/Scale: 1.0160352198911133   Bias/Offset: 0.05676371965093806
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [1951]: 0.00237
    │         Avg displacement in data: 0.04868
    │         Weight/Scale: 1.015763941178961   Bias/Offset: 0.05660311562772707
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [2001]: 0.00228
    │         Avg displacement in data: 0.04772
    │         Weight/Scale: 1.015522447414816   Bias/Offset: 0.05645826582215821
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info: Run: 1/2  Epoch: 5/5
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:3
    ┌ Info:   Loss [2051]: 0.00218
    │         Avg displacement in data: 0.04673
    │         Weight/Scale: 1.015287439796913   Bias/Offset: 0.05632043900351172
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [2101]: 0.00211
    │         Avg displacement in data: 0.04594
    │         Weight/Scale: 1.01506319413731   Bias/Offset: 0.056194022241978725
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [2151]: 0.00204
    │         Avg displacement in data: 0.04516
    │         Weight/Scale: 1.0148787735913567   Bias/Offset: 0.056099037436618725
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [2201]: 0.00196
    │         Avg displacement in data: 0.0443
    │         Weight/Scale: 1.0146776915402087   Bias/Offset: 0.05598911133008781
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [2251]: 0.0019
    │         Avg displacement in data: 0.04356
    │         Weight/Scale: 1.0145066089430042   Bias/Offset: 0.05590976850023108
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [2301]: 0.00184
    │         Avg displacement in data: 0.04289
    │         Weight/Scale: 1.014350929576157   Bias/Offset: 0.055840218056319765
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [2351]: 0.00178
    │         Avg displacement in data: 0.04222
    │         Weight/Scale: 1.0142014322657635   Bias/Offset: 0.05578181835941591
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [2401]: 0.00174
    │         Avg displacement in data: 0.04176
    │         Weight/Scale: 1.0140774594769901   Bias/Offset: 0.05573974700257181
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [2451]: 0.00168
    │         Avg displacement in data: 0.04105
    │         Weight/Scale: 1.0139400024096277   Bias/Offset: 0.05569726115418364
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [2501]: 0.00164
    │         Avg displacement in data: 0.04052
    │         Weight/Scale: 1.013838090789545   Bias/Offset: 0.055673695406279776
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15


    2076.542533 seconds (8.35 G allocations: 512.220 GiB, 3.94% gc time)



<?xml version="1.0" encoding="utf-8"?>
<svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" width="600" height="400" viewBox="0 0 2400 1600">
<defs>
  <clipPath id="clip390">
    <rect x="0" y="0" width="2400" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip390)" d="
M0 1600 L2400 1600 L2400 0 L0 0  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip391">
    <rect x="480" y="0" width="1681" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip390)" d="
M303.503 1401.58 L2352.76 1401.58 L2352.76 123.472 L303.503 123.472  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip392">
    <rect x="303" y="123" width="2050" height="1279"/>
  </clipPath>
</defs>
<polyline clip-path="url(#clip392)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  361.501,1401.58 361.501,123.472 
  "/>
<polyline clip-path="url(#clip392)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  844.815,1401.58 844.815,123.472 
  "/>
<polyline clip-path="url(#clip392)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1328.13,1401.58 1328.13,123.472 
  "/>
<polyline clip-path="url(#clip392)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1811.44,1401.58 1811.44,123.472 
  "/>
<polyline clip-path="url(#clip392)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  2294.76,1401.58 2294.76,123.472 
  "/>
<polyline clip-path="url(#clip390)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  303.503,1401.58 2352.76,1401.58 
  "/>
<polyline clip-path="url(#clip390)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  361.501,1401.58 361.501,1382.68 
  "/>
<polyline clip-path="url(#clip390)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  844.815,1401.58 844.815,1382.68 
  "/>
<polyline clip-path="url(#clip390)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1328.13,1401.58 1328.13,1382.68 
  "/>
<polyline clip-path="url(#clip390)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1811.44,1401.58 1811.44,1382.68 
  "/>
<polyline clip-path="url(#clip390)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  2294.76,1401.58 2294.76,1382.68 
  "/>
<path clip-path="url(#clip390)" d="M361.501 1438.36 Q356.084 1438.36 353.341 1443.71 Q350.633 1449.02 350.633 1459.71 Q350.633 1470.37 353.341 1475.72 Q356.084 1481.03 361.501 1481.03 Q366.952 1481.03 369.66 1475.72 Q372.403 1470.37 372.403 1459.71 Q372.403 1449.02 369.66 1443.71 Q366.952 1438.36 361.501 1438.36 M361.501 1432.8 Q370.216 1432.8 374.799 1439.71 Q379.417 1446.59 379.417 1459.71 Q379.417 1472.8 374.799 1479.71 Q370.216 1486.59 361.501 1486.59 Q352.785 1486.59 348.167 1479.71 Q343.584 1472.8 343.584 1459.71 Q343.584 1446.59 348.167 1439.71 Q352.785 1432.8 361.501 1432.8 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M830.388 1479.68 L841.846 1479.68 L841.846 1440.13 L829.381 1442.63 L829.381 1436.24 L841.777 1433.74 L848.791 1433.74 L848.791 1479.68 L860.249 1479.68 L860.249 1485.58 L830.388 1485.58 L830.388 1479.68 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M1320.11 1479.68 L1344.59 1479.68 L1344.59 1485.58 L1311.67 1485.58 L1311.67 1479.68 Q1315.66 1475.55 1322.54 1468.6 Q1329.45 1461.62 1331.22 1459.61 Q1334.59 1455.82 1335.91 1453.22 Q1337.26 1450.58 1337.26 1448.05 Q1337.26 1443.91 1334.34 1441.31 Q1331.46 1438.71 1326.81 1438.71 Q1323.51 1438.71 1319.83 1439.85 Q1316.19 1441 1312.02 1443.32 L1312.02 1436.24 Q1316.25 1434.54 1319.94 1433.67 Q1323.62 1432.8 1326.67 1432.8 Q1334.73 1432.8 1339.52 1436.83 Q1344.31 1440.86 1344.31 1447.59 Q1344.31 1450.79 1343.09 1453.67 Q1341.91 1456.52 1338.75 1460.41 Q1337.89 1461.41 1333.23 1466.24 Q1328.58 1471.03 1320.11 1479.68 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M1817.82 1457.63 Q1822.85 1458.7 1825.66 1462.11 Q1828.51 1465.51 1828.51 1470.51 Q1828.51 1478.18 1823.23 1482.39 Q1817.95 1486.59 1808.23 1486.59 Q1804.97 1486.59 1801.5 1485.93 Q1798.06 1485.3 1794.38 1484.02 L1794.38 1477.25 Q1797.29 1478.95 1800.77 1479.82 Q1804.24 1480.68 1808.02 1480.68 Q1814.62 1480.68 1818.06 1478.08 Q1821.53 1475.48 1821.53 1470.51 Q1821.53 1465.93 1818.3 1463.36 Q1815.11 1460.75 1809.38 1460.75 L1803.34 1460.75 L1803.34 1454.99 L1809.66 1454.99 Q1814.83 1454.99 1817.57 1452.94 Q1820.32 1450.86 1820.32 1446.97 Q1820.32 1442.98 1817.47 1440.86 Q1814.66 1438.71 1809.38 1438.71 Q1806.5 1438.71 1803.2 1439.33 Q1799.9 1439.96 1795.94 1441.27 L1795.94 1435.02 Q1799.93 1433.91 1803.41 1433.36 Q1806.91 1432.8 1810 1432.8 Q1817.99 1432.8 1822.64 1436.45 Q1827.29 1440.06 1827.29 1446.24 Q1827.29 1450.55 1824.83 1453.53 Q1822.36 1456.48 1817.82 1457.63 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M2299.27 1439.85 L2281.56 1467.52 L2299.27 1467.52 L2299.27 1439.85 M2297.43 1433.74 L2306.25 1433.74 L2306.25 1467.52 L2313.65 1467.52 L2313.65 1473.36 L2306.25 1473.36 L2306.25 1485.58 L2299.27 1485.58 L2299.27 1473.36 L2275.87 1473.36 L2275.87 1466.59 L2297.43 1433.74 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M1271.74 1527.62 L1271.74 1538.66 L1284.9 1538.66 L1284.9 1543.62 L1271.74 1543.62 L1271.74 1564.74 Q1271.74 1569.49 1273.03 1570.85 Q1274.35 1572.2 1278.34 1572.2 L1284.9 1572.2 L1284.9 1577.55 L1278.34 1577.55 Q1270.94 1577.55 1268.13 1574.8 Q1265.32 1572.03 1265.32 1564.74 L1265.32 1543.62 L1260.63 1543.62 L1260.63 1538.66 L1265.32 1538.66 L1265.32 1527.62 L1271.74 1527.62 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M1315.32 1523.52 L1330.04 1523.52 L1330.04 1528.49 L1321.71 1528.49 L1321.71 1581.96 L1330.04 1581.96 L1330.04 1586.92 L1315.32 1586.92 L1315.32 1523.52 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M1368.44 1539.8 L1368.44 1545.85 Q1365.73 1544.46 1362.82 1543.76 Q1359.9 1543.07 1356.78 1543.07 Q1352.02 1543.07 1349.62 1544.53 Q1347.26 1545.99 1347.26 1548.9 Q1347.26 1551.12 1348.96 1552.41 Q1350.66 1553.66 1355.8 1554.8 L1357.99 1555.29 Q1364.8 1556.75 1367.64 1559.42 Q1370.53 1562.06 1370.53 1566.82 Q1370.53 1572.24 1366.22 1575.4 Q1361.95 1578.55 1354.45 1578.55 Q1351.32 1578.55 1347.92 1577.93 Q1344.55 1577.34 1340.8 1576.12 L1340.8 1569.53 Q1344.34 1571.37 1347.78 1572.3 Q1351.22 1573.21 1354.59 1573.21 Q1359.1 1573.21 1361.53 1571.68 Q1363.96 1570.12 1363.96 1567.3 Q1363.96 1564.7 1362.19 1563.31 Q1360.46 1561.92 1354.52 1560.64 L1352.3 1560.12 Q1346.36 1558.87 1343.72 1556.3 Q1341.08 1553.69 1341.08 1549.18 Q1341.08 1543.69 1344.97 1540.71 Q1348.86 1537.72 1356.01 1537.72 Q1359.55 1537.72 1362.68 1538.24 Q1365.8 1538.76 1368.44 1539.8 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M1395.63 1523.52 L1395.63 1586.92 L1380.91 1586.92 L1380.91 1581.96 L1389.21 1581.96 L1389.21 1528.49 L1380.91 1528.49 L1380.91 1523.52 L1395.63 1523.52 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip392)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  303.503,1365.41 2352.76,1365.41 
  "/>
<polyline clip-path="url(#clip392)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  303.503,1114.21 2352.76,1114.21 
  "/>
<polyline clip-path="url(#clip392)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  303.503,863.02 2352.76,863.02 
  "/>
<polyline clip-path="url(#clip392)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  303.503,611.826 2352.76,611.826 
  "/>
<polyline clip-path="url(#clip392)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  303.503,360.632 2352.76,360.632 
  "/>
<polyline clip-path="url(#clip390)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  303.503,1401.58 303.503,123.472 
  "/>
<polyline clip-path="url(#clip390)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  303.503,1365.41 322.401,1365.41 
  "/>
<polyline clip-path="url(#clip390)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  303.503,1114.21 322.401,1114.21 
  "/>
<polyline clip-path="url(#clip390)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  303.503,863.02 322.401,863.02 
  "/>
<polyline clip-path="url(#clip390)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  303.503,611.826 322.401,611.826 
  "/>
<polyline clip-path="url(#clip390)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  303.503,360.632 322.401,360.632 
  "/>
<path clip-path="url(#clip390)" d="M136.497 1344.11 Q131.08 1344.11 128.337 1349.45 Q125.629 1354.76 125.629 1365.46 Q125.629 1376.12 128.337 1381.47 Q131.08 1386.78 136.497 1386.78 Q141.948 1386.78 144.656 1381.47 Q147.399 1376.12 147.399 1365.46 Q147.399 1354.76 144.656 1349.45 Q141.948 1344.11 136.497 1344.11 M136.497 1338.55 Q145.212 1338.55 149.795 1345.46 Q154.413 1352.33 154.413 1365.46 Q154.413 1378.55 149.795 1385.46 Q145.212 1392.33 136.497 1392.33 Q127.781 1392.33 123.163 1385.46 Q118.58 1378.55 118.58 1365.46 Q118.58 1352.33 123.163 1345.46 Q127.781 1338.55 136.497 1338.55 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M166.74 1382.51 L174.066 1382.51 L174.066 1391.33 L166.74 1391.33 L166.74 1382.51 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M189.413 1339.49 L216.948 1339.49 L216.948 1345.39 L195.837 1345.39 L195.837 1358.1 Q197.364 1357.58 198.892 1357.33 Q200.42 1357.06 201.948 1357.06 Q210.628 1357.06 215.698 1361.81 Q220.767 1366.57 220.767 1374.7 Q220.767 1383.06 215.559 1387.72 Q210.35 1392.33 200.871 1392.33 Q197.608 1392.33 194.205 1391.78 Q190.837 1391.22 187.226 1390.11 L187.226 1383.06 Q190.351 1384.76 193.684 1385.6 Q197.017 1386.43 200.732 1386.43 Q206.739 1386.43 210.246 1383.27 Q213.753 1380.11 213.753 1374.7 Q213.753 1369.28 210.246 1366.12 Q206.739 1362.96 200.732 1362.96 Q197.92 1362.96 195.108 1363.58 Q192.33 1364.21 189.413 1365.53 L189.413 1339.49 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M249.586 1344.11 Q244.17 1344.11 241.427 1349.45 Q238.718 1354.76 238.718 1365.46 Q238.718 1376.12 241.427 1381.47 Q244.17 1386.78 249.586 1386.78 Q255.038 1386.78 257.746 1381.47 Q260.489 1376.12 260.489 1365.46 Q260.489 1354.76 257.746 1349.45 Q255.038 1344.11 249.586 1344.11 M249.586 1338.55 Q258.302 1338.55 262.885 1345.46 Q267.503 1352.33 267.503 1365.46 Q267.503 1378.55 262.885 1385.46 Q258.302 1392.33 249.586 1392.33 Q240.871 1392.33 236.253 1385.46 Q231.67 1378.55 231.67 1365.46 Q231.67 1352.33 236.253 1345.46 Q240.871 1338.55 249.586 1338.55 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M137.99 1092.91 Q132.573 1092.91 129.83 1098.26 Q127.122 1103.57 127.122 1114.27 Q127.122 1124.93 129.83 1130.27 Q132.573 1135.58 137.99 1135.58 Q143.441 1135.58 146.149 1130.27 Q148.892 1124.93 148.892 1114.27 Q148.892 1103.57 146.149 1098.26 Q143.441 1092.91 137.99 1092.91 M137.99 1087.36 Q146.705 1087.36 151.288 1094.27 Q155.906 1101.14 155.906 1114.27 Q155.906 1127.36 151.288 1134.27 Q146.705 1141.14 137.99 1141.14 Q129.275 1141.14 124.657 1134.27 Q120.073 1127.36 120.073 1114.27 Q120.073 1101.14 124.657 1094.27 Q129.275 1087.36 137.99 1087.36 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M168.233 1131.31 L175.559 1131.31 L175.559 1140.13 L168.233 1140.13 L168.233 1131.31 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M189.066 1088.29 L222.399 1088.29 L222.399 1091.28 L203.58 1140.13 L196.253 1140.13 L213.962 1094.2 L189.066 1094.2 L189.066 1088.29 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M236.149 1088.29 L263.684 1088.29 L263.684 1094.2 L242.573 1094.2 L242.573 1106.9 Q244.1 1106.38 245.628 1106.14 Q247.156 1105.86 248.684 1105.86 Q257.364 1105.86 262.434 1110.62 Q267.503 1115.38 267.503 1123.5 Q267.503 1131.87 262.295 1136.52 Q257.086 1141.14 247.607 1141.14 Q244.343 1141.14 240.941 1140.58 Q237.573 1140.03 233.961 1138.92 L233.961 1131.87 Q237.086 1133.57 240.42 1134.4 Q243.753 1135.24 247.468 1135.24 Q253.475 1135.24 256.982 1132.08 Q260.489 1128.92 260.489 1123.5 Q260.489 1118.08 256.982 1114.93 Q253.475 1111.77 247.468 1111.77 Q244.656 1111.77 241.843 1112.39 Q239.066 1113.02 236.149 1114.33 L236.149 1088.29 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M122.712 883.037 L134.17 883.037 L134.17 843.488 L121.705 845.988 L121.705 839.6 L134.101 837.1 L141.115 837.1 L141.115 883.037 L152.573 883.037 L152.573 888.94 L122.712 888.94 L122.712 883.037 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M166.74 880.12 L174.066 880.12 L174.066 888.94 L166.74 888.94 L166.74 880.12 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M204.344 841.718 Q198.927 841.718 196.184 847.065 Q193.476 852.377 193.476 863.072 Q193.476 873.731 196.184 879.079 Q198.927 884.391 204.344 884.391 Q209.795 884.391 212.503 879.079 Q215.246 873.731 215.246 863.072 Q215.246 852.377 212.503 847.065 Q209.795 841.718 204.344 841.718 M204.344 836.162 Q213.059 836.162 217.642 843.072 Q222.26 849.947 222.26 863.072 Q222.26 876.162 217.642 883.072 Q213.059 889.947 204.344 889.947 Q195.628 889.947 191.01 883.072 Q186.427 876.162 186.427 863.072 Q186.427 849.947 191.01 843.072 Q195.628 836.162 204.344 836.162 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M249.586 841.718 Q244.17 841.718 241.427 847.065 Q238.718 852.377 238.718 863.072 Q238.718 873.731 241.427 879.079 Q244.17 884.391 249.586 884.391 Q255.038 884.391 257.746 879.079 Q260.489 873.731 260.489 863.072 Q260.489 852.377 257.746 847.065 Q255.038 841.718 249.586 841.718 M249.586 836.162 Q258.302 836.162 262.885 843.072 Q267.503 849.947 267.503 863.072 Q267.503 876.162 262.885 883.072 Q258.302 889.947 249.586 889.947 Q240.871 889.947 236.253 883.072 Q231.67 876.162 231.67 863.072 Q231.67 849.947 236.253 843.072 Q240.871 836.162 249.586 836.162 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M124.205 631.843 L135.663 631.843 L135.663 592.295 L123.198 594.795 L123.198 588.406 L135.594 585.906 L142.608 585.906 L142.608 631.843 L154.066 631.843 L154.066 637.746 L124.205 637.746 L124.205 631.843 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M168.233 628.927 L175.559 628.927 L175.559 637.746 L168.233 637.746 L168.233 628.927 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M196.878 631.843 L221.357 631.843 L221.357 637.746 L188.441 637.746 L188.441 631.843 Q192.434 627.711 199.309 620.767 Q206.219 613.788 207.989 611.774 Q211.357 607.989 212.677 605.385 Q214.031 602.746 214.031 600.211 Q214.031 596.08 211.114 593.475 Q208.232 590.871 203.58 590.871 Q200.281 590.871 196.601 592.017 Q192.955 593.163 188.788 595.489 L188.788 588.406 Q193.024 586.705 196.705 585.837 Q200.385 584.968 203.441 584.968 Q211.496 584.968 216.288 588.996 Q221.08 593.024 221.08 599.76 Q221.08 602.954 219.864 605.836 Q218.684 608.684 215.524 612.572 Q214.656 613.579 210.003 618.406 Q205.351 623.197 196.878 631.843 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M236.149 585.906 L263.684 585.906 L263.684 591.809 L242.573 591.809 L242.573 604.517 Q244.1 603.996 245.628 603.753 Q247.156 603.475 248.684 603.475 Q257.364 603.475 262.434 608.232 Q267.503 612.989 267.503 621.114 Q267.503 629.482 262.295 634.135 Q257.086 638.753 247.607 638.753 Q244.343 638.753 240.941 638.197 Q237.573 637.642 233.961 636.531 L233.961 629.482 Q237.086 631.184 240.42 632.017 Q243.753 632.85 247.468 632.85 Q253.475 632.85 256.982 629.69 Q260.489 626.531 260.489 621.114 Q260.489 615.697 256.982 612.538 Q253.475 609.378 247.468 609.378 Q244.656 609.378 241.843 610.003 Q239.066 610.628 236.149 611.948 L236.149 585.906 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M122.712 380.65 L134.17 380.65 L134.17 341.101 L121.705 343.601 L121.705 337.212 L134.101 334.712 L141.115 334.712 L141.115 380.65 L152.573 380.65 L152.573 386.552 L122.712 386.552 L122.712 380.65 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M166.74 377.733 L174.066 377.733 L174.066 386.552 L166.74 386.552 L166.74 377.733 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M189.413 334.712 L216.948 334.712 L216.948 340.615 L195.837 340.615 L195.837 353.323 Q197.364 352.802 198.892 352.559 Q200.42 352.282 201.948 352.282 Q210.628 352.282 215.698 357.039 Q220.767 361.795 220.767 369.92 Q220.767 378.288 215.559 382.941 Q210.35 387.559 200.871 387.559 Q197.608 387.559 194.205 387.004 Q190.837 386.448 187.226 385.337 L187.226 378.288 Q190.351 379.99 193.684 380.823 Q197.017 381.656 200.732 381.656 Q206.739 381.656 210.246 378.497 Q213.753 375.337 213.753 369.92 Q213.753 364.504 210.246 361.344 Q206.739 358.184 200.732 358.184 Q197.92 358.184 195.108 358.809 Q192.33 359.434 189.413 360.754 L189.413 334.712 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M249.586 339.33 Q244.17 339.33 241.427 344.678 Q238.718 349.99 238.718 360.684 Q238.718 371.344 241.427 376.691 Q244.17 382.004 249.586 382.004 Q255.038 382.004 257.746 376.691 Q260.489 371.344 260.489 360.684 Q260.489 349.99 257.746 344.678 Q255.038 339.33 249.586 339.33 M249.586 333.775 Q258.302 333.775 262.885 340.684 Q267.503 347.559 267.503 360.684 Q267.503 373.775 262.885 380.684 Q258.302 387.559 249.586 387.559 Q240.871 387.559 236.253 380.684 Q231.67 373.775 231.67 360.684 Q231.67 347.559 236.253 340.684 Q240.871 333.775 249.586 333.775 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M35.6047 1045.01 Q31.2992 1042.61 29.2506 1039.28 Q27.202 1035.94 27.202 1031.43 Q27.202 1025.35 31.4728 1022.06 Q35.7089 1018.76 43.5561 1018.76 L67.0282 1018.76 L67.0282 1025.18 L43.7644 1025.18 Q38.1741 1025.18 35.4658 1027.16 Q32.7575 1029.14 32.7575 1033.2 Q32.7575 1038.17 36.0561 1041.05 Q39.3547 1043.93 45.0491 1043.93 L67.0282 1043.93 L67.0282 1050.35 L43.7644 1050.35 Q38.1394 1050.35 35.4658 1052.33 Q32.7575 1054.31 32.7575 1058.44 Q32.7575 1063.34 36.0908 1066.22 Q39.3894 1069.1 45.0491 1069.1 L67.0282 1069.1 L67.0282 1075.53 L28.1395 1075.53 L28.1395 1069.1 L34.1811 1069.1 Q30.6047 1066.92 28.9034 1063.86 Q27.202 1060.81 27.202 1056.6 Q27.202 1052.37 29.3547 1049.42 Q31.5075 1046.43 35.6047 1045.01 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M47.4797 988.341 Q47.4797 996.084 49.2505 999.07 Q51.0213 1002.06 55.2921 1002.06 Q58.6949 1002.06 60.7087 999.834 Q62.6879 997.577 62.6879 993.723 Q62.6879 988.41 58.9379 985.216 Q55.1532 981.987 48.9033 981.987 L47.4797 981.987 L47.4797 988.341 M44.8408 975.598 L67.0282 975.598 L67.0282 981.987 L61.1254 981.987 Q64.6671 984.174 66.3684 987.438 Q68.0351 990.702 68.0351 995.424 Q68.0351 1001.4 64.7018 1004.94 Q61.3337 1008.44 55.7088 1008.44 Q49.1463 1008.44 45.813 1004.07 Q42.4797 999.66 42.4797 990.945 L42.4797 981.987 L41.8547 981.987 Q37.445 981.987 35.0492 984.903 Q32.6186 987.785 32.6186 993.028 Q32.6186 996.362 33.4172 999.521 Q34.2158 1002.68 35.813 1005.6 L29.9103 1005.6 Q28.5561 1002.09 27.8964 998.792 Q27.202 995.493 27.202 992.369 Q27.202 983.931 31.577 979.764 Q35.9519 975.598 44.8408 975.598 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M29.2853 937.647 L35.3269 937.647 Q33.9381 940.355 33.2436 943.272 Q32.5492 946.188 32.5492 949.313 Q32.5492 954.07 34.0075 956.466 Q35.4658 958.827 38.3825 958.827 Q40.6047 958.827 41.8894 957.126 Q43.1394 955.424 44.2852 950.285 L44.7713 948.098 Q46.2297 941.292 48.9033 938.445 Q51.5421 935.563 56.299 935.563 Q61.7157 935.563 64.8754 939.869 Q68.0351 944.14 68.0351 951.64 Q68.0351 954.765 67.4101 958.167 Q66.8198 961.535 65.6046 965.285 L59.0074 965.285 Q60.8476 961.744 61.7851 958.306 Q62.6879 954.869 62.6879 951.501 Q62.6879 946.987 61.1601 944.556 Q59.5976 942.126 56.7852 942.126 Q54.181 942.126 52.7921 943.897 Q51.4032 945.633 50.1185 951.57 L49.5977 953.792 Q48.3477 959.73 45.7783 962.369 Q43.1741 965.008 38.6603 965.008 Q33.1742 965.008 30.1881 961.119 Q27.202 957.23 27.202 950.077 Q27.202 946.535 27.7228 943.41 Q28.2436 940.285 29.2853 937.647 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M29.2853 900.598 L35.3269 900.598 Q33.9381 903.306 33.2436 906.223 Q32.5492 909.14 32.5492 912.265 Q32.5492 917.022 34.0075 919.418 Q35.4658 921.779 38.3825 921.779 Q40.6047 921.779 41.8894 920.077 Q43.1394 918.376 44.2852 913.237 L44.7713 911.049 Q46.2297 904.244 48.9033 901.397 Q51.5421 898.515 56.299 898.515 Q61.7157 898.515 64.8754 902.82 Q68.0351 907.091 68.0351 914.591 Q68.0351 917.716 67.4101 921.119 Q66.8198 924.487 65.6046 928.237 L59.0074 928.237 Q60.8476 924.695 61.7851 921.258 Q62.6879 917.82 62.6879 914.452 Q62.6879 909.938 61.1601 907.508 Q59.5976 905.077 56.7852 905.077 Q54.181 905.077 52.7921 906.848 Q51.4032 908.584 50.1185 914.522 L49.5977 916.744 Q48.3477 922.681 45.7783 925.32 Q43.1741 927.959 38.6603 927.959 Q33.1742 927.959 30.1881 924.07 Q27.202 920.181 27.202 913.029 Q27.202 909.487 27.7228 906.362 Q28.2436 903.237 29.2853 900.598 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M61.1949 859.557 L81.8197 859.557 L81.8197 865.98 L28.1395 865.98 L28.1395 859.557 L34.0422 859.557 Q30.57 857.543 28.9034 854.487 Q27.202 851.397 27.202 847.126 Q27.202 840.043 32.827 835.633 Q38.4519 831.189 47.6185 831.189 Q56.7852 831.189 62.4101 835.633 Q68.0351 840.043 68.0351 847.126 Q68.0351 851.397 66.3684 854.487 Q64.6671 857.543 61.1949 859.557 M47.6185 837.821 Q40.57 837.821 36.5769 840.737 Q32.5492 843.619 32.5492 848.689 Q32.5492 853.758 36.5769 856.675 Q40.57 859.557 47.6185 859.557 Q54.6671 859.557 58.6949 856.675 Q62.6879 853.758 62.6879 848.689 Q62.6879 843.619 58.6949 840.737 Q54.6671 837.821 47.6185 837.821 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M32.6186 805.529 Q32.6186 810.668 36.6464 813.654 Q40.6394 816.64 47.6185 816.64 Q54.5977 816.64 58.6254 813.689 Q62.6185 810.703 62.6185 805.529 Q62.6185 800.425 58.5907 797.439 Q54.5629 794.453 47.6185 794.453 Q40.7089 794.453 36.6811 797.439 Q32.6186 800.425 32.6186 805.529 M27.202 805.529 Q27.202 797.196 32.6186 792.439 Q38.0353 787.682 47.6185 787.682 Q57.1671 787.682 62.6185 792.439 Q68.0351 797.196 68.0351 805.529 Q68.0351 813.897 62.6185 818.654 Q57.1671 823.376 47.6185 823.376 Q38.0353 823.376 32.6186 818.654 Q27.202 813.897 27.202 805.529 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M29.2853 752.3 L35.3269 752.3 Q33.9381 755.009 33.2436 757.925 Q32.5492 760.842 32.5492 763.967 Q32.5492 768.724 34.0075 771.12 Q35.4658 773.481 38.3825 773.481 Q40.6047 773.481 41.8894 771.779 Q43.1394 770.078 44.2852 764.939 L44.7713 762.752 Q46.2297 755.946 48.9033 753.099 Q51.5421 750.217 56.299 750.217 Q61.7157 750.217 64.8754 754.523 Q68.0351 758.793 68.0351 766.293 Q68.0351 769.418 67.4101 772.821 Q66.8198 776.189 65.6046 779.939 L59.0074 779.939 Q60.8476 776.397 61.7851 772.96 Q62.6879 769.522 62.6879 766.154 Q62.6879 761.641 61.1601 759.21 Q59.5976 756.779 56.7852 756.779 Q54.181 756.779 52.7921 758.55 Q51.4032 760.286 50.1185 766.224 L49.5977 768.446 Q48.3477 774.384 45.7783 777.022 Q43.1741 779.661 38.6603 779.661 Q33.1742 779.661 30.1881 775.772 Q27.202 771.884 27.202 764.731 Q27.202 761.189 27.7228 758.064 Q28.2436 754.939 29.2853 752.3 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M28.1395 740.043 L28.1395 733.655 L67.0282 733.655 L67.0282 740.043 L28.1395 740.043 M13.0007 740.043 L13.0007 733.655 L21.0909 733.655 L21.0909 740.043 L13.0007 740.043 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M17.0979 713.967 L28.1395 713.967 L28.1395 700.808 L33.1047 700.808 L33.1047 713.967 L54.2157 713.967 Q58.9726 713.967 60.3268 712.683 Q61.681 711.363 61.681 707.37 L61.681 700.808 L67.0282 700.808 L67.0282 707.37 Q67.0282 714.766 64.2851 717.578 Q61.5074 720.391 54.2157 720.391 L33.1047 720.391 L33.1047 725.078 L28.1395 725.078 L28.1395 720.391 L17.0979 720.391 L17.0979 713.967 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M28.1395 692.405 L28.1395 686.016 L67.0282 686.016 L67.0282 692.405 L28.1395 692.405 M13.0007 692.405 L13.0007 686.016 L21.0909 686.016 L21.0909 692.405 L13.0007 692.405 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M32.6186 657.579 Q32.6186 662.717 36.6464 665.704 Q40.6394 668.69 47.6185 668.69 Q54.5977 668.69 58.6254 665.738 Q62.6185 662.752 62.6185 657.579 Q62.6185 652.474 58.5907 649.488 Q54.5629 646.502 47.6185 646.502 Q40.7089 646.502 36.6811 649.488 Q32.6186 652.474 32.6186 657.579 M27.202 657.579 Q27.202 649.245 32.6186 644.488 Q38.0353 639.732 47.6185 639.732 Q57.1671 639.732 62.6185 644.488 Q68.0351 649.245 68.0351 657.579 Q68.0351 665.947 62.6185 670.704 Q57.1671 675.426 47.6185 675.426 Q38.0353 675.426 32.6186 670.704 Q27.202 665.947 27.202 657.579 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M43.5561 596.815 L67.0282 596.815 L67.0282 603.204 L43.7644 603.204 Q38.2436 603.204 35.5005 605.357 Q32.7575 607.509 32.7575 611.815 Q32.7575 616.989 36.0561 619.975 Q39.3547 622.961 45.0491 622.961 L67.0282 622.961 L67.0282 629.384 L28.1395 629.384 L28.1395 622.961 L34.1811 622.961 Q30.6742 620.669 28.9381 617.579 Q27.202 614.454 27.202 610.391 Q27.202 603.69 31.3686 600.253 Q35.5005 596.815 43.5561 596.815 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M13.0007 562.058 L13.0007 547.336 L17.9659 547.336 L17.9659 555.669 L71.4379 555.669 L71.4379 547.336 L76.4031 547.336 L76.4031 562.058 L13.0007 562.058 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M35.6047 503.448 Q31.2992 501.052 29.2506 497.718 Q27.202 494.385 27.202 489.871 Q27.202 483.795 31.4728 480.496 Q35.7089 477.198 43.5561 477.198 L67.0282 477.198 L67.0282 483.621 L43.7644 483.621 Q38.1741 483.621 35.4658 485.6 Q32.7575 487.58 32.7575 491.642 Q32.7575 496.607 36.0561 499.489 Q39.3547 502.371 45.0491 502.371 L67.0282 502.371 L67.0282 508.795 L43.7644 508.795 Q38.1394 508.795 35.4658 510.774 Q32.7575 512.753 32.7575 516.885 Q32.7575 521.781 36.0908 524.663 Q39.3894 527.545 45.0491 527.545 L67.0282 527.545 L67.0282 533.968 L28.1395 533.968 L28.1395 527.545 L34.1811 527.545 Q30.6047 525.357 28.9034 522.302 Q27.202 519.246 27.202 515.045 Q27.202 510.809 29.3547 507.857 Q31.5075 504.871 35.6047 503.448 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M13.0007 449.524 L76.4031 449.524 L76.4031 464.246 L71.4379 464.246 L71.4379 455.948 L17.9659 455.948 L17.9659 464.246 L13.0007 464.246 L13.0007 449.524 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M625.966 18.8205 L625.966 65.8515 L635.85 65.8515 Q648.367 65.8515 654.16 60.1802 Q659.994 54.509 659.994 42.2752 Q659.994 30.1225 654.16 24.4918 Q648.367 18.8205 635.85 18.8205 L625.966 18.8205 M617.783 12.096 L634.594 12.096 Q652.175 12.096 660.399 19.4281 Q668.622 26.7198 668.622 42.2752 Q668.622 57.9117 660.358 65.2439 Q652.094 72.576 634.594 72.576 L617.783 72.576 L617.783 12.096 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M720.149 48.0275 L720.149 51.6733 L685.879 51.6733 Q686.365 59.3701 690.497 63.421 Q694.669 67.4314 702.082 67.4314 Q706.376 67.4314 710.387 66.3781 Q714.438 65.3249 718.408 63.2184 L718.408 70.267 Q714.397 71.9684 710.184 72.8596 Q705.971 73.7508 701.637 73.7508 Q690.78 73.7508 684.421 67.4314 Q678.101 61.1119 678.101 50.3365 Q678.101 39.1965 684.096 32.6746 Q690.132 26.1121 700.341 26.1121 Q709.496 26.1121 714.802 32.0264 Q720.149 37.9003 720.149 48.0275 M712.696 45.84 Q712.615 39.7232 709.253 36.0774 Q705.931 32.4315 700.422 32.4315 Q694.183 32.4315 690.416 35.9558 Q686.689 39.4801 686.122 45.8805 L712.696 45.84 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M755.352 9.54393 L755.352 15.7418 L748.222 15.7418 Q744.212 15.7418 742.632 17.3622 Q741.093 18.9825 741.093 23.1955 L741.093 27.2059 L753.367 27.2059 L753.367 32.9987 L741.093 32.9987 L741.093 72.576 L733.598 72.576 L733.598 32.9987 L726.469 32.9987 L726.469 27.2059 L733.598 27.2059 L733.598 24.0462 Q733.598 16.471 737.123 13.0277 Q740.647 9.54393 748.303 9.54393 L755.352 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M782.209 49.7694 Q773.176 49.7694 769.692 51.8354 Q766.208 53.9013 766.208 58.8839 Q766.208 62.8538 768.801 65.2034 Q771.434 67.5124 775.93 67.5124 Q782.128 67.5124 785.855 63.1374 Q789.622 58.7219 789.622 51.4303 L789.622 49.7694 L782.209 49.7694 M797.076 46.6907 L797.076 72.576 L789.622 72.576 L789.622 65.6895 Q787.07 69.8214 783.263 71.8063 Q779.455 73.7508 773.945 73.7508 Q766.978 73.7508 762.846 69.8619 Q758.755 65.9325 758.755 59.3701 Q758.755 51.7138 763.859 47.825 Q769.003 43.9361 779.171 43.9361 L789.622 43.9361 L789.622 43.2069 Q789.622 38.0623 786.22 35.2672 Q782.857 32.4315 776.741 32.4315 Q772.852 32.4315 769.165 33.3632 Q765.479 34.295 762.076 36.1584 L762.076 29.2718 Q766.168 27.692 770.016 26.9223 Q773.864 26.1121 777.51 26.1121 Q787.354 26.1121 792.215 31.2163 Q797.076 36.3204 797.076 46.6907 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M811.659 54.671 L811.659 27.2059 L819.113 27.2059 L819.113 54.3874 Q819.113 60.8284 821.625 64.0691 Q824.136 67.2693 829.159 67.2693 Q835.195 67.2693 838.679 63.421 Q842.203 59.5726 842.203 52.9291 L842.203 27.2059 L849.657 27.2059 L849.657 72.576 L842.203 72.576 L842.203 65.6084 Q839.489 69.7404 835.884 71.7658 Q832.319 73.7508 827.579 73.7508 Q819.761 73.7508 815.71 68.8897 Q811.659 64.0286 811.659 54.671 M830.415 26.1121 L830.415 26.1121 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M865.01 9.54393 L872.463 9.54393 L872.463 72.576 L865.01 72.576 L865.01 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M895.432 14.324 L895.432 27.2059 L910.785 27.2059 L910.785 32.9987 L895.432 32.9987 L895.432 57.6282 Q895.432 63.1779 896.931 64.7578 Q898.47 66.3376 903.129 66.3376 L910.785 66.3376 L910.785 72.576 L903.129 72.576 Q894.5 72.576 891.219 69.3758 Q887.938 66.1351 887.938 57.6282 L887.938 32.9987 L882.469 32.9987 L882.469 27.2059 L887.938 27.2059 L887.938 14.324 L895.432 14.324 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M922.492 62.2867 L931.04 62.2867 L931.04 72.576 L922.492 72.576 L922.492 62.2867 M922.492 29.6769 L931.04 29.6769 L931.04 39.9662 L922.492 39.9662 L922.492 29.6769 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M975.235 12.096 L987.428 12.096 L1002.86 53.2532 L1018.38 12.096 L1030.57 12.096 L1030.57 72.576 L1022.59 72.576 L1022.59 19.4686 L1006.99 60.9499 L998.771 60.9499 L983.175 19.4686 L983.175 72.576 L975.235 72.576 L975.235 12.096 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M1067.11 49.7694 Q1058.08 49.7694 1054.59 51.8354 Q1051.11 53.9013 1051.11 58.8839 Q1051.11 62.8538 1053.7 65.2034 Q1056.33 67.5124 1060.83 67.5124 Q1067.03 67.5124 1070.76 63.1374 Q1074.52 58.7219 1074.52 51.4303 L1074.52 49.7694 L1067.11 49.7694 M1081.98 46.6907 L1081.98 72.576 L1074.52 72.576 L1074.52 65.6895 Q1071.97 69.8214 1068.16 71.8063 Q1064.35 73.7508 1058.85 73.7508 Q1051.88 73.7508 1047.75 69.8619 Q1043.65 65.9325 1043.65 59.3701 Q1043.65 51.7138 1048.76 47.825 Q1053.9 43.9361 1064.07 43.9361 L1074.52 43.9361 L1074.52 43.2069 Q1074.52 38.0623 1071.12 35.2672 Q1067.76 32.4315 1061.64 32.4315 Q1057.75 32.4315 1054.07 33.3632 Q1050.38 34.295 1046.98 36.1584 L1046.98 29.2718 Q1051.07 27.692 1054.92 26.9223 Q1058.76 26.1121 1062.41 26.1121 Q1072.25 26.1121 1077.12 31.2163 Q1081.98 36.3204 1081.98 46.6907 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M1126.25 28.5427 L1126.25 35.5912 Q1123.09 33.9709 1119.69 33.1607 Q1116.29 32.3505 1112.64 32.3505 Q1107.09 32.3505 1104.3 34.0519 Q1101.54 35.7533 1101.54 39.156 Q1101.54 41.7486 1103.53 43.2475 Q1105.51 44.7058 1111.51 46.0426 L1114.06 46.6097 Q1122 48.3111 1125.32 51.4303 Q1128.68 54.509 1128.68 60.0587 Q1128.68 66.3781 1123.66 70.0644 Q1118.68 73.7508 1109.93 73.7508 Q1106.28 73.7508 1102.31 73.0216 Q1098.38 72.3329 1094.01 70.9151 L1094.01 63.2184 Q1098.14 65.3654 1102.15 66.4591 Q1106.16 67.5124 1110.09 67.5124 Q1115.36 67.5124 1118.19 65.73 Q1121.03 63.9071 1121.03 60.6258 Q1121.03 57.5877 1118.96 55.9673 Q1116.94 54.3469 1110.01 52.8481 L1107.42 52.2405 Q1100.49 50.7821 1097.41 47.7845 Q1094.33 44.7463 1094.33 39.4801 Q1094.33 33.0797 1098.87 29.5959 Q1103.41 26.1121 1111.75 26.1121 Q1115.88 26.1121 1119.53 26.7198 Q1123.17 27.3274 1126.25 28.5427 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M1169.48 28.5427 L1169.48 35.5912 Q1166.32 33.9709 1162.91 33.1607 Q1159.51 32.3505 1155.86 32.3505 Q1150.31 32.3505 1147.52 34.0519 Q1144.77 35.7533 1144.77 39.156 Q1144.77 41.7486 1146.75 43.2475 Q1148.74 44.7058 1154.73 46.0426 L1157.28 46.6097 Q1165.22 48.3111 1168.54 51.4303 Q1171.91 54.509 1171.91 60.0587 Q1171.91 66.3781 1166.88 70.0644 Q1161.9 73.7508 1153.15 73.7508 Q1149.5 73.7508 1145.53 73.0216 Q1141.61 72.3329 1137.23 70.9151 L1137.23 63.2184 Q1141.36 65.3654 1145.37 66.4591 Q1149.38 67.5124 1153.31 67.5124 Q1158.58 67.5124 1161.41 65.73 Q1164.25 63.9071 1164.25 60.6258 Q1164.25 57.5877 1162.18 55.9673 Q1160.16 54.3469 1153.23 52.8481 L1150.64 52.2405 Q1143.71 50.7821 1140.63 47.7845 Q1137.55 44.7463 1137.55 39.4801 Q1137.55 33.0797 1142.09 29.5959 Q1146.63 26.1121 1154.97 26.1121 Q1159.11 26.1121 1162.75 26.7198 Q1166.4 27.3274 1169.48 28.5427 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M1217.36 65.7705 L1217.36 89.8329 L1209.86 89.8329 L1209.86 27.2059 L1217.36 27.2059 L1217.36 34.0924 Q1219.71 30.0415 1223.27 28.0971 Q1226.88 26.1121 1231.86 26.1121 Q1240.12 26.1121 1245.27 32.6746 Q1250.45 39.2371 1250.45 49.9314 Q1250.45 60.6258 1245.27 67.1883 Q1240.12 73.7508 1231.86 73.7508 Q1226.88 73.7508 1223.27 71.8063 Q1219.71 69.8214 1217.36 65.7705 M1242.72 49.9314 Q1242.72 41.7081 1239.31 37.0496 Q1235.95 32.3505 1230.04 32.3505 Q1224.12 32.3505 1220.72 37.0496 Q1217.36 41.7081 1217.36 49.9314 Q1217.36 58.1548 1220.72 62.8538 Q1224.12 67.5124 1230.04 67.5124 Q1235.95 67.5124 1239.31 62.8538 Q1242.72 58.1548 1242.72 49.9314 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M1280.39 32.4315 Q1274.39 32.4315 1270.91 37.1306 Q1267.43 41.7891 1267.43 49.9314 Q1267.43 58.0738 1270.87 62.7728 Q1274.35 67.4314 1280.39 67.4314 Q1286.34 67.4314 1289.83 62.7323 Q1293.31 58.0333 1293.31 49.9314 Q1293.31 41.8701 1289.83 37.1711 Q1286.34 32.4315 1280.39 32.4315 M1280.39 26.1121 Q1290.11 26.1121 1295.66 32.4315 Q1301.21 38.7509 1301.21 49.9314 Q1301.21 61.0714 1295.66 67.4314 Q1290.11 73.7508 1280.39 73.7508 Q1270.63 73.7508 1265.08 67.4314 Q1259.57 61.0714 1259.57 49.9314 Q1259.57 38.7509 1265.08 32.4315 Q1270.63 26.1121 1280.39 26.1121 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M1342.49 28.5427 L1342.49 35.5912 Q1339.33 33.9709 1335.93 33.1607 Q1332.52 32.3505 1328.88 32.3505 Q1323.33 32.3505 1320.53 34.0519 Q1317.78 35.7533 1317.78 39.156 Q1317.78 41.7486 1319.76 43.2475 Q1321.75 44.7058 1327.74 46.0426 L1330.3 46.6097 Q1338.24 48.3111 1341.56 51.4303 Q1344.92 54.509 1344.92 60.0587 Q1344.92 66.3781 1339.9 70.0644 Q1334.91 73.7508 1326.16 73.7508 Q1322.52 73.7508 1318.55 73.0216 Q1314.62 72.3329 1310.24 70.9151 L1310.24 63.2184 Q1314.38 65.3654 1318.39 66.4591 Q1322.4 67.5124 1326.33 67.5124 Q1331.59 67.5124 1334.43 65.73 Q1337.26 63.9071 1337.26 60.6258 Q1337.26 57.5877 1335.2 55.9673 Q1333.17 54.3469 1326.25 52.8481 L1323.65 52.2405 Q1316.73 50.7821 1313.65 47.7845 Q1310.57 44.7463 1310.57 39.4801 Q1310.57 33.0797 1315.11 29.5959 Q1319.64 26.1121 1327.99 26.1121 Q1332.12 26.1121 1335.77 26.7198 Q1339.41 27.3274 1342.49 28.5427 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M1356.79 27.2059 L1364.24 27.2059 L1364.24 72.576 L1356.79 72.576 L1356.79 27.2059 M1356.79 9.54393 L1364.24 9.54393 L1364.24 18.9825 L1356.79 18.9825 L1356.79 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M1387.21 14.324 L1387.21 27.2059 L1402.56 27.2059 L1402.56 32.9987 L1387.21 32.9987 L1387.21 57.6282 Q1387.21 63.1779 1388.71 64.7578 Q1390.25 66.3376 1394.91 66.3376 L1402.56 66.3376 L1402.56 72.576 L1394.91 72.576 Q1386.28 72.576 1383 69.3758 Q1379.72 66.1351 1379.72 57.6282 L1379.72 32.9987 L1374.25 32.9987 L1374.25 27.2059 L1379.72 27.2059 L1379.72 14.324 L1387.21 14.324 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M1412.37 27.2059 L1419.82 27.2059 L1419.82 72.576 L1412.37 72.576 L1412.37 27.2059 M1412.37 9.54393 L1419.82 9.54393 L1419.82 18.9825 L1412.37 18.9825 L1412.37 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M1453 32.4315 Q1447 32.4315 1443.52 37.1306 Q1440.04 41.7891 1440.04 49.9314 Q1440.04 58.0738 1443.48 62.7728 Q1446.96 67.4314 1453 67.4314 Q1458.95 67.4314 1462.44 62.7323 Q1465.92 58.0333 1465.92 49.9314 Q1465.92 41.8701 1462.44 37.1711 Q1458.95 32.4315 1453 32.4315 M1453 26.1121 Q1462.72 26.1121 1468.27 32.4315 Q1473.82 38.7509 1473.82 49.9314 Q1473.82 61.0714 1468.27 67.4314 Q1462.72 73.7508 1453 73.7508 Q1443.24 73.7508 1437.69 67.4314 Q1432.18 61.0714 1432.18 49.9314 Q1432.18 38.7509 1437.69 32.4315 Q1443.24 26.1121 1453 26.1121 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M1523.89 45.1919 L1523.89 72.576 L1516.44 72.576 L1516.44 45.4349 Q1516.44 38.994 1513.92 35.7938 Q1511.41 32.5936 1506.39 32.5936 Q1500.35 32.5936 1496.87 36.4419 Q1493.39 40.2903 1493.39 46.9338 L1493.39 72.576 L1485.89 72.576 L1485.89 27.2059 L1493.39 27.2059 L1493.39 34.2544 Q1496.06 30.163 1499.66 28.1376 Q1503.31 26.1121 1508.05 26.1121 Q1515.87 26.1121 1519.88 30.9732 Q1523.89 35.7938 1523.89 45.1919 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M1585.75 49.7694 Q1576.71 49.7694 1573.23 51.8354 Q1569.75 53.9013 1569.75 58.8839 Q1569.75 62.8538 1572.34 65.2034 Q1574.97 67.5124 1579.47 67.5124 Q1585.67 67.5124 1589.39 63.1374 Q1593.16 58.7219 1593.16 51.4303 L1593.16 49.7694 L1585.75 49.7694 M1600.61 46.6907 L1600.61 72.576 L1593.16 72.576 L1593.16 65.6895 Q1590.61 69.8214 1586.8 71.8063 Q1582.99 73.7508 1577.48 73.7508 Q1570.52 73.7508 1566.38 69.8619 Q1562.29 65.9325 1562.29 59.3701 Q1562.29 51.7138 1567.4 47.825 Q1572.54 43.9361 1582.71 43.9361 L1593.16 43.9361 L1593.16 43.2069 Q1593.16 38.0623 1589.76 35.2672 Q1586.39 32.4315 1580.28 32.4315 Q1576.39 32.4315 1572.7 33.3632 Q1569.02 34.295 1565.61 36.1584 L1565.61 29.2718 Q1569.71 27.692 1573.55 26.9223 Q1577.4 26.1121 1581.05 26.1121 Q1590.89 26.1121 1595.75 31.2163 Q1600.61 36.3204 1600.61 46.6907 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M1638.94 9.54393 L1638.94 15.7418 L1631.81 15.7418 Q1627.8 15.7418 1626.22 17.3622 Q1624.68 18.9825 1624.68 23.1955 L1624.68 27.2059 L1636.95 27.2059 L1636.95 32.9987 L1624.68 32.9987 L1624.68 72.576 L1617.18 72.576 L1617.18 32.9987 L1610.05 32.9987 L1610.05 27.2059 L1617.18 27.2059 L1617.18 24.0462 Q1617.18 16.471 1620.71 13.0277 Q1624.23 9.54393 1631.89 9.54393 L1638.94 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M1651.09 14.324 L1651.09 27.2059 L1666.44 27.2059 L1666.44 32.9987 L1651.09 32.9987 L1651.09 57.6282 Q1651.09 63.1779 1652.59 64.7578 Q1654.13 66.3376 1658.78 66.3376 L1666.44 66.3376 L1666.44 72.576 L1658.78 72.576 Q1650.16 72.576 1646.87 69.3758 Q1643.59 66.1351 1643.59 57.6282 L1643.59 32.9987 L1638.12 32.9987 L1638.12 27.2059 L1643.59 27.2059 L1643.59 14.324 L1651.09 14.324 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M1715.05 48.0275 L1715.05 51.6733 L1680.78 51.6733 Q1681.27 59.3701 1685.4 63.421 Q1689.57 67.4314 1696.98 67.4314 Q1701.28 67.4314 1705.29 66.3781 Q1709.34 65.3249 1713.31 63.2184 L1713.31 70.267 Q1709.3 71.9684 1705.09 72.8596 Q1700.87 73.7508 1696.54 73.7508 Q1685.68 73.7508 1679.32 67.4314 Q1673 61.1119 1673 50.3365 Q1673 39.1965 1679 32.6746 Q1685.03 26.1121 1695.24 26.1121 Q1704.4 26.1121 1709.7 32.0264 Q1715.05 37.9003 1715.05 48.0275 M1707.6 45.84 Q1707.52 39.7232 1704.15 36.0774 Q1700.83 32.4315 1695.32 32.4315 Q1689.09 32.4315 1685.32 35.9558 Q1681.59 39.4801 1681.02 45.8805 L1707.6 45.84 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M1753.58 34.1734 Q1752.32 33.4443 1750.82 33.1202 Q1749.36 32.7556 1747.58 32.7556 Q1741.26 32.7556 1737.86 36.8875 Q1734.5 40.9789 1734.5 48.6757 L1734.5 72.576 L1727 72.576 L1727 27.2059 L1734.5 27.2059 L1734.5 34.2544 Q1736.85 30.1225 1740.61 28.1376 Q1744.38 26.1121 1749.77 26.1121 Q1750.54 26.1121 1751.47 26.2337 Q1752.4 26.3147 1753.54 26.5172 L1753.58 34.1734 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M1816.77 44.2197 Q1819.4 45.1109 1821.87 48.0275 Q1824.39 50.9442 1826.9 56.0483 L1835.2 72.576 L1826.41 72.576 L1818.67 57.061 Q1815.68 50.9847 1812.84 48.9997 Q1810.05 47.0148 1805.18 47.0148 L1796.27 47.0148 L1796.27 72.576 L1788.09 72.576 L1788.09 12.096 L1806.56 12.096 Q1816.93 12.096 1822.04 16.4305 Q1827.14 20.7649 1827.14 29.5149 Q1827.14 35.2267 1824.47 38.994 Q1821.83 42.7613 1816.77 44.2197 M1796.27 18.8205 L1796.27 40.2903 L1806.56 40.2903 Q1812.48 40.2903 1815.47 37.5762 Q1818.51 34.8216 1818.51 29.5149 Q1818.51 24.2082 1815.47 21.5346 Q1812.48 18.8205 1806.56 18.8205 L1796.27 18.8205 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M1840.91 54.671 L1840.91 27.2059 L1848.37 27.2059 L1848.37 54.3874 Q1848.37 60.8284 1850.88 64.0691 Q1853.39 67.2693 1858.41 67.2693 Q1864.45 67.2693 1867.93 63.421 Q1871.46 59.5726 1871.46 52.9291 L1871.46 27.2059 L1878.91 27.2059 L1878.91 72.576 L1871.46 72.576 L1871.46 65.6084 Q1868.74 69.7404 1865.14 71.7658 Q1861.57 73.7508 1856.83 73.7508 Q1849.01 73.7508 1844.96 68.8897 Q1840.91 64.0286 1840.91 54.671 M1859.67 26.1121 L1859.67 26.1121 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M1931.98 45.1919 L1931.98 72.576 L1924.52 72.576 L1924.52 45.4349 Q1924.52 38.994 1922.01 35.7938 Q1919.5 32.5936 1914.48 32.5936 Q1908.44 32.5936 1904.96 36.4419 Q1901.47 40.2903 1901.47 46.9338 L1901.47 72.576 L1893.98 72.576 L1893.98 27.2059 L1901.47 27.2059 L1901.47 34.2544 Q1904.15 30.163 1907.75 28.1376 Q1911.4 26.1121 1916.14 26.1121 Q1923.96 26.1121 1927.97 30.9732 Q1931.98 35.7938 1931.98 45.1919 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M1948.75 62.2867 L1957.3 62.2867 L1957.3 72.576 L1948.75 72.576 L1948.75 62.2867 M1948.75 29.6769 L1957.3 29.6769 L1957.3 39.9662 L1948.75 39.9662 L1948.75 29.6769 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M2003.64 65.6895 L2017.01 65.6895 L2017.01 19.5497 L2002.46 22.4663 L2002.46 15.0127 L2016.92 12.096 L2025.11 12.096 L2025.11 65.6895 L2038.48 65.6895 L2038.48 72.576 L2003.64 72.576 L2003.64 65.6895 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip392)" style="stroke:#009af9; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" points="
  361.501,1365.41 366.334,1365.1 371.167,1364.2 376,1362.69 380.833,1360.59 385.666,1357.89 390.5,1354.59 395.333,1350.7 400.166,1346.22 404.999,1341.16 
  409.832,1335.52 414.665,1329.31 419.498,1322.53 424.332,1315.19 429.165,1307.3 433.998,1298.87 438.831,1289.89 443.664,1280.39 448.497,1270.38 453.33,1259.86 
  458.164,1248.84 462.997,1237.33 467.83,1225.34 472.663,1212.9 477.496,1200 482.329,1186.66 487.162,1172.9 491.996,1158.73 496.829,1144.16 501.662,1129.21 
  506.495,1113.9 511.328,1098.23 516.161,1082.23 520.994,1065.9 525.828,1049.28 530.661,1032.36 535.494,1015.18 540.327,997.743 545.16,980.072 549.993,962.183 
  554.826,944.096 559.66,925.826 564.493,907.393 569.326,888.815 574.159,870.111 578.992,851.3 583.825,832.399 588.658,813.429 593.492,794.408 598.325,775.355 
  603.158,756.289 607.991,737.229 612.824,718.195 617.657,699.205 622.49,680.278 627.324,661.433 632.157,642.69 636.99,624.067 641.823,605.582 646.656,587.253 
  651.489,569.101 656.322,551.141 661.156,533.393 665.989,515.874 670.822,498.602 675.655,481.594 680.488,464.866 685.321,448.437 690.154,432.321 694.988,416.536 
  699.821,401.097 704.654,386.019 709.487,371.317 714.32,357.007 719.153,343.102 723.986,329.617 728.82,316.565 733.653,303.959 738.486,291.811 743.319,280.134 
  748.152,268.939 752.985,258.238 757.818,248.042 762.652,238.359 767.485,229.202 772.318,220.577 777.151,212.494 781.984,204.961 786.817,197.986 791.65,191.576 
  796.484,185.736 801.317,180.473 806.15,175.793 810.983,171.699 815.816,168.196 820.649,165.287 825.482,162.976 830.316,161.264 835.149,160.153 839.982,159.645 
  844.815,159.74 849.648,160.437 854.481,161.737 859.314,163.637 864.148,166.137 868.981,169.232 873.814,172.922 878.647,177.2 883.48,182.064 888.313,187.509 
  893.147,193.528 897.98,200.117 902.813,207.268 907.646,214.975 912.479,223.229 917.312,232.023 922.145,241.348 926.979,251.194 931.812,261.55 936.645,272.408 
  941.478,283.757 946.311,295.584 951.144,307.877 955.977,320.626 960.811,333.816 965.644,347.435 970.477,361.468 975.31,375.903 980.143,390.724 984.976,405.917 
  989.809,421.467 994.643,437.357 999.476,453.573 1004.31,470.098 1009.14,486.915 1013.98,504.008 1018.81,521.359 1023.64,538.951 1028.47,556.767 1033.31,574.789 
  1038.14,592.998 1042.97,611.377 1047.81,629.908 1052.64,648.571 1057.47,667.348 1062.31,686.221 1067.14,705.17 1071.97,724.176 1076.81,743.221 1081.64,762.284 
  1086.47,781.349 1091.31,800.395 1096.14,819.403 1100.97,838.355 1105.8,857.231 1110.64,876.012 1115.47,894.679 1120.3,913.213 1125.14,931.598 1129.97,949.813 
  1134.8,967.842 1139.64,985.664 1144.47,1003.26 1149.3,1020.62 1154.14,1037.72 1158.97,1054.55 1163.8,1071.08 1168.64,1087.3 1173.47,1103.2 1178.3,1118.76 
  1183.14,1133.96 1187.97,1148.79 1192.8,1163.23 1197.63,1177.27 1202.47,1190.9 1207.3,1204.1 1212.13,1216.85 1216.97,1229.16 1221.8,1240.99 1226.63,1252.35 
  1231.47,1263.22 1236.3,1273.58 1241.13,1283.44 1245.97,1292.77 1250.8,1301.57 1255.63,1309.84 1260.47,1317.56 1265.3,1324.72 1270.13,1331.32 1274.96,1337.35 
  1279.8,1342.81 1284.63,1347.68 1289.46,1351.97 1294.3,1355.67 1299.13,1358.78 1303.96,1361.29 1308.8,1363.21 1313.63,1364.52 1318.46,1365.23 1323.3,1365.34 
  1328.13,1364.84 1332.96,1363.74 1337.8,1362.04 1342.63,1359.75 1347.46,1356.85 1352.3,1353.36 1357.13,1349.28 1361.96,1344.61 1366.79,1339.36 1371.63,1333.54 
  1376.46,1327.14 1381.29,1320.18 1386.13,1312.66 1390.96,1304.59 1395.79,1295.98 1400.63,1286.83 1405.46,1277.16 1410.29,1266.98 1415.13,1256.29 1419.96,1245.11 
  1424.79,1233.45 1429.63,1221.31 1434.46,1208.71 1439.29,1195.67 1444.12,1182.2 1448.96,1168.31 1453.79,1154.01 1458.62,1139.32 1463.46,1124.25 1468.29,1108.83 
  1473.12,1093.05 1477.96,1076.95 1482.79,1060.53 1487.62,1043.81 1492.46,1026.81 1497.29,1009.55 1502.12,992.042 1506.96,974.303 1511.79,956.353 1516.62,938.209 
  1521.46,919.889 1526.29,901.414 1531.12,882.8 1535.95,864.064 1540.79,845.227 1545.62,826.307 1550.45,807.324 1555.29,788.296 1560.12,769.243 1564.95,750.184 
  1569.79,731.138 1574.62,712.122 1579.45,693.156 1584.29,674.261 1589.12,655.453 1593.95,636.754 1598.79,618.18 1603.62,599.751 1608.45,581.484 1613.28,563.399 
  1618.12,545.513 1622.95,527.844 1627.78,510.41 1632.62,493.229 1637.45,476.316 1642.28,459.691 1647.12,443.368 1651.95,427.364 1656.78,411.696 1661.62,396.379 
  1666.45,381.428 1671.28,366.859 1676.12,352.685 1680.95,338.92 1685.78,325.581 1690.62,312.678 1695.45,300.225 1700.28,288.235 1705.11,276.719 1709.95,265.689 
  1714.78,255.155 1719.61,245.129 1724.45,235.621 1729.28,226.64 1734.11,218.195 1738.95,210.295 1743.78,202.947 1748.61,196.157 1753.45,189.934 1758.28,184.284 
  1763.11,179.214 1767.95,174.727 1772.78,170.827 1777.61,167.519 1782.44,164.806 1787.28,162.69 1792.11,161.173 1796.94,160.26 1801.78,159.949 1806.61,160.24 
  1811.44,161.134 1816.28,162.629 1821.11,164.724 1825.94,167.415 1830.78,170.701 1835.61,174.582 1840.44,179.05 1845.28,184.101 1850.11,189.731 1854.94,195.934 
  1859.78,202.702 1864.61,210.029 1869.44,217.91 1874.27,226.336 1879.11,235.299 1883.94,244.788 1888.77,254.795 1893.61,265.31 1898.44,276.321 1903.27,287.817 
  1908.11,299.79 1912.94,312.225 1917.77,325.112 1922.61,338.435 1927.44,352.182 1932.27,366.339 1937.11,380.891 1941.94,395.824 1946.77,411.126 1951.6,426.78 
  1956.44,442.768 1961.27,459.077 1966.1,475.689 1970.94,492.586 1975.77,509.753 1980.6,527.173 1985.44,544.829 1990.27,562.703 1995.1,580.777 1999.94,599.032 
  2004.77,617.45 2009.6,636.013 2014.44,654.701 2019.27,673.498 2024.1,692.385 2028.94,711.341 2033.77,730.349 2038.6,749.388 2043.43,768.441 2048.27,787.486 
  2053.1,806.507 2057.93,825.484 2062.77,844.398 2067.6,863.23 2072.43,881.961 2077.27,900.573 2082.1,919.046 2086.93,937.363 2091.77,955.504 2096.6,973.452 
  2101.43,991.19 2106.27,1008.7 2111.1,1025.96 2115.93,1042.96 2120.77,1059.68 2125.6,1076.1 2130.43,1092.21 2135.26,1107.98 2140.1,1123.42 2144.93,1138.49 
  2149.76,1153.18 2154.6,1167.48 2159.43,1181.38 2164.26,1194.86 2169.1,1207.91 2173.93,1220.51 2178.76,1232.66 2183.6,1244.33 2188.43,1255.52 2193.26,1266.22 
  2198.1,1276.42 2202.93,1286.1 2207.76,1295.26 2212.59,1303.88 2217.43,1311.96 2222.26,1319.49 2227.09,1326.47 2231.93,1332.88 2236.76,1338.73 2241.59,1343.99 
  2246.43,1348.68 2251.26,1352.77 2256.09,1356.28 2260.93,1359.19 2265.76,1361.51 2270.59,1363.23 2275.43,1364.35 2280.26,1364.87 2285.09,1364.78 2289.93,1364.09 
  2294.76,1362.8 
  "/>
<polyline clip-path="url(#clip392)" style="stroke:#e26f46; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" points="
  361.501,1365.41 366.334,1365.19 371.167,1364.55 376,1363.47 380.833,1361.95 385.666,1360 390.5,1357.61 395.333,1354.79 400.166,1351.52 404.999,1347.82 
  409.832,1343.68 414.665,1339.12 419.498,1334.12 424.332,1328.7 429.165,1322.86 433.998,1316.6 438.831,1309.93 443.664,1302.85 448.497,1295.38 453.33,1287.51 
  458.164,1279.26 462.997,1270.62 467.83,1261.63 472.663,1252.27 477.496,1242.55 482.329,1232.5 487.162,1222.11 491.996,1211.4 496.829,1200.37 501.662,1189.05 
  506.495,1177.43 511.328,1165.53 516.161,1153.37 520.994,1140.95 525.828,1128.29 530.661,1115.4 535.494,1102.29 540.327,1088.97 545.16,1075.47 549.993,1061.79 
  554.826,1047.94 559.66,1033.94 564.493,1019.8 569.326,1005.54 574.159,991.18 578.992,976.726 583.825,962.197 588.658,947.601 593.492,932.956 598.325,918.278 
  603.158,903.581 607.991,888.88 612.824,874.188 617.657,859.52 622.49,844.891 627.324,830.316 632.157,815.808 636.99,801.383 641.823,787.054 646.656,772.837 
  651.489,758.745 656.322,744.792 661.156,730.992 665.989,717.359 670.822,703.908 675.655,690.652 680.488,677.602 685.321,664.773 690.154,652.178 694.988,639.83 
  699.821,627.741 704.654,615.922 709.487,604.386 714.32,593.145 719.153,582.211 723.986,571.594 728.82,561.305 733.653,551.354 738.486,541.753 743.319,532.511 
  748.152,523.637 752.985,515.141 757.818,507.032 762.652,499.318 767.485,492.009 772.318,485.111 777.151,478.634 781.984,472.583 786.817,466.967 791.65,461.793 
  796.484,457.066 801.317,452.793 806.15,448.98 810.983,445.632 815.816,442.756 820.649,440.356 825.482,438.439 830.316,437.007 835.149,436.067 839.982,435.623 
  844.815,435.654 849.648,436.048 854.481,436.799 859.314,437.909 864.148,439.378 868.981,441.209 873.814,443.402 878.647,445.956 883.48,448.872 888.313,452.148 
  893.147,455.782 897.98,459.772 902.813,464.116 907.646,468.811 912.479,473.853 917.312,479.239 922.145,484.964 926.979,491.025 931.812,497.414 936.645,504.126 
  941.478,511.157 946.311,518.499 951.144,526.145 955.977,534.088 960.811,542.322 965.644,550.837 970.477,559.624 975.31,568.676 980.143,577.986 984.976,587.544 
  989.809,597.341 994.643,607.367 999.476,617.612 1004.31,628.067 1009.14,638.72 1013.98,649.561 1018.81,660.58 1023.64,671.765 1028.47,683.105 1033.31,694.589 
  1038.14,706.205 1042.97,717.942 1047.81,729.789 1052.64,741.732 1057.47,753.761 1062.31,765.864 1067.14,778.029 1071.97,790.243 1076.81,802.494 1081.64,814.77 
  1086.47,827.059 1091.31,839.349 1096.14,851.628 1100.97,863.882 1105.8,876.099 1110.64,888.267 1115.47,900.376 1120.3,912.412 1125.14,924.362 1129.97,936.215 
  1134.8,947.958 1139.64,959.581 1144.47,971.072 1149.3,982.419 1154.14,993.61 1158.97,1004.63 1163.8,1015.48 1168.64,1026.14 1173.47,1036.6 1178.3,1046.84 
  1183.14,1056.87 1187.97,1066.66 1192.8,1076.22 1197.63,1085.52 1202.47,1094.56 1207.3,1103.34 1212.13,1111.83 1216.97,1120.04 1221.8,1127.95 1226.63,1135.56 
  1231.47,1142.85 1236.3,1149.82 1241.13,1156.46 1245.97,1162.77 1250.8,1168.74 1255.63,1174.35 1260.47,1179.61 1265.3,1184.51 1270.13,1189.03 1274.96,1193.19 
  1279.8,1196.96 1284.63,1200.34 1289.46,1203.34 1294.3,1205.94 1299.13,1208.14 1303.96,1209.93 1308.8,1211.32 1313.63,1212.29 1318.46,1212.85 1323.3,1212.98 
  1328.13,1212.79 1332.96,1212.32 1337.8,1211.57 1342.63,1210.54 1347.46,1209.24 1352.3,1207.64 1357.13,1205.77 1361.96,1203.61 1366.79,1201.17 1371.63,1198.45 
  1376.46,1195.45 1381.29,1192.17 1386.13,1188.61 1390.96,1184.78 1395.79,1180.68 1400.63,1176.31 1405.46,1171.67 1410.29,1166.77 1415.13,1161.62 1419.96,1156.21 
  1424.79,1150.56 1429.63,1144.65 1434.46,1138.52 1439.29,1132.14 1444.12,1125.54 1448.96,1118.72 1453.79,1111.68 1458.62,1104.44 1463.46,1096.99 1468.29,1089.35 
  1473.12,1081.52 1477.96,1073.52 1482.79,1065.34 1487.62,1057 1492.46,1048.5 1497.29,1039.86 1502.12,1031.08 1506.96,1022.17 1511.79,1013.13 1516.62,1003.99 
  1521.46,994.74 1526.29,985.398 1531.12,975.969 1535.95,966.464 1540.79,956.892 1545.62,947.264 1550.45,937.59 1555.29,927.88 1560.12,918.14 1564.95,908.381 
  1569.79,898.614 1574.62,888.848 1579.45,879.095 1584.29,869.364 1589.12,859.665 1593.95,850.005 1598.79,840.396 1603.62,830.847 1608.45,821.368 1613.28,811.968 
  1618.12,802.659 1622.95,793.447 1627.78,784.344 1632.62,775.357 1637.45,766.496 1642.28,757.771 1647.12,749.189 1651.95,740.76 1656.78,732.491 1661.62,724.393 
  1666.45,716.473 1671.28,708.74 1676.12,701.2 1680.95,693.863 1685.78,686.734 1690.62,679.823 1695.45,673.138 1700.28,666.684 1705.11,660.469 1709.95,654.499 
  1714.78,648.781 1719.61,643.322 1724.45,638.128 1729.28,633.205 1734.11,628.559 1738.95,624.195 1743.78,620.119 1748.61,616.336 1753.45,612.851 1758.28,609.669 
  1763.11,606.794 1767.95,604.232 1772.78,601.987 1777.61,600.061 1782.44,598.46 1787.28,597.186 1792.11,596.244 1796.94,595.637 1801.78,595.369 1806.61,595.408 
  1811.44,595.644 1816.28,596.075 1821.11,596.701 1825.94,597.525 1830.78,598.547 1835.61,599.767 1840.44,601.186 1845.28,602.803 1850.11,604.619 1854.94,606.632 
  1859.78,608.843 1864.61,611.25 1869.44,613.851 1874.27,616.645 1879.11,619.632 1883.94,622.806 1888.77,626.168 1893.61,629.716 1898.44,633.444 1903.27,637.353 
  1908.11,641.437 1912.94,645.692 1917.77,650.115 1922.61,654.704 1927.44,659.455 1932.27,664.364 1937.11,669.424 1941.94,674.633 1946.77,679.987 1951.6,685.478 
  1956.44,691.103 1961.27,696.856 1966.1,702.731 1970.94,708.724 1975.77,714.827 1980.6,721.035 1985.44,727.343 1990.27,733.744 1995.1,740.232 1999.94,746.8 
  2004.77,753.443 2009.6,760.153 2014.44,766.925 2019.27,773.752 2024.1,780.626 2028.94,787.542 2033.77,794.492 2038.6,801.47 2043.43,808.468 2048.27,815.479 
  2053.1,822.497 2057.93,829.514 2062.77,836.525 2067.6,843.52 2072.43,850.494 2077.27,857.439 2082.1,864.348 2086.93,871.214 2091.77,878.031 2096.6,884.791 
  2101.43,891.488 2106.27,898.114 2111.1,904.663 2115.93,911.128 2120.77,917.503 2125.6,923.781 2130.43,929.955 2135.26,936.019 2140.1,941.966 2144.93,947.791 
  2149.76,953.487 2154.6,959.049 2159.43,964.47 2164.26,969.744 2169.1,974.866 2173.93,979.83 2178.76,984.631 2183.6,989.263 2188.43,993.721 2193.26,998 
  2198.1,1002.09 2202.93,1006 2207.76,1009.71 2212.59,1013.23 2217.43,1016.54 2222.26,1019.65 2227.09,1022.54 2231.93,1025.22 2236.76,1027.68 2241.59,1029.92 
  2246.43,1031.93 2251.26,1033.71 2256.09,1035.26 2260.93,1036.58 2265.76,1037.65 2270.59,1038.49 2275.43,1039.07 2280.26,1039.42 2285.09,1039.51 2289.93,1039.44 
  2294.76,1039.27 
  "/>
<polyline clip-path="url(#clip392)" style="stroke:#008000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" stroke-dasharray="4, 8" points="
  361.501,1365.41 366.334,1365.05 371.167,1364.23 376,1362.94 380.833,1361.18 385.666,1358.96 390.5,1356.27 395.333,1353.12 400.166,1349.51 404.999,1345.44 
  409.832,1340.91 414.665,1335.92 419.498,1330.49 424.332,1324.61 429.165,1318.29 433.998,1311.55 438.831,1304.37 443.664,1296.78 448.497,1288.78 453.33,1280.37 
  458.164,1271.58 462.997,1262.4 467.83,1252.85 472.663,1242.94 477.496,1232.67 482.329,1222.08 487.162,1211.15 491.996,1199.92 496.829,1188.38 501.662,1176.56 
  506.495,1164.46 511.328,1152.11 516.161,1139.52 520.994,1126.69 525.828,1113.66 530.661,1100.42 535.494,1087.01 540.327,1073.43 545.16,1059.7 549.993,1045.83 
  554.826,1031.85 559.66,1017.76 564.493,1003.59 569.326,989.341 574.159,975.039 578.992,960.698 583.825,946.333 588.658,931.959 593.492,917.592 598.325,903.248 
  603.158,888.94 607.991,874.686 612.824,860.497 617.657,846.39 622.49,832.378 627.324,818.474 632.157,804.691 636.99,791.043 641.823,777.541 646.656,764.199 
  651.489,751.029 656.322,738.041 661.156,725.248 665.989,712.659 670.822,700.286 675.655,688.138 680.488,676.225 685.321,664.555 690.154,653.139 694.988,641.983 
  699.821,631.098 704.654,620.49 709.487,610.167 714.32,600.136 719.153,590.405 723.986,580.98 728.82,571.867 733.653,563.072 738.486,554.601 743.319,546.459 
  748.152,538.65 752.985,531.179 757.818,524.049 762.652,517.267 767.485,510.836 772.318,504.763 777.151,499.05 781.984,493.703 786.817,488.723 791.65,484.115 
  796.484,479.881 801.317,476.023 806.15,472.543 810.983,469.442 815.816,466.723 820.649,464.384 825.482,462.427 830.316,460.851 835.149,459.656 839.982,458.841 
  844.815,458.404 849.648,458.345 854.481,458.66 859.314,459.347 864.148,460.404 868.981,461.827 873.814,463.613 878.647,465.76 883.48,468.264 888.313,471.121 
  893.147,474.325 897.98,477.871 902.813,481.753 907.646,485.966 912.479,490.504 917.312,495.36 922.145,500.527 926.979,505.999 931.812,511.768 936.645,517.827 
  941.478,524.168 946.311,530.783 951.144,537.664 955.977,544.802 960.811,552.187 965.644,559.812 970.477,567.669 975.31,575.75 980.143,584.044 984.976,592.543 
  989.809,601.237 994.643,610.117 999.476,619.174 1004.31,628.398 1009.14,637.78 1013.98,647.311 1018.81,656.982 1023.64,666.785 1028.47,676.71 1033.31,686.749 
  1038.14,696.892 1042.97,707.131 1047.81,717.457 1052.64,727.863 1057.47,738.338 1062.31,748.876 1067.14,759.466 1071.97,770.102 1076.81,780.774 1081.64,791.475 
  1086.47,802.197 1091.31,812.93 1096.14,823.668 1100.97,834.402 1105.8,845.124 1110.64,855.827 1115.47,866.501 1120.3,877.14 1125.14,887.735 1129.97,898.279 
  1134.8,908.762 1139.64,919.178 1144.47,929.518 1149.3,939.774 1154.14,949.937 1158.97,960 1163.8,969.954 1168.64,979.792 1173.47,989.504 1178.3,999.083 
  1183.14,1008.52 1187.97,1017.81 1192.8,1026.94 1197.63,1035.9 1202.47,1044.68 1207.3,1053.29 1212.13,1061.7 1216.97,1069.93 1221.8,1077.96 1226.63,1085.77 
  1231.47,1093.37 1236.3,1100.73 1241.13,1107.86 1245.97,1114.73 1250.8,1121.35 1255.63,1127.69 1260.47,1133.77 1265.3,1139.56 1270.13,1145.05 1274.96,1150.26 
  1279.8,1155.15 1284.63,1159.74 1289.46,1164.02 1294.3,1167.97 1299.13,1171.6 1303.96,1174.9 1308.8,1177.86 1313.63,1180.49 1318.46,1182.78 1323.3,1184.72 
  1328.13,1186.32 1332.96,1187.57 1337.8,1188.48 1342.63,1189.03 1347.46,1189.23 1352.3,1189.08 1357.13,1188.58 1361.96,1187.73 1366.79,1186.53 1371.63,1184.98 
  1376.46,1183.08 1381.29,1180.82 1386.13,1178.22 1390.96,1175.28 1395.79,1171.99 1400.63,1168.37 1405.46,1164.41 1410.29,1160.13 1415.13,1155.52 1419.96,1150.59 
  1424.79,1145.35 1429.63,1139.8 1434.46,1133.96 1439.29,1127.83 1444.12,1121.41 1448.96,1114.72 1453.79,1107.76 1458.62,1100.55 1463.46,1093.09 1468.29,1085.39 
  1473.12,1077.47 1477.96,1069.33 1482.79,1060.99 1487.62,1052.45 1492.46,1043.73 1497.29,1034.84 1502.12,1025.78 1506.96,1016.58 1511.79,1007.25 1516.62,997.79 
  1521.46,988.222 1526.29,978.555 1531.12,968.802 1535.95,958.975 1540.79,949.086 1545.62,939.146 1550.45,929.169 1555.29,919.165 1560.12,909.147 1564.95,899.126 
  1569.79,889.115 1574.62,879.124 1579.45,869.166 1584.29,859.251 1589.12,849.393 1593.95,839.601 1598.79,829.888 1603.62,820.266 1608.45,810.744 1613.28,801.335 
  1618.12,792.051 1622.95,782.898 1627.78,773.881 1632.62,765.01 1637.45,756.295 1642.28,747.742 1647.12,739.362 1651.95,731.161 1656.78,723.149 1661.62,715.331 
  1666.45,707.715 1671.28,700.308 1676.12,693.116 1680.95,686.145 1685.78,679.402 1690.62,672.891 1695.45,666.618 1700.28,660.587 1705.11,654.804 1709.95,649.272 
  1714.78,643.995 1719.61,638.977 1724.45,634.221 1729.28,629.73 1734.11,625.507 1738.95,621.553 1743.78,617.871 1748.61,614.462 1753.45,611.329 1758.28,608.473 
  1763.11,605.898 1767.95,603.604 1772.78,601.591 1777.61,599.862 1782.44,598.416 1787.28,597.253 1792.11,596.374 1796.94,595.777 1801.78,595.462 1806.61,595.427 
  1811.44,595.671 1816.28,596.192 1821.11,596.988 1825.94,598.056 1830.78,599.393 1835.61,600.996 1840.44,602.862 1845.28,604.986 1850.11,607.365 1854.94,609.993 
  1859.78,612.867 1864.61,615.98 1869.44,619.327 1874.27,622.903 1879.11,626.7 1883.94,630.714 1888.77,634.939 1893.61,639.373 1898.44,644.011 1903.27,648.844 
  1908.11,653.868 1912.94,659.076 1917.77,664.46 1922.61,670.016 1927.44,675.735 1932.27,681.612 1937.11,687.639 1941.94,693.809 1946.77,700.117 1951.6,706.554 
  1956.44,713.115 1961.27,719.791 1966.1,726.576 1970.94,733.462 1975.77,740.442 1980.6,747.509 1985.44,754.656 1990.27,761.875 1995.1,769.158 1999.94,776.499 
  2004.77,783.89 2009.6,791.326 2014.44,798.798 2019.27,806.299 2024.1,813.823 2028.94,821.361 2033.77,828.907 2038.6,836.453 2043.43,843.993 2048.27,851.519 
  2053.1,859.024 2057.93,866.502 2062.77,873.944 2067.6,881.345 2072.43,888.698 2077.27,895.995 2082.1,903.23 2086.93,910.396 2091.77,917.485 2096.6,924.493 
  2101.43,931.41 2106.27,938.232 2111.1,944.952 2115.93,951.562 2120.77,958.056 2125.6,964.428 2130.43,970.671 2135.26,976.779 2140.1,982.745 2144.93,988.566 
  2149.76,994.239 2154.6,999.754 2159.43,1005.11 2164.26,1010.29 2169.1,1015.29 2173.93,1020.12 2178.76,1024.75 2183.6,1029.19 2188.43,1033.44 2193.26,1037.48 
  2198.1,1041.31 2202.93,1044.92 2207.76,1048.32 2212.59,1051.5 2217.43,1054.46 2222.26,1057.18 2227.09,1059.68 2231.93,1061.95 2236.76,1063.97 2241.59,1065.76 
  2246.43,1067.31 2251.26,1068.62 2256.09,1069.69 2260.93,1070.51 2265.76,1071.09 2270.59,1071.43 2275.43,1071.51 2280.26,1071.36 2285.09,1070.96 2289.93,1070.31 
  2294.76,1069.42 
  "/>
<path clip-path="url(#clip390)" d="
M1421.89 477.116 L2284.45 477.116 L2284.45 166.076 L1421.89 166.076  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<polyline clip-path="url(#clip390)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1421.89,477.116 2284.45,477.116 2284.45,166.076 1421.89,166.076 1421.89,477.116 
  "/>
<polyline clip-path="url(#clip390)" style="stroke:#009af9; stroke-linecap:butt; stroke-linejoin:round; stroke-width:6; stroke-opacity:1; fill:none" points="
  1444.66,243.836 1581.28,243.836 
  "/>
<path clip-path="url(#clip390)" d="M1637.42 219.617 L1637.42 226.457 Q1633.43 224.548 1629.88 223.61 Q1626.34 222.673 1623.04 222.673 Q1617.31 222.673 1614.19 224.895 Q1611.1 227.117 1611.1 231.214 Q1611.1 234.652 1613.15 236.423 Q1615.23 238.159 1621 239.235 L1625.23 240.103 Q1633.08 241.596 1636.79 245.381 Q1640.54 249.131 1640.54 255.45 Q1640.54 262.985 1635.47 266.874 Q1630.44 270.763 1620.68 270.763 Q1617 270.763 1612.84 269.929 Q1608.7 269.096 1604.26 267.464 L1604.26 260.242 Q1608.53 262.638 1612.63 263.853 Q1616.72 265.068 1620.68 265.068 Q1626.69 265.068 1629.95 262.707 Q1633.22 260.346 1633.22 255.971 Q1633.22 252.152 1630.86 249.999 Q1628.53 247.846 1623.18 246.77 L1618.91 245.936 Q1611.06 244.374 1607.56 241.041 Q1604.05 237.707 1604.05 231.77 Q1604.05 224.895 1608.88 220.937 Q1613.74 216.978 1622.25 216.978 Q1625.89 216.978 1629.68 217.638 Q1633.46 218.298 1637.42 219.617 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M1651.2 230.867 L1657.59 230.867 L1657.59 269.756 L1651.2 269.756 L1651.2 230.867 M1651.2 215.728 L1657.59 215.728 L1657.59 223.818 L1651.2 223.818 L1651.2 215.728 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M1701.24 238.332 Q1703.63 234.027 1706.97 231.978 Q1710.3 229.93 1714.81 229.93 Q1720.89 229.93 1724.19 234.2 Q1727.49 238.436 1727.49 246.284 L1727.49 269.756 L1721.06 269.756 L1721.06 246.492 Q1721.06 240.902 1719.08 238.193 Q1717.11 235.485 1713.04 235.485 Q1708.08 235.485 1705.2 238.784 Q1702.31 242.082 1702.31 247.777 L1702.31 269.756 L1695.89 269.756 L1695.89 246.492 Q1695.89 240.867 1693.91 238.193 Q1691.93 235.485 1687.8 235.485 Q1682.9 235.485 1680.02 238.818 Q1677.14 242.117 1677.14 247.777 L1677.14 269.756 L1670.72 269.756 L1670.72 230.867 L1677.14 230.867 L1677.14 236.909 Q1679.33 233.332 1682.38 231.631 Q1685.44 229.93 1689.64 229.93 Q1693.88 229.93 1696.83 232.082 Q1699.81 234.235 1701.24 238.332 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M1746.41 263.922 L1746.41 284.547 L1739.99 284.547 L1739.99 230.867 L1746.41 230.867 L1746.41 236.77 Q1748.43 233.298 1751.48 231.631 Q1754.57 229.93 1758.84 229.93 Q1765.92 229.93 1770.33 235.555 Q1774.78 241.179 1774.78 250.346 Q1774.78 259.513 1770.33 265.138 Q1765.92 270.763 1758.84 270.763 Q1754.57 270.763 1751.48 269.096 Q1748.43 267.395 1746.41 263.922 M1768.15 250.346 Q1768.15 243.298 1765.23 239.304 Q1762.35 235.277 1757.28 235.277 Q1752.21 235.277 1749.29 239.304 Q1746.41 243.298 1746.41 250.346 Q1746.41 257.395 1749.29 261.422 Q1752.21 265.415 1757.28 265.415 Q1762.35 265.415 1765.23 261.422 Q1768.15 257.395 1768.15 250.346 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M1785.37 215.728 L1791.76 215.728 L1791.76 269.756 L1785.37 269.756 L1785.37 215.728 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M1838.39 248.714 L1838.39 251.839 L1809.01 251.839 Q1809.43 258.436 1812.97 261.909 Q1816.55 265.346 1822.9 265.346 Q1826.58 265.346 1830.02 264.443 Q1833.49 263.54 1836.9 261.735 L1836.9 267.777 Q1833.46 269.235 1829.85 269.999 Q1826.24 270.763 1822.52 270.763 Q1813.22 270.763 1807.76 265.346 Q1802.35 259.929 1802.35 250.693 Q1802.35 241.145 1807.49 235.555 Q1812.66 229.93 1821.41 229.93 Q1829.26 229.93 1833.81 234.999 Q1838.39 240.034 1838.39 248.714 M1832 246.839 Q1831.93 241.596 1829.05 238.471 Q1826.2 235.346 1821.48 235.346 Q1816.13 235.346 1812.9 238.367 Q1809.71 241.388 1809.22 246.874 L1832 246.839 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M1849.15 217.916 L1878.95 217.916 L1878.95 223.818 L1856.17 223.818 L1856.17 239.096 L1876.72 239.096 L1876.72 244.999 L1856.17 244.999 L1856.17 269.756 L1849.15 269.756 L1849.15 217.916 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M1890.06 217.916 L1900.51 217.916 L1913.74 253.193 L1927.04 217.916 L1937.49 217.916 L1937.49 269.756 L1930.65 269.756 L1930.65 224.235 L1917.28 259.79 L1910.23 259.79 L1896.86 224.235 L1896.86 269.756 L1890.06 269.756 L1890.06 217.916 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M1950.61 217.916 L1957.66 217.916 L1957.66 249.409 Q1957.66 257.742 1960.68 261.422 Q1963.7 265.068 1970.47 265.068 Q1977.21 265.068 1980.23 261.422 Q1983.25 257.742 1983.25 249.409 L1983.25 217.916 L1990.3 217.916 L1990.3 250.277 Q1990.3 260.415 1985.26 265.589 Q1980.26 270.763 1970.47 270.763 Q1960.65 270.763 1955.61 265.589 Q1950.61 260.415 1950.61 250.277 L1950.61 217.916 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip390)" style="stroke:#e26f46; stroke-linecap:butt; stroke-linejoin:round; stroke-width:6; stroke-opacity:1; fill:none" points="
  1444.66,321.596 1581.28,321.596 
  "/>
<path clip-path="url(#clip390)" d="M1628.63 323.21 Q1630.89 323.974 1633.01 326.474 Q1635.16 328.974 1637.31 333.349 L1644.43 347.516 L1636.9 347.516 L1630.27 334.217 Q1627.7 329.009 1625.27 327.307 Q1622.87 325.606 1618.7 325.606 L1611.06 325.606 L1611.06 347.516 L1604.05 347.516 L1604.05 295.676 L1619.88 295.676 Q1628.77 295.676 1633.15 299.391 Q1637.52 303.106 1637.52 310.606 Q1637.52 315.502 1635.23 318.731 Q1632.97 321.96 1628.63 323.21 M1611.06 301.44 L1611.06 319.842 L1619.88 319.842 Q1624.95 319.842 1627.52 317.516 Q1630.13 315.155 1630.13 310.606 Q1630.13 306.058 1627.52 303.766 Q1624.95 301.44 1619.88 301.44 L1611.06 301.44 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M1683.25 326.474 L1683.25 329.599 L1653.88 329.599 Q1654.29 336.196 1657.84 339.669 Q1661.41 343.106 1667.77 343.106 Q1671.45 343.106 1674.88 342.203 Q1678.36 341.3 1681.76 339.495 L1681.76 345.537 Q1678.32 346.995 1674.71 347.759 Q1671.1 348.523 1667.38 348.523 Q1658.08 348.523 1652.63 343.106 Q1647.21 337.689 1647.21 328.453 Q1647.21 318.905 1652.35 313.315 Q1657.52 307.69 1666.27 307.69 Q1674.12 307.69 1678.67 312.759 Q1683.25 317.794 1683.25 326.474 M1676.86 324.599 Q1676.79 319.356 1673.91 316.231 Q1671.06 313.106 1666.34 313.106 Q1660.99 313.106 1657.77 316.127 Q1654.57 319.148 1654.09 324.634 L1676.86 324.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M1713.43 293.488 L1713.43 298.801 L1707.31 298.801 Q1703.88 298.801 1702.52 300.19 Q1701.2 301.578 1701.2 305.19 L1701.2 308.627 L1711.72 308.627 L1711.72 313.592 L1701.2 313.592 L1701.2 347.516 L1694.78 347.516 L1694.78 313.592 L1688.67 313.592 L1688.67 308.627 L1694.78 308.627 L1694.78 305.919 Q1694.78 299.426 1697.8 296.474 Q1700.82 293.488 1707.38 293.488 L1713.43 293.488 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M1752.04 326.474 L1752.04 329.599 L1722.66 329.599 Q1723.08 336.196 1726.62 339.669 Q1730.2 343.106 1736.55 343.106 Q1740.23 343.106 1743.67 342.203 Q1747.14 341.3 1750.54 339.495 L1750.54 345.537 Q1747.11 346.995 1743.49 347.759 Q1739.88 348.523 1736.17 348.523 Q1726.86 348.523 1721.41 343.106 Q1715.99 337.689 1715.99 328.453 Q1715.99 318.905 1721.13 313.315 Q1726.31 307.69 1735.06 307.69 Q1742.9 307.69 1747.45 312.759 Q1752.04 317.794 1752.04 326.474 M1745.65 324.599 Q1745.58 319.356 1742.7 316.231 Q1739.85 313.106 1735.13 313.106 Q1729.78 313.106 1726.55 316.127 Q1723.36 319.148 1722.87 324.634 L1745.65 324.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M1785.06 314.599 Q1783.98 313.974 1782.7 313.696 Q1781.45 313.384 1779.92 313.384 Q1774.5 313.384 1771.58 316.926 Q1768.7 320.433 1768.7 327.03 L1768.7 347.516 L1762.28 347.516 L1762.28 308.627 L1768.7 308.627 L1768.7 314.669 Q1770.72 311.127 1773.95 309.426 Q1777.17 307.69 1781.79 307.69 Q1782.45 307.69 1783.25 307.794 Q1784.05 307.863 1785.02 308.037 L1785.06 314.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M1823.46 326.474 L1823.46 329.599 L1794.08 329.599 Q1794.5 336.196 1798.04 339.669 Q1801.62 343.106 1807.97 343.106 Q1811.65 343.106 1815.09 342.203 Q1818.56 341.3 1821.97 339.495 L1821.97 345.537 Q1818.53 346.995 1814.92 347.759 Q1811.31 348.523 1807.59 348.523 Q1798.29 348.523 1792.83 343.106 Q1787.42 337.689 1787.42 328.453 Q1787.42 318.905 1792.56 313.315 Q1797.73 307.69 1806.48 307.69 Q1814.33 307.69 1818.88 312.759 Q1823.46 317.794 1823.46 326.474 M1817.07 324.599 Q1817 319.356 1814.12 316.231 Q1811.27 313.106 1806.55 313.106 Q1801.2 313.106 1797.97 316.127 Q1794.78 319.148 1794.29 324.634 L1817.07 324.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M1866.27 324.044 L1866.27 347.516 L1859.88 347.516 L1859.88 324.252 Q1859.88 318.731 1857.73 315.988 Q1855.58 313.245 1851.27 313.245 Q1846.1 313.245 1843.11 316.544 Q1840.13 319.842 1840.13 325.537 L1840.13 347.516 L1833.7 347.516 L1833.7 308.627 L1840.13 308.627 L1840.13 314.669 Q1842.42 311.162 1845.51 309.426 Q1848.63 307.69 1852.7 307.69 Q1859.4 307.69 1862.83 311.856 Q1866.27 315.988 1866.27 324.044 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M1907 310.12 L1907 316.092 Q1904.29 314.599 1901.55 313.87 Q1898.84 313.106 1896.06 313.106 Q1889.85 313.106 1886.41 317.064 Q1882.97 320.988 1882.97 328.106 Q1882.97 335.224 1886.41 339.182 Q1889.85 343.106 1896.06 343.106 Q1898.84 343.106 1901.55 342.377 Q1904.29 341.613 1907 340.12 L1907 346.023 Q1904.33 347.273 1901.45 347.898 Q1898.6 348.523 1895.37 348.523 Q1886.58 348.523 1881.41 343.002 Q1876.24 337.481 1876.24 328.106 Q1876.24 318.592 1881.45 313.141 Q1886.69 307.69 1895.79 307.69 Q1898.74 307.69 1901.55 308.315 Q1904.36 308.905 1907 310.12 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M1951.38 326.474 L1951.38 329.599 L1922 329.599 Q1922.42 336.196 1925.96 339.669 Q1929.54 343.106 1935.89 343.106 Q1939.57 343.106 1943.01 342.203 Q1946.48 341.3 1949.88 339.495 L1949.88 345.537 Q1946.44 346.995 1942.83 347.759 Q1939.22 348.523 1935.51 348.523 Q1926.2 348.523 1920.75 343.106 Q1915.33 337.689 1915.33 328.453 Q1915.33 318.905 1920.47 313.315 Q1925.65 307.69 1934.4 307.69 Q1942.24 307.69 1946.79 312.759 Q1951.38 317.794 1951.38 326.474 M1944.99 324.599 Q1944.92 319.356 1942.04 316.231 Q1939.19 313.106 1934.47 313.106 Q1929.12 313.106 1925.89 316.127 Q1922.69 319.148 1922.21 324.634 L1944.99 324.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip390)" style="stroke:#008000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:6; stroke-opacity:1; fill:none" stroke-dasharray="3, 6" points="
  1444.66,399.356 1581.28,399.356 
  "/>
<path clip-path="url(#clip390)" d="M1604.05 373.436 L1613.5 373.436 L1636.48 416.804 L1636.48 373.436 L1643.29 373.436 L1643.29 425.276 L1633.84 425.276 L1610.86 381.908 L1610.86 425.276 L1604.05 425.276 L1604.05 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M1690.23 404.234 L1690.23 407.359 L1660.86 407.359 Q1661.27 413.956 1664.81 417.429 Q1668.39 420.866 1674.74 420.866 Q1678.43 420.866 1681.86 419.963 Q1685.34 419.06 1688.74 417.255 L1688.74 423.297 Q1685.3 424.755 1681.69 425.519 Q1678.08 426.283 1674.36 426.283 Q1665.06 426.283 1659.61 420.866 Q1654.19 415.449 1654.19 406.213 Q1654.19 396.665 1659.33 391.075 Q1664.5 385.45 1673.25 385.45 Q1681.1 385.45 1685.65 390.519 Q1690.23 395.554 1690.23 404.234 M1683.84 402.359 Q1683.77 397.116 1680.89 393.991 Q1678.04 390.866 1673.32 390.866 Q1667.97 390.866 1664.74 393.887 Q1661.55 396.908 1661.06 402.394 L1683.84 402.359 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M1700.06 409.929 L1700.06 386.387 L1706.45 386.387 L1706.45 409.686 Q1706.45 415.206 1708.6 417.984 Q1710.75 420.727 1715.06 420.727 Q1720.23 420.727 1723.22 417.429 Q1726.24 414.13 1726.24 408.436 L1726.24 386.387 L1732.63 386.387 L1732.63 425.276 L1726.24 425.276 L1726.24 419.304 Q1723.91 422.845 1720.82 424.581 Q1717.77 426.283 1713.7 426.283 Q1707 426.283 1703.53 422.116 Q1700.06 417.949 1700.06 409.929 M1716.13 385.45 L1716.13 385.45 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M1768.32 392.359 Q1767.24 391.734 1765.96 391.456 Q1764.71 391.144 1763.18 391.144 Q1757.77 391.144 1754.85 394.686 Q1751.97 398.193 1751.97 404.79 L1751.97 425.276 L1745.54 425.276 L1745.54 386.387 L1751.97 386.387 L1751.97 392.429 Q1753.98 388.887 1757.21 387.186 Q1760.44 385.45 1765.06 385.45 Q1765.72 385.45 1766.52 385.554 Q1767.31 385.623 1768.29 385.797 L1768.32 392.359 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M1792.7 405.727 Q1784.95 405.727 1781.97 407.498 Q1778.98 409.269 1778.98 413.54 Q1778.98 416.942 1781.2 418.956 Q1783.46 420.935 1787.31 420.935 Q1792.63 420.935 1795.82 417.185 Q1799.05 413.401 1799.05 407.151 L1799.05 405.727 L1792.7 405.727 M1805.44 403.088 L1805.44 425.276 L1799.05 425.276 L1799.05 419.373 Q1796.86 422.915 1793.6 424.616 Q1790.33 426.283 1785.61 426.283 Q1779.64 426.283 1776.1 422.949 Q1772.59 419.581 1772.59 413.956 Q1772.59 407.394 1776.97 404.061 Q1781.38 400.727 1790.09 400.727 L1799.05 400.727 L1799.05 400.102 Q1799.05 395.693 1796.13 393.297 Q1793.25 390.866 1788.01 390.866 Q1784.67 390.866 1781.52 391.665 Q1778.36 392.463 1775.44 394.061 L1775.44 388.158 Q1778.95 386.804 1782.24 386.144 Q1785.54 385.45 1788.67 385.45 Q1797.11 385.45 1801.27 389.825 Q1805.44 394.199 1805.44 403.088 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M1818.6 371.248 L1824.99 371.248 L1824.99 425.276 L1818.6 425.276 L1818.6 371.248 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M1838.63 373.436 L1868.42 373.436 L1868.42 379.338 L1845.65 379.338 L1845.65 394.616 L1866.2 394.616 L1866.2 400.519 L1845.65 400.519 L1845.65 425.276 L1838.63 425.276 L1838.63 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M1879.54 373.436 L1889.99 373.436 L1903.22 408.713 L1916.51 373.436 L1926.97 373.436 L1926.97 425.276 L1920.13 425.276 L1920.13 379.755 L1906.76 415.31 L1899.71 415.31 L1886.34 379.755 L1886.34 425.276 L1879.54 425.276 L1879.54 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M1940.09 373.436 L1947.14 373.436 L1947.14 404.929 Q1947.14 413.262 1950.16 416.942 Q1953.18 420.588 1959.95 420.588 Q1966.69 420.588 1969.71 416.942 Q1972.73 413.262 1972.73 404.929 L1972.73 373.436 L1979.78 373.436 L1979.78 405.797 Q1979.78 415.935 1974.74 421.109 Q1969.74 426.283 1959.95 426.283 Q1950.13 426.283 1945.09 421.109 Q1940.09 415.935 1940.09 405.797 L1940.09 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M2030.61 371.318 Q2025.96 379.304 2023.7 387.116 Q2021.44 394.929 2021.44 402.949 Q2021.44 410.97 2023.7 418.852 Q2025.99 426.699 2030.61 434.651 L2025.06 434.651 Q2019.85 426.491 2017.24 418.609 Q2014.67 410.727 2014.67 402.949 Q2014.67 395.206 2017.24 387.359 Q2019.81 379.512 2025.06 371.318 L2030.61 371.318 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M2049.95 419.373 L2074.43 419.373 L2074.43 425.276 L2041.51 425.276 L2041.51 419.373 Q2045.51 415.241 2052.38 408.297 Q2059.29 401.318 2061.06 399.304 Q2064.43 395.519 2065.75 392.915 Q2067.1 390.276 2067.1 387.741 Q2067.1 383.609 2064.19 381.005 Q2061.31 378.401 2056.65 378.401 Q2053.35 378.401 2049.67 379.547 Q2046.03 380.693 2041.86 383.019 L2041.86 375.936 Q2046.1 374.234 2049.78 373.366 Q2053.46 372.498 2056.51 372.498 Q2064.57 372.498 2069.36 376.526 Q2074.15 380.554 2074.15 387.29 Q2074.15 390.484 2072.94 393.366 Q2071.76 396.213 2068.6 400.102 Q2067.73 401.109 2063.08 405.936 Q2058.42 410.727 2049.95 419.373 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M2089.22 373.436 L2116.76 373.436 L2116.76 379.338 L2095.65 379.338 L2095.65 392.047 Q2097.17 391.526 2098.7 391.283 Q2100.23 391.005 2101.76 391.005 Q2110.44 391.005 2115.51 395.762 Q2120.58 400.519 2120.58 408.644 Q2120.58 417.012 2115.37 421.665 Q2110.16 426.283 2100.68 426.283 Q2097.42 426.283 2094.01 425.727 Q2090.65 425.172 2087.03 424.06 L2087.03 417.012 Q2090.16 418.713 2093.49 419.547 Q2096.83 420.38 2100.54 420.38 Q2106.55 420.38 2110.06 417.22 Q2113.56 414.06 2113.56 408.644 Q2113.56 403.227 2110.06 400.068 Q2106.55 396.908 2100.54 396.908 Q2097.73 396.908 2094.92 397.533 Q2092.14 398.158 2089.22 399.477 L2089.22 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M2149.4 378.054 Q2143.98 378.054 2141.24 383.401 Q2138.53 388.713 2138.53 399.408 Q2138.53 410.067 2141.24 415.415 Q2143.98 420.727 2149.4 420.727 Q2154.85 420.727 2157.55 415.415 Q2160.3 410.067 2160.3 399.408 Q2160.3 388.713 2157.55 383.401 Q2154.85 378.054 2149.4 378.054 M2149.4 372.498 Q2158.11 372.498 2162.69 379.408 Q2167.31 386.283 2167.31 399.408 Q2167.31 412.498 2162.69 419.408 Q2158.11 426.283 2149.4 426.283 Q2140.68 426.283 2136.06 419.408 Q2131.48 412.498 2131.48 399.408 Q2131.48 386.283 2136.06 379.408 Q2140.68 372.498 2149.4 372.498 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M2194.64 378.054 Q2189.22 378.054 2186.48 383.401 Q2183.77 388.713 2183.77 399.408 Q2183.77 410.067 2186.48 415.415 Q2189.22 420.727 2194.64 420.727 Q2200.09 420.727 2202.8 415.415 Q2205.54 410.067 2205.54 399.408 Q2205.54 388.713 2202.8 383.401 Q2200.09 378.054 2194.64 378.054 M2194.64 372.498 Q2203.35 372.498 2207.94 379.408 Q2212.55 386.283 2212.55 399.408 Q2212.55 412.498 2207.94 419.408 Q2203.35 426.283 2194.64 426.283 Q2185.92 426.283 2181.3 419.408 Q2176.72 412.498 2176.72 399.408 Q2176.72 386.283 2181.3 379.408 Q2185.92 372.498 2194.64 372.498 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip390)" d="M2222.97 371.318 L2228.53 371.318 Q2233.73 379.512 2236.3 387.359 Q2238.91 395.206 2238.91 402.949 Q2238.91 410.727 2236.3 418.609 Q2233.73 426.491 2228.53 434.651 L2222.97 434.651 Q2227.59 426.699 2229.85 418.852 Q2232.14 410.97 2232.14 402.949 Q2232.14 394.929 2229.85 387.116 Q2227.59 379.304 2222.97 371.318 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /></svg>




<?xml version="1.0" encoding="utf-8"?>
<svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" width="600" height="400" viewBox="0 0 2400 1600">
<defs>
  <clipPath id="clip480">
    <rect x="0" y="0" width="2400" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip480)" d="
M0 1600 L2400 1600 L2400 0 L0 0  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip481">
    <rect x="480" y="0" width="1681" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip480)" d="
M303.503 1401.58 L2352.76 1401.58 L2352.76 123.472 L303.503 123.472  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip482">
    <rect x="303" y="123" width="2050" height="1279"/>
  </clipPath>
</defs>
<polyline clip-path="url(#clip482)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  361.501,1401.58 361.501,123.472 
  "/>
<polyline clip-path="url(#clip482)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  844.815,1401.58 844.815,123.472 
  "/>
<polyline clip-path="url(#clip482)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1328.13,1401.58 1328.13,123.472 
  "/>
<polyline clip-path="url(#clip482)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1811.44,1401.58 1811.44,123.472 
  "/>
<polyline clip-path="url(#clip482)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  2294.76,1401.58 2294.76,123.472 
  "/>
<polyline clip-path="url(#clip480)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  303.503,1401.58 2352.76,1401.58 
  "/>
<polyline clip-path="url(#clip480)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  361.501,1401.58 361.501,1382.68 
  "/>
<polyline clip-path="url(#clip480)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  844.815,1401.58 844.815,1382.68 
  "/>
<polyline clip-path="url(#clip480)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1328.13,1401.58 1328.13,1382.68 
  "/>
<polyline clip-path="url(#clip480)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1811.44,1401.58 1811.44,1382.68 
  "/>
<polyline clip-path="url(#clip480)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  2294.76,1401.58 2294.76,1382.68 
  "/>
<path clip-path="url(#clip480)" d="M361.501 1438.36 Q356.084 1438.36 353.341 1443.71 Q350.633 1449.02 350.633 1459.71 Q350.633 1470.37 353.341 1475.72 Q356.084 1481.03 361.501 1481.03 Q366.952 1481.03 369.66 1475.72 Q372.403 1470.37 372.403 1459.71 Q372.403 1449.02 369.66 1443.71 Q366.952 1438.36 361.501 1438.36 M361.501 1432.8 Q370.216 1432.8 374.799 1439.71 Q379.417 1446.59 379.417 1459.71 Q379.417 1472.8 374.799 1479.71 Q370.216 1486.59 361.501 1486.59 Q352.785 1486.59 348.167 1479.71 Q343.584 1472.8 343.584 1459.71 Q343.584 1446.59 348.167 1439.71 Q352.785 1432.8 361.501 1432.8 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M830.388 1479.68 L841.846 1479.68 L841.846 1440.13 L829.381 1442.63 L829.381 1436.24 L841.777 1433.74 L848.791 1433.74 L848.791 1479.68 L860.249 1479.68 L860.249 1485.58 L830.388 1485.58 L830.388 1479.68 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M1320.11 1479.68 L1344.59 1479.68 L1344.59 1485.58 L1311.67 1485.58 L1311.67 1479.68 Q1315.66 1475.55 1322.54 1468.6 Q1329.45 1461.62 1331.22 1459.61 Q1334.59 1455.82 1335.91 1453.22 Q1337.26 1450.58 1337.26 1448.05 Q1337.26 1443.91 1334.34 1441.31 Q1331.46 1438.71 1326.81 1438.71 Q1323.51 1438.71 1319.83 1439.85 Q1316.19 1441 1312.02 1443.32 L1312.02 1436.24 Q1316.25 1434.54 1319.94 1433.67 Q1323.62 1432.8 1326.67 1432.8 Q1334.73 1432.8 1339.52 1436.83 Q1344.31 1440.86 1344.31 1447.59 Q1344.31 1450.79 1343.09 1453.67 Q1341.91 1456.52 1338.75 1460.41 Q1337.89 1461.41 1333.23 1466.24 Q1328.58 1471.03 1320.11 1479.68 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M1817.82 1457.63 Q1822.85 1458.7 1825.66 1462.11 Q1828.51 1465.51 1828.51 1470.51 Q1828.51 1478.18 1823.23 1482.39 Q1817.95 1486.59 1808.23 1486.59 Q1804.97 1486.59 1801.5 1485.93 Q1798.06 1485.3 1794.38 1484.02 L1794.38 1477.25 Q1797.29 1478.95 1800.77 1479.82 Q1804.24 1480.68 1808.02 1480.68 Q1814.62 1480.68 1818.06 1478.08 Q1821.53 1475.48 1821.53 1470.51 Q1821.53 1465.93 1818.3 1463.36 Q1815.11 1460.75 1809.38 1460.75 L1803.34 1460.75 L1803.34 1454.99 L1809.66 1454.99 Q1814.83 1454.99 1817.57 1452.94 Q1820.32 1450.86 1820.32 1446.97 Q1820.32 1442.98 1817.47 1440.86 Q1814.66 1438.71 1809.38 1438.71 Q1806.5 1438.71 1803.2 1439.33 Q1799.9 1439.96 1795.94 1441.27 L1795.94 1435.02 Q1799.93 1433.91 1803.41 1433.36 Q1806.91 1432.8 1810 1432.8 Q1817.99 1432.8 1822.64 1436.45 Q1827.29 1440.06 1827.29 1446.24 Q1827.29 1450.55 1824.83 1453.53 Q1822.36 1456.48 1817.82 1457.63 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M2299.27 1439.85 L2281.56 1467.52 L2299.27 1467.52 L2299.27 1439.85 M2297.43 1433.74 L2306.25 1433.74 L2306.25 1467.52 L2313.65 1467.52 L2313.65 1473.36 L2306.25 1473.36 L2306.25 1485.58 L2299.27 1485.58 L2299.27 1473.36 L2275.87 1473.36 L2275.87 1466.59 L2297.43 1433.74 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M1271.74 1527.62 L1271.74 1538.66 L1284.9 1538.66 L1284.9 1543.62 L1271.74 1543.62 L1271.74 1564.74 Q1271.74 1569.49 1273.03 1570.85 Q1274.35 1572.2 1278.34 1572.2 L1284.9 1572.2 L1284.9 1577.55 L1278.34 1577.55 Q1270.94 1577.55 1268.13 1574.8 Q1265.32 1572.03 1265.32 1564.74 L1265.32 1543.62 L1260.63 1543.62 L1260.63 1538.66 L1265.32 1538.66 L1265.32 1527.62 L1271.74 1527.62 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M1315.32 1523.52 L1330.04 1523.52 L1330.04 1528.49 L1321.71 1528.49 L1321.71 1581.96 L1330.04 1581.96 L1330.04 1586.92 L1315.32 1586.92 L1315.32 1523.52 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M1368.44 1539.8 L1368.44 1545.85 Q1365.73 1544.46 1362.82 1543.76 Q1359.9 1543.07 1356.78 1543.07 Q1352.02 1543.07 1349.62 1544.53 Q1347.26 1545.99 1347.26 1548.9 Q1347.26 1551.12 1348.96 1552.41 Q1350.66 1553.66 1355.8 1554.8 L1357.99 1555.29 Q1364.8 1556.75 1367.64 1559.42 Q1370.53 1562.06 1370.53 1566.82 Q1370.53 1572.24 1366.22 1575.4 Q1361.95 1578.55 1354.45 1578.55 Q1351.32 1578.55 1347.92 1577.93 Q1344.55 1577.34 1340.8 1576.12 L1340.8 1569.53 Q1344.34 1571.37 1347.78 1572.3 Q1351.22 1573.21 1354.59 1573.21 Q1359.1 1573.21 1361.53 1571.68 Q1363.96 1570.12 1363.96 1567.3 Q1363.96 1564.7 1362.19 1563.31 Q1360.46 1561.92 1354.52 1560.64 L1352.3 1560.12 Q1346.36 1558.87 1343.72 1556.3 Q1341.08 1553.69 1341.08 1549.18 Q1341.08 1543.69 1344.97 1540.71 Q1348.86 1537.72 1356.01 1537.72 Q1359.55 1537.72 1362.68 1538.24 Q1365.8 1538.76 1368.44 1539.8 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M1395.63 1523.52 L1395.63 1586.92 L1380.91 1586.92 L1380.91 1581.96 L1389.21 1581.96 L1389.21 1528.49 L1380.91 1528.49 L1380.91 1523.52 L1395.63 1523.52 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip482)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  303.503,1365.41 2352.76,1365.41 
  "/>
<polyline clip-path="url(#clip482)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  303.503,1087.5 2352.76,1087.5 
  "/>
<polyline clip-path="url(#clip482)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  303.503,809.586 2352.76,809.586 
  "/>
<polyline clip-path="url(#clip482)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  303.503,531.676 2352.76,531.676 
  "/>
<polyline clip-path="url(#clip482)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  303.503,253.766 2352.76,253.766 
  "/>
<polyline clip-path="url(#clip480)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  303.503,1401.58 303.503,123.472 
  "/>
<polyline clip-path="url(#clip480)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  303.503,1365.41 322.401,1365.41 
  "/>
<polyline clip-path="url(#clip480)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  303.503,1087.5 322.401,1087.5 
  "/>
<polyline clip-path="url(#clip480)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  303.503,809.586 322.401,809.586 
  "/>
<polyline clip-path="url(#clip480)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  303.503,531.676 322.401,531.676 
  "/>
<polyline clip-path="url(#clip480)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  303.503,253.766 322.401,253.766 
  "/>
<path clip-path="url(#clip480)" d="M136.497 1344.11 Q131.08 1344.11 128.337 1349.45 Q125.629 1354.76 125.629 1365.46 Q125.629 1376.12 128.337 1381.47 Q131.08 1386.78 136.497 1386.78 Q141.948 1386.78 144.656 1381.47 Q147.399 1376.12 147.399 1365.46 Q147.399 1354.76 144.656 1349.45 Q141.948 1344.11 136.497 1344.11 M136.497 1338.55 Q145.212 1338.55 149.795 1345.46 Q154.413 1352.33 154.413 1365.46 Q154.413 1378.55 149.795 1385.46 Q145.212 1392.33 136.497 1392.33 Q127.781 1392.33 123.163 1385.46 Q118.58 1378.55 118.58 1365.46 Q118.58 1352.33 123.163 1345.46 Q127.781 1338.55 136.497 1338.55 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M166.74 1382.51 L174.066 1382.51 L174.066 1391.33 L166.74 1391.33 L166.74 1382.51 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M189.413 1339.49 L216.948 1339.49 L216.948 1345.39 L195.837 1345.39 L195.837 1358.1 Q197.364 1357.58 198.892 1357.33 Q200.42 1357.06 201.948 1357.06 Q210.628 1357.06 215.698 1361.81 Q220.767 1366.57 220.767 1374.7 Q220.767 1383.06 215.559 1387.72 Q210.35 1392.33 200.871 1392.33 Q197.608 1392.33 194.205 1391.78 Q190.837 1391.22 187.226 1390.11 L187.226 1383.06 Q190.351 1384.76 193.684 1385.6 Q197.017 1386.43 200.732 1386.43 Q206.739 1386.43 210.246 1383.27 Q213.753 1380.11 213.753 1374.7 Q213.753 1369.28 210.246 1366.12 Q206.739 1362.96 200.732 1362.96 Q197.92 1362.96 195.108 1363.58 Q192.33 1364.21 189.413 1365.53 L189.413 1339.49 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M249.586 1344.11 Q244.17 1344.11 241.427 1349.45 Q238.718 1354.76 238.718 1365.46 Q238.718 1376.12 241.427 1381.47 Q244.17 1386.78 249.586 1386.78 Q255.038 1386.78 257.746 1381.47 Q260.489 1376.12 260.489 1365.46 Q260.489 1354.76 257.746 1349.45 Q255.038 1344.11 249.586 1344.11 M249.586 1338.55 Q258.302 1338.55 262.885 1345.46 Q267.503 1352.33 267.503 1365.46 Q267.503 1378.55 262.885 1385.46 Q258.302 1392.33 249.586 1392.33 Q240.871 1392.33 236.253 1385.46 Q231.67 1378.55 231.67 1365.46 Q231.67 1352.33 236.253 1345.46 Q240.871 1338.55 249.586 1338.55 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M137.99 1066.19 Q132.573 1066.19 129.83 1071.54 Q127.122 1076.85 127.122 1087.55 Q127.122 1098.21 129.83 1103.56 Q132.573 1108.87 137.99 1108.87 Q143.441 1108.87 146.149 1103.56 Q148.892 1098.21 148.892 1087.55 Q148.892 1076.85 146.149 1071.54 Q143.441 1066.19 137.99 1066.19 M137.99 1060.64 Q146.705 1060.64 151.288 1067.55 Q155.906 1074.42 155.906 1087.55 Q155.906 1100.64 151.288 1107.55 Q146.705 1114.42 137.99 1114.42 Q129.275 1114.42 124.657 1107.55 Q120.073 1100.64 120.073 1087.55 Q120.073 1074.42 124.657 1067.55 Q129.275 1060.64 137.99 1060.64 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M168.233 1104.6 L175.559 1104.6 L175.559 1113.42 L168.233 1113.42 L168.233 1104.6 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M189.066 1061.58 L222.399 1061.58 L222.399 1064.56 L203.58 1113.42 L196.253 1113.42 L213.962 1067.48 L189.066 1067.48 L189.066 1061.58 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M236.149 1061.58 L263.684 1061.58 L263.684 1067.48 L242.573 1067.48 L242.573 1080.19 Q244.1 1079.67 245.628 1079.42 Q247.156 1079.15 248.684 1079.15 Q257.364 1079.15 262.434 1083.9 Q267.503 1088.66 267.503 1096.78 Q267.503 1105.15 262.295 1109.81 Q257.086 1114.42 247.607 1114.42 Q244.343 1114.42 240.941 1113.87 Q237.573 1113.31 233.961 1112.2 L233.961 1105.15 Q237.086 1106.85 240.42 1107.69 Q243.753 1108.52 247.468 1108.52 Q253.475 1108.52 256.982 1105.36 Q260.489 1102.2 260.489 1096.78 Q260.489 1091.37 256.982 1088.21 Q253.475 1085.05 247.468 1085.05 Q244.656 1085.05 241.843 1085.67 Q239.066 1086.3 236.149 1087.62 L236.149 1061.58 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M122.712 829.604 L134.17 829.604 L134.17 790.055 L121.705 792.555 L121.705 786.166 L134.101 783.666 L141.115 783.666 L141.115 829.604 L152.573 829.604 L152.573 835.506 L122.712 835.506 L122.712 829.604 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M166.74 826.687 L174.066 826.687 L174.066 835.506 L166.74 835.506 L166.74 826.687 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M204.344 788.284 Q198.927 788.284 196.184 793.631 Q193.476 798.944 193.476 809.638 Q193.476 820.298 196.184 825.645 Q198.927 830.958 204.344 830.958 Q209.795 830.958 212.503 825.645 Q215.246 820.298 215.246 809.638 Q215.246 798.944 212.503 793.631 Q209.795 788.284 204.344 788.284 M204.344 782.729 Q213.059 782.729 217.642 789.638 Q222.26 796.513 222.26 809.638 Q222.26 822.729 217.642 829.638 Q213.059 836.513 204.344 836.513 Q195.628 836.513 191.01 829.638 Q186.427 822.729 186.427 809.638 Q186.427 796.513 191.01 789.638 Q195.628 782.729 204.344 782.729 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M249.586 788.284 Q244.17 788.284 241.427 793.631 Q238.718 798.944 238.718 809.638 Q238.718 820.298 241.427 825.645 Q244.17 830.958 249.586 830.958 Q255.038 830.958 257.746 825.645 Q260.489 820.298 260.489 809.638 Q260.489 798.944 257.746 793.631 Q255.038 788.284 249.586 788.284 M249.586 782.729 Q258.302 782.729 262.885 789.638 Q267.503 796.513 267.503 809.638 Q267.503 822.729 262.885 829.638 Q258.302 836.513 249.586 836.513 Q240.871 836.513 236.253 829.638 Q231.67 822.729 231.67 809.638 Q231.67 796.513 236.253 789.638 Q240.871 782.729 249.586 782.729 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M124.205 551.693 L135.663 551.693 L135.663 512.145 L123.198 514.645 L123.198 508.256 L135.594 505.756 L142.608 505.756 L142.608 551.693 L154.066 551.693 L154.066 557.596 L124.205 557.596 L124.205 551.693 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M168.233 548.776 L175.559 548.776 L175.559 557.596 L168.233 557.596 L168.233 548.776 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M196.878 551.693 L221.357 551.693 L221.357 557.596 L188.441 557.596 L188.441 551.693 Q192.434 547.561 199.309 540.617 Q206.219 533.638 207.989 531.624 Q211.357 527.839 212.677 525.235 Q214.031 522.596 214.031 520.061 Q214.031 515.929 211.114 513.325 Q208.232 510.721 203.58 510.721 Q200.281 510.721 196.601 511.867 Q192.955 513.013 188.788 515.339 L188.788 508.256 Q193.024 506.554 196.705 505.686 Q200.385 504.818 203.441 504.818 Q211.496 504.818 216.288 508.846 Q221.08 512.874 221.08 519.61 Q221.08 522.804 219.864 525.686 Q218.684 528.534 215.524 532.422 Q214.656 533.429 210.003 538.256 Q205.351 543.047 196.878 551.693 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M236.149 505.756 L263.684 505.756 L263.684 511.659 L242.573 511.659 L242.573 524.367 Q244.1 523.846 245.628 523.603 Q247.156 523.325 248.684 523.325 Q257.364 523.325 262.434 528.082 Q267.503 532.839 267.503 540.964 Q267.503 549.332 262.295 553.985 Q257.086 558.603 247.607 558.603 Q244.343 558.603 240.941 558.047 Q237.573 557.492 233.961 556.381 L233.961 549.332 Q237.086 551.033 240.42 551.867 Q243.753 552.7 247.468 552.7 Q253.475 552.7 256.982 549.54 Q260.489 546.381 260.489 540.964 Q260.489 535.547 256.982 532.388 Q253.475 529.228 247.468 529.228 Q244.656 529.228 241.843 529.853 Q239.066 530.478 236.149 531.797 L236.149 505.756 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M122.712 273.783 L134.17 273.783 L134.17 234.234 L121.705 236.734 L121.705 230.346 L134.101 227.846 L141.115 227.846 L141.115 273.783 L152.573 273.783 L152.573 279.686 L122.712 279.686 L122.712 273.783 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M166.74 270.866 L174.066 270.866 L174.066 279.686 L166.74 279.686 L166.74 270.866 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M189.413 227.846 L216.948 227.846 L216.948 233.748 L195.837 233.748 L195.837 246.457 Q197.364 245.936 198.892 245.693 Q200.42 245.415 201.948 245.415 Q210.628 245.415 215.698 250.172 Q220.767 254.929 220.767 263.054 Q220.767 271.422 215.559 276.074 Q210.35 280.692 200.871 280.692 Q197.608 280.692 194.205 280.137 Q190.837 279.581 187.226 278.47 L187.226 271.422 Q190.351 273.123 193.684 273.956 Q197.017 274.79 200.732 274.79 Q206.739 274.79 210.246 271.63 Q213.753 268.47 213.753 263.054 Q213.753 257.637 210.246 254.477 Q206.739 251.318 200.732 251.318 Q197.92 251.318 195.108 251.943 Q192.33 252.568 189.413 253.887 L189.413 227.846 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M249.586 232.464 Q244.17 232.464 241.427 237.811 Q238.718 243.123 238.718 253.818 Q238.718 264.477 241.427 269.824 Q244.17 275.137 249.586 275.137 Q255.038 275.137 257.746 269.824 Q260.489 264.477 260.489 253.818 Q260.489 243.123 257.746 237.811 Q255.038 232.464 249.586 232.464 M249.586 226.908 Q258.302 226.908 262.885 233.818 Q267.503 240.693 267.503 253.818 Q267.503 266.908 262.885 273.817 Q258.302 280.692 249.586 280.692 Q240.871 280.692 236.253 273.817 Q231.67 266.908 231.67 253.818 Q231.67 240.693 236.253 233.818 Q240.871 226.908 249.586 226.908 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M35.6047 1045.01 Q31.2992 1042.61 29.2506 1039.28 Q27.202 1035.94 27.202 1031.43 Q27.202 1025.35 31.4728 1022.06 Q35.7089 1018.76 43.5561 1018.76 L67.0282 1018.76 L67.0282 1025.18 L43.7644 1025.18 Q38.1741 1025.18 35.4658 1027.16 Q32.7575 1029.14 32.7575 1033.2 Q32.7575 1038.17 36.0561 1041.05 Q39.3547 1043.93 45.0491 1043.93 L67.0282 1043.93 L67.0282 1050.35 L43.7644 1050.35 Q38.1394 1050.35 35.4658 1052.33 Q32.7575 1054.31 32.7575 1058.44 Q32.7575 1063.34 36.0908 1066.22 Q39.3894 1069.1 45.0491 1069.1 L67.0282 1069.1 L67.0282 1075.53 L28.1395 1075.53 L28.1395 1069.1 L34.1811 1069.1 Q30.6047 1066.92 28.9034 1063.86 Q27.202 1060.81 27.202 1056.6 Q27.202 1052.37 29.3547 1049.42 Q31.5075 1046.43 35.6047 1045.01 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M47.4797 988.341 Q47.4797 996.084 49.2505 999.07 Q51.0213 1002.06 55.2921 1002.06 Q58.6949 1002.06 60.7087 999.834 Q62.6879 997.577 62.6879 993.723 Q62.6879 988.41 58.9379 985.216 Q55.1532 981.987 48.9033 981.987 L47.4797 981.987 L47.4797 988.341 M44.8408 975.598 L67.0282 975.598 L67.0282 981.987 L61.1254 981.987 Q64.6671 984.174 66.3684 987.438 Q68.0351 990.702 68.0351 995.424 Q68.0351 1001.4 64.7018 1004.94 Q61.3337 1008.44 55.7088 1008.44 Q49.1463 1008.44 45.813 1004.07 Q42.4797 999.66 42.4797 990.945 L42.4797 981.987 L41.8547 981.987 Q37.445 981.987 35.0492 984.903 Q32.6186 987.785 32.6186 993.028 Q32.6186 996.362 33.4172 999.521 Q34.2158 1002.68 35.813 1005.6 L29.9103 1005.6 Q28.5561 1002.09 27.8964 998.792 Q27.202 995.493 27.202 992.369 Q27.202 983.931 31.577 979.764 Q35.9519 975.598 44.8408 975.598 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M29.2853 937.647 L35.3269 937.647 Q33.9381 940.355 33.2436 943.272 Q32.5492 946.188 32.5492 949.313 Q32.5492 954.07 34.0075 956.466 Q35.4658 958.827 38.3825 958.827 Q40.6047 958.827 41.8894 957.126 Q43.1394 955.424 44.2852 950.285 L44.7713 948.098 Q46.2297 941.292 48.9033 938.445 Q51.5421 935.563 56.299 935.563 Q61.7157 935.563 64.8754 939.869 Q68.0351 944.14 68.0351 951.64 Q68.0351 954.765 67.4101 958.167 Q66.8198 961.535 65.6046 965.285 L59.0074 965.285 Q60.8476 961.744 61.7851 958.306 Q62.6879 954.869 62.6879 951.501 Q62.6879 946.987 61.1601 944.556 Q59.5976 942.126 56.7852 942.126 Q54.181 942.126 52.7921 943.897 Q51.4032 945.633 50.1185 951.57 L49.5977 953.792 Q48.3477 959.73 45.7783 962.369 Q43.1741 965.008 38.6603 965.008 Q33.1742 965.008 30.1881 961.119 Q27.202 957.23 27.202 950.077 Q27.202 946.535 27.7228 943.41 Q28.2436 940.285 29.2853 937.647 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M29.2853 900.598 L35.3269 900.598 Q33.9381 903.306 33.2436 906.223 Q32.5492 909.14 32.5492 912.265 Q32.5492 917.022 34.0075 919.418 Q35.4658 921.779 38.3825 921.779 Q40.6047 921.779 41.8894 920.077 Q43.1394 918.376 44.2852 913.237 L44.7713 911.049 Q46.2297 904.244 48.9033 901.397 Q51.5421 898.515 56.299 898.515 Q61.7157 898.515 64.8754 902.82 Q68.0351 907.091 68.0351 914.591 Q68.0351 917.716 67.4101 921.119 Q66.8198 924.487 65.6046 928.237 L59.0074 928.237 Q60.8476 924.695 61.7851 921.258 Q62.6879 917.82 62.6879 914.452 Q62.6879 909.938 61.1601 907.508 Q59.5976 905.077 56.7852 905.077 Q54.181 905.077 52.7921 906.848 Q51.4032 908.584 50.1185 914.522 L49.5977 916.744 Q48.3477 922.681 45.7783 925.32 Q43.1741 927.959 38.6603 927.959 Q33.1742 927.959 30.1881 924.07 Q27.202 920.181 27.202 913.029 Q27.202 909.487 27.7228 906.362 Q28.2436 903.237 29.2853 900.598 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M61.1949 859.557 L81.8197 859.557 L81.8197 865.98 L28.1395 865.98 L28.1395 859.557 L34.0422 859.557 Q30.57 857.543 28.9034 854.487 Q27.202 851.397 27.202 847.126 Q27.202 840.043 32.827 835.633 Q38.4519 831.189 47.6185 831.189 Q56.7852 831.189 62.4101 835.633 Q68.0351 840.043 68.0351 847.126 Q68.0351 851.397 66.3684 854.487 Q64.6671 857.543 61.1949 859.557 M47.6185 837.821 Q40.57 837.821 36.5769 840.737 Q32.5492 843.619 32.5492 848.689 Q32.5492 853.758 36.5769 856.675 Q40.57 859.557 47.6185 859.557 Q54.6671 859.557 58.6949 856.675 Q62.6879 853.758 62.6879 848.689 Q62.6879 843.619 58.6949 840.737 Q54.6671 837.821 47.6185 837.821 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M32.6186 805.529 Q32.6186 810.668 36.6464 813.654 Q40.6394 816.64 47.6185 816.64 Q54.5977 816.64 58.6254 813.689 Q62.6185 810.703 62.6185 805.529 Q62.6185 800.425 58.5907 797.439 Q54.5629 794.453 47.6185 794.453 Q40.7089 794.453 36.6811 797.439 Q32.6186 800.425 32.6186 805.529 M27.202 805.529 Q27.202 797.196 32.6186 792.439 Q38.0353 787.682 47.6185 787.682 Q57.1671 787.682 62.6185 792.439 Q68.0351 797.196 68.0351 805.529 Q68.0351 813.897 62.6185 818.654 Q57.1671 823.376 47.6185 823.376 Q38.0353 823.376 32.6186 818.654 Q27.202 813.897 27.202 805.529 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M29.2853 752.3 L35.3269 752.3 Q33.9381 755.009 33.2436 757.925 Q32.5492 760.842 32.5492 763.967 Q32.5492 768.724 34.0075 771.12 Q35.4658 773.481 38.3825 773.481 Q40.6047 773.481 41.8894 771.779 Q43.1394 770.078 44.2852 764.939 L44.7713 762.752 Q46.2297 755.946 48.9033 753.099 Q51.5421 750.217 56.299 750.217 Q61.7157 750.217 64.8754 754.523 Q68.0351 758.793 68.0351 766.293 Q68.0351 769.418 67.4101 772.821 Q66.8198 776.189 65.6046 779.939 L59.0074 779.939 Q60.8476 776.397 61.7851 772.96 Q62.6879 769.522 62.6879 766.154 Q62.6879 761.641 61.1601 759.21 Q59.5976 756.779 56.7852 756.779 Q54.181 756.779 52.7921 758.55 Q51.4032 760.286 50.1185 766.224 L49.5977 768.446 Q48.3477 774.384 45.7783 777.022 Q43.1741 779.661 38.6603 779.661 Q33.1742 779.661 30.1881 775.772 Q27.202 771.884 27.202 764.731 Q27.202 761.189 27.7228 758.064 Q28.2436 754.939 29.2853 752.3 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M28.1395 740.043 L28.1395 733.655 L67.0282 733.655 L67.0282 740.043 L28.1395 740.043 M13.0007 740.043 L13.0007 733.655 L21.0909 733.655 L21.0909 740.043 L13.0007 740.043 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M17.0979 713.967 L28.1395 713.967 L28.1395 700.808 L33.1047 700.808 L33.1047 713.967 L54.2157 713.967 Q58.9726 713.967 60.3268 712.683 Q61.681 711.363 61.681 707.37 L61.681 700.808 L67.0282 700.808 L67.0282 707.37 Q67.0282 714.766 64.2851 717.578 Q61.5074 720.391 54.2157 720.391 L33.1047 720.391 L33.1047 725.078 L28.1395 725.078 L28.1395 720.391 L17.0979 720.391 L17.0979 713.967 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M28.1395 692.405 L28.1395 686.016 L67.0282 686.016 L67.0282 692.405 L28.1395 692.405 M13.0007 692.405 L13.0007 686.016 L21.0909 686.016 L21.0909 692.405 L13.0007 692.405 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M32.6186 657.579 Q32.6186 662.717 36.6464 665.704 Q40.6394 668.69 47.6185 668.69 Q54.5977 668.69 58.6254 665.738 Q62.6185 662.752 62.6185 657.579 Q62.6185 652.474 58.5907 649.488 Q54.5629 646.502 47.6185 646.502 Q40.7089 646.502 36.6811 649.488 Q32.6186 652.474 32.6186 657.579 M27.202 657.579 Q27.202 649.245 32.6186 644.488 Q38.0353 639.732 47.6185 639.732 Q57.1671 639.732 62.6185 644.488 Q68.0351 649.245 68.0351 657.579 Q68.0351 665.947 62.6185 670.704 Q57.1671 675.426 47.6185 675.426 Q38.0353 675.426 32.6186 670.704 Q27.202 665.947 27.202 657.579 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M43.5561 596.815 L67.0282 596.815 L67.0282 603.204 L43.7644 603.204 Q38.2436 603.204 35.5005 605.357 Q32.7575 607.509 32.7575 611.815 Q32.7575 616.989 36.0561 619.975 Q39.3547 622.961 45.0491 622.961 L67.0282 622.961 L67.0282 629.384 L28.1395 629.384 L28.1395 622.961 L34.1811 622.961 Q30.6742 620.669 28.9381 617.579 Q27.202 614.454 27.202 610.391 Q27.202 603.69 31.3686 600.253 Q35.5005 596.815 43.5561 596.815 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M13.0007 562.058 L13.0007 547.336 L17.9659 547.336 L17.9659 555.669 L71.4379 555.669 L71.4379 547.336 L76.4031 547.336 L76.4031 562.058 L13.0007 562.058 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M35.6047 503.448 Q31.2992 501.052 29.2506 497.718 Q27.202 494.385 27.202 489.871 Q27.202 483.795 31.4728 480.496 Q35.7089 477.198 43.5561 477.198 L67.0282 477.198 L67.0282 483.621 L43.7644 483.621 Q38.1741 483.621 35.4658 485.6 Q32.7575 487.58 32.7575 491.642 Q32.7575 496.607 36.0561 499.489 Q39.3547 502.371 45.0491 502.371 L67.0282 502.371 L67.0282 508.795 L43.7644 508.795 Q38.1394 508.795 35.4658 510.774 Q32.7575 512.753 32.7575 516.885 Q32.7575 521.781 36.0908 524.663 Q39.3894 527.545 45.0491 527.545 L67.0282 527.545 L67.0282 533.968 L28.1395 533.968 L28.1395 527.545 L34.1811 527.545 Q30.6047 525.357 28.9034 522.302 Q27.202 519.246 27.202 515.045 Q27.202 510.809 29.3547 507.857 Q31.5075 504.871 35.6047 503.448 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M13.0007 449.524 L76.4031 449.524 L76.4031 464.246 L71.4379 464.246 L71.4379 455.948 L17.9659 455.948 L17.9659 464.246 L13.0007 464.246 L13.0007 449.524 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M592.343 12.096 L604.537 12.096 L619.971 53.2532 L635.486 12.096 L647.679 12.096 L647.679 72.576 L639.699 72.576 L639.699 19.4686 L624.103 60.9499 L615.879 60.9499 L600.283 19.4686 L600.283 72.576 L592.343 72.576 L592.343 12.096 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M681.18 32.4315 Q675.184 32.4315 671.701 37.1306 Q668.217 41.7891 668.217 49.9314 Q668.217 58.0738 671.66 62.7728 Q675.144 67.4314 681.18 67.4314 Q687.135 67.4314 690.618 62.7323 Q694.102 58.0333 694.102 49.9314 Q694.102 41.8701 690.618 37.1711 Q687.135 32.4315 681.18 32.4315 M681.18 26.1121 Q690.902 26.1121 696.452 32.4315 Q702.001 38.7509 702.001 49.9314 Q702.001 61.0714 696.452 67.4314 Q690.902 73.7508 681.18 73.7508 Q671.417 73.7508 665.867 67.4314 Q660.358 61.0714 660.358 49.9314 Q660.358 38.7509 665.867 32.4315 Q671.417 26.1121 681.18 26.1121 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M744.212 34.0924 L744.212 9.54393 L751.666 9.54393 L751.666 72.576 L744.212 72.576 L744.212 65.7705 Q741.862 69.8214 738.257 71.8063 Q734.692 73.7508 729.669 73.7508 Q721.446 73.7508 716.261 67.1883 Q711.116 60.6258 711.116 49.9314 Q711.116 39.2371 716.261 32.6746 Q721.446 26.1121 729.669 26.1121 Q734.692 26.1121 738.257 28.0971 Q741.862 30.0415 744.212 34.0924 M718.813 49.9314 Q718.813 58.1548 722.175 62.8538 Q725.578 67.5124 731.492 67.5124 Q737.406 67.5124 740.809 62.8538 Q744.212 58.1548 744.212 49.9314 Q744.212 41.7081 740.809 37.0496 Q737.406 32.3505 731.492 32.3505 Q725.578 32.3505 722.175 37.0496 Q718.813 41.7081 718.813 49.9314 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M767.018 27.2059 L774.472 27.2059 L774.472 72.576 L767.018 72.576 L767.018 27.2059 M767.018 9.54393 L774.472 9.54393 L774.472 18.9825 L767.018 18.9825 L767.018 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M813.037 9.54393 L813.037 15.7418 L805.907 15.7418 Q801.897 15.7418 800.317 17.3622 Q798.778 18.9825 798.778 23.1955 L798.778 27.2059 L811.052 27.2059 L811.052 32.9987 L798.778 32.9987 L798.778 72.576 L791.283 72.576 L791.283 32.9987 L784.154 32.9987 L784.154 27.2059 L791.283 27.2059 L791.283 24.0462 Q791.283 16.471 794.808 13.0277 Q798.332 9.54393 805.988 9.54393 L813.037 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M819.275 27.2059 L826.729 27.2059 L826.729 72.576 L819.275 72.576 L819.275 27.2059 M819.275 9.54393 L826.729 9.54393 L826.729 18.9825 L819.275 18.9825 L819.275 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M881.132 48.0275 L881.132 51.6733 L846.862 51.6733 Q847.348 59.3701 851.48 63.421 Q855.652 67.4314 863.065 67.4314 Q867.359 67.4314 871.37 66.3781 Q875.421 65.3249 879.391 63.2184 L879.391 70.267 Q875.38 71.9684 871.167 72.8596 Q866.954 73.7508 862.62 73.7508 Q851.763 73.7508 845.403 67.4314 Q839.084 61.1119 839.084 50.3365 Q839.084 39.1965 845.079 32.6746 Q851.115 26.1121 861.323 26.1121 Q870.479 26.1121 875.785 32.0264 Q881.132 37.9003 881.132 48.0275 M873.679 45.84 Q873.598 39.7232 870.235 36.0774 Q866.914 32.4315 861.405 32.4315 Q855.166 32.4315 851.399 35.9558 Q847.672 39.4801 847.105 45.8805 L873.679 45.84 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M923.221 34.0924 L923.221 9.54393 L930.675 9.54393 L930.675 72.576 L923.221 72.576 L923.221 65.7705 Q920.872 69.8214 917.266 71.8063 Q913.702 73.7508 908.679 73.7508 Q900.455 73.7508 895.27 67.1883 Q890.125 60.6258 890.125 49.9314 Q890.125 39.2371 895.27 32.6746 Q900.455 26.1121 908.679 26.1121 Q913.702 26.1121 917.266 28.0971 Q920.872 30.0415 923.221 34.0924 M897.822 49.9314 Q897.822 58.1548 901.184 62.8538 Q904.587 67.5124 910.501 67.5124 Q916.416 67.5124 919.819 62.8538 Q923.221 58.1548 923.221 49.9314 Q923.221 41.7081 919.819 37.0496 Q916.416 32.3505 910.501 32.3505 Q904.587 32.3505 901.184 37.0496 Q897.822 41.7081 897.822 49.9314 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M947.932 62.2867 L956.479 62.2867 L956.479 72.576 L947.932 72.576 L947.932 62.2867 M947.932 29.6769 L956.479 29.6769 L956.479 39.9662 L947.932 39.9662 L947.932 29.6769 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M1000.67 12.096 L1012.87 12.096 L1028.3 53.2532 L1043.82 12.096 L1056.01 12.096 L1056.01 72.576 L1048.03 72.576 L1048.03 19.4686 L1032.43 60.9499 L1024.21 60.9499 L1008.61 19.4686 L1008.61 72.576 L1000.67 72.576 L1000.67 12.096 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M1092.55 49.7694 Q1083.52 49.7694 1080.03 51.8354 Q1076.55 53.9013 1076.55 58.8839 Q1076.55 62.8538 1079.14 65.2034 Q1081.77 67.5124 1086.27 67.5124 Q1092.47 67.5124 1096.19 63.1374 Q1099.96 58.7219 1099.96 51.4303 L1099.96 49.7694 L1092.55 49.7694 M1107.42 46.6907 L1107.42 72.576 L1099.96 72.576 L1099.96 65.6895 Q1097.41 69.8214 1093.6 71.8063 Q1089.79 73.7508 1084.29 73.7508 Q1077.32 73.7508 1073.19 69.8619 Q1069.09 65.9325 1069.09 59.3701 Q1069.09 51.7138 1074.2 47.825 Q1079.34 43.9361 1089.51 43.9361 L1099.96 43.9361 L1099.96 43.2069 Q1099.96 38.0623 1096.56 35.2672 Q1093.2 32.4315 1087.08 32.4315 Q1083.19 32.4315 1079.51 33.3632 Q1075.82 34.295 1072.42 36.1584 L1072.42 29.2718 Q1076.51 27.692 1080.36 26.9223 Q1084.2 26.1121 1087.85 26.1121 Q1097.69 26.1121 1102.55 31.2163 Q1107.42 36.3204 1107.42 46.6907 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M1151.69 28.5427 L1151.69 35.5912 Q1148.53 33.9709 1145.13 33.1607 Q1141.73 32.3505 1138.08 32.3505 Q1132.53 32.3505 1129.74 34.0519 Q1126.98 35.7533 1126.98 39.156 Q1126.98 41.7486 1128.97 43.2475 Q1130.95 44.7058 1136.95 46.0426 L1139.5 46.6097 Q1147.44 48.3111 1150.76 51.4303 Q1154.12 54.509 1154.12 60.0587 Q1154.12 66.3781 1149.1 70.0644 Q1144.12 73.7508 1135.37 73.7508 Q1131.72 73.7508 1127.75 73.0216 Q1123.82 72.3329 1119.45 70.9151 L1119.45 63.2184 Q1123.58 65.3654 1127.59 66.4591 Q1131.6 67.5124 1135.53 67.5124 Q1140.8 67.5124 1143.63 65.73 Q1146.47 63.9071 1146.47 60.6258 Q1146.47 57.5877 1144.4 55.9673 Q1142.38 54.3469 1135.45 52.8481 L1132.86 52.2405 Q1125.93 50.7821 1122.85 47.7845 Q1119.77 44.7463 1119.77 39.4801 Q1119.77 33.0797 1124.31 29.5959 Q1128.85 26.1121 1137.19 26.1121 Q1141.32 26.1121 1144.97 26.7198 Q1148.61 27.3274 1151.69 28.5427 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M1194.92 28.5427 L1194.92 35.5912 Q1191.76 33.9709 1188.35 33.1607 Q1184.95 32.3505 1181.3 32.3505 Q1175.75 32.3505 1172.96 34.0519 Q1170.2 35.7533 1170.2 39.156 Q1170.2 41.7486 1172.19 43.2475 Q1174.17 44.7058 1180.17 46.0426 L1182.72 46.6097 Q1190.66 48.3111 1193.98 51.4303 Q1197.35 54.509 1197.35 60.0587 Q1197.35 66.3781 1192.32 70.0644 Q1187.34 73.7508 1178.59 73.7508 Q1174.94 73.7508 1170.97 73.0216 Q1167.05 72.3329 1162.67 70.9151 L1162.67 63.2184 Q1166.8 65.3654 1170.81 66.4591 Q1174.82 67.5124 1178.75 67.5124 Q1184.02 67.5124 1186.85 65.73 Q1189.69 63.9071 1189.69 60.6258 Q1189.69 57.5877 1187.62 55.9673 Q1185.6 54.3469 1178.67 52.8481 L1176.08 52.2405 Q1169.15 50.7821 1166.07 47.7845 Q1162.99 44.7463 1162.99 39.4801 Q1162.99 33.0797 1167.53 29.5959 Q1172.07 26.1121 1180.41 26.1121 Q1184.55 26.1121 1188.19 26.7198 Q1191.84 27.3274 1194.92 28.5427 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M1242.8 65.7705 L1242.8 89.8329 L1235.3 89.8329 L1235.3 27.2059 L1242.8 27.2059 L1242.8 34.0924 Q1245.15 30.0415 1248.71 28.0971 Q1252.32 26.1121 1257.3 26.1121 Q1265.56 26.1121 1270.71 32.6746 Q1275.89 39.2371 1275.89 49.9314 Q1275.89 60.6258 1270.71 67.1883 Q1265.56 73.7508 1257.3 73.7508 Q1252.32 73.7508 1248.71 71.8063 Q1245.15 69.8214 1242.8 65.7705 M1268.16 49.9314 Q1268.16 41.7081 1264.75 37.0496 Q1261.39 32.3505 1255.48 32.3505 Q1249.56 32.3505 1246.16 37.0496 Q1242.8 41.7081 1242.8 49.9314 Q1242.8 58.1548 1246.16 62.8538 Q1249.56 67.5124 1255.48 67.5124 Q1261.39 67.5124 1264.75 62.8538 Q1268.16 58.1548 1268.16 49.9314 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M1305.83 32.4315 Q1299.83 32.4315 1296.35 37.1306 Q1292.87 41.7891 1292.87 49.9314 Q1292.87 58.0738 1296.31 62.7728 Q1299.79 67.4314 1305.83 67.4314 Q1311.78 67.4314 1315.27 62.7323 Q1318.75 58.0333 1318.75 49.9314 Q1318.75 41.8701 1315.27 37.1711 Q1311.78 32.4315 1305.83 32.4315 M1305.83 26.1121 Q1315.55 26.1121 1321.1 32.4315 Q1326.65 38.7509 1326.65 49.9314 Q1326.65 61.0714 1321.1 67.4314 Q1315.55 73.7508 1305.83 73.7508 Q1296.07 73.7508 1290.52 67.4314 Q1285.01 61.0714 1285.01 49.9314 Q1285.01 38.7509 1290.52 32.4315 Q1296.07 26.1121 1305.83 26.1121 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M1367.93 28.5427 L1367.93 35.5912 Q1364.77 33.9709 1361.37 33.1607 Q1357.96 32.3505 1354.32 32.3505 Q1348.77 32.3505 1345.97 34.0519 Q1343.22 35.7533 1343.22 39.156 Q1343.22 41.7486 1345.2 43.2475 Q1347.19 44.7058 1353.18 46.0426 L1355.74 46.6097 Q1363.68 48.3111 1367 51.4303 Q1370.36 54.509 1370.36 60.0587 Q1370.36 66.3781 1365.34 70.0644 Q1360.35 73.7508 1351.6 73.7508 Q1347.96 73.7508 1343.99 73.0216 Q1340.06 72.3329 1335.68 70.9151 L1335.68 63.2184 Q1339.82 65.3654 1343.83 66.4591 Q1347.84 67.5124 1351.77 67.5124 Q1357.03 67.5124 1359.87 65.73 Q1362.7 63.9071 1362.7 60.6258 Q1362.7 57.5877 1360.64 55.9673 Q1358.61 54.3469 1351.69 52.8481 L1349.09 52.2405 Q1342.17 50.7821 1339.09 47.7845 Q1336.01 44.7463 1336.01 39.4801 Q1336.01 33.0797 1340.55 29.5959 Q1345.08 26.1121 1353.43 26.1121 Q1357.56 26.1121 1361.21 26.7198 Q1364.85 27.3274 1367.93 28.5427 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M1382.23 27.2059 L1389.68 27.2059 L1389.68 72.576 L1382.23 72.576 L1382.23 27.2059 M1382.23 9.54393 L1389.68 9.54393 L1389.68 18.9825 L1382.23 18.9825 L1382.23 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M1412.65 14.324 L1412.65 27.2059 L1428 27.2059 L1428 32.9987 L1412.65 32.9987 L1412.65 57.6282 Q1412.65 63.1779 1414.15 64.7578 Q1415.69 66.3376 1420.35 66.3376 L1428 66.3376 L1428 72.576 L1420.35 72.576 Q1411.72 72.576 1408.44 69.3758 Q1405.16 66.1351 1405.16 57.6282 L1405.16 32.9987 L1399.69 32.9987 L1399.69 27.2059 L1405.16 27.2059 L1405.16 14.324 L1412.65 14.324 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M1437.81 27.2059 L1445.26 27.2059 L1445.26 72.576 L1437.81 72.576 L1437.81 27.2059 M1437.81 9.54393 L1445.26 9.54393 L1445.26 18.9825 L1437.81 18.9825 L1437.81 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M1478.44 32.4315 Q1472.44 32.4315 1468.96 37.1306 Q1465.48 41.7891 1465.48 49.9314 Q1465.48 58.0738 1468.92 62.7728 Q1472.4 67.4314 1478.44 67.4314 Q1484.39 67.4314 1487.88 62.7323 Q1491.36 58.0333 1491.36 49.9314 Q1491.36 41.8701 1487.88 37.1711 Q1484.39 32.4315 1478.44 32.4315 M1478.44 26.1121 Q1488.16 26.1121 1493.71 32.4315 Q1499.26 38.7509 1499.26 49.9314 Q1499.26 61.0714 1493.71 67.4314 Q1488.16 73.7508 1478.44 73.7508 Q1468.68 73.7508 1463.13 67.4314 Q1457.62 61.0714 1457.62 49.9314 Q1457.62 38.7509 1463.13 32.4315 Q1468.68 26.1121 1478.44 26.1121 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M1549.33 45.1919 L1549.33 72.576 L1541.88 72.576 L1541.88 45.4349 Q1541.88 38.994 1539.36 35.7938 Q1536.85 32.5936 1531.83 32.5936 Q1525.79 32.5936 1522.31 36.4419 Q1518.83 40.2903 1518.83 46.9338 L1518.83 72.576 L1511.33 72.576 L1511.33 27.2059 L1518.83 27.2059 L1518.83 34.2544 Q1521.5 30.163 1525.1 28.1376 Q1528.75 26.1121 1533.49 26.1121 Q1541.31 26.1121 1545.32 30.9732 Q1549.33 35.7938 1549.33 45.1919 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M1611.19 49.7694 Q1602.15 49.7694 1598.67 51.8354 Q1595.19 53.9013 1595.19 58.8839 Q1595.19 62.8538 1597.78 65.2034 Q1600.41 67.5124 1604.91 67.5124 Q1611.11 67.5124 1614.83 63.1374 Q1618.6 58.7219 1618.6 51.4303 L1618.6 49.7694 L1611.19 49.7694 M1626.05 46.6907 L1626.05 72.576 L1618.6 72.576 L1618.6 65.6895 Q1616.05 69.8214 1612.24 71.8063 Q1608.43 73.7508 1602.92 73.7508 Q1595.95 73.7508 1591.82 69.8619 Q1587.73 65.9325 1587.73 59.3701 Q1587.73 51.7138 1592.84 47.825 Q1597.98 43.9361 1608.15 43.9361 L1618.6 43.9361 L1618.6 43.2069 Q1618.6 38.0623 1615.2 35.2672 Q1611.83 32.4315 1605.72 32.4315 Q1601.83 32.4315 1598.14 33.3632 Q1594.46 34.295 1591.05 36.1584 L1591.05 29.2718 Q1595.14 27.692 1598.99 26.9223 Q1602.84 26.1121 1606.49 26.1121 Q1616.33 26.1121 1621.19 31.2163 Q1626.05 36.3204 1626.05 46.6907 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M1664.37 9.54393 L1664.37 15.7418 L1657.25 15.7418 Q1653.23 15.7418 1651.65 17.3622 Q1650.12 18.9825 1650.12 23.1955 L1650.12 27.2059 L1662.39 27.2059 L1662.39 32.9987 L1650.12 32.9987 L1650.12 72.576 L1642.62 72.576 L1642.62 32.9987 L1635.49 32.9987 L1635.49 27.2059 L1642.62 27.2059 L1642.62 24.0462 Q1642.62 16.471 1646.15 13.0277 Q1649.67 9.54393 1657.33 9.54393 L1664.37 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M1676.53 14.324 L1676.53 27.2059 L1691.88 27.2059 L1691.88 32.9987 L1676.53 32.9987 L1676.53 57.6282 Q1676.53 63.1779 1678.03 64.7578 Q1679.57 66.3376 1684.22 66.3376 L1691.88 66.3376 L1691.88 72.576 L1684.22 72.576 Q1675.6 72.576 1672.31 69.3758 Q1669.03 66.1351 1669.03 57.6282 L1669.03 32.9987 L1663.56 32.9987 L1663.56 27.2059 L1669.03 27.2059 L1669.03 14.324 L1676.53 14.324 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M1740.49 48.0275 L1740.49 51.6733 L1706.22 51.6733 Q1706.71 59.3701 1710.84 63.421 Q1715.01 67.4314 1722.42 67.4314 Q1726.72 67.4314 1730.73 66.3781 Q1734.78 65.3249 1738.75 63.2184 L1738.75 70.267 Q1734.74 71.9684 1730.53 72.8596 Q1726.31 73.7508 1721.98 73.7508 Q1711.12 73.7508 1704.76 67.4314 Q1698.44 61.1119 1698.44 50.3365 Q1698.44 39.1965 1704.44 32.6746 Q1710.47 26.1121 1720.68 26.1121 Q1729.84 26.1121 1735.14 32.0264 Q1740.49 37.9003 1740.49 48.0275 M1733.04 45.84 Q1732.96 39.7232 1729.59 36.0774 Q1726.27 32.4315 1720.76 32.4315 Q1714.52 32.4315 1710.76 35.9558 Q1707.03 39.4801 1706.46 45.8805 L1733.04 45.84 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M1779.02 34.1734 Q1777.76 33.4443 1776.26 33.1202 Q1774.8 32.7556 1773.02 32.7556 Q1766.7 32.7556 1763.3 36.8875 Q1759.94 40.9789 1759.94 48.6757 L1759.94 72.576 L1752.44 72.576 L1752.44 27.2059 L1759.94 27.2059 L1759.94 34.2544 Q1762.29 30.1225 1766.05 28.1376 Q1769.82 26.1121 1775.21 26.1121 Q1775.98 26.1121 1776.91 26.2337 Q1777.84 26.3147 1778.97 26.5172 L1779.02 34.1734 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M1842.21 44.2197 Q1844.84 45.1109 1847.31 48.0275 Q1849.83 50.9442 1852.34 56.0483 L1860.64 72.576 L1851.85 72.576 L1844.11 57.061 Q1841.12 50.9847 1838.28 48.9997 Q1835.48 47.0148 1830.62 47.0148 L1821.71 47.0148 L1821.71 72.576 L1813.53 72.576 L1813.53 12.096 L1832 12.096 Q1842.37 12.096 1847.48 16.4305 Q1852.58 20.7649 1852.58 29.5149 Q1852.58 35.2267 1849.91 38.994 Q1847.27 42.7613 1842.21 44.2197 M1821.71 18.8205 L1821.71 40.2903 L1832 40.2903 Q1837.92 40.2903 1840.91 37.5762 Q1843.95 34.8216 1843.95 29.5149 Q1843.95 24.2082 1840.91 21.5346 Q1837.92 18.8205 1832 18.8205 L1821.71 18.8205 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M1866.35 54.671 L1866.35 27.2059 L1873.81 27.2059 L1873.81 54.3874 Q1873.81 60.8284 1876.32 64.0691 Q1878.83 67.2693 1883.85 67.2693 Q1889.89 67.2693 1893.37 63.421 Q1896.9 59.5726 1896.9 52.9291 L1896.9 27.2059 L1904.35 27.2059 L1904.35 72.576 L1896.9 72.576 L1896.9 65.6084 Q1894.18 69.7404 1890.58 71.7658 Q1887.01 73.7508 1882.27 73.7508 Q1874.45 73.7508 1870.4 68.8897 Q1866.35 64.0286 1866.35 54.671 M1885.11 26.1121 L1885.11 26.1121 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M1957.42 45.1919 L1957.42 72.576 L1949.96 72.576 L1949.96 45.4349 Q1949.96 38.994 1947.45 35.7938 Q1944.94 32.5936 1939.92 32.5936 Q1933.88 32.5936 1930.4 36.4419 Q1926.91 40.2903 1926.91 46.9338 L1926.91 72.576 L1919.42 72.576 L1919.42 27.2059 L1926.91 27.2059 L1926.91 34.2544 Q1929.59 30.163 1933.19 28.1376 Q1936.84 26.1121 1941.58 26.1121 Q1949.4 26.1121 1953.41 30.9732 Q1957.42 35.7938 1957.42 45.1919 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M1974.19 62.2867 L1982.74 62.2867 L1982.74 72.576 L1974.19 72.576 L1974.19 62.2867 M1974.19 29.6769 L1982.74 29.6769 L1982.74 39.9662 L1974.19 39.9662 L1974.19 29.6769 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M2029.08 65.6895 L2042.45 65.6895 L2042.45 19.5497 L2027.9 22.4663 L2027.9 15.0127 L2042.36 12.096 L2050.55 12.096 L2050.55 65.6895 L2063.92 65.6895 L2063.92 72.576 L2029.08 72.576 L2029.08 65.6895 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip482)" style="stroke:#009af9; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" points="
  361.501,809.586 366.334,826.199 371.167,842.685 376,859.024 380.833,875.202 385.666,891.202 390.5,907.009 395.333,922.606 400.166,937.982 404.999,953.118 
  409.832,968.003 414.665,982.619 419.498,996.952 424.332,1010.99 429.165,1024.71 433.998,1038.11 438.831,1051.17 443.664,1063.87 448.497,1076.21 453.33,1088.17 
  458.164,1099.75 462.997,1110.92 467.83,1121.68 472.663,1132.02 477.496,1141.92 482.329,1151.38 487.162,1160.39 491.996,1168.93 496.829,1177 501.662,1184.6 
  506.495,1191.71 511.328,1198.32 516.161,1204.44 520.994,1210.05 525.828,1215.14 530.661,1219.72 535.494,1223.78 540.327,1227.32 545.16,1230.32 549.993,1232.8 
  554.826,1234.73 559.66,1236.14 564.493,1237 569.326,1237.33 574.159,1237.12 578.992,1236.36 583.825,1235.07 588.658,1233.25 593.492,1230.89 598.325,1228 
  603.158,1224.57 607.991,1220.62 612.824,1216.15 617.657,1211.16 622.49,1205.66 627.324,1199.65 632.157,1193.14 636.99,1186.14 641.823,1178.64 646.656,1170.67 
  651.489,1162.23 656.322,1153.32 661.156,1143.95 665.989,1134.14 670.822,1123.9 675.655,1113.23 680.488,1102.14 685.321,1090.66 690.154,1078.77 694.988,1066.51 
  699.821,1053.88 704.654,1040.9 709.487,1027.57 714.32,1013.92 719.153,999.944 723.986,985.671 728.82,971.11 733.653,956.277 738.486,941.186 743.319,925.852 
  748.152,910.291 752.985,894.518 757.818,878.548 762.652,862.398 767.485,846.085 772.318,829.624 777.151,813.031 781.984,796.324 786.817,779.518 791.65,762.631 
  796.484,745.68 801.317,728.682 806.15,711.654 810.983,694.613 815.816,677.575 820.649,660.558 825.482,643.579 830.316,626.655 835.149,609.803 839.982,593.039 
  844.815,576.382 849.648,559.846 854.481,543.449 859.314,527.207 864.148,511.137 868.981,495.254 873.814,479.574 878.647,464.114 883.48,448.887 888.313,433.911 
  893.147,419.199 897.98,404.767 902.813,390.628 907.646,376.797 912.479,363.288 917.312,350.114 922.145,337.289 926.979,324.825 931.812,312.734 936.645,301.029 
  941.478,289.722 946.311,278.823 951.144,268.343 955.977,258.295 960.811,248.686 965.644,239.527 970.477,230.827 975.31,222.594 980.143,214.837 984.976,207.564 
  989.809,200.782 994.643,194.498 999.476,188.717 1004.31,183.447 1009.14,178.691 1013.98,174.455 1018.81,170.742 1023.64,167.558 1028.47,164.905 1033.31,162.785 
  1038.14,161.201 1042.97,160.154 1047.81,159.645 1052.64,159.674 1057.47,160.243 1062.31,161.351 1067.14,162.995 1071.97,165.175 1076.81,167.888 1081.64,171.131 
  1086.47,174.9 1091.31,179.195 1096.14,184.009 1100.97,189.337 1105.8,195.174 1110.64,201.514 1115.47,208.35 1120.3,215.676 1125.14,223.486 1129.97,231.771 
  1134.8,240.522 1139.64,249.732 1144.47,259.389 1149.3,269.486 1154.14,280.01 1158.97,290.954 1163.8,302.305 1168.64,314.053 1173.47,326.185 1178.3,338.689 
  1183.14,351.552 1187.97,364.761 1192.8,378.305 1197.63,392.169 1202.47,406.339 1207.3,420.801 1212.13,435.541 1216.97,450.543 1221.8,465.793 1226.63,481.275 
  1231.47,496.975 1236.3,512.877 1241.13,528.964 1245.97,545.22 1250.8,561.629 1255.63,578.174 1260.47,594.84 1265.3,611.609 1270.13,628.466 1274.96,645.392 
  1279.8,662.371 1284.63,679.386 1289.46,696.42 1294.3,713.455 1299.13,730.476 1303.96,747.464 1308.8,764.404 1313.63,781.277 1318.46,798.068 1323.3,814.758 
  1328.13,831.332 1332.96,847.773 1337.8,864.065 1342.63,880.191 1347.46,896.135 1352.3,911.882 1357.13,927.415 1361.96,942.718 1366.79,957.778 1371.63,972.578 
  1376.46,987.103 1381.29,1001.34 1386.13,1015.27 1390.96,1028.89 1395.79,1042.18 1400.63,1055.12 1405.46,1067.71 1410.29,1079.92 1415.13,1091.76 1419.96,1103.2 
  1424.79,1114.24 1429.63,1124.86 1434.46,1135.05 1439.29,1144.81 1444.12,1154.12 1448.96,1162.97 1453.79,1171.37 1458.62,1179.28 1463.46,1186.72 1468.29,1193.67 
  1473.12,1200.12 1477.96,1206.07 1482.79,1211.52 1487.62,1216.45 1492.46,1220.86 1497.29,1224.75 1502.12,1228.12 1506.96,1230.95 1511.79,1233.25 1516.62,1235.02 
  1521.46,1236.25 1526.29,1236.94 1531.12,1237.09 1535.95,1236.71 1540.79,1235.79 1545.62,1234.33 1550.45,1232.33 1555.29,1229.8 1560.12,1226.74 1564.95,1223.15 
  1569.79,1219.03 1574.62,1214.4 1579.45,1209.25 1584.29,1203.58 1589.12,1197.42 1593.95,1190.75 1598.79,1183.59 1603.62,1175.95 1608.45,1167.83 1613.28,1159.24 
  1618.12,1150.19 1622.95,1140.68 1627.78,1130.74 1632.62,1120.36 1637.45,1109.56 1642.28,1098.35 1647.12,1086.74 1651.95,1074.74 1656.78,1062.37 1661.62,1049.63 
  1666.45,1036.54 1671.28,1023.11 1676.12,1009.36 1680.95,995.296 1685.78,980.936 1690.62,966.293 1695.45,951.383 1700.28,936.22 1705.11,920.819 1709.95,905.197 
  1714.78,889.366 1719.61,873.345 1724.45,857.149 1729.28,840.795 1734.11,824.299 1738.95,807.677 1743.78,790.946 1748.61,774.122 1753.45,757.222 1758.28,740.263 
  1763.11,723.263 1767.95,706.239 1772.78,689.206 1777.61,672.184 1782.44,655.187 1787.28,638.233 1792.11,621.34 1796.94,604.524 1801.78,587.802 1806.61,571.19 
  1811.44,554.707 1816.28,538.366 1821.11,522.186 1825.94,506.183 1830.78,490.371 1835.61,474.768 1840.44,459.388 1845.28,444.248 1850.11,429.362 1854.94,414.745 
  1859.78,400.412 1864.61,386.377 1869.44,372.654 1874.27,359.256 1879.11,346.198 1883.94,333.492 1888.77,321.152 1893.61,309.189 1898.44,297.615 1903.27,286.442 
  1908.11,275.681 1912.94,265.342 1917.77,255.437 1922.61,245.975 1927.44,236.965 1932.27,228.418 1937.11,220.34 1941.94,212.74 1946.77,205.625 1951.6,199.003 
  1956.44,192.882 1961.27,187.266 1966.1,182.161 1970.94,177.572 1975.77,173.504 1980.6,169.961 1985.44,166.945 1990.27,164.463 1995.1,162.514 1999.94,161.101 
  2004.77,160.226 2009.6,159.888 2014.44,160.089 2019.27,160.827 2024.1,162.104 2028.94,163.917 2033.77,166.266 2038.6,169.146 2043.43,172.555 2048.27,176.489 
  2053.1,180.945 2057.93,185.92 2062.77,191.407 2067.6,197.401 2072.43,203.896 2077.27,210.885 2082.1,218.362 2086.93,226.318 2091.77,234.746 2096.6,243.639 
  2101.43,252.987 2106.27,262.78 2111.1,273.008 2115.93,283.662 2120.77,294.729 2125.6,306.2 2130.43,318.065 2135.26,330.31 2140.1,342.922 2144.93,355.89 
  2149.76,369.2 2154.6,382.839 2159.43,396.794 2164.26,411.05 2169.1,425.595 2173.93,440.411 2178.76,455.486 2183.6,470.803 2188.43,486.348 2193.26,502.104 
  2198.1,518.057 2202.93,534.191 2207.76,550.489 2212.59,566.934 2217.43,583.511 2222.26,600.202 2227.09,616.991 2231.93,633.863 2236.76,650.798 2241.59,667.782 
  2246.43,684.795 2251.26,701.823 2256.09,718.846 2260.93,735.849 2265.76,752.814 2270.59,769.726 2275.43,786.565 2280.26,803.317 2285.09,819.963 2289.93,836.488 
  2294.76,852.875 
  "/>
<polyline clip-path="url(#clip482)" style="stroke:#e26f46; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" points="
  361.501,809.586 366.334,826.237 371.167,842.831 376,859.351 380.833,875.781 385.666,892.106 390.5,908.31 395.333,924.371 400.166,940.284 404.999,956.026 
  409.832,971.586 414.665,986.947 419.498,1002.09 424.332,1017.01 429.165,1031.68 433.998,1046.08 438.831,1060.21 443.664,1074.06 448.497,1087.6 453.33,1100.82 
  458.164,1113.71 462.997,1126.25 467.83,1138.45 472.663,1150.27 477.496,1161.71 482.329,1172.75 487.162,1183.4 491.996,1193.63 496.829,1203.43 501.662,1212.79 
  506.495,1221.71 511.328,1230.17 516.161,1238.17 520.994,1245.69 525.828,1252.73 530.661,1259.28 535.494,1265.33 540.327,1270.88 545.16,1275.91 549.993,1280.42 
  554.826,1284.41 559.66,1287.86 564.493,1290.77 569.326,1293.14 574.159,1294.96 578.992,1296.22 583.825,1296.92 588.658,1297.06 593.492,1296.74 598.325,1296.02 
  603.158,1294.89 607.991,1293.35 612.824,1291.39 617.657,1289.02 622.49,1286.23 627.324,1283.03 632.157,1279.41 636.99,1275.39 641.823,1270.95 646.656,1266.11 
  651.489,1260.86 656.322,1255.21 661.156,1249.17 665.989,1242.73 670.822,1235.92 675.655,1228.72 680.488,1221.15 685.321,1213.21 690.154,1204.92 694.988,1196.27 
  699.821,1187.28 704.654,1177.96 709.487,1168.31 714.32,1158.34 719.153,1148.06 723.986,1137.49 728.82,1126.63 733.653,1115.49 738.486,1104.09 743.319,1092.44 
  748.152,1080.54 752.985,1068.41 757.818,1056.06 762.652,1043.51 767.485,1030.76 772.318,1017.83 777.151,1004.73 781.984,991.477 786.817,978.081 791.65,964.556 
  796.484,950.915 801.317,937.172 806.15,923.342 810.983,909.436 815.816,895.47 820.649,881.458 825.482,867.413 830.316,853.349 835.149,839.281 839.982,825.222 
  844.815,811.187 849.648,797.188 854.481,783.241 859.314,769.358 864.148,755.556 868.981,741.847 873.814,728.247 878.647,714.764 883.48,701.414 888.313,688.212 
  893.147,675.171 897.98,662.305 902.813,649.627 907.646,637.15 912.479,624.882 917.312,612.839 922.145,601.034 926.979,589.478 931.812,578.184 936.645,567.163 
  941.478,556.427 946.311,545.985 951.144,535.848 955.977,526.028 960.811,516.534 965.644,507.377 970.477,498.566 975.31,490.11 980.143,482.019 984.976,474.301 
  989.809,466.965 994.643,460.018 999.476,453.468 1004.31,447.322 1009.14,441.589 1013.98,436.275 1018.81,431.387 1023.64,426.93 1028.47,422.912 1033.31,419.338 
  1038.14,416.214 1042.97,413.546 1047.81,411.339 1052.64,409.597 1057.47,408.327 1062.31,407.532 1067.14,407.217 1071.97,407.33 1076.81,407.764 1081.64,408.52 
  1086.47,409.603 1091.31,411.012 1096.14,412.749 1100.97,414.815 1105.8,417.209 1110.64,419.931 1115.47,422.98 1120.3,426.355 1125.14,430.054 1129.97,434.073 
  1134.8,438.409 1139.64,443.061 1144.47,448.028 1149.3,453.305 1154.14,458.888 1158.97,464.772 1163.8,470.952 1168.64,477.423 1173.47,484.178 1178.3,491.211 
  1183.14,498.516 1187.97,506.086 1192.8,513.913 1197.63,521.991 1202.47,530.312 1207.3,538.868 1212.13,547.651 1216.97,556.652 1221.8,565.863 1226.63,575.276 
  1231.47,584.88 1236.3,594.667 1241.13,604.627 1245.97,614.749 1250.8,625.025 1255.63,635.444 1260.47,645.996 1265.3,656.67 1270.13,667.455 1274.96,678.342 
  1279.8,689.318 1284.63,700.374 1289.46,711.498 1294.3,722.678 1299.13,733.905 1303.96,745.166 1308.8,756.451 1313.63,767.749 1318.46,779.047 1323.3,790.335 
  1328.13,801.601 1332.96,812.835 1337.8,824.025 1342.63,835.159 1347.46,846.227 1352.3,857.218 1357.13,868.12 1361.96,878.922 1366.79,889.614 1371.63,900.184 
  1376.46,910.623 1381.29,920.919 1386.13,931.062 1390.96,941.041 1395.79,950.847 1400.63,960.469 1405.46,969.898 1410.29,979.124 1415.13,988.137 1419.96,996.928 
  1424.79,1005.49 1429.63,1013.81 1434.46,1021.88 1439.29,1029.69 1444.12,1037.24 1448.96,1044.51 1453.79,1051.5 1458.62,1058.21 1463.46,1064.61 1468.29,1070.72 
  1473.12,1076.51 1477.96,1081.98 1482.79,1087.13 1487.62,1091.95 1492.46,1096.43 1497.29,1100.57 1502.12,1104.36 1506.96,1107.8 1511.79,1110.88 1516.62,1113.6 
  1521.46,1115.94 1526.29,1117.92 1531.12,1119.52 1535.95,1120.73 1540.79,1121.56 1545.62,1122 1550.45,1122.06 1555.29,1121.86 1560.12,1121.42 1564.95,1120.76 
  1569.79,1119.85 1574.62,1118.72 1579.45,1117.34 1584.29,1115.73 1589.12,1113.89 1593.95,1111.8 1598.79,1109.48 1603.62,1106.93 1608.45,1104.14 1613.28,1101.12 
  1618.12,1097.87 1622.95,1094.4 1627.78,1090.69 1632.62,1086.76 1637.45,1082.62 1642.28,1078.25 1647.12,1073.67 1651.95,1068.88 1656.78,1063.89 1661.62,1058.69 
  1666.45,1053.3 1671.28,1047.71 1676.12,1041.94 1680.95,1035.98 1685.78,1029.84 1690.62,1023.54 1695.45,1017.07 1700.28,1010.44 1705.11,1003.65 1709.95,996.72 
  1714.78,989.649 1719.61,982.445 1724.45,975.114 1729.28,967.664 1734.11,960.102 1738.95,952.435 1743.78,944.67 1748.61,936.816 1753.45,928.88 1758.28,920.87 
  1763.11,912.793 1767.95,904.658 1772.78,896.472 1777.61,888.244 1782.44,879.982 1787.28,871.693 1792.11,863.386 1796.94,855.07 1801.78,846.753 1806.61,838.444 
  1811.44,830.149 1816.28,821.878 1821.11,813.638 1825.94,805.438 1830.78,797.288 1835.61,789.195 1840.44,781.169 1845.28,773.215 1850.11,765.341 1854.94,757.557 
  1859.78,749.872 1864.61,742.291 1869.44,734.825 1874.27,727.481 1879.11,720.264 1883.94,713.184 1888.77,706.248 1893.61,699.462 1898.44,692.835 1903.27,686.373 
  1908.11,680.083 1912.94,673.972 1917.77,668.046 1922.61,662.313 1927.44,656.777 1932.27,651.446 1937.11,646.324 1941.94,641.419 1946.77,636.736 1951.6,632.281 
  1956.44,628.058 1961.27,624.073 1966.1,620.331 1970.94,616.836 1975.77,613.595 1980.6,610.61 1985.44,607.887 1990.27,605.43 1995.1,603.242 1999.94,601.327 
  2004.77,599.69 2009.6,598.333 2014.44,597.261 2019.27,596.476 2024.1,595.982 2028.94,595.782 2033.77,595.826 2038.6,596.001 2043.43,596.308 2048.27,596.747 
  2053.1,597.321 2057.93,598.029 2062.77,598.872 2067.6,599.85 2072.43,600.963 2077.27,602.211 2082.1,603.594 2086.93,605.112 2091.77,606.764 2096.6,608.55 
  2101.43,610.467 2106.27,612.515 2111.1,614.694 2115.93,617.001 2120.77,619.436 2125.6,621.995 2130.43,624.678 2135.26,627.484 2140.1,630.409 2144.93,633.45 
  2149.76,636.608 2154.6,639.879 2159.43,643.258 2164.26,646.744 2169.1,650.335 2173.93,654.027 2178.76,657.815 2183.6,661.698 2188.43,665.673 2193.26,669.734 
  2198.1,673.879 2202.93,678.103 2207.76,682.404 2212.59,686.776 2217.43,691.216 2222.26,695.72 2227.09,700.284 2231.93,704.901 2236.76,709.57 2241.59,714.285 
  2246.43,719.041 2251.26,723.834 2256.09,728.66 2260.93,733.513 2265.76,738.389 2270.59,743.283 2275.43,748.19 2280.26,753.106 2285.09,758.024 2289.93,762.942 
  2294.76,767.853 
  "/>
<polyline clip-path="url(#clip482)" style="stroke:#008000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" stroke-dasharray="4, 8" points="
  361.501,1365.41 366.334,1365.01 371.167,1364.1 376,1362.67 380.833,1360.73 385.666,1358.28 390.5,1355.3 395.333,1351.82 400.166,1347.82 404.999,1343.31 
  409.832,1338.3 414.665,1332.79 419.498,1326.77 424.332,1320.27 429.165,1313.28 433.998,1305.82 438.831,1297.88 443.664,1289.48 448.497,1280.63 453.33,1271.33 
  458.164,1261.6 462.997,1251.44 467.83,1240.88 472.663,1229.91 477.496,1218.56 482.329,1206.83 487.162,1194.75 491.996,1182.31 496.829,1169.55 501.662,1156.47 
  506.495,1143.09 511.328,1129.42 516.161,1115.49 520.994,1101.3 525.828,1086.88 530.661,1072.24 535.494,1057.4 540.327,1042.38 545.16,1027.19 549.993,1011.85 
  554.826,996.373 559.66,980.787 564.493,965.105 569.326,949.343 574.159,933.52 578.992,917.654 583.825,901.761 588.658,885.858 593.492,869.963 598.325,854.093 
  603.158,838.264 607.991,822.493 612.824,806.796 617.657,791.188 622.49,775.685 627.324,760.302 632.157,745.054 636.99,729.954 641.823,715.017 646.656,700.256 
  651.489,685.684 656.322,671.315 661.156,657.161 665.989,643.234 670.822,629.545 675.655,616.105 680.488,602.924 685.321,590.013 690.154,577.383 694.988,565.041 
  699.821,552.997 704.654,541.261 709.487,529.84 714.32,518.743 719.153,507.977 723.986,497.549 728.82,487.467 733.653,477.736 738.486,468.364 743.319,459.356 
  748.152,450.717 752.985,442.451 757.818,434.563 762.652,427.06 767.485,419.945 772.318,413.226 777.151,406.906 781.984,400.989 786.817,395.48 791.65,390.382 
  796.484,385.697 801.317,381.429 806.15,377.579 810.983,374.149 815.816,371.139 820.649,368.552 825.482,366.387 830.316,364.643 835.149,363.321 839.982,362.42 
  844.815,361.937 849.648,361.871 854.481,362.219 859.314,362.98 864.148,364.149 868.981,365.723 873.814,367.699 878.647,370.074 883.48,372.845 888.313,376.006 
  893.147,379.55 897.98,383.473 902.813,387.769 907.646,392.43 912.479,397.45 917.312,402.822 922.145,408.539 926.979,414.593 931.812,420.976 936.645,427.679 
  941.478,434.695 946.311,442.013 951.144,449.626 955.977,457.523 960.811,465.694 965.644,474.13 970.477,482.823 975.31,491.763 980.143,500.939 984.976,510.342 
  989.809,519.961 994.643,529.785 999.476,539.805 1004.31,550.01 1009.14,560.39 1013.98,570.935 1018.81,581.635 1023.64,592.48 1028.47,603.461 1033.31,614.567 
  1038.14,625.789 1042.97,637.117 1047.81,648.542 1052.64,660.054 1057.47,671.644 1062.31,683.302 1067.14,695.019 1071.97,706.786 1076.81,718.594 1081.64,730.433 
  1086.47,742.294 1091.31,754.169 1096.14,766.049 1100.97,777.925 1105.8,789.787 1110.64,801.628 1115.47,813.438 1120.3,825.208 1125.14,836.931 1129.97,848.595 
  1134.8,860.194 1139.64,871.718 1144.47,883.157 1149.3,894.504 1154.14,905.748 1158.97,916.881 1163.8,927.894 1168.64,938.778 1173.47,949.523 1178.3,960.121 
  1183.14,970.562 1187.97,980.837 1192.8,990.936 1197.63,1000.85 1202.47,1010.57 1207.3,1020.09 1212.13,1029.4 1216.97,1038.5 1221.8,1047.38 1226.63,1056.03 
  1231.47,1064.44 1236.3,1072.58 1241.13,1080.47 1245.97,1088.07 1250.8,1095.39 1255.63,1102.41 1260.47,1109.13 1265.3,1115.53 1270.13,1121.62 1274.96,1127.37 
  1279.8,1132.79 1284.63,1137.87 1289.46,1142.6 1294.3,1146.97 1299.13,1150.99 1303.96,1154.64 1308.8,1157.92 1313.63,1160.82 1318.46,1163.36 1323.3,1165.51 
  1328.13,1167.28 1332.96,1168.66 1337.8,1169.66 1342.63,1170.27 1347.46,1170.49 1352.3,1170.33 1357.13,1169.77 1361.96,1168.83 1366.79,1167.5 1371.63,1165.79 
  1376.46,1163.68 1381.29,1161.19 1386.13,1158.32 1390.96,1155.06 1395.79,1151.42 1400.63,1147.41 1405.46,1143.03 1410.29,1138.29 1415.13,1133.19 1419.96,1127.74 
  1424.79,1121.94 1429.63,1115.81 1434.46,1109.34 1439.29,1102.56 1444.12,1095.46 1448.96,1088.06 1453.79,1080.36 1458.62,1072.38 1463.46,1064.12 1468.29,1055.61 
  1473.12,1046.84 1477.96,1037.84 1482.79,1028.61 1487.62,1019.17 1492.46,1009.52 1497.29,999.679 1502.12,989.662 1506.96,979.483 1511.79,969.155 1516.62,958.691 
  1521.46,948.104 1526.29,937.41 1531.12,926.619 1535.95,915.747 1540.79,904.806 1545.62,893.81 1550.45,882.771 1555.29,871.703 1560.12,860.62 1564.95,849.533 
  1569.79,838.457 1574.62,827.403 1579.45,816.386 1584.29,805.417 1589.12,794.51 1593.95,783.677 1598.79,772.931 1603.62,762.285 1608.45,751.751 1613.28,741.341 
  1618.12,731.069 1622.95,720.942 1627.78,710.967 1632.62,701.153 1637.45,691.51 1642.28,682.048 1647.12,672.776 1651.95,663.704 1656.78,654.839 1661.62,646.189 
  1666.45,637.763 1671.28,629.568 1676.12,621.612 1680.95,613.9 1685.78,606.439 1690.62,599.236 1695.45,592.295 1700.28,585.624 1705.11,579.225 1709.95,573.105 
  1714.78,567.267 1719.61,561.715 1724.45,556.453 1729.28,551.484 1734.11,546.812 1738.95,542.438 1743.78,538.363 1748.61,534.592 1753.45,531.126 1758.28,527.967 
  1763.11,525.117 1767.95,522.579 1772.78,520.353 1777.61,518.439 1782.44,516.84 1787.28,515.553 1792.11,514.58 1796.94,513.92 1801.78,513.571 1806.61,513.533 
  1811.44,513.803 1816.28,514.379 1821.11,515.26 1825.94,516.441 1830.78,517.921 1835.61,519.694 1840.44,521.758 1845.28,524.109 1850.11,526.74 1854.94,529.648 
  1859.78,532.827 1864.61,536.271 1869.44,539.975 1874.27,543.931 1879.11,548.132 1883.94,552.573 1888.77,557.247 1893.61,562.153 1898.44,567.284 1903.27,572.632 
  1908.11,578.19 1912.94,583.951 1917.77,589.909 1922.61,596.055 1927.44,602.382 1932.27,608.884 1937.11,615.552 1941.94,622.379 1946.77,629.357 1951.6,636.48 
  1956.44,643.738 1961.27,651.124 1966.1,658.63 1970.94,666.249 1975.77,673.971 1980.6,681.79 1985.44,689.697 1990.27,697.684 1995.1,705.742 1999.94,713.863 
  2004.77,722.041 2009.6,730.267 2014.44,738.534 2019.27,746.833 2024.1,755.157 2028.94,763.497 2033.77,771.845 2038.6,780.194 2043.43,788.536 2048.27,796.862 
  2053.1,805.166 2057.93,813.439 2062.77,821.673 2067.6,829.861 2072.43,837.996 2077.27,846.069 2082.1,854.073 2086.93,862.001 2091.77,869.845 2096.6,877.597 
  2101.43,885.251 2106.27,892.798 2111.1,900.232 2115.93,907.545 2120.77,914.73 2125.6,921.78 2130.43,928.687 2135.26,935.445 2140.1,942.045 2144.93,948.486 
  2149.76,954.762 2154.6,960.864 2159.43,966.785 2164.26,972.519 2169.1,978.057 2173.93,983.393 2178.76,988.521 2183.6,993.435 2188.43,998.128 2193.26,1002.6 
  2198.1,1006.83 2202.93,1010.84 2207.76,1014.6 2212.59,1018.12 2217.43,1021.38 2222.26,1024.4 2227.09,1027.17 2231.93,1029.67 2236.76,1031.91 2241.59,1033.89 
  2246.43,1035.61 2251.26,1037.06 2256.09,1038.24 2260.93,1039.15 2265.76,1039.79 2270.59,1040.16 2275.43,1040.26 2280.26,1040.08 2285.09,1039.64 2289.93,1038.92 
  2294.76,1037.94 
  "/>
<path clip-path="url(#clip480)" d="
M1421.89 477.116 L2284.45 477.116 L2284.45 166.076 L1421.89 166.076  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<polyline clip-path="url(#clip480)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1421.89,477.116 2284.45,477.116 2284.45,166.076 1421.89,166.076 1421.89,477.116 
  "/>
<polyline clip-path="url(#clip480)" style="stroke:#009af9; stroke-linecap:butt; stroke-linejoin:round; stroke-width:6; stroke-opacity:1; fill:none" points="
  1444.66,243.836 1581.28,243.836 
  "/>
<path clip-path="url(#clip480)" d="M1637.42 219.617 L1637.42 226.457 Q1633.43 224.548 1629.88 223.61 Q1626.34 222.673 1623.04 222.673 Q1617.31 222.673 1614.19 224.895 Q1611.1 227.117 1611.1 231.214 Q1611.1 234.652 1613.15 236.423 Q1615.23 238.159 1621 239.235 L1625.23 240.103 Q1633.08 241.596 1636.79 245.381 Q1640.54 249.131 1640.54 255.45 Q1640.54 262.985 1635.47 266.874 Q1630.44 270.763 1620.68 270.763 Q1617 270.763 1612.84 269.929 Q1608.7 269.096 1604.26 267.464 L1604.26 260.242 Q1608.53 262.638 1612.63 263.853 Q1616.72 265.068 1620.68 265.068 Q1626.69 265.068 1629.95 262.707 Q1633.22 260.346 1633.22 255.971 Q1633.22 252.152 1630.86 249.999 Q1628.53 247.846 1623.18 246.77 L1618.91 245.936 Q1611.06 244.374 1607.56 241.041 Q1604.05 237.707 1604.05 231.77 Q1604.05 224.895 1608.88 220.937 Q1613.74 216.978 1622.25 216.978 Q1625.89 216.978 1629.68 217.638 Q1633.46 218.298 1637.42 219.617 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M1651.2 230.867 L1657.59 230.867 L1657.59 269.756 L1651.2 269.756 L1651.2 230.867 M1651.2 215.728 L1657.59 215.728 L1657.59 223.818 L1651.2 223.818 L1651.2 215.728 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M1701.24 238.332 Q1703.63 234.027 1706.97 231.978 Q1710.3 229.93 1714.81 229.93 Q1720.89 229.93 1724.19 234.2 Q1727.49 238.436 1727.49 246.284 L1727.49 269.756 L1721.06 269.756 L1721.06 246.492 Q1721.06 240.902 1719.08 238.193 Q1717.11 235.485 1713.04 235.485 Q1708.08 235.485 1705.2 238.784 Q1702.31 242.082 1702.31 247.777 L1702.31 269.756 L1695.89 269.756 L1695.89 246.492 Q1695.89 240.867 1693.91 238.193 Q1691.93 235.485 1687.8 235.485 Q1682.9 235.485 1680.02 238.818 Q1677.14 242.117 1677.14 247.777 L1677.14 269.756 L1670.72 269.756 L1670.72 230.867 L1677.14 230.867 L1677.14 236.909 Q1679.33 233.332 1682.38 231.631 Q1685.44 229.93 1689.64 229.93 Q1693.88 229.93 1696.83 232.082 Q1699.81 234.235 1701.24 238.332 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M1746.41 263.922 L1746.41 284.547 L1739.99 284.547 L1739.99 230.867 L1746.41 230.867 L1746.41 236.77 Q1748.43 233.298 1751.48 231.631 Q1754.57 229.93 1758.84 229.93 Q1765.92 229.93 1770.33 235.555 Q1774.78 241.179 1774.78 250.346 Q1774.78 259.513 1770.33 265.138 Q1765.92 270.763 1758.84 270.763 Q1754.57 270.763 1751.48 269.096 Q1748.43 267.395 1746.41 263.922 M1768.15 250.346 Q1768.15 243.298 1765.23 239.304 Q1762.35 235.277 1757.28 235.277 Q1752.21 235.277 1749.29 239.304 Q1746.41 243.298 1746.41 250.346 Q1746.41 257.395 1749.29 261.422 Q1752.21 265.415 1757.28 265.415 Q1762.35 265.415 1765.23 261.422 Q1768.15 257.395 1768.15 250.346 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M1785.37 215.728 L1791.76 215.728 L1791.76 269.756 L1785.37 269.756 L1785.37 215.728 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M1838.39 248.714 L1838.39 251.839 L1809.01 251.839 Q1809.43 258.436 1812.97 261.909 Q1816.55 265.346 1822.9 265.346 Q1826.58 265.346 1830.02 264.443 Q1833.49 263.54 1836.9 261.735 L1836.9 267.777 Q1833.46 269.235 1829.85 269.999 Q1826.24 270.763 1822.52 270.763 Q1813.22 270.763 1807.76 265.346 Q1802.35 259.929 1802.35 250.693 Q1802.35 241.145 1807.49 235.555 Q1812.66 229.93 1821.41 229.93 Q1829.26 229.93 1833.81 234.999 Q1838.39 240.034 1838.39 248.714 M1832 246.839 Q1831.93 241.596 1829.05 238.471 Q1826.2 235.346 1821.48 235.346 Q1816.13 235.346 1812.9 238.367 Q1809.71 241.388 1809.22 246.874 L1832 246.839 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M1849.15 217.916 L1878.95 217.916 L1878.95 223.818 L1856.17 223.818 L1856.17 239.096 L1876.72 239.096 L1876.72 244.999 L1856.17 244.999 L1856.17 269.756 L1849.15 269.756 L1849.15 217.916 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M1890.06 217.916 L1900.51 217.916 L1913.74 253.193 L1927.04 217.916 L1937.49 217.916 L1937.49 269.756 L1930.65 269.756 L1930.65 224.235 L1917.28 259.79 L1910.23 259.79 L1896.86 224.235 L1896.86 269.756 L1890.06 269.756 L1890.06 217.916 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M1950.61 217.916 L1957.66 217.916 L1957.66 249.409 Q1957.66 257.742 1960.68 261.422 Q1963.7 265.068 1970.47 265.068 Q1977.21 265.068 1980.23 261.422 Q1983.25 257.742 1983.25 249.409 L1983.25 217.916 L1990.3 217.916 L1990.3 250.277 Q1990.3 260.415 1985.26 265.589 Q1980.26 270.763 1970.47 270.763 Q1960.65 270.763 1955.61 265.589 Q1950.61 260.415 1950.61 250.277 L1950.61 217.916 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip480)" style="stroke:#e26f46; stroke-linecap:butt; stroke-linejoin:round; stroke-width:6; stroke-opacity:1; fill:none" points="
  1444.66,321.596 1581.28,321.596 
  "/>
<path clip-path="url(#clip480)" d="M1628.63 323.21 Q1630.89 323.974 1633.01 326.474 Q1635.16 328.974 1637.31 333.349 L1644.43 347.516 L1636.9 347.516 L1630.27 334.217 Q1627.7 329.009 1625.27 327.307 Q1622.87 325.606 1618.7 325.606 L1611.06 325.606 L1611.06 347.516 L1604.05 347.516 L1604.05 295.676 L1619.88 295.676 Q1628.77 295.676 1633.15 299.391 Q1637.52 303.106 1637.52 310.606 Q1637.52 315.502 1635.23 318.731 Q1632.97 321.96 1628.63 323.21 M1611.06 301.44 L1611.06 319.842 L1619.88 319.842 Q1624.95 319.842 1627.52 317.516 Q1630.13 315.155 1630.13 310.606 Q1630.13 306.058 1627.52 303.766 Q1624.95 301.44 1619.88 301.44 L1611.06 301.44 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M1683.25 326.474 L1683.25 329.599 L1653.88 329.599 Q1654.29 336.196 1657.84 339.669 Q1661.41 343.106 1667.77 343.106 Q1671.45 343.106 1674.88 342.203 Q1678.36 341.3 1681.76 339.495 L1681.76 345.537 Q1678.32 346.995 1674.71 347.759 Q1671.1 348.523 1667.38 348.523 Q1658.08 348.523 1652.63 343.106 Q1647.21 337.689 1647.21 328.453 Q1647.21 318.905 1652.35 313.315 Q1657.52 307.69 1666.27 307.69 Q1674.12 307.69 1678.67 312.759 Q1683.25 317.794 1683.25 326.474 M1676.86 324.599 Q1676.79 319.356 1673.91 316.231 Q1671.06 313.106 1666.34 313.106 Q1660.99 313.106 1657.77 316.127 Q1654.57 319.148 1654.09 324.634 L1676.86 324.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M1713.43 293.488 L1713.43 298.801 L1707.31 298.801 Q1703.88 298.801 1702.52 300.19 Q1701.2 301.578 1701.2 305.19 L1701.2 308.627 L1711.72 308.627 L1711.72 313.592 L1701.2 313.592 L1701.2 347.516 L1694.78 347.516 L1694.78 313.592 L1688.67 313.592 L1688.67 308.627 L1694.78 308.627 L1694.78 305.919 Q1694.78 299.426 1697.8 296.474 Q1700.82 293.488 1707.38 293.488 L1713.43 293.488 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M1752.04 326.474 L1752.04 329.599 L1722.66 329.599 Q1723.08 336.196 1726.62 339.669 Q1730.2 343.106 1736.55 343.106 Q1740.23 343.106 1743.67 342.203 Q1747.14 341.3 1750.54 339.495 L1750.54 345.537 Q1747.11 346.995 1743.49 347.759 Q1739.88 348.523 1736.17 348.523 Q1726.86 348.523 1721.41 343.106 Q1715.99 337.689 1715.99 328.453 Q1715.99 318.905 1721.13 313.315 Q1726.31 307.69 1735.06 307.69 Q1742.9 307.69 1747.45 312.759 Q1752.04 317.794 1752.04 326.474 M1745.65 324.599 Q1745.58 319.356 1742.7 316.231 Q1739.85 313.106 1735.13 313.106 Q1729.78 313.106 1726.55 316.127 Q1723.36 319.148 1722.87 324.634 L1745.65 324.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M1785.06 314.599 Q1783.98 313.974 1782.7 313.696 Q1781.45 313.384 1779.92 313.384 Q1774.5 313.384 1771.58 316.926 Q1768.7 320.433 1768.7 327.03 L1768.7 347.516 L1762.28 347.516 L1762.28 308.627 L1768.7 308.627 L1768.7 314.669 Q1770.72 311.127 1773.95 309.426 Q1777.17 307.69 1781.79 307.69 Q1782.45 307.69 1783.25 307.794 Q1784.05 307.863 1785.02 308.037 L1785.06 314.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M1823.46 326.474 L1823.46 329.599 L1794.08 329.599 Q1794.5 336.196 1798.04 339.669 Q1801.62 343.106 1807.97 343.106 Q1811.65 343.106 1815.09 342.203 Q1818.56 341.3 1821.97 339.495 L1821.97 345.537 Q1818.53 346.995 1814.92 347.759 Q1811.31 348.523 1807.59 348.523 Q1798.29 348.523 1792.83 343.106 Q1787.42 337.689 1787.42 328.453 Q1787.42 318.905 1792.56 313.315 Q1797.73 307.69 1806.48 307.69 Q1814.33 307.69 1818.88 312.759 Q1823.46 317.794 1823.46 326.474 M1817.07 324.599 Q1817 319.356 1814.12 316.231 Q1811.27 313.106 1806.55 313.106 Q1801.2 313.106 1797.97 316.127 Q1794.78 319.148 1794.29 324.634 L1817.07 324.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M1866.27 324.044 L1866.27 347.516 L1859.88 347.516 L1859.88 324.252 Q1859.88 318.731 1857.73 315.988 Q1855.58 313.245 1851.27 313.245 Q1846.1 313.245 1843.11 316.544 Q1840.13 319.842 1840.13 325.537 L1840.13 347.516 L1833.7 347.516 L1833.7 308.627 L1840.13 308.627 L1840.13 314.669 Q1842.42 311.162 1845.51 309.426 Q1848.63 307.69 1852.7 307.69 Q1859.4 307.69 1862.83 311.856 Q1866.27 315.988 1866.27 324.044 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M1907 310.12 L1907 316.092 Q1904.29 314.599 1901.55 313.87 Q1898.84 313.106 1896.06 313.106 Q1889.85 313.106 1886.41 317.064 Q1882.97 320.988 1882.97 328.106 Q1882.97 335.224 1886.41 339.182 Q1889.85 343.106 1896.06 343.106 Q1898.84 343.106 1901.55 342.377 Q1904.29 341.613 1907 340.12 L1907 346.023 Q1904.33 347.273 1901.45 347.898 Q1898.6 348.523 1895.37 348.523 Q1886.58 348.523 1881.41 343.002 Q1876.24 337.481 1876.24 328.106 Q1876.24 318.592 1881.45 313.141 Q1886.69 307.69 1895.79 307.69 Q1898.74 307.69 1901.55 308.315 Q1904.36 308.905 1907 310.12 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M1951.38 326.474 L1951.38 329.599 L1922 329.599 Q1922.42 336.196 1925.96 339.669 Q1929.54 343.106 1935.89 343.106 Q1939.57 343.106 1943.01 342.203 Q1946.48 341.3 1949.88 339.495 L1949.88 345.537 Q1946.44 346.995 1942.83 347.759 Q1939.22 348.523 1935.51 348.523 Q1926.2 348.523 1920.75 343.106 Q1915.33 337.689 1915.33 328.453 Q1915.33 318.905 1920.47 313.315 Q1925.65 307.69 1934.4 307.69 Q1942.24 307.69 1946.79 312.759 Q1951.38 317.794 1951.38 326.474 M1944.99 324.599 Q1944.92 319.356 1942.04 316.231 Q1939.19 313.106 1934.47 313.106 Q1929.12 313.106 1925.89 316.127 Q1922.69 319.148 1922.21 324.634 L1944.99 324.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip480)" style="stroke:#008000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:6; stroke-opacity:1; fill:none" stroke-dasharray="3, 6" points="
  1444.66,399.356 1581.28,399.356 
  "/>
<path clip-path="url(#clip480)" d="M1604.05 373.436 L1613.5 373.436 L1636.48 416.804 L1636.48 373.436 L1643.29 373.436 L1643.29 425.276 L1633.84 425.276 L1610.86 381.908 L1610.86 425.276 L1604.05 425.276 L1604.05 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M1690.23 404.234 L1690.23 407.359 L1660.86 407.359 Q1661.27 413.956 1664.81 417.429 Q1668.39 420.866 1674.74 420.866 Q1678.43 420.866 1681.86 419.963 Q1685.34 419.06 1688.74 417.255 L1688.74 423.297 Q1685.3 424.755 1681.69 425.519 Q1678.08 426.283 1674.36 426.283 Q1665.06 426.283 1659.61 420.866 Q1654.19 415.449 1654.19 406.213 Q1654.19 396.665 1659.33 391.075 Q1664.5 385.45 1673.25 385.45 Q1681.1 385.45 1685.65 390.519 Q1690.23 395.554 1690.23 404.234 M1683.84 402.359 Q1683.77 397.116 1680.89 393.991 Q1678.04 390.866 1673.32 390.866 Q1667.97 390.866 1664.74 393.887 Q1661.55 396.908 1661.06 402.394 L1683.84 402.359 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M1700.06 409.929 L1700.06 386.387 L1706.45 386.387 L1706.45 409.686 Q1706.45 415.206 1708.6 417.984 Q1710.75 420.727 1715.06 420.727 Q1720.23 420.727 1723.22 417.429 Q1726.24 414.13 1726.24 408.436 L1726.24 386.387 L1732.63 386.387 L1732.63 425.276 L1726.24 425.276 L1726.24 419.304 Q1723.91 422.845 1720.82 424.581 Q1717.77 426.283 1713.7 426.283 Q1707 426.283 1703.53 422.116 Q1700.06 417.949 1700.06 409.929 M1716.13 385.45 L1716.13 385.45 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M1768.32 392.359 Q1767.24 391.734 1765.96 391.456 Q1764.71 391.144 1763.18 391.144 Q1757.77 391.144 1754.85 394.686 Q1751.97 398.193 1751.97 404.79 L1751.97 425.276 L1745.54 425.276 L1745.54 386.387 L1751.97 386.387 L1751.97 392.429 Q1753.98 388.887 1757.21 387.186 Q1760.44 385.45 1765.06 385.45 Q1765.72 385.45 1766.52 385.554 Q1767.31 385.623 1768.29 385.797 L1768.32 392.359 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M1792.7 405.727 Q1784.95 405.727 1781.97 407.498 Q1778.98 409.269 1778.98 413.54 Q1778.98 416.942 1781.2 418.956 Q1783.46 420.935 1787.31 420.935 Q1792.63 420.935 1795.82 417.185 Q1799.05 413.401 1799.05 407.151 L1799.05 405.727 L1792.7 405.727 M1805.44 403.088 L1805.44 425.276 L1799.05 425.276 L1799.05 419.373 Q1796.86 422.915 1793.6 424.616 Q1790.33 426.283 1785.61 426.283 Q1779.64 426.283 1776.1 422.949 Q1772.59 419.581 1772.59 413.956 Q1772.59 407.394 1776.97 404.061 Q1781.38 400.727 1790.09 400.727 L1799.05 400.727 L1799.05 400.102 Q1799.05 395.693 1796.13 393.297 Q1793.25 390.866 1788.01 390.866 Q1784.67 390.866 1781.52 391.665 Q1778.36 392.463 1775.44 394.061 L1775.44 388.158 Q1778.95 386.804 1782.24 386.144 Q1785.54 385.45 1788.67 385.45 Q1797.11 385.45 1801.27 389.825 Q1805.44 394.199 1805.44 403.088 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M1818.6 371.248 L1824.99 371.248 L1824.99 425.276 L1818.6 425.276 L1818.6 371.248 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M1838.63 373.436 L1868.42 373.436 L1868.42 379.338 L1845.65 379.338 L1845.65 394.616 L1866.2 394.616 L1866.2 400.519 L1845.65 400.519 L1845.65 425.276 L1838.63 425.276 L1838.63 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M1879.54 373.436 L1889.99 373.436 L1903.22 408.713 L1916.51 373.436 L1926.97 373.436 L1926.97 425.276 L1920.13 425.276 L1920.13 379.755 L1906.76 415.31 L1899.71 415.31 L1886.34 379.755 L1886.34 425.276 L1879.54 425.276 L1879.54 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M1940.09 373.436 L1947.14 373.436 L1947.14 404.929 Q1947.14 413.262 1950.16 416.942 Q1953.18 420.588 1959.95 420.588 Q1966.69 420.588 1969.71 416.942 Q1972.73 413.262 1972.73 404.929 L1972.73 373.436 L1979.78 373.436 L1979.78 405.797 Q1979.78 415.935 1974.74 421.109 Q1969.74 426.283 1959.95 426.283 Q1950.13 426.283 1945.09 421.109 Q1940.09 415.935 1940.09 405.797 L1940.09 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M2030.61 371.318 Q2025.96 379.304 2023.7 387.116 Q2021.44 394.929 2021.44 402.949 Q2021.44 410.97 2023.7 418.852 Q2025.99 426.699 2030.61 434.651 L2025.06 434.651 Q2019.85 426.491 2017.24 418.609 Q2014.67 410.727 2014.67 402.949 Q2014.67 395.206 2017.24 387.359 Q2019.81 379.512 2025.06 371.318 L2030.61 371.318 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M2049.95 419.373 L2074.43 419.373 L2074.43 425.276 L2041.51 425.276 L2041.51 419.373 Q2045.51 415.241 2052.38 408.297 Q2059.29 401.318 2061.06 399.304 Q2064.43 395.519 2065.75 392.915 Q2067.1 390.276 2067.1 387.741 Q2067.1 383.609 2064.19 381.005 Q2061.31 378.401 2056.65 378.401 Q2053.35 378.401 2049.67 379.547 Q2046.03 380.693 2041.86 383.019 L2041.86 375.936 Q2046.1 374.234 2049.78 373.366 Q2053.46 372.498 2056.51 372.498 Q2064.57 372.498 2069.36 376.526 Q2074.15 380.554 2074.15 387.29 Q2074.15 390.484 2072.94 393.366 Q2071.76 396.213 2068.6 400.102 Q2067.73 401.109 2063.08 405.936 Q2058.42 410.727 2049.95 419.373 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M2089.22 373.436 L2116.76 373.436 L2116.76 379.338 L2095.65 379.338 L2095.65 392.047 Q2097.17 391.526 2098.7 391.283 Q2100.23 391.005 2101.76 391.005 Q2110.44 391.005 2115.51 395.762 Q2120.58 400.519 2120.58 408.644 Q2120.58 417.012 2115.37 421.665 Q2110.16 426.283 2100.68 426.283 Q2097.42 426.283 2094.01 425.727 Q2090.65 425.172 2087.03 424.06 L2087.03 417.012 Q2090.16 418.713 2093.49 419.547 Q2096.83 420.38 2100.54 420.38 Q2106.55 420.38 2110.06 417.22 Q2113.56 414.06 2113.56 408.644 Q2113.56 403.227 2110.06 400.068 Q2106.55 396.908 2100.54 396.908 Q2097.73 396.908 2094.92 397.533 Q2092.14 398.158 2089.22 399.477 L2089.22 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M2149.4 378.054 Q2143.98 378.054 2141.24 383.401 Q2138.53 388.713 2138.53 399.408 Q2138.53 410.067 2141.24 415.415 Q2143.98 420.727 2149.4 420.727 Q2154.85 420.727 2157.55 415.415 Q2160.3 410.067 2160.3 399.408 Q2160.3 388.713 2157.55 383.401 Q2154.85 378.054 2149.4 378.054 M2149.4 372.498 Q2158.11 372.498 2162.69 379.408 Q2167.31 386.283 2167.31 399.408 Q2167.31 412.498 2162.69 419.408 Q2158.11 426.283 2149.4 426.283 Q2140.68 426.283 2136.06 419.408 Q2131.48 412.498 2131.48 399.408 Q2131.48 386.283 2136.06 379.408 Q2140.68 372.498 2149.4 372.498 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M2194.64 378.054 Q2189.22 378.054 2186.48 383.401 Q2183.77 388.713 2183.77 399.408 Q2183.77 410.067 2186.48 415.415 Q2189.22 420.727 2194.64 420.727 Q2200.09 420.727 2202.8 415.415 Q2205.54 410.067 2205.54 399.408 Q2205.54 388.713 2202.8 383.401 Q2200.09 378.054 2194.64 378.054 M2194.64 372.498 Q2203.35 372.498 2207.94 379.408 Q2212.55 386.283 2212.55 399.408 Q2212.55 412.498 2207.94 419.408 Q2203.35 426.283 2194.64 426.283 Q2185.92 426.283 2181.3 419.408 Q2176.72 412.498 2176.72 399.408 Q2176.72 386.283 2181.3 379.408 Q2185.92 372.498 2194.64 372.498 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip480)" d="M2222.97 371.318 L2228.53 371.318 Q2233.73 379.512 2236.3 387.359 Q2238.91 395.206 2238.91 402.949 Q2238.91 410.727 2236.3 418.609 Q2233.73 426.491 2228.53 434.651 L2222.97 434.651 Q2227.59 426.699 2229.85 418.852 Q2232.14 410.97 2232.14 402.949 Q2232.14 394.929 2229.85 387.116 Q2227.59 379.304 2222.97 371.318 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /></svg>




<?xml version="1.0" encoding="utf-8"?>
<svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" width="600" height="400" viewBox="0 0 2400 1600">
<defs>
  <clipPath id="clip570">
    <rect x="0" y="0" width="2400" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip570)" d="
M0 1600 L2400 1600 L2400 0 L0 0  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip571">
    <rect x="480" y="0" width="1681" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip570)" d="
M245.309 1401.58 L2352.76 1401.58 L2352.76 123.472 L245.309 123.472  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip572">
    <rect x="245" y="123" width="2108" height="1279"/>
  </clipPath>
</defs>
<polyline clip-path="url(#clip572)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  304.954,1401.58 304.954,123.472 
  "/>
<polyline clip-path="url(#clip572)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  801.993,1401.58 801.993,123.472 
  "/>
<polyline clip-path="url(#clip572)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1299.03,1401.58 1299.03,123.472 
  "/>
<polyline clip-path="url(#clip572)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1796.07,1401.58 1796.07,123.472 
  "/>
<polyline clip-path="url(#clip572)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  2293.11,1401.58 2293.11,123.472 
  "/>
<polyline clip-path="url(#clip570)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  245.309,1401.58 2352.76,1401.58 
  "/>
<polyline clip-path="url(#clip570)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  304.954,1401.58 304.954,1382.68 
  "/>
<polyline clip-path="url(#clip570)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  801.993,1401.58 801.993,1382.68 
  "/>
<polyline clip-path="url(#clip570)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1299.03,1401.58 1299.03,1382.68 
  "/>
<polyline clip-path="url(#clip570)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1796.07,1401.58 1796.07,1382.68 
  "/>
<polyline clip-path="url(#clip570)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  2293.11,1401.58 2293.11,1382.68 
  "/>
<path clip-path="url(#clip570)" d="M304.954 1438.36 Q299.537 1438.36 296.794 1443.71 Q294.086 1449.02 294.086 1459.71 Q294.086 1470.37 296.794 1475.72 Q299.537 1481.03 304.954 1481.03 Q310.405 1481.03 313.113 1475.72 Q315.856 1470.37 315.856 1459.71 Q315.856 1449.02 313.113 1443.71 Q310.405 1438.36 304.954 1438.36 M304.954 1432.8 Q313.669 1432.8 318.252 1439.71 Q322.87 1446.59 322.87 1459.71 Q322.87 1472.8 318.252 1479.71 Q313.669 1486.59 304.954 1486.59 Q296.238 1486.59 291.62 1479.71 Q287.037 1472.8 287.037 1459.71 Q287.037 1446.59 291.62 1439.71 Q296.238 1432.8 304.954 1432.8 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M787.566 1479.68 L799.024 1479.68 L799.024 1440.13 L786.559 1442.63 L786.559 1436.24 L798.955 1433.74 L805.969 1433.74 L805.969 1479.68 L817.427 1479.68 L817.427 1485.58 L787.566 1485.58 L787.566 1479.68 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M1291.01 1479.68 L1315.49 1479.68 L1315.49 1485.58 L1282.57 1485.58 L1282.57 1479.68 Q1286.57 1475.55 1293.44 1468.6 Q1300.35 1461.62 1302.12 1459.61 Q1305.49 1455.82 1306.81 1453.22 Q1308.16 1450.58 1308.16 1448.05 Q1308.16 1443.91 1305.25 1441.31 Q1302.37 1438.71 1297.71 1438.71 Q1294.41 1438.71 1290.73 1439.85 Q1287.09 1441 1282.92 1443.32 L1282.92 1436.24 Q1287.16 1434.54 1290.84 1433.67 Q1294.52 1432.8 1297.57 1432.8 Q1305.63 1432.8 1310.42 1436.83 Q1315.21 1440.86 1315.21 1447.59 Q1315.21 1450.79 1314 1453.67 Q1312.82 1456.52 1309.66 1460.41 Q1308.79 1461.41 1304.14 1466.24 Q1299.48 1471.03 1291.01 1479.68 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M1802.44 1457.63 Q1807.48 1458.7 1810.29 1462.11 Q1813.14 1465.51 1813.14 1470.51 Q1813.14 1478.18 1807.86 1482.39 Q1802.58 1486.59 1792.86 1486.59 Q1789.6 1486.59 1786.12 1485.93 Q1782.69 1485.3 1779.01 1484.02 L1779.01 1477.25 Q1781.92 1478.95 1785.39 1479.82 Q1788.87 1480.68 1792.65 1480.68 Q1799.25 1480.68 1802.69 1478.08 Q1806.16 1475.48 1806.16 1470.51 Q1806.16 1465.93 1802.93 1463.36 Q1799.73 1460.75 1794.01 1460.75 L1787.96 1460.75 L1787.96 1454.99 L1794.28 1454.99 Q1799.46 1454.99 1802.2 1452.94 Q1804.94 1450.86 1804.94 1446.97 Q1804.94 1442.98 1802.1 1440.86 Q1799.28 1438.71 1794.01 1438.71 Q1791.12 1438.71 1787.83 1439.33 Q1784.53 1439.96 1780.57 1441.27 L1780.57 1435.02 Q1784.56 1433.91 1788.03 1433.36 Q1791.54 1432.8 1794.63 1432.8 Q1802.62 1432.8 1807.27 1436.45 Q1811.92 1440.06 1811.92 1446.24 Q1811.92 1450.55 1809.46 1453.53 Q1806.99 1456.48 1802.44 1457.63 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M2297.63 1439.85 L2279.92 1467.52 L2297.63 1467.52 L2297.63 1439.85 M2295.78 1433.74 L2304.6 1433.74 L2304.6 1467.52 L2312 1467.52 L2312 1473.36 L2304.6 1473.36 L2304.6 1485.58 L2297.63 1485.58 L2297.63 1473.36 L2274.22 1473.36 L2274.22 1466.59 L2295.78 1433.74 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M1242.64 1527.62 L1242.64 1538.66 L1255.8 1538.66 L1255.8 1543.62 L1242.64 1543.62 L1242.64 1564.74 Q1242.64 1569.49 1243.93 1570.85 Q1245.25 1572.2 1249.24 1572.2 L1255.8 1572.2 L1255.8 1577.55 L1249.24 1577.55 Q1241.85 1577.55 1239.03 1574.8 Q1236.22 1572.03 1236.22 1564.74 L1236.22 1543.62 L1231.53 1543.62 L1231.53 1538.66 L1236.22 1538.66 L1236.22 1527.62 L1242.64 1527.62 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M1286.22 1523.52 L1300.94 1523.52 L1300.94 1528.49 L1292.61 1528.49 L1292.61 1581.96 L1300.94 1581.96 L1300.94 1586.92 L1286.22 1586.92 L1286.22 1523.52 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M1339.34 1539.8 L1339.34 1545.85 Q1336.64 1544.46 1333.72 1543.76 Q1330.8 1543.07 1327.68 1543.07 Q1322.92 1543.07 1320.53 1544.53 Q1318.16 1545.99 1318.16 1548.9 Q1318.16 1551.12 1319.87 1552.41 Q1321.57 1553.66 1326.71 1554.8 L1328.89 1555.29 Q1335.7 1556.75 1338.55 1559.42 Q1341.43 1562.06 1341.43 1566.82 Q1341.43 1572.24 1337.12 1575.4 Q1332.85 1578.55 1325.35 1578.55 Q1322.23 1578.55 1318.82 1577.93 Q1315.46 1577.34 1311.71 1576.12 L1311.71 1569.53 Q1315.25 1571.37 1318.69 1572.3 Q1322.12 1573.21 1325.49 1573.21 Q1330 1573.21 1332.43 1571.68 Q1334.87 1570.12 1334.87 1567.3 Q1334.87 1564.7 1333.09 1563.31 Q1331.36 1561.92 1325.42 1560.64 L1323.2 1560.12 Q1317.26 1558.87 1314.62 1556.3 Q1311.98 1553.69 1311.98 1549.18 Q1311.98 1543.69 1315.87 1540.71 Q1319.76 1537.72 1326.91 1537.72 Q1330.46 1537.72 1333.58 1538.24 Q1336.71 1538.76 1339.34 1539.8 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M1366.53 1523.52 L1366.53 1586.92 L1351.81 1586.92 L1351.81 1581.96 L1360.11 1581.96 L1360.11 1528.49 L1351.81 1528.49 L1351.81 1523.52 L1366.53 1523.52 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip572)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  245.309,1397.95 2352.76,1397.95 
  "/>
<polyline clip-path="url(#clip572)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  245.309,1080.23 2352.76,1080.23 
  "/>
<polyline clip-path="url(#clip572)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  245.309,762.504 2352.76,762.504 
  "/>
<polyline clip-path="url(#clip572)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  245.309,444.783 2352.76,444.783 
  "/>
<polyline clip-path="url(#clip572)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  245.309,127.062 2352.76,127.062 
  "/>
<polyline clip-path="url(#clip570)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  245.309,1401.58 245.309,123.472 
  "/>
<polyline clip-path="url(#clip570)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  245.309,1397.95 264.206,1397.95 
  "/>
<polyline clip-path="url(#clip570)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  245.309,1080.23 264.206,1080.23 
  "/>
<polyline clip-path="url(#clip570)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  245.309,762.504 264.206,762.504 
  "/>
<polyline clip-path="url(#clip570)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  245.309,444.783 264.206,444.783 
  "/>
<polyline clip-path="url(#clip570)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  245.309,127.062 264.206,127.062 
  "/>
<path clip-path="url(#clip570)" d="M119.136 1398.62 L163.649 1398.62 L163.649 1404.53 L119.136 1404.53 L119.136 1398.62 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M184.83 1417.96 L209.309 1417.96 L209.309 1423.87 L176.392 1423.87 L176.392 1417.96 Q180.385 1413.83 187.26 1406.89 Q194.17 1399.91 195.941 1397.89 Q199.309 1394.11 200.628 1391.51 Q201.982 1388.87 201.982 1386.33 Q201.982 1382.2 199.066 1379.6 Q196.184 1376.99 191.531 1376.99 Q188.233 1376.99 184.552 1378.14 Q180.906 1379.28 176.74 1381.61 L176.74 1374.53 Q180.976 1372.82 184.656 1371.96 Q188.337 1371.09 191.392 1371.09 Q199.448 1371.09 204.239 1375.12 Q209.031 1379.14 209.031 1385.88 Q209.031 1389.07 207.816 1391.96 Q206.635 1394.8 203.476 1398.69 Q202.607 1399.7 197.955 1404.53 Q193.302 1409.32 184.83 1417.96 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M118.58 1080.9 L163.094 1080.9 L163.094 1086.8 L118.58 1086.8 L118.58 1080.9 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M179.448 1100.24 L190.906 1100.24 L190.906 1060.69 L178.441 1063.19 L178.441 1056.81 L190.837 1054.31 L197.851 1054.31 L197.851 1100.24 L209.309 1100.24 L209.309 1106.15 L179.448 1106.15 L179.448 1100.24 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M191.392 741.202 Q185.976 741.202 183.233 746.549 Q180.524 751.862 180.524 762.556 Q180.524 773.216 183.233 778.563 Q185.976 783.876 191.392 783.876 Q196.844 783.876 199.552 778.563 Q202.295 773.216 202.295 762.556 Q202.295 751.862 199.552 746.549 Q196.844 741.202 191.392 741.202 M191.392 735.647 Q200.107 735.647 204.691 742.556 Q209.309 749.431 209.309 762.556 Q209.309 775.646 204.691 782.556 Q200.107 789.431 191.392 789.431 Q182.677 789.431 178.059 782.556 Q173.476 775.646 173.476 762.556 Q173.476 749.431 178.059 742.556 Q182.677 735.647 191.392 735.647 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M179.448 464.8 L190.906 464.8 L190.906 425.252 L178.441 427.752 L178.441 421.363 L190.837 418.863 L197.851 418.863 L197.851 464.8 L209.309 464.8 L209.309 470.703 L179.448 470.703 L179.448 464.8 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M184.83 147.079 L209.309 147.079 L209.309 152.982 L176.392 152.982 L176.392 147.079 Q180.385 142.947 187.26 136.003 Q194.17 129.024 195.941 127.01 Q199.309 123.225 200.628 120.621 Q201.982 117.982 201.982 115.448 Q201.982 111.316 199.066 108.712 Q196.184 106.107 191.531 106.107 Q188.233 106.107 184.552 107.253 Q180.906 108.399 176.74 110.725 L176.74 103.642 Q180.976 101.941 184.656 101.073 Q188.337 100.205 191.392 100.205 Q199.448 100.205 204.239 104.232 Q209.031 108.26 209.031 114.996 Q209.031 118.191 207.816 121.073 Q206.635 123.92 203.476 127.809 Q202.607 128.816 197.955 133.642 Q193.302 138.434 184.83 147.079 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M35.6047 1073.64 Q31.2992 1071.24 29.2506 1067.91 Q27.202 1064.57 27.202 1060.06 Q27.202 1053.98 31.4728 1050.68 Q35.7089 1047.39 43.5561 1047.39 L67.0282 1047.39 L67.0282 1053.81 L43.7644 1053.81 Q38.1741 1053.81 35.4658 1055.79 Q32.7575 1057.77 32.7575 1061.83 Q32.7575 1066.8 36.0561 1069.68 Q39.3547 1072.56 45.0491 1072.56 L67.0282 1072.56 L67.0282 1078.98 L43.7644 1078.98 Q38.1394 1078.98 35.4658 1080.96 Q32.7575 1082.94 32.7575 1087.07 Q32.7575 1091.97 36.0908 1094.85 Q39.3894 1097.73 45.0491 1097.73 L67.0282 1097.73 L67.0282 1104.16 L28.1395 1104.16 L28.1395 1097.73 L34.1811 1097.73 Q30.6047 1095.55 28.9034 1092.49 Q27.202 1089.43 27.202 1085.23 Q27.202 1081 29.3547 1078.05 Q31.5075 1075.06 35.6047 1073.64 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M47.4797 1016.97 Q47.4797 1024.71 49.2505 1027.7 Q51.0213 1030.68 55.2921 1030.68 Q58.6949 1030.68 60.7087 1028.46 Q62.6879 1026.21 62.6879 1022.35 Q62.6879 1017.04 58.9379 1013.84 Q55.1532 1010.61 48.9033 1010.61 L47.4797 1010.61 L47.4797 1016.97 M44.8408 1004.23 L67.0282 1004.23 L67.0282 1010.61 L61.1254 1010.61 Q64.6671 1012.8 66.3684 1016.07 Q68.0351 1019.33 68.0351 1024.05 Q68.0351 1030.02 64.7018 1033.57 Q61.3337 1037.07 55.7088 1037.07 Q49.1463 1037.07 45.813 1032.7 Q42.4797 1028.29 42.4797 1019.57 L42.4797 1010.61 L41.8547 1010.61 Q37.445 1010.61 35.0492 1013.53 Q32.6186 1016.41 32.6186 1021.66 Q32.6186 1024.99 33.4172 1028.15 Q34.2158 1031.31 35.813 1034.23 L29.9103 1034.23 Q28.5561 1030.72 27.8964 1027.42 Q27.202 1024.12 27.202 1021 Q27.202 1012.56 31.577 1008.39 Q35.9519 1004.23 44.8408 1004.23 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M29.2853 966.275 L35.3269 966.275 Q33.9381 968.983 33.2436 971.9 Q32.5492 974.817 32.5492 977.941 Q32.5492 982.698 34.0075 985.094 Q35.4658 987.455 38.3825 987.455 Q40.6047 987.455 41.8894 985.754 Q43.1394 984.053 44.2852 978.914 L44.7713 976.726 Q46.2297 969.921 48.9033 967.073 Q51.5421 964.192 56.299 964.192 Q61.7157 964.192 64.8754 968.497 Q68.0351 972.768 68.0351 980.268 Q68.0351 983.393 67.4101 986.796 Q66.8198 990.164 65.6046 993.914 L59.0074 993.914 Q60.8476 990.372 61.7851 986.935 Q62.6879 983.497 62.6879 980.129 Q62.6879 975.615 61.1601 973.185 Q59.5976 970.754 56.7852 970.754 Q54.181 970.754 52.7921 972.525 Q51.4032 974.261 50.1185 980.198 L49.5977 982.421 Q48.3477 988.358 45.7783 990.997 Q43.1741 993.636 38.6603 993.636 Q33.1742 993.636 30.1881 989.747 Q27.202 985.858 27.202 978.705 Q27.202 975.164 27.7228 972.039 Q28.2436 968.914 29.2853 966.275 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M29.2853 929.226 L35.3269 929.226 Q33.9381 931.935 33.2436 934.851 Q32.5492 937.768 32.5492 940.893 Q32.5492 945.65 34.0075 948.046 Q35.4658 950.407 38.3825 950.407 Q40.6047 950.407 41.8894 948.706 Q43.1394 947.004 44.2852 941.865 L44.7713 939.678 Q46.2297 932.872 48.9033 930.025 Q51.5421 927.143 56.299 927.143 Q61.7157 927.143 64.8754 931.449 Q68.0351 935.72 68.0351 943.219 Q68.0351 946.344 67.4101 949.747 Q66.8198 953.115 65.6046 956.865 L59.0074 956.865 Q60.8476 953.324 61.7851 949.886 Q62.6879 946.449 62.6879 943.081 Q62.6879 938.567 61.1601 936.136 Q59.5976 933.706 56.7852 933.706 Q54.181 933.706 52.7921 935.476 Q51.4032 937.213 50.1185 943.15 L49.5977 945.372 Q48.3477 951.31 45.7783 953.949 Q43.1741 956.587 38.6603 956.587 Q33.1742 956.587 30.1881 952.699 Q27.202 948.81 27.202 941.657 Q27.202 938.115 27.7228 934.99 Q28.2436 931.865 29.2853 929.226 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M28.1395 898.949 L28.1395 892.178 L60.7782 880.025 L28.1395 867.873 L28.1395 861.102 L67.0282 875.685 L67.0282 884.366 L28.1395 898.949 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M45.9866 819.019 L49.1116 819.019 L49.1116 848.394 Q55.7088 847.977 59.181 844.435 Q62.6185 840.859 62.6185 834.505 Q62.6185 830.824 61.7157 827.387 Q60.8129 823.915 59.0074 820.512 L65.049 820.512 Q66.5073 823.949 67.2712 827.56 Q68.0351 831.171 68.0351 834.887 Q68.0351 844.192 62.6185 849.644 Q57.2018 855.06 47.9658 855.06 Q38.4172 855.06 32.827 849.921 Q27.202 844.748 27.202 835.998 Q27.202 828.151 32.2714 823.602 Q37.3061 819.019 45.9866 819.019 M44.1116 825.408 Q38.8686 825.477 35.7436 828.359 Q32.6186 831.206 32.6186 835.928 Q32.6186 841.276 35.6394 844.505 Q38.6603 847.699 44.1463 848.185 L44.1116 825.408 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M13.0007 808.533 L13.0007 802.144 L67.0282 802.144 L67.0282 808.533 L13.0007 808.533 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M32.6186 773.706 Q32.6186 778.845 36.6464 781.831 Q40.6394 784.818 47.6185 784.818 Q54.5977 784.818 58.6254 781.866 Q62.6185 778.88 62.6185 773.706 Q62.6185 768.602 58.5907 765.616 Q54.5629 762.63 47.6185 762.63 Q40.7089 762.63 36.6811 765.616 Q32.6186 768.602 32.6186 773.706 M27.202 773.706 Q27.202 765.373 32.6186 760.616 Q38.0353 755.859 47.6185 755.859 Q57.1671 755.859 62.6185 760.616 Q68.0351 765.373 68.0351 773.706 Q68.0351 782.074 62.6185 786.831 Q57.1671 791.554 47.6185 791.554 Q38.0353 791.554 32.6186 786.831 Q27.202 782.074 27.202 773.706 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M29.6325 717.283 L35.6047 717.283 Q34.1117 719.991 33.3825 722.735 Q32.6186 725.443 32.6186 728.221 Q32.6186 734.436 36.5769 737.873 Q40.5005 741.311 47.6185 741.311 Q54.7366 741.311 58.6949 737.873 Q62.6185 734.436 62.6185 728.221 Q62.6185 725.443 61.8893 722.735 Q61.1254 719.991 59.6324 717.283 L65.5351 717.283 Q66.7851 719.957 67.4101 722.839 Q68.0351 725.686 68.0351 728.915 Q68.0351 737.7 62.5143 742.873 Q56.9935 748.047 47.6185 748.047 Q38.1047 748.047 32.6533 742.839 Q27.202 737.596 27.202 728.498 Q27.202 725.547 27.827 722.735 Q28.4173 719.922 29.6325 717.283 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M28.1395 706.172 L28.1395 699.783 L67.0282 699.783 L67.0282 706.172 L28.1395 706.172 M13.0007 706.172 L13.0007 699.783 L21.0909 699.783 L21.0909 706.172 L13.0007 706.172 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M17.0979 680.096 L28.1395 680.096 L28.1395 666.936 L33.1047 666.936 L33.1047 680.096 L54.2157 680.096 Q58.9726 680.096 60.3268 678.811 Q61.681 677.492 61.681 673.499 L61.681 666.936 L67.0282 666.936 L67.0282 673.499 Q67.0282 680.894 64.2851 683.707 Q61.5074 686.519 54.2157 686.519 L33.1047 686.519 L33.1047 691.207 L28.1395 691.207 L28.1395 686.519 L17.0979 686.519 L17.0979 680.096 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M70.6392 642.353 Q77.5837 645.061 79.7017 647.631 Q81.8197 650.2 81.8197 654.506 L81.8197 659.61 L76.4726 659.61 L76.4726 655.86 Q76.4726 653.221 75.2226 651.763 Q73.9726 650.304 69.3198 648.534 L66.4032 647.388 L28.1395 663.117 L28.1395 656.346 L58.556 644.193 L28.1395 632.041 L28.1395 625.27 L70.6392 642.353 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M13.0007 594.437 L13.0007 579.714 L17.9659 579.714 L17.9659 588.048 L71.4379 588.048 L71.4379 579.714 L76.4031 579.714 L76.4031 594.437 L13.0007 594.437 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M35.6047 535.826 Q31.2992 533.43 29.2506 530.097 Q27.202 526.763 27.202 522.249 Q27.202 516.173 31.4728 512.875 Q35.7089 509.576 43.5561 509.576 L67.0282 509.576 L67.0282 516 L43.7644 516 Q38.1741 516 35.4658 517.979 Q32.7575 519.958 32.7575 524.02 Q32.7575 528.986 36.0561 531.867 Q39.3547 534.749 45.0491 534.749 L67.0282 534.749 L67.0282 541.173 L43.7644 541.173 Q38.1394 541.173 35.4658 543.152 Q32.7575 545.131 32.7575 549.263 Q32.7575 554.159 36.0908 557.041 Q39.3894 559.923 45.0491 559.923 L67.0282 559.923 L67.0282 566.346 L28.1395 566.346 L28.1395 559.923 L34.1811 559.923 Q30.6047 557.735 28.9034 554.68 Q27.202 551.624 27.202 547.423 Q27.202 543.187 29.3547 540.236 Q31.5075 537.249 35.6047 535.826 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M15.1882 485.479 L15.1882 479.576 L73.6253 497.632 L73.6253 503.534 L15.1882 485.479 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M29.2853 448.083 L35.3269 448.083 Q33.9381 450.792 33.2436 453.708 Q32.5492 456.625 32.5492 459.75 Q32.5492 464.507 34.0075 466.903 Q35.4658 469.264 38.3825 469.264 Q40.6047 469.264 41.8894 467.562 Q43.1394 465.861 44.2852 460.722 L44.7713 458.535 Q46.2297 451.729 48.9033 448.882 Q51.5421 446 56.299 446 Q61.7157 446 64.8754 450.305 Q68.0351 454.576 68.0351 462.076 Q68.0351 465.201 67.4101 468.604 Q66.8198 471.972 65.6046 475.722 L59.0074 475.722 Q60.8476 472.18 61.7851 468.743 Q62.6879 465.305 62.6879 461.937 Q62.6879 457.423 61.1601 454.993 Q59.5976 452.562 56.7852 452.562 Q54.181 452.562 52.7921 454.333 Q51.4032 456.069 50.1185 462.007 L49.5977 464.229 Q48.3477 470.166 45.7783 472.805 Q43.1741 475.444 38.6603 475.444 Q33.1742 475.444 30.1881 471.555 Q27.202 467.666 27.202 460.514 Q27.202 456.972 27.7228 453.847 Q28.2436 450.722 29.2853 448.083 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M13.0007 420.896 L76.4031 420.896 L76.4031 435.618 L71.4379 435.618 L71.4379 427.319 L17.9659 427.319 L17.9659 435.618 L13.0007 435.618 L13.0007 420.896 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M599.056 18.8205 L599.056 65.8515 L608.941 65.8515 Q621.458 65.8515 627.251 60.1802 Q633.084 54.509 633.084 42.2752 Q633.084 30.1225 627.251 24.4918 Q621.458 18.8205 608.941 18.8205 L599.056 18.8205 M590.874 12.096 L607.685 12.096 Q625.266 12.096 633.489 19.4281 Q641.712 26.7198 641.712 42.2752 Q641.712 57.9117 633.449 65.2439 Q625.185 72.576 607.685 72.576 L590.874 72.576 L590.874 12.096 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M693.24 48.0275 L693.24 51.6733 L658.969 51.6733 Q659.455 59.3701 663.587 63.421 Q667.76 67.4314 675.173 67.4314 Q679.467 67.4314 683.477 66.3781 Q687.528 65.3249 691.498 63.2184 L691.498 70.267 Q687.488 71.9684 683.275 72.8596 Q679.062 73.7508 674.727 73.7508 Q663.871 73.7508 657.511 67.4314 Q651.192 61.1119 651.192 50.3365 Q651.192 39.1965 657.187 32.6746 Q663.223 26.1121 673.431 26.1121 Q682.586 26.1121 687.893 32.0264 Q693.24 37.9003 693.24 48.0275 M685.786 45.84 Q685.705 39.7232 682.343 36.0774 Q679.021 32.4315 673.512 32.4315 Q667.274 32.4315 663.506 35.9558 Q659.779 39.4801 659.212 45.8805 L685.786 45.84 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M728.442 9.54393 L728.442 15.7418 L721.313 15.7418 Q717.302 15.7418 715.722 17.3622 Q714.183 18.9825 714.183 23.1955 L714.183 27.2059 L726.457 27.2059 L726.457 32.9987 L714.183 32.9987 L714.183 72.576 L706.689 72.576 L706.689 32.9987 L699.559 32.9987 L699.559 27.2059 L706.689 27.2059 L706.689 24.0462 Q706.689 16.471 710.213 13.0277 Q713.737 9.54393 721.394 9.54393 L728.442 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M755.3 49.7694 Q746.266 49.7694 742.782 51.8354 Q739.299 53.9013 739.299 58.8839 Q739.299 62.8538 741.891 65.2034 Q744.524 67.5124 749.021 67.5124 Q755.219 67.5124 758.946 63.1374 Q762.713 58.7219 762.713 51.4303 L762.713 49.7694 L755.3 49.7694 M770.167 46.6907 L770.167 72.576 L762.713 72.576 L762.713 65.6895 Q760.161 69.8214 756.353 71.8063 Q752.545 73.7508 747.036 73.7508 Q740.068 73.7508 735.936 69.8619 Q731.845 65.9325 731.845 59.3701 Q731.845 51.7138 736.949 47.825 Q742.094 43.9361 752.262 43.9361 L762.713 43.9361 L762.713 43.2069 Q762.713 38.0623 759.31 35.2672 Q755.948 32.4315 749.831 32.4315 Q745.942 32.4315 742.256 33.3632 Q738.57 34.295 735.167 36.1584 L735.167 29.2718 Q739.258 27.692 743.107 26.9223 Q746.955 26.1121 750.601 26.1121 Q760.444 26.1121 765.305 31.2163 Q770.167 36.3204 770.167 46.6907 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M784.75 54.671 L784.75 27.2059 L792.203 27.2059 L792.203 54.3874 Q792.203 60.8284 794.715 64.0691 Q797.227 67.2693 802.25 67.2693 Q808.286 67.2693 811.769 63.421 Q815.294 59.5726 815.294 52.9291 L815.294 27.2059 L822.747 27.2059 L822.747 72.576 L815.294 72.576 L815.294 65.6084 Q812.58 69.7404 808.974 71.7658 Q805.409 73.7508 800.67 73.7508 Q792.852 73.7508 788.801 68.8897 Q784.75 64.0286 784.75 54.671 M803.506 26.1121 L803.506 26.1121 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M838.1 9.54393 L845.554 9.54393 L845.554 72.576 L838.1 72.576 L838.1 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M868.523 14.324 L868.523 27.2059 L883.875 27.2059 L883.875 32.9987 L868.523 32.9987 L868.523 57.6282 Q868.523 63.1779 870.021 64.7578 Q871.561 66.3376 876.219 66.3376 L883.875 66.3376 L883.875 72.576 L876.219 72.576 Q867.591 72.576 864.31 69.3758 Q861.028 66.1351 861.028 57.6282 L861.028 32.9987 L855.56 32.9987 L855.56 27.2059 L861.028 27.2059 L861.028 14.324 L868.523 14.324 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M895.583 62.2867 L904.13 62.2867 L904.13 72.576 L895.583 72.576 L895.583 62.2867 M895.583 29.6769 L904.13 29.6769 L904.13 39.9662 L895.583 39.9662 L895.583 29.6769 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M948.325 12.096 L960.519 12.096 L975.953 53.2532 L991.467 12.096 L1003.66 12.096 L1003.66 72.576 L995.68 72.576 L995.68 19.4686 L980.084 60.9499 L971.861 60.9499 L956.265 19.4686 L956.265 72.576 L948.325 72.576 L948.325 12.096 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M1040.2 49.7694 Q1031.17 49.7694 1027.68 51.8354 Q1024.2 53.9013 1024.2 58.8839 Q1024.2 62.8538 1026.79 65.2034 Q1029.42 67.5124 1033.92 67.5124 Q1040.12 67.5124 1043.85 63.1374 Q1047.61 58.7219 1047.61 51.4303 L1047.61 49.7694 L1040.2 49.7694 M1055.07 46.6907 L1055.07 72.576 L1047.61 72.576 L1047.61 65.6895 Q1045.06 69.8214 1041.25 71.8063 Q1037.45 73.7508 1031.94 73.7508 Q1024.97 73.7508 1020.84 69.8619 Q1016.75 65.9325 1016.75 59.3701 Q1016.75 51.7138 1021.85 47.825 Q1026.99 43.9361 1037.16 43.9361 L1047.61 43.9361 L1047.61 43.2069 Q1047.61 38.0623 1044.21 35.2672 Q1040.85 32.4315 1034.73 32.4315 Q1030.84 32.4315 1027.16 33.3632 Q1023.47 34.295 1020.07 36.1584 L1020.07 29.2718 Q1024.16 27.692 1028.01 26.9223 Q1031.85 26.1121 1035.5 26.1121 Q1045.34 26.1121 1050.21 31.2163 Q1055.07 36.3204 1055.07 46.6907 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M1099.34 28.5427 L1099.34 35.5912 Q1096.18 33.9709 1092.78 33.1607 Q1089.38 32.3505 1085.73 32.3505 Q1080.18 32.3505 1077.39 34.0519 Q1074.63 35.7533 1074.63 39.156 Q1074.63 41.7486 1076.62 43.2475 Q1078.6 44.7058 1084.6 46.0426 L1087.15 46.6097 Q1095.09 48.3111 1098.41 51.4303 Q1101.77 54.509 1101.77 60.0587 Q1101.77 66.3781 1096.75 70.0644 Q1091.77 73.7508 1083.02 73.7508 Q1079.37 73.7508 1075.4 73.0216 Q1071.47 72.3329 1067.1 70.9151 L1067.1 63.2184 Q1071.23 65.3654 1075.24 66.4591 Q1079.25 67.5124 1083.18 67.5124 Q1088.45 67.5124 1091.28 65.73 Q1094.12 63.9071 1094.12 60.6258 Q1094.12 57.5877 1092.05 55.9673 Q1090.03 54.3469 1083.1 52.8481 L1080.51 52.2405 Q1073.58 50.7821 1070.5 47.7845 Q1067.42 44.7463 1067.42 39.4801 Q1067.42 33.0797 1071.96 29.5959 Q1076.5 26.1121 1084.84 26.1121 Q1088.97 26.1121 1092.62 26.7198 Q1096.26 27.3274 1099.34 28.5427 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M1142.57 28.5427 L1142.57 35.5912 Q1139.41 33.9709 1136 33.1607 Q1132.6 32.3505 1128.96 32.3505 Q1123.41 32.3505 1120.61 34.0519 Q1117.86 35.7533 1117.86 39.156 Q1117.86 41.7486 1119.84 43.2475 Q1121.83 44.7058 1127.82 46.0426 L1130.37 46.6097 Q1138.31 48.3111 1141.63 51.4303 Q1145 54.509 1145 60.0587 Q1145 66.3781 1139.97 70.0644 Q1134.99 73.7508 1126.24 73.7508 Q1122.6 73.7508 1118.63 73.0216 Q1114.7 72.3329 1110.32 70.9151 L1110.32 63.2184 Q1114.45 65.3654 1118.46 66.4591 Q1122.47 67.5124 1126.4 67.5124 Q1131.67 67.5124 1134.5 65.73 Q1137.34 63.9071 1137.34 60.6258 Q1137.34 57.5877 1135.27 55.9673 Q1133.25 54.3469 1126.32 52.8481 L1123.73 52.2405 Q1116.8 50.7821 1113.72 47.7845 Q1110.65 44.7463 1110.65 39.4801 Q1110.65 33.0797 1115.18 29.5959 Q1119.72 26.1121 1128.06 26.1121 Q1132.2 26.1121 1135.84 26.7198 Q1139.49 27.3274 1142.57 28.5427 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M1177.89 27.2059 L1185.79 27.2059 L1199.97 65.2844 L1214.15 27.2059 L1222.04 27.2059 L1205.03 72.576 L1194.9 72.576 L1177.89 27.2059 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M1271.14 48.0275 L1271.14 51.6733 L1236.87 51.6733 Q1237.36 59.3701 1241.49 63.421 Q1245.66 67.4314 1253.07 67.4314 Q1257.37 67.4314 1261.38 66.3781 Q1265.43 65.3249 1269.4 63.2184 L1269.4 70.267 Q1265.39 71.9684 1261.18 72.8596 Q1256.96 73.7508 1252.63 73.7508 Q1241.77 73.7508 1235.41 67.4314 Q1229.09 61.1119 1229.09 50.3365 Q1229.09 39.1965 1235.09 32.6746 Q1241.12 26.1121 1251.33 26.1121 Q1260.49 26.1121 1265.79 32.0264 Q1271.14 37.9003 1271.14 48.0275 M1263.69 45.84 Q1263.61 39.7232 1260.24 36.0774 Q1256.92 32.4315 1251.41 32.4315 Q1245.18 32.4315 1241.41 35.9558 Q1237.68 39.4801 1237.11 45.8805 L1263.69 45.84 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M1283.38 9.54393 L1290.83 9.54393 L1290.83 72.576 L1283.38 72.576 L1283.38 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M1324.01 32.4315 Q1318.01 32.4315 1314.53 37.1306 Q1311.04 41.7891 1311.04 49.9314 Q1311.04 58.0738 1314.49 62.7728 Q1317.97 67.4314 1324.01 67.4314 Q1329.96 67.4314 1333.44 62.7323 Q1336.93 58.0333 1336.93 49.9314 Q1336.93 41.8701 1333.44 37.1711 Q1329.96 32.4315 1324.01 32.4315 M1324.01 26.1121 Q1333.73 26.1121 1339.28 32.4315 Q1344.83 38.7509 1344.83 49.9314 Q1344.83 61.0714 1339.28 67.4314 Q1333.73 73.7508 1324.01 73.7508 Q1314.24 73.7508 1308.69 67.4314 Q1303.18 61.0714 1303.18 49.9314 Q1303.18 38.7509 1308.69 32.4315 Q1314.24 26.1121 1324.01 26.1121 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M1389.83 28.9478 L1389.83 35.9153 Q1386.67 34.1734 1383.47 33.3227 Q1380.31 32.4315 1377.07 32.4315 Q1369.82 32.4315 1365.81 37.0496 Q1361.8 41.6271 1361.8 49.9314 Q1361.8 58.2358 1365.81 62.8538 Q1369.82 67.4314 1377.07 67.4314 Q1380.31 67.4314 1383.47 66.5807 Q1386.67 65.6895 1389.83 63.9476 L1389.83 70.8341 Q1386.71 72.2924 1383.35 73.0216 Q1380.03 73.7508 1376.26 73.7508 Q1366.01 73.7508 1359.98 67.3098 Q1353.94 60.8689 1353.94 49.9314 Q1353.94 38.832 1360.02 32.472 Q1366.14 26.1121 1376.75 26.1121 Q1380.19 26.1121 1383.47 26.8413 Q1386.75 27.5299 1389.83 28.9478 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M1402.8 27.2059 L1410.25 27.2059 L1410.25 72.576 L1402.8 72.576 L1402.8 27.2059 M1402.8 9.54393 L1410.25 9.54393 L1410.25 18.9825 L1402.8 18.9825 L1402.8 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M1433.22 14.324 L1433.22 27.2059 L1448.57 27.2059 L1448.57 32.9987 L1433.22 32.9987 L1433.22 57.6282 Q1433.22 63.1779 1434.72 64.7578 Q1436.26 66.3376 1440.92 66.3376 L1448.57 66.3376 L1448.57 72.576 L1440.92 72.576 Q1432.29 72.576 1429.01 69.3758 Q1425.72 66.1351 1425.72 57.6282 L1425.72 32.9987 L1420.26 32.9987 L1420.26 27.2059 L1425.72 27.2059 L1425.72 14.324 L1433.22 14.324 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M1477.25 76.7889 Q1474.09 84.8907 1471.09 87.3618 Q1468.1 89.8329 1463.07 89.8329 L1457.12 89.8329 L1457.12 83.5945 L1461.49 83.5945 Q1464.57 83.5945 1466.27 82.1361 Q1467.98 80.6778 1470.04 75.2496 L1471.38 71.8468 L1453.03 27.2059 L1460.93 27.2059 L1475.1 62.6918 L1489.28 27.2059 L1497.18 27.2059 L1477.25 76.7889 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M1554.46 49.7694 Q1545.43 49.7694 1541.94 51.8354 Q1538.46 53.9013 1538.46 58.8839 Q1538.46 62.8538 1541.05 65.2034 Q1543.69 67.5124 1548.18 67.5124 Q1554.38 67.5124 1558.11 63.1374 Q1561.88 58.7219 1561.88 51.4303 L1561.88 49.7694 L1554.46 49.7694 M1569.33 46.6907 L1569.33 72.576 L1561.88 72.576 L1561.88 65.6895 Q1559.32 69.8214 1555.52 71.8063 Q1551.71 73.7508 1546.2 73.7508 Q1539.23 73.7508 1535.1 69.8619 Q1531.01 65.9325 1531.01 59.3701 Q1531.01 51.7138 1536.11 47.825 Q1541.26 43.9361 1551.42 43.9361 L1561.88 43.9361 L1561.88 43.2069 Q1561.88 38.0623 1558.47 35.2672 Q1555.11 32.4315 1548.99 32.4315 Q1545.1 32.4315 1541.42 33.3632 Q1537.73 34.295 1534.33 36.1584 L1534.33 29.2718 Q1538.42 27.692 1542.27 26.9223 Q1546.12 26.1121 1549.76 26.1121 Q1559.61 26.1121 1564.47 31.2163 Q1569.33 36.3204 1569.33 46.6907 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M1607.65 9.54393 L1607.65 15.7418 L1600.52 15.7418 Q1596.51 15.7418 1594.93 17.3622 Q1593.39 18.9825 1593.39 23.1955 L1593.39 27.2059 L1605.67 27.2059 L1605.67 32.9987 L1593.39 32.9987 L1593.39 72.576 L1585.9 72.576 L1585.9 32.9987 L1578.77 32.9987 L1578.77 27.2059 L1585.9 27.2059 L1585.9 24.0462 Q1585.9 16.471 1589.42 13.0277 Q1592.95 9.54393 1600.6 9.54393 L1607.65 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M1619.8 14.324 L1619.8 27.2059 L1635.16 27.2059 L1635.16 32.9987 L1619.8 32.9987 L1619.8 57.6282 Q1619.8 63.1779 1621.3 64.7578 Q1622.84 66.3376 1627.5 66.3376 L1635.16 66.3376 L1635.16 72.576 L1627.5 72.576 Q1618.87 72.576 1615.59 69.3758 Q1612.31 66.1351 1612.31 57.6282 L1612.31 32.9987 L1606.84 32.9987 L1606.84 27.2059 L1612.31 27.2059 L1612.31 14.324 L1619.8 14.324 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M1683.77 48.0275 L1683.77 51.6733 L1649.5 51.6733 Q1649.98 59.3701 1654.11 63.421 Q1658.29 67.4314 1665.7 67.4314 Q1669.99 67.4314 1674 66.3781 Q1678.06 65.3249 1682.03 63.2184 L1682.03 70.267 Q1678.01 71.9684 1673.8 72.8596 Q1669.59 73.7508 1665.25 73.7508 Q1654.4 73.7508 1648.04 67.4314 Q1641.72 61.1119 1641.72 50.3365 Q1641.72 39.1965 1647.71 32.6746 Q1653.75 26.1121 1663.96 26.1121 Q1673.11 26.1121 1678.42 32.0264 Q1683.77 37.9003 1683.77 48.0275 M1676.31 45.84 Q1676.23 39.7232 1672.87 36.0774 Q1669.55 32.4315 1664.04 32.4315 Q1657.8 32.4315 1654.03 35.9558 Q1650.31 39.4801 1649.74 45.8805 L1676.31 45.84 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M1722.29 34.1734 Q1721.04 33.4443 1719.54 33.1202 Q1718.08 32.7556 1716.3 32.7556 Q1709.98 32.7556 1706.57 36.8875 Q1703.21 40.9789 1703.21 48.6757 L1703.21 72.576 L1695.72 72.576 L1695.72 27.2059 L1703.21 27.2059 L1703.21 34.2544 Q1705.56 30.1225 1709.33 28.1376 Q1713.1 26.1121 1718.48 26.1121 Q1719.25 26.1121 1720.18 26.2337 Q1721.12 26.3147 1722.25 26.5172 L1722.29 34.1734 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M1785.49 44.2197 Q1788.12 45.1109 1790.59 48.0275 Q1793.1 50.9442 1795.61 56.0483 L1803.92 72.576 L1795.13 72.576 L1787.39 57.061 Q1784.39 50.9847 1781.56 48.9997 Q1778.76 47.0148 1773.9 47.0148 L1764.99 47.0148 L1764.99 72.576 L1756.8 72.576 L1756.8 12.096 L1775.28 12.096 Q1785.65 12.096 1790.75 16.4305 Q1795.86 20.7649 1795.86 29.5149 Q1795.86 35.2267 1793.18 38.994 Q1790.55 42.7613 1785.49 44.2197 M1764.99 18.8205 L1764.99 40.2903 L1775.28 40.2903 Q1781.19 40.2903 1784.19 37.5762 Q1787.23 34.8216 1787.23 29.5149 Q1787.23 24.2082 1784.19 21.5346 Q1781.19 18.8205 1775.28 18.8205 L1764.99 18.8205 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M1809.63 54.671 L1809.63 27.2059 L1817.08 27.2059 L1817.08 54.3874 Q1817.08 60.8284 1819.59 64.0691 Q1822.11 67.2693 1827.13 67.2693 Q1833.16 67.2693 1836.65 63.421 Q1840.17 59.5726 1840.17 52.9291 L1840.17 27.2059 L1847.63 27.2059 L1847.63 72.576 L1840.17 72.576 L1840.17 65.6084 Q1837.46 69.7404 1833.85 71.7658 Q1830.29 73.7508 1825.55 73.7508 Q1817.73 73.7508 1813.68 68.8897 Q1809.63 64.0286 1809.63 54.671 M1828.38 26.1121 L1828.38 26.1121 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M1900.69 45.1919 L1900.69 72.576 L1893.24 72.576 L1893.24 45.4349 Q1893.24 38.994 1890.73 35.7938 Q1888.22 32.5936 1883.19 32.5936 Q1877.16 32.5936 1873.67 36.4419 Q1870.19 40.2903 1870.19 46.9338 L1870.19 72.576 L1862.7 72.576 L1862.7 27.2059 L1870.19 27.2059 L1870.19 34.2544 Q1872.86 30.163 1876.47 28.1376 Q1880.11 26.1121 1884.85 26.1121 Q1892.67 26.1121 1896.68 30.9732 Q1900.69 35.7938 1900.69 45.1919 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M1917.46 62.2867 L1926.01 62.2867 L1926.01 72.576 L1917.46 72.576 L1917.46 62.2867 M1917.46 29.6769 L1926.01 29.6769 L1926.01 39.9662 L1917.46 39.9662 L1917.46 29.6769 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M1972.35 65.6895 L1985.72 65.6895 L1985.72 19.5497 L1971.18 22.4663 L1971.18 15.0127 L1985.64 12.096 L1993.82 12.096 L1993.82 65.6895 L2007.19 65.6895 L2007.19 72.576 L1972.35 72.576 L1972.35 65.6895 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip572)" style="stroke:#009af9; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" points="
  304.954,762.504 309.924,743.445 314.894,724.405 319.865,705.402 324.835,686.457 329.806,667.587 334.776,648.812 339.746,630.15 344.717,611.621 349.687,593.242 
  354.658,575.033 359.628,557.011 364.598,539.194 369.569,521.601 374.539,504.248 379.509,487.152 384.48,470.332 389.45,453.805 394.421,437.584 399.391,421.688 
  404.361,406.134 409.332,390.936 414.302,376.109 419.273,361.668 424.243,347.628 429.213,334.004 434.184,320.808 439.154,308.054 444.125,295.754 449.095,283.922 
  454.065,272.568 459.036,261.704 464.006,251.34 468.977,241.489 473.947,232.157 478.917,223.357 483.888,215.095 488.858,207.381 493.829,200.222 498.799,193.625 
  503.769,187.598 508.74,182.145 513.71,177.272 518.681,172.985 523.651,169.287 528.621,166.182 533.592,163.674 538.562,161.764 543.532,160.456 548.503,159.749 
  553.473,159.645 558.444,160.144 563.414,161.245 568.384,162.948 573.355,165.25 578.325,168.149 583.296,171.643 588.266,175.728 593.236,180.399 598.207,185.653 
  603.177,191.483 608.148,197.885 613.118,204.851 618.088,212.375 623.059,220.448 628.029,229.064 633,238.214 637.97,247.887 642.94,258.076 647.911,268.768 
  652.881,279.955 657.852,291.624 662.822,303.764 667.792,316.363 672.763,329.408 677.733,342.886 682.704,356.783 687.674,371.087 692.644,385.781 697.615,400.853 
  702.585,416.286 707.555,432.065 712.526,448.176 717.496,464.6 722.467,481.321 727.437,498.325 732.407,515.593 737.378,533.107 742.348,550.851 747.319,568.806 
  752.289,586.956 757.259,605.281 762.23,623.762 767.2,642.383 772.171,661.124 777.141,679.967 782.111,698.891 787.082,717.88 792.052,736.913 797.023,755.972 
  801.993,775.037 806.963,794.09 811.934,813.111 816.904,832.082 821.875,850.982 826.845,869.795 831.815,888.501 836.786,907.08 841.756,925.514 846.727,943.786 
  851.697,961.877 856.667,979.768 861.638,997.441 866.608,1014.88 871.578,1032.07 876.549,1048.99 881.519,1065.62 886.49,1081.95 891.46,1097.96 896.43,1113.63 
  901.401,1128.96 906.371,1143.91 911.342,1158.49 916.312,1172.66 921.282,1186.43 926.253,1199.77 931.223,1212.68 936.194,1225.13 941.164,1237.12 946.134,1248.64 
  951.105,1259.67 956.075,1270.21 961.046,1280.23 966.016,1289.74 970.986,1298.72 975.957,1307.17 980.927,1315.07 985.898,1322.42 990.868,1329.21 995.838,1335.43 
  1000.81,1341.08 1005.78,1346.15 1010.75,1350.64 1015.72,1354.54 1020.69,1357.85 1025.66,1360.56 1030.63,1362.67 1035.6,1364.19 1040.57,1365.1 1045.54,1365.41 
  1050.51,1365.11 1055.48,1364.22 1060.45,1362.72 1065.42,1360.62 1070.39,1357.92 1075.36,1354.63 1080.34,1350.74 1085.31,1346.27 1090.28,1341.21 1095.25,1335.57 
  1100.22,1329.36 1105.19,1322.59 1110.16,1315.25 1115.13,1307.36 1120.1,1298.92 1125.07,1289.96 1130.04,1280.46 1135.01,1270.44 1139.98,1259.91 1144.95,1248.89 
  1149.92,1237.38 1154.89,1225.4 1159.86,1212.95 1164.83,1200.06 1169.8,1186.72 1174.77,1172.97 1179.74,1158.8 1184.71,1144.23 1189.68,1129.29 1194.65,1113.97 
  1199.62,1098.31 1204.59,1082.31 1209.57,1065.99 1214.54,1049.37 1219.51,1032.46 1224.48,1015.28 1229.45,997.849 1234.42,980.183 1239.39,962.3 1244.36,944.215 
  1249.33,925.949 1254.3,907.52 1259.27,888.946 1264.24,870.246 1269.21,851.438 1274.18,832.542 1279.15,813.576 1284.12,794.559 1289.09,775.509 1294.06,756.446 
  1299.03,737.39 1304,718.359 1308.97,699.372 1313.94,680.45 1318.91,661.609 1323.88,642.869 1328.85,624.248 1333.83,605.766 1338.8,587.441 1343.77,569.291 
  1348.74,551.335 1353.71,533.59 1358.68,516.074 1363.65,498.804 1368.62,481.799 1373.59,465.074 1378.56,448.647 1383.53,432.534 1388.5,416.75 1393.47,401.313 
  1398.44,386.237 1403.41,371.538 1408.38,357.23 1413.35,343.328 1418.32,329.844 1423.29,316.793 1428.26,304.188 1433.23,292.041 1438.2,280.366 1443.17,269.172 
  1448.14,258.473 1453.11,248.277 1458.08,238.595 1463.06,229.437 1468.03,220.812 1473,212.729 1477.97,205.197 1482.94,198.222 1487.91,191.811 1492.88,185.971 
  1497.85,180.707 1502.82,176.025 1507.79,171.928 1512.76,168.424 1517.73,165.514 1522.7,163.201 1527.67,161.487 1532.64,160.374 1537.61,159.863 1542.58,159.954 
  1547.55,160.647 1552.52,161.944 1557.49,163.841 1562.46,166.338 1567.43,169.43 1572.4,173.115 1577.37,177.388 1582.34,182.246 1587.32,187.685 1592.29,193.7 
  1597.26,200.284 1602.23,207.43 1607.2,215.131 1612.17,223.379 1617.14,232.165 1622.11,241.48 1627.08,251.318 1632.05,261.668 1637.02,272.519 1641.99,283.86 
  1646.96,295.678 1651.93,307.964 1656.9,320.702 1661.87,333.882 1666.84,347.493 1671.81,361.519 1676.78,375.945 1681.75,390.758 1686.72,405.942 1691.69,421.482 
  1696.66,437.363 1701.63,453.568 1706.6,470.084 1711.58,486.892 1716.55,503.975 1721.52,521.317 1726.49,538.9 1731.46,556.705 1736.43,574.716 1741.4,592.915 
  1746.37,611.284 1751.34,629.804 1756.31,648.457 1761.28,667.224 1766.25,686.085 1771.22,705.023 1776.19,724.017 1781.16,743.05 1786.13,762.103 1791.1,781.157 
  1796.07,800.191 1801.04,819.187 1806.01,838.127 1810.98,856.99 1815.95,875.759 1820.92,894.415 1825.89,912.938 1830.86,931.312 1835.83,949.516 1840.81,967.533 
  1845.78,985.344 1850.75,1002.93 1855.72,1020.28 1860.69,1037.37 1865.66,1054.19 1870.63,1070.71 1875.6,1086.92 1880.57,1102.82 1885.54,1118.37 1890.51,1133.56 
  1895.48,1148.38 1900.45,1162.82 1905.42,1176.86 1910.39,1190.48 1915.36,1203.67 1920.33,1216.42 1925.3,1228.72 1930.27,1240.55 1935.24,1251.91 1940.21,1262.77 
  1945.18,1273.14 1950.15,1282.99 1955.12,1292.32 1960.09,1301.12 1965.07,1309.39 1970.04,1317.11 1975.01,1324.27 1979.98,1330.87 1984.95,1336.9 1989.92,1342.36 
  1994.89,1347.24 1999.86,1351.53 2004.83,1355.23 2009.8,1358.35 2014.77,1360.86 2019.74,1362.78 2024.71,1364.09 2029.68,1364.81 2034.65,1364.92 2039.62,1364.43 
  2044.59,1363.34 2049.56,1361.65 2054.53,1359.35 2059.5,1356.46 2064.47,1352.98 2069.44,1348.91 2074.41,1344.25 2079.38,1339.01 2084.35,1333.19 2089.33,1326.8 
  2094.3,1319.85 2099.27,1312.34 2104.24,1304.28 2109.21,1295.67 2114.18,1286.54 2119.15,1276.88 2124.12,1266.71 2129.09,1256.03 2134.06,1244.86 2139.03,1233.21 
  2144,1221.08 2148.97,1208.5 2153.94,1195.47 2158.91,1182.01 2163.88,1168.13 2168.85,1153.84 2173.82,1139.17 2178.79,1124.11 2183.76,1108.7 2188.73,1092.94 
  2193.7,1076.84 2198.67,1060.44 2203.64,1043.74 2208.61,1026.75 2213.58,1009.5 2218.56,992.004 2223.53,974.278 2228.5,956.341 2233.47,938.211 2238.44,919.906 
  2243.41,901.442 2248.38,882.839 2253.35,864.116 2258.32,845.292 2263.29,826.385 2268.26,807.415 2273.23,788.4 2278.2,769.36 2283.17,750.313 2288.14,731.278 
  2293.11,712.275 
  "/>
<polyline clip-path="url(#clip572)" style="stroke:#e26f46; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" points="
  304.954,762.504 309.924,748.942 314.894,735.277 319.865,721.531 324.835,707.728 329.806,693.887 334.776,680.031 339.746,666.179 344.717,652.355 349.687,638.579 
  354.658,624.869 359.628,611.246 364.598,597.728 369.569,584.334 374.539,571.081 379.509,557.987 384.48,545.067 389.45,532.339 394.421,519.817 399.391,507.516 
  404.361,495.452 409.332,483.64 414.302,472.091 419.273,460.819 424.243,449.839 429.213,439.163 434.184,428.8 439.154,418.764 444.125,409.065 449.095,399.715 
  454.065,390.723 459.036,382.099 464.006,373.852 468.977,365.991 473.947,358.524 478.917,351.459 483.888,344.804 488.858,338.565 493.829,332.748 498.799,327.361 
  503.769,322.408 508.74,317.895 513.71,313.826 518.681,310.206 523.651,307.037 528.621,304.325 533.592,302.071 538.562,300.277 543.532,298.945 548.503,298.077 
  553.473,297.674 558.444,297.736 563.414,298.262 568.384,299.253 573.355,300.707 578.325,302.623 583.296,304.999 588.266,307.833 593.236,311.122 598.207,314.863 
  603.177,319.051 608.148,323.684 613.118,328.756 618.088,334.263 623.059,340.199 628.029,346.558 633,353.336 637.97,360.524 642.94,368.117 647.911,376.106 
  652.881,384.486 657.852,393.247 662.822,402.382 667.792,411.882 672.763,421.738 677.733,431.943 682.704,442.486 687.674,453.359 692.644,464.55 697.615,476.051 
  702.585,487.853 707.555,499.945 712.526,512.316 717.496,524.956 722.467,537.856 727.437,551.009 732.407,564.401 737.378,578.024 742.348,591.867 747.319,605.921 
  752.289,620.177 757.259,634.625 762.23,649.258 767.2,664.071 772.171,679.053 777.141,694.201 782.111,709.509 787.082,724.973 792.052,740.589 797.023,756.355 
  801.993,769.342 806.963,780.578 811.934,791.891 816.904,803.262 821.875,814.675 826.845,826.112 831.815,837.558 836.786,848.996 841.756,860.409 846.727,871.78 
  851.697,883.095 856.667,894.336 861.638,905.488 866.608,916.539 871.578,927.47 876.549,938.27 881.519,948.925 886.49,959.421 891.46,969.745 896.43,979.884 
  901.401,989.827 906.371,999.561 911.342,1009.07 916.312,1018.36 921.282,1027.4 926.253,1036.19 931.223,1044.71 936.194,1052.97 941.164,1060.94 946.134,1068.63 
  951.105,1076.01 956.075,1083.09 961.046,1089.85 966.016,1096.28 970.986,1102.39 975.957,1108.17 980.927,1113.6 985.898,1118.68 990.868,1123.41 995.838,1127.78 
  1000.81,1131.78 1005.78,1135.42 1010.75,1138.69 1015.72,1141.59 1020.69,1144.1 1025.66,1146.24 1030.63,1147.99 1035.6,1149.36 1040.57,1150.34 1045.54,1150.93 
  1050.51,1151.14 1055.48,1150.95 1060.45,1150.38 1065.42,1149.42 1070.39,1148.08 1075.36,1146.34 1080.34,1144.22 1085.31,1141.72 1090.28,1138.84 1095.25,1135.58 
  1100.22,1131.95 1105.19,1127.94 1110.16,1123.57 1115.13,1118.83 1120.1,1113.74 1125.07,1108.29 1130.04,1102.49 1135.01,1096.34 1139.98,1089.86 1144.95,1083.04 
  1149.92,1075.9 1154.89,1068.44 1159.86,1060.66 1164.83,1052.58 1169.8,1044.2 1174.77,1035.53 1179.74,1026.57 1184.71,1017.34 1189.68,1007.83 1194.65,998.065 
  1199.62,988.047 1204.59,977.784 1209.57,967.284 1214.54,956.553 1219.51,945.602 1224.48,934.438 1229.45,923.072 1234.42,911.509 1239.39,899.757 1244.36,887.825 
  1249.33,875.719 1254.3,863.446 1259.27,851.013 1264.24,838.427 1269.21,825.691 1274.18,812.813 1279.15,799.797 1284.12,786.647 1289.09,773.366 1294.06,760.867 
  1299.03,752.13 1304,743.328 1308.97,734.471 1313.94,725.573 1318.91,716.646 1323.88,707.701 1328.85,698.751 1333.83,689.809 1338.8,680.887 1343.77,671.998 
  1348.74,663.153 1353.71,654.365 1358.68,645.644 1363.65,637.004 1368.62,628.454 1373.59,620.005 1378.56,611.669 1383.53,603.454 1388.5,595.372 1393.47,587.432 
  1398.44,579.646 1403.41,572.023 1408.38,564.571 1413.35,557.3 1418.32,550.219 1423.29,543.335 1428.26,536.657 1433.23,530.192 1438.2,523.949 1443.17,517.933 
  1448.14,512.153 1453.11,506.613 1458.08,501.322 1463.06,496.284 1468.03,491.506 1473,486.992 1477.97,482.748 1482.94,478.779 1487.91,475.089 1492.88,471.682 
  1497.85,468.563 1502.82,465.734 1507.79,463.199 1512.76,460.961 1517.73,459.022 1522.7,457.386 1527.67,456.053 1532.64,455.026 1537.61,454.305 1542.58,453.892 
  1547.55,453.788 1552.52,453.992 1557.49,454.506 1562.46,455.329 1567.43,456.46 1572.4,457.898 1577.37,459.643 1582.34,461.693 1587.32,464.046 1592.29,466.7 
  1597.26,469.653 1602.23,472.901 1607.2,476.444 1612.17,480.278 1617.14,484.398 1622.11,488.802 1627.08,493.485 1632.05,498.443 1637.02,503.673 1641.99,509.169 
  1646.96,514.928 1651.93,520.944 1656.9,527.211 1661.87,533.725 1666.84,540.479 1671.81,547.469 1676.78,554.69 1681.75,562.134 1686.72,569.796 1691.69,577.671 
  1696.66,585.75 1701.63,594.029 1706.6,602.501 1711.58,611.161 1716.55,620.002 1721.52,629.017 1726.49,638.201 1731.46,647.546 1736.43,657.046 1741.4,666.698 
  1746.37,676.494 1751.34,686.429 1756.31,696.496 1761.28,706.692 1766.25,717.009 1771.22,727.444 1776.19,737.993 1781.16,748.651 1786.13,759.416 1791.1,766.866 
  1796.07,773.01 1801.04,779.198 1806.01,785.421 1810.98,791.671 1815.95,797.941 1820.92,804.223 1825.89,810.51 1830.86,816.792 1835.83,823.063 1840.81,829.312 
  1845.78,835.533 1850.75,841.716 1855.72,847.856 1860.69,853.942 1865.66,859.969 1870.63,865.927 1875.6,871.811 1880.57,877.611 1885.54,883.322 1890.51,888.935 
  1895.48,894.444 1900.45,899.844 1905.42,905.127 1910.39,910.286 1915.36,915.314 1920.33,920.205 1925.3,924.954 1930.27,929.555 1935.24,934.002 1940.21,938.29 
  1945.18,942.414 1950.15,946.369 1955.12,950.15 1960.09,953.753 1965.07,957.174 1970.04,960.408 1975.01,963.451 1979.98,966.301 1984.95,968.952 1989.92,971.404 
  1994.89,973.651 1999.86,975.692 2004.83,977.524 2009.8,979.145 2014.77,980.552 2019.74,981.743 2024.71,982.718 2029.68,983.474 2034.65,984.01 2039.62,984.326 
  2044.59,984.42 2049.56,984.293 2054.53,983.943 2059.5,983.37 2064.47,982.575 2069.44,981.559 2074.41,980.321 2079.38,978.862 2084.35,977.184 2089.33,975.287 
  2094.3,973.173 2099.27,970.843 2104.24,968.301 2109.21,965.546 2114.18,962.581 2119.15,959.409 2124.12,956.033 2129.09,952.455 2134.06,948.677 2139.03,944.703 
  2144,940.536 2148.97,936.18 2153.94,931.637 2158.91,926.911 2163.88,922.006 2168.85,916.926 2173.82,911.674 2178.79,906.255 2183.76,900.673 2188.73,894.932 
  2193.7,889.036 2198.67,882.99 2203.64,876.797 2208.61,870.463 2213.58,863.991 2218.56,857.387 2223.53,850.655 2228.5,843.799 2233.47,836.823 2238.44,829.733 
  2243.41,822.532 2248.38,815.225 2253.35,807.816 2258.32,800.31 2263.29,792.709 2268.26,785.02 2273.23,777.245 2278.2,769.387 2283.17,762.081 2288.14,758.864 
  2293.11,755.622 
  "/>
<polyline clip-path="url(#clip572)" style="stroke:#008000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" stroke-dasharray="4, 8" points="
  304.954,762.504 309.924,748.602 314.894,734.691 319.865,720.781 324.835,706.88 329.806,692.999 334.776,679.145 339.746,665.33 344.717,651.564 349.687,637.857 
  354.658,624.219 359.628,610.663 364.598,597.199 369.569,583.84 374.539,570.599 379.509,557.487 384.48,544.519 389.45,531.707 394.421,519.067 399.391,506.611 
  404.361,494.356 409.332,482.316 414.302,470.507 419.273,458.945 424.243,447.647 429.213,436.627 434.184,425.903 439.154,415.491 444.125,405.408 449.095,395.671 
  454.065,386.296 459.036,377.296 464.006,368.692 468.977,360.498 473.947,352.732 478.917,345.407 483.888,338.537 488.858,332.136 493.829,326.214 498.799,320.781 
  503.769,315.848 508.74,311.422 513.71,307.51 518.681,304.119 523.651,301.253 528.621,298.917 533.592,297.113 538.562,295.841 543.532,295.1 548.503,294.887 
  553.473,295.198 558.444,296.029 563.414,297.375 568.384,299.23 573.355,301.585 578.325,304.434 583.296,307.767 588.266,311.575 593.236,315.846 598.207,320.568 
  603.177,325.73 608.148,331.316 613.118,337.314 618.088,343.712 623.059,350.498 628.029,357.658 633,365.177 637.97,373.044 642.94,381.243 647.911,389.763 
  652.881,398.588 657.852,407.707 662.822,417.107 667.792,426.774 672.763,436.696 677.733,446.86 682.704,457.254 687.674,467.866 692.644,478.683 697.615,489.692 
  702.585,500.883 707.555,512.244 712.526,523.764 717.496,535.432 722.467,547.237 727.437,559.17 732.407,571.219 737.378,583.375 742.348,595.627 747.319,607.965 
  752.289,620.378 757.259,632.856 762.23,645.39 767.2,657.968 772.171,670.583 777.141,683.222 782.111,695.877 787.082,708.538 792.052,721.196 797.023,733.839 
  801.993,746.46 806.963,759.048 811.934,771.594 816.904,784.089 821.875,796.523 826.845,808.887 831.815,821.172 836.786,833.366 841.756,845.458 846.727,857.439 
  851.697,869.297 856.667,881.021 861.638,892.602 866.608,904.029 871.578,915.292 876.549,926.379 881.519,937.281 886.49,947.986 891.46,958.486 896.43,968.768 
  901.401,978.824 906.371,988.642 911.342,998.213 916.312,1007.53 921.282,1016.57 926.253,1025.34 931.223,1033.82 936.194,1042.01 941.164,1049.91 946.134,1057.49 
  951.105,1064.75 956.075,1071.68 961.046,1078.28 966.016,1084.53 970.986,1090.43 975.957,1095.97 980.927,1101.16 985.898,1105.98 990.868,1110.43 995.838,1114.5 
  1000.81,1118.21 1005.78,1121.53 1010.75,1124.48 1015.72,1127.05 1020.69,1129.25 1025.66,1131.07 1030.63,1132.51 1035.6,1133.58 1040.57,1134.28 1045.54,1134.62 
  1050.51,1134.6 1055.48,1134.22 1060.45,1133.49 1065.42,1132.41 1070.39,1130.98 1075.36,1129.22 1080.34,1127.11 1085.31,1124.68 1090.28,1121.93 1095.25,1118.86 
  1100.22,1115.48 1105.19,1111.79 1110.16,1107.81 1115.13,1103.54 1120.1,1098.98 1125.07,1094.14 1130.04,1089.04 1135.01,1083.67 1139.98,1078.04 1144.95,1072.16 
  1149.92,1066.04 1154.89,1059.67 1159.86,1053.08 1164.83,1046.27 1169.8,1039.24 1174.77,1032 1179.74,1024.56 1184.71,1016.92 1189.68,1009.09 1194.65,1001.08 
  1199.62,992.887 1204.59,984.522 1209.57,975.992 1214.54,967.302 1219.51,958.461 1224.48,949.473 1229.45,940.347 1234.42,931.089 1239.39,921.705 1244.36,912.204 
  1249.33,902.591 1254.3,892.875 1259.27,883.061 1264.24,873.159 1269.21,863.174 1274.18,853.114 1279.15,842.988 1284.12,832.802 1289.09,822.563 1294.06,812.281 
  1299.03,801.962 1304,791.614 1308.97,781.246 1313.94,770.865 1318.91,760.479 1323.88,750.097 1328.85,739.727 1333.83,729.376 1338.8,719.054 1343.77,708.769 
  1348.74,698.531 1353.71,688.35 1358.68,678.238 1363.65,668.205 1368.62,658.261 1373.59,648.418 1378.56,638.686 1383.53,629.076 1388.5,619.599 1393.47,610.265 
  1398.44,601.086 1403.41,592.072 1408.38,583.235 1413.35,574.584 1418.32,566.132 1423.29,557.888 1428.26,549.865 1433.23,542.073 1438.2,534.523 1443.17,527.226 
  1448.14,520.193 1453.11,513.436 1458.08,506.962 1463.06,500.776 1468.03,494.887 1473,489.307 1477.97,484.046 1482.94,479.113 1487.91,474.516 1492.88,470.263 
  1497.85,466.361 1502.82,462.817 1507.79,459.635 1512.76,456.82 1517.73,454.376 1522.7,452.305 1527.67,450.609 1532.64,449.29 1537.61,448.349 1542.58,447.784 
  1547.55,447.595 1552.52,447.779 1557.49,448.334 1562.46,449.257 1567.43,450.542 1572.4,452.185 1577.37,454.179 1582.34,456.518 1587.32,459.194 1592.29,462.199 
  1597.26,465.523 1602.23,469.162 1607.2,473.113 1612.17,477.369 1617.14,481.92 1622.11,486.757 1627.08,491.871 1632.05,497.252 1637.02,502.893 1641.99,508.783 
  1646.96,514.914 1651.93,521.276 1656.9,527.859 1661.87,534.654 1666.84,541.652 1671.81,548.842 1676.78,556.216 1681.75,563.764 1686.72,571.475 1691.69,579.341 
  1696.66,587.35 1701.63,595.493 1706.6,603.761 1711.58,612.142 1716.55,620.627 1721.52,629.206 1726.49,637.872 1731.46,646.615 1736.43,655.427 1741.4,664.297 
  1746.37,673.216 1751.34,682.176 1756.31,691.166 1761.28,700.178 1766.25,709.202 1771.22,718.229 1776.19,727.251 1781.16,736.257 1786.13,745.24 1791.1,754.191 
  1796.07,763.1 1801.04,771.959 1806.01,780.759 1810.98,789.492 1815.95,798.149 1820.92,806.721 1825.89,815.2 1830.86,823.579 1835.83,831.847 1840.81,839.998 
  1845.78,848.023 1850.75,855.913 1855.72,863.662 1860.69,871.26 1865.66,878.7 1870.63,885.973 1875.6,893.075 1880.57,900 1885.54,906.739 1890.51,913.285 
  1895.48,919.631 1900.45,925.77 1905.42,931.697 1910.39,937.405 1915.36,942.888 1920.33,948.142 1925.3,953.161 1930.27,957.942 1935.24,962.478 1940.21,966.768 
  1945.18,970.807 1950.15,974.593 1955.12,978.122 1960.09,981.393 1965.07,984.403 1970.04,987.15 1975.01,989.634 1979.98,991.854 1984.95,993.808 1989.92,995.496 
  1994.89,996.918 1999.86,998.074 2004.83,998.965 2009.8,999.592 2014.77,999.957 2019.74,1000.06 2024.71,999.906 2029.68,999.494 2034.65,998.827 2039.62,997.909 
  2044.59,996.741 2049.56,995.326 2054.53,993.667 2059.5,991.769 2064.47,989.634 2069.44,987.265 2074.41,984.668 2079.38,981.847 2084.35,978.805 2089.33,975.547 
  2094.3,972.078 2099.27,968.404 2104.24,964.529 2109.21,960.459 2114.18,956.199 2119.15,951.756 2124.12,947.136 2129.09,942.344 2134.06,937.388 2139.03,932.271 
  2144,926.997 2148.97,921.57 2153.94,915.997 2158.91,910.283 2163.88,904.434 2168.85,898.457 2173.82,892.355 2178.79,886.137 2183.76,879.807 2188.73,873.372 
  2193.7,866.838 2198.67,860.211 2203.64,853.497 2208.61,846.703 2213.58,839.835 2218.56,832.9 2223.53,825.903 2228.5,818.853 2233.47,811.755 2238.44,804.616 
  2243.41,797.444 2248.38,790.244 2253.35,783.024 2258.32,775.791 2263.29,768.551 2268.26,761.313 2273.23,754.083 2278.2,746.869 2283.17,739.677 2288.14,732.516 
  2293.11,725.392 
  "/>
<path clip-path="url(#clip570)" d="
M1413.49 477.116 L2282.51 477.116 L2282.51 166.076 L1413.49 166.076  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<polyline clip-path="url(#clip570)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1413.49,477.116 2282.51,477.116 2282.51,166.076 1413.49,166.076 1413.49,477.116 
  "/>
<polyline clip-path="url(#clip570)" style="stroke:#009af9; stroke-linecap:butt; stroke-linejoin:round; stroke-width:6; stroke-opacity:1; fill:none" points="
  1436.91,243.836 1577.4,243.836 
  "/>
<path clip-path="url(#clip570)" d="M1634.19 219.617 L1634.19 226.457 Q1630.19 224.548 1626.65 223.61 Q1623.11 222.673 1619.81 222.673 Q1614.08 222.673 1610.96 224.895 Q1607.87 227.117 1607.87 231.214 Q1607.87 234.652 1609.91 236.423 Q1612 238.159 1617.76 239.235 L1622 240.103 Q1629.85 241.596 1633.56 245.381 Q1637.31 249.131 1637.31 255.45 Q1637.31 262.985 1632.24 266.874 Q1627.21 270.763 1617.45 270.763 Q1613.77 270.763 1609.6 269.929 Q1605.47 269.096 1601.03 267.464 L1601.03 260.242 Q1605.3 262.638 1609.39 263.853 Q1613.49 265.068 1617.45 265.068 Q1623.46 265.068 1626.72 262.707 Q1629.98 260.346 1629.98 255.971 Q1629.98 252.152 1627.62 249.999 Q1625.3 247.846 1619.95 246.77 L1615.68 245.936 Q1607.83 244.374 1604.32 241.041 Q1600.82 237.707 1600.82 231.77 Q1600.82 224.895 1605.64 220.937 Q1610.51 216.978 1619.01 216.978 Q1622.66 216.978 1626.44 217.638 Q1630.23 218.298 1634.19 219.617 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M1647.97 230.867 L1654.36 230.867 L1654.36 269.756 L1647.97 269.756 L1647.97 230.867 M1647.97 215.728 L1654.36 215.728 L1654.36 223.818 L1647.97 223.818 L1647.97 215.728 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M1698 238.332 Q1700.4 234.027 1703.73 231.978 Q1707.07 229.93 1711.58 229.93 Q1717.66 229.93 1720.96 234.2 Q1724.25 238.436 1724.25 246.284 L1724.25 269.756 L1717.83 269.756 L1717.83 246.492 Q1717.83 240.902 1715.85 238.193 Q1713.87 235.485 1709.81 235.485 Q1704.85 235.485 1701.96 238.784 Q1699.08 242.082 1699.08 247.777 L1699.08 269.756 L1692.66 269.756 L1692.66 246.492 Q1692.66 240.867 1690.68 238.193 Q1688.7 235.485 1684.57 235.485 Q1679.67 235.485 1676.79 238.818 Q1673.91 242.117 1673.91 247.777 L1673.91 269.756 L1667.48 269.756 L1667.48 230.867 L1673.91 230.867 L1673.91 236.909 Q1676.1 233.332 1679.15 231.631 Q1682.21 229.93 1686.41 229.93 Q1690.64 229.93 1693.6 232.082 Q1696.58 234.235 1698 238.332 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M1743.18 263.922 L1743.18 284.547 L1736.75 284.547 L1736.75 230.867 L1743.18 230.867 L1743.18 236.77 Q1745.19 233.298 1748.25 231.631 Q1751.34 229.93 1755.61 229.93 Q1762.69 229.93 1767.1 235.555 Q1771.55 241.179 1771.55 250.346 Q1771.55 259.513 1767.1 265.138 Q1762.69 270.763 1755.61 270.763 Q1751.34 270.763 1748.25 269.096 Q1745.19 267.395 1743.18 263.922 M1764.91 250.346 Q1764.91 243.298 1762 239.304 Q1759.12 235.277 1754.05 235.277 Q1748.98 235.277 1746.06 239.304 Q1743.18 243.298 1743.18 250.346 Q1743.18 257.395 1746.06 261.422 Q1748.98 265.415 1754.05 265.415 Q1759.12 265.415 1762 261.422 Q1764.91 257.395 1764.91 250.346 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M1782.14 215.728 L1788.53 215.728 L1788.53 269.756 L1782.14 269.756 L1782.14 215.728 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M1835.16 248.714 L1835.16 251.839 L1805.78 251.839 Q1806.2 258.436 1809.74 261.909 Q1813.32 265.346 1819.67 265.346 Q1823.35 265.346 1826.79 264.443 Q1830.26 263.54 1833.66 261.735 L1833.66 267.777 Q1830.23 269.235 1826.62 269.999 Q1823 270.763 1819.29 270.763 Q1809.98 270.763 1804.53 265.346 Q1799.12 259.929 1799.12 250.693 Q1799.12 241.145 1804.25 235.555 Q1809.43 229.93 1818.18 229.93 Q1826.02 229.93 1830.57 234.999 Q1835.16 240.034 1835.16 248.714 M1828.77 246.839 Q1828.7 241.596 1825.82 238.471 Q1822.97 235.346 1818.25 235.346 Q1812.9 235.346 1809.67 238.367 Q1806.48 241.388 1805.99 246.874 L1828.77 246.839 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M1845.92 217.916 L1875.71 217.916 L1875.71 223.818 L1852.93 223.818 L1852.93 239.096 L1873.49 239.096 L1873.49 244.999 L1852.93 244.999 L1852.93 269.756 L1845.92 269.756 L1845.92 217.916 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M1886.82 217.916 L1897.27 217.916 L1910.5 253.193 L1923.8 217.916 L1934.25 217.916 L1934.25 269.756 L1927.41 269.756 L1927.41 224.235 L1914.05 259.79 L1907 259.79 L1893.63 224.235 L1893.63 269.756 L1886.82 269.756 L1886.82 217.916 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M1947.38 217.916 L1954.43 217.916 L1954.43 249.409 Q1954.43 257.742 1957.45 261.422 Q1960.47 265.068 1967.24 265.068 Q1973.98 265.068 1977 261.422 Q1980.02 257.742 1980.02 249.409 L1980.02 217.916 L1987.07 217.916 L1987.07 250.277 Q1987.07 260.415 1982.03 265.589 Q1977.03 270.763 1967.24 270.763 Q1957.41 270.763 1952.38 265.589 Q1947.38 260.415 1947.38 250.277 L1947.38 217.916 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip570)" style="stroke:#e26f46; stroke-linecap:butt; stroke-linejoin:round; stroke-width:6; stroke-opacity:1; fill:none" points="
  1436.91,321.596 1577.4,321.596 
  "/>
<path clip-path="url(#clip570)" d="M1625.4 323.21 Q1627.66 323.974 1629.78 326.474 Q1631.93 328.974 1634.08 333.349 L1641.2 347.516 L1633.66 347.516 L1627.03 334.217 Q1624.46 329.009 1622.03 327.307 Q1619.64 325.606 1615.47 325.606 L1607.83 325.606 L1607.83 347.516 L1600.82 347.516 L1600.82 295.676 L1616.65 295.676 Q1625.54 295.676 1629.91 299.391 Q1634.29 303.106 1634.29 310.606 Q1634.29 315.502 1632 318.731 Q1629.74 321.96 1625.4 323.21 M1607.83 301.44 L1607.83 319.842 L1616.65 319.842 Q1621.72 319.842 1624.29 317.516 Q1626.89 315.155 1626.89 310.606 Q1626.89 306.058 1624.29 303.766 Q1621.72 301.44 1616.65 301.44 L1607.83 301.44 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M1680.02 326.474 L1680.02 329.599 L1650.64 329.599 Q1651.06 336.196 1654.6 339.669 Q1658.18 343.106 1664.53 343.106 Q1668.21 343.106 1671.65 342.203 Q1675.12 341.3 1678.53 339.495 L1678.53 345.537 Q1675.09 346.995 1671.48 347.759 Q1667.87 348.523 1664.15 348.523 Q1654.85 348.523 1649.39 343.106 Q1643.98 337.689 1643.98 328.453 Q1643.98 318.905 1649.12 313.315 Q1654.29 307.69 1663.04 307.69 Q1670.89 307.69 1675.44 312.759 Q1680.02 317.794 1680.02 326.474 M1673.63 324.599 Q1673.56 319.356 1670.68 316.231 Q1667.83 313.106 1663.11 313.106 Q1657.76 313.106 1654.53 316.127 Q1651.34 319.148 1650.85 324.634 L1673.63 324.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M1710.19 293.488 L1710.19 298.801 L1704.08 298.801 Q1700.64 298.801 1699.29 300.19 Q1697.97 301.578 1697.97 305.19 L1697.97 308.627 L1708.49 308.627 L1708.49 313.592 L1697.97 313.592 L1697.97 347.516 L1691.55 347.516 L1691.55 313.592 L1685.44 313.592 L1685.44 308.627 L1691.55 308.627 L1691.55 305.919 Q1691.55 299.426 1694.57 296.474 Q1697.59 293.488 1704.15 293.488 L1710.19 293.488 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M1748.8 326.474 L1748.8 329.599 L1719.43 329.599 Q1719.84 336.196 1723.39 339.669 Q1726.96 343.106 1733.32 343.106 Q1737 343.106 1740.44 342.203 Q1743.91 341.3 1747.31 339.495 L1747.31 345.537 Q1743.87 346.995 1740.26 347.759 Q1736.65 348.523 1732.94 348.523 Q1723.63 348.523 1718.18 343.106 Q1712.76 337.689 1712.76 328.453 Q1712.76 318.905 1717.9 313.315 Q1723.07 307.69 1731.82 307.69 Q1739.67 307.69 1744.22 312.759 Q1748.8 317.794 1748.8 326.474 M1742.41 324.599 Q1742.34 319.356 1739.46 316.231 Q1736.62 313.106 1731.89 313.106 Q1726.55 313.106 1723.32 316.127 Q1720.12 319.148 1719.64 324.634 L1742.41 324.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M1781.82 314.599 Q1780.75 313.974 1779.46 313.696 Q1778.21 313.384 1776.68 313.384 Q1771.27 313.384 1768.35 316.926 Q1765.47 320.433 1765.47 327.03 L1765.47 347.516 L1759.05 347.516 L1759.05 308.627 L1765.47 308.627 L1765.47 314.669 Q1767.48 311.127 1770.71 309.426 Q1773.94 307.69 1778.56 307.69 Q1779.22 307.69 1780.02 307.794 Q1780.82 307.863 1781.79 308.037 L1781.82 314.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M1820.23 326.474 L1820.23 329.599 L1790.85 329.599 Q1791.27 336.196 1794.81 339.669 Q1798.39 343.106 1804.74 343.106 Q1808.42 343.106 1811.86 342.203 Q1815.33 341.3 1818.73 339.495 L1818.73 345.537 Q1815.3 346.995 1811.68 347.759 Q1808.07 348.523 1804.36 348.523 Q1795.05 348.523 1789.6 343.106 Q1784.18 337.689 1784.18 328.453 Q1784.18 318.905 1789.32 313.315 Q1794.5 307.69 1803.25 307.69 Q1811.09 307.69 1815.64 312.759 Q1820.23 317.794 1820.23 326.474 M1813.84 324.599 Q1813.77 319.356 1810.89 316.231 Q1808.04 313.106 1803.32 313.106 Q1797.97 313.106 1794.74 316.127 Q1791.55 319.148 1791.06 324.634 L1813.84 324.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M1863.04 324.044 L1863.04 347.516 L1856.65 347.516 L1856.65 324.252 Q1856.65 318.731 1854.5 315.988 Q1852.34 313.245 1848.04 313.245 Q1842.87 313.245 1839.88 316.544 Q1836.89 319.842 1836.89 325.537 L1836.89 347.516 L1830.47 347.516 L1830.47 308.627 L1836.89 308.627 L1836.89 314.669 Q1839.18 311.162 1842.27 309.426 Q1845.4 307.69 1849.46 307.69 Q1856.16 307.69 1859.6 311.856 Q1863.04 315.988 1863.04 324.044 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M1903.77 310.12 L1903.77 316.092 Q1901.06 314.599 1898.32 313.87 Q1895.61 313.106 1892.83 313.106 Q1886.61 313.106 1883.18 317.064 Q1879.74 320.988 1879.74 328.106 Q1879.74 335.224 1883.18 339.182 Q1886.61 343.106 1892.83 343.106 Q1895.61 343.106 1898.32 342.377 Q1901.06 341.613 1903.77 340.12 L1903.77 346.023 Q1901.09 347.273 1898.21 347.898 Q1895.36 348.523 1892.14 348.523 Q1883.35 348.523 1878.18 343.002 Q1873 337.481 1873 328.106 Q1873 318.592 1878.21 313.141 Q1883.46 307.69 1892.55 307.69 Q1895.5 307.69 1898.32 308.315 Q1901.13 308.905 1903.77 310.12 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M1948.14 326.474 L1948.14 329.599 L1918.77 329.599 Q1919.18 336.196 1922.73 339.669 Q1926.3 343.106 1932.66 343.106 Q1936.34 343.106 1939.77 342.203 Q1943.25 341.3 1946.65 339.495 L1946.65 345.537 Q1943.21 346.995 1939.6 347.759 Q1935.99 348.523 1932.27 348.523 Q1922.97 348.523 1917.52 343.106 Q1912.1 337.689 1912.1 328.453 Q1912.1 318.905 1917.24 313.315 Q1922.41 307.69 1931.16 307.69 Q1939.01 307.69 1943.56 312.759 Q1948.14 317.794 1948.14 326.474 M1941.75 324.599 Q1941.68 319.356 1938.8 316.231 Q1935.95 313.106 1931.23 313.106 Q1925.89 313.106 1922.66 316.127 Q1919.46 319.148 1918.98 324.634 L1941.75 324.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip570)" style="stroke:#008000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:6; stroke-opacity:1; fill:none" stroke-dasharray="3, 6" points="
  1436.91,399.356 1577.4,399.356 
  "/>
<path clip-path="url(#clip570)" d="M1600.82 373.436 L1610.26 373.436 L1633.25 416.804 L1633.25 373.436 L1640.05 373.436 L1640.05 425.276 L1630.61 425.276 L1607.62 381.908 L1607.62 425.276 L1600.82 425.276 L1600.82 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M1687 404.234 L1687 407.359 L1657.62 407.359 Q1658.04 413.956 1661.58 417.429 Q1665.16 420.866 1671.51 420.866 Q1675.19 420.866 1678.63 419.963 Q1682.1 419.06 1685.5 417.255 L1685.5 423.297 Q1682.07 424.755 1678.46 425.519 Q1674.85 426.283 1671.13 426.283 Q1661.82 426.283 1656.37 420.866 Q1650.96 415.449 1650.96 406.213 Q1650.96 396.665 1656.1 391.075 Q1661.27 385.45 1670.02 385.45 Q1677.87 385.45 1682.41 390.519 Q1687 395.554 1687 404.234 M1680.61 402.359 Q1680.54 397.116 1677.66 393.991 Q1674.81 390.866 1670.09 390.866 Q1664.74 390.866 1661.51 393.887 Q1658.32 396.908 1657.83 402.394 L1680.61 402.359 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M1696.82 409.929 L1696.82 386.387 L1703.21 386.387 L1703.21 409.686 Q1703.21 415.206 1705.37 417.984 Q1707.52 420.727 1711.82 420.727 Q1717 420.727 1719.98 417.429 Q1723 414.13 1723 408.436 L1723 386.387 L1729.39 386.387 L1729.39 425.276 L1723 425.276 L1723 419.304 Q1720.68 422.845 1717.59 424.581 Q1714.53 426.283 1710.47 426.283 Q1703.77 426.283 1700.3 422.116 Q1696.82 417.949 1696.82 409.929 M1712.9 385.45 L1712.9 385.45 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M1765.09 392.359 Q1764.01 391.734 1762.73 391.456 Q1761.48 391.144 1759.95 391.144 Q1754.53 391.144 1751.62 394.686 Q1748.73 398.193 1748.73 404.79 L1748.73 425.276 L1742.31 425.276 L1742.31 386.387 L1748.73 386.387 L1748.73 392.429 Q1750.75 388.887 1753.98 387.186 Q1757.21 385.45 1761.82 385.45 Q1762.48 385.45 1763.28 385.554 Q1764.08 385.623 1765.05 385.797 L1765.09 392.359 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M1789.46 405.727 Q1781.72 405.727 1778.73 407.498 Q1775.75 409.269 1775.75 413.54 Q1775.75 416.942 1777.97 418.956 Q1780.23 420.935 1784.08 420.935 Q1789.39 420.935 1792.59 417.185 Q1795.82 413.401 1795.82 407.151 L1795.82 405.727 L1789.46 405.727 M1802.21 403.088 L1802.21 425.276 L1795.82 425.276 L1795.82 419.373 Q1793.63 422.915 1790.37 424.616 Q1787.1 426.283 1782.38 426.283 Q1776.41 426.283 1772.87 422.949 Q1769.36 419.581 1769.36 413.956 Q1769.36 407.394 1773.73 404.061 Q1778.14 400.727 1786.86 400.727 L1795.82 400.727 L1795.82 400.102 Q1795.82 395.693 1792.9 393.297 Q1790.02 390.866 1784.78 390.866 Q1781.44 390.866 1778.28 391.665 Q1775.12 392.463 1772.21 394.061 L1772.21 388.158 Q1775.71 386.804 1779.01 386.144 Q1782.31 385.45 1785.43 385.45 Q1793.87 385.45 1798.04 389.825 Q1802.21 394.199 1802.21 403.088 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M1815.37 371.248 L1821.75 371.248 L1821.75 425.276 L1815.37 425.276 L1815.37 371.248 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M1835.4 373.436 L1865.19 373.436 L1865.19 379.338 L1842.41 379.338 L1842.41 394.616 L1862.97 394.616 L1862.97 400.519 L1842.41 400.519 L1842.41 425.276 L1835.4 425.276 L1835.4 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M1876.3 373.436 L1886.75 373.436 L1899.98 408.713 L1913.28 373.436 L1923.73 373.436 L1923.73 425.276 L1916.89 425.276 L1916.89 379.755 L1903.52 415.31 L1896.48 415.31 L1883.11 379.755 L1883.11 425.276 L1876.3 425.276 L1876.3 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M1936.86 373.436 L1943.91 373.436 L1943.91 404.929 Q1943.91 413.262 1946.93 416.942 Q1949.95 420.588 1956.72 420.588 Q1963.45 420.588 1966.48 416.942 Q1969.5 413.262 1969.5 404.929 L1969.5 373.436 L1976.54 373.436 L1976.54 405.797 Q1976.54 415.935 1971.51 421.109 Q1966.51 426.283 1956.72 426.283 Q1946.89 426.283 1941.86 421.109 Q1936.86 415.935 1936.86 405.797 L1936.86 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M2027.38 371.318 Q2022.73 379.304 2020.47 387.116 Q2018.21 394.929 2018.21 402.949 Q2018.21 410.97 2020.47 418.852 Q2022.76 426.699 2027.38 434.651 L2021.82 434.651 Q2016.61 426.491 2014.01 418.609 Q2011.44 410.727 2011.44 402.949 Q2011.44 395.206 2014.01 387.359 Q2016.58 379.512 2021.82 371.318 L2027.38 371.318 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M2046.72 419.373 L2071.2 419.373 L2071.2 425.276 L2038.28 425.276 L2038.28 419.373 Q2042.27 415.241 2049.15 408.297 Q2056.06 401.318 2057.83 399.304 Q2061.2 395.519 2062.52 392.915 Q2063.87 390.276 2063.87 387.741 Q2063.87 383.609 2060.95 381.005 Q2058.07 378.401 2053.42 378.401 Q2050.12 378.401 2046.44 379.547 Q2042.79 380.693 2038.63 383.019 L2038.63 375.936 Q2042.86 374.234 2046.54 373.366 Q2050.23 372.498 2053.28 372.498 Q2061.34 372.498 2066.13 376.526 Q2070.92 380.554 2070.92 387.29 Q2070.92 390.484 2069.7 393.366 Q2068.52 396.213 2065.36 400.102 Q2064.5 401.109 2059.84 405.936 Q2055.19 410.727 2046.72 419.373 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M2085.99 373.436 L2113.52 373.436 L2113.52 379.338 L2092.41 379.338 L2092.41 392.047 Q2093.94 391.526 2095.47 391.283 Q2097 391.005 2098.52 391.005 Q2107.2 391.005 2112.27 395.762 Q2117.34 400.519 2117.34 408.644 Q2117.34 417.012 2112.13 421.665 Q2106.93 426.283 2097.45 426.283 Q2094.18 426.283 2090.78 425.727 Q2087.41 425.172 2083.8 424.06 L2083.8 417.012 Q2086.93 418.713 2090.26 419.547 Q2093.59 420.38 2097.31 420.38 Q2103.32 420.38 2106.82 417.22 Q2110.33 414.06 2110.33 408.644 Q2110.33 403.227 2106.82 400.068 Q2103.32 396.908 2097.31 396.908 Q2094.5 396.908 2091.68 397.533 Q2088.91 398.158 2085.99 399.477 L2085.99 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M2146.16 378.054 Q2140.75 378.054 2138 383.401 Q2135.29 388.713 2135.29 399.408 Q2135.29 410.067 2138 415.415 Q2140.75 420.727 2146.16 420.727 Q2151.61 420.727 2154.32 415.415 Q2157.06 410.067 2157.06 399.408 Q2157.06 388.713 2154.32 383.401 Q2151.61 378.054 2146.16 378.054 M2146.16 372.498 Q2154.88 372.498 2159.46 379.408 Q2164.08 386.283 2164.08 399.408 Q2164.08 412.498 2159.46 419.408 Q2154.88 426.283 2146.16 426.283 Q2137.45 426.283 2132.83 419.408 Q2128.25 412.498 2128.25 399.408 Q2128.25 386.283 2132.83 379.408 Q2137.45 372.498 2146.16 372.498 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M2191.4 378.054 Q2185.99 378.054 2183.25 383.401 Q2180.54 388.713 2180.54 399.408 Q2180.54 410.067 2183.25 415.415 Q2185.99 420.727 2191.4 420.727 Q2196.86 420.727 2199.56 415.415 Q2202.31 410.067 2202.31 399.408 Q2202.31 388.713 2199.56 383.401 Q2196.86 378.054 2191.4 378.054 M2191.4 372.498 Q2200.12 372.498 2204.7 379.408 Q2209.32 386.283 2209.32 399.408 Q2209.32 412.498 2204.7 419.408 Q2200.12 426.283 2191.4 426.283 Q2182.69 426.283 2178.07 419.408 Q2173.49 412.498 2173.49 399.408 Q2173.49 386.283 2178.07 379.408 Q2182.69 372.498 2191.4 372.498 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip570)" d="M2219.74 371.318 L2225.29 371.318 Q2230.5 379.512 2233.07 387.359 Q2235.68 395.206 2235.68 402.949 Q2235.68 410.727 2233.07 418.609 Q2230.5 426.491 2225.29 434.651 L2219.74 434.651 Q2224.36 426.699 2226.61 418.852 Q2228.9 410.97 2228.9 402.949 Q2228.9 394.929 2226.61 387.116 Q2224.36 379.304 2219.74 371.318 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /></svg>




<?xml version="1.0" encoding="utf-8"?>
<svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" width="600" height="400" viewBox="0 0 2400 1600">
<defs>
  <clipPath id="clip660">
    <rect x="0" y="0" width="2400" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip660)" d="
M0 1600 L2400 1600 L2400 0 L0 0  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip661">
    <rect x="480" y="0" width="1681" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip660)" d="
M314.996 1401.58 L2352.76 1401.58 L2352.76 123.472 L314.996 123.472  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip662">
    <rect x="314" y="123" width="2039" height="1279"/>
  </clipPath>
</defs>
<polyline clip-path="url(#clip662)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  372.668,1401.58 372.668,123.472 
  "/>
<polyline clip-path="url(#clip662)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  853.272,1401.58 853.272,123.472 
  "/>
<polyline clip-path="url(#clip662)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1333.88,1401.58 1333.88,123.472 
  "/>
<polyline clip-path="url(#clip662)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1814.48,1401.58 1814.48,123.472 
  "/>
<polyline clip-path="url(#clip662)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  2295.08,1401.58 2295.08,123.472 
  "/>
<polyline clip-path="url(#clip660)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  314.996,1401.58 2352.76,1401.58 
  "/>
<polyline clip-path="url(#clip660)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  372.668,1401.58 372.668,1382.68 
  "/>
<polyline clip-path="url(#clip660)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  853.272,1401.58 853.272,1382.68 
  "/>
<polyline clip-path="url(#clip660)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1333.88,1401.58 1333.88,1382.68 
  "/>
<polyline clip-path="url(#clip660)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1814.48,1401.58 1814.48,1382.68 
  "/>
<polyline clip-path="url(#clip660)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  2295.08,1401.58 2295.08,1382.68 
  "/>
<path clip-path="url(#clip660)" d="M372.668 1438.36 Q367.252 1438.36 364.509 1443.71 Q361.8 1449.02 361.8 1459.71 Q361.8 1470.37 364.509 1475.72 Q367.252 1481.03 372.668 1481.03 Q378.12 1481.03 380.828 1475.72 Q383.571 1470.37 383.571 1459.71 Q383.571 1449.02 380.828 1443.71 Q378.12 1438.36 372.668 1438.36 M372.668 1432.8 Q381.384 1432.8 385.967 1439.71 Q390.585 1446.59 390.585 1459.71 Q390.585 1472.8 385.967 1479.71 Q381.384 1486.59 372.668 1486.59 Q363.953 1486.59 359.335 1479.71 Q354.752 1472.8 354.752 1459.71 Q354.752 1446.59 359.335 1439.71 Q363.953 1432.8 372.668 1432.8 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M838.845 1479.68 L850.303 1479.68 L850.303 1440.13 L837.838 1442.63 L837.838 1436.24 L850.234 1433.74 L857.248 1433.74 L857.248 1479.68 L868.706 1479.68 L868.706 1485.58 L838.845 1485.58 L838.845 1479.68 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M1325.86 1479.68 L1350.33 1479.68 L1350.33 1485.58 L1317.42 1485.58 L1317.42 1479.68 Q1321.41 1475.55 1328.29 1468.6 Q1335.2 1461.62 1336.97 1459.61 Q1340.33 1455.82 1341.65 1453.22 Q1343.01 1450.58 1343.01 1448.05 Q1343.01 1443.91 1340.09 1441.31 Q1337.21 1438.71 1332.56 1438.71 Q1329.26 1438.71 1325.58 1439.85 Q1321.93 1441 1317.76 1443.32 L1317.76 1436.24 Q1322 1434.54 1325.68 1433.67 Q1329.36 1432.8 1332.42 1432.8 Q1340.47 1432.8 1345.26 1436.83 Q1350.06 1440.86 1350.06 1447.59 Q1350.06 1450.79 1348.84 1453.67 Q1347.66 1456.52 1344.5 1460.41 Q1343.63 1461.41 1338.98 1466.24 Q1334.33 1471.03 1325.86 1479.68 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M1820.85 1457.63 Q1825.89 1458.7 1828.7 1462.11 Q1831.55 1465.51 1831.55 1470.51 Q1831.55 1478.18 1826.27 1482.39 Q1820.99 1486.59 1811.27 1486.59 Q1808 1486.59 1804.53 1485.93 Q1801.09 1485.3 1797.41 1484.02 L1797.41 1477.25 Q1800.33 1478.95 1803.8 1479.82 Q1807.27 1480.68 1811.06 1480.68 Q1817.66 1480.68 1821.09 1478.08 Q1824.57 1475.48 1824.57 1470.51 Q1824.57 1465.93 1821.34 1463.36 Q1818.14 1460.75 1812.41 1460.75 L1806.37 1460.75 L1806.37 1454.99 L1812.69 1454.99 Q1817.87 1454.99 1820.61 1452.94 Q1823.35 1450.86 1823.35 1446.97 Q1823.35 1442.98 1820.5 1440.86 Q1817.69 1438.71 1812.41 1438.71 Q1809.53 1438.71 1806.23 1439.33 Q1802.93 1439.96 1798.98 1441.27 L1798.98 1435.02 Q1802.97 1433.91 1806.44 1433.36 Q1809.95 1432.8 1813.04 1432.8 Q1821.02 1432.8 1825.68 1436.45 Q1830.33 1440.06 1830.33 1446.24 Q1830.33 1450.55 1827.87 1453.53 Q1825.4 1456.48 1820.85 1457.63 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M2299.6 1439.85 L2281.89 1467.52 L2299.6 1467.52 L2299.6 1439.85 M2297.76 1433.74 L2306.58 1433.74 L2306.58 1467.52 L2313.97 1467.52 L2313.97 1473.36 L2306.58 1473.36 L2306.58 1485.58 L2299.6 1485.58 L2299.6 1473.36 L2276.19 1473.36 L2276.19 1466.59 L2297.76 1433.74 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M1277.49 1527.62 L1277.49 1538.66 L1290.65 1538.66 L1290.65 1543.62 L1277.49 1543.62 L1277.49 1564.74 Q1277.49 1569.49 1278.77 1570.85 Q1280.09 1572.2 1284.08 1572.2 L1290.65 1572.2 L1290.65 1577.55 L1284.08 1577.55 Q1276.69 1577.55 1273.88 1574.8 Q1271.06 1572.03 1271.06 1564.74 L1271.06 1543.62 L1266.38 1543.62 L1266.38 1538.66 L1271.06 1538.66 L1271.06 1527.62 L1277.49 1527.62 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M1321.06 1523.52 L1335.79 1523.52 L1335.79 1528.49 L1327.45 1528.49 L1327.45 1581.96 L1335.79 1581.96 L1335.79 1586.92 L1321.06 1586.92 L1321.06 1523.52 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M1374.19 1539.8 L1374.19 1545.85 Q1371.48 1544.46 1368.56 1543.76 Q1365.65 1543.07 1362.52 1543.07 Q1357.76 1543.07 1355.37 1544.53 Q1353.01 1545.99 1353.01 1548.9 Q1353.01 1551.12 1354.71 1552.41 Q1356.41 1553.66 1361.55 1554.8 L1363.74 1555.29 Q1370.54 1556.75 1373.39 1559.42 Q1376.27 1562.06 1376.27 1566.82 Q1376.27 1572.24 1371.97 1575.4 Q1367.7 1578.55 1360.2 1578.55 Q1357.07 1578.55 1353.67 1577.93 Q1350.3 1577.34 1346.55 1576.12 L1346.55 1569.53 Q1350.09 1571.37 1353.53 1572.3 Q1356.97 1573.21 1360.33 1573.21 Q1364.85 1573.21 1367.28 1571.68 Q1369.71 1570.12 1369.71 1567.3 Q1369.71 1564.7 1367.94 1563.31 Q1366.2 1561.92 1360.26 1560.64 L1358.04 1560.12 Q1352.11 1558.87 1349.47 1556.3 Q1346.83 1553.69 1346.83 1549.18 Q1346.83 1543.69 1350.72 1540.71 Q1354.6 1537.72 1361.76 1537.72 Q1365.3 1537.72 1368.42 1538.24 Q1371.55 1538.76 1374.19 1539.8 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M1401.38 1523.52 L1401.38 1586.92 L1386.65 1586.92 L1386.65 1581.96 L1394.95 1581.96 L1394.95 1528.49 L1386.65 1528.49 L1386.65 1523.52 L1401.38 1523.52 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip662)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  314.996,1352.59 2352.76,1352.59 
  "/>
<polyline clip-path="url(#clip662)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  314.996,1155.93 2352.76,1155.93 
  "/>
<polyline clip-path="url(#clip662)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  314.996,959.265 2352.76,959.265 
  "/>
<polyline clip-path="url(#clip662)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  314.996,762.603 2352.76,762.603 
  "/>
<polyline clip-path="url(#clip662)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  314.996,565.941 2352.76,565.941 
  "/>
<polyline clip-path="url(#clip662)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  314.996,369.278 2352.76,369.278 
  "/>
<polyline clip-path="url(#clip662)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  314.996,172.616 2352.76,172.616 
  "/>
<polyline clip-path="url(#clip660)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  314.996,1401.58 314.996,123.472 
  "/>
<polyline clip-path="url(#clip660)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  314.996,1352.59 333.894,1352.59 
  "/>
<polyline clip-path="url(#clip660)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  314.996,1155.93 333.894,1155.93 
  "/>
<polyline clip-path="url(#clip660)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  314.996,959.265 333.894,959.265 
  "/>
<polyline clip-path="url(#clip660)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  314.996,762.603 333.894,762.603 
  "/>
<polyline clip-path="url(#clip660)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  314.996,565.941 333.894,565.941 
  "/>
<polyline clip-path="url(#clip660)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  314.996,369.278 333.894,369.278 
  "/>
<polyline clip-path="url(#clip660)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  314.996,172.616 333.894,172.616 
  "/>
<path clip-path="url(#clip660)" d="M120.073 1353.27 L164.587 1353.27 L164.587 1359.17 L120.073 1359.17 L120.073 1353.27 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M180.941 1372.61 L192.399 1372.61 L192.399 1333.06 L179.934 1335.56 L179.934 1329.17 L192.33 1326.67 L199.344 1326.67 L199.344 1372.61 L210.802 1372.61 L210.802 1378.51 L180.941 1378.51 L180.941 1372.61 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M224.968 1369.69 L232.295 1369.69 L232.295 1378.51 L224.968 1378.51 L224.968 1369.69 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M247.642 1326.67 L275.177 1326.67 L275.177 1332.57 L254.066 1332.57 L254.066 1345.28 Q255.593 1344.76 257.121 1344.52 Q258.649 1344.24 260.177 1344.24 Q268.857 1344.24 273.927 1349 Q278.996 1353.75 278.996 1361.88 Q278.996 1370.25 273.788 1374.9 Q268.579 1379.52 259.1 1379.52 Q255.836 1379.52 252.434 1378.96 Q249.066 1378.41 245.454 1377.29 L245.454 1370.25 Q248.579 1371.95 251.913 1372.78 Q255.246 1373.61 258.961 1373.61 Q264.968 1373.61 268.475 1370.45 Q271.982 1367.29 271.982 1361.88 Q271.982 1356.46 268.475 1353.3 Q264.968 1350.14 258.961 1350.14 Q256.149 1350.14 253.336 1350.77 Q250.559 1351.39 247.642 1352.71 L247.642 1326.67 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M118.58 1156.6 L163.094 1156.6 L163.094 1162.51 L118.58 1162.51 L118.58 1156.6 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M179.448 1175.94 L190.906 1175.94 L190.906 1136.4 L178.441 1138.9 L178.441 1132.51 L190.837 1130.01 L197.851 1130.01 L197.851 1175.94 L209.309 1175.94 L209.309 1181.85 L179.448 1181.85 L179.448 1175.94 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M223.475 1173.03 L230.802 1173.03 L230.802 1181.85 L223.475 1181.85 L223.475 1173.03 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M261.079 1134.63 Q255.663 1134.63 252.92 1139.97 Q250.211 1145.28 250.211 1155.98 Q250.211 1166.64 252.92 1171.99 Q255.663 1177.3 261.079 1177.3 Q266.531 1177.3 269.239 1171.99 Q271.982 1166.64 271.982 1155.98 Q271.982 1145.28 269.239 1139.97 Q266.531 1134.63 261.079 1134.63 M261.079 1129.07 Q269.795 1129.07 274.378 1135.98 Q278.996 1142.85 278.996 1155.98 Q278.996 1169.07 274.378 1175.98 Q269.795 1182.85 261.079 1182.85 Q252.364 1182.85 247.746 1175.98 Q243.163 1169.07 243.163 1155.98 Q243.163 1142.85 247.746 1135.98 Q252.364 1129.07 261.079 1129.07 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M120.073 959.942 L164.587 959.942 L164.587 965.845 L120.073 965.845 L120.073 959.942 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M194.726 937.963 Q189.309 937.963 186.566 943.31 Q183.858 948.623 183.858 959.317 Q183.858 969.977 186.566 975.324 Q189.309 980.636 194.726 980.636 Q200.177 980.636 202.885 975.324 Q205.628 969.977 205.628 959.317 Q205.628 948.623 202.885 943.31 Q200.177 937.963 194.726 937.963 M194.726 932.407 Q203.441 932.407 208.024 939.317 Q212.642 946.192 212.642 959.317 Q212.642 972.407 208.024 979.317 Q203.441 986.192 194.726 986.192 Q186.01 986.192 181.392 979.317 Q176.809 972.407 176.809 959.317 Q176.809 946.192 181.392 939.317 Q186.01 932.407 194.726 932.407 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M224.968 976.366 L232.295 976.366 L232.295 985.185 L224.968 985.185 L224.968 976.366 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M247.642 933.345 L275.177 933.345 L275.177 939.248 L254.066 939.248 L254.066 951.956 Q255.593 951.435 257.121 951.192 Q258.649 950.914 260.177 950.914 Q268.857 950.914 273.927 955.671 Q278.996 960.428 278.996 968.553 Q278.996 976.921 273.788 981.574 Q268.579 986.192 259.1 986.192 Q255.836 986.192 252.434 985.636 Q249.066 985.081 245.454 983.97 L245.454 976.921 Q248.579 978.622 251.913 979.456 Q255.246 980.289 258.961 980.289 Q264.968 980.289 268.475 977.129 Q271.982 973.97 271.982 968.553 Q271.982 963.136 268.475 959.977 Q264.968 956.817 258.961 956.817 Q256.149 956.817 253.336 957.442 Q250.559 958.067 247.642 959.386 L247.642 933.345 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M193.233 741.301 Q187.816 741.301 185.073 746.648 Q182.365 751.96 182.365 762.655 Q182.365 773.315 185.073 778.662 Q187.816 783.974 193.233 783.974 Q198.684 783.974 201.392 778.662 Q204.135 773.315 204.135 762.655 Q204.135 751.96 201.392 746.648 Q198.684 741.301 193.233 741.301 M193.233 735.745 Q201.948 735.745 206.531 742.655 Q211.149 749.53 211.149 762.655 Q211.149 775.745 206.531 782.655 Q201.948 789.53 193.233 789.53 Q184.517 789.53 179.899 782.655 Q175.316 775.745 175.316 762.655 Q175.316 749.53 179.899 742.655 Q184.517 735.745 193.233 735.745 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M223.475 779.703 L230.802 779.703 L230.802 788.523 L223.475 788.523 L223.475 779.703 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M261.079 741.301 Q255.663 741.301 252.92 746.648 Q250.211 751.96 250.211 762.655 Q250.211 773.315 252.92 778.662 Q255.663 783.974 261.079 783.974 Q266.531 783.974 269.239 778.662 Q271.982 773.315 271.982 762.655 Q271.982 751.96 269.239 746.648 Q266.531 741.301 261.079 741.301 M261.079 735.745 Q269.795 735.745 274.378 742.655 Q278.996 749.53 278.996 762.655 Q278.996 775.745 274.378 782.655 Q269.795 789.53 261.079 789.53 Q252.364 789.53 247.746 782.655 Q243.163 775.745 243.163 762.655 Q243.163 749.53 247.746 742.655 Q252.364 735.745 261.079 735.745 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M194.726 544.639 Q189.309 544.639 186.566 549.986 Q183.858 555.298 183.858 565.993 Q183.858 576.652 186.566 582 Q189.309 587.312 194.726 587.312 Q200.177 587.312 202.885 582 Q205.628 576.652 205.628 565.993 Q205.628 555.298 202.885 549.986 Q200.177 544.639 194.726 544.639 M194.726 539.083 Q203.441 539.083 208.024 545.993 Q212.642 552.868 212.642 565.993 Q212.642 579.083 208.024 585.993 Q203.441 592.868 194.726 592.868 Q186.01 592.868 181.392 585.993 Q176.809 579.083 176.809 565.993 Q176.809 552.868 181.392 545.993 Q186.01 539.083 194.726 539.083 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M224.968 583.041 L232.295 583.041 L232.295 591.861 L224.968 591.861 L224.968 583.041 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M247.642 540.021 L275.177 540.021 L275.177 545.923 L254.066 545.923 L254.066 558.632 Q255.593 558.111 257.121 557.868 Q258.649 557.59 260.177 557.59 Q268.857 557.59 273.927 562.347 Q278.996 567.104 278.996 575.229 Q278.996 583.597 273.788 588.25 Q268.579 592.868 259.1 592.868 Q255.836 592.868 252.434 592.312 Q249.066 591.756 245.454 590.645 L245.454 583.597 Q248.579 585.298 251.913 586.131 Q255.246 586.965 258.961 586.965 Q264.968 586.965 268.475 583.805 Q271.982 580.645 271.982 575.229 Q271.982 569.812 268.475 566.652 Q264.968 563.493 258.961 563.493 Q256.149 563.493 253.336 564.118 Q250.559 564.743 247.642 566.062 L247.642 540.021 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M179.448 389.296 L190.906 389.296 L190.906 349.747 L178.441 352.247 L178.441 345.858 L190.837 343.358 L197.851 343.358 L197.851 389.296 L209.309 389.296 L209.309 395.198 L179.448 395.198 L179.448 389.296 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M223.475 386.379 L230.802 386.379 L230.802 395.198 L223.475 395.198 L223.475 386.379 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M261.079 347.976 Q255.663 347.976 252.92 353.324 Q250.211 358.636 250.211 369.331 Q250.211 379.99 252.92 385.337 Q255.663 390.65 261.079 390.65 Q266.531 390.65 269.239 385.337 Q271.982 379.99 271.982 369.331 Q271.982 358.636 269.239 353.324 Q266.531 347.976 261.079 347.976 M261.079 342.421 Q269.795 342.421 274.378 349.331 Q278.996 356.206 278.996 369.331 Q278.996 382.421 274.378 389.33 Q269.795 396.205 261.079 396.205 Q252.364 396.205 247.746 389.33 Q243.163 382.421 243.163 369.331 Q243.163 356.206 247.746 349.331 Q252.364 342.421 261.079 342.421 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M180.941 192.634 L192.399 192.634 L192.399 153.085 L179.934 155.585 L179.934 149.196 L192.33 146.696 L199.344 146.696 L199.344 192.634 L210.802 192.634 L210.802 198.536 L180.941 198.536 L180.941 192.634 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M224.968 189.717 L232.295 189.717 L232.295 198.536 L224.968 198.536 L224.968 189.717 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M247.642 146.696 L275.177 146.696 L275.177 152.599 L254.066 152.599 L254.066 165.307 Q255.593 164.786 257.121 164.543 Q258.649 164.266 260.177 164.266 Q268.857 164.266 273.927 169.023 Q278.996 173.779 278.996 181.904 Q278.996 190.272 273.788 194.925 Q268.579 199.543 259.1 199.543 Q255.836 199.543 252.434 198.988 Q249.066 198.432 245.454 197.321 L245.454 190.272 Q248.579 191.974 251.913 192.807 Q255.246 193.64 258.961 193.64 Q264.968 193.64 268.475 190.481 Q271.982 187.321 271.982 181.904 Q271.982 176.488 268.475 173.328 Q264.968 170.168 258.961 170.168 Q256.149 170.168 253.336 170.793 Q250.559 171.418 247.642 172.738 L247.642 146.696 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M35.6047 1073.64 Q31.2992 1071.24 29.2506 1067.91 Q27.202 1064.57 27.202 1060.06 Q27.202 1053.98 31.4728 1050.68 Q35.7089 1047.39 43.5561 1047.39 L67.0282 1047.39 L67.0282 1053.81 L43.7644 1053.81 Q38.1741 1053.81 35.4658 1055.79 Q32.7575 1057.77 32.7575 1061.83 Q32.7575 1066.8 36.0561 1069.68 Q39.3547 1072.56 45.0491 1072.56 L67.0282 1072.56 L67.0282 1078.98 L43.7644 1078.98 Q38.1394 1078.98 35.4658 1080.96 Q32.7575 1082.94 32.7575 1087.07 Q32.7575 1091.97 36.0908 1094.85 Q39.3894 1097.73 45.0491 1097.73 L67.0282 1097.73 L67.0282 1104.16 L28.1395 1104.16 L28.1395 1097.73 L34.1811 1097.73 Q30.6047 1095.55 28.9034 1092.49 Q27.202 1089.43 27.202 1085.23 Q27.202 1081 29.3547 1078.05 Q31.5075 1075.06 35.6047 1073.64 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M47.4797 1016.97 Q47.4797 1024.71 49.2505 1027.7 Q51.0213 1030.68 55.2921 1030.68 Q58.6949 1030.68 60.7087 1028.46 Q62.6879 1026.21 62.6879 1022.35 Q62.6879 1017.04 58.9379 1013.84 Q55.1532 1010.61 48.9033 1010.61 L47.4797 1010.61 L47.4797 1016.97 M44.8408 1004.23 L67.0282 1004.23 L67.0282 1010.61 L61.1254 1010.61 Q64.6671 1012.8 66.3684 1016.07 Q68.0351 1019.33 68.0351 1024.05 Q68.0351 1030.02 64.7018 1033.57 Q61.3337 1037.07 55.7088 1037.07 Q49.1463 1037.07 45.813 1032.7 Q42.4797 1028.29 42.4797 1019.57 L42.4797 1010.61 L41.8547 1010.61 Q37.445 1010.61 35.0492 1013.53 Q32.6186 1016.41 32.6186 1021.66 Q32.6186 1024.99 33.4172 1028.15 Q34.2158 1031.31 35.813 1034.23 L29.9103 1034.23 Q28.5561 1030.72 27.8964 1027.42 Q27.202 1024.12 27.202 1021 Q27.202 1012.56 31.577 1008.39 Q35.9519 1004.23 44.8408 1004.23 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M29.2853 966.275 L35.3269 966.275 Q33.9381 968.983 33.2436 971.9 Q32.5492 974.817 32.5492 977.941 Q32.5492 982.698 34.0075 985.094 Q35.4658 987.455 38.3825 987.455 Q40.6047 987.455 41.8894 985.754 Q43.1394 984.053 44.2852 978.914 L44.7713 976.726 Q46.2297 969.921 48.9033 967.073 Q51.5421 964.192 56.299 964.192 Q61.7157 964.192 64.8754 968.497 Q68.0351 972.768 68.0351 980.268 Q68.0351 983.393 67.4101 986.796 Q66.8198 990.164 65.6046 993.914 L59.0074 993.914 Q60.8476 990.372 61.7851 986.935 Q62.6879 983.497 62.6879 980.129 Q62.6879 975.615 61.1601 973.185 Q59.5976 970.754 56.7852 970.754 Q54.181 970.754 52.7921 972.525 Q51.4032 974.261 50.1185 980.198 L49.5977 982.421 Q48.3477 988.358 45.7783 990.997 Q43.1741 993.636 38.6603 993.636 Q33.1742 993.636 30.1881 989.747 Q27.202 985.858 27.202 978.705 Q27.202 975.164 27.7228 972.039 Q28.2436 968.914 29.2853 966.275 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M29.2853 929.226 L35.3269 929.226 Q33.9381 931.935 33.2436 934.851 Q32.5492 937.768 32.5492 940.893 Q32.5492 945.65 34.0075 948.046 Q35.4658 950.407 38.3825 950.407 Q40.6047 950.407 41.8894 948.706 Q43.1394 947.004 44.2852 941.865 L44.7713 939.678 Q46.2297 932.872 48.9033 930.025 Q51.5421 927.143 56.299 927.143 Q61.7157 927.143 64.8754 931.449 Q68.0351 935.72 68.0351 943.219 Q68.0351 946.344 67.4101 949.747 Q66.8198 953.115 65.6046 956.865 L59.0074 956.865 Q60.8476 953.324 61.7851 949.886 Q62.6879 946.449 62.6879 943.081 Q62.6879 938.567 61.1601 936.136 Q59.5976 933.706 56.7852 933.706 Q54.181 933.706 52.7921 935.476 Q51.4032 937.213 50.1185 943.15 L49.5977 945.372 Q48.3477 951.31 45.7783 953.949 Q43.1741 956.587 38.6603 956.587 Q33.1742 956.587 30.1881 952.699 Q27.202 948.81 27.202 941.657 Q27.202 938.115 27.7228 934.99 Q28.2436 931.865 29.2853 929.226 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M28.1395 898.949 L28.1395 892.178 L60.7782 880.025 L28.1395 867.873 L28.1395 861.102 L67.0282 875.685 L67.0282 884.366 L28.1395 898.949 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M45.9866 819.019 L49.1116 819.019 L49.1116 848.394 Q55.7088 847.977 59.181 844.435 Q62.6185 840.859 62.6185 834.505 Q62.6185 830.824 61.7157 827.387 Q60.8129 823.915 59.0074 820.512 L65.049 820.512 Q66.5073 823.949 67.2712 827.56 Q68.0351 831.171 68.0351 834.887 Q68.0351 844.192 62.6185 849.644 Q57.2018 855.06 47.9658 855.06 Q38.4172 855.06 32.827 849.921 Q27.202 844.748 27.202 835.998 Q27.202 828.151 32.2714 823.602 Q37.3061 819.019 45.9866 819.019 M44.1116 825.408 Q38.8686 825.477 35.7436 828.359 Q32.6186 831.206 32.6186 835.928 Q32.6186 841.276 35.6394 844.505 Q38.6603 847.699 44.1463 848.185 L44.1116 825.408 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M13.0007 808.533 L13.0007 802.144 L67.0282 802.144 L67.0282 808.533 L13.0007 808.533 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M32.6186 773.706 Q32.6186 778.845 36.6464 781.831 Q40.6394 784.818 47.6185 784.818 Q54.5977 784.818 58.6254 781.866 Q62.6185 778.88 62.6185 773.706 Q62.6185 768.602 58.5907 765.616 Q54.5629 762.63 47.6185 762.63 Q40.7089 762.63 36.6811 765.616 Q32.6186 768.602 32.6186 773.706 M27.202 773.706 Q27.202 765.373 32.6186 760.616 Q38.0353 755.859 47.6185 755.859 Q57.1671 755.859 62.6185 760.616 Q68.0351 765.373 68.0351 773.706 Q68.0351 782.074 62.6185 786.831 Q57.1671 791.554 47.6185 791.554 Q38.0353 791.554 32.6186 786.831 Q27.202 782.074 27.202 773.706 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M29.6325 717.283 L35.6047 717.283 Q34.1117 719.991 33.3825 722.735 Q32.6186 725.443 32.6186 728.221 Q32.6186 734.436 36.5769 737.873 Q40.5005 741.311 47.6185 741.311 Q54.7366 741.311 58.6949 737.873 Q62.6185 734.436 62.6185 728.221 Q62.6185 725.443 61.8893 722.735 Q61.1254 719.991 59.6324 717.283 L65.5351 717.283 Q66.7851 719.957 67.4101 722.839 Q68.0351 725.686 68.0351 728.915 Q68.0351 737.7 62.5143 742.873 Q56.9935 748.047 47.6185 748.047 Q38.1047 748.047 32.6533 742.839 Q27.202 737.596 27.202 728.498 Q27.202 725.547 27.827 722.735 Q28.4173 719.922 29.6325 717.283 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M28.1395 706.172 L28.1395 699.783 L67.0282 699.783 L67.0282 706.172 L28.1395 706.172 M13.0007 706.172 L13.0007 699.783 L21.0909 699.783 L21.0909 706.172 L13.0007 706.172 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M17.0979 680.096 L28.1395 680.096 L28.1395 666.936 L33.1047 666.936 L33.1047 680.096 L54.2157 680.096 Q58.9726 680.096 60.3268 678.811 Q61.681 677.492 61.681 673.499 L61.681 666.936 L67.0282 666.936 L67.0282 673.499 Q67.0282 680.894 64.2851 683.707 Q61.5074 686.519 54.2157 686.519 L33.1047 686.519 L33.1047 691.207 L28.1395 691.207 L28.1395 686.519 L17.0979 686.519 L17.0979 680.096 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M70.6392 642.353 Q77.5837 645.061 79.7017 647.631 Q81.8197 650.2 81.8197 654.506 L81.8197 659.61 L76.4726 659.61 L76.4726 655.86 Q76.4726 653.221 75.2226 651.763 Q73.9726 650.304 69.3198 648.534 L66.4032 647.388 L28.1395 663.117 L28.1395 656.346 L58.556 644.193 L28.1395 632.041 L28.1395 625.27 L70.6392 642.353 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M13.0007 594.437 L13.0007 579.714 L17.9659 579.714 L17.9659 588.048 L71.4379 588.048 L71.4379 579.714 L76.4031 579.714 L76.4031 594.437 L13.0007 594.437 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M35.6047 535.826 Q31.2992 533.43 29.2506 530.097 Q27.202 526.763 27.202 522.249 Q27.202 516.173 31.4728 512.875 Q35.7089 509.576 43.5561 509.576 L67.0282 509.576 L67.0282 516 L43.7644 516 Q38.1741 516 35.4658 517.979 Q32.7575 519.958 32.7575 524.02 Q32.7575 528.986 36.0561 531.867 Q39.3547 534.749 45.0491 534.749 L67.0282 534.749 L67.0282 541.173 L43.7644 541.173 Q38.1394 541.173 35.4658 543.152 Q32.7575 545.131 32.7575 549.263 Q32.7575 554.159 36.0908 557.041 Q39.3894 559.923 45.0491 559.923 L67.0282 559.923 L67.0282 566.346 L28.1395 566.346 L28.1395 559.923 L34.1811 559.923 Q30.6047 557.735 28.9034 554.68 Q27.202 551.624 27.202 547.423 Q27.202 543.187 29.3547 540.236 Q31.5075 537.249 35.6047 535.826 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M15.1882 485.479 L15.1882 479.576 L73.6253 497.632 L73.6253 503.534 L15.1882 485.479 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M29.2853 448.083 L35.3269 448.083 Q33.9381 450.792 33.2436 453.708 Q32.5492 456.625 32.5492 459.75 Q32.5492 464.507 34.0075 466.903 Q35.4658 469.264 38.3825 469.264 Q40.6047 469.264 41.8894 467.562 Q43.1394 465.861 44.2852 460.722 L44.7713 458.535 Q46.2297 451.729 48.9033 448.882 Q51.5421 446 56.299 446 Q61.7157 446 64.8754 450.305 Q68.0351 454.576 68.0351 462.076 Q68.0351 465.201 67.4101 468.604 Q66.8198 471.972 65.6046 475.722 L59.0074 475.722 Q60.8476 472.18 61.7851 468.743 Q62.6879 465.305 62.6879 461.937 Q62.6879 457.423 61.1601 454.993 Q59.5976 452.562 56.7852 452.562 Q54.181 452.562 52.7921 454.333 Q51.4032 456.069 50.1185 462.007 L49.5977 464.229 Q48.3477 470.166 45.7783 472.805 Q43.1741 475.444 38.6603 475.444 Q33.1742 475.444 30.1881 471.555 Q27.202 467.666 27.202 460.514 Q27.202 456.972 27.7228 453.847 Q28.2436 450.722 29.2853 448.083 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M13.0007 420.896 L76.4031 420.896 L76.4031 435.618 L71.4379 435.618 L71.4379 427.319 L17.9659 427.319 L17.9659 435.618 L13.0007 435.618 L13.0007 420.896 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M600.277 12.096 L612.471 12.096 L627.905 53.2532 L643.42 12.096 L655.613 12.096 L655.613 72.576 L647.633 72.576 L647.633 19.4686 L632.037 60.9499 L623.813 60.9499 L608.217 19.4686 L608.217 72.576 L600.277 72.576 L600.277 12.096 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M689.114 32.4315 Q683.118 32.4315 679.635 37.1306 Q676.151 41.7891 676.151 49.9314 Q676.151 58.0738 679.594 62.7728 Q683.078 67.4314 689.114 67.4314 Q695.069 67.4314 698.552 62.7323 Q702.036 58.0333 702.036 49.9314 Q702.036 41.8701 698.552 37.1711 Q695.069 32.4315 689.114 32.4315 M689.114 26.1121 Q698.836 26.1121 704.386 32.4315 Q709.935 38.7509 709.935 49.9314 Q709.935 61.0714 704.386 67.4314 Q698.836 73.7508 689.114 73.7508 Q679.351 73.7508 673.801 67.4314 Q668.292 61.0714 668.292 49.9314 Q668.292 38.7509 673.801 32.4315 Q679.351 26.1121 689.114 26.1121 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M752.146 34.0924 L752.146 9.54393 L759.599 9.54393 L759.599 72.576 L752.146 72.576 L752.146 65.7705 Q749.796 69.8214 746.191 71.8063 Q742.626 73.7508 737.603 73.7508 Q729.38 73.7508 724.195 67.1883 Q719.05 60.6258 719.05 49.9314 Q719.05 39.2371 724.195 32.6746 Q729.38 26.1121 737.603 26.1121 Q742.626 26.1121 746.191 28.0971 Q749.796 30.0415 752.146 34.0924 M726.747 49.9314 Q726.747 58.1548 730.109 62.8538 Q733.512 67.5124 739.426 67.5124 Q745.34 67.5124 748.743 62.8538 Q752.146 58.1548 752.146 49.9314 Q752.146 41.7081 748.743 37.0496 Q745.34 32.3505 739.426 32.3505 Q733.512 32.3505 730.109 37.0496 Q726.747 41.7081 726.747 49.9314 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M774.952 27.2059 L782.406 27.2059 L782.406 72.576 L774.952 72.576 L774.952 27.2059 M774.952 9.54393 L782.406 9.54393 L782.406 18.9825 L774.952 18.9825 L774.952 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M820.971 9.54393 L820.971 15.7418 L813.841 15.7418 Q809.831 15.7418 808.251 17.3622 Q806.712 18.9825 806.712 23.1955 L806.712 27.2059 L818.986 27.2059 L818.986 32.9987 L806.712 32.9987 L806.712 72.576 L799.217 72.576 L799.217 32.9987 L792.088 32.9987 L792.088 27.2059 L799.217 27.2059 L799.217 24.0462 Q799.217 16.471 802.742 13.0277 Q806.266 9.54393 813.922 9.54393 L820.971 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M827.209 27.2059 L834.663 27.2059 L834.663 72.576 L827.209 72.576 L827.209 27.2059 M827.209 9.54393 L834.663 9.54393 L834.663 18.9825 L827.209 18.9825 L827.209 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M889.066 48.0275 L889.066 51.6733 L854.796 51.6733 Q855.282 59.3701 859.414 63.421 Q863.586 67.4314 870.999 67.4314 Q875.293 67.4314 879.304 66.3781 Q883.355 65.3249 887.325 63.2184 L887.325 70.267 Q883.314 71.9684 879.101 72.8596 Q874.888 73.7508 870.554 73.7508 Q859.697 73.7508 853.337 67.4314 Q847.018 61.1119 847.018 50.3365 Q847.018 39.1965 853.013 32.6746 Q859.049 26.1121 869.257 26.1121 Q878.413 26.1121 883.719 32.0264 Q889.066 37.9003 889.066 48.0275 M881.613 45.84 Q881.532 39.7232 878.169 36.0774 Q874.848 32.4315 869.338 32.4315 Q863.1 32.4315 859.333 35.9558 Q855.606 39.4801 855.039 45.8805 L881.613 45.84 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M931.155 34.0924 L931.155 9.54393 L938.609 9.54393 L938.609 72.576 L931.155 72.576 L931.155 65.7705 Q928.806 69.8214 925.2 71.8063 Q921.636 73.7508 916.613 73.7508 Q908.389 73.7508 903.204 67.1883 Q898.059 60.6258 898.059 49.9314 Q898.059 39.2371 903.204 32.6746 Q908.389 26.1121 916.613 26.1121 Q921.636 26.1121 925.2 28.0971 Q928.806 30.0415 931.155 34.0924 M905.756 49.9314 Q905.756 58.1548 909.118 62.8538 Q912.521 67.5124 918.435 67.5124 Q924.35 67.5124 927.753 62.8538 Q931.155 58.1548 931.155 49.9314 Q931.155 41.7081 927.753 37.0496 Q924.35 32.3505 918.435 32.3505 Q912.521 32.3505 909.118 37.0496 Q905.756 41.7081 905.756 49.9314 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M955.866 62.2867 L964.413 62.2867 L964.413 72.576 L955.866 72.576 L955.866 62.2867 M955.866 29.6769 L964.413 29.6769 L964.413 39.9662 L955.866 39.9662 L955.866 29.6769 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M1008.61 12.096 L1020.8 12.096 L1036.24 53.2532 L1051.75 12.096 L1063.94 12.096 L1063.94 72.576 L1055.96 72.576 L1055.96 19.4686 L1040.37 60.9499 L1032.14 60.9499 L1016.55 19.4686 L1016.55 72.576 L1008.61 72.576 L1008.61 12.096 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M1100.48 49.7694 Q1091.45 49.7694 1087.97 51.8354 Q1084.48 53.9013 1084.48 58.8839 Q1084.48 62.8538 1087.07 65.2034 Q1089.71 67.5124 1094.2 67.5124 Q1100.4 67.5124 1104.13 63.1374 Q1107.9 58.7219 1107.9 51.4303 L1107.9 49.7694 L1100.48 49.7694 M1115.35 46.6907 L1115.35 72.576 L1107.9 72.576 L1107.9 65.6895 Q1105.34 69.8214 1101.54 71.8063 Q1097.73 73.7508 1092.22 73.7508 Q1085.25 73.7508 1081.12 69.8619 Q1077.03 65.9325 1077.03 59.3701 Q1077.03 51.7138 1082.13 47.825 Q1087.28 43.9361 1097.44 43.9361 L1107.9 43.9361 L1107.9 43.2069 Q1107.9 38.0623 1104.49 35.2672 Q1101.13 32.4315 1095.01 32.4315 Q1091.13 32.4315 1087.44 33.3632 Q1083.75 34.295 1080.35 36.1584 L1080.35 29.2718 Q1084.44 27.692 1088.29 26.9223 Q1092.14 26.1121 1095.78 26.1121 Q1105.63 26.1121 1110.49 31.2163 Q1115.35 36.3204 1115.35 46.6907 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M1159.63 28.5427 L1159.63 35.5912 Q1156.47 33.9709 1153.06 33.1607 Q1149.66 32.3505 1146.02 32.3505 Q1140.47 32.3505 1137.67 34.0519 Q1134.92 35.7533 1134.92 39.156 Q1134.92 41.7486 1136.9 43.2475 Q1138.89 44.7058 1144.88 46.0426 L1147.43 46.6097 Q1155.37 48.3111 1158.69 51.4303 Q1162.06 54.509 1162.06 60.0587 Q1162.06 66.3781 1157.03 70.0644 Q1152.05 73.7508 1143.3 73.7508 Q1139.66 73.7508 1135.69 73.0216 Q1131.76 72.3329 1127.38 70.9151 L1127.38 63.2184 Q1131.51 65.3654 1135.52 66.4591 Q1139.53 67.5124 1143.46 67.5124 Q1148.73 67.5124 1151.56 65.73 Q1154.4 63.9071 1154.4 60.6258 Q1154.4 57.5877 1152.33 55.9673 Q1150.31 54.3469 1143.38 52.8481 L1140.79 52.2405 Q1133.86 50.7821 1130.78 47.7845 Q1127.71 44.7463 1127.71 39.4801 Q1127.71 33.0797 1132.24 29.5959 Q1136.78 26.1121 1145.12 26.1121 Q1149.26 26.1121 1152.9 26.7198 Q1156.55 27.3274 1159.63 28.5427 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M1202.85 28.5427 L1202.85 35.5912 Q1199.69 33.9709 1196.29 33.1607 Q1192.88 32.3505 1189.24 32.3505 Q1183.69 32.3505 1180.89 34.0519 Q1178.14 35.7533 1178.14 39.156 Q1178.14 41.7486 1180.12 43.2475 Q1182.11 44.7058 1188.1 46.0426 L1190.66 46.6097 Q1198.6 48.3111 1201.92 51.4303 Q1205.28 54.509 1205.28 60.0587 Q1205.28 66.3781 1200.26 70.0644 Q1195.27 73.7508 1186.52 73.7508 Q1182.88 73.7508 1178.91 73.0216 Q1174.98 72.3329 1170.6 70.9151 L1170.6 63.2184 Q1174.74 65.3654 1178.75 66.4591 Q1182.76 67.5124 1186.69 67.5124 Q1191.95 67.5124 1194.79 65.73 Q1197.62 63.9071 1197.62 60.6258 Q1197.62 57.5877 1195.56 55.9673 Q1193.53 54.3469 1186.61 52.8481 L1184.01 52.2405 Q1177.09 50.7821 1174.01 47.7845 Q1170.93 44.7463 1170.93 39.4801 Q1170.93 33.0797 1175.47 29.5959 Q1180 26.1121 1188.35 26.1121 Q1192.48 26.1121 1196.12 26.7198 Q1199.77 27.3274 1202.85 28.5427 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M1238.17 27.2059 L1246.07 27.2059 L1260.25 65.2844 L1274.43 27.2059 L1282.33 27.2059 L1265.31 72.576 L1255.19 72.576 L1238.17 27.2059 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M1331.43 48.0275 L1331.43 51.6733 L1297.15 51.6733 Q1297.64 59.3701 1301.77 63.421 Q1305.94 67.4314 1313.36 67.4314 Q1317.65 67.4314 1321.66 66.3781 Q1325.71 65.3249 1329.68 63.2184 L1329.68 70.267 Q1325.67 71.9684 1321.46 72.8596 Q1317.25 73.7508 1312.91 73.7508 Q1302.06 73.7508 1295.7 67.4314 Q1289.38 61.1119 1289.38 50.3365 Q1289.38 39.1965 1295.37 32.6746 Q1301.41 26.1121 1311.62 26.1121 Q1320.77 26.1121 1326.08 32.0264 Q1331.43 37.9003 1331.43 48.0275 M1323.97 45.84 Q1323.89 39.7232 1320.53 36.0774 Q1317.21 32.4315 1311.7 32.4315 Q1305.46 32.4315 1301.69 35.9558 Q1297.96 39.4801 1297.4 45.8805 L1323.97 45.84 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M1343.66 9.54393 L1351.11 9.54393 L1351.11 72.576 L1343.66 72.576 L1343.66 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M1384.29 32.4315 Q1378.29 32.4315 1374.81 37.1306 Q1371.33 41.7891 1371.33 49.9314 Q1371.33 58.0738 1374.77 62.7728 Q1378.25 67.4314 1384.29 67.4314 Q1390.24 67.4314 1393.73 62.7323 Q1397.21 58.0333 1397.21 49.9314 Q1397.21 41.8701 1393.73 37.1711 Q1390.24 32.4315 1384.29 32.4315 M1384.29 26.1121 Q1394.01 26.1121 1399.56 32.4315 Q1405.11 38.7509 1405.11 49.9314 Q1405.11 61.0714 1399.56 67.4314 Q1394.01 73.7508 1384.29 73.7508 Q1374.53 73.7508 1368.98 67.4314 Q1363.47 61.0714 1363.47 49.9314 Q1363.47 38.7509 1368.98 32.4315 Q1374.53 26.1121 1384.29 26.1121 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M1450.12 28.9478 L1450.12 35.9153 Q1446.96 34.1734 1443.76 33.3227 Q1440.6 32.4315 1437.36 32.4315 Q1430.11 32.4315 1426.09 37.0496 Q1422.08 41.6271 1422.08 49.9314 Q1422.08 58.2358 1426.09 62.8538 Q1430.11 67.4314 1437.36 67.4314 Q1440.6 67.4314 1443.76 66.5807 Q1446.96 65.6895 1450.12 63.9476 L1450.12 70.8341 Q1447 72.2924 1443.64 73.0216 Q1440.31 73.7508 1436.55 73.7508 Q1426.3 73.7508 1420.26 67.3098 Q1414.23 60.8689 1414.23 49.9314 Q1414.23 38.832 1420.3 32.472 Q1426.42 26.1121 1437.03 26.1121 Q1440.48 26.1121 1443.76 26.8413 Q1447.04 27.5299 1450.12 28.9478 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M1463.08 27.2059 L1470.53 27.2059 L1470.53 72.576 L1463.08 72.576 L1463.08 27.2059 M1463.08 9.54393 L1470.53 9.54393 L1470.53 18.9825 L1463.08 18.9825 L1463.08 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M1493.5 14.324 L1493.5 27.2059 L1508.85 27.2059 L1508.85 32.9987 L1493.5 32.9987 L1493.5 57.6282 Q1493.5 63.1779 1495 64.7578 Q1496.54 66.3376 1501.2 66.3376 L1508.85 66.3376 L1508.85 72.576 L1501.2 72.576 Q1492.57 72.576 1489.29 69.3758 Q1486.01 66.1351 1486.01 57.6282 L1486.01 32.9987 L1480.54 32.9987 L1480.54 27.2059 L1486.01 27.2059 L1486.01 14.324 L1493.5 14.324 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M1537.54 76.7889 Q1534.38 84.8907 1531.38 87.3618 Q1528.38 89.8329 1523.36 89.8329 L1517.4 89.8329 L1517.4 83.5945 L1521.78 83.5945 Q1524.86 83.5945 1526.56 82.1361 Q1528.26 80.6778 1530.32 75.2496 L1531.66 71.8468 L1513.31 27.2059 L1521.21 27.2059 L1535.39 62.6918 L1549.57 27.2059 L1557.47 27.2059 L1537.54 76.7889 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M1614.75 49.7694 Q1605.71 49.7694 1602.23 51.8354 Q1598.74 53.9013 1598.74 58.8839 Q1598.74 62.8538 1601.34 65.2034 Q1603.97 67.5124 1608.47 67.5124 Q1614.66 67.5124 1618.39 63.1374 Q1622.16 58.7219 1622.16 51.4303 L1622.16 49.7694 L1614.75 49.7694 M1629.61 46.6907 L1629.61 72.576 L1622.16 72.576 L1622.16 65.6895 Q1619.61 69.8214 1615.8 71.8063 Q1611.99 73.7508 1606.48 73.7508 Q1599.51 73.7508 1595.38 69.8619 Q1591.29 65.9325 1591.29 59.3701 Q1591.29 51.7138 1596.39 47.825 Q1601.54 43.9361 1611.71 43.9361 L1622.16 43.9361 L1622.16 43.2069 Q1622.16 38.0623 1618.76 35.2672 Q1615.39 32.4315 1609.28 32.4315 Q1605.39 32.4315 1601.7 33.3632 Q1598.02 34.295 1594.61 36.1584 L1594.61 29.2718 Q1598.7 27.692 1602.55 26.9223 Q1606.4 26.1121 1610.05 26.1121 Q1619.89 26.1121 1624.75 31.2163 Q1629.61 36.3204 1629.61 46.6907 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M1667.93 9.54393 L1667.93 15.7418 L1660.8 15.7418 Q1656.79 15.7418 1655.21 17.3622 Q1653.67 18.9825 1653.67 23.1955 L1653.67 27.2059 L1665.95 27.2059 L1665.95 32.9987 L1653.67 32.9987 L1653.67 72.576 L1646.18 72.576 L1646.18 32.9987 L1639.05 32.9987 L1639.05 27.2059 L1646.18 27.2059 L1646.18 24.0462 Q1646.18 16.471 1649.7 13.0277 Q1653.23 9.54393 1660.89 9.54393 L1667.93 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M1680.09 14.324 L1680.09 27.2059 L1695.44 27.2059 L1695.44 32.9987 L1680.09 32.9987 L1680.09 57.6282 Q1680.09 63.1779 1681.59 64.7578 Q1683.12 66.3376 1687.78 66.3376 L1695.44 66.3376 L1695.44 72.576 L1687.78 72.576 Q1679.15 72.576 1675.87 69.3758 Q1672.59 66.1351 1672.59 57.6282 L1672.59 32.9987 L1667.12 32.9987 L1667.12 27.2059 L1672.59 27.2059 L1672.59 14.324 L1680.09 14.324 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M1744.05 48.0275 L1744.05 51.6733 L1709.78 51.6733 Q1710.27 59.3701 1714.4 63.421 Q1718.57 67.4314 1725.98 67.4314 Q1730.28 67.4314 1734.29 66.3781 Q1738.34 65.3249 1742.31 63.2184 L1742.31 70.267 Q1738.3 71.9684 1734.09 72.8596 Q1729.87 73.7508 1725.54 73.7508 Q1714.68 73.7508 1708.32 67.4314 Q1702 61.1119 1702 50.3365 Q1702 39.1965 1708 32.6746 Q1714.03 26.1121 1724.24 26.1121 Q1733.4 26.1121 1738.7 32.0264 Q1744.05 37.9003 1744.05 48.0275 M1736.6 45.84 Q1736.52 39.7232 1733.15 36.0774 Q1729.83 32.4315 1724.32 32.4315 Q1718.08 32.4315 1714.32 35.9558 Q1710.59 39.4801 1710.02 45.8805 L1736.6 45.84 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M1782.57 34.1734 Q1781.32 33.4443 1779.82 33.1202 Q1778.36 32.7556 1776.58 32.7556 Q1770.26 32.7556 1766.86 36.8875 Q1763.49 40.9789 1763.49 48.6757 L1763.49 72.576 L1756 72.576 L1756 27.2059 L1763.49 27.2059 L1763.49 34.2544 Q1765.84 30.1225 1769.61 28.1376 Q1773.38 26.1121 1778.77 26.1121 Q1779.54 26.1121 1780.47 26.2337 Q1781.4 26.3147 1782.53 26.5172 L1782.57 34.1734 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M1845.77 44.2197 Q1848.4 45.1109 1850.87 48.0275 Q1853.38 50.9442 1855.9 56.0483 L1864.2 72.576 L1855.41 72.576 L1847.67 57.061 Q1844.67 50.9847 1841.84 48.9997 Q1839.04 47.0148 1834.18 47.0148 L1825.27 47.0148 L1825.27 72.576 L1817.09 72.576 L1817.09 12.096 L1835.56 12.096 Q1845.93 12.096 1851.03 16.4305 Q1856.14 20.7649 1856.14 29.5149 Q1856.14 35.2267 1853.47 38.994 Q1850.83 42.7613 1845.77 44.2197 M1825.27 18.8205 L1825.27 40.2903 L1835.56 40.2903 Q1841.47 40.2903 1844.47 37.5762 Q1847.51 34.8216 1847.51 29.5149 Q1847.51 24.2082 1844.47 21.5346 Q1841.47 18.8205 1835.56 18.8205 L1825.27 18.8205 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M1869.91 54.671 L1869.91 27.2059 L1877.37 27.2059 L1877.37 54.3874 Q1877.37 60.8284 1879.88 64.0691 Q1882.39 67.2693 1887.41 67.2693 Q1893.45 67.2693 1896.93 63.421 Q1900.46 59.5726 1900.46 52.9291 L1900.46 27.2059 L1907.91 27.2059 L1907.91 72.576 L1900.46 72.576 L1900.46 65.6084 Q1897.74 69.7404 1894.14 71.7658 Q1890.57 73.7508 1885.83 73.7508 Q1878.01 73.7508 1873.96 68.8897 Q1869.91 64.0286 1869.91 54.671 M1888.67 26.1121 L1888.67 26.1121 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M1960.98 45.1919 L1960.98 72.576 L1953.52 72.576 L1953.52 45.4349 Q1953.52 38.994 1951.01 35.7938 Q1948.5 32.5936 1943.48 32.5936 Q1937.44 32.5936 1933.96 36.4419 Q1930.47 40.2903 1930.47 46.9338 L1930.47 72.576 L1922.98 72.576 L1922.98 27.2059 L1930.47 27.2059 L1930.47 34.2544 Q1933.15 30.163 1936.75 28.1376 Q1940.4 26.1121 1945.14 26.1121 Q1952.96 26.1121 1956.97 30.9732 Q1960.98 35.7938 1960.98 45.1919 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M1977.75 62.2867 L1986.29 62.2867 L1986.29 72.576 L1977.75 72.576 L1977.75 62.2867 M1977.75 29.6769 L1986.29 29.6769 L1986.29 39.9662 L1977.75 39.9662 L1977.75 29.6769 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M2032.64 65.6895 L2046 65.6895 L2046 19.5497 L2031.46 22.4663 L2031.46 15.0127 L2045.92 12.096 L2054.11 12.096 L2054.11 65.6895 L2067.47 65.6895 L2067.47 72.576 L2032.64 72.576 L2032.64 65.6895 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip662)" style="stroke:#009af9; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" points="
  372.668,1352.59 377.474,1348.35 382.28,1343.54 387.087,1338.14 391.893,1332.18 396.699,1325.64 401.505,1318.54 406.311,1310.89 411.117,1302.69 415.923,1293.95 
  420.729,1284.67 425.535,1274.88 430.341,1264.57 435.147,1253.76 439.953,1242.45 444.759,1230.67 449.565,1218.42 454.371,1205.72 459.177,1192.57 463.983,1178.99 
  468.789,1165 473.595,1150.6 478.401,1135.81 483.207,1120.66 488.013,1105.14 492.819,1089.28 497.625,1073.1 502.431,1056.6 507.237,1039.81 512.044,1022.74 
  516.85,1005.41 521.656,987.837 526.462,970.039 531.268,952.034 536.074,933.839 540.88,915.473 545.686,896.954 550.492,878.3 555.298,859.531 560.104,840.666 
  564.91,821.722 569.716,802.719 574.522,783.676 579.328,764.612 584.134,745.545 588.94,726.496 593.746,707.484 598.552,688.526 603.358,669.642 608.164,650.851 
  612.97,632.172 617.776,613.622 622.582,595.221 627.388,576.986 632.194,558.938 637,541.093 641.807,523.471 646.613,506.087 651.419,488.96 656.225,472.106 
  661.031,455.544 665.837,439.289 670.643,423.358 675.449,407.766 680.255,392.53 685.061,377.663 689.867,363.182 694.673,349.1 699.479,335.432 704.285,322.19 
  709.091,309.389 713.897,297.042 718.703,285.159 723.509,273.754 728.315,262.837 733.121,252.421 737.927,242.514 742.733,233.127 747.539,224.27 752.345,215.95 
  757.151,208.178 761.957,200.96 766.763,194.303 771.57,188.215 776.376,182.701 781.182,177.767 785.988,173.418 790.794,169.659 795.6,166.492 800.406,163.922 
  805.212,161.951 810.018,160.58 814.824,159.811 819.63,159.645 824.436,160.083 829.242,161.124 834.048,162.766 838.854,165.008 843.66,167.848 848.466,171.281 
  853.272,175.306 858.078,179.919 862.884,185.115 867.69,190.889 872.496,197.235 877.302,204.145 882.108,211.613 886.914,219.632 891.72,228.194 896.527,237.292 
  901.333,246.914 906.139,257.052 910.945,267.696 915.751,278.834 920.557,290.454 925.363,302.548 930.169,315.102 934.975,328.103 939.781,341.539 944.587,355.395 
  949.393,369.659 954.199,384.314 959.005,399.349 963.811,414.747 968.617,430.493 973.423,446.57 978.229,462.964 983.035,479.657 987.841,496.633 992.647,513.874 
  997.453,531.365 1002.26,549.087 1007.07,567.022 1011.87,585.153 1016.68,603.461 1021.48,621.928 1026.29,640.534 1031.1,659.264 1035.9,678.098 1040.71,697.015 
  1045.51,715.998 1050.32,735.027 1055.13,754.084 1059.93,773.148 1064.74,792.203 1069.54,811.228 1074.35,830.204 1079.16,849.113 1083.96,867.934 1088.77,886.651 
  1093.57,905.242 1098.38,923.692 1103.19,941.98 1107.99,960.089 1112.8,978 1117.6,995.695 1122.41,1013.16 1127.22,1030.37 1132.02,1047.31 1136.83,1063.97 
  1141.63,1080.33 1146.44,1096.37 1151.25,1112.08 1156.05,1127.43 1160.86,1142.42 1165.66,1157.04 1170.47,1171.25 1175.28,1185.06 1180.08,1198.45 1184.89,1211.4 
  1189.69,1223.9 1194.5,1235.94 1199.31,1247.5 1204.11,1258.58 1208.92,1269.17 1213.72,1279.25 1218.53,1288.81 1223.34,1297.85 1228.14,1306.35 1232.95,1314.3 
  1237.76,1321.71 1242.56,1328.55 1247.37,1334.83 1252.17,1340.54 1256.98,1345.67 1261.79,1350.22 1266.59,1354.18 1271.4,1357.54 1276.2,1360.31 1281.01,1362.49 
  1285.82,1364.06 1290.62,1365.04 1295.43,1365.41 1300.23,1365.18 1305.04,1364.34 1309.85,1362.9 1314.65,1360.87 1319.46,1358.23 1324.26,1355 1329.07,1351.18 
  1333.88,1346.77 1338.68,1341.77 1343.49,1336.2 1348.29,1330.05 1353.1,1323.33 1357.91,1316.05 1362.71,1308.22 1367.52,1299.85 1372.32,1290.94 1377.13,1281.5 
  1381.94,1271.54 1386.74,1261.07 1391.55,1250.1 1396.35,1238.64 1401.16,1226.71 1405.97,1214.32 1410.77,1201.47 1415.58,1188.19 1420.38,1174.48 1425.19,1160.35 
  1430,1145.83 1434.8,1130.93 1439.61,1115.66 1444.41,1100.04 1449.22,1084.07 1454.03,1067.79 1458.83,1051.2 1463.64,1034.32 1468.44,1017.18 1473.25,999.775 
  1478.06,982.134 1482.86,964.275 1487.67,946.213 1492.48,927.969 1497.28,909.559 1502.09,891.002 1506.89,872.318 1511.7,853.523 1516.51,834.638 1521.31,815.68 
  1526.12,796.67 1530.92,777.625 1535.73,758.566 1540.54,739.512 1545.34,720.48 1550.15,701.49 1554.95,682.561 1559.76,663.713 1564.57,644.964 1569.37,626.332 
  1574.18,607.837 1578.98,589.497 1583.79,571.33 1588.6,553.354 1593.4,535.588 1598.21,518.048 1603.01,500.754 1607.82,483.721 1612.63,466.967 1617.43,450.509 
  1622.24,434.364 1627.04,418.546 1631.85,403.073 1636.66,387.959 1641.46,373.22 1646.27,358.87 1651.07,344.925 1655.88,331.397 1660.69,318.301 1665.49,305.648 
  1670.3,293.453 1675.1,281.727 1679.91,270.481 1684.72,259.729 1689.52,249.479 1694.33,239.742 1699.13,230.529 1703.94,221.847 1708.75,213.705 1713.55,206.113 
  1718.36,199.078 1723.16,192.606 1727.97,186.705 1732.78,181.379 1737.58,176.634 1742.39,172.475 1747.2,168.906 1752,165.931 1756.81,163.553 1761.61,161.775 
  1766.42,160.597 1771.23,160.021 1776.03,160.047 1780.84,160.674 1785.64,161.906 1790.45,163.738 1795.26,166.169 1800.06,169.196 1804.87,172.816 1809.67,177.026 
  1814.48,181.82 1819.29,187.196 1824.09,193.149 1828.9,199.67 1833.7,206.755 1838.51,214.395 1843.32,222.582 1848.12,231.309 1852.93,240.568 1857.73,250.35 
  1862.54,260.644 1867.35,271.44 1872.15,282.726 1876.96,294.492 1881.76,306.726 1886.57,319.415 1891.38,332.549 1896.18,346.113 1900.99,360.093 1905.79,374.475 
  1910.6,389.245 1915.41,404.388 1920.21,419.889 1925.02,435.733 1929.82,451.905 1934.63,468.387 1939.44,485.162 1944.24,502.215 1949.05,519.528 1953.85,537.084 
  1958.66,554.866 1963.47,572.856 1968.27,591.035 1973.08,609.386 1977.88,627.889 1982.69,646.527 1987.5,665.281 1992.3,684.132 1997.11,703.062 2001.92,722.052 
  2006.72,741.082 2011.53,760.133 2016.33,779.186 2021.14,798.223 2025.95,817.224 2030.75,836.17 2035.56,855.043 2040.36,873.823 2045.17,892.492 2049.98,911.031 
  2054.78,929.421 2059.59,947.644 2064.39,965.682 2069.2,983.518 2074.01,1001.13 2078.81,1018.51 2083.62,1035.63 2088.42,1052.47 2093.23,1069.03 2098.04,1085.28 
  2102.84,1101.21 2107.65,1116.79 2112.45,1132.03 2117.26,1146.89 2122.07,1161.37 2126.87,1175.45 2131.68,1189.12 2136.48,1202.36 2141.29,1215.17 2146.1,1227.52 
  2150.9,1239.4 2155.71,1250.81 2160.51,1261.73 2165.32,1272.15 2170.13,1282.06 2174.93,1291.45 2179.74,1300.31 2184.54,1308.63 2189.35,1316.41 2194.16,1323.64 
  2198.96,1330.3 2203.77,1336.4 2208.57,1341.92 2213.38,1346.86 2218.19,1351.22 2222.99,1354.99 2227.8,1358.17 2232.6,1360.75 2237.41,1362.73 2242.22,1364.11 
  2247.02,1364.89 2251.83,1365.07 2256.64,1364.65 2261.44,1363.63 2266.25,1362 2271.05,1359.77 2275.86,1356.95 2280.67,1353.53 2285.47,1349.52 2290.28,1344.93 
  2295.08,1339.75 
  "/>
<polyline clip-path="url(#clip662)" style="stroke:#e26f46; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" points="
  372.668,1352.59 377.474,1350.91 382.28,1348.65 387.087,1345.8 391.893,1342.37 396.699,1338.37 401.505,1333.79 406.311,1328.63 411.117,1322.92 415.923,1316.64 
  420.729,1309.8 425.535,1302.42 430.341,1294.5 435.147,1286.04 439.953,1277.06 444.759,1267.56 449.565,1257.56 454.371,1247.05 459.177,1236.06 463.983,1224.59 
  468.789,1212.65 473.595,1200.26 478.401,1187.42 483.207,1174.15 488.013,1160.46 492.819,1146.36 497.625,1131.86 502.431,1116.97 507.237,1101.71 512.044,1086.1 
  516.85,1070.13 521.656,1053.84 526.462,1037.22 531.268,1020.3 536.074,1003.08 540.88,985.575 545.686,967.8 550.492,949.766 555.298,931.485 560.104,912.965 
  564.91,894.216 569.716,875.25 574.522,856.073 579.328,836.69 584.134,817.109 588.94,797.336 593.746,777.373 598.552,758.799 603.358,744.406 608.164,729.909 
  612.97,715.329 617.776,700.69 622.582,686.012 627.388,671.317 632.194,656.627 637,641.963 641.807,627.348 646.613,612.8 651.419,598.342 656.225,583.993 
  661.031,569.771 665.837,555.696 670.643,541.785 675.449,528.058 680.255,514.53 685.061,501.219 689.867,488.141 694.673,475.311 699.479,462.744 704.285,450.458 
  709.091,438.464 713.897,426.776 718.703,415.41 723.509,404.379 728.315,393.693 733.121,383.365 737.927,373.407 742.733,363.831 747.539,354.645 752.345,345.861 
  757.151,337.487 761.957,329.534 766.763,322.009 771.57,314.921 776.376,308.276 781.182,302.083 785.988,296.348 790.794,291.076 795.6,286.274 800.406,281.946 
  805.212,278.097 810.018,274.732 814.824,271.854 819.63,269.465 824.436,267.568 829.242,266.167 834.048,265.261 838.854,264.852 843.66,264.941 848.466,265.528 
  853.272,266.613 858.078,268.193 862.884,270.269 867.69,272.838 872.496,275.898 877.302,279.446 882.108,283.478 886.914,287.991 891.72,292.982 896.527,298.444 
  901.333,304.373 906.139,310.764 910.945,317.609 915.751,324.903 920.557,332.641 925.363,340.814 930.169,349.415 934.975,358.435 939.781,367.867 944.587,377.701 
  949.393,387.93 954.199,398.545 959.005,409.536 963.811,420.893 968.617,432.607 973.423,444.666 978.229,457.06 983.035,469.781 987.841,482.819 992.647,496.162 
  997.453,509.799 1002.26,523.721 1007.07,537.914 1011.87,552.37 1016.68,567.08 1021.48,582.032 1026.29,597.217 1031.1,612.624 1035.9,628.246 1040.71,644.074 
  1045.51,660.099 1050.32,676.315 1055.13,692.714 1059.93,709.29 1064.74,726.037 1069.54,742.953 1074.35,760.009 1079.16,772.216 1083.96,783.63 1088.77,795.119 
  1093.57,806.667 1098.38,818.257 1103.19,829.875 1107.99,841.503 1112.8,853.125 1117.6,864.724 1122.41,876.285 1127.22,887.792 1132.02,899.228 1136.83,910.58 
  1141.63,921.835 1146.44,932.974 1151.25,943.983 1156.05,954.848 1160.86,965.556 1165.66,976.092 1170.47,986.445 1175.28,996.601 1180.08,1006.55 1184.89,1016.28 
  1189.69,1025.77 1194.5,1035.02 1199.31,1044.02 1204.11,1052.75 1208.92,1061.21 1213.72,1069.39 1218.53,1077.27 1223.34,1084.85 1228.14,1092.11 1232.95,1099.06 
  1237.76,1105.68 1242.56,1111.96 1247.37,1117.9 1252.17,1123.5 1256.98,1128.73 1261.79,1133.61 1266.59,1138.12 1271.4,1142.26 1276.2,1146.02 1281.01,1149.41 
  1285.82,1152.41 1290.62,1155.02 1295.43,1157.24 1300.23,1159.07 1305.04,1160.5 1309.85,1161.54 1314.65,1162.18 1319.46,1162.42 1324.26,1162.26 1329.07,1161.69 
  1333.88,1160.73 1338.68,1159.38 1343.49,1157.62 1348.29,1155.46 1353.1,1152.92 1357.91,1149.97 1362.71,1146.64 1367.52,1142.92 1372.32,1138.82 1377.13,1134.34 
  1381.94,1129.48 1386.74,1124.25 1391.55,1118.65 1396.35,1112.69 1401.16,1106.37 1405.97,1099.7 1410.77,1092.68 1415.58,1085.32 1420.38,1077.63 1425.19,1069.61 
  1430,1061.28 1434.8,1052.63 1439.61,1043.67 1444.41,1034.42 1449.22,1024.87 1454.03,1015.04 1458.83,1004.94 1463.64,994.568 1468.44,983.939 1473.25,973.059 
  1478.06,961.935 1482.86,950.578 1487.67,938.994 1492.48,927.192 1497.28,915.18 1502.09,902.965 1506.89,890.556 1511.7,877.96 1516.51,865.184 1521.31,852.236 
  1526.12,839.122 1530.92,825.848 1535.73,812.42 1540.54,798.845 1545.34,785.128 1550.15,771.273 1554.95,759.534 1559.76,751.365 1564.57,743.135 1569.37,734.854 
  1574.18,726.534 1578.98,718.184 1583.79,709.817 1588.6,701.44 1593.4,693.068 1598.21,684.709 1603.01,676.377 1607.82,668.08 1612.63,659.831 1617.43,651.64 
  1622.24,643.518 1627.04,635.474 1631.85,627.519 1636.66,619.664 1641.46,611.917 1646.27,604.289 1651.07,596.789 1655.88,589.426 1660.69,582.21 1665.49,575.15 
  1670.3,568.253 1675.1,561.529 1679.91,554.985 1684.72,548.63 1689.52,542.471 1694.33,536.515 1699.13,530.77 1703.94,525.242 1708.75,519.938 1713.55,514.865 
  1718.36,510.028 1723.16,505.432 1727.97,501.084 1732.78,496.989 1737.58,493.15 1742.39,489.574 1747.2,486.264 1752,483.224 1756.81,480.458 1761.61,477.969 
  1766.42,475.76 1771.23,473.834 1776.03,472.194 1780.84,470.842 1785.64,469.779 1790.45,469.007 1795.26,468.528 1800.06,468.342 1804.87,468.451 1809.67,468.854 
  1814.48,469.552 1819.29,470.545 1824.09,471.832 1828.9,473.412 1833.7,475.285 1838.51,477.448 1843.32,479.901 1848.12,482.642 1852.93,485.669 1857.73,488.979 
  1862.54,492.57 1867.35,496.438 1872.15,500.581 1876.96,504.995 1881.76,509.677 1886.57,514.623 1891.38,519.829 1896.18,525.29 1900.99,531.003 1905.79,536.962 
  1910.6,543.163 1915.41,549.602 1920.21,556.273 1925.02,563.17 1929.82,570.288 1934.63,577.622 1939.44,585.166 1944.24,592.915 1949.05,600.863 1953.85,609.004 
  1958.66,617.332 1963.47,625.841 1968.27,634.526 1973.08,643.379 1977.88,652.396 1982.69,661.572 1987.5,670.899 1992.3,680.372 1997.11,689.985 2001.92,699.732 
  2006.72,709.608 2011.53,719.608 2016.33,729.727 2021.14,739.96 2025.95,750.3 2030.75,760.739 2035.56,766.414 2040.36,771.082 2045.17,775.782 2049.98,780.512 
  2054.78,785.264 2059.59,790.032 2064.39,794.812 2069.2,799.599 2074.01,804.386 2078.81,809.169 2083.62,813.941 2088.42,818.696 2093.23,823.428 2098.04,828.132 
  2102.84,832.802 2107.65,837.432 2112.45,842.017 2117.26,846.55 2122.07,851.027 2126.87,855.442 2131.68,859.789 2136.48,864.063 2141.29,868.259 2146.1,872.373 
  2150.9,876.397 2155.71,880.328 2160.51,884.161 2165.32,887.891 2170.13,891.514 2174.93,895.024 2179.74,898.418 2184.54,901.692 2189.35,904.84 2194.16,907.86 
  2198.96,910.749 2203.77,913.5 2208.57,916.111 2213.38,918.58 2218.19,920.903 2222.99,923.077 2227.8,925.097 2232.6,926.964 2237.41,928.673 2242.22,930.222 
  2247.02,931.608 2251.83,932.832 2256.64,933.89 2261.44,934.778 2266.25,935.498 2271.05,936.049 2275.86,936.427 2280.67,936.631 2285.47,936.664 2290.28,936.522 
  2295.08,936.205 
  "/>
<polyline clip-path="url(#clip662)" style="stroke:#008000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" stroke-dasharray="4, 8" points="
  372.668,762.603 377.474,745.392 382.28,728.171 387.087,710.951 391.893,693.743 396.699,676.558 401.505,659.408 406.311,642.306 411.117,625.264 415.923,608.295 
  420.729,591.412 425.535,574.63 430.341,557.963 435.147,541.425 439.953,525.032 444.759,508.801 449.565,492.747 454.371,476.887 459.177,461.238 463.983,445.819 
  468.789,430.647 473.595,415.742 478.401,401.123 483.207,386.811 488.013,372.823 492.819,359.181 497.625,345.906 502.431,333.016 507.237,320.534 512.044,308.48 
  516.85,296.873 521.656,285.733 526.462,275.081 531.268,264.937 536.074,255.323 540.88,246.255 545.686,237.751 550.492,229.826 555.298,222.495 560.104,215.77 
  564.91,209.662 569.716,204.183 574.522,199.341 579.328,195.142 584.134,191.595 588.94,188.703 593.746,186.469 598.552,184.895 603.358,183.977 608.164,183.713 
  612.97,184.099 617.776,185.128 622.582,186.794 627.388,189.09 632.194,192.006 637,195.533 641.807,199.659 646.613,204.373 651.419,209.66 656.225,215.506 
  661.031,221.896 665.837,228.812 670.643,236.237 675.449,244.157 680.255,252.558 685.061,261.421 689.867,270.73 694.673,280.468 699.479,290.619 704.285,301.165 
  709.091,312.091 713.897,323.38 718.703,335.017 723.509,346.984 728.315,359.267 733.121,371.85 737.927,384.717 742.733,397.854 747.539,411.244 752.345,424.874 
  757.151,438.728 761.957,452.792 766.763,467.053 771.57,481.497 776.376,496.112 781.182,510.884 785.988,525.801 790.794,540.849 795.6,556.017 800.406,571.29 
  805.212,586.657 810.018,602.104 814.824,617.62 819.63,633.192 824.436,648.808 829.242,664.455 834.048,680.122 838.854,695.796 843.66,711.465 848.466,727.117 
  853.272,742.741 858.078,758.324 862.884,773.856 867.69,789.324 872.496,804.717 877.302,820.023 882.108,835.231 886.914,850.326 891.72,865.296 896.527,880.127 
  901.333,894.807 906.139,909.322 910.945,923.659 915.751,937.805 920.557,951.747 925.363,965.472 930.169,978.968 934.975,992.221 939.781,1005.22 944.587,1017.95 
  949.393,1030.4 954.199,1042.55 959.005,1054.4 963.811,1065.93 968.617,1077.12 973.423,1087.98 978.229,1098.48 983.035,1108.62 987.841,1118.39 992.647,1127.78 
  997.453,1136.77 1002.26,1145.35 1007.07,1153.51 1011.87,1161.25 1016.68,1168.56 1021.48,1175.42 1026.29,1181.84 1031.1,1187.81 1035.9,1193.32 1040.71,1198.36 
  1045.51,1202.95 1050.32,1207.06 1055.13,1210.72 1059.93,1213.9 1064.74,1216.62 1069.54,1218.87 1074.35,1220.65 1079.16,1221.98 1083.96,1222.85 1088.77,1223.27 
  1093.57,1223.24 1098.38,1222.77 1103.19,1221.86 1107.99,1220.53 1112.8,1218.76 1117.6,1216.58 1122.41,1213.97 1127.22,1210.96 1132.02,1207.56 1136.83,1203.75 
  1141.63,1199.57 1146.44,1195.01 1151.25,1190.08 1156.05,1184.79 1160.86,1179.14 1165.66,1173.16 1170.47,1166.84 1175.28,1160.19 1180.08,1153.22 1184.89,1145.94 
  1189.69,1138.36 1194.5,1130.49 1199.31,1122.33 1204.11,1113.89 1208.92,1105.19 1213.72,1096.23 1218.53,1087.02 1223.34,1077.56 1228.14,1067.87 1232.95,1057.95 
  1237.76,1047.81 1242.56,1037.45 1247.37,1026.89 1252.17,1016.13 1256.98,1005.19 1261.79,994.062 1266.59,982.764 1271.4,971.303 1276.2,959.687 1281.01,947.924 
  1285.82,936.024 1290.62,923.996 1295.43,911.847 1300.23,899.588 1305.04,887.228 1309.85,874.774 1314.65,862.238 1319.46,849.628 1324.26,836.953 1329.07,824.224 
  1333.88,811.45 1338.68,798.64 1343.49,785.804 1348.29,772.953 1353.1,760.096 1357.91,747.243 1362.71,734.405 1367.52,721.592 1372.32,708.814 1377.13,696.081 
  1381.94,683.407 1386.74,670.803 1391.55,658.285 1396.35,645.864 1401.16,633.554 1405.97,621.369 1410.77,609.321 1415.58,597.425 1420.38,585.692 1425.19,574.138 
  1430,562.774 1434.8,551.616 1439.61,540.675 1444.41,529.966 1449.22,519.503 1454.03,509.298 1458.83,499.365 1463.64,489.719 1468.44,480.372 1473.25,471.339 
  1478.06,462.632 1482.86,454.267 1487.67,446.254 1492.48,438.595 1497.28,431.305 1502.09,424.397 1506.89,417.884 1511.7,411.777 1516.51,406.086 1521.31,400.821 
  1526.12,395.991 1530.92,391.604 1535.73,387.664 1540.54,384.18 1545.34,381.153 1550.15,378.59 1554.95,376.491 1559.76,374.858 1564.57,373.693 1569.37,372.993 
  1574.18,372.759 1578.98,372.987 1583.79,373.675 1588.6,374.817 1593.4,376.408 1598.21,378.441 1603.01,380.91 1607.82,383.806 1612.63,387.119 1617.43,390.838 
  1622.24,394.953 1627.04,399.458 1631.85,404.35 1636.66,409.618 1641.46,415.252 1646.27,421.24 1651.07,427.571 1655.88,434.233 1660.69,441.216 1665.49,448.508 
  1670.3,456.097 1675.1,463.973 1679.91,472.122 1684.72,480.534 1689.52,489.197 1694.33,498.099 1699.13,507.228 1703.94,516.571 1708.75,526.118 1713.55,535.855 
  1718.36,545.77 1723.16,555.851 1727.97,566.086 1732.78,576.462 1737.58,586.966 1742.39,597.586 1747.2,608.314 1752,619.137 1756.81,630.046 1761.61,641.027 
  1766.42,652.069 1771.23,663.16 1776.03,674.29 1780.84,685.446 1785.64,696.617 1790.45,707.793 1795.26,718.961 1800.06,730.111 1804.87,741.231 1809.67,752.311 
  1814.48,763.34 1819.29,774.307 1824.09,785.202 1828.9,796.012 1833.7,806.729 1838.51,817.341 1843.32,827.838 1848.12,838.21 1852.93,848.446 1857.73,858.537 
  1862.54,868.471 1867.35,878.239 1872.15,887.831 1876.96,897.237 1881.76,906.448 1886.57,915.452 1891.38,924.244 1896.18,932.816 1900.99,941.159 1905.79,949.262 
  1910.6,957.118 1915.41,964.719 1920.21,972.056 1925.02,979.122 1929.82,985.911 1934.63,992.414 1939.44,998.628 1944.24,1004.55 1949.05,1010.16 1953.85,1015.47 
  1958.66,1020.47 1963.47,1025.16 1968.27,1029.53 1973.08,1033.58 1977.88,1037.3 1982.69,1040.7 1987.5,1043.78 1992.3,1046.53 1997.11,1048.95 2001.92,1051.04 
  2006.72,1052.8 2011.53,1054.23 2016.33,1055.33 2021.14,1056.11 2025.95,1056.56 2030.75,1056.69 2035.56,1056.5 2040.36,1055.99 2045.17,1055.16 2049.98,1054.02 
  2054.78,1052.58 2059.59,1050.83 2064.39,1048.77 2069.2,1046.42 2074.01,1043.78 2078.81,1040.85 2083.62,1037.63 2088.42,1034.14 2093.23,1030.37 2098.04,1026.34 
  2102.84,1022.05 2107.65,1017.5 2112.45,1012.7 2117.26,1007.66 2122.07,1002.39 2126.87,996.888 2131.68,991.168 2136.48,985.237 2141.29,979.102 2146.1,972.767 
  2150.9,966.237 2155.71,959.519 2160.51,952.62 2165.32,945.547 2170.13,938.306 2174.93,930.906 2179.74,923.353 2184.54,915.655 2189.35,907.819 2194.16,899.852 
  2198.96,891.763 2203.77,883.559 2208.57,875.248 2213.38,866.837 2218.19,858.335 2222.99,849.749 2227.8,841.088 2232.6,832.36 2237.41,823.573 2242.22,814.736 
  2247.02,805.856 2251.83,796.943 2256.64,788.005 2261.44,779.051 2266.25,770.089 2271.05,761.128 2275.86,752.178 2280.67,743.247 2285.47,734.344 2290.28,725.479 
  2295.08,716.659 
  "/>
<path clip-path="url(#clip660)" d="
M1423.56 477.116 L2284.83 477.116 L2284.83 166.076 L1423.56 166.076  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<polyline clip-path="url(#clip660)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1423.56,477.116 2284.83,477.116 2284.83,166.076 1423.56,166.076 1423.56,477.116 
  "/>
<polyline clip-path="url(#clip660)" style="stroke:#009af9; stroke-linecap:butt; stroke-linejoin:round; stroke-width:6; stroke-opacity:1; fill:none" points="
  1446.2,243.836 1582.05,243.836 
  "/>
<path clip-path="url(#clip660)" d="M1638.06 219.617 L1638.06 226.457 Q1634.06 224.548 1630.52 223.61 Q1626.98 222.673 1623.68 222.673 Q1617.95 222.673 1614.83 224.895 Q1611.74 227.117 1611.74 231.214 Q1611.74 234.652 1613.79 236.423 Q1615.87 238.159 1621.63 239.235 L1625.87 240.103 Q1633.72 241.596 1637.43 245.381 Q1641.18 249.131 1641.18 255.45 Q1641.18 262.985 1636.11 266.874 Q1631.08 270.763 1621.32 270.763 Q1617.64 270.763 1613.47 269.929 Q1609.34 269.096 1604.9 267.464 L1604.9 260.242 Q1609.17 262.638 1613.27 263.853 Q1617.36 265.068 1621.32 265.068 Q1627.33 265.068 1630.59 262.707 Q1633.86 260.346 1633.86 255.971 Q1633.86 252.152 1631.49 249.999 Q1629.17 247.846 1623.82 246.77 L1619.55 245.936 Q1611.7 244.374 1608.2 241.041 Q1604.69 237.707 1604.69 231.77 Q1604.69 224.895 1609.52 220.937 Q1614.38 216.978 1622.88 216.978 Q1626.53 216.978 1630.31 217.638 Q1634.1 218.298 1638.06 219.617 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M1651.84 230.867 L1658.23 230.867 L1658.23 269.756 L1651.84 269.756 L1651.84 230.867 M1651.84 215.728 L1658.23 215.728 L1658.23 223.818 L1651.84 223.818 L1651.84 215.728 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M1701.88 238.332 Q1704.27 234.027 1707.61 231.978 Q1710.94 229.93 1715.45 229.93 Q1721.53 229.93 1724.83 234.2 Q1728.13 238.436 1728.13 246.284 L1728.13 269.756 L1721.7 269.756 L1721.7 246.492 Q1721.7 240.902 1719.72 238.193 Q1717.74 235.485 1713.68 235.485 Q1708.72 235.485 1705.83 238.784 Q1702.95 242.082 1702.95 247.777 L1702.95 269.756 L1696.53 269.756 L1696.53 246.492 Q1696.53 240.867 1694.55 238.193 Q1692.57 235.485 1688.44 235.485 Q1683.54 235.485 1680.66 238.818 Q1677.78 242.117 1677.78 247.777 L1677.78 269.756 L1671.36 269.756 L1671.36 230.867 L1677.78 230.867 L1677.78 236.909 Q1679.97 233.332 1683.02 231.631 Q1686.08 229.93 1690.28 229.93 Q1694.52 229.93 1697.47 232.082 Q1700.45 234.235 1701.88 238.332 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M1747.05 263.922 L1747.05 284.547 L1740.63 284.547 L1740.63 230.867 L1747.05 230.867 L1747.05 236.77 Q1749.06 233.298 1752.12 231.631 Q1755.21 229.93 1759.48 229.93 Q1766.56 229.93 1770.97 235.555 Q1775.42 241.179 1775.42 250.346 Q1775.42 259.513 1770.97 265.138 Q1766.56 270.763 1759.48 270.763 Q1755.21 270.763 1752.12 269.096 Q1749.06 267.395 1747.05 263.922 M1768.79 250.346 Q1768.79 243.298 1765.87 239.304 Q1762.99 235.277 1757.92 235.277 Q1752.85 235.277 1749.93 239.304 Q1747.05 243.298 1747.05 250.346 Q1747.05 257.395 1749.93 261.422 Q1752.85 265.415 1757.92 265.415 Q1762.99 265.415 1765.87 261.422 Q1768.79 257.395 1768.79 250.346 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M1786.01 215.728 L1792.4 215.728 L1792.4 269.756 L1786.01 269.756 L1786.01 215.728 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M1839.03 248.714 L1839.03 251.839 L1809.65 251.839 Q1810.07 258.436 1813.61 261.909 Q1817.19 265.346 1823.54 265.346 Q1827.22 265.346 1830.66 264.443 Q1834.13 263.54 1837.54 261.735 L1837.54 267.777 Q1834.1 269.235 1830.49 269.999 Q1826.88 270.763 1823.16 270.763 Q1813.85 270.763 1808.4 265.346 Q1802.99 259.929 1802.99 250.693 Q1802.99 241.145 1808.13 235.555 Q1813.3 229.93 1822.05 229.93 Q1829.9 229.93 1834.45 234.999 Q1839.03 240.034 1839.03 248.714 M1832.64 246.839 Q1832.57 241.596 1829.69 238.471 Q1826.84 235.346 1822.12 235.346 Q1816.77 235.346 1813.54 238.367 Q1810.35 241.388 1809.86 246.874 L1832.64 246.839 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M1849.79 217.916 L1879.58 217.916 L1879.58 223.818 L1856.81 223.818 L1856.81 239.096 L1877.36 239.096 L1877.36 244.999 L1856.81 244.999 L1856.81 269.756 L1849.79 269.756 L1849.79 217.916 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M1890.69 217.916 L1901.15 217.916 L1914.38 253.193 L1927.67 217.916 L1938.13 217.916 L1938.13 269.756 L1931.28 269.756 L1931.28 224.235 L1917.92 259.79 L1910.87 259.79 L1897.5 224.235 L1897.5 269.756 L1890.69 269.756 L1890.69 217.916 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M1951.25 217.916 L1958.3 217.916 L1958.3 249.409 Q1958.3 257.742 1961.32 261.422 Q1964.34 265.068 1971.11 265.068 Q1977.85 265.068 1980.87 261.422 Q1983.89 257.742 1983.89 249.409 L1983.89 217.916 L1990.94 217.916 L1990.94 250.277 Q1990.94 260.415 1985.9 265.589 Q1980.9 270.763 1971.11 270.763 Q1961.28 270.763 1956.25 265.589 Q1951.25 260.415 1951.25 250.277 L1951.25 217.916 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip660)" style="stroke:#e26f46; stroke-linecap:butt; stroke-linejoin:round; stroke-width:6; stroke-opacity:1; fill:none" points="
  1446.2,321.596 1582.05,321.596 
  "/>
<path clip-path="url(#clip660)" d="M1629.27 323.21 Q1631.53 323.974 1633.65 326.474 Q1635.8 328.974 1637.95 333.349 L1645.07 347.516 L1637.54 347.516 L1630.9 334.217 Q1628.34 329.009 1625.9 327.307 Q1623.51 325.606 1619.34 325.606 L1611.7 325.606 L1611.7 347.516 L1604.69 347.516 L1604.69 295.676 L1620.52 295.676 Q1629.41 295.676 1633.79 299.391 Q1638.16 303.106 1638.16 310.606 Q1638.16 315.502 1635.87 318.731 Q1633.61 321.96 1629.27 323.21 M1611.7 301.44 L1611.7 319.842 L1620.52 319.842 Q1625.59 319.842 1628.16 317.516 Q1630.77 315.155 1630.77 310.606 Q1630.77 306.058 1628.16 303.766 Q1625.59 301.44 1620.52 301.44 L1611.7 301.44 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M1683.89 326.474 L1683.89 329.599 L1654.52 329.599 Q1654.93 336.196 1658.47 339.669 Q1662.05 343.106 1668.4 343.106 Q1672.08 343.106 1675.52 342.203 Q1678.99 341.3 1682.4 339.495 L1682.4 345.537 Q1678.96 346.995 1675.35 347.759 Q1671.74 348.523 1668.02 348.523 Q1658.72 348.523 1653.27 343.106 Q1647.85 337.689 1647.85 328.453 Q1647.85 318.905 1652.99 313.315 Q1658.16 307.69 1666.91 307.69 Q1674.76 307.69 1679.31 312.759 Q1683.89 317.794 1683.89 326.474 M1677.5 324.599 Q1677.43 319.356 1674.55 316.231 Q1671.7 313.106 1666.98 313.106 Q1661.63 313.106 1658.4 316.127 Q1655.21 319.148 1654.72 324.634 L1677.5 324.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M1714.06 293.488 L1714.06 298.801 L1707.95 298.801 Q1704.52 298.801 1703.16 300.19 Q1701.84 301.578 1701.84 305.19 L1701.84 308.627 L1712.36 308.627 L1712.36 313.592 L1701.84 313.592 L1701.84 347.516 L1695.42 347.516 L1695.42 313.592 L1689.31 313.592 L1689.31 308.627 L1695.42 308.627 L1695.42 305.919 Q1695.42 299.426 1698.44 296.474 Q1701.46 293.488 1708.02 293.488 L1714.06 293.488 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M1752.67 326.474 L1752.67 329.599 L1723.3 329.599 Q1723.72 336.196 1727.26 339.669 Q1730.83 343.106 1737.19 343.106 Q1740.87 343.106 1744.31 342.203 Q1747.78 341.3 1751.18 339.495 L1751.18 345.537 Q1747.74 346.995 1744.13 347.759 Q1740.52 348.523 1736.81 348.523 Q1727.5 348.523 1722.05 343.106 Q1716.63 337.689 1716.63 328.453 Q1716.63 318.905 1721.77 313.315 Q1726.95 307.69 1735.7 307.69 Q1743.54 307.69 1748.09 312.759 Q1752.67 317.794 1752.67 326.474 M1746.29 324.599 Q1746.22 319.356 1743.33 316.231 Q1740.49 313.106 1735.76 313.106 Q1730.42 313.106 1727.19 316.127 Q1723.99 319.148 1723.51 324.634 L1746.29 324.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M1785.7 314.599 Q1784.62 313.974 1783.33 313.696 Q1782.08 313.384 1780.56 313.384 Q1775.14 313.384 1772.22 316.926 Q1769.34 320.433 1769.34 327.03 L1769.34 347.516 L1762.92 347.516 L1762.92 308.627 L1769.34 308.627 L1769.34 314.669 Q1771.36 311.127 1774.58 309.426 Q1777.81 307.69 1782.43 307.69 Q1783.09 307.69 1783.89 307.794 Q1784.69 307.863 1785.66 308.037 L1785.7 314.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M1824.1 326.474 L1824.1 329.599 L1794.72 329.599 Q1795.14 336.196 1798.68 339.669 Q1802.26 343.106 1808.61 343.106 Q1812.29 343.106 1815.73 342.203 Q1819.2 341.3 1822.6 339.495 L1822.6 345.537 Q1819.17 346.995 1815.56 347.759 Q1811.95 348.523 1808.23 348.523 Q1798.92 348.523 1793.47 343.106 Q1788.06 337.689 1788.06 328.453 Q1788.06 318.905 1793.2 313.315 Q1798.37 307.69 1807.12 307.69 Q1814.97 307.69 1819.51 312.759 Q1824.1 317.794 1824.1 326.474 M1817.71 324.599 Q1817.64 319.356 1814.76 316.231 Q1811.91 313.106 1807.19 313.106 Q1801.84 313.106 1798.61 316.127 Q1795.42 319.148 1794.93 324.634 L1817.71 324.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M1866.91 324.044 L1866.91 347.516 L1860.52 347.516 L1860.52 324.252 Q1860.52 318.731 1858.37 315.988 Q1856.22 313.245 1851.91 313.245 Q1846.74 313.245 1843.75 316.544 Q1840.76 319.842 1840.76 325.537 L1840.76 347.516 L1834.34 347.516 L1834.34 308.627 L1840.76 308.627 L1840.76 314.669 Q1843.06 311.162 1846.15 309.426 Q1849.27 307.69 1853.33 307.69 Q1860.04 307.69 1863.47 311.856 Q1866.91 315.988 1866.91 324.044 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M1907.64 310.12 L1907.64 316.092 Q1904.93 314.599 1902.19 313.87 Q1899.48 313.106 1896.7 313.106 Q1890.49 313.106 1887.05 317.064 Q1883.61 320.988 1883.61 328.106 Q1883.61 335.224 1887.05 339.182 Q1890.49 343.106 1896.7 343.106 Q1899.48 343.106 1902.19 342.377 Q1904.93 341.613 1907.64 340.12 L1907.64 346.023 Q1904.97 347.273 1902.08 347.898 Q1899.24 348.523 1896.01 348.523 Q1887.22 348.523 1882.05 343.002 Q1876.88 337.481 1876.88 328.106 Q1876.88 318.592 1882.08 313.141 Q1887.33 307.69 1896.42 307.69 Q1899.38 307.69 1902.19 308.315 Q1905 308.905 1907.64 310.12 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M1952.01 326.474 L1952.01 329.599 L1922.64 329.599 Q1923.06 336.196 1926.6 339.669 Q1930.17 343.106 1936.53 343.106 Q1940.21 343.106 1943.65 342.203 Q1947.12 341.3 1950.52 339.495 L1950.52 345.537 Q1947.08 346.995 1943.47 347.759 Q1939.86 348.523 1936.15 348.523 Q1926.84 348.523 1921.39 343.106 Q1915.97 337.689 1915.97 328.453 Q1915.97 318.905 1921.11 313.315 Q1926.28 307.69 1935.03 307.69 Q1942.88 307.69 1947.43 312.759 Q1952.01 317.794 1952.01 326.474 M1945.62 324.599 Q1945.56 319.356 1942.67 316.231 Q1939.83 313.106 1935.1 313.106 Q1929.76 313.106 1926.53 316.127 Q1923.33 319.148 1922.85 324.634 L1945.62 324.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip660)" style="stroke:#008000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:6; stroke-opacity:1; fill:none" stroke-dasharray="3, 6" points="
  1446.2,399.356 1582.05,399.356 
  "/>
<path clip-path="url(#clip660)" d="M1604.69 373.436 L1614.13 373.436 L1637.12 416.804 L1637.12 373.436 L1643.93 373.436 L1643.93 425.276 L1634.48 425.276 L1611.49 381.908 L1611.49 425.276 L1604.69 425.276 L1604.69 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M1690.87 404.234 L1690.87 407.359 L1661.49 407.359 Q1661.91 413.956 1665.45 417.429 Q1669.03 420.866 1675.38 420.866 Q1679.06 420.866 1682.5 419.963 Q1685.97 419.06 1689.38 417.255 L1689.38 423.297 Q1685.94 424.755 1682.33 425.519 Q1678.72 426.283 1675 426.283 Q1665.7 426.283 1660.24 420.866 Q1654.83 415.449 1654.83 406.213 Q1654.83 396.665 1659.97 391.075 Q1665.14 385.45 1673.89 385.45 Q1681.74 385.45 1686.29 390.519 Q1690.87 395.554 1690.87 404.234 M1684.48 402.359 Q1684.41 397.116 1681.53 393.991 Q1678.68 390.866 1673.96 390.866 Q1668.61 390.866 1665.38 393.887 Q1662.19 396.908 1661.7 402.394 L1684.48 402.359 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M1700.7 409.929 L1700.7 386.387 L1707.08 386.387 L1707.08 409.686 Q1707.08 415.206 1709.24 417.984 Q1711.39 420.727 1715.7 420.727 Q1720.87 420.727 1723.86 417.429 Q1726.88 414.13 1726.88 408.436 L1726.88 386.387 L1733.27 386.387 L1733.27 425.276 L1726.88 425.276 L1726.88 419.304 Q1724.55 422.845 1721.46 424.581 Q1718.4 426.283 1714.34 426.283 Q1707.64 426.283 1704.17 422.116 Q1700.7 417.949 1700.7 409.929 M1716.77 385.45 L1716.77 385.45 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M1768.96 392.359 Q1767.88 391.734 1766.6 391.456 Q1765.35 391.144 1763.82 391.144 Q1758.4 391.144 1755.49 394.686 Q1752.61 398.193 1752.61 404.79 L1752.61 425.276 L1746.18 425.276 L1746.18 386.387 L1752.61 386.387 L1752.61 392.429 Q1754.62 388.887 1757.85 387.186 Q1761.08 385.45 1765.7 385.45 Q1766.36 385.45 1767.15 385.554 Q1767.95 385.623 1768.92 385.797 L1768.96 392.359 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M1793.33 405.727 Q1785.59 405.727 1782.61 407.498 Q1779.62 409.269 1779.62 413.54 Q1779.62 416.942 1781.84 418.956 Q1784.1 420.935 1787.95 420.935 Q1793.26 420.935 1796.46 417.185 Q1799.69 413.401 1799.69 407.151 L1799.69 405.727 L1793.33 405.727 M1806.08 403.088 L1806.08 425.276 L1799.69 425.276 L1799.69 419.373 Q1797.5 422.915 1794.24 424.616 Q1790.97 426.283 1786.25 426.283 Q1780.28 426.283 1776.74 422.949 Q1773.23 419.581 1773.23 413.956 Q1773.23 407.394 1777.61 404.061 Q1782.01 400.727 1790.73 400.727 L1799.69 400.727 L1799.69 400.102 Q1799.69 395.693 1796.77 393.297 Q1793.89 390.866 1788.65 390.866 Q1785.31 390.866 1782.15 391.665 Q1778.99 392.463 1776.08 394.061 L1776.08 388.158 Q1779.58 386.804 1782.88 386.144 Q1786.18 385.45 1789.31 385.45 Q1797.74 385.45 1801.91 389.825 Q1806.08 394.199 1806.08 403.088 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M1819.24 371.248 L1825.63 371.248 L1825.63 425.276 L1819.24 425.276 L1819.24 371.248 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M1839.27 373.436 L1869.06 373.436 L1869.06 379.338 L1846.29 379.338 L1846.29 394.616 L1866.84 394.616 L1866.84 400.519 L1846.29 400.519 L1846.29 425.276 L1839.27 425.276 L1839.27 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M1880.17 373.436 L1890.63 373.436 L1903.85 408.713 L1917.15 373.436 L1927.6 373.436 L1927.6 425.276 L1920.76 425.276 L1920.76 379.755 L1907.4 415.31 L1900.35 415.31 L1886.98 379.755 L1886.98 425.276 L1880.17 425.276 L1880.17 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M1940.73 373.436 L1947.78 373.436 L1947.78 404.929 Q1947.78 413.262 1950.8 416.942 Q1953.82 420.588 1960.59 420.588 Q1967.33 420.588 1970.35 416.942 Q1973.37 413.262 1973.37 404.929 L1973.37 373.436 L1980.42 373.436 L1980.42 405.797 Q1980.42 415.935 1975.38 421.109 Q1970.38 426.283 1960.59 426.283 Q1950.76 426.283 1945.73 421.109 Q1940.73 415.935 1940.73 405.797 L1940.73 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M2031.25 371.318 Q2026.6 379.304 2024.34 387.116 Q2022.08 394.929 2022.08 402.949 Q2022.08 410.97 2024.34 418.852 Q2026.63 426.699 2031.25 434.651 L2025.69 434.651 Q2020.49 426.491 2017.88 418.609 Q2015.31 410.727 2015.31 402.949 Q2015.31 395.206 2017.88 387.359 Q2020.45 379.512 2025.69 371.318 L2031.25 371.318 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M2050.59 419.373 L2075.07 419.373 L2075.07 425.276 L2042.15 425.276 L2042.15 419.373 Q2046.15 415.241 2053.02 408.297 Q2059.93 401.318 2061.7 399.304 Q2065.07 395.519 2066.39 392.915 Q2067.74 390.276 2067.74 387.741 Q2067.74 383.609 2064.83 381.005 Q2061.94 378.401 2057.29 378.401 Q2053.99 378.401 2050.31 379.547 Q2046.67 380.693 2042.5 383.019 L2042.5 375.936 Q2046.74 374.234 2050.42 373.366 Q2054.1 372.498 2057.15 372.498 Q2065.21 372.498 2070 376.526 Q2074.79 380.554 2074.79 387.29 Q2074.79 390.484 2073.58 393.366 Q2072.4 396.213 2069.24 400.102 Q2068.37 401.109 2063.71 405.936 Q2059.06 410.727 2050.59 419.373 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M2089.86 373.436 L2117.39 373.436 L2117.39 379.338 L2096.28 379.338 L2096.28 392.047 Q2097.81 391.526 2099.34 391.283 Q2100.87 391.005 2102.39 391.005 Q2111.08 391.005 2116.14 395.762 Q2121.21 400.519 2121.21 408.644 Q2121.21 417.012 2116.01 421.665 Q2110.8 426.283 2101.32 426.283 Q2098.05 426.283 2094.65 425.727 Q2091.28 425.172 2087.67 424.06 L2087.67 417.012 Q2090.8 418.713 2094.13 419.547 Q2097.46 420.38 2101.18 420.38 Q2107.19 420.38 2110.69 417.22 Q2114.2 414.06 2114.2 408.644 Q2114.2 403.227 2110.69 400.068 Q2107.19 396.908 2101.18 396.908 Q2098.37 396.908 2095.55 397.533 Q2092.78 398.158 2089.86 399.477 L2089.86 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M2150.03 378.054 Q2144.62 378.054 2141.87 383.401 Q2139.17 388.713 2139.17 399.408 Q2139.17 410.067 2141.87 415.415 Q2144.62 420.727 2150.03 420.727 Q2155.48 420.727 2158.19 415.415 Q2160.94 410.067 2160.94 399.408 Q2160.94 388.713 2158.19 383.401 Q2155.48 378.054 2150.03 378.054 M2150.03 372.498 Q2158.75 372.498 2163.33 379.408 Q2167.95 386.283 2167.95 399.408 Q2167.95 412.498 2163.33 419.408 Q2158.75 426.283 2150.03 426.283 Q2141.32 426.283 2136.7 419.408 Q2132.12 412.498 2132.12 399.408 Q2132.12 386.283 2136.7 379.408 Q2141.32 372.498 2150.03 372.498 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M2195.28 378.054 Q2189.86 378.054 2187.12 383.401 Q2184.41 388.713 2184.41 399.408 Q2184.41 410.067 2187.12 415.415 Q2189.86 420.727 2195.28 420.727 Q2200.73 420.727 2203.44 415.415 Q2206.18 410.067 2206.18 399.408 Q2206.18 388.713 2203.44 383.401 Q2200.73 378.054 2195.28 378.054 M2195.28 372.498 Q2203.99 372.498 2208.57 379.408 Q2213.19 386.283 2213.19 399.408 Q2213.19 412.498 2208.57 419.408 Q2203.99 426.283 2195.28 426.283 Q2186.56 426.283 2181.94 419.408 Q2177.36 412.498 2177.36 399.408 Q2177.36 386.283 2181.94 379.408 Q2186.56 372.498 2195.28 372.498 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip660)" d="M2223.61 371.318 L2229.17 371.318 Q2234.37 379.512 2236.94 387.359 Q2239.55 395.206 2239.55 402.949 Q2239.55 410.727 2236.94 418.609 Q2234.37 426.491 2229.17 434.651 L2223.61 434.651 Q2228.23 426.699 2230.48 418.852 Q2232.78 410.97 2232.78 402.949 Q2232.78 394.929 2230.48 387.116 Q2228.23 379.304 2223.61 371.318 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /></svg>



    ┌ Info: Friction model 1 mse: 0.0942545853263148
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:29



<?xml version="1.0" encoding="utf-8"?>
<svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" width="600" height="400" viewBox="0 0 2400 1600">
<defs>
  <clipPath id="clip750">
    <rect x="0" y="0" width="2400" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip750)" d="
M0 1600 L2400 1600 L2400 0 L0 0  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip751">
    <rect x="480" y="0" width="1681" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip750)" d="
M315.725 1401.58 L2352.76 1401.58 L2352.76 123.472 L315.725 123.472  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip752">
    <rect x="315" y="123" width="2038" height="1279"/>
  </clipPath>
</defs>
<polyline clip-path="url(#clip752)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  519.428,1401.58 519.428,123.472 
  "/>
<polyline clip-path="url(#clip752)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  926.834,1401.58 926.834,123.472 
  "/>
<polyline clip-path="url(#clip752)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1334.24,1401.58 1334.24,123.472 
  "/>
<polyline clip-path="url(#clip752)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1741.65,1401.58 1741.65,123.472 
  "/>
<polyline clip-path="url(#clip752)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  2149.05,1401.58 2149.05,123.472 
  "/>
<polyline clip-path="url(#clip750)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  315.725,1401.58 2352.76,1401.58 
  "/>
<polyline clip-path="url(#clip750)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  519.428,1401.58 519.428,1382.68 
  "/>
<polyline clip-path="url(#clip750)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  926.834,1401.58 926.834,1382.68 
  "/>
<polyline clip-path="url(#clip750)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1334.24,1401.58 1334.24,1382.68 
  "/>
<polyline clip-path="url(#clip750)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1741.65,1401.58 1741.65,1382.68 
  "/>
<polyline clip-path="url(#clip750)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  2149.05,1401.58 2149.05,1382.68 
  "/>
<path clip-path="url(#clip750)" d="M439.22 1460.34 L483.734 1460.34 L483.734 1466.24 L439.22 1466.24 L439.22 1460.34 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M500.088 1479.68 L511.546 1479.68 L511.546 1440.13 L499.081 1442.63 L499.081 1436.24 L511.477 1433.74 L518.491 1433.74 L518.491 1479.68 L529.949 1479.68 L529.949 1485.58 L500.088 1485.58 L500.088 1479.68 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M544.116 1476.76 L551.442 1476.76 L551.442 1485.58 L544.116 1485.58 L544.116 1476.76 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M581.72 1438.36 Q576.303 1438.36 573.56 1443.71 Q570.852 1449.02 570.852 1459.71 Q570.852 1470.37 573.56 1475.72 Q576.303 1481.03 581.72 1481.03 Q587.171 1481.03 589.879 1475.72 Q592.622 1470.37 592.622 1459.71 Q592.622 1449.02 589.879 1443.71 Q587.171 1438.36 581.72 1438.36 M581.72 1432.8 Q590.435 1432.8 595.018 1439.71 Q599.636 1446.59 599.636 1459.71 Q599.636 1472.8 595.018 1479.71 Q590.435 1486.59 581.72 1486.59 Q573.004 1486.59 568.386 1479.71 Q563.803 1472.8 563.803 1459.71 Q563.803 1446.59 568.386 1439.71 Q573.004 1432.8 581.72 1432.8 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M847.373 1460.34 L891.887 1460.34 L891.887 1466.24 L847.373 1466.24 L847.373 1460.34 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M922.025 1438.36 Q916.609 1438.36 913.866 1443.71 Q911.157 1449.02 911.157 1459.71 Q911.157 1470.37 913.866 1475.72 Q916.609 1481.03 922.025 1481.03 Q927.477 1481.03 930.185 1475.72 Q932.928 1470.37 932.928 1459.71 Q932.928 1449.02 930.185 1443.71 Q927.477 1438.36 922.025 1438.36 M922.025 1432.8 Q930.741 1432.8 935.324 1439.71 Q939.942 1446.59 939.942 1459.71 Q939.942 1472.8 935.324 1479.71 Q930.741 1486.59 922.025 1486.59 Q913.31 1486.59 908.692 1479.71 Q904.109 1472.8 904.109 1459.71 Q904.109 1446.59 908.692 1439.71 Q913.31 1432.8 922.025 1432.8 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M952.268 1476.76 L959.595 1476.76 L959.595 1485.58 L952.268 1485.58 L952.268 1476.76 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M974.942 1433.74 L1002.48 1433.74 L1002.48 1439.64 L981.365 1439.64 L981.365 1452.35 Q982.893 1451.83 984.421 1451.59 Q985.949 1451.31 987.476 1451.31 Q996.157 1451.31 1001.23 1456.07 Q1006.3 1460.82 1006.3 1468.95 Q1006.3 1477.32 1001.09 1481.97 Q995.879 1486.59 986.4 1486.59 Q983.136 1486.59 979.733 1486.03 Q976.365 1485.48 972.754 1484.36 L972.754 1477.32 Q975.879 1479.02 979.213 1479.85 Q982.546 1480.68 986.261 1480.68 Q992.268 1480.68 995.775 1477.52 Q999.282 1474.36 999.282 1468.95 Q999.282 1463.53 995.775 1460.37 Q992.268 1457.21 986.261 1457.21 Q983.449 1457.21 980.636 1457.84 Q977.858 1458.46 974.942 1459.78 L974.942 1433.74 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M1300.32 1438.36 Q1294.9 1438.36 1292.16 1443.71 Q1289.45 1449.02 1289.45 1459.71 Q1289.45 1470.37 1292.16 1475.72 Q1294.9 1481.03 1300.32 1481.03 Q1305.77 1481.03 1308.48 1475.72 Q1311.22 1470.37 1311.22 1459.71 Q1311.22 1449.02 1308.48 1443.71 Q1305.77 1438.36 1300.32 1438.36 M1300.32 1432.8 Q1309.03 1432.8 1313.62 1439.71 Q1318.23 1446.59 1318.23 1459.71 Q1318.23 1472.8 1313.62 1479.71 Q1309.03 1486.59 1300.32 1486.59 Q1291.6 1486.59 1286.98 1479.71 Q1282.4 1472.8 1282.4 1459.71 Q1282.4 1446.59 1286.98 1439.71 Q1291.6 1432.8 1300.32 1432.8 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M1330.56 1476.76 L1337.89 1476.76 L1337.89 1485.58 L1330.56 1485.58 L1330.56 1476.76 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M1368.16 1438.36 Q1362.75 1438.36 1360 1443.71 Q1357.3 1449.02 1357.3 1459.71 Q1357.3 1470.37 1360 1475.72 Q1362.75 1481.03 1368.16 1481.03 Q1373.62 1481.03 1376.32 1475.72 Q1379.07 1470.37 1379.07 1459.71 Q1379.07 1449.02 1376.32 1443.71 Q1373.62 1438.36 1368.16 1438.36 M1368.16 1432.8 Q1376.88 1432.8 1381.46 1439.71 Q1386.08 1446.59 1386.08 1459.71 Q1386.08 1472.8 1381.46 1479.71 Q1376.88 1486.59 1368.16 1486.59 Q1359.45 1486.59 1354.83 1479.71 Q1350.25 1472.8 1350.25 1459.71 Q1350.25 1446.59 1354.83 1439.71 Q1359.45 1432.8 1368.16 1432.8 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M1708.47 1438.36 Q1703.05 1438.36 1700.31 1443.71 Q1697.6 1449.02 1697.6 1459.71 Q1697.6 1470.37 1700.31 1475.72 Q1703.05 1481.03 1708.47 1481.03 Q1713.92 1481.03 1716.63 1475.72 Q1719.37 1470.37 1719.37 1459.71 Q1719.37 1449.02 1716.63 1443.71 Q1713.92 1438.36 1708.47 1438.36 M1708.47 1432.8 Q1717.18 1432.8 1721.77 1439.71 Q1726.39 1446.59 1726.39 1459.71 Q1726.39 1472.8 1721.77 1479.71 Q1717.18 1486.59 1708.47 1486.59 Q1699.75 1486.59 1695.14 1479.71 Q1690.55 1472.8 1690.55 1459.71 Q1690.55 1446.59 1695.14 1439.71 Q1699.75 1432.8 1708.47 1432.8 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M1738.71 1476.76 L1746.04 1476.76 L1746.04 1485.58 L1738.71 1485.58 L1738.71 1476.76 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M1761.39 1433.74 L1788.92 1433.74 L1788.92 1439.64 L1767.81 1439.64 L1767.81 1452.35 Q1769.34 1451.83 1770.87 1451.59 Q1772.39 1451.31 1773.92 1451.31 Q1782.6 1451.31 1787.67 1456.07 Q1792.74 1460.82 1792.74 1468.95 Q1792.74 1477.32 1787.53 1481.97 Q1782.32 1486.59 1772.84 1486.59 Q1769.58 1486.59 1766.18 1486.03 Q1762.81 1485.48 1759.2 1484.36 L1759.2 1477.32 Q1762.32 1479.02 1765.66 1479.85 Q1768.99 1480.68 1772.71 1480.68 Q1778.71 1480.68 1782.22 1477.52 Q1785.73 1474.36 1785.73 1468.95 Q1785.73 1463.53 1782.22 1460.37 Q1778.71 1457.21 1772.71 1457.21 Q1769.89 1457.21 1767.08 1457.84 Q1764.3 1458.46 1761.39 1459.78 L1761.39 1433.74 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M2099.78 1479.68 L2111.24 1479.68 L2111.24 1440.13 L2098.78 1442.63 L2098.78 1436.24 L2111.17 1433.74 L2118.18 1433.74 L2118.18 1479.68 L2129.64 1479.68 L2129.64 1485.58 L2099.78 1485.58 L2099.78 1479.68 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M2143.81 1476.76 L2151.14 1476.76 L2151.14 1485.58 L2143.81 1485.58 L2143.81 1476.76 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M2181.41 1438.36 Q2176 1438.36 2173.25 1443.71 Q2170.55 1449.02 2170.55 1459.71 Q2170.55 1470.37 2173.25 1475.72 Q2176 1481.03 2181.41 1481.03 Q2186.87 1481.03 2189.57 1475.72 Q2192.32 1470.37 2192.32 1459.71 Q2192.32 1449.02 2189.57 1443.71 Q2186.87 1438.36 2181.41 1438.36 M2181.41 1432.8 Q2190.13 1432.8 2194.71 1439.71 Q2199.33 1446.59 2199.33 1459.71 Q2199.33 1472.8 2194.71 1479.71 Q2190.13 1486.59 2181.41 1486.59 Q2172.7 1486.59 2168.08 1479.71 Q2163.5 1472.8 2163.5 1459.71 Q2163.5 1446.59 2168.08 1439.71 Q2172.7 1432.8 2181.41 1432.8 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M1213.13 1538.66 L1219.9 1538.66 L1232.05 1571.3 L1244.21 1538.66 L1250.98 1538.66 L1236.39 1577.55 L1227.71 1577.55 L1213.13 1538.66 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M1281.81 1523.52 L1296.53 1523.52 L1296.53 1528.49 L1288.2 1528.49 L1288.2 1581.96 L1296.53 1581.96 L1296.53 1586.92 L1281.81 1586.92 L1281.81 1523.52 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M1340.42 1546.12 Q1342.82 1541.82 1346.15 1539.77 Q1349.48 1537.72 1354 1537.72 Q1360.07 1537.72 1363.37 1541.99 Q1366.67 1546.23 1366.67 1554.08 L1366.67 1577.55 L1360.25 1577.55 L1360.25 1554.28 Q1360.25 1548.69 1358.27 1545.99 Q1356.29 1543.28 1352.23 1543.28 Q1347.26 1543.28 1344.38 1546.58 Q1341.5 1549.87 1341.5 1555.57 L1341.5 1577.55 L1335.07 1577.55 L1335.07 1554.28 Q1335.07 1548.66 1333.09 1545.99 Q1331.12 1543.28 1326.98 1543.28 Q1322.09 1543.28 1319.21 1546.61 Q1316.32 1549.91 1316.32 1555.57 L1316.32 1577.55 L1309.9 1577.55 L1309.9 1538.66 L1316.32 1538.66 L1316.32 1544.7 Q1318.51 1541.12 1321.57 1539.42 Q1324.62 1537.72 1328.82 1537.72 Q1333.06 1537.72 1336.01 1539.87 Q1339 1542.03 1340.42 1546.12 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M1390.77 1525.71 L1396.67 1525.71 L1378.62 1584.15 L1372.71 1584.15 L1390.77 1525.71 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M1428.16 1539.8 L1428.16 1545.85 Q1425.46 1544.46 1422.54 1543.76 Q1419.62 1543.07 1416.5 1543.07 Q1411.74 1543.07 1409.34 1544.53 Q1406.98 1545.99 1406.98 1548.9 Q1406.98 1551.12 1408.68 1552.41 Q1410.39 1553.66 1415.52 1554.8 L1417.71 1555.29 Q1424.52 1556.75 1427.37 1559.42 Q1430.25 1562.06 1430.25 1566.82 Q1430.25 1572.24 1425.94 1575.4 Q1421.67 1578.55 1414.17 1578.55 Q1411.05 1578.55 1407.64 1577.93 Q1404.27 1577.34 1400.52 1576.12 L1400.52 1569.53 Q1404.07 1571.37 1407.5 1572.3 Q1410.94 1573.21 1414.31 1573.21 Q1418.82 1573.21 1421.25 1571.68 Q1423.68 1570.12 1423.68 1567.3 Q1423.68 1564.7 1421.91 1563.31 Q1420.18 1561.92 1414.24 1560.64 L1412.02 1560.12 Q1406.08 1558.87 1403.44 1556.3 Q1400.8 1553.69 1400.8 1549.18 Q1400.8 1543.69 1404.69 1540.71 Q1408.58 1537.72 1415.73 1537.72 Q1419.27 1537.72 1422.4 1538.24 Q1425.52 1538.76 1428.16 1539.8 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M1455.35 1523.52 L1455.35 1586.92 L1440.63 1586.92 L1440.63 1581.96 L1448.93 1581.96 L1448.93 1528.49 L1440.63 1528.49 L1440.63 1523.52 L1455.35 1523.52 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip752)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  315.725,1254.21 2352.76,1254.21 
  "/>
<polyline clip-path="url(#clip752)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  315.725,1092.05 2352.76,1092.05 
  "/>
<polyline clip-path="url(#clip752)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  315.725,929.896 2352.76,929.896 
  "/>
<polyline clip-path="url(#clip752)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  315.725,767.738 2352.76,767.738 
  "/>
<polyline clip-path="url(#clip752)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  315.725,605.58 2352.76,605.58 
  "/>
<polyline clip-path="url(#clip752)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  315.725,443.422 2352.76,443.422 
  "/>
<polyline clip-path="url(#clip752)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  315.725,281.263 2352.76,281.263 
  "/>
<polyline clip-path="url(#clip750)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  315.725,1401.58 315.725,123.472 
  "/>
<polyline clip-path="url(#clip750)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  315.725,1254.21 334.623,1254.21 
  "/>
<polyline clip-path="url(#clip750)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  315.725,1092.05 334.623,1092.05 
  "/>
<polyline clip-path="url(#clip750)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  315.725,929.896 334.623,929.896 
  "/>
<polyline clip-path="url(#clip750)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  315.725,767.738 334.623,767.738 
  "/>
<polyline clip-path="url(#clip750)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  315.725,605.58 334.623,605.58 
  "/>
<polyline clip-path="url(#clip750)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  315.725,443.422 334.623,443.422 
  "/>
<polyline clip-path="url(#clip750)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  315.725,281.263 334.623,281.263 
  "/>
<path clip-path="url(#clip750)" d="M119.066 1254.89 L163.58 1254.89 L163.58 1260.79 L119.066 1260.79 L119.066 1254.89 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M193.719 1232.91 Q188.302 1232.91 185.559 1238.26 Q182.851 1243.57 182.851 1254.26 Q182.851 1264.92 185.559 1270.27 Q188.302 1275.58 193.719 1275.58 Q199.17 1275.58 201.878 1270.27 Q204.621 1264.92 204.621 1254.26 Q204.621 1243.57 201.878 1238.26 Q199.17 1232.91 193.719 1232.91 M193.719 1227.35 Q202.434 1227.35 207.017 1234.26 Q211.635 1241.14 211.635 1254.26 Q211.635 1267.35 207.017 1274.26 Q202.434 1281.14 193.719 1281.14 Q185.003 1281.14 180.385 1274.26 Q175.802 1267.35 175.802 1254.26 Q175.802 1241.14 180.385 1234.26 Q185.003 1227.35 193.719 1227.35 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M223.962 1271.31 L231.288 1271.31 L231.288 1280.13 L223.962 1280.13 L223.962 1271.31 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M262.434 1251.42 Q257.711 1251.42 254.934 1254.65 Q252.191 1257.87 252.191 1263.5 Q252.191 1269.09 254.934 1272.35 Q257.711 1275.58 262.434 1275.58 Q267.156 1275.58 269.899 1272.35 Q272.677 1269.09 272.677 1263.5 Q272.677 1257.87 269.899 1254.65 Q267.156 1251.42 262.434 1251.42 M276.357 1229.44 L276.357 1235.83 Q273.718 1234.58 271.01 1233.92 Q268.336 1233.26 265.697 1233.26 Q258.753 1233.26 255.072 1237.94 Q251.427 1242.63 250.906 1252.11 Q252.954 1249.09 256.045 1247.49 Q259.135 1245.86 262.85 1245.86 Q270.663 1245.86 275.177 1250.62 Q279.725 1255.34 279.725 1263.5 Q279.725 1271.49 275.003 1276.31 Q270.281 1281.14 262.434 1281.14 Q253.441 1281.14 248.684 1274.26 Q243.927 1267.35 243.927 1254.26 Q243.927 1241.97 249.76 1234.68 Q255.593 1227.35 265.42 1227.35 Q268.059 1227.35 270.732 1227.87 Q273.44 1228.4 276.357 1229.44 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M118.58 1092.73 L163.094 1092.73 L163.094 1098.63 L118.58 1098.63 L118.58 1092.73 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M193.233 1070.75 Q187.816 1070.75 185.073 1076.1 Q182.365 1081.41 182.365 1092.11 Q182.365 1102.77 185.073 1108.11 Q187.816 1113.43 193.233 1113.43 Q198.684 1113.43 201.392 1108.11 Q204.135 1102.77 204.135 1092.11 Q204.135 1081.41 201.392 1076.1 Q198.684 1070.75 193.233 1070.75 M193.233 1065.2 Q201.948 1065.2 206.531 1072.11 Q211.149 1078.98 211.149 1092.11 Q211.149 1105.2 206.531 1112.11 Q201.948 1118.98 193.233 1118.98 Q184.517 1118.98 179.899 1112.11 Q175.316 1105.2 175.316 1092.11 Q175.316 1078.98 179.899 1072.11 Q184.517 1065.2 193.233 1065.2 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M223.475 1109.15 L230.802 1109.15 L230.802 1117.97 L223.475 1117.97 L223.475 1109.15 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M265.35 1072.24 L247.642 1099.92 L265.35 1099.92 L265.35 1072.24 M263.51 1066.13 L272.329 1066.13 L272.329 1099.92 L279.725 1099.92 L279.725 1105.75 L272.329 1105.75 L272.329 1117.97 L265.35 1117.97 L265.35 1105.75 L241.948 1105.75 L241.948 1098.98 L263.51 1066.13 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M121.705 930.573 L166.219 930.573 L166.219 936.475 L121.705 936.475 L121.705 930.573 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M196.358 908.594 Q190.941 908.594 188.198 913.941 Q185.49 919.253 185.49 929.948 Q185.49 940.607 188.198 945.955 Q190.941 951.267 196.358 951.267 Q201.809 951.267 204.517 945.955 Q207.26 940.607 207.26 929.948 Q207.26 919.253 204.517 913.941 Q201.809 908.594 196.358 908.594 M196.358 903.038 Q205.073 903.038 209.656 909.948 Q214.274 916.823 214.274 929.948 Q214.274 943.038 209.656 949.948 Q205.073 956.823 196.358 956.823 Q187.642 956.823 183.024 949.948 Q178.441 943.038 178.441 929.948 Q178.441 916.823 183.024 909.948 Q187.642 903.038 196.358 903.038 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M226.6 946.996 L233.927 946.996 L233.927 955.816 L226.6 955.816 L226.6 946.996 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M255.246 949.913 L279.725 949.913 L279.725 955.816 L246.809 955.816 L246.809 949.913 Q250.802 945.781 257.677 938.836 Q264.586 931.857 266.357 929.843 Q269.725 926.059 271.045 923.455 Q272.399 920.816 272.399 918.281 Q272.399 914.149 269.482 911.545 Q266.6 908.941 261.947 908.941 Q258.649 908.941 254.968 910.087 Q251.323 911.232 247.156 913.559 L247.156 906.476 Q251.392 904.774 255.072 903.906 Q258.753 903.038 261.809 903.038 Q269.864 903.038 274.656 907.066 Q279.447 911.094 279.447 917.83 Q279.447 921.024 278.232 923.906 Q277.052 926.753 273.892 930.642 Q273.024 931.649 268.371 936.475 Q263.718 941.267 255.246 949.913 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M193.962 746.436 Q188.545 746.436 185.802 751.783 Q183.094 757.095 183.094 767.79 Q183.094 778.449 185.802 783.796 Q188.545 789.109 193.962 789.109 Q199.413 789.109 202.121 783.796 Q204.864 778.449 204.864 767.79 Q204.864 757.095 202.121 751.783 Q199.413 746.436 193.962 746.436 M193.962 740.88 Q202.677 740.88 207.26 747.79 Q211.878 754.665 211.878 767.79 Q211.878 780.88 207.26 787.79 Q202.677 794.664 193.962 794.664 Q185.246 794.664 180.628 787.79 Q176.045 780.88 176.045 767.79 Q176.045 754.665 180.628 747.79 Q185.246 740.88 193.962 740.88 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M224.205 784.838 L231.531 784.838 L231.531 793.658 L224.205 793.658 L224.205 784.838 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M261.809 746.436 Q256.392 746.436 253.649 751.783 Q250.941 757.095 250.941 767.79 Q250.941 778.449 253.649 783.796 Q256.392 789.109 261.809 789.109 Q267.26 789.109 269.968 783.796 Q272.711 778.449 272.711 767.79 Q272.711 757.095 269.968 751.783 Q267.26 746.436 261.809 746.436 M261.809 740.88 Q270.524 740.88 275.107 747.79 Q279.725 754.665 279.725 767.79 Q279.725 780.88 275.107 787.79 Q270.524 794.664 261.809 794.664 Q253.093 794.664 248.475 787.79 Q243.892 780.88 243.892 767.79 Q243.892 754.665 248.475 747.79 Q253.093 740.88 261.809 740.88 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M196.358 584.278 Q190.941 584.278 188.198 589.625 Q185.49 594.937 185.49 605.632 Q185.49 616.291 188.198 621.638 Q190.941 626.951 196.358 626.951 Q201.809 626.951 204.517 621.638 Q207.26 616.291 207.26 605.632 Q207.26 594.937 204.517 589.625 Q201.809 584.278 196.358 584.278 M196.358 578.722 Q205.073 578.722 209.656 585.632 Q214.274 592.507 214.274 605.632 Q214.274 618.722 209.656 625.632 Q205.073 632.506 196.358 632.506 Q187.642 632.506 183.024 625.632 Q178.441 618.722 178.441 605.632 Q178.441 592.507 183.024 585.632 Q187.642 578.722 196.358 578.722 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M226.6 622.68 L233.927 622.68 L233.927 631.5 L226.6 631.5 L226.6 622.68 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M255.246 625.597 L279.725 625.597 L279.725 631.5 L246.809 631.5 L246.809 625.597 Q250.802 621.465 257.677 614.52 Q264.586 607.541 266.357 605.527 Q269.725 601.743 271.045 599.139 Q272.399 596.5 272.399 593.965 Q272.399 589.833 269.482 587.229 Q266.6 584.625 261.947 584.625 Q258.649 584.625 254.968 585.771 Q251.323 586.916 247.156 589.243 L247.156 582.16 Q251.392 580.458 255.072 579.59 Q258.753 578.722 261.809 578.722 Q269.864 578.722 274.656 582.75 Q279.447 586.778 279.447 593.514 Q279.447 596.708 278.232 599.59 Q277.052 602.437 273.892 606.326 Q273.024 607.333 268.371 612.159 Q263.718 616.951 255.246 625.597 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M193.233 422.12 Q187.816 422.12 185.073 427.467 Q182.365 432.779 182.365 443.474 Q182.365 454.133 185.073 459.48 Q187.816 464.793 193.233 464.793 Q198.684 464.793 201.392 459.48 Q204.135 454.133 204.135 443.474 Q204.135 432.779 201.392 427.467 Q198.684 422.12 193.233 422.12 M193.233 416.564 Q201.948 416.564 206.531 423.474 Q211.149 430.349 211.149 443.474 Q211.149 456.564 206.531 463.473 Q201.948 470.348 193.233 470.348 Q184.517 470.348 179.899 463.473 Q175.316 456.564 175.316 443.474 Q175.316 430.349 179.899 423.474 Q184.517 416.564 193.233 416.564 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M223.475 460.522 L230.802 460.522 L230.802 469.342 L223.475 469.342 L223.475 460.522 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M265.35 423.613 L247.642 451.286 L265.35 451.286 L265.35 423.613 M263.51 417.502 L272.329 417.502 L272.329 451.286 L279.725 451.286 L279.725 457.119 L272.329 457.119 L272.329 469.342 L265.35 469.342 L265.35 457.119 L241.948 457.119 L241.948 450.349 L263.51 417.502 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M193.719 259.962 Q188.302 259.962 185.559 265.309 Q182.851 270.621 182.851 281.316 Q182.851 291.975 185.559 297.322 Q188.302 302.635 193.719 302.635 Q199.17 302.635 201.878 297.322 Q204.621 291.975 204.621 281.316 Q204.621 270.621 201.878 265.309 Q199.17 259.962 193.719 259.962 M193.719 254.406 Q202.434 254.406 207.017 261.316 Q211.635 268.191 211.635 281.316 Q211.635 294.406 207.017 301.315 Q202.434 308.19 193.719 308.19 Q185.003 308.19 180.385 301.315 Q175.802 294.406 175.802 281.316 Q175.802 268.191 180.385 261.316 Q185.003 254.406 193.719 254.406 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M223.962 298.364 L231.288 298.364 L231.288 307.183 L223.962 307.183 L223.962 298.364 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M262.434 278.468 Q257.711 278.468 254.934 281.698 Q252.191 284.927 252.191 290.552 Q252.191 296.142 254.934 299.406 Q257.711 302.635 262.434 302.635 Q267.156 302.635 269.899 299.406 Q272.677 296.142 272.677 290.552 Q272.677 284.927 269.899 281.698 Q267.156 278.468 262.434 278.468 M276.357 256.489 L276.357 262.878 Q273.718 261.628 271.01 260.968 Q268.336 260.309 265.697 260.309 Q258.753 260.309 255.072 264.996 Q251.427 269.684 250.906 279.163 Q252.954 276.142 256.045 274.545 Q259.135 272.913 262.85 272.913 Q270.663 272.913 275.177 277.67 Q279.725 282.392 279.725 290.552 Q279.725 298.538 275.003 303.364 Q270.281 308.19 262.434 308.19 Q253.441 308.19 248.684 301.315 Q243.927 294.406 243.927 281.316 Q243.927 269.024 249.76 261.732 Q255.593 254.406 265.42 254.406 Q268.059 254.406 270.732 254.927 Q273.44 255.448 276.357 256.489 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M13.0007 1025.04 L18.3131 1025.04 L18.3131 1031.15 Q18.3131 1034.59 19.702 1035.94 Q21.0909 1037.26 24.702 1037.26 L28.1395 1037.26 L28.1395 1026.74 L33.1047 1026.74 L33.1047 1037.26 L67.0282 1037.26 L67.0282 1043.69 L33.1047 1043.69 L33.1047 1049.8 L28.1395 1049.8 L28.1395 1043.69 L25.4312 1043.69 Q18.9381 1043.69 15.9868 1040.67 Q13.0007 1037.65 13.0007 1031.08 L13.0007 1025.04 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M34.1117 997.16 Q33.4867 998.237 33.2089 999.521 Q32.8964 1000.77 32.8964 1002.3 Q32.8964 1007.72 36.438 1010.63 Q39.945 1013.51 46.5422 1013.51 L67.0282 1013.51 L67.0282 1019.94 L28.1395 1019.94 L28.1395 1013.51 L34.1811 1013.51 Q30.6395 1011.5 28.9381 1008.27 Q27.202 1005.04 27.202 1000.42 Q27.202 999.764 27.3061 998.966 Q27.3756 998.167 27.5492 997.195 L34.1117 997.16 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M28.1395 990.459 L28.1395 984.07 L67.0282 984.07 L67.0282 990.459 L28.1395 990.459 M13.0007 990.459 L13.0007 984.07 L21.0909 984.07 L21.0909 990.459 L13.0007 990.459 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M29.6325 942.716 L35.6047 942.716 Q34.1117 945.424 33.3825 948.167 Q32.6186 950.876 32.6186 953.653 Q32.6186 959.869 36.5769 963.306 Q40.5005 966.744 47.6185 966.744 Q54.7366 966.744 58.6949 963.306 Q62.6185 959.869 62.6185 953.653 Q62.6185 950.876 61.8893 948.167 Q61.1254 945.424 59.6324 942.716 L65.5351 942.716 Q66.7851 945.39 67.4101 948.272 Q68.0351 951.119 68.0351 954.348 Q68.0351 963.133 62.5143 968.306 Q56.9935 973.48 47.6185 973.48 Q38.1047 973.48 32.6533 968.271 Q27.202 963.028 27.202 953.931 Q27.202 950.98 27.827 948.167 Q28.4173 945.355 29.6325 942.716 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M17.0979 925.286 L28.1395 925.286 L28.1395 912.126 L33.1047 912.126 L33.1047 925.286 L54.2157 925.286 Q58.9726 925.286 60.3268 924.001 Q61.681 922.681 61.681 918.688 L61.681 912.126 L67.0282 912.126 L67.0282 918.688 Q67.0282 926.084 64.2851 928.897 Q61.5074 931.709 54.2157 931.709 L33.1047 931.709 L33.1047 936.397 L28.1395 936.397 L28.1395 931.709 L17.0979 931.709 L17.0979 925.286 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M28.1395 903.723 L28.1395 897.334 L67.0282 897.334 L67.0282 903.723 L28.1395 903.723 M13.0007 903.723 L13.0007 897.334 L21.0909 897.334 L21.0909 903.723 L13.0007 903.723 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M32.6186 868.897 Q32.6186 874.036 36.6464 877.022 Q40.6394 880.008 47.6185 880.008 Q54.5977 880.008 58.6254 877.057 Q62.6185 874.071 62.6185 868.897 Q62.6185 863.793 58.5907 860.807 Q54.5629 857.821 47.6185 857.821 Q40.7089 857.821 36.6811 860.807 Q32.6186 863.793 32.6186 868.897 M27.202 868.897 Q27.202 860.564 32.6186 855.807 Q38.0353 851.05 47.6185 851.05 Q57.1671 851.05 62.6185 855.807 Q68.0351 860.564 68.0351 868.897 Q68.0351 877.265 62.6185 882.022 Q57.1671 886.744 47.6185 886.744 Q38.0353 886.744 32.6186 882.022 Q27.202 877.265 27.202 868.897 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M43.5561 808.133 L67.0282 808.133 L67.0282 814.522 L43.7644 814.522 Q38.2436 814.522 35.5005 816.675 Q32.7575 818.828 32.7575 823.133 Q32.7575 828.307 36.0561 831.293 Q39.3547 834.279 45.0491 834.279 L67.0282 834.279 L67.0282 840.703 L28.1395 840.703 L28.1395 834.279 L34.1811 834.279 Q30.6742 831.987 28.9381 828.897 Q27.202 825.772 27.202 821.71 Q27.202 815.008 31.3686 811.571 Q35.5005 808.133 43.5561 808.133 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M13.0007 753.099 L18.3131 753.099 L18.3131 759.21 Q18.3131 762.648 19.702 764.002 Q21.0909 765.321 24.702 765.321 L28.1395 765.321 L28.1395 754.8 L33.1047 754.8 L33.1047 765.321 L67.0282 765.321 L67.0282 771.745 L33.1047 771.745 L33.1047 777.856 L28.1395 777.856 L28.1395 771.745 L25.4312 771.745 Q18.9381 771.745 15.9868 768.724 Q13.0007 765.703 13.0007 759.141 L13.0007 753.099 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M32.6186 732.682 Q32.6186 737.821 36.6464 740.807 Q40.6394 743.793 47.6185 743.793 Q54.5977 743.793 58.6254 740.842 Q62.6185 737.856 62.6185 732.682 Q62.6185 727.578 58.5907 724.592 Q54.5629 721.606 47.6185 721.606 Q40.7089 721.606 36.6811 724.592 Q32.6186 727.578 32.6186 732.682 M27.202 732.682 Q27.202 724.349 32.6186 719.592 Q38.0353 714.835 47.6185 714.835 Q57.1671 714.835 62.6185 719.592 Q68.0351 724.349 68.0351 732.682 Q68.0351 741.05 62.6185 745.807 Q57.1671 750.53 47.6185 750.53 Q38.0353 750.53 32.6186 745.807 Q27.202 741.05 27.202 732.682 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M34.1117 681.71 Q33.4867 682.787 33.2089 684.072 Q32.8964 685.322 32.8964 686.849 Q32.8964 692.266 36.438 695.183 Q39.945 698.065 46.5422 698.065 L67.0282 698.065 L67.0282 704.488 L28.1395 704.488 L28.1395 698.065 L34.1811 698.065 Q30.6395 696.051 28.9381 692.821 Q27.202 689.592 27.202 684.974 Q27.202 684.315 27.3061 683.516 Q27.3756 682.717 27.5492 681.745 L34.1117 681.71 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M29.6325 648.586 L35.6047 648.586 Q34.1117 651.294 33.3825 654.037 Q32.6186 656.745 32.6186 659.523 Q32.6186 665.738 36.5769 669.176 Q40.5005 672.613 47.6185 672.613 Q54.7366 672.613 58.6949 669.176 Q62.6185 665.738 62.6185 659.523 Q62.6185 656.745 61.8893 654.037 Q61.1254 651.294 59.6324 648.586 L65.5351 648.586 Q66.7851 651.259 67.4101 654.141 Q68.0351 656.988 68.0351 660.218 Q68.0351 669.002 62.5143 674.176 Q56.9935 679.349 47.6185 679.349 Q38.1047 679.349 32.6533 674.141 Q27.202 668.898 27.202 659.801 Q27.202 656.849 27.827 654.037 Q28.4173 651.224 29.6325 648.586 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M45.9866 604.211 L49.1116 604.211 L49.1116 633.586 Q55.7088 633.169 59.181 629.627 Q62.6185 626.051 62.6185 619.697 Q62.6185 616.016 61.7157 612.579 Q60.8129 609.107 59.0074 605.704 L65.049 605.704 Q66.5073 609.141 67.2712 612.752 Q68.0351 616.364 68.0351 620.079 Q68.0351 629.384 62.6185 634.836 Q57.2018 640.252 47.9658 640.252 Q38.4172 640.252 32.827 635.113 Q27.202 629.94 27.202 621.19 Q27.202 613.343 32.2714 608.794 Q37.3061 604.211 45.9866 604.211 M44.1116 610.6 Q38.8686 610.669 35.7436 613.551 Q32.6186 616.398 32.6186 621.12 Q32.6186 626.468 35.6394 629.697 Q38.6603 632.891 44.1463 633.377 L44.1116 610.6 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M13.0007 571.711 L13.0007 556.989 L17.9659 556.989 L17.9659 565.322 L71.4379 565.322 L71.4379 556.989 L76.4031 556.989 L76.4031 571.711 L13.0007 571.711 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M15.1882 543.1 L15.1882 533.656 L58.556 510.67 L15.1882 510.67 L15.1882 503.864 L67.0282 503.864 L67.0282 513.309 L23.6603 536.295 L67.0282 536.295 L67.0282 543.1 L15.1882 543.1 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M13.0007 475.253 L76.4031 475.253 L76.4031 489.975 L71.4379 489.975 L71.4379 481.677 L17.9659 481.677 L17.9659 489.975 L13.0007 489.975 L13.0007 475.253 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M1008.53 12.096 L1043.28 12.096 L1043.28 18.9825 L1016.71 18.9825 L1016.71 36.8065 L1040.69 36.8065 L1040.69 43.6931 L1016.71 43.6931 L1016.71 72.576 L1008.53 72.576 L1008.53 12.096 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M1076.18 34.1734 Q1074.92 33.4443 1073.42 33.1202 Q1071.96 32.7556 1070.18 32.7556 Q1063.86 32.7556 1060.46 36.8875 Q1057.1 40.9789 1057.1 48.6757 L1057.1 72.576 L1049.6 72.576 L1049.6 27.2059 L1057.1 27.2059 L1057.1 34.2544 Q1059.45 30.1225 1063.21 28.1376 Q1066.98 26.1121 1072.37 26.1121 Q1073.14 26.1121 1074.07 26.2337 Q1075 26.3147 1076.14 26.5172 L1076.18 34.1734 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M1084 27.2059 L1091.45 27.2059 L1091.45 72.576 L1084 72.576 L1084 27.2059 M1084 9.54393 L1091.45 9.54393 L1091.45 18.9825 L1084 18.9825 L1084 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M1139.7 28.9478 L1139.7 35.9153 Q1136.54 34.1734 1133.34 33.3227 Q1130.18 32.4315 1126.94 32.4315 Q1119.68 32.4315 1115.67 37.0496 Q1111.66 41.6271 1111.66 49.9314 Q1111.66 58.2358 1115.67 62.8538 Q1119.68 67.4314 1126.94 67.4314 Q1130.18 67.4314 1133.34 66.5807 Q1136.54 65.6895 1139.7 63.9476 L1139.7 70.8341 Q1136.58 72.2924 1133.21 73.0216 Q1129.89 73.7508 1126.13 73.7508 Q1115.88 73.7508 1109.84 67.3098 Q1103.8 60.8689 1103.8 49.9314 Q1103.8 38.832 1109.88 32.472 Q1116 26.1121 1126.61 26.1121 Q1130.05 26.1121 1133.34 26.8413 Q1136.62 27.5299 1139.7 28.9478 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M1160.03 14.324 L1160.03 27.2059 L1175.38 27.2059 L1175.38 32.9987 L1160.03 32.9987 L1160.03 57.6282 Q1160.03 63.1779 1161.53 64.7578 Q1163.07 66.3376 1167.73 66.3376 L1175.38 66.3376 L1175.38 72.576 L1167.73 72.576 Q1159.1 72.576 1155.82 69.3758 Q1152.54 66.1351 1152.54 57.6282 L1152.54 32.9987 L1147.07 32.9987 L1147.07 27.2059 L1152.54 27.2059 L1152.54 14.324 L1160.03 14.324 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M1185.19 27.2059 L1192.64 27.2059 L1192.64 72.576 L1185.19 72.576 L1185.19 27.2059 M1185.19 9.54393 L1192.64 9.54393 L1192.64 18.9825 L1185.19 18.9825 L1185.19 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M1225.82 32.4315 Q1219.82 32.4315 1216.34 37.1306 Q1212.86 41.7891 1212.86 49.9314 Q1212.86 58.0738 1216.3 62.7728 Q1219.78 67.4314 1225.82 67.4314 Q1231.77 67.4314 1235.26 62.7323 Q1238.74 58.0333 1238.74 49.9314 Q1238.74 41.8701 1235.26 37.1711 Q1231.77 32.4315 1225.82 32.4315 M1225.82 26.1121 Q1235.54 26.1121 1241.09 32.4315 Q1246.64 38.7509 1246.64 49.9314 Q1246.64 61.0714 1241.09 67.4314 Q1235.54 73.7508 1225.82 73.7508 Q1216.06 73.7508 1210.51 67.4314 Q1205 61.0714 1205 49.9314 Q1205 38.7509 1210.51 32.4315 Q1216.06 26.1121 1225.82 26.1121 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M1296.71 45.1919 L1296.71 72.576 L1289.26 72.576 L1289.26 45.4349 Q1289.26 38.994 1286.74 35.7938 Q1284.23 32.5936 1279.21 32.5936 Q1273.17 32.5936 1269.69 36.4419 Q1266.21 40.2903 1266.21 46.9338 L1266.21 72.576 L1258.71 72.576 L1258.71 27.2059 L1266.21 27.2059 L1266.21 34.2544 Q1268.88 30.163 1272.48 28.1376 Q1276.13 26.1121 1280.87 26.1121 Q1288.69 26.1121 1292.7 30.9732 Q1296.71 35.7938 1296.71 45.1919 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M1373.27 35.9153 Q1376.07 30.8922 1379.95 28.5022 Q1383.84 26.1121 1389.11 26.1121 Q1396.2 26.1121 1400.05 31.0947 Q1403.9 36.0368 1403.9 45.1919 L1403.9 72.576 L1396.4 72.576 L1396.4 45.4349 Q1396.4 38.913 1394.09 35.7533 Q1391.78 32.5936 1387.04 32.5936 Q1381.25 32.5936 1377.89 36.4419 Q1374.53 40.2903 1374.53 46.9338 L1374.53 72.576 L1367.03 72.576 L1367.03 45.4349 Q1367.03 38.8725 1364.72 35.7533 Q1362.41 32.5936 1357.59 32.5936 Q1351.88 32.5936 1348.52 36.4824 Q1345.16 40.3308 1345.16 46.9338 L1345.16 72.576 L1337.66 72.576 L1337.66 27.2059 L1345.16 27.2059 L1345.16 34.2544 Q1347.71 30.082 1351.27 28.0971 Q1354.84 26.1121 1359.74 26.1121 Q1364.68 26.1121 1368.13 28.6237 Q1371.61 31.1352 1373.27 35.9153 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M1436.34 32.4315 Q1430.35 32.4315 1426.86 37.1306 Q1423.38 41.7891 1423.38 49.9314 Q1423.38 58.0738 1426.82 62.7728 Q1430.31 67.4314 1436.34 67.4314 Q1442.3 67.4314 1445.78 62.7323 Q1449.27 58.0333 1449.27 49.9314 Q1449.27 41.8701 1445.78 37.1711 Q1442.3 32.4315 1436.34 32.4315 M1436.34 26.1121 Q1446.07 26.1121 1451.62 32.4315 Q1457.17 38.7509 1457.17 49.9314 Q1457.17 61.0714 1451.62 67.4314 Q1446.07 73.7508 1436.34 73.7508 Q1426.58 73.7508 1421.03 67.4314 Q1415.52 61.0714 1415.52 49.9314 Q1415.52 38.7509 1421.03 32.4315 Q1426.58 26.1121 1436.34 26.1121 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M1499.38 34.0924 L1499.38 9.54393 L1506.83 9.54393 L1506.83 72.576 L1499.38 72.576 L1499.38 65.7705 Q1497.03 69.8214 1493.42 71.8063 Q1489.86 73.7508 1484.83 73.7508 Q1476.61 73.7508 1471.42 67.1883 Q1466.28 60.6258 1466.28 49.9314 Q1466.28 39.2371 1471.42 32.6746 Q1476.61 26.1121 1484.83 26.1121 Q1489.86 26.1121 1493.42 28.0971 Q1497.03 30.0415 1499.38 34.0924 M1473.98 49.9314 Q1473.98 58.1548 1477.34 62.8538 Q1480.74 67.5124 1486.66 67.5124 Q1492.57 67.5124 1495.97 62.8538 Q1499.38 58.1548 1499.38 49.9314 Q1499.38 41.7081 1495.97 37.0496 Q1492.57 32.3505 1486.66 32.3505 Q1480.74 32.3505 1477.34 37.0496 Q1473.98 41.7081 1473.98 49.9314 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M1560.99 48.0275 L1560.99 51.6733 L1526.72 51.6733 Q1527.21 59.3701 1531.34 63.421 Q1535.51 67.4314 1542.92 67.4314 Q1547.22 67.4314 1551.23 66.3781 Q1555.28 65.3249 1559.25 63.2184 L1559.25 70.267 Q1555.24 71.9684 1551.02 72.8596 Q1546.81 73.7508 1542.48 73.7508 Q1531.62 73.7508 1525.26 67.4314 Q1518.94 61.1119 1518.94 50.3365 Q1518.94 39.1965 1524.94 32.6746 Q1530.97 26.1121 1541.18 26.1121 Q1550.34 26.1121 1555.64 32.0264 Q1560.99 37.9003 1560.99 48.0275 M1553.54 45.84 Q1553.46 39.7232 1550.09 36.0774 Q1546.77 32.4315 1541.26 32.4315 Q1535.02 32.4315 1531.26 35.9558 Q1527.53 39.4801 1526.96 45.8805 L1553.54 45.84 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M1573.22 9.54393 L1580.68 9.54393 L1580.68 72.576 L1573.22 72.576 L1573.22 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M1625.12 65.6895 L1638.48 65.6895 L1638.48 19.5497 L1623.94 22.4663 L1623.94 15.0127 L1638.4 12.096 L1646.59 12.096 L1646.59 65.6895 L1659.95 65.6895 L1659.95 72.576 L1625.12 72.576 L1625.12 65.6895 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip752)" style="stroke:#009af9; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" points="
  1334.24,767.738 1369.02,767.738 1404.07,767.738 1439.32,767.738 1474.72,767.738 1510.21,767.738 1545.75,767.738 1581.27,767.738 1616.72,767.738 1652.05,767.738 
  1687.21,767.738 1722.15,767.738 1756.82,767.738 1791.17,767.738 1825.16,767.738 1858.74,767.738 1891.87,767.738 1924.51,767.738 1956.62,767.738 1988.17,767.738 
  2019.11,767.738 2049.4,767.738 2079.02,767.738 2107.93,767.738 2136.09,767.738 2163.47,767.738 2190.04,767.738 2215.78,767.738 2240.65,767.738 2264.63,767.738 
  2287.69,767.738 2309.81,767.738 2330.96,767.738 2351.12,767.738 2370.27,767.738 2388.39,767.738 2405.46,767.738 2421.46,767.738 2436.37,767.738 2450.19,767.738 
  2462.89,767.738 2474.46,767.738 2484.9,767.738 2494.18,767.738 2502.31,767.738 2509.27,767.738 2515.05,767.738 2519.65,767.738 2523.06,767.738 2525.29,767.738 
  2526.32,767.738 2526.16,767.738 2524.81,767.738 2522.27,767.738 2518.54,767.738 2513.63,767.738 2507.54,767.738 2500.27,767.738 2491.83,767.738 2482.24,767.738 
  2471.5,767.738 2459.62,767.738 2446.61,767.738 2432.49,767.738 2417.26,767.738 2400.96,767.738 2383.57,767.738 2365.14,767.738 2345.67,767.738 2325.18,767.738 
  2303.69,767.738 2281.22,767.738 2257.79,767.738 2233.43,767.738 2208.15,767.738 2181.98,767.738 2154.94,767.738 2127.06,767.738 2098.36,767.738 2068.86,767.738 
  2038.6,767.738 2007.59,767.738 1975.86,767.738 1943.44,767.738 1910.36,767.738 1876.63,767.738 1842.29,767.738 1807.35,767.738 1771.85,767.738 1735.81,767.738 
  1699.25,767.738 1662.19,767.738 1624.67,767.738 1586.68,767.738 1548.25,767.738 1509.41,767.738 1470.15,767.738 1430.49,767.738 1390.44,767.738 1350.01,767.738 
  1316.71,767.738 1287.89,767.738 1258.88,767.738 1229.71,767.738 1200.45,767.738 1171.11,767.738 1141.76,767.738 1112.43,767.738 1083.16,767.738 1054,767.738 
  1024.98,767.738 996.151,767.738 967.55,767.738 939.211,767.738 911.176,767.738 883.478,767.738 856.153,767.738 829.237,767.738 802.76,767.738 776.757,767.738 
  751.259,767.738 726.295,767.738 701.898,767.738 678.091,767.738 654.906,767.738 632.368,767.738 610.498,767.738 589.325,767.738 568.874,767.738 549.174,767.738 
  530.237,767.738 512.09,767.738 494.752,767.738 478.242,767.738 462.577,767.738 447.774,767.738 433.848,767.738 420.815,767.738 408.688,767.738 397.481,767.738 
  387.204,767.738 377.87,767.738 369.488,767.738 362.066,767.738 355.614,767.738 350.138,767.738 345.644,767.738 342.136,767.738 339.619,767.738 338.096,767.738 
  337.57,767.738 338.039,767.738 339.505,767.738 341.967,767.738 345.422,767.738 349.868,767.738 355.301,767.738 361.715,767.738 369.105,767.738 377.464,767.738 
  386.783,767.738 397.056,767.738 408.271,767.738 420.42,767.738 433.489,767.738 447.467,767.738 462.341,767.738 478.099,767.738 494.725,767.738 512.204,767.738 
  530.52,767.738 549.655,767.738 569.596,767.738 590.323,767.738 611.818,767.738 634.06,767.738 657.029,767.738 680.713,767.738 705.088,767.738 730.132,767.738 
  755.824,767.738 782.143,767.738 809.073,767.738 836.592,767.738 864.677,767.738 893.306,767.738 922.457,767.738 952.11,767.738 982.248,767.738 1012.85,767.738 
  1043.89,767.738 1075.37,767.738 1107.25,767.738 1139.53,767.738 1172.19,767.738 1205.22,767.738 1238.6,767.738 1272.33,767.738 1306.38,767.738 1338.44,767.738 
  1360.84,767.738 1383.42,767.738 1406.13,767.738 1428.95,767.738 1451.85,767.738 1474.79,767.738 1497.74,767.738 1520.67,767.738 1543.55,767.738 1566.35,767.738 
  1589.03,767.738 1611.57,767.738 1633.93,767.738 1656.09,767.738 1678.02,767.738 1699.69,767.738 1721.06,767.738 1742.13,767.738 1762.86,767.738 1783.22,767.738 
  1803.19,767.738 1822.74,767.738 1841.85,767.738 1860.5,767.738 1878.66,767.738 1896.31,767.738 1913.44,767.738 1930.02,767.738 1946.03,767.738 1961.46,767.738 
  1976.28,767.738 1990.49,767.738 2004.06,767.738 2016.98,767.738 2029.23,767.738 2040.81,767.738 2051.69,767.738 2061.87,767.738 2071.33,767.738 2080.07,767.738 
  2088.07,767.738 2095.32,767.738 2101.82,767.738 2107.56,767.738 2112.54,767.738 2116.73,767.738 2120.15,767.738 2122.79,767.738 2124.63,767.738 2125.69,767.738 
  2125.96,767.738 2125.43,767.738 2124.12,767.738 2122.01,767.738 2119.11,767.738 2115.42,767.738 2110.94,767.738 2105.69,767.738 2099.65,767.738 2092.85,767.738 
  2085.27,767.738 2076.94,767.738 2067.86,767.738 2058.03,767.738 2047.46,767.738 2036.16,767.738 2024.15,767.738 2011.44,767.738 1998.03,767.738 1983.93,767.738 
  1969.16,767.738 1953.74,767.738 1937.66,767.738 1920.96,767.738 1903.64,767.738 1885.71,767.738 1867.19,767.738 1848.1,767.738 1828.45,767.738 1808.26,767.738 
  1787.54,767.738 1766.3,767.738 1744.58,767.738 1722.37,767.738 1699.69,767.738 1676.57,767.738 1653.02,767.738 1629.06,767.738 1604.69,767.738 1579.94,767.738 
  1554.82,767.738 1529.34,767.738 1503.52,767.738 1477.37,767.738 1450.91,767.738 1424.15,767.738 1397.1,767.738 1369.77,767.738 1342.16,767.738 1323.05,767.738 
  1307.3,767.738 1291.43,767.738 1275.47,767.738 1259.44,767.738 1243.36,767.738 1227.25,767.738 1211.13,767.738 1195.02,767.738 1178.94,767.738 1162.91,767.738 
  1146.95,767.738 1131.1,767.738 1115.35,767.738 1099.74,767.738 1084.29,767.738 1069.01,767.738 1053.92,767.738 1039.04,767.738 1024.4,767.738 1010,767.738 
  995.873,767.738 982.025,767.738 968.477,767.738 955.247,767.738 942.352,767.738 929.809,767.738 917.629,767.738 905.829,767.738 894.424,767.738 883.427,767.738 
  872.851,767.738 862.709,767.738 853.012,767.738 843.772,767.738 835,767.738 826.706,767.738 818.901,767.738 811.594,767.738 804.793,767.738 798.507,767.738 
  792.743,767.738 787.509,767.738 782.811,767.738 778.655,767.738 775.046,767.738 771.99,767.738 769.491,767.738 767.552,767.738 766.176,767.738 765.366,767.738 
  765.124,767.738 765.452,767.738 766.35,767.738 767.818,767.738 769.856,767.738 772.463,767.738 775.638,767.738 779.379,767.738 783.683,767.738 788.548,767.738 
  793.97,767.738 799.943,767.738 806.465,767.738 813.53,767.738 821.132,767.738 829.266,767.738 837.925,767.738 847.102,767.738 856.79,767.738 866.981,767.738 
  877.667,767.738 888.841,767.738 900.491,767.738 912.611,767.738 925.189,767.738 938.218,767.738 951.686,767.738 965.583,767.738 979.899,767.738 994.622,767.738 
  1009.74,767.738 1025.25,767.738 1041.13,767.738 1057.37,767.738 1073.97,767.738 1090.91,767.738 1108.17,767.738 1125.76,767.738 1143.65,767.738 1161.83,767.738 
  1180.3,767.738 1199.03,767.738 1218.04,767.738 1237.29,767.738 1256.78,767.738 1276.5,767.738 1296.44,767.738 1316.59,767.738 1335.33,767.738 1343.58,767.738 
  1351.89,767.738 
  "/>
<polyline clip-path="url(#clip752)" style="stroke:#e26f46; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" points="
  1334.24,159.645 1369.02,191.088 1404.07,220.023 1439.32,246.58 1474.72,270.887 1510.21,293.081 1545.75,313.298 1581.27,331.672 1616.72,348.333 1652.05,363.409 
  1687.21,377.023 1722.15,389.293 1756.82,400.33 1791.17,410.239 1825.16,419.119 1858.74,427.061 1891.87,434.151 1924.51,440.469 1956.62,446.087 1988.17,451.074 
  2019.11,455.49 2049.4,459.391 2079.02,462.83 2107.93,465.855 2136.09,468.507 2163.47,470.826 2190.04,472.848 2215.78,474.605 2240.65,476.126 2264.63,477.438 
  2287.69,478.564 2309.81,479.526 2330.96,480.344 2351.12,481.036 2370.27,481.618 2388.39,482.103 2405.46,482.505 2421.46,482.835 2436.37,483.103 2450.19,483.319 
  2462.89,483.492 2474.46,483.627 2484.9,483.732 2494.18,483.812 2502.31,483.872 2509.27,483.917 2515.05,483.948 2519.65,483.97 2523.06,483.985 2525.29,483.993 
  2526.32,483.997 2526.16,483.997 2524.81,483.992 2522.27,483.981 2518.54,483.965 2513.63,483.941 2507.54,483.906 2500.27,483.858 2491.83,483.793 2482.24,483.707 
  2471.5,483.594 2459.62,483.45 2446.61,483.266 2432.49,483.037 2417.26,482.752 2400.96,482.404 2383.57,481.98 2365.14,481.469 2345.67,480.857 2325.18,480.13 
  2303.69,479.271 2281.22,478.261 2257.79,477.078 2233.43,475.702 2208.15,474.104 2181.98,472.258 2154.94,470.131 2127.06,467.687 2098.36,464.889 2068.86,461.691 
  2038.6,458.045 2007.59,453.898 1975.86,449.188 1943.44,443.85 1910.36,437.809 1876.63,430.979 1842.29,423.269 1807.35,414.577 1771.85,404.787 1735.81,393.771 
  1699.25,381.386 1662.19,367.472 1624.67,351.85 1586.68,334.318 1548.25,314.652 1509.41,292.601 1470.15,267.876 1430.49,240.159 1390.44,209.088 1350.01,174.252 
  1316.71,1359.62 1287.89,1334.54 1258.88,1311.12 1229.71,1289.29 1200.45,1269.01 1171.11,1250.19 1141.76,1232.76 1112.43,1216.66 1083.16,1201.82 1054,1188.15 
  1024.98,1175.58 996.151,1164.05 967.55,1153.49 939.211,1143.83 911.176,1135 883.478,1126.94 856.153,1119.6 829.237,1112.93 802.76,1106.86 776.757,1101.35 
  751.259,1096.36 726.295,1091.84 701.898,1087.76 678.091,1084.07 654.906,1080.74 632.368,1077.75 610.498,1075.05 589.325,1072.63 568.874,1070.46 549.174,1068.51 
  530.237,1066.78 512.09,1065.22 494.752,1063.84 478.242,1062.61 462.577,1061.52 447.774,1060.56 433.848,1059.7 420.815,1058.95 408.688,1058.29 397.481,1057.71 
  387.204,1057.21 377.87,1056.78 369.488,1056.41 362.066,1056.09 355.614,1055.82 350.138,1055.61 345.644,1055.43 342.136,1055.3 339.619,1055.21 338.096,1055.15 
  337.57,1055.13 338.039,1055.15 339.505,1055.2 341.967,1055.29 345.422,1055.43 349.868,1055.6 355.301,1055.81 361.715,1056.07 369.105,1056.39 377.464,1056.76 
  386.783,1057.19 397.056,1057.69 408.271,1058.27 420.42,1058.93 433.489,1059.68 447.467,1060.54 462.341,1061.51 478.099,1062.6 494.725,1063.84 512.204,1065.23 
  530.52,1066.8 549.655,1068.56 569.596,1070.53 590.323,1072.74 611.818,1075.21 634.06,1077.96 657.029,1081.04 680.713,1084.46 705.088,1088.27 730.132,1092.52 
  755.824,1097.23 782.143,1102.46 809.073,1108.26 836.592,1114.7 864.677,1121.83 893.306,1129.73 922.457,1138.47 952.11,1148.13 982.248,1158.81 1012.85,1170.62 
  1043.89,1183.66 1075.37,1198.06 1107.25,1213.95 1139.53,1231.5 1172.19,1250.85 1205.22,1272.21 1238.6,1295.77 1272.33,1321.75 1306.38,1350.42 1338.44,163.593 
  1360.84,183.948 1383.42,203.296 1406.13,221.65 1428.95,239.023 1451.85,255.442 1474.79,270.932 1497.74,285.518 1520.67,299.231 1543.55,312.103 1566.35,324.167 
  1589.03,335.457 1611.57,346.009 1633.93,355.859 1656.09,365.041 1678.02,373.591 1699.69,381.543 1721.06,388.93 1742.13,395.787 1762.86,402.145 1783.22,408.034 
  1803.19,413.482 1822.74,418.516 1841.85,423.167 1860.5,427.456 1878.66,431.409 1896.31,435.049 1913.44,438.397 1930.02,441.473 1946.03,444.296 1961.46,446.884 
  1976.28,449.254 1990.49,451.42 2004.06,453.398 2016.98,455.199 2029.23,456.838 2040.81,458.324 2051.69,459.669 2061.87,460.882 2071.33,461.972 2080.07,462.946 
  2088.07,463.811 2095.32,464.575 2101.82,465.243 2107.56,465.819 2112.54,466.308 2116.73,466.715 2120.15,467.041 2122.79,467.289 2124.63,467.462 2125.69,467.561 
  2125.96,467.586 2125.43,467.537 2124.12,467.414 2122.01,467.216 2119.11,466.942 2115.42,466.588 2110.94,466.153 2105.69,465.632 2099.65,465.021 2092.85,464.317 
  2085.27,463.512 2076.94,462.601 2067.86,461.576 2058.03,460.429 2047.46,459.152 2036.16,457.735 2024.15,456.167 2011.44,454.437 1998.03,452.53 1983.93,450.433 
  1969.16,448.13 1953.74,445.605 1937.66,442.839 1920.96,439.812 1903.64,436.502 1885.71,432.887 1867.19,428.939 1848.1,424.631 1828.45,419.933 1808.26,414.814 
  1787.54,409.237 1766.3,403.165 1744.58,396.557 1722.37,389.366 1699.69,381.545 1676.57,373.043 1653.02,363.802 1629.06,353.763 1604.69,342.858 1579.94,331.016 
  1554.82,318.158 1529.34,304.201 1503.52,289.055 1477.37,272.621 1450.91,254.793 1424.15,235.456 1397.1,214.481 1369.77,191.732 1342.16,167.056 1323.05,1365.41 
  1307.3,1351.23 1291.43,1337.52 1275.47,1324.3 1259.44,1311.55 1243.36,1299.3 1227.25,1287.52 1211.13,1276.23 1195.02,1265.41 1178.94,1255.06 1162.91,1245.18 
  1146.95,1235.75 1131.1,1226.76 1115.35,1218.21 1099.74,1210.09 1084.29,1202.37 1069.01,1195.05 1053.92,1188.11 1039.04,1181.55 1024.4,1175.34 1010,1169.48 
  995.873,1163.95 982.025,1158.73 968.477,1153.82 955.247,1149.2 942.352,1144.86 929.809,1140.79 917.629,1136.97 905.829,1133.39 894.424,1130.05 883.427,1126.93 
  872.851,1124.02 862.709,1121.31 853.012,1118.8 843.772,1116.47 835,1114.31 826.706,1112.33 818.901,1110.5 811.594,1108.83 804.793,1107.31 798.507,1105.93 
  792.743,1104.69 787.509,1103.58 782.811,1102.6 778.655,1101.74 775.046,1101.01 771.99,1100.39 769.491,1099.89 767.552,1099.51 766.176,1099.23 765.366,1099.07 
  765.124,1099.03 765.452,1099.09 766.35,1099.27 767.818,1099.56 769.856,1099.96 772.463,1100.48 775.638,1101.13 779.379,1101.89 783.683,1102.78 788.548,1103.8 
  793.97,1104.95 799.943,1106.24 806.465,1107.68 813.53,1109.27 821.132,1111.02 829.266,1112.93 837.925,1115.03 847.102,1117.3 856.79,1119.77 866.981,1122.44 
  877.667,1125.33 888.841,1128.45 900.491,1131.81 912.611,1135.43 925.189,1139.32 938.218,1143.5 951.686,1147.99 965.583,1152.79 979.899,1157.95 994.622,1163.47 
  1009.74,1169.37 1025.25,1175.69 1041.13,1182.45 1057.37,1189.68 1073.97,1197.39 1090.91,1205.63 1108.17,1214.43 1125.76,1223.82 1143.65,1233.84 1161.83,1244.53 
  1180.3,1255.92 1199.03,1268.07 1218.04,1281.01 1237.29,1294.8 1256.78,1309.49 1276.5,1325.13 1296.44,1341.79 1316.59,1359.52 1335.33,160.669 1343.58,168.364 
  1351.89,175.954 
  "/>
<polyline clip-path="url(#clip752)" style="stroke:#008000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" stroke-dasharray="4, 8" points="
  1334.24,767.738 1369.02,751.647 1404.07,735.443 1439.32,719.162 1474.72,702.841 1510.21,686.514 1545.75,670.214 1581.27,653.978 1616.72,637.838 1652.05,621.827 
  1687.21,605.975 1722.15,590.311 1756.82,574.863 1791.17,559.657 1825.16,544.715 1858.74,530.061 1891.87,515.714 1924.51,501.694 1956.62,488.016 1988.17,474.694 
  2019.11,461.744 2049.4,449.179 2079.02,437.005 2107.93,425.233 2136.09,413.871 2163.47,402.927 2190.04,392.403 2215.78,382.304 2240.65,372.635 2264.63,363.396 
  2287.69,354.59 2309.81,346.217 2330.96,338.278 2351.12,330.773 2370.27,323.7 2388.39,317.059 2405.46,310.848 2421.46,305.066 2436.37,299.711 2450.19,294.781 
  2462.89,290.274 2474.46,286.189 2484.9,282.524 2494.18,279.277 2502.31,276.447 2509.27,274.031 2515.05,272.03 2519.65,270.441 2523.06,269.263 2525.29,268.497 
  2526.32,268.141 2526.16,268.195 2524.81,268.66 2522.27,269.536 2518.54,270.822 2513.63,272.52 2507.54,274.631 2500.27,277.157 2491.83,280.098 2482.24,283.456 
  2471.5,287.234 2459.62,291.432 2446.61,296.054 2432.49,301.102 2417.26,306.576 2400.96,312.481 2383.57,318.818 2365.14,325.589 2345.67,332.797 2325.18,340.442 
  2303.69,348.527 2281.22,357.054 2257.79,366.023 2233.43,375.434 2208.15,385.288 2181.98,395.585 2154.94,406.324 2127.06,417.502 2098.36,429.117 2068.86,441.166 
  2038.6,453.647 2007.59,466.553 1975.86,479.879 1943.44,493.615 1910.36,507.757 1876.63,522.298 1842.29,537.226 1807.35,552.529 1771.85,568.196 1735.81,584.214 
  1699.25,600.569 1662.19,617.247 1624.67,634.232 1586.68,651.511 1548.25,669.067 1509.41,686.883 1470.15,704.945 1430.49,723.237 1390.44,741.74 1350.01,760.442 
  1316.71,775.851 1287.89,789.18 1258.88,802.591 1229.71,816.059 1200.45,829.556 1171.11,843.058 1141.76,856.54 1112.43,869.978 1083.16,883.346 1054,896.621 
  1024.98,909.78 996.151,922.801 967.55,935.662 939.211,948.345 911.176,960.829 883.478,973.097 856.153,985.134 829.237,996.922 802.76,1008.45 776.757,1019.7 
  751.259,1030.66 726.295,1041.32 701.898,1051.68 678.091,1061.71 654.906,1071.42 632.368,1080.8 610.498,1089.83 589.325,1098.53 568.874,1106.86 549.174,1114.85 
  530.237,1122.47 512.09,1129.73 494.752,1136.62 478.242,1143.15 462.577,1149.31 447.774,1155.09 433.848,1160.51 420.815,1165.55 408.688,1170.22 397.481,1174.52 
  387.204,1178.44 377.87,1181.99 369.488,1185.17 362.066,1187.98 355.614,1190.41 350.138,1192.47 345.644,1194.16 342.136,1195.47 339.619,1196.41 338.096,1196.98 
  337.57,1197.18 338.039,1197 339.505,1196.45 341.967,1195.53 345.422,1194.24 349.868,1192.57 355.301,1190.53 361.715,1188.11 369.105,1185.32 377.464,1182.15 
  386.783,1178.6 397.056,1174.68 408.271,1170.38 420.42,1165.7 433.489,1160.65 447.467,1155.21 462.341,1149.4 478.099,1143.21 494.725,1136.63 512.204,1129.68 
  530.52,1122.36 549.655,1114.65 569.596,1106.57 590.323,1098.12 611.818,1089.29 634.06,1080.1 657.029,1070.54 680.713,1060.61 705.088,1050.33 730.132,1039.69 
  755.824,1028.7 782.143,1017.37 809.073,1005.71 836.592,993.708 864.677,981.386 893.306,968.752 922.457,955.813 952.11,942.58 982.248,929.06 1012.85,915.265 
  1043.89,901.208 1075.37,886.896 1107.25,872.343 1139.53,857.561 1172.19,842.562 1205.22,827.355 1238.6,811.956 1272.33,796.376 1306.38,780.625 1338.44,765.795 
  1360.84,755.429 1383.42,744.988 1406.13,734.487 1428.95,723.947 1451.85,713.382 1474.79,702.809 1497.74,692.246 1520.67,681.711 1543.55,671.22 1566.35,660.791 
  1589.03,650.438 1611.57,640.18 1633.93,630.029 1656.09,620.001 1678.02,610.111 1699.69,600.372 1721.06,590.796 1742.13,581.395 1762.86,572.181 1783.22,563.165 
  1803.19,554.359 1822.74,545.773 1841.85,537.416 1860.5,529.295 1878.66,521.421 1896.31,513.799 1913.44,506.437 1930.02,499.341 1946.03,492.516 1961.46,485.968 
  1976.28,479.702 1990.49,473.721 2004.06,468.031 2016.98,462.634 2029.23,457.533 2040.81,452.733 2051.69,448.235 2061.87,444.041 2071.33,440.154 2080.07,436.576 
  2088.07,433.308 2095.32,430.353 2101.82,427.709 2107.56,425.38 2112.54,423.366 2116.73,421.669 2120.15,420.288 2122.79,419.225 2124.63,418.48 2125.69,418.053 
  2125.96,417.945 2125.43,418.157 2124.12,418.688 2122.01,419.539 2119.11,420.709 2115.42,422.2 2110.94,424.011 2105.69,426.141 2099.65,428.591 2092.85,431.361 
  2085.27,434.449 2076.94,437.855 2067.86,441.581 2058.03,445.623 2047.46,449.982 2036.16,454.656 2024.15,459.644 2011.44,464.944 1998.03,470.556 1983.93,476.478 
  1969.16,482.707 1953.74,489.241 1937.66,496.078 1920.96,503.215 1903.64,510.647 1885.71,518.373 1867.19,526.389 1848.1,534.69 1828.45,543.272 1808.26,552.131 
  1787.54,561.259 1766.3,570.653 1744.58,580.307 1722.37,590.214 1699.69,600.368 1676.57,610.762 1653.02,621.389 1629.06,632.239 1604.69,643.307 1579.94,654.584 
  1554.82,666.063 1529.34,677.734 1503.52,689.588 1477.37,701.618 1450.91,713.812 1424.15,726.163 1397.1,738.662 1369.77,751.301 1342.16,764.073 1323.05,772.913 
  1307.3,780.203 1291.43,787.543 1275.47,794.922 1259.44,802.331 1243.36,809.759 1227.25,817.196 1211.13,824.633 1195.02,832.058 1178.94,839.46 1162.91,846.83 
  1146.95,854.156 1131.1,861.429 1115.35,868.64 1099.74,875.777 1084.29,882.831 1069.01,889.794 1053.92,896.656 1039.04,903.408 1024.4,910.043 1010,916.552 
  995.873,922.926 982.025,929.16 968.477,935.246 955.247,941.176 942.352,946.942 929.809,952.539 917.629,957.961 905.829,963.202 894.424,968.257 883.427,973.12 
  872.851,977.787 862.709,982.252 853.012,986.513 843.772,990.565 835,994.404 826.706,998.026 818.901,1001.43 811.594,1004.61 804.793,1007.57 798.507,1010.29 
  792.743,1012.79 787.509,1015.06 782.811,1017.09 778.655,1018.88 775.046,1020.44 771.99,1021.75 769.491,1022.83 767.552,1023.66 766.176,1024.26 765.366,1024.6 
  765.124,1024.71 765.452,1024.57 766.35,1024.18 767.818,1023.55 769.856,1022.67 772.463,1021.55 775.638,1020.18 779.379,1018.57 783.683,1016.71 788.548,1014.61 
  793.97,1012.26 799.943,1009.67 806.465,1006.84 813.53,1003.77 821.132,1000.46 829.266,996.909 837.925,993.124 847.102,989.105 856.79,984.854 866.981,980.373 
  877.667,975.663 888.841,970.727 900.491,965.569 912.611,960.191 925.189,954.597 938.218,948.788 951.686,942.769 965.583,936.544 979.899,930.116 994.622,923.49 
  1009.74,916.669 1025.25,909.658 1041.13,902.462 1057.37,895.085 1073.97,887.533 1090.91,879.811 1108.17,871.923 1125.76,863.876 1143.65,855.675 1161.83,847.326 
  1180.3,838.834 1199.03,830.206 1218.04,821.447 1237.29,812.564 1256.78,803.561 1276.5,794.447 1296.44,785.226 1316.59,775.905 1335.33,767.236 1343.58,763.418 
  1351.89,759.572 
  "/>
<path clip-path="url(#clip750)" d="
M1423.66 477.116 L2284.85 477.116 L2284.85 166.076 L1423.66 166.076  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<polyline clip-path="url(#clip750)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1423.66,477.116 2284.85,477.116 2284.85,166.076 1423.66,166.076 1423.66,477.116 
  "/>
<polyline clip-path="url(#clip750)" style="stroke:#009af9; stroke-linecap:butt; stroke-linejoin:round; stroke-width:6; stroke-opacity:1; fill:none" points="
  1446.29,243.836 1582.1,243.836 
  "/>
<path clip-path="url(#clip750)" d="M1638.1 219.617 L1638.1 226.457 Q1634.1 224.548 1630.56 223.61 Q1627.02 222.673 1623.72 222.673 Q1617.99 222.673 1614.87 224.895 Q1611.78 227.117 1611.78 231.214 Q1611.78 234.652 1613.83 236.423 Q1615.91 238.159 1621.67 239.235 L1625.91 240.103 Q1633.76 241.596 1637.47 245.381 Q1641.22 249.131 1641.22 255.45 Q1641.22 262.985 1636.15 266.874 Q1631.12 270.763 1621.36 270.763 Q1617.68 270.763 1613.51 269.929 Q1609.38 269.096 1604.94 267.464 L1604.94 260.242 Q1609.21 262.638 1613.31 263.853 Q1617.4 265.068 1621.36 265.068 Q1627.37 265.068 1630.63 262.707 Q1633.9 260.346 1633.9 255.971 Q1633.9 252.152 1631.54 249.999 Q1629.21 247.846 1623.86 246.77 L1619.59 245.936 Q1611.74 244.374 1608.24 241.041 Q1604.73 237.707 1604.73 231.77 Q1604.73 224.895 1609.56 220.937 Q1614.42 216.978 1622.92 216.978 Q1626.57 216.978 1630.35 217.638 Q1634.14 218.298 1638.1 219.617 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M1651.88 230.867 L1658.27 230.867 L1658.27 269.756 L1651.88 269.756 L1651.88 230.867 M1651.88 215.728 L1658.27 215.728 L1658.27 223.818 L1651.88 223.818 L1651.88 215.728 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M1701.92 238.332 Q1704.31 234.027 1707.65 231.978 Q1710.98 229.93 1715.49 229.93 Q1721.57 229.93 1724.87 234.2 Q1728.17 238.436 1728.17 246.284 L1728.17 269.756 L1721.74 269.756 L1721.74 246.492 Q1721.74 240.902 1719.76 238.193 Q1717.78 235.485 1713.72 235.485 Q1708.76 235.485 1705.88 238.784 Q1702.99 242.082 1702.99 247.777 L1702.99 269.756 L1696.57 269.756 L1696.57 246.492 Q1696.57 240.867 1694.59 238.193 Q1692.61 235.485 1688.48 235.485 Q1683.58 235.485 1680.7 238.818 Q1677.82 242.117 1677.82 247.777 L1677.82 269.756 L1671.4 269.756 L1671.4 230.867 L1677.82 230.867 L1677.82 236.909 Q1680.01 233.332 1683.06 231.631 Q1686.12 229.93 1690.32 229.93 Q1694.56 229.93 1697.51 232.082 Q1700.49 234.235 1701.92 238.332 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M1747.09 263.922 L1747.09 284.547 L1740.67 284.547 L1740.67 230.867 L1747.09 230.867 L1747.09 236.77 Q1749.1 233.298 1752.16 231.631 Q1755.25 229.93 1759.52 229.93 Q1766.6 229.93 1771.01 235.555 Q1775.46 241.179 1775.46 250.346 Q1775.46 259.513 1771.01 265.138 Q1766.6 270.763 1759.52 270.763 Q1755.25 270.763 1752.16 269.096 Q1749.1 267.395 1747.09 263.922 M1768.83 250.346 Q1768.83 243.298 1765.91 239.304 Q1763.03 235.277 1757.96 235.277 Q1752.89 235.277 1749.97 239.304 Q1747.09 243.298 1747.09 250.346 Q1747.09 257.395 1749.97 261.422 Q1752.89 265.415 1757.96 265.415 Q1763.03 265.415 1765.91 261.422 Q1768.83 257.395 1768.83 250.346 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M1786.05 215.728 L1792.44 215.728 L1792.44 269.756 L1786.05 269.756 L1786.05 215.728 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M1839.07 248.714 L1839.07 251.839 L1809.69 251.839 Q1810.11 258.436 1813.65 261.909 Q1817.23 265.346 1823.58 265.346 Q1827.26 265.346 1830.7 264.443 Q1834.17 263.54 1837.58 261.735 L1837.58 267.777 Q1834.14 269.235 1830.53 269.999 Q1826.92 270.763 1823.2 270.763 Q1813.9 270.763 1808.44 265.346 Q1803.03 259.929 1803.03 250.693 Q1803.03 241.145 1808.17 235.555 Q1813.34 229.93 1822.09 229.93 Q1829.94 229.93 1834.49 234.999 Q1839.07 240.034 1839.07 248.714 M1832.68 246.839 Q1832.61 241.596 1829.73 238.471 Q1826.88 235.346 1822.16 235.346 Q1816.81 235.346 1813.58 238.367 Q1810.39 241.388 1809.9 246.874 L1832.68 246.839 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M1849.83 217.916 L1879.62 217.916 L1879.62 223.818 L1856.85 223.818 L1856.85 239.096 L1877.4 239.096 L1877.4 244.999 L1856.85 244.999 L1856.85 269.756 L1849.83 269.756 L1849.83 217.916 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M1890.74 217.916 L1901.19 217.916 L1914.42 253.193 L1927.71 217.916 L1938.17 217.916 L1938.17 269.756 L1931.33 269.756 L1931.33 224.235 L1917.96 259.79 L1910.91 259.79 L1897.54 224.235 L1897.54 269.756 L1890.74 269.756 L1890.74 217.916 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M1951.29 217.916 L1958.34 217.916 L1958.34 249.409 Q1958.34 257.742 1961.36 261.422 Q1964.38 265.068 1971.15 265.068 Q1977.89 265.068 1980.91 261.422 Q1983.93 257.742 1983.93 249.409 L1983.93 217.916 L1990.98 217.916 L1990.98 250.277 Q1990.98 260.415 1985.94 265.589 Q1980.94 270.763 1971.15 270.763 Q1961.33 270.763 1956.29 265.589 Q1951.29 260.415 1951.29 250.277 L1951.29 217.916 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip750)" style="stroke:#e26f46; stroke-linecap:butt; stroke-linejoin:round; stroke-width:6; stroke-opacity:1; fill:none" points="
  1446.29,321.596 1582.1,321.596 
  "/>
<path clip-path="url(#clip750)" d="M1627.51 314.599 Q1626.43 313.974 1625.15 313.696 Q1623.9 313.384 1622.37 313.384 Q1616.95 313.384 1614.04 316.926 Q1611.15 320.433 1611.15 327.03 L1611.15 347.516 L1604.73 347.516 L1604.73 308.627 L1611.15 308.627 L1611.15 314.669 Q1613.17 311.127 1616.4 309.426 Q1619.63 307.69 1624.24 307.69 Q1624.9 307.69 1625.7 307.794 Q1626.5 307.863 1627.47 308.037 L1627.51 314.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M1665.91 326.474 L1665.91 329.599 L1636.54 329.599 Q1636.95 336.196 1640.49 339.669 Q1644.07 343.106 1650.42 343.106 Q1654.1 343.106 1657.54 342.203 Q1661.01 341.3 1664.42 339.495 L1664.42 345.537 Q1660.98 346.995 1657.37 347.759 Q1653.76 348.523 1650.04 348.523 Q1640.74 348.523 1635.29 343.106 Q1629.87 337.689 1629.87 328.453 Q1629.87 318.905 1635.01 313.315 Q1640.18 307.69 1648.93 307.69 Q1656.78 307.69 1661.33 312.759 Q1665.91 317.794 1665.91 326.474 M1659.52 324.599 Q1659.45 319.356 1656.57 316.231 Q1653.72 313.106 1649 313.106 Q1643.65 313.106 1640.42 316.127 Q1637.23 319.148 1636.74 324.634 L1659.52 324.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M1696.08 293.488 L1696.08 298.801 L1689.97 298.801 Q1686.53 298.801 1685.18 300.19 Q1683.86 301.578 1683.86 305.19 L1683.86 308.627 L1694.38 308.627 L1694.38 313.592 L1683.86 313.592 L1683.86 347.516 L1677.44 347.516 L1677.44 313.592 L1671.33 313.592 L1671.33 308.627 L1677.44 308.627 L1677.44 305.919 Q1677.44 299.426 1680.46 296.474 Q1683.48 293.488 1690.04 293.488 L1696.08 293.488 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M1734.69 326.474 L1734.69 329.599 L1705.32 329.599 Q1705.74 336.196 1709.28 339.669 Q1712.85 343.106 1719.21 343.106 Q1722.89 343.106 1726.33 342.203 Q1729.8 341.3 1733.2 339.495 L1733.2 345.537 Q1729.76 346.995 1726.15 347.759 Q1722.54 348.523 1718.83 348.523 Q1709.52 348.523 1704.07 343.106 Q1698.65 337.689 1698.65 328.453 Q1698.65 318.905 1703.79 313.315 Q1708.97 307.69 1717.72 307.69 Q1725.56 307.69 1730.11 312.759 Q1734.69 317.794 1734.69 326.474 M1728.31 324.599 Q1728.24 319.356 1725.35 316.231 Q1722.51 313.106 1717.78 313.106 Q1712.44 313.106 1709.21 316.127 Q1706.01 319.148 1705.53 324.634 L1728.31 324.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M1767.72 314.599 Q1766.64 313.974 1765.35 313.696 Q1764.1 313.384 1762.58 313.384 Q1757.16 313.384 1754.24 316.926 Q1751.36 320.433 1751.36 327.03 L1751.36 347.516 L1744.94 347.516 L1744.94 308.627 L1751.36 308.627 L1751.36 314.669 Q1753.37 311.127 1756.6 309.426 Q1759.83 307.69 1764.45 307.69 Q1765.11 307.69 1765.91 307.794 Q1766.71 307.863 1767.68 308.037 L1767.72 314.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M1806.12 326.474 L1806.12 329.599 L1776.74 329.599 Q1777.16 336.196 1780.7 339.669 Q1784.28 343.106 1790.63 343.106 Q1794.31 343.106 1797.75 342.203 Q1801.22 341.3 1804.62 339.495 L1804.62 345.537 Q1801.19 346.995 1797.58 347.759 Q1793.96 348.523 1790.25 348.523 Q1780.94 348.523 1775.49 343.106 Q1770.08 337.689 1770.08 328.453 Q1770.08 318.905 1775.22 313.315 Q1780.39 307.69 1789.14 307.69 Q1796.99 307.69 1801.53 312.759 Q1806.12 317.794 1806.12 326.474 M1799.73 324.599 Q1799.66 319.356 1796.78 316.231 Q1793.93 313.106 1789.21 313.106 Q1783.86 313.106 1780.63 316.127 Q1777.44 319.148 1776.95 324.634 L1799.73 324.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M1848.93 324.044 L1848.93 347.516 L1842.54 347.516 L1842.54 324.252 Q1842.54 318.731 1840.39 315.988 Q1838.24 313.245 1833.93 313.245 Q1828.76 313.245 1825.77 316.544 Q1822.78 319.842 1822.78 325.537 L1822.78 347.516 L1816.36 347.516 L1816.36 308.627 L1822.78 308.627 L1822.78 314.669 Q1825.08 311.162 1828.17 309.426 Q1831.29 307.69 1835.35 307.69 Q1842.05 307.69 1845.49 311.856 Q1848.93 315.988 1848.93 324.044 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M1889.66 310.12 L1889.66 316.092 Q1886.95 314.599 1884.21 313.87 Q1881.5 313.106 1878.72 313.106 Q1872.51 313.106 1869.07 317.064 Q1865.63 320.988 1865.63 328.106 Q1865.63 335.224 1869.07 339.182 Q1872.51 343.106 1878.72 343.106 Q1881.5 343.106 1884.21 342.377 Q1886.95 341.613 1889.66 340.12 L1889.66 346.023 Q1886.99 347.273 1884.1 347.898 Q1881.26 348.523 1878.03 348.523 Q1869.24 348.523 1864.07 343.002 Q1858.9 337.481 1858.9 328.106 Q1858.9 318.592 1864.1 313.141 Q1869.35 307.69 1878.44 307.69 Q1881.4 307.69 1884.21 308.315 Q1887.02 308.905 1889.66 310.12 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M1934.03 326.474 L1934.03 329.599 L1904.66 329.599 Q1905.08 336.196 1908.62 339.669 Q1912.19 343.106 1918.55 343.106 Q1922.23 343.106 1925.67 342.203 Q1929.14 341.3 1932.54 339.495 L1932.54 345.537 Q1929.1 346.995 1925.49 347.759 Q1921.88 348.523 1918.17 348.523 Q1908.86 348.523 1903.41 343.106 Q1897.99 337.689 1897.99 328.453 Q1897.99 318.905 1903.13 313.315 Q1908.3 307.69 1917.05 307.69 Q1924.9 307.69 1929.45 312.759 Q1934.03 317.794 1934.03 326.474 M1927.64 324.599 Q1927.58 319.356 1924.69 316.231 Q1921.85 313.106 1917.12 313.106 Q1911.78 313.106 1908.55 316.127 Q1905.35 319.148 1904.87 324.634 L1927.64 324.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip750)" style="stroke:#008000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:6; stroke-opacity:1; fill:none" stroke-dasharray="3, 6" points="
  1446.29,399.356 1582.1,399.356 
  "/>
<path clip-path="url(#clip750)" d="M1604.73 373.436 L1614.17 373.436 L1637.16 416.804 L1637.16 373.436 L1643.97 373.436 L1643.97 425.276 L1634.52 425.276 L1611.54 381.908 L1611.54 425.276 L1604.73 425.276 L1604.73 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M1690.91 404.234 L1690.91 407.359 L1661.54 407.359 Q1661.95 413.956 1665.49 417.429 Q1669.07 420.866 1675.42 420.866 Q1679.1 420.866 1682.54 419.963 Q1686.01 419.06 1689.42 417.255 L1689.42 423.297 Q1685.98 424.755 1682.37 425.519 Q1678.76 426.283 1675.04 426.283 Q1665.74 426.283 1660.29 420.866 Q1654.87 415.449 1654.87 406.213 Q1654.87 396.665 1660.01 391.075 Q1665.18 385.45 1673.93 385.45 Q1681.78 385.45 1686.33 390.519 Q1690.91 395.554 1690.91 404.234 M1684.52 402.359 Q1684.45 397.116 1681.57 393.991 Q1678.72 390.866 1674 390.866 Q1668.65 390.866 1665.42 393.887 Q1662.23 396.908 1661.74 402.394 L1684.52 402.359 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M1700.74 409.929 L1700.74 386.387 L1707.13 386.387 L1707.13 409.686 Q1707.13 415.206 1709.28 417.984 Q1711.43 420.727 1715.74 420.727 Q1720.91 420.727 1723.9 417.429 Q1726.92 414.13 1726.92 408.436 L1726.92 386.387 L1733.31 386.387 L1733.31 425.276 L1726.92 425.276 L1726.92 419.304 Q1724.59 422.845 1721.5 424.581 Q1718.44 426.283 1714.38 426.283 Q1707.68 426.283 1704.21 422.116 Q1700.74 417.949 1700.74 409.929 M1716.81 385.45 L1716.81 385.45 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M1769 392.359 Q1767.92 391.734 1766.64 391.456 Q1765.39 391.144 1763.86 391.144 Q1758.44 391.144 1755.53 394.686 Q1752.65 398.193 1752.65 404.79 L1752.65 425.276 L1746.22 425.276 L1746.22 386.387 L1752.65 386.387 L1752.65 392.429 Q1754.66 388.887 1757.89 387.186 Q1761.12 385.45 1765.74 385.45 Q1766.4 385.45 1767.19 385.554 Q1767.99 385.623 1768.97 385.797 L1769 392.359 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M1793.37 405.727 Q1785.63 405.727 1782.65 407.498 Q1779.66 409.269 1779.66 413.54 Q1779.66 416.942 1781.88 418.956 Q1784.14 420.935 1787.99 420.935 Q1793.31 420.935 1796.5 417.185 Q1799.73 413.401 1799.73 407.151 L1799.73 405.727 L1793.37 405.727 M1806.12 403.088 L1806.12 425.276 L1799.73 425.276 L1799.73 419.373 Q1797.54 422.915 1794.28 424.616 Q1791.01 426.283 1786.29 426.283 Q1780.32 426.283 1776.78 422.949 Q1773.27 419.581 1773.27 413.956 Q1773.27 407.394 1777.65 404.061 Q1782.06 400.727 1790.77 400.727 L1799.73 400.727 L1799.73 400.102 Q1799.73 395.693 1796.81 393.297 Q1793.93 390.866 1788.69 390.866 Q1785.35 390.866 1782.19 391.665 Q1779.03 392.463 1776.12 394.061 L1776.12 388.158 Q1779.62 386.804 1782.92 386.144 Q1786.22 385.45 1789.35 385.45 Q1797.78 385.45 1801.95 389.825 Q1806.12 394.199 1806.12 403.088 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M1819.28 371.248 L1825.67 371.248 L1825.67 425.276 L1819.28 425.276 L1819.28 371.248 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M1839.31 373.436 L1869.1 373.436 L1869.1 379.338 L1846.33 379.338 L1846.33 394.616 L1866.88 394.616 L1866.88 400.519 L1846.33 400.519 L1846.33 425.276 L1839.31 425.276 L1839.31 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M1880.21 373.436 L1890.67 373.436 L1903.89 408.713 L1917.19 373.436 L1927.64 373.436 L1927.64 425.276 L1920.8 425.276 L1920.8 379.755 L1907.44 415.31 L1900.39 415.31 L1887.02 379.755 L1887.02 425.276 L1880.21 425.276 L1880.21 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M1940.77 373.436 L1947.82 373.436 L1947.82 404.929 Q1947.82 413.262 1950.84 416.942 Q1953.86 420.588 1960.63 420.588 Q1967.37 420.588 1970.39 416.942 Q1973.41 413.262 1973.41 404.929 L1973.41 373.436 L1980.46 373.436 L1980.46 405.797 Q1980.46 415.935 1975.42 421.109 Q1970.42 426.283 1960.63 426.283 Q1950.8 426.283 1945.77 421.109 Q1940.77 415.935 1940.77 405.797 L1940.77 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M2031.29 371.318 Q2026.64 379.304 2024.38 387.116 Q2022.12 394.929 2022.12 402.949 Q2022.12 410.97 2024.38 418.852 Q2026.67 426.699 2031.29 434.651 L2025.73 434.651 Q2020.53 426.491 2017.92 418.609 Q2015.35 410.727 2015.35 402.949 Q2015.35 395.206 2017.92 387.359 Q2020.49 379.512 2025.73 371.318 L2031.29 371.318 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M2050.63 419.373 L2075.11 419.373 L2075.11 425.276 L2042.19 425.276 L2042.19 419.373 Q2046.19 415.241 2053.06 408.297 Q2059.97 401.318 2061.74 399.304 Q2065.11 395.519 2066.43 392.915 Q2067.78 390.276 2067.78 387.741 Q2067.78 383.609 2064.87 381.005 Q2061.98 378.401 2057.33 378.401 Q2054.03 378.401 2050.35 379.547 Q2046.71 380.693 2042.54 383.019 L2042.54 375.936 Q2046.78 374.234 2050.46 373.366 Q2054.14 372.498 2057.19 372.498 Q2065.25 372.498 2070.04 376.526 Q2074.83 380.554 2074.83 387.29 Q2074.83 390.484 2073.62 393.366 Q2072.44 396.213 2069.28 400.102 Q2068.41 401.109 2063.76 405.936 Q2059.1 410.727 2050.63 419.373 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M2089.9 373.436 L2117.44 373.436 L2117.44 379.338 L2096.32 379.338 L2096.32 392.047 Q2097.85 391.526 2099.38 391.283 Q2100.91 391.005 2102.44 391.005 Q2111.12 391.005 2116.19 395.762 Q2121.25 400.519 2121.25 408.644 Q2121.25 417.012 2116.05 421.665 Q2110.84 426.283 2101.36 426.283 Q2098.1 426.283 2094.69 425.727 Q2091.32 425.172 2087.71 424.06 L2087.71 417.012 Q2090.84 418.713 2094.17 419.547 Q2097.5 420.38 2101.22 420.38 Q2107.23 420.38 2110.73 417.22 Q2114.24 414.06 2114.24 408.644 Q2114.24 403.227 2110.73 400.068 Q2107.23 396.908 2101.22 396.908 Q2098.41 396.908 2095.6 397.533 Q2092.82 398.158 2089.9 399.477 L2089.9 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M2150.07 378.054 Q2144.66 378.054 2141.91 383.401 Q2139.21 388.713 2139.21 399.408 Q2139.21 410.067 2141.91 415.415 Q2144.66 420.727 2150.07 420.727 Q2155.53 420.727 2158.23 415.415 Q2160.98 410.067 2160.98 399.408 Q2160.98 388.713 2158.23 383.401 Q2155.53 378.054 2150.07 378.054 M2150.07 372.498 Q2158.79 372.498 2163.37 379.408 Q2167.99 386.283 2167.99 399.408 Q2167.99 412.498 2163.37 419.408 Q2158.79 426.283 2150.07 426.283 Q2141.36 426.283 2136.74 419.408 Q2132.16 412.498 2132.16 399.408 Q2132.16 386.283 2136.74 379.408 Q2141.36 372.498 2150.07 372.498 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M2195.32 378.054 Q2189.9 378.054 2187.16 383.401 Q2184.45 388.713 2184.45 399.408 Q2184.45 410.067 2187.16 415.415 Q2189.9 420.727 2195.32 420.727 Q2200.77 420.727 2203.48 415.415 Q2206.22 410.067 2206.22 399.408 Q2206.22 388.713 2203.48 383.401 Q2200.77 378.054 2195.32 378.054 M2195.32 372.498 Q2204.03 372.498 2208.62 379.408 Q2213.23 386.283 2213.23 399.408 Q2213.23 412.498 2208.62 419.408 Q2204.03 426.283 2195.32 426.283 Q2186.6 426.283 2181.98 419.408 Q2177.4 412.498 2177.4 399.408 Q2177.4 386.283 2181.98 379.408 Q2186.6 372.498 2195.32 372.498 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip750)" d="M2223.65 371.318 L2229.21 371.318 Q2234.41 379.512 2236.98 387.359 Q2239.59 395.206 2239.59 402.949 Q2239.59 410.727 2236.98 418.609 Q2234.41 426.491 2229.21 434.651 L2223.65 434.651 Q2228.27 426.699 2230.53 418.852 Q2232.82 410.97 2232.82 402.949 Q2232.82 394.929 2230.53 387.116 Q2228.27 379.304 2223.65 371.318 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /></svg>




<?xml version="1.0" encoding="utf-8"?>
<svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" width="600" height="400" viewBox="0 0 2400 1600">
<defs>
  <clipPath id="clip840">
    <rect x="0" y="0" width="2400" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip840)" d="
M0 1600 L2400 1600 L2400 0 L0 0  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip841">
    <rect x="480" y="0" width="1681" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip840)" d="
M360.239 1401.58 L2352.76 1401.58 L2352.76 123.472 L360.239 123.472  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip842">
    <rect x="360" y="123" width="1994" height="1279"/>
  </clipPath>
</defs>
<polyline clip-path="url(#clip842)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  416.631,1401.58 416.631,123.472 
  "/>
<polyline clip-path="url(#clip842)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  886.564,1401.58 886.564,123.472 
  "/>
<polyline clip-path="url(#clip842)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1356.5,1401.58 1356.5,123.472 
  "/>
<polyline clip-path="url(#clip842)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1826.43,1401.58 1826.43,123.472 
  "/>
<polyline clip-path="url(#clip842)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  2296.36,1401.58 2296.36,123.472 
  "/>
<polyline clip-path="url(#clip840)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  360.239,1401.58 2352.76,1401.58 
  "/>
<polyline clip-path="url(#clip840)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  416.631,1401.58 416.631,1382.68 
  "/>
<polyline clip-path="url(#clip840)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  886.564,1401.58 886.564,1382.68 
  "/>
<polyline clip-path="url(#clip840)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1356.5,1401.58 1356.5,1382.68 
  "/>
<polyline clip-path="url(#clip840)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1826.43,1401.58 1826.43,1382.68 
  "/>
<polyline clip-path="url(#clip840)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  2296.36,1401.58 2296.36,1382.68 
  "/>
<path clip-path="url(#clip840)" d="M416.631 1438.36 Q411.214 1438.36 408.471 1443.71 Q405.763 1449.02 405.763 1459.71 Q405.763 1470.37 408.471 1475.72 Q411.214 1481.03 416.631 1481.03 Q422.082 1481.03 424.79 1475.72 Q427.533 1470.37 427.533 1459.71 Q427.533 1449.02 424.79 1443.71 Q422.082 1438.36 416.631 1438.36 M416.631 1432.8 Q425.346 1432.8 429.929 1439.71 Q434.547 1446.59 434.547 1459.71 Q434.547 1472.8 429.929 1479.71 Q425.346 1486.59 416.631 1486.59 Q407.916 1486.59 403.298 1479.71 Q398.714 1472.8 398.714 1459.71 Q398.714 1446.59 403.298 1439.71 Q407.916 1432.8 416.631 1432.8 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M872.137 1479.68 L883.595 1479.68 L883.595 1440.13 L871.13 1442.63 L871.13 1436.24 L883.526 1433.74 L890.54 1433.74 L890.54 1479.68 L901.998 1479.68 L901.998 1485.58 L872.137 1485.58 L872.137 1479.68 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M1348.48 1479.68 L1372.96 1479.68 L1372.96 1485.58 L1340.04 1485.58 L1340.04 1479.68 Q1344.03 1475.55 1350.91 1468.6 Q1357.82 1461.62 1359.59 1459.61 Q1362.96 1455.82 1364.28 1453.22 Q1365.63 1450.58 1365.63 1448.05 Q1365.63 1443.91 1362.71 1441.31 Q1359.83 1438.71 1355.18 1438.71 Q1351.88 1438.71 1348.2 1439.85 Q1344.55 1441 1340.39 1443.32 L1340.39 1436.24 Q1344.62 1434.54 1348.3 1433.67 Q1351.98 1432.8 1355.04 1432.8 Q1363.09 1432.8 1367.89 1436.83 Q1372.68 1440.86 1372.68 1447.59 Q1372.68 1450.79 1371.46 1453.67 Q1370.28 1456.52 1367.12 1460.41 Q1366.25 1461.41 1361.6 1466.24 Q1356.95 1471.03 1348.48 1479.68 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M1832.8 1457.63 Q1837.84 1458.7 1840.65 1462.11 Q1843.5 1465.51 1843.5 1470.51 Q1843.5 1478.18 1838.22 1482.39 Q1832.94 1486.59 1823.22 1486.59 Q1819.95 1486.59 1816.48 1485.93 Q1813.05 1485.3 1809.36 1484.02 L1809.36 1477.25 Q1812.28 1478.95 1815.75 1479.82 Q1819.23 1480.68 1823.01 1480.68 Q1829.61 1480.68 1833.05 1478.08 Q1836.52 1475.48 1836.52 1470.51 Q1836.52 1465.93 1833.29 1463.36 Q1830.09 1460.75 1824.36 1460.75 L1818.32 1460.75 L1818.32 1454.99 L1824.64 1454.99 Q1829.82 1454.99 1832.56 1452.94 Q1835.3 1450.86 1835.3 1446.97 Q1835.3 1442.98 1832.45 1440.86 Q1829.64 1438.71 1824.36 1438.71 Q1821.48 1438.71 1818.18 1439.33 Q1814.89 1439.96 1810.93 1441.27 L1810.93 1435.02 Q1814.92 1433.91 1818.39 1433.36 Q1821.9 1432.8 1824.99 1432.8 Q1832.98 1432.8 1837.63 1436.45 Q1842.28 1440.06 1842.28 1446.24 Q1842.28 1450.55 1839.82 1453.53 Q1837.35 1456.48 1832.8 1457.63 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M2300.88 1439.85 L2283.17 1467.52 L2300.88 1467.52 L2300.88 1439.85 M2299.04 1433.74 L2307.86 1433.74 L2307.86 1467.52 L2315.25 1467.52 L2315.25 1473.36 L2307.86 1473.36 L2307.86 1485.58 L2300.88 1485.58 L2300.88 1473.36 L2277.48 1473.36 L2277.48 1466.59 L2299.04 1433.74 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M1300.11 1527.62 L1300.11 1538.66 L1313.27 1538.66 L1313.27 1543.62 L1300.11 1543.62 L1300.11 1564.74 Q1300.11 1569.49 1301.39 1570.85 Q1302.71 1572.2 1306.71 1572.2 L1313.27 1572.2 L1313.27 1577.55 L1306.71 1577.55 Q1299.31 1577.55 1296.5 1574.8 Q1293.69 1572.03 1293.69 1564.74 L1293.69 1543.62 L1289 1543.62 L1289 1538.66 L1293.69 1538.66 L1293.69 1527.62 L1300.11 1527.62 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M1343.68 1523.52 L1358.41 1523.52 L1358.41 1528.49 L1350.07 1528.49 L1350.07 1581.96 L1358.41 1581.96 L1358.41 1586.92 L1343.68 1586.92 L1343.68 1523.52 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M1396.81 1539.8 L1396.81 1545.85 Q1394.1 1544.46 1391.18 1543.76 Q1388.27 1543.07 1385.14 1543.07 Q1380.39 1543.07 1377.99 1544.53 Q1375.63 1545.99 1375.63 1548.9 Q1375.63 1551.12 1377.33 1552.41 Q1379.03 1553.66 1384.17 1554.8 L1386.36 1555.29 Q1393.16 1556.75 1396.01 1559.42 Q1398.89 1562.06 1398.89 1566.82 Q1398.89 1572.24 1394.59 1575.4 Q1390.32 1578.55 1382.82 1578.55 Q1379.69 1578.55 1376.29 1577.93 Q1372.92 1577.34 1369.17 1576.12 L1369.17 1569.53 Q1372.71 1571.37 1376.15 1572.3 Q1379.59 1573.21 1382.96 1573.21 Q1387.47 1573.21 1389.9 1571.68 Q1392.33 1570.12 1392.33 1567.3 Q1392.33 1564.7 1390.56 1563.31 Q1388.82 1561.92 1382.89 1560.64 L1380.66 1560.12 Q1374.73 1558.87 1372.09 1556.3 Q1369.45 1553.69 1369.45 1549.18 Q1369.45 1543.69 1373.34 1540.71 Q1377.23 1537.72 1384.38 1537.72 Q1387.92 1537.72 1391.05 1538.24 Q1394.17 1538.76 1396.81 1539.8 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M1424 1523.52 L1424 1586.92 L1409.27 1586.92 L1409.27 1581.96 L1417.57 1581.96 L1417.57 1528.49 L1409.27 1528.49 L1409.27 1523.52 L1424 1523.52 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip842)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  360.239,1365.41 2352.76,1365.41 
  "/>
<polyline clip-path="url(#clip842)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  360.239,1063.97 2352.76,1063.97 
  "/>
<polyline clip-path="url(#clip842)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  360.239,762.526 2352.76,762.526 
  "/>
<polyline clip-path="url(#clip842)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  360.239,461.085 2352.76,461.085 
  "/>
<polyline clip-path="url(#clip842)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  360.239,159.645 2352.76,159.645 
  "/>
<polyline clip-path="url(#clip840)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  360.239,1401.58 360.239,123.472 
  "/>
<polyline clip-path="url(#clip840)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  360.239,1365.41 379.136,1365.41 
  "/>
<polyline clip-path="url(#clip840)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  360.239,1063.97 379.136,1063.97 
  "/>
<polyline clip-path="url(#clip840)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  360.239,762.526 379.136,762.526 
  "/>
<polyline clip-path="url(#clip840)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  360.239,461.085 379.136,461.085 
  "/>
<polyline clip-path="url(#clip840)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  360.239,159.645 379.136,159.645 
  "/>
<path clip-path="url(#clip840)" d="M118.58 1366.08 L163.094 1366.08 L163.094 1371.99 L118.58 1371.99 L118.58 1366.08 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M193.233 1344.11 Q187.816 1344.11 185.073 1349.45 Q182.365 1354.76 182.365 1365.46 Q182.365 1376.12 185.073 1381.47 Q187.816 1386.78 193.233 1386.78 Q198.684 1386.78 201.392 1381.47 Q204.135 1376.12 204.135 1365.46 Q204.135 1354.76 201.392 1349.45 Q198.684 1344.11 193.233 1344.11 M193.233 1338.55 Q201.948 1338.55 206.531 1345.46 Q211.149 1352.33 211.149 1365.46 Q211.149 1378.55 206.531 1385.46 Q201.948 1392.33 193.233 1392.33 Q184.517 1392.33 179.899 1385.46 Q175.316 1378.55 175.316 1365.46 Q175.316 1352.33 179.899 1345.46 Q184.517 1338.55 193.233 1338.55 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M223.475 1382.51 L230.802 1382.51 L230.802 1391.33 L223.475 1391.33 L223.475 1382.51 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M247.295 1385.42 L258.753 1385.42 L258.753 1345.88 L246.288 1348.38 L246.288 1341.99 L258.684 1339.49 L265.697 1339.49 L265.697 1385.42 L277.156 1385.42 L277.156 1391.33 L247.295 1391.33 L247.295 1385.42 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M306.322 1344.11 Q300.906 1344.11 298.163 1349.45 Q295.454 1354.76 295.454 1365.46 Q295.454 1376.12 298.163 1381.47 Q300.906 1386.78 306.322 1386.78 Q311.774 1386.78 314.482 1381.47 Q317.225 1376.12 317.225 1365.46 Q317.225 1354.76 314.482 1349.45 Q311.774 1344.11 306.322 1344.11 M306.322 1338.55 Q315.037 1338.55 319.621 1345.46 Q324.239 1352.33 324.239 1365.46 Q324.239 1378.55 319.621 1385.46 Q315.037 1392.33 306.322 1392.33 Q297.607 1392.33 292.989 1385.46 Q288.406 1378.55 288.406 1365.46 Q288.406 1352.33 292.989 1345.46 Q297.607 1338.55 306.322 1338.55 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M120.073 1064.64 L164.587 1064.64 L164.587 1070.55 L120.073 1070.55 L120.073 1064.64 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M194.726 1042.66 Q189.309 1042.66 186.566 1048.01 Q183.858 1053.32 183.858 1064.02 Q183.858 1074.68 186.566 1080.03 Q189.309 1085.34 194.726 1085.34 Q200.177 1085.34 202.885 1080.03 Q205.628 1074.68 205.628 1064.02 Q205.628 1053.32 202.885 1048.01 Q200.177 1042.66 194.726 1042.66 M194.726 1037.11 Q203.441 1037.11 208.024 1044.02 Q212.642 1050.89 212.642 1064.02 Q212.642 1077.11 208.024 1084.02 Q203.441 1090.89 194.726 1090.89 Q186.01 1090.89 181.392 1084.02 Q176.809 1077.11 176.809 1064.02 Q176.809 1050.89 181.392 1044.02 Q186.01 1037.11 194.726 1037.11 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M224.968 1081.07 L232.295 1081.07 L232.295 1089.89 L224.968 1089.89 L224.968 1081.07 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M262.572 1042.66 Q257.156 1042.66 254.413 1048.01 Q251.704 1053.32 251.704 1064.02 Q251.704 1074.68 254.413 1080.03 Q257.156 1085.34 262.572 1085.34 Q268.024 1085.34 270.732 1080.03 Q273.475 1074.68 273.475 1064.02 Q273.475 1053.32 270.732 1048.01 Q268.024 1042.66 262.572 1042.66 M262.572 1037.11 Q271.288 1037.11 275.871 1044.02 Q280.489 1050.89 280.489 1064.02 Q280.489 1077.11 275.871 1084.02 Q271.288 1090.89 262.572 1090.89 Q253.857 1090.89 249.239 1084.02 Q244.656 1077.11 244.656 1064.02 Q244.656 1050.89 249.239 1044.02 Q253.857 1037.11 262.572 1037.11 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M292.885 1038.05 L320.419 1038.05 L320.419 1043.95 L299.308 1043.95 L299.308 1056.66 Q300.836 1056.14 302.364 1055.89 Q303.892 1055.62 305.419 1055.62 Q314.1 1055.62 319.169 1060.37 Q324.239 1065.13 324.239 1073.25 Q324.239 1081.62 319.03 1086.28 Q313.822 1090.89 304.343 1090.89 Q301.079 1090.89 297.676 1090.34 Q294.308 1089.78 290.697 1088.67 L290.697 1081.62 Q293.822 1083.32 297.156 1084.16 Q300.489 1084.99 304.204 1084.99 Q310.211 1084.99 313.718 1081.83 Q317.225 1078.67 317.225 1073.25 Q317.225 1067.84 313.718 1064.68 Q310.211 1061.52 304.204 1061.52 Q301.392 1061.52 298.579 1062.14 Q295.801 1062.77 292.885 1064.09 L292.885 1038.05 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M193.233 741.224 Q187.816 741.224 185.073 746.571 Q182.365 751.884 182.365 762.578 Q182.365 773.238 185.073 778.585 Q187.816 783.897 193.233 783.897 Q198.684 783.897 201.392 778.585 Q204.135 773.238 204.135 762.578 Q204.135 751.884 201.392 746.571 Q198.684 741.224 193.233 741.224 M193.233 735.668 Q201.948 735.668 206.531 742.578 Q211.149 749.453 211.149 762.578 Q211.149 775.668 206.531 782.578 Q201.948 789.453 193.233 789.453 Q184.517 789.453 179.899 782.578 Q175.316 775.668 175.316 762.578 Q175.316 749.453 179.899 742.578 Q184.517 735.668 193.233 735.668 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M223.475 779.627 L230.802 779.627 L230.802 788.446 L223.475 788.446 L223.475 779.627 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M261.079 741.224 Q255.663 741.224 252.92 746.571 Q250.211 751.884 250.211 762.578 Q250.211 773.238 252.92 778.585 Q255.663 783.897 261.079 783.897 Q266.531 783.897 269.239 778.585 Q271.982 773.238 271.982 762.578 Q271.982 751.884 269.239 746.571 Q266.531 741.224 261.079 741.224 M261.079 735.668 Q269.795 735.668 274.378 742.578 Q278.996 749.453 278.996 762.578 Q278.996 775.668 274.378 782.578 Q269.795 789.453 261.079 789.453 Q252.364 789.453 247.746 782.578 Q243.163 775.668 243.163 762.578 Q243.163 749.453 247.746 742.578 Q252.364 735.668 261.079 735.668 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M306.322 741.224 Q300.906 741.224 298.163 746.571 Q295.454 751.884 295.454 762.578 Q295.454 773.238 298.163 778.585 Q300.906 783.897 306.322 783.897 Q311.774 783.897 314.482 778.585 Q317.225 773.238 317.225 762.578 Q317.225 751.884 314.482 746.571 Q311.774 741.224 306.322 741.224 M306.322 735.668 Q315.037 735.668 319.621 742.578 Q324.239 749.453 324.239 762.578 Q324.239 775.668 319.621 782.578 Q315.037 789.453 306.322 789.453 Q297.607 789.453 292.989 782.578 Q288.406 775.668 288.406 762.578 Q288.406 749.453 292.989 742.578 Q297.607 735.668 306.322 735.668 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M194.726 439.784 Q189.309 439.784 186.566 445.131 Q183.858 450.443 183.858 461.138 Q183.858 471.797 186.566 477.144 Q189.309 482.457 194.726 482.457 Q200.177 482.457 202.885 477.144 Q205.628 471.797 205.628 461.138 Q205.628 450.443 202.885 445.131 Q200.177 439.784 194.726 439.784 M194.726 434.228 Q203.441 434.228 208.024 441.138 Q212.642 448.013 212.642 461.138 Q212.642 474.228 208.024 481.137 Q203.441 488.012 194.726 488.012 Q186.01 488.012 181.392 481.137 Q176.809 474.228 176.809 461.138 Q176.809 448.013 181.392 441.138 Q186.01 434.228 194.726 434.228 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M224.968 478.186 L232.295 478.186 L232.295 487.005 L224.968 487.005 L224.968 478.186 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M262.572 439.784 Q257.156 439.784 254.413 445.131 Q251.704 450.443 251.704 461.138 Q251.704 471.797 254.413 477.144 Q257.156 482.457 262.572 482.457 Q268.024 482.457 270.732 477.144 Q273.475 471.797 273.475 461.138 Q273.475 450.443 270.732 445.131 Q268.024 439.784 262.572 439.784 M262.572 434.228 Q271.288 434.228 275.871 441.138 Q280.489 448.013 280.489 461.138 Q280.489 474.228 275.871 481.137 Q271.288 488.012 262.572 488.012 Q253.857 488.012 249.239 481.137 Q244.656 474.228 244.656 461.138 Q244.656 448.013 249.239 441.138 Q253.857 434.228 262.572 434.228 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M292.885 435.165 L320.419 435.165 L320.419 441.068 L299.308 441.068 L299.308 453.776 Q300.836 453.256 302.364 453.013 Q303.892 452.735 305.419 452.735 Q314.1 452.735 319.169 457.492 Q324.239 462.249 324.239 470.374 Q324.239 478.742 319.03 483.394 Q313.822 488.012 304.343 488.012 Q301.079 488.012 297.676 487.457 Q294.308 486.901 290.697 485.79 L290.697 478.742 Q293.822 480.443 297.156 481.276 Q300.489 482.11 304.204 482.11 Q310.211 482.11 313.718 478.95 Q317.225 475.79 317.225 470.374 Q317.225 464.957 313.718 461.797 Q310.211 458.638 304.204 458.638 Q301.392 458.638 298.579 459.263 Q295.801 459.888 292.885 461.207 L292.885 435.165 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M193.233 138.343 Q187.816 138.343 185.073 143.69 Q182.365 149.003 182.365 159.697 Q182.365 170.357 185.073 175.704 Q187.816 181.016 193.233 181.016 Q198.684 181.016 201.392 175.704 Q204.135 170.357 204.135 159.697 Q204.135 149.003 201.392 143.69 Q198.684 138.343 193.233 138.343 M193.233 132.787 Q201.948 132.787 206.531 139.697 Q211.149 146.572 211.149 159.697 Q211.149 172.787 206.531 179.697 Q201.948 186.572 193.233 186.572 Q184.517 186.572 179.899 179.697 Q175.316 172.787 175.316 159.697 Q175.316 146.572 179.899 139.697 Q184.517 132.787 193.233 132.787 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M223.475 176.746 L230.802 176.746 L230.802 185.565 L223.475 185.565 L223.475 176.746 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M247.295 179.662 L258.753 179.662 L258.753 140.114 L246.288 142.614 L246.288 136.225 L258.684 133.725 L265.697 133.725 L265.697 179.662 L277.156 179.662 L277.156 185.565 L247.295 185.565 L247.295 179.662 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M306.322 138.343 Q300.906 138.343 298.163 143.69 Q295.454 149.003 295.454 159.697 Q295.454 170.357 298.163 175.704 Q300.906 181.016 306.322 181.016 Q311.774 181.016 314.482 175.704 Q317.225 170.357 317.225 159.697 Q317.225 149.003 314.482 143.69 Q311.774 138.343 306.322 138.343 M306.322 132.787 Q315.037 132.787 319.621 139.697 Q324.239 146.572 324.239 159.697 Q324.239 172.787 319.621 179.697 Q315.037 186.572 306.322 186.572 Q297.607 186.572 292.989 179.697 Q288.406 172.787 288.406 159.697 Q288.406 146.572 292.989 139.697 Q297.607 132.787 306.322 132.787 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M34.0422 1042.44 L13.0007 1042.44 L13.0007 1036.05 L67.0282 1036.05 L67.0282 1042.44 L61.1949 1042.44 Q64.6671 1044.45 66.3684 1047.54 Q68.0351 1050.6 68.0351 1054.9 Q68.0351 1061.95 62.4101 1066.4 Q56.7852 1070.81 47.6185 1070.81 Q38.4519 1070.81 32.827 1066.4 Q27.202 1061.95 27.202 1054.9 Q27.202 1050.6 28.9034 1047.54 Q30.57 1044.45 34.0422 1042.44 M47.6185 1064.21 Q54.6671 1064.21 58.6949 1061.33 Q62.6879 1058.41 62.6879 1053.34 Q62.6879 1048.27 58.6949 1045.35 Q54.6671 1042.44 47.6185 1042.44 Q40.57 1042.44 36.5769 1045.35 Q32.5492 1048.27 32.5492 1053.34 Q32.5492 1058.41 36.5769 1061.33 Q40.57 1064.21 47.6185 1064.21 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M28.1395 1022.89 L28.1395 1016.5 L67.0282 1016.5 L67.0282 1022.89 L28.1395 1022.89 M13.0007 1022.89 L13.0007 1016.5 L21.0909 1016.5 L21.0909 1022.89 L13.0007 1022.89 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M29.2853 978.341 L35.3269 978.341 Q33.9381 981.049 33.2436 983.966 Q32.5492 986.882 32.5492 990.007 Q32.5492 994.764 34.0075 997.16 Q35.4658 999.521 38.3825 999.521 Q40.6047 999.521 41.8894 997.82 Q43.1394 996.118 44.2852 990.98 L44.7713 988.792 Q46.2297 981.987 48.9033 979.139 Q51.5421 976.257 56.299 976.257 Q61.7157 976.257 64.8754 980.563 Q68.0351 984.834 68.0351 992.334 Q68.0351 995.459 67.4101 998.862 Q66.8198 1002.23 65.6046 1005.98 L59.0074 1005.98 Q60.8476 1002.44 61.7851 999 Q62.6879 995.563 62.6879 992.195 Q62.6879 987.681 61.1601 985.25 Q59.5976 982.82 56.7852 982.82 Q54.181 982.82 52.7921 984.591 Q51.4032 986.327 50.1185 992.264 L49.5977 994.487 Q48.3477 1000.42 45.7783 1003.06 Q43.1741 1005.7 38.6603 1005.7 Q33.1742 1005.7 30.1881 1001.81 Q27.202 997.924 27.202 990.771 Q27.202 987.23 27.7228 984.105 Q28.2436 980.98 29.2853 978.341 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M61.1949 959.903 L81.8197 959.903 L81.8197 966.327 L28.1395 966.327 L28.1395 959.903 L34.0422 959.903 Q30.57 957.89 28.9034 954.834 Q27.202 951.744 27.202 947.473 Q27.202 940.39 32.827 935.98 Q38.4519 931.535 47.6185 931.535 Q56.7852 931.535 62.4101 935.98 Q68.0351 940.39 68.0351 947.473 Q68.0351 951.744 66.3684 954.834 Q64.6671 957.89 61.1949 959.903 M47.6185 938.167 Q40.57 938.167 36.5769 941.084 Q32.5492 943.966 32.5492 949.035 Q32.5492 954.105 36.5769 957.021 Q40.57 959.903 47.6185 959.903 Q54.6671 959.903 58.6949 957.021 Q62.6879 954.105 62.6879 949.035 Q62.6879 943.966 58.6949 941.084 Q54.6671 938.167 47.6185 938.167 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M13.0007 920.945 L13.0007 914.556 L67.0282 914.556 L67.0282 920.945 L13.0007 920.945 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M47.4797 883.515 Q47.4797 891.258 49.2505 894.244 Q51.0213 897.23 55.2921 897.23 Q58.6949 897.23 60.7087 895.008 Q62.6879 892.751 62.6879 888.897 Q62.6879 883.584 58.9379 880.39 Q55.1532 877.161 48.9033 877.161 L47.4797 877.161 L47.4797 883.515 M44.8408 870.772 L67.0282 870.772 L67.0282 877.161 L61.1254 877.161 Q64.6671 879.348 66.3684 882.612 Q68.0351 885.876 68.0351 890.598 Q68.0351 896.57 64.7018 900.112 Q61.3337 903.619 55.7088 903.619 Q49.1463 903.619 45.813 899.244 Q42.4797 894.834 42.4797 886.119 L42.4797 877.161 L41.8547 877.161 Q37.445 877.161 35.0492 880.077 Q32.6186 882.959 32.6186 888.202 Q32.6186 891.536 33.4172 894.695 Q34.2158 897.855 35.813 900.772 L29.9103 900.772 Q28.5561 897.265 27.8964 893.966 Q27.202 890.668 27.202 887.543 Q27.202 879.105 31.577 874.939 Q35.9519 870.772 44.8408 870.772 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M29.6325 829.626 L35.6047 829.626 Q34.1117 832.335 33.3825 835.078 Q32.6186 837.786 32.6186 840.564 Q32.6186 846.779 36.5769 850.216 Q40.5005 853.654 47.6185 853.654 Q54.7366 853.654 58.6949 850.216 Q62.6185 846.779 62.6185 840.564 Q62.6185 837.786 61.8893 835.078 Q61.1254 832.335 59.6324 829.626 L65.5351 829.626 Q66.7851 832.3 67.4101 835.182 Q68.0351 838.029 68.0351 841.258 Q68.0351 850.043 62.5143 855.216 Q56.9935 860.39 47.6185 860.39 Q38.1047 860.39 32.6533 855.182 Q27.202 849.939 27.202 840.842 Q27.202 837.89 27.827 835.078 Q28.4173 832.265 29.6325 829.626 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M45.9866 785.252 L49.1116 785.252 L49.1116 814.626 Q55.7088 814.21 59.181 810.668 Q62.6185 807.092 62.6185 800.738 Q62.6185 797.057 61.7157 793.62 Q60.8129 790.147 59.0074 786.745 L65.049 786.745 Q66.5073 790.182 67.2712 793.793 Q68.0351 797.404 68.0351 801.12 Q68.0351 810.425 62.6185 815.876 Q57.2018 821.293 47.9658 821.293 Q38.4172 821.293 32.827 816.154 Q27.202 810.981 27.202 802.231 Q27.202 794.383 32.2714 789.835 Q37.3061 785.252 45.9866 785.252 M44.1116 791.64 Q38.8686 791.71 35.7436 794.592 Q32.6186 797.439 32.6186 802.161 Q32.6186 807.508 35.6394 810.738 Q38.6603 813.932 44.1463 814.418 L44.1116 791.64 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M35.6047 744.488 Q31.2992 742.092 29.2506 738.759 Q27.202 735.425 27.202 730.912 Q27.202 724.835 31.4728 721.537 Q35.7089 718.238 43.5561 718.238 L67.0282 718.238 L67.0282 724.662 L43.7644 724.662 Q38.1741 724.662 35.4658 726.641 Q32.7575 728.62 32.7575 732.682 Q32.7575 737.648 36.0561 740.53 Q39.3547 743.412 45.0491 743.412 L67.0282 743.412 L67.0282 749.835 L43.7644 749.835 Q38.1394 749.835 35.4658 751.814 Q32.7575 753.793 32.7575 757.925 Q32.7575 762.821 36.0908 765.703 Q39.3894 768.585 45.0491 768.585 L67.0282 768.585 L67.0282 775.009 L28.1395 775.009 L28.1395 768.585 L34.1811 768.585 Q30.6047 766.397 28.9034 763.342 Q27.202 760.286 27.202 756.085 Q27.202 751.849 29.3547 748.898 Q31.5075 745.911 35.6047 744.488 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M45.9866 672.231 L49.1116 672.231 L49.1116 701.606 Q55.7088 701.19 59.181 697.648 Q62.6185 694.071 62.6185 687.717 Q62.6185 684.037 61.7157 680.599 Q60.8129 677.127 59.0074 673.724 L65.049 673.724 Q66.5073 677.162 67.2712 680.773 Q68.0351 684.384 68.0351 688.099 Q68.0351 697.405 62.6185 702.856 Q57.2018 708.273 47.9658 708.273 Q38.4172 708.273 32.827 703.134 Q27.202 697.96 27.202 689.21 Q27.202 681.363 32.2714 676.815 Q37.3061 672.231 45.9866 672.231 M44.1116 678.62 Q38.8686 678.69 35.7436 681.572 Q32.6186 684.419 32.6186 689.141 Q32.6186 694.488 35.6394 697.717 Q38.6603 700.912 44.1463 701.398 L44.1116 678.62 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M43.5561 629.419 L67.0282 629.419 L67.0282 635.808 L43.7644 635.808 Q38.2436 635.808 35.5005 637.961 Q32.7575 640.113 32.7575 644.419 Q32.7575 649.593 36.0561 652.579 Q39.3547 655.565 45.0491 655.565 L67.0282 655.565 L67.0282 661.988 L28.1395 661.988 L28.1395 655.565 L34.1811 655.565 Q30.6742 653.273 28.9381 650.183 Q27.202 647.058 27.202 642.995 Q27.202 636.294 31.3686 632.857 Q35.5005 629.419 43.5561 629.419 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M17.0979 610.357 L28.1395 610.357 L28.1395 597.197 L33.1047 597.197 L33.1047 610.357 L54.2157 610.357 Q58.9726 610.357 60.3268 609.072 Q61.681 607.753 61.681 603.759 L61.681 597.197 L67.0282 597.197 L67.0282 603.759 Q67.0282 611.155 64.2851 613.968 Q61.5074 616.78 54.2157 616.78 L33.1047 616.78 L33.1047 621.468 L28.1395 621.468 L28.1395 616.78 L17.0979 616.78 L17.0979 610.357 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M13.0007 566.781 L13.0007 552.058 L17.9659 552.058 L17.9659 560.392 L71.4379 560.392 L71.4379 552.058 L76.4031 552.058 L76.4031 566.781 L13.0007 566.781 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M35.6047 508.17 Q31.2992 505.774 29.2506 502.441 Q27.202 499.107 27.202 494.593 Q27.202 488.517 31.4728 485.218 Q35.7089 481.92 43.5561 481.92 L67.0282 481.92 L67.0282 488.343 L43.7644 488.343 Q38.1741 488.343 35.4658 490.323 Q32.7575 492.302 32.7575 496.364 Q32.7575 501.329 36.0561 504.211 Q39.3547 507.093 45.0491 507.093 L67.0282 507.093 L67.0282 513.517 L43.7644 513.517 Q38.1394 513.517 35.4658 515.496 Q32.7575 517.475 32.7575 521.607 Q32.7575 526.503 36.0908 529.385 Q39.3894 532.267 45.0491 532.267 L67.0282 532.267 L67.0282 538.69 L28.1395 538.69 L28.1395 532.267 L34.1811 532.267 Q30.6047 530.079 28.9034 527.024 Q27.202 523.968 27.202 519.767 Q27.202 515.531 29.3547 512.579 Q31.5075 509.593 35.6047 508.17 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M13.0007 454.246 L76.4031 454.246 L76.4031 468.969 L71.4379 468.969 L71.4379 460.67 L17.9659 460.67 L17.9659 468.969 L13.0007 468.969 L13.0007 454.246 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M905.49 18.8205 L905.49 65.8515 L915.374 65.8515 Q927.891 65.8515 933.684 60.1802 Q939.518 54.509 939.518 42.2752 Q939.518 30.1225 933.684 24.4918 Q927.891 18.8205 915.374 18.8205 L905.49 18.8205 M897.307 12.096 L914.118 12.096 Q931.699 12.096 939.923 19.4281 Q948.146 26.7198 948.146 42.2752 Q948.146 57.9117 939.882 65.2439 Q931.618 72.576 914.118 72.576 L897.307 72.576 L897.307 12.096 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M960.866 27.2059 L968.319 27.2059 L968.319 72.576 L960.866 72.576 L960.866 27.2059 M960.866 9.54393 L968.319 9.54393 L968.319 18.9825 L960.866 18.9825 L960.866 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M1012.84 28.5427 L1012.84 35.5912 Q1009.68 33.9709 1006.28 33.1607 Q1002.87 32.3505 999.228 32.3505 Q993.678 32.3505 990.883 34.0519 Q988.128 35.7533 988.128 39.156 Q988.128 41.7486 990.113 43.2475 Q992.098 44.7058 998.094 46.0426 L1000.65 46.6097 Q1008.59 48.3111 1011.91 51.4303 Q1015.27 54.509 1015.27 60.0587 Q1015.27 66.3781 1010.25 70.0644 Q1005.26 73.7508 996.514 73.7508 Q992.868 73.7508 988.898 73.0216 Q984.969 72.3329 980.594 70.9151 L980.594 63.2184 Q984.726 65.3654 988.736 66.4591 Q992.746 67.5124 996.676 67.5124 Q1001.94 67.5124 1004.78 65.73 Q1007.61 63.9071 1007.61 60.6258 Q1007.61 57.5877 1005.55 55.9673 Q1003.52 54.3469 996.595 52.8481 L994.002 52.2405 Q987.075 50.7821 983.996 47.7845 Q980.918 44.7463 980.918 39.4801 Q980.918 33.0797 985.455 29.5959 Q989.992 26.1121 998.337 26.1121 Q1002.47 26.1121 1006.11 26.7198 Q1009.76 27.3274 1012.84 28.5427 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M1034.35 65.7705 L1034.35 89.8329 L1026.86 89.8329 L1026.86 27.2059 L1034.35 27.2059 L1034.35 34.0924 Q1036.7 30.0415 1040.26 28.0971 Q1043.87 26.1121 1048.85 26.1121 Q1057.12 26.1121 1062.26 32.6746 Q1067.45 39.2371 1067.45 49.9314 Q1067.45 60.6258 1062.26 67.1883 Q1057.12 73.7508 1048.85 73.7508 Q1043.87 73.7508 1040.26 71.8063 Q1036.7 69.8214 1034.35 65.7705 M1059.71 49.9314 Q1059.71 41.7081 1056.31 37.0496 Q1052.94 32.3505 1047.03 32.3505 Q1041.11 32.3505 1037.71 37.0496 Q1034.35 41.7081 1034.35 49.9314 Q1034.35 58.1548 1037.71 62.8538 Q1041.11 67.5124 1047.03 67.5124 Q1052.94 67.5124 1056.31 62.8538 Q1059.71 58.1548 1059.71 49.9314 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M1079.8 9.54393 L1087.25 9.54393 L1087.25 72.576 L1079.8 72.576 L1079.8 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M1123.47 49.7694 Q1114.44 49.7694 1110.95 51.8354 Q1107.47 53.9013 1107.47 58.8839 Q1107.47 62.8538 1110.06 65.2034 Q1112.69 67.5124 1117.19 67.5124 Q1123.39 67.5124 1127.11 63.1374 Q1130.88 58.7219 1130.88 51.4303 L1130.88 49.7694 L1123.47 49.7694 M1138.34 46.6907 L1138.34 72.576 L1130.88 72.576 L1130.88 65.6895 Q1128.33 69.8214 1124.52 71.8063 Q1120.71 73.7508 1115.21 73.7508 Q1108.24 73.7508 1104.11 69.8619 Q1100.01 65.9325 1100.01 59.3701 Q1100.01 51.7138 1105.12 47.825 Q1110.26 43.9361 1120.43 43.9361 L1130.88 43.9361 L1130.88 43.2069 Q1130.88 38.0623 1127.48 35.2672 Q1124.12 32.4315 1118 32.4315 Q1114.11 32.4315 1110.43 33.3632 Q1106.74 34.295 1103.34 36.1584 L1103.34 29.2718 Q1107.43 27.692 1111.28 26.9223 Q1115.12 26.1121 1118.77 26.1121 Q1128.61 26.1121 1133.47 31.2163 Q1138.34 36.3204 1138.34 46.6907 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M1186.34 28.9478 L1186.34 35.9153 Q1183.18 34.1734 1179.98 33.3227 Q1176.82 32.4315 1173.58 32.4315 Q1166.33 32.4315 1162.32 37.0496 Q1158.31 41.6271 1158.31 49.9314 Q1158.31 58.2358 1162.32 62.8538 Q1166.33 67.4314 1173.58 67.4314 Q1176.82 67.4314 1179.98 66.5807 Q1183.18 65.6895 1186.34 63.9476 L1186.34 70.8341 Q1183.22 72.2924 1179.86 73.0216 Q1176.54 73.7508 1172.77 73.7508 Q1162.52 73.7508 1156.48 67.3098 Q1150.45 60.8689 1150.45 49.9314 Q1150.45 38.832 1156.52 32.472 Q1162.64 26.1121 1173.25 26.1121 Q1176.7 26.1121 1179.98 26.8413 Q1183.26 27.5299 1186.34 28.9478 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M1238.11 48.0275 L1238.11 51.6733 L1203.84 51.6733 Q1204.33 59.3701 1208.46 63.421 Q1212.63 67.4314 1220.04 67.4314 Q1224.34 67.4314 1228.35 66.3781 Q1232.4 65.3249 1236.37 63.2184 L1236.37 70.267 Q1232.36 71.9684 1228.14 72.8596 Q1223.93 73.7508 1219.6 73.7508 Q1208.74 73.7508 1202.38 67.4314 Q1196.06 61.1119 1196.06 50.3365 Q1196.06 39.1965 1202.06 32.6746 Q1208.09 26.1121 1218.3 26.1121 Q1227.46 26.1121 1232.76 32.0264 Q1238.11 37.9003 1238.11 48.0275 M1230.66 45.84 Q1230.57 39.7232 1227.21 36.0774 Q1223.89 32.4315 1218.38 32.4315 Q1212.14 32.4315 1208.38 35.9558 Q1204.65 39.4801 1204.08 45.8805 L1230.66 45.84 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M1285.67 35.9153 Q1288.46 30.8922 1292.35 28.5022 Q1296.24 26.1121 1301.51 26.1121 Q1308.6 26.1121 1312.44 31.0947 Q1316.29 36.0368 1316.29 45.1919 L1316.29 72.576 L1308.8 72.576 L1308.8 45.4349 Q1308.8 38.913 1306.49 35.7533 Q1304.18 32.5936 1299.44 32.5936 Q1293.65 32.5936 1290.29 36.4419 Q1286.92 40.2903 1286.92 46.9338 L1286.92 72.576 L1279.43 72.576 L1279.43 45.4349 Q1279.43 38.8725 1277.12 35.7533 Q1274.81 32.5936 1269.99 32.5936 Q1264.28 32.5936 1260.92 36.4824 Q1257.55 40.3308 1257.55 46.9338 L1257.55 72.576 L1250.06 72.576 L1250.06 27.2059 L1257.55 27.2059 L1257.55 34.2544 Q1260.11 30.082 1263.67 28.0971 Q1267.24 26.1121 1272.14 26.1121 Q1277.08 26.1121 1280.52 28.6237 Q1284.01 31.1352 1285.67 35.9153 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M1369.97 48.0275 L1369.97 51.6733 L1335.7 51.6733 Q1336.18 59.3701 1340.31 63.421 Q1344.49 67.4314 1351.9 67.4314 Q1356.19 67.4314 1360.2 66.3781 Q1364.25 65.3249 1368.22 63.2184 L1368.22 70.267 Q1364.21 71.9684 1360 72.8596 Q1355.79 73.7508 1351.45 73.7508 Q1340.6 73.7508 1334.24 67.4314 Q1327.92 61.1119 1327.92 50.3365 Q1327.92 39.1965 1333.91 32.6746 Q1339.95 26.1121 1350.16 26.1121 Q1359.31 26.1121 1364.62 32.0264 Q1369.97 37.9003 1369.97 48.0275 M1362.51 45.84 Q1362.43 39.7232 1359.07 36.0774 Q1355.75 32.4315 1350.24 32.4315 Q1344 32.4315 1340.23 35.9558 Q1336.51 39.4801 1335.94 45.8805 L1362.51 45.84 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M1419.91 45.1919 L1419.91 72.576 L1412.46 72.576 L1412.46 45.4349 Q1412.46 38.994 1409.95 35.7938 Q1407.44 32.5936 1402.41 32.5936 Q1396.38 32.5936 1392.89 36.4419 Q1389.41 40.2903 1389.41 46.9338 L1389.41 72.576 L1381.92 72.576 L1381.92 27.2059 L1389.41 27.2059 L1389.41 34.2544 Q1392.08 30.163 1395.69 28.1376 Q1399.34 26.1121 1404.08 26.1121 Q1411.89 26.1121 1415.9 30.9732 Q1419.91 35.7938 1419.91 45.1919 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M1442.15 14.324 L1442.15 27.2059 L1457.51 27.2059 L1457.51 32.9987 L1442.15 32.9987 L1442.15 57.6282 Q1442.15 63.1779 1443.65 64.7578 Q1445.19 66.3376 1449.85 66.3376 L1457.51 66.3376 L1457.51 72.576 L1449.85 72.576 Q1441.22 72.576 1437.94 69.3758 Q1434.66 66.1351 1434.66 57.6282 L1434.66 32.9987 L1429.19 32.9987 L1429.19 27.2059 L1434.66 27.2059 L1434.66 14.324 L1442.15 14.324 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M1529.01 35.9153 Q1531.8 30.8922 1535.69 28.5022 Q1539.58 26.1121 1544.84 26.1121 Q1551.93 26.1121 1555.78 31.0947 Q1559.63 36.0368 1559.63 45.1919 L1559.63 72.576 L1552.14 72.576 L1552.14 45.4349 Q1552.14 38.913 1549.83 35.7533 Q1547.52 32.5936 1542.78 32.5936 Q1536.99 32.5936 1533.62 36.4419 Q1530.26 40.2903 1530.26 46.9338 L1530.26 72.576 L1522.77 72.576 L1522.77 45.4349 Q1522.77 38.8725 1520.46 35.7533 Q1518.15 32.5936 1513.33 32.5936 Q1507.62 32.5936 1504.25 36.4824 Q1500.89 40.3308 1500.89 46.9338 L1500.89 72.576 L1493.4 72.576 L1493.4 27.2059 L1500.89 27.2059 L1500.89 34.2544 Q1503.44 30.082 1507.01 28.0971 Q1510.57 26.1121 1515.48 26.1121 Q1520.42 26.1121 1523.86 28.6237 Q1527.34 31.1352 1529.01 35.9153 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M1592.08 32.4315 Q1586.08 32.4315 1582.6 37.1306 Q1579.11 41.7891 1579.11 49.9314 Q1579.11 58.0738 1582.56 62.7728 Q1586.04 67.4314 1592.08 67.4314 Q1598.03 67.4314 1601.52 62.7323 Q1605 58.0333 1605 49.9314 Q1605 41.8701 1601.52 37.1711 Q1598.03 32.4315 1592.08 32.4315 M1592.08 26.1121 Q1601.8 26.1121 1607.35 32.4315 Q1612.9 38.7509 1612.9 49.9314 Q1612.9 61.0714 1607.35 67.4314 Q1601.8 73.7508 1592.08 73.7508 Q1582.31 73.7508 1576.77 67.4314 Q1571.26 61.0714 1571.26 49.9314 Q1571.26 38.7509 1576.77 32.4315 Q1582.31 26.1121 1592.08 26.1121 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M1655.11 34.0924 L1655.11 9.54393 L1662.56 9.54393 L1662.56 72.576 L1655.11 72.576 L1655.11 65.7705 Q1652.76 69.8214 1649.15 71.8063 Q1645.59 73.7508 1640.57 73.7508 Q1632.34 73.7508 1627.16 67.1883 Q1622.01 60.6258 1622.01 49.9314 Q1622.01 39.2371 1627.16 32.6746 Q1632.34 26.1121 1640.57 26.1121 Q1645.59 26.1121 1649.15 28.0971 Q1652.76 30.0415 1655.11 34.0924 M1629.71 49.9314 Q1629.71 58.1548 1633.07 62.8538 Q1636.48 67.5124 1642.39 67.5124 Q1648.3 67.5124 1651.71 62.8538 Q1655.11 58.1548 1655.11 49.9314 Q1655.11 41.7081 1651.71 37.0496 Q1648.3 32.3505 1642.39 32.3505 Q1636.48 32.3505 1633.07 37.0496 Q1629.71 41.7081 1629.71 49.9314 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M1716.72 48.0275 L1716.72 51.6733 L1682.45 51.6733 Q1682.94 59.3701 1687.07 63.421 Q1691.24 67.4314 1698.66 67.4314 Q1702.95 67.4314 1706.96 66.3781 Q1711.01 65.3249 1714.98 63.2184 L1714.98 70.267 Q1710.97 71.9684 1706.76 72.8596 Q1702.55 73.7508 1698.21 73.7508 Q1687.35 73.7508 1681 67.4314 Q1674.68 61.1119 1674.68 50.3365 Q1674.68 39.1965 1680.67 32.6746 Q1686.71 26.1121 1696.92 26.1121 Q1706.07 26.1121 1711.38 32.0264 Q1716.72 37.9003 1716.72 48.0275 M1709.27 45.84 Q1709.19 39.7232 1705.83 36.0774 Q1702.51 32.4315 1697 32.4315 Q1690.76 32.4315 1686.99 35.9558 Q1683.26 39.4801 1682.7 45.8805 L1709.27 45.84 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M1728.96 9.54393 L1736.41 9.54393 L1736.41 72.576 L1728.96 72.576 L1728.96 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M1780.85 65.6895 L1794.22 65.6895 L1794.22 19.5497 L1779.68 22.4663 L1779.68 15.0127 L1794.14 12.096 L1802.32 12.096 L1802.32 65.6895 L1815.69 65.6895 L1815.69 72.576 L1780.85 72.576 L1780.85 65.6895 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip842)" style="stroke:#009af9; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" points="
  416.631,159.645 2296.36,159.645 
  "/>
<polyline clip-path="url(#clip842)" style="stroke:#e26f46; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" points="
  416.631,762.526 2296.36,762.526 
  "/>
<polyline clip-path="url(#clip842)" style="stroke:#008000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" stroke-dasharray="4, 8" points="
  416.631,1365.41 421.33,1365.41 426.029,1365.41 430.729,1365.41 435.428,1365.41 440.127,1365.41 444.827,1365.41 449.526,1365.41 454.225,1365.41 458.925,1365.41 
  463.624,1365.41 468.323,1365.41 473.023,1365.41 477.722,1365.41 482.421,1365.41 487.121,1365.41 491.82,1365.41 496.519,1365.41 501.219,1365.41 505.918,1365.41 
  510.617,1365.41 515.317,1365.41 520.016,1365.41 524.715,1365.41 529.415,1365.41 534.114,1365.41 538.813,1365.41 543.513,1365.41 548.212,1365.41 552.911,1365.41 
  557.611,1365.41 562.31,1365.41 567.009,1365.41 571.709,1365.41 576.408,1365.41 581.107,1365.41 585.807,1365.41 590.506,1365.41 595.205,1365.41 599.905,1365.41 
  604.604,1365.41 609.303,1365.41 614.003,1365.41 618.702,1365.41 623.401,1365.41 628.101,1365.41 632.8,1365.41 637.499,1365.41 642.199,1365.41 646.898,1365.41 
  651.597,1365.41 656.297,1365.41 660.996,1365.41 665.695,1365.41 670.395,1365.41 675.094,1365.41 679.793,1365.41 684.493,1365.41 689.192,1365.41 693.891,1365.41 
  698.591,1365.41 703.29,1365.41 707.989,1365.41 712.689,1365.41 717.388,1365.41 722.087,1365.41 726.787,1365.41 731.486,1365.41 736.185,1365.41 740.885,1365.41 
  745.584,1365.41 750.283,1365.41 754.983,1365.41 759.682,1365.41 764.381,1365.41 769.081,1365.41 773.78,1365.41 778.479,1365.41 783.179,1365.41 787.878,1365.41 
  792.577,1365.41 797.277,1365.41 801.976,1365.41 806.675,1365.41 811.375,1365.41 816.074,1365.41 820.773,1365.41 825.473,1365.41 830.172,1365.41 834.871,1365.41 
  839.571,1365.41 844.27,1365.41 848.969,1365.41 853.669,1365.41 858.368,1365.41 863.067,1365.41 867.767,1365.41 872.466,1365.41 877.165,1365.41 881.865,1365.41 
  886.564,1365.41 891.263,1365.41 895.963,1365.41 900.662,1365.41 905.361,1365.41 910.061,1365.41 914.76,1365.41 919.459,1365.41 924.159,1365.41 928.858,1365.41 
  933.557,1365.41 938.257,1365.41 942.956,1365.41 947.655,1365.41 952.355,1365.41 957.054,1365.41 961.753,1365.41 966.453,1365.41 971.152,1365.41 975.851,1365.41 
  980.551,1365.41 985.25,1365.41 989.949,1365.41 994.649,1365.41 999.348,1365.41 1004.05,1365.41 1008.75,1365.41 1013.45,1365.41 1018.15,1365.41 1022.84,1365.41 
  1027.54,1365.41 1032.24,1365.41 1036.94,1365.41 1041.64,1365.41 1046.34,1365.41 1051.04,1365.41 1055.74,1365.41 1060.44,1365.41 1065.14,1365.41 1069.84,1365.41 
  1074.54,1365.41 1079.24,1365.41 1083.94,1365.41 1088.64,1365.41 1093.33,1365.41 1098.03,1365.41 1102.73,1365.41 1107.43,1365.41 1112.13,1365.41 1116.83,1365.41 
  1121.53,1365.41 1126.23,1365.41 1130.93,1365.41 1135.63,1365.41 1140.33,1365.41 1145.03,1365.41 1149.73,1365.41 1154.43,1365.41 1159.13,1365.41 1163.82,1365.41 
  1168.52,1365.41 1173.22,1365.41 1177.92,1365.41 1182.62,1365.41 1187.32,1365.41 1192.02,1365.41 1196.72,1365.41 1201.42,1365.41 1206.12,1365.41 1210.82,1365.41 
  1215.52,1365.41 1220.22,1365.41 1224.92,1365.41 1229.62,1365.41 1234.31,1365.41 1239.01,1365.41 1243.71,1365.41 1248.41,1365.41 1253.11,1365.41 1257.81,1365.41 
  1262.51,1365.41 1267.21,1365.41 1271.91,1365.41 1276.61,1365.41 1281.31,1365.41 1286.01,1365.41 1290.71,1365.41 1295.41,1365.41 1300.11,1365.41 1304.8,1365.41 
  1309.5,1365.41 1314.2,1365.41 1318.9,1365.41 1323.6,1365.41 1328.3,1365.41 1333,1365.41 1337.7,1365.41 1342.4,1365.41 1347.1,1365.41 1351.8,1365.41 
  1356.5,1365.41 1361.2,1365.41 1365.9,1365.41 1370.6,1365.41 1375.29,1365.41 1379.99,1365.41 1384.69,1365.41 1389.39,1365.41 1394.09,1365.41 1398.79,1365.41 
  1403.49,1365.41 1408.19,1365.41 1412.89,1365.41 1417.59,1365.41 1422.29,1365.41 1426.99,1365.41 1431.69,1365.41 1436.39,1365.41 1441.09,1365.41 1445.78,1365.41 
  1450.48,1365.41 1455.18,1365.41 1459.88,1365.41 1464.58,1365.41 1469.28,1365.41 1473.98,1365.41 1478.68,1365.41 1483.38,1365.41 1488.08,1365.41 1492.78,1365.41 
  1497.48,1365.41 1502.18,1365.41 1506.88,1365.41 1511.58,1365.41 1516.27,1365.41 1520.97,1365.41 1525.67,1365.41 1530.37,1365.41 1535.07,1365.41 1539.77,1365.41 
  1544.47,1365.41 1549.17,1365.41 1553.87,1365.41 1558.57,1365.41 1563.27,1365.41 1567.97,1365.41 1572.67,1365.41 1577.37,1365.41 1582.07,1365.41 1586.76,1365.41 
  1591.46,1365.41 1596.16,1365.41 1600.86,1365.41 1605.56,1365.41 1610.26,1365.41 1614.96,1365.41 1619.66,1365.41 1624.36,1365.41 1629.06,1365.41 1633.76,1365.41 
  1638.46,1365.41 1643.16,1365.41 1647.86,1365.41 1652.56,1365.41 1657.25,1365.41 1661.95,1365.41 1666.65,1365.41 1671.35,1365.41 1676.05,1365.41 1680.75,1365.41 
  1685.45,1365.41 1690.15,1365.41 1694.85,1365.41 1699.55,1365.41 1704.25,1365.41 1708.95,1365.41 1713.65,1365.41 1718.35,1365.41 1723.05,1365.41 1727.74,1365.41 
  1732.44,1365.41 1737.14,1365.41 1741.84,1365.41 1746.54,1365.41 1751.24,1365.41 1755.94,1365.41 1760.64,1365.41 1765.34,1365.41 1770.04,1365.41 1774.74,1365.41 
  1779.44,1365.41 1784.14,1365.41 1788.84,1365.41 1793.54,1365.41 1798.23,1365.41 1802.93,1365.41 1807.63,1365.41 1812.33,1365.41 1817.03,1365.41 1821.73,1365.41 
  1826.43,1365.41 1831.13,1365.41 1835.83,1365.41 1840.53,1365.41 1845.23,1365.41 1849.93,1365.41 1854.63,1365.41 1859.33,1365.41 1864.03,1365.41 1868.72,1365.41 
  1873.42,1365.41 1878.12,1365.41 1882.82,1365.41 1887.52,1365.41 1892.22,1365.41 1896.92,1365.41 1901.62,1365.41 1906.32,1365.41 1911.02,1365.41 1915.72,1365.41 
  1920.42,1365.41 1925.12,1365.41 1929.82,1365.41 1934.52,1365.41 1939.21,1365.41 1943.91,1365.41 1948.61,1365.41 1953.31,1365.41 1958.01,1365.41 1962.71,1365.41 
  1967.41,1365.41 1972.11,1365.41 1976.81,1365.41 1981.51,1365.41 1986.21,1365.41 1990.91,1365.41 1995.61,1365.41 2000.31,1365.41 2005.01,1365.41 2009.7,1365.41 
  2014.4,1365.41 2019.1,1365.41 2023.8,1365.41 2028.5,1365.41 2033.2,1365.41 2037.9,1365.41 2042.6,1365.41 2047.3,1365.41 2052,1365.41 2056.7,1365.41 
  2061.4,1365.41 2066.1,1365.41 2070.8,1365.41 2075.5,1365.41 2080.19,1365.41 2084.89,1365.41 2089.59,1365.41 2094.29,1365.41 2098.99,1365.41 2103.69,1365.41 
  2108.39,1365.41 2113.09,1365.41 2117.79,1365.41 2122.49,1365.41 2127.19,1365.41 2131.89,1365.41 2136.59,1365.41 2141.29,1365.41 2145.99,1365.41 2150.68,1365.41 
  2155.38,1365.41 2160.08,1365.41 2164.78,1365.41 2169.48,1365.41 2174.18,1365.41 2178.88,1365.41 2183.58,1365.41 2188.28,1365.41 2192.98,1365.41 2197.68,1365.41 
  2202.38,1365.41 2207.08,1365.41 2211.78,1365.41 2216.48,1365.41 2221.17,1365.41 2225.87,1365.41 2230.57,1365.41 2235.27,1365.41 2239.97,1365.41 2244.67,1365.41 
  2249.37,1365.41 2254.07,1365.41 2258.77,1365.41 2263.47,1365.41 2268.17,1365.41 2272.87,1365.41 2277.57,1365.41 2282.27,1365.41 2286.97,1365.41 2291.66,1365.41 
  2296.36,1365.41 
  "/>
<path clip-path="url(#clip840)" d="
M1430.09 477.116 L2286.34 477.116 L2286.34 166.076 L1430.09 166.076  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<polyline clip-path="url(#clip840)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1430.09,477.116 2286.34,477.116 2286.34,166.076 1430.09,166.076 1430.09,477.116 
  "/>
<polyline clip-path="url(#clip840)" style="stroke:#009af9; stroke-linecap:butt; stroke-linejoin:round; stroke-width:6; stroke-opacity:1; fill:none" points="
  1452.23,243.836 1585.06,243.836 
  "/>
<path clip-path="url(#clip840)" d="M1634.84 232.013 L1634.84 238.054 Q1632.13 236.666 1629.22 235.971 Q1626.3 235.277 1623.17 235.277 Q1618.42 235.277 1616.02 236.735 Q1613.66 238.193 1613.66 241.11 Q1613.66 243.332 1615.36 244.617 Q1617.06 245.867 1622.2 247.013 L1624.39 247.499 Q1631.2 248.957 1634.04 251.631 Q1636.92 254.27 1636.92 259.027 Q1636.92 264.443 1632.62 267.603 Q1628.35 270.763 1620.85 270.763 Q1617.72 270.763 1614.32 270.138 Q1610.95 269.547 1607.2 268.332 L1607.2 261.735 Q1610.74 263.575 1614.18 264.513 Q1617.62 265.415 1620.99 265.415 Q1625.5 265.415 1627.93 263.888 Q1630.36 262.325 1630.36 259.513 Q1630.36 256.909 1628.59 255.52 Q1626.86 254.131 1620.92 252.846 L1618.7 252.325 Q1612.76 251.075 1610.12 248.506 Q1607.48 245.902 1607.48 241.388 Q1607.48 235.902 1611.37 232.916 Q1615.26 229.93 1622.41 229.93 Q1625.95 229.93 1629.08 230.45 Q1632.2 230.971 1634.84 232.013 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M1647.1 230.867 L1653.49 230.867 L1653.49 269.756 L1647.1 269.756 L1647.1 230.867 M1647.1 215.728 L1653.49 215.728 L1653.49 223.818 L1647.1 223.818 L1647.1 215.728 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M1697.13 238.332 Q1699.53 234.027 1702.86 231.978 Q1706.2 229.93 1710.71 229.93 Q1716.79 229.93 1720.08 234.2 Q1723.38 238.436 1723.38 246.284 L1723.38 269.756 L1716.96 269.756 L1716.96 246.492 Q1716.96 240.902 1714.98 238.193 Q1713 235.485 1708.94 235.485 Q1703.97 235.485 1701.09 238.784 Q1698.21 242.082 1698.21 247.777 L1698.21 269.756 L1691.79 269.756 L1691.79 246.492 Q1691.79 240.867 1689.81 238.193 Q1687.83 235.485 1683.7 235.485 Q1678.8 235.485 1675.92 238.818 Q1673.04 242.117 1673.04 247.777 L1673.04 269.756 L1666.61 269.756 L1666.61 230.867 L1673.04 230.867 L1673.04 236.909 Q1675.22 233.332 1678.28 231.631 Q1681.33 229.93 1685.54 229.93 Q1689.77 229.93 1692.72 232.082 Q1695.71 234.235 1697.13 238.332 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M1742.31 263.922 L1742.31 284.547 L1735.88 284.547 L1735.88 230.867 L1742.31 230.867 L1742.31 236.77 Q1744.32 233.298 1747.38 231.631 Q1750.47 229.93 1754.74 229.93 Q1761.82 229.93 1766.23 235.555 Q1770.67 241.179 1770.67 250.346 Q1770.67 259.513 1766.23 265.138 Q1761.82 270.763 1754.74 270.763 Q1750.47 270.763 1747.38 269.096 Q1744.32 267.395 1742.31 263.922 M1764.04 250.346 Q1764.04 243.298 1761.13 239.304 Q1758.24 235.277 1753.17 235.277 Q1748.1 235.277 1745.19 239.304 Q1742.31 243.298 1742.31 250.346 Q1742.31 257.395 1745.19 261.422 Q1748.1 265.415 1753.17 265.415 Q1758.24 265.415 1761.13 261.422 Q1764.04 257.395 1764.04 250.346 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M1781.26 215.728 L1787.65 215.728 L1787.65 269.756 L1781.26 269.756 L1781.26 215.728 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M1834.28 248.714 L1834.28 251.839 L1804.91 251.839 Q1805.33 258.436 1808.87 261.909 Q1812.44 265.346 1818.8 265.346 Q1822.48 265.346 1825.92 264.443 Q1829.39 263.54 1832.79 261.735 L1832.79 267.777 Q1829.35 269.235 1825.74 269.999 Q1822.13 270.763 1818.42 270.763 Q1809.11 270.763 1803.66 265.346 Q1798.24 259.929 1798.24 250.693 Q1798.24 241.145 1803.38 235.555 Q1808.56 229.93 1817.31 229.93 Q1825.15 229.93 1829.7 234.999 Q1834.28 240.034 1834.28 248.714 M1827.9 246.839 Q1827.83 241.596 1824.94 238.471 Q1822.1 235.346 1817.38 235.346 Q1812.03 235.346 1808.8 238.367 Q1805.6 241.388 1805.12 246.874 L1827.9 246.839 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M1845.05 217.916 L1874.84 217.916 L1874.84 223.818 L1852.06 223.818 L1852.06 239.096 L1872.62 239.096 L1872.62 244.999 L1852.06 244.999 L1852.06 269.756 L1845.05 269.756 L1845.05 217.916 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M1885.95 217.916 L1896.4 217.916 L1909.63 253.193 L1922.93 217.916 L1933.38 217.916 L1933.38 269.756 L1926.54 269.756 L1926.54 224.235 L1913.17 259.79 L1906.12 259.79 L1892.76 224.235 L1892.76 269.756 L1885.95 269.756 L1885.95 217.916 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M1946.51 217.916 L1953.56 217.916 L1953.56 249.409 Q1953.56 257.742 1956.58 261.422 Q1959.6 265.068 1966.37 265.068 Q1973.1 265.068 1976.12 261.422 Q1979.15 257.742 1979.15 249.409 L1979.15 217.916 L1986.19 217.916 L1986.19 250.277 Q1986.19 260.415 1981.16 265.589 Q1976.16 270.763 1966.37 270.763 Q1956.54 270.763 1951.51 265.589 Q1946.51 260.415 1946.51 250.277 L1946.51 217.916 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip840)" style="stroke:#e26f46; stroke-linecap:butt; stroke-linejoin:round; stroke-width:6; stroke-opacity:1; fill:none" points="
  1452.23,321.596 1585.06,321.596 
  "/>
<path clip-path="url(#clip840)" d="M1629.98 314.599 Q1628.9 313.974 1627.62 313.696 Q1626.37 313.384 1624.84 313.384 Q1619.42 313.384 1616.51 316.926 Q1613.63 320.433 1613.63 327.03 L1613.63 347.516 L1607.2 347.516 L1607.2 308.627 L1613.63 308.627 L1613.63 314.669 Q1615.64 311.127 1618.87 309.426 Q1622.1 307.69 1626.72 307.69 Q1627.38 307.69 1628.17 307.794 Q1628.97 307.863 1629.95 308.037 L1629.98 314.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M1668.38 326.474 L1668.38 329.599 L1639.01 329.599 Q1639.42 336.196 1642.97 339.669 Q1646.54 343.106 1652.9 343.106 Q1656.58 343.106 1660.02 342.203 Q1663.49 341.3 1666.89 339.495 L1666.89 345.537 Q1663.45 346.995 1659.84 347.759 Q1656.23 348.523 1652.52 348.523 Q1643.21 348.523 1637.76 343.106 Q1632.34 337.689 1632.34 328.453 Q1632.34 318.905 1637.48 313.315 Q1642.65 307.69 1651.4 307.69 Q1659.25 307.69 1663.8 312.759 Q1668.38 317.794 1668.38 326.474 M1661.99 324.599 Q1661.92 319.356 1659.04 316.231 Q1656.2 313.106 1651.47 313.106 Q1646.13 313.106 1642.9 316.127 Q1639.7 319.148 1639.22 324.634 L1661.99 324.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M1698.56 293.488 L1698.56 298.801 L1692.45 298.801 Q1689.01 298.801 1687.65 300.19 Q1686.33 301.578 1686.33 305.19 L1686.33 308.627 L1696.86 308.627 L1696.86 313.592 L1686.33 313.592 L1686.33 347.516 L1679.91 347.516 L1679.91 313.592 L1673.8 313.592 L1673.8 308.627 L1679.91 308.627 L1679.91 305.919 Q1679.91 299.426 1682.93 296.474 Q1685.95 293.488 1692.51 293.488 L1698.56 293.488 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M1737.17 326.474 L1737.17 329.599 L1707.79 329.599 Q1708.21 336.196 1711.75 339.669 Q1715.33 343.106 1721.68 343.106 Q1725.36 343.106 1728.8 342.203 Q1732.27 341.3 1735.67 339.495 L1735.67 345.537 Q1732.24 346.995 1728.63 347.759 Q1725.01 348.523 1721.3 348.523 Q1711.99 348.523 1706.54 343.106 Q1701.13 337.689 1701.13 328.453 Q1701.13 318.905 1706.26 313.315 Q1711.44 307.69 1720.19 307.69 Q1728.04 307.69 1732.58 312.759 Q1737.17 317.794 1737.17 326.474 M1730.78 324.599 Q1730.71 319.356 1727.83 316.231 Q1724.98 313.106 1720.26 313.106 Q1714.91 313.106 1711.68 316.127 Q1708.49 319.148 1708 324.634 L1730.78 324.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M1770.19 314.599 Q1769.11 313.974 1767.83 313.696 Q1766.58 313.384 1765.05 313.384 Q1759.63 313.384 1756.72 316.926 Q1753.83 320.433 1753.83 327.03 L1753.83 347.516 L1747.41 347.516 L1747.41 308.627 L1753.83 308.627 L1753.83 314.669 Q1755.85 311.127 1759.08 309.426 Q1762.31 307.69 1766.92 307.69 Q1767.58 307.69 1768.38 307.794 Q1769.18 307.863 1770.15 308.037 L1770.19 314.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M1808.59 326.474 L1808.59 329.599 L1779.22 329.599 Q1779.63 336.196 1783.17 339.669 Q1786.75 343.106 1793.1 343.106 Q1796.79 343.106 1800.22 342.203 Q1803.69 341.3 1807.1 339.495 L1807.1 345.537 Q1803.66 346.995 1800.05 347.759 Q1796.44 348.523 1792.72 348.523 Q1783.42 348.523 1777.97 343.106 Q1772.55 337.689 1772.55 328.453 Q1772.55 318.905 1777.69 313.315 Q1782.86 307.69 1791.61 307.69 Q1799.46 307.69 1804.01 312.759 Q1808.59 317.794 1808.59 326.474 M1802.2 324.599 Q1802.13 319.356 1799.25 316.231 Q1796.4 313.106 1791.68 313.106 Q1786.33 313.106 1783.1 316.127 Q1779.91 319.148 1779.42 324.634 L1802.2 324.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M1851.4 324.044 L1851.4 347.516 L1845.01 347.516 L1845.01 324.252 Q1845.01 318.731 1842.86 315.988 Q1840.71 313.245 1836.4 313.245 Q1831.23 313.245 1828.24 316.544 Q1825.26 319.842 1825.26 325.537 L1825.26 347.516 L1818.83 347.516 L1818.83 308.627 L1825.26 308.627 L1825.26 314.669 Q1827.55 311.162 1830.64 309.426 Q1833.76 307.69 1837.83 307.69 Q1844.53 307.69 1847.97 311.856 Q1851.4 315.988 1851.4 324.044 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M1892.13 310.12 L1892.13 316.092 Q1889.42 314.599 1886.68 313.87 Q1883.97 313.106 1881.19 313.106 Q1874.98 313.106 1871.54 317.064 Q1868.1 320.988 1868.1 328.106 Q1868.1 335.224 1871.54 339.182 Q1874.98 343.106 1881.19 343.106 Q1883.97 343.106 1886.68 342.377 Q1889.42 341.613 1892.13 340.12 L1892.13 346.023 Q1889.46 347.273 1886.58 347.898 Q1883.73 348.523 1880.5 348.523 Q1871.72 348.523 1866.54 343.002 Q1861.37 337.481 1861.37 328.106 Q1861.37 318.592 1866.58 313.141 Q1871.82 307.69 1880.92 307.69 Q1883.87 307.69 1886.68 308.315 Q1889.49 308.905 1892.13 310.12 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M1936.51 326.474 L1936.51 329.599 L1907.13 329.599 Q1907.55 336.196 1911.09 339.669 Q1914.67 343.106 1921.02 343.106 Q1924.7 343.106 1928.14 342.203 Q1931.61 341.3 1935.01 339.495 L1935.01 345.537 Q1931.58 346.995 1927.96 347.759 Q1924.35 348.523 1920.64 348.523 Q1911.33 348.523 1905.88 343.106 Q1900.47 337.689 1900.47 328.453 Q1900.47 318.905 1905.6 313.315 Q1910.78 307.69 1919.53 307.69 Q1927.37 307.69 1931.92 312.759 Q1936.51 317.794 1936.51 326.474 M1930.12 324.599 Q1930.05 319.356 1927.17 316.231 Q1924.32 313.106 1919.6 313.106 Q1914.25 313.106 1911.02 316.127 Q1907.83 319.148 1907.34 324.634 L1930.12 324.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip840)" style="stroke:#008000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:6; stroke-opacity:1; fill:none" stroke-dasharray="3, 6" points="
  1452.23,399.356 1585.06,399.356 
  "/>
<path clip-path="url(#clip840)" d="M1607.2 373.436 L1616.65 373.436 L1639.63 416.804 L1639.63 373.436 L1646.44 373.436 L1646.44 425.276 L1636.99 425.276 L1614.01 381.908 L1614.01 425.276 L1607.2 425.276 L1607.2 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M1693.38 404.234 L1693.38 407.359 L1664.01 407.359 Q1664.42 413.956 1667.97 417.429 Q1671.54 420.866 1677.9 420.866 Q1681.58 420.866 1685.01 419.963 Q1688.49 419.06 1691.89 417.255 L1691.89 423.297 Q1688.45 424.755 1684.84 425.519 Q1681.23 426.283 1677.51 426.283 Q1668.21 426.283 1662.76 420.866 Q1657.34 415.449 1657.34 406.213 Q1657.34 396.665 1662.48 391.075 Q1667.65 385.45 1676.4 385.45 Q1684.25 385.45 1688.8 390.519 Q1693.38 395.554 1693.38 404.234 M1686.99 402.359 Q1686.92 397.116 1684.04 393.991 Q1681.2 390.866 1676.47 390.866 Q1671.13 390.866 1667.9 393.887 Q1664.7 396.908 1664.22 402.394 L1686.99 402.359 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M1703.21 409.929 L1703.21 386.387 L1709.6 386.387 L1709.6 409.686 Q1709.6 415.206 1711.75 417.984 Q1713.9 420.727 1718.21 420.727 Q1723.38 420.727 1726.37 417.429 Q1729.39 414.13 1729.39 408.436 L1729.39 386.387 L1735.78 386.387 L1735.78 425.276 L1729.39 425.276 L1729.39 419.304 Q1727.06 422.845 1723.97 424.581 Q1720.92 426.283 1716.85 426.283 Q1710.15 426.283 1706.68 422.116 Q1703.21 417.949 1703.21 409.929 M1719.29 385.45 L1719.29 385.45 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M1771.47 392.359 Q1770.4 391.734 1769.11 391.456 Q1767.86 391.144 1766.33 391.144 Q1760.92 391.144 1758 394.686 Q1755.12 398.193 1755.12 404.79 L1755.12 425.276 L1748.7 425.276 L1748.7 386.387 L1755.12 386.387 L1755.12 392.429 Q1757.13 388.887 1760.36 387.186 Q1763.59 385.45 1768.21 385.45 Q1768.87 385.45 1769.67 385.554 Q1770.47 385.623 1771.44 385.797 L1771.47 392.359 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M1795.85 405.727 Q1788.1 405.727 1785.12 407.498 Q1782.13 409.269 1782.13 413.54 Q1782.13 416.942 1784.35 418.956 Q1786.61 420.935 1790.47 420.935 Q1795.78 420.935 1798.97 417.185 Q1802.2 413.401 1802.2 407.151 L1802.2 405.727 L1795.85 405.727 M1808.59 403.088 L1808.59 425.276 L1802.2 425.276 L1802.2 419.373 Q1800.01 422.915 1796.75 424.616 Q1793.49 426.283 1788.76 426.283 Q1782.79 426.283 1779.25 422.949 Q1775.74 419.581 1775.74 413.956 Q1775.74 407.394 1780.12 404.061 Q1784.53 400.727 1793.24 400.727 L1802.2 400.727 L1802.2 400.102 Q1802.2 395.693 1799.29 393.297 Q1796.4 390.866 1791.16 390.866 Q1787.83 390.866 1784.67 391.665 Q1781.51 392.463 1778.59 394.061 L1778.59 388.158 Q1782.1 386.804 1785.4 386.144 Q1788.69 385.45 1791.82 385.45 Q1800.26 385.45 1804.42 389.825 Q1808.59 394.199 1808.59 403.088 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M1821.75 371.248 L1828.14 371.248 L1828.14 425.276 L1821.75 425.276 L1821.75 371.248 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M1841.78 373.436 L1871.58 373.436 L1871.58 379.338 L1848.8 379.338 L1848.8 394.616 L1869.35 394.616 L1869.35 400.519 L1848.8 400.519 L1848.8 425.276 L1841.78 425.276 L1841.78 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M1882.69 373.436 L1893.14 373.436 L1906.37 408.713 L1919.67 373.436 L1930.12 373.436 L1930.12 425.276 L1923.28 425.276 L1923.28 379.755 L1909.91 415.31 L1902.86 415.31 L1889.49 379.755 L1889.49 425.276 L1882.69 425.276 L1882.69 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M1943.24 373.436 L1950.29 373.436 L1950.29 404.929 Q1950.29 413.262 1953.31 416.942 Q1956.33 420.588 1963.1 420.588 Q1969.84 420.588 1972.86 416.942 Q1975.88 413.262 1975.88 404.929 L1975.88 373.436 L1982.93 373.436 L1982.93 405.797 Q1982.93 415.935 1977.9 421.109 Q1972.9 426.283 1963.1 426.283 Q1953.28 426.283 1948.24 421.109 Q1943.24 415.935 1943.24 405.797 L1943.24 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M2033.76 371.318 Q2029.11 379.304 2026.85 387.116 Q2024.6 394.929 2024.6 402.949 Q2024.6 410.97 2026.85 418.852 Q2029.14 426.699 2033.76 434.651 L2028.21 434.651 Q2023 426.491 2020.4 418.609 Q2017.83 410.727 2017.83 402.949 Q2017.83 395.206 2020.4 387.359 Q2022.96 379.512 2028.21 371.318 L2033.76 371.318 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M2053.1 419.373 L2077.58 419.373 L2077.58 425.276 L2044.67 425.276 L2044.67 419.373 Q2048.66 415.241 2055.53 408.297 Q2062.44 401.318 2064.21 399.304 Q2067.58 395.519 2068.9 392.915 Q2070.26 390.276 2070.26 387.741 Q2070.26 383.609 2067.34 381.005 Q2064.46 378.401 2059.8 378.401 Q2056.51 378.401 2052.83 379.547 Q2049.18 380.693 2045.01 383.019 L2045.01 375.936 Q2049.25 374.234 2052.93 373.366 Q2056.61 372.498 2059.67 372.498 Q2067.72 372.498 2072.51 376.526 Q2077.3 380.554 2077.3 387.29 Q2077.3 390.484 2076.09 393.366 Q2074.91 396.213 2071.75 400.102 Q2070.88 401.109 2066.23 405.936 Q2061.58 410.727 2053.1 419.373 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M2092.37 373.436 L2119.91 373.436 L2119.91 379.338 L2098.8 379.338 L2098.8 392.047 Q2100.33 391.526 2101.85 391.283 Q2103.38 391.005 2104.91 391.005 Q2113.59 391.005 2118.66 395.762 Q2123.73 400.519 2123.73 408.644 Q2123.73 417.012 2118.52 421.665 Q2113.31 426.283 2103.83 426.283 Q2100.57 426.283 2097.17 425.727 Q2093.8 425.172 2090.19 424.06 L2090.19 417.012 Q2093.31 418.713 2096.64 419.547 Q2099.98 420.38 2103.69 420.38 Q2109.7 420.38 2113.21 417.22 Q2116.71 414.06 2116.71 408.644 Q2116.71 403.227 2113.21 400.068 Q2109.7 396.908 2103.69 396.908 Q2100.88 396.908 2098.07 397.533 Q2095.29 398.158 2092.37 399.477 L2092.37 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M2152.55 378.054 Q2147.13 378.054 2144.39 383.401 Q2141.68 388.713 2141.68 399.408 Q2141.68 410.067 2144.39 415.415 Q2147.13 420.727 2152.55 420.727 Q2158 420.727 2160.71 415.415 Q2163.45 410.067 2163.45 399.408 Q2163.45 388.713 2160.71 383.401 Q2158 378.054 2152.55 378.054 M2152.55 372.498 Q2161.26 372.498 2165.85 379.408 Q2170.46 386.283 2170.46 399.408 Q2170.46 412.498 2165.85 419.408 Q2161.26 426.283 2152.55 426.283 Q2143.83 426.283 2139.21 419.408 Q2134.63 412.498 2134.63 399.408 Q2134.63 386.283 2139.21 379.408 Q2143.83 372.498 2152.55 372.498 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M2197.79 378.054 Q2192.37 378.054 2189.63 383.401 Q2186.92 388.713 2186.92 399.408 Q2186.92 410.067 2189.63 415.415 Q2192.37 420.727 2197.79 420.727 Q2203.24 420.727 2205.95 415.415 Q2208.69 410.067 2208.69 399.408 Q2208.69 388.713 2205.95 383.401 Q2203.24 378.054 2197.79 378.054 M2197.79 372.498 Q2206.51 372.498 2211.09 379.408 Q2215.71 386.283 2215.71 399.408 Q2215.71 412.498 2211.09 419.408 Q2206.51 426.283 2197.79 426.283 Q2189.07 426.283 2184.46 419.408 Q2179.87 412.498 2179.87 399.408 Q2179.87 386.283 2184.46 379.408 Q2189.07 372.498 2197.79 372.498 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip840)" d="M2226.12 371.318 L2231.68 371.318 Q2236.89 379.512 2239.46 387.359 Q2242.06 395.206 2242.06 402.949 Q2242.06 410.727 2239.46 418.609 Q2236.89 426.491 2231.68 434.651 L2226.12 434.651 Q2230.74 426.699 2233 418.852 Q2235.29 410.97 2235.29 402.949 Q2235.29 394.929 2233 387.116 Q2230.74 379.304 2226.12 371.318 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /></svg>



    ┌ Info: Run: 2/2  Epoch: 1/5
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:3
    ┌ Info:   Loss [2551]: 0.0016
    │         Avg displacement in data: 0.04001
    │         Weight/Scale: 1.013728233988419   Bias/Offset: 0.05566377285400733
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [2601]: 0.00156
    │         Avg displacement in data: 0.03951
    │         Weight/Scale: 1.0136341011255399   Bias/Offset: 0.0556502098051191
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [2651]: 0.00153
    │         Avg displacement in data: 0.03907
    │         Weight/Scale: 1.013560875012561   Bias/Offset: 0.05565333590293733
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [2701]: 0.00149
    │         Avg displacement in data: 0.03863
    │         Weight/Scale: 1.0134601199953166   Bias/Offset: 0.0556657403913297
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [2751]: 0.00146
    │         Avg displacement in data: 0.0382
    │         Weight/Scale: 1.013396250856142   Bias/Offset: 0.055683368188246374
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [2801]: 0.00143
    │         Avg displacement in data: 0.03781
    │         Weight/Scale: 1.0133404862975057   Bias/Offset: 0.055705280650676946
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [2851]: 0.0014
    │         Avg displacement in data: 0.03744
    │         Weight/Scale: 1.0132958994198862   Bias/Offset: 0.055735627409539475
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [2901]: 0.00138
    │         Avg displacement in data: 0.03709
    │         Weight/Scale: 1.013218012967237   Bias/Offset: 0.055775173190348346
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [2951]: 0.00135
    │         Avg displacement in data: 0.03673
    │         Weight/Scale: 1.0131793143007137   Bias/Offset: 0.05581649513763439
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [3001]: 0.00133
    │         Avg displacement in data: 0.0364
    │         Weight/Scale: 1.0131443466176002   Bias/Offset: 0.055857013802385426
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info: Run: 2/2  Epoch: 2/5
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:3
    ┌ Info:   Loss [3051]: 0.0013
    │         Avg displacement in data: 0.03609
    │         Weight/Scale: 1.0131154167594163   Bias/Offset: 0.05590079068969623
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [3101]: 0.0013
    │         Avg displacement in data: 0.03608
    │         Weight/Scale: 1.0130868421031263   Bias/Offset: 0.05597175462264453
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [3151]: 0.00126
    │         Avg displacement in data: 0.0355
    │         Weight/Scale: 1.0130242869107013   Bias/Offset: 0.05600750255719061
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [3201]: 0.00124
    │         Avg displacement in data: 0.03522
    │         Weight/Scale: 1.0129981472180718   Bias/Offset: 0.05605249279101586
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [3251]: 0.00122
    │         Avg displacement in data: 0.03495
    │         Weight/Scale: 1.0129752912931875   Bias/Offset: 0.05609678362259138
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [3301]: 0.0012
    │         Avg displacement in data: 0.03468
    │         Weight/Scale: 1.0129541238201665   Bias/Offset: 0.05613907071683375
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [3351]: 0.00121
    │         Avg displacement in data: 0.03476
    │         Weight/Scale: 1.0129233408651779   Bias/Offset: 0.05618713624086441
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [3401]: 0.00117
    │         Avg displacement in data: 0.03418
    │         Weight/Scale: 1.0128507101841244   Bias/Offset: 0.05622701271598478
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [3451]: 0.00115
    │         Avg displacement in data: 0.03394
    │         Weight/Scale: 1.0128230720455083   Bias/Offset: 0.056256029991234954
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [3501]: 0.00114
    │         Avg displacement in data: 0.03371
    │         Weight/Scale: 1.0127914715565853   Bias/Offset: 0.05627554973268133
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info: Run: 2/2  Epoch: 3/5
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:3
    ┌ Info:   Loss [3551]: 0.00112
    │         Avg displacement in data: 0.03348
    │         Weight/Scale: 1.012757278142104   Bias/Offset: 0.05628694227936519
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [3601]: 0.00111
    │         Avg displacement in data: 0.03325
    │         Weight/Scale: 1.0127201517557438   Bias/Offset: 0.05628937874777039
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [3651]: 0.0011
    │         Avg displacement in data: 0.03309
    │         Weight/Scale: 1.0126345235803296   Bias/Offset: 0.05629084856116462
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [3701]: 0.00108
    │         Avg displacement in data: 0.03281
    │         Weight/Scale: 1.0125650080866402   Bias/Offset: 0.05627297159316166
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [3751]: 0.00106
    │         Avg displacement in data: 0.03259
    │         Weight/Scale: 1.0125069364922694   Bias/Offset: 0.05623790664671173
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [3801]: 0.00105
    │         Avg displacement in data: 0.03238
    │         Weight/Scale: 1.012440018256317   Bias/Offset: 0.05618535407915965
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [3851]: 0.00103
    │         Avg displacement in data: 0.03216
    │         Weight/Scale: 1.0123651258104895   Bias/Offset: 0.056116618194637584
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [3901]: 0.00102
    │         Avg displacement in data: 0.03195
    │         Weight/Scale: 1.0122845147841195   Bias/Offset: 0.056033033676680194
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [3951]: 0.00101
    │         Avg displacement in data: 0.03178
    │         Weight/Scale: 1.012132035508753   Bias/Offset: 0.05593973782043158
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [4001]: 0.00099
    │         Avg displacement in data: 0.03152
    │         Weight/Scale: 1.0120157339136429   Bias/Offset: 0.05581333934046344
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info: Run: 2/2  Epoch: 4/5
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:3
    ┌ Info:   Loss [4051]: 0.00098
    │         Avg displacement in data: 0.03131
    │         Weight/Scale: 1.0118936733964896   Bias/Offset: 0.055662491134584845
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [4101]: 0.00097
    │         Avg displacement in data: 0.0311
    │         Weight/Scale: 1.0117599071600911   Bias/Offset: 0.055490188477823596
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [4151]: 0.00095
    │         Avg displacement in data: 0.03088
    │         Weight/Scale: 1.0116129294111869   Bias/Offset: 0.05529529414996092
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [4201]: 0.00095
    │         Avg displacement in data: 0.03076
    │         Weight/Scale: 1.011425585464772   Bias/Offset: 0.05508364941793416
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [4251]: 0.00093
    │         Avg displacement in data: 0.03045
    │         Weight/Scale: 1.0111962852669019   Bias/Offset: 0.054848371180265906
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [4301]: 0.00091
    │         Avg displacement in data: 0.03024
    │         Weight/Scale: 1.0110049921877355   Bias/Offset: 0.05458325712342302
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [4351]: 0.0009
    │         Avg displacement in data: 0.03002
    │         Weight/Scale: 1.0107965909610808   Bias/Offset: 0.05429070902053792
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [4401]: 0.00089
    │         Avg displacement in data: 0.0298
    │         Weight/Scale: 1.0105727134736049   Bias/Offset: 0.05397476469676082
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [4451]: 0.00087
    │         Avg displacement in data: 0.02958
    │         Weight/Scale: 1.0103331895968897   Bias/Offset: 0.053635938599082576
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [4501]: 0.00089
    │         Avg displacement in data: 0.02988
    │         Weight/Scale: 1.0100765189166843   Bias/Offset: 0.05331948805541876
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info: Run: 2/2  Epoch: 5/5
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:3
    ┌ Info:   Loss [4551]: 0.00085
    │         Avg displacement in data: 0.02914
    │         Weight/Scale: 1.0097234988397652   Bias/Offset: 0.0529088909351857
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [4601]: 0.00084
    │         Avg displacement in data: 0.02891
    │         Weight/Scale: 1.0094459499374637   Bias/Offset: 0.05251138969091385
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [4651]: 0.00082
    │         Avg displacement in data: 0.02868
    │         Weight/Scale: 1.0091491921370825   Bias/Offset: 0.05208612493384556
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [4701]: 0.00081
    │         Avg displacement in data: 0.02844
    │         Weight/Scale: 1.0088374186481304   Bias/Offset: 0.05163781571190417
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [4751]: 0.00079
    │         Avg displacement in data: 0.0282
    │         Weight/Scale: 1.0085110697416406   Bias/Offset: 0.051165640824350545
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [4801]: 0.00082
    │         Avg displacement in data: 0.02872
    │         Weight/Scale: 1.0080617904933833   Bias/Offset: 0.05062984229140617
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [4851]: 0.00077
    │         Avg displacement in data: 0.02768
    │         Weight/Scale: 1.0077217372486384   Bias/Offset: 0.0501953088775428
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [4901]: 0.00075
    │         Avg displacement in data: 0.0274
    │         Weight/Scale: 1.0073718416946547   Bias/Offset: 0.049662341595064316
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [4951]: 0.00073
    │         Avg displacement in data: 0.0271
    │         Weight/Scale: 1.007008606239314   Bias/Offset: 0.04909309895026161
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15
    ┌ Info:   Loss [5001]: 0.00072
    │         Avg displacement in data: 0.02679
    │         Weight/Scale: 1.0066330227835532   Bias/Offset: 0.04848636968197639
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:15


    1973.045912 seconds (8.34 G allocations: 512.036 GiB, 3.99% gc time)



<?xml version="1.0" encoding="utf-8"?>
<svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" width="600" height="400" viewBox="0 0 2400 1600">
<defs>
  <clipPath id="clip930">
    <rect x="0" y="0" width="2400" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip930)" d="
M0 1600 L2400 1600 L2400 0 L0 0  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip931">
    <rect x="480" y="0" width="1681" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip930)" d="
M303.503 1401.58 L2352.76 1401.58 L2352.76 123.472 L303.503 123.472  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip932">
    <rect x="303" y="123" width="2050" height="1279"/>
  </clipPath>
</defs>
<polyline clip-path="url(#clip932)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  361.501,1401.58 361.501,123.472 
  "/>
<polyline clip-path="url(#clip932)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  844.815,1401.58 844.815,123.472 
  "/>
<polyline clip-path="url(#clip932)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1328.13,1401.58 1328.13,123.472 
  "/>
<polyline clip-path="url(#clip932)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1811.44,1401.58 1811.44,123.472 
  "/>
<polyline clip-path="url(#clip932)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  2294.76,1401.58 2294.76,123.472 
  "/>
<polyline clip-path="url(#clip930)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  303.503,1401.58 2352.76,1401.58 
  "/>
<polyline clip-path="url(#clip930)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  361.501,1401.58 361.501,1382.68 
  "/>
<polyline clip-path="url(#clip930)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  844.815,1401.58 844.815,1382.68 
  "/>
<polyline clip-path="url(#clip930)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1328.13,1401.58 1328.13,1382.68 
  "/>
<polyline clip-path="url(#clip930)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1811.44,1401.58 1811.44,1382.68 
  "/>
<polyline clip-path="url(#clip930)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  2294.76,1401.58 2294.76,1382.68 
  "/>
<path clip-path="url(#clip930)" d="M361.501 1438.36 Q356.084 1438.36 353.341 1443.71 Q350.633 1449.02 350.633 1459.71 Q350.633 1470.37 353.341 1475.72 Q356.084 1481.03 361.501 1481.03 Q366.952 1481.03 369.66 1475.72 Q372.403 1470.37 372.403 1459.71 Q372.403 1449.02 369.66 1443.71 Q366.952 1438.36 361.501 1438.36 M361.501 1432.8 Q370.216 1432.8 374.799 1439.71 Q379.417 1446.59 379.417 1459.71 Q379.417 1472.8 374.799 1479.71 Q370.216 1486.59 361.501 1486.59 Q352.785 1486.59 348.167 1479.71 Q343.584 1472.8 343.584 1459.71 Q343.584 1446.59 348.167 1439.71 Q352.785 1432.8 361.501 1432.8 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M830.388 1479.68 L841.846 1479.68 L841.846 1440.13 L829.381 1442.63 L829.381 1436.24 L841.777 1433.74 L848.791 1433.74 L848.791 1479.68 L860.249 1479.68 L860.249 1485.58 L830.388 1485.58 L830.388 1479.68 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1320.11 1479.68 L1344.59 1479.68 L1344.59 1485.58 L1311.67 1485.58 L1311.67 1479.68 Q1315.66 1475.55 1322.54 1468.6 Q1329.45 1461.62 1331.22 1459.61 Q1334.59 1455.82 1335.91 1453.22 Q1337.26 1450.58 1337.26 1448.05 Q1337.26 1443.91 1334.34 1441.31 Q1331.46 1438.71 1326.81 1438.71 Q1323.51 1438.71 1319.83 1439.85 Q1316.19 1441 1312.02 1443.32 L1312.02 1436.24 Q1316.25 1434.54 1319.94 1433.67 Q1323.62 1432.8 1326.67 1432.8 Q1334.73 1432.8 1339.52 1436.83 Q1344.31 1440.86 1344.31 1447.59 Q1344.31 1450.79 1343.09 1453.67 Q1341.91 1456.52 1338.75 1460.41 Q1337.89 1461.41 1333.23 1466.24 Q1328.58 1471.03 1320.11 1479.68 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1817.82 1457.63 Q1822.85 1458.7 1825.66 1462.11 Q1828.51 1465.51 1828.51 1470.51 Q1828.51 1478.18 1823.23 1482.39 Q1817.95 1486.59 1808.23 1486.59 Q1804.97 1486.59 1801.5 1485.93 Q1798.06 1485.3 1794.38 1484.02 L1794.38 1477.25 Q1797.29 1478.95 1800.77 1479.82 Q1804.24 1480.68 1808.02 1480.68 Q1814.62 1480.68 1818.06 1478.08 Q1821.53 1475.48 1821.53 1470.51 Q1821.53 1465.93 1818.3 1463.36 Q1815.11 1460.75 1809.38 1460.75 L1803.34 1460.75 L1803.34 1454.99 L1809.66 1454.99 Q1814.83 1454.99 1817.57 1452.94 Q1820.32 1450.86 1820.32 1446.97 Q1820.32 1442.98 1817.47 1440.86 Q1814.66 1438.71 1809.38 1438.71 Q1806.5 1438.71 1803.2 1439.33 Q1799.9 1439.96 1795.94 1441.27 L1795.94 1435.02 Q1799.93 1433.91 1803.41 1433.36 Q1806.91 1432.8 1810 1432.8 Q1817.99 1432.8 1822.64 1436.45 Q1827.29 1440.06 1827.29 1446.24 Q1827.29 1450.55 1824.83 1453.53 Q1822.36 1456.48 1817.82 1457.63 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M2299.27 1439.85 L2281.56 1467.52 L2299.27 1467.52 L2299.27 1439.85 M2297.43 1433.74 L2306.25 1433.74 L2306.25 1467.52 L2313.65 1467.52 L2313.65 1473.36 L2306.25 1473.36 L2306.25 1485.58 L2299.27 1485.58 L2299.27 1473.36 L2275.87 1473.36 L2275.87 1466.59 L2297.43 1433.74 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1271.74 1527.62 L1271.74 1538.66 L1284.9 1538.66 L1284.9 1543.62 L1271.74 1543.62 L1271.74 1564.74 Q1271.74 1569.49 1273.03 1570.85 Q1274.35 1572.2 1278.34 1572.2 L1284.9 1572.2 L1284.9 1577.55 L1278.34 1577.55 Q1270.94 1577.55 1268.13 1574.8 Q1265.32 1572.03 1265.32 1564.74 L1265.32 1543.62 L1260.63 1543.62 L1260.63 1538.66 L1265.32 1538.66 L1265.32 1527.62 L1271.74 1527.62 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1315.32 1523.52 L1330.04 1523.52 L1330.04 1528.49 L1321.71 1528.49 L1321.71 1581.96 L1330.04 1581.96 L1330.04 1586.92 L1315.32 1586.92 L1315.32 1523.52 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1368.44 1539.8 L1368.44 1545.85 Q1365.73 1544.46 1362.82 1543.76 Q1359.9 1543.07 1356.78 1543.07 Q1352.02 1543.07 1349.62 1544.53 Q1347.26 1545.99 1347.26 1548.9 Q1347.26 1551.12 1348.96 1552.41 Q1350.66 1553.66 1355.8 1554.8 L1357.99 1555.29 Q1364.8 1556.75 1367.64 1559.42 Q1370.53 1562.06 1370.53 1566.82 Q1370.53 1572.24 1366.22 1575.4 Q1361.95 1578.55 1354.45 1578.55 Q1351.32 1578.55 1347.92 1577.93 Q1344.55 1577.34 1340.8 1576.12 L1340.8 1569.53 Q1344.34 1571.37 1347.78 1572.3 Q1351.22 1573.21 1354.59 1573.21 Q1359.1 1573.21 1361.53 1571.68 Q1363.96 1570.12 1363.96 1567.3 Q1363.96 1564.7 1362.19 1563.31 Q1360.46 1561.92 1354.52 1560.64 L1352.3 1560.12 Q1346.36 1558.87 1343.72 1556.3 Q1341.08 1553.69 1341.08 1549.18 Q1341.08 1543.69 1344.97 1540.71 Q1348.86 1537.72 1356.01 1537.72 Q1359.55 1537.72 1362.68 1538.24 Q1365.8 1538.76 1368.44 1539.8 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1395.63 1523.52 L1395.63 1586.92 L1380.91 1586.92 L1380.91 1581.96 L1389.21 1581.96 L1389.21 1528.49 L1380.91 1528.49 L1380.91 1523.52 L1395.63 1523.52 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip932)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  303.503,1365.41 2352.76,1365.41 
  "/>
<polyline clip-path="url(#clip932)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  303.503,1114.21 2352.76,1114.21 
  "/>
<polyline clip-path="url(#clip932)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  303.503,863.02 2352.76,863.02 
  "/>
<polyline clip-path="url(#clip932)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  303.503,611.826 2352.76,611.826 
  "/>
<polyline clip-path="url(#clip932)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  303.503,360.632 2352.76,360.632 
  "/>
<polyline clip-path="url(#clip930)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  303.503,1401.58 303.503,123.472 
  "/>
<polyline clip-path="url(#clip930)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  303.503,1365.41 322.401,1365.41 
  "/>
<polyline clip-path="url(#clip930)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  303.503,1114.21 322.401,1114.21 
  "/>
<polyline clip-path="url(#clip930)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  303.503,863.02 322.401,863.02 
  "/>
<polyline clip-path="url(#clip930)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  303.503,611.826 322.401,611.826 
  "/>
<polyline clip-path="url(#clip930)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  303.503,360.632 322.401,360.632 
  "/>
<path clip-path="url(#clip930)" d="M136.497 1344.11 Q131.08 1344.11 128.337 1349.45 Q125.629 1354.76 125.629 1365.46 Q125.629 1376.12 128.337 1381.47 Q131.08 1386.78 136.497 1386.78 Q141.948 1386.78 144.656 1381.47 Q147.399 1376.12 147.399 1365.46 Q147.399 1354.76 144.656 1349.45 Q141.948 1344.11 136.497 1344.11 M136.497 1338.55 Q145.212 1338.55 149.795 1345.46 Q154.413 1352.33 154.413 1365.46 Q154.413 1378.55 149.795 1385.46 Q145.212 1392.33 136.497 1392.33 Q127.781 1392.33 123.163 1385.46 Q118.58 1378.55 118.58 1365.46 Q118.58 1352.33 123.163 1345.46 Q127.781 1338.55 136.497 1338.55 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M166.74 1382.51 L174.066 1382.51 L174.066 1391.33 L166.74 1391.33 L166.74 1382.51 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M189.413 1339.49 L216.948 1339.49 L216.948 1345.39 L195.837 1345.39 L195.837 1358.1 Q197.364 1357.58 198.892 1357.33 Q200.42 1357.06 201.948 1357.06 Q210.628 1357.06 215.698 1361.81 Q220.767 1366.57 220.767 1374.7 Q220.767 1383.06 215.559 1387.72 Q210.35 1392.33 200.871 1392.33 Q197.608 1392.33 194.205 1391.78 Q190.837 1391.22 187.226 1390.11 L187.226 1383.06 Q190.351 1384.76 193.684 1385.6 Q197.017 1386.43 200.732 1386.43 Q206.739 1386.43 210.246 1383.27 Q213.753 1380.11 213.753 1374.7 Q213.753 1369.28 210.246 1366.12 Q206.739 1362.96 200.732 1362.96 Q197.92 1362.96 195.108 1363.58 Q192.33 1364.21 189.413 1365.53 L189.413 1339.49 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M249.586 1344.11 Q244.17 1344.11 241.427 1349.45 Q238.718 1354.76 238.718 1365.46 Q238.718 1376.12 241.427 1381.47 Q244.17 1386.78 249.586 1386.78 Q255.038 1386.78 257.746 1381.47 Q260.489 1376.12 260.489 1365.46 Q260.489 1354.76 257.746 1349.45 Q255.038 1344.11 249.586 1344.11 M249.586 1338.55 Q258.302 1338.55 262.885 1345.46 Q267.503 1352.33 267.503 1365.46 Q267.503 1378.55 262.885 1385.46 Q258.302 1392.33 249.586 1392.33 Q240.871 1392.33 236.253 1385.46 Q231.67 1378.55 231.67 1365.46 Q231.67 1352.33 236.253 1345.46 Q240.871 1338.55 249.586 1338.55 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M137.99 1092.91 Q132.573 1092.91 129.83 1098.26 Q127.122 1103.57 127.122 1114.27 Q127.122 1124.93 129.83 1130.27 Q132.573 1135.58 137.99 1135.58 Q143.441 1135.58 146.149 1130.27 Q148.892 1124.93 148.892 1114.27 Q148.892 1103.57 146.149 1098.26 Q143.441 1092.91 137.99 1092.91 M137.99 1087.36 Q146.705 1087.36 151.288 1094.27 Q155.906 1101.14 155.906 1114.27 Q155.906 1127.36 151.288 1134.27 Q146.705 1141.14 137.99 1141.14 Q129.275 1141.14 124.657 1134.27 Q120.073 1127.36 120.073 1114.27 Q120.073 1101.14 124.657 1094.27 Q129.275 1087.36 137.99 1087.36 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M168.233 1131.31 L175.559 1131.31 L175.559 1140.13 L168.233 1140.13 L168.233 1131.31 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M189.066 1088.29 L222.399 1088.29 L222.399 1091.28 L203.58 1140.13 L196.253 1140.13 L213.962 1094.2 L189.066 1094.2 L189.066 1088.29 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M236.149 1088.29 L263.684 1088.29 L263.684 1094.2 L242.573 1094.2 L242.573 1106.9 Q244.1 1106.38 245.628 1106.14 Q247.156 1105.86 248.684 1105.86 Q257.364 1105.86 262.434 1110.62 Q267.503 1115.38 267.503 1123.5 Q267.503 1131.87 262.295 1136.52 Q257.086 1141.14 247.607 1141.14 Q244.343 1141.14 240.941 1140.58 Q237.573 1140.03 233.961 1138.92 L233.961 1131.87 Q237.086 1133.57 240.42 1134.4 Q243.753 1135.24 247.468 1135.24 Q253.475 1135.24 256.982 1132.08 Q260.489 1128.92 260.489 1123.5 Q260.489 1118.08 256.982 1114.93 Q253.475 1111.77 247.468 1111.77 Q244.656 1111.77 241.843 1112.39 Q239.066 1113.02 236.149 1114.33 L236.149 1088.29 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M122.712 883.037 L134.17 883.037 L134.17 843.488 L121.705 845.988 L121.705 839.6 L134.101 837.1 L141.115 837.1 L141.115 883.037 L152.573 883.037 L152.573 888.94 L122.712 888.94 L122.712 883.037 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M166.74 880.12 L174.066 880.12 L174.066 888.94 L166.74 888.94 L166.74 880.12 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M204.344 841.718 Q198.927 841.718 196.184 847.065 Q193.476 852.377 193.476 863.072 Q193.476 873.731 196.184 879.079 Q198.927 884.391 204.344 884.391 Q209.795 884.391 212.503 879.079 Q215.246 873.731 215.246 863.072 Q215.246 852.377 212.503 847.065 Q209.795 841.718 204.344 841.718 M204.344 836.162 Q213.059 836.162 217.642 843.072 Q222.26 849.947 222.26 863.072 Q222.26 876.162 217.642 883.072 Q213.059 889.947 204.344 889.947 Q195.628 889.947 191.01 883.072 Q186.427 876.162 186.427 863.072 Q186.427 849.947 191.01 843.072 Q195.628 836.162 204.344 836.162 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M249.586 841.718 Q244.17 841.718 241.427 847.065 Q238.718 852.377 238.718 863.072 Q238.718 873.731 241.427 879.079 Q244.17 884.391 249.586 884.391 Q255.038 884.391 257.746 879.079 Q260.489 873.731 260.489 863.072 Q260.489 852.377 257.746 847.065 Q255.038 841.718 249.586 841.718 M249.586 836.162 Q258.302 836.162 262.885 843.072 Q267.503 849.947 267.503 863.072 Q267.503 876.162 262.885 883.072 Q258.302 889.947 249.586 889.947 Q240.871 889.947 236.253 883.072 Q231.67 876.162 231.67 863.072 Q231.67 849.947 236.253 843.072 Q240.871 836.162 249.586 836.162 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M124.205 631.843 L135.663 631.843 L135.663 592.295 L123.198 594.795 L123.198 588.406 L135.594 585.906 L142.608 585.906 L142.608 631.843 L154.066 631.843 L154.066 637.746 L124.205 637.746 L124.205 631.843 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M168.233 628.927 L175.559 628.927 L175.559 637.746 L168.233 637.746 L168.233 628.927 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M196.878 631.843 L221.357 631.843 L221.357 637.746 L188.441 637.746 L188.441 631.843 Q192.434 627.711 199.309 620.767 Q206.219 613.788 207.989 611.774 Q211.357 607.989 212.677 605.385 Q214.031 602.746 214.031 600.211 Q214.031 596.08 211.114 593.475 Q208.232 590.871 203.58 590.871 Q200.281 590.871 196.601 592.017 Q192.955 593.163 188.788 595.489 L188.788 588.406 Q193.024 586.705 196.705 585.837 Q200.385 584.968 203.441 584.968 Q211.496 584.968 216.288 588.996 Q221.08 593.024 221.08 599.76 Q221.08 602.954 219.864 605.836 Q218.684 608.684 215.524 612.572 Q214.656 613.579 210.003 618.406 Q205.351 623.197 196.878 631.843 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M236.149 585.906 L263.684 585.906 L263.684 591.809 L242.573 591.809 L242.573 604.517 Q244.1 603.996 245.628 603.753 Q247.156 603.475 248.684 603.475 Q257.364 603.475 262.434 608.232 Q267.503 612.989 267.503 621.114 Q267.503 629.482 262.295 634.135 Q257.086 638.753 247.607 638.753 Q244.343 638.753 240.941 638.197 Q237.573 637.642 233.961 636.531 L233.961 629.482 Q237.086 631.184 240.42 632.017 Q243.753 632.85 247.468 632.85 Q253.475 632.85 256.982 629.69 Q260.489 626.531 260.489 621.114 Q260.489 615.697 256.982 612.538 Q253.475 609.378 247.468 609.378 Q244.656 609.378 241.843 610.003 Q239.066 610.628 236.149 611.948 L236.149 585.906 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M122.712 380.65 L134.17 380.65 L134.17 341.101 L121.705 343.601 L121.705 337.212 L134.101 334.712 L141.115 334.712 L141.115 380.65 L152.573 380.65 L152.573 386.552 L122.712 386.552 L122.712 380.65 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M166.74 377.733 L174.066 377.733 L174.066 386.552 L166.74 386.552 L166.74 377.733 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M189.413 334.712 L216.948 334.712 L216.948 340.615 L195.837 340.615 L195.837 353.323 Q197.364 352.802 198.892 352.559 Q200.42 352.282 201.948 352.282 Q210.628 352.282 215.698 357.039 Q220.767 361.795 220.767 369.92 Q220.767 378.288 215.559 382.941 Q210.35 387.559 200.871 387.559 Q197.608 387.559 194.205 387.004 Q190.837 386.448 187.226 385.337 L187.226 378.288 Q190.351 379.99 193.684 380.823 Q197.017 381.656 200.732 381.656 Q206.739 381.656 210.246 378.497 Q213.753 375.337 213.753 369.92 Q213.753 364.504 210.246 361.344 Q206.739 358.184 200.732 358.184 Q197.92 358.184 195.108 358.809 Q192.33 359.434 189.413 360.754 L189.413 334.712 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M249.586 339.33 Q244.17 339.33 241.427 344.678 Q238.718 349.99 238.718 360.684 Q238.718 371.344 241.427 376.691 Q244.17 382.004 249.586 382.004 Q255.038 382.004 257.746 376.691 Q260.489 371.344 260.489 360.684 Q260.489 349.99 257.746 344.678 Q255.038 339.33 249.586 339.33 M249.586 333.775 Q258.302 333.775 262.885 340.684 Q267.503 347.559 267.503 360.684 Q267.503 373.775 262.885 380.684 Q258.302 387.559 249.586 387.559 Q240.871 387.559 236.253 380.684 Q231.67 373.775 231.67 360.684 Q231.67 347.559 236.253 340.684 Q240.871 333.775 249.586 333.775 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M35.6047 1045.01 Q31.2992 1042.61 29.2506 1039.28 Q27.202 1035.94 27.202 1031.43 Q27.202 1025.35 31.4728 1022.06 Q35.7089 1018.76 43.5561 1018.76 L67.0282 1018.76 L67.0282 1025.18 L43.7644 1025.18 Q38.1741 1025.18 35.4658 1027.16 Q32.7575 1029.14 32.7575 1033.2 Q32.7575 1038.17 36.0561 1041.05 Q39.3547 1043.93 45.0491 1043.93 L67.0282 1043.93 L67.0282 1050.35 L43.7644 1050.35 Q38.1394 1050.35 35.4658 1052.33 Q32.7575 1054.31 32.7575 1058.44 Q32.7575 1063.34 36.0908 1066.22 Q39.3894 1069.1 45.0491 1069.1 L67.0282 1069.1 L67.0282 1075.53 L28.1395 1075.53 L28.1395 1069.1 L34.1811 1069.1 Q30.6047 1066.92 28.9034 1063.86 Q27.202 1060.81 27.202 1056.6 Q27.202 1052.37 29.3547 1049.42 Q31.5075 1046.43 35.6047 1045.01 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M47.4797 988.341 Q47.4797 996.084 49.2505 999.07 Q51.0213 1002.06 55.2921 1002.06 Q58.6949 1002.06 60.7087 999.834 Q62.6879 997.577 62.6879 993.723 Q62.6879 988.41 58.9379 985.216 Q55.1532 981.987 48.9033 981.987 L47.4797 981.987 L47.4797 988.341 M44.8408 975.598 L67.0282 975.598 L67.0282 981.987 L61.1254 981.987 Q64.6671 984.174 66.3684 987.438 Q68.0351 990.702 68.0351 995.424 Q68.0351 1001.4 64.7018 1004.94 Q61.3337 1008.44 55.7088 1008.44 Q49.1463 1008.44 45.813 1004.07 Q42.4797 999.66 42.4797 990.945 L42.4797 981.987 L41.8547 981.987 Q37.445 981.987 35.0492 984.903 Q32.6186 987.785 32.6186 993.028 Q32.6186 996.362 33.4172 999.521 Q34.2158 1002.68 35.813 1005.6 L29.9103 1005.6 Q28.5561 1002.09 27.8964 998.792 Q27.202 995.493 27.202 992.369 Q27.202 983.931 31.577 979.764 Q35.9519 975.598 44.8408 975.598 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M29.2853 937.647 L35.3269 937.647 Q33.9381 940.355 33.2436 943.272 Q32.5492 946.188 32.5492 949.313 Q32.5492 954.07 34.0075 956.466 Q35.4658 958.827 38.3825 958.827 Q40.6047 958.827 41.8894 957.126 Q43.1394 955.424 44.2852 950.285 L44.7713 948.098 Q46.2297 941.292 48.9033 938.445 Q51.5421 935.563 56.299 935.563 Q61.7157 935.563 64.8754 939.869 Q68.0351 944.14 68.0351 951.64 Q68.0351 954.765 67.4101 958.167 Q66.8198 961.535 65.6046 965.285 L59.0074 965.285 Q60.8476 961.744 61.7851 958.306 Q62.6879 954.869 62.6879 951.501 Q62.6879 946.987 61.1601 944.556 Q59.5976 942.126 56.7852 942.126 Q54.181 942.126 52.7921 943.897 Q51.4032 945.633 50.1185 951.57 L49.5977 953.792 Q48.3477 959.73 45.7783 962.369 Q43.1741 965.008 38.6603 965.008 Q33.1742 965.008 30.1881 961.119 Q27.202 957.23 27.202 950.077 Q27.202 946.535 27.7228 943.41 Q28.2436 940.285 29.2853 937.647 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M29.2853 900.598 L35.3269 900.598 Q33.9381 903.306 33.2436 906.223 Q32.5492 909.14 32.5492 912.265 Q32.5492 917.022 34.0075 919.418 Q35.4658 921.779 38.3825 921.779 Q40.6047 921.779 41.8894 920.077 Q43.1394 918.376 44.2852 913.237 L44.7713 911.049 Q46.2297 904.244 48.9033 901.397 Q51.5421 898.515 56.299 898.515 Q61.7157 898.515 64.8754 902.82 Q68.0351 907.091 68.0351 914.591 Q68.0351 917.716 67.4101 921.119 Q66.8198 924.487 65.6046 928.237 L59.0074 928.237 Q60.8476 924.695 61.7851 921.258 Q62.6879 917.82 62.6879 914.452 Q62.6879 909.938 61.1601 907.508 Q59.5976 905.077 56.7852 905.077 Q54.181 905.077 52.7921 906.848 Q51.4032 908.584 50.1185 914.522 L49.5977 916.744 Q48.3477 922.681 45.7783 925.32 Q43.1741 927.959 38.6603 927.959 Q33.1742 927.959 30.1881 924.07 Q27.202 920.181 27.202 913.029 Q27.202 909.487 27.7228 906.362 Q28.2436 903.237 29.2853 900.598 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M61.1949 859.557 L81.8197 859.557 L81.8197 865.98 L28.1395 865.98 L28.1395 859.557 L34.0422 859.557 Q30.57 857.543 28.9034 854.487 Q27.202 851.397 27.202 847.126 Q27.202 840.043 32.827 835.633 Q38.4519 831.189 47.6185 831.189 Q56.7852 831.189 62.4101 835.633 Q68.0351 840.043 68.0351 847.126 Q68.0351 851.397 66.3684 854.487 Q64.6671 857.543 61.1949 859.557 M47.6185 837.821 Q40.57 837.821 36.5769 840.737 Q32.5492 843.619 32.5492 848.689 Q32.5492 853.758 36.5769 856.675 Q40.57 859.557 47.6185 859.557 Q54.6671 859.557 58.6949 856.675 Q62.6879 853.758 62.6879 848.689 Q62.6879 843.619 58.6949 840.737 Q54.6671 837.821 47.6185 837.821 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M32.6186 805.529 Q32.6186 810.668 36.6464 813.654 Q40.6394 816.64 47.6185 816.64 Q54.5977 816.64 58.6254 813.689 Q62.6185 810.703 62.6185 805.529 Q62.6185 800.425 58.5907 797.439 Q54.5629 794.453 47.6185 794.453 Q40.7089 794.453 36.6811 797.439 Q32.6186 800.425 32.6186 805.529 M27.202 805.529 Q27.202 797.196 32.6186 792.439 Q38.0353 787.682 47.6185 787.682 Q57.1671 787.682 62.6185 792.439 Q68.0351 797.196 68.0351 805.529 Q68.0351 813.897 62.6185 818.654 Q57.1671 823.376 47.6185 823.376 Q38.0353 823.376 32.6186 818.654 Q27.202 813.897 27.202 805.529 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M29.2853 752.3 L35.3269 752.3 Q33.9381 755.009 33.2436 757.925 Q32.5492 760.842 32.5492 763.967 Q32.5492 768.724 34.0075 771.12 Q35.4658 773.481 38.3825 773.481 Q40.6047 773.481 41.8894 771.779 Q43.1394 770.078 44.2852 764.939 L44.7713 762.752 Q46.2297 755.946 48.9033 753.099 Q51.5421 750.217 56.299 750.217 Q61.7157 750.217 64.8754 754.523 Q68.0351 758.793 68.0351 766.293 Q68.0351 769.418 67.4101 772.821 Q66.8198 776.189 65.6046 779.939 L59.0074 779.939 Q60.8476 776.397 61.7851 772.96 Q62.6879 769.522 62.6879 766.154 Q62.6879 761.641 61.1601 759.21 Q59.5976 756.779 56.7852 756.779 Q54.181 756.779 52.7921 758.55 Q51.4032 760.286 50.1185 766.224 L49.5977 768.446 Q48.3477 774.384 45.7783 777.022 Q43.1741 779.661 38.6603 779.661 Q33.1742 779.661 30.1881 775.772 Q27.202 771.884 27.202 764.731 Q27.202 761.189 27.7228 758.064 Q28.2436 754.939 29.2853 752.3 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M28.1395 740.043 L28.1395 733.655 L67.0282 733.655 L67.0282 740.043 L28.1395 740.043 M13.0007 740.043 L13.0007 733.655 L21.0909 733.655 L21.0909 740.043 L13.0007 740.043 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M17.0979 713.967 L28.1395 713.967 L28.1395 700.808 L33.1047 700.808 L33.1047 713.967 L54.2157 713.967 Q58.9726 713.967 60.3268 712.683 Q61.681 711.363 61.681 707.37 L61.681 700.808 L67.0282 700.808 L67.0282 707.37 Q67.0282 714.766 64.2851 717.578 Q61.5074 720.391 54.2157 720.391 L33.1047 720.391 L33.1047 725.078 L28.1395 725.078 L28.1395 720.391 L17.0979 720.391 L17.0979 713.967 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M28.1395 692.405 L28.1395 686.016 L67.0282 686.016 L67.0282 692.405 L28.1395 692.405 M13.0007 692.405 L13.0007 686.016 L21.0909 686.016 L21.0909 692.405 L13.0007 692.405 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M32.6186 657.579 Q32.6186 662.717 36.6464 665.704 Q40.6394 668.69 47.6185 668.69 Q54.5977 668.69 58.6254 665.738 Q62.6185 662.752 62.6185 657.579 Q62.6185 652.474 58.5907 649.488 Q54.5629 646.502 47.6185 646.502 Q40.7089 646.502 36.6811 649.488 Q32.6186 652.474 32.6186 657.579 M27.202 657.579 Q27.202 649.245 32.6186 644.488 Q38.0353 639.732 47.6185 639.732 Q57.1671 639.732 62.6185 644.488 Q68.0351 649.245 68.0351 657.579 Q68.0351 665.947 62.6185 670.704 Q57.1671 675.426 47.6185 675.426 Q38.0353 675.426 32.6186 670.704 Q27.202 665.947 27.202 657.579 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M43.5561 596.815 L67.0282 596.815 L67.0282 603.204 L43.7644 603.204 Q38.2436 603.204 35.5005 605.357 Q32.7575 607.509 32.7575 611.815 Q32.7575 616.989 36.0561 619.975 Q39.3547 622.961 45.0491 622.961 L67.0282 622.961 L67.0282 629.384 L28.1395 629.384 L28.1395 622.961 L34.1811 622.961 Q30.6742 620.669 28.9381 617.579 Q27.202 614.454 27.202 610.391 Q27.202 603.69 31.3686 600.253 Q35.5005 596.815 43.5561 596.815 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M13.0007 562.058 L13.0007 547.336 L17.9659 547.336 L17.9659 555.669 L71.4379 555.669 L71.4379 547.336 L76.4031 547.336 L76.4031 562.058 L13.0007 562.058 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M35.6047 503.448 Q31.2992 501.052 29.2506 497.718 Q27.202 494.385 27.202 489.871 Q27.202 483.795 31.4728 480.496 Q35.7089 477.198 43.5561 477.198 L67.0282 477.198 L67.0282 483.621 L43.7644 483.621 Q38.1741 483.621 35.4658 485.6 Q32.7575 487.58 32.7575 491.642 Q32.7575 496.607 36.0561 499.489 Q39.3547 502.371 45.0491 502.371 L67.0282 502.371 L67.0282 508.795 L43.7644 508.795 Q38.1394 508.795 35.4658 510.774 Q32.7575 512.753 32.7575 516.885 Q32.7575 521.781 36.0908 524.663 Q39.3894 527.545 45.0491 527.545 L67.0282 527.545 L67.0282 533.968 L28.1395 533.968 L28.1395 527.545 L34.1811 527.545 Q30.6047 525.357 28.9034 522.302 Q27.202 519.246 27.202 515.045 Q27.202 510.809 29.3547 507.857 Q31.5075 504.871 35.6047 503.448 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M13.0007 449.524 L76.4031 449.524 L76.4031 464.246 L71.4379 464.246 L71.4379 455.948 L17.9659 455.948 L17.9659 464.246 L13.0007 464.246 L13.0007 449.524 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M626.29 18.8205 L626.29 65.8515 L636.174 65.8515 Q648.692 65.8515 654.484 60.1802 Q660.318 54.509 660.318 42.2752 Q660.318 30.1225 654.484 24.4918 Q648.692 18.8205 636.174 18.8205 L626.29 18.8205 M618.107 12.096 L634.918 12.096 Q652.499 12.096 660.723 19.4281 Q668.946 26.7198 668.946 42.2752 Q668.946 57.9117 660.682 65.2439 Q652.418 72.576 634.918 72.576 L618.107 72.576 L618.107 12.096 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M720.474 48.0275 L720.474 51.6733 L686.203 51.6733 Q686.689 59.3701 690.821 63.421 Q694.993 67.4314 702.407 67.4314 Q706.7 67.4314 710.711 66.3781 Q714.762 65.3249 718.732 63.2184 L718.732 70.267 Q714.721 71.9684 710.508 72.8596 Q706.295 73.7508 701.961 73.7508 Q691.104 73.7508 684.745 67.4314 Q678.425 61.1119 678.425 50.3365 Q678.425 39.1965 684.421 32.6746 Q690.456 26.1121 700.665 26.1121 Q709.82 26.1121 715.126 32.0264 Q720.474 37.9003 720.474 48.0275 M713.02 45.84 Q712.939 39.7232 709.577 36.0774 Q706.255 32.4315 700.746 32.4315 Q694.507 32.4315 690.74 35.9558 Q687.013 39.4801 686.446 45.8805 L713.02 45.84 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M755.676 9.54393 L755.676 15.7418 L748.546 15.7418 Q744.536 15.7418 742.956 17.3622 Q741.417 18.9825 741.417 23.1955 L741.417 27.2059 L753.691 27.2059 L753.691 32.9987 L741.417 32.9987 L741.417 72.576 L733.923 72.576 L733.923 32.9987 L726.793 32.9987 L726.793 27.2059 L733.923 27.2059 L733.923 24.0462 Q733.923 16.471 737.447 13.0277 Q740.971 9.54393 748.627 9.54393 L755.676 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M782.533 49.7694 Q773.5 49.7694 770.016 51.8354 Q766.532 53.9013 766.532 58.8839 Q766.532 62.8538 769.125 65.2034 Q771.758 67.5124 776.255 67.5124 Q782.452 67.5124 786.179 63.1374 Q789.947 58.7219 789.947 51.4303 L789.947 49.7694 L782.533 49.7694 M797.4 46.6907 L797.4 72.576 L789.947 72.576 L789.947 65.6895 Q787.394 69.8214 783.587 71.8063 Q779.779 73.7508 774.27 73.7508 Q767.302 73.7508 763.17 69.8619 Q759.079 65.9325 759.079 59.3701 Q759.079 51.7138 764.183 47.825 Q769.327 43.9361 779.495 43.9361 L789.947 43.9361 L789.947 43.2069 Q789.947 38.0623 786.544 35.2672 Q783.182 32.4315 777.065 32.4315 Q773.176 32.4315 769.489 33.3632 Q765.803 34.295 762.4 36.1584 L762.4 29.2718 Q766.492 27.692 770.34 26.9223 Q774.189 26.1121 777.834 26.1121 Q787.678 26.1121 792.539 31.2163 Q797.4 36.3204 797.4 46.6907 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M811.983 54.671 L811.983 27.2059 L819.437 27.2059 L819.437 54.3874 Q819.437 60.8284 821.949 64.0691 Q824.46 67.2693 829.483 67.2693 Q835.519 67.2693 839.003 63.421 Q842.527 59.5726 842.527 52.9291 L842.527 27.2059 L849.981 27.2059 L849.981 72.576 L842.527 72.576 L842.527 65.6084 Q839.813 69.7404 836.208 71.7658 Q832.643 73.7508 827.904 73.7508 Q820.085 73.7508 816.034 68.8897 Q811.983 64.0286 811.983 54.671 M830.739 26.1121 L830.739 26.1121 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M865.334 9.54393 L872.788 9.54393 L872.788 72.576 L865.334 72.576 L865.334 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M895.756 14.324 L895.756 27.2059 L911.109 27.2059 L911.109 32.9987 L895.756 32.9987 L895.756 57.6282 Q895.756 63.1779 897.255 64.7578 Q898.794 66.3376 903.453 66.3376 L911.109 66.3376 L911.109 72.576 L903.453 72.576 Q894.824 72.576 891.543 69.3758 Q888.262 66.1351 888.262 57.6282 L888.262 32.9987 L882.793 32.9987 L882.793 27.2059 L888.262 27.2059 L888.262 14.324 L895.756 14.324 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M922.816 62.2867 L931.364 62.2867 L931.364 72.576 L922.816 72.576 L922.816 62.2867 M922.816 29.6769 L931.364 29.6769 L931.364 39.9662 L922.816 39.9662 L922.816 29.6769 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M975.559 12.096 L987.752 12.096 L1003.19 53.2532 L1018.7 12.096 L1030.89 12.096 L1030.89 72.576 L1022.91 72.576 L1022.91 19.4686 L1007.32 60.9499 L999.095 60.9499 L983.499 19.4686 L983.499 72.576 L975.559 72.576 L975.559 12.096 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1067.43 49.7694 Q1058.4 49.7694 1054.92 51.8354 Q1051.43 53.9013 1051.43 58.8839 Q1051.43 62.8538 1054.03 65.2034 Q1056.66 67.5124 1061.15 67.5124 Q1067.35 67.5124 1071.08 63.1374 Q1074.85 58.7219 1074.85 51.4303 L1074.85 49.7694 L1067.43 49.7694 M1082.3 46.6907 L1082.3 72.576 L1074.85 72.576 L1074.85 65.6895 Q1072.29 69.8214 1068.49 71.8063 Q1064.68 73.7508 1059.17 73.7508 Q1052.2 73.7508 1048.07 69.8619 Q1043.98 65.9325 1043.98 59.3701 Q1043.98 51.7138 1049.08 47.825 Q1054.23 43.9361 1064.4 43.9361 L1074.85 43.9361 L1074.85 43.2069 Q1074.85 38.0623 1071.44 35.2672 Q1068.08 32.4315 1061.96 32.4315 Q1058.08 32.4315 1054.39 33.3632 Q1050.7 34.295 1047.3 36.1584 L1047.3 29.2718 Q1051.39 27.692 1055.24 26.9223 Q1059.09 26.1121 1062.73 26.1121 Q1072.58 26.1121 1077.44 31.2163 Q1082.3 36.3204 1082.3 46.6907 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1126.58 28.5427 L1126.58 35.5912 Q1123.42 33.9709 1120.01 33.1607 Q1116.61 32.3505 1112.97 32.3505 Q1107.42 32.3505 1104.62 34.0519 Q1101.87 35.7533 1101.87 39.156 Q1101.87 41.7486 1103.85 43.2475 Q1105.84 44.7058 1111.83 46.0426 L1114.38 46.6097 Q1122.32 48.3111 1125.64 51.4303 Q1129.01 54.509 1129.01 60.0587 Q1129.01 66.3781 1123.98 70.0644 Q1119 73.7508 1110.25 73.7508 Q1106.61 73.7508 1102.64 73.0216 Q1098.71 72.3329 1094.33 70.9151 L1094.33 63.2184 Q1098.46 65.3654 1102.47 66.4591 Q1106.48 67.5124 1110.41 67.5124 Q1115.68 67.5124 1118.52 65.73 Q1121.35 63.9071 1121.35 60.6258 Q1121.35 57.5877 1119.29 55.9673 Q1117.26 54.3469 1110.33 52.8481 L1107.74 52.2405 Q1100.81 50.7821 1097.73 47.7845 Q1094.66 44.7463 1094.66 39.4801 Q1094.66 33.0797 1099.19 29.5959 Q1103.73 26.1121 1112.07 26.1121 Q1116.21 26.1121 1119.85 26.7198 Q1123.5 27.3274 1126.58 28.5427 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1169.8 28.5427 L1169.8 35.5912 Q1166.64 33.9709 1163.24 33.1607 Q1159.83 32.3505 1156.19 32.3505 Q1150.64 32.3505 1147.84 34.0519 Q1145.09 35.7533 1145.09 39.156 Q1145.09 41.7486 1147.07 43.2475 Q1149.06 44.7058 1155.05 46.0426 L1157.61 46.6097 Q1165.55 48.3111 1168.87 51.4303 Q1172.23 54.509 1172.23 60.0587 Q1172.23 66.3781 1167.21 70.0644 Q1162.22 73.7508 1153.47 73.7508 Q1149.83 73.7508 1145.86 73.0216 Q1141.93 72.3329 1137.55 70.9151 L1137.55 63.2184 Q1141.69 65.3654 1145.7 66.4591 Q1149.71 67.5124 1153.64 67.5124 Q1158.9 67.5124 1161.74 65.73 Q1164.57 63.9071 1164.57 60.6258 Q1164.57 57.5877 1162.51 55.9673 Q1160.48 54.3469 1153.56 52.8481 L1150.96 52.2405 Q1144.04 50.7821 1140.96 47.7845 Q1137.88 44.7463 1137.88 39.4801 Q1137.88 33.0797 1142.42 29.5959 Q1146.95 26.1121 1155.3 26.1121 Q1159.43 26.1121 1163.08 26.7198 Q1166.72 27.3274 1169.8 28.5427 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1217.68 65.7705 L1217.68 89.8329 L1210.19 89.8329 L1210.19 27.2059 L1217.68 27.2059 L1217.68 34.0924 Q1220.03 30.0415 1223.6 28.0971 Q1227.2 26.1121 1232.18 26.1121 Q1240.45 26.1121 1245.59 32.6746 Q1250.78 39.2371 1250.78 49.9314 Q1250.78 60.6258 1245.59 67.1883 Q1240.45 73.7508 1232.18 73.7508 Q1227.2 73.7508 1223.6 71.8063 Q1220.03 69.8214 1217.68 65.7705 M1243.04 49.9314 Q1243.04 41.7081 1239.64 37.0496 Q1236.28 32.3505 1230.36 32.3505 Q1224.45 32.3505 1221.04 37.0496 Q1217.68 41.7081 1217.68 49.9314 Q1217.68 58.1548 1221.04 62.8538 Q1224.45 67.5124 1230.36 67.5124 Q1236.28 67.5124 1239.64 62.8538 Q1243.04 58.1548 1243.04 49.9314 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1280.71 32.4315 Q1274.72 32.4315 1271.23 37.1306 Q1267.75 41.7891 1267.75 49.9314 Q1267.75 58.0738 1271.19 62.7728 Q1274.68 67.4314 1280.71 67.4314 Q1286.67 67.4314 1290.15 62.7323 Q1293.64 58.0333 1293.64 49.9314 Q1293.64 41.8701 1290.15 37.1711 Q1286.67 32.4315 1280.71 32.4315 M1280.71 26.1121 Q1290.44 26.1121 1295.99 32.4315 Q1301.54 38.7509 1301.54 49.9314 Q1301.54 61.0714 1295.99 67.4314 Q1290.44 73.7508 1280.71 73.7508 Q1270.95 73.7508 1265.4 67.4314 Q1259.89 61.0714 1259.89 49.9314 Q1259.89 38.7509 1265.4 32.4315 Q1270.95 26.1121 1280.71 26.1121 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1342.81 28.5427 L1342.81 35.5912 Q1339.65 33.9709 1336.25 33.1607 Q1332.85 32.3505 1329.2 32.3505 Q1323.65 32.3505 1320.86 34.0519 Q1318.1 35.7533 1318.1 39.156 Q1318.1 41.7486 1320.09 43.2475 Q1322.07 44.7058 1328.07 46.0426 L1330.62 46.6097 Q1338.56 48.3111 1341.88 51.4303 Q1345.24 54.509 1345.24 60.0587 Q1345.24 66.3781 1340.22 70.0644 Q1335.24 73.7508 1326.49 73.7508 Q1322.84 73.7508 1318.87 73.0216 Q1314.94 72.3329 1310.57 70.9151 L1310.57 63.2184 Q1314.7 65.3654 1318.71 66.4591 Q1322.72 67.5124 1326.65 67.5124 Q1331.92 67.5124 1334.75 65.73 Q1337.59 63.9071 1337.59 60.6258 Q1337.59 57.5877 1335.52 55.9673 Q1333.5 54.3469 1326.57 52.8481 L1323.98 52.2405 Q1317.05 50.7821 1313.97 47.7845 Q1310.89 44.7463 1310.89 39.4801 Q1310.89 33.0797 1315.43 29.5959 Q1319.97 26.1121 1328.31 26.1121 Q1332.44 26.1121 1336.09 26.7198 Q1339.74 27.3274 1342.81 28.5427 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1357.11 27.2059 L1364.57 27.2059 L1364.57 72.576 L1357.11 72.576 L1357.11 27.2059 M1357.11 9.54393 L1364.57 9.54393 L1364.57 18.9825 L1357.11 18.9825 L1357.11 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1387.54 14.324 L1387.54 27.2059 L1402.89 27.2059 L1402.89 32.9987 L1387.54 32.9987 L1387.54 57.6282 Q1387.54 63.1779 1389.03 64.7578 Q1390.57 66.3376 1395.23 66.3376 L1402.89 66.3376 L1402.89 72.576 L1395.23 72.576 Q1386.6 72.576 1383.32 69.3758 Q1380.04 66.1351 1380.04 57.6282 L1380.04 32.9987 L1374.57 32.9987 L1374.57 27.2059 L1380.04 27.2059 L1380.04 14.324 L1387.54 14.324 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1412.69 27.2059 L1420.15 27.2059 L1420.15 72.576 L1412.69 72.576 L1412.69 27.2059 M1412.69 9.54393 L1420.15 9.54393 L1420.15 18.9825 L1412.69 18.9825 L1412.69 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1453.32 32.4315 Q1447.33 32.4315 1443.84 37.1306 Q1440.36 41.7891 1440.36 49.9314 Q1440.36 58.0738 1443.8 62.7728 Q1447.29 67.4314 1453.32 67.4314 Q1459.28 67.4314 1462.76 62.7323 Q1466.25 58.0333 1466.25 49.9314 Q1466.25 41.8701 1462.76 37.1711 Q1459.28 32.4315 1453.32 32.4315 M1453.32 26.1121 Q1463.04 26.1121 1468.59 32.4315 Q1474.14 38.7509 1474.14 49.9314 Q1474.14 61.0714 1468.59 67.4314 Q1463.04 73.7508 1453.32 73.7508 Q1443.56 73.7508 1438.01 67.4314 Q1432.5 61.0714 1432.5 49.9314 Q1432.5 38.7509 1438.01 32.4315 Q1443.56 26.1121 1453.32 26.1121 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1524.21 45.1919 L1524.21 72.576 L1516.76 72.576 L1516.76 45.4349 Q1516.76 38.994 1514.25 35.7938 Q1511.74 32.5936 1506.71 32.5936 Q1500.68 32.5936 1497.19 36.4419 Q1493.71 40.2903 1493.71 46.9338 L1493.71 72.576 L1486.22 72.576 L1486.22 27.2059 L1493.71 27.2059 L1493.71 34.2544 Q1496.38 30.163 1499.99 28.1376 Q1503.63 26.1121 1508.37 26.1121 Q1516.19 26.1121 1520.2 30.9732 Q1524.21 35.7938 1524.21 45.1919 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1586.07 49.7694 Q1577.04 49.7694 1573.55 51.8354 Q1570.07 53.9013 1570.07 58.8839 Q1570.07 62.8538 1572.66 65.2034 Q1575.3 67.5124 1579.79 67.5124 Q1585.99 67.5124 1589.72 63.1374 Q1593.48 58.7219 1593.48 51.4303 L1593.48 49.7694 L1586.07 49.7694 M1600.94 46.6907 L1600.94 72.576 L1593.48 72.576 L1593.48 65.6895 Q1590.93 69.8214 1587.12 71.8063 Q1583.32 73.7508 1577.81 73.7508 Q1570.84 73.7508 1566.71 69.8619 Q1562.62 65.9325 1562.62 59.3701 Q1562.62 51.7138 1567.72 47.825 Q1572.86 43.9361 1583.03 43.9361 L1593.48 43.9361 L1593.48 43.2069 Q1593.48 38.0623 1590.08 35.2672 Q1586.72 32.4315 1580.6 32.4315 Q1576.71 32.4315 1573.03 33.3632 Q1569.34 34.295 1565.94 36.1584 L1565.94 29.2718 Q1570.03 27.692 1573.88 26.9223 Q1577.73 26.1121 1581.37 26.1121 Q1591.22 26.1121 1596.08 31.2163 Q1600.94 36.3204 1600.94 46.6907 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1639.26 9.54393 L1639.26 15.7418 L1632.13 15.7418 Q1628.12 15.7418 1626.54 17.3622 Q1625 18.9825 1625 23.1955 L1625 27.2059 L1637.27 27.2059 L1637.27 32.9987 L1625 32.9987 L1625 72.576 L1617.51 72.576 L1617.51 32.9987 L1610.38 32.9987 L1610.38 27.2059 L1617.51 27.2059 L1617.51 24.0462 Q1617.51 16.471 1621.03 13.0277 Q1624.55 9.54393 1632.21 9.54393 L1639.26 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1651.41 14.324 L1651.41 27.2059 L1666.76 27.2059 L1666.76 32.9987 L1651.41 32.9987 L1651.41 57.6282 Q1651.41 63.1779 1652.91 64.7578 Q1654.45 66.3376 1659.11 66.3376 L1666.76 66.3376 L1666.76 72.576 L1659.11 72.576 Q1650.48 72.576 1647.2 69.3758 Q1643.92 66.1351 1643.92 57.6282 L1643.92 32.9987 L1638.45 32.9987 L1638.45 27.2059 L1643.92 27.2059 L1643.92 14.324 L1651.41 14.324 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1715.38 48.0275 L1715.38 51.6733 L1681.1 51.6733 Q1681.59 59.3701 1685.72 63.421 Q1689.9 67.4314 1697.31 67.4314 Q1701.6 67.4314 1705.61 66.3781 Q1709.66 65.3249 1713.63 63.2184 L1713.63 70.267 Q1709.62 71.9684 1705.41 72.8596 Q1701.2 73.7508 1696.86 73.7508 Q1686.01 73.7508 1679.65 67.4314 Q1673.33 61.1119 1673.33 50.3365 Q1673.33 39.1965 1679.32 32.6746 Q1685.36 26.1121 1695.57 26.1121 Q1704.72 26.1121 1710.03 32.0264 Q1715.38 37.9003 1715.38 48.0275 M1707.92 45.84 Q1707.84 39.7232 1704.48 36.0774 Q1701.16 32.4315 1695.65 32.4315 Q1689.41 32.4315 1685.64 35.9558 Q1681.92 39.4801 1681.35 45.8805 L1707.92 45.84 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1753.9 34.1734 Q1752.64 33.4443 1751.15 33.1202 Q1749.69 32.7556 1747.9 32.7556 Q1741.58 32.7556 1738.18 36.8875 Q1734.82 40.9789 1734.82 48.6757 L1734.82 72.576 L1727.33 72.576 L1727.33 27.2059 L1734.82 27.2059 L1734.82 34.2544 Q1737.17 30.1225 1740.94 28.1376 Q1744.7 26.1121 1750.09 26.1121 Q1750.86 26.1121 1751.79 26.2337 Q1752.72 26.3147 1753.86 26.5172 L1753.9 34.1734 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1817.09 44.2197 Q1819.73 45.1109 1822.2 48.0275 Q1824.71 50.9442 1827.22 56.0483 L1835.53 72.576 L1826.73 72.576 L1819 57.061 Q1816 50.9847 1813.16 48.9997 Q1810.37 47.0148 1805.51 47.0148 L1796.6 47.0148 L1796.6 72.576 L1788.41 72.576 L1788.41 12.096 L1806.89 12.096 Q1817.26 12.096 1822.36 16.4305 Q1827.46 20.7649 1827.46 29.5149 Q1827.46 35.2267 1824.79 38.994 Q1822.16 42.7613 1817.09 44.2197 M1796.6 18.8205 L1796.6 40.2903 L1806.89 40.2903 Q1812.8 40.2903 1815.8 37.5762 Q1818.84 34.8216 1818.84 29.5149 Q1818.84 24.2082 1815.8 21.5346 Q1812.8 18.8205 1806.89 18.8205 L1796.6 18.8205 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1841.24 54.671 L1841.24 27.2059 L1848.69 27.2059 L1848.69 54.3874 Q1848.69 60.8284 1851.2 64.0691 Q1853.71 67.2693 1858.74 67.2693 Q1864.77 67.2693 1868.26 63.421 Q1871.78 59.5726 1871.78 52.9291 L1871.78 27.2059 L1879.23 27.2059 L1879.23 72.576 L1871.78 72.576 L1871.78 65.6084 Q1869.07 69.7404 1865.46 71.7658 Q1861.9 73.7508 1857.16 73.7508 Q1849.34 73.7508 1845.29 68.8897 Q1841.24 64.0286 1841.24 54.671 M1859.99 26.1121 L1859.99 26.1121 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1932.3 45.1919 L1932.3 72.576 L1924.85 72.576 L1924.85 45.4349 Q1924.85 38.994 1922.34 35.7938 Q1919.82 32.5936 1914.8 32.5936 Q1908.77 32.5936 1905.28 36.4419 Q1901.8 40.2903 1901.8 46.9338 L1901.8 72.576 L1894.3 72.576 L1894.3 27.2059 L1901.8 27.2059 L1901.8 34.2544 Q1904.47 30.163 1908.08 28.1376 Q1911.72 26.1121 1916.46 26.1121 Q1924.28 26.1121 1928.29 30.9732 Q1932.3 35.7938 1932.3 45.1919 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1949.07 62.2867 L1957.62 62.2867 L1957.62 72.576 L1949.07 72.576 L1949.07 62.2867 M1949.07 29.6769 L1957.62 29.6769 L1957.62 39.9662 L1949.07 39.9662 L1949.07 29.6769 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M2009.59 65.6895 L2038.15 65.6895 L2038.15 72.576 L1999.75 72.576 L1999.75 65.6895 Q2004.41 60.8689 2012.43 52.7671 Q2020.49 44.6248 2022.56 42.2752 Q2026.49 37.8598 2028.02 34.8216 Q2029.6 31.7429 2029.6 28.7857 Q2029.6 23.9651 2026.2 20.927 Q2022.84 17.8888 2017.41 17.8888 Q2013.56 17.8888 2009.27 19.2256 Q2005.02 20.5624 2000.15 23.2765 L2000.15 15.0127 Q2005.1 13.0277 2009.39 12.015 Q2013.68 11.0023 2017.25 11.0023 Q2026.65 11.0023 2032.24 15.7013 Q2037.83 20.4004 2037.83 28.2591 Q2037.83 31.9859 2036.41 35.3482 Q2035.03 38.6699 2031.35 43.2069 Q2030.33 44.3817 2024.91 50.0125 Q2019.48 55.6027 2009.59 65.6895 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip932)" style="stroke:#009af9; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" points="
  361.501,1365.41 366.334,1365.1 371.167,1364.2 376,1362.69 380.833,1360.59 385.666,1357.89 390.5,1354.59 395.333,1350.7 400.166,1346.22 404.999,1341.16 
  409.832,1335.52 414.665,1329.31 419.498,1322.53 424.332,1315.19 429.165,1307.3 433.998,1298.87 438.831,1289.89 443.664,1280.39 448.497,1270.38 453.33,1259.86 
  458.164,1248.84 462.997,1237.33 467.83,1225.34 472.663,1212.9 477.496,1200 482.329,1186.66 487.162,1172.9 491.996,1158.73 496.829,1144.16 501.662,1129.21 
  506.495,1113.9 511.328,1098.23 516.161,1082.23 520.994,1065.9 525.828,1049.28 530.661,1032.36 535.494,1015.18 540.327,997.743 545.16,980.072 549.993,962.183 
  554.826,944.096 559.66,925.826 564.493,907.393 569.326,888.815 574.159,870.111 578.992,851.3 583.825,832.399 588.658,813.429 593.492,794.408 598.325,775.355 
  603.158,756.289 607.991,737.229 612.824,718.195 617.657,699.205 622.49,680.278 627.324,661.433 632.157,642.69 636.99,624.067 641.823,605.582 646.656,587.253 
  651.489,569.101 656.322,551.141 661.156,533.393 665.989,515.874 670.822,498.602 675.655,481.594 680.488,464.866 685.321,448.437 690.154,432.321 694.988,416.536 
  699.821,401.097 704.654,386.019 709.487,371.317 714.32,357.007 719.153,343.102 723.986,329.617 728.82,316.565 733.653,303.959 738.486,291.811 743.319,280.134 
  748.152,268.939 752.985,258.238 757.818,248.042 762.652,238.359 767.485,229.202 772.318,220.577 777.151,212.494 781.984,204.961 786.817,197.986 791.65,191.576 
  796.484,185.736 801.317,180.473 806.15,175.793 810.983,171.699 815.816,168.196 820.649,165.287 825.482,162.976 830.316,161.264 835.149,160.153 839.982,159.645 
  844.815,159.74 849.648,160.437 854.481,161.737 859.314,163.637 864.148,166.137 868.981,169.232 873.814,172.922 878.647,177.2 883.48,182.064 888.313,187.509 
  893.147,193.528 897.98,200.117 902.813,207.268 907.646,214.975 912.479,223.229 917.312,232.023 922.145,241.348 926.979,251.194 931.812,261.55 936.645,272.408 
  941.478,283.757 946.311,295.584 951.144,307.877 955.977,320.626 960.811,333.816 965.644,347.435 970.477,361.468 975.31,375.903 980.143,390.724 984.976,405.917 
  989.809,421.467 994.643,437.357 999.476,453.573 1004.31,470.098 1009.14,486.915 1013.98,504.008 1018.81,521.359 1023.64,538.951 1028.47,556.767 1033.31,574.789 
  1038.14,592.998 1042.97,611.377 1047.81,629.908 1052.64,648.571 1057.47,667.348 1062.31,686.221 1067.14,705.17 1071.97,724.176 1076.81,743.221 1081.64,762.284 
  1086.47,781.349 1091.31,800.395 1096.14,819.403 1100.97,838.355 1105.8,857.231 1110.64,876.012 1115.47,894.679 1120.3,913.213 1125.14,931.598 1129.97,949.813 
  1134.8,967.842 1139.64,985.664 1144.47,1003.26 1149.3,1020.62 1154.14,1037.72 1158.97,1054.55 1163.8,1071.08 1168.64,1087.3 1173.47,1103.2 1178.3,1118.76 
  1183.14,1133.96 1187.97,1148.79 1192.8,1163.23 1197.63,1177.27 1202.47,1190.9 1207.3,1204.1 1212.13,1216.85 1216.97,1229.16 1221.8,1240.99 1226.63,1252.35 
  1231.47,1263.22 1236.3,1273.58 1241.13,1283.44 1245.97,1292.77 1250.8,1301.57 1255.63,1309.84 1260.47,1317.56 1265.3,1324.72 1270.13,1331.32 1274.96,1337.35 
  1279.8,1342.81 1284.63,1347.68 1289.46,1351.97 1294.3,1355.67 1299.13,1358.78 1303.96,1361.29 1308.8,1363.21 1313.63,1364.52 1318.46,1365.23 1323.3,1365.34 
  1328.13,1364.84 1332.96,1363.74 1337.8,1362.04 1342.63,1359.75 1347.46,1356.85 1352.3,1353.36 1357.13,1349.28 1361.96,1344.61 1366.79,1339.36 1371.63,1333.54 
  1376.46,1327.14 1381.29,1320.18 1386.13,1312.66 1390.96,1304.59 1395.79,1295.98 1400.63,1286.83 1405.46,1277.16 1410.29,1266.98 1415.13,1256.29 1419.96,1245.11 
  1424.79,1233.45 1429.63,1221.31 1434.46,1208.71 1439.29,1195.67 1444.12,1182.2 1448.96,1168.31 1453.79,1154.01 1458.62,1139.32 1463.46,1124.25 1468.29,1108.83 
  1473.12,1093.05 1477.96,1076.95 1482.79,1060.53 1487.62,1043.81 1492.46,1026.81 1497.29,1009.55 1502.12,992.042 1506.96,974.303 1511.79,956.353 1516.62,938.209 
  1521.46,919.889 1526.29,901.414 1531.12,882.8 1535.95,864.064 1540.79,845.227 1545.62,826.307 1550.45,807.324 1555.29,788.296 1560.12,769.243 1564.95,750.184 
  1569.79,731.138 1574.62,712.122 1579.45,693.156 1584.29,674.261 1589.12,655.453 1593.95,636.754 1598.79,618.18 1603.62,599.751 1608.45,581.484 1613.28,563.399 
  1618.12,545.513 1622.95,527.844 1627.78,510.41 1632.62,493.229 1637.45,476.316 1642.28,459.691 1647.12,443.368 1651.95,427.364 1656.78,411.696 1661.62,396.379 
  1666.45,381.428 1671.28,366.859 1676.12,352.685 1680.95,338.92 1685.78,325.581 1690.62,312.678 1695.45,300.225 1700.28,288.235 1705.11,276.719 1709.95,265.689 
  1714.78,255.155 1719.61,245.129 1724.45,235.621 1729.28,226.64 1734.11,218.195 1738.95,210.295 1743.78,202.947 1748.61,196.157 1753.45,189.934 1758.28,184.284 
  1763.11,179.214 1767.95,174.727 1772.78,170.827 1777.61,167.519 1782.44,164.806 1787.28,162.69 1792.11,161.173 1796.94,160.26 1801.78,159.949 1806.61,160.24 
  1811.44,161.134 1816.28,162.629 1821.11,164.724 1825.94,167.415 1830.78,170.701 1835.61,174.582 1840.44,179.05 1845.28,184.101 1850.11,189.731 1854.94,195.934 
  1859.78,202.702 1864.61,210.029 1869.44,217.91 1874.27,226.336 1879.11,235.299 1883.94,244.788 1888.77,254.795 1893.61,265.31 1898.44,276.321 1903.27,287.817 
  1908.11,299.79 1912.94,312.225 1917.77,325.112 1922.61,338.435 1927.44,352.182 1932.27,366.339 1937.11,380.891 1941.94,395.824 1946.77,411.126 1951.6,426.78 
  1956.44,442.768 1961.27,459.077 1966.1,475.689 1970.94,492.586 1975.77,509.753 1980.6,527.173 1985.44,544.829 1990.27,562.703 1995.1,580.777 1999.94,599.032 
  2004.77,617.45 2009.6,636.013 2014.44,654.701 2019.27,673.498 2024.1,692.385 2028.94,711.341 2033.77,730.349 2038.6,749.388 2043.43,768.441 2048.27,787.486 
  2053.1,806.507 2057.93,825.484 2062.77,844.398 2067.6,863.23 2072.43,881.961 2077.27,900.573 2082.1,919.046 2086.93,937.363 2091.77,955.504 2096.6,973.452 
  2101.43,991.19 2106.27,1008.7 2111.1,1025.96 2115.93,1042.96 2120.77,1059.68 2125.6,1076.1 2130.43,1092.21 2135.26,1107.98 2140.1,1123.42 2144.93,1138.49 
  2149.76,1153.18 2154.6,1167.48 2159.43,1181.38 2164.26,1194.86 2169.1,1207.91 2173.93,1220.51 2178.76,1232.66 2183.6,1244.33 2188.43,1255.52 2193.26,1266.22 
  2198.1,1276.42 2202.93,1286.1 2207.76,1295.26 2212.59,1303.88 2217.43,1311.96 2222.26,1319.49 2227.09,1326.47 2231.93,1332.88 2236.76,1338.73 2241.59,1343.99 
  2246.43,1348.68 2251.26,1352.77 2256.09,1356.28 2260.93,1359.19 2265.76,1361.51 2270.59,1363.23 2275.43,1364.35 2280.26,1364.87 2285.09,1364.78 2289.93,1364.09 
  2294.76,1362.8 
  "/>
<polyline clip-path="url(#clip932)" style="stroke:#e26f46; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" points="
  361.501,1365.41 366.334,1365.19 371.167,1364.55 376,1363.47 380.833,1361.95 385.666,1360 390.5,1357.61 395.333,1354.79 400.166,1351.52 404.999,1347.82 
  409.832,1343.68 414.665,1339.12 419.498,1334.12 424.332,1328.7 429.165,1322.86 433.998,1316.6 438.831,1309.93 443.664,1302.85 448.497,1295.38 453.33,1287.51 
  458.164,1279.26 462.997,1270.62 467.83,1261.63 472.663,1252.27 477.496,1242.55 482.329,1232.5 487.162,1222.11 491.996,1211.4 496.829,1200.37 501.662,1189.05 
  506.495,1177.43 511.328,1165.53 516.161,1153.37 520.994,1140.95 525.828,1128.29 530.661,1115.4 535.494,1102.29 540.327,1088.97 545.16,1075.47 549.993,1061.79 
  554.826,1047.94 559.66,1033.94 564.493,1019.8 569.326,1005.54 574.159,991.18 578.992,976.726 583.825,962.197 588.658,947.601 593.492,932.956 598.325,918.278 
  603.158,903.581 607.991,888.88 612.824,874.188 617.657,859.52 622.49,844.891 627.324,830.316 632.157,815.808 636.99,801.383 641.823,787.054 646.656,772.837 
  651.489,758.745 656.322,744.792 661.156,730.992 665.989,717.359 670.822,703.908 675.655,690.652 680.488,677.602 685.321,664.773 690.154,652.178 694.988,639.83 
  699.821,627.741 704.654,615.922 709.487,604.386 714.32,593.145 719.153,582.211 723.986,571.594 728.82,561.305 733.653,551.354 738.486,541.753 743.319,532.511 
  748.152,523.637 752.985,515.141 757.818,507.032 762.652,499.318 767.485,492.009 772.318,485.111 777.151,478.634 781.984,472.583 786.817,466.967 791.65,461.793 
  796.484,457.066 801.317,452.793 806.15,448.98 810.983,445.632 815.816,442.756 820.649,440.356 825.482,438.439 830.316,437.007 835.149,436.067 839.982,435.623 
  844.815,435.654 849.648,436.048 854.481,436.799 859.314,437.909 864.148,439.378 868.981,441.209 873.814,443.402 878.647,445.956 883.48,448.872 888.313,452.148 
  893.147,455.782 897.98,459.772 902.813,464.116 907.646,468.811 912.479,473.853 917.312,479.239 922.145,484.964 926.979,491.025 931.812,497.414 936.645,504.126 
  941.478,511.157 946.311,518.499 951.144,526.145 955.977,534.088 960.811,542.322 965.644,550.837 970.477,559.624 975.31,568.676 980.143,577.986 984.976,587.544 
  989.809,597.341 994.643,607.367 999.476,617.612 1004.31,628.067 1009.14,638.72 1013.98,649.561 1018.81,660.58 1023.64,671.765 1028.47,683.105 1033.31,694.589 
  1038.14,706.205 1042.97,717.942 1047.81,729.789 1052.64,741.732 1057.47,753.761 1062.31,765.864 1067.14,778.029 1071.97,790.243 1076.81,802.494 1081.64,814.77 
  1086.47,827.059 1091.31,839.349 1096.14,851.628 1100.97,863.882 1105.8,876.099 1110.64,888.267 1115.47,900.376 1120.3,912.412 1125.14,924.362 1129.97,936.215 
  1134.8,947.958 1139.64,959.581 1144.47,971.072 1149.3,982.419 1154.14,993.61 1158.97,1004.63 1163.8,1015.48 1168.64,1026.14 1173.47,1036.6 1178.3,1046.84 
  1183.14,1056.87 1187.97,1066.66 1192.8,1076.22 1197.63,1085.52 1202.47,1094.56 1207.3,1103.34 1212.13,1111.83 1216.97,1120.04 1221.8,1127.95 1226.63,1135.56 
  1231.47,1142.85 1236.3,1149.82 1241.13,1156.46 1245.97,1162.77 1250.8,1168.74 1255.63,1174.35 1260.47,1179.61 1265.3,1184.51 1270.13,1189.03 1274.96,1193.19 
  1279.8,1196.96 1284.63,1200.34 1289.46,1203.34 1294.3,1205.94 1299.13,1208.14 1303.96,1209.93 1308.8,1211.32 1313.63,1212.29 1318.46,1212.85 1323.3,1212.98 
  1328.13,1212.79 1332.96,1212.32 1337.8,1211.57 1342.63,1210.54 1347.46,1209.24 1352.3,1207.64 1357.13,1205.77 1361.96,1203.61 1366.79,1201.17 1371.63,1198.45 
  1376.46,1195.45 1381.29,1192.17 1386.13,1188.61 1390.96,1184.78 1395.79,1180.68 1400.63,1176.31 1405.46,1171.67 1410.29,1166.77 1415.13,1161.62 1419.96,1156.21 
  1424.79,1150.56 1429.63,1144.65 1434.46,1138.52 1439.29,1132.14 1444.12,1125.54 1448.96,1118.72 1453.79,1111.68 1458.62,1104.44 1463.46,1096.99 1468.29,1089.35 
  1473.12,1081.52 1477.96,1073.52 1482.79,1065.34 1487.62,1057 1492.46,1048.5 1497.29,1039.86 1502.12,1031.08 1506.96,1022.17 1511.79,1013.13 1516.62,1003.99 
  1521.46,994.74 1526.29,985.398 1531.12,975.969 1535.95,966.464 1540.79,956.892 1545.62,947.264 1550.45,937.59 1555.29,927.88 1560.12,918.14 1564.95,908.381 
  1569.79,898.614 1574.62,888.848 1579.45,879.095 1584.29,869.364 1589.12,859.665 1593.95,850.005 1598.79,840.396 1603.62,830.847 1608.45,821.368 1613.28,811.968 
  1618.12,802.659 1622.95,793.447 1627.78,784.344 1632.62,775.357 1637.45,766.496 1642.28,757.771 1647.12,749.189 1651.95,740.76 1656.78,732.491 1661.62,724.393 
  1666.45,716.473 1671.28,708.74 1676.12,701.2 1680.95,693.863 1685.78,686.734 1690.62,679.823 1695.45,673.138 1700.28,666.684 1705.11,660.469 1709.95,654.499 
  1714.78,648.781 1719.61,643.322 1724.45,638.128 1729.28,633.205 1734.11,628.559 1738.95,624.195 1743.78,620.119 1748.61,616.336 1753.45,612.851 1758.28,609.669 
  1763.11,606.794 1767.95,604.232 1772.78,601.987 1777.61,600.061 1782.44,598.46 1787.28,597.186 1792.11,596.244 1796.94,595.637 1801.78,595.369 1806.61,595.408 
  1811.44,595.644 1816.28,596.075 1821.11,596.701 1825.94,597.525 1830.78,598.547 1835.61,599.767 1840.44,601.186 1845.28,602.803 1850.11,604.619 1854.94,606.632 
  1859.78,608.843 1864.61,611.25 1869.44,613.851 1874.27,616.645 1879.11,619.632 1883.94,622.806 1888.77,626.168 1893.61,629.716 1898.44,633.444 1903.27,637.353 
  1908.11,641.437 1912.94,645.692 1917.77,650.115 1922.61,654.704 1927.44,659.455 1932.27,664.364 1937.11,669.424 1941.94,674.633 1946.77,679.987 1951.6,685.478 
  1956.44,691.103 1961.27,696.856 1966.1,702.731 1970.94,708.724 1975.77,714.827 1980.6,721.035 1985.44,727.343 1990.27,733.744 1995.1,740.232 1999.94,746.8 
  2004.77,753.443 2009.6,760.153 2014.44,766.925 2019.27,773.752 2024.1,780.626 2028.94,787.542 2033.77,794.492 2038.6,801.47 2043.43,808.468 2048.27,815.479 
  2053.1,822.497 2057.93,829.514 2062.77,836.525 2067.6,843.52 2072.43,850.494 2077.27,857.439 2082.1,864.348 2086.93,871.214 2091.77,878.031 2096.6,884.791 
  2101.43,891.488 2106.27,898.114 2111.1,904.663 2115.93,911.128 2120.77,917.503 2125.6,923.781 2130.43,929.955 2135.26,936.019 2140.1,941.966 2144.93,947.791 
  2149.76,953.487 2154.6,959.049 2159.43,964.47 2164.26,969.744 2169.1,974.866 2173.93,979.83 2178.76,984.631 2183.6,989.263 2188.43,993.721 2193.26,998 
  2198.1,1002.09 2202.93,1006 2207.76,1009.71 2212.59,1013.23 2217.43,1016.54 2222.26,1019.65 2227.09,1022.54 2231.93,1025.22 2236.76,1027.68 2241.59,1029.92 
  2246.43,1031.93 2251.26,1033.71 2256.09,1035.26 2260.93,1036.58 2265.76,1037.65 2270.59,1038.49 2275.43,1039.07 2280.26,1039.42 2285.09,1039.51 2289.93,1039.44 
  2294.76,1039.27 
  "/>
<polyline clip-path="url(#clip932)" style="stroke:#008000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" stroke-dasharray="4, 8" points="
  361.501,1365.41 366.334,1365.05 371.167,1364.23 376,1362.94 380.833,1361.18 385.666,1358.96 390.5,1356.27 395.333,1353.12 400.166,1349.51 404.999,1345.44 
  409.832,1340.91 414.665,1335.92 419.498,1330.49 424.332,1324.61 429.165,1318.29 433.998,1311.55 438.831,1304.37 443.664,1296.78 448.497,1288.78 453.33,1280.37 
  458.164,1271.58 462.997,1262.4 467.83,1252.85 472.663,1242.94 477.496,1232.67 482.329,1222.08 487.162,1211.15 491.996,1199.92 496.829,1188.38 501.662,1176.56 
  506.495,1164.46 511.328,1152.11 516.161,1139.52 520.994,1126.69 525.828,1113.66 530.661,1100.42 535.494,1087.01 540.327,1073.43 545.16,1059.7 549.993,1045.83 
  554.826,1031.85 559.66,1017.76 564.493,1003.59 569.326,989.341 574.159,975.039 578.992,960.698 583.825,946.333 588.658,931.959 593.492,917.592 598.325,903.248 
  603.158,888.94 607.991,874.686 612.824,860.497 617.657,846.39 622.49,832.378 627.324,818.474 632.157,804.691 636.99,791.043 641.823,777.541 646.656,764.199 
  651.489,751.029 656.322,738.041 661.156,725.248 665.989,712.659 670.822,700.286 675.655,688.138 680.488,676.225 685.321,664.555 690.154,653.139 694.988,641.983 
  699.821,631.098 704.654,620.49 709.487,610.167 714.32,600.136 719.153,590.405 723.986,580.98 728.82,571.867 733.653,563.072 738.486,554.601 743.319,546.459 
  748.152,538.65 752.985,531.179 757.818,524.049 762.652,517.267 767.485,510.836 772.318,504.763 777.151,499.05 781.984,493.703 786.817,488.723 791.65,484.115 
  796.484,479.881 801.317,476.023 806.15,472.543 810.983,469.442 815.816,466.723 820.649,464.384 825.482,462.427 830.316,460.851 835.149,459.656 839.982,458.841 
  844.815,458.404 849.648,458.345 854.481,458.66 859.314,459.347 864.148,460.404 868.981,461.827 873.814,463.613 878.647,465.76 883.48,468.264 888.313,471.121 
  893.147,474.325 897.98,477.871 902.813,481.753 907.646,485.966 912.479,490.504 917.312,495.36 922.145,500.527 926.979,505.999 931.812,511.768 936.645,517.827 
  941.478,524.168 946.311,530.783 951.144,537.664 955.977,544.802 960.811,552.187 965.644,559.812 970.477,567.669 975.31,575.75 980.143,584.044 984.976,592.543 
  989.809,601.237 994.643,610.117 999.476,619.174 1004.31,628.398 1009.14,637.78 1013.98,647.311 1018.81,656.982 1023.64,666.785 1028.47,676.71 1033.31,686.749 
  1038.14,696.892 1042.97,707.131 1047.81,717.457 1052.64,727.863 1057.47,738.338 1062.31,748.876 1067.14,759.466 1071.97,770.102 1076.81,780.774 1081.64,791.475 
  1086.47,802.197 1091.31,812.93 1096.14,823.668 1100.97,834.402 1105.8,845.124 1110.64,855.827 1115.47,866.501 1120.3,877.14 1125.14,887.735 1129.97,898.279 
  1134.8,908.762 1139.64,919.178 1144.47,929.518 1149.3,939.774 1154.14,949.937 1158.97,960 1163.8,969.954 1168.64,979.792 1173.47,989.504 1178.3,999.083 
  1183.14,1008.52 1187.97,1017.81 1192.8,1026.94 1197.63,1035.9 1202.47,1044.68 1207.3,1053.29 1212.13,1061.7 1216.97,1069.93 1221.8,1077.96 1226.63,1085.77 
  1231.47,1093.37 1236.3,1100.73 1241.13,1107.86 1245.97,1114.73 1250.8,1121.35 1255.63,1127.69 1260.47,1133.77 1265.3,1139.56 1270.13,1145.05 1274.96,1150.26 
  1279.8,1155.15 1284.63,1159.74 1289.46,1164.02 1294.3,1167.97 1299.13,1171.6 1303.96,1174.9 1308.8,1177.86 1313.63,1180.49 1318.46,1182.78 1323.3,1184.72 
  1328.13,1186.32 1332.96,1187.57 1337.8,1188.48 1342.63,1189.03 1347.46,1189.23 1352.3,1189.08 1357.13,1188.58 1361.96,1187.73 1366.79,1186.53 1371.63,1184.98 
  1376.46,1183.08 1381.29,1180.82 1386.13,1178.22 1390.96,1175.28 1395.79,1171.99 1400.63,1168.37 1405.46,1164.41 1410.29,1160.13 1415.13,1155.52 1419.96,1150.59 
  1424.79,1145.35 1429.63,1139.8 1434.46,1133.96 1439.29,1127.83 1444.12,1121.41 1448.96,1114.72 1453.79,1107.76 1458.62,1100.55 1463.46,1093.09 1468.29,1085.39 
  1473.12,1077.47 1477.96,1069.33 1482.79,1060.99 1487.62,1052.45 1492.46,1043.73 1497.29,1034.84 1502.12,1025.78 1506.96,1016.58 1511.79,1007.25 1516.62,997.79 
  1521.46,988.222 1526.29,978.555 1531.12,968.802 1535.95,958.975 1540.79,949.086 1545.62,939.146 1550.45,929.169 1555.29,919.165 1560.12,909.147 1564.95,899.126 
  1569.79,889.115 1574.62,879.124 1579.45,869.166 1584.29,859.251 1589.12,849.393 1593.95,839.601 1598.79,829.888 1603.62,820.266 1608.45,810.744 1613.28,801.335 
  1618.12,792.051 1622.95,782.898 1627.78,773.881 1632.62,765.01 1637.45,756.295 1642.28,747.742 1647.12,739.362 1651.95,731.161 1656.78,723.149 1661.62,715.331 
  1666.45,707.715 1671.28,700.308 1676.12,693.116 1680.95,686.145 1685.78,679.402 1690.62,672.891 1695.45,666.618 1700.28,660.587 1705.11,654.804 1709.95,649.272 
  1714.78,643.995 1719.61,638.977 1724.45,634.221 1729.28,629.73 1734.11,625.507 1738.95,621.553 1743.78,617.871 1748.61,614.462 1753.45,611.329 1758.28,608.473 
  1763.11,605.898 1767.95,603.604 1772.78,601.591 1777.61,599.862 1782.44,598.416 1787.28,597.253 1792.11,596.374 1796.94,595.777 1801.78,595.462 1806.61,595.427 
  1811.44,595.671 1816.28,596.192 1821.11,596.988 1825.94,598.056 1830.78,599.393 1835.61,600.996 1840.44,602.862 1845.28,604.986 1850.11,607.365 1854.94,609.993 
  1859.78,612.867 1864.61,615.98 1869.44,619.327 1874.27,622.903 1879.11,626.7 1883.94,630.714 1888.77,634.939 1893.61,639.373 1898.44,644.011 1903.27,648.844 
  1908.11,653.868 1912.94,659.076 1917.77,664.46 1922.61,670.016 1927.44,675.735 1932.27,681.612 1937.11,687.639 1941.94,693.809 1946.77,700.117 1951.6,706.554 
  1956.44,713.115 1961.27,719.791 1966.1,726.576 1970.94,733.462 1975.77,740.442 1980.6,747.509 1985.44,754.656 1990.27,761.875 1995.1,769.158 1999.94,776.499 
  2004.77,783.89 2009.6,791.326 2014.44,798.798 2019.27,806.299 2024.1,813.823 2028.94,821.361 2033.77,828.907 2038.6,836.453 2043.43,843.993 2048.27,851.519 
  2053.1,859.024 2057.93,866.502 2062.77,873.944 2067.6,881.345 2072.43,888.698 2077.27,895.995 2082.1,903.23 2086.93,910.396 2091.77,917.485 2096.6,924.493 
  2101.43,931.41 2106.27,938.232 2111.1,944.952 2115.93,951.562 2120.77,958.056 2125.6,964.428 2130.43,970.671 2135.26,976.779 2140.1,982.745 2144.93,988.566 
  2149.76,994.239 2154.6,999.754 2159.43,1005.11 2164.26,1010.29 2169.1,1015.29 2173.93,1020.12 2178.76,1024.75 2183.6,1029.19 2188.43,1033.44 2193.26,1037.48 
  2198.1,1041.31 2202.93,1044.92 2207.76,1048.32 2212.59,1051.5 2217.43,1054.46 2222.26,1057.18 2227.09,1059.68 2231.93,1061.95 2236.76,1063.97 2241.59,1065.76 
  2246.43,1067.31 2251.26,1068.62 2256.09,1069.69 2260.93,1070.51 2265.76,1071.09 2270.59,1071.43 2275.43,1071.51 2280.26,1071.36 2285.09,1070.96 2289.93,1070.31 
  2294.76,1069.42 
  "/>
<polyline clip-path="url(#clip932)" style="stroke:#008000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" points="
  361.501,1365.41 366.334,1364.77 371.167,1363.65 376,1362.06 380.833,1359.99 385.666,1357.46 390.5,1354.45 395.333,1350.98 400.166,1347.04 404.999,1342.65 
  409.832,1337.81 414.665,1332.53 419.498,1326.8 424.332,1320.65 429.165,1314.07 433.998,1307.07 438.831,1299.67 443.664,1291.87 448.497,1283.68 453.33,1275.11 
  458.164,1266.17 462.997,1256.88 467.83,1247.23 472.663,1237.26 477.496,1226.95 482.329,1216.34 487.162,1205.43 491.996,1194.23 496.829,1182.76 501.662,1171.03 
  506.495,1159.05 511.328,1146.84 516.161,1134.42 520.994,1121.79 525.828,1108.97 530.661,1095.97 535.494,1082.82 540.327,1069.52 545.16,1056.09 549.993,1042.55 
  554.826,1028.9 559.66,1015.17 564.493,1001.36 569.326,987.503 574.159,973.597 578.992,959.664 583.825,945.716 588.658,931.767 593.492,917.831 598.325,903.921 
  603.158,890.05 607.991,876.231 612.824,862.477 617.657,848.798 622.49,835.208 627.324,821.717 632.157,808.337 636.99,795.077 641.823,781.95 646.656,768.964 
  651.489,756.129 656.322,743.456 661.156,730.952 665.989,718.629 670.822,706.493 675.655,694.552 680.488,682.816 685.321,671.293 690.154,659.989 694.988,648.912 
  699.821,638.072 704.654,627.474 709.487,617.126 714.32,607.036 719.153,597.21 723.986,587.655 728.82,578.38 733.653,569.389 738.486,560.69 743.319,552.289 
  748.152,544.192 752.985,536.406 757.818,528.935 762.652,521.785 767.485,514.964 772.318,508.477 777.151,502.331 781.984,496.532 786.817,491.084 791.65,485.994 
  796.484,481.265 801.317,476.902 806.15,472.909 810.983,469.289 815.816,466.045 820.649,463.181 825.482,460.698 830.316,458.598 835.149,456.885 839.982,455.557 
  844.815,454.618 849.648,454.066 854.481,453.903 859.314,454.128 864.148,454.741 868.981,455.741 873.814,457.127 878.647,458.896 883.48,461.047 888.313,463.577 
  893.147,466.483 897.98,469.761 902.813,473.407 907.646,477.416 912.479,481.784 917.312,486.506 922.145,491.575 926.979,496.986 931.812,502.733 936.645,508.807 
  941.478,515.202 946.311,521.911 951.144,528.924 955.977,536.233 960.811,543.83 965.644,551.707 970.477,559.854 975.31,568.262 980.143,576.92 984.976,585.818 
  989.809,594.944 994.643,604.288 999.476,613.841 1004.31,623.589 1009.14,633.524 1013.98,643.633 1018.81,653.906 1023.64,664.332 1028.47,674.898 1033.31,685.594 
  1038.14,696.409 1042.97,707.33 1047.81,718.345 1052.64,729.444 1057.47,740.615 1062.31,751.844 1067.14,763.121 1071.97,774.432 1076.81,785.768 1081.64,797.118 
  1086.47,808.471 1091.31,819.817 1096.14,831.143 1100.97,842.441 1105.8,853.699 1110.64,864.907 1115.47,876.055 1120.3,887.132 1125.14,898.13 1129.97,909.037 
  1134.8,919.846 1139.64,930.546 1144.47,941.127 1149.3,951.582 1154.14,961.901 1158.97,972.075 1163.8,982.096 1168.64,991.955 1173.47,1001.64 1178.3,1011.16 
  1183.14,1020.48 1187.97,1029.61 1192.8,1038.54 1197.63,1047.27 1202.47,1055.79 1207.3,1064.08 1212.13,1072.14 1216.97,1079.97 1221.8,1087.55 1226.63,1094.88 
  1231.47,1101.95 1236.3,1108.75 1241.13,1115.28 1245.97,1121.53 1250.8,1127.49 1255.63,1133.17 1260.47,1138.55 1265.3,1143.63 1270.13,1148.41 1274.96,1152.88 
  1279.8,1157.03 1284.63,1160.88 1289.46,1164.4 1294.3,1167.6 1299.13,1170.48 1303.96,1173.03 1308.8,1175.25 1313.63,1177.14 1318.46,1178.71 1323.3,1179.94 
  1328.13,1180.83 1332.96,1181.4 1337.8,1181.63 1342.63,1181.53 1347.46,1181.09 1352.3,1180.32 1357.13,1179.22 1361.96,1177.79 1366.79,1176.04 1371.63,1173.95 
  1376.46,1171.55 1381.29,1168.83 1386.13,1165.8 1390.96,1162.45 1395.79,1158.8 1400.63,1154.85 1405.46,1150.6 1410.29,1146.07 1415.13,1141.25 1419.96,1136.15 
  1424.79,1130.79 1429.63,1125.16 1434.46,1119.27 1439.29,1113.14 1444.12,1106.78 1448.96,1100.18 1453.79,1093.36 1458.62,1086.32 1463.46,1079.08 1468.29,1071.64 
  1473.12,1064.02 1477.96,1056.22 1482.79,1048.26 1487.62,1040.13 1492.46,1031.86 1497.29,1023.45 1502.12,1014.92 1506.96,1006.26 1511.79,997.5 1516.62,988.639 
  1521.46,979.69 1526.29,970.662 1531.12,961.565 1535.95,952.41 1540.79,943.207 1545.62,933.965 1550.45,924.694 1555.29,915.405 1560.12,906.108 1564.95,896.812 
  1569.79,887.529 1574.62,878.267 1579.45,869.037 1584.29,859.844 1589.12,850.698 1593.95,841.606 1598.79,832.579 1603.62,823.623 1608.45,814.748 1613.28,805.96 
  1618.12,797.269 1622.95,788.681 1627.78,780.204 1632.62,771.845 1637.45,763.612 1642.28,755.511 1647.12,747.549 1651.95,739.733 1656.78,732.069 1661.62,724.563 
  1666.45,717.222 1671.28,710.051 1676.12,703.056 1680.95,696.242 1685.78,689.616 1690.62,683.182 1695.45,676.945 1700.28,670.91 1705.11,665.082 1709.95,659.463 
  1714.78,654.058 1719.61,648.871 1724.45,643.908 1729.28,639.172 1734.11,634.669 1738.95,630.403 1743.78,626.378 1748.61,622.596 1753.45,619.062 1758.28,615.779 
  1763.11,612.749 1767.95,609.975 1772.78,607.458 1777.61,605.201 1782.44,603.205 1787.28,601.472 1792.11,600.003 1796.94,598.798 1801.78,597.857 1806.61,597.181 
  1811.44,596.77 1816.28,596.623 1821.11,596.74 1825.94,597.118 1830.78,597.758 1835.61,598.658 1840.44,599.815 1845.28,601.227 1850.11,602.892 1854.94,604.808 
  1859.78,606.97 1864.61,609.379 1869.44,612.03 1874.27,614.92 1879.11,618.046 1883.94,621.401 1888.77,624.984 1893.61,628.788 1898.44,632.809 1903.27,637.042 
  1908.11,641.481 1912.94,646.122 1917.77,650.959 1922.61,655.985 1927.44,661.195 1932.27,666.582 1937.11,672.141 1941.94,677.864 1946.77,683.744 1951.6,689.774 
  1956.44,695.948 1961.27,702.257 1966.1,708.694 1970.94,715.251 1975.77,721.919 1980.6,728.692 1985.44,735.56 1990.27,742.52 1995.1,749.561 1999.94,756.677 
  2004.77,763.858 2009.6,771.098 2014.44,778.388 2019.27,785.719 2024.1,793.085 2028.94,800.477 2033.77,807.887 2038.6,815.309 2043.43,822.734 2048.27,830.155 
  2053.1,837.564 2057.93,844.953 2062.77,852.317 2067.6,859.646 2072.43,866.935 2077.27,874.175 2082.1,881.36 2086.93,888.482 2091.77,895.535 2096.6,902.512 
  2101.43,909.405 2106.27,916.208 2111.1,922.915 2115.93,929.518 2120.77,936.012 2125.6,942.394 2130.43,948.658 2135.26,954.796 2140.1,960.803 2144.93,966.673 
  2149.76,972.4 2154.6,977.978 2159.43,983.402 2164.26,988.667 2169.1,993.767 2173.93,998.699 2178.76,1003.46 2183.6,1008.04 2188.43,1012.44 2193.26,1016.65 
  2198.1,1020.68 2202.93,1024.51 2207.76,1028.15 2212.59,1031.59 2217.43,1034.83 2222.26,1037.86 2227.09,1040.69 2231.93,1043.31 2236.76,1045.72 2241.59,1047.92 
  2246.43,1049.91 2251.26,1051.68 2256.09,1053.23 2260.93,1054.58 2265.76,1055.7 2270.59,1056.6 2275.43,1057.29 2280.26,1057.76 2285.09,1058.01 2289.93,1058.04 
  2294.76,1057.86 
  "/>
<path clip-path="url(#clip930)" d="
M1421.89 554.876 L2284.45 554.876 L2284.45 166.076 L1421.89 166.076  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<polyline clip-path="url(#clip930)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1421.89,554.876 2284.45,554.876 2284.45,166.076 1421.89,166.076 1421.89,554.876 
  "/>
<polyline clip-path="url(#clip930)" style="stroke:#009af9; stroke-linecap:butt; stroke-linejoin:round; stroke-width:6; stroke-opacity:1; fill:none" points="
  1444.66,243.836 1581.28,243.836 
  "/>
<path clip-path="url(#clip930)" d="M1637.42 219.617 L1637.42 226.457 Q1633.43 224.548 1629.88 223.61 Q1626.34 222.673 1623.04 222.673 Q1617.31 222.673 1614.19 224.895 Q1611.1 227.117 1611.1 231.214 Q1611.1 234.652 1613.15 236.423 Q1615.23 238.159 1621 239.235 L1625.23 240.103 Q1633.08 241.596 1636.79 245.381 Q1640.54 249.131 1640.54 255.45 Q1640.54 262.985 1635.47 266.874 Q1630.44 270.763 1620.68 270.763 Q1617 270.763 1612.84 269.929 Q1608.7 269.096 1604.26 267.464 L1604.26 260.242 Q1608.53 262.638 1612.63 263.853 Q1616.72 265.068 1620.68 265.068 Q1626.69 265.068 1629.95 262.707 Q1633.22 260.346 1633.22 255.971 Q1633.22 252.152 1630.86 249.999 Q1628.53 247.846 1623.18 246.77 L1618.91 245.936 Q1611.06 244.374 1607.56 241.041 Q1604.05 237.707 1604.05 231.77 Q1604.05 224.895 1608.88 220.937 Q1613.74 216.978 1622.25 216.978 Q1625.89 216.978 1629.68 217.638 Q1633.46 218.298 1637.42 219.617 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1651.2 230.867 L1657.59 230.867 L1657.59 269.756 L1651.2 269.756 L1651.2 230.867 M1651.2 215.728 L1657.59 215.728 L1657.59 223.818 L1651.2 223.818 L1651.2 215.728 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1701.24 238.332 Q1703.63 234.027 1706.97 231.978 Q1710.3 229.93 1714.81 229.93 Q1720.89 229.93 1724.19 234.2 Q1727.49 238.436 1727.49 246.284 L1727.49 269.756 L1721.06 269.756 L1721.06 246.492 Q1721.06 240.902 1719.08 238.193 Q1717.11 235.485 1713.04 235.485 Q1708.08 235.485 1705.2 238.784 Q1702.31 242.082 1702.31 247.777 L1702.31 269.756 L1695.89 269.756 L1695.89 246.492 Q1695.89 240.867 1693.91 238.193 Q1691.93 235.485 1687.8 235.485 Q1682.9 235.485 1680.02 238.818 Q1677.14 242.117 1677.14 247.777 L1677.14 269.756 L1670.72 269.756 L1670.72 230.867 L1677.14 230.867 L1677.14 236.909 Q1679.33 233.332 1682.38 231.631 Q1685.44 229.93 1689.64 229.93 Q1693.88 229.93 1696.83 232.082 Q1699.81 234.235 1701.24 238.332 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1746.41 263.922 L1746.41 284.547 L1739.99 284.547 L1739.99 230.867 L1746.41 230.867 L1746.41 236.77 Q1748.43 233.298 1751.48 231.631 Q1754.57 229.93 1758.84 229.93 Q1765.92 229.93 1770.33 235.555 Q1774.78 241.179 1774.78 250.346 Q1774.78 259.513 1770.33 265.138 Q1765.92 270.763 1758.84 270.763 Q1754.57 270.763 1751.48 269.096 Q1748.43 267.395 1746.41 263.922 M1768.15 250.346 Q1768.15 243.298 1765.23 239.304 Q1762.35 235.277 1757.28 235.277 Q1752.21 235.277 1749.29 239.304 Q1746.41 243.298 1746.41 250.346 Q1746.41 257.395 1749.29 261.422 Q1752.21 265.415 1757.28 265.415 Q1762.35 265.415 1765.23 261.422 Q1768.15 257.395 1768.15 250.346 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1785.37 215.728 L1791.76 215.728 L1791.76 269.756 L1785.37 269.756 L1785.37 215.728 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1838.39 248.714 L1838.39 251.839 L1809.01 251.839 Q1809.43 258.436 1812.97 261.909 Q1816.55 265.346 1822.9 265.346 Q1826.58 265.346 1830.02 264.443 Q1833.49 263.54 1836.9 261.735 L1836.9 267.777 Q1833.46 269.235 1829.85 269.999 Q1826.24 270.763 1822.52 270.763 Q1813.22 270.763 1807.76 265.346 Q1802.35 259.929 1802.35 250.693 Q1802.35 241.145 1807.49 235.555 Q1812.66 229.93 1821.41 229.93 Q1829.26 229.93 1833.81 234.999 Q1838.39 240.034 1838.39 248.714 M1832 246.839 Q1831.93 241.596 1829.05 238.471 Q1826.2 235.346 1821.48 235.346 Q1816.13 235.346 1812.9 238.367 Q1809.71 241.388 1809.22 246.874 L1832 246.839 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1849.15 217.916 L1878.95 217.916 L1878.95 223.818 L1856.17 223.818 L1856.17 239.096 L1876.72 239.096 L1876.72 244.999 L1856.17 244.999 L1856.17 269.756 L1849.15 269.756 L1849.15 217.916 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1890.06 217.916 L1900.51 217.916 L1913.74 253.193 L1927.04 217.916 L1937.49 217.916 L1937.49 269.756 L1930.65 269.756 L1930.65 224.235 L1917.28 259.79 L1910.23 259.79 L1896.86 224.235 L1896.86 269.756 L1890.06 269.756 L1890.06 217.916 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1950.61 217.916 L1957.66 217.916 L1957.66 249.409 Q1957.66 257.742 1960.68 261.422 Q1963.7 265.068 1970.47 265.068 Q1977.21 265.068 1980.23 261.422 Q1983.25 257.742 1983.25 249.409 L1983.25 217.916 L1990.3 217.916 L1990.3 250.277 Q1990.3 260.415 1985.26 265.589 Q1980.26 270.763 1970.47 270.763 Q1960.65 270.763 1955.61 265.589 Q1950.61 260.415 1950.61 250.277 L1950.61 217.916 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip930)" style="stroke:#e26f46; stroke-linecap:butt; stroke-linejoin:round; stroke-width:6; stroke-opacity:1; fill:none" points="
  1444.66,321.596 1581.28,321.596 
  "/>
<path clip-path="url(#clip930)" d="M1628.63 323.21 Q1630.89 323.974 1633.01 326.474 Q1635.16 328.974 1637.31 333.349 L1644.43 347.516 L1636.9 347.516 L1630.27 334.217 Q1627.7 329.009 1625.27 327.307 Q1622.87 325.606 1618.7 325.606 L1611.06 325.606 L1611.06 347.516 L1604.05 347.516 L1604.05 295.676 L1619.88 295.676 Q1628.77 295.676 1633.15 299.391 Q1637.52 303.106 1637.52 310.606 Q1637.52 315.502 1635.23 318.731 Q1632.97 321.96 1628.63 323.21 M1611.06 301.44 L1611.06 319.842 L1619.88 319.842 Q1624.95 319.842 1627.52 317.516 Q1630.13 315.155 1630.13 310.606 Q1630.13 306.058 1627.52 303.766 Q1624.95 301.44 1619.88 301.44 L1611.06 301.44 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1683.25 326.474 L1683.25 329.599 L1653.88 329.599 Q1654.29 336.196 1657.84 339.669 Q1661.41 343.106 1667.77 343.106 Q1671.45 343.106 1674.88 342.203 Q1678.36 341.3 1681.76 339.495 L1681.76 345.537 Q1678.32 346.995 1674.71 347.759 Q1671.1 348.523 1667.38 348.523 Q1658.08 348.523 1652.63 343.106 Q1647.21 337.689 1647.21 328.453 Q1647.21 318.905 1652.35 313.315 Q1657.52 307.69 1666.27 307.69 Q1674.12 307.69 1678.67 312.759 Q1683.25 317.794 1683.25 326.474 M1676.86 324.599 Q1676.79 319.356 1673.91 316.231 Q1671.06 313.106 1666.34 313.106 Q1660.99 313.106 1657.77 316.127 Q1654.57 319.148 1654.09 324.634 L1676.86 324.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1713.43 293.488 L1713.43 298.801 L1707.31 298.801 Q1703.88 298.801 1702.52 300.19 Q1701.2 301.578 1701.2 305.19 L1701.2 308.627 L1711.72 308.627 L1711.72 313.592 L1701.2 313.592 L1701.2 347.516 L1694.78 347.516 L1694.78 313.592 L1688.67 313.592 L1688.67 308.627 L1694.78 308.627 L1694.78 305.919 Q1694.78 299.426 1697.8 296.474 Q1700.82 293.488 1707.38 293.488 L1713.43 293.488 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1752.04 326.474 L1752.04 329.599 L1722.66 329.599 Q1723.08 336.196 1726.62 339.669 Q1730.2 343.106 1736.55 343.106 Q1740.23 343.106 1743.67 342.203 Q1747.14 341.3 1750.54 339.495 L1750.54 345.537 Q1747.11 346.995 1743.49 347.759 Q1739.88 348.523 1736.17 348.523 Q1726.86 348.523 1721.41 343.106 Q1715.99 337.689 1715.99 328.453 Q1715.99 318.905 1721.13 313.315 Q1726.31 307.69 1735.06 307.69 Q1742.9 307.69 1747.45 312.759 Q1752.04 317.794 1752.04 326.474 M1745.65 324.599 Q1745.58 319.356 1742.7 316.231 Q1739.85 313.106 1735.13 313.106 Q1729.78 313.106 1726.55 316.127 Q1723.36 319.148 1722.87 324.634 L1745.65 324.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1785.06 314.599 Q1783.98 313.974 1782.7 313.696 Q1781.45 313.384 1779.92 313.384 Q1774.5 313.384 1771.58 316.926 Q1768.7 320.433 1768.7 327.03 L1768.7 347.516 L1762.28 347.516 L1762.28 308.627 L1768.7 308.627 L1768.7 314.669 Q1770.72 311.127 1773.95 309.426 Q1777.17 307.69 1781.79 307.69 Q1782.45 307.69 1783.25 307.794 Q1784.05 307.863 1785.02 308.037 L1785.06 314.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1823.46 326.474 L1823.46 329.599 L1794.08 329.599 Q1794.5 336.196 1798.04 339.669 Q1801.62 343.106 1807.97 343.106 Q1811.65 343.106 1815.09 342.203 Q1818.56 341.3 1821.97 339.495 L1821.97 345.537 Q1818.53 346.995 1814.92 347.759 Q1811.31 348.523 1807.59 348.523 Q1798.29 348.523 1792.83 343.106 Q1787.42 337.689 1787.42 328.453 Q1787.42 318.905 1792.56 313.315 Q1797.73 307.69 1806.48 307.69 Q1814.33 307.69 1818.88 312.759 Q1823.46 317.794 1823.46 326.474 M1817.07 324.599 Q1817 319.356 1814.12 316.231 Q1811.27 313.106 1806.55 313.106 Q1801.2 313.106 1797.97 316.127 Q1794.78 319.148 1794.29 324.634 L1817.07 324.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1866.27 324.044 L1866.27 347.516 L1859.88 347.516 L1859.88 324.252 Q1859.88 318.731 1857.73 315.988 Q1855.58 313.245 1851.27 313.245 Q1846.1 313.245 1843.11 316.544 Q1840.13 319.842 1840.13 325.537 L1840.13 347.516 L1833.7 347.516 L1833.7 308.627 L1840.13 308.627 L1840.13 314.669 Q1842.42 311.162 1845.51 309.426 Q1848.63 307.69 1852.7 307.69 Q1859.4 307.69 1862.83 311.856 Q1866.27 315.988 1866.27 324.044 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1907 310.12 L1907 316.092 Q1904.29 314.599 1901.55 313.87 Q1898.84 313.106 1896.06 313.106 Q1889.85 313.106 1886.41 317.064 Q1882.97 320.988 1882.97 328.106 Q1882.97 335.224 1886.41 339.182 Q1889.85 343.106 1896.06 343.106 Q1898.84 343.106 1901.55 342.377 Q1904.29 341.613 1907 340.12 L1907 346.023 Q1904.33 347.273 1901.45 347.898 Q1898.6 348.523 1895.37 348.523 Q1886.58 348.523 1881.41 343.002 Q1876.24 337.481 1876.24 328.106 Q1876.24 318.592 1881.45 313.141 Q1886.69 307.69 1895.79 307.69 Q1898.74 307.69 1901.55 308.315 Q1904.36 308.905 1907 310.12 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1951.38 326.474 L1951.38 329.599 L1922 329.599 Q1922.42 336.196 1925.96 339.669 Q1929.54 343.106 1935.89 343.106 Q1939.57 343.106 1943.01 342.203 Q1946.48 341.3 1949.88 339.495 L1949.88 345.537 Q1946.44 346.995 1942.83 347.759 Q1939.22 348.523 1935.51 348.523 Q1926.2 348.523 1920.75 343.106 Q1915.33 337.689 1915.33 328.453 Q1915.33 318.905 1920.47 313.315 Q1925.65 307.69 1934.4 307.69 Q1942.24 307.69 1946.79 312.759 Q1951.38 317.794 1951.38 326.474 M1944.99 324.599 Q1944.92 319.356 1942.04 316.231 Q1939.19 313.106 1934.47 313.106 Q1929.12 313.106 1925.89 316.127 Q1922.69 319.148 1922.21 324.634 L1944.99 324.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip930)" style="stroke:#008000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:6; stroke-opacity:1; fill:none" stroke-dasharray="3, 6" points="
  1444.66,399.356 1581.28,399.356 
  "/>
<path clip-path="url(#clip930)" d="M1604.05 373.436 L1613.5 373.436 L1636.48 416.804 L1636.48 373.436 L1643.29 373.436 L1643.29 425.276 L1633.84 425.276 L1610.86 381.908 L1610.86 425.276 L1604.05 425.276 L1604.05 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1690.23 404.234 L1690.23 407.359 L1660.86 407.359 Q1661.27 413.956 1664.81 417.429 Q1668.39 420.866 1674.74 420.866 Q1678.43 420.866 1681.86 419.963 Q1685.34 419.06 1688.74 417.255 L1688.74 423.297 Q1685.3 424.755 1681.69 425.519 Q1678.08 426.283 1674.36 426.283 Q1665.06 426.283 1659.61 420.866 Q1654.19 415.449 1654.19 406.213 Q1654.19 396.665 1659.33 391.075 Q1664.5 385.45 1673.25 385.45 Q1681.1 385.45 1685.65 390.519 Q1690.23 395.554 1690.23 404.234 M1683.84 402.359 Q1683.77 397.116 1680.89 393.991 Q1678.04 390.866 1673.32 390.866 Q1667.97 390.866 1664.74 393.887 Q1661.55 396.908 1661.06 402.394 L1683.84 402.359 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1700.06 409.929 L1700.06 386.387 L1706.45 386.387 L1706.45 409.686 Q1706.45 415.206 1708.6 417.984 Q1710.75 420.727 1715.06 420.727 Q1720.23 420.727 1723.22 417.429 Q1726.24 414.13 1726.24 408.436 L1726.24 386.387 L1732.63 386.387 L1732.63 425.276 L1726.24 425.276 L1726.24 419.304 Q1723.91 422.845 1720.82 424.581 Q1717.77 426.283 1713.7 426.283 Q1707 426.283 1703.53 422.116 Q1700.06 417.949 1700.06 409.929 M1716.13 385.45 L1716.13 385.45 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1768.32 392.359 Q1767.24 391.734 1765.96 391.456 Q1764.71 391.144 1763.18 391.144 Q1757.77 391.144 1754.85 394.686 Q1751.97 398.193 1751.97 404.79 L1751.97 425.276 L1745.54 425.276 L1745.54 386.387 L1751.97 386.387 L1751.97 392.429 Q1753.98 388.887 1757.21 387.186 Q1760.44 385.45 1765.06 385.45 Q1765.72 385.45 1766.52 385.554 Q1767.31 385.623 1768.29 385.797 L1768.32 392.359 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1792.7 405.727 Q1784.95 405.727 1781.97 407.498 Q1778.98 409.269 1778.98 413.54 Q1778.98 416.942 1781.2 418.956 Q1783.46 420.935 1787.31 420.935 Q1792.63 420.935 1795.82 417.185 Q1799.05 413.401 1799.05 407.151 L1799.05 405.727 L1792.7 405.727 M1805.44 403.088 L1805.44 425.276 L1799.05 425.276 L1799.05 419.373 Q1796.86 422.915 1793.6 424.616 Q1790.33 426.283 1785.61 426.283 Q1779.64 426.283 1776.1 422.949 Q1772.59 419.581 1772.59 413.956 Q1772.59 407.394 1776.97 404.061 Q1781.38 400.727 1790.09 400.727 L1799.05 400.727 L1799.05 400.102 Q1799.05 395.693 1796.13 393.297 Q1793.25 390.866 1788.01 390.866 Q1784.67 390.866 1781.52 391.665 Q1778.36 392.463 1775.44 394.061 L1775.44 388.158 Q1778.95 386.804 1782.24 386.144 Q1785.54 385.45 1788.67 385.45 Q1797.11 385.45 1801.27 389.825 Q1805.44 394.199 1805.44 403.088 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1818.6 371.248 L1824.99 371.248 L1824.99 425.276 L1818.6 425.276 L1818.6 371.248 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1838.63 373.436 L1868.42 373.436 L1868.42 379.338 L1845.65 379.338 L1845.65 394.616 L1866.2 394.616 L1866.2 400.519 L1845.65 400.519 L1845.65 425.276 L1838.63 425.276 L1838.63 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1879.54 373.436 L1889.99 373.436 L1903.22 408.713 L1916.51 373.436 L1926.97 373.436 L1926.97 425.276 L1920.13 425.276 L1920.13 379.755 L1906.76 415.31 L1899.71 415.31 L1886.34 379.755 L1886.34 425.276 L1879.54 425.276 L1879.54 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1940.09 373.436 L1947.14 373.436 L1947.14 404.929 Q1947.14 413.262 1950.16 416.942 Q1953.18 420.588 1959.95 420.588 Q1966.69 420.588 1969.71 416.942 Q1972.73 413.262 1972.73 404.929 L1972.73 373.436 L1979.78 373.436 L1979.78 405.797 Q1979.78 415.935 1974.74 421.109 Q1969.74 426.283 1959.95 426.283 Q1950.13 426.283 1945.09 421.109 Q1940.09 415.935 1940.09 405.797 L1940.09 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M2030.61 371.318 Q2025.96 379.304 2023.7 387.116 Q2021.44 394.929 2021.44 402.949 Q2021.44 410.97 2023.7 418.852 Q2025.99 426.699 2030.61 434.651 L2025.06 434.651 Q2019.85 426.491 2017.24 418.609 Q2014.67 410.727 2014.67 402.949 Q2014.67 395.206 2017.24 387.359 Q2019.81 379.512 2025.06 371.318 L2030.61 371.318 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M2049.95 419.373 L2074.43 419.373 L2074.43 425.276 L2041.51 425.276 L2041.51 419.373 Q2045.51 415.241 2052.38 408.297 Q2059.29 401.318 2061.06 399.304 Q2064.43 395.519 2065.75 392.915 Q2067.1 390.276 2067.1 387.741 Q2067.1 383.609 2064.19 381.005 Q2061.31 378.401 2056.65 378.401 Q2053.35 378.401 2049.67 379.547 Q2046.03 380.693 2041.86 383.019 L2041.86 375.936 Q2046.1 374.234 2049.78 373.366 Q2053.46 372.498 2056.51 372.498 Q2064.57 372.498 2069.36 376.526 Q2074.15 380.554 2074.15 387.29 Q2074.15 390.484 2072.94 393.366 Q2071.76 396.213 2068.6 400.102 Q2067.73 401.109 2063.08 405.936 Q2058.42 410.727 2049.95 419.373 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M2089.22 373.436 L2116.76 373.436 L2116.76 379.338 L2095.65 379.338 L2095.65 392.047 Q2097.17 391.526 2098.7 391.283 Q2100.23 391.005 2101.76 391.005 Q2110.44 391.005 2115.51 395.762 Q2120.58 400.519 2120.58 408.644 Q2120.58 417.012 2115.37 421.665 Q2110.16 426.283 2100.68 426.283 Q2097.42 426.283 2094.01 425.727 Q2090.65 425.172 2087.03 424.06 L2087.03 417.012 Q2090.16 418.713 2093.49 419.547 Q2096.83 420.38 2100.54 420.38 Q2106.55 420.38 2110.06 417.22 Q2113.56 414.06 2113.56 408.644 Q2113.56 403.227 2110.06 400.068 Q2106.55 396.908 2100.54 396.908 Q2097.73 396.908 2094.92 397.533 Q2092.14 398.158 2089.22 399.477 L2089.22 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M2149.4 378.054 Q2143.98 378.054 2141.24 383.401 Q2138.53 388.713 2138.53 399.408 Q2138.53 410.067 2141.24 415.415 Q2143.98 420.727 2149.4 420.727 Q2154.85 420.727 2157.55 415.415 Q2160.3 410.067 2160.3 399.408 Q2160.3 388.713 2157.55 383.401 Q2154.85 378.054 2149.4 378.054 M2149.4 372.498 Q2158.11 372.498 2162.69 379.408 Q2167.31 386.283 2167.31 399.408 Q2167.31 412.498 2162.69 419.408 Q2158.11 426.283 2149.4 426.283 Q2140.68 426.283 2136.06 419.408 Q2131.48 412.498 2131.48 399.408 Q2131.48 386.283 2136.06 379.408 Q2140.68 372.498 2149.4 372.498 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M2194.64 378.054 Q2189.22 378.054 2186.48 383.401 Q2183.77 388.713 2183.77 399.408 Q2183.77 410.067 2186.48 415.415 Q2189.22 420.727 2194.64 420.727 Q2200.09 420.727 2202.8 415.415 Q2205.54 410.067 2205.54 399.408 Q2205.54 388.713 2202.8 383.401 Q2200.09 378.054 2194.64 378.054 M2194.64 372.498 Q2203.35 372.498 2207.94 379.408 Q2212.55 386.283 2212.55 399.408 Q2212.55 412.498 2207.94 419.408 Q2203.35 426.283 2194.64 426.283 Q2185.92 426.283 2181.3 419.408 Q2176.72 412.498 2176.72 399.408 Q2176.72 386.283 2181.3 379.408 Q2185.92 372.498 2194.64 372.498 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M2222.97 371.318 L2228.53 371.318 Q2233.73 379.512 2236.3 387.359 Q2238.91 395.206 2238.91 402.949 Q2238.91 410.727 2236.3 418.609 Q2233.73 426.491 2228.53 434.651 L2222.97 434.651 Q2227.59 426.699 2229.85 418.852 Q2232.14 410.97 2232.14 402.949 Q2232.14 394.929 2229.85 387.116 Q2227.59 379.304 2222.97 371.318 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip930)" style="stroke:#008000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:6; stroke-opacity:1; fill:none" points="
  1444.66,477.116 1581.28,477.116 
  "/>
<path clip-path="url(#clip930)" d="M1604.05 451.196 L1613.5 451.196 L1636.48 494.564 L1636.48 451.196 L1643.29 451.196 L1643.29 503.036 L1633.84 503.036 L1610.86 459.668 L1610.86 503.036 L1604.05 503.036 L1604.05 451.196 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1690.23 481.994 L1690.23 485.119 L1660.86 485.119 Q1661.27 491.716 1664.81 495.189 Q1668.39 498.626 1674.74 498.626 Q1678.43 498.626 1681.86 497.723 Q1685.34 496.82 1688.74 495.015 L1688.74 501.057 Q1685.3 502.515 1681.69 503.279 Q1678.08 504.043 1674.36 504.043 Q1665.06 504.043 1659.61 498.626 Q1654.19 493.209 1654.19 483.973 Q1654.19 474.425 1659.33 468.835 Q1664.5 463.21 1673.25 463.21 Q1681.1 463.21 1685.65 468.279 Q1690.23 473.314 1690.23 481.994 M1683.84 480.119 Q1683.77 474.876 1680.89 471.751 Q1678.04 468.626 1673.32 468.626 Q1667.97 468.626 1664.74 471.647 Q1661.55 474.668 1661.06 480.154 L1683.84 480.119 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1700.06 487.689 L1700.06 464.147 L1706.45 464.147 L1706.45 487.446 Q1706.45 492.966 1708.6 495.744 Q1710.75 498.487 1715.06 498.487 Q1720.23 498.487 1723.22 495.189 Q1726.24 491.89 1726.24 486.196 L1726.24 464.147 L1732.63 464.147 L1732.63 503.036 L1726.24 503.036 L1726.24 497.064 Q1723.91 500.605 1720.82 502.341 Q1717.77 504.043 1713.7 504.043 Q1707 504.043 1703.53 499.876 Q1700.06 495.709 1700.06 487.689 M1716.13 463.21 L1716.13 463.21 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1768.32 470.119 Q1767.24 469.494 1765.96 469.216 Q1764.71 468.904 1763.18 468.904 Q1757.77 468.904 1754.85 472.446 Q1751.97 475.953 1751.97 482.55 L1751.97 503.036 L1745.54 503.036 L1745.54 464.147 L1751.97 464.147 L1751.97 470.189 Q1753.98 466.647 1757.21 464.946 Q1760.44 463.21 1765.06 463.21 Q1765.72 463.21 1766.52 463.314 Q1767.31 463.383 1768.29 463.557 L1768.32 470.119 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1792.7 483.487 Q1784.95 483.487 1781.97 485.258 Q1778.98 487.029 1778.98 491.3 Q1778.98 494.702 1781.2 496.716 Q1783.46 498.695 1787.31 498.695 Q1792.63 498.695 1795.82 494.945 Q1799.05 491.161 1799.05 484.911 L1799.05 483.487 L1792.7 483.487 M1805.44 480.848 L1805.44 503.036 L1799.05 503.036 L1799.05 497.133 Q1796.86 500.675 1793.6 502.376 Q1790.33 504.043 1785.61 504.043 Q1779.64 504.043 1776.1 500.709 Q1772.59 497.341 1772.59 491.716 Q1772.59 485.154 1776.97 481.821 Q1781.38 478.487 1790.09 478.487 L1799.05 478.487 L1799.05 477.862 Q1799.05 473.453 1796.13 471.057 Q1793.25 468.626 1788.01 468.626 Q1784.67 468.626 1781.52 469.425 Q1778.36 470.223 1775.44 471.821 L1775.44 465.918 Q1778.95 464.564 1782.24 463.904 Q1785.54 463.21 1788.67 463.21 Q1797.11 463.21 1801.27 467.585 Q1805.44 471.959 1805.44 480.848 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1818.6 449.008 L1824.99 449.008 L1824.99 503.036 L1818.6 503.036 L1818.6 449.008 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1838.63 451.196 L1868.42 451.196 L1868.42 457.098 L1845.65 457.098 L1845.65 472.376 L1866.2 472.376 L1866.2 478.279 L1845.65 478.279 L1845.65 503.036 L1838.63 503.036 L1838.63 451.196 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1879.54 451.196 L1889.99 451.196 L1903.22 486.473 L1916.51 451.196 L1926.97 451.196 L1926.97 503.036 L1920.13 503.036 L1920.13 457.515 L1906.76 493.07 L1899.71 493.07 L1886.34 457.515 L1886.34 503.036 L1879.54 503.036 L1879.54 451.196 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M1940.09 451.196 L1947.14 451.196 L1947.14 482.689 Q1947.14 491.022 1950.16 494.702 Q1953.18 498.348 1959.95 498.348 Q1966.69 498.348 1969.71 494.702 Q1972.73 491.022 1972.73 482.689 L1972.73 451.196 L1979.78 451.196 L1979.78 483.557 Q1979.78 493.695 1974.74 498.869 Q1969.74 504.043 1959.95 504.043 Q1950.13 504.043 1945.09 498.869 Q1940.09 493.695 1940.09 483.557 L1940.09 451.196 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M2030.61 449.078 Q2025.96 457.064 2023.7 464.876 Q2021.44 472.689 2021.44 480.709 Q2021.44 488.73 2023.7 496.612 Q2025.99 504.459 2030.61 512.411 L2025.06 512.411 Q2019.85 504.251 2017.24 496.369 Q2014.67 488.487 2014.67 480.709 Q2014.67 472.966 2017.24 465.119 Q2019.81 457.272 2025.06 449.078 L2030.61 449.078 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M2043.98 451.196 L2071.51 451.196 L2071.51 457.098 L2050.4 457.098 L2050.4 469.807 Q2051.93 469.286 2053.46 469.043 Q2054.99 468.765 2056.51 468.765 Q2065.19 468.765 2070.26 473.522 Q2075.33 478.279 2075.33 486.404 Q2075.33 494.772 2070.12 499.425 Q2064.92 504.043 2055.44 504.043 Q2052.17 504.043 2048.77 503.487 Q2045.4 502.932 2041.79 501.82 L2041.79 494.772 Q2044.92 496.473 2048.25 497.307 Q2051.58 498.14 2055.3 498.14 Q2061.31 498.14 2064.81 494.98 Q2068.32 491.82 2068.32 486.404 Q2068.32 480.987 2064.81 477.828 Q2061.31 474.668 2055.3 474.668 Q2052.49 474.668 2049.67 475.293 Q2046.9 475.918 2043.98 477.237 L2043.98 451.196 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M2104.15 455.814 Q2098.74 455.814 2095.99 461.161 Q2093.28 466.473 2093.28 477.168 Q2093.28 487.827 2095.99 493.175 Q2098.74 498.487 2104.15 498.487 Q2109.6 498.487 2112.31 493.175 Q2115.06 487.827 2115.06 477.168 Q2115.06 466.473 2112.31 461.161 Q2109.6 455.814 2104.15 455.814 M2104.15 450.258 Q2112.87 450.258 2117.45 457.168 Q2122.07 464.043 2122.07 477.168 Q2122.07 490.258 2117.45 497.168 Q2112.87 504.043 2104.15 504.043 Q2095.44 504.043 2090.82 497.168 Q2086.24 490.258 2086.24 477.168 Q2086.24 464.043 2090.82 457.168 Q2095.44 450.258 2104.15 450.258 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M2149.4 455.814 Q2143.98 455.814 2141.24 461.161 Q2138.53 466.473 2138.53 477.168 Q2138.53 487.827 2141.24 493.175 Q2143.98 498.487 2149.4 498.487 Q2154.85 498.487 2157.55 493.175 Q2160.3 487.827 2160.3 477.168 Q2160.3 466.473 2157.55 461.161 Q2154.85 455.814 2149.4 455.814 M2149.4 450.258 Q2158.11 450.258 2162.69 457.168 Q2167.31 464.043 2167.31 477.168 Q2167.31 490.258 2162.69 497.168 Q2158.11 504.043 2149.4 504.043 Q2140.68 504.043 2136.06 497.168 Q2131.48 490.258 2131.48 477.168 Q2131.48 464.043 2136.06 457.168 Q2140.68 450.258 2149.4 450.258 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M2194.64 455.814 Q2189.22 455.814 2186.48 461.161 Q2183.77 466.473 2183.77 477.168 Q2183.77 487.827 2186.48 493.175 Q2189.22 498.487 2194.64 498.487 Q2200.09 498.487 2202.8 493.175 Q2205.54 487.827 2205.54 477.168 Q2205.54 466.473 2202.8 461.161 Q2200.09 455.814 2194.64 455.814 M2194.64 450.258 Q2203.35 450.258 2207.94 457.168 Q2212.55 464.043 2212.55 477.168 Q2212.55 490.258 2207.94 497.168 Q2203.35 504.043 2194.64 504.043 Q2185.92 504.043 2181.3 497.168 Q2176.72 490.258 2176.72 477.168 Q2176.72 464.043 2181.3 457.168 Q2185.92 450.258 2194.64 450.258 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip930)" d="M2222.97 449.078 L2228.53 449.078 Q2233.73 457.272 2236.3 465.119 Q2238.91 472.966 2238.91 480.709 Q2238.91 488.487 2236.3 496.369 Q2233.73 504.251 2228.53 512.411 L2222.97 512.411 Q2227.59 504.459 2229.85 496.612 Q2232.14 488.73 2232.14 480.709 Q2232.14 472.689 2229.85 464.876 Q2227.59 457.064 2222.97 449.078 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /></svg>




<?xml version="1.0" encoding="utf-8"?>
<svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" width="600" height="400" viewBox="0 0 2400 1600">
<defs>
  <clipPath id="clip020">
    <rect x="0" y="0" width="2400" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip020)" d="
M0 1600 L2400 1600 L2400 0 L0 0  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip021">
    <rect x="480" y="0" width="1681" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip020)" d="
M303.503 1401.58 L2352.76 1401.58 L2352.76 123.472 L303.503 123.472  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip022">
    <rect x="303" y="123" width="2050" height="1279"/>
  </clipPath>
</defs>
<polyline clip-path="url(#clip022)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  361.501,1401.58 361.501,123.472 
  "/>
<polyline clip-path="url(#clip022)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  844.815,1401.58 844.815,123.472 
  "/>
<polyline clip-path="url(#clip022)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1328.13,1401.58 1328.13,123.472 
  "/>
<polyline clip-path="url(#clip022)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1811.44,1401.58 1811.44,123.472 
  "/>
<polyline clip-path="url(#clip022)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  2294.76,1401.58 2294.76,123.472 
  "/>
<polyline clip-path="url(#clip020)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  303.503,1401.58 2352.76,1401.58 
  "/>
<polyline clip-path="url(#clip020)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  361.501,1401.58 361.501,1382.68 
  "/>
<polyline clip-path="url(#clip020)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  844.815,1401.58 844.815,1382.68 
  "/>
<polyline clip-path="url(#clip020)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1328.13,1401.58 1328.13,1382.68 
  "/>
<polyline clip-path="url(#clip020)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1811.44,1401.58 1811.44,1382.68 
  "/>
<polyline clip-path="url(#clip020)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  2294.76,1401.58 2294.76,1382.68 
  "/>
<path clip-path="url(#clip020)" d="M361.501 1438.36 Q356.084 1438.36 353.341 1443.71 Q350.633 1449.02 350.633 1459.71 Q350.633 1470.37 353.341 1475.72 Q356.084 1481.03 361.501 1481.03 Q366.952 1481.03 369.66 1475.72 Q372.403 1470.37 372.403 1459.71 Q372.403 1449.02 369.66 1443.71 Q366.952 1438.36 361.501 1438.36 M361.501 1432.8 Q370.216 1432.8 374.799 1439.71 Q379.417 1446.59 379.417 1459.71 Q379.417 1472.8 374.799 1479.71 Q370.216 1486.59 361.501 1486.59 Q352.785 1486.59 348.167 1479.71 Q343.584 1472.8 343.584 1459.71 Q343.584 1446.59 348.167 1439.71 Q352.785 1432.8 361.501 1432.8 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M830.388 1479.68 L841.846 1479.68 L841.846 1440.13 L829.381 1442.63 L829.381 1436.24 L841.777 1433.74 L848.791 1433.74 L848.791 1479.68 L860.249 1479.68 L860.249 1485.58 L830.388 1485.58 L830.388 1479.68 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1320.11 1479.68 L1344.59 1479.68 L1344.59 1485.58 L1311.67 1485.58 L1311.67 1479.68 Q1315.66 1475.55 1322.54 1468.6 Q1329.45 1461.62 1331.22 1459.61 Q1334.59 1455.82 1335.91 1453.22 Q1337.26 1450.58 1337.26 1448.05 Q1337.26 1443.91 1334.34 1441.31 Q1331.46 1438.71 1326.81 1438.71 Q1323.51 1438.71 1319.83 1439.85 Q1316.19 1441 1312.02 1443.32 L1312.02 1436.24 Q1316.25 1434.54 1319.94 1433.67 Q1323.62 1432.8 1326.67 1432.8 Q1334.73 1432.8 1339.52 1436.83 Q1344.31 1440.86 1344.31 1447.59 Q1344.31 1450.79 1343.09 1453.67 Q1341.91 1456.52 1338.75 1460.41 Q1337.89 1461.41 1333.23 1466.24 Q1328.58 1471.03 1320.11 1479.68 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1817.82 1457.63 Q1822.85 1458.7 1825.66 1462.11 Q1828.51 1465.51 1828.51 1470.51 Q1828.51 1478.18 1823.23 1482.39 Q1817.95 1486.59 1808.23 1486.59 Q1804.97 1486.59 1801.5 1485.93 Q1798.06 1485.3 1794.38 1484.02 L1794.38 1477.25 Q1797.29 1478.95 1800.77 1479.82 Q1804.24 1480.68 1808.02 1480.68 Q1814.62 1480.68 1818.06 1478.08 Q1821.53 1475.48 1821.53 1470.51 Q1821.53 1465.93 1818.3 1463.36 Q1815.11 1460.75 1809.38 1460.75 L1803.34 1460.75 L1803.34 1454.99 L1809.66 1454.99 Q1814.83 1454.99 1817.57 1452.94 Q1820.32 1450.86 1820.32 1446.97 Q1820.32 1442.98 1817.47 1440.86 Q1814.66 1438.71 1809.38 1438.71 Q1806.5 1438.71 1803.2 1439.33 Q1799.9 1439.96 1795.94 1441.27 L1795.94 1435.02 Q1799.93 1433.91 1803.41 1433.36 Q1806.91 1432.8 1810 1432.8 Q1817.99 1432.8 1822.64 1436.45 Q1827.29 1440.06 1827.29 1446.24 Q1827.29 1450.55 1824.83 1453.53 Q1822.36 1456.48 1817.82 1457.63 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M2299.27 1439.85 L2281.56 1467.52 L2299.27 1467.52 L2299.27 1439.85 M2297.43 1433.74 L2306.25 1433.74 L2306.25 1467.52 L2313.65 1467.52 L2313.65 1473.36 L2306.25 1473.36 L2306.25 1485.58 L2299.27 1485.58 L2299.27 1473.36 L2275.87 1473.36 L2275.87 1466.59 L2297.43 1433.74 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1271.74 1527.62 L1271.74 1538.66 L1284.9 1538.66 L1284.9 1543.62 L1271.74 1543.62 L1271.74 1564.74 Q1271.74 1569.49 1273.03 1570.85 Q1274.35 1572.2 1278.34 1572.2 L1284.9 1572.2 L1284.9 1577.55 L1278.34 1577.55 Q1270.94 1577.55 1268.13 1574.8 Q1265.32 1572.03 1265.32 1564.74 L1265.32 1543.62 L1260.63 1543.62 L1260.63 1538.66 L1265.32 1538.66 L1265.32 1527.62 L1271.74 1527.62 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1315.32 1523.52 L1330.04 1523.52 L1330.04 1528.49 L1321.71 1528.49 L1321.71 1581.96 L1330.04 1581.96 L1330.04 1586.92 L1315.32 1586.92 L1315.32 1523.52 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1368.44 1539.8 L1368.44 1545.85 Q1365.73 1544.46 1362.82 1543.76 Q1359.9 1543.07 1356.78 1543.07 Q1352.02 1543.07 1349.62 1544.53 Q1347.26 1545.99 1347.26 1548.9 Q1347.26 1551.12 1348.96 1552.41 Q1350.66 1553.66 1355.8 1554.8 L1357.99 1555.29 Q1364.8 1556.75 1367.64 1559.42 Q1370.53 1562.06 1370.53 1566.82 Q1370.53 1572.24 1366.22 1575.4 Q1361.95 1578.55 1354.45 1578.55 Q1351.32 1578.55 1347.92 1577.93 Q1344.55 1577.34 1340.8 1576.12 L1340.8 1569.53 Q1344.34 1571.37 1347.78 1572.3 Q1351.22 1573.21 1354.59 1573.21 Q1359.1 1573.21 1361.53 1571.68 Q1363.96 1570.12 1363.96 1567.3 Q1363.96 1564.7 1362.19 1563.31 Q1360.46 1561.92 1354.52 1560.64 L1352.3 1560.12 Q1346.36 1558.87 1343.72 1556.3 Q1341.08 1553.69 1341.08 1549.18 Q1341.08 1543.69 1344.97 1540.71 Q1348.86 1537.72 1356.01 1537.72 Q1359.55 1537.72 1362.68 1538.24 Q1365.8 1538.76 1368.44 1539.8 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1395.63 1523.52 L1395.63 1586.92 L1380.91 1586.92 L1380.91 1581.96 L1389.21 1581.96 L1389.21 1528.49 L1380.91 1528.49 L1380.91 1523.52 L1395.63 1523.52 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip022)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  303.503,1365.41 2352.76,1365.41 
  "/>
<polyline clip-path="url(#clip022)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  303.503,1087.5 2352.76,1087.5 
  "/>
<polyline clip-path="url(#clip022)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  303.503,809.586 2352.76,809.586 
  "/>
<polyline clip-path="url(#clip022)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  303.503,531.676 2352.76,531.676 
  "/>
<polyline clip-path="url(#clip022)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  303.503,253.766 2352.76,253.766 
  "/>
<polyline clip-path="url(#clip020)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  303.503,1401.58 303.503,123.472 
  "/>
<polyline clip-path="url(#clip020)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  303.503,1365.41 322.401,1365.41 
  "/>
<polyline clip-path="url(#clip020)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  303.503,1087.5 322.401,1087.5 
  "/>
<polyline clip-path="url(#clip020)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  303.503,809.586 322.401,809.586 
  "/>
<polyline clip-path="url(#clip020)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  303.503,531.676 322.401,531.676 
  "/>
<polyline clip-path="url(#clip020)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  303.503,253.766 322.401,253.766 
  "/>
<path clip-path="url(#clip020)" d="M136.497 1344.11 Q131.08 1344.11 128.337 1349.45 Q125.629 1354.76 125.629 1365.46 Q125.629 1376.12 128.337 1381.47 Q131.08 1386.78 136.497 1386.78 Q141.948 1386.78 144.656 1381.47 Q147.399 1376.12 147.399 1365.46 Q147.399 1354.76 144.656 1349.45 Q141.948 1344.11 136.497 1344.11 M136.497 1338.55 Q145.212 1338.55 149.795 1345.46 Q154.413 1352.33 154.413 1365.46 Q154.413 1378.55 149.795 1385.46 Q145.212 1392.33 136.497 1392.33 Q127.781 1392.33 123.163 1385.46 Q118.58 1378.55 118.58 1365.46 Q118.58 1352.33 123.163 1345.46 Q127.781 1338.55 136.497 1338.55 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M166.74 1382.51 L174.066 1382.51 L174.066 1391.33 L166.74 1391.33 L166.74 1382.51 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M189.413 1339.49 L216.948 1339.49 L216.948 1345.39 L195.837 1345.39 L195.837 1358.1 Q197.364 1357.58 198.892 1357.33 Q200.42 1357.06 201.948 1357.06 Q210.628 1357.06 215.698 1361.81 Q220.767 1366.57 220.767 1374.7 Q220.767 1383.06 215.559 1387.72 Q210.35 1392.33 200.871 1392.33 Q197.608 1392.33 194.205 1391.78 Q190.837 1391.22 187.226 1390.11 L187.226 1383.06 Q190.351 1384.76 193.684 1385.6 Q197.017 1386.43 200.732 1386.43 Q206.739 1386.43 210.246 1383.27 Q213.753 1380.11 213.753 1374.7 Q213.753 1369.28 210.246 1366.12 Q206.739 1362.96 200.732 1362.96 Q197.92 1362.96 195.108 1363.58 Q192.33 1364.21 189.413 1365.53 L189.413 1339.49 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M249.586 1344.11 Q244.17 1344.11 241.427 1349.45 Q238.718 1354.76 238.718 1365.46 Q238.718 1376.12 241.427 1381.47 Q244.17 1386.78 249.586 1386.78 Q255.038 1386.78 257.746 1381.47 Q260.489 1376.12 260.489 1365.46 Q260.489 1354.76 257.746 1349.45 Q255.038 1344.11 249.586 1344.11 M249.586 1338.55 Q258.302 1338.55 262.885 1345.46 Q267.503 1352.33 267.503 1365.46 Q267.503 1378.55 262.885 1385.46 Q258.302 1392.33 249.586 1392.33 Q240.871 1392.33 236.253 1385.46 Q231.67 1378.55 231.67 1365.46 Q231.67 1352.33 236.253 1345.46 Q240.871 1338.55 249.586 1338.55 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M137.99 1066.19 Q132.573 1066.19 129.83 1071.54 Q127.122 1076.85 127.122 1087.55 Q127.122 1098.21 129.83 1103.56 Q132.573 1108.87 137.99 1108.87 Q143.441 1108.87 146.149 1103.56 Q148.892 1098.21 148.892 1087.55 Q148.892 1076.85 146.149 1071.54 Q143.441 1066.19 137.99 1066.19 M137.99 1060.64 Q146.705 1060.64 151.288 1067.55 Q155.906 1074.42 155.906 1087.55 Q155.906 1100.64 151.288 1107.55 Q146.705 1114.42 137.99 1114.42 Q129.275 1114.42 124.657 1107.55 Q120.073 1100.64 120.073 1087.55 Q120.073 1074.42 124.657 1067.55 Q129.275 1060.64 137.99 1060.64 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M168.233 1104.6 L175.559 1104.6 L175.559 1113.42 L168.233 1113.42 L168.233 1104.6 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M189.066 1061.58 L222.399 1061.58 L222.399 1064.56 L203.58 1113.42 L196.253 1113.42 L213.962 1067.48 L189.066 1067.48 L189.066 1061.58 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M236.149 1061.58 L263.684 1061.58 L263.684 1067.48 L242.573 1067.48 L242.573 1080.19 Q244.1 1079.67 245.628 1079.42 Q247.156 1079.15 248.684 1079.15 Q257.364 1079.15 262.434 1083.9 Q267.503 1088.66 267.503 1096.78 Q267.503 1105.15 262.295 1109.81 Q257.086 1114.42 247.607 1114.42 Q244.343 1114.42 240.941 1113.87 Q237.573 1113.31 233.961 1112.2 L233.961 1105.15 Q237.086 1106.85 240.42 1107.69 Q243.753 1108.52 247.468 1108.52 Q253.475 1108.52 256.982 1105.36 Q260.489 1102.2 260.489 1096.78 Q260.489 1091.37 256.982 1088.21 Q253.475 1085.05 247.468 1085.05 Q244.656 1085.05 241.843 1085.67 Q239.066 1086.3 236.149 1087.62 L236.149 1061.58 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M122.712 829.604 L134.17 829.604 L134.17 790.055 L121.705 792.555 L121.705 786.166 L134.101 783.666 L141.115 783.666 L141.115 829.604 L152.573 829.604 L152.573 835.506 L122.712 835.506 L122.712 829.604 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M166.74 826.687 L174.066 826.687 L174.066 835.506 L166.74 835.506 L166.74 826.687 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M204.344 788.284 Q198.927 788.284 196.184 793.631 Q193.476 798.944 193.476 809.638 Q193.476 820.298 196.184 825.645 Q198.927 830.958 204.344 830.958 Q209.795 830.958 212.503 825.645 Q215.246 820.298 215.246 809.638 Q215.246 798.944 212.503 793.631 Q209.795 788.284 204.344 788.284 M204.344 782.729 Q213.059 782.729 217.642 789.638 Q222.26 796.513 222.26 809.638 Q222.26 822.729 217.642 829.638 Q213.059 836.513 204.344 836.513 Q195.628 836.513 191.01 829.638 Q186.427 822.729 186.427 809.638 Q186.427 796.513 191.01 789.638 Q195.628 782.729 204.344 782.729 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M249.586 788.284 Q244.17 788.284 241.427 793.631 Q238.718 798.944 238.718 809.638 Q238.718 820.298 241.427 825.645 Q244.17 830.958 249.586 830.958 Q255.038 830.958 257.746 825.645 Q260.489 820.298 260.489 809.638 Q260.489 798.944 257.746 793.631 Q255.038 788.284 249.586 788.284 M249.586 782.729 Q258.302 782.729 262.885 789.638 Q267.503 796.513 267.503 809.638 Q267.503 822.729 262.885 829.638 Q258.302 836.513 249.586 836.513 Q240.871 836.513 236.253 829.638 Q231.67 822.729 231.67 809.638 Q231.67 796.513 236.253 789.638 Q240.871 782.729 249.586 782.729 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M124.205 551.693 L135.663 551.693 L135.663 512.145 L123.198 514.645 L123.198 508.256 L135.594 505.756 L142.608 505.756 L142.608 551.693 L154.066 551.693 L154.066 557.596 L124.205 557.596 L124.205 551.693 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M168.233 548.776 L175.559 548.776 L175.559 557.596 L168.233 557.596 L168.233 548.776 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M196.878 551.693 L221.357 551.693 L221.357 557.596 L188.441 557.596 L188.441 551.693 Q192.434 547.561 199.309 540.617 Q206.219 533.638 207.989 531.624 Q211.357 527.839 212.677 525.235 Q214.031 522.596 214.031 520.061 Q214.031 515.929 211.114 513.325 Q208.232 510.721 203.58 510.721 Q200.281 510.721 196.601 511.867 Q192.955 513.013 188.788 515.339 L188.788 508.256 Q193.024 506.554 196.705 505.686 Q200.385 504.818 203.441 504.818 Q211.496 504.818 216.288 508.846 Q221.08 512.874 221.08 519.61 Q221.08 522.804 219.864 525.686 Q218.684 528.534 215.524 532.422 Q214.656 533.429 210.003 538.256 Q205.351 543.047 196.878 551.693 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M236.149 505.756 L263.684 505.756 L263.684 511.659 L242.573 511.659 L242.573 524.367 Q244.1 523.846 245.628 523.603 Q247.156 523.325 248.684 523.325 Q257.364 523.325 262.434 528.082 Q267.503 532.839 267.503 540.964 Q267.503 549.332 262.295 553.985 Q257.086 558.603 247.607 558.603 Q244.343 558.603 240.941 558.047 Q237.573 557.492 233.961 556.381 L233.961 549.332 Q237.086 551.033 240.42 551.867 Q243.753 552.7 247.468 552.7 Q253.475 552.7 256.982 549.54 Q260.489 546.381 260.489 540.964 Q260.489 535.547 256.982 532.388 Q253.475 529.228 247.468 529.228 Q244.656 529.228 241.843 529.853 Q239.066 530.478 236.149 531.797 L236.149 505.756 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M122.712 273.783 L134.17 273.783 L134.17 234.234 L121.705 236.734 L121.705 230.346 L134.101 227.846 L141.115 227.846 L141.115 273.783 L152.573 273.783 L152.573 279.686 L122.712 279.686 L122.712 273.783 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M166.74 270.866 L174.066 270.866 L174.066 279.686 L166.74 279.686 L166.74 270.866 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M189.413 227.846 L216.948 227.846 L216.948 233.748 L195.837 233.748 L195.837 246.457 Q197.364 245.936 198.892 245.693 Q200.42 245.415 201.948 245.415 Q210.628 245.415 215.698 250.172 Q220.767 254.929 220.767 263.054 Q220.767 271.422 215.559 276.074 Q210.35 280.692 200.871 280.692 Q197.608 280.692 194.205 280.137 Q190.837 279.581 187.226 278.47 L187.226 271.422 Q190.351 273.123 193.684 273.956 Q197.017 274.79 200.732 274.79 Q206.739 274.79 210.246 271.63 Q213.753 268.47 213.753 263.054 Q213.753 257.637 210.246 254.477 Q206.739 251.318 200.732 251.318 Q197.92 251.318 195.108 251.943 Q192.33 252.568 189.413 253.887 L189.413 227.846 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M249.586 232.464 Q244.17 232.464 241.427 237.811 Q238.718 243.123 238.718 253.818 Q238.718 264.477 241.427 269.824 Q244.17 275.137 249.586 275.137 Q255.038 275.137 257.746 269.824 Q260.489 264.477 260.489 253.818 Q260.489 243.123 257.746 237.811 Q255.038 232.464 249.586 232.464 M249.586 226.908 Q258.302 226.908 262.885 233.818 Q267.503 240.693 267.503 253.818 Q267.503 266.908 262.885 273.817 Q258.302 280.692 249.586 280.692 Q240.871 280.692 236.253 273.817 Q231.67 266.908 231.67 253.818 Q231.67 240.693 236.253 233.818 Q240.871 226.908 249.586 226.908 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M35.6047 1045.01 Q31.2992 1042.61 29.2506 1039.28 Q27.202 1035.94 27.202 1031.43 Q27.202 1025.35 31.4728 1022.06 Q35.7089 1018.76 43.5561 1018.76 L67.0282 1018.76 L67.0282 1025.18 L43.7644 1025.18 Q38.1741 1025.18 35.4658 1027.16 Q32.7575 1029.14 32.7575 1033.2 Q32.7575 1038.17 36.0561 1041.05 Q39.3547 1043.93 45.0491 1043.93 L67.0282 1043.93 L67.0282 1050.35 L43.7644 1050.35 Q38.1394 1050.35 35.4658 1052.33 Q32.7575 1054.31 32.7575 1058.44 Q32.7575 1063.34 36.0908 1066.22 Q39.3894 1069.1 45.0491 1069.1 L67.0282 1069.1 L67.0282 1075.53 L28.1395 1075.53 L28.1395 1069.1 L34.1811 1069.1 Q30.6047 1066.92 28.9034 1063.86 Q27.202 1060.81 27.202 1056.6 Q27.202 1052.37 29.3547 1049.42 Q31.5075 1046.43 35.6047 1045.01 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M47.4797 988.341 Q47.4797 996.084 49.2505 999.07 Q51.0213 1002.06 55.2921 1002.06 Q58.6949 1002.06 60.7087 999.834 Q62.6879 997.577 62.6879 993.723 Q62.6879 988.41 58.9379 985.216 Q55.1532 981.987 48.9033 981.987 L47.4797 981.987 L47.4797 988.341 M44.8408 975.598 L67.0282 975.598 L67.0282 981.987 L61.1254 981.987 Q64.6671 984.174 66.3684 987.438 Q68.0351 990.702 68.0351 995.424 Q68.0351 1001.4 64.7018 1004.94 Q61.3337 1008.44 55.7088 1008.44 Q49.1463 1008.44 45.813 1004.07 Q42.4797 999.66 42.4797 990.945 L42.4797 981.987 L41.8547 981.987 Q37.445 981.987 35.0492 984.903 Q32.6186 987.785 32.6186 993.028 Q32.6186 996.362 33.4172 999.521 Q34.2158 1002.68 35.813 1005.6 L29.9103 1005.6 Q28.5561 1002.09 27.8964 998.792 Q27.202 995.493 27.202 992.369 Q27.202 983.931 31.577 979.764 Q35.9519 975.598 44.8408 975.598 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M29.2853 937.647 L35.3269 937.647 Q33.9381 940.355 33.2436 943.272 Q32.5492 946.188 32.5492 949.313 Q32.5492 954.07 34.0075 956.466 Q35.4658 958.827 38.3825 958.827 Q40.6047 958.827 41.8894 957.126 Q43.1394 955.424 44.2852 950.285 L44.7713 948.098 Q46.2297 941.292 48.9033 938.445 Q51.5421 935.563 56.299 935.563 Q61.7157 935.563 64.8754 939.869 Q68.0351 944.14 68.0351 951.64 Q68.0351 954.765 67.4101 958.167 Q66.8198 961.535 65.6046 965.285 L59.0074 965.285 Q60.8476 961.744 61.7851 958.306 Q62.6879 954.869 62.6879 951.501 Q62.6879 946.987 61.1601 944.556 Q59.5976 942.126 56.7852 942.126 Q54.181 942.126 52.7921 943.897 Q51.4032 945.633 50.1185 951.57 L49.5977 953.792 Q48.3477 959.73 45.7783 962.369 Q43.1741 965.008 38.6603 965.008 Q33.1742 965.008 30.1881 961.119 Q27.202 957.23 27.202 950.077 Q27.202 946.535 27.7228 943.41 Q28.2436 940.285 29.2853 937.647 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M29.2853 900.598 L35.3269 900.598 Q33.9381 903.306 33.2436 906.223 Q32.5492 909.14 32.5492 912.265 Q32.5492 917.022 34.0075 919.418 Q35.4658 921.779 38.3825 921.779 Q40.6047 921.779 41.8894 920.077 Q43.1394 918.376 44.2852 913.237 L44.7713 911.049 Q46.2297 904.244 48.9033 901.397 Q51.5421 898.515 56.299 898.515 Q61.7157 898.515 64.8754 902.82 Q68.0351 907.091 68.0351 914.591 Q68.0351 917.716 67.4101 921.119 Q66.8198 924.487 65.6046 928.237 L59.0074 928.237 Q60.8476 924.695 61.7851 921.258 Q62.6879 917.82 62.6879 914.452 Q62.6879 909.938 61.1601 907.508 Q59.5976 905.077 56.7852 905.077 Q54.181 905.077 52.7921 906.848 Q51.4032 908.584 50.1185 914.522 L49.5977 916.744 Q48.3477 922.681 45.7783 925.32 Q43.1741 927.959 38.6603 927.959 Q33.1742 927.959 30.1881 924.07 Q27.202 920.181 27.202 913.029 Q27.202 909.487 27.7228 906.362 Q28.2436 903.237 29.2853 900.598 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M61.1949 859.557 L81.8197 859.557 L81.8197 865.98 L28.1395 865.98 L28.1395 859.557 L34.0422 859.557 Q30.57 857.543 28.9034 854.487 Q27.202 851.397 27.202 847.126 Q27.202 840.043 32.827 835.633 Q38.4519 831.189 47.6185 831.189 Q56.7852 831.189 62.4101 835.633 Q68.0351 840.043 68.0351 847.126 Q68.0351 851.397 66.3684 854.487 Q64.6671 857.543 61.1949 859.557 M47.6185 837.821 Q40.57 837.821 36.5769 840.737 Q32.5492 843.619 32.5492 848.689 Q32.5492 853.758 36.5769 856.675 Q40.57 859.557 47.6185 859.557 Q54.6671 859.557 58.6949 856.675 Q62.6879 853.758 62.6879 848.689 Q62.6879 843.619 58.6949 840.737 Q54.6671 837.821 47.6185 837.821 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M32.6186 805.529 Q32.6186 810.668 36.6464 813.654 Q40.6394 816.64 47.6185 816.64 Q54.5977 816.64 58.6254 813.689 Q62.6185 810.703 62.6185 805.529 Q62.6185 800.425 58.5907 797.439 Q54.5629 794.453 47.6185 794.453 Q40.7089 794.453 36.6811 797.439 Q32.6186 800.425 32.6186 805.529 M27.202 805.529 Q27.202 797.196 32.6186 792.439 Q38.0353 787.682 47.6185 787.682 Q57.1671 787.682 62.6185 792.439 Q68.0351 797.196 68.0351 805.529 Q68.0351 813.897 62.6185 818.654 Q57.1671 823.376 47.6185 823.376 Q38.0353 823.376 32.6186 818.654 Q27.202 813.897 27.202 805.529 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M29.2853 752.3 L35.3269 752.3 Q33.9381 755.009 33.2436 757.925 Q32.5492 760.842 32.5492 763.967 Q32.5492 768.724 34.0075 771.12 Q35.4658 773.481 38.3825 773.481 Q40.6047 773.481 41.8894 771.779 Q43.1394 770.078 44.2852 764.939 L44.7713 762.752 Q46.2297 755.946 48.9033 753.099 Q51.5421 750.217 56.299 750.217 Q61.7157 750.217 64.8754 754.523 Q68.0351 758.793 68.0351 766.293 Q68.0351 769.418 67.4101 772.821 Q66.8198 776.189 65.6046 779.939 L59.0074 779.939 Q60.8476 776.397 61.7851 772.96 Q62.6879 769.522 62.6879 766.154 Q62.6879 761.641 61.1601 759.21 Q59.5976 756.779 56.7852 756.779 Q54.181 756.779 52.7921 758.55 Q51.4032 760.286 50.1185 766.224 L49.5977 768.446 Q48.3477 774.384 45.7783 777.022 Q43.1741 779.661 38.6603 779.661 Q33.1742 779.661 30.1881 775.772 Q27.202 771.884 27.202 764.731 Q27.202 761.189 27.7228 758.064 Q28.2436 754.939 29.2853 752.3 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M28.1395 740.043 L28.1395 733.655 L67.0282 733.655 L67.0282 740.043 L28.1395 740.043 M13.0007 740.043 L13.0007 733.655 L21.0909 733.655 L21.0909 740.043 L13.0007 740.043 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M17.0979 713.967 L28.1395 713.967 L28.1395 700.808 L33.1047 700.808 L33.1047 713.967 L54.2157 713.967 Q58.9726 713.967 60.3268 712.683 Q61.681 711.363 61.681 707.37 L61.681 700.808 L67.0282 700.808 L67.0282 707.37 Q67.0282 714.766 64.2851 717.578 Q61.5074 720.391 54.2157 720.391 L33.1047 720.391 L33.1047 725.078 L28.1395 725.078 L28.1395 720.391 L17.0979 720.391 L17.0979 713.967 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M28.1395 692.405 L28.1395 686.016 L67.0282 686.016 L67.0282 692.405 L28.1395 692.405 M13.0007 692.405 L13.0007 686.016 L21.0909 686.016 L21.0909 692.405 L13.0007 692.405 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M32.6186 657.579 Q32.6186 662.717 36.6464 665.704 Q40.6394 668.69 47.6185 668.69 Q54.5977 668.69 58.6254 665.738 Q62.6185 662.752 62.6185 657.579 Q62.6185 652.474 58.5907 649.488 Q54.5629 646.502 47.6185 646.502 Q40.7089 646.502 36.6811 649.488 Q32.6186 652.474 32.6186 657.579 M27.202 657.579 Q27.202 649.245 32.6186 644.488 Q38.0353 639.732 47.6185 639.732 Q57.1671 639.732 62.6185 644.488 Q68.0351 649.245 68.0351 657.579 Q68.0351 665.947 62.6185 670.704 Q57.1671 675.426 47.6185 675.426 Q38.0353 675.426 32.6186 670.704 Q27.202 665.947 27.202 657.579 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M43.5561 596.815 L67.0282 596.815 L67.0282 603.204 L43.7644 603.204 Q38.2436 603.204 35.5005 605.357 Q32.7575 607.509 32.7575 611.815 Q32.7575 616.989 36.0561 619.975 Q39.3547 622.961 45.0491 622.961 L67.0282 622.961 L67.0282 629.384 L28.1395 629.384 L28.1395 622.961 L34.1811 622.961 Q30.6742 620.669 28.9381 617.579 Q27.202 614.454 27.202 610.391 Q27.202 603.69 31.3686 600.253 Q35.5005 596.815 43.5561 596.815 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M13.0007 562.058 L13.0007 547.336 L17.9659 547.336 L17.9659 555.669 L71.4379 555.669 L71.4379 547.336 L76.4031 547.336 L76.4031 562.058 L13.0007 562.058 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M35.6047 503.448 Q31.2992 501.052 29.2506 497.718 Q27.202 494.385 27.202 489.871 Q27.202 483.795 31.4728 480.496 Q35.7089 477.198 43.5561 477.198 L67.0282 477.198 L67.0282 483.621 L43.7644 483.621 Q38.1741 483.621 35.4658 485.6 Q32.7575 487.58 32.7575 491.642 Q32.7575 496.607 36.0561 499.489 Q39.3547 502.371 45.0491 502.371 L67.0282 502.371 L67.0282 508.795 L43.7644 508.795 Q38.1394 508.795 35.4658 510.774 Q32.7575 512.753 32.7575 516.885 Q32.7575 521.781 36.0908 524.663 Q39.3894 527.545 45.0491 527.545 L67.0282 527.545 L67.0282 533.968 L28.1395 533.968 L28.1395 527.545 L34.1811 527.545 Q30.6047 525.357 28.9034 522.302 Q27.202 519.246 27.202 515.045 Q27.202 510.809 29.3547 507.857 Q31.5075 504.871 35.6047 503.448 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M13.0007 449.524 L76.4031 449.524 L76.4031 464.246 L71.4379 464.246 L71.4379 455.948 L17.9659 455.948 L17.9659 464.246 L13.0007 464.246 L13.0007 449.524 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M592.668 12.096 L604.861 12.096 L620.295 53.2532 L635.81 12.096 L648.003 12.096 L648.003 72.576 L640.023 72.576 L640.023 19.4686 L624.427 60.9499 L616.203 60.9499 L600.607 19.4686 L600.607 72.576 L592.668 72.576 L592.668 12.096 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M681.504 32.4315 Q675.509 32.4315 672.025 37.1306 Q668.541 41.7891 668.541 49.9314 Q668.541 58.0738 671.984 62.7728 Q675.468 67.4314 681.504 67.4314 Q687.459 67.4314 690.942 62.7323 Q694.426 58.0333 694.426 49.9314 Q694.426 41.8701 690.942 37.1711 Q687.459 32.4315 681.504 32.4315 M681.504 26.1121 Q691.226 26.1121 696.776 32.4315 Q702.325 38.7509 702.325 49.9314 Q702.325 61.0714 696.776 67.4314 Q691.226 73.7508 681.504 73.7508 Q671.741 73.7508 666.191 67.4314 Q660.682 61.0714 660.682 49.9314 Q660.682 38.7509 666.191 32.4315 Q671.741 26.1121 681.504 26.1121 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M744.536 34.0924 L744.536 9.54393 L751.99 9.54393 L751.99 72.576 L744.536 72.576 L744.536 65.7705 Q742.186 69.8214 738.581 71.8063 Q735.016 73.7508 729.993 73.7508 Q721.77 73.7508 716.585 67.1883 Q711.44 60.6258 711.44 49.9314 Q711.44 39.2371 716.585 32.6746 Q721.77 26.1121 729.993 26.1121 Q735.016 26.1121 738.581 28.0971 Q742.186 30.0415 744.536 34.0924 M719.137 49.9314 Q719.137 58.1548 722.499 62.8538 Q725.902 67.5124 731.816 67.5124 Q737.73 67.5124 741.133 62.8538 Q744.536 58.1548 744.536 49.9314 Q744.536 41.7081 741.133 37.0496 Q737.73 32.3505 731.816 32.3505 Q725.902 32.3505 722.499 37.0496 Q719.137 41.7081 719.137 49.9314 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M767.343 27.2059 L774.796 27.2059 L774.796 72.576 L767.343 72.576 L767.343 27.2059 M767.343 9.54393 L774.796 9.54393 L774.796 18.9825 L767.343 18.9825 L767.343 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M813.361 9.54393 L813.361 15.7418 L806.231 15.7418 Q802.221 15.7418 800.641 17.3622 Q799.102 18.9825 799.102 23.1955 L799.102 27.2059 L811.376 27.2059 L811.376 32.9987 L799.102 32.9987 L799.102 72.576 L791.607 72.576 L791.607 32.9987 L784.478 32.9987 L784.478 27.2059 L791.607 27.2059 L791.607 24.0462 Q791.607 16.471 795.132 13.0277 Q798.656 9.54393 806.312 9.54393 L813.361 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M819.599 27.2059 L827.053 27.2059 L827.053 72.576 L819.599 72.576 L819.599 27.2059 M819.599 9.54393 L827.053 9.54393 L827.053 18.9825 L819.599 18.9825 L819.599 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M881.456 48.0275 L881.456 51.6733 L847.186 51.6733 Q847.672 59.3701 851.804 63.421 Q855.976 67.4314 863.389 67.4314 Q867.683 67.4314 871.694 66.3781 Q875.745 65.3249 879.715 63.2184 L879.715 70.267 Q875.704 71.9684 871.491 72.8596 Q867.278 73.7508 862.944 73.7508 Q852.087 73.7508 845.728 67.4314 Q839.408 61.1119 839.408 50.3365 Q839.408 39.1965 845.403 32.6746 Q851.439 26.1121 861.648 26.1121 Q870.803 26.1121 876.109 32.0264 Q881.456 37.9003 881.456 48.0275 M874.003 45.84 Q873.922 39.7232 870.56 36.0774 Q867.238 32.4315 861.729 32.4315 Q855.49 32.4315 851.723 35.9558 Q847.996 39.4801 847.429 45.8805 L874.003 45.84 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M923.545 34.0924 L923.545 9.54393 L930.999 9.54393 L930.999 72.576 L923.545 72.576 L923.545 65.7705 Q921.196 69.8214 917.591 71.8063 Q914.026 73.7508 909.003 73.7508 Q900.779 73.7508 895.594 67.1883 Q890.449 60.6258 890.449 49.9314 Q890.449 39.2371 895.594 32.6746 Q900.779 26.1121 909.003 26.1121 Q914.026 26.1121 917.591 28.0971 Q921.196 30.0415 923.545 34.0924 M898.146 49.9314 Q898.146 58.1548 901.508 62.8538 Q904.911 67.5124 910.826 67.5124 Q916.74 67.5124 920.143 62.8538 Q923.545 58.1548 923.545 49.9314 Q923.545 41.7081 920.143 37.0496 Q916.74 32.3505 910.826 32.3505 Q904.911 32.3505 901.508 37.0496 Q898.146 41.7081 898.146 49.9314 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M948.256 62.2867 L956.803 62.2867 L956.803 72.576 L948.256 72.576 L948.256 62.2867 M948.256 29.6769 L956.803 29.6769 L956.803 39.9662 L948.256 39.9662 L948.256 29.6769 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1001 12.096 L1013.19 12.096 L1028.63 53.2532 L1044.14 12.096 L1056.33 12.096 L1056.33 72.576 L1048.35 72.576 L1048.35 19.4686 L1032.76 60.9499 L1024.53 60.9499 L1008.94 19.4686 L1008.94 72.576 L1001 72.576 L1001 12.096 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1092.87 49.7694 Q1083.84 49.7694 1080.36 51.8354 Q1076.87 53.9013 1076.87 58.8839 Q1076.87 62.8538 1079.46 65.2034 Q1082.1 67.5124 1086.59 67.5124 Q1092.79 67.5124 1096.52 63.1374 Q1100.29 58.7219 1100.29 51.4303 L1100.29 49.7694 L1092.87 49.7694 M1107.74 46.6907 L1107.74 72.576 L1100.29 72.576 L1100.29 65.6895 Q1097.73 69.8214 1093.93 71.8063 Q1090.12 73.7508 1084.61 73.7508 Q1077.64 73.7508 1073.51 69.8619 Q1069.42 65.9325 1069.42 59.3701 Q1069.42 51.7138 1074.52 47.825 Q1079.67 43.9361 1089.83 43.9361 L1100.29 43.9361 L1100.29 43.2069 Q1100.29 38.0623 1096.88 35.2672 Q1093.52 32.4315 1087.4 32.4315 Q1083.52 32.4315 1079.83 33.3632 Q1076.14 34.295 1072.74 36.1584 L1072.74 29.2718 Q1076.83 27.692 1080.68 26.9223 Q1084.53 26.1121 1088.17 26.1121 Q1098.02 26.1121 1102.88 31.2163 Q1107.74 36.3204 1107.74 46.6907 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1152.02 28.5427 L1152.02 35.5912 Q1148.86 33.9709 1145.45 33.1607 Q1142.05 32.3505 1138.41 32.3505 Q1132.86 32.3505 1130.06 34.0519 Q1127.31 35.7533 1127.31 39.156 Q1127.31 41.7486 1129.29 43.2475 Q1131.28 44.7058 1137.27 46.0426 L1139.82 46.6097 Q1147.76 48.3111 1151.08 51.4303 Q1154.45 54.509 1154.45 60.0587 Q1154.45 66.3781 1149.42 70.0644 Q1144.44 73.7508 1135.69 73.7508 Q1132.05 73.7508 1128.08 73.0216 Q1124.15 72.3329 1119.77 70.9151 L1119.77 63.2184 Q1123.9 65.3654 1127.91 66.4591 Q1131.92 67.5124 1135.85 67.5124 Q1141.12 67.5124 1143.96 65.73 Q1146.79 63.9071 1146.79 60.6258 Q1146.79 57.5877 1144.72 55.9673 Q1142.7 54.3469 1135.77 52.8481 L1133.18 52.2405 Q1126.25 50.7821 1123.17 47.7845 Q1120.1 44.7463 1120.1 39.4801 Q1120.1 33.0797 1124.63 29.5959 Q1129.17 26.1121 1137.51 26.1121 Q1141.65 26.1121 1145.29 26.7198 Q1148.94 27.3274 1152.02 28.5427 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1195.24 28.5427 L1195.24 35.5912 Q1192.08 33.9709 1188.68 33.1607 Q1185.27 32.3505 1181.63 32.3505 Q1176.08 32.3505 1173.28 34.0519 Q1170.53 35.7533 1170.53 39.156 Q1170.53 41.7486 1172.51 43.2475 Q1174.5 44.7058 1180.49 46.0426 L1183.05 46.6097 Q1190.99 48.3111 1194.31 51.4303 Q1197.67 54.509 1197.67 60.0587 Q1197.67 66.3781 1192.65 70.0644 Q1187.66 73.7508 1178.91 73.7508 Q1175.27 73.7508 1171.3 73.0216 Q1167.37 72.3329 1162.99 70.9151 L1162.99 63.2184 Q1167.13 65.3654 1171.14 66.4591 Q1175.15 67.5124 1179.08 67.5124 Q1184.34 67.5124 1187.18 65.73 Q1190.01 63.9071 1190.01 60.6258 Q1190.01 57.5877 1187.95 55.9673 Q1185.92 54.3469 1179 52.8481 L1176.4 52.2405 Q1169.48 50.7821 1166.4 47.7845 Q1163.32 44.7463 1163.32 39.4801 Q1163.32 33.0797 1167.86 29.5959 Q1172.39 26.1121 1180.74 26.1121 Q1184.87 26.1121 1188.52 26.7198 Q1192.16 27.3274 1195.24 28.5427 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1243.12 65.7705 L1243.12 89.8329 L1235.63 89.8329 L1235.63 27.2059 L1243.12 27.2059 L1243.12 34.0924 Q1245.47 30.0415 1249.04 28.0971 Q1252.64 26.1121 1257.62 26.1121 Q1265.89 26.1121 1271.03 32.6746 Q1276.22 39.2371 1276.22 49.9314 Q1276.22 60.6258 1271.03 67.1883 Q1265.89 73.7508 1257.62 73.7508 Q1252.64 73.7508 1249.04 71.8063 Q1245.47 69.8214 1243.12 65.7705 M1268.48 49.9314 Q1268.48 41.7081 1265.08 37.0496 Q1261.71 32.3505 1255.8 32.3505 Q1249.89 32.3505 1246.48 37.0496 Q1243.12 41.7081 1243.12 49.9314 Q1243.12 58.1548 1246.48 62.8538 Q1249.89 67.5124 1255.8 67.5124 Q1261.71 67.5124 1265.08 62.8538 Q1268.48 58.1548 1268.48 49.9314 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1306.15 32.4315 Q1300.16 32.4315 1296.67 37.1306 Q1293.19 41.7891 1293.19 49.9314 Q1293.19 58.0738 1296.63 62.7728 Q1300.12 67.4314 1306.15 67.4314 Q1312.11 67.4314 1315.59 62.7323 Q1319.08 58.0333 1319.08 49.9314 Q1319.08 41.8701 1315.59 37.1711 Q1312.11 32.4315 1306.15 32.4315 M1306.15 26.1121 Q1315.88 26.1121 1321.43 32.4315 Q1326.97 38.7509 1326.97 49.9314 Q1326.97 61.0714 1321.43 67.4314 Q1315.88 73.7508 1306.15 73.7508 Q1296.39 73.7508 1290.84 67.4314 Q1285.33 61.0714 1285.33 49.9314 Q1285.33 38.7509 1290.84 32.4315 Q1296.39 26.1121 1306.15 26.1121 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1368.25 28.5427 L1368.25 35.5912 Q1365.09 33.9709 1361.69 33.1607 Q1358.29 32.3505 1354.64 32.3505 Q1349.09 32.3505 1346.3 34.0519 Q1343.54 35.7533 1343.54 39.156 Q1343.54 41.7486 1345.53 43.2475 Q1347.51 44.7058 1353.51 46.0426 L1356.06 46.6097 Q1364 48.3111 1367.32 51.4303 Q1370.68 54.509 1370.68 60.0587 Q1370.68 66.3781 1365.66 70.0644 Q1360.68 73.7508 1351.93 73.7508 Q1348.28 73.7508 1344.31 73.0216 Q1340.38 72.3329 1336.01 70.9151 L1336.01 63.2184 Q1340.14 65.3654 1344.15 66.4591 Q1348.16 67.5124 1352.09 67.5124 Q1357.36 67.5124 1360.19 65.73 Q1363.03 63.9071 1363.03 60.6258 Q1363.03 57.5877 1360.96 55.9673 Q1358.94 54.3469 1352.01 52.8481 L1349.42 52.2405 Q1342.49 50.7821 1339.41 47.7845 Q1336.33 44.7463 1336.33 39.4801 Q1336.33 33.0797 1340.87 29.5959 Q1345.41 26.1121 1353.75 26.1121 Q1357.88 26.1121 1361.53 26.7198 Q1365.17 27.3274 1368.25 28.5427 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1382.55 27.2059 L1390.01 27.2059 L1390.01 72.576 L1382.55 72.576 L1382.55 27.2059 M1382.55 9.54393 L1390.01 9.54393 L1390.01 18.9825 L1382.55 18.9825 L1382.55 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1412.98 14.324 L1412.98 27.2059 L1428.33 27.2059 L1428.33 32.9987 L1412.98 32.9987 L1412.98 57.6282 Q1412.98 63.1779 1414.47 64.7578 Q1416.01 66.3376 1420.67 66.3376 L1428.33 66.3376 L1428.33 72.576 L1420.67 72.576 Q1412.04 72.576 1408.76 69.3758 Q1405.48 66.1351 1405.48 57.6282 L1405.48 32.9987 L1400.01 32.9987 L1400.01 27.2059 L1405.48 27.2059 L1405.48 14.324 L1412.98 14.324 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1438.13 27.2059 L1445.59 27.2059 L1445.59 72.576 L1438.13 72.576 L1438.13 27.2059 M1438.13 9.54393 L1445.59 9.54393 L1445.59 18.9825 L1438.13 18.9825 L1438.13 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1478.76 32.4315 Q1472.77 32.4315 1469.28 37.1306 Q1465.8 41.7891 1465.8 49.9314 Q1465.8 58.0738 1469.24 62.7728 Q1472.73 67.4314 1478.76 67.4314 Q1484.72 67.4314 1488.2 62.7323 Q1491.68 58.0333 1491.68 49.9314 Q1491.68 41.8701 1488.2 37.1711 Q1484.72 32.4315 1478.76 32.4315 M1478.76 26.1121 Q1488.48 26.1121 1494.03 32.4315 Q1499.58 38.7509 1499.58 49.9314 Q1499.58 61.0714 1494.03 67.4314 Q1488.48 73.7508 1478.76 73.7508 Q1469 73.7508 1463.45 67.4314 Q1457.94 61.0714 1457.94 49.9314 Q1457.94 38.7509 1463.45 32.4315 Q1469 26.1121 1478.76 26.1121 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1549.65 45.1919 L1549.65 72.576 L1542.2 72.576 L1542.2 45.4349 Q1542.2 38.994 1539.69 35.7938 Q1537.18 32.5936 1532.15 32.5936 Q1526.12 32.5936 1522.63 36.4419 Q1519.15 40.2903 1519.15 46.9338 L1519.15 72.576 L1511.66 72.576 L1511.66 27.2059 L1519.15 27.2059 L1519.15 34.2544 Q1521.82 30.163 1525.43 28.1376 Q1529.07 26.1121 1533.81 26.1121 Q1541.63 26.1121 1545.64 30.9732 Q1549.65 35.7938 1549.65 45.1919 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1611.51 49.7694 Q1602.48 49.7694 1598.99 51.8354 Q1595.51 53.9013 1595.51 58.8839 Q1595.51 62.8538 1598.1 65.2034 Q1600.74 67.5124 1605.23 67.5124 Q1611.43 67.5124 1615.16 63.1374 Q1618.92 58.7219 1618.92 51.4303 L1618.92 49.7694 L1611.51 49.7694 M1626.38 46.6907 L1626.38 72.576 L1618.92 72.576 L1618.92 65.6895 Q1616.37 69.8214 1612.56 71.8063 Q1608.76 73.7508 1603.25 73.7508 Q1596.28 73.7508 1592.15 69.8619 Q1588.06 65.9325 1588.06 59.3701 Q1588.06 51.7138 1593.16 47.825 Q1598.3 43.9361 1608.47 43.9361 L1618.92 43.9361 L1618.92 43.2069 Q1618.92 38.0623 1615.52 35.2672 Q1612.16 32.4315 1606.04 32.4315 Q1602.15 32.4315 1598.47 33.3632 Q1594.78 34.295 1591.38 36.1584 L1591.38 29.2718 Q1595.47 27.692 1599.32 26.9223 Q1603.17 26.1121 1606.81 26.1121 Q1616.66 26.1121 1621.52 31.2163 Q1626.38 36.3204 1626.38 46.6907 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1664.7 9.54393 L1664.7 15.7418 L1657.57 15.7418 Q1653.56 15.7418 1651.98 17.3622 Q1650.44 18.9825 1650.44 23.1955 L1650.44 27.2059 L1662.71 27.2059 L1662.71 32.9987 L1650.44 32.9987 L1650.44 72.576 L1642.95 72.576 L1642.95 32.9987 L1635.82 32.9987 L1635.82 27.2059 L1642.95 27.2059 L1642.95 24.0462 Q1642.95 16.471 1646.47 13.0277 Q1649.99 9.54393 1657.65 9.54393 L1664.7 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1676.85 14.324 L1676.85 27.2059 L1692.2 27.2059 L1692.2 32.9987 L1676.85 32.9987 L1676.85 57.6282 Q1676.85 63.1779 1678.35 64.7578 Q1679.89 66.3376 1684.55 66.3376 L1692.2 66.3376 L1692.2 72.576 L1684.55 72.576 Q1675.92 72.576 1672.64 69.3758 Q1669.36 66.1351 1669.36 57.6282 L1669.36 32.9987 L1663.89 32.9987 L1663.89 27.2059 L1669.36 27.2059 L1669.36 14.324 L1676.85 14.324 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1740.82 48.0275 L1740.82 51.6733 L1706.54 51.6733 Q1707.03 59.3701 1711.16 63.421 Q1715.34 67.4314 1722.75 67.4314 Q1727.04 67.4314 1731.05 66.3781 Q1735.1 65.3249 1739.07 63.2184 L1739.07 70.267 Q1735.06 71.9684 1730.85 72.8596 Q1726.64 73.7508 1722.3 73.7508 Q1711.45 73.7508 1705.09 67.4314 Q1698.77 61.1119 1698.77 50.3365 Q1698.77 39.1965 1704.76 32.6746 Q1710.8 26.1121 1721.01 26.1121 Q1730.16 26.1121 1735.47 32.0264 Q1740.82 37.9003 1740.82 48.0275 M1733.36 45.84 Q1733.28 39.7232 1729.92 36.0774 Q1726.6 32.4315 1721.09 32.4315 Q1714.85 32.4315 1711.08 35.9558 Q1707.35 39.4801 1706.79 45.8805 L1733.36 45.84 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1779.34 34.1734 Q1778.08 33.4443 1776.58 33.1202 Q1775.13 32.7556 1773.34 32.7556 Q1767.02 32.7556 1763.62 36.8875 Q1760.26 40.9789 1760.26 48.6757 L1760.26 72.576 L1752.77 72.576 L1752.77 27.2059 L1760.26 27.2059 L1760.26 34.2544 Q1762.61 30.1225 1766.38 28.1376 Q1770.14 26.1121 1775.53 26.1121 Q1776.3 26.1121 1777.23 26.2337 Q1778.16 26.3147 1779.3 26.5172 L1779.34 34.1734 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1842.53 44.2197 Q1845.17 45.1109 1847.64 48.0275 Q1850.15 50.9442 1852.66 56.0483 L1860.97 72.576 L1852.17 72.576 L1844.44 57.061 Q1841.44 50.9847 1838.6 48.9997 Q1835.81 47.0148 1830.95 47.0148 L1822.04 47.0148 L1822.04 72.576 L1813.85 72.576 L1813.85 12.096 L1832.33 12.096 Q1842.7 12.096 1847.8 16.4305 Q1852.9 20.7649 1852.9 29.5149 Q1852.9 35.2267 1850.23 38.994 Q1847.6 42.7613 1842.53 44.2197 M1822.04 18.8205 L1822.04 40.2903 L1832.33 40.2903 Q1838.24 40.2903 1841.24 37.5762 Q1844.28 34.8216 1844.28 29.5149 Q1844.28 24.2082 1841.24 21.5346 Q1838.24 18.8205 1832.33 18.8205 L1822.04 18.8205 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1866.68 54.671 L1866.68 27.2059 L1874.13 27.2059 L1874.13 54.3874 Q1874.13 60.8284 1876.64 64.0691 Q1879.15 67.2693 1884.18 67.2693 Q1890.21 67.2693 1893.7 63.421 Q1897.22 59.5726 1897.22 52.9291 L1897.22 27.2059 L1904.67 27.2059 L1904.67 72.576 L1897.22 72.576 L1897.22 65.6084 Q1894.51 69.7404 1890.9 71.7658 Q1887.34 73.7508 1882.6 73.7508 Q1874.78 73.7508 1870.73 68.8897 Q1866.68 64.0286 1866.68 54.671 M1885.43 26.1121 L1885.43 26.1121 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1957.74 45.1919 L1957.74 72.576 L1950.29 72.576 L1950.29 45.4349 Q1950.29 38.994 1947.78 35.7938 Q1945.26 32.5936 1940.24 32.5936 Q1934.21 32.5936 1930.72 36.4419 Q1927.24 40.2903 1927.24 46.9338 L1927.24 72.576 L1919.74 72.576 L1919.74 27.2059 L1927.24 27.2059 L1927.24 34.2544 Q1929.91 30.163 1933.52 28.1376 Q1937.16 26.1121 1941.9 26.1121 Q1949.72 26.1121 1953.73 30.9732 Q1957.74 35.7938 1957.74 45.1919 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1974.51 62.2867 L1983.06 62.2867 L1983.06 72.576 L1974.51 72.576 L1974.51 62.2867 M1974.51 29.6769 L1983.06 29.6769 L1983.06 39.9662 L1974.51 39.9662 L1974.51 29.6769 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M2035.03 65.6895 L2063.59 65.6895 L2063.59 72.576 L2025.19 72.576 L2025.19 65.6895 Q2029.85 60.8689 2037.87 52.7671 Q2045.93 44.6248 2048 42.2752 Q2051.92 37.8598 2053.46 34.8216 Q2055.04 31.7429 2055.04 28.7857 Q2055.04 23.9651 2051.64 20.927 Q2048.28 17.8888 2042.85 17.8888 Q2039 17.8888 2034.71 19.2256 Q2030.45 20.5624 2025.59 23.2765 L2025.59 15.0127 Q2030.54 13.0277 2034.83 12.015 Q2039.12 11.0023 2042.69 11.0023 Q2052.09 11.0023 2057.68 15.7013 Q2063.27 20.4004 2063.27 28.2591 Q2063.27 31.9859 2061.85 35.3482 Q2060.47 38.6699 2056.79 43.2069 Q2055.77 44.3817 2050.34 50.0125 Q2044.92 55.6027 2035.03 65.6895 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip022)" style="stroke:#009af9; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" points="
  361.501,809.586 366.334,826.199 371.167,842.685 376,859.024 380.833,875.202 385.666,891.202 390.5,907.009 395.333,922.606 400.166,937.982 404.999,953.118 
  409.832,968.003 414.665,982.619 419.498,996.952 424.332,1010.99 429.165,1024.71 433.998,1038.11 438.831,1051.17 443.664,1063.87 448.497,1076.21 453.33,1088.17 
  458.164,1099.75 462.997,1110.92 467.83,1121.68 472.663,1132.02 477.496,1141.92 482.329,1151.38 487.162,1160.39 491.996,1168.93 496.829,1177 501.662,1184.6 
  506.495,1191.71 511.328,1198.32 516.161,1204.44 520.994,1210.05 525.828,1215.14 530.661,1219.72 535.494,1223.78 540.327,1227.32 545.16,1230.32 549.993,1232.8 
  554.826,1234.73 559.66,1236.14 564.493,1237 569.326,1237.33 574.159,1237.12 578.992,1236.36 583.825,1235.07 588.658,1233.25 593.492,1230.89 598.325,1228 
  603.158,1224.57 607.991,1220.62 612.824,1216.15 617.657,1211.16 622.49,1205.66 627.324,1199.65 632.157,1193.14 636.99,1186.14 641.823,1178.64 646.656,1170.67 
  651.489,1162.23 656.322,1153.32 661.156,1143.95 665.989,1134.14 670.822,1123.9 675.655,1113.23 680.488,1102.14 685.321,1090.66 690.154,1078.77 694.988,1066.51 
  699.821,1053.88 704.654,1040.9 709.487,1027.57 714.32,1013.92 719.153,999.944 723.986,985.671 728.82,971.11 733.653,956.277 738.486,941.186 743.319,925.852 
  748.152,910.291 752.985,894.518 757.818,878.548 762.652,862.398 767.485,846.085 772.318,829.624 777.151,813.031 781.984,796.324 786.817,779.518 791.65,762.631 
  796.484,745.68 801.317,728.682 806.15,711.654 810.983,694.613 815.816,677.575 820.649,660.558 825.482,643.579 830.316,626.655 835.149,609.803 839.982,593.039 
  844.815,576.382 849.648,559.846 854.481,543.449 859.314,527.207 864.148,511.137 868.981,495.254 873.814,479.574 878.647,464.114 883.48,448.887 888.313,433.911 
  893.147,419.199 897.98,404.767 902.813,390.628 907.646,376.797 912.479,363.288 917.312,350.114 922.145,337.289 926.979,324.825 931.812,312.734 936.645,301.029 
  941.478,289.722 946.311,278.823 951.144,268.343 955.977,258.295 960.811,248.686 965.644,239.527 970.477,230.827 975.31,222.594 980.143,214.837 984.976,207.564 
  989.809,200.782 994.643,194.498 999.476,188.717 1004.31,183.447 1009.14,178.691 1013.98,174.455 1018.81,170.742 1023.64,167.558 1028.47,164.905 1033.31,162.785 
  1038.14,161.201 1042.97,160.154 1047.81,159.645 1052.64,159.674 1057.47,160.243 1062.31,161.351 1067.14,162.995 1071.97,165.175 1076.81,167.888 1081.64,171.131 
  1086.47,174.9 1091.31,179.195 1096.14,184.009 1100.97,189.337 1105.8,195.174 1110.64,201.514 1115.47,208.35 1120.3,215.676 1125.14,223.486 1129.97,231.771 
  1134.8,240.522 1139.64,249.732 1144.47,259.389 1149.3,269.486 1154.14,280.01 1158.97,290.954 1163.8,302.305 1168.64,314.053 1173.47,326.185 1178.3,338.689 
  1183.14,351.552 1187.97,364.761 1192.8,378.305 1197.63,392.169 1202.47,406.339 1207.3,420.801 1212.13,435.541 1216.97,450.543 1221.8,465.793 1226.63,481.275 
  1231.47,496.975 1236.3,512.877 1241.13,528.964 1245.97,545.22 1250.8,561.629 1255.63,578.174 1260.47,594.84 1265.3,611.609 1270.13,628.466 1274.96,645.392 
  1279.8,662.371 1284.63,679.386 1289.46,696.42 1294.3,713.455 1299.13,730.476 1303.96,747.464 1308.8,764.404 1313.63,781.277 1318.46,798.068 1323.3,814.758 
  1328.13,831.332 1332.96,847.773 1337.8,864.065 1342.63,880.191 1347.46,896.135 1352.3,911.882 1357.13,927.415 1361.96,942.718 1366.79,957.778 1371.63,972.578 
  1376.46,987.103 1381.29,1001.34 1386.13,1015.27 1390.96,1028.89 1395.79,1042.18 1400.63,1055.12 1405.46,1067.71 1410.29,1079.92 1415.13,1091.76 1419.96,1103.2 
  1424.79,1114.24 1429.63,1124.86 1434.46,1135.05 1439.29,1144.81 1444.12,1154.12 1448.96,1162.97 1453.79,1171.37 1458.62,1179.28 1463.46,1186.72 1468.29,1193.67 
  1473.12,1200.12 1477.96,1206.07 1482.79,1211.52 1487.62,1216.45 1492.46,1220.86 1497.29,1224.75 1502.12,1228.12 1506.96,1230.95 1511.79,1233.25 1516.62,1235.02 
  1521.46,1236.25 1526.29,1236.94 1531.12,1237.09 1535.95,1236.71 1540.79,1235.79 1545.62,1234.33 1550.45,1232.33 1555.29,1229.8 1560.12,1226.74 1564.95,1223.15 
  1569.79,1219.03 1574.62,1214.4 1579.45,1209.25 1584.29,1203.58 1589.12,1197.42 1593.95,1190.75 1598.79,1183.59 1603.62,1175.95 1608.45,1167.83 1613.28,1159.24 
  1618.12,1150.19 1622.95,1140.68 1627.78,1130.74 1632.62,1120.36 1637.45,1109.56 1642.28,1098.35 1647.12,1086.74 1651.95,1074.74 1656.78,1062.37 1661.62,1049.63 
  1666.45,1036.54 1671.28,1023.11 1676.12,1009.36 1680.95,995.296 1685.78,980.936 1690.62,966.293 1695.45,951.383 1700.28,936.22 1705.11,920.819 1709.95,905.197 
  1714.78,889.366 1719.61,873.345 1724.45,857.149 1729.28,840.795 1734.11,824.299 1738.95,807.677 1743.78,790.946 1748.61,774.122 1753.45,757.222 1758.28,740.263 
  1763.11,723.263 1767.95,706.239 1772.78,689.206 1777.61,672.184 1782.44,655.187 1787.28,638.233 1792.11,621.34 1796.94,604.524 1801.78,587.802 1806.61,571.19 
  1811.44,554.707 1816.28,538.366 1821.11,522.186 1825.94,506.183 1830.78,490.371 1835.61,474.768 1840.44,459.388 1845.28,444.248 1850.11,429.362 1854.94,414.745 
  1859.78,400.412 1864.61,386.377 1869.44,372.654 1874.27,359.256 1879.11,346.198 1883.94,333.492 1888.77,321.152 1893.61,309.189 1898.44,297.615 1903.27,286.442 
  1908.11,275.681 1912.94,265.342 1917.77,255.437 1922.61,245.975 1927.44,236.965 1932.27,228.418 1937.11,220.34 1941.94,212.74 1946.77,205.625 1951.6,199.003 
  1956.44,192.882 1961.27,187.266 1966.1,182.161 1970.94,177.572 1975.77,173.504 1980.6,169.961 1985.44,166.945 1990.27,164.463 1995.1,162.514 1999.94,161.101 
  2004.77,160.226 2009.6,159.888 2014.44,160.089 2019.27,160.827 2024.1,162.104 2028.94,163.917 2033.77,166.266 2038.6,169.146 2043.43,172.555 2048.27,176.489 
  2053.1,180.945 2057.93,185.92 2062.77,191.407 2067.6,197.401 2072.43,203.896 2077.27,210.885 2082.1,218.362 2086.93,226.318 2091.77,234.746 2096.6,243.639 
  2101.43,252.987 2106.27,262.78 2111.1,273.008 2115.93,283.662 2120.77,294.729 2125.6,306.2 2130.43,318.065 2135.26,330.31 2140.1,342.922 2144.93,355.89 
  2149.76,369.2 2154.6,382.839 2159.43,396.794 2164.26,411.05 2169.1,425.595 2173.93,440.411 2178.76,455.486 2183.6,470.803 2188.43,486.348 2193.26,502.104 
  2198.1,518.057 2202.93,534.191 2207.76,550.489 2212.59,566.934 2217.43,583.511 2222.26,600.202 2227.09,616.991 2231.93,633.863 2236.76,650.798 2241.59,667.782 
  2246.43,684.795 2251.26,701.823 2256.09,718.846 2260.93,735.849 2265.76,752.814 2270.59,769.726 2275.43,786.565 2280.26,803.317 2285.09,819.963 2289.93,836.488 
  2294.76,852.875 
  "/>
<polyline clip-path="url(#clip022)" style="stroke:#e26f46; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" points="
  361.501,809.586 366.334,826.237 371.167,842.831 376,859.351 380.833,875.781 385.666,892.106 390.5,908.31 395.333,924.371 400.166,940.284 404.999,956.026 
  409.832,971.586 414.665,986.947 419.498,1002.09 424.332,1017.01 429.165,1031.68 433.998,1046.08 438.831,1060.21 443.664,1074.06 448.497,1087.6 453.33,1100.82 
  458.164,1113.71 462.997,1126.25 467.83,1138.45 472.663,1150.27 477.496,1161.71 482.329,1172.75 487.162,1183.4 491.996,1193.63 496.829,1203.43 501.662,1212.79 
  506.495,1221.71 511.328,1230.17 516.161,1238.17 520.994,1245.69 525.828,1252.73 530.661,1259.28 535.494,1265.33 540.327,1270.88 545.16,1275.91 549.993,1280.42 
  554.826,1284.41 559.66,1287.86 564.493,1290.77 569.326,1293.14 574.159,1294.96 578.992,1296.22 583.825,1296.92 588.658,1297.06 593.492,1296.74 598.325,1296.02 
  603.158,1294.89 607.991,1293.35 612.824,1291.39 617.657,1289.02 622.49,1286.23 627.324,1283.03 632.157,1279.41 636.99,1275.39 641.823,1270.95 646.656,1266.11 
  651.489,1260.86 656.322,1255.21 661.156,1249.17 665.989,1242.73 670.822,1235.92 675.655,1228.72 680.488,1221.15 685.321,1213.21 690.154,1204.92 694.988,1196.27 
  699.821,1187.28 704.654,1177.96 709.487,1168.31 714.32,1158.34 719.153,1148.06 723.986,1137.49 728.82,1126.63 733.653,1115.49 738.486,1104.09 743.319,1092.44 
  748.152,1080.54 752.985,1068.41 757.818,1056.06 762.652,1043.51 767.485,1030.76 772.318,1017.83 777.151,1004.73 781.984,991.477 786.817,978.081 791.65,964.556 
  796.484,950.915 801.317,937.172 806.15,923.342 810.983,909.436 815.816,895.47 820.649,881.458 825.482,867.413 830.316,853.349 835.149,839.281 839.982,825.222 
  844.815,811.187 849.648,797.188 854.481,783.241 859.314,769.358 864.148,755.556 868.981,741.847 873.814,728.247 878.647,714.764 883.48,701.414 888.313,688.212 
  893.147,675.171 897.98,662.305 902.813,649.627 907.646,637.15 912.479,624.882 917.312,612.839 922.145,601.034 926.979,589.478 931.812,578.184 936.645,567.163 
  941.478,556.427 946.311,545.985 951.144,535.848 955.977,526.028 960.811,516.534 965.644,507.377 970.477,498.566 975.31,490.11 980.143,482.019 984.976,474.301 
  989.809,466.965 994.643,460.018 999.476,453.468 1004.31,447.322 1009.14,441.589 1013.98,436.275 1018.81,431.387 1023.64,426.93 1028.47,422.912 1033.31,419.338 
  1038.14,416.214 1042.97,413.546 1047.81,411.339 1052.64,409.597 1057.47,408.327 1062.31,407.532 1067.14,407.217 1071.97,407.33 1076.81,407.764 1081.64,408.52 
  1086.47,409.603 1091.31,411.012 1096.14,412.749 1100.97,414.815 1105.8,417.209 1110.64,419.931 1115.47,422.98 1120.3,426.355 1125.14,430.054 1129.97,434.073 
  1134.8,438.409 1139.64,443.061 1144.47,448.028 1149.3,453.305 1154.14,458.888 1158.97,464.772 1163.8,470.952 1168.64,477.423 1173.47,484.178 1178.3,491.211 
  1183.14,498.516 1187.97,506.086 1192.8,513.913 1197.63,521.991 1202.47,530.312 1207.3,538.868 1212.13,547.651 1216.97,556.652 1221.8,565.863 1226.63,575.276 
  1231.47,584.88 1236.3,594.667 1241.13,604.627 1245.97,614.749 1250.8,625.025 1255.63,635.444 1260.47,645.996 1265.3,656.67 1270.13,667.455 1274.96,678.342 
  1279.8,689.318 1284.63,700.374 1289.46,711.498 1294.3,722.678 1299.13,733.905 1303.96,745.166 1308.8,756.451 1313.63,767.749 1318.46,779.047 1323.3,790.335 
  1328.13,801.601 1332.96,812.835 1337.8,824.025 1342.63,835.159 1347.46,846.227 1352.3,857.218 1357.13,868.12 1361.96,878.922 1366.79,889.614 1371.63,900.184 
  1376.46,910.623 1381.29,920.919 1386.13,931.062 1390.96,941.041 1395.79,950.847 1400.63,960.469 1405.46,969.898 1410.29,979.124 1415.13,988.137 1419.96,996.928 
  1424.79,1005.49 1429.63,1013.81 1434.46,1021.88 1439.29,1029.69 1444.12,1037.24 1448.96,1044.51 1453.79,1051.5 1458.62,1058.21 1463.46,1064.61 1468.29,1070.72 
  1473.12,1076.51 1477.96,1081.98 1482.79,1087.13 1487.62,1091.95 1492.46,1096.43 1497.29,1100.57 1502.12,1104.36 1506.96,1107.8 1511.79,1110.88 1516.62,1113.6 
  1521.46,1115.94 1526.29,1117.92 1531.12,1119.52 1535.95,1120.73 1540.79,1121.56 1545.62,1122 1550.45,1122.06 1555.29,1121.86 1560.12,1121.42 1564.95,1120.76 
  1569.79,1119.85 1574.62,1118.72 1579.45,1117.34 1584.29,1115.73 1589.12,1113.89 1593.95,1111.8 1598.79,1109.48 1603.62,1106.93 1608.45,1104.14 1613.28,1101.12 
  1618.12,1097.87 1622.95,1094.4 1627.78,1090.69 1632.62,1086.76 1637.45,1082.62 1642.28,1078.25 1647.12,1073.67 1651.95,1068.88 1656.78,1063.89 1661.62,1058.69 
  1666.45,1053.3 1671.28,1047.71 1676.12,1041.94 1680.95,1035.98 1685.78,1029.84 1690.62,1023.54 1695.45,1017.07 1700.28,1010.44 1705.11,1003.65 1709.95,996.72 
  1714.78,989.649 1719.61,982.445 1724.45,975.114 1729.28,967.664 1734.11,960.102 1738.95,952.435 1743.78,944.67 1748.61,936.816 1753.45,928.88 1758.28,920.87 
  1763.11,912.793 1767.95,904.658 1772.78,896.472 1777.61,888.244 1782.44,879.982 1787.28,871.693 1792.11,863.386 1796.94,855.07 1801.78,846.753 1806.61,838.444 
  1811.44,830.149 1816.28,821.878 1821.11,813.638 1825.94,805.438 1830.78,797.288 1835.61,789.195 1840.44,781.169 1845.28,773.215 1850.11,765.341 1854.94,757.557 
  1859.78,749.872 1864.61,742.291 1869.44,734.825 1874.27,727.481 1879.11,720.264 1883.94,713.184 1888.77,706.248 1893.61,699.462 1898.44,692.835 1903.27,686.373 
  1908.11,680.083 1912.94,673.972 1917.77,668.046 1922.61,662.313 1927.44,656.777 1932.27,651.446 1937.11,646.324 1941.94,641.419 1946.77,636.736 1951.6,632.281 
  1956.44,628.058 1961.27,624.073 1966.1,620.331 1970.94,616.836 1975.77,613.595 1980.6,610.61 1985.44,607.887 1990.27,605.43 1995.1,603.242 1999.94,601.327 
  2004.77,599.69 2009.6,598.333 2014.44,597.261 2019.27,596.476 2024.1,595.982 2028.94,595.782 2033.77,595.826 2038.6,596.001 2043.43,596.308 2048.27,596.747 
  2053.1,597.321 2057.93,598.029 2062.77,598.872 2067.6,599.85 2072.43,600.963 2077.27,602.211 2082.1,603.594 2086.93,605.112 2091.77,606.764 2096.6,608.55 
  2101.43,610.467 2106.27,612.515 2111.1,614.694 2115.93,617.001 2120.77,619.436 2125.6,621.995 2130.43,624.678 2135.26,627.484 2140.1,630.409 2144.93,633.45 
  2149.76,636.608 2154.6,639.879 2159.43,643.258 2164.26,646.744 2169.1,650.335 2173.93,654.027 2178.76,657.815 2183.6,661.698 2188.43,665.673 2193.26,669.734 
  2198.1,673.879 2202.93,678.103 2207.76,682.404 2212.59,686.776 2217.43,691.216 2222.26,695.72 2227.09,700.284 2231.93,704.901 2236.76,709.57 2241.59,714.285 
  2246.43,719.041 2251.26,723.834 2256.09,728.66 2260.93,733.513 2265.76,738.389 2270.59,743.283 2275.43,748.19 2280.26,753.106 2285.09,758.024 2289.93,762.942 
  2294.76,767.853 
  "/>
<polyline clip-path="url(#clip022)" style="stroke:#008000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" stroke-dasharray="4, 8" points="
  361.501,1365.41 366.334,1365.01 371.167,1364.1 376,1362.67 380.833,1360.73 385.666,1358.28 390.5,1355.3 395.333,1351.82 400.166,1347.82 404.999,1343.31 
  409.832,1338.3 414.665,1332.79 419.498,1326.77 424.332,1320.27 429.165,1313.28 433.998,1305.82 438.831,1297.88 443.664,1289.48 448.497,1280.63 453.33,1271.33 
  458.164,1261.6 462.997,1251.44 467.83,1240.88 472.663,1229.91 477.496,1218.56 482.329,1206.83 487.162,1194.75 491.996,1182.31 496.829,1169.55 501.662,1156.47 
  506.495,1143.09 511.328,1129.42 516.161,1115.49 520.994,1101.3 525.828,1086.88 530.661,1072.24 535.494,1057.4 540.327,1042.38 545.16,1027.19 549.993,1011.85 
  554.826,996.373 559.66,980.787 564.493,965.105 569.326,949.343 574.159,933.52 578.992,917.654 583.825,901.761 588.658,885.858 593.492,869.963 598.325,854.093 
  603.158,838.264 607.991,822.493 612.824,806.796 617.657,791.188 622.49,775.685 627.324,760.302 632.157,745.054 636.99,729.954 641.823,715.017 646.656,700.256 
  651.489,685.684 656.322,671.315 661.156,657.161 665.989,643.234 670.822,629.545 675.655,616.105 680.488,602.924 685.321,590.013 690.154,577.383 694.988,565.041 
  699.821,552.997 704.654,541.261 709.487,529.84 714.32,518.743 719.153,507.977 723.986,497.549 728.82,487.467 733.653,477.736 738.486,468.364 743.319,459.356 
  748.152,450.717 752.985,442.451 757.818,434.563 762.652,427.06 767.485,419.945 772.318,413.226 777.151,406.906 781.984,400.989 786.817,395.48 791.65,390.382 
  796.484,385.697 801.317,381.429 806.15,377.579 810.983,374.149 815.816,371.139 820.649,368.552 825.482,366.387 830.316,364.643 835.149,363.321 839.982,362.42 
  844.815,361.937 849.648,361.871 854.481,362.219 859.314,362.98 864.148,364.149 868.981,365.723 873.814,367.699 878.647,370.074 883.48,372.845 888.313,376.006 
  893.147,379.55 897.98,383.473 902.813,387.769 907.646,392.43 912.479,397.45 917.312,402.822 922.145,408.539 926.979,414.593 931.812,420.976 936.645,427.679 
  941.478,434.695 946.311,442.013 951.144,449.626 955.977,457.523 960.811,465.694 965.644,474.13 970.477,482.823 975.31,491.763 980.143,500.939 984.976,510.342 
  989.809,519.961 994.643,529.785 999.476,539.805 1004.31,550.01 1009.14,560.39 1013.98,570.935 1018.81,581.635 1023.64,592.48 1028.47,603.461 1033.31,614.567 
  1038.14,625.789 1042.97,637.117 1047.81,648.542 1052.64,660.054 1057.47,671.644 1062.31,683.302 1067.14,695.019 1071.97,706.786 1076.81,718.594 1081.64,730.433 
  1086.47,742.294 1091.31,754.169 1096.14,766.049 1100.97,777.925 1105.8,789.787 1110.64,801.628 1115.47,813.438 1120.3,825.208 1125.14,836.931 1129.97,848.595 
  1134.8,860.194 1139.64,871.718 1144.47,883.157 1149.3,894.504 1154.14,905.748 1158.97,916.881 1163.8,927.894 1168.64,938.778 1173.47,949.523 1178.3,960.121 
  1183.14,970.562 1187.97,980.837 1192.8,990.936 1197.63,1000.85 1202.47,1010.57 1207.3,1020.09 1212.13,1029.4 1216.97,1038.5 1221.8,1047.38 1226.63,1056.03 
  1231.47,1064.44 1236.3,1072.58 1241.13,1080.47 1245.97,1088.07 1250.8,1095.39 1255.63,1102.41 1260.47,1109.13 1265.3,1115.53 1270.13,1121.62 1274.96,1127.37 
  1279.8,1132.79 1284.63,1137.87 1289.46,1142.6 1294.3,1146.97 1299.13,1150.99 1303.96,1154.64 1308.8,1157.92 1313.63,1160.82 1318.46,1163.36 1323.3,1165.51 
  1328.13,1167.28 1332.96,1168.66 1337.8,1169.66 1342.63,1170.27 1347.46,1170.49 1352.3,1170.33 1357.13,1169.77 1361.96,1168.83 1366.79,1167.5 1371.63,1165.79 
  1376.46,1163.68 1381.29,1161.19 1386.13,1158.32 1390.96,1155.06 1395.79,1151.42 1400.63,1147.41 1405.46,1143.03 1410.29,1138.29 1415.13,1133.19 1419.96,1127.74 
  1424.79,1121.94 1429.63,1115.81 1434.46,1109.34 1439.29,1102.56 1444.12,1095.46 1448.96,1088.06 1453.79,1080.36 1458.62,1072.38 1463.46,1064.12 1468.29,1055.61 
  1473.12,1046.84 1477.96,1037.84 1482.79,1028.61 1487.62,1019.17 1492.46,1009.52 1497.29,999.679 1502.12,989.662 1506.96,979.483 1511.79,969.155 1516.62,958.691 
  1521.46,948.104 1526.29,937.41 1531.12,926.619 1535.95,915.747 1540.79,904.806 1545.62,893.81 1550.45,882.771 1555.29,871.703 1560.12,860.62 1564.95,849.533 
  1569.79,838.457 1574.62,827.403 1579.45,816.386 1584.29,805.417 1589.12,794.51 1593.95,783.677 1598.79,772.931 1603.62,762.285 1608.45,751.751 1613.28,741.341 
  1618.12,731.069 1622.95,720.942 1627.78,710.967 1632.62,701.153 1637.45,691.51 1642.28,682.048 1647.12,672.776 1651.95,663.704 1656.78,654.839 1661.62,646.189 
  1666.45,637.763 1671.28,629.568 1676.12,621.612 1680.95,613.9 1685.78,606.439 1690.62,599.236 1695.45,592.295 1700.28,585.624 1705.11,579.225 1709.95,573.105 
  1714.78,567.267 1719.61,561.715 1724.45,556.453 1729.28,551.484 1734.11,546.812 1738.95,542.438 1743.78,538.363 1748.61,534.592 1753.45,531.126 1758.28,527.967 
  1763.11,525.117 1767.95,522.579 1772.78,520.353 1777.61,518.439 1782.44,516.84 1787.28,515.553 1792.11,514.58 1796.94,513.92 1801.78,513.571 1806.61,513.533 
  1811.44,513.803 1816.28,514.379 1821.11,515.26 1825.94,516.441 1830.78,517.921 1835.61,519.694 1840.44,521.758 1845.28,524.109 1850.11,526.74 1854.94,529.648 
  1859.78,532.827 1864.61,536.271 1869.44,539.975 1874.27,543.931 1879.11,548.132 1883.94,552.573 1888.77,557.247 1893.61,562.153 1898.44,567.284 1903.27,572.632 
  1908.11,578.19 1912.94,583.951 1917.77,589.909 1922.61,596.055 1927.44,602.382 1932.27,608.884 1937.11,615.552 1941.94,622.379 1946.77,629.357 1951.6,636.48 
  1956.44,643.738 1961.27,651.124 1966.1,658.63 1970.94,666.249 1975.77,673.971 1980.6,681.79 1985.44,689.697 1990.27,697.684 1995.1,705.742 1999.94,713.863 
  2004.77,722.041 2009.6,730.267 2014.44,738.534 2019.27,746.833 2024.1,755.157 2028.94,763.497 2033.77,771.845 2038.6,780.194 2043.43,788.536 2048.27,796.862 
  2053.1,805.166 2057.93,813.439 2062.77,821.673 2067.6,829.861 2072.43,837.996 2077.27,846.069 2082.1,854.073 2086.93,862.001 2091.77,869.845 2096.6,877.597 
  2101.43,885.251 2106.27,892.798 2111.1,900.232 2115.93,907.545 2120.77,914.73 2125.6,921.78 2130.43,928.687 2135.26,935.445 2140.1,942.045 2144.93,948.486 
  2149.76,954.762 2154.6,960.864 2159.43,966.785 2164.26,972.519 2169.1,978.057 2173.93,983.393 2178.76,988.521 2183.6,993.435 2188.43,998.128 2193.26,1002.6 
  2198.1,1006.83 2202.93,1010.84 2207.76,1014.6 2212.59,1018.12 2217.43,1021.38 2222.26,1024.4 2227.09,1027.17 2231.93,1029.67 2236.76,1031.91 2241.59,1033.89 
  2246.43,1035.61 2251.26,1037.06 2256.09,1038.24 2260.93,1039.15 2265.76,1039.79 2270.59,1040.16 2275.43,1040.26 2280.26,1040.08 2285.09,1039.64 2289.93,1038.92 
  2294.76,1037.94 
  "/>
<polyline clip-path="url(#clip022)" style="stroke:#008000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" points="
  361.501,1365.41 366.334,1364.7 371.167,1363.47 376,1361.71 380.833,1359.42 385.666,1356.61 390.5,1353.28 395.333,1349.44 400.166,1345.09 404.999,1340.23 
  409.832,1334.88 414.665,1329.03 419.498,1322.7 424.332,1315.89 429.165,1308.61 433.998,1300.87 438.831,1292.68 443.664,1284.05 448.497,1274.99 453.33,1265.51 
  458.164,1255.62 462.997,1245.33 467.83,1234.66 472.663,1223.63 477.496,1212.23 482.329,1200.49 487.162,1188.41 491.996,1176.03 496.829,1163.33 501.662,1150.36 
  506.495,1137.1 511.328,1123.6 516.161,1109.85 520.994,1095.87 525.828,1081.69 530.661,1067.32 535.494,1052.76 540.327,1038.05 545.16,1023.19 549.993,1008.21 
  554.826,993.112 559.66,977.919 564.493,962.646 569.326,947.309 574.159,931.925 578.992,916.51 583.825,901.078 588.658,885.646 593.492,870.228 598.325,854.838 
  603.158,839.492 607.991,824.203 612.824,808.986 617.657,793.852 622.49,778.817 627.324,763.891 632.157,749.087 636.99,734.418 641.823,719.894 646.656,705.527 
  651.489,691.327 656.322,677.305 661.156,663.472 665.989,649.838 670.822,636.411 675.655,623.201 680.488,610.217 685.321,597.467 690.154,584.961 694.988,572.707 
  699.821,560.713 704.654,548.988 709.487,537.539 714.32,526.376 719.153,515.505 723.986,504.935 728.82,494.672 733.653,484.725 738.486,475.101 743.319,465.806 
  748.152,456.849 752.985,448.235 757.818,439.969 762.652,432.058 767.485,424.511 772.318,417.335 777.151,410.535 781.984,404.119 786.817,398.092 791.65,392.46 
  796.484,387.229 801.317,382.402 806.15,377.984 810.983,373.979 815.816,370.39 820.649,367.221 825.482,364.474 830.316,362.151 835.149,360.255 839.982,358.787 
  844.815,357.747 849.648,357.137 854.481,356.956 859.314,357.205 864.148,357.883 868.981,358.99 873.814,360.523 878.647,362.481 883.48,364.861 888.313,367.66 
  893.147,370.874 897.98,374.501 902.813,378.534 907.646,382.97 912.479,387.803 917.312,393.027 922.145,398.635 926.979,404.622 931.812,410.979 936.645,417.7 
  941.478,424.776 946.311,432.197 951.144,439.957 955.977,448.043 960.811,456.448 965.644,465.163 970.477,474.177 975.31,483.479 980.143,493.058 984.976,502.901 
  989.809,512.998 994.643,523.337 999.476,533.905 1004.31,544.69 1009.14,555.682 1013.98,566.866 1018.81,578.232 1023.64,589.766 1028.47,601.456 1033.31,613.29 
  1038.14,625.255 1042.97,637.337 1047.81,649.525 1052.64,661.804 1057.47,674.162 1062.31,686.586 1067.14,699.062 1071.97,711.577 1076.81,724.118 1081.64,736.675 
  1086.47,749.236 1091.31,761.788 1096.14,774.32 1100.97,786.819 1105.8,799.274 1110.64,811.674 1115.47,824.008 1120.3,836.263 1125.14,848.431 1129.97,860.498 
  1134.8,872.456 1139.64,884.294 1144.47,896.001 1149.3,907.568 1154.14,918.985 1158.97,930.241 1163.8,941.328 1168.64,952.236 1173.47,962.955 1178.3,973.478 
  1183.14,983.795 1187.97,993.897 1192.8,1003.78 1197.63,1013.43 1202.47,1022.86 1207.3,1032.03 1212.13,1040.95 1216.97,1049.61 1221.8,1058 1226.63,1066.1 
  1231.47,1073.92 1236.3,1081.45 1241.13,1088.67 1245.97,1095.59 1250.8,1102.19 1255.63,1108.47 1260.47,1114.42 1265.3,1120.04 1270.13,1125.33 1274.96,1130.27 
  1279.8,1134.87 1284.63,1139.12 1289.46,1143.02 1294.3,1146.56 1299.13,1149.74 1303.96,1152.57 1308.8,1155.02 1313.63,1157.12 1318.46,1158.85 1323.3,1160.21 
  1328.13,1161.2 1332.96,1161.83 1337.8,1162.08 1342.63,1161.97 1347.46,1161.49 1352.3,1160.64 1357.13,1159.42 1361.96,1157.84 1366.79,1155.89 1371.63,1153.59 
  1376.46,1150.93 1381.29,1147.92 1386.13,1144.57 1390.96,1140.87 1395.79,1136.83 1400.63,1132.45 1405.46,1127.76 1410.29,1122.74 1415.13,1117.41 1419.96,1111.77 
  1424.79,1105.83 1429.63,1099.6 1434.46,1093.1 1439.29,1086.31 1444.12,1079.27 1448.96,1071.97 1453.79,1064.42 1458.62,1056.64 1463.46,1048.63 1468.29,1040.4 
  1473.12,1031.97 1477.96,1023.34 1482.79,1014.52 1487.62,1005.54 1492.46,996.386 1497.29,987.083 1502.12,977.639 1506.96,968.064 1511.79,958.369 1516.62,948.566 
  1521.46,938.665 1526.29,928.677 1531.12,918.613 1535.95,908.484 1540.79,898.302 1545.62,888.077 1550.45,877.82 1555.29,867.543 1560.12,857.257 1564.95,846.973 
  1569.79,836.702 1574.62,826.456 1579.45,816.244 1584.29,806.073 1589.12,795.954 1593.95,785.896 1598.79,775.908 1603.62,766 1608.45,756.18 1613.28,746.458 
  1618.12,736.842 1622.95,727.341 1627.78,717.962 1632.62,708.715 1637.45,699.606 1642.28,690.643 1647.12,681.834 1651.95,673.187 1656.78,664.707 1661.62,656.403 
  1666.45,648.281 1671.28,640.347 1676.12,632.609 1680.95,625.07 1685.78,617.739 1690.62,610.621 1695.45,603.72 1700.28,597.044 1705.11,590.596 1709.95,584.38 
  1714.78,578.4 1719.61,572.661 1724.45,567.17 1729.28,561.931 1734.11,556.949 1738.95,552.229 1743.78,547.775 1748.61,543.592 1753.45,539.682 1758.28,536.049 
  1763.11,532.697 1767.95,529.628 1772.78,526.843 1777.61,524.346 1782.44,522.139 1787.28,520.221 1792.11,518.595 1796.94,517.262 1801.78,516.221 1806.61,515.474 
  1811.44,515.019 1816.28,514.856 1821.11,514.985 1825.94,515.404 1830.78,516.112 1835.61,517.107 1840.44,518.387 1845.28,519.95 1850.11,521.792 1854.94,523.911 
  1859.78,526.304 1864.61,528.968 1869.44,531.902 1874.27,535.099 1879.11,538.557 1883.94,542.27 1888.77,546.233 1893.61,550.442 1898.44,554.89 1903.27,559.574 
  1908.11,564.485 1912.94,569.62 1917.77,574.971 1922.61,580.532 1927.44,586.296 1932.27,592.256 1937.11,598.406 1941.94,604.737 1946.77,611.243 1951.6,617.915 
  1956.44,624.745 1961.27,631.725 1966.1,638.846 1970.94,646.101 1975.77,653.479 1980.6,660.971 1985.44,668.571 1990.27,676.27 1995.1,684.06 1999.94,691.933 
  2004.77,699.878 2009.6,707.888 2014.44,715.953 2019.27,724.064 2024.1,732.213 2028.94,740.391 2033.77,748.59 2038.6,756.801 2043.43,765.016 2048.27,773.226 
  2053.1,781.423 2057.93,789.599 2062.77,797.745 2067.6,805.854 2072.43,813.918 2077.27,821.928 2082.1,829.877 2086.93,837.757 2091.77,845.56 2096.6,853.278 
  2101.43,860.905 2106.27,868.432 2111.1,875.852 2115.93,883.157 2120.77,890.342 2125.6,897.403 2130.43,904.333 2135.26,911.124 2140.1,917.77 2144.93,924.264 
  2149.76,930.6 2154.6,936.771 2159.43,942.772 2164.26,948.597 2169.1,954.24 2173.93,959.697 2178.76,964.962 2183.6,970.03 2188.43,974.898 2193.26,979.561 
  2198.1,984.014 2202.93,988.255 2207.76,992.279 2212.59,996.085 2217.43,999.667 2222.26,1003.02 2227.09,1006.15 2231.93,1009.05 2236.76,1011.72 2241.59,1014.15 
  2246.43,1016.35 2251.26,1018.31 2256.09,1020.03 2260.93,1021.52 2265.76,1022.76 2270.59,1023.76 2275.43,1024.52 2280.26,1025.04 2285.09,1025.31 2289.93,1025.35 
  2294.76,1025.14 
  "/>
<path clip-path="url(#clip020)" d="
M1421.89 554.876 L2284.45 554.876 L2284.45 166.076 L1421.89 166.076  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<polyline clip-path="url(#clip020)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1421.89,554.876 2284.45,554.876 2284.45,166.076 1421.89,166.076 1421.89,554.876 
  "/>
<polyline clip-path="url(#clip020)" style="stroke:#009af9; stroke-linecap:butt; stroke-linejoin:round; stroke-width:6; stroke-opacity:1; fill:none" points="
  1444.66,243.836 1581.28,243.836 
  "/>
<path clip-path="url(#clip020)" d="M1637.42 219.617 L1637.42 226.457 Q1633.43 224.548 1629.88 223.61 Q1626.34 222.673 1623.04 222.673 Q1617.31 222.673 1614.19 224.895 Q1611.1 227.117 1611.1 231.214 Q1611.1 234.652 1613.15 236.423 Q1615.23 238.159 1621 239.235 L1625.23 240.103 Q1633.08 241.596 1636.79 245.381 Q1640.54 249.131 1640.54 255.45 Q1640.54 262.985 1635.47 266.874 Q1630.44 270.763 1620.68 270.763 Q1617 270.763 1612.84 269.929 Q1608.7 269.096 1604.26 267.464 L1604.26 260.242 Q1608.53 262.638 1612.63 263.853 Q1616.72 265.068 1620.68 265.068 Q1626.69 265.068 1629.95 262.707 Q1633.22 260.346 1633.22 255.971 Q1633.22 252.152 1630.86 249.999 Q1628.53 247.846 1623.18 246.77 L1618.91 245.936 Q1611.06 244.374 1607.56 241.041 Q1604.05 237.707 1604.05 231.77 Q1604.05 224.895 1608.88 220.937 Q1613.74 216.978 1622.25 216.978 Q1625.89 216.978 1629.68 217.638 Q1633.46 218.298 1637.42 219.617 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1651.2 230.867 L1657.59 230.867 L1657.59 269.756 L1651.2 269.756 L1651.2 230.867 M1651.2 215.728 L1657.59 215.728 L1657.59 223.818 L1651.2 223.818 L1651.2 215.728 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1701.24 238.332 Q1703.63 234.027 1706.97 231.978 Q1710.3 229.93 1714.81 229.93 Q1720.89 229.93 1724.19 234.2 Q1727.49 238.436 1727.49 246.284 L1727.49 269.756 L1721.06 269.756 L1721.06 246.492 Q1721.06 240.902 1719.08 238.193 Q1717.11 235.485 1713.04 235.485 Q1708.08 235.485 1705.2 238.784 Q1702.31 242.082 1702.31 247.777 L1702.31 269.756 L1695.89 269.756 L1695.89 246.492 Q1695.89 240.867 1693.91 238.193 Q1691.93 235.485 1687.8 235.485 Q1682.9 235.485 1680.02 238.818 Q1677.14 242.117 1677.14 247.777 L1677.14 269.756 L1670.72 269.756 L1670.72 230.867 L1677.14 230.867 L1677.14 236.909 Q1679.33 233.332 1682.38 231.631 Q1685.44 229.93 1689.64 229.93 Q1693.88 229.93 1696.83 232.082 Q1699.81 234.235 1701.24 238.332 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1746.41 263.922 L1746.41 284.547 L1739.99 284.547 L1739.99 230.867 L1746.41 230.867 L1746.41 236.77 Q1748.43 233.298 1751.48 231.631 Q1754.57 229.93 1758.84 229.93 Q1765.92 229.93 1770.33 235.555 Q1774.78 241.179 1774.78 250.346 Q1774.78 259.513 1770.33 265.138 Q1765.92 270.763 1758.84 270.763 Q1754.57 270.763 1751.48 269.096 Q1748.43 267.395 1746.41 263.922 M1768.15 250.346 Q1768.15 243.298 1765.23 239.304 Q1762.35 235.277 1757.28 235.277 Q1752.21 235.277 1749.29 239.304 Q1746.41 243.298 1746.41 250.346 Q1746.41 257.395 1749.29 261.422 Q1752.21 265.415 1757.28 265.415 Q1762.35 265.415 1765.23 261.422 Q1768.15 257.395 1768.15 250.346 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1785.37 215.728 L1791.76 215.728 L1791.76 269.756 L1785.37 269.756 L1785.37 215.728 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1838.39 248.714 L1838.39 251.839 L1809.01 251.839 Q1809.43 258.436 1812.97 261.909 Q1816.55 265.346 1822.9 265.346 Q1826.58 265.346 1830.02 264.443 Q1833.49 263.54 1836.9 261.735 L1836.9 267.777 Q1833.46 269.235 1829.85 269.999 Q1826.24 270.763 1822.52 270.763 Q1813.22 270.763 1807.76 265.346 Q1802.35 259.929 1802.35 250.693 Q1802.35 241.145 1807.49 235.555 Q1812.66 229.93 1821.41 229.93 Q1829.26 229.93 1833.81 234.999 Q1838.39 240.034 1838.39 248.714 M1832 246.839 Q1831.93 241.596 1829.05 238.471 Q1826.2 235.346 1821.48 235.346 Q1816.13 235.346 1812.9 238.367 Q1809.71 241.388 1809.22 246.874 L1832 246.839 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1849.15 217.916 L1878.95 217.916 L1878.95 223.818 L1856.17 223.818 L1856.17 239.096 L1876.72 239.096 L1876.72 244.999 L1856.17 244.999 L1856.17 269.756 L1849.15 269.756 L1849.15 217.916 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1890.06 217.916 L1900.51 217.916 L1913.74 253.193 L1927.04 217.916 L1937.49 217.916 L1937.49 269.756 L1930.65 269.756 L1930.65 224.235 L1917.28 259.79 L1910.23 259.79 L1896.86 224.235 L1896.86 269.756 L1890.06 269.756 L1890.06 217.916 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1950.61 217.916 L1957.66 217.916 L1957.66 249.409 Q1957.66 257.742 1960.68 261.422 Q1963.7 265.068 1970.47 265.068 Q1977.21 265.068 1980.23 261.422 Q1983.25 257.742 1983.25 249.409 L1983.25 217.916 L1990.3 217.916 L1990.3 250.277 Q1990.3 260.415 1985.26 265.589 Q1980.26 270.763 1970.47 270.763 Q1960.65 270.763 1955.61 265.589 Q1950.61 260.415 1950.61 250.277 L1950.61 217.916 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip020)" style="stroke:#e26f46; stroke-linecap:butt; stroke-linejoin:round; stroke-width:6; stroke-opacity:1; fill:none" points="
  1444.66,321.596 1581.28,321.596 
  "/>
<path clip-path="url(#clip020)" d="M1628.63 323.21 Q1630.89 323.974 1633.01 326.474 Q1635.16 328.974 1637.31 333.349 L1644.43 347.516 L1636.9 347.516 L1630.27 334.217 Q1627.7 329.009 1625.27 327.307 Q1622.87 325.606 1618.7 325.606 L1611.06 325.606 L1611.06 347.516 L1604.05 347.516 L1604.05 295.676 L1619.88 295.676 Q1628.77 295.676 1633.15 299.391 Q1637.52 303.106 1637.52 310.606 Q1637.52 315.502 1635.23 318.731 Q1632.97 321.96 1628.63 323.21 M1611.06 301.44 L1611.06 319.842 L1619.88 319.842 Q1624.95 319.842 1627.52 317.516 Q1630.13 315.155 1630.13 310.606 Q1630.13 306.058 1627.52 303.766 Q1624.95 301.44 1619.88 301.44 L1611.06 301.44 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1683.25 326.474 L1683.25 329.599 L1653.88 329.599 Q1654.29 336.196 1657.84 339.669 Q1661.41 343.106 1667.77 343.106 Q1671.45 343.106 1674.88 342.203 Q1678.36 341.3 1681.76 339.495 L1681.76 345.537 Q1678.32 346.995 1674.71 347.759 Q1671.1 348.523 1667.38 348.523 Q1658.08 348.523 1652.63 343.106 Q1647.21 337.689 1647.21 328.453 Q1647.21 318.905 1652.35 313.315 Q1657.52 307.69 1666.27 307.69 Q1674.12 307.69 1678.67 312.759 Q1683.25 317.794 1683.25 326.474 M1676.86 324.599 Q1676.79 319.356 1673.91 316.231 Q1671.06 313.106 1666.34 313.106 Q1660.99 313.106 1657.77 316.127 Q1654.57 319.148 1654.09 324.634 L1676.86 324.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1713.43 293.488 L1713.43 298.801 L1707.31 298.801 Q1703.88 298.801 1702.52 300.19 Q1701.2 301.578 1701.2 305.19 L1701.2 308.627 L1711.72 308.627 L1711.72 313.592 L1701.2 313.592 L1701.2 347.516 L1694.78 347.516 L1694.78 313.592 L1688.67 313.592 L1688.67 308.627 L1694.78 308.627 L1694.78 305.919 Q1694.78 299.426 1697.8 296.474 Q1700.82 293.488 1707.38 293.488 L1713.43 293.488 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1752.04 326.474 L1752.04 329.599 L1722.66 329.599 Q1723.08 336.196 1726.62 339.669 Q1730.2 343.106 1736.55 343.106 Q1740.23 343.106 1743.67 342.203 Q1747.14 341.3 1750.54 339.495 L1750.54 345.537 Q1747.11 346.995 1743.49 347.759 Q1739.88 348.523 1736.17 348.523 Q1726.86 348.523 1721.41 343.106 Q1715.99 337.689 1715.99 328.453 Q1715.99 318.905 1721.13 313.315 Q1726.31 307.69 1735.06 307.69 Q1742.9 307.69 1747.45 312.759 Q1752.04 317.794 1752.04 326.474 M1745.65 324.599 Q1745.58 319.356 1742.7 316.231 Q1739.85 313.106 1735.13 313.106 Q1729.78 313.106 1726.55 316.127 Q1723.36 319.148 1722.87 324.634 L1745.65 324.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1785.06 314.599 Q1783.98 313.974 1782.7 313.696 Q1781.45 313.384 1779.92 313.384 Q1774.5 313.384 1771.58 316.926 Q1768.7 320.433 1768.7 327.03 L1768.7 347.516 L1762.28 347.516 L1762.28 308.627 L1768.7 308.627 L1768.7 314.669 Q1770.72 311.127 1773.95 309.426 Q1777.17 307.69 1781.79 307.69 Q1782.45 307.69 1783.25 307.794 Q1784.05 307.863 1785.02 308.037 L1785.06 314.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1823.46 326.474 L1823.46 329.599 L1794.08 329.599 Q1794.5 336.196 1798.04 339.669 Q1801.62 343.106 1807.97 343.106 Q1811.65 343.106 1815.09 342.203 Q1818.56 341.3 1821.97 339.495 L1821.97 345.537 Q1818.53 346.995 1814.92 347.759 Q1811.31 348.523 1807.59 348.523 Q1798.29 348.523 1792.83 343.106 Q1787.42 337.689 1787.42 328.453 Q1787.42 318.905 1792.56 313.315 Q1797.73 307.69 1806.48 307.69 Q1814.33 307.69 1818.88 312.759 Q1823.46 317.794 1823.46 326.474 M1817.07 324.599 Q1817 319.356 1814.12 316.231 Q1811.27 313.106 1806.55 313.106 Q1801.2 313.106 1797.97 316.127 Q1794.78 319.148 1794.29 324.634 L1817.07 324.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1866.27 324.044 L1866.27 347.516 L1859.88 347.516 L1859.88 324.252 Q1859.88 318.731 1857.73 315.988 Q1855.58 313.245 1851.27 313.245 Q1846.1 313.245 1843.11 316.544 Q1840.13 319.842 1840.13 325.537 L1840.13 347.516 L1833.7 347.516 L1833.7 308.627 L1840.13 308.627 L1840.13 314.669 Q1842.42 311.162 1845.51 309.426 Q1848.63 307.69 1852.7 307.69 Q1859.4 307.69 1862.83 311.856 Q1866.27 315.988 1866.27 324.044 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1907 310.12 L1907 316.092 Q1904.29 314.599 1901.55 313.87 Q1898.84 313.106 1896.06 313.106 Q1889.85 313.106 1886.41 317.064 Q1882.97 320.988 1882.97 328.106 Q1882.97 335.224 1886.41 339.182 Q1889.85 343.106 1896.06 343.106 Q1898.84 343.106 1901.55 342.377 Q1904.29 341.613 1907 340.12 L1907 346.023 Q1904.33 347.273 1901.45 347.898 Q1898.6 348.523 1895.37 348.523 Q1886.58 348.523 1881.41 343.002 Q1876.24 337.481 1876.24 328.106 Q1876.24 318.592 1881.45 313.141 Q1886.69 307.69 1895.79 307.69 Q1898.74 307.69 1901.55 308.315 Q1904.36 308.905 1907 310.12 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1951.38 326.474 L1951.38 329.599 L1922 329.599 Q1922.42 336.196 1925.96 339.669 Q1929.54 343.106 1935.89 343.106 Q1939.57 343.106 1943.01 342.203 Q1946.48 341.3 1949.88 339.495 L1949.88 345.537 Q1946.44 346.995 1942.83 347.759 Q1939.22 348.523 1935.51 348.523 Q1926.2 348.523 1920.75 343.106 Q1915.33 337.689 1915.33 328.453 Q1915.33 318.905 1920.47 313.315 Q1925.65 307.69 1934.4 307.69 Q1942.24 307.69 1946.79 312.759 Q1951.38 317.794 1951.38 326.474 M1944.99 324.599 Q1944.92 319.356 1942.04 316.231 Q1939.19 313.106 1934.47 313.106 Q1929.12 313.106 1925.89 316.127 Q1922.69 319.148 1922.21 324.634 L1944.99 324.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip020)" style="stroke:#008000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:6; stroke-opacity:1; fill:none" stroke-dasharray="3, 6" points="
  1444.66,399.356 1581.28,399.356 
  "/>
<path clip-path="url(#clip020)" d="M1604.05 373.436 L1613.5 373.436 L1636.48 416.804 L1636.48 373.436 L1643.29 373.436 L1643.29 425.276 L1633.84 425.276 L1610.86 381.908 L1610.86 425.276 L1604.05 425.276 L1604.05 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1690.23 404.234 L1690.23 407.359 L1660.86 407.359 Q1661.27 413.956 1664.81 417.429 Q1668.39 420.866 1674.74 420.866 Q1678.43 420.866 1681.86 419.963 Q1685.34 419.06 1688.74 417.255 L1688.74 423.297 Q1685.3 424.755 1681.69 425.519 Q1678.08 426.283 1674.36 426.283 Q1665.06 426.283 1659.61 420.866 Q1654.19 415.449 1654.19 406.213 Q1654.19 396.665 1659.33 391.075 Q1664.5 385.45 1673.25 385.45 Q1681.1 385.45 1685.65 390.519 Q1690.23 395.554 1690.23 404.234 M1683.84 402.359 Q1683.77 397.116 1680.89 393.991 Q1678.04 390.866 1673.32 390.866 Q1667.97 390.866 1664.74 393.887 Q1661.55 396.908 1661.06 402.394 L1683.84 402.359 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1700.06 409.929 L1700.06 386.387 L1706.45 386.387 L1706.45 409.686 Q1706.45 415.206 1708.6 417.984 Q1710.75 420.727 1715.06 420.727 Q1720.23 420.727 1723.22 417.429 Q1726.24 414.13 1726.24 408.436 L1726.24 386.387 L1732.63 386.387 L1732.63 425.276 L1726.24 425.276 L1726.24 419.304 Q1723.91 422.845 1720.82 424.581 Q1717.77 426.283 1713.7 426.283 Q1707 426.283 1703.53 422.116 Q1700.06 417.949 1700.06 409.929 M1716.13 385.45 L1716.13 385.45 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1768.32 392.359 Q1767.24 391.734 1765.96 391.456 Q1764.71 391.144 1763.18 391.144 Q1757.77 391.144 1754.85 394.686 Q1751.97 398.193 1751.97 404.79 L1751.97 425.276 L1745.54 425.276 L1745.54 386.387 L1751.97 386.387 L1751.97 392.429 Q1753.98 388.887 1757.21 387.186 Q1760.44 385.45 1765.06 385.45 Q1765.72 385.45 1766.52 385.554 Q1767.31 385.623 1768.29 385.797 L1768.32 392.359 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1792.7 405.727 Q1784.95 405.727 1781.97 407.498 Q1778.98 409.269 1778.98 413.54 Q1778.98 416.942 1781.2 418.956 Q1783.46 420.935 1787.31 420.935 Q1792.63 420.935 1795.82 417.185 Q1799.05 413.401 1799.05 407.151 L1799.05 405.727 L1792.7 405.727 M1805.44 403.088 L1805.44 425.276 L1799.05 425.276 L1799.05 419.373 Q1796.86 422.915 1793.6 424.616 Q1790.33 426.283 1785.61 426.283 Q1779.64 426.283 1776.1 422.949 Q1772.59 419.581 1772.59 413.956 Q1772.59 407.394 1776.97 404.061 Q1781.38 400.727 1790.09 400.727 L1799.05 400.727 L1799.05 400.102 Q1799.05 395.693 1796.13 393.297 Q1793.25 390.866 1788.01 390.866 Q1784.67 390.866 1781.52 391.665 Q1778.36 392.463 1775.44 394.061 L1775.44 388.158 Q1778.95 386.804 1782.24 386.144 Q1785.54 385.45 1788.67 385.45 Q1797.11 385.45 1801.27 389.825 Q1805.44 394.199 1805.44 403.088 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1818.6 371.248 L1824.99 371.248 L1824.99 425.276 L1818.6 425.276 L1818.6 371.248 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1838.63 373.436 L1868.42 373.436 L1868.42 379.338 L1845.65 379.338 L1845.65 394.616 L1866.2 394.616 L1866.2 400.519 L1845.65 400.519 L1845.65 425.276 L1838.63 425.276 L1838.63 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1879.54 373.436 L1889.99 373.436 L1903.22 408.713 L1916.51 373.436 L1926.97 373.436 L1926.97 425.276 L1920.13 425.276 L1920.13 379.755 L1906.76 415.31 L1899.71 415.31 L1886.34 379.755 L1886.34 425.276 L1879.54 425.276 L1879.54 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1940.09 373.436 L1947.14 373.436 L1947.14 404.929 Q1947.14 413.262 1950.16 416.942 Q1953.18 420.588 1959.95 420.588 Q1966.69 420.588 1969.71 416.942 Q1972.73 413.262 1972.73 404.929 L1972.73 373.436 L1979.78 373.436 L1979.78 405.797 Q1979.78 415.935 1974.74 421.109 Q1969.74 426.283 1959.95 426.283 Q1950.13 426.283 1945.09 421.109 Q1940.09 415.935 1940.09 405.797 L1940.09 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M2030.61 371.318 Q2025.96 379.304 2023.7 387.116 Q2021.44 394.929 2021.44 402.949 Q2021.44 410.97 2023.7 418.852 Q2025.99 426.699 2030.61 434.651 L2025.06 434.651 Q2019.85 426.491 2017.24 418.609 Q2014.67 410.727 2014.67 402.949 Q2014.67 395.206 2017.24 387.359 Q2019.81 379.512 2025.06 371.318 L2030.61 371.318 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M2049.95 419.373 L2074.43 419.373 L2074.43 425.276 L2041.51 425.276 L2041.51 419.373 Q2045.51 415.241 2052.38 408.297 Q2059.29 401.318 2061.06 399.304 Q2064.43 395.519 2065.75 392.915 Q2067.1 390.276 2067.1 387.741 Q2067.1 383.609 2064.19 381.005 Q2061.31 378.401 2056.65 378.401 Q2053.35 378.401 2049.67 379.547 Q2046.03 380.693 2041.86 383.019 L2041.86 375.936 Q2046.1 374.234 2049.78 373.366 Q2053.46 372.498 2056.51 372.498 Q2064.57 372.498 2069.36 376.526 Q2074.15 380.554 2074.15 387.29 Q2074.15 390.484 2072.94 393.366 Q2071.76 396.213 2068.6 400.102 Q2067.73 401.109 2063.08 405.936 Q2058.42 410.727 2049.95 419.373 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M2089.22 373.436 L2116.76 373.436 L2116.76 379.338 L2095.65 379.338 L2095.65 392.047 Q2097.17 391.526 2098.7 391.283 Q2100.23 391.005 2101.76 391.005 Q2110.44 391.005 2115.51 395.762 Q2120.58 400.519 2120.58 408.644 Q2120.58 417.012 2115.37 421.665 Q2110.16 426.283 2100.68 426.283 Q2097.42 426.283 2094.01 425.727 Q2090.65 425.172 2087.03 424.06 L2087.03 417.012 Q2090.16 418.713 2093.49 419.547 Q2096.83 420.38 2100.54 420.38 Q2106.55 420.38 2110.06 417.22 Q2113.56 414.06 2113.56 408.644 Q2113.56 403.227 2110.06 400.068 Q2106.55 396.908 2100.54 396.908 Q2097.73 396.908 2094.92 397.533 Q2092.14 398.158 2089.22 399.477 L2089.22 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M2149.4 378.054 Q2143.98 378.054 2141.24 383.401 Q2138.53 388.713 2138.53 399.408 Q2138.53 410.067 2141.24 415.415 Q2143.98 420.727 2149.4 420.727 Q2154.85 420.727 2157.55 415.415 Q2160.3 410.067 2160.3 399.408 Q2160.3 388.713 2157.55 383.401 Q2154.85 378.054 2149.4 378.054 M2149.4 372.498 Q2158.11 372.498 2162.69 379.408 Q2167.31 386.283 2167.31 399.408 Q2167.31 412.498 2162.69 419.408 Q2158.11 426.283 2149.4 426.283 Q2140.68 426.283 2136.06 419.408 Q2131.48 412.498 2131.48 399.408 Q2131.48 386.283 2136.06 379.408 Q2140.68 372.498 2149.4 372.498 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M2194.64 378.054 Q2189.22 378.054 2186.48 383.401 Q2183.77 388.713 2183.77 399.408 Q2183.77 410.067 2186.48 415.415 Q2189.22 420.727 2194.64 420.727 Q2200.09 420.727 2202.8 415.415 Q2205.54 410.067 2205.54 399.408 Q2205.54 388.713 2202.8 383.401 Q2200.09 378.054 2194.64 378.054 M2194.64 372.498 Q2203.35 372.498 2207.94 379.408 Q2212.55 386.283 2212.55 399.408 Q2212.55 412.498 2207.94 419.408 Q2203.35 426.283 2194.64 426.283 Q2185.92 426.283 2181.3 419.408 Q2176.72 412.498 2176.72 399.408 Q2176.72 386.283 2181.3 379.408 Q2185.92 372.498 2194.64 372.498 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M2222.97 371.318 L2228.53 371.318 Q2233.73 379.512 2236.3 387.359 Q2238.91 395.206 2238.91 402.949 Q2238.91 410.727 2236.3 418.609 Q2233.73 426.491 2228.53 434.651 L2222.97 434.651 Q2227.59 426.699 2229.85 418.852 Q2232.14 410.97 2232.14 402.949 Q2232.14 394.929 2229.85 387.116 Q2227.59 379.304 2222.97 371.318 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip020)" style="stroke:#008000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:6; stroke-opacity:1; fill:none" points="
  1444.66,477.116 1581.28,477.116 
  "/>
<path clip-path="url(#clip020)" d="M1604.05 451.196 L1613.5 451.196 L1636.48 494.564 L1636.48 451.196 L1643.29 451.196 L1643.29 503.036 L1633.84 503.036 L1610.86 459.668 L1610.86 503.036 L1604.05 503.036 L1604.05 451.196 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1690.23 481.994 L1690.23 485.119 L1660.86 485.119 Q1661.27 491.716 1664.81 495.189 Q1668.39 498.626 1674.74 498.626 Q1678.43 498.626 1681.86 497.723 Q1685.34 496.82 1688.74 495.015 L1688.74 501.057 Q1685.3 502.515 1681.69 503.279 Q1678.08 504.043 1674.36 504.043 Q1665.06 504.043 1659.61 498.626 Q1654.19 493.209 1654.19 483.973 Q1654.19 474.425 1659.33 468.835 Q1664.5 463.21 1673.25 463.21 Q1681.1 463.21 1685.65 468.279 Q1690.23 473.314 1690.23 481.994 M1683.84 480.119 Q1683.77 474.876 1680.89 471.751 Q1678.04 468.626 1673.32 468.626 Q1667.97 468.626 1664.74 471.647 Q1661.55 474.668 1661.06 480.154 L1683.84 480.119 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1700.06 487.689 L1700.06 464.147 L1706.45 464.147 L1706.45 487.446 Q1706.45 492.966 1708.6 495.744 Q1710.75 498.487 1715.06 498.487 Q1720.23 498.487 1723.22 495.189 Q1726.24 491.89 1726.24 486.196 L1726.24 464.147 L1732.63 464.147 L1732.63 503.036 L1726.24 503.036 L1726.24 497.064 Q1723.91 500.605 1720.82 502.341 Q1717.77 504.043 1713.7 504.043 Q1707 504.043 1703.53 499.876 Q1700.06 495.709 1700.06 487.689 M1716.13 463.21 L1716.13 463.21 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1768.32 470.119 Q1767.24 469.494 1765.96 469.216 Q1764.71 468.904 1763.18 468.904 Q1757.77 468.904 1754.85 472.446 Q1751.97 475.953 1751.97 482.55 L1751.97 503.036 L1745.54 503.036 L1745.54 464.147 L1751.97 464.147 L1751.97 470.189 Q1753.98 466.647 1757.21 464.946 Q1760.44 463.21 1765.06 463.21 Q1765.72 463.21 1766.52 463.314 Q1767.31 463.383 1768.29 463.557 L1768.32 470.119 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1792.7 483.487 Q1784.95 483.487 1781.97 485.258 Q1778.98 487.029 1778.98 491.3 Q1778.98 494.702 1781.2 496.716 Q1783.46 498.695 1787.31 498.695 Q1792.63 498.695 1795.82 494.945 Q1799.05 491.161 1799.05 484.911 L1799.05 483.487 L1792.7 483.487 M1805.44 480.848 L1805.44 503.036 L1799.05 503.036 L1799.05 497.133 Q1796.86 500.675 1793.6 502.376 Q1790.33 504.043 1785.61 504.043 Q1779.64 504.043 1776.1 500.709 Q1772.59 497.341 1772.59 491.716 Q1772.59 485.154 1776.97 481.821 Q1781.38 478.487 1790.09 478.487 L1799.05 478.487 L1799.05 477.862 Q1799.05 473.453 1796.13 471.057 Q1793.25 468.626 1788.01 468.626 Q1784.67 468.626 1781.52 469.425 Q1778.36 470.223 1775.44 471.821 L1775.44 465.918 Q1778.95 464.564 1782.24 463.904 Q1785.54 463.21 1788.67 463.21 Q1797.11 463.21 1801.27 467.585 Q1805.44 471.959 1805.44 480.848 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1818.6 449.008 L1824.99 449.008 L1824.99 503.036 L1818.6 503.036 L1818.6 449.008 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1838.63 451.196 L1868.42 451.196 L1868.42 457.098 L1845.65 457.098 L1845.65 472.376 L1866.2 472.376 L1866.2 478.279 L1845.65 478.279 L1845.65 503.036 L1838.63 503.036 L1838.63 451.196 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1879.54 451.196 L1889.99 451.196 L1903.22 486.473 L1916.51 451.196 L1926.97 451.196 L1926.97 503.036 L1920.13 503.036 L1920.13 457.515 L1906.76 493.07 L1899.71 493.07 L1886.34 457.515 L1886.34 503.036 L1879.54 503.036 L1879.54 451.196 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M1940.09 451.196 L1947.14 451.196 L1947.14 482.689 Q1947.14 491.022 1950.16 494.702 Q1953.18 498.348 1959.95 498.348 Q1966.69 498.348 1969.71 494.702 Q1972.73 491.022 1972.73 482.689 L1972.73 451.196 L1979.78 451.196 L1979.78 483.557 Q1979.78 493.695 1974.74 498.869 Q1969.74 504.043 1959.95 504.043 Q1950.13 504.043 1945.09 498.869 Q1940.09 493.695 1940.09 483.557 L1940.09 451.196 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M2030.61 449.078 Q2025.96 457.064 2023.7 464.876 Q2021.44 472.689 2021.44 480.709 Q2021.44 488.73 2023.7 496.612 Q2025.99 504.459 2030.61 512.411 L2025.06 512.411 Q2019.85 504.251 2017.24 496.369 Q2014.67 488.487 2014.67 480.709 Q2014.67 472.966 2017.24 465.119 Q2019.81 457.272 2025.06 449.078 L2030.61 449.078 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M2043.98 451.196 L2071.51 451.196 L2071.51 457.098 L2050.4 457.098 L2050.4 469.807 Q2051.93 469.286 2053.46 469.043 Q2054.99 468.765 2056.51 468.765 Q2065.19 468.765 2070.26 473.522 Q2075.33 478.279 2075.33 486.404 Q2075.33 494.772 2070.12 499.425 Q2064.92 504.043 2055.44 504.043 Q2052.17 504.043 2048.77 503.487 Q2045.4 502.932 2041.79 501.82 L2041.79 494.772 Q2044.92 496.473 2048.25 497.307 Q2051.58 498.14 2055.3 498.14 Q2061.31 498.14 2064.81 494.98 Q2068.32 491.82 2068.32 486.404 Q2068.32 480.987 2064.81 477.828 Q2061.31 474.668 2055.3 474.668 Q2052.49 474.668 2049.67 475.293 Q2046.9 475.918 2043.98 477.237 L2043.98 451.196 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M2104.15 455.814 Q2098.74 455.814 2095.99 461.161 Q2093.28 466.473 2093.28 477.168 Q2093.28 487.827 2095.99 493.175 Q2098.74 498.487 2104.15 498.487 Q2109.6 498.487 2112.31 493.175 Q2115.06 487.827 2115.06 477.168 Q2115.06 466.473 2112.31 461.161 Q2109.6 455.814 2104.15 455.814 M2104.15 450.258 Q2112.87 450.258 2117.45 457.168 Q2122.07 464.043 2122.07 477.168 Q2122.07 490.258 2117.45 497.168 Q2112.87 504.043 2104.15 504.043 Q2095.44 504.043 2090.82 497.168 Q2086.24 490.258 2086.24 477.168 Q2086.24 464.043 2090.82 457.168 Q2095.44 450.258 2104.15 450.258 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M2149.4 455.814 Q2143.98 455.814 2141.24 461.161 Q2138.53 466.473 2138.53 477.168 Q2138.53 487.827 2141.24 493.175 Q2143.98 498.487 2149.4 498.487 Q2154.85 498.487 2157.55 493.175 Q2160.3 487.827 2160.3 477.168 Q2160.3 466.473 2157.55 461.161 Q2154.85 455.814 2149.4 455.814 M2149.4 450.258 Q2158.11 450.258 2162.69 457.168 Q2167.31 464.043 2167.31 477.168 Q2167.31 490.258 2162.69 497.168 Q2158.11 504.043 2149.4 504.043 Q2140.68 504.043 2136.06 497.168 Q2131.48 490.258 2131.48 477.168 Q2131.48 464.043 2136.06 457.168 Q2140.68 450.258 2149.4 450.258 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M2194.64 455.814 Q2189.22 455.814 2186.48 461.161 Q2183.77 466.473 2183.77 477.168 Q2183.77 487.827 2186.48 493.175 Q2189.22 498.487 2194.64 498.487 Q2200.09 498.487 2202.8 493.175 Q2205.54 487.827 2205.54 477.168 Q2205.54 466.473 2202.8 461.161 Q2200.09 455.814 2194.64 455.814 M2194.64 450.258 Q2203.35 450.258 2207.94 457.168 Q2212.55 464.043 2212.55 477.168 Q2212.55 490.258 2207.94 497.168 Q2203.35 504.043 2194.64 504.043 Q2185.92 504.043 2181.3 497.168 Q2176.72 490.258 2176.72 477.168 Q2176.72 464.043 2181.3 457.168 Q2185.92 450.258 2194.64 450.258 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip020)" d="M2222.97 449.078 L2228.53 449.078 Q2233.73 457.272 2236.3 465.119 Q2238.91 472.966 2238.91 480.709 Q2238.91 488.487 2236.3 496.369 Q2233.73 504.251 2228.53 512.411 L2222.97 512.411 Q2227.59 504.459 2229.85 496.612 Q2232.14 488.73 2232.14 480.709 Q2232.14 472.689 2229.85 464.876 Q2227.59 457.064 2222.97 449.078 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /></svg>




<?xml version="1.0" encoding="utf-8"?>
<svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" width="600" height="400" viewBox="0 0 2400 1600">
<defs>
  <clipPath id="clip110">
    <rect x="0" y="0" width="2400" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip110)" d="
M0 1600 L2400 1600 L2400 0 L0 0  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip111">
    <rect x="480" y="0" width="1681" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip110)" d="
M245.309 1401.58 L2352.76 1401.58 L2352.76 123.472 L245.309 123.472  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip112">
    <rect x="245" y="123" width="2108" height="1279"/>
  </clipPath>
</defs>
<polyline clip-path="url(#clip112)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  304.954,1401.58 304.954,123.472 
  "/>
<polyline clip-path="url(#clip112)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  801.993,1401.58 801.993,123.472 
  "/>
<polyline clip-path="url(#clip112)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1299.03,1401.58 1299.03,123.472 
  "/>
<polyline clip-path="url(#clip112)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1796.07,1401.58 1796.07,123.472 
  "/>
<polyline clip-path="url(#clip112)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  2293.11,1401.58 2293.11,123.472 
  "/>
<polyline clip-path="url(#clip110)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  245.309,1401.58 2352.76,1401.58 
  "/>
<polyline clip-path="url(#clip110)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  304.954,1401.58 304.954,1382.68 
  "/>
<polyline clip-path="url(#clip110)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  801.993,1401.58 801.993,1382.68 
  "/>
<polyline clip-path="url(#clip110)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1299.03,1401.58 1299.03,1382.68 
  "/>
<polyline clip-path="url(#clip110)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1796.07,1401.58 1796.07,1382.68 
  "/>
<polyline clip-path="url(#clip110)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  2293.11,1401.58 2293.11,1382.68 
  "/>
<path clip-path="url(#clip110)" d="M304.954 1438.36 Q299.537 1438.36 296.794 1443.71 Q294.086 1449.02 294.086 1459.71 Q294.086 1470.37 296.794 1475.72 Q299.537 1481.03 304.954 1481.03 Q310.405 1481.03 313.113 1475.72 Q315.856 1470.37 315.856 1459.71 Q315.856 1449.02 313.113 1443.71 Q310.405 1438.36 304.954 1438.36 M304.954 1432.8 Q313.669 1432.8 318.252 1439.71 Q322.87 1446.59 322.87 1459.71 Q322.87 1472.8 318.252 1479.71 Q313.669 1486.59 304.954 1486.59 Q296.238 1486.59 291.62 1479.71 Q287.037 1472.8 287.037 1459.71 Q287.037 1446.59 291.62 1439.71 Q296.238 1432.8 304.954 1432.8 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M787.566 1479.68 L799.024 1479.68 L799.024 1440.13 L786.559 1442.63 L786.559 1436.24 L798.955 1433.74 L805.969 1433.74 L805.969 1479.68 L817.427 1479.68 L817.427 1485.58 L787.566 1485.58 L787.566 1479.68 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1291.01 1479.68 L1315.49 1479.68 L1315.49 1485.58 L1282.57 1485.58 L1282.57 1479.68 Q1286.57 1475.55 1293.44 1468.6 Q1300.35 1461.62 1302.12 1459.61 Q1305.49 1455.82 1306.81 1453.22 Q1308.16 1450.58 1308.16 1448.05 Q1308.16 1443.91 1305.25 1441.31 Q1302.37 1438.71 1297.71 1438.71 Q1294.41 1438.71 1290.73 1439.85 Q1287.09 1441 1282.92 1443.32 L1282.92 1436.24 Q1287.16 1434.54 1290.84 1433.67 Q1294.52 1432.8 1297.57 1432.8 Q1305.63 1432.8 1310.42 1436.83 Q1315.21 1440.86 1315.21 1447.59 Q1315.21 1450.79 1314 1453.67 Q1312.82 1456.52 1309.66 1460.41 Q1308.79 1461.41 1304.14 1466.24 Q1299.48 1471.03 1291.01 1479.68 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1802.44 1457.63 Q1807.48 1458.7 1810.29 1462.11 Q1813.14 1465.51 1813.14 1470.51 Q1813.14 1478.18 1807.86 1482.39 Q1802.58 1486.59 1792.86 1486.59 Q1789.6 1486.59 1786.12 1485.93 Q1782.69 1485.3 1779.01 1484.02 L1779.01 1477.25 Q1781.92 1478.95 1785.39 1479.82 Q1788.87 1480.68 1792.65 1480.68 Q1799.25 1480.68 1802.69 1478.08 Q1806.16 1475.48 1806.16 1470.51 Q1806.16 1465.93 1802.93 1463.36 Q1799.73 1460.75 1794.01 1460.75 L1787.96 1460.75 L1787.96 1454.99 L1794.28 1454.99 Q1799.46 1454.99 1802.2 1452.94 Q1804.94 1450.86 1804.94 1446.97 Q1804.94 1442.98 1802.1 1440.86 Q1799.28 1438.71 1794.01 1438.71 Q1791.12 1438.71 1787.83 1439.33 Q1784.53 1439.96 1780.57 1441.27 L1780.57 1435.02 Q1784.56 1433.91 1788.03 1433.36 Q1791.54 1432.8 1794.63 1432.8 Q1802.62 1432.8 1807.27 1436.45 Q1811.92 1440.06 1811.92 1446.24 Q1811.92 1450.55 1809.46 1453.53 Q1806.99 1456.48 1802.44 1457.63 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M2297.63 1439.85 L2279.92 1467.52 L2297.63 1467.52 L2297.63 1439.85 M2295.78 1433.74 L2304.6 1433.74 L2304.6 1467.52 L2312 1467.52 L2312 1473.36 L2304.6 1473.36 L2304.6 1485.58 L2297.63 1485.58 L2297.63 1473.36 L2274.22 1473.36 L2274.22 1466.59 L2295.78 1433.74 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1242.64 1527.62 L1242.64 1538.66 L1255.8 1538.66 L1255.8 1543.62 L1242.64 1543.62 L1242.64 1564.74 Q1242.64 1569.49 1243.93 1570.85 Q1245.25 1572.2 1249.24 1572.2 L1255.8 1572.2 L1255.8 1577.55 L1249.24 1577.55 Q1241.85 1577.55 1239.03 1574.8 Q1236.22 1572.03 1236.22 1564.74 L1236.22 1543.62 L1231.53 1543.62 L1231.53 1538.66 L1236.22 1538.66 L1236.22 1527.62 L1242.64 1527.62 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1286.22 1523.52 L1300.94 1523.52 L1300.94 1528.49 L1292.61 1528.49 L1292.61 1581.96 L1300.94 1581.96 L1300.94 1586.92 L1286.22 1586.92 L1286.22 1523.52 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1339.34 1539.8 L1339.34 1545.85 Q1336.64 1544.46 1333.72 1543.76 Q1330.8 1543.07 1327.68 1543.07 Q1322.92 1543.07 1320.53 1544.53 Q1318.16 1545.99 1318.16 1548.9 Q1318.16 1551.12 1319.87 1552.41 Q1321.57 1553.66 1326.71 1554.8 L1328.89 1555.29 Q1335.7 1556.75 1338.55 1559.42 Q1341.43 1562.06 1341.43 1566.82 Q1341.43 1572.24 1337.12 1575.4 Q1332.85 1578.55 1325.35 1578.55 Q1322.23 1578.55 1318.82 1577.93 Q1315.46 1577.34 1311.71 1576.12 L1311.71 1569.53 Q1315.25 1571.37 1318.69 1572.3 Q1322.12 1573.21 1325.49 1573.21 Q1330 1573.21 1332.43 1571.68 Q1334.87 1570.12 1334.87 1567.3 Q1334.87 1564.7 1333.09 1563.31 Q1331.36 1561.92 1325.42 1560.64 L1323.2 1560.12 Q1317.26 1558.87 1314.62 1556.3 Q1311.98 1553.69 1311.98 1549.18 Q1311.98 1543.69 1315.87 1540.71 Q1319.76 1537.72 1326.91 1537.72 Q1330.46 1537.72 1333.58 1538.24 Q1336.71 1538.76 1339.34 1539.8 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1366.53 1523.52 L1366.53 1586.92 L1351.81 1586.92 L1351.81 1581.96 L1360.11 1581.96 L1360.11 1528.49 L1351.81 1528.49 L1351.81 1523.52 L1366.53 1523.52 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip112)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  245.309,1397.95 2352.76,1397.95 
  "/>
<polyline clip-path="url(#clip112)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  245.309,1080.23 2352.76,1080.23 
  "/>
<polyline clip-path="url(#clip112)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  245.309,762.504 2352.76,762.504 
  "/>
<polyline clip-path="url(#clip112)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  245.309,444.783 2352.76,444.783 
  "/>
<polyline clip-path="url(#clip112)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  245.309,127.062 2352.76,127.062 
  "/>
<polyline clip-path="url(#clip110)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  245.309,1401.58 245.309,123.472 
  "/>
<polyline clip-path="url(#clip110)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  245.309,1397.95 264.206,1397.95 
  "/>
<polyline clip-path="url(#clip110)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  245.309,1080.23 264.206,1080.23 
  "/>
<polyline clip-path="url(#clip110)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  245.309,762.504 264.206,762.504 
  "/>
<polyline clip-path="url(#clip110)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  245.309,444.783 264.206,444.783 
  "/>
<polyline clip-path="url(#clip110)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  245.309,127.062 264.206,127.062 
  "/>
<path clip-path="url(#clip110)" d="M119.136 1398.62 L163.649 1398.62 L163.649 1404.53 L119.136 1404.53 L119.136 1398.62 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M184.83 1417.96 L209.309 1417.96 L209.309 1423.87 L176.392 1423.87 L176.392 1417.96 Q180.385 1413.83 187.26 1406.89 Q194.17 1399.91 195.941 1397.89 Q199.309 1394.11 200.628 1391.51 Q201.982 1388.87 201.982 1386.33 Q201.982 1382.2 199.066 1379.6 Q196.184 1376.99 191.531 1376.99 Q188.233 1376.99 184.552 1378.14 Q180.906 1379.28 176.74 1381.61 L176.74 1374.53 Q180.976 1372.82 184.656 1371.96 Q188.337 1371.09 191.392 1371.09 Q199.448 1371.09 204.239 1375.12 Q209.031 1379.14 209.031 1385.88 Q209.031 1389.07 207.816 1391.96 Q206.635 1394.8 203.476 1398.69 Q202.607 1399.7 197.955 1404.53 Q193.302 1409.32 184.83 1417.96 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M118.58 1080.9 L163.094 1080.9 L163.094 1086.8 L118.58 1086.8 L118.58 1080.9 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M179.448 1100.24 L190.906 1100.24 L190.906 1060.69 L178.441 1063.19 L178.441 1056.81 L190.837 1054.31 L197.851 1054.31 L197.851 1100.24 L209.309 1100.24 L209.309 1106.15 L179.448 1106.15 L179.448 1100.24 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M191.392 741.202 Q185.976 741.202 183.233 746.549 Q180.524 751.862 180.524 762.556 Q180.524 773.216 183.233 778.563 Q185.976 783.876 191.392 783.876 Q196.844 783.876 199.552 778.563 Q202.295 773.216 202.295 762.556 Q202.295 751.862 199.552 746.549 Q196.844 741.202 191.392 741.202 M191.392 735.647 Q200.107 735.647 204.691 742.556 Q209.309 749.431 209.309 762.556 Q209.309 775.646 204.691 782.556 Q200.107 789.431 191.392 789.431 Q182.677 789.431 178.059 782.556 Q173.476 775.646 173.476 762.556 Q173.476 749.431 178.059 742.556 Q182.677 735.647 191.392 735.647 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M179.448 464.8 L190.906 464.8 L190.906 425.252 L178.441 427.752 L178.441 421.363 L190.837 418.863 L197.851 418.863 L197.851 464.8 L209.309 464.8 L209.309 470.703 L179.448 470.703 L179.448 464.8 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M184.83 147.079 L209.309 147.079 L209.309 152.982 L176.392 152.982 L176.392 147.079 Q180.385 142.947 187.26 136.003 Q194.17 129.024 195.941 127.01 Q199.309 123.225 200.628 120.621 Q201.982 117.982 201.982 115.448 Q201.982 111.316 199.066 108.712 Q196.184 106.107 191.531 106.107 Q188.233 106.107 184.552 107.253 Q180.906 108.399 176.74 110.725 L176.74 103.642 Q180.976 101.941 184.656 101.073 Q188.337 100.205 191.392 100.205 Q199.448 100.205 204.239 104.232 Q209.031 108.26 209.031 114.996 Q209.031 118.191 207.816 121.073 Q206.635 123.92 203.476 127.809 Q202.607 128.816 197.955 133.642 Q193.302 138.434 184.83 147.079 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M35.6047 1073.64 Q31.2992 1071.24 29.2506 1067.91 Q27.202 1064.57 27.202 1060.06 Q27.202 1053.98 31.4728 1050.68 Q35.7089 1047.39 43.5561 1047.39 L67.0282 1047.39 L67.0282 1053.81 L43.7644 1053.81 Q38.1741 1053.81 35.4658 1055.79 Q32.7575 1057.77 32.7575 1061.83 Q32.7575 1066.8 36.0561 1069.68 Q39.3547 1072.56 45.0491 1072.56 L67.0282 1072.56 L67.0282 1078.98 L43.7644 1078.98 Q38.1394 1078.98 35.4658 1080.96 Q32.7575 1082.94 32.7575 1087.07 Q32.7575 1091.97 36.0908 1094.85 Q39.3894 1097.73 45.0491 1097.73 L67.0282 1097.73 L67.0282 1104.16 L28.1395 1104.16 L28.1395 1097.73 L34.1811 1097.73 Q30.6047 1095.55 28.9034 1092.49 Q27.202 1089.43 27.202 1085.23 Q27.202 1081 29.3547 1078.05 Q31.5075 1075.06 35.6047 1073.64 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M47.4797 1016.97 Q47.4797 1024.71 49.2505 1027.7 Q51.0213 1030.68 55.2921 1030.68 Q58.6949 1030.68 60.7087 1028.46 Q62.6879 1026.21 62.6879 1022.35 Q62.6879 1017.04 58.9379 1013.84 Q55.1532 1010.61 48.9033 1010.61 L47.4797 1010.61 L47.4797 1016.97 M44.8408 1004.23 L67.0282 1004.23 L67.0282 1010.61 L61.1254 1010.61 Q64.6671 1012.8 66.3684 1016.07 Q68.0351 1019.33 68.0351 1024.05 Q68.0351 1030.02 64.7018 1033.57 Q61.3337 1037.07 55.7088 1037.07 Q49.1463 1037.07 45.813 1032.7 Q42.4797 1028.29 42.4797 1019.57 L42.4797 1010.61 L41.8547 1010.61 Q37.445 1010.61 35.0492 1013.53 Q32.6186 1016.41 32.6186 1021.66 Q32.6186 1024.99 33.4172 1028.15 Q34.2158 1031.31 35.813 1034.23 L29.9103 1034.23 Q28.5561 1030.72 27.8964 1027.42 Q27.202 1024.12 27.202 1021 Q27.202 1012.56 31.577 1008.39 Q35.9519 1004.23 44.8408 1004.23 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M29.2853 966.275 L35.3269 966.275 Q33.9381 968.983 33.2436 971.9 Q32.5492 974.817 32.5492 977.941 Q32.5492 982.698 34.0075 985.094 Q35.4658 987.455 38.3825 987.455 Q40.6047 987.455 41.8894 985.754 Q43.1394 984.053 44.2852 978.914 L44.7713 976.726 Q46.2297 969.921 48.9033 967.073 Q51.5421 964.192 56.299 964.192 Q61.7157 964.192 64.8754 968.497 Q68.0351 972.768 68.0351 980.268 Q68.0351 983.393 67.4101 986.796 Q66.8198 990.164 65.6046 993.914 L59.0074 993.914 Q60.8476 990.372 61.7851 986.935 Q62.6879 983.497 62.6879 980.129 Q62.6879 975.615 61.1601 973.185 Q59.5976 970.754 56.7852 970.754 Q54.181 970.754 52.7921 972.525 Q51.4032 974.261 50.1185 980.198 L49.5977 982.421 Q48.3477 988.358 45.7783 990.997 Q43.1741 993.636 38.6603 993.636 Q33.1742 993.636 30.1881 989.747 Q27.202 985.858 27.202 978.705 Q27.202 975.164 27.7228 972.039 Q28.2436 968.914 29.2853 966.275 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M29.2853 929.226 L35.3269 929.226 Q33.9381 931.935 33.2436 934.851 Q32.5492 937.768 32.5492 940.893 Q32.5492 945.65 34.0075 948.046 Q35.4658 950.407 38.3825 950.407 Q40.6047 950.407 41.8894 948.706 Q43.1394 947.004 44.2852 941.865 L44.7713 939.678 Q46.2297 932.872 48.9033 930.025 Q51.5421 927.143 56.299 927.143 Q61.7157 927.143 64.8754 931.449 Q68.0351 935.72 68.0351 943.219 Q68.0351 946.344 67.4101 949.747 Q66.8198 953.115 65.6046 956.865 L59.0074 956.865 Q60.8476 953.324 61.7851 949.886 Q62.6879 946.449 62.6879 943.081 Q62.6879 938.567 61.1601 936.136 Q59.5976 933.706 56.7852 933.706 Q54.181 933.706 52.7921 935.476 Q51.4032 937.213 50.1185 943.15 L49.5977 945.372 Q48.3477 951.31 45.7783 953.949 Q43.1741 956.587 38.6603 956.587 Q33.1742 956.587 30.1881 952.699 Q27.202 948.81 27.202 941.657 Q27.202 938.115 27.7228 934.99 Q28.2436 931.865 29.2853 929.226 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M28.1395 898.949 L28.1395 892.178 L60.7782 880.025 L28.1395 867.873 L28.1395 861.102 L67.0282 875.685 L67.0282 884.366 L28.1395 898.949 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M45.9866 819.019 L49.1116 819.019 L49.1116 848.394 Q55.7088 847.977 59.181 844.435 Q62.6185 840.859 62.6185 834.505 Q62.6185 830.824 61.7157 827.387 Q60.8129 823.915 59.0074 820.512 L65.049 820.512 Q66.5073 823.949 67.2712 827.56 Q68.0351 831.171 68.0351 834.887 Q68.0351 844.192 62.6185 849.644 Q57.2018 855.06 47.9658 855.06 Q38.4172 855.06 32.827 849.921 Q27.202 844.748 27.202 835.998 Q27.202 828.151 32.2714 823.602 Q37.3061 819.019 45.9866 819.019 M44.1116 825.408 Q38.8686 825.477 35.7436 828.359 Q32.6186 831.206 32.6186 835.928 Q32.6186 841.276 35.6394 844.505 Q38.6603 847.699 44.1463 848.185 L44.1116 825.408 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M13.0007 808.533 L13.0007 802.144 L67.0282 802.144 L67.0282 808.533 L13.0007 808.533 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M32.6186 773.706 Q32.6186 778.845 36.6464 781.831 Q40.6394 784.818 47.6185 784.818 Q54.5977 784.818 58.6254 781.866 Q62.6185 778.88 62.6185 773.706 Q62.6185 768.602 58.5907 765.616 Q54.5629 762.63 47.6185 762.63 Q40.7089 762.63 36.6811 765.616 Q32.6186 768.602 32.6186 773.706 M27.202 773.706 Q27.202 765.373 32.6186 760.616 Q38.0353 755.859 47.6185 755.859 Q57.1671 755.859 62.6185 760.616 Q68.0351 765.373 68.0351 773.706 Q68.0351 782.074 62.6185 786.831 Q57.1671 791.554 47.6185 791.554 Q38.0353 791.554 32.6186 786.831 Q27.202 782.074 27.202 773.706 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M29.6325 717.283 L35.6047 717.283 Q34.1117 719.991 33.3825 722.735 Q32.6186 725.443 32.6186 728.221 Q32.6186 734.436 36.5769 737.873 Q40.5005 741.311 47.6185 741.311 Q54.7366 741.311 58.6949 737.873 Q62.6185 734.436 62.6185 728.221 Q62.6185 725.443 61.8893 722.735 Q61.1254 719.991 59.6324 717.283 L65.5351 717.283 Q66.7851 719.957 67.4101 722.839 Q68.0351 725.686 68.0351 728.915 Q68.0351 737.7 62.5143 742.873 Q56.9935 748.047 47.6185 748.047 Q38.1047 748.047 32.6533 742.839 Q27.202 737.596 27.202 728.498 Q27.202 725.547 27.827 722.735 Q28.4173 719.922 29.6325 717.283 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M28.1395 706.172 L28.1395 699.783 L67.0282 699.783 L67.0282 706.172 L28.1395 706.172 M13.0007 706.172 L13.0007 699.783 L21.0909 699.783 L21.0909 706.172 L13.0007 706.172 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M17.0979 680.096 L28.1395 680.096 L28.1395 666.936 L33.1047 666.936 L33.1047 680.096 L54.2157 680.096 Q58.9726 680.096 60.3268 678.811 Q61.681 677.492 61.681 673.499 L61.681 666.936 L67.0282 666.936 L67.0282 673.499 Q67.0282 680.894 64.2851 683.707 Q61.5074 686.519 54.2157 686.519 L33.1047 686.519 L33.1047 691.207 L28.1395 691.207 L28.1395 686.519 L17.0979 686.519 L17.0979 680.096 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M70.6392 642.353 Q77.5837 645.061 79.7017 647.631 Q81.8197 650.2 81.8197 654.506 L81.8197 659.61 L76.4726 659.61 L76.4726 655.86 Q76.4726 653.221 75.2226 651.763 Q73.9726 650.304 69.3198 648.534 L66.4032 647.388 L28.1395 663.117 L28.1395 656.346 L58.556 644.193 L28.1395 632.041 L28.1395 625.27 L70.6392 642.353 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M13.0007 594.437 L13.0007 579.714 L17.9659 579.714 L17.9659 588.048 L71.4379 588.048 L71.4379 579.714 L76.4031 579.714 L76.4031 594.437 L13.0007 594.437 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M35.6047 535.826 Q31.2992 533.43 29.2506 530.097 Q27.202 526.763 27.202 522.249 Q27.202 516.173 31.4728 512.875 Q35.7089 509.576 43.5561 509.576 L67.0282 509.576 L67.0282 516 L43.7644 516 Q38.1741 516 35.4658 517.979 Q32.7575 519.958 32.7575 524.02 Q32.7575 528.986 36.0561 531.867 Q39.3547 534.749 45.0491 534.749 L67.0282 534.749 L67.0282 541.173 L43.7644 541.173 Q38.1394 541.173 35.4658 543.152 Q32.7575 545.131 32.7575 549.263 Q32.7575 554.159 36.0908 557.041 Q39.3894 559.923 45.0491 559.923 L67.0282 559.923 L67.0282 566.346 L28.1395 566.346 L28.1395 559.923 L34.1811 559.923 Q30.6047 557.735 28.9034 554.68 Q27.202 551.624 27.202 547.423 Q27.202 543.187 29.3547 540.236 Q31.5075 537.249 35.6047 535.826 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M15.1882 485.479 L15.1882 479.576 L73.6253 497.632 L73.6253 503.534 L15.1882 485.479 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M29.2853 448.083 L35.3269 448.083 Q33.9381 450.792 33.2436 453.708 Q32.5492 456.625 32.5492 459.75 Q32.5492 464.507 34.0075 466.903 Q35.4658 469.264 38.3825 469.264 Q40.6047 469.264 41.8894 467.562 Q43.1394 465.861 44.2852 460.722 L44.7713 458.535 Q46.2297 451.729 48.9033 448.882 Q51.5421 446 56.299 446 Q61.7157 446 64.8754 450.305 Q68.0351 454.576 68.0351 462.076 Q68.0351 465.201 67.4101 468.604 Q66.8198 471.972 65.6046 475.722 L59.0074 475.722 Q60.8476 472.18 61.7851 468.743 Q62.6879 465.305 62.6879 461.937 Q62.6879 457.423 61.1601 454.993 Q59.5976 452.562 56.7852 452.562 Q54.181 452.562 52.7921 454.333 Q51.4032 456.069 50.1185 462.007 L49.5977 464.229 Q48.3477 470.166 45.7783 472.805 Q43.1741 475.444 38.6603 475.444 Q33.1742 475.444 30.1881 471.555 Q27.202 467.666 27.202 460.514 Q27.202 456.972 27.7228 453.847 Q28.2436 450.722 29.2853 448.083 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M13.0007 420.896 L76.4031 420.896 L76.4031 435.618 L71.4379 435.618 L71.4379 427.319 L17.9659 427.319 L17.9659 435.618 L13.0007 435.618 L13.0007 420.896 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M599.38 18.8205 L599.38 65.8515 L609.265 65.8515 Q621.782 65.8515 627.575 60.1802 Q633.408 54.509 633.408 42.2752 Q633.408 30.1225 627.575 24.4918 Q621.782 18.8205 609.265 18.8205 L599.38 18.8205 M591.198 12.096 L608.009 12.096 Q625.59 12.096 633.813 19.4281 Q642.036 26.7198 642.036 42.2752 Q642.036 57.9117 633.773 65.2439 Q625.509 72.576 608.009 72.576 L591.198 72.576 L591.198 12.096 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M693.564 48.0275 L693.564 51.6733 L659.293 51.6733 Q659.779 59.3701 663.911 63.421 Q668.084 67.4314 675.497 67.4314 Q679.791 67.4314 683.801 66.3781 Q687.852 65.3249 691.822 63.2184 L691.822 70.267 Q687.812 71.9684 683.599 72.8596 Q679.386 73.7508 675.051 73.7508 Q664.195 73.7508 657.835 67.4314 Q651.516 61.1119 651.516 50.3365 Q651.516 39.1965 657.511 32.6746 Q663.547 26.1121 673.755 26.1121 Q682.91 26.1121 688.217 32.0264 Q693.564 37.9003 693.564 48.0275 M686.11 45.84 Q686.029 39.7232 682.667 36.0774 Q679.345 32.4315 673.836 32.4315 Q667.598 32.4315 663.83 35.9558 Q660.104 39.4801 659.536 45.8805 L686.11 45.84 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M728.766 9.54393 L728.766 15.7418 L721.637 15.7418 Q717.626 15.7418 716.046 17.3622 Q714.507 18.9825 714.507 23.1955 L714.507 27.2059 L726.781 27.2059 L726.781 32.9987 L714.507 32.9987 L714.507 72.576 L707.013 72.576 L707.013 32.9987 L699.883 32.9987 L699.883 27.2059 L707.013 27.2059 L707.013 24.0462 Q707.013 16.471 710.537 13.0277 Q714.062 9.54393 721.718 9.54393 L728.766 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M755.624 49.7694 Q746.59 49.7694 743.107 51.8354 Q739.623 53.9013 739.623 58.8839 Q739.623 62.8538 742.215 65.2034 Q744.848 67.5124 749.345 67.5124 Q755.543 67.5124 759.27 63.1374 Q763.037 58.7219 763.037 51.4303 L763.037 49.7694 L755.624 49.7694 M770.491 46.6907 L770.491 72.576 L763.037 72.576 L763.037 65.6895 Q760.485 69.8214 756.677 71.8063 Q752.869 73.7508 747.36 73.7508 Q740.392 73.7508 736.261 69.8619 Q732.169 65.9325 732.169 59.3701 Q732.169 51.7138 737.273 47.825 Q742.418 43.9361 752.586 43.9361 L763.037 43.9361 L763.037 43.2069 Q763.037 38.0623 759.634 35.2672 Q756.272 32.4315 750.155 32.4315 Q746.266 32.4315 742.58 33.3632 Q738.894 34.295 735.491 36.1584 L735.491 29.2718 Q739.582 27.692 743.431 26.9223 Q747.279 26.1121 750.925 26.1121 Q760.768 26.1121 765.63 31.2163 Q770.491 36.3204 770.491 46.6907 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M785.074 54.671 L785.074 27.2059 L792.528 27.2059 L792.528 54.3874 Q792.528 60.8284 795.039 64.0691 Q797.551 67.2693 802.574 67.2693 Q808.61 67.2693 812.093 63.421 Q815.618 59.5726 815.618 52.9291 L815.618 27.2059 L823.071 27.2059 L823.071 72.576 L815.618 72.576 L815.618 65.6084 Q812.904 69.7404 809.298 71.7658 Q805.734 73.7508 800.994 73.7508 Q793.176 73.7508 789.125 68.8897 Q785.074 64.0286 785.074 54.671 M803.83 26.1121 L803.83 26.1121 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M838.424 9.54393 L845.878 9.54393 L845.878 72.576 L838.424 72.576 L838.424 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M868.847 14.324 L868.847 27.2059 L884.2 27.2059 L884.2 32.9987 L868.847 32.9987 L868.847 57.6282 Q868.847 63.1779 870.345 64.7578 Q871.885 66.3376 876.543 66.3376 L884.2 66.3376 L884.2 72.576 L876.543 72.576 Q867.915 72.576 864.634 69.3758 Q861.352 66.1351 861.352 57.6282 L861.352 32.9987 L855.884 32.9987 L855.884 27.2059 L861.352 27.2059 L861.352 14.324 L868.847 14.324 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M895.907 62.2867 L904.454 62.2867 L904.454 72.576 L895.907 72.576 L895.907 62.2867 M895.907 29.6769 L904.454 29.6769 L904.454 39.9662 L895.907 39.9662 L895.907 29.6769 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M948.649 12.096 L960.843 12.096 L976.277 53.2532 L991.792 12.096 L1003.98 12.096 L1003.98 72.576 L996.004 72.576 L996.004 19.4686 L980.409 60.9499 L972.185 60.9499 L956.589 19.4686 L956.589 72.576 L948.649 72.576 L948.649 12.096 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1040.52 49.7694 Q1031.49 49.7694 1028.01 51.8354 Q1024.52 53.9013 1024.52 58.8839 Q1024.52 62.8538 1027.12 65.2034 Q1029.75 67.5124 1034.25 67.5124 Q1040.44 67.5124 1044.17 63.1374 Q1047.94 58.7219 1047.94 51.4303 L1047.94 49.7694 L1040.52 49.7694 M1055.39 46.6907 L1055.39 72.576 L1047.94 72.576 L1047.94 65.6895 Q1045.39 69.8214 1041.58 71.8063 Q1037.77 73.7508 1032.26 73.7508 Q1025.29 73.7508 1021.16 69.8619 Q1017.07 65.9325 1017.07 59.3701 Q1017.07 51.7138 1022.17 47.825 Q1027.32 43.9361 1037.49 43.9361 L1047.94 43.9361 L1047.94 43.2069 Q1047.94 38.0623 1044.53 35.2672 Q1041.17 32.4315 1035.06 32.4315 Q1031.17 32.4315 1027.48 33.3632 Q1023.79 34.295 1020.39 36.1584 L1020.39 29.2718 Q1024.48 27.692 1028.33 26.9223 Q1032.18 26.1121 1035.82 26.1121 Q1045.67 26.1121 1050.53 31.2163 Q1055.39 36.3204 1055.39 46.6907 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1099.67 28.5427 L1099.67 35.5912 Q1096.51 33.9709 1093.1 33.1607 Q1089.7 32.3505 1086.06 32.3505 Q1080.51 32.3505 1077.71 34.0519 Q1074.96 35.7533 1074.96 39.156 Q1074.96 41.7486 1076.94 43.2475 Q1078.93 44.7058 1084.92 46.0426 L1087.47 46.6097 Q1095.41 48.3111 1098.74 51.4303 Q1102.1 54.509 1102.1 60.0587 Q1102.1 66.3781 1097.07 70.0644 Q1092.09 73.7508 1083.34 73.7508 Q1079.7 73.7508 1075.73 73.0216 Q1071.8 72.3329 1067.42 70.9151 L1067.42 63.2184 Q1071.55 65.3654 1075.56 66.4591 Q1079.57 67.5124 1083.5 67.5124 Q1088.77 67.5124 1091.61 65.73 Q1094.44 63.9071 1094.44 60.6258 Q1094.44 57.5877 1092.38 55.9673 Q1090.35 54.3469 1083.42 52.8481 L1080.83 52.2405 Q1073.9 50.7821 1070.82 47.7845 Q1067.75 44.7463 1067.75 39.4801 Q1067.75 33.0797 1072.28 29.5959 Q1076.82 26.1121 1085.16 26.1121 Q1089.3 26.1121 1092.94 26.7198 Q1096.59 27.3274 1099.67 28.5427 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1142.89 28.5427 L1142.89 35.5912 Q1139.73 33.9709 1136.33 33.1607 Q1132.93 32.3505 1129.28 32.3505 Q1123.73 32.3505 1120.93 34.0519 Q1118.18 35.7533 1118.18 39.156 Q1118.18 41.7486 1120.16 43.2475 Q1122.15 44.7058 1128.14 46.0426 L1130.7 46.6097 Q1138.64 48.3111 1141.96 51.4303 Q1145.32 54.509 1145.32 60.0587 Q1145.32 66.3781 1140.3 70.0644 Q1135.32 73.7508 1126.57 73.7508 Q1122.92 73.7508 1118.95 73.0216 Q1115.02 72.3329 1110.65 70.9151 L1110.65 63.2184 Q1114.78 65.3654 1118.79 66.4591 Q1122.8 67.5124 1126.73 67.5124 Q1131.99 67.5124 1134.83 65.73 Q1137.66 63.9071 1137.66 60.6258 Q1137.66 57.5877 1135.6 55.9673 Q1133.57 54.3469 1126.65 52.8481 L1124.05 52.2405 Q1117.13 50.7821 1114.05 47.7845 Q1110.97 44.7463 1110.97 39.4801 Q1110.97 33.0797 1115.51 29.5959 Q1120.04 26.1121 1128.39 26.1121 Q1132.52 26.1121 1136.17 26.7198 Q1139.81 27.3274 1142.89 28.5427 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1178.21 27.2059 L1186.11 27.2059 L1200.29 65.2844 L1214.47 27.2059 L1222.37 27.2059 L1205.36 72.576 L1195.23 72.576 L1178.21 27.2059 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1271.47 48.0275 L1271.47 51.6733 L1237.2 51.6733 Q1237.68 59.3701 1241.81 63.421 Q1245.99 67.4314 1253.4 67.4314 Q1257.69 67.4314 1261.7 66.3781 Q1265.75 65.3249 1269.72 63.2184 L1269.72 70.267 Q1265.71 71.9684 1261.5 72.8596 Q1257.29 73.7508 1252.95 73.7508 Q1242.1 73.7508 1235.74 67.4314 Q1229.42 61.1119 1229.42 50.3365 Q1229.42 39.1965 1235.41 32.6746 Q1241.45 26.1121 1251.66 26.1121 Q1260.81 26.1121 1266.12 32.0264 Q1271.47 37.9003 1271.47 48.0275 M1264.01 45.84 Q1263.93 39.7232 1260.57 36.0774 Q1257.25 32.4315 1251.74 32.4315 Q1245.5 32.4315 1241.73 35.9558 Q1238.01 39.4801 1237.44 45.8805 L1264.01 45.84 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1283.7 9.54393 L1291.15 9.54393 L1291.15 72.576 L1283.7 72.576 L1283.7 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1324.33 32.4315 Q1318.33 32.4315 1314.85 37.1306 Q1311.37 41.7891 1311.37 49.9314 Q1311.37 58.0738 1314.81 62.7728 Q1318.29 67.4314 1324.33 67.4314 Q1330.29 67.4314 1333.77 62.7323 Q1337.25 58.0333 1337.25 49.9314 Q1337.25 41.8701 1333.77 37.1711 Q1330.29 32.4315 1324.33 32.4315 M1324.33 26.1121 Q1334.05 26.1121 1339.6 32.4315 Q1345.15 38.7509 1345.15 49.9314 Q1345.15 61.0714 1339.6 67.4314 Q1334.05 73.7508 1324.33 73.7508 Q1314.57 73.7508 1309.02 67.4314 Q1303.51 61.0714 1303.51 49.9314 Q1303.51 38.7509 1309.02 32.4315 Q1314.57 26.1121 1324.33 26.1121 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1390.16 28.9478 L1390.16 35.9153 Q1387 34.1734 1383.8 33.3227 Q1380.64 32.4315 1377.4 32.4315 Q1370.15 32.4315 1366.14 37.0496 Q1362.13 41.6271 1362.13 49.9314 Q1362.13 58.2358 1366.14 62.8538 Q1370.15 67.4314 1377.4 67.4314 Q1380.64 67.4314 1383.8 66.5807 Q1387 65.6895 1390.16 63.9476 L1390.16 70.8341 Q1387.04 72.2924 1383.68 73.0216 Q1380.35 73.7508 1376.59 73.7508 Q1366.34 73.7508 1360.3 67.3098 Q1354.27 60.8689 1354.27 49.9314 Q1354.27 38.832 1360.34 32.472 Q1366.46 26.1121 1377.07 26.1121 Q1380.52 26.1121 1383.8 26.8413 Q1387.08 27.5299 1390.16 28.9478 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1403.12 27.2059 L1410.57 27.2059 L1410.57 72.576 L1403.12 72.576 L1403.12 27.2059 M1403.12 9.54393 L1410.57 9.54393 L1410.57 18.9825 L1403.12 18.9825 L1403.12 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1433.54 14.324 L1433.54 27.2059 L1448.9 27.2059 L1448.9 32.9987 L1433.54 32.9987 L1433.54 57.6282 Q1433.54 63.1779 1435.04 64.7578 Q1436.58 66.3376 1441.24 66.3376 L1448.9 66.3376 L1448.9 72.576 L1441.24 72.576 Q1432.61 72.576 1429.33 69.3758 Q1426.05 66.1351 1426.05 57.6282 L1426.05 32.9987 L1420.58 32.9987 L1420.58 27.2059 L1426.05 27.2059 L1426.05 14.324 L1433.54 14.324 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1477.58 76.7889 Q1474.42 84.8907 1471.42 87.3618 Q1468.42 89.8329 1463.4 89.8329 L1457.44 89.8329 L1457.44 83.5945 L1461.82 83.5945 Q1464.9 83.5945 1466.6 82.1361 Q1468.3 80.6778 1470.37 75.2496 L1471.7 71.8468 L1453.35 27.2059 L1461.25 27.2059 L1475.43 62.6918 L1489.61 27.2059 L1497.51 27.2059 L1477.58 76.7889 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1554.79 49.7694 Q1545.75 49.7694 1542.27 51.8354 Q1538.79 53.9013 1538.79 58.8839 Q1538.79 62.8538 1541.38 65.2034 Q1544.01 67.5124 1548.51 67.5124 Q1554.71 67.5124 1558.43 63.1374 Q1562.2 58.7219 1562.2 51.4303 L1562.2 49.7694 L1554.79 49.7694 M1569.65 46.6907 L1569.65 72.576 L1562.2 72.576 L1562.2 65.6895 Q1559.65 69.8214 1555.84 71.8063 Q1552.03 73.7508 1546.52 73.7508 Q1539.55 73.7508 1535.42 69.8619 Q1531.33 65.9325 1531.33 59.3701 Q1531.33 51.7138 1536.44 47.825 Q1541.58 43.9361 1551.75 43.9361 L1562.2 43.9361 L1562.2 43.2069 Q1562.2 38.0623 1558.8 35.2672 Q1555.43 32.4315 1549.32 32.4315 Q1545.43 32.4315 1541.74 33.3632 Q1538.06 34.295 1534.65 36.1584 L1534.65 29.2718 Q1538.74 27.692 1542.59 26.9223 Q1546.44 26.1121 1550.09 26.1121 Q1559.93 26.1121 1564.79 31.2163 Q1569.65 36.3204 1569.65 46.6907 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1607.97 9.54393 L1607.97 15.7418 L1600.84 15.7418 Q1596.83 15.7418 1595.25 17.3622 Q1593.72 18.9825 1593.72 23.1955 L1593.72 27.2059 L1605.99 27.2059 L1605.99 32.9987 L1593.72 32.9987 L1593.72 72.576 L1586.22 72.576 L1586.22 32.9987 L1579.09 32.9987 L1579.09 27.2059 L1586.22 27.2059 L1586.22 24.0462 Q1586.22 16.471 1589.75 13.0277 Q1593.27 9.54393 1600.93 9.54393 L1607.97 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1620.13 14.324 L1620.13 27.2059 L1635.48 27.2059 L1635.48 32.9987 L1620.13 32.9987 L1620.13 57.6282 Q1620.13 63.1779 1621.63 64.7578 Q1623.17 66.3376 1627.82 66.3376 L1635.48 66.3376 L1635.48 72.576 L1627.82 72.576 Q1619.2 72.576 1615.91 69.3758 Q1612.63 66.1351 1612.63 57.6282 L1612.63 32.9987 L1607.16 32.9987 L1607.16 27.2059 L1612.63 27.2059 L1612.63 14.324 L1620.13 14.324 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1684.09 48.0275 L1684.09 51.6733 L1649.82 51.6733 Q1650.31 59.3701 1654.44 63.421 Q1658.61 67.4314 1666.02 67.4314 Q1670.32 67.4314 1674.33 66.3781 Q1678.38 65.3249 1682.35 63.2184 L1682.35 70.267 Q1678.34 71.9684 1674.13 72.8596 Q1669.91 73.7508 1665.58 73.7508 Q1654.72 73.7508 1648.36 67.4314 Q1642.04 61.1119 1642.04 50.3365 Q1642.04 39.1965 1648.04 32.6746 Q1654.07 26.1121 1664.28 26.1121 Q1673.44 26.1121 1678.74 32.0264 Q1684.09 37.9003 1684.09 48.0275 M1676.64 45.84 Q1676.56 39.7232 1673.19 36.0774 Q1669.87 32.4315 1664.36 32.4315 Q1658.12 32.4315 1654.36 35.9558 Q1650.63 39.4801 1650.06 45.8805 L1676.64 45.84 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1722.62 34.1734 Q1721.36 33.4443 1719.86 33.1202 Q1718.4 32.7556 1716.62 32.7556 Q1710.3 32.7556 1706.9 36.8875 Q1703.54 40.9789 1703.54 48.6757 L1703.54 72.576 L1696.04 72.576 L1696.04 27.2059 L1703.54 27.2059 L1703.54 34.2544 Q1705.88 30.1225 1709.65 28.1376 Q1713.42 26.1121 1718.81 26.1121 Q1719.58 26.1121 1720.51 26.2337 Q1721.44 26.3147 1722.57 26.5172 L1722.62 34.1734 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1785.81 44.2197 Q1788.44 45.1109 1790.91 48.0275 Q1793.42 50.9442 1795.94 56.0483 L1804.24 72.576 L1795.45 72.576 L1787.71 57.061 Q1784.72 50.9847 1781.88 48.9997 Q1779.08 47.0148 1774.22 47.0148 L1765.31 47.0148 L1765.31 72.576 L1757.13 72.576 L1757.13 12.096 L1775.6 12.096 Q1785.97 12.096 1791.08 16.4305 Q1796.18 20.7649 1796.18 29.5149 Q1796.18 35.2267 1793.51 38.994 Q1790.87 42.7613 1785.81 44.2197 M1765.31 18.8205 L1765.31 40.2903 L1775.6 40.2903 Q1781.52 40.2903 1784.51 37.5762 Q1787.55 34.8216 1787.55 29.5149 Q1787.55 24.2082 1784.51 21.5346 Q1781.52 18.8205 1775.6 18.8205 L1765.31 18.8205 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1809.95 54.671 L1809.95 27.2059 L1817.41 27.2059 L1817.41 54.3874 Q1817.41 60.8284 1819.92 64.0691 Q1822.43 67.2693 1827.45 67.2693 Q1833.49 67.2693 1836.97 63.421 Q1840.5 59.5726 1840.5 52.9291 L1840.5 27.2059 L1847.95 27.2059 L1847.95 72.576 L1840.5 72.576 L1840.5 65.6084 Q1837.78 69.7404 1834.18 71.7658 Q1830.61 73.7508 1825.87 73.7508 Q1818.05 73.7508 1814 68.8897 Q1809.95 64.0286 1809.95 54.671 M1828.71 26.1121 L1828.71 26.1121 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1901.02 45.1919 L1901.02 72.576 L1893.56 72.576 L1893.56 45.4349 Q1893.56 38.994 1891.05 35.7938 Q1888.54 32.5936 1883.52 32.5936 Q1877.48 32.5936 1874 36.4419 Q1870.51 40.2903 1870.51 46.9338 L1870.51 72.576 L1863.02 72.576 L1863.02 27.2059 L1870.51 27.2059 L1870.51 34.2544 Q1873.19 30.163 1876.79 28.1376 Q1880.44 26.1121 1885.18 26.1121 Q1893 26.1121 1897.01 30.9732 Q1901.02 35.7938 1901.02 45.1919 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1917.79 62.2867 L1926.34 62.2867 L1926.34 72.576 L1917.79 72.576 L1917.79 62.2867 M1917.79 29.6769 L1926.34 29.6769 L1926.34 39.9662 L1917.79 39.9662 L1917.79 29.6769 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1978.31 65.6895 L2006.87 65.6895 L2006.87 72.576 L1968.46 72.576 L1968.46 65.6895 Q1973.12 60.8689 1981.14 52.7671 Q1989.21 44.6248 1991.27 42.2752 Q1995.2 37.8598 1996.74 34.8216 Q1998.32 31.7429 1998.32 28.7857 Q1998.32 23.9651 1994.92 20.927 Q1991.55 17.8888 1986.13 17.8888 Q1982.28 17.8888 1977.98 19.2256 Q1973.73 20.5624 1968.87 23.2765 L1968.87 15.0127 Q1973.81 13.0277 1978.11 12.015 Q1982.4 11.0023 1985.96 11.0023 Q1995.36 11.0023 2000.95 15.7013 Q2006.54 20.4004 2006.54 28.2591 Q2006.54 31.9859 2005.13 35.3482 Q2003.75 38.6699 2000.06 43.2069 Q1999.05 44.3817 1993.62 50.0125 Q1988.19 55.6027 1978.31 65.6895 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip112)" style="stroke:#009af9; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" points="
  304.954,762.504 309.924,743.445 314.894,724.405 319.865,705.402 324.835,686.457 329.806,667.587 334.776,648.812 339.746,630.15 344.717,611.621 349.687,593.242 
  354.658,575.033 359.628,557.011 364.598,539.194 369.569,521.601 374.539,504.248 379.509,487.152 384.48,470.332 389.45,453.805 394.421,437.584 399.391,421.688 
  404.361,406.134 409.332,390.936 414.302,376.109 419.273,361.668 424.243,347.628 429.213,334.004 434.184,320.808 439.154,308.054 444.125,295.754 449.095,283.922 
  454.065,272.568 459.036,261.704 464.006,251.34 468.977,241.489 473.947,232.157 478.917,223.357 483.888,215.095 488.858,207.381 493.829,200.222 498.799,193.625 
  503.769,187.598 508.74,182.145 513.71,177.272 518.681,172.985 523.651,169.287 528.621,166.182 533.592,163.674 538.562,161.764 543.532,160.456 548.503,159.749 
  553.473,159.645 558.444,160.144 563.414,161.245 568.384,162.948 573.355,165.25 578.325,168.149 583.296,171.643 588.266,175.728 593.236,180.399 598.207,185.653 
  603.177,191.483 608.148,197.885 613.118,204.851 618.088,212.375 623.059,220.448 628.029,229.064 633,238.214 637.97,247.887 642.94,258.076 647.911,268.768 
  652.881,279.955 657.852,291.624 662.822,303.764 667.792,316.363 672.763,329.408 677.733,342.886 682.704,356.783 687.674,371.087 692.644,385.781 697.615,400.853 
  702.585,416.286 707.555,432.065 712.526,448.176 717.496,464.6 722.467,481.321 727.437,498.325 732.407,515.593 737.378,533.107 742.348,550.851 747.319,568.806 
  752.289,586.956 757.259,605.281 762.23,623.762 767.2,642.383 772.171,661.124 777.141,679.967 782.111,698.891 787.082,717.88 792.052,736.913 797.023,755.972 
  801.993,775.037 806.963,794.09 811.934,813.111 816.904,832.082 821.875,850.982 826.845,869.795 831.815,888.501 836.786,907.08 841.756,925.514 846.727,943.786 
  851.697,961.877 856.667,979.768 861.638,997.441 866.608,1014.88 871.578,1032.07 876.549,1048.99 881.519,1065.62 886.49,1081.95 891.46,1097.96 896.43,1113.63 
  901.401,1128.96 906.371,1143.91 911.342,1158.49 916.312,1172.66 921.282,1186.43 926.253,1199.77 931.223,1212.68 936.194,1225.13 941.164,1237.12 946.134,1248.64 
  951.105,1259.67 956.075,1270.21 961.046,1280.23 966.016,1289.74 970.986,1298.72 975.957,1307.17 980.927,1315.07 985.898,1322.42 990.868,1329.21 995.838,1335.43 
  1000.81,1341.08 1005.78,1346.15 1010.75,1350.64 1015.72,1354.54 1020.69,1357.85 1025.66,1360.56 1030.63,1362.67 1035.6,1364.19 1040.57,1365.1 1045.54,1365.41 
  1050.51,1365.11 1055.48,1364.22 1060.45,1362.72 1065.42,1360.62 1070.39,1357.92 1075.36,1354.63 1080.34,1350.74 1085.31,1346.27 1090.28,1341.21 1095.25,1335.57 
  1100.22,1329.36 1105.19,1322.59 1110.16,1315.25 1115.13,1307.36 1120.1,1298.92 1125.07,1289.96 1130.04,1280.46 1135.01,1270.44 1139.98,1259.91 1144.95,1248.89 
  1149.92,1237.38 1154.89,1225.4 1159.86,1212.95 1164.83,1200.06 1169.8,1186.72 1174.77,1172.97 1179.74,1158.8 1184.71,1144.23 1189.68,1129.29 1194.65,1113.97 
  1199.62,1098.31 1204.59,1082.31 1209.57,1065.99 1214.54,1049.37 1219.51,1032.46 1224.48,1015.28 1229.45,997.849 1234.42,980.183 1239.39,962.3 1244.36,944.215 
  1249.33,925.949 1254.3,907.52 1259.27,888.946 1264.24,870.246 1269.21,851.438 1274.18,832.542 1279.15,813.576 1284.12,794.559 1289.09,775.509 1294.06,756.446 
  1299.03,737.39 1304,718.359 1308.97,699.372 1313.94,680.45 1318.91,661.609 1323.88,642.869 1328.85,624.248 1333.83,605.766 1338.8,587.441 1343.77,569.291 
  1348.74,551.335 1353.71,533.59 1358.68,516.074 1363.65,498.804 1368.62,481.799 1373.59,465.074 1378.56,448.647 1383.53,432.534 1388.5,416.75 1393.47,401.313 
  1398.44,386.237 1403.41,371.538 1408.38,357.23 1413.35,343.328 1418.32,329.844 1423.29,316.793 1428.26,304.188 1433.23,292.041 1438.2,280.366 1443.17,269.172 
  1448.14,258.473 1453.11,248.277 1458.08,238.595 1463.06,229.437 1468.03,220.812 1473,212.729 1477.97,205.197 1482.94,198.222 1487.91,191.811 1492.88,185.971 
  1497.85,180.707 1502.82,176.025 1507.79,171.928 1512.76,168.424 1517.73,165.514 1522.7,163.201 1527.67,161.487 1532.64,160.374 1537.61,159.863 1542.58,159.954 
  1547.55,160.647 1552.52,161.944 1557.49,163.841 1562.46,166.338 1567.43,169.43 1572.4,173.115 1577.37,177.388 1582.34,182.246 1587.32,187.685 1592.29,193.7 
  1597.26,200.284 1602.23,207.43 1607.2,215.131 1612.17,223.379 1617.14,232.165 1622.11,241.48 1627.08,251.318 1632.05,261.668 1637.02,272.519 1641.99,283.86 
  1646.96,295.678 1651.93,307.964 1656.9,320.702 1661.87,333.882 1666.84,347.493 1671.81,361.519 1676.78,375.945 1681.75,390.758 1686.72,405.942 1691.69,421.482 
  1696.66,437.363 1701.63,453.568 1706.6,470.084 1711.58,486.892 1716.55,503.975 1721.52,521.317 1726.49,538.9 1731.46,556.705 1736.43,574.716 1741.4,592.915 
  1746.37,611.284 1751.34,629.804 1756.31,648.457 1761.28,667.224 1766.25,686.085 1771.22,705.023 1776.19,724.017 1781.16,743.05 1786.13,762.103 1791.1,781.157 
  1796.07,800.191 1801.04,819.187 1806.01,838.127 1810.98,856.99 1815.95,875.759 1820.92,894.415 1825.89,912.938 1830.86,931.312 1835.83,949.516 1840.81,967.533 
  1845.78,985.344 1850.75,1002.93 1855.72,1020.28 1860.69,1037.37 1865.66,1054.19 1870.63,1070.71 1875.6,1086.92 1880.57,1102.82 1885.54,1118.37 1890.51,1133.56 
  1895.48,1148.38 1900.45,1162.82 1905.42,1176.86 1910.39,1190.48 1915.36,1203.67 1920.33,1216.42 1925.3,1228.72 1930.27,1240.55 1935.24,1251.91 1940.21,1262.77 
  1945.18,1273.14 1950.15,1282.99 1955.12,1292.32 1960.09,1301.12 1965.07,1309.39 1970.04,1317.11 1975.01,1324.27 1979.98,1330.87 1984.95,1336.9 1989.92,1342.36 
  1994.89,1347.24 1999.86,1351.53 2004.83,1355.23 2009.8,1358.35 2014.77,1360.86 2019.74,1362.78 2024.71,1364.09 2029.68,1364.81 2034.65,1364.92 2039.62,1364.43 
  2044.59,1363.34 2049.56,1361.65 2054.53,1359.35 2059.5,1356.46 2064.47,1352.98 2069.44,1348.91 2074.41,1344.25 2079.38,1339.01 2084.35,1333.19 2089.33,1326.8 
  2094.3,1319.85 2099.27,1312.34 2104.24,1304.28 2109.21,1295.67 2114.18,1286.54 2119.15,1276.88 2124.12,1266.71 2129.09,1256.03 2134.06,1244.86 2139.03,1233.21 
  2144,1221.08 2148.97,1208.5 2153.94,1195.47 2158.91,1182.01 2163.88,1168.13 2168.85,1153.84 2173.82,1139.17 2178.79,1124.11 2183.76,1108.7 2188.73,1092.94 
  2193.7,1076.84 2198.67,1060.44 2203.64,1043.74 2208.61,1026.75 2213.58,1009.5 2218.56,992.004 2223.53,974.278 2228.5,956.341 2233.47,938.211 2238.44,919.906 
  2243.41,901.442 2248.38,882.839 2253.35,864.116 2258.32,845.292 2263.29,826.385 2268.26,807.415 2273.23,788.4 2278.2,769.36 2283.17,750.313 2288.14,731.278 
  2293.11,712.275 
  "/>
<polyline clip-path="url(#clip112)" style="stroke:#e26f46; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" points="
  304.954,762.504 309.924,748.942 314.894,735.277 319.865,721.531 324.835,707.728 329.806,693.887 334.776,680.031 339.746,666.179 344.717,652.355 349.687,638.579 
  354.658,624.869 359.628,611.246 364.598,597.728 369.569,584.334 374.539,571.081 379.509,557.987 384.48,545.067 389.45,532.339 394.421,519.817 399.391,507.516 
  404.361,495.452 409.332,483.64 414.302,472.091 419.273,460.819 424.243,449.839 429.213,439.163 434.184,428.8 439.154,418.764 444.125,409.065 449.095,399.715 
  454.065,390.723 459.036,382.099 464.006,373.852 468.977,365.991 473.947,358.524 478.917,351.459 483.888,344.804 488.858,338.565 493.829,332.748 498.799,327.361 
  503.769,322.408 508.74,317.895 513.71,313.826 518.681,310.206 523.651,307.037 528.621,304.325 533.592,302.071 538.562,300.277 543.532,298.945 548.503,298.077 
  553.473,297.674 558.444,297.736 563.414,298.262 568.384,299.253 573.355,300.707 578.325,302.623 583.296,304.999 588.266,307.833 593.236,311.122 598.207,314.863 
  603.177,319.051 608.148,323.684 613.118,328.756 618.088,334.263 623.059,340.199 628.029,346.558 633,353.336 637.97,360.524 642.94,368.117 647.911,376.106 
  652.881,384.486 657.852,393.247 662.822,402.382 667.792,411.882 672.763,421.738 677.733,431.943 682.704,442.486 687.674,453.359 692.644,464.55 697.615,476.051 
  702.585,487.853 707.555,499.945 712.526,512.316 717.496,524.956 722.467,537.856 727.437,551.009 732.407,564.401 737.378,578.024 742.348,591.867 747.319,605.921 
  752.289,620.177 757.259,634.625 762.23,649.258 767.2,664.071 772.171,679.053 777.141,694.201 782.111,709.509 787.082,724.973 792.052,740.589 797.023,756.355 
  801.993,769.342 806.963,780.578 811.934,791.891 816.904,803.262 821.875,814.675 826.845,826.112 831.815,837.558 836.786,848.996 841.756,860.409 846.727,871.78 
  851.697,883.095 856.667,894.336 861.638,905.488 866.608,916.539 871.578,927.47 876.549,938.27 881.519,948.925 886.49,959.421 891.46,969.745 896.43,979.884 
  901.401,989.827 906.371,999.561 911.342,1009.07 916.312,1018.36 921.282,1027.4 926.253,1036.19 931.223,1044.71 936.194,1052.97 941.164,1060.94 946.134,1068.63 
  951.105,1076.01 956.075,1083.09 961.046,1089.85 966.016,1096.28 970.986,1102.39 975.957,1108.17 980.927,1113.6 985.898,1118.68 990.868,1123.41 995.838,1127.78 
  1000.81,1131.78 1005.78,1135.42 1010.75,1138.69 1015.72,1141.59 1020.69,1144.1 1025.66,1146.24 1030.63,1147.99 1035.6,1149.36 1040.57,1150.34 1045.54,1150.93 
  1050.51,1151.14 1055.48,1150.95 1060.45,1150.38 1065.42,1149.42 1070.39,1148.08 1075.36,1146.34 1080.34,1144.22 1085.31,1141.72 1090.28,1138.84 1095.25,1135.58 
  1100.22,1131.95 1105.19,1127.94 1110.16,1123.57 1115.13,1118.83 1120.1,1113.74 1125.07,1108.29 1130.04,1102.49 1135.01,1096.34 1139.98,1089.86 1144.95,1083.04 
  1149.92,1075.9 1154.89,1068.44 1159.86,1060.66 1164.83,1052.58 1169.8,1044.2 1174.77,1035.53 1179.74,1026.57 1184.71,1017.34 1189.68,1007.83 1194.65,998.065 
  1199.62,988.047 1204.59,977.784 1209.57,967.284 1214.54,956.553 1219.51,945.602 1224.48,934.438 1229.45,923.072 1234.42,911.509 1239.39,899.757 1244.36,887.825 
  1249.33,875.719 1254.3,863.446 1259.27,851.013 1264.24,838.427 1269.21,825.691 1274.18,812.813 1279.15,799.797 1284.12,786.647 1289.09,773.366 1294.06,760.867 
  1299.03,752.13 1304,743.328 1308.97,734.471 1313.94,725.573 1318.91,716.646 1323.88,707.701 1328.85,698.751 1333.83,689.809 1338.8,680.887 1343.77,671.998 
  1348.74,663.153 1353.71,654.365 1358.68,645.644 1363.65,637.004 1368.62,628.454 1373.59,620.005 1378.56,611.669 1383.53,603.454 1388.5,595.372 1393.47,587.432 
  1398.44,579.646 1403.41,572.023 1408.38,564.571 1413.35,557.3 1418.32,550.219 1423.29,543.335 1428.26,536.657 1433.23,530.192 1438.2,523.949 1443.17,517.933 
  1448.14,512.153 1453.11,506.613 1458.08,501.322 1463.06,496.284 1468.03,491.506 1473,486.992 1477.97,482.748 1482.94,478.779 1487.91,475.089 1492.88,471.682 
  1497.85,468.563 1502.82,465.734 1507.79,463.199 1512.76,460.961 1517.73,459.022 1522.7,457.386 1527.67,456.053 1532.64,455.026 1537.61,454.305 1542.58,453.892 
  1547.55,453.788 1552.52,453.992 1557.49,454.506 1562.46,455.329 1567.43,456.46 1572.4,457.898 1577.37,459.643 1582.34,461.693 1587.32,464.046 1592.29,466.7 
  1597.26,469.653 1602.23,472.901 1607.2,476.444 1612.17,480.278 1617.14,484.398 1622.11,488.802 1627.08,493.485 1632.05,498.443 1637.02,503.673 1641.99,509.169 
  1646.96,514.928 1651.93,520.944 1656.9,527.211 1661.87,533.725 1666.84,540.479 1671.81,547.469 1676.78,554.69 1681.75,562.134 1686.72,569.796 1691.69,577.671 
  1696.66,585.75 1701.63,594.029 1706.6,602.501 1711.58,611.161 1716.55,620.002 1721.52,629.017 1726.49,638.201 1731.46,647.546 1736.43,657.046 1741.4,666.698 
  1746.37,676.494 1751.34,686.429 1756.31,696.496 1761.28,706.692 1766.25,717.009 1771.22,727.444 1776.19,737.993 1781.16,748.651 1786.13,759.416 1791.1,766.866 
  1796.07,773.01 1801.04,779.198 1806.01,785.421 1810.98,791.671 1815.95,797.941 1820.92,804.223 1825.89,810.51 1830.86,816.792 1835.83,823.063 1840.81,829.312 
  1845.78,835.533 1850.75,841.716 1855.72,847.856 1860.69,853.942 1865.66,859.969 1870.63,865.927 1875.6,871.811 1880.57,877.611 1885.54,883.322 1890.51,888.935 
  1895.48,894.444 1900.45,899.844 1905.42,905.127 1910.39,910.286 1915.36,915.314 1920.33,920.205 1925.3,924.954 1930.27,929.555 1935.24,934.002 1940.21,938.29 
  1945.18,942.414 1950.15,946.369 1955.12,950.15 1960.09,953.753 1965.07,957.174 1970.04,960.408 1975.01,963.451 1979.98,966.301 1984.95,968.952 1989.92,971.404 
  1994.89,973.651 1999.86,975.692 2004.83,977.524 2009.8,979.145 2014.77,980.552 2019.74,981.743 2024.71,982.718 2029.68,983.474 2034.65,984.01 2039.62,984.326 
  2044.59,984.42 2049.56,984.293 2054.53,983.943 2059.5,983.37 2064.47,982.575 2069.44,981.559 2074.41,980.321 2079.38,978.862 2084.35,977.184 2089.33,975.287 
  2094.3,973.173 2099.27,970.843 2104.24,968.301 2109.21,965.546 2114.18,962.581 2119.15,959.409 2124.12,956.033 2129.09,952.455 2134.06,948.677 2139.03,944.703 
  2144,940.536 2148.97,936.18 2153.94,931.637 2158.91,926.911 2163.88,922.006 2168.85,916.926 2173.82,911.674 2178.79,906.255 2183.76,900.673 2188.73,894.932 
  2193.7,889.036 2198.67,882.99 2203.64,876.797 2208.61,870.463 2213.58,863.991 2218.56,857.387 2223.53,850.655 2228.5,843.799 2233.47,836.823 2238.44,829.733 
  2243.41,822.532 2248.38,815.225 2253.35,807.816 2258.32,800.31 2263.29,792.709 2268.26,785.02 2273.23,777.245 2278.2,769.387 2283.17,762.081 2288.14,758.864 
  2293.11,755.622 
  "/>
<polyline clip-path="url(#clip112)" style="stroke:#008000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" stroke-dasharray="4, 8" points="
  304.954,762.504 309.924,748.602 314.894,734.691 319.865,720.781 324.835,706.88 329.806,692.999 334.776,679.145 339.746,665.33 344.717,651.564 349.687,637.857 
  354.658,624.219 359.628,610.663 364.598,597.199 369.569,583.84 374.539,570.599 379.509,557.487 384.48,544.519 389.45,531.707 394.421,519.067 399.391,506.611 
  404.361,494.356 409.332,482.316 414.302,470.507 419.273,458.945 424.243,447.647 429.213,436.627 434.184,425.903 439.154,415.491 444.125,405.408 449.095,395.671 
  454.065,386.296 459.036,377.296 464.006,368.692 468.977,360.498 473.947,352.732 478.917,345.407 483.888,338.537 488.858,332.136 493.829,326.214 498.799,320.781 
  503.769,315.848 508.74,311.422 513.71,307.51 518.681,304.119 523.651,301.253 528.621,298.917 533.592,297.113 538.562,295.841 543.532,295.1 548.503,294.887 
  553.473,295.198 558.444,296.029 563.414,297.375 568.384,299.23 573.355,301.585 578.325,304.434 583.296,307.767 588.266,311.575 593.236,315.846 598.207,320.568 
  603.177,325.73 608.148,331.316 613.118,337.314 618.088,343.712 623.059,350.498 628.029,357.658 633,365.177 637.97,373.044 642.94,381.243 647.911,389.763 
  652.881,398.588 657.852,407.707 662.822,417.107 667.792,426.774 672.763,436.696 677.733,446.86 682.704,457.254 687.674,467.866 692.644,478.683 697.615,489.692 
  702.585,500.883 707.555,512.244 712.526,523.764 717.496,535.432 722.467,547.237 727.437,559.17 732.407,571.219 737.378,583.375 742.348,595.627 747.319,607.965 
  752.289,620.378 757.259,632.856 762.23,645.39 767.2,657.968 772.171,670.583 777.141,683.222 782.111,695.877 787.082,708.538 792.052,721.196 797.023,733.839 
  801.993,746.46 806.963,759.048 811.934,771.594 816.904,784.089 821.875,796.523 826.845,808.887 831.815,821.172 836.786,833.366 841.756,845.458 846.727,857.439 
  851.697,869.297 856.667,881.021 861.638,892.602 866.608,904.029 871.578,915.292 876.549,926.379 881.519,937.281 886.49,947.986 891.46,958.486 896.43,968.768 
  901.401,978.824 906.371,988.642 911.342,998.213 916.312,1007.53 921.282,1016.57 926.253,1025.34 931.223,1033.82 936.194,1042.01 941.164,1049.91 946.134,1057.49 
  951.105,1064.75 956.075,1071.68 961.046,1078.28 966.016,1084.53 970.986,1090.43 975.957,1095.97 980.927,1101.16 985.898,1105.98 990.868,1110.43 995.838,1114.5 
  1000.81,1118.21 1005.78,1121.53 1010.75,1124.48 1015.72,1127.05 1020.69,1129.25 1025.66,1131.07 1030.63,1132.51 1035.6,1133.58 1040.57,1134.28 1045.54,1134.62 
  1050.51,1134.6 1055.48,1134.22 1060.45,1133.49 1065.42,1132.41 1070.39,1130.98 1075.36,1129.22 1080.34,1127.11 1085.31,1124.68 1090.28,1121.93 1095.25,1118.86 
  1100.22,1115.48 1105.19,1111.79 1110.16,1107.81 1115.13,1103.54 1120.1,1098.98 1125.07,1094.14 1130.04,1089.04 1135.01,1083.67 1139.98,1078.04 1144.95,1072.16 
  1149.92,1066.04 1154.89,1059.67 1159.86,1053.08 1164.83,1046.27 1169.8,1039.24 1174.77,1032 1179.74,1024.56 1184.71,1016.92 1189.68,1009.09 1194.65,1001.08 
  1199.62,992.887 1204.59,984.522 1209.57,975.992 1214.54,967.302 1219.51,958.461 1224.48,949.473 1229.45,940.347 1234.42,931.089 1239.39,921.705 1244.36,912.204 
  1249.33,902.591 1254.3,892.875 1259.27,883.061 1264.24,873.159 1269.21,863.174 1274.18,853.114 1279.15,842.988 1284.12,832.802 1289.09,822.563 1294.06,812.281 
  1299.03,801.962 1304,791.614 1308.97,781.246 1313.94,770.865 1318.91,760.479 1323.88,750.097 1328.85,739.727 1333.83,729.376 1338.8,719.054 1343.77,708.769 
  1348.74,698.531 1353.71,688.35 1358.68,678.238 1363.65,668.205 1368.62,658.261 1373.59,648.418 1378.56,638.686 1383.53,629.076 1388.5,619.599 1393.47,610.265 
  1398.44,601.086 1403.41,592.072 1408.38,583.235 1413.35,574.584 1418.32,566.132 1423.29,557.888 1428.26,549.865 1433.23,542.073 1438.2,534.523 1443.17,527.226 
  1448.14,520.193 1453.11,513.436 1458.08,506.962 1463.06,500.776 1468.03,494.887 1473,489.307 1477.97,484.046 1482.94,479.113 1487.91,474.516 1492.88,470.263 
  1497.85,466.361 1502.82,462.817 1507.79,459.635 1512.76,456.82 1517.73,454.376 1522.7,452.305 1527.67,450.609 1532.64,449.29 1537.61,448.349 1542.58,447.784 
  1547.55,447.595 1552.52,447.779 1557.49,448.334 1562.46,449.257 1567.43,450.542 1572.4,452.185 1577.37,454.179 1582.34,456.518 1587.32,459.194 1592.29,462.199 
  1597.26,465.523 1602.23,469.162 1607.2,473.113 1612.17,477.369 1617.14,481.92 1622.11,486.757 1627.08,491.871 1632.05,497.252 1637.02,502.893 1641.99,508.783 
  1646.96,514.914 1651.93,521.276 1656.9,527.859 1661.87,534.654 1666.84,541.652 1671.81,548.842 1676.78,556.216 1681.75,563.764 1686.72,571.475 1691.69,579.341 
  1696.66,587.35 1701.63,595.493 1706.6,603.761 1711.58,612.142 1716.55,620.627 1721.52,629.206 1726.49,637.872 1731.46,646.615 1736.43,655.427 1741.4,664.297 
  1746.37,673.216 1751.34,682.176 1756.31,691.166 1761.28,700.178 1766.25,709.202 1771.22,718.229 1776.19,727.251 1781.16,736.257 1786.13,745.24 1791.1,754.191 
  1796.07,763.1 1801.04,771.959 1806.01,780.759 1810.98,789.492 1815.95,798.149 1820.92,806.721 1825.89,815.2 1830.86,823.579 1835.83,831.847 1840.81,839.998 
  1845.78,848.023 1850.75,855.913 1855.72,863.662 1860.69,871.26 1865.66,878.7 1870.63,885.973 1875.6,893.075 1880.57,900 1885.54,906.739 1890.51,913.285 
  1895.48,919.631 1900.45,925.77 1905.42,931.697 1910.39,937.405 1915.36,942.888 1920.33,948.142 1925.3,953.161 1930.27,957.942 1935.24,962.478 1940.21,966.768 
  1945.18,970.807 1950.15,974.593 1955.12,978.122 1960.09,981.393 1965.07,984.403 1970.04,987.15 1975.01,989.634 1979.98,991.854 1984.95,993.808 1989.92,995.496 
  1994.89,996.918 1999.86,998.074 2004.83,998.965 2009.8,999.592 2014.77,999.957 2019.74,1000.06 2024.71,999.906 2029.68,999.494 2034.65,998.827 2039.62,997.909 
  2044.59,996.741 2049.56,995.326 2054.53,993.667 2059.5,991.769 2064.47,989.634 2069.44,987.265 2074.41,984.668 2079.38,981.847 2084.35,978.805 2089.33,975.547 
  2094.3,972.078 2099.27,968.404 2104.24,964.529 2109.21,960.459 2114.18,956.199 2119.15,951.756 2124.12,947.136 2129.09,942.344 2134.06,937.388 2139.03,932.271 
  2144,926.997 2148.97,921.57 2153.94,915.997 2158.91,910.283 2163.88,904.434 2168.85,898.457 2173.82,892.355 2178.79,886.137 2183.76,879.807 2188.73,873.372 
  2193.7,866.838 2198.67,860.211 2203.64,853.497 2208.61,846.703 2213.58,839.835 2218.56,832.9 2223.53,825.903 2228.5,818.853 2233.47,811.755 2238.44,804.616 
  2243.41,797.444 2248.38,790.244 2253.35,783.024 2258.32,775.791 2263.29,768.551 2268.26,761.313 2273.23,754.083 2278.2,746.869 2283.17,739.677 2288.14,732.516 
  2293.11,725.392 
  "/>
<polyline clip-path="url(#clip112)" style="stroke:#008000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" points="
  304.954,762.504 309.924,748.112 314.894,733.736 319.865,719.387 324.835,705.073 329.806,690.806 334.776,676.595 339.746,662.451 344.717,648.385 349.687,634.41 
  354.658,620.536 359.628,606.775 364.598,593.141 369.569,579.646 374.539,566.303 379.509,553.124 384.48,540.124 389.45,527.317 394.421,514.715 399.391,502.333 
  404.361,490.186 409.332,478.287 414.302,466.651 419.273,455.293 424.243,444.226 429.213,433.463 434.184,423.02 439.154,412.909 444.125,403.144 449.095,393.739 
  454.065,384.704 459.036,376.054 464.006,367.8 468.977,359.954 473.947,352.528 478.917,345.532 483.888,338.975 488.858,332.865 493.829,327.21 498.799,322.017 
  503.769,317.293 508.74,313.042 513.71,309.27 518.681,305.979 523.651,303.173 528.621,300.854 533.592,299.024 538.562,297.683 543.532,296.831 548.503,296.467 
  553.473,296.588 558.444,297.194 563.414,298.28 568.384,299.843 573.355,301.88 578.325,304.385 583.296,307.353 588.266,310.779 593.236,314.655 598.207,318.976 
  603.177,323.732 608.148,328.915 613.118,334.52 618.088,340.536 623.059,346.956 628.029,353.77 633,360.968 637.97,368.541 642.94,376.479 647.911,384.772 
  652.881,393.409 657.852,402.379 662.822,411.672 667.792,421.276 672.763,431.181 677.733,441.373 682.704,451.841 687.674,462.573 692.644,473.557 697.615,484.779 
  702.585,496.227 707.555,507.888 712.526,519.751 717.496,531.803 722.467,544.032 727.437,556.424 732.407,568.967 737.378,581.648 742.348,594.454 747.319,607.372 
  752.289,620.391 757.259,633.496 762.23,646.677 767.2,659.92 772.171,673.212 777.141,686.542 782.111,699.896 787.082,713.264 792.052,726.631 797.023,739.987 
  801.993,753.318 806.963,766.613 811.934,779.86 816.904,793.046 821.875,806.159 826.845,819.189 831.815,832.121 836.786,844.944 841.756,857.644 846.727,870.21 
  851.697,882.63 856.667,894.89 861.638,906.979 866.608,918.885 871.578,930.597 876.549,942.102 881.519,953.388 886.49,964.444 891.46,975.26 896.43,985.822 
  901.401,996.121 906.371,1006.15 911.342,1015.88 916.312,1025.32 921.282,1034.46 926.253,1043.28 931.223,1051.77 936.194,1059.94 941.164,1067.76 946.134,1075.22 
  951.105,1082.33 956.075,1089.08 961.046,1095.44 966.016,1101.44 970.986,1107.04 975.957,1112.26 980.927,1117.09 985.898,1121.52 990.868,1125.55 995.838,1129.18 
  1000.81,1132.41 1005.78,1135.24 1010.75,1137.67 1015.72,1139.7 1020.69,1141.33 1025.66,1142.57 1030.63,1143.41 1035.6,1143.85 1040.57,1143.92 1045.54,1143.59 
  1050.51,1142.88 1055.48,1141.8 1060.45,1140.35 1065.42,1138.54 1070.39,1136.36 1075.36,1133.84 1080.34,1130.97 1085.31,1127.76 1090.28,1124.22 1095.25,1120.36 
  1100.22,1116.17 1105.19,1111.68 1110.16,1106.89 1115.13,1101.8 1120.1,1096.43 1125.07,1090.77 1130.04,1084.85 1135.01,1078.67 1139.98,1072.23 1144.95,1065.56 
  1149.92,1058.64 1154.89,1051.51 1159.86,1044.15 1164.83,1036.59 1169.8,1028.82 1174.77,1020.86 1179.74,1012.71 1184.71,1004.39 1189.68,995.89 1194.65,987.232 
  1199.62,978.419 1204.59,969.461 1209.57,960.363 1214.54,951.136 1219.51,941.785 1224.48,932.32 1229.45,922.748 1234.42,913.076 1239.39,903.313 1244.36,893.465 
  1249.33,883.541 1254.3,873.547 1259.27,863.493 1264.24,853.384 1269.21,843.229 1274.18,833.034 1279.15,822.808 1284.12,812.557 1289.09,802.29 1294.06,792.012 
  1299.03,781.731 1304,771.455 1308.97,761.192 1313.94,750.951 1318.91,740.741 1323.88,730.57 1328.85,720.448 1333.83,710.383 1338.8,700.384 1343.77,690.46 
  1348.74,680.62 1353.71,670.872 1358.68,661.226 1363.65,651.691 1368.62,642.275 1373.59,632.988 1378.56,623.838 1383.53,614.835 1388.5,605.988 1393.47,597.305 
  1398.44,588.797 1403.41,580.471 1408.38,572.338 1413.35,564.406 1418.32,556.684 1423.29,549.182 1428.26,541.903 1433.23,534.857 1438.2,528.053 1443.17,521.499 
  1448.14,515.203 1453.11,509.174 1458.08,503.418 1463.06,497.943 1468.03,492.755 1473,487.86 1477.97,483.263 1482.94,478.969 1487.91,474.983 1492.88,471.308 
  1497.85,467.948 1502.82,464.907 1507.79,462.186 1512.76,459.788 1517.73,457.715 1522.7,455.968 1527.67,454.547 1532.64,453.452 1537.61,452.684 1542.58,452.242 
  1547.55,452.124 1552.52,452.329 1557.49,452.855 1562.46,453.702 1567.43,454.865 1572.4,456.344 1577.37,458.133 1582.34,460.23 1587.32,462.631 1592.29,465.332 
  1597.26,468.328 1602.23,471.616 1607.2,475.189 1612.17,479.044 1617.14,483.174 1622.11,487.574 1627.08,492.237 1632.05,497.158 1637.02,502.33 1641.99,507.746 
  1646.96,513.399 1651.93,519.282 1656.9,525.386 1661.87,531.705 1666.84,538.229 1671.81,544.95 1676.78,551.86 1681.75,558.949 1686.72,566.207 1691.69,573.629 
  1696.66,581.207 1701.63,588.932 1706.6,596.797 1711.58,604.79 1716.55,612.905 1721.52,621.131 1726.49,629.46 1731.46,637.882 1736.43,646.389 1741.4,654.971 
  1746.37,663.619 1751.34,672.325 1756.31,681.078 1761.28,689.871 1766.25,698.693 1771.22,707.536 1776.19,716.39 1781.16,725.247 1786.13,734.097 1791.1,742.931 
  1796.07,751.74 1801.04,760.515 1806.01,769.245 1810.98,777.923 1815.95,786.539 1820.92,795.083 1825.89,803.547 1830.86,811.921 1835.83,820.195 1840.81,828.361 
  1845.78,836.41 1850.75,844.337 1855.72,852.134 1860.69,859.794 1865.66,867.307 1870.63,874.665 1875.6,881.861 1880.57,888.886 1885.54,895.735 1890.51,902.399 
  1895.48,908.873 1900.45,915.15 1905.42,921.224 1910.39,927.089 1915.36,932.74 1920.33,938.171 1925.3,943.378 1930.27,948.357 1935.24,953.102 1940.21,957.611 
  1945.18,961.879 1950.15,965.903 1955.12,969.679 1960.09,973.207 1965.07,976.481 1970.04,979.502 1975.01,982.266 1979.98,984.773 1984.95,987.023 1989.92,989.013 
  1994.89,990.745 1999.86,992.217 2004.83,993.431 2009.8,994.386 2014.77,995.083 2019.74,995.524 2024.71,995.709 2029.68,995.641 2034.65,995.32 2039.62,994.75 
  2044.59,993.933 2049.56,992.871 2054.53,991.568 2059.5,990.026 2064.47,988.25 2069.44,986.242 2074.41,984.009 2079.38,981.553 2084.35,978.879 2089.33,975.993 
  2094.3,972.9 2099.27,969.605 2104.24,966.114 2109.21,962.433 2114.18,958.567 2119.15,954.519 2124.12,950.293 2129.09,945.895 2134.06,941.332 2139.03,936.608 
  2144,931.729 2148.97,926.702 2153.94,921.531 2158.91,916.223 2163.88,910.783 2168.85,905.218 2173.82,899.533 2178.79,893.735 2183.76,887.828 2188.73,881.819 
  2193.7,875.715 2198.67,869.521 2203.64,863.243 2208.61,856.888 2213.58,850.461 2218.56,843.968 2223.53,837.417 2228.5,830.813 2233.47,824.162 2238.44,817.471 
  2243.41,810.745 2248.38,803.992 2253.35,797.218 2258.32,790.429 2263.29,783.631 2268.26,776.83 2273.23,770.031 2278.2,763.242 2283.17,756.469 2288.14,749.718 
  2293.11,742.995 
  "/>
<path clip-path="url(#clip110)" d="
M1413.49 554.876 L2282.51 554.876 L2282.51 166.076 L1413.49 166.076  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<polyline clip-path="url(#clip110)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1413.49,554.876 2282.51,554.876 2282.51,166.076 1413.49,166.076 1413.49,554.876 
  "/>
<polyline clip-path="url(#clip110)" style="stroke:#009af9; stroke-linecap:butt; stroke-linejoin:round; stroke-width:6; stroke-opacity:1; fill:none" points="
  1436.91,243.836 1577.4,243.836 
  "/>
<path clip-path="url(#clip110)" d="M1634.19 219.617 L1634.19 226.457 Q1630.19 224.548 1626.65 223.61 Q1623.11 222.673 1619.81 222.673 Q1614.08 222.673 1610.96 224.895 Q1607.87 227.117 1607.87 231.214 Q1607.87 234.652 1609.91 236.423 Q1612 238.159 1617.76 239.235 L1622 240.103 Q1629.85 241.596 1633.56 245.381 Q1637.31 249.131 1637.31 255.45 Q1637.31 262.985 1632.24 266.874 Q1627.21 270.763 1617.45 270.763 Q1613.77 270.763 1609.6 269.929 Q1605.47 269.096 1601.03 267.464 L1601.03 260.242 Q1605.3 262.638 1609.39 263.853 Q1613.49 265.068 1617.45 265.068 Q1623.46 265.068 1626.72 262.707 Q1629.98 260.346 1629.98 255.971 Q1629.98 252.152 1627.62 249.999 Q1625.3 247.846 1619.95 246.77 L1615.68 245.936 Q1607.83 244.374 1604.32 241.041 Q1600.82 237.707 1600.82 231.77 Q1600.82 224.895 1605.64 220.937 Q1610.51 216.978 1619.01 216.978 Q1622.66 216.978 1626.44 217.638 Q1630.23 218.298 1634.19 219.617 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1647.97 230.867 L1654.36 230.867 L1654.36 269.756 L1647.97 269.756 L1647.97 230.867 M1647.97 215.728 L1654.36 215.728 L1654.36 223.818 L1647.97 223.818 L1647.97 215.728 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1698 238.332 Q1700.4 234.027 1703.73 231.978 Q1707.07 229.93 1711.58 229.93 Q1717.66 229.93 1720.96 234.2 Q1724.25 238.436 1724.25 246.284 L1724.25 269.756 L1717.83 269.756 L1717.83 246.492 Q1717.83 240.902 1715.85 238.193 Q1713.87 235.485 1709.81 235.485 Q1704.85 235.485 1701.96 238.784 Q1699.08 242.082 1699.08 247.777 L1699.08 269.756 L1692.66 269.756 L1692.66 246.492 Q1692.66 240.867 1690.68 238.193 Q1688.7 235.485 1684.57 235.485 Q1679.67 235.485 1676.79 238.818 Q1673.91 242.117 1673.91 247.777 L1673.91 269.756 L1667.48 269.756 L1667.48 230.867 L1673.91 230.867 L1673.91 236.909 Q1676.1 233.332 1679.15 231.631 Q1682.21 229.93 1686.41 229.93 Q1690.64 229.93 1693.6 232.082 Q1696.58 234.235 1698 238.332 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1743.18 263.922 L1743.18 284.547 L1736.75 284.547 L1736.75 230.867 L1743.18 230.867 L1743.18 236.77 Q1745.19 233.298 1748.25 231.631 Q1751.34 229.93 1755.61 229.93 Q1762.69 229.93 1767.1 235.555 Q1771.55 241.179 1771.55 250.346 Q1771.55 259.513 1767.1 265.138 Q1762.69 270.763 1755.61 270.763 Q1751.34 270.763 1748.25 269.096 Q1745.19 267.395 1743.18 263.922 M1764.91 250.346 Q1764.91 243.298 1762 239.304 Q1759.12 235.277 1754.05 235.277 Q1748.98 235.277 1746.06 239.304 Q1743.18 243.298 1743.18 250.346 Q1743.18 257.395 1746.06 261.422 Q1748.98 265.415 1754.05 265.415 Q1759.12 265.415 1762 261.422 Q1764.91 257.395 1764.91 250.346 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1782.14 215.728 L1788.53 215.728 L1788.53 269.756 L1782.14 269.756 L1782.14 215.728 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1835.16 248.714 L1835.16 251.839 L1805.78 251.839 Q1806.2 258.436 1809.74 261.909 Q1813.32 265.346 1819.67 265.346 Q1823.35 265.346 1826.79 264.443 Q1830.26 263.54 1833.66 261.735 L1833.66 267.777 Q1830.23 269.235 1826.62 269.999 Q1823 270.763 1819.29 270.763 Q1809.98 270.763 1804.53 265.346 Q1799.12 259.929 1799.12 250.693 Q1799.12 241.145 1804.25 235.555 Q1809.43 229.93 1818.18 229.93 Q1826.02 229.93 1830.57 234.999 Q1835.16 240.034 1835.16 248.714 M1828.77 246.839 Q1828.7 241.596 1825.82 238.471 Q1822.97 235.346 1818.25 235.346 Q1812.9 235.346 1809.67 238.367 Q1806.48 241.388 1805.99 246.874 L1828.77 246.839 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1845.92 217.916 L1875.71 217.916 L1875.71 223.818 L1852.93 223.818 L1852.93 239.096 L1873.49 239.096 L1873.49 244.999 L1852.93 244.999 L1852.93 269.756 L1845.92 269.756 L1845.92 217.916 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1886.82 217.916 L1897.27 217.916 L1910.5 253.193 L1923.8 217.916 L1934.25 217.916 L1934.25 269.756 L1927.41 269.756 L1927.41 224.235 L1914.05 259.79 L1907 259.79 L1893.63 224.235 L1893.63 269.756 L1886.82 269.756 L1886.82 217.916 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1947.38 217.916 L1954.43 217.916 L1954.43 249.409 Q1954.43 257.742 1957.45 261.422 Q1960.47 265.068 1967.24 265.068 Q1973.98 265.068 1977 261.422 Q1980.02 257.742 1980.02 249.409 L1980.02 217.916 L1987.07 217.916 L1987.07 250.277 Q1987.07 260.415 1982.03 265.589 Q1977.03 270.763 1967.24 270.763 Q1957.41 270.763 1952.38 265.589 Q1947.38 260.415 1947.38 250.277 L1947.38 217.916 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip110)" style="stroke:#e26f46; stroke-linecap:butt; stroke-linejoin:round; stroke-width:6; stroke-opacity:1; fill:none" points="
  1436.91,321.596 1577.4,321.596 
  "/>
<path clip-path="url(#clip110)" d="M1625.4 323.21 Q1627.66 323.974 1629.78 326.474 Q1631.93 328.974 1634.08 333.349 L1641.2 347.516 L1633.66 347.516 L1627.03 334.217 Q1624.46 329.009 1622.03 327.307 Q1619.64 325.606 1615.47 325.606 L1607.83 325.606 L1607.83 347.516 L1600.82 347.516 L1600.82 295.676 L1616.65 295.676 Q1625.54 295.676 1629.91 299.391 Q1634.29 303.106 1634.29 310.606 Q1634.29 315.502 1632 318.731 Q1629.74 321.96 1625.4 323.21 M1607.83 301.44 L1607.83 319.842 L1616.65 319.842 Q1621.72 319.842 1624.29 317.516 Q1626.89 315.155 1626.89 310.606 Q1626.89 306.058 1624.29 303.766 Q1621.72 301.44 1616.65 301.44 L1607.83 301.44 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1680.02 326.474 L1680.02 329.599 L1650.64 329.599 Q1651.06 336.196 1654.6 339.669 Q1658.18 343.106 1664.53 343.106 Q1668.21 343.106 1671.65 342.203 Q1675.12 341.3 1678.53 339.495 L1678.53 345.537 Q1675.09 346.995 1671.48 347.759 Q1667.87 348.523 1664.15 348.523 Q1654.85 348.523 1649.39 343.106 Q1643.98 337.689 1643.98 328.453 Q1643.98 318.905 1649.12 313.315 Q1654.29 307.69 1663.04 307.69 Q1670.89 307.69 1675.44 312.759 Q1680.02 317.794 1680.02 326.474 M1673.63 324.599 Q1673.56 319.356 1670.68 316.231 Q1667.83 313.106 1663.11 313.106 Q1657.76 313.106 1654.53 316.127 Q1651.34 319.148 1650.85 324.634 L1673.63 324.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1710.19 293.488 L1710.19 298.801 L1704.08 298.801 Q1700.64 298.801 1699.29 300.19 Q1697.97 301.578 1697.97 305.19 L1697.97 308.627 L1708.49 308.627 L1708.49 313.592 L1697.97 313.592 L1697.97 347.516 L1691.55 347.516 L1691.55 313.592 L1685.44 313.592 L1685.44 308.627 L1691.55 308.627 L1691.55 305.919 Q1691.55 299.426 1694.57 296.474 Q1697.59 293.488 1704.15 293.488 L1710.19 293.488 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1748.8 326.474 L1748.8 329.599 L1719.43 329.599 Q1719.84 336.196 1723.39 339.669 Q1726.96 343.106 1733.32 343.106 Q1737 343.106 1740.44 342.203 Q1743.91 341.3 1747.31 339.495 L1747.31 345.537 Q1743.87 346.995 1740.26 347.759 Q1736.65 348.523 1732.94 348.523 Q1723.63 348.523 1718.18 343.106 Q1712.76 337.689 1712.76 328.453 Q1712.76 318.905 1717.9 313.315 Q1723.07 307.69 1731.82 307.69 Q1739.67 307.69 1744.22 312.759 Q1748.8 317.794 1748.8 326.474 M1742.41 324.599 Q1742.34 319.356 1739.46 316.231 Q1736.62 313.106 1731.89 313.106 Q1726.55 313.106 1723.32 316.127 Q1720.12 319.148 1719.64 324.634 L1742.41 324.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1781.82 314.599 Q1780.75 313.974 1779.46 313.696 Q1778.21 313.384 1776.68 313.384 Q1771.27 313.384 1768.35 316.926 Q1765.47 320.433 1765.47 327.03 L1765.47 347.516 L1759.05 347.516 L1759.05 308.627 L1765.47 308.627 L1765.47 314.669 Q1767.48 311.127 1770.71 309.426 Q1773.94 307.69 1778.56 307.69 Q1779.22 307.69 1780.02 307.794 Q1780.82 307.863 1781.79 308.037 L1781.82 314.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1820.23 326.474 L1820.23 329.599 L1790.85 329.599 Q1791.27 336.196 1794.81 339.669 Q1798.39 343.106 1804.74 343.106 Q1808.42 343.106 1811.86 342.203 Q1815.33 341.3 1818.73 339.495 L1818.73 345.537 Q1815.3 346.995 1811.68 347.759 Q1808.07 348.523 1804.36 348.523 Q1795.05 348.523 1789.6 343.106 Q1784.18 337.689 1784.18 328.453 Q1784.18 318.905 1789.32 313.315 Q1794.5 307.69 1803.25 307.69 Q1811.09 307.69 1815.64 312.759 Q1820.23 317.794 1820.23 326.474 M1813.84 324.599 Q1813.77 319.356 1810.89 316.231 Q1808.04 313.106 1803.32 313.106 Q1797.97 313.106 1794.74 316.127 Q1791.55 319.148 1791.06 324.634 L1813.84 324.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1863.04 324.044 L1863.04 347.516 L1856.65 347.516 L1856.65 324.252 Q1856.65 318.731 1854.5 315.988 Q1852.34 313.245 1848.04 313.245 Q1842.87 313.245 1839.88 316.544 Q1836.89 319.842 1836.89 325.537 L1836.89 347.516 L1830.47 347.516 L1830.47 308.627 L1836.89 308.627 L1836.89 314.669 Q1839.18 311.162 1842.27 309.426 Q1845.4 307.69 1849.46 307.69 Q1856.16 307.69 1859.6 311.856 Q1863.04 315.988 1863.04 324.044 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1903.77 310.12 L1903.77 316.092 Q1901.06 314.599 1898.32 313.87 Q1895.61 313.106 1892.83 313.106 Q1886.61 313.106 1883.18 317.064 Q1879.74 320.988 1879.74 328.106 Q1879.74 335.224 1883.18 339.182 Q1886.61 343.106 1892.83 343.106 Q1895.61 343.106 1898.32 342.377 Q1901.06 341.613 1903.77 340.12 L1903.77 346.023 Q1901.09 347.273 1898.21 347.898 Q1895.36 348.523 1892.14 348.523 Q1883.35 348.523 1878.18 343.002 Q1873 337.481 1873 328.106 Q1873 318.592 1878.21 313.141 Q1883.46 307.69 1892.55 307.69 Q1895.5 307.69 1898.32 308.315 Q1901.13 308.905 1903.77 310.12 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1948.14 326.474 L1948.14 329.599 L1918.77 329.599 Q1919.18 336.196 1922.73 339.669 Q1926.3 343.106 1932.66 343.106 Q1936.34 343.106 1939.77 342.203 Q1943.25 341.3 1946.65 339.495 L1946.65 345.537 Q1943.21 346.995 1939.6 347.759 Q1935.99 348.523 1932.27 348.523 Q1922.97 348.523 1917.52 343.106 Q1912.1 337.689 1912.1 328.453 Q1912.1 318.905 1917.24 313.315 Q1922.41 307.69 1931.16 307.69 Q1939.01 307.69 1943.56 312.759 Q1948.14 317.794 1948.14 326.474 M1941.75 324.599 Q1941.68 319.356 1938.8 316.231 Q1935.95 313.106 1931.23 313.106 Q1925.89 313.106 1922.66 316.127 Q1919.46 319.148 1918.98 324.634 L1941.75 324.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip110)" style="stroke:#008000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:6; stroke-opacity:1; fill:none" stroke-dasharray="3, 6" points="
  1436.91,399.356 1577.4,399.356 
  "/>
<path clip-path="url(#clip110)" d="M1600.82 373.436 L1610.26 373.436 L1633.25 416.804 L1633.25 373.436 L1640.05 373.436 L1640.05 425.276 L1630.61 425.276 L1607.62 381.908 L1607.62 425.276 L1600.82 425.276 L1600.82 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1687 404.234 L1687 407.359 L1657.62 407.359 Q1658.04 413.956 1661.58 417.429 Q1665.16 420.866 1671.51 420.866 Q1675.19 420.866 1678.63 419.963 Q1682.1 419.06 1685.5 417.255 L1685.5 423.297 Q1682.07 424.755 1678.46 425.519 Q1674.85 426.283 1671.13 426.283 Q1661.82 426.283 1656.37 420.866 Q1650.96 415.449 1650.96 406.213 Q1650.96 396.665 1656.1 391.075 Q1661.27 385.45 1670.02 385.45 Q1677.87 385.45 1682.41 390.519 Q1687 395.554 1687 404.234 M1680.61 402.359 Q1680.54 397.116 1677.66 393.991 Q1674.81 390.866 1670.09 390.866 Q1664.74 390.866 1661.51 393.887 Q1658.32 396.908 1657.83 402.394 L1680.61 402.359 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1696.82 409.929 L1696.82 386.387 L1703.21 386.387 L1703.21 409.686 Q1703.21 415.206 1705.37 417.984 Q1707.52 420.727 1711.82 420.727 Q1717 420.727 1719.98 417.429 Q1723 414.13 1723 408.436 L1723 386.387 L1729.39 386.387 L1729.39 425.276 L1723 425.276 L1723 419.304 Q1720.68 422.845 1717.59 424.581 Q1714.53 426.283 1710.47 426.283 Q1703.77 426.283 1700.3 422.116 Q1696.82 417.949 1696.82 409.929 M1712.9 385.45 L1712.9 385.45 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1765.09 392.359 Q1764.01 391.734 1762.73 391.456 Q1761.48 391.144 1759.95 391.144 Q1754.53 391.144 1751.62 394.686 Q1748.73 398.193 1748.73 404.79 L1748.73 425.276 L1742.31 425.276 L1742.31 386.387 L1748.73 386.387 L1748.73 392.429 Q1750.75 388.887 1753.98 387.186 Q1757.21 385.45 1761.82 385.45 Q1762.48 385.45 1763.28 385.554 Q1764.08 385.623 1765.05 385.797 L1765.09 392.359 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1789.46 405.727 Q1781.72 405.727 1778.73 407.498 Q1775.75 409.269 1775.75 413.54 Q1775.75 416.942 1777.97 418.956 Q1780.23 420.935 1784.08 420.935 Q1789.39 420.935 1792.59 417.185 Q1795.82 413.401 1795.82 407.151 L1795.82 405.727 L1789.46 405.727 M1802.21 403.088 L1802.21 425.276 L1795.82 425.276 L1795.82 419.373 Q1793.63 422.915 1790.37 424.616 Q1787.1 426.283 1782.38 426.283 Q1776.41 426.283 1772.87 422.949 Q1769.36 419.581 1769.36 413.956 Q1769.36 407.394 1773.73 404.061 Q1778.14 400.727 1786.86 400.727 L1795.82 400.727 L1795.82 400.102 Q1795.82 395.693 1792.9 393.297 Q1790.02 390.866 1784.78 390.866 Q1781.44 390.866 1778.28 391.665 Q1775.12 392.463 1772.21 394.061 L1772.21 388.158 Q1775.71 386.804 1779.01 386.144 Q1782.31 385.45 1785.43 385.45 Q1793.87 385.45 1798.04 389.825 Q1802.21 394.199 1802.21 403.088 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1815.37 371.248 L1821.75 371.248 L1821.75 425.276 L1815.37 425.276 L1815.37 371.248 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1835.4 373.436 L1865.19 373.436 L1865.19 379.338 L1842.41 379.338 L1842.41 394.616 L1862.97 394.616 L1862.97 400.519 L1842.41 400.519 L1842.41 425.276 L1835.4 425.276 L1835.4 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1876.3 373.436 L1886.75 373.436 L1899.98 408.713 L1913.28 373.436 L1923.73 373.436 L1923.73 425.276 L1916.89 425.276 L1916.89 379.755 L1903.52 415.31 L1896.48 415.31 L1883.11 379.755 L1883.11 425.276 L1876.3 425.276 L1876.3 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1936.86 373.436 L1943.91 373.436 L1943.91 404.929 Q1943.91 413.262 1946.93 416.942 Q1949.95 420.588 1956.72 420.588 Q1963.45 420.588 1966.48 416.942 Q1969.5 413.262 1969.5 404.929 L1969.5 373.436 L1976.54 373.436 L1976.54 405.797 Q1976.54 415.935 1971.51 421.109 Q1966.51 426.283 1956.72 426.283 Q1946.89 426.283 1941.86 421.109 Q1936.86 415.935 1936.86 405.797 L1936.86 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M2027.38 371.318 Q2022.73 379.304 2020.47 387.116 Q2018.21 394.929 2018.21 402.949 Q2018.21 410.97 2020.47 418.852 Q2022.76 426.699 2027.38 434.651 L2021.82 434.651 Q2016.61 426.491 2014.01 418.609 Q2011.44 410.727 2011.44 402.949 Q2011.44 395.206 2014.01 387.359 Q2016.58 379.512 2021.82 371.318 L2027.38 371.318 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M2046.72 419.373 L2071.2 419.373 L2071.2 425.276 L2038.28 425.276 L2038.28 419.373 Q2042.27 415.241 2049.15 408.297 Q2056.06 401.318 2057.83 399.304 Q2061.2 395.519 2062.52 392.915 Q2063.87 390.276 2063.87 387.741 Q2063.87 383.609 2060.95 381.005 Q2058.07 378.401 2053.42 378.401 Q2050.12 378.401 2046.44 379.547 Q2042.79 380.693 2038.63 383.019 L2038.63 375.936 Q2042.86 374.234 2046.54 373.366 Q2050.23 372.498 2053.28 372.498 Q2061.34 372.498 2066.13 376.526 Q2070.92 380.554 2070.92 387.29 Q2070.92 390.484 2069.7 393.366 Q2068.52 396.213 2065.36 400.102 Q2064.5 401.109 2059.84 405.936 Q2055.19 410.727 2046.72 419.373 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M2085.99 373.436 L2113.52 373.436 L2113.52 379.338 L2092.41 379.338 L2092.41 392.047 Q2093.94 391.526 2095.47 391.283 Q2097 391.005 2098.52 391.005 Q2107.2 391.005 2112.27 395.762 Q2117.34 400.519 2117.34 408.644 Q2117.34 417.012 2112.13 421.665 Q2106.93 426.283 2097.45 426.283 Q2094.18 426.283 2090.78 425.727 Q2087.41 425.172 2083.8 424.06 L2083.8 417.012 Q2086.93 418.713 2090.26 419.547 Q2093.59 420.38 2097.31 420.38 Q2103.32 420.38 2106.82 417.22 Q2110.33 414.06 2110.33 408.644 Q2110.33 403.227 2106.82 400.068 Q2103.32 396.908 2097.31 396.908 Q2094.5 396.908 2091.68 397.533 Q2088.91 398.158 2085.99 399.477 L2085.99 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M2146.16 378.054 Q2140.75 378.054 2138 383.401 Q2135.29 388.713 2135.29 399.408 Q2135.29 410.067 2138 415.415 Q2140.75 420.727 2146.16 420.727 Q2151.61 420.727 2154.32 415.415 Q2157.06 410.067 2157.06 399.408 Q2157.06 388.713 2154.32 383.401 Q2151.61 378.054 2146.16 378.054 M2146.16 372.498 Q2154.88 372.498 2159.46 379.408 Q2164.08 386.283 2164.08 399.408 Q2164.08 412.498 2159.46 419.408 Q2154.88 426.283 2146.16 426.283 Q2137.45 426.283 2132.83 419.408 Q2128.25 412.498 2128.25 399.408 Q2128.25 386.283 2132.83 379.408 Q2137.45 372.498 2146.16 372.498 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M2191.4 378.054 Q2185.99 378.054 2183.25 383.401 Q2180.54 388.713 2180.54 399.408 Q2180.54 410.067 2183.25 415.415 Q2185.99 420.727 2191.4 420.727 Q2196.86 420.727 2199.56 415.415 Q2202.31 410.067 2202.31 399.408 Q2202.31 388.713 2199.56 383.401 Q2196.86 378.054 2191.4 378.054 M2191.4 372.498 Q2200.12 372.498 2204.7 379.408 Q2209.32 386.283 2209.32 399.408 Q2209.32 412.498 2204.7 419.408 Q2200.12 426.283 2191.4 426.283 Q2182.69 426.283 2178.07 419.408 Q2173.49 412.498 2173.49 399.408 Q2173.49 386.283 2178.07 379.408 Q2182.69 372.498 2191.4 372.498 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M2219.74 371.318 L2225.29 371.318 Q2230.5 379.512 2233.07 387.359 Q2235.68 395.206 2235.68 402.949 Q2235.68 410.727 2233.07 418.609 Q2230.5 426.491 2225.29 434.651 L2219.74 434.651 Q2224.36 426.699 2226.61 418.852 Q2228.9 410.97 2228.9 402.949 Q2228.9 394.929 2226.61 387.116 Q2224.36 379.304 2219.74 371.318 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip110)" style="stroke:#008000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:6; stroke-opacity:1; fill:none" points="
  1436.91,477.116 1577.4,477.116 
  "/>
<path clip-path="url(#clip110)" d="M1600.82 451.196 L1610.26 451.196 L1633.25 494.564 L1633.25 451.196 L1640.05 451.196 L1640.05 503.036 L1630.61 503.036 L1607.62 459.668 L1607.62 503.036 L1600.82 503.036 L1600.82 451.196 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1687 481.994 L1687 485.119 L1657.62 485.119 Q1658.04 491.716 1661.58 495.189 Q1665.16 498.626 1671.51 498.626 Q1675.19 498.626 1678.63 497.723 Q1682.1 496.82 1685.5 495.015 L1685.5 501.057 Q1682.07 502.515 1678.46 503.279 Q1674.85 504.043 1671.13 504.043 Q1661.82 504.043 1656.37 498.626 Q1650.96 493.209 1650.96 483.973 Q1650.96 474.425 1656.1 468.835 Q1661.27 463.21 1670.02 463.21 Q1677.87 463.21 1682.41 468.279 Q1687 473.314 1687 481.994 M1680.61 480.119 Q1680.54 474.876 1677.66 471.751 Q1674.81 468.626 1670.09 468.626 Q1664.74 468.626 1661.51 471.647 Q1658.32 474.668 1657.83 480.154 L1680.61 480.119 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1696.82 487.689 L1696.82 464.147 L1703.21 464.147 L1703.21 487.446 Q1703.21 492.966 1705.37 495.744 Q1707.52 498.487 1711.82 498.487 Q1717 498.487 1719.98 495.189 Q1723 491.89 1723 486.196 L1723 464.147 L1729.39 464.147 L1729.39 503.036 L1723 503.036 L1723 497.064 Q1720.68 500.605 1717.59 502.341 Q1714.53 504.043 1710.47 504.043 Q1703.77 504.043 1700.3 499.876 Q1696.82 495.709 1696.82 487.689 M1712.9 463.21 L1712.9 463.21 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1765.09 470.119 Q1764.01 469.494 1762.73 469.216 Q1761.48 468.904 1759.95 468.904 Q1754.53 468.904 1751.62 472.446 Q1748.73 475.953 1748.73 482.55 L1748.73 503.036 L1742.31 503.036 L1742.31 464.147 L1748.73 464.147 L1748.73 470.189 Q1750.75 466.647 1753.98 464.946 Q1757.21 463.21 1761.82 463.21 Q1762.48 463.21 1763.28 463.314 Q1764.08 463.383 1765.05 463.557 L1765.09 470.119 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1789.46 483.487 Q1781.72 483.487 1778.73 485.258 Q1775.75 487.029 1775.75 491.3 Q1775.75 494.702 1777.97 496.716 Q1780.23 498.695 1784.08 498.695 Q1789.39 498.695 1792.59 494.945 Q1795.82 491.161 1795.82 484.911 L1795.82 483.487 L1789.46 483.487 M1802.21 480.848 L1802.21 503.036 L1795.82 503.036 L1795.82 497.133 Q1793.63 500.675 1790.37 502.376 Q1787.1 504.043 1782.38 504.043 Q1776.41 504.043 1772.87 500.709 Q1769.36 497.341 1769.36 491.716 Q1769.36 485.154 1773.73 481.821 Q1778.14 478.487 1786.86 478.487 L1795.82 478.487 L1795.82 477.862 Q1795.82 473.453 1792.9 471.057 Q1790.02 468.626 1784.78 468.626 Q1781.44 468.626 1778.28 469.425 Q1775.12 470.223 1772.21 471.821 L1772.21 465.918 Q1775.71 464.564 1779.01 463.904 Q1782.31 463.21 1785.43 463.21 Q1793.87 463.21 1798.04 467.585 Q1802.21 471.959 1802.21 480.848 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1815.37 449.008 L1821.75 449.008 L1821.75 503.036 L1815.37 503.036 L1815.37 449.008 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1835.4 451.196 L1865.19 451.196 L1865.19 457.098 L1842.41 457.098 L1842.41 472.376 L1862.97 472.376 L1862.97 478.279 L1842.41 478.279 L1842.41 503.036 L1835.4 503.036 L1835.4 451.196 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1876.3 451.196 L1886.75 451.196 L1899.98 486.473 L1913.28 451.196 L1923.73 451.196 L1923.73 503.036 L1916.89 503.036 L1916.89 457.515 L1903.52 493.07 L1896.48 493.07 L1883.11 457.515 L1883.11 503.036 L1876.3 503.036 L1876.3 451.196 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M1936.86 451.196 L1943.91 451.196 L1943.91 482.689 Q1943.91 491.022 1946.93 494.702 Q1949.95 498.348 1956.72 498.348 Q1963.45 498.348 1966.48 494.702 Q1969.5 491.022 1969.5 482.689 L1969.5 451.196 L1976.54 451.196 L1976.54 483.557 Q1976.54 493.695 1971.51 498.869 Q1966.51 504.043 1956.72 504.043 Q1946.89 504.043 1941.86 498.869 Q1936.86 493.695 1936.86 483.557 L1936.86 451.196 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M2027.38 449.078 Q2022.73 457.064 2020.47 464.876 Q2018.21 472.689 2018.21 480.709 Q2018.21 488.73 2020.47 496.612 Q2022.76 504.459 2027.38 512.411 L2021.82 512.411 Q2016.61 504.251 2014.01 496.369 Q2011.44 488.487 2011.44 480.709 Q2011.44 472.966 2014.01 465.119 Q2016.58 457.272 2021.82 449.078 L2027.38 449.078 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M2040.75 451.196 L2068.28 451.196 L2068.28 457.098 L2047.17 457.098 L2047.17 469.807 Q2048.7 469.286 2050.23 469.043 Q2051.75 468.765 2053.28 468.765 Q2061.96 468.765 2067.03 473.522 Q2072.1 478.279 2072.1 486.404 Q2072.1 494.772 2066.89 499.425 Q2061.68 504.043 2052.2 504.043 Q2048.94 504.043 2045.54 503.487 Q2042.17 502.932 2038.56 501.82 L2038.56 494.772 Q2041.68 496.473 2045.02 497.307 Q2048.35 498.14 2052.07 498.14 Q2058.07 498.14 2061.58 494.98 Q2065.09 491.82 2065.09 486.404 Q2065.09 480.987 2061.58 477.828 Q2058.07 474.668 2052.07 474.668 Q2049.25 474.668 2046.44 475.293 Q2043.66 475.918 2040.75 477.237 L2040.75 451.196 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M2100.92 455.814 Q2095.5 455.814 2092.76 461.161 Q2090.05 466.473 2090.05 477.168 Q2090.05 487.827 2092.76 493.175 Q2095.5 498.487 2100.92 498.487 Q2106.37 498.487 2109.08 493.175 Q2111.82 487.827 2111.82 477.168 Q2111.82 466.473 2109.08 461.161 Q2106.37 455.814 2100.92 455.814 M2100.92 450.258 Q2109.63 450.258 2114.22 457.168 Q2118.84 464.043 2118.84 477.168 Q2118.84 490.258 2114.22 497.168 Q2109.63 504.043 2100.92 504.043 Q2092.2 504.043 2087.59 497.168 Q2083 490.258 2083 477.168 Q2083 464.043 2087.59 457.168 Q2092.2 450.258 2100.92 450.258 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M2146.16 455.814 Q2140.75 455.814 2138 461.161 Q2135.29 466.473 2135.29 477.168 Q2135.29 487.827 2138 493.175 Q2140.75 498.487 2146.16 498.487 Q2151.61 498.487 2154.32 493.175 Q2157.06 487.827 2157.06 477.168 Q2157.06 466.473 2154.32 461.161 Q2151.61 455.814 2146.16 455.814 M2146.16 450.258 Q2154.88 450.258 2159.46 457.168 Q2164.08 464.043 2164.08 477.168 Q2164.08 490.258 2159.46 497.168 Q2154.88 504.043 2146.16 504.043 Q2137.45 504.043 2132.83 497.168 Q2128.25 490.258 2128.25 477.168 Q2128.25 464.043 2132.83 457.168 Q2137.45 450.258 2146.16 450.258 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M2191.4 455.814 Q2185.99 455.814 2183.25 461.161 Q2180.54 466.473 2180.54 477.168 Q2180.54 487.827 2183.25 493.175 Q2185.99 498.487 2191.4 498.487 Q2196.86 498.487 2199.56 493.175 Q2202.31 487.827 2202.31 477.168 Q2202.31 466.473 2199.56 461.161 Q2196.86 455.814 2191.4 455.814 M2191.4 450.258 Q2200.12 450.258 2204.7 457.168 Q2209.32 464.043 2209.32 477.168 Q2209.32 490.258 2204.7 497.168 Q2200.12 504.043 2191.4 504.043 Q2182.69 504.043 2178.07 497.168 Q2173.49 490.258 2173.49 477.168 Q2173.49 464.043 2178.07 457.168 Q2182.69 450.258 2191.4 450.258 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip110)" d="M2219.74 449.078 L2225.29 449.078 Q2230.5 457.272 2233.07 465.119 Q2235.68 472.966 2235.68 480.709 Q2235.68 488.487 2233.07 496.369 Q2230.5 504.251 2225.29 512.411 L2219.74 512.411 Q2224.36 504.459 2226.61 496.612 Q2228.9 488.73 2228.9 480.709 Q2228.9 472.689 2226.61 464.876 Q2224.36 457.064 2219.74 449.078 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /></svg>




<?xml version="1.0" encoding="utf-8"?>
<svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" width="600" height="400" viewBox="0 0 2400 1600">
<defs>
  <clipPath id="clip200">
    <rect x="0" y="0" width="2400" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip200)" d="
M0 1600 L2400 1600 L2400 0 L0 0  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip201">
    <rect x="480" y="0" width="1681" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip200)" d="
M314.996 1401.58 L2352.76 1401.58 L2352.76 123.472 L314.996 123.472  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip202">
    <rect x="314" y="123" width="2039" height="1279"/>
  </clipPath>
</defs>
<polyline clip-path="url(#clip202)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  372.668,1401.58 372.668,123.472 
  "/>
<polyline clip-path="url(#clip202)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  853.272,1401.58 853.272,123.472 
  "/>
<polyline clip-path="url(#clip202)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1333.88,1401.58 1333.88,123.472 
  "/>
<polyline clip-path="url(#clip202)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1814.48,1401.58 1814.48,123.472 
  "/>
<polyline clip-path="url(#clip202)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  2295.08,1401.58 2295.08,123.472 
  "/>
<polyline clip-path="url(#clip200)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  314.996,1401.58 2352.76,1401.58 
  "/>
<polyline clip-path="url(#clip200)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  372.668,1401.58 372.668,1382.68 
  "/>
<polyline clip-path="url(#clip200)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  853.272,1401.58 853.272,1382.68 
  "/>
<polyline clip-path="url(#clip200)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1333.88,1401.58 1333.88,1382.68 
  "/>
<polyline clip-path="url(#clip200)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1814.48,1401.58 1814.48,1382.68 
  "/>
<polyline clip-path="url(#clip200)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  2295.08,1401.58 2295.08,1382.68 
  "/>
<path clip-path="url(#clip200)" d="M372.668 1438.36 Q367.252 1438.36 364.509 1443.71 Q361.8 1449.02 361.8 1459.71 Q361.8 1470.37 364.509 1475.72 Q367.252 1481.03 372.668 1481.03 Q378.12 1481.03 380.828 1475.72 Q383.571 1470.37 383.571 1459.71 Q383.571 1449.02 380.828 1443.71 Q378.12 1438.36 372.668 1438.36 M372.668 1432.8 Q381.384 1432.8 385.967 1439.71 Q390.585 1446.59 390.585 1459.71 Q390.585 1472.8 385.967 1479.71 Q381.384 1486.59 372.668 1486.59 Q363.953 1486.59 359.335 1479.71 Q354.752 1472.8 354.752 1459.71 Q354.752 1446.59 359.335 1439.71 Q363.953 1432.8 372.668 1432.8 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M838.845 1479.68 L850.303 1479.68 L850.303 1440.13 L837.838 1442.63 L837.838 1436.24 L850.234 1433.74 L857.248 1433.74 L857.248 1479.68 L868.706 1479.68 L868.706 1485.58 L838.845 1485.58 L838.845 1479.68 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1325.86 1479.68 L1350.33 1479.68 L1350.33 1485.58 L1317.42 1485.58 L1317.42 1479.68 Q1321.41 1475.55 1328.29 1468.6 Q1335.2 1461.62 1336.97 1459.61 Q1340.33 1455.82 1341.65 1453.22 Q1343.01 1450.58 1343.01 1448.05 Q1343.01 1443.91 1340.09 1441.31 Q1337.21 1438.71 1332.56 1438.71 Q1329.26 1438.71 1325.58 1439.85 Q1321.93 1441 1317.76 1443.32 L1317.76 1436.24 Q1322 1434.54 1325.68 1433.67 Q1329.36 1432.8 1332.42 1432.8 Q1340.47 1432.8 1345.26 1436.83 Q1350.06 1440.86 1350.06 1447.59 Q1350.06 1450.79 1348.84 1453.67 Q1347.66 1456.52 1344.5 1460.41 Q1343.63 1461.41 1338.98 1466.24 Q1334.33 1471.03 1325.86 1479.68 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1820.85 1457.63 Q1825.89 1458.7 1828.7 1462.11 Q1831.55 1465.51 1831.55 1470.51 Q1831.55 1478.18 1826.27 1482.39 Q1820.99 1486.59 1811.27 1486.59 Q1808 1486.59 1804.53 1485.93 Q1801.09 1485.3 1797.41 1484.02 L1797.41 1477.25 Q1800.33 1478.95 1803.8 1479.82 Q1807.27 1480.68 1811.06 1480.68 Q1817.66 1480.68 1821.09 1478.08 Q1824.57 1475.48 1824.57 1470.51 Q1824.57 1465.93 1821.34 1463.36 Q1818.14 1460.75 1812.41 1460.75 L1806.37 1460.75 L1806.37 1454.99 L1812.69 1454.99 Q1817.87 1454.99 1820.61 1452.94 Q1823.35 1450.86 1823.35 1446.97 Q1823.35 1442.98 1820.5 1440.86 Q1817.69 1438.71 1812.41 1438.71 Q1809.53 1438.71 1806.23 1439.33 Q1802.93 1439.96 1798.98 1441.27 L1798.98 1435.02 Q1802.97 1433.91 1806.44 1433.36 Q1809.95 1432.8 1813.04 1432.8 Q1821.02 1432.8 1825.68 1436.45 Q1830.33 1440.06 1830.33 1446.24 Q1830.33 1450.55 1827.87 1453.53 Q1825.4 1456.48 1820.85 1457.63 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M2299.6 1439.85 L2281.89 1467.52 L2299.6 1467.52 L2299.6 1439.85 M2297.76 1433.74 L2306.58 1433.74 L2306.58 1467.52 L2313.97 1467.52 L2313.97 1473.36 L2306.58 1473.36 L2306.58 1485.58 L2299.6 1485.58 L2299.6 1473.36 L2276.19 1473.36 L2276.19 1466.59 L2297.76 1433.74 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1277.49 1527.62 L1277.49 1538.66 L1290.65 1538.66 L1290.65 1543.62 L1277.49 1543.62 L1277.49 1564.74 Q1277.49 1569.49 1278.77 1570.85 Q1280.09 1572.2 1284.08 1572.2 L1290.65 1572.2 L1290.65 1577.55 L1284.08 1577.55 Q1276.69 1577.55 1273.88 1574.8 Q1271.06 1572.03 1271.06 1564.74 L1271.06 1543.62 L1266.38 1543.62 L1266.38 1538.66 L1271.06 1538.66 L1271.06 1527.62 L1277.49 1527.62 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1321.06 1523.52 L1335.79 1523.52 L1335.79 1528.49 L1327.45 1528.49 L1327.45 1581.96 L1335.79 1581.96 L1335.79 1586.92 L1321.06 1586.92 L1321.06 1523.52 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1374.19 1539.8 L1374.19 1545.85 Q1371.48 1544.46 1368.56 1543.76 Q1365.65 1543.07 1362.52 1543.07 Q1357.76 1543.07 1355.37 1544.53 Q1353.01 1545.99 1353.01 1548.9 Q1353.01 1551.12 1354.71 1552.41 Q1356.41 1553.66 1361.55 1554.8 L1363.74 1555.29 Q1370.54 1556.75 1373.39 1559.42 Q1376.27 1562.06 1376.27 1566.82 Q1376.27 1572.24 1371.97 1575.4 Q1367.7 1578.55 1360.2 1578.55 Q1357.07 1578.55 1353.67 1577.93 Q1350.3 1577.34 1346.55 1576.12 L1346.55 1569.53 Q1350.09 1571.37 1353.53 1572.3 Q1356.97 1573.21 1360.33 1573.21 Q1364.85 1573.21 1367.28 1571.68 Q1369.71 1570.12 1369.71 1567.3 Q1369.71 1564.7 1367.94 1563.31 Q1366.2 1561.92 1360.26 1560.64 L1358.04 1560.12 Q1352.11 1558.87 1349.47 1556.3 Q1346.83 1553.69 1346.83 1549.18 Q1346.83 1543.69 1350.72 1540.71 Q1354.6 1537.72 1361.76 1537.72 Q1365.3 1537.72 1368.42 1538.24 Q1371.55 1538.76 1374.19 1539.8 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1401.38 1523.52 L1401.38 1586.92 L1386.65 1586.92 L1386.65 1581.96 L1394.95 1581.96 L1394.95 1528.49 L1386.65 1528.49 L1386.65 1523.52 L1401.38 1523.52 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip202)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  314.996,1352.59 2352.76,1352.59 
  "/>
<polyline clip-path="url(#clip202)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  314.996,1155.93 2352.76,1155.93 
  "/>
<polyline clip-path="url(#clip202)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  314.996,959.265 2352.76,959.265 
  "/>
<polyline clip-path="url(#clip202)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  314.996,762.603 2352.76,762.603 
  "/>
<polyline clip-path="url(#clip202)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  314.996,565.941 2352.76,565.941 
  "/>
<polyline clip-path="url(#clip202)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  314.996,369.278 2352.76,369.278 
  "/>
<polyline clip-path="url(#clip202)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  314.996,172.616 2352.76,172.616 
  "/>
<polyline clip-path="url(#clip200)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  314.996,1401.58 314.996,123.472 
  "/>
<polyline clip-path="url(#clip200)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  314.996,1352.59 333.894,1352.59 
  "/>
<polyline clip-path="url(#clip200)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  314.996,1155.93 333.894,1155.93 
  "/>
<polyline clip-path="url(#clip200)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  314.996,959.265 333.894,959.265 
  "/>
<polyline clip-path="url(#clip200)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  314.996,762.603 333.894,762.603 
  "/>
<polyline clip-path="url(#clip200)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  314.996,565.941 333.894,565.941 
  "/>
<polyline clip-path="url(#clip200)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  314.996,369.278 333.894,369.278 
  "/>
<polyline clip-path="url(#clip200)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  314.996,172.616 333.894,172.616 
  "/>
<path clip-path="url(#clip200)" d="M120.073 1353.27 L164.587 1353.27 L164.587 1359.17 L120.073 1359.17 L120.073 1353.27 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M180.941 1372.61 L192.399 1372.61 L192.399 1333.06 L179.934 1335.56 L179.934 1329.17 L192.33 1326.67 L199.344 1326.67 L199.344 1372.61 L210.802 1372.61 L210.802 1378.51 L180.941 1378.51 L180.941 1372.61 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M224.968 1369.69 L232.295 1369.69 L232.295 1378.51 L224.968 1378.51 L224.968 1369.69 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M247.642 1326.67 L275.177 1326.67 L275.177 1332.57 L254.066 1332.57 L254.066 1345.28 Q255.593 1344.76 257.121 1344.52 Q258.649 1344.24 260.177 1344.24 Q268.857 1344.24 273.927 1349 Q278.996 1353.75 278.996 1361.88 Q278.996 1370.25 273.788 1374.9 Q268.579 1379.52 259.1 1379.52 Q255.836 1379.52 252.434 1378.96 Q249.066 1378.41 245.454 1377.29 L245.454 1370.25 Q248.579 1371.95 251.913 1372.78 Q255.246 1373.61 258.961 1373.61 Q264.968 1373.61 268.475 1370.45 Q271.982 1367.29 271.982 1361.88 Q271.982 1356.46 268.475 1353.3 Q264.968 1350.14 258.961 1350.14 Q256.149 1350.14 253.336 1350.77 Q250.559 1351.39 247.642 1352.71 L247.642 1326.67 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M118.58 1156.6 L163.094 1156.6 L163.094 1162.51 L118.58 1162.51 L118.58 1156.6 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M179.448 1175.94 L190.906 1175.94 L190.906 1136.4 L178.441 1138.9 L178.441 1132.51 L190.837 1130.01 L197.851 1130.01 L197.851 1175.94 L209.309 1175.94 L209.309 1181.85 L179.448 1181.85 L179.448 1175.94 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M223.475 1173.03 L230.802 1173.03 L230.802 1181.85 L223.475 1181.85 L223.475 1173.03 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M261.079 1134.63 Q255.663 1134.63 252.92 1139.97 Q250.211 1145.28 250.211 1155.98 Q250.211 1166.64 252.92 1171.99 Q255.663 1177.3 261.079 1177.3 Q266.531 1177.3 269.239 1171.99 Q271.982 1166.64 271.982 1155.98 Q271.982 1145.28 269.239 1139.97 Q266.531 1134.63 261.079 1134.63 M261.079 1129.07 Q269.795 1129.07 274.378 1135.98 Q278.996 1142.85 278.996 1155.98 Q278.996 1169.07 274.378 1175.98 Q269.795 1182.85 261.079 1182.85 Q252.364 1182.85 247.746 1175.98 Q243.163 1169.07 243.163 1155.98 Q243.163 1142.85 247.746 1135.98 Q252.364 1129.07 261.079 1129.07 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M120.073 959.942 L164.587 959.942 L164.587 965.845 L120.073 965.845 L120.073 959.942 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M194.726 937.963 Q189.309 937.963 186.566 943.31 Q183.858 948.623 183.858 959.317 Q183.858 969.977 186.566 975.324 Q189.309 980.636 194.726 980.636 Q200.177 980.636 202.885 975.324 Q205.628 969.977 205.628 959.317 Q205.628 948.623 202.885 943.31 Q200.177 937.963 194.726 937.963 M194.726 932.407 Q203.441 932.407 208.024 939.317 Q212.642 946.192 212.642 959.317 Q212.642 972.407 208.024 979.317 Q203.441 986.192 194.726 986.192 Q186.01 986.192 181.392 979.317 Q176.809 972.407 176.809 959.317 Q176.809 946.192 181.392 939.317 Q186.01 932.407 194.726 932.407 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M224.968 976.366 L232.295 976.366 L232.295 985.185 L224.968 985.185 L224.968 976.366 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M247.642 933.345 L275.177 933.345 L275.177 939.248 L254.066 939.248 L254.066 951.956 Q255.593 951.435 257.121 951.192 Q258.649 950.914 260.177 950.914 Q268.857 950.914 273.927 955.671 Q278.996 960.428 278.996 968.553 Q278.996 976.921 273.788 981.574 Q268.579 986.192 259.1 986.192 Q255.836 986.192 252.434 985.636 Q249.066 985.081 245.454 983.97 L245.454 976.921 Q248.579 978.622 251.913 979.456 Q255.246 980.289 258.961 980.289 Q264.968 980.289 268.475 977.129 Q271.982 973.97 271.982 968.553 Q271.982 963.136 268.475 959.977 Q264.968 956.817 258.961 956.817 Q256.149 956.817 253.336 957.442 Q250.559 958.067 247.642 959.386 L247.642 933.345 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M193.233 741.301 Q187.816 741.301 185.073 746.648 Q182.365 751.96 182.365 762.655 Q182.365 773.315 185.073 778.662 Q187.816 783.974 193.233 783.974 Q198.684 783.974 201.392 778.662 Q204.135 773.315 204.135 762.655 Q204.135 751.96 201.392 746.648 Q198.684 741.301 193.233 741.301 M193.233 735.745 Q201.948 735.745 206.531 742.655 Q211.149 749.53 211.149 762.655 Q211.149 775.745 206.531 782.655 Q201.948 789.53 193.233 789.53 Q184.517 789.53 179.899 782.655 Q175.316 775.745 175.316 762.655 Q175.316 749.53 179.899 742.655 Q184.517 735.745 193.233 735.745 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M223.475 779.703 L230.802 779.703 L230.802 788.523 L223.475 788.523 L223.475 779.703 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M261.079 741.301 Q255.663 741.301 252.92 746.648 Q250.211 751.96 250.211 762.655 Q250.211 773.315 252.92 778.662 Q255.663 783.974 261.079 783.974 Q266.531 783.974 269.239 778.662 Q271.982 773.315 271.982 762.655 Q271.982 751.96 269.239 746.648 Q266.531 741.301 261.079 741.301 M261.079 735.745 Q269.795 735.745 274.378 742.655 Q278.996 749.53 278.996 762.655 Q278.996 775.745 274.378 782.655 Q269.795 789.53 261.079 789.53 Q252.364 789.53 247.746 782.655 Q243.163 775.745 243.163 762.655 Q243.163 749.53 247.746 742.655 Q252.364 735.745 261.079 735.745 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M194.726 544.639 Q189.309 544.639 186.566 549.986 Q183.858 555.298 183.858 565.993 Q183.858 576.652 186.566 582 Q189.309 587.312 194.726 587.312 Q200.177 587.312 202.885 582 Q205.628 576.652 205.628 565.993 Q205.628 555.298 202.885 549.986 Q200.177 544.639 194.726 544.639 M194.726 539.083 Q203.441 539.083 208.024 545.993 Q212.642 552.868 212.642 565.993 Q212.642 579.083 208.024 585.993 Q203.441 592.868 194.726 592.868 Q186.01 592.868 181.392 585.993 Q176.809 579.083 176.809 565.993 Q176.809 552.868 181.392 545.993 Q186.01 539.083 194.726 539.083 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M224.968 583.041 L232.295 583.041 L232.295 591.861 L224.968 591.861 L224.968 583.041 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M247.642 540.021 L275.177 540.021 L275.177 545.923 L254.066 545.923 L254.066 558.632 Q255.593 558.111 257.121 557.868 Q258.649 557.59 260.177 557.59 Q268.857 557.59 273.927 562.347 Q278.996 567.104 278.996 575.229 Q278.996 583.597 273.788 588.25 Q268.579 592.868 259.1 592.868 Q255.836 592.868 252.434 592.312 Q249.066 591.756 245.454 590.645 L245.454 583.597 Q248.579 585.298 251.913 586.131 Q255.246 586.965 258.961 586.965 Q264.968 586.965 268.475 583.805 Q271.982 580.645 271.982 575.229 Q271.982 569.812 268.475 566.652 Q264.968 563.493 258.961 563.493 Q256.149 563.493 253.336 564.118 Q250.559 564.743 247.642 566.062 L247.642 540.021 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M179.448 389.296 L190.906 389.296 L190.906 349.747 L178.441 352.247 L178.441 345.858 L190.837 343.358 L197.851 343.358 L197.851 389.296 L209.309 389.296 L209.309 395.198 L179.448 395.198 L179.448 389.296 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M223.475 386.379 L230.802 386.379 L230.802 395.198 L223.475 395.198 L223.475 386.379 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M261.079 347.976 Q255.663 347.976 252.92 353.324 Q250.211 358.636 250.211 369.331 Q250.211 379.99 252.92 385.337 Q255.663 390.65 261.079 390.65 Q266.531 390.65 269.239 385.337 Q271.982 379.99 271.982 369.331 Q271.982 358.636 269.239 353.324 Q266.531 347.976 261.079 347.976 M261.079 342.421 Q269.795 342.421 274.378 349.331 Q278.996 356.206 278.996 369.331 Q278.996 382.421 274.378 389.33 Q269.795 396.205 261.079 396.205 Q252.364 396.205 247.746 389.33 Q243.163 382.421 243.163 369.331 Q243.163 356.206 247.746 349.331 Q252.364 342.421 261.079 342.421 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M180.941 192.634 L192.399 192.634 L192.399 153.085 L179.934 155.585 L179.934 149.196 L192.33 146.696 L199.344 146.696 L199.344 192.634 L210.802 192.634 L210.802 198.536 L180.941 198.536 L180.941 192.634 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M224.968 189.717 L232.295 189.717 L232.295 198.536 L224.968 198.536 L224.968 189.717 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M247.642 146.696 L275.177 146.696 L275.177 152.599 L254.066 152.599 L254.066 165.307 Q255.593 164.786 257.121 164.543 Q258.649 164.266 260.177 164.266 Q268.857 164.266 273.927 169.023 Q278.996 173.779 278.996 181.904 Q278.996 190.272 273.788 194.925 Q268.579 199.543 259.1 199.543 Q255.836 199.543 252.434 198.988 Q249.066 198.432 245.454 197.321 L245.454 190.272 Q248.579 191.974 251.913 192.807 Q255.246 193.64 258.961 193.64 Q264.968 193.64 268.475 190.481 Q271.982 187.321 271.982 181.904 Q271.982 176.488 268.475 173.328 Q264.968 170.168 258.961 170.168 Q256.149 170.168 253.336 170.793 Q250.559 171.418 247.642 172.738 L247.642 146.696 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M35.6047 1073.64 Q31.2992 1071.24 29.2506 1067.91 Q27.202 1064.57 27.202 1060.06 Q27.202 1053.98 31.4728 1050.68 Q35.7089 1047.39 43.5561 1047.39 L67.0282 1047.39 L67.0282 1053.81 L43.7644 1053.81 Q38.1741 1053.81 35.4658 1055.79 Q32.7575 1057.77 32.7575 1061.83 Q32.7575 1066.8 36.0561 1069.68 Q39.3547 1072.56 45.0491 1072.56 L67.0282 1072.56 L67.0282 1078.98 L43.7644 1078.98 Q38.1394 1078.98 35.4658 1080.96 Q32.7575 1082.94 32.7575 1087.07 Q32.7575 1091.97 36.0908 1094.85 Q39.3894 1097.73 45.0491 1097.73 L67.0282 1097.73 L67.0282 1104.16 L28.1395 1104.16 L28.1395 1097.73 L34.1811 1097.73 Q30.6047 1095.55 28.9034 1092.49 Q27.202 1089.43 27.202 1085.23 Q27.202 1081 29.3547 1078.05 Q31.5075 1075.06 35.6047 1073.64 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M47.4797 1016.97 Q47.4797 1024.71 49.2505 1027.7 Q51.0213 1030.68 55.2921 1030.68 Q58.6949 1030.68 60.7087 1028.46 Q62.6879 1026.21 62.6879 1022.35 Q62.6879 1017.04 58.9379 1013.84 Q55.1532 1010.61 48.9033 1010.61 L47.4797 1010.61 L47.4797 1016.97 M44.8408 1004.23 L67.0282 1004.23 L67.0282 1010.61 L61.1254 1010.61 Q64.6671 1012.8 66.3684 1016.07 Q68.0351 1019.33 68.0351 1024.05 Q68.0351 1030.02 64.7018 1033.57 Q61.3337 1037.07 55.7088 1037.07 Q49.1463 1037.07 45.813 1032.7 Q42.4797 1028.29 42.4797 1019.57 L42.4797 1010.61 L41.8547 1010.61 Q37.445 1010.61 35.0492 1013.53 Q32.6186 1016.41 32.6186 1021.66 Q32.6186 1024.99 33.4172 1028.15 Q34.2158 1031.31 35.813 1034.23 L29.9103 1034.23 Q28.5561 1030.72 27.8964 1027.42 Q27.202 1024.12 27.202 1021 Q27.202 1012.56 31.577 1008.39 Q35.9519 1004.23 44.8408 1004.23 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M29.2853 966.275 L35.3269 966.275 Q33.9381 968.983 33.2436 971.9 Q32.5492 974.817 32.5492 977.941 Q32.5492 982.698 34.0075 985.094 Q35.4658 987.455 38.3825 987.455 Q40.6047 987.455 41.8894 985.754 Q43.1394 984.053 44.2852 978.914 L44.7713 976.726 Q46.2297 969.921 48.9033 967.073 Q51.5421 964.192 56.299 964.192 Q61.7157 964.192 64.8754 968.497 Q68.0351 972.768 68.0351 980.268 Q68.0351 983.393 67.4101 986.796 Q66.8198 990.164 65.6046 993.914 L59.0074 993.914 Q60.8476 990.372 61.7851 986.935 Q62.6879 983.497 62.6879 980.129 Q62.6879 975.615 61.1601 973.185 Q59.5976 970.754 56.7852 970.754 Q54.181 970.754 52.7921 972.525 Q51.4032 974.261 50.1185 980.198 L49.5977 982.421 Q48.3477 988.358 45.7783 990.997 Q43.1741 993.636 38.6603 993.636 Q33.1742 993.636 30.1881 989.747 Q27.202 985.858 27.202 978.705 Q27.202 975.164 27.7228 972.039 Q28.2436 968.914 29.2853 966.275 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M29.2853 929.226 L35.3269 929.226 Q33.9381 931.935 33.2436 934.851 Q32.5492 937.768 32.5492 940.893 Q32.5492 945.65 34.0075 948.046 Q35.4658 950.407 38.3825 950.407 Q40.6047 950.407 41.8894 948.706 Q43.1394 947.004 44.2852 941.865 L44.7713 939.678 Q46.2297 932.872 48.9033 930.025 Q51.5421 927.143 56.299 927.143 Q61.7157 927.143 64.8754 931.449 Q68.0351 935.72 68.0351 943.219 Q68.0351 946.344 67.4101 949.747 Q66.8198 953.115 65.6046 956.865 L59.0074 956.865 Q60.8476 953.324 61.7851 949.886 Q62.6879 946.449 62.6879 943.081 Q62.6879 938.567 61.1601 936.136 Q59.5976 933.706 56.7852 933.706 Q54.181 933.706 52.7921 935.476 Q51.4032 937.213 50.1185 943.15 L49.5977 945.372 Q48.3477 951.31 45.7783 953.949 Q43.1741 956.587 38.6603 956.587 Q33.1742 956.587 30.1881 952.699 Q27.202 948.81 27.202 941.657 Q27.202 938.115 27.7228 934.99 Q28.2436 931.865 29.2853 929.226 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M28.1395 898.949 L28.1395 892.178 L60.7782 880.025 L28.1395 867.873 L28.1395 861.102 L67.0282 875.685 L67.0282 884.366 L28.1395 898.949 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M45.9866 819.019 L49.1116 819.019 L49.1116 848.394 Q55.7088 847.977 59.181 844.435 Q62.6185 840.859 62.6185 834.505 Q62.6185 830.824 61.7157 827.387 Q60.8129 823.915 59.0074 820.512 L65.049 820.512 Q66.5073 823.949 67.2712 827.56 Q68.0351 831.171 68.0351 834.887 Q68.0351 844.192 62.6185 849.644 Q57.2018 855.06 47.9658 855.06 Q38.4172 855.06 32.827 849.921 Q27.202 844.748 27.202 835.998 Q27.202 828.151 32.2714 823.602 Q37.3061 819.019 45.9866 819.019 M44.1116 825.408 Q38.8686 825.477 35.7436 828.359 Q32.6186 831.206 32.6186 835.928 Q32.6186 841.276 35.6394 844.505 Q38.6603 847.699 44.1463 848.185 L44.1116 825.408 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M13.0007 808.533 L13.0007 802.144 L67.0282 802.144 L67.0282 808.533 L13.0007 808.533 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M32.6186 773.706 Q32.6186 778.845 36.6464 781.831 Q40.6394 784.818 47.6185 784.818 Q54.5977 784.818 58.6254 781.866 Q62.6185 778.88 62.6185 773.706 Q62.6185 768.602 58.5907 765.616 Q54.5629 762.63 47.6185 762.63 Q40.7089 762.63 36.6811 765.616 Q32.6186 768.602 32.6186 773.706 M27.202 773.706 Q27.202 765.373 32.6186 760.616 Q38.0353 755.859 47.6185 755.859 Q57.1671 755.859 62.6185 760.616 Q68.0351 765.373 68.0351 773.706 Q68.0351 782.074 62.6185 786.831 Q57.1671 791.554 47.6185 791.554 Q38.0353 791.554 32.6186 786.831 Q27.202 782.074 27.202 773.706 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M29.6325 717.283 L35.6047 717.283 Q34.1117 719.991 33.3825 722.735 Q32.6186 725.443 32.6186 728.221 Q32.6186 734.436 36.5769 737.873 Q40.5005 741.311 47.6185 741.311 Q54.7366 741.311 58.6949 737.873 Q62.6185 734.436 62.6185 728.221 Q62.6185 725.443 61.8893 722.735 Q61.1254 719.991 59.6324 717.283 L65.5351 717.283 Q66.7851 719.957 67.4101 722.839 Q68.0351 725.686 68.0351 728.915 Q68.0351 737.7 62.5143 742.873 Q56.9935 748.047 47.6185 748.047 Q38.1047 748.047 32.6533 742.839 Q27.202 737.596 27.202 728.498 Q27.202 725.547 27.827 722.735 Q28.4173 719.922 29.6325 717.283 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M28.1395 706.172 L28.1395 699.783 L67.0282 699.783 L67.0282 706.172 L28.1395 706.172 M13.0007 706.172 L13.0007 699.783 L21.0909 699.783 L21.0909 706.172 L13.0007 706.172 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M17.0979 680.096 L28.1395 680.096 L28.1395 666.936 L33.1047 666.936 L33.1047 680.096 L54.2157 680.096 Q58.9726 680.096 60.3268 678.811 Q61.681 677.492 61.681 673.499 L61.681 666.936 L67.0282 666.936 L67.0282 673.499 Q67.0282 680.894 64.2851 683.707 Q61.5074 686.519 54.2157 686.519 L33.1047 686.519 L33.1047 691.207 L28.1395 691.207 L28.1395 686.519 L17.0979 686.519 L17.0979 680.096 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M70.6392 642.353 Q77.5837 645.061 79.7017 647.631 Q81.8197 650.2 81.8197 654.506 L81.8197 659.61 L76.4726 659.61 L76.4726 655.86 Q76.4726 653.221 75.2226 651.763 Q73.9726 650.304 69.3198 648.534 L66.4032 647.388 L28.1395 663.117 L28.1395 656.346 L58.556 644.193 L28.1395 632.041 L28.1395 625.27 L70.6392 642.353 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M13.0007 594.437 L13.0007 579.714 L17.9659 579.714 L17.9659 588.048 L71.4379 588.048 L71.4379 579.714 L76.4031 579.714 L76.4031 594.437 L13.0007 594.437 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M35.6047 535.826 Q31.2992 533.43 29.2506 530.097 Q27.202 526.763 27.202 522.249 Q27.202 516.173 31.4728 512.875 Q35.7089 509.576 43.5561 509.576 L67.0282 509.576 L67.0282 516 L43.7644 516 Q38.1741 516 35.4658 517.979 Q32.7575 519.958 32.7575 524.02 Q32.7575 528.986 36.0561 531.867 Q39.3547 534.749 45.0491 534.749 L67.0282 534.749 L67.0282 541.173 L43.7644 541.173 Q38.1394 541.173 35.4658 543.152 Q32.7575 545.131 32.7575 549.263 Q32.7575 554.159 36.0908 557.041 Q39.3894 559.923 45.0491 559.923 L67.0282 559.923 L67.0282 566.346 L28.1395 566.346 L28.1395 559.923 L34.1811 559.923 Q30.6047 557.735 28.9034 554.68 Q27.202 551.624 27.202 547.423 Q27.202 543.187 29.3547 540.236 Q31.5075 537.249 35.6047 535.826 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M15.1882 485.479 L15.1882 479.576 L73.6253 497.632 L73.6253 503.534 L15.1882 485.479 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M29.2853 448.083 L35.3269 448.083 Q33.9381 450.792 33.2436 453.708 Q32.5492 456.625 32.5492 459.75 Q32.5492 464.507 34.0075 466.903 Q35.4658 469.264 38.3825 469.264 Q40.6047 469.264 41.8894 467.562 Q43.1394 465.861 44.2852 460.722 L44.7713 458.535 Q46.2297 451.729 48.9033 448.882 Q51.5421 446 56.299 446 Q61.7157 446 64.8754 450.305 Q68.0351 454.576 68.0351 462.076 Q68.0351 465.201 67.4101 468.604 Q66.8198 471.972 65.6046 475.722 L59.0074 475.722 Q60.8476 472.18 61.7851 468.743 Q62.6879 465.305 62.6879 461.937 Q62.6879 457.423 61.1601 454.993 Q59.5976 452.562 56.7852 452.562 Q54.181 452.562 52.7921 454.333 Q51.4032 456.069 50.1185 462.007 L49.5977 464.229 Q48.3477 470.166 45.7783 472.805 Q43.1741 475.444 38.6603 475.444 Q33.1742 475.444 30.1881 471.555 Q27.202 467.666 27.202 460.514 Q27.202 456.972 27.7228 453.847 Q28.2436 450.722 29.2853 448.083 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M13.0007 420.896 L76.4031 420.896 L76.4031 435.618 L71.4379 435.618 L71.4379 427.319 L17.9659 427.319 L17.9659 435.618 L13.0007 435.618 L13.0007 420.896 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M600.602 12.096 L612.795 12.096 L628.229 53.2532 L643.744 12.096 L655.937 12.096 L655.937 72.576 L647.957 72.576 L647.957 19.4686 L632.361 60.9499 L624.137 60.9499 L608.541 19.4686 L608.541 72.576 L600.602 72.576 L600.602 12.096 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M689.438 32.4315 Q683.442 32.4315 679.959 37.1306 Q676.475 41.7891 676.475 49.9314 Q676.475 58.0738 679.918 62.7728 Q683.402 67.4314 689.438 67.4314 Q695.393 67.4314 698.876 62.7323 Q702.36 58.0333 702.36 49.9314 Q702.36 41.8701 698.876 37.1711 Q695.393 32.4315 689.438 32.4315 M689.438 26.1121 Q699.16 26.1121 704.71 32.4315 Q710.259 38.7509 710.259 49.9314 Q710.259 61.0714 704.71 67.4314 Q699.16 73.7508 689.438 73.7508 Q679.675 73.7508 674.125 67.4314 Q668.616 61.0714 668.616 49.9314 Q668.616 38.7509 674.125 32.4315 Q679.675 26.1121 689.438 26.1121 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M752.47 34.0924 L752.47 9.54393 L759.924 9.54393 L759.924 72.576 L752.47 72.576 L752.47 65.7705 Q750.12 69.8214 746.515 71.8063 Q742.95 73.7508 737.927 73.7508 Q729.704 73.7508 724.519 67.1883 Q719.374 60.6258 719.374 49.9314 Q719.374 39.2371 724.519 32.6746 Q729.704 26.1121 737.927 26.1121 Q742.95 26.1121 746.515 28.0971 Q750.12 30.0415 752.47 34.0924 M727.071 49.9314 Q727.071 58.1548 730.433 62.8538 Q733.836 67.5124 739.75 67.5124 Q745.664 67.5124 749.067 62.8538 Q752.47 58.1548 752.47 49.9314 Q752.47 41.7081 749.067 37.0496 Q745.664 32.3505 739.75 32.3505 Q733.836 32.3505 730.433 37.0496 Q727.071 41.7081 727.071 49.9314 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M775.276 27.2059 L782.73 27.2059 L782.73 72.576 L775.276 72.576 L775.276 27.2059 M775.276 9.54393 L782.73 9.54393 L782.73 18.9825 L775.276 18.9825 L775.276 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M821.295 9.54393 L821.295 15.7418 L814.165 15.7418 Q810.155 15.7418 808.575 17.3622 Q807.036 18.9825 807.036 23.1955 L807.036 27.2059 L819.31 27.2059 L819.31 32.9987 L807.036 32.9987 L807.036 72.576 L799.541 72.576 L799.541 32.9987 L792.412 32.9987 L792.412 27.2059 L799.541 27.2059 L799.541 24.0462 Q799.541 16.471 803.066 13.0277 Q806.59 9.54393 814.246 9.54393 L821.295 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M827.533 27.2059 L834.987 27.2059 L834.987 72.576 L827.533 72.576 L827.533 27.2059 M827.533 9.54393 L834.987 9.54393 L834.987 18.9825 L827.533 18.9825 L827.533 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M889.39 48.0275 L889.39 51.6733 L855.12 51.6733 Q855.606 59.3701 859.738 63.421 Q863.91 67.4314 871.323 67.4314 Q875.617 67.4314 879.628 66.3781 Q883.679 65.3249 887.649 63.2184 L887.649 70.267 Q883.638 71.9684 879.425 72.8596 Q875.212 73.7508 870.878 73.7508 Q860.021 73.7508 853.661 67.4314 Q847.342 61.1119 847.342 50.3365 Q847.342 39.1965 853.337 32.6746 Q859.373 26.1121 869.582 26.1121 Q878.737 26.1121 884.043 32.0264 Q889.39 37.9003 889.39 48.0275 M881.937 45.84 Q881.856 39.7232 878.494 36.0774 Q875.172 32.4315 869.663 32.4315 Q863.424 32.4315 859.657 35.9558 Q855.93 39.4801 855.363 45.8805 L881.937 45.84 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M931.479 34.0924 L931.479 9.54393 L938.933 9.54393 L938.933 72.576 L931.479 72.576 L931.479 65.7705 Q929.13 69.8214 925.525 71.8063 Q921.96 73.7508 916.937 73.7508 Q908.713 73.7508 903.528 67.1883 Q898.383 60.6258 898.383 49.9314 Q898.383 39.2371 903.528 32.6746 Q908.713 26.1121 916.937 26.1121 Q921.96 26.1121 925.525 28.0971 Q929.13 30.0415 931.479 34.0924 M906.08 49.9314 Q906.08 58.1548 909.442 62.8538 Q912.845 67.5124 918.76 67.5124 Q924.674 67.5124 928.077 62.8538 Q931.479 58.1548 931.479 49.9314 Q931.479 41.7081 928.077 37.0496 Q924.674 32.3505 918.76 32.3505 Q912.845 32.3505 909.442 37.0496 Q906.08 41.7081 906.08 49.9314 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M956.19 62.2867 L964.737 62.2867 L964.737 72.576 L956.19 72.576 L956.19 62.2867 M956.19 29.6769 L964.737 29.6769 L964.737 39.9662 L956.19 39.9662 L956.19 29.6769 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1008.93 12.096 L1021.13 12.096 L1036.56 53.2532 L1052.07 12.096 L1064.27 12.096 L1064.27 72.576 L1056.29 72.576 L1056.29 19.4686 L1040.69 60.9499 L1032.47 60.9499 L1016.87 19.4686 L1016.87 72.576 L1008.93 72.576 L1008.93 12.096 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1100.81 49.7694 Q1091.77 49.7694 1088.29 51.8354 Q1084.81 53.9013 1084.81 58.8839 Q1084.81 62.8538 1087.4 65.2034 Q1090.03 67.5124 1094.53 67.5124 Q1100.73 67.5124 1104.45 63.1374 Q1108.22 58.7219 1108.22 51.4303 L1108.22 49.7694 L1100.81 49.7694 M1115.67 46.6907 L1115.67 72.576 L1108.22 72.576 L1108.22 65.6895 Q1105.67 69.8214 1101.86 71.8063 Q1098.05 73.7508 1092.54 73.7508 Q1085.58 73.7508 1081.44 69.8619 Q1077.35 65.9325 1077.35 59.3701 Q1077.35 51.7138 1082.46 47.825 Q1087.6 43.9361 1097.77 43.9361 L1108.22 43.9361 L1108.22 43.2069 Q1108.22 38.0623 1104.82 35.2672 Q1101.46 32.4315 1095.34 32.4315 Q1091.45 32.4315 1087.76 33.3632 Q1084.08 34.295 1080.67 36.1584 L1080.67 29.2718 Q1084.77 27.692 1088.61 26.9223 Q1092.46 26.1121 1096.11 26.1121 Q1105.95 26.1121 1110.81 31.2163 Q1115.67 36.3204 1115.67 46.6907 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1159.95 28.5427 L1159.95 35.5912 Q1156.79 33.9709 1153.39 33.1607 Q1149.99 32.3505 1146.34 32.3505 Q1140.79 32.3505 1137.99 34.0519 Q1135.24 35.7533 1135.24 39.156 Q1135.24 41.7486 1137.22 43.2475 Q1139.21 44.7058 1145.21 46.0426 L1147.76 46.6097 Q1155.7 48.3111 1159.02 51.4303 Q1162.38 54.509 1162.38 60.0587 Q1162.38 66.3781 1157.36 70.0644 Q1152.38 73.7508 1143.63 73.7508 Q1139.98 73.7508 1136.01 73.0216 Q1132.08 72.3329 1127.71 70.9151 L1127.71 63.2184 Q1131.84 65.3654 1135.85 66.4591 Q1139.86 67.5124 1143.79 67.5124 Q1149.05 67.5124 1151.89 65.73 Q1154.72 63.9071 1154.72 60.6258 Q1154.72 57.5877 1152.66 55.9673 Q1150.63 54.3469 1143.71 52.8481 L1141.11 52.2405 Q1134.19 50.7821 1131.11 47.7845 Q1128.03 44.7463 1128.03 39.4801 Q1128.03 33.0797 1132.57 29.5959 Q1137.1 26.1121 1145.45 26.1121 Q1149.58 26.1121 1153.23 26.7198 Q1156.87 27.3274 1159.95 28.5427 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1203.17 28.5427 L1203.17 35.5912 Q1200.01 33.9709 1196.61 33.1607 Q1193.21 32.3505 1189.56 32.3505 Q1184.01 32.3505 1181.22 34.0519 Q1178.46 35.7533 1178.46 39.156 Q1178.46 41.7486 1180.45 43.2475 Q1182.43 44.7058 1188.43 46.0426 L1190.98 46.6097 Q1198.92 48.3111 1202.24 51.4303 Q1205.6 54.509 1205.6 60.0587 Q1205.6 66.3781 1200.58 70.0644 Q1195.6 73.7508 1186.85 73.7508 Q1183.2 73.7508 1179.23 73.0216 Q1175.3 72.3329 1170.93 70.9151 L1170.93 63.2184 Q1175.06 65.3654 1179.07 66.4591 Q1183.08 67.5124 1187.01 67.5124 Q1192.28 67.5124 1195.11 65.73 Q1197.95 63.9071 1197.95 60.6258 Q1197.95 57.5877 1195.88 55.9673 Q1193.86 54.3469 1186.93 52.8481 L1184.34 52.2405 Q1177.41 50.7821 1174.33 47.7845 Q1171.25 44.7463 1171.25 39.4801 Q1171.25 33.0797 1175.79 29.5959 Q1180.33 26.1121 1188.67 26.1121 Q1192.8 26.1121 1196.45 26.7198 Q1200.09 27.3274 1203.17 28.5427 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1238.5 27.2059 L1246.4 27.2059 L1260.57 65.2844 L1274.75 27.2059 L1282.65 27.2059 L1265.64 72.576 L1255.51 72.576 L1238.5 27.2059 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1331.75 48.0275 L1331.75 51.6733 L1297.48 51.6733 Q1297.96 59.3701 1302.1 63.421 Q1306.27 67.4314 1313.68 67.4314 Q1317.98 67.4314 1321.99 66.3781 Q1326.04 65.3249 1330.01 63.2184 L1330.01 70.267 Q1326 71.9684 1321.78 72.8596 Q1317.57 73.7508 1313.24 73.7508 Q1302.38 73.7508 1296.02 67.4314 Q1289.7 61.1119 1289.7 50.3365 Q1289.7 39.1965 1295.7 32.6746 Q1301.73 26.1121 1311.94 26.1121 Q1321.1 26.1121 1326.4 32.0264 Q1331.75 37.9003 1331.75 48.0275 M1324.3 45.84 Q1324.21 39.7232 1320.85 36.0774 Q1317.53 32.4315 1312.02 32.4315 Q1305.78 32.4315 1302.02 35.9558 Q1298.29 39.4801 1297.72 45.8805 L1324.3 45.84 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1343.98 9.54393 L1351.44 9.54393 L1351.44 72.576 L1343.98 72.576 L1343.98 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1384.61 32.4315 Q1378.62 32.4315 1375.13 37.1306 Q1371.65 41.7891 1371.65 49.9314 Q1371.65 58.0738 1375.09 62.7728 Q1378.58 67.4314 1384.61 67.4314 Q1390.57 67.4314 1394.05 62.7323 Q1397.54 58.0333 1397.54 49.9314 Q1397.54 41.8701 1394.05 37.1711 Q1390.57 32.4315 1384.61 32.4315 M1384.61 26.1121 Q1394.34 26.1121 1399.89 32.4315 Q1405.44 38.7509 1405.44 49.9314 Q1405.44 61.0714 1399.89 67.4314 Q1394.34 73.7508 1384.61 73.7508 Q1374.85 73.7508 1369.3 67.4314 Q1363.79 61.0714 1363.79 49.9314 Q1363.79 38.7509 1369.3 32.4315 Q1374.85 26.1121 1384.61 26.1121 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1450.44 28.9478 L1450.44 35.9153 Q1447.28 34.1734 1444.08 33.3227 Q1440.92 32.4315 1437.68 32.4315 Q1430.43 32.4315 1426.42 37.0496 Q1422.41 41.6271 1422.41 49.9314 Q1422.41 58.2358 1426.42 62.8538 Q1430.43 67.4314 1437.68 67.4314 Q1440.92 67.4314 1444.08 66.5807 Q1447.28 65.6895 1450.44 63.9476 L1450.44 70.8341 Q1447.32 72.2924 1443.96 73.0216 Q1440.64 73.7508 1436.87 73.7508 Q1426.62 73.7508 1420.59 67.3098 Q1414.55 60.8689 1414.55 49.9314 Q1414.55 38.832 1420.63 32.472 Q1426.74 26.1121 1437.36 26.1121 Q1440.8 26.1121 1444.08 26.8413 Q1447.36 27.5299 1450.44 28.9478 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1463.4 27.2059 L1470.86 27.2059 L1470.86 72.576 L1463.4 72.576 L1463.4 27.2059 M1463.4 9.54393 L1470.86 9.54393 L1470.86 18.9825 L1463.4 18.9825 L1463.4 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1493.83 14.324 L1493.83 27.2059 L1509.18 27.2059 L1509.18 32.9987 L1493.83 32.9987 L1493.83 57.6282 Q1493.83 63.1779 1495.32 64.7578 Q1496.86 66.3376 1501.52 66.3376 L1509.18 66.3376 L1509.18 72.576 L1501.52 72.576 Q1492.89 72.576 1489.61 69.3758 Q1486.33 66.1351 1486.33 57.6282 L1486.33 32.9987 L1480.86 32.9987 L1480.86 27.2059 L1486.33 27.2059 L1486.33 14.324 L1493.83 14.324 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1537.86 76.7889 Q1534.7 84.8907 1531.7 87.3618 Q1528.7 89.8329 1523.68 89.8329 L1517.73 89.8329 L1517.73 83.5945 L1522.1 83.5945 Q1525.18 83.5945 1526.88 82.1361 Q1528.58 80.6778 1530.65 75.2496 L1531.99 71.8468 L1513.63 27.2059 L1521.53 27.2059 L1535.71 62.6918 L1549.89 27.2059 L1557.79 27.2059 L1537.86 76.7889 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1615.07 49.7694 Q1606.04 49.7694 1602.55 51.8354 Q1599.07 53.9013 1599.07 58.8839 Q1599.07 62.8538 1601.66 65.2034 Q1604.29 67.5124 1608.79 67.5124 Q1614.99 67.5124 1618.72 63.1374 Q1622.48 58.7219 1622.48 51.4303 L1622.48 49.7694 L1615.07 49.7694 M1629.94 46.6907 L1629.94 72.576 L1622.48 72.576 L1622.48 65.6895 Q1619.93 69.8214 1616.12 71.8063 Q1612.31 73.7508 1606.81 73.7508 Q1599.84 73.7508 1595.71 69.8619 Q1591.61 65.9325 1591.61 59.3701 Q1591.61 51.7138 1596.72 47.825 Q1601.86 43.9361 1612.03 43.9361 L1622.48 43.9361 L1622.48 43.2069 Q1622.48 38.0623 1619.08 35.2672 Q1615.72 32.4315 1609.6 32.4315 Q1605.71 32.4315 1602.03 33.3632 Q1598.34 34.295 1594.94 36.1584 L1594.94 29.2718 Q1599.03 27.692 1602.88 26.9223 Q1606.72 26.1121 1610.37 26.1121 Q1620.21 26.1121 1625.08 31.2163 Q1629.94 36.3204 1629.94 46.6907 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1668.26 9.54393 L1668.26 15.7418 L1661.13 15.7418 Q1657.12 15.7418 1655.54 17.3622 Q1654 18.9825 1654 23.1955 L1654 27.2059 L1666.27 27.2059 L1666.27 32.9987 L1654 32.9987 L1654 72.576 L1646.5 72.576 L1646.5 32.9987 L1639.37 32.9987 L1639.37 27.2059 L1646.5 27.2059 L1646.5 24.0462 Q1646.5 16.471 1650.03 13.0277 Q1653.55 9.54393 1661.21 9.54393 L1668.26 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1680.41 14.324 L1680.41 27.2059 L1695.76 27.2059 L1695.76 32.9987 L1680.41 32.9987 L1680.41 57.6282 Q1680.41 63.1779 1681.91 64.7578 Q1683.45 66.3376 1688.11 66.3376 L1695.76 66.3376 L1695.76 72.576 L1688.11 72.576 Q1679.48 72.576 1676.2 69.3758 Q1672.92 66.1351 1672.92 57.6282 L1672.92 32.9987 L1667.45 32.9987 L1667.45 27.2059 L1672.92 27.2059 L1672.92 14.324 L1680.41 14.324 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1744.37 48.0275 L1744.37 51.6733 L1710.1 51.6733 Q1710.59 59.3701 1714.72 63.421 Q1718.89 67.4314 1726.31 67.4314 Q1730.6 67.4314 1734.61 66.3781 Q1738.66 65.3249 1742.63 63.2184 L1742.63 70.267 Q1738.62 71.9684 1734.41 72.8596 Q1730.2 73.7508 1725.86 73.7508 Q1715.01 73.7508 1708.65 67.4314 Q1702.33 61.1119 1702.33 50.3365 Q1702.33 39.1965 1708.32 32.6746 Q1714.36 26.1121 1724.57 26.1121 Q1733.72 26.1121 1739.03 32.0264 Q1744.37 37.9003 1744.37 48.0275 M1736.92 45.84 Q1736.84 39.7232 1733.48 36.0774 Q1730.16 32.4315 1724.65 32.4315 Q1718.41 32.4315 1714.64 35.9558 Q1710.91 39.4801 1710.35 45.8805 L1736.92 45.84 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1782.9 34.1734 Q1781.64 33.4443 1780.14 33.1202 Q1778.69 32.7556 1776.9 32.7556 Q1770.58 32.7556 1767.18 36.8875 Q1763.82 40.9789 1763.82 48.6757 L1763.82 72.576 L1756.32 72.576 L1756.32 27.2059 L1763.82 27.2059 L1763.82 34.2544 Q1766.17 30.1225 1769.94 28.1376 Q1773.7 26.1121 1779.09 26.1121 Q1779.86 26.1121 1780.79 26.2337 Q1781.72 26.3147 1782.86 26.5172 L1782.9 34.1734 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1846.09 44.2197 Q1848.73 45.1109 1851.2 48.0275 Q1853.71 50.9442 1856.22 56.0483 L1864.52 72.576 L1855.73 72.576 L1848 57.061 Q1845 50.9847 1842.16 48.9997 Q1839.37 47.0148 1834.51 47.0148 L1825.59 47.0148 L1825.59 72.576 L1817.41 72.576 L1817.41 12.096 L1835.88 12.096 Q1846.25 12.096 1851.36 16.4305 Q1856.46 20.7649 1856.46 29.5149 Q1856.46 35.2267 1853.79 38.994 Q1851.16 42.7613 1846.09 44.2197 M1825.59 18.8205 L1825.59 40.2903 L1835.88 40.2903 Q1841.8 40.2903 1844.8 37.5762 Q1847.83 34.8216 1847.83 29.5149 Q1847.83 24.2082 1844.8 21.5346 Q1841.8 18.8205 1835.88 18.8205 L1825.59 18.8205 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1870.24 54.671 L1870.24 27.2059 L1877.69 27.2059 L1877.69 54.3874 Q1877.69 60.8284 1880.2 64.0691 Q1882.71 67.2693 1887.74 67.2693 Q1893.77 67.2693 1897.26 63.421 Q1900.78 59.5726 1900.78 52.9291 L1900.78 27.2059 L1908.23 27.2059 L1908.23 72.576 L1900.78 72.576 L1900.78 65.6084 Q1898.07 69.7404 1894.46 71.7658 Q1890.9 73.7508 1886.16 73.7508 Q1878.34 73.7508 1874.29 68.8897 Q1870.24 64.0286 1870.24 54.671 M1888.99 26.1121 L1888.99 26.1121 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1961.3 45.1919 L1961.3 72.576 L1953.85 72.576 L1953.85 45.4349 Q1953.85 38.994 1951.34 35.7938 Q1948.82 32.5936 1943.8 32.5936 Q1937.76 32.5936 1934.28 36.4419 Q1930.8 40.2903 1930.8 46.9338 L1930.8 72.576 L1923.3 72.576 L1923.3 27.2059 L1930.8 27.2059 L1930.8 34.2544 Q1933.47 30.163 1937.08 28.1376 Q1940.72 26.1121 1945.46 26.1121 Q1953.28 26.1121 1957.29 30.9732 Q1961.3 35.7938 1961.3 45.1919 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1978.07 62.2867 L1986.62 62.2867 L1986.62 72.576 L1978.07 72.576 L1978.07 62.2867 M1978.07 29.6769 L1986.62 29.6769 L1986.62 39.9662 L1978.07 39.9662 L1978.07 29.6769 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M2038.59 65.6895 L2067.15 65.6895 L2067.15 72.576 L2028.75 72.576 L2028.75 65.6895 Q2033.41 60.8689 2041.43 52.7671 Q2049.49 44.6248 2051.55 42.2752 Q2055.48 37.8598 2057.02 34.8216 Q2058.6 31.7429 2058.6 28.7857 Q2058.6 23.9651 2055.2 20.927 Q2051.84 17.8888 2046.41 17.8888 Q2042.56 17.8888 2038.27 19.2256 Q2034.01 20.5624 2029.15 23.2765 L2029.15 15.0127 Q2034.09 13.0277 2038.39 12.015 Q2042.68 11.0023 2046.25 11.0023 Q2055.65 11.0023 2061.24 15.7013 Q2066.83 20.4004 2066.83 28.2591 Q2066.83 31.9859 2065.41 35.3482 Q2064.03 38.6699 2060.34 43.2069 Q2059.33 44.3817 2053.9 50.0125 Q2048.48 55.6027 2038.59 65.6895 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip202)" style="stroke:#009af9; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" points="
  372.668,1352.59 377.474,1348.35 382.28,1343.54 387.087,1338.14 391.893,1332.18 396.699,1325.64 401.505,1318.54 406.311,1310.89 411.117,1302.69 415.923,1293.95 
  420.729,1284.67 425.535,1274.88 430.341,1264.57 435.147,1253.76 439.953,1242.45 444.759,1230.67 449.565,1218.42 454.371,1205.72 459.177,1192.57 463.983,1178.99 
  468.789,1165 473.595,1150.6 478.401,1135.81 483.207,1120.66 488.013,1105.14 492.819,1089.28 497.625,1073.1 502.431,1056.6 507.237,1039.81 512.044,1022.74 
  516.85,1005.41 521.656,987.837 526.462,970.039 531.268,952.034 536.074,933.839 540.88,915.473 545.686,896.954 550.492,878.3 555.298,859.531 560.104,840.666 
  564.91,821.722 569.716,802.719 574.522,783.676 579.328,764.612 584.134,745.545 588.94,726.496 593.746,707.484 598.552,688.526 603.358,669.642 608.164,650.851 
  612.97,632.172 617.776,613.622 622.582,595.221 627.388,576.986 632.194,558.938 637,541.093 641.807,523.471 646.613,506.087 651.419,488.96 656.225,472.106 
  661.031,455.544 665.837,439.289 670.643,423.358 675.449,407.766 680.255,392.53 685.061,377.663 689.867,363.182 694.673,349.1 699.479,335.432 704.285,322.19 
  709.091,309.389 713.897,297.042 718.703,285.159 723.509,273.754 728.315,262.837 733.121,252.421 737.927,242.514 742.733,233.127 747.539,224.27 752.345,215.95 
  757.151,208.178 761.957,200.96 766.763,194.303 771.57,188.215 776.376,182.701 781.182,177.767 785.988,173.418 790.794,169.659 795.6,166.492 800.406,163.922 
  805.212,161.951 810.018,160.58 814.824,159.811 819.63,159.645 824.436,160.083 829.242,161.124 834.048,162.766 838.854,165.008 843.66,167.848 848.466,171.281 
  853.272,175.306 858.078,179.919 862.884,185.115 867.69,190.889 872.496,197.235 877.302,204.145 882.108,211.613 886.914,219.632 891.72,228.194 896.527,237.292 
  901.333,246.914 906.139,257.052 910.945,267.696 915.751,278.834 920.557,290.454 925.363,302.548 930.169,315.102 934.975,328.103 939.781,341.539 944.587,355.395 
  949.393,369.659 954.199,384.314 959.005,399.349 963.811,414.747 968.617,430.493 973.423,446.57 978.229,462.964 983.035,479.657 987.841,496.633 992.647,513.874 
  997.453,531.365 1002.26,549.087 1007.07,567.022 1011.87,585.153 1016.68,603.461 1021.48,621.928 1026.29,640.534 1031.1,659.264 1035.9,678.098 1040.71,697.015 
  1045.51,715.998 1050.32,735.027 1055.13,754.084 1059.93,773.148 1064.74,792.203 1069.54,811.228 1074.35,830.204 1079.16,849.113 1083.96,867.934 1088.77,886.651 
  1093.57,905.242 1098.38,923.692 1103.19,941.98 1107.99,960.089 1112.8,978 1117.6,995.695 1122.41,1013.16 1127.22,1030.37 1132.02,1047.31 1136.83,1063.97 
  1141.63,1080.33 1146.44,1096.37 1151.25,1112.08 1156.05,1127.43 1160.86,1142.42 1165.66,1157.04 1170.47,1171.25 1175.28,1185.06 1180.08,1198.45 1184.89,1211.4 
  1189.69,1223.9 1194.5,1235.94 1199.31,1247.5 1204.11,1258.58 1208.92,1269.17 1213.72,1279.25 1218.53,1288.81 1223.34,1297.85 1228.14,1306.35 1232.95,1314.3 
  1237.76,1321.71 1242.56,1328.55 1247.37,1334.83 1252.17,1340.54 1256.98,1345.67 1261.79,1350.22 1266.59,1354.18 1271.4,1357.54 1276.2,1360.31 1281.01,1362.49 
  1285.82,1364.06 1290.62,1365.04 1295.43,1365.41 1300.23,1365.18 1305.04,1364.34 1309.85,1362.9 1314.65,1360.87 1319.46,1358.23 1324.26,1355 1329.07,1351.18 
  1333.88,1346.77 1338.68,1341.77 1343.49,1336.2 1348.29,1330.05 1353.1,1323.33 1357.91,1316.05 1362.71,1308.22 1367.52,1299.85 1372.32,1290.94 1377.13,1281.5 
  1381.94,1271.54 1386.74,1261.07 1391.55,1250.1 1396.35,1238.64 1401.16,1226.71 1405.97,1214.32 1410.77,1201.47 1415.58,1188.19 1420.38,1174.48 1425.19,1160.35 
  1430,1145.83 1434.8,1130.93 1439.61,1115.66 1444.41,1100.04 1449.22,1084.07 1454.03,1067.79 1458.83,1051.2 1463.64,1034.32 1468.44,1017.18 1473.25,999.775 
  1478.06,982.134 1482.86,964.275 1487.67,946.213 1492.48,927.969 1497.28,909.559 1502.09,891.002 1506.89,872.318 1511.7,853.523 1516.51,834.638 1521.31,815.68 
  1526.12,796.67 1530.92,777.625 1535.73,758.566 1540.54,739.512 1545.34,720.48 1550.15,701.49 1554.95,682.561 1559.76,663.713 1564.57,644.964 1569.37,626.332 
  1574.18,607.837 1578.98,589.497 1583.79,571.33 1588.6,553.354 1593.4,535.588 1598.21,518.048 1603.01,500.754 1607.82,483.721 1612.63,466.967 1617.43,450.509 
  1622.24,434.364 1627.04,418.546 1631.85,403.073 1636.66,387.959 1641.46,373.22 1646.27,358.87 1651.07,344.925 1655.88,331.397 1660.69,318.301 1665.49,305.648 
  1670.3,293.453 1675.1,281.727 1679.91,270.481 1684.72,259.729 1689.52,249.479 1694.33,239.742 1699.13,230.529 1703.94,221.847 1708.75,213.705 1713.55,206.113 
  1718.36,199.078 1723.16,192.606 1727.97,186.705 1732.78,181.379 1737.58,176.634 1742.39,172.475 1747.2,168.906 1752,165.931 1756.81,163.553 1761.61,161.775 
  1766.42,160.597 1771.23,160.021 1776.03,160.047 1780.84,160.674 1785.64,161.906 1790.45,163.738 1795.26,166.169 1800.06,169.196 1804.87,172.816 1809.67,177.026 
  1814.48,181.82 1819.29,187.196 1824.09,193.149 1828.9,199.67 1833.7,206.755 1838.51,214.395 1843.32,222.582 1848.12,231.309 1852.93,240.568 1857.73,250.35 
  1862.54,260.644 1867.35,271.44 1872.15,282.726 1876.96,294.492 1881.76,306.726 1886.57,319.415 1891.38,332.549 1896.18,346.113 1900.99,360.093 1905.79,374.475 
  1910.6,389.245 1915.41,404.388 1920.21,419.889 1925.02,435.733 1929.82,451.905 1934.63,468.387 1939.44,485.162 1944.24,502.215 1949.05,519.528 1953.85,537.084 
  1958.66,554.866 1963.47,572.856 1968.27,591.035 1973.08,609.386 1977.88,627.889 1982.69,646.527 1987.5,665.281 1992.3,684.132 1997.11,703.062 2001.92,722.052 
  2006.72,741.082 2011.53,760.133 2016.33,779.186 2021.14,798.223 2025.95,817.224 2030.75,836.17 2035.56,855.043 2040.36,873.823 2045.17,892.492 2049.98,911.031 
  2054.78,929.421 2059.59,947.644 2064.39,965.682 2069.2,983.518 2074.01,1001.13 2078.81,1018.51 2083.62,1035.63 2088.42,1052.47 2093.23,1069.03 2098.04,1085.28 
  2102.84,1101.21 2107.65,1116.79 2112.45,1132.03 2117.26,1146.89 2122.07,1161.37 2126.87,1175.45 2131.68,1189.12 2136.48,1202.36 2141.29,1215.17 2146.1,1227.52 
  2150.9,1239.4 2155.71,1250.81 2160.51,1261.73 2165.32,1272.15 2170.13,1282.06 2174.93,1291.45 2179.74,1300.31 2184.54,1308.63 2189.35,1316.41 2194.16,1323.64 
  2198.96,1330.3 2203.77,1336.4 2208.57,1341.92 2213.38,1346.86 2218.19,1351.22 2222.99,1354.99 2227.8,1358.17 2232.6,1360.75 2237.41,1362.73 2242.22,1364.11 
  2247.02,1364.89 2251.83,1365.07 2256.64,1364.65 2261.44,1363.63 2266.25,1362 2271.05,1359.77 2275.86,1356.95 2280.67,1353.53 2285.47,1349.52 2290.28,1344.93 
  2295.08,1339.75 
  "/>
<polyline clip-path="url(#clip202)" style="stroke:#e26f46; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" points="
  372.668,1352.59 377.474,1350.91 382.28,1348.65 387.087,1345.8 391.893,1342.37 396.699,1338.37 401.505,1333.79 406.311,1328.63 411.117,1322.92 415.923,1316.64 
  420.729,1309.8 425.535,1302.42 430.341,1294.5 435.147,1286.04 439.953,1277.06 444.759,1267.56 449.565,1257.56 454.371,1247.05 459.177,1236.06 463.983,1224.59 
  468.789,1212.65 473.595,1200.26 478.401,1187.42 483.207,1174.15 488.013,1160.46 492.819,1146.36 497.625,1131.86 502.431,1116.97 507.237,1101.71 512.044,1086.1 
  516.85,1070.13 521.656,1053.84 526.462,1037.22 531.268,1020.3 536.074,1003.08 540.88,985.575 545.686,967.8 550.492,949.766 555.298,931.485 560.104,912.965 
  564.91,894.216 569.716,875.25 574.522,856.073 579.328,836.69 584.134,817.109 588.94,797.336 593.746,777.373 598.552,758.799 603.358,744.406 608.164,729.909 
  612.97,715.329 617.776,700.69 622.582,686.012 627.388,671.317 632.194,656.627 637,641.963 641.807,627.348 646.613,612.8 651.419,598.342 656.225,583.993 
  661.031,569.771 665.837,555.696 670.643,541.785 675.449,528.058 680.255,514.53 685.061,501.219 689.867,488.141 694.673,475.311 699.479,462.744 704.285,450.458 
  709.091,438.464 713.897,426.776 718.703,415.41 723.509,404.379 728.315,393.693 733.121,383.365 737.927,373.407 742.733,363.831 747.539,354.645 752.345,345.861 
  757.151,337.487 761.957,329.534 766.763,322.009 771.57,314.921 776.376,308.276 781.182,302.083 785.988,296.348 790.794,291.076 795.6,286.274 800.406,281.946 
  805.212,278.097 810.018,274.732 814.824,271.854 819.63,269.465 824.436,267.568 829.242,266.167 834.048,265.261 838.854,264.852 843.66,264.941 848.466,265.528 
  853.272,266.613 858.078,268.193 862.884,270.269 867.69,272.838 872.496,275.898 877.302,279.446 882.108,283.478 886.914,287.991 891.72,292.982 896.527,298.444 
  901.333,304.373 906.139,310.764 910.945,317.609 915.751,324.903 920.557,332.641 925.363,340.814 930.169,349.415 934.975,358.435 939.781,367.867 944.587,377.701 
  949.393,387.93 954.199,398.545 959.005,409.536 963.811,420.893 968.617,432.607 973.423,444.666 978.229,457.06 983.035,469.781 987.841,482.819 992.647,496.162 
  997.453,509.799 1002.26,523.721 1007.07,537.914 1011.87,552.37 1016.68,567.08 1021.48,582.032 1026.29,597.217 1031.1,612.624 1035.9,628.246 1040.71,644.074 
  1045.51,660.099 1050.32,676.315 1055.13,692.714 1059.93,709.29 1064.74,726.037 1069.54,742.953 1074.35,760.009 1079.16,772.216 1083.96,783.63 1088.77,795.119 
  1093.57,806.667 1098.38,818.257 1103.19,829.875 1107.99,841.503 1112.8,853.125 1117.6,864.724 1122.41,876.285 1127.22,887.792 1132.02,899.228 1136.83,910.58 
  1141.63,921.835 1146.44,932.974 1151.25,943.983 1156.05,954.848 1160.86,965.556 1165.66,976.092 1170.47,986.445 1175.28,996.601 1180.08,1006.55 1184.89,1016.28 
  1189.69,1025.77 1194.5,1035.02 1199.31,1044.02 1204.11,1052.75 1208.92,1061.21 1213.72,1069.39 1218.53,1077.27 1223.34,1084.85 1228.14,1092.11 1232.95,1099.06 
  1237.76,1105.68 1242.56,1111.96 1247.37,1117.9 1252.17,1123.5 1256.98,1128.73 1261.79,1133.61 1266.59,1138.12 1271.4,1142.26 1276.2,1146.02 1281.01,1149.41 
  1285.82,1152.41 1290.62,1155.02 1295.43,1157.24 1300.23,1159.07 1305.04,1160.5 1309.85,1161.54 1314.65,1162.18 1319.46,1162.42 1324.26,1162.26 1329.07,1161.69 
  1333.88,1160.73 1338.68,1159.38 1343.49,1157.62 1348.29,1155.46 1353.1,1152.92 1357.91,1149.97 1362.71,1146.64 1367.52,1142.92 1372.32,1138.82 1377.13,1134.34 
  1381.94,1129.48 1386.74,1124.25 1391.55,1118.65 1396.35,1112.69 1401.16,1106.37 1405.97,1099.7 1410.77,1092.68 1415.58,1085.32 1420.38,1077.63 1425.19,1069.61 
  1430,1061.28 1434.8,1052.63 1439.61,1043.67 1444.41,1034.42 1449.22,1024.87 1454.03,1015.04 1458.83,1004.94 1463.64,994.568 1468.44,983.939 1473.25,973.059 
  1478.06,961.935 1482.86,950.578 1487.67,938.994 1492.48,927.192 1497.28,915.18 1502.09,902.965 1506.89,890.556 1511.7,877.96 1516.51,865.184 1521.31,852.236 
  1526.12,839.122 1530.92,825.848 1535.73,812.42 1540.54,798.845 1545.34,785.128 1550.15,771.273 1554.95,759.534 1559.76,751.365 1564.57,743.135 1569.37,734.854 
  1574.18,726.534 1578.98,718.184 1583.79,709.817 1588.6,701.44 1593.4,693.068 1598.21,684.709 1603.01,676.377 1607.82,668.08 1612.63,659.831 1617.43,651.64 
  1622.24,643.518 1627.04,635.474 1631.85,627.519 1636.66,619.664 1641.46,611.917 1646.27,604.289 1651.07,596.789 1655.88,589.426 1660.69,582.21 1665.49,575.15 
  1670.3,568.253 1675.1,561.529 1679.91,554.985 1684.72,548.63 1689.52,542.471 1694.33,536.515 1699.13,530.77 1703.94,525.242 1708.75,519.938 1713.55,514.865 
  1718.36,510.028 1723.16,505.432 1727.97,501.084 1732.78,496.989 1737.58,493.15 1742.39,489.574 1747.2,486.264 1752,483.224 1756.81,480.458 1761.61,477.969 
  1766.42,475.76 1771.23,473.834 1776.03,472.194 1780.84,470.842 1785.64,469.779 1790.45,469.007 1795.26,468.528 1800.06,468.342 1804.87,468.451 1809.67,468.854 
  1814.48,469.552 1819.29,470.545 1824.09,471.832 1828.9,473.412 1833.7,475.285 1838.51,477.448 1843.32,479.901 1848.12,482.642 1852.93,485.669 1857.73,488.979 
  1862.54,492.57 1867.35,496.438 1872.15,500.581 1876.96,504.995 1881.76,509.677 1886.57,514.623 1891.38,519.829 1896.18,525.29 1900.99,531.003 1905.79,536.962 
  1910.6,543.163 1915.41,549.602 1920.21,556.273 1925.02,563.17 1929.82,570.288 1934.63,577.622 1939.44,585.166 1944.24,592.915 1949.05,600.863 1953.85,609.004 
  1958.66,617.332 1963.47,625.841 1968.27,634.526 1973.08,643.379 1977.88,652.396 1982.69,661.572 1987.5,670.899 1992.3,680.372 1997.11,689.985 2001.92,699.732 
  2006.72,709.608 2011.53,719.608 2016.33,729.727 2021.14,739.96 2025.95,750.3 2030.75,760.739 2035.56,766.414 2040.36,771.082 2045.17,775.782 2049.98,780.512 
  2054.78,785.264 2059.59,790.032 2064.39,794.812 2069.2,799.599 2074.01,804.386 2078.81,809.169 2083.62,813.941 2088.42,818.696 2093.23,823.428 2098.04,828.132 
  2102.84,832.802 2107.65,837.432 2112.45,842.017 2117.26,846.55 2122.07,851.027 2126.87,855.442 2131.68,859.789 2136.48,864.063 2141.29,868.259 2146.1,872.373 
  2150.9,876.397 2155.71,880.328 2160.51,884.161 2165.32,887.891 2170.13,891.514 2174.93,895.024 2179.74,898.418 2184.54,901.692 2189.35,904.84 2194.16,907.86 
  2198.96,910.749 2203.77,913.5 2208.57,916.111 2213.38,918.58 2218.19,920.903 2222.99,923.077 2227.8,925.097 2232.6,926.964 2237.41,928.673 2242.22,930.222 
  2247.02,931.608 2251.83,932.832 2256.64,933.89 2261.44,934.778 2266.25,935.498 2271.05,936.049 2275.86,936.427 2280.67,936.631 2285.47,936.664 2290.28,936.522 
  2295.08,936.205 
  "/>
<polyline clip-path="url(#clip202)" style="stroke:#008000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" stroke-dasharray="4, 8" points="
  372.668,762.603 377.474,745.392 382.28,728.171 387.087,710.951 391.893,693.743 396.699,676.558 401.505,659.408 406.311,642.306 411.117,625.264 415.923,608.295 
  420.729,591.412 425.535,574.63 430.341,557.963 435.147,541.425 439.953,525.032 444.759,508.801 449.565,492.747 454.371,476.887 459.177,461.238 463.983,445.819 
  468.789,430.647 473.595,415.742 478.401,401.123 483.207,386.811 488.013,372.823 492.819,359.181 497.625,345.906 502.431,333.016 507.237,320.534 512.044,308.48 
  516.85,296.873 521.656,285.733 526.462,275.081 531.268,264.937 536.074,255.323 540.88,246.255 545.686,237.751 550.492,229.826 555.298,222.495 560.104,215.77 
  564.91,209.662 569.716,204.183 574.522,199.341 579.328,195.142 584.134,191.595 588.94,188.703 593.746,186.469 598.552,184.895 603.358,183.977 608.164,183.713 
  612.97,184.099 617.776,185.128 622.582,186.794 627.388,189.09 632.194,192.006 637,195.533 641.807,199.659 646.613,204.373 651.419,209.66 656.225,215.506 
  661.031,221.896 665.837,228.812 670.643,236.237 675.449,244.157 680.255,252.558 685.061,261.421 689.867,270.73 694.673,280.468 699.479,290.619 704.285,301.165 
  709.091,312.091 713.897,323.38 718.703,335.017 723.509,346.984 728.315,359.267 733.121,371.85 737.927,384.717 742.733,397.854 747.539,411.244 752.345,424.874 
  757.151,438.728 761.957,452.792 766.763,467.053 771.57,481.497 776.376,496.112 781.182,510.884 785.988,525.801 790.794,540.849 795.6,556.017 800.406,571.29 
  805.212,586.657 810.018,602.104 814.824,617.62 819.63,633.192 824.436,648.808 829.242,664.455 834.048,680.122 838.854,695.796 843.66,711.465 848.466,727.117 
  853.272,742.741 858.078,758.324 862.884,773.856 867.69,789.324 872.496,804.717 877.302,820.023 882.108,835.231 886.914,850.326 891.72,865.296 896.527,880.127 
  901.333,894.807 906.139,909.322 910.945,923.659 915.751,937.805 920.557,951.747 925.363,965.472 930.169,978.968 934.975,992.221 939.781,1005.22 944.587,1017.95 
  949.393,1030.4 954.199,1042.55 959.005,1054.4 963.811,1065.93 968.617,1077.12 973.423,1087.98 978.229,1098.48 983.035,1108.62 987.841,1118.39 992.647,1127.78 
  997.453,1136.77 1002.26,1145.35 1007.07,1153.51 1011.87,1161.25 1016.68,1168.56 1021.48,1175.42 1026.29,1181.84 1031.1,1187.81 1035.9,1193.32 1040.71,1198.36 
  1045.51,1202.95 1050.32,1207.06 1055.13,1210.72 1059.93,1213.9 1064.74,1216.62 1069.54,1218.87 1074.35,1220.65 1079.16,1221.98 1083.96,1222.85 1088.77,1223.27 
  1093.57,1223.24 1098.38,1222.77 1103.19,1221.86 1107.99,1220.53 1112.8,1218.76 1117.6,1216.58 1122.41,1213.97 1127.22,1210.96 1132.02,1207.56 1136.83,1203.75 
  1141.63,1199.57 1146.44,1195.01 1151.25,1190.08 1156.05,1184.79 1160.86,1179.14 1165.66,1173.16 1170.47,1166.84 1175.28,1160.19 1180.08,1153.22 1184.89,1145.94 
  1189.69,1138.36 1194.5,1130.49 1199.31,1122.33 1204.11,1113.89 1208.92,1105.19 1213.72,1096.23 1218.53,1087.02 1223.34,1077.56 1228.14,1067.87 1232.95,1057.95 
  1237.76,1047.81 1242.56,1037.45 1247.37,1026.89 1252.17,1016.13 1256.98,1005.19 1261.79,994.062 1266.59,982.764 1271.4,971.303 1276.2,959.687 1281.01,947.924 
  1285.82,936.024 1290.62,923.996 1295.43,911.847 1300.23,899.588 1305.04,887.228 1309.85,874.774 1314.65,862.238 1319.46,849.628 1324.26,836.953 1329.07,824.224 
  1333.88,811.45 1338.68,798.64 1343.49,785.804 1348.29,772.953 1353.1,760.096 1357.91,747.243 1362.71,734.405 1367.52,721.592 1372.32,708.814 1377.13,696.081 
  1381.94,683.407 1386.74,670.803 1391.55,658.285 1396.35,645.864 1401.16,633.554 1405.97,621.369 1410.77,609.321 1415.58,597.425 1420.38,585.692 1425.19,574.138 
  1430,562.774 1434.8,551.616 1439.61,540.675 1444.41,529.966 1449.22,519.503 1454.03,509.298 1458.83,499.365 1463.64,489.719 1468.44,480.372 1473.25,471.339 
  1478.06,462.632 1482.86,454.267 1487.67,446.254 1492.48,438.595 1497.28,431.305 1502.09,424.397 1506.89,417.884 1511.7,411.777 1516.51,406.086 1521.31,400.821 
  1526.12,395.991 1530.92,391.604 1535.73,387.664 1540.54,384.18 1545.34,381.153 1550.15,378.59 1554.95,376.491 1559.76,374.858 1564.57,373.693 1569.37,372.993 
  1574.18,372.759 1578.98,372.987 1583.79,373.675 1588.6,374.817 1593.4,376.408 1598.21,378.441 1603.01,380.91 1607.82,383.806 1612.63,387.119 1617.43,390.838 
  1622.24,394.953 1627.04,399.458 1631.85,404.35 1636.66,409.618 1641.46,415.252 1646.27,421.24 1651.07,427.571 1655.88,434.233 1660.69,441.216 1665.49,448.508 
  1670.3,456.097 1675.1,463.973 1679.91,472.122 1684.72,480.534 1689.52,489.197 1694.33,498.099 1699.13,507.228 1703.94,516.571 1708.75,526.118 1713.55,535.855 
  1718.36,545.77 1723.16,555.851 1727.97,566.086 1732.78,576.462 1737.58,586.966 1742.39,597.586 1747.2,608.314 1752,619.137 1756.81,630.046 1761.61,641.027 
  1766.42,652.069 1771.23,663.16 1776.03,674.29 1780.84,685.446 1785.64,696.617 1790.45,707.793 1795.26,718.961 1800.06,730.111 1804.87,741.231 1809.67,752.311 
  1814.48,763.34 1819.29,774.307 1824.09,785.202 1828.9,796.012 1833.7,806.729 1838.51,817.341 1843.32,827.838 1848.12,838.21 1852.93,848.446 1857.73,858.537 
  1862.54,868.471 1867.35,878.239 1872.15,887.831 1876.96,897.237 1881.76,906.448 1886.57,915.452 1891.38,924.244 1896.18,932.816 1900.99,941.159 1905.79,949.262 
  1910.6,957.118 1915.41,964.719 1920.21,972.056 1925.02,979.122 1929.82,985.911 1934.63,992.414 1939.44,998.628 1944.24,1004.55 1949.05,1010.16 1953.85,1015.47 
  1958.66,1020.47 1963.47,1025.16 1968.27,1029.53 1973.08,1033.58 1977.88,1037.3 1982.69,1040.7 1987.5,1043.78 1992.3,1046.53 1997.11,1048.95 2001.92,1051.04 
  2006.72,1052.8 2011.53,1054.23 2016.33,1055.33 2021.14,1056.11 2025.95,1056.56 2030.75,1056.69 2035.56,1056.5 2040.36,1055.99 2045.17,1055.16 2049.98,1054.02 
  2054.78,1052.58 2059.59,1050.83 2064.39,1048.77 2069.2,1046.42 2074.01,1043.78 2078.81,1040.85 2083.62,1037.63 2088.42,1034.14 2093.23,1030.37 2098.04,1026.34 
  2102.84,1022.05 2107.65,1017.5 2112.45,1012.7 2117.26,1007.66 2122.07,1002.39 2126.87,996.888 2131.68,991.168 2136.48,985.237 2141.29,979.102 2146.1,972.767 
  2150.9,966.237 2155.71,959.519 2160.51,952.62 2165.32,945.547 2170.13,938.306 2174.93,930.906 2179.74,923.353 2184.54,915.655 2189.35,907.819 2194.16,899.852 
  2198.96,891.763 2203.77,883.559 2208.57,875.248 2213.38,866.837 2218.19,858.335 2222.99,849.749 2227.8,841.088 2232.6,832.36 2237.41,823.573 2242.22,814.736 
  2247.02,805.856 2251.83,796.943 2256.64,788.005 2261.44,779.051 2266.25,770.089 2271.05,761.128 2275.86,752.178 2280.67,743.247 2285.47,734.344 2290.28,725.479 
  2295.08,716.659 
  "/>
<polyline clip-path="url(#clip202)" style="stroke:#008000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" points="
  372.668,762.603 377.474,744.786 382.28,726.99 387.087,709.226 391.893,691.506 396.699,673.843 401.505,656.251 406.311,638.741 411.117,621.329 415.923,604.028 
  420.729,586.852 425.535,569.817 430.341,552.939 435.147,536.233 439.953,519.714 444.759,503.4 449.565,487.307 454.371,471.451 459.177,455.85 463.983,440.522 
  468.789,425.485 473.595,410.755 478.401,396.35 483.207,382.289 488.013,368.588 492.819,355.265 497.625,342.336 502.431,329.819 507.237,317.732 512.044,306.088 
  516.85,294.903 521.656,284.194 526.462,273.976 531.268,264.264 536.074,255.071 540.88,246.41 545.686,238.292 550.492,230.728 555.298,223.728 560.104,217.3 
  564.91,211.451 569.716,206.189 574.522,201.519 579.328,197.445 584.134,193.972 588.94,191.101 593.746,188.835 598.552,187.175 603.358,186.12 608.164,185.669 
  612.97,185.82 617.776,186.569 622.582,187.914 627.388,189.849 632.194,192.371 637,195.472 641.807,199.147 646.613,203.387 651.419,208.186 656.225,213.534 
  661.031,219.422 665.837,225.839 670.643,232.777 675.449,240.226 680.255,248.173 685.061,256.608 689.867,265.519 694.673,274.895 699.479,284.722 704.285,294.987 
  709.091,305.68 713.897,316.784 718.703,328.289 723.509,340.178 728.315,352.439 733.121,365.056 737.927,378.016 742.733,391.302 747.539,404.899 752.345,418.792 
  757.151,432.964 761.957,447.4 766.763,462.085 771.57,477.005 776.376,492.144 781.182,507.485 785.988,523.012 790.794,538.711 795.6,554.564 800.406,570.556 
  805.212,586.673 810.018,602.897 814.824,619.214 819.63,635.608 824.436,652.063 829.242,668.565 834.048,685.097 838.854,701.645 843.66,718.194 848.466,734.727 
  853.272,751.231 858.078,767.69 862.884,784.088 867.69,800.412 872.496,816.646 877.302,832.776 882.108,848.785 886.914,864.659 891.72,880.382 896.527,895.938 
  901.333,911.313 906.139,926.491 910.945,941.457 915.751,956.196 920.557,970.694 925.363,984.936 930.169,998.908 934.975,1012.6 939.781,1025.98 944.587,1039.06 
  949.393,1051.81 954.199,1064.22 959.005,1076.27 963.811,1087.96 968.617,1099.27 973.423,1110.19 978.229,1120.71 983.035,1130.81 987.841,1140.49 992.647,1149.74 
  997.453,1158.54 1002.26,1166.88 1007.07,1174.77 1011.87,1182.18 1016.68,1189.13 1021.48,1195.58 1026.29,1201.56 1031.1,1207.04 1035.9,1212.04 1040.71,1216.53 
  1045.51,1220.53 1050.32,1224.04 1055.13,1227.05 1059.93,1229.56 1064.74,1231.58 1069.54,1233.1 1074.35,1234.14 1079.16,1234.7 1083.96,1234.77 1088.77,1234.37 
  1093.57,1233.5 1098.38,1232.16 1103.19,1230.36 1107.99,1228.11 1112.8,1225.42 1117.6,1222.3 1122.41,1218.74 1127.22,1214.77 1132.02,1210.39 1136.83,1205.61 
  1141.63,1200.43 1146.44,1194.87 1151.25,1188.93 1156.05,1182.64 1160.86,1175.98 1165.66,1168.99 1170.47,1161.65 1175.28,1154 1180.08,1146.03 1184.89,1137.77 
  1189.69,1129.21 1194.5,1120.38 1199.31,1111.27 1204.11,1101.91 1208.92,1092.29 1213.72,1082.44 1218.53,1072.35 1223.34,1062.04 1228.14,1051.52 1232.95,1040.81 
  1237.76,1029.9 1242.56,1018.81 1247.37,1007.54 1252.17,996.12 1256.98,984.545 1261.79,972.828 1266.59,960.977 1271.4,949.004 1276.2,936.917 1281.01,924.726 
  1285.82,912.441 1290.62,900.069 1295.43,887.622 1300.23,875.108 1305.04,862.536 1309.85,849.916 1314.65,837.257 1319.46,824.567 1324.26,811.855 1329.07,799.132 
  1333.88,786.405 1338.68,773.684 1343.49,760.979 1348.29,748.301 1353.1,735.661 1357.91,723.07 1362.71,710.539 1367.52,698.079 1372.32,685.701 1377.13,673.415 
  1381.94,661.233 1386.74,649.166 1391.55,637.225 1396.35,625.421 1401.16,613.764 1405.97,602.267 1410.77,590.941 1415.58,579.795 1420.38,568.843 1425.19,558.094 
  1430,547.561 1434.8,537.254 1439.61,527.185 1444.41,517.366 1449.22,507.807 1454.03,498.519 1458.83,489.509 1463.64,480.786 1468.44,472.362 1473.25,464.249 
  1478.06,456.455 1482.86,448.991 1487.67,441.866 1492.48,435.088 1497.28,428.666 1502.09,422.606 1506.89,416.915 1511.7,411.599 1516.51,406.664 1521.31,402.115 
  1526.12,397.956 1530.92,394.19 1535.73,390.822 1540.54,387.854 1545.34,385.288 1550.15,383.124 1554.95,381.365 1559.76,380.01 1564.57,379.06 1569.37,378.512 
  1574.18,378.366 1578.98,378.62 1583.79,379.272 1588.6,380.319 1593.4,381.76 1598.21,383.59 1603.01,385.805 1607.82,388.401 1612.63,391.373 1617.43,394.717 
  1622.24,398.426 1627.04,402.496 1631.85,406.92 1636.66,411.692 1641.46,416.805 1646.27,422.251 1651.07,428.025 1655.88,434.117 1660.69,440.519 1665.49,447.224 
  1670.3,454.222 1675.1,461.505 1679.91,469.062 1684.72,476.884 1689.52,484.96 1694.33,493.281 1699.13,501.835 1703.94,510.61 1708.75,519.596 1713.55,528.784 
  1718.36,538.165 1723.16,547.729 1727.97,557.464 1732.78,567.36 1737.58,577.406 1742.39,587.589 1747.2,597.9 1752,608.326 1756.81,618.857 1761.61,629.481 
  1766.42,640.188 1771.23,650.965 1776.03,661.801 1780.84,672.686 1785.64,683.607 1790.45,694.555 1795.26,705.516 1800.06,716.481 1804.87,727.436 1809.67,738.373 
  1814.48,749.278 1819.29,760.14 1824.09,770.948 1828.9,781.691 1833.7,792.357 1838.51,802.934 1843.32,813.412 1848.12,823.778 1852.93,834.022 1857.73,844.131 
  1862.54,854.095 1867.35,863.908 1872.15,873.561 1876.96,883.043 1881.76,892.344 1886.57,901.453 1891.38,910.361 1896.18,919.058 1900.99,927.537 1905.79,935.787 
  1910.6,943.801 1915.41,951.572 1920.21,959.091 1925.02,966.351 1929.82,973.347 1934.63,980.07 1939.44,986.517 1944.24,992.68 1949.05,998.555 1953.85,1004.14 
  1958.66,1009.42 1963.47,1014.4 1968.27,1019.08 1973.08,1023.44 1977.88,1027.5 1982.69,1031.24 1987.5,1034.66 1992.3,1037.76 1997.11,1040.55 2001.92,1043.01 
  2006.72,1045.15 2011.53,1046.98 2016.33,1048.48 2021.14,1049.66 2025.95,1050.53 2030.75,1051.07 2035.56,1051.3 2040.36,1051.22 2045.17,1050.82 2049.98,1050.11 
  2054.78,1049.1 2059.59,1047.79 2064.39,1046.17 2069.2,1044.26 2074.01,1042.07 2078.81,1039.58 2083.62,1036.82 2088.42,1033.78 2093.23,1030.47 2098.04,1026.89 
  2102.84,1023.06 2107.65,1018.98 2112.45,1014.66 2117.26,1010.11 2122.07,1005.32 2126.87,1000.31 2131.68,995.077 2136.48,989.633 2141.29,983.983 2146.1,978.135 
  2150.9,972.096 2155.71,965.872 2160.51,959.471 2165.32,952.9 2170.13,946.166 2174.93,939.276 2179.74,932.239 2184.54,925.06 2189.35,917.748 2194.16,910.31 
  2198.96,902.753 2203.77,895.085 2208.57,887.313 2213.38,879.445 2218.19,871.489 2222.99,863.452 2227.8,855.342 2232.6,847.166 2237.41,838.932 2242.22,830.649 
  2247.02,822.323 2251.83,813.963 2256.64,805.577 2261.44,797.173 2266.25,788.757 2271.05,780.337 2275.86,771.921 2280.67,763.517 2285.47,755.132 2290.28,746.774 
  2295.08,738.451 
  "/>
<path clip-path="url(#clip200)" d="
M1423.56 554.876 L2284.83 554.876 L2284.83 166.076 L1423.56 166.076  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<polyline clip-path="url(#clip200)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1423.56,554.876 2284.83,554.876 2284.83,166.076 1423.56,166.076 1423.56,554.876 
  "/>
<polyline clip-path="url(#clip200)" style="stroke:#009af9; stroke-linecap:butt; stroke-linejoin:round; stroke-width:6; stroke-opacity:1; fill:none" points="
  1446.2,243.836 1582.05,243.836 
  "/>
<path clip-path="url(#clip200)" d="M1638.06 219.617 L1638.06 226.457 Q1634.06 224.548 1630.52 223.61 Q1626.98 222.673 1623.68 222.673 Q1617.95 222.673 1614.83 224.895 Q1611.74 227.117 1611.74 231.214 Q1611.74 234.652 1613.79 236.423 Q1615.87 238.159 1621.63 239.235 L1625.87 240.103 Q1633.72 241.596 1637.43 245.381 Q1641.18 249.131 1641.18 255.45 Q1641.18 262.985 1636.11 266.874 Q1631.08 270.763 1621.32 270.763 Q1617.64 270.763 1613.47 269.929 Q1609.34 269.096 1604.9 267.464 L1604.9 260.242 Q1609.17 262.638 1613.27 263.853 Q1617.36 265.068 1621.32 265.068 Q1627.33 265.068 1630.59 262.707 Q1633.86 260.346 1633.86 255.971 Q1633.86 252.152 1631.49 249.999 Q1629.17 247.846 1623.82 246.77 L1619.55 245.936 Q1611.7 244.374 1608.2 241.041 Q1604.69 237.707 1604.69 231.77 Q1604.69 224.895 1609.52 220.937 Q1614.38 216.978 1622.88 216.978 Q1626.53 216.978 1630.31 217.638 Q1634.1 218.298 1638.06 219.617 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1651.84 230.867 L1658.23 230.867 L1658.23 269.756 L1651.84 269.756 L1651.84 230.867 M1651.84 215.728 L1658.23 215.728 L1658.23 223.818 L1651.84 223.818 L1651.84 215.728 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1701.88 238.332 Q1704.27 234.027 1707.61 231.978 Q1710.94 229.93 1715.45 229.93 Q1721.53 229.93 1724.83 234.2 Q1728.13 238.436 1728.13 246.284 L1728.13 269.756 L1721.7 269.756 L1721.7 246.492 Q1721.7 240.902 1719.72 238.193 Q1717.74 235.485 1713.68 235.485 Q1708.72 235.485 1705.83 238.784 Q1702.95 242.082 1702.95 247.777 L1702.95 269.756 L1696.53 269.756 L1696.53 246.492 Q1696.53 240.867 1694.55 238.193 Q1692.57 235.485 1688.44 235.485 Q1683.54 235.485 1680.66 238.818 Q1677.78 242.117 1677.78 247.777 L1677.78 269.756 L1671.36 269.756 L1671.36 230.867 L1677.78 230.867 L1677.78 236.909 Q1679.97 233.332 1683.02 231.631 Q1686.08 229.93 1690.28 229.93 Q1694.52 229.93 1697.47 232.082 Q1700.45 234.235 1701.88 238.332 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1747.05 263.922 L1747.05 284.547 L1740.63 284.547 L1740.63 230.867 L1747.05 230.867 L1747.05 236.77 Q1749.06 233.298 1752.12 231.631 Q1755.21 229.93 1759.48 229.93 Q1766.56 229.93 1770.97 235.555 Q1775.42 241.179 1775.42 250.346 Q1775.42 259.513 1770.97 265.138 Q1766.56 270.763 1759.48 270.763 Q1755.21 270.763 1752.12 269.096 Q1749.06 267.395 1747.05 263.922 M1768.79 250.346 Q1768.79 243.298 1765.87 239.304 Q1762.99 235.277 1757.92 235.277 Q1752.85 235.277 1749.93 239.304 Q1747.05 243.298 1747.05 250.346 Q1747.05 257.395 1749.93 261.422 Q1752.85 265.415 1757.92 265.415 Q1762.99 265.415 1765.87 261.422 Q1768.79 257.395 1768.79 250.346 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1786.01 215.728 L1792.4 215.728 L1792.4 269.756 L1786.01 269.756 L1786.01 215.728 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1839.03 248.714 L1839.03 251.839 L1809.65 251.839 Q1810.07 258.436 1813.61 261.909 Q1817.19 265.346 1823.54 265.346 Q1827.22 265.346 1830.66 264.443 Q1834.13 263.54 1837.54 261.735 L1837.54 267.777 Q1834.1 269.235 1830.49 269.999 Q1826.88 270.763 1823.16 270.763 Q1813.85 270.763 1808.4 265.346 Q1802.99 259.929 1802.99 250.693 Q1802.99 241.145 1808.13 235.555 Q1813.3 229.93 1822.05 229.93 Q1829.9 229.93 1834.45 234.999 Q1839.03 240.034 1839.03 248.714 M1832.64 246.839 Q1832.57 241.596 1829.69 238.471 Q1826.84 235.346 1822.12 235.346 Q1816.77 235.346 1813.54 238.367 Q1810.35 241.388 1809.86 246.874 L1832.64 246.839 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1849.79 217.916 L1879.58 217.916 L1879.58 223.818 L1856.81 223.818 L1856.81 239.096 L1877.36 239.096 L1877.36 244.999 L1856.81 244.999 L1856.81 269.756 L1849.79 269.756 L1849.79 217.916 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1890.69 217.916 L1901.15 217.916 L1914.38 253.193 L1927.67 217.916 L1938.13 217.916 L1938.13 269.756 L1931.28 269.756 L1931.28 224.235 L1917.92 259.79 L1910.87 259.79 L1897.5 224.235 L1897.5 269.756 L1890.69 269.756 L1890.69 217.916 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1951.25 217.916 L1958.3 217.916 L1958.3 249.409 Q1958.3 257.742 1961.32 261.422 Q1964.34 265.068 1971.11 265.068 Q1977.85 265.068 1980.87 261.422 Q1983.89 257.742 1983.89 249.409 L1983.89 217.916 L1990.94 217.916 L1990.94 250.277 Q1990.94 260.415 1985.9 265.589 Q1980.9 270.763 1971.11 270.763 Q1961.28 270.763 1956.25 265.589 Q1951.25 260.415 1951.25 250.277 L1951.25 217.916 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip200)" style="stroke:#e26f46; stroke-linecap:butt; stroke-linejoin:round; stroke-width:6; stroke-opacity:1; fill:none" points="
  1446.2,321.596 1582.05,321.596 
  "/>
<path clip-path="url(#clip200)" d="M1629.27 323.21 Q1631.53 323.974 1633.65 326.474 Q1635.8 328.974 1637.95 333.349 L1645.07 347.516 L1637.54 347.516 L1630.9 334.217 Q1628.34 329.009 1625.9 327.307 Q1623.51 325.606 1619.34 325.606 L1611.7 325.606 L1611.7 347.516 L1604.69 347.516 L1604.69 295.676 L1620.52 295.676 Q1629.41 295.676 1633.79 299.391 Q1638.16 303.106 1638.16 310.606 Q1638.16 315.502 1635.87 318.731 Q1633.61 321.96 1629.27 323.21 M1611.7 301.44 L1611.7 319.842 L1620.52 319.842 Q1625.59 319.842 1628.16 317.516 Q1630.77 315.155 1630.77 310.606 Q1630.77 306.058 1628.16 303.766 Q1625.59 301.44 1620.52 301.44 L1611.7 301.44 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1683.89 326.474 L1683.89 329.599 L1654.52 329.599 Q1654.93 336.196 1658.47 339.669 Q1662.05 343.106 1668.4 343.106 Q1672.08 343.106 1675.52 342.203 Q1678.99 341.3 1682.4 339.495 L1682.4 345.537 Q1678.96 346.995 1675.35 347.759 Q1671.74 348.523 1668.02 348.523 Q1658.72 348.523 1653.27 343.106 Q1647.85 337.689 1647.85 328.453 Q1647.85 318.905 1652.99 313.315 Q1658.16 307.69 1666.91 307.69 Q1674.76 307.69 1679.31 312.759 Q1683.89 317.794 1683.89 326.474 M1677.5 324.599 Q1677.43 319.356 1674.55 316.231 Q1671.7 313.106 1666.98 313.106 Q1661.63 313.106 1658.4 316.127 Q1655.21 319.148 1654.72 324.634 L1677.5 324.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1714.06 293.488 L1714.06 298.801 L1707.95 298.801 Q1704.52 298.801 1703.16 300.19 Q1701.84 301.578 1701.84 305.19 L1701.84 308.627 L1712.36 308.627 L1712.36 313.592 L1701.84 313.592 L1701.84 347.516 L1695.42 347.516 L1695.42 313.592 L1689.31 313.592 L1689.31 308.627 L1695.42 308.627 L1695.42 305.919 Q1695.42 299.426 1698.44 296.474 Q1701.46 293.488 1708.02 293.488 L1714.06 293.488 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1752.67 326.474 L1752.67 329.599 L1723.3 329.599 Q1723.72 336.196 1727.26 339.669 Q1730.83 343.106 1737.19 343.106 Q1740.87 343.106 1744.31 342.203 Q1747.78 341.3 1751.18 339.495 L1751.18 345.537 Q1747.74 346.995 1744.13 347.759 Q1740.52 348.523 1736.81 348.523 Q1727.5 348.523 1722.05 343.106 Q1716.63 337.689 1716.63 328.453 Q1716.63 318.905 1721.77 313.315 Q1726.95 307.69 1735.7 307.69 Q1743.54 307.69 1748.09 312.759 Q1752.67 317.794 1752.67 326.474 M1746.29 324.599 Q1746.22 319.356 1743.33 316.231 Q1740.49 313.106 1735.76 313.106 Q1730.42 313.106 1727.19 316.127 Q1723.99 319.148 1723.51 324.634 L1746.29 324.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1785.7 314.599 Q1784.62 313.974 1783.33 313.696 Q1782.08 313.384 1780.56 313.384 Q1775.14 313.384 1772.22 316.926 Q1769.34 320.433 1769.34 327.03 L1769.34 347.516 L1762.92 347.516 L1762.92 308.627 L1769.34 308.627 L1769.34 314.669 Q1771.36 311.127 1774.58 309.426 Q1777.81 307.69 1782.43 307.69 Q1783.09 307.69 1783.89 307.794 Q1784.69 307.863 1785.66 308.037 L1785.7 314.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1824.1 326.474 L1824.1 329.599 L1794.72 329.599 Q1795.14 336.196 1798.68 339.669 Q1802.26 343.106 1808.61 343.106 Q1812.29 343.106 1815.73 342.203 Q1819.2 341.3 1822.6 339.495 L1822.6 345.537 Q1819.17 346.995 1815.56 347.759 Q1811.95 348.523 1808.23 348.523 Q1798.92 348.523 1793.47 343.106 Q1788.06 337.689 1788.06 328.453 Q1788.06 318.905 1793.2 313.315 Q1798.37 307.69 1807.12 307.69 Q1814.97 307.69 1819.51 312.759 Q1824.1 317.794 1824.1 326.474 M1817.71 324.599 Q1817.64 319.356 1814.76 316.231 Q1811.91 313.106 1807.19 313.106 Q1801.84 313.106 1798.61 316.127 Q1795.42 319.148 1794.93 324.634 L1817.71 324.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1866.91 324.044 L1866.91 347.516 L1860.52 347.516 L1860.52 324.252 Q1860.52 318.731 1858.37 315.988 Q1856.22 313.245 1851.91 313.245 Q1846.74 313.245 1843.75 316.544 Q1840.76 319.842 1840.76 325.537 L1840.76 347.516 L1834.34 347.516 L1834.34 308.627 L1840.76 308.627 L1840.76 314.669 Q1843.06 311.162 1846.15 309.426 Q1849.27 307.69 1853.33 307.69 Q1860.04 307.69 1863.47 311.856 Q1866.91 315.988 1866.91 324.044 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1907.64 310.12 L1907.64 316.092 Q1904.93 314.599 1902.19 313.87 Q1899.48 313.106 1896.7 313.106 Q1890.49 313.106 1887.05 317.064 Q1883.61 320.988 1883.61 328.106 Q1883.61 335.224 1887.05 339.182 Q1890.49 343.106 1896.7 343.106 Q1899.48 343.106 1902.19 342.377 Q1904.93 341.613 1907.64 340.12 L1907.64 346.023 Q1904.97 347.273 1902.08 347.898 Q1899.24 348.523 1896.01 348.523 Q1887.22 348.523 1882.05 343.002 Q1876.88 337.481 1876.88 328.106 Q1876.88 318.592 1882.08 313.141 Q1887.33 307.69 1896.42 307.69 Q1899.38 307.69 1902.19 308.315 Q1905 308.905 1907.64 310.12 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1952.01 326.474 L1952.01 329.599 L1922.64 329.599 Q1923.06 336.196 1926.6 339.669 Q1930.17 343.106 1936.53 343.106 Q1940.21 343.106 1943.65 342.203 Q1947.12 341.3 1950.52 339.495 L1950.52 345.537 Q1947.08 346.995 1943.47 347.759 Q1939.86 348.523 1936.15 348.523 Q1926.84 348.523 1921.39 343.106 Q1915.97 337.689 1915.97 328.453 Q1915.97 318.905 1921.11 313.315 Q1926.28 307.69 1935.03 307.69 Q1942.88 307.69 1947.43 312.759 Q1952.01 317.794 1952.01 326.474 M1945.62 324.599 Q1945.56 319.356 1942.67 316.231 Q1939.83 313.106 1935.1 313.106 Q1929.76 313.106 1926.53 316.127 Q1923.33 319.148 1922.85 324.634 L1945.62 324.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip200)" style="stroke:#008000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:6; stroke-opacity:1; fill:none" stroke-dasharray="3, 6" points="
  1446.2,399.356 1582.05,399.356 
  "/>
<path clip-path="url(#clip200)" d="M1604.69 373.436 L1614.13 373.436 L1637.12 416.804 L1637.12 373.436 L1643.93 373.436 L1643.93 425.276 L1634.48 425.276 L1611.49 381.908 L1611.49 425.276 L1604.69 425.276 L1604.69 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1690.87 404.234 L1690.87 407.359 L1661.49 407.359 Q1661.91 413.956 1665.45 417.429 Q1669.03 420.866 1675.38 420.866 Q1679.06 420.866 1682.5 419.963 Q1685.97 419.06 1689.38 417.255 L1689.38 423.297 Q1685.94 424.755 1682.33 425.519 Q1678.72 426.283 1675 426.283 Q1665.7 426.283 1660.24 420.866 Q1654.83 415.449 1654.83 406.213 Q1654.83 396.665 1659.97 391.075 Q1665.14 385.45 1673.89 385.45 Q1681.74 385.45 1686.29 390.519 Q1690.87 395.554 1690.87 404.234 M1684.48 402.359 Q1684.41 397.116 1681.53 393.991 Q1678.68 390.866 1673.96 390.866 Q1668.61 390.866 1665.38 393.887 Q1662.19 396.908 1661.7 402.394 L1684.48 402.359 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1700.7 409.929 L1700.7 386.387 L1707.08 386.387 L1707.08 409.686 Q1707.08 415.206 1709.24 417.984 Q1711.39 420.727 1715.7 420.727 Q1720.87 420.727 1723.86 417.429 Q1726.88 414.13 1726.88 408.436 L1726.88 386.387 L1733.27 386.387 L1733.27 425.276 L1726.88 425.276 L1726.88 419.304 Q1724.55 422.845 1721.46 424.581 Q1718.4 426.283 1714.34 426.283 Q1707.64 426.283 1704.17 422.116 Q1700.7 417.949 1700.7 409.929 M1716.77 385.45 L1716.77 385.45 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1768.96 392.359 Q1767.88 391.734 1766.6 391.456 Q1765.35 391.144 1763.82 391.144 Q1758.4 391.144 1755.49 394.686 Q1752.61 398.193 1752.61 404.79 L1752.61 425.276 L1746.18 425.276 L1746.18 386.387 L1752.61 386.387 L1752.61 392.429 Q1754.62 388.887 1757.85 387.186 Q1761.08 385.45 1765.7 385.45 Q1766.36 385.45 1767.15 385.554 Q1767.95 385.623 1768.92 385.797 L1768.96 392.359 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1793.33 405.727 Q1785.59 405.727 1782.61 407.498 Q1779.62 409.269 1779.62 413.54 Q1779.62 416.942 1781.84 418.956 Q1784.1 420.935 1787.95 420.935 Q1793.26 420.935 1796.46 417.185 Q1799.69 413.401 1799.69 407.151 L1799.69 405.727 L1793.33 405.727 M1806.08 403.088 L1806.08 425.276 L1799.69 425.276 L1799.69 419.373 Q1797.5 422.915 1794.24 424.616 Q1790.97 426.283 1786.25 426.283 Q1780.28 426.283 1776.74 422.949 Q1773.23 419.581 1773.23 413.956 Q1773.23 407.394 1777.61 404.061 Q1782.01 400.727 1790.73 400.727 L1799.69 400.727 L1799.69 400.102 Q1799.69 395.693 1796.77 393.297 Q1793.89 390.866 1788.65 390.866 Q1785.31 390.866 1782.15 391.665 Q1778.99 392.463 1776.08 394.061 L1776.08 388.158 Q1779.58 386.804 1782.88 386.144 Q1786.18 385.45 1789.31 385.45 Q1797.74 385.45 1801.91 389.825 Q1806.08 394.199 1806.08 403.088 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1819.24 371.248 L1825.63 371.248 L1825.63 425.276 L1819.24 425.276 L1819.24 371.248 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1839.27 373.436 L1869.06 373.436 L1869.06 379.338 L1846.29 379.338 L1846.29 394.616 L1866.84 394.616 L1866.84 400.519 L1846.29 400.519 L1846.29 425.276 L1839.27 425.276 L1839.27 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1880.17 373.436 L1890.63 373.436 L1903.85 408.713 L1917.15 373.436 L1927.6 373.436 L1927.6 425.276 L1920.76 425.276 L1920.76 379.755 L1907.4 415.31 L1900.35 415.31 L1886.98 379.755 L1886.98 425.276 L1880.17 425.276 L1880.17 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1940.73 373.436 L1947.78 373.436 L1947.78 404.929 Q1947.78 413.262 1950.8 416.942 Q1953.82 420.588 1960.59 420.588 Q1967.33 420.588 1970.35 416.942 Q1973.37 413.262 1973.37 404.929 L1973.37 373.436 L1980.42 373.436 L1980.42 405.797 Q1980.42 415.935 1975.38 421.109 Q1970.38 426.283 1960.59 426.283 Q1950.76 426.283 1945.73 421.109 Q1940.73 415.935 1940.73 405.797 L1940.73 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M2031.25 371.318 Q2026.6 379.304 2024.34 387.116 Q2022.08 394.929 2022.08 402.949 Q2022.08 410.97 2024.34 418.852 Q2026.63 426.699 2031.25 434.651 L2025.69 434.651 Q2020.49 426.491 2017.88 418.609 Q2015.31 410.727 2015.31 402.949 Q2015.31 395.206 2017.88 387.359 Q2020.45 379.512 2025.69 371.318 L2031.25 371.318 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M2050.59 419.373 L2075.07 419.373 L2075.07 425.276 L2042.15 425.276 L2042.15 419.373 Q2046.15 415.241 2053.02 408.297 Q2059.93 401.318 2061.7 399.304 Q2065.07 395.519 2066.39 392.915 Q2067.74 390.276 2067.74 387.741 Q2067.74 383.609 2064.83 381.005 Q2061.94 378.401 2057.29 378.401 Q2053.99 378.401 2050.31 379.547 Q2046.67 380.693 2042.5 383.019 L2042.5 375.936 Q2046.74 374.234 2050.42 373.366 Q2054.1 372.498 2057.15 372.498 Q2065.21 372.498 2070 376.526 Q2074.79 380.554 2074.79 387.29 Q2074.79 390.484 2073.58 393.366 Q2072.4 396.213 2069.24 400.102 Q2068.37 401.109 2063.71 405.936 Q2059.06 410.727 2050.59 419.373 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M2089.86 373.436 L2117.39 373.436 L2117.39 379.338 L2096.28 379.338 L2096.28 392.047 Q2097.81 391.526 2099.34 391.283 Q2100.87 391.005 2102.39 391.005 Q2111.08 391.005 2116.14 395.762 Q2121.21 400.519 2121.21 408.644 Q2121.21 417.012 2116.01 421.665 Q2110.8 426.283 2101.32 426.283 Q2098.05 426.283 2094.65 425.727 Q2091.28 425.172 2087.67 424.06 L2087.67 417.012 Q2090.8 418.713 2094.13 419.547 Q2097.46 420.38 2101.18 420.38 Q2107.19 420.38 2110.69 417.22 Q2114.2 414.06 2114.2 408.644 Q2114.2 403.227 2110.69 400.068 Q2107.19 396.908 2101.18 396.908 Q2098.37 396.908 2095.55 397.533 Q2092.78 398.158 2089.86 399.477 L2089.86 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M2150.03 378.054 Q2144.62 378.054 2141.87 383.401 Q2139.17 388.713 2139.17 399.408 Q2139.17 410.067 2141.87 415.415 Q2144.62 420.727 2150.03 420.727 Q2155.48 420.727 2158.19 415.415 Q2160.94 410.067 2160.94 399.408 Q2160.94 388.713 2158.19 383.401 Q2155.48 378.054 2150.03 378.054 M2150.03 372.498 Q2158.75 372.498 2163.33 379.408 Q2167.95 386.283 2167.95 399.408 Q2167.95 412.498 2163.33 419.408 Q2158.75 426.283 2150.03 426.283 Q2141.32 426.283 2136.7 419.408 Q2132.12 412.498 2132.12 399.408 Q2132.12 386.283 2136.7 379.408 Q2141.32 372.498 2150.03 372.498 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M2195.28 378.054 Q2189.86 378.054 2187.12 383.401 Q2184.41 388.713 2184.41 399.408 Q2184.41 410.067 2187.12 415.415 Q2189.86 420.727 2195.28 420.727 Q2200.73 420.727 2203.44 415.415 Q2206.18 410.067 2206.18 399.408 Q2206.18 388.713 2203.44 383.401 Q2200.73 378.054 2195.28 378.054 M2195.28 372.498 Q2203.99 372.498 2208.57 379.408 Q2213.19 386.283 2213.19 399.408 Q2213.19 412.498 2208.57 419.408 Q2203.99 426.283 2195.28 426.283 Q2186.56 426.283 2181.94 419.408 Q2177.36 412.498 2177.36 399.408 Q2177.36 386.283 2181.94 379.408 Q2186.56 372.498 2195.28 372.498 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M2223.61 371.318 L2229.17 371.318 Q2234.37 379.512 2236.94 387.359 Q2239.55 395.206 2239.55 402.949 Q2239.55 410.727 2236.94 418.609 Q2234.37 426.491 2229.17 434.651 L2223.61 434.651 Q2228.23 426.699 2230.48 418.852 Q2232.78 410.97 2232.78 402.949 Q2232.78 394.929 2230.48 387.116 Q2228.23 379.304 2223.61 371.318 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip200)" style="stroke:#008000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:6; stroke-opacity:1; fill:none" points="
  1446.2,477.116 1582.05,477.116 
  "/>
<path clip-path="url(#clip200)" d="M1604.69 451.196 L1614.13 451.196 L1637.12 494.564 L1637.12 451.196 L1643.93 451.196 L1643.93 503.036 L1634.48 503.036 L1611.49 459.668 L1611.49 503.036 L1604.69 503.036 L1604.69 451.196 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1690.87 481.994 L1690.87 485.119 L1661.49 485.119 Q1661.91 491.716 1665.45 495.189 Q1669.03 498.626 1675.38 498.626 Q1679.06 498.626 1682.5 497.723 Q1685.97 496.82 1689.38 495.015 L1689.38 501.057 Q1685.94 502.515 1682.33 503.279 Q1678.72 504.043 1675 504.043 Q1665.7 504.043 1660.24 498.626 Q1654.83 493.209 1654.83 483.973 Q1654.83 474.425 1659.97 468.835 Q1665.14 463.21 1673.89 463.21 Q1681.74 463.21 1686.29 468.279 Q1690.87 473.314 1690.87 481.994 M1684.48 480.119 Q1684.41 474.876 1681.53 471.751 Q1678.68 468.626 1673.96 468.626 Q1668.61 468.626 1665.38 471.647 Q1662.19 474.668 1661.7 480.154 L1684.48 480.119 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1700.7 487.689 L1700.7 464.147 L1707.08 464.147 L1707.08 487.446 Q1707.08 492.966 1709.24 495.744 Q1711.39 498.487 1715.7 498.487 Q1720.87 498.487 1723.86 495.189 Q1726.88 491.89 1726.88 486.196 L1726.88 464.147 L1733.27 464.147 L1733.27 503.036 L1726.88 503.036 L1726.88 497.064 Q1724.55 500.605 1721.46 502.341 Q1718.4 504.043 1714.34 504.043 Q1707.64 504.043 1704.17 499.876 Q1700.7 495.709 1700.7 487.689 M1716.77 463.21 L1716.77 463.21 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1768.96 470.119 Q1767.88 469.494 1766.6 469.216 Q1765.35 468.904 1763.82 468.904 Q1758.4 468.904 1755.49 472.446 Q1752.61 475.953 1752.61 482.55 L1752.61 503.036 L1746.18 503.036 L1746.18 464.147 L1752.61 464.147 L1752.61 470.189 Q1754.62 466.647 1757.85 464.946 Q1761.08 463.21 1765.7 463.21 Q1766.36 463.21 1767.15 463.314 Q1767.95 463.383 1768.92 463.557 L1768.96 470.119 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1793.33 483.487 Q1785.59 483.487 1782.61 485.258 Q1779.62 487.029 1779.62 491.3 Q1779.62 494.702 1781.84 496.716 Q1784.1 498.695 1787.95 498.695 Q1793.26 498.695 1796.46 494.945 Q1799.69 491.161 1799.69 484.911 L1799.69 483.487 L1793.33 483.487 M1806.08 480.848 L1806.08 503.036 L1799.69 503.036 L1799.69 497.133 Q1797.5 500.675 1794.24 502.376 Q1790.97 504.043 1786.25 504.043 Q1780.28 504.043 1776.74 500.709 Q1773.23 497.341 1773.23 491.716 Q1773.23 485.154 1777.61 481.821 Q1782.01 478.487 1790.73 478.487 L1799.69 478.487 L1799.69 477.862 Q1799.69 473.453 1796.77 471.057 Q1793.89 468.626 1788.65 468.626 Q1785.31 468.626 1782.15 469.425 Q1778.99 470.223 1776.08 471.821 L1776.08 465.918 Q1779.58 464.564 1782.88 463.904 Q1786.18 463.21 1789.31 463.21 Q1797.74 463.21 1801.91 467.585 Q1806.08 471.959 1806.08 480.848 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1819.24 449.008 L1825.63 449.008 L1825.63 503.036 L1819.24 503.036 L1819.24 449.008 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1839.27 451.196 L1869.06 451.196 L1869.06 457.098 L1846.29 457.098 L1846.29 472.376 L1866.84 472.376 L1866.84 478.279 L1846.29 478.279 L1846.29 503.036 L1839.27 503.036 L1839.27 451.196 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1880.17 451.196 L1890.63 451.196 L1903.85 486.473 L1917.15 451.196 L1927.6 451.196 L1927.6 503.036 L1920.76 503.036 L1920.76 457.515 L1907.4 493.07 L1900.35 493.07 L1886.98 457.515 L1886.98 503.036 L1880.17 503.036 L1880.17 451.196 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M1940.73 451.196 L1947.78 451.196 L1947.78 482.689 Q1947.78 491.022 1950.8 494.702 Q1953.82 498.348 1960.59 498.348 Q1967.33 498.348 1970.35 494.702 Q1973.37 491.022 1973.37 482.689 L1973.37 451.196 L1980.42 451.196 L1980.42 483.557 Q1980.42 493.695 1975.38 498.869 Q1970.38 504.043 1960.59 504.043 Q1950.76 504.043 1945.73 498.869 Q1940.73 493.695 1940.73 483.557 L1940.73 451.196 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M2031.25 449.078 Q2026.6 457.064 2024.34 464.876 Q2022.08 472.689 2022.08 480.709 Q2022.08 488.73 2024.34 496.612 Q2026.63 504.459 2031.25 512.411 L2025.69 512.411 Q2020.49 504.251 2017.88 496.369 Q2015.31 488.487 2015.31 480.709 Q2015.31 472.966 2017.88 465.119 Q2020.45 457.272 2025.69 449.078 L2031.25 449.078 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M2044.62 451.196 L2072.15 451.196 L2072.15 457.098 L2051.04 457.098 L2051.04 469.807 Q2052.57 469.286 2054.1 469.043 Q2055.62 468.765 2057.15 468.765 Q2065.83 468.765 2070.9 473.522 Q2075.97 478.279 2075.97 486.404 Q2075.97 494.772 2070.76 499.425 Q2065.55 504.043 2056.08 504.043 Q2052.81 504.043 2049.41 503.487 Q2046.04 502.932 2042.43 501.82 L2042.43 494.772 Q2045.56 496.473 2048.89 497.307 Q2052.22 498.14 2055.94 498.14 Q2061.94 498.14 2065.45 494.98 Q2068.96 491.82 2068.96 486.404 Q2068.96 480.987 2065.45 477.828 Q2061.94 474.668 2055.94 474.668 Q2053.12 474.668 2050.31 475.293 Q2047.53 475.918 2044.62 477.237 L2044.62 451.196 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M2104.79 455.814 Q2099.37 455.814 2096.63 461.161 Q2093.92 466.473 2093.92 477.168 Q2093.92 487.827 2096.63 493.175 Q2099.37 498.487 2104.79 498.487 Q2110.24 498.487 2112.95 493.175 Q2115.69 487.827 2115.69 477.168 Q2115.69 466.473 2112.95 461.161 Q2110.24 455.814 2104.79 455.814 M2104.79 450.258 Q2113.51 450.258 2118.09 457.168 Q2122.71 464.043 2122.71 477.168 Q2122.71 490.258 2118.09 497.168 Q2113.51 504.043 2104.79 504.043 Q2096.08 504.043 2091.46 497.168 Q2086.87 490.258 2086.87 477.168 Q2086.87 464.043 2091.46 457.168 Q2096.08 450.258 2104.79 450.258 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M2150.03 455.814 Q2144.62 455.814 2141.87 461.161 Q2139.17 466.473 2139.17 477.168 Q2139.17 487.827 2141.87 493.175 Q2144.62 498.487 2150.03 498.487 Q2155.48 498.487 2158.19 493.175 Q2160.94 487.827 2160.94 477.168 Q2160.94 466.473 2158.19 461.161 Q2155.48 455.814 2150.03 455.814 M2150.03 450.258 Q2158.75 450.258 2163.33 457.168 Q2167.95 464.043 2167.95 477.168 Q2167.95 490.258 2163.33 497.168 Q2158.75 504.043 2150.03 504.043 Q2141.32 504.043 2136.7 497.168 Q2132.12 490.258 2132.12 477.168 Q2132.12 464.043 2136.7 457.168 Q2141.32 450.258 2150.03 450.258 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M2195.28 455.814 Q2189.86 455.814 2187.12 461.161 Q2184.41 466.473 2184.41 477.168 Q2184.41 487.827 2187.12 493.175 Q2189.86 498.487 2195.28 498.487 Q2200.73 498.487 2203.44 493.175 Q2206.18 487.827 2206.18 477.168 Q2206.18 466.473 2203.44 461.161 Q2200.73 455.814 2195.28 455.814 M2195.28 450.258 Q2203.99 450.258 2208.57 457.168 Q2213.19 464.043 2213.19 477.168 Q2213.19 490.258 2208.57 497.168 Q2203.99 504.043 2195.28 504.043 Q2186.56 504.043 2181.94 497.168 Q2177.36 490.258 2177.36 477.168 Q2177.36 464.043 2181.94 457.168 Q2186.56 450.258 2195.28 450.258 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip200)" d="M2223.61 449.078 L2229.17 449.078 Q2234.37 457.272 2236.94 465.119 Q2239.55 472.966 2239.55 480.709 Q2239.55 488.487 2236.94 496.369 Q2234.37 504.251 2229.17 512.411 L2223.61 512.411 Q2228.23 504.459 2230.48 496.612 Q2232.78 488.73 2232.78 480.709 Q2232.78 472.689 2230.48 464.876 Q2228.23 457.064 2223.61 449.078 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /></svg>



    ┌ Info: Friction model 1 mse: 0.0942545853263148
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:29
    ┌ Info: Friction model 2 mse: 0.0942545853263148
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\modelica_conference_2021.ipynb:29



<?xml version="1.0" encoding="utf-8"?>
<svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" width="600" height="400" viewBox="0 0 2400 1600">
<defs>
  <clipPath id="clip290">
    <rect x="0" y="0" width="2400" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip290)" d="
M0 1600 L2400 1600 L2400 0 L0 0  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip291">
    <rect x="480" y="0" width="1681" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip290)" d="
M315.725 1401.58 L2352.76 1401.58 L2352.76 123.472 L315.725 123.472  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip292">
    <rect x="315" y="123" width="2038" height="1279"/>
  </clipPath>
</defs>
<polyline clip-path="url(#clip292)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  519.428,1401.58 519.428,123.472 
  "/>
<polyline clip-path="url(#clip292)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  926.834,1401.58 926.834,123.472 
  "/>
<polyline clip-path="url(#clip292)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1334.24,1401.58 1334.24,123.472 
  "/>
<polyline clip-path="url(#clip292)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1741.65,1401.58 1741.65,123.472 
  "/>
<polyline clip-path="url(#clip292)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  2149.05,1401.58 2149.05,123.472 
  "/>
<polyline clip-path="url(#clip290)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  315.725,1401.58 2352.76,1401.58 
  "/>
<polyline clip-path="url(#clip290)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  519.428,1401.58 519.428,1382.68 
  "/>
<polyline clip-path="url(#clip290)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  926.834,1401.58 926.834,1382.68 
  "/>
<polyline clip-path="url(#clip290)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1334.24,1401.58 1334.24,1382.68 
  "/>
<polyline clip-path="url(#clip290)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1741.65,1401.58 1741.65,1382.68 
  "/>
<polyline clip-path="url(#clip290)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  2149.05,1401.58 2149.05,1382.68 
  "/>
<path clip-path="url(#clip290)" d="M439.22 1460.34 L483.734 1460.34 L483.734 1466.24 L439.22 1466.24 L439.22 1460.34 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M500.088 1479.68 L511.546 1479.68 L511.546 1440.13 L499.081 1442.63 L499.081 1436.24 L511.477 1433.74 L518.491 1433.74 L518.491 1479.68 L529.949 1479.68 L529.949 1485.58 L500.088 1485.58 L500.088 1479.68 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M544.116 1476.76 L551.442 1476.76 L551.442 1485.58 L544.116 1485.58 L544.116 1476.76 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M581.72 1438.36 Q576.303 1438.36 573.56 1443.71 Q570.852 1449.02 570.852 1459.71 Q570.852 1470.37 573.56 1475.72 Q576.303 1481.03 581.72 1481.03 Q587.171 1481.03 589.879 1475.72 Q592.622 1470.37 592.622 1459.71 Q592.622 1449.02 589.879 1443.71 Q587.171 1438.36 581.72 1438.36 M581.72 1432.8 Q590.435 1432.8 595.018 1439.71 Q599.636 1446.59 599.636 1459.71 Q599.636 1472.8 595.018 1479.71 Q590.435 1486.59 581.72 1486.59 Q573.004 1486.59 568.386 1479.71 Q563.803 1472.8 563.803 1459.71 Q563.803 1446.59 568.386 1439.71 Q573.004 1432.8 581.72 1432.8 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M847.373 1460.34 L891.887 1460.34 L891.887 1466.24 L847.373 1466.24 L847.373 1460.34 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M922.025 1438.36 Q916.609 1438.36 913.866 1443.71 Q911.157 1449.02 911.157 1459.71 Q911.157 1470.37 913.866 1475.72 Q916.609 1481.03 922.025 1481.03 Q927.477 1481.03 930.185 1475.72 Q932.928 1470.37 932.928 1459.71 Q932.928 1449.02 930.185 1443.71 Q927.477 1438.36 922.025 1438.36 M922.025 1432.8 Q930.741 1432.8 935.324 1439.71 Q939.942 1446.59 939.942 1459.71 Q939.942 1472.8 935.324 1479.71 Q930.741 1486.59 922.025 1486.59 Q913.31 1486.59 908.692 1479.71 Q904.109 1472.8 904.109 1459.71 Q904.109 1446.59 908.692 1439.71 Q913.31 1432.8 922.025 1432.8 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M952.268 1476.76 L959.595 1476.76 L959.595 1485.58 L952.268 1485.58 L952.268 1476.76 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M974.942 1433.74 L1002.48 1433.74 L1002.48 1439.64 L981.365 1439.64 L981.365 1452.35 Q982.893 1451.83 984.421 1451.59 Q985.949 1451.31 987.476 1451.31 Q996.157 1451.31 1001.23 1456.07 Q1006.3 1460.82 1006.3 1468.95 Q1006.3 1477.32 1001.09 1481.97 Q995.879 1486.59 986.4 1486.59 Q983.136 1486.59 979.733 1486.03 Q976.365 1485.48 972.754 1484.36 L972.754 1477.32 Q975.879 1479.02 979.213 1479.85 Q982.546 1480.68 986.261 1480.68 Q992.268 1480.68 995.775 1477.52 Q999.282 1474.36 999.282 1468.95 Q999.282 1463.53 995.775 1460.37 Q992.268 1457.21 986.261 1457.21 Q983.449 1457.21 980.636 1457.84 Q977.858 1458.46 974.942 1459.78 L974.942 1433.74 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1300.32 1438.36 Q1294.9 1438.36 1292.16 1443.71 Q1289.45 1449.02 1289.45 1459.71 Q1289.45 1470.37 1292.16 1475.72 Q1294.9 1481.03 1300.32 1481.03 Q1305.77 1481.03 1308.48 1475.72 Q1311.22 1470.37 1311.22 1459.71 Q1311.22 1449.02 1308.48 1443.71 Q1305.77 1438.36 1300.32 1438.36 M1300.32 1432.8 Q1309.03 1432.8 1313.62 1439.71 Q1318.23 1446.59 1318.23 1459.71 Q1318.23 1472.8 1313.62 1479.71 Q1309.03 1486.59 1300.32 1486.59 Q1291.6 1486.59 1286.98 1479.71 Q1282.4 1472.8 1282.4 1459.71 Q1282.4 1446.59 1286.98 1439.71 Q1291.6 1432.8 1300.32 1432.8 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1330.56 1476.76 L1337.89 1476.76 L1337.89 1485.58 L1330.56 1485.58 L1330.56 1476.76 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1368.16 1438.36 Q1362.75 1438.36 1360 1443.71 Q1357.3 1449.02 1357.3 1459.71 Q1357.3 1470.37 1360 1475.72 Q1362.75 1481.03 1368.16 1481.03 Q1373.62 1481.03 1376.32 1475.72 Q1379.07 1470.37 1379.07 1459.71 Q1379.07 1449.02 1376.32 1443.71 Q1373.62 1438.36 1368.16 1438.36 M1368.16 1432.8 Q1376.88 1432.8 1381.46 1439.71 Q1386.08 1446.59 1386.08 1459.71 Q1386.08 1472.8 1381.46 1479.71 Q1376.88 1486.59 1368.16 1486.59 Q1359.45 1486.59 1354.83 1479.71 Q1350.25 1472.8 1350.25 1459.71 Q1350.25 1446.59 1354.83 1439.71 Q1359.45 1432.8 1368.16 1432.8 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1708.47 1438.36 Q1703.05 1438.36 1700.31 1443.71 Q1697.6 1449.02 1697.6 1459.71 Q1697.6 1470.37 1700.31 1475.72 Q1703.05 1481.03 1708.47 1481.03 Q1713.92 1481.03 1716.63 1475.72 Q1719.37 1470.37 1719.37 1459.71 Q1719.37 1449.02 1716.63 1443.71 Q1713.92 1438.36 1708.47 1438.36 M1708.47 1432.8 Q1717.18 1432.8 1721.77 1439.71 Q1726.39 1446.59 1726.39 1459.71 Q1726.39 1472.8 1721.77 1479.71 Q1717.18 1486.59 1708.47 1486.59 Q1699.75 1486.59 1695.14 1479.71 Q1690.55 1472.8 1690.55 1459.71 Q1690.55 1446.59 1695.14 1439.71 Q1699.75 1432.8 1708.47 1432.8 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1738.71 1476.76 L1746.04 1476.76 L1746.04 1485.58 L1738.71 1485.58 L1738.71 1476.76 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1761.39 1433.74 L1788.92 1433.74 L1788.92 1439.64 L1767.81 1439.64 L1767.81 1452.35 Q1769.34 1451.83 1770.87 1451.59 Q1772.39 1451.31 1773.92 1451.31 Q1782.6 1451.31 1787.67 1456.07 Q1792.74 1460.82 1792.74 1468.95 Q1792.74 1477.32 1787.53 1481.97 Q1782.32 1486.59 1772.84 1486.59 Q1769.58 1486.59 1766.18 1486.03 Q1762.81 1485.48 1759.2 1484.36 L1759.2 1477.32 Q1762.32 1479.02 1765.66 1479.85 Q1768.99 1480.68 1772.71 1480.68 Q1778.71 1480.68 1782.22 1477.52 Q1785.73 1474.36 1785.73 1468.95 Q1785.73 1463.53 1782.22 1460.37 Q1778.71 1457.21 1772.71 1457.21 Q1769.89 1457.21 1767.08 1457.84 Q1764.3 1458.46 1761.39 1459.78 L1761.39 1433.74 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M2099.78 1479.68 L2111.24 1479.68 L2111.24 1440.13 L2098.78 1442.63 L2098.78 1436.24 L2111.17 1433.74 L2118.18 1433.74 L2118.18 1479.68 L2129.64 1479.68 L2129.64 1485.58 L2099.78 1485.58 L2099.78 1479.68 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M2143.81 1476.76 L2151.14 1476.76 L2151.14 1485.58 L2143.81 1485.58 L2143.81 1476.76 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M2181.41 1438.36 Q2176 1438.36 2173.25 1443.71 Q2170.55 1449.02 2170.55 1459.71 Q2170.55 1470.37 2173.25 1475.72 Q2176 1481.03 2181.41 1481.03 Q2186.87 1481.03 2189.57 1475.72 Q2192.32 1470.37 2192.32 1459.71 Q2192.32 1449.02 2189.57 1443.71 Q2186.87 1438.36 2181.41 1438.36 M2181.41 1432.8 Q2190.13 1432.8 2194.71 1439.71 Q2199.33 1446.59 2199.33 1459.71 Q2199.33 1472.8 2194.71 1479.71 Q2190.13 1486.59 2181.41 1486.59 Q2172.7 1486.59 2168.08 1479.71 Q2163.5 1472.8 2163.5 1459.71 Q2163.5 1446.59 2168.08 1439.71 Q2172.7 1432.8 2181.41 1432.8 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1213.13 1538.66 L1219.9 1538.66 L1232.05 1571.3 L1244.21 1538.66 L1250.98 1538.66 L1236.39 1577.55 L1227.71 1577.55 L1213.13 1538.66 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1281.81 1523.52 L1296.53 1523.52 L1296.53 1528.49 L1288.2 1528.49 L1288.2 1581.96 L1296.53 1581.96 L1296.53 1586.92 L1281.81 1586.92 L1281.81 1523.52 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1340.42 1546.12 Q1342.82 1541.82 1346.15 1539.77 Q1349.48 1537.72 1354 1537.72 Q1360.07 1537.72 1363.37 1541.99 Q1366.67 1546.23 1366.67 1554.08 L1366.67 1577.55 L1360.25 1577.55 L1360.25 1554.28 Q1360.25 1548.69 1358.27 1545.99 Q1356.29 1543.28 1352.23 1543.28 Q1347.26 1543.28 1344.38 1546.58 Q1341.5 1549.87 1341.5 1555.57 L1341.5 1577.55 L1335.07 1577.55 L1335.07 1554.28 Q1335.07 1548.66 1333.09 1545.99 Q1331.12 1543.28 1326.98 1543.28 Q1322.09 1543.28 1319.21 1546.61 Q1316.32 1549.91 1316.32 1555.57 L1316.32 1577.55 L1309.9 1577.55 L1309.9 1538.66 L1316.32 1538.66 L1316.32 1544.7 Q1318.51 1541.12 1321.57 1539.42 Q1324.62 1537.72 1328.82 1537.72 Q1333.06 1537.72 1336.01 1539.87 Q1339 1542.03 1340.42 1546.12 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1390.77 1525.71 L1396.67 1525.71 L1378.62 1584.15 L1372.71 1584.15 L1390.77 1525.71 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1428.16 1539.8 L1428.16 1545.85 Q1425.46 1544.46 1422.54 1543.76 Q1419.62 1543.07 1416.5 1543.07 Q1411.74 1543.07 1409.34 1544.53 Q1406.98 1545.99 1406.98 1548.9 Q1406.98 1551.12 1408.68 1552.41 Q1410.39 1553.66 1415.52 1554.8 L1417.71 1555.29 Q1424.52 1556.75 1427.37 1559.42 Q1430.25 1562.06 1430.25 1566.82 Q1430.25 1572.24 1425.94 1575.4 Q1421.67 1578.55 1414.17 1578.55 Q1411.05 1578.55 1407.64 1577.93 Q1404.27 1577.34 1400.52 1576.12 L1400.52 1569.53 Q1404.07 1571.37 1407.5 1572.3 Q1410.94 1573.21 1414.31 1573.21 Q1418.82 1573.21 1421.25 1571.68 Q1423.68 1570.12 1423.68 1567.3 Q1423.68 1564.7 1421.91 1563.31 Q1420.18 1561.92 1414.24 1560.64 L1412.02 1560.12 Q1406.08 1558.87 1403.44 1556.3 Q1400.8 1553.69 1400.8 1549.18 Q1400.8 1543.69 1404.69 1540.71 Q1408.58 1537.72 1415.73 1537.72 Q1419.27 1537.72 1422.4 1538.24 Q1425.52 1538.76 1428.16 1539.8 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1455.35 1523.52 L1455.35 1586.92 L1440.63 1586.92 L1440.63 1581.96 L1448.93 1581.96 L1448.93 1528.49 L1440.63 1528.49 L1440.63 1523.52 L1455.35 1523.52 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip292)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  315.725,1254.21 2352.76,1254.21 
  "/>
<polyline clip-path="url(#clip292)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  315.725,1092.05 2352.76,1092.05 
  "/>
<polyline clip-path="url(#clip292)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  315.725,929.896 2352.76,929.896 
  "/>
<polyline clip-path="url(#clip292)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  315.725,767.738 2352.76,767.738 
  "/>
<polyline clip-path="url(#clip292)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  315.725,605.58 2352.76,605.58 
  "/>
<polyline clip-path="url(#clip292)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  315.725,443.422 2352.76,443.422 
  "/>
<polyline clip-path="url(#clip292)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  315.725,281.263 2352.76,281.263 
  "/>
<polyline clip-path="url(#clip290)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  315.725,1401.58 315.725,123.472 
  "/>
<polyline clip-path="url(#clip290)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  315.725,1254.21 334.623,1254.21 
  "/>
<polyline clip-path="url(#clip290)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  315.725,1092.05 334.623,1092.05 
  "/>
<polyline clip-path="url(#clip290)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  315.725,929.896 334.623,929.896 
  "/>
<polyline clip-path="url(#clip290)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  315.725,767.738 334.623,767.738 
  "/>
<polyline clip-path="url(#clip290)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  315.725,605.58 334.623,605.58 
  "/>
<polyline clip-path="url(#clip290)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  315.725,443.422 334.623,443.422 
  "/>
<polyline clip-path="url(#clip290)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  315.725,281.263 334.623,281.263 
  "/>
<path clip-path="url(#clip290)" d="M119.066 1254.89 L163.58 1254.89 L163.58 1260.79 L119.066 1260.79 L119.066 1254.89 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M193.719 1232.91 Q188.302 1232.91 185.559 1238.26 Q182.851 1243.57 182.851 1254.26 Q182.851 1264.92 185.559 1270.27 Q188.302 1275.58 193.719 1275.58 Q199.17 1275.58 201.878 1270.27 Q204.621 1264.92 204.621 1254.26 Q204.621 1243.57 201.878 1238.26 Q199.17 1232.91 193.719 1232.91 M193.719 1227.35 Q202.434 1227.35 207.017 1234.26 Q211.635 1241.14 211.635 1254.26 Q211.635 1267.35 207.017 1274.26 Q202.434 1281.14 193.719 1281.14 Q185.003 1281.14 180.385 1274.26 Q175.802 1267.35 175.802 1254.26 Q175.802 1241.14 180.385 1234.26 Q185.003 1227.35 193.719 1227.35 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M223.962 1271.31 L231.288 1271.31 L231.288 1280.13 L223.962 1280.13 L223.962 1271.31 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M262.434 1251.42 Q257.711 1251.42 254.934 1254.65 Q252.191 1257.87 252.191 1263.5 Q252.191 1269.09 254.934 1272.35 Q257.711 1275.58 262.434 1275.58 Q267.156 1275.58 269.899 1272.35 Q272.677 1269.09 272.677 1263.5 Q272.677 1257.87 269.899 1254.65 Q267.156 1251.42 262.434 1251.42 M276.357 1229.44 L276.357 1235.83 Q273.718 1234.58 271.01 1233.92 Q268.336 1233.26 265.697 1233.26 Q258.753 1233.26 255.072 1237.94 Q251.427 1242.63 250.906 1252.11 Q252.954 1249.09 256.045 1247.49 Q259.135 1245.86 262.85 1245.86 Q270.663 1245.86 275.177 1250.62 Q279.725 1255.34 279.725 1263.5 Q279.725 1271.49 275.003 1276.31 Q270.281 1281.14 262.434 1281.14 Q253.441 1281.14 248.684 1274.26 Q243.927 1267.35 243.927 1254.26 Q243.927 1241.97 249.76 1234.68 Q255.593 1227.35 265.42 1227.35 Q268.059 1227.35 270.732 1227.87 Q273.44 1228.4 276.357 1229.44 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M118.58 1092.73 L163.094 1092.73 L163.094 1098.63 L118.58 1098.63 L118.58 1092.73 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M193.233 1070.75 Q187.816 1070.75 185.073 1076.1 Q182.365 1081.41 182.365 1092.11 Q182.365 1102.77 185.073 1108.11 Q187.816 1113.43 193.233 1113.43 Q198.684 1113.43 201.392 1108.11 Q204.135 1102.77 204.135 1092.11 Q204.135 1081.41 201.392 1076.1 Q198.684 1070.75 193.233 1070.75 M193.233 1065.2 Q201.948 1065.2 206.531 1072.11 Q211.149 1078.98 211.149 1092.11 Q211.149 1105.2 206.531 1112.11 Q201.948 1118.98 193.233 1118.98 Q184.517 1118.98 179.899 1112.11 Q175.316 1105.2 175.316 1092.11 Q175.316 1078.98 179.899 1072.11 Q184.517 1065.2 193.233 1065.2 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M223.475 1109.15 L230.802 1109.15 L230.802 1117.97 L223.475 1117.97 L223.475 1109.15 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M265.35 1072.24 L247.642 1099.92 L265.35 1099.92 L265.35 1072.24 M263.51 1066.13 L272.329 1066.13 L272.329 1099.92 L279.725 1099.92 L279.725 1105.75 L272.329 1105.75 L272.329 1117.97 L265.35 1117.97 L265.35 1105.75 L241.948 1105.75 L241.948 1098.98 L263.51 1066.13 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M121.705 930.573 L166.219 930.573 L166.219 936.475 L121.705 936.475 L121.705 930.573 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M196.358 908.594 Q190.941 908.594 188.198 913.941 Q185.49 919.253 185.49 929.948 Q185.49 940.607 188.198 945.955 Q190.941 951.267 196.358 951.267 Q201.809 951.267 204.517 945.955 Q207.26 940.607 207.26 929.948 Q207.26 919.253 204.517 913.941 Q201.809 908.594 196.358 908.594 M196.358 903.038 Q205.073 903.038 209.656 909.948 Q214.274 916.823 214.274 929.948 Q214.274 943.038 209.656 949.948 Q205.073 956.823 196.358 956.823 Q187.642 956.823 183.024 949.948 Q178.441 943.038 178.441 929.948 Q178.441 916.823 183.024 909.948 Q187.642 903.038 196.358 903.038 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M226.6 946.996 L233.927 946.996 L233.927 955.816 L226.6 955.816 L226.6 946.996 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M255.246 949.913 L279.725 949.913 L279.725 955.816 L246.809 955.816 L246.809 949.913 Q250.802 945.781 257.677 938.836 Q264.586 931.857 266.357 929.843 Q269.725 926.059 271.045 923.455 Q272.399 920.816 272.399 918.281 Q272.399 914.149 269.482 911.545 Q266.6 908.941 261.947 908.941 Q258.649 908.941 254.968 910.087 Q251.323 911.232 247.156 913.559 L247.156 906.476 Q251.392 904.774 255.072 903.906 Q258.753 903.038 261.809 903.038 Q269.864 903.038 274.656 907.066 Q279.447 911.094 279.447 917.83 Q279.447 921.024 278.232 923.906 Q277.052 926.753 273.892 930.642 Q273.024 931.649 268.371 936.475 Q263.718 941.267 255.246 949.913 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M193.962 746.436 Q188.545 746.436 185.802 751.783 Q183.094 757.095 183.094 767.79 Q183.094 778.449 185.802 783.796 Q188.545 789.109 193.962 789.109 Q199.413 789.109 202.121 783.796 Q204.864 778.449 204.864 767.79 Q204.864 757.095 202.121 751.783 Q199.413 746.436 193.962 746.436 M193.962 740.88 Q202.677 740.88 207.26 747.79 Q211.878 754.665 211.878 767.79 Q211.878 780.88 207.26 787.79 Q202.677 794.664 193.962 794.664 Q185.246 794.664 180.628 787.79 Q176.045 780.88 176.045 767.79 Q176.045 754.665 180.628 747.79 Q185.246 740.88 193.962 740.88 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M224.205 784.838 L231.531 784.838 L231.531 793.658 L224.205 793.658 L224.205 784.838 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M261.809 746.436 Q256.392 746.436 253.649 751.783 Q250.941 757.095 250.941 767.79 Q250.941 778.449 253.649 783.796 Q256.392 789.109 261.809 789.109 Q267.26 789.109 269.968 783.796 Q272.711 778.449 272.711 767.79 Q272.711 757.095 269.968 751.783 Q267.26 746.436 261.809 746.436 M261.809 740.88 Q270.524 740.88 275.107 747.79 Q279.725 754.665 279.725 767.79 Q279.725 780.88 275.107 787.79 Q270.524 794.664 261.809 794.664 Q253.093 794.664 248.475 787.79 Q243.892 780.88 243.892 767.79 Q243.892 754.665 248.475 747.79 Q253.093 740.88 261.809 740.88 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M196.358 584.278 Q190.941 584.278 188.198 589.625 Q185.49 594.937 185.49 605.632 Q185.49 616.291 188.198 621.638 Q190.941 626.951 196.358 626.951 Q201.809 626.951 204.517 621.638 Q207.26 616.291 207.26 605.632 Q207.26 594.937 204.517 589.625 Q201.809 584.278 196.358 584.278 M196.358 578.722 Q205.073 578.722 209.656 585.632 Q214.274 592.507 214.274 605.632 Q214.274 618.722 209.656 625.632 Q205.073 632.506 196.358 632.506 Q187.642 632.506 183.024 625.632 Q178.441 618.722 178.441 605.632 Q178.441 592.507 183.024 585.632 Q187.642 578.722 196.358 578.722 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M226.6 622.68 L233.927 622.68 L233.927 631.5 L226.6 631.5 L226.6 622.68 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M255.246 625.597 L279.725 625.597 L279.725 631.5 L246.809 631.5 L246.809 625.597 Q250.802 621.465 257.677 614.52 Q264.586 607.541 266.357 605.527 Q269.725 601.743 271.045 599.139 Q272.399 596.5 272.399 593.965 Q272.399 589.833 269.482 587.229 Q266.6 584.625 261.947 584.625 Q258.649 584.625 254.968 585.771 Q251.323 586.916 247.156 589.243 L247.156 582.16 Q251.392 580.458 255.072 579.59 Q258.753 578.722 261.809 578.722 Q269.864 578.722 274.656 582.75 Q279.447 586.778 279.447 593.514 Q279.447 596.708 278.232 599.59 Q277.052 602.437 273.892 606.326 Q273.024 607.333 268.371 612.159 Q263.718 616.951 255.246 625.597 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M193.233 422.12 Q187.816 422.12 185.073 427.467 Q182.365 432.779 182.365 443.474 Q182.365 454.133 185.073 459.48 Q187.816 464.793 193.233 464.793 Q198.684 464.793 201.392 459.48 Q204.135 454.133 204.135 443.474 Q204.135 432.779 201.392 427.467 Q198.684 422.12 193.233 422.12 M193.233 416.564 Q201.948 416.564 206.531 423.474 Q211.149 430.349 211.149 443.474 Q211.149 456.564 206.531 463.473 Q201.948 470.348 193.233 470.348 Q184.517 470.348 179.899 463.473 Q175.316 456.564 175.316 443.474 Q175.316 430.349 179.899 423.474 Q184.517 416.564 193.233 416.564 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M223.475 460.522 L230.802 460.522 L230.802 469.342 L223.475 469.342 L223.475 460.522 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M265.35 423.613 L247.642 451.286 L265.35 451.286 L265.35 423.613 M263.51 417.502 L272.329 417.502 L272.329 451.286 L279.725 451.286 L279.725 457.119 L272.329 457.119 L272.329 469.342 L265.35 469.342 L265.35 457.119 L241.948 457.119 L241.948 450.349 L263.51 417.502 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M193.719 259.962 Q188.302 259.962 185.559 265.309 Q182.851 270.621 182.851 281.316 Q182.851 291.975 185.559 297.322 Q188.302 302.635 193.719 302.635 Q199.17 302.635 201.878 297.322 Q204.621 291.975 204.621 281.316 Q204.621 270.621 201.878 265.309 Q199.17 259.962 193.719 259.962 M193.719 254.406 Q202.434 254.406 207.017 261.316 Q211.635 268.191 211.635 281.316 Q211.635 294.406 207.017 301.315 Q202.434 308.19 193.719 308.19 Q185.003 308.19 180.385 301.315 Q175.802 294.406 175.802 281.316 Q175.802 268.191 180.385 261.316 Q185.003 254.406 193.719 254.406 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M223.962 298.364 L231.288 298.364 L231.288 307.183 L223.962 307.183 L223.962 298.364 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M262.434 278.468 Q257.711 278.468 254.934 281.698 Q252.191 284.927 252.191 290.552 Q252.191 296.142 254.934 299.406 Q257.711 302.635 262.434 302.635 Q267.156 302.635 269.899 299.406 Q272.677 296.142 272.677 290.552 Q272.677 284.927 269.899 281.698 Q267.156 278.468 262.434 278.468 M276.357 256.489 L276.357 262.878 Q273.718 261.628 271.01 260.968 Q268.336 260.309 265.697 260.309 Q258.753 260.309 255.072 264.996 Q251.427 269.684 250.906 279.163 Q252.954 276.142 256.045 274.545 Q259.135 272.913 262.85 272.913 Q270.663 272.913 275.177 277.67 Q279.725 282.392 279.725 290.552 Q279.725 298.538 275.003 303.364 Q270.281 308.19 262.434 308.19 Q253.441 308.19 248.684 301.315 Q243.927 294.406 243.927 281.316 Q243.927 269.024 249.76 261.732 Q255.593 254.406 265.42 254.406 Q268.059 254.406 270.732 254.927 Q273.44 255.448 276.357 256.489 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M13.0007 1025.04 L18.3131 1025.04 L18.3131 1031.15 Q18.3131 1034.59 19.702 1035.94 Q21.0909 1037.26 24.702 1037.26 L28.1395 1037.26 L28.1395 1026.74 L33.1047 1026.74 L33.1047 1037.26 L67.0282 1037.26 L67.0282 1043.69 L33.1047 1043.69 L33.1047 1049.8 L28.1395 1049.8 L28.1395 1043.69 L25.4312 1043.69 Q18.9381 1043.69 15.9868 1040.67 Q13.0007 1037.65 13.0007 1031.08 L13.0007 1025.04 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M34.1117 997.16 Q33.4867 998.237 33.2089 999.521 Q32.8964 1000.77 32.8964 1002.3 Q32.8964 1007.72 36.438 1010.63 Q39.945 1013.51 46.5422 1013.51 L67.0282 1013.51 L67.0282 1019.94 L28.1395 1019.94 L28.1395 1013.51 L34.1811 1013.51 Q30.6395 1011.5 28.9381 1008.27 Q27.202 1005.04 27.202 1000.42 Q27.202 999.764 27.3061 998.966 Q27.3756 998.167 27.5492 997.195 L34.1117 997.16 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M28.1395 990.459 L28.1395 984.07 L67.0282 984.07 L67.0282 990.459 L28.1395 990.459 M13.0007 990.459 L13.0007 984.07 L21.0909 984.07 L21.0909 990.459 L13.0007 990.459 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M29.6325 942.716 L35.6047 942.716 Q34.1117 945.424 33.3825 948.167 Q32.6186 950.876 32.6186 953.653 Q32.6186 959.869 36.5769 963.306 Q40.5005 966.744 47.6185 966.744 Q54.7366 966.744 58.6949 963.306 Q62.6185 959.869 62.6185 953.653 Q62.6185 950.876 61.8893 948.167 Q61.1254 945.424 59.6324 942.716 L65.5351 942.716 Q66.7851 945.39 67.4101 948.272 Q68.0351 951.119 68.0351 954.348 Q68.0351 963.133 62.5143 968.306 Q56.9935 973.48 47.6185 973.48 Q38.1047 973.48 32.6533 968.271 Q27.202 963.028 27.202 953.931 Q27.202 950.98 27.827 948.167 Q28.4173 945.355 29.6325 942.716 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M17.0979 925.286 L28.1395 925.286 L28.1395 912.126 L33.1047 912.126 L33.1047 925.286 L54.2157 925.286 Q58.9726 925.286 60.3268 924.001 Q61.681 922.681 61.681 918.688 L61.681 912.126 L67.0282 912.126 L67.0282 918.688 Q67.0282 926.084 64.2851 928.897 Q61.5074 931.709 54.2157 931.709 L33.1047 931.709 L33.1047 936.397 L28.1395 936.397 L28.1395 931.709 L17.0979 931.709 L17.0979 925.286 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M28.1395 903.723 L28.1395 897.334 L67.0282 897.334 L67.0282 903.723 L28.1395 903.723 M13.0007 903.723 L13.0007 897.334 L21.0909 897.334 L21.0909 903.723 L13.0007 903.723 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M32.6186 868.897 Q32.6186 874.036 36.6464 877.022 Q40.6394 880.008 47.6185 880.008 Q54.5977 880.008 58.6254 877.057 Q62.6185 874.071 62.6185 868.897 Q62.6185 863.793 58.5907 860.807 Q54.5629 857.821 47.6185 857.821 Q40.7089 857.821 36.6811 860.807 Q32.6186 863.793 32.6186 868.897 M27.202 868.897 Q27.202 860.564 32.6186 855.807 Q38.0353 851.05 47.6185 851.05 Q57.1671 851.05 62.6185 855.807 Q68.0351 860.564 68.0351 868.897 Q68.0351 877.265 62.6185 882.022 Q57.1671 886.744 47.6185 886.744 Q38.0353 886.744 32.6186 882.022 Q27.202 877.265 27.202 868.897 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M43.5561 808.133 L67.0282 808.133 L67.0282 814.522 L43.7644 814.522 Q38.2436 814.522 35.5005 816.675 Q32.7575 818.828 32.7575 823.133 Q32.7575 828.307 36.0561 831.293 Q39.3547 834.279 45.0491 834.279 L67.0282 834.279 L67.0282 840.703 L28.1395 840.703 L28.1395 834.279 L34.1811 834.279 Q30.6742 831.987 28.9381 828.897 Q27.202 825.772 27.202 821.71 Q27.202 815.008 31.3686 811.571 Q35.5005 808.133 43.5561 808.133 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M13.0007 753.099 L18.3131 753.099 L18.3131 759.21 Q18.3131 762.648 19.702 764.002 Q21.0909 765.321 24.702 765.321 L28.1395 765.321 L28.1395 754.8 L33.1047 754.8 L33.1047 765.321 L67.0282 765.321 L67.0282 771.745 L33.1047 771.745 L33.1047 777.856 L28.1395 777.856 L28.1395 771.745 L25.4312 771.745 Q18.9381 771.745 15.9868 768.724 Q13.0007 765.703 13.0007 759.141 L13.0007 753.099 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M32.6186 732.682 Q32.6186 737.821 36.6464 740.807 Q40.6394 743.793 47.6185 743.793 Q54.5977 743.793 58.6254 740.842 Q62.6185 737.856 62.6185 732.682 Q62.6185 727.578 58.5907 724.592 Q54.5629 721.606 47.6185 721.606 Q40.7089 721.606 36.6811 724.592 Q32.6186 727.578 32.6186 732.682 M27.202 732.682 Q27.202 724.349 32.6186 719.592 Q38.0353 714.835 47.6185 714.835 Q57.1671 714.835 62.6185 719.592 Q68.0351 724.349 68.0351 732.682 Q68.0351 741.05 62.6185 745.807 Q57.1671 750.53 47.6185 750.53 Q38.0353 750.53 32.6186 745.807 Q27.202 741.05 27.202 732.682 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M34.1117 681.71 Q33.4867 682.787 33.2089 684.072 Q32.8964 685.322 32.8964 686.849 Q32.8964 692.266 36.438 695.183 Q39.945 698.065 46.5422 698.065 L67.0282 698.065 L67.0282 704.488 L28.1395 704.488 L28.1395 698.065 L34.1811 698.065 Q30.6395 696.051 28.9381 692.821 Q27.202 689.592 27.202 684.974 Q27.202 684.315 27.3061 683.516 Q27.3756 682.717 27.5492 681.745 L34.1117 681.71 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M29.6325 648.586 L35.6047 648.586 Q34.1117 651.294 33.3825 654.037 Q32.6186 656.745 32.6186 659.523 Q32.6186 665.738 36.5769 669.176 Q40.5005 672.613 47.6185 672.613 Q54.7366 672.613 58.6949 669.176 Q62.6185 665.738 62.6185 659.523 Q62.6185 656.745 61.8893 654.037 Q61.1254 651.294 59.6324 648.586 L65.5351 648.586 Q66.7851 651.259 67.4101 654.141 Q68.0351 656.988 68.0351 660.218 Q68.0351 669.002 62.5143 674.176 Q56.9935 679.349 47.6185 679.349 Q38.1047 679.349 32.6533 674.141 Q27.202 668.898 27.202 659.801 Q27.202 656.849 27.827 654.037 Q28.4173 651.224 29.6325 648.586 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M45.9866 604.211 L49.1116 604.211 L49.1116 633.586 Q55.7088 633.169 59.181 629.627 Q62.6185 626.051 62.6185 619.697 Q62.6185 616.016 61.7157 612.579 Q60.8129 609.107 59.0074 605.704 L65.049 605.704 Q66.5073 609.141 67.2712 612.752 Q68.0351 616.364 68.0351 620.079 Q68.0351 629.384 62.6185 634.836 Q57.2018 640.252 47.9658 640.252 Q38.4172 640.252 32.827 635.113 Q27.202 629.94 27.202 621.19 Q27.202 613.343 32.2714 608.794 Q37.3061 604.211 45.9866 604.211 M44.1116 610.6 Q38.8686 610.669 35.7436 613.551 Q32.6186 616.398 32.6186 621.12 Q32.6186 626.468 35.6394 629.697 Q38.6603 632.891 44.1463 633.377 L44.1116 610.6 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M13.0007 571.711 L13.0007 556.989 L17.9659 556.989 L17.9659 565.322 L71.4379 565.322 L71.4379 556.989 L76.4031 556.989 L76.4031 571.711 L13.0007 571.711 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M15.1882 543.1 L15.1882 533.656 L58.556 510.67 L15.1882 510.67 L15.1882 503.864 L67.0282 503.864 L67.0282 513.309 L23.6603 536.295 L67.0282 536.295 L67.0282 543.1 L15.1882 543.1 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M13.0007 475.253 L76.4031 475.253 L76.4031 489.975 L71.4379 489.975 L71.4379 481.677 L17.9659 481.677 L17.9659 489.975 L13.0007 489.975 L13.0007 475.253 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1008.85 12.096 L1043.61 12.096 L1043.61 18.9825 L1017.03 18.9825 L1017.03 36.8065 L1041.02 36.8065 L1041.02 43.6931 L1017.03 43.6931 L1017.03 72.576 L1008.85 72.576 L1008.85 12.096 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1076.5 34.1734 Q1075.25 33.4443 1073.75 33.1202 Q1072.29 32.7556 1070.51 32.7556 Q1064.19 32.7556 1060.78 36.8875 Q1057.42 40.9789 1057.42 48.6757 L1057.42 72.576 L1049.93 72.576 L1049.93 27.2059 L1057.42 27.2059 L1057.42 34.2544 Q1059.77 30.1225 1063.54 28.1376 Q1067.31 26.1121 1072.69 26.1121 Q1073.46 26.1121 1074.4 26.2337 Q1075.33 26.3147 1076.46 26.5172 L1076.5 34.1734 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1084.32 27.2059 L1091.77 27.2059 L1091.77 72.576 L1084.32 72.576 L1084.32 27.2059 M1084.32 9.54393 L1091.77 9.54393 L1091.77 18.9825 L1084.32 18.9825 L1084.32 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1140.02 28.9478 L1140.02 35.9153 Q1136.86 34.1734 1133.66 33.3227 Q1130.5 32.4315 1127.26 32.4315 Q1120.01 32.4315 1116 37.0496 Q1111.99 41.6271 1111.99 49.9314 Q1111.99 58.2358 1116 62.8538 Q1120.01 67.4314 1127.26 67.4314 Q1130.5 67.4314 1133.66 66.5807 Q1136.86 65.6895 1140.02 63.9476 L1140.02 70.8341 Q1136.9 72.2924 1133.54 73.0216 Q1130.22 73.7508 1126.45 73.7508 Q1116.2 73.7508 1110.16 67.3098 Q1104.13 60.8689 1104.13 49.9314 Q1104.13 38.832 1110.21 32.472 Q1116.32 26.1121 1126.94 26.1121 Q1130.38 26.1121 1133.66 26.8413 Q1136.94 27.5299 1140.02 28.9478 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1160.36 14.324 L1160.36 27.2059 L1175.71 27.2059 L1175.71 32.9987 L1160.36 32.9987 L1160.36 57.6282 Q1160.36 63.1779 1161.85 64.7578 Q1163.39 66.3376 1168.05 66.3376 L1175.71 66.3376 L1175.71 72.576 L1168.05 72.576 Q1159.42 72.576 1156.14 69.3758 Q1152.86 66.1351 1152.86 57.6282 L1152.86 32.9987 L1147.39 32.9987 L1147.39 27.2059 L1152.86 27.2059 L1152.86 14.324 L1160.36 14.324 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1185.51 27.2059 L1192.97 27.2059 L1192.97 72.576 L1185.51 72.576 L1185.51 27.2059 M1185.51 9.54393 L1192.97 9.54393 L1192.97 18.9825 L1185.51 18.9825 L1185.51 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1226.14 32.4315 Q1220.15 32.4315 1216.66 37.1306 Q1213.18 41.7891 1213.18 49.9314 Q1213.18 58.0738 1216.62 62.7728 Q1220.11 67.4314 1226.14 67.4314 Q1232.1 67.4314 1235.58 62.7323 Q1239.06 58.0333 1239.06 49.9314 Q1239.06 41.8701 1235.58 37.1711 Q1232.1 32.4315 1226.14 32.4315 M1226.14 26.1121 Q1235.86 26.1121 1241.41 32.4315 Q1246.96 38.7509 1246.96 49.9314 Q1246.96 61.0714 1241.41 67.4314 Q1235.86 73.7508 1226.14 73.7508 Q1216.38 73.7508 1210.83 67.4314 Q1205.32 61.0714 1205.32 49.9314 Q1205.32 38.7509 1210.83 32.4315 Q1216.38 26.1121 1226.14 26.1121 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1297.03 45.1919 L1297.03 72.576 L1289.58 72.576 L1289.58 45.4349 Q1289.58 38.994 1287.07 35.7938 Q1284.56 32.5936 1279.53 32.5936 Q1273.5 32.5936 1270.01 36.4419 Q1266.53 40.2903 1266.53 46.9338 L1266.53 72.576 L1259.04 72.576 L1259.04 27.2059 L1266.53 27.2059 L1266.53 34.2544 Q1269.2 30.163 1272.81 28.1376 Q1276.45 26.1121 1281.19 26.1121 Q1289.01 26.1121 1293.02 30.9732 Q1297.03 35.7938 1297.03 45.1919 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1373.6 35.9153 Q1376.39 30.8922 1380.28 28.5022 Q1384.17 26.1121 1389.43 26.1121 Q1396.52 26.1121 1400.37 31.0947 Q1404.22 36.0368 1404.22 45.1919 L1404.22 72.576 L1396.73 72.576 L1396.73 45.4349 Q1396.73 38.913 1394.42 35.7533 Q1392.11 32.5936 1387.37 32.5936 Q1381.58 32.5936 1378.21 36.4419 Q1374.85 40.2903 1374.85 46.9338 L1374.85 72.576 L1367.36 72.576 L1367.36 45.4349 Q1367.36 38.8725 1365.05 35.7533 Q1362.74 32.5936 1357.92 32.5936 Q1352.21 32.5936 1348.84 36.4824 Q1345.48 40.3308 1345.48 46.9338 L1345.48 72.576 L1337.99 72.576 L1337.99 27.2059 L1345.48 27.2059 L1345.48 34.2544 Q1348.03 30.082 1351.6 28.0971 Q1355.16 26.1121 1360.07 26.1121 Q1365.01 26.1121 1368.45 28.6237 Q1371.93 31.1352 1373.6 35.9153 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1436.67 32.4315 Q1430.67 32.4315 1427.19 37.1306 Q1423.7 41.7891 1423.7 49.9314 Q1423.7 58.0738 1427.15 62.7728 Q1430.63 67.4314 1436.67 67.4314 Q1442.62 67.4314 1446.11 62.7323 Q1449.59 58.0333 1449.59 49.9314 Q1449.59 41.8701 1446.11 37.1711 Q1442.62 32.4315 1436.67 32.4315 M1436.67 26.1121 Q1446.39 26.1121 1451.94 32.4315 Q1457.49 38.7509 1457.49 49.9314 Q1457.49 61.0714 1451.94 67.4314 Q1446.39 73.7508 1436.67 73.7508 Q1426.9 73.7508 1421.36 67.4314 Q1415.85 61.0714 1415.85 49.9314 Q1415.85 38.7509 1421.36 32.4315 Q1426.9 26.1121 1436.67 26.1121 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1499.7 34.0924 L1499.7 9.54393 L1507.15 9.54393 L1507.15 72.576 L1499.7 72.576 L1499.7 65.7705 Q1497.35 69.8214 1493.74 71.8063 Q1490.18 73.7508 1485.16 73.7508 Q1476.93 73.7508 1471.75 67.1883 Q1466.6 60.6258 1466.6 49.9314 Q1466.6 39.2371 1471.75 32.6746 Q1476.93 26.1121 1485.16 26.1121 Q1490.18 26.1121 1493.74 28.0971 Q1497.35 30.0415 1499.7 34.0924 M1474.3 49.9314 Q1474.3 58.1548 1477.66 62.8538 Q1481.07 67.5124 1486.98 67.5124 Q1492.89 67.5124 1496.3 62.8538 Q1499.7 58.1548 1499.7 49.9314 Q1499.7 41.7081 1496.3 37.0496 Q1492.89 32.3505 1486.98 32.3505 Q1481.07 32.3505 1477.66 37.0496 Q1474.3 41.7081 1474.3 49.9314 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1561.31 48.0275 L1561.31 51.6733 L1527.04 51.6733 Q1527.53 59.3701 1531.66 63.421 Q1535.83 67.4314 1543.25 67.4314 Q1547.54 67.4314 1551.55 66.3781 Q1555.6 65.3249 1559.57 63.2184 L1559.57 70.267 Q1555.56 71.9684 1551.35 72.8596 Q1547.14 73.7508 1542.8 73.7508 Q1531.94 73.7508 1525.58 67.4314 Q1519.27 61.1119 1519.27 50.3365 Q1519.27 39.1965 1525.26 32.6746 Q1531.3 26.1121 1541.51 26.1121 Q1550.66 26.1121 1555.97 32.0264 Q1561.31 37.9003 1561.31 48.0275 M1553.86 45.84 Q1553.78 39.7232 1550.42 36.0774 Q1547.1 32.4315 1541.59 32.4315 Q1535.35 32.4315 1531.58 35.9558 Q1527.85 39.4801 1527.29 45.8805 L1553.86 45.84 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1573.55 9.54393 L1581 9.54393 L1581 72.576 L1573.55 72.576 L1573.55 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1631.07 65.6895 L1659.63 65.6895 L1659.63 72.576 L1621.23 72.576 L1621.23 65.6895 Q1625.89 60.8689 1633.91 52.7671 Q1641.97 44.6248 1644.03 42.2752 Q1647.96 37.8598 1649.5 34.8216 Q1651.08 31.7429 1651.08 28.7857 Q1651.08 23.9651 1647.68 20.927 Q1644.32 17.8888 1638.89 17.8888 Q1635.04 17.8888 1630.75 19.2256 Q1626.49 20.5624 1621.63 23.2765 L1621.63 15.0127 Q1626.57 13.0277 1630.87 12.015 Q1635.16 11.0023 1638.73 11.0023 Q1648.12 11.0023 1653.72 15.7013 Q1659.31 20.4004 1659.31 28.2591 Q1659.31 31.9859 1657.89 35.3482 Q1656.51 38.6699 1652.82 43.2069 Q1651.81 44.3817 1646.38 50.0125 Q1640.95 55.6027 1631.07 65.6895 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip292)" style="stroke:#009af9; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" points="
  1334.24,767.738 1369.02,767.738 1404.07,767.738 1439.32,767.738 1474.72,767.738 1510.21,767.738 1545.75,767.738 1581.27,767.738 1616.72,767.738 1652.05,767.738 
  1687.21,767.738 1722.15,767.738 1756.82,767.738 1791.17,767.738 1825.16,767.738 1858.74,767.738 1891.87,767.738 1924.51,767.738 1956.62,767.738 1988.17,767.738 
  2019.11,767.738 2049.4,767.738 2079.02,767.738 2107.93,767.738 2136.09,767.738 2163.47,767.738 2190.04,767.738 2215.78,767.738 2240.65,767.738 2264.63,767.738 
  2287.69,767.738 2309.81,767.738 2330.96,767.738 2351.12,767.738 2370.27,767.738 2388.39,767.738 2405.46,767.738 2421.46,767.738 2436.37,767.738 2450.19,767.738 
  2462.89,767.738 2474.46,767.738 2484.9,767.738 2494.18,767.738 2502.31,767.738 2509.27,767.738 2515.05,767.738 2519.65,767.738 2523.06,767.738 2525.29,767.738 
  2526.32,767.738 2526.16,767.738 2524.81,767.738 2522.27,767.738 2518.54,767.738 2513.63,767.738 2507.54,767.738 2500.27,767.738 2491.83,767.738 2482.24,767.738 
  2471.5,767.738 2459.62,767.738 2446.61,767.738 2432.49,767.738 2417.26,767.738 2400.96,767.738 2383.57,767.738 2365.14,767.738 2345.67,767.738 2325.18,767.738 
  2303.69,767.738 2281.22,767.738 2257.79,767.738 2233.43,767.738 2208.15,767.738 2181.98,767.738 2154.94,767.738 2127.06,767.738 2098.36,767.738 2068.86,767.738 
  2038.6,767.738 2007.59,767.738 1975.86,767.738 1943.44,767.738 1910.36,767.738 1876.63,767.738 1842.29,767.738 1807.35,767.738 1771.85,767.738 1735.81,767.738 
  1699.25,767.738 1662.19,767.738 1624.67,767.738 1586.68,767.738 1548.25,767.738 1509.41,767.738 1470.15,767.738 1430.49,767.738 1390.44,767.738 1350.01,767.738 
  1316.71,767.738 1287.89,767.738 1258.88,767.738 1229.71,767.738 1200.45,767.738 1171.11,767.738 1141.76,767.738 1112.43,767.738 1083.16,767.738 1054,767.738 
  1024.98,767.738 996.151,767.738 967.55,767.738 939.211,767.738 911.176,767.738 883.478,767.738 856.153,767.738 829.237,767.738 802.76,767.738 776.757,767.738 
  751.259,767.738 726.295,767.738 701.898,767.738 678.091,767.738 654.906,767.738 632.368,767.738 610.498,767.738 589.325,767.738 568.874,767.738 549.174,767.738 
  530.237,767.738 512.09,767.738 494.752,767.738 478.242,767.738 462.577,767.738 447.774,767.738 433.848,767.738 420.815,767.738 408.688,767.738 397.481,767.738 
  387.204,767.738 377.87,767.738 369.488,767.738 362.066,767.738 355.614,767.738 350.138,767.738 345.644,767.738 342.136,767.738 339.619,767.738 338.096,767.738 
  337.57,767.738 338.039,767.738 339.505,767.738 341.967,767.738 345.422,767.738 349.868,767.738 355.301,767.738 361.715,767.738 369.105,767.738 377.464,767.738 
  386.783,767.738 397.056,767.738 408.271,767.738 420.42,767.738 433.489,767.738 447.467,767.738 462.341,767.738 478.099,767.738 494.725,767.738 512.204,767.738 
  530.52,767.738 549.655,767.738 569.596,767.738 590.323,767.738 611.818,767.738 634.06,767.738 657.029,767.738 680.713,767.738 705.088,767.738 730.132,767.738 
  755.824,767.738 782.143,767.738 809.073,767.738 836.592,767.738 864.677,767.738 893.306,767.738 922.457,767.738 952.11,767.738 982.248,767.738 1012.85,767.738 
  1043.89,767.738 1075.37,767.738 1107.25,767.738 1139.53,767.738 1172.19,767.738 1205.22,767.738 1238.6,767.738 1272.33,767.738 1306.38,767.738 1338.44,767.738 
  1360.84,767.738 1383.42,767.738 1406.13,767.738 1428.95,767.738 1451.85,767.738 1474.79,767.738 1497.74,767.738 1520.67,767.738 1543.55,767.738 1566.35,767.738 
  1589.03,767.738 1611.57,767.738 1633.93,767.738 1656.09,767.738 1678.02,767.738 1699.69,767.738 1721.06,767.738 1742.13,767.738 1762.86,767.738 1783.22,767.738 
  1803.19,767.738 1822.74,767.738 1841.85,767.738 1860.5,767.738 1878.66,767.738 1896.31,767.738 1913.44,767.738 1930.02,767.738 1946.03,767.738 1961.46,767.738 
  1976.28,767.738 1990.49,767.738 2004.06,767.738 2016.98,767.738 2029.23,767.738 2040.81,767.738 2051.69,767.738 2061.87,767.738 2071.33,767.738 2080.07,767.738 
  2088.07,767.738 2095.32,767.738 2101.82,767.738 2107.56,767.738 2112.54,767.738 2116.73,767.738 2120.15,767.738 2122.79,767.738 2124.63,767.738 2125.69,767.738 
  2125.96,767.738 2125.43,767.738 2124.12,767.738 2122.01,767.738 2119.11,767.738 2115.42,767.738 2110.94,767.738 2105.69,767.738 2099.65,767.738 2092.85,767.738 
  2085.27,767.738 2076.94,767.738 2067.86,767.738 2058.03,767.738 2047.46,767.738 2036.16,767.738 2024.15,767.738 2011.44,767.738 1998.03,767.738 1983.93,767.738 
  1969.16,767.738 1953.74,767.738 1937.66,767.738 1920.96,767.738 1903.64,767.738 1885.71,767.738 1867.19,767.738 1848.1,767.738 1828.45,767.738 1808.26,767.738 
  1787.54,767.738 1766.3,767.738 1744.58,767.738 1722.37,767.738 1699.69,767.738 1676.57,767.738 1653.02,767.738 1629.06,767.738 1604.69,767.738 1579.94,767.738 
  1554.82,767.738 1529.34,767.738 1503.52,767.738 1477.37,767.738 1450.91,767.738 1424.15,767.738 1397.1,767.738 1369.77,767.738 1342.16,767.738 1323.05,767.738 
  1307.3,767.738 1291.43,767.738 1275.47,767.738 1259.44,767.738 1243.36,767.738 1227.25,767.738 1211.13,767.738 1195.02,767.738 1178.94,767.738 1162.91,767.738 
  1146.95,767.738 1131.1,767.738 1115.35,767.738 1099.74,767.738 1084.29,767.738 1069.01,767.738 1053.92,767.738 1039.04,767.738 1024.4,767.738 1010,767.738 
  995.873,767.738 982.025,767.738 968.477,767.738 955.247,767.738 942.352,767.738 929.809,767.738 917.629,767.738 905.829,767.738 894.424,767.738 883.427,767.738 
  872.851,767.738 862.709,767.738 853.012,767.738 843.772,767.738 835,767.738 826.706,767.738 818.901,767.738 811.594,767.738 804.793,767.738 798.507,767.738 
  792.743,767.738 787.509,767.738 782.811,767.738 778.655,767.738 775.046,767.738 771.99,767.738 769.491,767.738 767.552,767.738 766.176,767.738 765.366,767.738 
  765.124,767.738 765.452,767.738 766.35,767.738 767.818,767.738 769.856,767.738 772.463,767.738 775.638,767.738 779.379,767.738 783.683,767.738 788.548,767.738 
  793.97,767.738 799.943,767.738 806.465,767.738 813.53,767.738 821.132,767.738 829.266,767.738 837.925,767.738 847.102,767.738 856.79,767.738 866.981,767.738 
  877.667,767.738 888.841,767.738 900.491,767.738 912.611,767.738 925.189,767.738 938.218,767.738 951.686,767.738 965.583,767.738 979.899,767.738 994.622,767.738 
  1009.74,767.738 1025.25,767.738 1041.13,767.738 1057.37,767.738 1073.97,767.738 1090.91,767.738 1108.17,767.738 1125.76,767.738 1143.65,767.738 1161.83,767.738 
  1180.3,767.738 1199.03,767.738 1218.04,767.738 1237.29,767.738 1256.78,767.738 1276.5,767.738 1296.44,767.738 1316.59,767.738 1335.33,767.738 1343.58,767.738 
  1351.89,767.738 
  "/>
<polyline clip-path="url(#clip292)" style="stroke:#e26f46; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" points="
  1334.24,159.645 1369.02,191.088 1404.07,220.023 1439.32,246.58 1474.72,270.887 1510.21,293.081 1545.75,313.298 1581.27,331.672 1616.72,348.333 1652.05,363.409 
  1687.21,377.023 1722.15,389.293 1756.82,400.33 1791.17,410.239 1825.16,419.119 1858.74,427.061 1891.87,434.151 1924.51,440.469 1956.62,446.087 1988.17,451.074 
  2019.11,455.49 2049.4,459.391 2079.02,462.83 2107.93,465.855 2136.09,468.507 2163.47,470.826 2190.04,472.848 2215.78,474.605 2240.65,476.126 2264.63,477.438 
  2287.69,478.564 2309.81,479.526 2330.96,480.344 2351.12,481.036 2370.27,481.618 2388.39,482.103 2405.46,482.505 2421.46,482.835 2436.37,483.103 2450.19,483.319 
  2462.89,483.492 2474.46,483.627 2484.9,483.732 2494.18,483.812 2502.31,483.872 2509.27,483.917 2515.05,483.948 2519.65,483.97 2523.06,483.985 2525.29,483.993 
  2526.32,483.997 2526.16,483.997 2524.81,483.992 2522.27,483.981 2518.54,483.965 2513.63,483.941 2507.54,483.906 2500.27,483.858 2491.83,483.793 2482.24,483.707 
  2471.5,483.594 2459.62,483.45 2446.61,483.266 2432.49,483.037 2417.26,482.752 2400.96,482.404 2383.57,481.98 2365.14,481.469 2345.67,480.857 2325.18,480.13 
  2303.69,479.271 2281.22,478.261 2257.79,477.078 2233.43,475.702 2208.15,474.104 2181.98,472.258 2154.94,470.131 2127.06,467.687 2098.36,464.889 2068.86,461.691 
  2038.6,458.045 2007.59,453.898 1975.86,449.188 1943.44,443.85 1910.36,437.809 1876.63,430.979 1842.29,423.269 1807.35,414.577 1771.85,404.787 1735.81,393.771 
  1699.25,381.386 1662.19,367.472 1624.67,351.85 1586.68,334.318 1548.25,314.652 1509.41,292.601 1470.15,267.876 1430.49,240.159 1390.44,209.088 1350.01,174.252 
  1316.71,1359.62 1287.89,1334.54 1258.88,1311.12 1229.71,1289.29 1200.45,1269.01 1171.11,1250.19 1141.76,1232.76 1112.43,1216.66 1083.16,1201.82 1054,1188.15 
  1024.98,1175.58 996.151,1164.05 967.55,1153.49 939.211,1143.83 911.176,1135 883.478,1126.94 856.153,1119.6 829.237,1112.93 802.76,1106.86 776.757,1101.35 
  751.259,1096.36 726.295,1091.84 701.898,1087.76 678.091,1084.07 654.906,1080.74 632.368,1077.75 610.498,1075.05 589.325,1072.63 568.874,1070.46 549.174,1068.51 
  530.237,1066.78 512.09,1065.22 494.752,1063.84 478.242,1062.61 462.577,1061.52 447.774,1060.56 433.848,1059.7 420.815,1058.95 408.688,1058.29 397.481,1057.71 
  387.204,1057.21 377.87,1056.78 369.488,1056.41 362.066,1056.09 355.614,1055.82 350.138,1055.61 345.644,1055.43 342.136,1055.3 339.619,1055.21 338.096,1055.15 
  337.57,1055.13 338.039,1055.15 339.505,1055.2 341.967,1055.29 345.422,1055.43 349.868,1055.6 355.301,1055.81 361.715,1056.07 369.105,1056.39 377.464,1056.76 
  386.783,1057.19 397.056,1057.69 408.271,1058.27 420.42,1058.93 433.489,1059.68 447.467,1060.54 462.341,1061.51 478.099,1062.6 494.725,1063.84 512.204,1065.23 
  530.52,1066.8 549.655,1068.56 569.596,1070.53 590.323,1072.74 611.818,1075.21 634.06,1077.96 657.029,1081.04 680.713,1084.46 705.088,1088.27 730.132,1092.52 
  755.824,1097.23 782.143,1102.46 809.073,1108.26 836.592,1114.7 864.677,1121.83 893.306,1129.73 922.457,1138.47 952.11,1148.13 982.248,1158.81 1012.85,1170.62 
  1043.89,1183.66 1075.37,1198.06 1107.25,1213.95 1139.53,1231.5 1172.19,1250.85 1205.22,1272.21 1238.6,1295.77 1272.33,1321.75 1306.38,1350.42 1338.44,163.593 
  1360.84,183.948 1383.42,203.296 1406.13,221.65 1428.95,239.023 1451.85,255.442 1474.79,270.932 1497.74,285.518 1520.67,299.231 1543.55,312.103 1566.35,324.167 
  1589.03,335.457 1611.57,346.009 1633.93,355.859 1656.09,365.041 1678.02,373.591 1699.69,381.543 1721.06,388.93 1742.13,395.787 1762.86,402.145 1783.22,408.034 
  1803.19,413.482 1822.74,418.516 1841.85,423.167 1860.5,427.456 1878.66,431.409 1896.31,435.049 1913.44,438.397 1930.02,441.473 1946.03,444.296 1961.46,446.884 
  1976.28,449.254 1990.49,451.42 2004.06,453.398 2016.98,455.199 2029.23,456.838 2040.81,458.324 2051.69,459.669 2061.87,460.882 2071.33,461.972 2080.07,462.946 
  2088.07,463.811 2095.32,464.575 2101.82,465.243 2107.56,465.819 2112.54,466.308 2116.73,466.715 2120.15,467.041 2122.79,467.289 2124.63,467.462 2125.69,467.561 
  2125.96,467.586 2125.43,467.537 2124.12,467.414 2122.01,467.216 2119.11,466.942 2115.42,466.588 2110.94,466.153 2105.69,465.632 2099.65,465.021 2092.85,464.317 
  2085.27,463.512 2076.94,462.601 2067.86,461.576 2058.03,460.429 2047.46,459.152 2036.16,457.735 2024.15,456.167 2011.44,454.437 1998.03,452.53 1983.93,450.433 
  1969.16,448.13 1953.74,445.605 1937.66,442.839 1920.96,439.812 1903.64,436.502 1885.71,432.887 1867.19,428.939 1848.1,424.631 1828.45,419.933 1808.26,414.814 
  1787.54,409.237 1766.3,403.165 1744.58,396.557 1722.37,389.366 1699.69,381.545 1676.57,373.043 1653.02,363.802 1629.06,353.763 1604.69,342.858 1579.94,331.016 
  1554.82,318.158 1529.34,304.201 1503.52,289.055 1477.37,272.621 1450.91,254.793 1424.15,235.456 1397.1,214.481 1369.77,191.732 1342.16,167.056 1323.05,1365.41 
  1307.3,1351.23 1291.43,1337.52 1275.47,1324.3 1259.44,1311.55 1243.36,1299.3 1227.25,1287.52 1211.13,1276.23 1195.02,1265.41 1178.94,1255.06 1162.91,1245.18 
  1146.95,1235.75 1131.1,1226.76 1115.35,1218.21 1099.74,1210.09 1084.29,1202.37 1069.01,1195.05 1053.92,1188.11 1039.04,1181.55 1024.4,1175.34 1010,1169.48 
  995.873,1163.95 982.025,1158.73 968.477,1153.82 955.247,1149.2 942.352,1144.86 929.809,1140.79 917.629,1136.97 905.829,1133.39 894.424,1130.05 883.427,1126.93 
  872.851,1124.02 862.709,1121.31 853.012,1118.8 843.772,1116.47 835,1114.31 826.706,1112.33 818.901,1110.5 811.594,1108.83 804.793,1107.31 798.507,1105.93 
  792.743,1104.69 787.509,1103.58 782.811,1102.6 778.655,1101.74 775.046,1101.01 771.99,1100.39 769.491,1099.89 767.552,1099.51 766.176,1099.23 765.366,1099.07 
  765.124,1099.03 765.452,1099.09 766.35,1099.27 767.818,1099.56 769.856,1099.96 772.463,1100.48 775.638,1101.13 779.379,1101.89 783.683,1102.78 788.548,1103.8 
  793.97,1104.95 799.943,1106.24 806.465,1107.68 813.53,1109.27 821.132,1111.02 829.266,1112.93 837.925,1115.03 847.102,1117.3 856.79,1119.77 866.981,1122.44 
  877.667,1125.33 888.841,1128.45 900.491,1131.81 912.611,1135.43 925.189,1139.32 938.218,1143.5 951.686,1147.99 965.583,1152.79 979.899,1157.95 994.622,1163.47 
  1009.74,1169.37 1025.25,1175.69 1041.13,1182.45 1057.37,1189.68 1073.97,1197.39 1090.91,1205.63 1108.17,1214.43 1125.76,1223.82 1143.65,1233.84 1161.83,1244.53 
  1180.3,1255.92 1199.03,1268.07 1218.04,1281.01 1237.29,1294.8 1256.78,1309.49 1276.5,1325.13 1296.44,1341.79 1316.59,1359.52 1335.33,160.669 1343.58,168.364 
  1351.89,175.954 
  "/>
<polyline clip-path="url(#clip292)" style="stroke:#008000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" stroke-dasharray="4, 8" points="
  1334.24,767.738 1369.02,751.647 1404.07,735.443 1439.32,719.162 1474.72,702.841 1510.21,686.514 1545.75,670.214 1581.27,653.978 1616.72,637.838 1652.05,621.827 
  1687.21,605.975 1722.15,590.311 1756.82,574.863 1791.17,559.657 1825.16,544.715 1858.74,530.061 1891.87,515.714 1924.51,501.694 1956.62,488.016 1988.17,474.694 
  2019.11,461.744 2049.4,449.179 2079.02,437.005 2107.93,425.233 2136.09,413.871 2163.47,402.927 2190.04,392.403 2215.78,382.304 2240.65,372.635 2264.63,363.396 
  2287.69,354.59 2309.81,346.217 2330.96,338.278 2351.12,330.773 2370.27,323.7 2388.39,317.059 2405.46,310.848 2421.46,305.066 2436.37,299.711 2450.19,294.781 
  2462.89,290.274 2474.46,286.189 2484.9,282.524 2494.18,279.277 2502.31,276.447 2509.27,274.031 2515.05,272.03 2519.65,270.441 2523.06,269.263 2525.29,268.497 
  2526.32,268.141 2526.16,268.195 2524.81,268.66 2522.27,269.536 2518.54,270.822 2513.63,272.52 2507.54,274.631 2500.27,277.157 2491.83,280.098 2482.24,283.456 
  2471.5,287.234 2459.62,291.432 2446.61,296.054 2432.49,301.102 2417.26,306.576 2400.96,312.481 2383.57,318.818 2365.14,325.589 2345.67,332.797 2325.18,340.442 
  2303.69,348.527 2281.22,357.054 2257.79,366.023 2233.43,375.434 2208.15,385.288 2181.98,395.585 2154.94,406.324 2127.06,417.502 2098.36,429.117 2068.86,441.166 
  2038.6,453.647 2007.59,466.553 1975.86,479.879 1943.44,493.615 1910.36,507.757 1876.63,522.298 1842.29,537.226 1807.35,552.529 1771.85,568.196 1735.81,584.214 
  1699.25,600.569 1662.19,617.247 1624.67,634.232 1586.68,651.511 1548.25,669.067 1509.41,686.883 1470.15,704.945 1430.49,723.237 1390.44,741.74 1350.01,760.442 
  1316.71,775.851 1287.89,789.18 1258.88,802.591 1229.71,816.059 1200.45,829.556 1171.11,843.058 1141.76,856.54 1112.43,869.978 1083.16,883.346 1054,896.621 
  1024.98,909.78 996.151,922.801 967.55,935.662 939.211,948.345 911.176,960.829 883.478,973.097 856.153,985.134 829.237,996.922 802.76,1008.45 776.757,1019.7 
  751.259,1030.66 726.295,1041.32 701.898,1051.68 678.091,1061.71 654.906,1071.42 632.368,1080.8 610.498,1089.83 589.325,1098.53 568.874,1106.86 549.174,1114.85 
  530.237,1122.47 512.09,1129.73 494.752,1136.62 478.242,1143.15 462.577,1149.31 447.774,1155.09 433.848,1160.51 420.815,1165.55 408.688,1170.22 397.481,1174.52 
  387.204,1178.44 377.87,1181.99 369.488,1185.17 362.066,1187.98 355.614,1190.41 350.138,1192.47 345.644,1194.16 342.136,1195.47 339.619,1196.41 338.096,1196.98 
  337.57,1197.18 338.039,1197 339.505,1196.45 341.967,1195.53 345.422,1194.24 349.868,1192.57 355.301,1190.53 361.715,1188.11 369.105,1185.32 377.464,1182.15 
  386.783,1178.6 397.056,1174.68 408.271,1170.38 420.42,1165.7 433.489,1160.65 447.467,1155.21 462.341,1149.4 478.099,1143.21 494.725,1136.63 512.204,1129.68 
  530.52,1122.36 549.655,1114.65 569.596,1106.57 590.323,1098.12 611.818,1089.29 634.06,1080.1 657.029,1070.54 680.713,1060.61 705.088,1050.33 730.132,1039.69 
  755.824,1028.7 782.143,1017.37 809.073,1005.71 836.592,993.708 864.677,981.386 893.306,968.752 922.457,955.813 952.11,942.58 982.248,929.06 1012.85,915.265 
  1043.89,901.208 1075.37,886.896 1107.25,872.343 1139.53,857.561 1172.19,842.562 1205.22,827.355 1238.6,811.956 1272.33,796.376 1306.38,780.625 1338.44,765.795 
  1360.84,755.429 1383.42,744.988 1406.13,734.487 1428.95,723.947 1451.85,713.382 1474.79,702.809 1497.74,692.246 1520.67,681.711 1543.55,671.22 1566.35,660.791 
  1589.03,650.438 1611.57,640.18 1633.93,630.029 1656.09,620.001 1678.02,610.111 1699.69,600.372 1721.06,590.796 1742.13,581.395 1762.86,572.181 1783.22,563.165 
  1803.19,554.359 1822.74,545.773 1841.85,537.416 1860.5,529.295 1878.66,521.421 1896.31,513.799 1913.44,506.437 1930.02,499.341 1946.03,492.516 1961.46,485.968 
  1976.28,479.702 1990.49,473.721 2004.06,468.031 2016.98,462.634 2029.23,457.533 2040.81,452.733 2051.69,448.235 2061.87,444.041 2071.33,440.154 2080.07,436.576 
  2088.07,433.308 2095.32,430.353 2101.82,427.709 2107.56,425.38 2112.54,423.366 2116.73,421.669 2120.15,420.288 2122.79,419.225 2124.63,418.48 2125.69,418.053 
  2125.96,417.945 2125.43,418.157 2124.12,418.688 2122.01,419.539 2119.11,420.709 2115.42,422.2 2110.94,424.011 2105.69,426.141 2099.65,428.591 2092.85,431.361 
  2085.27,434.449 2076.94,437.855 2067.86,441.581 2058.03,445.623 2047.46,449.982 2036.16,454.656 2024.15,459.644 2011.44,464.944 1998.03,470.556 1983.93,476.478 
  1969.16,482.707 1953.74,489.241 1937.66,496.078 1920.96,503.215 1903.64,510.647 1885.71,518.373 1867.19,526.389 1848.1,534.69 1828.45,543.272 1808.26,552.131 
  1787.54,561.259 1766.3,570.653 1744.58,580.307 1722.37,590.214 1699.69,600.368 1676.57,610.762 1653.02,621.389 1629.06,632.239 1604.69,643.307 1579.94,654.584 
  1554.82,666.063 1529.34,677.734 1503.52,689.588 1477.37,701.618 1450.91,713.812 1424.15,726.163 1397.1,738.662 1369.77,751.301 1342.16,764.073 1323.05,772.913 
  1307.3,780.203 1291.43,787.543 1275.47,794.922 1259.44,802.331 1243.36,809.759 1227.25,817.196 1211.13,824.633 1195.02,832.058 1178.94,839.46 1162.91,846.83 
  1146.95,854.156 1131.1,861.429 1115.35,868.64 1099.74,875.777 1084.29,882.831 1069.01,889.794 1053.92,896.656 1039.04,903.408 1024.4,910.043 1010,916.552 
  995.873,922.926 982.025,929.16 968.477,935.246 955.247,941.176 942.352,946.942 929.809,952.539 917.629,957.961 905.829,963.202 894.424,968.257 883.427,973.12 
  872.851,977.787 862.709,982.252 853.012,986.513 843.772,990.565 835,994.404 826.706,998.026 818.901,1001.43 811.594,1004.61 804.793,1007.57 798.507,1010.29 
  792.743,1012.79 787.509,1015.06 782.811,1017.09 778.655,1018.88 775.046,1020.44 771.99,1021.75 769.491,1022.83 767.552,1023.66 766.176,1024.26 765.366,1024.6 
  765.124,1024.71 765.452,1024.57 766.35,1024.18 767.818,1023.55 769.856,1022.67 772.463,1021.55 775.638,1020.18 779.379,1018.57 783.683,1016.71 788.548,1014.61 
  793.97,1012.26 799.943,1009.67 806.465,1006.84 813.53,1003.77 821.132,1000.46 829.266,996.909 837.925,993.124 847.102,989.105 856.79,984.854 866.981,980.373 
  877.667,975.663 888.841,970.727 900.491,965.569 912.611,960.191 925.189,954.597 938.218,948.788 951.686,942.769 965.583,936.544 979.899,930.116 994.622,923.49 
  1009.74,916.669 1025.25,909.658 1041.13,902.462 1057.37,895.085 1073.97,887.533 1090.91,879.811 1108.17,871.923 1125.76,863.876 1143.65,855.675 1161.83,847.326 
  1180.3,838.834 1199.03,830.206 1218.04,821.447 1237.29,812.564 1256.78,803.561 1276.5,794.447 1296.44,785.226 1316.59,775.905 1335.33,767.236 1343.58,763.418 
  1351.89,759.572 
  "/>
<polyline clip-path="url(#clip292)" style="stroke:#008000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" points="
  1334.24,767.738 1369.02,751.647 1404.07,735.443 1439.32,719.162 1474.72,702.841 1510.21,686.514 1545.75,670.214 1581.27,653.978 1616.72,637.838 1652.05,621.827 
  1687.21,605.975 1722.15,590.311 1756.82,574.863 1791.17,559.657 1825.16,544.715 1858.74,530.061 1891.87,515.714 1924.51,501.694 1956.62,488.016 1988.17,474.694 
  2019.11,461.744 2049.4,449.179 2079.02,437.005 2107.93,425.233 2136.09,413.871 2163.47,402.927 2190.04,392.403 2215.78,382.304 2240.65,372.635 2264.63,363.396 
  2287.69,354.59 2309.81,346.217 2330.96,338.278 2351.12,330.773 2370.27,323.7 2388.39,317.059 2405.46,310.848 2421.46,305.066 2436.37,299.711 2450.19,294.781 
  2462.89,290.274 2474.46,286.189 2484.9,282.524 2494.18,279.277 2502.31,276.447 2509.27,274.031 2515.05,272.03 2519.65,270.441 2523.06,269.263 2525.29,268.497 
  2526.32,268.141 2526.16,268.195 2524.81,268.66 2522.27,269.536 2518.54,270.822 2513.63,272.52 2507.54,274.631 2500.27,277.157 2491.83,280.098 2482.24,283.456 
  2471.5,287.234 2459.62,291.432 2446.61,296.054 2432.49,301.102 2417.26,306.576 2400.96,312.481 2383.57,318.818 2365.14,325.589 2345.67,332.797 2325.18,340.442 
  2303.69,348.527 2281.22,357.054 2257.79,366.023 2233.43,375.434 2208.15,385.288 2181.98,395.585 2154.94,406.324 2127.06,417.502 2098.36,429.117 2068.86,441.166 
  2038.6,453.647 2007.59,466.553 1975.86,479.879 1943.44,493.615 1910.36,507.757 1876.63,522.298 1842.29,537.226 1807.35,552.529 1771.85,568.196 1735.81,584.214 
  1699.25,600.569 1662.19,617.247 1624.67,634.232 1586.68,651.511 1548.25,669.067 1509.41,686.883 1470.15,704.945 1430.49,723.237 1390.44,741.74 1350.01,760.442 
  1316.71,775.851 1287.89,789.18 1258.88,802.591 1229.71,816.059 1200.45,829.556 1171.11,843.058 1141.76,856.54 1112.43,869.978 1083.16,883.346 1054,896.621 
  1024.98,909.78 996.151,922.801 967.55,935.662 939.211,948.345 911.176,960.829 883.478,973.097 856.153,985.134 829.237,996.922 802.76,1008.45 776.757,1019.7 
  751.259,1030.66 726.295,1041.32 701.898,1051.68 678.091,1061.71 654.906,1071.42 632.368,1080.8 610.498,1089.83 589.325,1098.53 568.874,1106.86 549.174,1114.85 
  530.237,1122.47 512.09,1129.73 494.752,1136.62 478.242,1143.15 462.577,1149.31 447.774,1155.09 433.848,1160.51 420.815,1165.55 408.688,1170.22 397.481,1174.52 
  387.204,1178.44 377.87,1181.99 369.488,1185.17 362.066,1187.98 355.614,1190.41 350.138,1192.47 345.644,1194.16 342.136,1195.47 339.619,1196.41 338.096,1196.98 
  337.57,1197.18 338.039,1197 339.505,1196.45 341.967,1195.53 345.422,1194.24 349.868,1192.57 355.301,1190.53 361.715,1188.11 369.105,1185.32 377.464,1182.15 
  386.783,1178.6 397.056,1174.68 408.271,1170.38 420.42,1165.7 433.489,1160.65 447.467,1155.21 462.341,1149.4 478.099,1143.21 494.725,1136.63 512.204,1129.68 
  530.52,1122.36 549.655,1114.65 569.596,1106.57 590.323,1098.12 611.818,1089.29 634.06,1080.1 657.029,1070.54 680.713,1060.61 705.088,1050.33 730.132,1039.69 
  755.824,1028.7 782.143,1017.37 809.073,1005.71 836.592,993.708 864.677,981.386 893.306,968.752 922.457,955.813 952.11,942.58 982.248,929.06 1012.85,915.265 
  1043.89,901.208 1075.37,886.896 1107.25,872.343 1139.53,857.561 1172.19,842.562 1205.22,827.355 1238.6,811.956 1272.33,796.376 1306.38,780.625 1338.44,765.795 
  1360.84,755.429 1383.42,744.988 1406.13,734.487 1428.95,723.947 1451.85,713.382 1474.79,702.809 1497.74,692.246 1520.67,681.711 1543.55,671.22 1566.35,660.791 
  1589.03,650.438 1611.57,640.18 1633.93,630.029 1656.09,620.001 1678.02,610.111 1699.69,600.372 1721.06,590.796 1742.13,581.395 1762.86,572.181 1783.22,563.165 
  1803.19,554.359 1822.74,545.773 1841.85,537.416 1860.5,529.295 1878.66,521.421 1896.31,513.799 1913.44,506.437 1930.02,499.341 1946.03,492.516 1961.46,485.968 
  1976.28,479.702 1990.49,473.721 2004.06,468.031 2016.98,462.634 2029.23,457.533 2040.81,452.733 2051.69,448.235 2061.87,444.041 2071.33,440.154 2080.07,436.576 
  2088.07,433.308 2095.32,430.353 2101.82,427.709 2107.56,425.38 2112.54,423.366 2116.73,421.669 2120.15,420.288 2122.79,419.225 2124.63,418.48 2125.69,418.053 
  2125.96,417.945 2125.43,418.157 2124.12,418.688 2122.01,419.539 2119.11,420.709 2115.42,422.2 2110.94,424.011 2105.69,426.141 2099.65,428.591 2092.85,431.361 
  2085.27,434.449 2076.94,437.855 2067.86,441.581 2058.03,445.623 2047.46,449.982 2036.16,454.656 2024.15,459.644 2011.44,464.944 1998.03,470.556 1983.93,476.478 
  1969.16,482.707 1953.74,489.241 1937.66,496.078 1920.96,503.215 1903.64,510.647 1885.71,518.373 1867.19,526.389 1848.1,534.69 1828.45,543.272 1808.26,552.131 
  1787.54,561.259 1766.3,570.653 1744.58,580.307 1722.37,590.214 1699.69,600.368 1676.57,610.762 1653.02,621.389 1629.06,632.239 1604.69,643.307 1579.94,654.584 
  1554.82,666.063 1529.34,677.734 1503.52,689.588 1477.37,701.618 1450.91,713.812 1424.15,726.163 1397.1,738.662 1369.77,751.301 1342.16,764.073 1323.05,772.913 
  1307.3,780.203 1291.43,787.543 1275.47,794.922 1259.44,802.331 1243.36,809.759 1227.25,817.196 1211.13,824.633 1195.02,832.058 1178.94,839.46 1162.91,846.83 
  1146.95,854.156 1131.1,861.429 1115.35,868.64 1099.74,875.777 1084.29,882.831 1069.01,889.794 1053.92,896.656 1039.04,903.408 1024.4,910.043 1010,916.552 
  995.873,922.926 982.025,929.16 968.477,935.246 955.247,941.176 942.352,946.942 929.809,952.539 917.629,957.961 905.829,963.202 894.424,968.257 883.427,973.12 
  872.851,977.787 862.709,982.252 853.012,986.513 843.772,990.565 835,994.404 826.706,998.026 818.901,1001.43 811.594,1004.61 804.793,1007.57 798.507,1010.29 
  792.743,1012.79 787.509,1015.06 782.811,1017.09 778.655,1018.88 775.046,1020.44 771.99,1021.75 769.491,1022.83 767.552,1023.66 766.176,1024.26 765.366,1024.6 
  765.124,1024.71 765.452,1024.57 766.35,1024.18 767.818,1023.55 769.856,1022.67 772.463,1021.55 775.638,1020.18 779.379,1018.57 783.683,1016.71 788.548,1014.61 
  793.97,1012.26 799.943,1009.67 806.465,1006.84 813.53,1003.77 821.132,1000.46 829.266,996.909 837.925,993.124 847.102,989.105 856.79,984.854 866.981,980.373 
  877.667,975.663 888.841,970.727 900.491,965.569 912.611,960.191 925.189,954.597 938.218,948.788 951.686,942.769 965.583,936.544 979.899,930.116 994.622,923.49 
  1009.74,916.669 1025.25,909.658 1041.13,902.462 1057.37,895.085 1073.97,887.533 1090.91,879.811 1108.17,871.923 1125.76,863.876 1143.65,855.675 1161.83,847.326 
  1180.3,838.834 1199.03,830.206 1218.04,821.447 1237.29,812.564 1256.78,803.561 1276.5,794.447 1296.44,785.226 1316.59,775.905 1335.33,767.236 1343.58,763.418 
  1351.89,759.572 
  "/>
<path clip-path="url(#clip290)" d="
M1423.66 554.876 L2284.85 554.876 L2284.85 166.076 L1423.66 166.076  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<polyline clip-path="url(#clip290)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1423.66,554.876 2284.85,554.876 2284.85,166.076 1423.66,166.076 1423.66,554.876 
  "/>
<polyline clip-path="url(#clip290)" style="stroke:#009af9; stroke-linecap:butt; stroke-linejoin:round; stroke-width:6; stroke-opacity:1; fill:none" points="
  1446.29,243.836 1582.1,243.836 
  "/>
<path clip-path="url(#clip290)" d="M1638.1 219.617 L1638.1 226.457 Q1634.1 224.548 1630.56 223.61 Q1627.02 222.673 1623.72 222.673 Q1617.99 222.673 1614.87 224.895 Q1611.78 227.117 1611.78 231.214 Q1611.78 234.652 1613.83 236.423 Q1615.91 238.159 1621.67 239.235 L1625.91 240.103 Q1633.76 241.596 1637.47 245.381 Q1641.22 249.131 1641.22 255.45 Q1641.22 262.985 1636.15 266.874 Q1631.12 270.763 1621.36 270.763 Q1617.68 270.763 1613.51 269.929 Q1609.38 269.096 1604.94 267.464 L1604.94 260.242 Q1609.21 262.638 1613.31 263.853 Q1617.4 265.068 1621.36 265.068 Q1627.37 265.068 1630.63 262.707 Q1633.9 260.346 1633.9 255.971 Q1633.9 252.152 1631.54 249.999 Q1629.21 247.846 1623.86 246.77 L1619.59 245.936 Q1611.74 244.374 1608.24 241.041 Q1604.73 237.707 1604.73 231.77 Q1604.73 224.895 1609.56 220.937 Q1614.42 216.978 1622.92 216.978 Q1626.57 216.978 1630.35 217.638 Q1634.14 218.298 1638.1 219.617 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1651.88 230.867 L1658.27 230.867 L1658.27 269.756 L1651.88 269.756 L1651.88 230.867 M1651.88 215.728 L1658.27 215.728 L1658.27 223.818 L1651.88 223.818 L1651.88 215.728 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1701.92 238.332 Q1704.31 234.027 1707.65 231.978 Q1710.98 229.93 1715.49 229.93 Q1721.57 229.93 1724.87 234.2 Q1728.17 238.436 1728.17 246.284 L1728.17 269.756 L1721.74 269.756 L1721.74 246.492 Q1721.74 240.902 1719.76 238.193 Q1717.78 235.485 1713.72 235.485 Q1708.76 235.485 1705.88 238.784 Q1702.99 242.082 1702.99 247.777 L1702.99 269.756 L1696.57 269.756 L1696.57 246.492 Q1696.57 240.867 1694.59 238.193 Q1692.61 235.485 1688.48 235.485 Q1683.58 235.485 1680.7 238.818 Q1677.82 242.117 1677.82 247.777 L1677.82 269.756 L1671.4 269.756 L1671.4 230.867 L1677.82 230.867 L1677.82 236.909 Q1680.01 233.332 1683.06 231.631 Q1686.12 229.93 1690.32 229.93 Q1694.56 229.93 1697.51 232.082 Q1700.49 234.235 1701.92 238.332 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1747.09 263.922 L1747.09 284.547 L1740.67 284.547 L1740.67 230.867 L1747.09 230.867 L1747.09 236.77 Q1749.1 233.298 1752.16 231.631 Q1755.25 229.93 1759.52 229.93 Q1766.6 229.93 1771.01 235.555 Q1775.46 241.179 1775.46 250.346 Q1775.46 259.513 1771.01 265.138 Q1766.6 270.763 1759.52 270.763 Q1755.25 270.763 1752.16 269.096 Q1749.1 267.395 1747.09 263.922 M1768.83 250.346 Q1768.83 243.298 1765.91 239.304 Q1763.03 235.277 1757.96 235.277 Q1752.89 235.277 1749.97 239.304 Q1747.09 243.298 1747.09 250.346 Q1747.09 257.395 1749.97 261.422 Q1752.89 265.415 1757.96 265.415 Q1763.03 265.415 1765.91 261.422 Q1768.83 257.395 1768.83 250.346 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1786.05 215.728 L1792.44 215.728 L1792.44 269.756 L1786.05 269.756 L1786.05 215.728 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1839.07 248.714 L1839.07 251.839 L1809.69 251.839 Q1810.11 258.436 1813.65 261.909 Q1817.23 265.346 1823.58 265.346 Q1827.26 265.346 1830.7 264.443 Q1834.17 263.54 1837.58 261.735 L1837.58 267.777 Q1834.14 269.235 1830.53 269.999 Q1826.92 270.763 1823.2 270.763 Q1813.9 270.763 1808.44 265.346 Q1803.03 259.929 1803.03 250.693 Q1803.03 241.145 1808.17 235.555 Q1813.34 229.93 1822.09 229.93 Q1829.94 229.93 1834.49 234.999 Q1839.07 240.034 1839.07 248.714 M1832.68 246.839 Q1832.61 241.596 1829.73 238.471 Q1826.88 235.346 1822.16 235.346 Q1816.81 235.346 1813.58 238.367 Q1810.39 241.388 1809.9 246.874 L1832.68 246.839 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1849.83 217.916 L1879.62 217.916 L1879.62 223.818 L1856.85 223.818 L1856.85 239.096 L1877.4 239.096 L1877.4 244.999 L1856.85 244.999 L1856.85 269.756 L1849.83 269.756 L1849.83 217.916 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1890.74 217.916 L1901.19 217.916 L1914.42 253.193 L1927.71 217.916 L1938.17 217.916 L1938.17 269.756 L1931.33 269.756 L1931.33 224.235 L1917.96 259.79 L1910.91 259.79 L1897.54 224.235 L1897.54 269.756 L1890.74 269.756 L1890.74 217.916 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1951.29 217.916 L1958.34 217.916 L1958.34 249.409 Q1958.34 257.742 1961.36 261.422 Q1964.38 265.068 1971.15 265.068 Q1977.89 265.068 1980.91 261.422 Q1983.93 257.742 1983.93 249.409 L1983.93 217.916 L1990.98 217.916 L1990.98 250.277 Q1990.98 260.415 1985.94 265.589 Q1980.94 270.763 1971.15 270.763 Q1961.33 270.763 1956.29 265.589 Q1951.29 260.415 1951.29 250.277 L1951.29 217.916 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip290)" style="stroke:#e26f46; stroke-linecap:butt; stroke-linejoin:round; stroke-width:6; stroke-opacity:1; fill:none" points="
  1446.29,321.596 1582.1,321.596 
  "/>
<path clip-path="url(#clip290)" d="M1627.51 314.599 Q1626.43 313.974 1625.15 313.696 Q1623.9 313.384 1622.37 313.384 Q1616.95 313.384 1614.04 316.926 Q1611.15 320.433 1611.15 327.03 L1611.15 347.516 L1604.73 347.516 L1604.73 308.627 L1611.15 308.627 L1611.15 314.669 Q1613.17 311.127 1616.4 309.426 Q1619.63 307.69 1624.24 307.69 Q1624.9 307.69 1625.7 307.794 Q1626.5 307.863 1627.47 308.037 L1627.51 314.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1665.91 326.474 L1665.91 329.599 L1636.54 329.599 Q1636.95 336.196 1640.49 339.669 Q1644.07 343.106 1650.42 343.106 Q1654.1 343.106 1657.54 342.203 Q1661.01 341.3 1664.42 339.495 L1664.42 345.537 Q1660.98 346.995 1657.37 347.759 Q1653.76 348.523 1650.04 348.523 Q1640.74 348.523 1635.29 343.106 Q1629.87 337.689 1629.87 328.453 Q1629.87 318.905 1635.01 313.315 Q1640.18 307.69 1648.93 307.69 Q1656.78 307.69 1661.33 312.759 Q1665.91 317.794 1665.91 326.474 M1659.52 324.599 Q1659.45 319.356 1656.57 316.231 Q1653.72 313.106 1649 313.106 Q1643.65 313.106 1640.42 316.127 Q1637.23 319.148 1636.74 324.634 L1659.52 324.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1696.08 293.488 L1696.08 298.801 L1689.97 298.801 Q1686.53 298.801 1685.18 300.19 Q1683.86 301.578 1683.86 305.19 L1683.86 308.627 L1694.38 308.627 L1694.38 313.592 L1683.86 313.592 L1683.86 347.516 L1677.44 347.516 L1677.44 313.592 L1671.33 313.592 L1671.33 308.627 L1677.44 308.627 L1677.44 305.919 Q1677.44 299.426 1680.46 296.474 Q1683.48 293.488 1690.04 293.488 L1696.08 293.488 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1734.69 326.474 L1734.69 329.599 L1705.32 329.599 Q1705.74 336.196 1709.28 339.669 Q1712.85 343.106 1719.21 343.106 Q1722.89 343.106 1726.33 342.203 Q1729.8 341.3 1733.2 339.495 L1733.2 345.537 Q1729.76 346.995 1726.15 347.759 Q1722.54 348.523 1718.83 348.523 Q1709.52 348.523 1704.07 343.106 Q1698.65 337.689 1698.65 328.453 Q1698.65 318.905 1703.79 313.315 Q1708.97 307.69 1717.72 307.69 Q1725.56 307.69 1730.11 312.759 Q1734.69 317.794 1734.69 326.474 M1728.31 324.599 Q1728.24 319.356 1725.35 316.231 Q1722.51 313.106 1717.78 313.106 Q1712.44 313.106 1709.21 316.127 Q1706.01 319.148 1705.53 324.634 L1728.31 324.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1767.72 314.599 Q1766.64 313.974 1765.35 313.696 Q1764.1 313.384 1762.58 313.384 Q1757.16 313.384 1754.24 316.926 Q1751.36 320.433 1751.36 327.03 L1751.36 347.516 L1744.94 347.516 L1744.94 308.627 L1751.36 308.627 L1751.36 314.669 Q1753.37 311.127 1756.6 309.426 Q1759.83 307.69 1764.45 307.69 Q1765.11 307.69 1765.91 307.794 Q1766.71 307.863 1767.68 308.037 L1767.72 314.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1806.12 326.474 L1806.12 329.599 L1776.74 329.599 Q1777.16 336.196 1780.7 339.669 Q1784.28 343.106 1790.63 343.106 Q1794.31 343.106 1797.75 342.203 Q1801.22 341.3 1804.62 339.495 L1804.62 345.537 Q1801.19 346.995 1797.58 347.759 Q1793.96 348.523 1790.25 348.523 Q1780.94 348.523 1775.49 343.106 Q1770.08 337.689 1770.08 328.453 Q1770.08 318.905 1775.22 313.315 Q1780.39 307.69 1789.14 307.69 Q1796.99 307.69 1801.53 312.759 Q1806.12 317.794 1806.12 326.474 M1799.73 324.599 Q1799.66 319.356 1796.78 316.231 Q1793.93 313.106 1789.21 313.106 Q1783.86 313.106 1780.63 316.127 Q1777.44 319.148 1776.95 324.634 L1799.73 324.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1848.93 324.044 L1848.93 347.516 L1842.54 347.516 L1842.54 324.252 Q1842.54 318.731 1840.39 315.988 Q1838.24 313.245 1833.93 313.245 Q1828.76 313.245 1825.77 316.544 Q1822.78 319.842 1822.78 325.537 L1822.78 347.516 L1816.36 347.516 L1816.36 308.627 L1822.78 308.627 L1822.78 314.669 Q1825.08 311.162 1828.17 309.426 Q1831.29 307.69 1835.35 307.69 Q1842.05 307.69 1845.49 311.856 Q1848.93 315.988 1848.93 324.044 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1889.66 310.12 L1889.66 316.092 Q1886.95 314.599 1884.21 313.87 Q1881.5 313.106 1878.72 313.106 Q1872.51 313.106 1869.07 317.064 Q1865.63 320.988 1865.63 328.106 Q1865.63 335.224 1869.07 339.182 Q1872.51 343.106 1878.72 343.106 Q1881.5 343.106 1884.21 342.377 Q1886.95 341.613 1889.66 340.12 L1889.66 346.023 Q1886.99 347.273 1884.1 347.898 Q1881.26 348.523 1878.03 348.523 Q1869.24 348.523 1864.07 343.002 Q1858.9 337.481 1858.9 328.106 Q1858.9 318.592 1864.1 313.141 Q1869.35 307.69 1878.44 307.69 Q1881.4 307.69 1884.21 308.315 Q1887.02 308.905 1889.66 310.12 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1934.03 326.474 L1934.03 329.599 L1904.66 329.599 Q1905.08 336.196 1908.62 339.669 Q1912.19 343.106 1918.55 343.106 Q1922.23 343.106 1925.67 342.203 Q1929.14 341.3 1932.54 339.495 L1932.54 345.537 Q1929.1 346.995 1925.49 347.759 Q1921.88 348.523 1918.17 348.523 Q1908.86 348.523 1903.41 343.106 Q1897.99 337.689 1897.99 328.453 Q1897.99 318.905 1903.13 313.315 Q1908.3 307.69 1917.05 307.69 Q1924.9 307.69 1929.45 312.759 Q1934.03 317.794 1934.03 326.474 M1927.64 324.599 Q1927.58 319.356 1924.69 316.231 Q1921.85 313.106 1917.12 313.106 Q1911.78 313.106 1908.55 316.127 Q1905.35 319.148 1904.87 324.634 L1927.64 324.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip290)" style="stroke:#008000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:6; stroke-opacity:1; fill:none" stroke-dasharray="3, 6" points="
  1446.29,399.356 1582.1,399.356 
  "/>
<path clip-path="url(#clip290)" d="M1604.73 373.436 L1614.17 373.436 L1637.16 416.804 L1637.16 373.436 L1643.97 373.436 L1643.97 425.276 L1634.52 425.276 L1611.54 381.908 L1611.54 425.276 L1604.73 425.276 L1604.73 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1690.91 404.234 L1690.91 407.359 L1661.54 407.359 Q1661.95 413.956 1665.49 417.429 Q1669.07 420.866 1675.42 420.866 Q1679.1 420.866 1682.54 419.963 Q1686.01 419.06 1689.42 417.255 L1689.42 423.297 Q1685.98 424.755 1682.37 425.519 Q1678.76 426.283 1675.04 426.283 Q1665.74 426.283 1660.29 420.866 Q1654.87 415.449 1654.87 406.213 Q1654.87 396.665 1660.01 391.075 Q1665.18 385.45 1673.93 385.45 Q1681.78 385.45 1686.33 390.519 Q1690.91 395.554 1690.91 404.234 M1684.52 402.359 Q1684.45 397.116 1681.57 393.991 Q1678.72 390.866 1674 390.866 Q1668.65 390.866 1665.42 393.887 Q1662.23 396.908 1661.74 402.394 L1684.52 402.359 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1700.74 409.929 L1700.74 386.387 L1707.13 386.387 L1707.13 409.686 Q1707.13 415.206 1709.28 417.984 Q1711.43 420.727 1715.74 420.727 Q1720.91 420.727 1723.9 417.429 Q1726.92 414.13 1726.92 408.436 L1726.92 386.387 L1733.31 386.387 L1733.31 425.276 L1726.92 425.276 L1726.92 419.304 Q1724.59 422.845 1721.5 424.581 Q1718.44 426.283 1714.38 426.283 Q1707.68 426.283 1704.21 422.116 Q1700.74 417.949 1700.74 409.929 M1716.81 385.45 L1716.81 385.45 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1769 392.359 Q1767.92 391.734 1766.64 391.456 Q1765.39 391.144 1763.86 391.144 Q1758.44 391.144 1755.53 394.686 Q1752.65 398.193 1752.65 404.79 L1752.65 425.276 L1746.22 425.276 L1746.22 386.387 L1752.65 386.387 L1752.65 392.429 Q1754.66 388.887 1757.89 387.186 Q1761.12 385.45 1765.74 385.45 Q1766.4 385.45 1767.19 385.554 Q1767.99 385.623 1768.97 385.797 L1769 392.359 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1793.37 405.727 Q1785.63 405.727 1782.65 407.498 Q1779.66 409.269 1779.66 413.54 Q1779.66 416.942 1781.88 418.956 Q1784.14 420.935 1787.99 420.935 Q1793.31 420.935 1796.5 417.185 Q1799.73 413.401 1799.73 407.151 L1799.73 405.727 L1793.37 405.727 M1806.12 403.088 L1806.12 425.276 L1799.73 425.276 L1799.73 419.373 Q1797.54 422.915 1794.28 424.616 Q1791.01 426.283 1786.29 426.283 Q1780.32 426.283 1776.78 422.949 Q1773.27 419.581 1773.27 413.956 Q1773.27 407.394 1777.65 404.061 Q1782.06 400.727 1790.77 400.727 L1799.73 400.727 L1799.73 400.102 Q1799.73 395.693 1796.81 393.297 Q1793.93 390.866 1788.69 390.866 Q1785.35 390.866 1782.19 391.665 Q1779.03 392.463 1776.12 394.061 L1776.12 388.158 Q1779.62 386.804 1782.92 386.144 Q1786.22 385.45 1789.35 385.45 Q1797.78 385.45 1801.95 389.825 Q1806.12 394.199 1806.12 403.088 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1819.28 371.248 L1825.67 371.248 L1825.67 425.276 L1819.28 425.276 L1819.28 371.248 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1839.31 373.436 L1869.1 373.436 L1869.1 379.338 L1846.33 379.338 L1846.33 394.616 L1866.88 394.616 L1866.88 400.519 L1846.33 400.519 L1846.33 425.276 L1839.31 425.276 L1839.31 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1880.21 373.436 L1890.67 373.436 L1903.89 408.713 L1917.19 373.436 L1927.64 373.436 L1927.64 425.276 L1920.8 425.276 L1920.8 379.755 L1907.44 415.31 L1900.39 415.31 L1887.02 379.755 L1887.02 425.276 L1880.21 425.276 L1880.21 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1940.77 373.436 L1947.82 373.436 L1947.82 404.929 Q1947.82 413.262 1950.84 416.942 Q1953.86 420.588 1960.63 420.588 Q1967.37 420.588 1970.39 416.942 Q1973.41 413.262 1973.41 404.929 L1973.41 373.436 L1980.46 373.436 L1980.46 405.797 Q1980.46 415.935 1975.42 421.109 Q1970.42 426.283 1960.63 426.283 Q1950.8 426.283 1945.77 421.109 Q1940.77 415.935 1940.77 405.797 L1940.77 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M2031.29 371.318 Q2026.64 379.304 2024.38 387.116 Q2022.12 394.929 2022.12 402.949 Q2022.12 410.97 2024.38 418.852 Q2026.67 426.699 2031.29 434.651 L2025.73 434.651 Q2020.53 426.491 2017.92 418.609 Q2015.35 410.727 2015.35 402.949 Q2015.35 395.206 2017.92 387.359 Q2020.49 379.512 2025.73 371.318 L2031.29 371.318 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M2050.63 419.373 L2075.11 419.373 L2075.11 425.276 L2042.19 425.276 L2042.19 419.373 Q2046.19 415.241 2053.06 408.297 Q2059.97 401.318 2061.74 399.304 Q2065.11 395.519 2066.43 392.915 Q2067.78 390.276 2067.78 387.741 Q2067.78 383.609 2064.87 381.005 Q2061.98 378.401 2057.33 378.401 Q2054.03 378.401 2050.35 379.547 Q2046.71 380.693 2042.54 383.019 L2042.54 375.936 Q2046.78 374.234 2050.46 373.366 Q2054.14 372.498 2057.19 372.498 Q2065.25 372.498 2070.04 376.526 Q2074.83 380.554 2074.83 387.29 Q2074.83 390.484 2073.62 393.366 Q2072.44 396.213 2069.28 400.102 Q2068.41 401.109 2063.76 405.936 Q2059.1 410.727 2050.63 419.373 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M2089.9 373.436 L2117.44 373.436 L2117.44 379.338 L2096.32 379.338 L2096.32 392.047 Q2097.85 391.526 2099.38 391.283 Q2100.91 391.005 2102.44 391.005 Q2111.12 391.005 2116.19 395.762 Q2121.25 400.519 2121.25 408.644 Q2121.25 417.012 2116.05 421.665 Q2110.84 426.283 2101.36 426.283 Q2098.1 426.283 2094.69 425.727 Q2091.32 425.172 2087.71 424.06 L2087.71 417.012 Q2090.84 418.713 2094.17 419.547 Q2097.5 420.38 2101.22 420.38 Q2107.23 420.38 2110.73 417.22 Q2114.24 414.06 2114.24 408.644 Q2114.24 403.227 2110.73 400.068 Q2107.23 396.908 2101.22 396.908 Q2098.41 396.908 2095.6 397.533 Q2092.82 398.158 2089.9 399.477 L2089.9 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M2150.07 378.054 Q2144.66 378.054 2141.91 383.401 Q2139.21 388.713 2139.21 399.408 Q2139.21 410.067 2141.91 415.415 Q2144.66 420.727 2150.07 420.727 Q2155.53 420.727 2158.23 415.415 Q2160.98 410.067 2160.98 399.408 Q2160.98 388.713 2158.23 383.401 Q2155.53 378.054 2150.07 378.054 M2150.07 372.498 Q2158.79 372.498 2163.37 379.408 Q2167.99 386.283 2167.99 399.408 Q2167.99 412.498 2163.37 419.408 Q2158.79 426.283 2150.07 426.283 Q2141.36 426.283 2136.74 419.408 Q2132.16 412.498 2132.16 399.408 Q2132.16 386.283 2136.74 379.408 Q2141.36 372.498 2150.07 372.498 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M2195.32 378.054 Q2189.9 378.054 2187.16 383.401 Q2184.45 388.713 2184.45 399.408 Q2184.45 410.067 2187.16 415.415 Q2189.9 420.727 2195.32 420.727 Q2200.77 420.727 2203.48 415.415 Q2206.22 410.067 2206.22 399.408 Q2206.22 388.713 2203.48 383.401 Q2200.77 378.054 2195.32 378.054 M2195.32 372.498 Q2204.03 372.498 2208.62 379.408 Q2213.23 386.283 2213.23 399.408 Q2213.23 412.498 2208.62 419.408 Q2204.03 426.283 2195.32 426.283 Q2186.6 426.283 2181.98 419.408 Q2177.4 412.498 2177.4 399.408 Q2177.4 386.283 2181.98 379.408 Q2186.6 372.498 2195.32 372.498 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M2223.65 371.318 L2229.21 371.318 Q2234.41 379.512 2236.98 387.359 Q2239.59 395.206 2239.59 402.949 Q2239.59 410.727 2236.98 418.609 Q2234.41 426.491 2229.21 434.651 L2223.65 434.651 Q2228.27 426.699 2230.53 418.852 Q2232.82 410.97 2232.82 402.949 Q2232.82 394.929 2230.53 387.116 Q2228.27 379.304 2223.65 371.318 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip290)" style="stroke:#008000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:6; stroke-opacity:1; fill:none" points="
  1446.29,477.116 1582.1,477.116 
  "/>
<path clip-path="url(#clip290)" d="M1604.73 451.196 L1614.17 451.196 L1637.16 494.564 L1637.16 451.196 L1643.97 451.196 L1643.97 503.036 L1634.52 503.036 L1611.54 459.668 L1611.54 503.036 L1604.73 503.036 L1604.73 451.196 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1690.91 481.994 L1690.91 485.119 L1661.54 485.119 Q1661.95 491.716 1665.49 495.189 Q1669.07 498.626 1675.42 498.626 Q1679.1 498.626 1682.54 497.723 Q1686.01 496.82 1689.42 495.015 L1689.42 501.057 Q1685.98 502.515 1682.37 503.279 Q1678.76 504.043 1675.04 504.043 Q1665.74 504.043 1660.29 498.626 Q1654.87 493.209 1654.87 483.973 Q1654.87 474.425 1660.01 468.835 Q1665.18 463.21 1673.93 463.21 Q1681.78 463.21 1686.33 468.279 Q1690.91 473.314 1690.91 481.994 M1684.52 480.119 Q1684.45 474.876 1681.57 471.751 Q1678.72 468.626 1674 468.626 Q1668.65 468.626 1665.42 471.647 Q1662.23 474.668 1661.74 480.154 L1684.52 480.119 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1700.74 487.689 L1700.74 464.147 L1707.13 464.147 L1707.13 487.446 Q1707.13 492.966 1709.28 495.744 Q1711.43 498.487 1715.74 498.487 Q1720.91 498.487 1723.9 495.189 Q1726.92 491.89 1726.92 486.196 L1726.92 464.147 L1733.31 464.147 L1733.31 503.036 L1726.92 503.036 L1726.92 497.064 Q1724.59 500.605 1721.5 502.341 Q1718.44 504.043 1714.38 504.043 Q1707.68 504.043 1704.21 499.876 Q1700.74 495.709 1700.74 487.689 M1716.81 463.21 L1716.81 463.21 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1769 470.119 Q1767.92 469.494 1766.64 469.216 Q1765.39 468.904 1763.86 468.904 Q1758.44 468.904 1755.53 472.446 Q1752.65 475.953 1752.65 482.55 L1752.65 503.036 L1746.22 503.036 L1746.22 464.147 L1752.65 464.147 L1752.65 470.189 Q1754.66 466.647 1757.89 464.946 Q1761.12 463.21 1765.74 463.21 Q1766.4 463.21 1767.19 463.314 Q1767.99 463.383 1768.97 463.557 L1769 470.119 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1793.37 483.487 Q1785.63 483.487 1782.65 485.258 Q1779.66 487.029 1779.66 491.3 Q1779.66 494.702 1781.88 496.716 Q1784.14 498.695 1787.99 498.695 Q1793.31 498.695 1796.5 494.945 Q1799.73 491.161 1799.73 484.911 L1799.73 483.487 L1793.37 483.487 M1806.12 480.848 L1806.12 503.036 L1799.73 503.036 L1799.73 497.133 Q1797.54 500.675 1794.28 502.376 Q1791.01 504.043 1786.29 504.043 Q1780.32 504.043 1776.78 500.709 Q1773.27 497.341 1773.27 491.716 Q1773.27 485.154 1777.65 481.821 Q1782.06 478.487 1790.77 478.487 L1799.73 478.487 L1799.73 477.862 Q1799.73 473.453 1796.81 471.057 Q1793.93 468.626 1788.69 468.626 Q1785.35 468.626 1782.19 469.425 Q1779.03 470.223 1776.12 471.821 L1776.12 465.918 Q1779.62 464.564 1782.92 463.904 Q1786.22 463.21 1789.35 463.21 Q1797.78 463.21 1801.95 467.585 Q1806.12 471.959 1806.12 480.848 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1819.28 449.008 L1825.67 449.008 L1825.67 503.036 L1819.28 503.036 L1819.28 449.008 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1839.31 451.196 L1869.1 451.196 L1869.1 457.098 L1846.33 457.098 L1846.33 472.376 L1866.88 472.376 L1866.88 478.279 L1846.33 478.279 L1846.33 503.036 L1839.31 503.036 L1839.31 451.196 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1880.21 451.196 L1890.67 451.196 L1903.89 486.473 L1917.19 451.196 L1927.64 451.196 L1927.64 503.036 L1920.8 503.036 L1920.8 457.515 L1907.44 493.07 L1900.39 493.07 L1887.02 457.515 L1887.02 503.036 L1880.21 503.036 L1880.21 451.196 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M1940.77 451.196 L1947.82 451.196 L1947.82 482.689 Q1947.82 491.022 1950.84 494.702 Q1953.86 498.348 1960.63 498.348 Q1967.37 498.348 1970.39 494.702 Q1973.41 491.022 1973.41 482.689 L1973.41 451.196 L1980.46 451.196 L1980.46 483.557 Q1980.46 493.695 1975.42 498.869 Q1970.42 504.043 1960.63 504.043 Q1950.8 504.043 1945.77 498.869 Q1940.77 493.695 1940.77 483.557 L1940.77 451.196 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M2031.29 449.078 Q2026.64 457.064 2024.38 464.876 Q2022.12 472.689 2022.12 480.709 Q2022.12 488.73 2024.38 496.612 Q2026.67 504.459 2031.29 512.411 L2025.73 512.411 Q2020.53 504.251 2017.92 496.369 Q2015.35 488.487 2015.35 480.709 Q2015.35 472.966 2017.92 465.119 Q2020.49 457.272 2025.73 449.078 L2031.29 449.078 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M2044.66 451.196 L2072.19 451.196 L2072.19 457.098 L2051.08 457.098 L2051.08 469.807 Q2052.61 469.286 2054.14 469.043 Q2055.66 468.765 2057.19 468.765 Q2065.87 468.765 2070.94 473.522 Q2076.01 478.279 2076.01 486.404 Q2076.01 494.772 2070.8 499.425 Q2065.6 504.043 2056.12 504.043 Q2052.85 504.043 2049.45 503.487 Q2046.08 502.932 2042.47 501.82 L2042.47 494.772 Q2045.6 496.473 2048.93 497.307 Q2052.26 498.14 2055.98 498.14 Q2061.98 498.14 2065.49 494.98 Q2069 491.82 2069 486.404 Q2069 480.987 2065.49 477.828 Q2061.98 474.668 2055.98 474.668 Q2053.16 474.668 2050.35 475.293 Q2047.57 475.918 2044.66 477.237 L2044.66 451.196 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M2104.83 455.814 Q2099.41 455.814 2096.67 461.161 Q2093.96 466.473 2093.96 477.168 Q2093.96 487.827 2096.67 493.175 Q2099.41 498.487 2104.83 498.487 Q2110.28 498.487 2112.99 493.175 Q2115.73 487.827 2115.73 477.168 Q2115.73 466.473 2112.99 461.161 Q2110.28 455.814 2104.83 455.814 M2104.83 450.258 Q2113.55 450.258 2118.13 457.168 Q2122.75 464.043 2122.75 477.168 Q2122.75 490.258 2118.13 497.168 Q2113.55 504.043 2104.83 504.043 Q2096.12 504.043 2091.5 497.168 Q2086.91 490.258 2086.91 477.168 Q2086.91 464.043 2091.5 457.168 Q2096.12 450.258 2104.83 450.258 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M2150.07 455.814 Q2144.66 455.814 2141.91 461.161 Q2139.21 466.473 2139.21 477.168 Q2139.21 487.827 2141.91 493.175 Q2144.66 498.487 2150.07 498.487 Q2155.53 498.487 2158.23 493.175 Q2160.98 487.827 2160.98 477.168 Q2160.98 466.473 2158.23 461.161 Q2155.53 455.814 2150.07 455.814 M2150.07 450.258 Q2158.79 450.258 2163.37 457.168 Q2167.99 464.043 2167.99 477.168 Q2167.99 490.258 2163.37 497.168 Q2158.79 504.043 2150.07 504.043 Q2141.36 504.043 2136.74 497.168 Q2132.16 490.258 2132.16 477.168 Q2132.16 464.043 2136.74 457.168 Q2141.36 450.258 2150.07 450.258 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M2195.32 455.814 Q2189.9 455.814 2187.16 461.161 Q2184.45 466.473 2184.45 477.168 Q2184.45 487.827 2187.16 493.175 Q2189.9 498.487 2195.32 498.487 Q2200.77 498.487 2203.48 493.175 Q2206.22 487.827 2206.22 477.168 Q2206.22 466.473 2203.48 461.161 Q2200.77 455.814 2195.32 455.814 M2195.32 450.258 Q2204.03 450.258 2208.62 457.168 Q2213.23 464.043 2213.23 477.168 Q2213.23 490.258 2208.62 497.168 Q2204.03 504.043 2195.32 504.043 Q2186.6 504.043 2181.98 497.168 Q2177.4 490.258 2177.4 477.168 Q2177.4 464.043 2181.98 457.168 Q2186.6 450.258 2195.32 450.258 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip290)" d="M2223.65 449.078 L2229.21 449.078 Q2234.41 457.272 2236.98 465.119 Q2239.59 472.966 2239.59 480.709 Q2239.59 488.487 2236.98 496.369 Q2234.41 504.251 2229.21 512.411 L2223.65 512.411 Q2228.27 504.459 2230.53 496.612 Q2232.82 488.73 2232.82 480.709 Q2232.82 472.689 2230.53 464.876 Q2228.27 457.064 2223.65 449.078 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /></svg>




<?xml version="1.0" encoding="utf-8"?>
<svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" width="600" height="400" viewBox="0 0 2400 1600">
<defs>
  <clipPath id="clip380">
    <rect x="0" y="0" width="2400" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip380)" d="
M0 1600 L2400 1600 L2400 0 L0 0  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip381">
    <rect x="480" y="0" width="1681" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip380)" d="
M360.239 1401.58 L2352.76 1401.58 L2352.76 123.472 L360.239 123.472  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip382">
    <rect x="360" y="123" width="1994" height="1279"/>
  </clipPath>
</defs>
<polyline clip-path="url(#clip382)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  416.631,1401.58 416.631,123.472 
  "/>
<polyline clip-path="url(#clip382)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  886.564,1401.58 886.564,123.472 
  "/>
<polyline clip-path="url(#clip382)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1356.5,1401.58 1356.5,123.472 
  "/>
<polyline clip-path="url(#clip382)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1826.43,1401.58 1826.43,123.472 
  "/>
<polyline clip-path="url(#clip382)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  2296.36,1401.58 2296.36,123.472 
  "/>
<polyline clip-path="url(#clip380)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  360.239,1401.58 2352.76,1401.58 
  "/>
<polyline clip-path="url(#clip380)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  416.631,1401.58 416.631,1382.68 
  "/>
<polyline clip-path="url(#clip380)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  886.564,1401.58 886.564,1382.68 
  "/>
<polyline clip-path="url(#clip380)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1356.5,1401.58 1356.5,1382.68 
  "/>
<polyline clip-path="url(#clip380)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1826.43,1401.58 1826.43,1382.68 
  "/>
<polyline clip-path="url(#clip380)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  2296.36,1401.58 2296.36,1382.68 
  "/>
<path clip-path="url(#clip380)" d="M416.631 1438.36 Q411.214 1438.36 408.471 1443.71 Q405.763 1449.02 405.763 1459.71 Q405.763 1470.37 408.471 1475.72 Q411.214 1481.03 416.631 1481.03 Q422.082 1481.03 424.79 1475.72 Q427.533 1470.37 427.533 1459.71 Q427.533 1449.02 424.79 1443.71 Q422.082 1438.36 416.631 1438.36 M416.631 1432.8 Q425.346 1432.8 429.929 1439.71 Q434.547 1446.59 434.547 1459.71 Q434.547 1472.8 429.929 1479.71 Q425.346 1486.59 416.631 1486.59 Q407.916 1486.59 403.298 1479.71 Q398.714 1472.8 398.714 1459.71 Q398.714 1446.59 403.298 1439.71 Q407.916 1432.8 416.631 1432.8 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M872.137 1479.68 L883.595 1479.68 L883.595 1440.13 L871.13 1442.63 L871.13 1436.24 L883.526 1433.74 L890.54 1433.74 L890.54 1479.68 L901.998 1479.68 L901.998 1485.58 L872.137 1485.58 L872.137 1479.68 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1348.48 1479.68 L1372.96 1479.68 L1372.96 1485.58 L1340.04 1485.58 L1340.04 1479.68 Q1344.03 1475.55 1350.91 1468.6 Q1357.82 1461.62 1359.59 1459.61 Q1362.96 1455.82 1364.28 1453.22 Q1365.63 1450.58 1365.63 1448.05 Q1365.63 1443.91 1362.71 1441.31 Q1359.83 1438.71 1355.18 1438.71 Q1351.88 1438.71 1348.2 1439.85 Q1344.55 1441 1340.39 1443.32 L1340.39 1436.24 Q1344.62 1434.54 1348.3 1433.67 Q1351.98 1432.8 1355.04 1432.8 Q1363.09 1432.8 1367.89 1436.83 Q1372.68 1440.86 1372.68 1447.59 Q1372.68 1450.79 1371.46 1453.67 Q1370.28 1456.52 1367.12 1460.41 Q1366.25 1461.41 1361.6 1466.24 Q1356.95 1471.03 1348.48 1479.68 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1832.8 1457.63 Q1837.84 1458.7 1840.65 1462.11 Q1843.5 1465.51 1843.5 1470.51 Q1843.5 1478.18 1838.22 1482.39 Q1832.94 1486.59 1823.22 1486.59 Q1819.95 1486.59 1816.48 1485.93 Q1813.05 1485.3 1809.36 1484.02 L1809.36 1477.25 Q1812.28 1478.95 1815.75 1479.82 Q1819.23 1480.68 1823.01 1480.68 Q1829.61 1480.68 1833.05 1478.08 Q1836.52 1475.48 1836.52 1470.51 Q1836.52 1465.93 1833.29 1463.36 Q1830.09 1460.75 1824.36 1460.75 L1818.32 1460.75 L1818.32 1454.99 L1824.64 1454.99 Q1829.82 1454.99 1832.56 1452.94 Q1835.3 1450.86 1835.3 1446.97 Q1835.3 1442.98 1832.45 1440.86 Q1829.64 1438.71 1824.36 1438.71 Q1821.48 1438.71 1818.18 1439.33 Q1814.89 1439.96 1810.93 1441.27 L1810.93 1435.02 Q1814.92 1433.91 1818.39 1433.36 Q1821.9 1432.8 1824.99 1432.8 Q1832.98 1432.8 1837.63 1436.45 Q1842.28 1440.06 1842.28 1446.24 Q1842.28 1450.55 1839.82 1453.53 Q1837.35 1456.48 1832.8 1457.63 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M2300.88 1439.85 L2283.17 1467.52 L2300.88 1467.52 L2300.88 1439.85 M2299.04 1433.74 L2307.86 1433.74 L2307.86 1467.52 L2315.25 1467.52 L2315.25 1473.36 L2307.86 1473.36 L2307.86 1485.58 L2300.88 1485.58 L2300.88 1473.36 L2277.48 1473.36 L2277.48 1466.59 L2299.04 1433.74 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1300.11 1527.62 L1300.11 1538.66 L1313.27 1538.66 L1313.27 1543.62 L1300.11 1543.62 L1300.11 1564.74 Q1300.11 1569.49 1301.39 1570.85 Q1302.71 1572.2 1306.71 1572.2 L1313.27 1572.2 L1313.27 1577.55 L1306.71 1577.55 Q1299.31 1577.55 1296.5 1574.8 Q1293.69 1572.03 1293.69 1564.74 L1293.69 1543.62 L1289 1543.62 L1289 1538.66 L1293.69 1538.66 L1293.69 1527.62 L1300.11 1527.62 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1343.68 1523.52 L1358.41 1523.52 L1358.41 1528.49 L1350.07 1528.49 L1350.07 1581.96 L1358.41 1581.96 L1358.41 1586.92 L1343.68 1586.92 L1343.68 1523.52 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1396.81 1539.8 L1396.81 1545.85 Q1394.1 1544.46 1391.18 1543.76 Q1388.27 1543.07 1385.14 1543.07 Q1380.39 1543.07 1377.99 1544.53 Q1375.63 1545.99 1375.63 1548.9 Q1375.63 1551.12 1377.33 1552.41 Q1379.03 1553.66 1384.17 1554.8 L1386.36 1555.29 Q1393.16 1556.75 1396.01 1559.42 Q1398.89 1562.06 1398.89 1566.82 Q1398.89 1572.24 1394.59 1575.4 Q1390.32 1578.55 1382.82 1578.55 Q1379.69 1578.55 1376.29 1577.93 Q1372.92 1577.34 1369.17 1576.12 L1369.17 1569.53 Q1372.71 1571.37 1376.15 1572.3 Q1379.59 1573.21 1382.96 1573.21 Q1387.47 1573.21 1389.9 1571.68 Q1392.33 1570.12 1392.33 1567.3 Q1392.33 1564.7 1390.56 1563.31 Q1388.82 1561.92 1382.89 1560.64 L1380.66 1560.12 Q1374.73 1558.87 1372.09 1556.3 Q1369.45 1553.69 1369.45 1549.18 Q1369.45 1543.69 1373.34 1540.71 Q1377.23 1537.72 1384.38 1537.72 Q1387.92 1537.72 1391.05 1538.24 Q1394.17 1538.76 1396.81 1539.8 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1424 1523.52 L1424 1586.92 L1409.27 1586.92 L1409.27 1581.96 L1417.57 1581.96 L1417.57 1528.49 L1409.27 1528.49 L1409.27 1523.52 L1424 1523.52 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip382)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  360.239,1365.41 2352.76,1365.41 
  "/>
<polyline clip-path="url(#clip382)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  360.239,1063.97 2352.76,1063.97 
  "/>
<polyline clip-path="url(#clip382)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  360.239,762.526 2352.76,762.526 
  "/>
<polyline clip-path="url(#clip382)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  360.239,461.085 2352.76,461.085 
  "/>
<polyline clip-path="url(#clip382)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  360.239,159.645 2352.76,159.645 
  "/>
<polyline clip-path="url(#clip380)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  360.239,1401.58 360.239,123.472 
  "/>
<polyline clip-path="url(#clip380)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  360.239,1365.41 379.136,1365.41 
  "/>
<polyline clip-path="url(#clip380)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  360.239,1063.97 379.136,1063.97 
  "/>
<polyline clip-path="url(#clip380)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  360.239,762.526 379.136,762.526 
  "/>
<polyline clip-path="url(#clip380)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  360.239,461.085 379.136,461.085 
  "/>
<polyline clip-path="url(#clip380)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  360.239,159.645 379.136,159.645 
  "/>
<path clip-path="url(#clip380)" d="M118.58 1366.08 L163.094 1366.08 L163.094 1371.99 L118.58 1371.99 L118.58 1366.08 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M193.233 1344.11 Q187.816 1344.11 185.073 1349.45 Q182.365 1354.76 182.365 1365.46 Q182.365 1376.12 185.073 1381.47 Q187.816 1386.78 193.233 1386.78 Q198.684 1386.78 201.392 1381.47 Q204.135 1376.12 204.135 1365.46 Q204.135 1354.76 201.392 1349.45 Q198.684 1344.11 193.233 1344.11 M193.233 1338.55 Q201.948 1338.55 206.531 1345.46 Q211.149 1352.33 211.149 1365.46 Q211.149 1378.55 206.531 1385.46 Q201.948 1392.33 193.233 1392.33 Q184.517 1392.33 179.899 1385.46 Q175.316 1378.55 175.316 1365.46 Q175.316 1352.33 179.899 1345.46 Q184.517 1338.55 193.233 1338.55 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M223.475 1382.51 L230.802 1382.51 L230.802 1391.33 L223.475 1391.33 L223.475 1382.51 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M247.295 1385.42 L258.753 1385.42 L258.753 1345.88 L246.288 1348.38 L246.288 1341.99 L258.684 1339.49 L265.697 1339.49 L265.697 1385.42 L277.156 1385.42 L277.156 1391.33 L247.295 1391.33 L247.295 1385.42 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M306.322 1344.11 Q300.906 1344.11 298.163 1349.45 Q295.454 1354.76 295.454 1365.46 Q295.454 1376.12 298.163 1381.47 Q300.906 1386.78 306.322 1386.78 Q311.774 1386.78 314.482 1381.47 Q317.225 1376.12 317.225 1365.46 Q317.225 1354.76 314.482 1349.45 Q311.774 1344.11 306.322 1344.11 M306.322 1338.55 Q315.037 1338.55 319.621 1345.46 Q324.239 1352.33 324.239 1365.46 Q324.239 1378.55 319.621 1385.46 Q315.037 1392.33 306.322 1392.33 Q297.607 1392.33 292.989 1385.46 Q288.406 1378.55 288.406 1365.46 Q288.406 1352.33 292.989 1345.46 Q297.607 1338.55 306.322 1338.55 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M120.073 1064.64 L164.587 1064.64 L164.587 1070.55 L120.073 1070.55 L120.073 1064.64 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M194.726 1042.66 Q189.309 1042.66 186.566 1048.01 Q183.858 1053.32 183.858 1064.02 Q183.858 1074.68 186.566 1080.03 Q189.309 1085.34 194.726 1085.34 Q200.177 1085.34 202.885 1080.03 Q205.628 1074.68 205.628 1064.02 Q205.628 1053.32 202.885 1048.01 Q200.177 1042.66 194.726 1042.66 M194.726 1037.11 Q203.441 1037.11 208.024 1044.02 Q212.642 1050.89 212.642 1064.02 Q212.642 1077.11 208.024 1084.02 Q203.441 1090.89 194.726 1090.89 Q186.01 1090.89 181.392 1084.02 Q176.809 1077.11 176.809 1064.02 Q176.809 1050.89 181.392 1044.02 Q186.01 1037.11 194.726 1037.11 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M224.968 1081.07 L232.295 1081.07 L232.295 1089.89 L224.968 1089.89 L224.968 1081.07 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M262.572 1042.66 Q257.156 1042.66 254.413 1048.01 Q251.704 1053.32 251.704 1064.02 Q251.704 1074.68 254.413 1080.03 Q257.156 1085.34 262.572 1085.34 Q268.024 1085.34 270.732 1080.03 Q273.475 1074.68 273.475 1064.02 Q273.475 1053.32 270.732 1048.01 Q268.024 1042.66 262.572 1042.66 M262.572 1037.11 Q271.288 1037.11 275.871 1044.02 Q280.489 1050.89 280.489 1064.02 Q280.489 1077.11 275.871 1084.02 Q271.288 1090.89 262.572 1090.89 Q253.857 1090.89 249.239 1084.02 Q244.656 1077.11 244.656 1064.02 Q244.656 1050.89 249.239 1044.02 Q253.857 1037.11 262.572 1037.11 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M292.885 1038.05 L320.419 1038.05 L320.419 1043.95 L299.308 1043.95 L299.308 1056.66 Q300.836 1056.14 302.364 1055.89 Q303.892 1055.62 305.419 1055.62 Q314.1 1055.62 319.169 1060.37 Q324.239 1065.13 324.239 1073.25 Q324.239 1081.62 319.03 1086.28 Q313.822 1090.89 304.343 1090.89 Q301.079 1090.89 297.676 1090.34 Q294.308 1089.78 290.697 1088.67 L290.697 1081.62 Q293.822 1083.32 297.156 1084.16 Q300.489 1084.99 304.204 1084.99 Q310.211 1084.99 313.718 1081.83 Q317.225 1078.67 317.225 1073.25 Q317.225 1067.84 313.718 1064.68 Q310.211 1061.52 304.204 1061.52 Q301.392 1061.52 298.579 1062.14 Q295.801 1062.77 292.885 1064.09 L292.885 1038.05 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M193.233 741.224 Q187.816 741.224 185.073 746.571 Q182.365 751.884 182.365 762.578 Q182.365 773.238 185.073 778.585 Q187.816 783.897 193.233 783.897 Q198.684 783.897 201.392 778.585 Q204.135 773.238 204.135 762.578 Q204.135 751.884 201.392 746.571 Q198.684 741.224 193.233 741.224 M193.233 735.668 Q201.948 735.668 206.531 742.578 Q211.149 749.453 211.149 762.578 Q211.149 775.668 206.531 782.578 Q201.948 789.453 193.233 789.453 Q184.517 789.453 179.899 782.578 Q175.316 775.668 175.316 762.578 Q175.316 749.453 179.899 742.578 Q184.517 735.668 193.233 735.668 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M223.475 779.627 L230.802 779.627 L230.802 788.446 L223.475 788.446 L223.475 779.627 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M261.079 741.224 Q255.663 741.224 252.92 746.571 Q250.211 751.884 250.211 762.578 Q250.211 773.238 252.92 778.585 Q255.663 783.897 261.079 783.897 Q266.531 783.897 269.239 778.585 Q271.982 773.238 271.982 762.578 Q271.982 751.884 269.239 746.571 Q266.531 741.224 261.079 741.224 M261.079 735.668 Q269.795 735.668 274.378 742.578 Q278.996 749.453 278.996 762.578 Q278.996 775.668 274.378 782.578 Q269.795 789.453 261.079 789.453 Q252.364 789.453 247.746 782.578 Q243.163 775.668 243.163 762.578 Q243.163 749.453 247.746 742.578 Q252.364 735.668 261.079 735.668 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M306.322 741.224 Q300.906 741.224 298.163 746.571 Q295.454 751.884 295.454 762.578 Q295.454 773.238 298.163 778.585 Q300.906 783.897 306.322 783.897 Q311.774 783.897 314.482 778.585 Q317.225 773.238 317.225 762.578 Q317.225 751.884 314.482 746.571 Q311.774 741.224 306.322 741.224 M306.322 735.668 Q315.037 735.668 319.621 742.578 Q324.239 749.453 324.239 762.578 Q324.239 775.668 319.621 782.578 Q315.037 789.453 306.322 789.453 Q297.607 789.453 292.989 782.578 Q288.406 775.668 288.406 762.578 Q288.406 749.453 292.989 742.578 Q297.607 735.668 306.322 735.668 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M194.726 439.784 Q189.309 439.784 186.566 445.131 Q183.858 450.443 183.858 461.138 Q183.858 471.797 186.566 477.144 Q189.309 482.457 194.726 482.457 Q200.177 482.457 202.885 477.144 Q205.628 471.797 205.628 461.138 Q205.628 450.443 202.885 445.131 Q200.177 439.784 194.726 439.784 M194.726 434.228 Q203.441 434.228 208.024 441.138 Q212.642 448.013 212.642 461.138 Q212.642 474.228 208.024 481.137 Q203.441 488.012 194.726 488.012 Q186.01 488.012 181.392 481.137 Q176.809 474.228 176.809 461.138 Q176.809 448.013 181.392 441.138 Q186.01 434.228 194.726 434.228 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M224.968 478.186 L232.295 478.186 L232.295 487.005 L224.968 487.005 L224.968 478.186 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M262.572 439.784 Q257.156 439.784 254.413 445.131 Q251.704 450.443 251.704 461.138 Q251.704 471.797 254.413 477.144 Q257.156 482.457 262.572 482.457 Q268.024 482.457 270.732 477.144 Q273.475 471.797 273.475 461.138 Q273.475 450.443 270.732 445.131 Q268.024 439.784 262.572 439.784 M262.572 434.228 Q271.288 434.228 275.871 441.138 Q280.489 448.013 280.489 461.138 Q280.489 474.228 275.871 481.137 Q271.288 488.012 262.572 488.012 Q253.857 488.012 249.239 481.137 Q244.656 474.228 244.656 461.138 Q244.656 448.013 249.239 441.138 Q253.857 434.228 262.572 434.228 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M292.885 435.165 L320.419 435.165 L320.419 441.068 L299.308 441.068 L299.308 453.776 Q300.836 453.256 302.364 453.013 Q303.892 452.735 305.419 452.735 Q314.1 452.735 319.169 457.492 Q324.239 462.249 324.239 470.374 Q324.239 478.742 319.03 483.394 Q313.822 488.012 304.343 488.012 Q301.079 488.012 297.676 487.457 Q294.308 486.901 290.697 485.79 L290.697 478.742 Q293.822 480.443 297.156 481.276 Q300.489 482.11 304.204 482.11 Q310.211 482.11 313.718 478.95 Q317.225 475.79 317.225 470.374 Q317.225 464.957 313.718 461.797 Q310.211 458.638 304.204 458.638 Q301.392 458.638 298.579 459.263 Q295.801 459.888 292.885 461.207 L292.885 435.165 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M193.233 138.343 Q187.816 138.343 185.073 143.69 Q182.365 149.003 182.365 159.697 Q182.365 170.357 185.073 175.704 Q187.816 181.016 193.233 181.016 Q198.684 181.016 201.392 175.704 Q204.135 170.357 204.135 159.697 Q204.135 149.003 201.392 143.69 Q198.684 138.343 193.233 138.343 M193.233 132.787 Q201.948 132.787 206.531 139.697 Q211.149 146.572 211.149 159.697 Q211.149 172.787 206.531 179.697 Q201.948 186.572 193.233 186.572 Q184.517 186.572 179.899 179.697 Q175.316 172.787 175.316 159.697 Q175.316 146.572 179.899 139.697 Q184.517 132.787 193.233 132.787 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M223.475 176.746 L230.802 176.746 L230.802 185.565 L223.475 185.565 L223.475 176.746 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M247.295 179.662 L258.753 179.662 L258.753 140.114 L246.288 142.614 L246.288 136.225 L258.684 133.725 L265.697 133.725 L265.697 179.662 L277.156 179.662 L277.156 185.565 L247.295 185.565 L247.295 179.662 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M306.322 138.343 Q300.906 138.343 298.163 143.69 Q295.454 149.003 295.454 159.697 Q295.454 170.357 298.163 175.704 Q300.906 181.016 306.322 181.016 Q311.774 181.016 314.482 175.704 Q317.225 170.357 317.225 159.697 Q317.225 149.003 314.482 143.69 Q311.774 138.343 306.322 138.343 M306.322 132.787 Q315.037 132.787 319.621 139.697 Q324.239 146.572 324.239 159.697 Q324.239 172.787 319.621 179.697 Q315.037 186.572 306.322 186.572 Q297.607 186.572 292.989 179.697 Q288.406 172.787 288.406 159.697 Q288.406 146.572 292.989 139.697 Q297.607 132.787 306.322 132.787 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M34.0422 1042.44 L13.0007 1042.44 L13.0007 1036.05 L67.0282 1036.05 L67.0282 1042.44 L61.1949 1042.44 Q64.6671 1044.45 66.3684 1047.54 Q68.0351 1050.6 68.0351 1054.9 Q68.0351 1061.95 62.4101 1066.4 Q56.7852 1070.81 47.6185 1070.81 Q38.4519 1070.81 32.827 1066.4 Q27.202 1061.95 27.202 1054.9 Q27.202 1050.6 28.9034 1047.54 Q30.57 1044.45 34.0422 1042.44 M47.6185 1064.21 Q54.6671 1064.21 58.6949 1061.33 Q62.6879 1058.41 62.6879 1053.34 Q62.6879 1048.27 58.6949 1045.35 Q54.6671 1042.44 47.6185 1042.44 Q40.57 1042.44 36.5769 1045.35 Q32.5492 1048.27 32.5492 1053.34 Q32.5492 1058.41 36.5769 1061.33 Q40.57 1064.21 47.6185 1064.21 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M28.1395 1022.89 L28.1395 1016.5 L67.0282 1016.5 L67.0282 1022.89 L28.1395 1022.89 M13.0007 1022.89 L13.0007 1016.5 L21.0909 1016.5 L21.0909 1022.89 L13.0007 1022.89 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M29.2853 978.341 L35.3269 978.341 Q33.9381 981.049 33.2436 983.966 Q32.5492 986.882 32.5492 990.007 Q32.5492 994.764 34.0075 997.16 Q35.4658 999.521 38.3825 999.521 Q40.6047 999.521 41.8894 997.82 Q43.1394 996.118 44.2852 990.98 L44.7713 988.792 Q46.2297 981.987 48.9033 979.139 Q51.5421 976.257 56.299 976.257 Q61.7157 976.257 64.8754 980.563 Q68.0351 984.834 68.0351 992.334 Q68.0351 995.459 67.4101 998.862 Q66.8198 1002.23 65.6046 1005.98 L59.0074 1005.98 Q60.8476 1002.44 61.7851 999 Q62.6879 995.563 62.6879 992.195 Q62.6879 987.681 61.1601 985.25 Q59.5976 982.82 56.7852 982.82 Q54.181 982.82 52.7921 984.591 Q51.4032 986.327 50.1185 992.264 L49.5977 994.487 Q48.3477 1000.42 45.7783 1003.06 Q43.1741 1005.7 38.6603 1005.7 Q33.1742 1005.7 30.1881 1001.81 Q27.202 997.924 27.202 990.771 Q27.202 987.23 27.7228 984.105 Q28.2436 980.98 29.2853 978.341 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M61.1949 959.903 L81.8197 959.903 L81.8197 966.327 L28.1395 966.327 L28.1395 959.903 L34.0422 959.903 Q30.57 957.89 28.9034 954.834 Q27.202 951.744 27.202 947.473 Q27.202 940.39 32.827 935.98 Q38.4519 931.535 47.6185 931.535 Q56.7852 931.535 62.4101 935.98 Q68.0351 940.39 68.0351 947.473 Q68.0351 951.744 66.3684 954.834 Q64.6671 957.89 61.1949 959.903 M47.6185 938.167 Q40.57 938.167 36.5769 941.084 Q32.5492 943.966 32.5492 949.035 Q32.5492 954.105 36.5769 957.021 Q40.57 959.903 47.6185 959.903 Q54.6671 959.903 58.6949 957.021 Q62.6879 954.105 62.6879 949.035 Q62.6879 943.966 58.6949 941.084 Q54.6671 938.167 47.6185 938.167 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M13.0007 920.945 L13.0007 914.556 L67.0282 914.556 L67.0282 920.945 L13.0007 920.945 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M47.4797 883.515 Q47.4797 891.258 49.2505 894.244 Q51.0213 897.23 55.2921 897.23 Q58.6949 897.23 60.7087 895.008 Q62.6879 892.751 62.6879 888.897 Q62.6879 883.584 58.9379 880.39 Q55.1532 877.161 48.9033 877.161 L47.4797 877.161 L47.4797 883.515 M44.8408 870.772 L67.0282 870.772 L67.0282 877.161 L61.1254 877.161 Q64.6671 879.348 66.3684 882.612 Q68.0351 885.876 68.0351 890.598 Q68.0351 896.57 64.7018 900.112 Q61.3337 903.619 55.7088 903.619 Q49.1463 903.619 45.813 899.244 Q42.4797 894.834 42.4797 886.119 L42.4797 877.161 L41.8547 877.161 Q37.445 877.161 35.0492 880.077 Q32.6186 882.959 32.6186 888.202 Q32.6186 891.536 33.4172 894.695 Q34.2158 897.855 35.813 900.772 L29.9103 900.772 Q28.5561 897.265 27.8964 893.966 Q27.202 890.668 27.202 887.543 Q27.202 879.105 31.577 874.939 Q35.9519 870.772 44.8408 870.772 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M29.6325 829.626 L35.6047 829.626 Q34.1117 832.335 33.3825 835.078 Q32.6186 837.786 32.6186 840.564 Q32.6186 846.779 36.5769 850.216 Q40.5005 853.654 47.6185 853.654 Q54.7366 853.654 58.6949 850.216 Q62.6185 846.779 62.6185 840.564 Q62.6185 837.786 61.8893 835.078 Q61.1254 832.335 59.6324 829.626 L65.5351 829.626 Q66.7851 832.3 67.4101 835.182 Q68.0351 838.029 68.0351 841.258 Q68.0351 850.043 62.5143 855.216 Q56.9935 860.39 47.6185 860.39 Q38.1047 860.39 32.6533 855.182 Q27.202 849.939 27.202 840.842 Q27.202 837.89 27.827 835.078 Q28.4173 832.265 29.6325 829.626 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M45.9866 785.252 L49.1116 785.252 L49.1116 814.626 Q55.7088 814.21 59.181 810.668 Q62.6185 807.092 62.6185 800.738 Q62.6185 797.057 61.7157 793.62 Q60.8129 790.147 59.0074 786.745 L65.049 786.745 Q66.5073 790.182 67.2712 793.793 Q68.0351 797.404 68.0351 801.12 Q68.0351 810.425 62.6185 815.876 Q57.2018 821.293 47.9658 821.293 Q38.4172 821.293 32.827 816.154 Q27.202 810.981 27.202 802.231 Q27.202 794.383 32.2714 789.835 Q37.3061 785.252 45.9866 785.252 M44.1116 791.64 Q38.8686 791.71 35.7436 794.592 Q32.6186 797.439 32.6186 802.161 Q32.6186 807.508 35.6394 810.738 Q38.6603 813.932 44.1463 814.418 L44.1116 791.64 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M35.6047 744.488 Q31.2992 742.092 29.2506 738.759 Q27.202 735.425 27.202 730.912 Q27.202 724.835 31.4728 721.537 Q35.7089 718.238 43.5561 718.238 L67.0282 718.238 L67.0282 724.662 L43.7644 724.662 Q38.1741 724.662 35.4658 726.641 Q32.7575 728.62 32.7575 732.682 Q32.7575 737.648 36.0561 740.53 Q39.3547 743.412 45.0491 743.412 L67.0282 743.412 L67.0282 749.835 L43.7644 749.835 Q38.1394 749.835 35.4658 751.814 Q32.7575 753.793 32.7575 757.925 Q32.7575 762.821 36.0908 765.703 Q39.3894 768.585 45.0491 768.585 L67.0282 768.585 L67.0282 775.009 L28.1395 775.009 L28.1395 768.585 L34.1811 768.585 Q30.6047 766.397 28.9034 763.342 Q27.202 760.286 27.202 756.085 Q27.202 751.849 29.3547 748.898 Q31.5075 745.911 35.6047 744.488 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M45.9866 672.231 L49.1116 672.231 L49.1116 701.606 Q55.7088 701.19 59.181 697.648 Q62.6185 694.071 62.6185 687.717 Q62.6185 684.037 61.7157 680.599 Q60.8129 677.127 59.0074 673.724 L65.049 673.724 Q66.5073 677.162 67.2712 680.773 Q68.0351 684.384 68.0351 688.099 Q68.0351 697.405 62.6185 702.856 Q57.2018 708.273 47.9658 708.273 Q38.4172 708.273 32.827 703.134 Q27.202 697.96 27.202 689.21 Q27.202 681.363 32.2714 676.815 Q37.3061 672.231 45.9866 672.231 M44.1116 678.62 Q38.8686 678.69 35.7436 681.572 Q32.6186 684.419 32.6186 689.141 Q32.6186 694.488 35.6394 697.717 Q38.6603 700.912 44.1463 701.398 L44.1116 678.62 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M43.5561 629.419 L67.0282 629.419 L67.0282 635.808 L43.7644 635.808 Q38.2436 635.808 35.5005 637.961 Q32.7575 640.113 32.7575 644.419 Q32.7575 649.593 36.0561 652.579 Q39.3547 655.565 45.0491 655.565 L67.0282 655.565 L67.0282 661.988 L28.1395 661.988 L28.1395 655.565 L34.1811 655.565 Q30.6742 653.273 28.9381 650.183 Q27.202 647.058 27.202 642.995 Q27.202 636.294 31.3686 632.857 Q35.5005 629.419 43.5561 629.419 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M17.0979 610.357 L28.1395 610.357 L28.1395 597.197 L33.1047 597.197 L33.1047 610.357 L54.2157 610.357 Q58.9726 610.357 60.3268 609.072 Q61.681 607.753 61.681 603.759 L61.681 597.197 L67.0282 597.197 L67.0282 603.759 Q67.0282 611.155 64.2851 613.968 Q61.5074 616.78 54.2157 616.78 L33.1047 616.78 L33.1047 621.468 L28.1395 621.468 L28.1395 616.78 L17.0979 616.78 L17.0979 610.357 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M13.0007 566.781 L13.0007 552.058 L17.9659 552.058 L17.9659 560.392 L71.4379 560.392 L71.4379 552.058 L76.4031 552.058 L76.4031 566.781 L13.0007 566.781 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M35.6047 508.17 Q31.2992 505.774 29.2506 502.441 Q27.202 499.107 27.202 494.593 Q27.202 488.517 31.4728 485.218 Q35.7089 481.92 43.5561 481.92 L67.0282 481.92 L67.0282 488.343 L43.7644 488.343 Q38.1741 488.343 35.4658 490.323 Q32.7575 492.302 32.7575 496.364 Q32.7575 501.329 36.0561 504.211 Q39.3547 507.093 45.0491 507.093 L67.0282 507.093 L67.0282 513.517 L43.7644 513.517 Q38.1394 513.517 35.4658 515.496 Q32.7575 517.475 32.7575 521.607 Q32.7575 526.503 36.0908 529.385 Q39.3894 532.267 45.0491 532.267 L67.0282 532.267 L67.0282 538.69 L28.1395 538.69 L28.1395 532.267 L34.1811 532.267 Q30.6047 530.079 28.9034 527.024 Q27.202 523.968 27.202 519.767 Q27.202 515.531 29.3547 512.579 Q31.5075 509.593 35.6047 508.17 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M13.0007 454.246 L76.4031 454.246 L76.4031 468.969 L71.4379 468.969 L71.4379 460.67 L17.9659 460.67 L17.9659 468.969 L13.0007 468.969 L13.0007 454.246 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M905.814 18.8205 L905.814 65.8515 L915.698 65.8515 Q928.215 65.8515 934.008 60.1802 Q939.842 54.509 939.842 42.2752 Q939.842 30.1225 934.008 24.4918 Q928.215 18.8205 915.698 18.8205 L905.814 18.8205 M897.631 12.096 L914.442 12.096 Q932.023 12.096 940.247 19.4281 Q948.47 26.7198 948.47 42.2752 Q948.47 57.9117 940.206 65.2439 Q931.942 72.576 914.442 72.576 L897.631 72.576 L897.631 12.096 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M961.19 27.2059 L968.644 27.2059 L968.644 72.576 L961.19 72.576 L961.19 27.2059 M961.19 9.54393 L968.644 9.54393 L968.644 18.9825 L961.19 18.9825 L961.19 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1013.16 28.5427 L1013.16 35.5912 Q1010 33.9709 1006.6 33.1607 Q1003.2 32.3505 999.552 32.3505 Q994.002 32.3505 991.207 34.0519 Q988.452 35.7533 988.452 39.156 Q988.452 41.7486 990.437 43.2475 Q992.422 44.7058 998.418 46.0426 L1000.97 46.6097 Q1008.91 48.3111 1012.23 51.4303 Q1015.59 54.509 1015.59 60.0587 Q1015.59 66.3781 1010.57 70.0644 Q1005.59 73.7508 996.838 73.7508 Q993.192 73.7508 989.222 73.0216 Q985.293 72.3329 980.918 70.9151 L980.918 63.2184 Q985.05 65.3654 989.06 66.4591 Q993.07 67.5124 997 67.5124 Q1002.27 67.5124 1005.1 65.73 Q1007.94 63.9071 1007.94 60.6258 Q1007.94 57.5877 1005.87 55.9673 Q1003.85 54.3469 996.919 52.8481 L994.326 52.2405 Q987.399 50.7821 984.321 47.7845 Q981.242 44.7463 981.242 39.4801 Q981.242 33.0797 985.779 29.5959 Q990.316 26.1121 998.661 26.1121 Q1002.79 26.1121 1006.44 26.7198 Q1010.08 27.3274 1013.16 28.5427 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1034.67 65.7705 L1034.67 89.8329 L1027.18 89.8329 L1027.18 27.2059 L1034.67 27.2059 L1034.67 34.0924 Q1037.02 30.0415 1040.59 28.0971 Q1044.19 26.1121 1049.18 26.1121 Q1057.44 26.1121 1062.58 32.6746 Q1067.77 39.2371 1067.77 49.9314 Q1067.77 60.6258 1062.58 67.1883 Q1057.44 73.7508 1049.18 73.7508 Q1044.19 73.7508 1040.59 71.8063 Q1037.02 69.8214 1034.67 65.7705 M1060.03 49.9314 Q1060.03 41.7081 1056.63 37.0496 Q1053.27 32.3505 1047.35 32.3505 Q1041.44 32.3505 1038.04 37.0496 Q1034.67 41.7081 1034.67 49.9314 Q1034.67 58.1548 1038.04 62.8538 Q1041.44 67.5124 1047.35 67.5124 Q1053.27 67.5124 1056.63 62.8538 Q1060.03 58.1548 1060.03 49.9314 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1080.12 9.54393 L1087.58 9.54393 L1087.58 72.576 L1080.12 72.576 L1080.12 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1123.79 49.7694 Q1114.76 49.7694 1111.28 51.8354 Q1107.79 53.9013 1107.79 58.8839 Q1107.79 62.8538 1110.38 65.2034 Q1113.02 67.5124 1117.51 67.5124 Q1123.71 67.5124 1127.44 63.1374 Q1131.21 58.7219 1131.21 51.4303 L1131.21 49.7694 L1123.79 49.7694 M1138.66 46.6907 L1138.66 72.576 L1131.21 72.576 L1131.21 65.6895 Q1128.65 69.8214 1124.85 71.8063 Q1121.04 73.7508 1115.53 73.7508 Q1108.56 73.7508 1104.43 69.8619 Q1100.34 65.9325 1100.34 59.3701 Q1100.34 51.7138 1105.44 47.825 Q1110.59 43.9361 1120.75 43.9361 L1131.21 43.9361 L1131.21 43.2069 Q1131.21 38.0623 1127.8 35.2672 Q1124.44 32.4315 1118.32 32.4315 Q1114.44 32.4315 1110.75 33.3632 Q1107.06 34.295 1103.66 36.1584 L1103.66 29.2718 Q1107.75 27.692 1111.6 26.9223 Q1115.45 26.1121 1119.09 26.1121 Q1128.94 26.1121 1133.8 31.2163 Q1138.66 36.3204 1138.66 46.6907 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1186.66 28.9478 L1186.66 35.9153 Q1183.5 34.1734 1180.3 33.3227 Q1177.14 32.4315 1173.9 32.4315 Q1166.65 32.4315 1162.64 37.0496 Q1158.63 41.6271 1158.63 49.9314 Q1158.63 58.2358 1162.64 62.8538 Q1166.65 67.4314 1173.9 67.4314 Q1177.14 67.4314 1180.3 66.5807 Q1183.5 65.6895 1186.66 63.9476 L1186.66 70.8341 Q1183.54 72.2924 1180.18 73.0216 Q1176.86 73.7508 1173.09 73.7508 Q1162.84 73.7508 1156.81 67.3098 Q1150.77 60.8689 1150.77 49.9314 Q1150.77 38.832 1156.85 32.472 Q1162.97 26.1121 1173.58 26.1121 Q1177.02 26.1121 1180.3 26.8413 Q1183.58 27.5299 1186.66 28.9478 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1238.43 48.0275 L1238.43 51.6733 L1204.16 51.6733 Q1204.65 59.3701 1208.78 63.421 Q1212.95 67.4314 1220.37 67.4314 Q1224.66 67.4314 1228.67 66.3781 Q1232.72 65.3249 1236.69 63.2184 L1236.69 70.267 Q1232.68 71.9684 1228.47 72.8596 Q1224.26 73.7508 1219.92 73.7508 Q1209.06 73.7508 1202.7 67.4314 Q1196.39 61.1119 1196.39 50.3365 Q1196.39 39.1965 1202.38 32.6746 Q1208.42 26.1121 1218.62 26.1121 Q1227.78 26.1121 1233.09 32.0264 Q1238.43 37.9003 1238.43 48.0275 M1230.98 45.84 Q1230.9 39.7232 1227.54 36.0774 Q1224.22 32.4315 1218.71 32.4315 Q1212.47 32.4315 1208.7 35.9558 Q1204.97 39.4801 1204.41 45.8805 L1230.98 45.84 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1285.99 35.9153 Q1288.79 30.8922 1292.68 28.5022 Q1296.56 26.1121 1301.83 26.1121 Q1308.92 26.1121 1312.77 31.0947 Q1316.62 36.0368 1316.62 45.1919 L1316.62 72.576 L1309.12 72.576 L1309.12 45.4349 Q1309.12 38.913 1306.81 35.7533 Q1304.5 32.5936 1299.76 32.5936 Q1293.97 32.5936 1290.61 36.4419 Q1287.25 40.2903 1287.25 46.9338 L1287.25 72.576 L1279.75 72.576 L1279.75 45.4349 Q1279.75 38.8725 1277.44 35.7533 Q1275.13 32.5936 1270.31 32.5936 Q1264.6 32.5936 1261.24 36.4824 Q1257.88 40.3308 1257.88 46.9338 L1257.88 72.576 L1250.38 72.576 L1250.38 27.2059 L1257.88 27.2059 L1257.88 34.2544 Q1260.43 30.082 1263.99 28.0971 Q1267.56 26.1121 1272.46 26.1121 Q1277.4 26.1121 1280.85 28.6237 Q1284.33 31.1352 1285.99 35.9153 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1370.29 48.0275 L1370.29 51.6733 L1336.02 51.6733 Q1336.51 59.3701 1340.64 63.421 Q1344.81 67.4314 1352.22 67.4314 Q1356.52 67.4314 1360.53 66.3781 Q1364.58 65.3249 1368.55 63.2184 L1368.55 70.267 Q1364.54 71.9684 1360.33 72.8596 Q1356.11 73.7508 1351.78 73.7508 Q1340.92 73.7508 1334.56 67.4314 Q1328.24 61.1119 1328.24 50.3365 Q1328.24 39.1965 1334.24 32.6746 Q1340.27 26.1121 1350.48 26.1121 Q1359.64 26.1121 1364.94 32.0264 Q1370.29 37.9003 1370.29 48.0275 M1362.84 45.84 Q1362.76 39.7232 1359.39 36.0774 Q1356.07 32.4315 1350.56 32.4315 Q1344.32 32.4315 1340.56 35.9558 Q1336.83 39.4801 1336.26 45.8805 L1362.84 45.84 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1420.24 45.1919 L1420.24 72.576 L1412.78 72.576 L1412.78 45.4349 Q1412.78 38.994 1410.27 35.7938 Q1407.76 32.5936 1402.74 32.5936 Q1396.7 32.5936 1393.22 36.4419 Q1389.74 40.2903 1389.74 46.9338 L1389.74 72.576 L1382.24 72.576 L1382.24 27.2059 L1389.74 27.2059 L1389.74 34.2544 Q1392.41 30.163 1396.01 28.1376 Q1399.66 26.1121 1404.4 26.1121 Q1412.22 26.1121 1416.23 30.9732 Q1420.24 35.7938 1420.24 45.1919 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1442.48 14.324 L1442.48 27.2059 L1457.83 27.2059 L1457.83 32.9987 L1442.48 32.9987 L1442.48 57.6282 Q1442.48 63.1779 1443.98 64.7578 Q1445.52 66.3376 1450.17 66.3376 L1457.83 66.3376 L1457.83 72.576 L1450.17 72.576 Q1441.55 72.576 1438.26 69.3758 Q1434.98 66.1351 1434.98 57.6282 L1434.98 32.9987 L1429.51 32.9987 L1429.51 27.2059 L1434.98 27.2059 L1434.98 14.324 L1442.48 14.324 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1529.33 35.9153 Q1532.12 30.8922 1536.01 28.5022 Q1539.9 26.1121 1545.17 26.1121 Q1552.26 26.1121 1556.11 31.0947 Q1559.95 36.0368 1559.95 45.1919 L1559.95 72.576 L1552.46 72.576 L1552.46 45.4349 Q1552.46 38.913 1550.15 35.7533 Q1547.84 32.5936 1543.1 32.5936 Q1537.31 32.5936 1533.95 36.4419 Q1530.58 40.2903 1530.58 46.9338 L1530.58 72.576 L1523.09 72.576 L1523.09 45.4349 Q1523.09 38.8725 1520.78 35.7533 Q1518.47 32.5936 1513.65 32.5936 Q1507.94 32.5936 1504.58 36.4824 Q1501.22 40.3308 1501.22 46.9338 L1501.22 72.576 L1493.72 72.576 L1493.72 27.2059 L1501.22 27.2059 L1501.22 34.2544 Q1503.77 30.082 1507.33 28.0971 Q1510.9 26.1121 1515.8 26.1121 Q1520.74 26.1121 1524.18 28.6237 Q1527.67 31.1352 1529.33 35.9153 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1592.4 32.4315 Q1586.41 32.4315 1582.92 37.1306 Q1579.44 41.7891 1579.44 49.9314 Q1579.44 58.0738 1582.88 62.7728 Q1586.37 67.4314 1592.4 67.4314 Q1598.36 67.4314 1601.84 62.7323 Q1605.32 58.0333 1605.32 49.9314 Q1605.32 41.8701 1601.84 37.1711 Q1598.36 32.4315 1592.4 32.4315 M1592.4 26.1121 Q1602.12 26.1121 1607.67 32.4315 Q1613.22 38.7509 1613.22 49.9314 Q1613.22 61.0714 1607.67 67.4314 Q1602.12 73.7508 1592.4 73.7508 Q1582.64 73.7508 1577.09 67.4314 Q1571.58 61.0714 1571.58 49.9314 Q1571.58 38.7509 1577.09 32.4315 Q1582.64 26.1121 1592.4 26.1121 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1655.43 34.0924 L1655.43 9.54393 L1662.89 9.54393 L1662.89 72.576 L1655.43 72.576 L1655.43 65.7705 Q1653.08 69.8214 1649.48 71.8063 Q1645.91 73.7508 1640.89 73.7508 Q1632.67 73.7508 1627.48 67.1883 Q1622.34 60.6258 1622.34 49.9314 Q1622.34 39.2371 1627.48 32.6746 Q1632.67 26.1121 1640.89 26.1121 Q1645.91 26.1121 1649.48 28.0971 Q1653.08 30.0415 1655.43 34.0924 M1630.03 49.9314 Q1630.03 58.1548 1633.4 62.8538 Q1636.8 67.5124 1642.71 67.5124 Q1648.63 67.5124 1652.03 62.8538 Q1655.43 58.1548 1655.43 49.9314 Q1655.43 41.7081 1652.03 37.0496 Q1648.63 32.3505 1642.71 32.3505 Q1636.8 32.3505 1633.4 37.0496 Q1630.03 41.7081 1630.03 49.9314 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1717.05 48.0275 L1717.05 51.6733 L1682.78 51.6733 Q1683.26 59.3701 1687.4 63.421 Q1691.57 67.4314 1698.98 67.4314 Q1703.27 67.4314 1707.29 66.3781 Q1711.34 65.3249 1715.31 63.2184 L1715.31 70.267 Q1711.3 71.9684 1707.08 72.8596 Q1702.87 73.7508 1698.54 73.7508 Q1687.68 73.7508 1681.32 67.4314 Q1675 61.1119 1675 50.3365 Q1675 39.1965 1681 32.6746 Q1687.03 26.1121 1697.24 26.1121 Q1706.39 26.1121 1711.7 32.0264 Q1717.05 37.9003 1717.05 48.0275 M1709.59 45.84 Q1709.51 39.7232 1706.15 36.0774 Q1702.83 32.4315 1697.32 32.4315 Q1691.08 32.4315 1687.31 35.9558 Q1683.59 39.4801 1683.02 45.8805 L1709.59 45.84 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1729.28 9.54393 L1736.74 9.54393 L1736.74 72.576 L1729.28 72.576 L1729.28 9.54393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1786.8 65.6895 L1815.36 65.6895 L1815.36 72.576 L1776.96 72.576 L1776.96 65.6895 Q1781.62 60.8689 1789.64 52.7671 Q1797.7 44.6248 1799.77 42.2752 Q1803.7 37.8598 1805.24 34.8216 Q1806.82 31.7429 1806.82 28.7857 Q1806.82 23.9651 1803.41 20.927 Q1800.05 17.8888 1794.62 17.8888 Q1790.77 17.8888 1786.48 19.2256 Q1782.23 20.5624 1777.37 23.2765 L1777.37 15.0127 Q1782.31 13.0277 1786.6 12.015 Q1790.9 11.0023 1794.46 11.0023 Q1803.86 11.0023 1809.45 15.7013 Q1815.04 20.4004 1815.04 28.2591 Q1815.04 31.9859 1813.62 35.3482 Q1812.24 38.6699 1808.56 43.2069 Q1807.55 44.3817 1802.12 50.0125 Q1796.69 55.6027 1786.8 65.6895 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip382)" style="stroke:#009af9; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" points="
  416.631,159.645 2296.36,159.645 
  "/>
<polyline clip-path="url(#clip382)" style="stroke:#e26f46; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" points="
  416.631,762.526 2296.36,762.526 
  "/>
<polyline clip-path="url(#clip382)" style="stroke:#008000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" stroke-dasharray="4, 8" points="
  416.631,1365.41 421.33,1365.41 426.029,1365.41 430.729,1365.41 435.428,1365.41 440.127,1365.41 444.827,1365.41 449.526,1365.41 454.225,1365.41 458.925,1365.41 
  463.624,1365.41 468.323,1365.41 473.023,1365.41 477.722,1365.41 482.421,1365.41 487.121,1365.41 491.82,1365.41 496.519,1365.41 501.219,1365.41 505.918,1365.41 
  510.617,1365.41 515.317,1365.41 520.016,1365.41 524.715,1365.41 529.415,1365.41 534.114,1365.41 538.813,1365.41 543.513,1365.41 548.212,1365.41 552.911,1365.41 
  557.611,1365.41 562.31,1365.41 567.009,1365.41 571.709,1365.41 576.408,1365.41 581.107,1365.41 585.807,1365.41 590.506,1365.41 595.205,1365.41 599.905,1365.41 
  604.604,1365.41 609.303,1365.41 614.003,1365.41 618.702,1365.41 623.401,1365.41 628.101,1365.41 632.8,1365.41 637.499,1365.41 642.199,1365.41 646.898,1365.41 
  651.597,1365.41 656.297,1365.41 660.996,1365.41 665.695,1365.41 670.395,1365.41 675.094,1365.41 679.793,1365.41 684.493,1365.41 689.192,1365.41 693.891,1365.41 
  698.591,1365.41 703.29,1365.41 707.989,1365.41 712.689,1365.41 717.388,1365.41 722.087,1365.41 726.787,1365.41 731.486,1365.41 736.185,1365.41 740.885,1365.41 
  745.584,1365.41 750.283,1365.41 754.983,1365.41 759.682,1365.41 764.381,1365.41 769.081,1365.41 773.78,1365.41 778.479,1365.41 783.179,1365.41 787.878,1365.41 
  792.577,1365.41 797.277,1365.41 801.976,1365.41 806.675,1365.41 811.375,1365.41 816.074,1365.41 820.773,1365.41 825.473,1365.41 830.172,1365.41 834.871,1365.41 
  839.571,1365.41 844.27,1365.41 848.969,1365.41 853.669,1365.41 858.368,1365.41 863.067,1365.41 867.767,1365.41 872.466,1365.41 877.165,1365.41 881.865,1365.41 
  886.564,1365.41 891.263,1365.41 895.963,1365.41 900.662,1365.41 905.361,1365.41 910.061,1365.41 914.76,1365.41 919.459,1365.41 924.159,1365.41 928.858,1365.41 
  933.557,1365.41 938.257,1365.41 942.956,1365.41 947.655,1365.41 952.355,1365.41 957.054,1365.41 961.753,1365.41 966.453,1365.41 971.152,1365.41 975.851,1365.41 
  980.551,1365.41 985.25,1365.41 989.949,1365.41 994.649,1365.41 999.348,1365.41 1004.05,1365.41 1008.75,1365.41 1013.45,1365.41 1018.15,1365.41 1022.84,1365.41 
  1027.54,1365.41 1032.24,1365.41 1036.94,1365.41 1041.64,1365.41 1046.34,1365.41 1051.04,1365.41 1055.74,1365.41 1060.44,1365.41 1065.14,1365.41 1069.84,1365.41 
  1074.54,1365.41 1079.24,1365.41 1083.94,1365.41 1088.64,1365.41 1093.33,1365.41 1098.03,1365.41 1102.73,1365.41 1107.43,1365.41 1112.13,1365.41 1116.83,1365.41 
  1121.53,1365.41 1126.23,1365.41 1130.93,1365.41 1135.63,1365.41 1140.33,1365.41 1145.03,1365.41 1149.73,1365.41 1154.43,1365.41 1159.13,1365.41 1163.82,1365.41 
  1168.52,1365.41 1173.22,1365.41 1177.92,1365.41 1182.62,1365.41 1187.32,1365.41 1192.02,1365.41 1196.72,1365.41 1201.42,1365.41 1206.12,1365.41 1210.82,1365.41 
  1215.52,1365.41 1220.22,1365.41 1224.92,1365.41 1229.62,1365.41 1234.31,1365.41 1239.01,1365.41 1243.71,1365.41 1248.41,1365.41 1253.11,1365.41 1257.81,1365.41 
  1262.51,1365.41 1267.21,1365.41 1271.91,1365.41 1276.61,1365.41 1281.31,1365.41 1286.01,1365.41 1290.71,1365.41 1295.41,1365.41 1300.11,1365.41 1304.8,1365.41 
  1309.5,1365.41 1314.2,1365.41 1318.9,1365.41 1323.6,1365.41 1328.3,1365.41 1333,1365.41 1337.7,1365.41 1342.4,1365.41 1347.1,1365.41 1351.8,1365.41 
  1356.5,1365.41 1361.2,1365.41 1365.9,1365.41 1370.6,1365.41 1375.29,1365.41 1379.99,1365.41 1384.69,1365.41 1389.39,1365.41 1394.09,1365.41 1398.79,1365.41 
  1403.49,1365.41 1408.19,1365.41 1412.89,1365.41 1417.59,1365.41 1422.29,1365.41 1426.99,1365.41 1431.69,1365.41 1436.39,1365.41 1441.09,1365.41 1445.78,1365.41 
  1450.48,1365.41 1455.18,1365.41 1459.88,1365.41 1464.58,1365.41 1469.28,1365.41 1473.98,1365.41 1478.68,1365.41 1483.38,1365.41 1488.08,1365.41 1492.78,1365.41 
  1497.48,1365.41 1502.18,1365.41 1506.88,1365.41 1511.58,1365.41 1516.27,1365.41 1520.97,1365.41 1525.67,1365.41 1530.37,1365.41 1535.07,1365.41 1539.77,1365.41 
  1544.47,1365.41 1549.17,1365.41 1553.87,1365.41 1558.57,1365.41 1563.27,1365.41 1567.97,1365.41 1572.67,1365.41 1577.37,1365.41 1582.07,1365.41 1586.76,1365.41 
  1591.46,1365.41 1596.16,1365.41 1600.86,1365.41 1605.56,1365.41 1610.26,1365.41 1614.96,1365.41 1619.66,1365.41 1624.36,1365.41 1629.06,1365.41 1633.76,1365.41 
  1638.46,1365.41 1643.16,1365.41 1647.86,1365.41 1652.56,1365.41 1657.25,1365.41 1661.95,1365.41 1666.65,1365.41 1671.35,1365.41 1676.05,1365.41 1680.75,1365.41 
  1685.45,1365.41 1690.15,1365.41 1694.85,1365.41 1699.55,1365.41 1704.25,1365.41 1708.95,1365.41 1713.65,1365.41 1718.35,1365.41 1723.05,1365.41 1727.74,1365.41 
  1732.44,1365.41 1737.14,1365.41 1741.84,1365.41 1746.54,1365.41 1751.24,1365.41 1755.94,1365.41 1760.64,1365.41 1765.34,1365.41 1770.04,1365.41 1774.74,1365.41 
  1779.44,1365.41 1784.14,1365.41 1788.84,1365.41 1793.54,1365.41 1798.23,1365.41 1802.93,1365.41 1807.63,1365.41 1812.33,1365.41 1817.03,1365.41 1821.73,1365.41 
  1826.43,1365.41 1831.13,1365.41 1835.83,1365.41 1840.53,1365.41 1845.23,1365.41 1849.93,1365.41 1854.63,1365.41 1859.33,1365.41 1864.03,1365.41 1868.72,1365.41 
  1873.42,1365.41 1878.12,1365.41 1882.82,1365.41 1887.52,1365.41 1892.22,1365.41 1896.92,1365.41 1901.62,1365.41 1906.32,1365.41 1911.02,1365.41 1915.72,1365.41 
  1920.42,1365.41 1925.12,1365.41 1929.82,1365.41 1934.52,1365.41 1939.21,1365.41 1943.91,1365.41 1948.61,1365.41 1953.31,1365.41 1958.01,1365.41 1962.71,1365.41 
  1967.41,1365.41 1972.11,1365.41 1976.81,1365.41 1981.51,1365.41 1986.21,1365.41 1990.91,1365.41 1995.61,1365.41 2000.31,1365.41 2005.01,1365.41 2009.7,1365.41 
  2014.4,1365.41 2019.1,1365.41 2023.8,1365.41 2028.5,1365.41 2033.2,1365.41 2037.9,1365.41 2042.6,1365.41 2047.3,1365.41 2052,1365.41 2056.7,1365.41 
  2061.4,1365.41 2066.1,1365.41 2070.8,1365.41 2075.5,1365.41 2080.19,1365.41 2084.89,1365.41 2089.59,1365.41 2094.29,1365.41 2098.99,1365.41 2103.69,1365.41 
  2108.39,1365.41 2113.09,1365.41 2117.79,1365.41 2122.49,1365.41 2127.19,1365.41 2131.89,1365.41 2136.59,1365.41 2141.29,1365.41 2145.99,1365.41 2150.68,1365.41 
  2155.38,1365.41 2160.08,1365.41 2164.78,1365.41 2169.48,1365.41 2174.18,1365.41 2178.88,1365.41 2183.58,1365.41 2188.28,1365.41 2192.98,1365.41 2197.68,1365.41 
  2202.38,1365.41 2207.08,1365.41 2211.78,1365.41 2216.48,1365.41 2221.17,1365.41 2225.87,1365.41 2230.57,1365.41 2235.27,1365.41 2239.97,1365.41 2244.67,1365.41 
  2249.37,1365.41 2254.07,1365.41 2258.77,1365.41 2263.47,1365.41 2268.17,1365.41 2272.87,1365.41 2277.57,1365.41 2282.27,1365.41 2286.97,1365.41 2291.66,1365.41 
  2296.36,1365.41 
  "/>
<polyline clip-path="url(#clip382)" style="stroke:#008000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" points="
  416.631,1365.41 421.33,1365.41 426.029,1365.41 430.729,1365.41 435.428,1365.41 440.127,1365.41 444.827,1365.41 449.526,1365.41 454.225,1365.41 458.925,1365.41 
  463.624,1365.41 468.323,1365.41 473.023,1365.41 477.722,1365.41 482.421,1365.41 487.121,1365.41 491.82,1365.41 496.519,1365.41 501.219,1365.41 505.918,1365.41 
  510.617,1365.41 515.317,1365.41 520.016,1365.41 524.715,1365.41 529.415,1365.41 534.114,1365.41 538.813,1365.41 543.513,1365.41 548.212,1365.41 552.911,1365.41 
  557.611,1365.41 562.31,1365.41 567.009,1365.41 571.709,1365.41 576.408,1365.41 581.107,1365.41 585.807,1365.41 590.506,1365.41 595.205,1365.41 599.905,1365.41 
  604.604,1365.41 609.303,1365.41 614.003,1365.41 618.702,1365.41 623.401,1365.41 628.101,1365.41 632.8,1365.41 637.499,1365.41 642.199,1365.41 646.898,1365.41 
  651.597,1365.41 656.297,1365.41 660.996,1365.41 665.695,1365.41 670.395,1365.41 675.094,1365.41 679.793,1365.41 684.493,1365.41 689.192,1365.41 693.891,1365.41 
  698.591,1365.41 703.29,1365.41 707.989,1365.41 712.689,1365.41 717.388,1365.41 722.087,1365.41 726.787,1365.41 731.486,1365.41 736.185,1365.41 740.885,1365.41 
  745.584,1365.41 750.283,1365.41 754.983,1365.41 759.682,1365.41 764.381,1365.41 769.081,1365.41 773.78,1365.41 778.479,1365.41 783.179,1365.41 787.878,1365.41 
  792.577,1365.41 797.277,1365.41 801.976,1365.41 806.675,1365.41 811.375,1365.41 816.074,1365.41 820.773,1365.41 825.473,1365.41 830.172,1365.41 834.871,1365.41 
  839.571,1365.41 844.27,1365.41 848.969,1365.41 853.669,1365.41 858.368,1365.41 863.067,1365.41 867.767,1365.41 872.466,1365.41 877.165,1365.41 881.865,1365.41 
  886.564,1365.41 891.263,1365.41 895.963,1365.41 900.662,1365.41 905.361,1365.41 910.061,1365.41 914.76,1365.41 919.459,1365.41 924.159,1365.41 928.858,1365.41 
  933.557,1365.41 938.257,1365.41 942.956,1365.41 947.655,1365.41 952.355,1365.41 957.054,1365.41 961.753,1365.41 966.453,1365.41 971.152,1365.41 975.851,1365.41 
  980.551,1365.41 985.25,1365.41 989.949,1365.41 994.649,1365.41 999.348,1365.41 1004.05,1365.41 1008.75,1365.41 1013.45,1365.41 1018.15,1365.41 1022.84,1365.41 
  1027.54,1365.41 1032.24,1365.41 1036.94,1365.41 1041.64,1365.41 1046.34,1365.41 1051.04,1365.41 1055.74,1365.41 1060.44,1365.41 1065.14,1365.41 1069.84,1365.41 
  1074.54,1365.41 1079.24,1365.41 1083.94,1365.41 1088.64,1365.41 1093.33,1365.41 1098.03,1365.41 1102.73,1365.41 1107.43,1365.41 1112.13,1365.41 1116.83,1365.41 
  1121.53,1365.41 1126.23,1365.41 1130.93,1365.41 1135.63,1365.41 1140.33,1365.41 1145.03,1365.41 1149.73,1365.41 1154.43,1365.41 1159.13,1365.41 1163.82,1365.41 
  1168.52,1365.41 1173.22,1365.41 1177.92,1365.41 1182.62,1365.41 1187.32,1365.41 1192.02,1365.41 1196.72,1365.41 1201.42,1365.41 1206.12,1365.41 1210.82,1365.41 
  1215.52,1365.41 1220.22,1365.41 1224.92,1365.41 1229.62,1365.41 1234.31,1365.41 1239.01,1365.41 1243.71,1365.41 1248.41,1365.41 1253.11,1365.41 1257.81,1365.41 
  1262.51,1365.41 1267.21,1365.41 1271.91,1365.41 1276.61,1365.41 1281.31,1365.41 1286.01,1365.41 1290.71,1365.41 1295.41,1365.41 1300.11,1365.41 1304.8,1365.41 
  1309.5,1365.41 1314.2,1365.41 1318.9,1365.41 1323.6,1365.41 1328.3,1365.41 1333,1365.41 1337.7,1365.41 1342.4,1365.41 1347.1,1365.41 1351.8,1365.41 
  1356.5,1365.41 1361.2,1365.41 1365.9,1365.41 1370.6,1365.41 1375.29,1365.41 1379.99,1365.41 1384.69,1365.41 1389.39,1365.41 1394.09,1365.41 1398.79,1365.41 
  1403.49,1365.41 1408.19,1365.41 1412.89,1365.41 1417.59,1365.41 1422.29,1365.41 1426.99,1365.41 1431.69,1365.41 1436.39,1365.41 1441.09,1365.41 1445.78,1365.41 
  1450.48,1365.41 1455.18,1365.41 1459.88,1365.41 1464.58,1365.41 1469.28,1365.41 1473.98,1365.41 1478.68,1365.41 1483.38,1365.41 1488.08,1365.41 1492.78,1365.41 
  1497.48,1365.41 1502.18,1365.41 1506.88,1365.41 1511.58,1365.41 1516.27,1365.41 1520.97,1365.41 1525.67,1365.41 1530.37,1365.41 1535.07,1365.41 1539.77,1365.41 
  1544.47,1365.41 1549.17,1365.41 1553.87,1365.41 1558.57,1365.41 1563.27,1365.41 1567.97,1365.41 1572.67,1365.41 1577.37,1365.41 1582.07,1365.41 1586.76,1365.41 
  1591.46,1365.41 1596.16,1365.41 1600.86,1365.41 1605.56,1365.41 1610.26,1365.41 1614.96,1365.41 1619.66,1365.41 1624.36,1365.41 1629.06,1365.41 1633.76,1365.41 
  1638.46,1365.41 1643.16,1365.41 1647.86,1365.41 1652.56,1365.41 1657.25,1365.41 1661.95,1365.41 1666.65,1365.41 1671.35,1365.41 1676.05,1365.41 1680.75,1365.41 
  1685.45,1365.41 1690.15,1365.41 1694.85,1365.41 1699.55,1365.41 1704.25,1365.41 1708.95,1365.41 1713.65,1365.41 1718.35,1365.41 1723.05,1365.41 1727.74,1365.41 
  1732.44,1365.41 1737.14,1365.41 1741.84,1365.41 1746.54,1365.41 1751.24,1365.41 1755.94,1365.41 1760.64,1365.41 1765.34,1365.41 1770.04,1365.41 1774.74,1365.41 
  1779.44,1365.41 1784.14,1365.41 1788.84,1365.41 1793.54,1365.41 1798.23,1365.41 1802.93,1365.41 1807.63,1365.41 1812.33,1365.41 1817.03,1365.41 1821.73,1365.41 
  1826.43,1365.41 1831.13,1365.41 1835.83,1365.41 1840.53,1365.41 1845.23,1365.41 1849.93,1365.41 1854.63,1365.41 1859.33,1365.41 1864.03,1365.41 1868.72,1365.41 
  1873.42,1365.41 1878.12,1365.41 1882.82,1365.41 1887.52,1365.41 1892.22,1365.41 1896.92,1365.41 1901.62,1365.41 1906.32,1365.41 1911.02,1365.41 1915.72,1365.41 
  1920.42,1365.41 1925.12,1365.41 1929.82,1365.41 1934.52,1365.41 1939.21,1365.41 1943.91,1365.41 1948.61,1365.41 1953.31,1365.41 1958.01,1365.41 1962.71,1365.41 
  1967.41,1365.41 1972.11,1365.41 1976.81,1365.41 1981.51,1365.41 1986.21,1365.41 1990.91,1365.41 1995.61,1365.41 2000.31,1365.41 2005.01,1365.41 2009.7,1365.41 
  2014.4,1365.41 2019.1,1365.41 2023.8,1365.41 2028.5,1365.41 2033.2,1365.41 2037.9,1365.41 2042.6,1365.41 2047.3,1365.41 2052,1365.41 2056.7,1365.41 
  2061.4,1365.41 2066.1,1365.41 2070.8,1365.41 2075.5,1365.41 2080.19,1365.41 2084.89,1365.41 2089.59,1365.41 2094.29,1365.41 2098.99,1365.41 2103.69,1365.41 
  2108.39,1365.41 2113.09,1365.41 2117.79,1365.41 2122.49,1365.41 2127.19,1365.41 2131.89,1365.41 2136.59,1365.41 2141.29,1365.41 2145.99,1365.41 2150.68,1365.41 
  2155.38,1365.41 2160.08,1365.41 2164.78,1365.41 2169.48,1365.41 2174.18,1365.41 2178.88,1365.41 2183.58,1365.41 2188.28,1365.41 2192.98,1365.41 2197.68,1365.41 
  2202.38,1365.41 2207.08,1365.41 2211.78,1365.41 2216.48,1365.41 2221.17,1365.41 2225.87,1365.41 2230.57,1365.41 2235.27,1365.41 2239.97,1365.41 2244.67,1365.41 
  2249.37,1365.41 2254.07,1365.41 2258.77,1365.41 2263.47,1365.41 2268.17,1365.41 2272.87,1365.41 2277.57,1365.41 2282.27,1365.41 2286.97,1365.41 2291.66,1365.41 
  2296.36,1365.41 
  "/>
<path clip-path="url(#clip380)" d="
M1430.09 554.876 L2286.34 554.876 L2286.34 166.076 L1430.09 166.076  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<polyline clip-path="url(#clip380)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1430.09,554.876 2286.34,554.876 2286.34,166.076 1430.09,166.076 1430.09,554.876 
  "/>
<polyline clip-path="url(#clip380)" style="stroke:#009af9; stroke-linecap:butt; stroke-linejoin:round; stroke-width:6; stroke-opacity:1; fill:none" points="
  1452.23,243.836 1585.06,243.836 
  "/>
<path clip-path="url(#clip380)" d="M1634.84 232.013 L1634.84 238.054 Q1632.13 236.666 1629.22 235.971 Q1626.3 235.277 1623.17 235.277 Q1618.42 235.277 1616.02 236.735 Q1613.66 238.193 1613.66 241.11 Q1613.66 243.332 1615.36 244.617 Q1617.06 245.867 1622.2 247.013 L1624.39 247.499 Q1631.2 248.957 1634.04 251.631 Q1636.92 254.27 1636.92 259.027 Q1636.92 264.443 1632.62 267.603 Q1628.35 270.763 1620.85 270.763 Q1617.72 270.763 1614.32 270.138 Q1610.95 269.547 1607.2 268.332 L1607.2 261.735 Q1610.74 263.575 1614.18 264.513 Q1617.62 265.415 1620.99 265.415 Q1625.5 265.415 1627.93 263.888 Q1630.36 262.325 1630.36 259.513 Q1630.36 256.909 1628.59 255.52 Q1626.86 254.131 1620.92 252.846 L1618.7 252.325 Q1612.76 251.075 1610.12 248.506 Q1607.48 245.902 1607.48 241.388 Q1607.48 235.902 1611.37 232.916 Q1615.26 229.93 1622.41 229.93 Q1625.95 229.93 1629.08 230.45 Q1632.2 230.971 1634.84 232.013 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1647.1 230.867 L1653.49 230.867 L1653.49 269.756 L1647.1 269.756 L1647.1 230.867 M1647.1 215.728 L1653.49 215.728 L1653.49 223.818 L1647.1 223.818 L1647.1 215.728 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1697.13 238.332 Q1699.53 234.027 1702.86 231.978 Q1706.2 229.93 1710.71 229.93 Q1716.79 229.93 1720.08 234.2 Q1723.38 238.436 1723.38 246.284 L1723.38 269.756 L1716.96 269.756 L1716.96 246.492 Q1716.96 240.902 1714.98 238.193 Q1713 235.485 1708.94 235.485 Q1703.97 235.485 1701.09 238.784 Q1698.21 242.082 1698.21 247.777 L1698.21 269.756 L1691.79 269.756 L1691.79 246.492 Q1691.79 240.867 1689.81 238.193 Q1687.83 235.485 1683.7 235.485 Q1678.8 235.485 1675.92 238.818 Q1673.04 242.117 1673.04 247.777 L1673.04 269.756 L1666.61 269.756 L1666.61 230.867 L1673.04 230.867 L1673.04 236.909 Q1675.22 233.332 1678.28 231.631 Q1681.33 229.93 1685.54 229.93 Q1689.77 229.93 1692.72 232.082 Q1695.71 234.235 1697.13 238.332 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1742.31 263.922 L1742.31 284.547 L1735.88 284.547 L1735.88 230.867 L1742.31 230.867 L1742.31 236.77 Q1744.32 233.298 1747.38 231.631 Q1750.47 229.93 1754.74 229.93 Q1761.82 229.93 1766.23 235.555 Q1770.67 241.179 1770.67 250.346 Q1770.67 259.513 1766.23 265.138 Q1761.82 270.763 1754.74 270.763 Q1750.47 270.763 1747.38 269.096 Q1744.32 267.395 1742.31 263.922 M1764.04 250.346 Q1764.04 243.298 1761.13 239.304 Q1758.24 235.277 1753.17 235.277 Q1748.1 235.277 1745.19 239.304 Q1742.31 243.298 1742.31 250.346 Q1742.31 257.395 1745.19 261.422 Q1748.1 265.415 1753.17 265.415 Q1758.24 265.415 1761.13 261.422 Q1764.04 257.395 1764.04 250.346 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1781.26 215.728 L1787.65 215.728 L1787.65 269.756 L1781.26 269.756 L1781.26 215.728 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1834.28 248.714 L1834.28 251.839 L1804.91 251.839 Q1805.33 258.436 1808.87 261.909 Q1812.44 265.346 1818.8 265.346 Q1822.48 265.346 1825.92 264.443 Q1829.39 263.54 1832.79 261.735 L1832.79 267.777 Q1829.35 269.235 1825.74 269.999 Q1822.13 270.763 1818.42 270.763 Q1809.11 270.763 1803.66 265.346 Q1798.24 259.929 1798.24 250.693 Q1798.24 241.145 1803.38 235.555 Q1808.56 229.93 1817.31 229.93 Q1825.15 229.93 1829.7 234.999 Q1834.28 240.034 1834.28 248.714 M1827.9 246.839 Q1827.83 241.596 1824.94 238.471 Q1822.1 235.346 1817.38 235.346 Q1812.03 235.346 1808.8 238.367 Q1805.6 241.388 1805.12 246.874 L1827.9 246.839 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1845.05 217.916 L1874.84 217.916 L1874.84 223.818 L1852.06 223.818 L1852.06 239.096 L1872.62 239.096 L1872.62 244.999 L1852.06 244.999 L1852.06 269.756 L1845.05 269.756 L1845.05 217.916 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1885.95 217.916 L1896.4 217.916 L1909.63 253.193 L1922.93 217.916 L1933.38 217.916 L1933.38 269.756 L1926.54 269.756 L1926.54 224.235 L1913.17 259.79 L1906.12 259.79 L1892.76 224.235 L1892.76 269.756 L1885.95 269.756 L1885.95 217.916 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1946.51 217.916 L1953.56 217.916 L1953.56 249.409 Q1953.56 257.742 1956.58 261.422 Q1959.6 265.068 1966.37 265.068 Q1973.1 265.068 1976.12 261.422 Q1979.15 257.742 1979.15 249.409 L1979.15 217.916 L1986.19 217.916 L1986.19 250.277 Q1986.19 260.415 1981.16 265.589 Q1976.16 270.763 1966.37 270.763 Q1956.54 270.763 1951.51 265.589 Q1946.51 260.415 1946.51 250.277 L1946.51 217.916 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip380)" style="stroke:#e26f46; stroke-linecap:butt; stroke-linejoin:round; stroke-width:6; stroke-opacity:1; fill:none" points="
  1452.23,321.596 1585.06,321.596 
  "/>
<path clip-path="url(#clip380)" d="M1629.98 314.599 Q1628.9 313.974 1627.62 313.696 Q1626.37 313.384 1624.84 313.384 Q1619.42 313.384 1616.51 316.926 Q1613.63 320.433 1613.63 327.03 L1613.63 347.516 L1607.2 347.516 L1607.2 308.627 L1613.63 308.627 L1613.63 314.669 Q1615.64 311.127 1618.87 309.426 Q1622.1 307.69 1626.72 307.69 Q1627.38 307.69 1628.17 307.794 Q1628.97 307.863 1629.95 308.037 L1629.98 314.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1668.38 326.474 L1668.38 329.599 L1639.01 329.599 Q1639.42 336.196 1642.97 339.669 Q1646.54 343.106 1652.9 343.106 Q1656.58 343.106 1660.02 342.203 Q1663.49 341.3 1666.89 339.495 L1666.89 345.537 Q1663.45 346.995 1659.84 347.759 Q1656.23 348.523 1652.52 348.523 Q1643.21 348.523 1637.76 343.106 Q1632.34 337.689 1632.34 328.453 Q1632.34 318.905 1637.48 313.315 Q1642.65 307.69 1651.4 307.69 Q1659.25 307.69 1663.8 312.759 Q1668.38 317.794 1668.38 326.474 M1661.99 324.599 Q1661.92 319.356 1659.04 316.231 Q1656.2 313.106 1651.47 313.106 Q1646.13 313.106 1642.9 316.127 Q1639.7 319.148 1639.22 324.634 L1661.99 324.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1698.56 293.488 L1698.56 298.801 L1692.45 298.801 Q1689.01 298.801 1687.65 300.19 Q1686.33 301.578 1686.33 305.19 L1686.33 308.627 L1696.86 308.627 L1696.86 313.592 L1686.33 313.592 L1686.33 347.516 L1679.91 347.516 L1679.91 313.592 L1673.8 313.592 L1673.8 308.627 L1679.91 308.627 L1679.91 305.919 Q1679.91 299.426 1682.93 296.474 Q1685.95 293.488 1692.51 293.488 L1698.56 293.488 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1737.17 326.474 L1737.17 329.599 L1707.79 329.599 Q1708.21 336.196 1711.75 339.669 Q1715.33 343.106 1721.68 343.106 Q1725.36 343.106 1728.8 342.203 Q1732.27 341.3 1735.67 339.495 L1735.67 345.537 Q1732.24 346.995 1728.63 347.759 Q1725.01 348.523 1721.3 348.523 Q1711.99 348.523 1706.54 343.106 Q1701.13 337.689 1701.13 328.453 Q1701.13 318.905 1706.26 313.315 Q1711.44 307.69 1720.19 307.69 Q1728.04 307.69 1732.58 312.759 Q1737.17 317.794 1737.17 326.474 M1730.78 324.599 Q1730.71 319.356 1727.83 316.231 Q1724.98 313.106 1720.26 313.106 Q1714.91 313.106 1711.68 316.127 Q1708.49 319.148 1708 324.634 L1730.78 324.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1770.19 314.599 Q1769.11 313.974 1767.83 313.696 Q1766.58 313.384 1765.05 313.384 Q1759.63 313.384 1756.72 316.926 Q1753.83 320.433 1753.83 327.03 L1753.83 347.516 L1747.41 347.516 L1747.41 308.627 L1753.83 308.627 L1753.83 314.669 Q1755.85 311.127 1759.08 309.426 Q1762.31 307.69 1766.92 307.69 Q1767.58 307.69 1768.38 307.794 Q1769.18 307.863 1770.15 308.037 L1770.19 314.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1808.59 326.474 L1808.59 329.599 L1779.22 329.599 Q1779.63 336.196 1783.17 339.669 Q1786.75 343.106 1793.1 343.106 Q1796.79 343.106 1800.22 342.203 Q1803.69 341.3 1807.1 339.495 L1807.1 345.537 Q1803.66 346.995 1800.05 347.759 Q1796.44 348.523 1792.72 348.523 Q1783.42 348.523 1777.97 343.106 Q1772.55 337.689 1772.55 328.453 Q1772.55 318.905 1777.69 313.315 Q1782.86 307.69 1791.61 307.69 Q1799.46 307.69 1804.01 312.759 Q1808.59 317.794 1808.59 326.474 M1802.2 324.599 Q1802.13 319.356 1799.25 316.231 Q1796.4 313.106 1791.68 313.106 Q1786.33 313.106 1783.1 316.127 Q1779.91 319.148 1779.42 324.634 L1802.2 324.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1851.4 324.044 L1851.4 347.516 L1845.01 347.516 L1845.01 324.252 Q1845.01 318.731 1842.86 315.988 Q1840.71 313.245 1836.4 313.245 Q1831.23 313.245 1828.24 316.544 Q1825.26 319.842 1825.26 325.537 L1825.26 347.516 L1818.83 347.516 L1818.83 308.627 L1825.26 308.627 L1825.26 314.669 Q1827.55 311.162 1830.64 309.426 Q1833.76 307.69 1837.83 307.69 Q1844.53 307.69 1847.97 311.856 Q1851.4 315.988 1851.4 324.044 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1892.13 310.12 L1892.13 316.092 Q1889.42 314.599 1886.68 313.87 Q1883.97 313.106 1881.19 313.106 Q1874.98 313.106 1871.54 317.064 Q1868.1 320.988 1868.1 328.106 Q1868.1 335.224 1871.54 339.182 Q1874.98 343.106 1881.19 343.106 Q1883.97 343.106 1886.68 342.377 Q1889.42 341.613 1892.13 340.12 L1892.13 346.023 Q1889.46 347.273 1886.58 347.898 Q1883.73 348.523 1880.5 348.523 Q1871.72 348.523 1866.54 343.002 Q1861.37 337.481 1861.37 328.106 Q1861.37 318.592 1866.58 313.141 Q1871.82 307.69 1880.92 307.69 Q1883.87 307.69 1886.68 308.315 Q1889.49 308.905 1892.13 310.12 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1936.51 326.474 L1936.51 329.599 L1907.13 329.599 Q1907.55 336.196 1911.09 339.669 Q1914.67 343.106 1921.02 343.106 Q1924.7 343.106 1928.14 342.203 Q1931.61 341.3 1935.01 339.495 L1935.01 345.537 Q1931.58 346.995 1927.96 347.759 Q1924.35 348.523 1920.64 348.523 Q1911.33 348.523 1905.88 343.106 Q1900.47 337.689 1900.47 328.453 Q1900.47 318.905 1905.6 313.315 Q1910.78 307.69 1919.53 307.69 Q1927.37 307.69 1931.92 312.759 Q1936.51 317.794 1936.51 326.474 M1930.12 324.599 Q1930.05 319.356 1927.17 316.231 Q1924.32 313.106 1919.6 313.106 Q1914.25 313.106 1911.02 316.127 Q1907.83 319.148 1907.34 324.634 L1930.12 324.599 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip380)" style="stroke:#008000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:6; stroke-opacity:1; fill:none" stroke-dasharray="3, 6" points="
  1452.23,399.356 1585.06,399.356 
  "/>
<path clip-path="url(#clip380)" d="M1607.2 373.436 L1616.65 373.436 L1639.63 416.804 L1639.63 373.436 L1646.44 373.436 L1646.44 425.276 L1636.99 425.276 L1614.01 381.908 L1614.01 425.276 L1607.2 425.276 L1607.2 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1693.38 404.234 L1693.38 407.359 L1664.01 407.359 Q1664.42 413.956 1667.97 417.429 Q1671.54 420.866 1677.9 420.866 Q1681.58 420.866 1685.01 419.963 Q1688.49 419.06 1691.89 417.255 L1691.89 423.297 Q1688.45 424.755 1684.84 425.519 Q1681.23 426.283 1677.51 426.283 Q1668.21 426.283 1662.76 420.866 Q1657.34 415.449 1657.34 406.213 Q1657.34 396.665 1662.48 391.075 Q1667.65 385.45 1676.4 385.45 Q1684.25 385.45 1688.8 390.519 Q1693.38 395.554 1693.38 404.234 M1686.99 402.359 Q1686.92 397.116 1684.04 393.991 Q1681.2 390.866 1676.47 390.866 Q1671.13 390.866 1667.9 393.887 Q1664.7 396.908 1664.22 402.394 L1686.99 402.359 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1703.21 409.929 L1703.21 386.387 L1709.6 386.387 L1709.6 409.686 Q1709.6 415.206 1711.75 417.984 Q1713.9 420.727 1718.21 420.727 Q1723.38 420.727 1726.37 417.429 Q1729.39 414.13 1729.39 408.436 L1729.39 386.387 L1735.78 386.387 L1735.78 425.276 L1729.39 425.276 L1729.39 419.304 Q1727.06 422.845 1723.97 424.581 Q1720.92 426.283 1716.85 426.283 Q1710.15 426.283 1706.68 422.116 Q1703.21 417.949 1703.21 409.929 M1719.29 385.45 L1719.29 385.45 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1771.47 392.359 Q1770.4 391.734 1769.11 391.456 Q1767.86 391.144 1766.33 391.144 Q1760.92 391.144 1758 394.686 Q1755.12 398.193 1755.12 404.79 L1755.12 425.276 L1748.7 425.276 L1748.7 386.387 L1755.12 386.387 L1755.12 392.429 Q1757.13 388.887 1760.36 387.186 Q1763.59 385.45 1768.21 385.45 Q1768.87 385.45 1769.67 385.554 Q1770.47 385.623 1771.44 385.797 L1771.47 392.359 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1795.85 405.727 Q1788.1 405.727 1785.12 407.498 Q1782.13 409.269 1782.13 413.54 Q1782.13 416.942 1784.35 418.956 Q1786.61 420.935 1790.47 420.935 Q1795.78 420.935 1798.97 417.185 Q1802.2 413.401 1802.2 407.151 L1802.2 405.727 L1795.85 405.727 M1808.59 403.088 L1808.59 425.276 L1802.2 425.276 L1802.2 419.373 Q1800.01 422.915 1796.75 424.616 Q1793.49 426.283 1788.76 426.283 Q1782.79 426.283 1779.25 422.949 Q1775.74 419.581 1775.74 413.956 Q1775.74 407.394 1780.12 404.061 Q1784.53 400.727 1793.24 400.727 L1802.2 400.727 L1802.2 400.102 Q1802.2 395.693 1799.29 393.297 Q1796.4 390.866 1791.16 390.866 Q1787.83 390.866 1784.67 391.665 Q1781.51 392.463 1778.59 394.061 L1778.59 388.158 Q1782.1 386.804 1785.4 386.144 Q1788.69 385.45 1791.82 385.45 Q1800.26 385.45 1804.42 389.825 Q1808.59 394.199 1808.59 403.088 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1821.75 371.248 L1828.14 371.248 L1828.14 425.276 L1821.75 425.276 L1821.75 371.248 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1841.78 373.436 L1871.58 373.436 L1871.58 379.338 L1848.8 379.338 L1848.8 394.616 L1869.35 394.616 L1869.35 400.519 L1848.8 400.519 L1848.8 425.276 L1841.78 425.276 L1841.78 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1882.69 373.436 L1893.14 373.436 L1906.37 408.713 L1919.67 373.436 L1930.12 373.436 L1930.12 425.276 L1923.28 425.276 L1923.28 379.755 L1909.91 415.31 L1902.86 415.31 L1889.49 379.755 L1889.49 425.276 L1882.69 425.276 L1882.69 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1943.24 373.436 L1950.29 373.436 L1950.29 404.929 Q1950.29 413.262 1953.31 416.942 Q1956.33 420.588 1963.1 420.588 Q1969.84 420.588 1972.86 416.942 Q1975.88 413.262 1975.88 404.929 L1975.88 373.436 L1982.93 373.436 L1982.93 405.797 Q1982.93 415.935 1977.9 421.109 Q1972.9 426.283 1963.1 426.283 Q1953.28 426.283 1948.24 421.109 Q1943.24 415.935 1943.24 405.797 L1943.24 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M2033.76 371.318 Q2029.11 379.304 2026.85 387.116 Q2024.6 394.929 2024.6 402.949 Q2024.6 410.97 2026.85 418.852 Q2029.14 426.699 2033.76 434.651 L2028.21 434.651 Q2023 426.491 2020.4 418.609 Q2017.83 410.727 2017.83 402.949 Q2017.83 395.206 2020.4 387.359 Q2022.96 379.512 2028.21 371.318 L2033.76 371.318 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M2053.1 419.373 L2077.58 419.373 L2077.58 425.276 L2044.67 425.276 L2044.67 419.373 Q2048.66 415.241 2055.53 408.297 Q2062.44 401.318 2064.21 399.304 Q2067.58 395.519 2068.9 392.915 Q2070.26 390.276 2070.26 387.741 Q2070.26 383.609 2067.34 381.005 Q2064.46 378.401 2059.8 378.401 Q2056.51 378.401 2052.83 379.547 Q2049.18 380.693 2045.01 383.019 L2045.01 375.936 Q2049.25 374.234 2052.93 373.366 Q2056.61 372.498 2059.67 372.498 Q2067.72 372.498 2072.51 376.526 Q2077.3 380.554 2077.3 387.29 Q2077.3 390.484 2076.09 393.366 Q2074.91 396.213 2071.75 400.102 Q2070.88 401.109 2066.23 405.936 Q2061.58 410.727 2053.1 419.373 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M2092.37 373.436 L2119.91 373.436 L2119.91 379.338 L2098.8 379.338 L2098.8 392.047 Q2100.33 391.526 2101.85 391.283 Q2103.38 391.005 2104.91 391.005 Q2113.59 391.005 2118.66 395.762 Q2123.73 400.519 2123.73 408.644 Q2123.73 417.012 2118.52 421.665 Q2113.31 426.283 2103.83 426.283 Q2100.57 426.283 2097.17 425.727 Q2093.8 425.172 2090.19 424.06 L2090.19 417.012 Q2093.31 418.713 2096.64 419.547 Q2099.98 420.38 2103.69 420.38 Q2109.7 420.38 2113.21 417.22 Q2116.71 414.06 2116.71 408.644 Q2116.71 403.227 2113.21 400.068 Q2109.7 396.908 2103.69 396.908 Q2100.88 396.908 2098.07 397.533 Q2095.29 398.158 2092.37 399.477 L2092.37 373.436 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M2152.55 378.054 Q2147.13 378.054 2144.39 383.401 Q2141.68 388.713 2141.68 399.408 Q2141.68 410.067 2144.39 415.415 Q2147.13 420.727 2152.55 420.727 Q2158 420.727 2160.71 415.415 Q2163.45 410.067 2163.45 399.408 Q2163.45 388.713 2160.71 383.401 Q2158 378.054 2152.55 378.054 M2152.55 372.498 Q2161.26 372.498 2165.85 379.408 Q2170.46 386.283 2170.46 399.408 Q2170.46 412.498 2165.85 419.408 Q2161.26 426.283 2152.55 426.283 Q2143.83 426.283 2139.21 419.408 Q2134.63 412.498 2134.63 399.408 Q2134.63 386.283 2139.21 379.408 Q2143.83 372.498 2152.55 372.498 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M2197.79 378.054 Q2192.37 378.054 2189.63 383.401 Q2186.92 388.713 2186.92 399.408 Q2186.92 410.067 2189.63 415.415 Q2192.37 420.727 2197.79 420.727 Q2203.24 420.727 2205.95 415.415 Q2208.69 410.067 2208.69 399.408 Q2208.69 388.713 2205.95 383.401 Q2203.24 378.054 2197.79 378.054 M2197.79 372.498 Q2206.51 372.498 2211.09 379.408 Q2215.71 386.283 2215.71 399.408 Q2215.71 412.498 2211.09 419.408 Q2206.51 426.283 2197.79 426.283 Q2189.07 426.283 2184.46 419.408 Q2179.87 412.498 2179.87 399.408 Q2179.87 386.283 2184.46 379.408 Q2189.07 372.498 2197.79 372.498 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M2226.12 371.318 L2231.68 371.318 Q2236.89 379.512 2239.46 387.359 Q2242.06 395.206 2242.06 402.949 Q2242.06 410.727 2239.46 418.609 Q2236.89 426.491 2231.68 434.651 L2226.12 434.651 Q2230.74 426.699 2233 418.852 Q2235.29 410.97 2235.29 402.949 Q2235.29 394.929 2233 387.116 Q2230.74 379.304 2226.12 371.318 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip380)" style="stroke:#008000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:6; stroke-opacity:1; fill:none" points="
  1452.23,477.116 1585.06,477.116 
  "/>
<path clip-path="url(#clip380)" d="M1607.2 451.196 L1616.65 451.196 L1639.63 494.564 L1639.63 451.196 L1646.44 451.196 L1646.44 503.036 L1636.99 503.036 L1614.01 459.668 L1614.01 503.036 L1607.2 503.036 L1607.2 451.196 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1693.38 481.994 L1693.38 485.119 L1664.01 485.119 Q1664.42 491.716 1667.97 495.189 Q1671.54 498.626 1677.9 498.626 Q1681.58 498.626 1685.01 497.723 Q1688.49 496.82 1691.89 495.015 L1691.89 501.057 Q1688.45 502.515 1684.84 503.279 Q1681.23 504.043 1677.51 504.043 Q1668.21 504.043 1662.76 498.626 Q1657.34 493.209 1657.34 483.973 Q1657.34 474.425 1662.48 468.835 Q1667.65 463.21 1676.4 463.21 Q1684.25 463.21 1688.8 468.279 Q1693.38 473.314 1693.38 481.994 M1686.99 480.119 Q1686.92 474.876 1684.04 471.751 Q1681.2 468.626 1676.47 468.626 Q1671.13 468.626 1667.9 471.647 Q1664.7 474.668 1664.22 480.154 L1686.99 480.119 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1703.21 487.689 L1703.21 464.147 L1709.6 464.147 L1709.6 487.446 Q1709.6 492.966 1711.75 495.744 Q1713.9 498.487 1718.21 498.487 Q1723.38 498.487 1726.37 495.189 Q1729.39 491.89 1729.39 486.196 L1729.39 464.147 L1735.78 464.147 L1735.78 503.036 L1729.39 503.036 L1729.39 497.064 Q1727.06 500.605 1723.97 502.341 Q1720.92 504.043 1716.85 504.043 Q1710.15 504.043 1706.68 499.876 Q1703.21 495.709 1703.21 487.689 M1719.29 463.21 L1719.29 463.21 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1771.47 470.119 Q1770.4 469.494 1769.11 469.216 Q1767.86 468.904 1766.33 468.904 Q1760.92 468.904 1758 472.446 Q1755.12 475.953 1755.12 482.55 L1755.12 503.036 L1748.7 503.036 L1748.7 464.147 L1755.12 464.147 L1755.12 470.189 Q1757.13 466.647 1760.36 464.946 Q1763.59 463.21 1768.21 463.21 Q1768.87 463.21 1769.67 463.314 Q1770.47 463.383 1771.44 463.557 L1771.47 470.119 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1795.85 483.487 Q1788.1 483.487 1785.12 485.258 Q1782.13 487.029 1782.13 491.3 Q1782.13 494.702 1784.35 496.716 Q1786.61 498.695 1790.47 498.695 Q1795.78 498.695 1798.97 494.945 Q1802.2 491.161 1802.2 484.911 L1802.2 483.487 L1795.85 483.487 M1808.59 480.848 L1808.59 503.036 L1802.2 503.036 L1802.2 497.133 Q1800.01 500.675 1796.75 502.376 Q1793.49 504.043 1788.76 504.043 Q1782.79 504.043 1779.25 500.709 Q1775.74 497.341 1775.74 491.716 Q1775.74 485.154 1780.12 481.821 Q1784.53 478.487 1793.24 478.487 L1802.2 478.487 L1802.2 477.862 Q1802.2 473.453 1799.29 471.057 Q1796.4 468.626 1791.16 468.626 Q1787.83 468.626 1784.67 469.425 Q1781.51 470.223 1778.59 471.821 L1778.59 465.918 Q1782.1 464.564 1785.4 463.904 Q1788.69 463.21 1791.82 463.21 Q1800.26 463.21 1804.42 467.585 Q1808.59 471.959 1808.59 480.848 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1821.75 449.008 L1828.14 449.008 L1828.14 503.036 L1821.75 503.036 L1821.75 449.008 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1841.78 451.196 L1871.58 451.196 L1871.58 457.098 L1848.8 457.098 L1848.8 472.376 L1869.35 472.376 L1869.35 478.279 L1848.8 478.279 L1848.8 503.036 L1841.78 503.036 L1841.78 451.196 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1882.69 451.196 L1893.14 451.196 L1906.37 486.473 L1919.67 451.196 L1930.12 451.196 L1930.12 503.036 L1923.28 503.036 L1923.28 457.515 L1909.91 493.07 L1902.86 493.07 L1889.49 457.515 L1889.49 503.036 L1882.69 503.036 L1882.69 451.196 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M1943.24 451.196 L1950.29 451.196 L1950.29 482.689 Q1950.29 491.022 1953.31 494.702 Q1956.33 498.348 1963.1 498.348 Q1969.84 498.348 1972.86 494.702 Q1975.88 491.022 1975.88 482.689 L1975.88 451.196 L1982.93 451.196 L1982.93 483.557 Q1982.93 493.695 1977.9 498.869 Q1972.9 504.043 1963.1 504.043 Q1953.28 504.043 1948.24 498.869 Q1943.24 493.695 1943.24 483.557 L1943.24 451.196 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M2033.76 449.078 Q2029.11 457.064 2026.85 464.876 Q2024.6 472.689 2024.6 480.709 Q2024.6 488.73 2026.85 496.612 Q2029.14 504.459 2033.76 512.411 L2028.21 512.411 Q2023 504.251 2020.4 496.369 Q2017.83 488.487 2017.83 480.709 Q2017.83 472.966 2020.4 465.119 Q2022.96 457.272 2028.21 449.078 L2033.76 449.078 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M2047.13 451.196 L2074.67 451.196 L2074.67 457.098 L2053.55 457.098 L2053.55 469.807 Q2055.08 469.286 2056.61 469.043 Q2058.14 468.765 2059.67 468.765 Q2068.35 468.765 2073.42 473.522 Q2078.48 478.279 2078.48 486.404 Q2078.48 494.772 2073.28 499.425 Q2068.07 504.043 2058.59 504.043 Q2055.33 504.043 2051.92 503.487 Q2048.55 502.932 2044.94 501.82 L2044.94 494.772 Q2048.07 496.473 2051.4 497.307 Q2054.74 498.14 2058.45 498.14 Q2064.46 498.14 2067.96 494.98 Q2071.47 491.82 2071.47 486.404 Q2071.47 480.987 2067.96 477.828 Q2064.46 474.668 2058.45 474.668 Q2055.64 474.668 2052.83 475.293 Q2050.05 475.918 2047.13 477.237 L2047.13 451.196 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M2107.3 455.814 Q2101.89 455.814 2099.14 461.161 Q2096.44 466.473 2096.44 477.168 Q2096.44 487.827 2099.14 493.175 Q2101.89 498.487 2107.3 498.487 Q2112.76 498.487 2115.46 493.175 Q2118.21 487.827 2118.21 477.168 Q2118.21 466.473 2115.46 461.161 Q2112.76 455.814 2107.3 455.814 M2107.3 450.258 Q2116.02 450.258 2120.6 457.168 Q2125.22 464.043 2125.22 477.168 Q2125.22 490.258 2120.6 497.168 Q2116.02 504.043 2107.3 504.043 Q2098.59 504.043 2093.97 497.168 Q2089.39 490.258 2089.39 477.168 Q2089.39 464.043 2093.97 457.168 Q2098.59 450.258 2107.3 450.258 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M2152.55 455.814 Q2147.13 455.814 2144.39 461.161 Q2141.68 466.473 2141.68 477.168 Q2141.68 487.827 2144.39 493.175 Q2147.13 498.487 2152.55 498.487 Q2158 498.487 2160.71 493.175 Q2163.45 487.827 2163.45 477.168 Q2163.45 466.473 2160.71 461.161 Q2158 455.814 2152.55 455.814 M2152.55 450.258 Q2161.26 450.258 2165.85 457.168 Q2170.46 464.043 2170.46 477.168 Q2170.46 490.258 2165.85 497.168 Q2161.26 504.043 2152.55 504.043 Q2143.83 504.043 2139.21 497.168 Q2134.63 490.258 2134.63 477.168 Q2134.63 464.043 2139.21 457.168 Q2143.83 450.258 2152.55 450.258 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M2197.79 455.814 Q2192.37 455.814 2189.63 461.161 Q2186.92 466.473 2186.92 477.168 Q2186.92 487.827 2189.63 493.175 Q2192.37 498.487 2197.79 498.487 Q2203.24 498.487 2205.95 493.175 Q2208.69 487.827 2208.69 477.168 Q2208.69 466.473 2205.95 461.161 Q2203.24 455.814 2197.79 455.814 M2197.79 450.258 Q2206.51 450.258 2211.09 457.168 Q2215.71 464.043 2215.71 477.168 Q2215.71 490.258 2211.09 497.168 Q2206.51 504.043 2197.79 504.043 Q2189.07 504.043 2184.46 497.168 Q2179.87 490.258 2179.87 477.168 Q2179.87 464.043 2184.46 457.168 Q2189.07 450.258 2197.79 450.258 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip380)" d="M2226.12 449.078 L2231.68 449.078 Q2236.89 457.272 2239.46 465.119 Q2242.06 472.966 2242.06 480.709 Q2242.06 488.487 2239.46 496.369 Q2236.89 504.251 2231.68 512.411 L2226.12 512.411 Q2230.74 504.459 2233 496.612 Q2235.29 488.73 2235.29 480.709 Q2235.29 472.689 2233 464.876 Q2230.74 457.064 2226.12 449.078 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /></svg>



Finally, the FMU is cleaned-up.


```julia
fmiUnload(simpleFMU)
```

### Summar

Based on the plots, it can be seen that the curves of the *realFMU* and the *neuralFMU* are very close. The *neuralFMU* is able to learn the friction and displacement model.

### Source

[1] Tobias Thummerer, Lars Mikelsons and Josef Kircher. 2021. **NeuralFMU: towards structural integration of FMUs into neural networks.** Martin Sjölund, Lena Buffoni, Adrian Pop and Lennart Ochel (Ed.). Proceedings of 14th Modelica Conference 2021, Linköping, Sweden, September 20-24, 2021. Linköping University Electronic Press, Linköping (Linköping Electronic Conference Proceedings ; 181), 297-306. [DOI: 10.3384/ecp21181297](https://doi.org/10.3384/ecp21181297)

