# Creation and training of CS-NeuralFMUs
Tutorial by Johannes Stoljar, Tobias Thummerer

## License
Copyright (c) 2021 Tobias Thummerer, Lars Mikelsons, Johannes Stoljar

Licensed under the MIT license. See [LICENSE](https://github.com/thummeto/FMIFlux.jl/blob/main/LICENSE) file in the project root for details.

## Motivation
The Julia Package *FMIFlux.jl* is motivated by the application of hybrid modeling. This package enables the user to integrate his simulation model between neural networks (NeuralFMU). For this, the simulation model must be exported as FMU (functional mock-up unit), which corresponds to a widely used standard. The big advantage of hybrid modeling with artificial neural networks is, that effects that are difficult to model (because they might be unknown) can be easily learned by the neural networks. For this purpose, the NeuralFMU is trained with measurement data containing the unmodeled physical effect. The final product is a simulation model including the orignially unmodeled effects. Another big advantage of the NeuralFMU is that it works with little data, because the FMU already contains the characterisitic functionality of the simulation and only the missing effects are added.

NeuralFMUs need not to be as easy as in this example. Basically a NeuralFMU can combine different ANN topologies that manipulate any FMU-input (system state, system inputs, time) and any FMU-output (system state derivative, system outputs, other system variables). However, for this example a NeuralFMU topology as shown in the following picture is used.

![CS-NeuralFMU.svg](https://github.com/thummeto/FMIFlux.jl/blob/main/docs/src/examples/pics/CSNeuralFMU.svg?raw=true)

*NeuralFMU (CS) from* [[1]](#Source).

## Introduction to the example
In this example, the model of a one-dimensional spring pendulum (with an external acting force) is used to learn the initial states. For this purpose, on the one hand the initial position of the mass of the pendulum is shifted and on the other hand the default position of the mass from the model is used. The model with the shifted initial position serves as reference and is called *referenceFMU* in the following. The model with the default position is further referenced with *defaultFMU*. At the beginning, the actual state of both simulations is shown, whereby clear deviations can be seen in the graphs. Afterwards, the *defaultFMU* is integrated into a co-simulation NeuralFMU (CS-NeuralFMU) architecture. By training the NeuralFMU, an attempt is made to learn the initial displacement of the *referenceFMU*. It can be clearly seen that the NeuralFMU learns this shift well in just a few training steps. 


## Target group
The example is primarily intended for users who work in the field of first principle and/or hybrid modeling and are further interested in hybrid model building. The example wants to show how simple it is to combine FMUs with machine learning and to illustrate the advantages of this approach.


## Other formats
Besides this [Jupyter Notebook](https://github.com/thummeto/FMIFlux.jl/blob/main/example/simple_hybrid_CS.ipynb) there is also a [Julia file](https://github.com/thummeto/FMIFlux.jl/blob/main/example/simple_hybrid_CS.jl) with the same name, which contains only the code cells and for the documentation there is a [Markdown file](https://github.com/thummeto/FMIFlux.jl/blob/main/docs/src/examples/simple_hybrid_CS.md) corresponding to the notebook.  


## Getting started

### Installation prerequisites
|    | Description                       | Command     |  Alternative  |   
|:--- |:---                               |:---        |:---|
|1.  | Enter Package Manager via         |     ]       |     |
|2.  | Install FMI via                   |   add FMI   | add " https://github.com/ThummeTo/FMI.jl "   |
|3.  | Install FMIFlux via               | add FMIFlux | add " https://github.com/ThummeTo/FMIFlux.jl " |
|4.  | Install FMIZoo via                | add FMIZoo  | add " https://github.com/ThummeTo/FMIZoo.jl " |
|5.  | Install Flux via                  |  add Flux   |     |
|6.  | Install DifferentialEquations via | add DifferentialEquations |  |
|7.  | Install Plots via                 | add Plots   |     |

## Code section

To run the example, the previously installed packages must be included. 


```julia
# imports
using FMI
using FMIFlux
using FMIZoo
using Flux
using DifferentialEquations: Tsit5
import Plots
```

    WARNING: Method definition (::Type{DiffEqFlux.DeterministicCNF{M, P, RE, D, T, A, K} where K where A where T where D where RE where P where M})(M, P, RE, D, T, A, K) where {M, P, RE, D, T, A, K} in module DiffEqFlux at deprecated.jl:70 overwritten at C:\Users\thummeto\.julia\packages\DiffEqFlux\A5156\src\ffjord.jl:41.
      ** incremental compilation may be fatally broken for this module **
    


After importing the packages, the path to the *Functional Mock-up Units* (FMUs) is set. The FMU is a model exported meeting the *Functional Mock-up Interface* (FMI) Standard. The FMI is a free standard ([fmi-standard.org](http://fmi-standard.org/)) that defines a container and an interface to exchange dynamic models using a combination of XML files, binaries and C code zipped into a single file. 

The objec-orientated structure of the *SpringPendulumExtForce1D* can be seen in the following graphic. This model is a simple spring pendulum without friction, but with an external force. 

![svg](https://github.com/thummeto/FMIFlux.jl/blob/main/docs/src/examples/pics/SpringPendulumExtForce1D.svg?raw=true)

Next, the start time and end time of the simulation are set. Finally, a step size is specified to store the results of the simulation at these time steps.


```julia
tStart = 0.0
tStep = 0.01
tStop = 5.0
tSave = tStart:tStep:tStop
```


    0.0:0.01:5.0


### ReferenceFMU

In the next lines of code the FMU of the *referenceFMU* model is loaded from *FMIZoo.jl* and instantiated.  


```julia
referenceFMU = fmiLoad("SpringPendulumExtForce1D", "Dymola", "2022x")
fmiInstantiate!(referenceFMU; loggingOn=false)
fmiInfo(referenceFMU)
```

    ┌ Info: fmi2Unzip(...): Successfully unzipped 153 files at `C:\Users\thummeto\AppData\Local\Temp\fmijl_4Cv65Q\SpringPendulumExtForce1D`.
    └ @ FMIImport C:\Users\thummeto\Documents\FMIImport.jl\src\FMI2_ext.jl:77
    ┌ Info: fmi2Load(...): FMU resources location is `file:///C:/Users/thummeto/AppData/Local/Temp/fmijl_4Cv65Q/SpringPendulumExtForce1D/resources`
    └ @ FMIImport C:\Users\thummeto\Documents\FMIImport.jl\src\FMI2_ext.jl:192
    ┌ Info: fmi2Load(...): FMU supports both CS and ME, using CS as default if nothing specified.
    └ @ FMIImport C:\Users\thummeto\Documents\FMIImport.jl\src\FMI2_ext.jl:195


    #################### Begin information for FMU ####################
    	Model name:			SpringPendulumExtForce1D
    	FMI-Version:			2.0
    	GUID:				{7d007818-e354-47c7-9222-cdc1c9e53a24}
    	Generation tool:		Dymola Version 2022x (64-bit), 2021-10-08
    	Generation time:		2022-03-18T07:57:54Z
    	Var. naming conv.:		structured
    	Event indicators:		0
    	Inputs:				1
    		352321536 ["extForce"]
    	Outputs:			2
    		335544320 ["accSensor.v", "der(accSensor.flange.s)", "v", "der(speedSensor.flange.s)", "speedSensor.v"]
    		335544321 ["der(accSensor.v)", "a", "accSensor.a"]
    	States:				2
    		33554432 ["mass.s"]
    		33554433 ["mass.v"]
    	Supports Co-Simulation:		true
    		Model identifier:	SpringPendulumExtForce1D
    		Get/Set State:		true
    		Serialize State:	true
    		Dir. Derivatives:	true
    		Var. com. steps:	true
    		Input interpol.:	true
    		Max order out. der.:	1
    	Supports Model-Exchange:	true
    		Model identifier:	SpringPendulumExtForce1D
    		Get/Set State:		true
    		Serialize State:	true
    		Dir. Derivatives:	true
    ##################### End information for FMU #####################


Both the start and end time are set via the *fmiSetupExperiment()* function. In addition, the initial position of the mass is set to a value of $1.3m$  The experiment is initialized to get the information of the continuous states. You can get all continuous states of a FMU by the function *fmiGetContinuousStates()* and this is also done for the *referenceFMU*. It has two states: The first state is the previously initialized position of the mass, the second state is the velocity, which is initialized with $0\frac{m}{s}$.   


```julia
fmiSetupExperiment(referenceFMU, tStart, tStop)
fmiSetReal(referenceFMU, "mass_s0", 1.3)   # increase amplitude, invert phase
fmiEnterInitializationMode(referenceFMU)
fmiExitInitializationMode(referenceFMU)

x₀ = fmiGetContinuousStates(referenceFMU)
```


    2-element Vector{Float64}:
     1.3
     0.0


In the following code block the *referenceFMU* is simulated, still specifying which variables are included. After the simulation is finished the result of the *referenceFMU* can be plotted. This plot also serves as a reference for the later CS-NeuralFMU model.


```julia
vrs = ["mass.s", "mass.v", "mass.a"]
_, referenceSimData = fmiSimulate(referenceFMU, tStart, tStop; recordValues=vrs, setup=false, reset=false, saveat=tSave)
fmiPlot(referenceFMU, vrs, referenceSimData)
```


<?xml version="1.0" encoding="utf-8"?>
<svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" width="600" height="400" viewBox="0 0 2400 1600">
<defs>
  <clipPath id="clip970">
    <rect x="0" y="0" width="2400" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip970)" d="
M0 1600 L2400 1600 L2400 0 L0 0  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip971">
    <rect x="480" y="0" width="1681" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip970)" d="
M147.478 1423.18 L2352.76 1423.18 L2352.76 47.2441 L147.478 47.2441  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip972">
    <rect x="147" y="47" width="2206" height="1377"/>
  </clipPath>
</defs>
<polyline clip-path="url(#clip972)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  209.891,1423.18 209.891,47.2441 
  "/>
<polyline clip-path="url(#clip972)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  625.982,1423.18 625.982,47.2441 
  "/>
<polyline clip-path="url(#clip972)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1042.07,1423.18 1042.07,47.2441 
  "/>
<polyline clip-path="url(#clip972)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1458.16,1423.18 1458.16,47.2441 
  "/>
<polyline clip-path="url(#clip972)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1874.25,1423.18 1874.25,47.2441 
  "/>
<polyline clip-path="url(#clip972)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  2290.34,1423.18 2290.34,47.2441 
  "/>
<polyline clip-path="url(#clip970)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  147.478,1423.18 2352.76,1423.18 
  "/>
<polyline clip-path="url(#clip970)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  209.891,1423.18 209.891,1404.28 
  "/>
<polyline clip-path="url(#clip970)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  625.982,1423.18 625.982,1404.28 
  "/>
<polyline clip-path="url(#clip970)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1042.07,1423.18 1042.07,1404.28 
  "/>
<polyline clip-path="url(#clip970)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1458.16,1423.18 1458.16,1404.28 
  "/>
<polyline clip-path="url(#clip970)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1874.25,1423.18 1874.25,1404.28 
  "/>
<polyline clip-path="url(#clip970)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  2290.34,1423.18 2290.34,1404.28 
  "/>
<path clip-path="url(#clip970)" d="M209.891 1454.1 Q206.28 1454.1 204.452 1457.66 Q202.646 1461.2 202.646 1468.33 Q202.646 1475.44 204.452 1479.01 Q206.28 1482.55 209.891 1482.55 Q213.526 1482.55 215.331 1479.01 Q217.16 1475.44 217.16 1468.33 Q217.16 1461.2 215.331 1457.66 Q213.526 1454.1 209.891 1454.1 M209.891 1450.39 Q215.702 1450.39 218.757 1455 Q221.836 1459.58 221.836 1468.33 Q221.836 1477.06 218.757 1481.67 Q215.702 1486.25 209.891 1486.25 Q204.081 1486.25 201.003 1481.67 Q197.947 1477.06 197.947 1468.33 Q197.947 1459.58 201.003 1455 Q204.081 1450.39 209.891 1450.39 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M616.364 1481.64 L624.002 1481.64 L624.002 1455.28 L615.692 1456.95 L615.692 1452.69 L623.956 1451.02 L628.632 1451.02 L628.632 1481.64 L636.271 1481.64 L636.271 1485.58 L616.364 1485.58 L616.364 1481.64 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M1036.72 1481.64 L1053.04 1481.64 L1053.04 1485.58 L1031.1 1485.58 L1031.1 1481.64 Q1033.76 1478.89 1038.34 1474.26 Q1042.95 1469.61 1044.13 1468.27 Q1046.38 1465.74 1047.26 1464.01 Q1048.16 1462.25 1048.16 1460.56 Q1048.16 1457.8 1046.22 1456.07 Q1044.29 1454.33 1041.19 1454.33 Q1038.99 1454.33 1036.54 1455.09 Q1034.11 1455.86 1031.33 1457.41 L1031.33 1452.69 Q1034.16 1451.55 1036.61 1450.97 Q1039.06 1450.39 1041.1 1450.39 Q1046.47 1450.39 1049.66 1453.08 Q1052.86 1455.77 1052.86 1460.26 Q1052.86 1462.39 1052.05 1464.31 Q1051.26 1466.2 1049.16 1468.8 Q1048.58 1469.47 1045.47 1472.69 Q1042.37 1475.88 1036.72 1481.64 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M1462.41 1466.95 Q1465.77 1467.66 1467.64 1469.93 Q1469.54 1472.2 1469.54 1475.53 Q1469.54 1480.65 1466.02 1483.45 Q1462.5 1486.25 1456.02 1486.25 Q1453.84 1486.25 1451.53 1485.81 Q1449.24 1485.39 1446.78 1484.54 L1446.78 1480.02 Q1448.73 1481.16 1451.04 1481.74 Q1453.36 1482.32 1455.88 1482.32 Q1460.28 1482.32 1462.57 1480.58 Q1464.89 1478.84 1464.89 1475.53 Q1464.89 1472.48 1462.73 1470.77 Q1460.6 1469.03 1456.78 1469.03 L1452.76 1469.03 L1452.76 1465.19 L1456.97 1465.19 Q1460.42 1465.19 1462.25 1463.82 Q1464.08 1462.43 1464.08 1459.84 Q1464.08 1457.18 1462.18 1455.77 Q1460.3 1454.33 1456.78 1454.33 Q1454.86 1454.33 1452.66 1454.75 Q1450.47 1455.16 1447.83 1456.04 L1447.83 1451.88 Q1450.49 1451.14 1452.8 1450.77 Q1455.14 1450.39 1457.2 1450.39 Q1462.53 1450.39 1465.63 1452.83 Q1468.73 1455.23 1468.73 1459.35 Q1468.73 1462.22 1467.09 1464.21 Q1465.44 1466.18 1462.41 1466.95 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M1877.26 1455.09 L1865.46 1473.54 L1877.26 1473.54 L1877.26 1455.09 M1876.03 1451.02 L1881.91 1451.02 L1881.91 1473.54 L1886.84 1473.54 L1886.84 1477.43 L1881.91 1477.43 L1881.91 1485.58 L1877.26 1485.58 L1877.26 1477.43 L1861.66 1477.43 L1861.66 1472.92 L1876.03 1451.02 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M2280.62 1451.02 L2298.98 1451.02 L2298.98 1454.96 L2284.9 1454.96 L2284.9 1463.43 Q2285.92 1463.08 2286.94 1462.92 Q2287.96 1462.73 2288.98 1462.73 Q2294.76 1462.73 2298.14 1465.9 Q2301.52 1469.08 2301.52 1474.49 Q2301.52 1480.07 2298.05 1483.17 Q2294.58 1486.25 2288.26 1486.25 Q2286.08 1486.25 2283.81 1485.88 Q2281.57 1485.51 2279.16 1484.77 L2279.16 1480.07 Q2281.25 1481.2 2283.47 1481.76 Q2285.69 1482.32 2288.17 1482.32 Q2292.17 1482.32 2294.51 1480.21 Q2296.85 1478.1 2296.85 1474.49 Q2296.85 1470.88 2294.51 1468.77 Q2292.17 1466.67 2288.17 1466.67 Q2286.29 1466.67 2284.42 1467.08 Q2282.56 1467.5 2280.62 1468.38 L2280.62 1451.02 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M1198.43 1522.27 L1198.43 1532.4 L1210.49 1532.4 L1210.49 1536.95 L1198.43 1536.95 L1198.43 1556.3 Q1198.43 1560.66 1199.61 1561.9 Q1200.81 1563.14 1204.47 1563.14 L1210.49 1563.14 L1210.49 1568.04 L1204.47 1568.04 Q1197.7 1568.04 1195.12 1565.53 Q1192.54 1562.98 1192.54 1556.3 L1192.54 1536.95 L1188.24 1536.95 L1188.24 1532.4 L1192.54 1532.4 L1192.54 1522.27 L1198.43 1522.27 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M1238.37 1518.52 L1251.87 1518.52 L1251.87 1523.07 L1244.23 1523.07 L1244.23 1572.09 L1251.87 1572.09 L1251.87 1576.64 L1238.37 1576.64 L1238.37 1518.52 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M1287.07 1533.45 L1287.07 1538.98 Q1284.59 1537.71 1281.91 1537.07 Q1279.24 1536.44 1276.38 1536.44 Q1272.01 1536.44 1269.82 1537.77 Q1267.65 1539.11 1267.65 1541.79 Q1267.65 1543.82 1269.21 1545 Q1270.77 1546.15 1275.48 1547.2 L1277.49 1547.64 Q1283.73 1548.98 1286.34 1551.43 Q1288.98 1553.85 1288.98 1558.21 Q1288.98 1563.17 1285.03 1566.07 Q1281.12 1568.97 1274.24 1568.97 Q1271.38 1568.97 1268.26 1568.39 Q1265.17 1567.85 1261.73 1566.74 L1261.73 1560.69 Q1264.98 1562.38 1268.13 1563.24 Q1271.28 1564.07 1274.37 1564.07 Q1278.51 1564.07 1280.74 1562.66 Q1282.96 1561.23 1282.96 1558.65 Q1282.96 1556.27 1281.34 1554.99 Q1279.75 1553.72 1274.31 1552.54 L1272.27 1552.07 Q1266.83 1550.92 1264.41 1548.56 Q1261.99 1546.18 1261.99 1542.04 Q1261.99 1537.01 1265.55 1534.27 Q1269.12 1531.54 1275.68 1531.54 Q1278.92 1531.54 1281.79 1532.01 Q1284.65 1532.49 1287.07 1533.45 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M1311.99 1518.52 L1311.99 1576.64 L1298.5 1576.64 L1298.5 1572.09 L1306.1 1572.09 L1306.1 1523.07 L1298.5 1523.07 L1298.5 1518.52 L1311.99 1518.52 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip972)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  147.478,1383.72 2352.76,1383.72 
  "/>
<polyline clip-path="url(#clip972)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  147.478,1059.5 2352.76,1059.5 
  "/>
<polyline clip-path="url(#clip972)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  147.478,735.273 2352.76,735.273 
  "/>
<polyline clip-path="url(#clip972)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  147.478,411.05 2352.76,411.05 
  "/>
<polyline clip-path="url(#clip972)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  147.478,86.8263 2352.76,86.8263 
  "/>
<polyline clip-path="url(#clip970)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  147.478,1423.18 147.478,47.2441 
  "/>
<polyline clip-path="url(#clip970)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  147.478,1383.72 166.376,1383.72 
  "/>
<polyline clip-path="url(#clip970)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  147.478,1059.5 166.376,1059.5 
  "/>
<polyline clip-path="url(#clip970)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  147.478,735.273 166.376,735.273 
  "/>
<polyline clip-path="url(#clip970)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  147.478,411.05 166.376,411.05 
  "/>
<polyline clip-path="url(#clip970)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  147.478,86.8263 166.376,86.8263 
  "/>
<path clip-path="url(#clip970)" d="M51.3625 1384.17 L81.0383 1384.17 L81.0383 1388.11 L51.3625 1388.11 L51.3625 1384.17 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M95.1586 1397.06 L111.478 1397.06 L111.478 1401 L89.5336 1401 L89.5336 1397.06 Q92.1956 1394.31 96.7789 1389.68 Q101.385 1385.03 102.566 1383.68 Q104.811 1381.16 105.691 1379.43 Q106.594 1377.67 106.594 1375.98 Q106.594 1373.22 104.649 1371.49 Q102.728 1369.75 99.6261 1369.75 Q97.4271 1369.75 94.9734 1370.51 Q92.5428 1371.28 89.7651 1372.83 L89.7651 1368.11 Q92.5891 1366.97 95.0428 1366.39 Q97.4965 1365.81 99.5335 1365.81 Q104.904 1365.81 108.098 1368.5 Q111.293 1371.18 111.293 1375.68 Q111.293 1377.8 110.483 1379.73 Q109.696 1381.62 107.589 1384.22 Q107.01 1384.89 103.909 1388.11 Q100.807 1391.3 95.1586 1397.06 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M50.9921 1059.95 L80.6679 1059.95 L80.6679 1063.88 L50.9921 1063.88 L50.9921 1059.95 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M91.5706 1072.84 L99.2095 1072.84 L99.2095 1046.48 L90.8993 1048.14 L90.8993 1043.88 L99.1632 1042.22 L103.839 1042.22 L103.839 1072.84 L111.478 1072.84 L111.478 1076.78 L91.5706 1076.78 L91.5706 1072.84 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M99.5335 721.071 Q95.9224 721.071 94.0937 724.636 Q92.2882 728.178 92.2882 735.307 Q92.2882 742.414 94.0937 745.979 Q95.9224 749.52 99.5335 749.52 Q103.168 749.52 104.973 745.979 Q106.802 742.414 106.802 735.307 Q106.802 728.178 104.973 724.636 Q103.168 721.071 99.5335 721.071 M99.5335 717.368 Q105.344 717.368 108.399 721.974 Q111.478 726.558 111.478 735.307 Q111.478 744.034 108.399 748.641 Q105.344 753.224 99.5335 753.224 Q93.7234 753.224 90.6447 748.641 Q87.5892 744.034 87.5892 735.307 Q87.5892 726.558 90.6447 721.974 Q93.7234 717.368 99.5335 717.368 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M91.5706 424.394 L99.2095 424.394 L99.2095 398.029 L90.8993 399.695 L90.8993 395.436 L99.1632 393.77 L103.839 393.77 L103.839 424.394 L111.478 424.394 L111.478 428.33 L91.5706 428.33 L91.5706 424.394 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M95.1586 100.171 L111.478 100.171 L111.478 104.106 L89.5336 104.106 L89.5336 100.171 Q92.1956 97.4165 96.7789 92.7869 Q101.385 88.1342 102.566 86.7916 Q104.811 84.2685 105.691 82.5324 Q106.594 80.7731 106.594 79.0833 Q106.594 76.3287 104.649 74.5926 Q102.728 72.8565 99.6261 72.8565 Q97.4271 72.8565 94.9734 73.6204 Q92.5428 74.3843 89.7651 75.9352 L89.7651 71.213 Q92.5891 70.0787 95.0428 69.5 Q97.4965 68.9213 99.5335 68.9213 Q104.904 68.9213 108.098 71.6065 Q111.293 74.2917 111.293 78.7824 Q111.293 80.912 110.483 82.8333 Q109.696 84.7314 107.589 87.324 Q107.01 87.9953 103.909 91.2129 Q100.807 94.4073 95.1586 100.171 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip972)" style="stroke:#009af9; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  209.891,313.783 214.052,313.816 218.213,313.913 222.374,314.075 226.535,314.301 230.696,314.592 234.857,314.947 239.018,315.365 243.179,315.847 247.34,316.392 
  251.5,316.998 255.661,317.666 259.822,318.395 263.983,319.185 268.144,320.034 272.305,320.941 276.466,321.906 280.627,322.928 284.788,324.005 288.949,325.137 
  293.109,326.321 297.27,327.559 301.431,328.847 305.592,330.186 309.753,331.572 313.914,333.006 318.075,334.486 322.236,336.01 326.397,337.577 330.558,339.185 
  334.718,340.832 338.879,342.518 343.04,344.239 347.201,345.995 351.362,347.784 355.523,349.603 359.684,351.451 363.845,353.327 368.006,355.228 372.167,357.152 
  376.328,359.098 380.488,361.063 384.649,363.045 388.81,365.044 392.971,367.056 397.132,369.079 401.293,371.112 405.454,373.153 409.615,375.199 413.776,377.248 
  417.937,379.299 422.097,381.349 426.258,383.396 430.419,385.439 434.58,387.475 438.741,389.502 442.902,391.519 447.063,393.522 451.224,395.511 455.385,397.482 
  459.546,399.435 463.706,401.368 467.867,403.277 472.028,405.162 476.189,407.02 480.35,408.85 484.511,410.65 488.672,412.418 492.833,414.152 496.994,415.85 
  501.155,417.511 505.315,419.133 509.476,420.715 513.637,422.254 517.798,423.75 521.959,425.201 526.12,426.605 530.281,427.961 534.442,429.268 538.603,430.525 
  542.764,431.729 546.924,432.88 551.085,433.977 555.246,435.019 559.407,436.004 563.568,436.932 567.729,437.801 571.89,438.611 576.051,439.361 580.212,440.051 
  584.373,440.679 588.534,441.245 592.694,441.748 596.855,442.188 601.016,442.565 605.177,442.878 609.338,443.126 613.499,443.31 617.66,443.429 621.821,443.483 
  625.982,443.473 630.143,443.398 634.303,443.257 638.464,443.053 642.625,442.783 646.786,442.45 650.947,442.053 655.108,441.592 659.269,441.068 663.43,440.482 
  667.591,439.834 671.752,439.124 675.912,438.355 680.073,437.525 684.234,436.636 688.395,435.69 692.556,434.686 696.717,433.627 700.878,432.512 705.039,431.343 
  709.2,430.122 713.361,428.849 717.521,427.526 721.682,426.154 725.843,424.734 730.004,423.268 734.165,421.758 738.326,420.205 742.487,418.61 746.648,416.974 
  750.809,415.301 754.97,413.59 759.13,411.845 763.291,410.067 767.452,408.257 771.613,406.418 775.774,404.55 779.935,402.657 784.096,400.74 788.257,398.8 
  792.418,396.841 796.579,394.863 800.74,392.868 804.9,390.86 809.061,388.84 813.222,386.809 817.383,384.77 821.544,382.725 825.705,380.675 829.866,378.624 
  834.027,376.573 838.188,374.523 842.349,372.478 846.509,370.439 850.67,368.408 854.831,366.387 858.992,364.379 863.153,362.385 867.314,360.407 871.475,358.447 
  875.636,356.507 879.797,354.59 883.958,352.697 888.118,350.829 892.279,348.99 896.44,347.18 900.601,345.401 904.762,343.656 908.923,341.946 913.084,340.272 
  917.245,338.636 921.406,337.041 925.567,335.487 929.727,333.977 933.888,332.511 938.049,331.091 942.21,329.719 946.371,328.395 950.532,327.122 954.693,325.901 
  958.854,324.732 963.015,323.617 967.176,322.557 971.336,321.553 975.497,320.607 979.658,319.718 983.819,318.888 987.98,318.118 992.141,317.409 996.302,316.76 
  1000.46,316.174 1004.62,315.65 1008.78,315.189 1012.95,314.792 1017.11,314.458 1021.27,314.189 1025.43,313.984 1029.59,313.844 1033.75,313.768 1037.91,313.758 
  1042.07,313.812 1046.23,313.931 1050.39,314.115 1054.55,314.363 1058.72,314.676 1062.88,315.053 1067.04,315.493 1071.2,315.996 1075.36,316.562 1079.52,317.191 
  1083.68,317.88 1087.84,318.631 1092,319.441 1096.16,320.311 1100.32,321.239 1104.49,322.224 1108.65,323.266 1112.81,324.363 1116.97,325.514 1121.13,326.719 
  1125.29,327.975 1129.45,329.283 1133.61,330.639 1137.77,332.044 1141.93,333.495 1146.09,334.991 1150.26,336.531 1154.42,338.113 1158.58,339.736 1162.74,341.398 
  1166.9,343.096 1171.06,344.831 1175.22,346.599 1179.38,348.399 1183.54,350.23 1187.7,352.089 1191.86,353.974 1196.03,355.884 1200.19,357.817 1204.35,359.771 
  1208.51,361.743 1212.67,363.733 1216.83,365.737 1220.99,367.754 1225.15,369.782 1229.31,371.819 1233.47,373.863 1237.63,375.912 1241.8,377.963 1245.96,380.015 
  1250.12,382.066 1254.28,384.113 1258.44,386.155 1262.6,388.189 1266.76,390.213 1270.92,392.226 1275.08,394.226 1279.24,396.21 1283.4,398.176 1287.57,400.123 
  1291.73,402.048 1295.89,403.95 1300.05,405.826 1304.21,407.676 1308.37,409.496 1312.53,411.285 1316.69,413.042 1320.85,414.764 1325.01,416.45 1329.17,418.099 
  1333.33,419.707 1337.5,421.275 1341.66,422.8 1345.82,424.281 1349.98,425.716 1354.14,427.105 1358.3,428.444 1362.46,429.734 1366.62,430.973 1370.78,432.159 
  1374.94,433.292 1379.1,434.37 1383.27,435.392 1387.43,436.358 1391.59,437.266 1395.75,438.115 1399.91,438.905 1404.07,439.634 1408.23,440.302 1412.39,440.909 
  1416.55,441.454 1420.71,441.935 1424.87,442.353 1429.04,442.708 1433.2,442.998 1437.36,443.224 1441.52,443.385 1445.68,443.482 1449.84,443.514 1454,443.481 
  1458.16,443.382 1462.32,443.22 1466.48,442.992 1470.64,442.7 1474.81,442.344 1478.97,441.925 1483.13,441.442 1487.29,440.896 1491.45,440.288 1495.61,439.618 
  1499.77,438.887 1503.93,438.096 1508.09,437.245 1512.25,436.336 1516.41,435.369 1520.58,434.345 1524.74,433.266 1528.9,432.131 1533.06,430.944 1537.22,429.704 
  1541.38,428.413 1545.54,427.072 1549.7,425.682 1553.86,424.246 1558.02,422.764 1562.18,421.237 1566.35,419.668 1570.51,418.058 1574.67,416.409 1578.83,414.722 
  1582.99,412.999 1587.15,411.241 1591.31,409.45 1595.47,407.629 1599.63,405.779 1603.79,403.902 1607.95,401.999 1612.12,400.073 1616.28,398.125 1620.44,396.158 
  1624.6,394.174 1628.76,392.174 1632.92,390.16 1637.08,388.135 1641.24,386.1 1645.4,384.058 1649.56,382.01 1653.72,379.959 1657.89,377.907 1662.05,375.855 
  1666.21,373.806 1670.37,371.762 1674.53,369.725 1678.69,367.696 1682.85,365.679 1687.01,363.674 1691.17,361.685 1695.33,359.712 1699.49,357.759 1703.66,355.826 
  1707.82,353.916 1711.98,352.03 1716.14,350.171 1720.3,348.341 1724.46,346.542 1728.62,344.774 1732.78,343.039 1736.94,341.341 1741.1,339.68 1745.26,338.058 
  1749.43,336.476 1753.59,334.937 1757.75,333.441 1761.91,331.99 1766.07,330.586 1770.23,329.231 1774.39,327.924 1778.55,326.668 1782.71,325.465 1786.87,324.314 
  1791.03,323.218 1795.2,322.177 1799.36,321.193 1803.52,320.266 1807.68,319.397 1811.84,318.588 1816,317.839 1820.16,317.15 1824.32,316.523 1828.48,315.958 
  1832.64,315.456 1836.8,315.017 1840.96,314.642 1845.13,314.331 1849.29,314.084 1853.45,313.901 1857.61,313.783 1861.77,313.731 1865.93,313.743 1870.09,313.82 
  1874.25,313.962 1878.41,314.168 1882.57,314.439 1886.73,314.774 1890.9,315.173 1895.06,315.636 1899.22,316.162 1903.38,316.75 1907.54,317.4 1911.7,318.111 
  1915.86,318.883 1920.02,319.714 1924.18,320.605 1928.34,321.553 1932.5,322.559 1936.67,323.62 1940.83,324.737 1944.99,325.908 1949.15,327.131 1953.31,328.406 
  1957.47,329.731 1961.63,331.105 1965.79,332.527 1969.95,333.995 1974.11,335.507 1978.27,337.063 1982.44,338.66 1986.6,340.297 1990.76,341.972 1994.92,343.684 
  1999.08,345.431 2003.24,347.212 2007.4,349.023 2011.56,350.865 2015.72,352.734 2019.88,354.629 2024.04,356.548 2028.21,358.489 2032.37,360.45 2036.53,362.43 
  2040.69,364.426 2044.85,366.436 2049.01,368.458 2053.17,370.49 2057.33,372.53 2061.49,374.577 2065.65,376.628 2069.81,378.68 2073.98,380.733 2078.14,382.783 
  2082.3,384.829 2086.46,386.87 2090.62,388.902 2094.78,390.923 2098.94,392.933 2103.1,394.928 2107.26,396.906 2111.42,398.867 2115.58,400.807 2119.75,402.725 
  2123.91,404.619 2128.07,406.487 2132.23,408.327 2136.39,410.137 2140.55,411.916 2144.71,413.662 2148.87,415.373 2153.03,417.046 2157.19,418.682 2161.35,420.277 
  2165.52,421.831 2169.68,423.341 2173.84,424.807 2178,426.226 2182.16,427.598 2186.32,428.921 2190.48,430.194 2194.64,431.415 2198.8,432.583 2202.96,433.697 
  2207.12,434.757 2211.29,435.76 2215.45,436.706 2219.61,437.593 2223.77,438.422 2227.93,439.191 2232.09,439.9 2236.25,440.547 2240.41,441.132 2244.57,441.655 
  2248.73,442.114 2252.89,442.51 2257.06,442.843 2261.22,443.111 2265.38,443.314 2269.54,443.453 2273.7,443.527 2277.86,443.536 2282.02,443.48 2286.18,443.359 
  2290.34,443.174 
  "/>
<polyline clip-path="url(#clip972)" style="stroke:#e26f46; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  209.891,735.273 214.052,741.756 218.213,748.231 222.374,754.694 226.535,761.138 230.696,767.557 234.857,773.944 239.018,780.292 243.179,786.595 247.34,792.847 
  251.5,799.041 255.661,805.171 259.822,811.232 263.983,817.216 268.144,823.119 272.305,828.933 276.466,834.655 280.627,840.276 284.788,845.794 288.949,851.201 
  293.109,856.493 297.27,861.665 301.431,866.709 305.592,871.621 309.753,876.398 313.914,881.034 318.075,885.523 322.236,889.862 326.397,894.046 330.558,898.071 
  334.718,901.934 338.879,905.629 343.04,909.154 347.201,912.506 351.362,915.68 355.523,918.673 359.684,921.484 363.845,924.108 368.006,926.543 372.167,928.787 
  376.328,930.838 380.488,932.693 384.649,934.35 388.81,935.809 392.971,937.067 397.132,938.123 401.293,938.976 405.454,939.626 409.615,940.071 413.776,940.312 
  417.937,940.347 422.097,940.177 426.258,939.803 430.419,939.224 434.58,938.441 438.741,937.455 442.902,936.266 447.063,934.877 451.224,933.288 455.385,931.501 
  459.546,929.517 463.706,927.34 467.867,924.97 472.028,922.41 476.189,919.664 480.35,916.733 484.511,913.62 488.672,910.328 492.833,906.862 496.994,903.224 
  501.155,899.418 505.315,895.448 509.476,891.317 513.637,887.03 517.798,882.592 521.959,878.006 526.12,873.278 530.281,868.411 534.442,863.411 538.603,858.282 
  542.764,853.031 546.924,847.662 551.085,842.181 555.246,836.592 559.407,830.902 563.568,825.117 567.729,819.241 571.89,813.282 576.051,807.245 580.212,801.135 
  584.373,794.959 588.534,788.724 592.694,782.435 596.855,776.1 601.016,769.724 605.177,763.312 609.338,756.872 613.499,750.411 617.66,743.935 621.821,737.452 
  625.982,730.964 630.143,724.481 634.303,718.009 638.464,711.554 642.625,705.124 646.786,698.723 650.947,692.359 655.108,686.037 659.269,679.764 663.43,673.548 
  667.591,667.393 671.752,661.306 675.912,655.293 680.073,649.359 684.234,643.512 688.395,637.757 692.556,632.099 696.717,626.545 700.878,621.098 705.039,615.766 
  709.2,610.554 713.361,605.466 717.521,600.509 721.682,595.686 725.843,591.002 730.004,586.463 734.165,582.073 738.326,577.836 742.487,573.757 746.648,569.839 
  750.809,566.086 754.97,562.503 759.13,559.092 763.291,555.859 767.452,552.804 771.613,549.932 775.774,547.245 779.935,544.746 784.096,542.438 788.257,540.323 
  792.418,538.402 796.579,536.679 800.74,535.154 804.9,533.829 809.061,532.707 813.222,531.787 817.383,531.07 821.544,530.557 825.705,530.249 829.866,530.146 
  834.027,530.249 838.188,530.557 842.349,531.069 846.509,531.786 850.67,532.706 854.831,533.829 858.992,535.153 863.153,536.678 867.314,538.401 871.475,540.321 
  875.636,542.436 879.797,544.744 883.958,547.243 888.118,549.93 892.279,552.802 896.44,555.857 900.601,559.091 904.762,562.502 908.923,566.085 913.084,569.838 
  917.245,573.756 921.406,577.836 925.567,582.073 929.727,586.463 933.888,591.002 938.049,595.686 942.21,600.51 946.371,605.468 950.532,610.556 954.693,615.768 
  958.854,621.101 963.015,626.548 967.176,632.103 971.336,637.762 975.497,643.518 979.658,649.365 983.819,655.3 987.98,661.314 992.141,667.402 996.302,673.558 
  1000.46,679.776 1004.62,686.048 1008.78,692.371 1012.95,698.737 1017.11,705.139 1021.27,711.571 1025.43,718.027 1029.59,724.499 1033.75,730.984 1037.91,737.472 
  1042.07,743.959 1046.23,750.436 1050.39,756.898 1054.55,763.338 1058.72,769.751 1062.88,776.13 1067.04,782.468 1071.2,788.758 1075.36,794.994 1079.52,801.171 
  1083.68,807.282 1087.84,813.322 1092,819.283 1096.16,825.159 1100.32,830.946 1104.49,836.638 1108.65,842.228 1112.81,847.711 1116.97,853.082 1121.13,858.335 
  1125.29,863.464 1129.45,868.466 1133.61,873.334 1137.77,878.065 1141.93,882.652 1146.09,887.092 1150.26,891.38 1154.42,895.512 1158.58,899.484 1162.74,903.292 
  1166.9,906.931 1171.06,910.398 1175.22,913.691 1179.38,916.805 1183.54,919.738 1187.7,922.486 1191.86,925.046 1196.03,927.417 1200.19,929.596 1204.35,931.58 
  1208.51,933.368 1212.67,934.958 1216.83,936.348 1220.99,937.537 1225.15,938.523 1229.31,939.307 1233.47,939.886 1237.63,940.261 1241.8,940.43 1245.96,940.394 
  1250.12,940.153 1254.28,939.707 1258.44,939.057 1262.6,938.203 1266.76,937.145 1270.92,935.886 1275.08,934.426 1279.24,932.767 1283.4,930.91 1287.57,928.858 
  1291.73,926.612 1295.89,924.175 1300.05,921.548 1304.21,918.735 1308.37,915.739 1312.53,912.562 1316.69,909.208 1320.85,905.68 1325.01,901.981 1329.17,898.115 
  1333.33,894.087 1337.5,889.9 1341.66,885.558 1345.82,881.065 1349.98,876.427 1354.14,871.647 1358.3,866.732 1362.46,861.685 1366.62,856.511 1370.78,851.215 
  1374.94,845.804 1379.1,840.282 1383.27,834.656 1387.43,828.93 1391.59,823.11 1395.75,817.201 1399.91,811.211 1404.07,805.146 1408.23,799.01 1412.39,792.811 
  1416.55,786.554 1420.71,780.245 1424.87,773.891 1429.04,767.499 1433.2,761.075 1437.36,754.625 1441.52,748.155 1445.68,741.672 1449.84,735.183 1454,728.695 
  1458.16,722.213 1462.32,715.744 1466.48,709.294 1470.64,702.869 1474.81,696.478 1478.97,690.126 1483.13,683.819 1487.29,677.562 1491.45,671.364 1495.61,665.229 
  1499.77,659.164 1503.93,653.177 1508.09,647.271 1512.25,641.452 1516.41,635.728 1520.58,630.102 1524.74,624.583 1528.9,619.175 1533.06,613.882 1537.22,608.71 
  1541.38,603.665 1545.54,598.752 1549.7,593.975 1553.86,589.341 1558.02,584.851 1562.18,580.512 1566.35,576.327 1570.51,572.302 1574.67,568.44 1578.83,564.746 
  1582.99,561.221 1587.15,557.87 1591.31,554.697 1595.47,551.704 1599.63,548.895 1603.79,546.273 1607.95,543.84 1612.12,541.597 1616.28,539.549 1620.44,537.697 
  1624.6,536.042 1628.76,534.587 1632.92,533.332 1637.08,532.279 1641.24,531.429 1645.4,530.784 1649.56,530.343 1653.72,530.106 1657.89,530.075 1662.05,530.249 
  1666.21,530.629 1670.37,531.213 1674.53,532.001 1678.69,532.993 1682.85,534.187 1687.01,535.582 1691.17,537.176 1695.33,538.969 1699.49,540.959 1703.66,543.143 
  1707.82,545.519 1711.98,548.084 1716.14,550.837 1720.3,553.775 1724.46,556.894 1728.62,560.192 1732.78,563.665 1736.94,567.309 1741.1,571.121 1745.26,575.098 
  1749.43,579.235 1753.59,583.528 1757.75,587.973 1761.91,592.565 1766.07,597.3 1770.23,602.172 1774.39,607.179 1778.55,612.313 1782.71,617.571 1786.87,622.946 
  1791.03,628.433 1795.2,634.027 1799.36,639.723 1803.52,645.515 1807.68,651.396 1811.84,657.361 1816,663.404 1820.16,669.518 1824.32,675.7 1828.48,681.94 
  1832.64,688.234 1836.8,694.575 1840.96,700.955 1845.13,707.372 1849.29,713.816 1853.45,720.281 1857.61,726.762 1861.77,733.25 1865.93,739.741 1870.09,746.227 
  1874.25,752.703 1878.41,759.162 1882.57,765.596 1886.73,772 1890.9,778.366 1895.06,784.691 1899.22,790.966 1903.38,797.185 1907.54,803.343 1911.7,809.431 
  1915.86,815.446 1920.02,821.381 1924.18,827.23 1928.34,832.987 1932.5,838.646 1936.67,844.2 1940.83,849.647 1944.99,854.98 1949.15,860.193 1953.31,865.28 
  1957.47,870.238 1961.63,875.06 1965.79,879.743 1969.95,884.281 1974.11,888.671 1978.27,892.907 1982.44,896.984 1986.6,900.9 1990.76,904.651 1994.92,908.233 
  1999.08,911.641 2003.24,914.873 2007.4,917.925 2011.56,920.794 2015.72,923.479 2019.88,925.974 2024.04,928.28 2028.21,930.391 2032.37,932.308 2036.53,934.027 
  2040.69,935.548 2044.85,936.869 2049.01,937.988 2053.17,938.904 2057.33,939.616 2061.49,940.124 2065.65,940.427 2069.81,940.525 2073.98,940.417 2078.14,940.104 
  2082.3,939.587 2086.46,938.865 2090.62,937.939 2094.78,936.811 2098.94,935.48 2103.1,933.95 2107.26,932.221 2111.42,930.295 2115.58,928.174 2119.75,925.86 
  2123.91,923.355 2128.07,920.662 2132.23,917.783 2136.39,914.722 2140.55,911.482 2144.71,908.065 2148.87,904.475 2153.03,900.717 2157.19,896.792 2161.35,892.706 
  2165.52,888.462 2169.68,884.065 2173.84,879.52 2178,874.83 2182.16,870 2186.32,865.035 2190.48,859.941 2194.64,854.722 2198.8,849.384 2202.96,843.932 
  2207.12,838.37 2211.29,832.705 2215.45,826.943 2219.61,821.089 2223.77,815.151 2227.93,809.131 2232.09,803.037 2236.25,796.875 2240.41,790.652 2244.57,784.374 
  2248.73,778.048 2252.89,771.677 2257.06,765.27 2261.22,758.833 2265.38,752.372 2269.54,745.896 2273.7,739.408 2277.86,732.915 2282.02,726.425 2286.18,719.943 
  2290.34,713.478 
  "/>
<polyline clip-path="url(#clip972)" style="stroke:#3da44d; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  209.891,1383.72 214.052,1383.39 218.213,1382.41 222.374,1380.8 226.535,1378.53 230.696,1375.62 234.857,1372.08 239.018,1367.89 243.179,1363.07 247.34,1357.63 
  251.5,1351.56 255.661,1344.88 259.822,1337.59 263.983,1329.7 268.144,1321.21 272.305,1312.13 276.466,1302.48 280.627,1292.27 284.788,1281.49 288.949,1270.18 
  293.109,1258.33 297.27,1245.96 301.431,1233.08 305.592,1219.69 309.753,1205.82 313.914,1191.48 318.075,1176.68 322.236,1161.44 326.397,1145.77 330.558,1129.69 
  334.718,1113.22 338.879,1096.37 343.04,1079.15 347.201,1061.59 351.362,1043.71 355.523,1025.51 359.684,1007.03 363.845,988.275 368.006,969.268 372.167,950.026 
  376.328,930.569 380.488,910.918 384.649,891.091 388.81,871.108 392.971,850.989 397.132,830.755 401.293,810.425 405.454,790.02 409.615,769.56 413.776,749.066 
  417.937,728.558 422.097,708.056 426.258,687.582 430.419,667.156 434.58,646.795 438.741,626.523 442.902,606.359 447.063,586.325 451.224,566.44 455.385,546.724 
  459.546,527.193 463.706,507.869 467.867,488.774 472.028,469.926 476.189,451.345 480.35,433.046 484.511,415.048 488.672,397.369 492.833,380.03 496.994,363.047 
  501.155,346.438 505.315,330.216 509.476,314.397 513.637,299 517.798,284.04 521.959,269.533 526.12,255.493 530.281,241.93 534.442,228.86 538.603,216.298 
  542.764,204.254 546.924,192.744 551.085,181.776 555.246,171.36 559.407,161.508 563.568,152.23 567.729,143.535 571.89,135.434 576.051,127.933 580.212,121.037 
  584.373,114.756 588.534,109.096 592.694,104.062 596.855,99.6616 601.016,95.8953 605.177,92.7681 609.338,90.2836 613.499,88.4446 617.66,87.2531 621.821,86.7104 
  625.982,86.8153 630.143,87.5689 634.303,88.9705 638.464,91.0188 642.625,93.7117 646.786,97.0466 650.947,101.02 655.108,105.627 659.269,110.864 663.43,116.726 
  667.591,123.207 671.752,130.3 675.912,137.999 680.073,146.295 684.234,155.18 688.395,164.645 692.556,174.681 696.717,185.278 700.878,196.426 705.039,208.113 
  709.2,220.326 713.361,233.055 717.521,246.285 721.682,260.006 725.843,274.203 730.004,288.861 734.165,303.965 738.326,319.499 742.487,335.45 746.648,351.802 
  750.809,368.538 754.97,385.64 759.13,403.092 763.291,420.875 767.452,438.973 771.613,457.369 775.774,476.044 779.935,494.977 784.096,514.149 788.257,533.542 
  792.418,553.138 796.579,572.918 800.74,592.86 804.9,612.944 809.061,633.149 813.222,653.455 817.383,673.847 821.544,694.3 825.705,714.794 829.866,735.307 
  834.027,755.819 838.188,776.31 842.349,796.764 846.509,817.156 850.67,837.465 854.831,857.672 858.992,877.754 863.153,897.695 867.314,917.476 871.475,937.074 
  875.636,956.47 879.797,975.644 883.958,994.575 888.118,1013.25 892.279,1031.65 896.44,1049.75 900.601,1067.54 904.762,1084.99 908.923,1102.09 913.084,1118.83 
  917.245,1135.18 921.406,1151.14 925.567,1166.67 929.727,1181.78 933.888,1196.43 938.049,1210.63 942.21,1224.36 946.371,1237.59 950.532,1250.32 954.693,1262.54 
  958.854,1274.22 963.015,1285.37 967.176,1295.97 971.336,1306.01 975.497,1315.48 979.658,1324.36 983.819,1332.66 987.98,1340.36 992.141,1347.46 996.302,1353.94 
  1000.46,1359.8 1004.62,1365.04 1008.78,1369.65 1012.95,1373.63 1017.11,1376.96 1021.27,1379.66 1025.43,1381.71 1029.59,1383.11 1033.75,1383.86 1037.91,1383.97 
  1042.07,1383.43 1046.23,1382.23 1050.39,1380.4 1054.55,1377.91 1058.72,1374.78 1062.88,1371.02 1067.04,1366.62 1071.2,1361.58 1075.36,1355.92 1079.52,1349.64 
  1083.68,1342.74 1087.84,1335.24 1092,1327.13 1096.16,1318.44 1100.32,1309.16 1104.49,1299.31 1108.65,1288.89 1112.81,1277.92 1116.97,1266.4 1121.13,1254.36 
  1125.29,1241.79 1129.45,1228.72 1133.61,1215.15 1137.77,1201.11 1141.93,1186.59 1146.09,1171.63 1150.26,1156.23 1154.42,1140.41 1158.58,1124.18 1162.74,1107.57 
  1166.9,1090.58 1171.06,1073.24 1175.22,1055.56 1179.38,1037.55 1183.54,1019.25 1187.7,1000.66 1191.86,981.806 1196.03,962.707 1200.19,943.377 1204.35,923.839 
  1208.51,904.113 1212.67,884.218 1216.83,864.176 1220.99,844.005 1225.15,823.722 1229.31,803.351 1233.47,782.912 1237.63,762.427 1241.8,741.915 1245.96,721.396 
  1250.12,700.889 1254.28,680.416 1258.44,659.998 1262.6,639.657 1266.76,619.413 1270.92,599.283 1275.08,579.287 1279.24,559.447 1283.4,539.784 1287.57,520.318 
  1291.73,501.068 1295.89,482.049 1300.05,463.283 1304.21,444.789 1308.37,426.587 1312.53,408.694 1316.69,391.129 1320.85,373.905 1325.01,357.042 1329.17,340.559 
  1333.33,324.47 1337.5,308.795 1341.66,293.545 1345.82,278.734 1349.98,264.381 1354.14,250.499 1358.3,237.103 1362.46,224.207 1366.62,211.819 1370.78,199.954 
  1374.94,188.626 1379.1,177.844 1383.27,167.621 1387.43,157.967 1391.59,148.887 1395.75,140.395 1399.91,132.497 1404.07,125.203 1408.23,118.52 1412.39,112.454 
  1416.55,107.01 1420.71,102.194 1424.87,98.0115 1429.04,94.467 1433.2,91.5642 1437.36,89.3048 1441.52,87.6908 1445.68,86.7247 1449.84,86.4076 1454,86.7398 
  1458.16,87.7212 1462.32,89.3498 1466.48,91.6244 1470.64,94.5429 1474.81,98.1024 1478.97,102.299 1483.13,107.13 1487.29,112.588 1491.45,118.67 1495.61,125.368 
  1499.77,132.677 1503.93,140.588 1508.09,149.094 1512.25,158.187 1516.41,167.857 1520.58,178.095 1524.74,188.889 1528.9,200.23 1533.06,212.107 1537.22,224.508 
  1541.38,237.419 1545.54,250.829 1549.7,264.723 1553.86,279.086 1558.02,293.907 1562.18,309.171 1566.35,324.861 1570.51,340.96 1574.67,357.454 1578.83,374.324 
  1582.99,391.558 1587.15,409.136 1591.31,427.041 1595.47,445.253 1599.63,463.753 1603.79,482.526 1607.95,501.554 1612.12,520.815 1616.28,540.291 1620.44,559.961 
  1624.6,579.805 1628.76,599.806 1632.92,619.944 1637.08,640.197 1641.24,660.546 1645.4,680.967 1649.56,701.442 1653.72,721.953 1657.89,742.478 1662.05,762.997 
  1666.21,783.486 1670.37,803.926 1674.53,824.297 1678.69,844.581 1682.85,864.757 1687.01,884.803 1691.17,904.699 1695.33,924.423 1699.49,943.958 1703.66,963.288 
  1707.82,982.389 1711.98,1001.24 1716.14,1019.83 1720.3,1038.13 1724.46,1056.13 1728.62,1073.81 1732.78,1091.15 1736.94,1108.14 1741.1,1124.75 1745.26,1140.97 
  1749.43,1156.78 1753.59,1172.18 1757.75,1187.14 1761.91,1201.64 1766.07,1215.68 1770.23,1229.24 1774.39,1242.3 1778.55,1254.86 1782.71,1266.9 1786.87,1278.41 
  1791.03,1289.37 1795.2,1299.77 1799.36,1309.62 1803.52,1318.89 1807.68,1327.58 1811.84,1335.67 1816,1343.16 1820.16,1350.04 1824.32,1356.31 1828.48,1361.96 
  1832.64,1366.98 1836.8,1371.37 1840.96,1375.12 1845.13,1378.24 1849.29,1380.71 1853.45,1382.53 1857.61,1383.71 1861.77,1384.24 1865.93,1384.12 1870.09,1383.35 
  1874.25,1381.93 1878.41,1379.86 1882.57,1377.15 1886.73,1373.8 1890.9,1369.81 1895.06,1365.18 1899.22,1359.93 1903.38,1354.05 1907.54,1347.55 1911.7,1340.44 
  1915.86,1332.72 1920.02,1324.4 1924.18,1315.5 1928.34,1306.01 1932.5,1295.96 1936.67,1285.34 1940.83,1274.17 1944.99,1262.47 1949.15,1250.23 1953.31,1237.48 
  1957.47,1224.23 1961.63,1210.49 1965.79,1196.28 1969.95,1181.6 1974.11,1166.47 1978.27,1150.92 1982.44,1134.95 1986.6,1118.58 1990.76,1101.82 1994.92,1084.7 
  1999.08,1067.23 2003.24,1049.43 2007.4,1031.31 2011.56,1012.9 2015.72,994.206 2019.88,975.255 2024.04,956.064 2028.21,936.653 2032.37,917.043 2036.53,897.247 
  2040.69,877.289 2044.85,857.189 2049.01,836.967 2053.17,816.645 2057.33,796.243 2061.49,775.776 2065.65,755.269 2069.81,734.742 2073.98,714.216 2078.14,693.713 
  2082.3,673.251 2086.46,652.848 2090.62,632.528 2094.78,612.311 2098.94,592.218 2103.1,572.27 2107.26,552.483 2111.42,532.877 2115.58,513.474 2119.75,494.293 
  2123.91,475.354 2128.07,456.677 2132.23,438.275 2136.39,420.17 2140.55,402.38 2144.71,384.924 2148.87,367.82 2153.03,351.084 2157.19,334.729 2161.35,318.775 
  2165.52,303.238 2169.68,288.133 2173.84,273.478 2178,259.284 2182.16,245.563 2186.32,232.333 2190.48,219.606 2194.64,207.396 2198.8,195.715 2202.96,184.572 
  2207.12,173.98 2211.29,163.948 2215.45,154.489 2219.61,145.611 2223.77,137.324 2227.93,129.634 2232.09,122.549 2236.25,116.076 2240.41,110.224 2244.57,104.998 
  2248.73,100.403 2252.89,96.4406 2257.06,93.1178 2261.22,90.4374 2265.38,88.4024 2269.54,87.0151 2273.7,86.2762 2277.86,86.1857 2282.02,86.7445 2286.18,87.9522 
  2290.34,89.8076 
  "/>
<path clip-path="url(#clip970)" d="
M1550.33 300.469 L2279.25 300.469 L2279.25 93.1086 L1550.33 93.1086  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<polyline clip-path="url(#clip970)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1550.33,300.469 2279.25,300.469 2279.25,93.1086 1550.33,93.1086 1550.33,300.469 
  "/>
<polyline clip-path="url(#clip970)" style="stroke:#009af9; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1574.83,144.949 1721.85,144.949 
  "/>
<path clip-path="url(#clip970)" d="M1766.7 141.28 Q1768.3 138.409 1770.52 137.044 Q1772.74 135.678 1775.75 135.678 Q1779.8 135.678 1782 138.525 Q1784.2 141.349 1784.2 146.581 L1784.2 162.229 L1779.92 162.229 L1779.92 146.719 Q1779.92 142.993 1778.6 141.187 Q1777.28 139.382 1774.57 139.382 Q1771.26 139.382 1769.34 141.581 Q1767.42 143.78 1767.42 147.576 L1767.42 162.229 L1763.14 162.229 L1763.14 146.719 Q1763.14 142.969 1761.82 141.187 Q1760.5 139.382 1757.74 139.382 Q1754.48 139.382 1752.56 141.604 Q1750.64 143.803 1750.64 147.576 L1750.64 162.229 L1746.35 162.229 L1746.35 136.303 L1750.64 136.303 L1750.64 140.331 Q1752.09 137.946 1754.13 136.812 Q1756.17 135.678 1758.97 135.678 Q1761.79 135.678 1763.76 137.113 Q1765.75 138.548 1766.7 141.28 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M1804.48 149.196 Q1799.32 149.196 1797.33 150.377 Q1795.34 151.557 1795.34 154.405 Q1795.34 156.673 1796.82 158.016 Q1798.32 159.335 1800.89 159.335 Q1804.43 159.335 1806.56 156.835 Q1808.71 154.312 1808.71 150.145 L1808.71 149.196 L1804.48 149.196 M1812.97 147.437 L1812.97 162.229 L1808.71 162.229 L1808.71 158.293 Q1807.26 160.655 1805.08 161.789 Q1802.9 162.9 1799.76 162.9 Q1795.78 162.9 1793.41 160.678 Q1791.08 158.432 1791.08 154.682 Q1791.08 150.307 1793.99 148.085 Q1796.93 145.863 1802.74 145.863 L1808.71 145.863 L1808.71 145.446 Q1808.71 142.507 1806.77 140.909 Q1804.85 139.289 1801.35 139.289 Q1799.13 139.289 1797.03 139.821 Q1794.92 140.354 1792.97 141.419 L1792.97 137.483 Q1795.31 136.581 1797.51 136.141 Q1799.71 135.678 1801.79 135.678 Q1807.42 135.678 1810.2 138.594 Q1812.97 141.511 1812.97 147.437 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M1838.27 137.067 L1838.27 141.094 Q1836.47 140.169 1834.52 139.706 Q1832.58 139.243 1830.5 139.243 Q1827.33 139.243 1825.73 140.215 Q1824.15 141.187 1824.15 143.131 Q1824.15 144.613 1825.29 145.469 Q1826.42 146.303 1829.85 147.067 L1831.31 147.391 Q1835.84 148.363 1837.74 150.145 Q1839.66 151.905 1839.66 155.076 Q1839.66 158.687 1836.79 160.793 Q1833.95 162.9 1828.95 162.9 Q1826.86 162.9 1824.59 162.483 Q1822.35 162.09 1819.85 161.28 L1819.85 156.881 Q1822.21 158.108 1824.5 158.733 Q1826.79 159.335 1829.04 159.335 Q1832.05 159.335 1833.67 158.317 Q1835.29 157.275 1835.29 155.4 Q1835.29 153.664 1834.11 152.738 Q1832.95 151.812 1828.99 150.956 L1827.51 150.608 Q1823.55 149.775 1821.79 148.062 Q1820.03 146.326 1820.03 143.317 Q1820.03 139.659 1822.63 137.669 Q1825.22 135.678 1829.99 135.678 Q1832.35 135.678 1834.43 136.025 Q1836.52 136.372 1838.27 137.067 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M1862.97 137.067 L1862.97 141.094 Q1861.17 140.169 1859.22 139.706 Q1857.28 139.243 1855.2 139.243 Q1852.02 139.243 1850.43 140.215 Q1848.85 141.187 1848.85 143.131 Q1848.85 144.613 1849.99 145.469 Q1851.12 146.303 1854.55 147.067 L1856.01 147.391 Q1860.54 148.363 1862.44 150.145 Q1864.36 151.905 1864.36 155.076 Q1864.36 158.687 1861.49 160.793 Q1858.65 162.9 1853.65 162.9 Q1851.56 162.9 1849.29 162.483 Q1847.05 162.09 1844.55 161.28 L1844.55 156.881 Q1846.91 158.108 1849.2 158.733 Q1851.49 159.335 1853.74 159.335 Q1856.75 159.335 1858.37 158.317 Q1859.99 157.275 1859.99 155.4 Q1859.99 153.664 1858.81 152.738 Q1857.65 151.812 1853.69 150.956 L1852.21 150.608 Q1848.25 149.775 1846.49 148.062 Q1844.73 146.326 1844.73 143.317 Q1844.73 139.659 1847.33 137.669 Q1849.92 135.678 1854.69 135.678 Q1857.05 135.678 1859.13 136.025 Q1861.21 136.372 1862.97 137.067 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M1871.75 156.349 L1876.63 156.349 L1876.63 162.229 L1871.75 162.229 L1871.75 156.349 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M1902.74 137.067 L1902.74 141.094 Q1900.94 140.169 1898.99 139.706 Q1897.05 139.243 1894.96 139.243 Q1891.79 139.243 1890.2 140.215 Q1888.62 141.187 1888.62 143.131 Q1888.62 144.613 1889.76 145.469 Q1890.89 146.303 1894.32 147.067 L1895.77 147.391 Q1900.31 148.363 1902.21 150.145 Q1904.13 151.905 1904.13 155.076 Q1904.13 158.687 1901.26 160.793 Q1898.41 162.9 1893.41 162.9 Q1891.33 162.9 1889.06 162.483 Q1886.82 162.09 1884.32 161.28 L1884.32 156.881 Q1886.68 158.108 1888.97 158.733 Q1891.26 159.335 1893.51 159.335 Q1896.52 159.335 1898.14 158.317 Q1899.76 157.275 1899.76 155.4 Q1899.76 153.664 1898.58 152.738 Q1897.42 151.812 1893.46 150.956 L1891.98 150.608 Q1888.02 149.775 1886.26 148.062 Q1884.5 146.326 1884.5 143.317 Q1884.5 139.659 1887.09 137.669 Q1889.69 135.678 1894.46 135.678 Q1896.82 135.678 1898.9 136.025 Q1900.98 136.372 1902.74 137.067 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M1936.21 126.257 Q1933.11 131.581 1931.61 136.789 Q1930.1 141.997 1930.1 147.344 Q1930.1 152.692 1931.61 157.946 Q1933.14 163.178 1936.21 168.479 L1932.51 168.479 Q1929.04 163.039 1927.3 157.784 Q1925.59 152.53 1925.59 147.344 Q1925.59 142.182 1927.3 136.951 Q1929.02 131.72 1932.51 126.257 L1936.21 126.257 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M1959.25 143.594 Q1962.6 144.312 1964.48 146.581 Q1966.38 148.849 1966.38 152.182 Q1966.38 157.298 1962.86 160.099 Q1959.34 162.9 1952.86 162.9 Q1950.68 162.9 1948.37 162.46 Q1946.08 162.043 1943.62 161.187 L1943.62 156.673 Q1945.57 157.807 1947.88 158.386 Q1950.2 158.965 1952.72 158.965 Q1957.12 158.965 1959.41 157.229 Q1961.72 155.493 1961.72 152.182 Q1961.72 149.127 1959.57 147.414 Q1957.44 145.678 1953.62 145.678 L1949.59 145.678 L1949.59 141.835 L1953.81 141.835 Q1957.26 141.835 1959.08 140.469 Q1960.91 139.081 1960.91 136.488 Q1960.91 133.826 1959.02 132.414 Q1957.14 130.979 1953.62 130.979 Q1951.7 130.979 1949.5 131.395 Q1947.3 131.812 1944.66 132.692 L1944.66 128.525 Q1947.33 127.784 1949.64 127.414 Q1951.98 127.044 1954.04 127.044 Q1959.36 127.044 1962.46 129.474 Q1965.57 131.882 1965.57 136.002 Q1965.57 138.872 1963.92 140.863 Q1962.28 142.831 1959.25 143.594 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M1989.41 143.594 Q1992.76 144.312 1994.64 146.581 Q1996.54 148.849 1996.54 152.182 Q1996.54 157.298 1993.02 160.099 Q1989.5 162.9 1983.02 162.9 Q1980.84 162.9 1978.53 162.46 Q1976.24 162.043 1973.78 161.187 L1973.78 156.673 Q1975.73 157.807 1978.04 158.386 Q1980.36 158.965 1982.88 158.965 Q1987.28 158.965 1989.57 157.229 Q1991.89 155.493 1991.89 152.182 Q1991.89 149.127 1989.73 147.414 Q1987.6 145.678 1983.78 145.678 L1979.76 145.678 L1979.76 141.835 L1983.97 141.835 Q1987.42 141.835 1989.25 140.469 Q1991.08 139.081 1991.08 136.488 Q1991.08 133.826 1989.18 132.414 Q1987.3 130.979 1983.78 130.979 Q1981.86 130.979 1979.66 131.395 Q1977.46 131.812 1974.83 132.692 L1974.83 128.525 Q1977.49 127.784 1979.8 127.414 Q1982.14 127.044 1984.2 127.044 Q1989.52 127.044 1992.63 129.474 Q1995.73 131.882 1995.73 136.002 Q1995.73 138.872 1994.08 140.863 Q1992.44 142.831 1989.41 143.594 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M2005.45 127.669 L2023.81 127.669 L2023.81 131.604 L2009.73 131.604 L2009.73 140.076 Q2010.75 139.729 2011.77 139.567 Q2012.79 139.382 2013.81 139.382 Q2019.59 139.382 2022.97 142.553 Q2026.35 145.724 2026.35 151.141 Q2026.35 156.719 2022.88 159.821 Q2019.41 162.9 2013.09 162.9 Q2010.91 162.9 2008.64 162.53 Q2006.4 162.159 2003.99 161.418 L2003.99 156.719 Q2006.07 157.854 2008.3 158.409 Q2010.52 158.965 2013 158.965 Q2017 158.965 2019.34 156.858 Q2021.68 154.752 2021.68 151.141 Q2021.68 147.53 2019.34 145.423 Q2017 143.317 2013 143.317 Q2011.12 143.317 2009.25 143.733 Q2007.39 144.15 2005.45 145.03 L2005.45 127.669 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M2035.61 127.669 L2053.97 127.669 L2053.97 131.604 L2039.89 131.604 L2039.89 140.076 Q2040.91 139.729 2041.93 139.567 Q2042.95 139.382 2043.97 139.382 Q2049.76 139.382 2053.13 142.553 Q2056.51 145.724 2056.51 151.141 Q2056.51 156.719 2053.04 159.821 Q2049.57 162.9 2043.25 162.9 Q2041.07 162.9 2038.81 162.53 Q2036.56 162.159 2034.15 161.418 L2034.15 156.719 Q2036.24 157.854 2038.46 158.409 Q2040.68 158.965 2043.16 158.965 Q2047.16 158.965 2049.5 156.858 Q2051.84 154.752 2051.84 151.141 Q2051.84 147.53 2049.5 145.423 Q2047.16 143.317 2043.16 143.317 Q2041.28 143.317 2039.41 143.733 Q2037.56 144.15 2035.61 145.03 L2035.61 127.669 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M2078.57 131.743 L2066.77 150.192 L2078.57 150.192 L2078.57 131.743 M2077.35 127.669 L2083.23 127.669 L2083.23 150.192 L2088.16 150.192 L2088.16 154.081 L2083.23 154.081 L2083.23 162.229 L2078.57 162.229 L2078.57 154.081 L2062.97 154.081 L2062.97 149.567 L2077.35 127.669 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M2108.74 131.743 L2096.93 150.192 L2108.74 150.192 L2108.74 131.743 M2107.51 127.669 L2113.39 127.669 L2113.39 150.192 L2118.32 150.192 L2118.32 154.081 L2113.39 154.081 L2113.39 162.229 L2108.74 162.229 L2108.74 154.081 L2093.13 154.081 L2093.13 149.567 L2107.51 127.669 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M2140.22 143.594 Q2143.57 144.312 2145.45 146.581 Q2147.35 148.849 2147.35 152.182 Q2147.35 157.298 2143.83 160.099 Q2140.31 162.9 2133.83 162.9 Q2131.65 162.9 2129.34 162.46 Q2127.05 162.043 2124.59 161.187 L2124.59 156.673 Q2126.54 157.807 2128.85 158.386 Q2131.17 158.965 2133.69 158.965 Q2138.09 158.965 2140.38 157.229 Q2142.69 155.493 2142.69 152.182 Q2142.69 149.127 2140.54 147.414 Q2138.41 145.678 2134.59 145.678 L2130.57 145.678 L2130.57 141.835 L2134.78 141.835 Q2138.23 141.835 2140.06 140.469 Q2141.88 139.081 2141.88 136.488 Q2141.88 133.826 2139.99 132.414 Q2138.11 130.979 2134.59 130.979 Q2132.67 130.979 2130.47 131.395 Q2128.27 131.812 2125.63 132.692 L2125.63 128.525 Q2128.3 127.784 2130.61 127.414 Q2132.95 127.044 2135.01 127.044 Q2140.33 127.044 2143.44 129.474 Q2146.54 131.882 2146.54 136.002 Q2146.54 138.872 2144.89 140.863 Q2143.25 142.831 2140.22 143.594 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M2160.24 158.293 L2176.56 158.293 L2176.56 162.229 L2154.62 162.229 L2154.62 158.293 Q2157.28 155.539 2161.86 150.909 Q2166.47 146.256 2167.65 144.914 Q2169.89 142.391 2170.77 140.655 Q2171.68 138.895 2171.68 137.206 Q2171.68 134.451 2169.73 132.715 Q2167.81 130.979 2164.71 130.979 Q2162.51 130.979 2160.06 131.743 Q2157.63 132.507 2154.85 134.057 L2154.85 129.335 Q2157.67 128.201 2160.13 127.622 Q2162.58 127.044 2164.62 127.044 Q2169.99 127.044 2173.18 129.729 Q2176.38 132.414 2176.38 136.905 Q2176.38 139.034 2175.56 140.956 Q2174.78 142.854 2172.67 145.446 Q2172.09 146.118 2168.99 149.335 Q2165.89 152.53 2160.24 158.293 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M2185.1 126.257 L2188.81 126.257 Q2192.28 131.72 2193.99 136.951 Q2195.73 142.182 2195.73 147.344 Q2195.73 152.53 2193.99 157.784 Q2192.28 163.039 2188.81 168.479 L2185.1 168.479 Q2188.18 163.178 2189.69 157.946 Q2191.21 152.692 2191.21 147.344 Q2191.21 141.997 2189.69 136.789 Q2188.18 131.581 2185.1 126.257 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip970)" style="stroke:#e26f46; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1574.83,196.789 1721.85,196.789 
  "/>
<path clip-path="url(#clip970)" d="M1766.7 193.12 Q1768.3 190.249 1770.52 188.884 Q1772.74 187.518 1775.75 187.518 Q1779.8 187.518 1782 190.365 Q1784.2 193.189 1784.2 198.421 L1784.2 214.069 L1779.92 214.069 L1779.92 198.559 Q1779.92 194.833 1778.6 193.027 Q1777.28 191.222 1774.57 191.222 Q1771.26 191.222 1769.34 193.421 Q1767.42 195.62 1767.42 199.416 L1767.42 214.069 L1763.14 214.069 L1763.14 198.559 Q1763.14 194.809 1761.82 193.027 Q1760.5 191.222 1757.74 191.222 Q1754.48 191.222 1752.56 193.444 Q1750.64 195.643 1750.64 199.416 L1750.64 214.069 L1746.35 214.069 L1746.35 188.143 L1750.64 188.143 L1750.64 192.171 Q1752.09 189.786 1754.13 188.652 Q1756.17 187.518 1758.97 187.518 Q1761.79 187.518 1763.76 188.953 Q1765.75 190.388 1766.7 193.12 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M1804.48 201.036 Q1799.32 201.036 1797.33 202.217 Q1795.34 203.397 1795.34 206.245 Q1795.34 208.513 1796.82 209.856 Q1798.32 211.175 1800.89 211.175 Q1804.43 211.175 1806.56 208.675 Q1808.71 206.152 1808.71 201.985 L1808.71 201.036 L1804.48 201.036 M1812.97 199.277 L1812.97 214.069 L1808.71 214.069 L1808.71 210.133 Q1807.26 212.495 1805.08 213.629 Q1802.9 214.74 1799.76 214.74 Q1795.78 214.74 1793.41 212.518 Q1791.08 210.272 1791.08 206.522 Q1791.08 202.147 1793.99 199.925 Q1796.93 197.703 1802.74 197.703 L1808.71 197.703 L1808.71 197.286 Q1808.71 194.347 1806.77 192.749 Q1804.85 191.129 1801.35 191.129 Q1799.13 191.129 1797.03 191.661 Q1794.92 192.194 1792.97 193.259 L1792.97 189.323 Q1795.31 188.421 1797.51 187.981 Q1799.71 187.518 1801.79 187.518 Q1807.42 187.518 1810.2 190.434 Q1812.97 193.351 1812.97 199.277 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M1838.27 188.907 L1838.27 192.934 Q1836.47 192.009 1834.52 191.546 Q1832.58 191.083 1830.5 191.083 Q1827.33 191.083 1825.73 192.055 Q1824.15 193.027 1824.15 194.971 Q1824.15 196.453 1825.29 197.309 Q1826.42 198.143 1829.85 198.907 L1831.31 199.231 Q1835.84 200.203 1837.74 201.985 Q1839.66 203.745 1839.66 206.916 Q1839.66 210.527 1836.79 212.633 Q1833.95 214.74 1828.95 214.74 Q1826.86 214.74 1824.59 214.323 Q1822.35 213.93 1819.85 213.12 L1819.85 208.721 Q1822.21 209.948 1824.5 210.573 Q1826.79 211.175 1829.04 211.175 Q1832.05 211.175 1833.67 210.157 Q1835.29 209.115 1835.29 207.24 Q1835.29 205.504 1834.11 204.578 Q1832.95 203.652 1828.99 202.796 L1827.51 202.448 Q1823.55 201.615 1821.79 199.902 Q1820.03 198.166 1820.03 195.157 Q1820.03 191.499 1822.63 189.509 Q1825.22 187.518 1829.99 187.518 Q1832.35 187.518 1834.43 187.865 Q1836.52 188.212 1838.27 188.907 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M1862.97 188.907 L1862.97 192.934 Q1861.17 192.009 1859.22 191.546 Q1857.28 191.083 1855.2 191.083 Q1852.02 191.083 1850.43 192.055 Q1848.85 193.027 1848.85 194.971 Q1848.85 196.453 1849.99 197.309 Q1851.12 198.143 1854.55 198.907 L1856.01 199.231 Q1860.54 200.203 1862.44 201.985 Q1864.36 203.745 1864.36 206.916 Q1864.36 210.527 1861.49 212.633 Q1858.65 214.74 1853.65 214.74 Q1851.56 214.74 1849.29 214.323 Q1847.05 213.93 1844.55 213.12 L1844.55 208.721 Q1846.91 209.948 1849.2 210.573 Q1851.49 211.175 1853.74 211.175 Q1856.75 211.175 1858.37 210.157 Q1859.99 209.115 1859.99 207.24 Q1859.99 205.504 1858.81 204.578 Q1857.65 203.652 1853.69 202.796 L1852.21 202.448 Q1848.25 201.615 1846.49 199.902 Q1844.73 198.166 1844.73 195.157 Q1844.73 191.499 1847.33 189.509 Q1849.92 187.518 1854.69 187.518 Q1857.05 187.518 1859.13 187.865 Q1861.21 188.212 1862.97 188.907 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M1871.75 208.189 L1876.63 208.189 L1876.63 214.069 L1871.75 214.069 L1871.75 208.189 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M1883.16 188.143 L1887.67 188.143 L1895.77 209.902 L1903.88 188.143 L1908.39 188.143 L1898.67 214.069 L1892.88 214.069 L1883.16 188.143 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M1939.57 178.097 Q1936.47 183.421 1934.96 188.629 Q1933.46 193.837 1933.46 199.184 Q1933.46 204.532 1934.96 209.786 Q1936.49 215.018 1939.57 220.319 L1935.87 220.319 Q1932.39 214.879 1930.66 209.624 Q1928.95 204.37 1928.95 199.184 Q1928.95 194.022 1930.66 188.791 Q1932.37 183.56 1935.87 178.097 L1939.57 178.097 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M1962.6 195.434 Q1965.96 196.152 1967.83 198.421 Q1969.73 200.689 1969.73 204.022 Q1969.73 209.138 1966.21 211.939 Q1962.7 214.74 1956.21 214.74 Q1954.04 214.74 1951.72 214.3 Q1949.43 213.883 1946.98 213.027 L1946.98 208.513 Q1948.92 209.647 1951.24 210.226 Q1953.55 210.805 1956.08 210.805 Q1960.47 210.805 1962.77 209.069 Q1965.08 207.333 1965.08 204.022 Q1965.08 200.967 1962.93 199.254 Q1960.8 197.518 1956.98 197.518 L1952.95 197.518 L1952.95 193.675 L1957.16 193.675 Q1960.61 193.675 1962.44 192.309 Q1964.27 190.921 1964.27 188.328 Q1964.27 185.666 1962.37 184.254 Q1960.5 182.819 1956.98 182.819 Q1955.06 182.819 1952.86 183.235 Q1950.66 183.652 1948.02 184.532 L1948.02 180.365 Q1950.68 179.624 1953 179.254 Q1955.33 178.884 1957.39 178.884 Q1962.72 178.884 1965.82 181.314 Q1968.92 183.722 1968.92 187.842 Q1968.92 190.712 1967.28 192.703 Q1965.64 194.671 1962.6 195.434 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M1992.76 195.434 Q1996.12 196.152 1998 198.421 Q1999.89 200.689 1999.89 204.022 Q1999.89 209.138 1996.38 211.939 Q1992.86 214.74 1986.38 214.74 Q1984.2 214.74 1981.89 214.3 Q1979.59 213.883 1977.14 213.027 L1977.14 208.513 Q1979.08 209.647 1981.4 210.226 Q1983.71 210.805 1986.24 210.805 Q1990.64 210.805 1992.93 209.069 Q1995.24 207.333 1995.24 204.022 Q1995.24 200.967 1993.09 199.254 Q1990.96 197.518 1987.14 197.518 L1983.11 197.518 L1983.11 193.675 L1987.33 193.675 Q1990.77 193.675 1992.6 192.309 Q1994.43 190.921 1994.43 188.328 Q1994.43 185.666 1992.53 184.254 Q1990.66 182.819 1987.14 182.819 Q1985.22 182.819 1983.02 183.235 Q1980.82 183.652 1978.18 184.532 L1978.18 180.365 Q1980.84 179.624 1983.16 179.254 Q1985.5 178.884 1987.56 178.884 Q1992.88 178.884 1995.98 181.314 Q1999.08 183.722 1999.08 187.842 Q1999.08 190.712 1997.44 192.703 Q1995.8 194.671 1992.76 195.434 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M2008.81 179.509 L2027.16 179.509 L2027.16 183.444 L2013.09 183.444 L2013.09 191.916 Q2014.11 191.569 2015.13 191.407 Q2016.14 191.222 2017.16 191.222 Q2022.95 191.222 2026.33 194.393 Q2029.71 197.564 2029.71 202.981 Q2029.71 208.559 2026.24 211.661 Q2022.76 214.74 2016.45 214.74 Q2014.27 214.74 2012 214.37 Q2009.76 213.999 2007.35 213.258 L2007.35 208.559 Q2009.43 209.694 2011.65 210.249 Q2013.88 210.805 2016.35 210.805 Q2020.36 210.805 2022.7 208.698 Q2025.03 206.592 2025.03 202.981 Q2025.03 199.37 2022.7 197.263 Q2020.36 195.157 2016.35 195.157 Q2014.48 195.157 2012.6 195.573 Q2010.75 195.99 2008.81 196.87 L2008.81 179.509 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M2038.97 179.509 L2057.32 179.509 L2057.32 183.444 L2043.25 183.444 L2043.25 191.916 Q2044.27 191.569 2045.29 191.407 Q2046.31 191.222 2047.32 191.222 Q2053.11 191.222 2056.49 194.393 Q2059.87 197.564 2059.87 202.981 Q2059.87 208.559 2056.4 211.661 Q2052.93 214.74 2046.61 214.74 Q2044.43 214.74 2042.16 214.37 Q2039.92 213.999 2037.51 213.258 L2037.51 208.559 Q2039.59 209.694 2041.82 210.249 Q2044.04 210.805 2046.51 210.805 Q2050.52 210.805 2052.86 208.698 Q2055.2 206.592 2055.2 202.981 Q2055.2 199.37 2052.86 197.263 Q2050.52 195.157 2046.51 195.157 Q2044.64 195.157 2042.76 195.573 Q2040.91 195.99 2038.97 196.87 L2038.97 179.509 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M2081.93 183.583 L2070.13 202.032 L2081.93 202.032 L2081.93 183.583 M2080.7 179.509 L2086.58 179.509 L2086.58 202.032 L2091.51 202.032 L2091.51 205.921 L2086.58 205.921 L2086.58 214.069 L2081.93 214.069 L2081.93 205.921 L2066.33 205.921 L2066.33 201.407 L2080.7 179.509 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M2112.09 183.583 L2100.29 202.032 L2112.09 202.032 L2112.09 183.583 M2110.87 179.509 L2116.75 179.509 L2116.75 202.032 L2121.68 202.032 L2121.68 205.921 L2116.75 205.921 L2116.75 214.069 L2112.09 214.069 L2112.09 205.921 L2096.49 205.921 L2096.49 201.407 L2110.87 179.509 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M2143.57 195.434 Q2146.93 196.152 2148.81 198.421 Q2150.7 200.689 2150.7 204.022 Q2150.7 209.138 2147.19 211.939 Q2143.67 214.74 2137.19 214.74 Q2135.01 214.74 2132.69 214.3 Q2130.4 213.883 2127.95 213.027 L2127.95 208.513 Q2129.89 209.647 2132.21 210.226 Q2134.52 210.805 2137.05 210.805 Q2141.44 210.805 2143.74 209.069 Q2146.05 207.333 2146.05 204.022 Q2146.05 200.967 2143.9 199.254 Q2141.77 197.518 2137.95 197.518 L2133.92 197.518 L2133.92 193.675 L2138.13 193.675 Q2141.58 193.675 2143.41 192.309 Q2145.24 190.921 2145.24 188.328 Q2145.24 185.666 2143.34 184.254 Q2141.47 182.819 2137.95 182.819 Q2136.03 182.819 2133.83 183.235 Q2131.63 183.652 2128.99 184.532 L2128.99 180.365 Q2131.65 179.624 2133.97 179.254 Q2136.31 178.884 2138.37 178.884 Q2143.69 178.884 2146.79 181.314 Q2149.89 183.722 2149.89 187.842 Q2149.89 190.712 2148.25 192.703 Q2146.61 194.671 2143.57 195.434 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M2173.74 195.434 Q2177.09 196.152 2178.97 198.421 Q2180.87 200.689 2180.87 204.022 Q2180.87 209.138 2177.35 211.939 Q2173.83 214.74 2167.35 214.74 Q2165.17 214.74 2162.86 214.3 Q2160.56 213.883 2158.11 213.027 L2158.11 208.513 Q2160.06 209.647 2162.37 210.226 Q2164.69 210.805 2167.21 210.805 Q2171.61 210.805 2173.9 209.069 Q2176.21 207.333 2176.21 204.022 Q2176.21 200.967 2174.06 199.254 Q2171.93 197.518 2168.11 197.518 L2164.08 197.518 L2164.08 193.675 L2168.3 193.675 Q2171.75 193.675 2173.57 192.309 Q2175.4 190.921 2175.4 188.328 Q2175.4 185.666 2173.5 184.254 Q2171.63 182.819 2168.11 182.819 Q2166.19 182.819 2163.99 183.235 Q2161.79 183.652 2159.15 184.532 L2159.15 180.365 Q2161.81 179.624 2164.13 179.254 Q2166.47 178.884 2168.53 178.884 Q2173.85 178.884 2176.95 181.314 Q2180.06 183.722 2180.06 187.842 Q2180.06 190.712 2178.41 192.703 Q2176.77 194.671 2173.74 195.434 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M2188.46 178.097 L2192.16 178.097 Q2195.63 183.56 2197.35 188.791 Q2199.08 194.022 2199.08 199.184 Q2199.08 204.37 2197.35 209.624 Q2195.63 214.879 2192.16 220.319 L2188.46 220.319 Q2191.54 215.018 2193.04 209.786 Q2194.57 204.532 2194.57 199.184 Q2194.57 193.837 2193.04 188.629 Q2191.54 183.421 2188.46 178.097 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip970)" style="stroke:#3da44d; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1574.83,248.629 1721.85,248.629 
  "/>
<path clip-path="url(#clip970)" d="M1766.7 244.96 Q1768.3 242.089 1770.52 240.724 Q1772.74 239.358 1775.75 239.358 Q1779.8 239.358 1782 242.205 Q1784.2 245.029 1784.2 250.261 L1784.2 265.909 L1779.92 265.909 L1779.92 250.399 Q1779.92 246.673 1778.6 244.867 Q1777.28 243.062 1774.57 243.062 Q1771.26 243.062 1769.34 245.261 Q1767.42 247.46 1767.42 251.256 L1767.42 265.909 L1763.14 265.909 L1763.14 250.399 Q1763.14 246.649 1761.82 244.867 Q1760.5 243.062 1757.74 243.062 Q1754.48 243.062 1752.56 245.284 Q1750.64 247.483 1750.64 251.256 L1750.64 265.909 L1746.35 265.909 L1746.35 239.983 L1750.64 239.983 L1750.64 244.011 Q1752.09 241.626 1754.13 240.492 Q1756.17 239.358 1758.97 239.358 Q1761.79 239.358 1763.76 240.793 Q1765.75 242.228 1766.7 244.96 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M1804.48 252.876 Q1799.32 252.876 1797.33 254.057 Q1795.34 255.237 1795.34 258.085 Q1795.34 260.353 1796.82 261.696 Q1798.32 263.015 1800.89 263.015 Q1804.43 263.015 1806.56 260.515 Q1808.71 257.992 1808.71 253.825 L1808.71 252.876 L1804.48 252.876 M1812.97 251.117 L1812.97 265.909 L1808.71 265.909 L1808.71 261.973 Q1807.26 264.335 1805.08 265.469 Q1802.9 266.58 1799.76 266.58 Q1795.78 266.58 1793.41 264.358 Q1791.08 262.112 1791.08 258.362 Q1791.08 253.987 1793.99 251.765 Q1796.93 249.543 1802.74 249.543 L1808.71 249.543 L1808.71 249.126 Q1808.71 246.187 1806.77 244.589 Q1804.85 242.969 1801.35 242.969 Q1799.13 242.969 1797.03 243.501 Q1794.92 244.034 1792.97 245.099 L1792.97 241.163 Q1795.31 240.261 1797.51 239.821 Q1799.71 239.358 1801.79 239.358 Q1807.42 239.358 1810.2 242.274 Q1812.97 245.191 1812.97 251.117 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M1838.27 240.747 L1838.27 244.774 Q1836.47 243.849 1834.52 243.386 Q1832.58 242.923 1830.5 242.923 Q1827.33 242.923 1825.73 243.895 Q1824.15 244.867 1824.15 246.811 Q1824.15 248.293 1825.29 249.149 Q1826.42 249.983 1829.85 250.747 L1831.31 251.071 Q1835.84 252.043 1837.74 253.825 Q1839.66 255.585 1839.66 258.756 Q1839.66 262.367 1836.79 264.473 Q1833.95 266.58 1828.95 266.58 Q1826.86 266.58 1824.59 266.163 Q1822.35 265.77 1819.85 264.96 L1819.85 260.561 Q1822.21 261.788 1824.5 262.413 Q1826.79 263.015 1829.04 263.015 Q1832.05 263.015 1833.67 261.997 Q1835.29 260.955 1835.29 259.08 Q1835.29 257.344 1834.11 256.418 Q1832.95 255.492 1828.99 254.636 L1827.51 254.288 Q1823.55 253.455 1821.79 251.742 Q1820.03 250.006 1820.03 246.997 Q1820.03 243.339 1822.63 241.349 Q1825.22 239.358 1829.99 239.358 Q1832.35 239.358 1834.43 239.705 Q1836.52 240.052 1838.27 240.747 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M1862.97 240.747 L1862.97 244.774 Q1861.17 243.849 1859.22 243.386 Q1857.28 242.923 1855.2 242.923 Q1852.02 242.923 1850.43 243.895 Q1848.85 244.867 1848.85 246.811 Q1848.85 248.293 1849.99 249.149 Q1851.12 249.983 1854.55 250.747 L1856.01 251.071 Q1860.54 252.043 1862.44 253.825 Q1864.36 255.585 1864.36 258.756 Q1864.36 262.367 1861.49 264.473 Q1858.65 266.58 1853.65 266.58 Q1851.56 266.58 1849.29 266.163 Q1847.05 265.77 1844.55 264.96 L1844.55 260.561 Q1846.91 261.788 1849.2 262.413 Q1851.49 263.015 1853.74 263.015 Q1856.75 263.015 1858.37 261.997 Q1859.99 260.955 1859.99 259.08 Q1859.99 257.344 1858.81 256.418 Q1857.65 255.492 1853.69 254.636 L1852.21 254.288 Q1848.25 253.455 1846.49 251.742 Q1844.73 250.006 1844.73 246.997 Q1844.73 243.339 1847.33 241.349 Q1849.92 239.358 1854.69 239.358 Q1857.05 239.358 1859.13 239.705 Q1861.21 240.052 1862.97 240.747 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M1871.75 260.029 L1876.63 260.029 L1876.63 265.909 L1871.75 265.909 L1871.75 260.029 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M1898 252.876 Q1892.83 252.876 1890.84 254.057 Q1888.85 255.237 1888.85 258.085 Q1888.85 260.353 1890.33 261.696 Q1891.84 263.015 1894.41 263.015 Q1897.95 263.015 1900.08 260.515 Q1902.23 257.992 1902.23 253.825 L1902.23 252.876 L1898 252.876 M1906.49 251.117 L1906.49 265.909 L1902.23 265.909 L1902.23 261.973 Q1900.77 264.335 1898.6 265.469 Q1896.42 266.58 1893.27 266.58 Q1889.29 266.58 1886.93 264.358 Q1884.59 262.112 1884.59 258.362 Q1884.59 253.987 1887.51 251.765 Q1890.45 249.543 1896.26 249.543 L1902.23 249.543 L1902.23 249.126 Q1902.23 246.187 1900.29 244.589 Q1898.37 242.969 1894.87 242.969 Q1892.65 242.969 1890.54 243.501 Q1888.44 244.034 1886.49 245.099 L1886.49 241.163 Q1888.83 240.261 1891.03 239.821 Q1893.23 239.358 1895.31 239.358 Q1900.94 239.358 1903.71 242.274 Q1906.49 245.191 1906.49 251.117 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M1940.57 229.937 Q1937.46 235.261 1935.96 240.469 Q1934.46 245.677 1934.46 251.024 Q1934.46 256.372 1935.96 261.626 Q1937.49 266.858 1940.57 272.159 L1936.86 272.159 Q1933.39 266.719 1931.65 261.464 Q1929.94 256.21 1929.94 251.024 Q1929.94 245.862 1931.65 240.631 Q1933.37 235.4 1936.86 229.937 L1940.57 229.937 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M1949.48 231.349 L1967.83 231.349 L1967.83 235.284 L1953.76 235.284 L1953.76 243.756 Q1954.78 243.409 1955.8 243.247 Q1956.82 243.062 1957.83 243.062 Q1963.62 243.062 1967 246.233 Q1970.38 249.404 1970.38 254.821 Q1970.38 260.399 1966.91 263.501 Q1963.44 266.58 1957.12 266.58 Q1954.94 266.58 1952.67 266.21 Q1950.43 265.839 1948.02 265.098 L1948.02 260.399 Q1950.1 261.534 1952.33 262.089 Q1954.55 262.645 1957.02 262.645 Q1961.03 262.645 1963.37 260.538 Q1965.7 258.432 1965.7 254.821 Q1965.7 251.21 1963.37 249.103 Q1961.03 246.997 1957.02 246.997 Q1955.15 246.997 1953.27 247.413 Q1951.42 247.83 1949.48 248.71 L1949.48 231.349 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M1989.59 249.497 Q1986.26 249.497 1984.34 251.279 Q1982.44 253.061 1982.44 256.186 Q1982.44 259.311 1984.34 261.094 Q1986.26 262.876 1989.59 262.876 Q1992.93 262.876 1994.85 261.094 Q1996.77 259.288 1996.77 256.186 Q1996.77 253.061 1994.85 251.279 Q1992.95 249.497 1989.59 249.497 M1984.92 247.506 Q1981.91 246.765 1980.22 244.705 Q1978.55 242.645 1978.55 239.682 Q1978.55 235.538 1981.49 233.131 Q1984.45 230.724 1989.59 230.724 Q1994.76 230.724 1997.7 233.131 Q2000.64 235.538 2000.64 239.682 Q2000.64 242.645 1998.95 244.705 Q1997.28 246.765 1994.29 247.506 Q1997.67 248.293 1999.55 250.585 Q2001.45 252.876 2001.45 256.186 Q2001.45 261.21 1998.37 263.895 Q1995.31 266.58 1989.59 266.58 Q1983.88 266.58 1980.8 263.895 Q1977.74 261.21 1977.74 256.186 Q1977.74 252.876 1979.64 250.585 Q1981.54 248.293 1984.92 247.506 M1983.2 240.122 Q1983.2 242.807 1984.87 244.312 Q1986.56 245.816 1989.59 245.816 Q1992.6 245.816 1994.29 244.312 Q1996.01 242.807 1996.01 240.122 Q1996.01 237.437 1994.29 235.932 Q1992.6 234.427 1989.59 234.427 Q1986.56 234.427 1984.87 235.932 Q1983.2 237.437 1983.2 240.122 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M2008.57 231.349 L2030.8 231.349 L2030.8 233.339 L2018.25 265.909 L2013.37 265.909 L2025.17 235.284 L2008.57 235.284 L2008.57 231.349 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M2043.95 261.973 L2060.26 261.973 L2060.26 265.909 L2038.32 265.909 L2038.32 261.973 Q2040.98 259.219 2045.57 254.589 Q2050.17 249.936 2051.35 248.594 Q2053.6 246.071 2054.48 244.335 Q2055.38 242.575 2055.38 240.886 Q2055.38 238.131 2053.44 236.395 Q2051.51 234.659 2048.41 234.659 Q2046.21 234.659 2043.76 235.423 Q2041.33 236.187 2038.55 237.737 L2038.55 233.015 Q2041.38 231.881 2043.83 231.302 Q2046.28 230.724 2048.32 230.724 Q2053.69 230.724 2056.88 233.409 Q2060.08 236.094 2060.08 240.585 Q2060.08 242.714 2059.27 244.636 Q2058.48 246.534 2056.38 249.126 Q2055.8 249.798 2052.7 253.015 Q2049.59 256.21 2043.95 261.973 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M2080.08 234.427 Q2076.47 234.427 2074.64 237.992 Q2072.83 241.534 2072.83 248.663 Q2072.83 255.77 2074.64 259.335 Q2076.47 262.876 2080.08 262.876 Q2083.71 262.876 2085.52 259.335 Q2087.35 255.77 2087.35 248.663 Q2087.35 241.534 2085.52 237.992 Q2083.71 234.427 2080.08 234.427 M2080.08 230.724 Q2085.89 230.724 2088.94 235.33 Q2092.02 239.913 2092.02 248.663 Q2092.02 257.39 2088.94 261.997 Q2085.89 266.58 2080.08 266.58 Q2074.27 266.58 2071.19 261.997 Q2068.13 257.39 2068.13 248.663 Q2068.13 239.913 2071.19 235.33 Q2074.27 230.724 2080.08 230.724 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M2104.27 261.973 L2120.59 261.973 L2120.59 265.909 L2098.64 265.909 L2098.64 261.973 Q2101.31 259.219 2105.89 254.589 Q2110.5 249.936 2111.68 248.594 Q2113.92 246.071 2114.8 244.335 Q2115.7 242.575 2115.7 240.886 Q2115.7 238.131 2113.76 236.395 Q2111.84 234.659 2108.74 234.659 Q2106.54 234.659 2104.08 235.423 Q2101.65 236.187 2098.88 237.737 L2098.88 233.015 Q2101.7 231.881 2104.15 231.302 Q2106.61 230.724 2108.64 230.724 Q2114.01 230.724 2117.21 233.409 Q2120.4 236.094 2120.4 240.585 Q2120.4 242.714 2119.59 244.636 Q2118.81 246.534 2116.7 249.126 Q2116.12 249.798 2113.02 253.015 Q2109.92 256.21 2104.27 261.973 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M2130.45 231.349 L2148.81 231.349 L2148.81 235.284 L2134.73 235.284 L2134.73 243.756 Q2135.75 243.409 2136.77 243.247 Q2137.79 243.062 2138.81 243.062 Q2144.59 243.062 2147.97 246.233 Q2151.35 249.404 2151.35 254.821 Q2151.35 260.399 2147.88 263.501 Q2144.41 266.58 2138.09 266.58 Q2135.91 266.58 2133.64 266.21 Q2131.4 265.839 2128.99 265.098 L2128.99 260.399 Q2131.07 261.534 2133.3 262.089 Q2135.52 262.645 2138 262.645 Q2142 262.645 2144.34 260.538 Q2146.68 258.432 2146.68 254.821 Q2146.68 251.21 2144.34 249.103 Q2142 246.997 2138 246.997 Q2136.12 246.997 2134.25 247.413 Q2132.39 247.83 2130.45 248.71 L2130.45 231.349 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M2171.14 246.765 Q2168 246.765 2166.14 248.918 Q2164.31 251.071 2164.31 254.821 Q2164.31 258.548 2166.14 260.723 Q2168 262.876 2171.14 262.876 Q2174.29 262.876 2176.12 260.723 Q2177.97 258.548 2177.97 254.821 Q2177.97 251.071 2176.12 248.918 Q2174.29 246.765 2171.14 246.765 M2180.43 232.113 L2180.43 236.372 Q2178.67 235.538 2176.86 235.099 Q2175.08 234.659 2173.32 234.659 Q2168.69 234.659 2166.24 237.784 Q2163.81 240.909 2163.46 247.228 Q2164.82 245.214 2166.88 244.149 Q2168.94 243.062 2171.42 243.062 Q2176.63 243.062 2179.64 246.233 Q2182.67 249.381 2182.67 254.821 Q2182.67 260.145 2179.52 263.362 Q2176.38 266.58 2171.14 266.58 Q2165.15 266.58 2161.98 261.997 Q2158.81 257.39 2158.81 248.663 Q2158.81 240.469 2162.69 235.608 Q2166.58 230.724 2173.13 230.724 Q2174.89 230.724 2176.68 231.071 Q2178.48 231.418 2180.43 232.113 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M2191.54 261.973 L2199.18 261.973 L2199.18 235.608 L2190.87 237.275 L2190.87 233.015 L2199.13 231.349 L2203.81 231.349 L2203.81 261.973 L2211.44 261.973 L2211.44 265.909 L2191.54 265.909 L2191.54 261.973 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip970)" d="M2219.62 229.937 L2223.32 229.937 Q2226.79 235.4 2228.5 240.631 Q2230.24 245.862 2230.24 251.024 Q2230.24 256.21 2228.5 261.464 Q2226.79 266.719 2223.32 272.159 L2219.62 272.159 Q2222.69 266.858 2224.2 261.626 Q2225.73 256.372 2225.73 251.024 Q2225.73 245.677 2224.2 240.469 Q2222.69 235.261 2219.62 229.937 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /></svg>



The data from the simualtion of the *referenceFMU*, are divided into position, velocity and acceleration data. The data for the acceleration will be needed later. 


```julia
posReference = collect(data[1] for data in referenceSimData.saveval)
velReference = collect(data[2] for data in referenceSimData.saveval)
accReference = collect(data[3] for data in referenceSimData.saveval)
```


    501-element Vector{Float64}:
     -1.9999999999999996
     -1.9989808107156004
     -1.995976332371232
     -1.9909821938997307
     -1.9839989801021418
     -1.9750314004124547
     -1.9640884504035183
     -1.951180580066516
     -1.9363227134824257
     -1.9195319560818125
      ⋮
      1.9805952930006132
      1.9888623187994514
      1.9951388459819808
      1.9994178648958127
      2.0016968375647415
      2.0019759530917005
      2.0002523498984894
      1.9965275218318568
      1.9908049090723823


### DefaultFMU

The following is a reset for the *referenceFMU* and a renaming to *defaultFMU*. After the reset, the previous initial position of the mass is not set, so the default position of the *defaultFMU* is used. The first state indicates the position of the mass, which is initilized with $0.5𝑚$.


```julia
fmiReset(referenceFMU)
defaultFMU = referenceFMU

fmiSetupExperiment(defaultFMU, tStart, tStop)
fmiEnterInitializationMode(defaultFMU)
fmiExitInitializationMode(defaultFMU)

x₀ = fmiGetContinuousStates(defaultFMU)
```

    ┌ Warning: fmi2Reset(_): Should be only called in FMU state `modelSetableFMUstate`.
    └ @ FMIImport C:\Users\thummeto\Documents\FMIImport.jl\src\FMI2_c.jl:219



    2-element Vector{Float64}:
     0.5
     0.0


The following simulate and plot the *defaultFMU* just like the *referenceFMU*. The differences between both systems can be clearly seen from the plots. In the plots for the *defaultFMU* you can see that other oscillations occur due to the different starting positions. On the one hand the oscillation of the *defaultFMU* starts in the opposite direction of the *referenceFMU* and on the other hand the graphs for the velocity and acceleration differ clearly in the amplitude. In the following we try to learn the initial shift of the position so that the graphs for the acceleration of both graphs match.


```julia
_, defaultSimData = fmiSimulate(defaultFMU, tStart, tStop; recordValues=vrs, setup=false, reset=false, saveat=tSave)
fmiPlot(defaultFMU, vrs, defaultSimData)
```


<?xml version="1.0" encoding="utf-8"?>
<svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" width="600" height="400" viewBox="0 0 2400 1600">
<defs>
  <clipPath id="clip060">
    <rect x="0" y="0" width="2400" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip060)" d="
M0 1600 L2400 1600 L2400 0 L0 0  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip061">
    <rect x="480" y="0" width="1681" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip060)" d="
M193.936 1423.18 L2352.76 1423.18 L2352.76 47.2441 L193.936 47.2441  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip062">
    <rect x="193" y="47" width="2160" height="1377"/>
  </clipPath>
</defs>
<polyline clip-path="url(#clip062)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  255.035,1423.18 255.035,47.2441 
  "/>
<polyline clip-path="url(#clip062)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  662.359,1423.18 662.359,47.2441 
  "/>
<polyline clip-path="url(#clip062)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1069.68,1423.18 1069.68,47.2441 
  "/>
<polyline clip-path="url(#clip062)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1477.01,1423.18 1477.01,47.2441 
  "/>
<polyline clip-path="url(#clip062)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1884.33,1423.18 1884.33,47.2441 
  "/>
<polyline clip-path="url(#clip062)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  2291.66,1423.18 2291.66,47.2441 
  "/>
<polyline clip-path="url(#clip060)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  193.936,1423.18 2352.76,1423.18 
  "/>
<polyline clip-path="url(#clip060)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  255.035,1423.18 255.035,1404.28 
  "/>
<polyline clip-path="url(#clip060)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  662.359,1423.18 662.359,1404.28 
  "/>
<polyline clip-path="url(#clip060)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1069.68,1423.18 1069.68,1404.28 
  "/>
<polyline clip-path="url(#clip060)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1477.01,1423.18 1477.01,1404.28 
  "/>
<polyline clip-path="url(#clip060)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1884.33,1423.18 1884.33,1404.28 
  "/>
<polyline clip-path="url(#clip060)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  2291.66,1423.18 2291.66,1404.28 
  "/>
<path clip-path="url(#clip060)" d="M255.035 1454.1 Q251.424 1454.1 249.595 1457.66 Q247.789 1461.2 247.789 1468.33 Q247.789 1475.44 249.595 1479.01 Q251.424 1482.55 255.035 1482.55 Q258.669 1482.55 260.474 1479.01 Q262.303 1475.44 262.303 1468.33 Q262.303 1461.2 260.474 1457.66 Q258.669 1454.1 255.035 1454.1 M255.035 1450.39 Q260.845 1450.39 263.9 1455 Q266.979 1459.58 266.979 1468.33 Q266.979 1477.06 263.9 1481.67 Q260.845 1486.25 255.035 1486.25 Q249.225 1486.25 246.146 1481.67 Q243.09 1477.06 243.09 1468.33 Q243.09 1459.58 246.146 1455 Q249.225 1450.39 255.035 1450.39 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M652.741 1481.64 L660.38 1481.64 L660.38 1455.28 L652.07 1456.95 L652.07 1452.69 L660.334 1451.02 L665.01 1451.02 L665.01 1481.64 L672.648 1481.64 L672.648 1485.58 L652.741 1485.58 L652.741 1481.64 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M1064.34 1481.64 L1080.66 1481.64 L1080.66 1485.58 L1058.71 1485.58 L1058.71 1481.64 Q1061.37 1478.89 1065.96 1474.26 Q1070.56 1469.61 1071.74 1468.27 Q1073.99 1465.74 1074.87 1464.01 Q1075.77 1462.25 1075.77 1460.56 Q1075.77 1457.8 1073.83 1456.07 Q1071.91 1454.33 1068.8 1454.33 Q1066.61 1454.33 1064.15 1455.09 Q1061.72 1455.86 1058.94 1457.41 L1058.94 1452.69 Q1061.77 1451.55 1064.22 1450.97 Q1066.67 1450.39 1068.71 1450.39 Q1074.08 1450.39 1077.28 1453.08 Q1080.47 1455.77 1080.47 1460.26 Q1080.47 1462.39 1079.66 1464.31 Q1078.87 1466.2 1076.77 1468.8 Q1076.19 1469.47 1073.09 1472.69 Q1069.98 1475.88 1064.34 1481.64 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M1481.26 1466.95 Q1484.61 1467.66 1486.49 1469.93 Q1488.39 1472.2 1488.39 1475.53 Q1488.39 1480.65 1484.87 1483.45 Q1481.35 1486.25 1474.87 1486.25 Q1472.69 1486.25 1470.38 1485.81 Q1468.08 1485.39 1465.63 1484.54 L1465.63 1480.02 Q1467.58 1481.16 1469.89 1481.74 Q1472.2 1482.32 1474.73 1482.32 Q1479.13 1482.32 1481.42 1480.58 Q1483.73 1478.84 1483.73 1475.53 Q1483.73 1472.48 1481.58 1470.77 Q1479.45 1469.03 1475.63 1469.03 L1471.6 1469.03 L1471.6 1465.19 L1475.82 1465.19 Q1479.27 1465.19 1481.09 1463.82 Q1482.92 1462.43 1482.92 1459.84 Q1482.92 1457.18 1481.02 1455.77 Q1479.15 1454.33 1475.63 1454.33 Q1473.71 1454.33 1471.51 1454.75 Q1469.31 1455.16 1466.67 1456.04 L1466.67 1451.88 Q1469.33 1451.14 1471.65 1450.77 Q1473.99 1450.39 1476.05 1450.39 Q1481.37 1450.39 1484.47 1452.83 Q1487.58 1455.23 1487.58 1459.35 Q1487.58 1462.22 1485.93 1464.21 Q1484.29 1466.18 1481.26 1466.95 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M1887.34 1455.09 L1875.54 1473.54 L1887.34 1473.54 L1887.34 1455.09 M1886.12 1451.02 L1891.99 1451.02 L1891.99 1473.54 L1896.93 1473.54 L1896.93 1477.43 L1891.99 1477.43 L1891.99 1485.58 L1887.34 1485.58 L1887.34 1477.43 L1871.74 1477.43 L1871.74 1472.92 L1886.12 1451.02 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M2281.94 1451.02 L2300.29 1451.02 L2300.29 1454.96 L2286.22 1454.96 L2286.22 1463.43 Q2287.24 1463.08 2288.25 1462.92 Q2289.27 1462.73 2290.29 1462.73 Q2296.08 1462.73 2299.46 1465.9 Q2302.84 1469.08 2302.84 1474.49 Q2302.84 1480.07 2299.37 1483.17 Q2295.89 1486.25 2289.57 1486.25 Q2287.4 1486.25 2285.13 1485.88 Q2282.88 1485.51 2280.48 1484.77 L2280.48 1480.07 Q2282.56 1481.2 2284.78 1481.76 Q2287 1482.32 2289.48 1482.32 Q2293.49 1482.32 2295.82 1480.21 Q2298.16 1478.1 2298.16 1474.49 Q2298.16 1470.88 2295.82 1468.77 Q2293.49 1466.67 2289.48 1466.67 Q2287.61 1466.67 2285.73 1467.08 Q2283.88 1467.5 2281.94 1468.38 L2281.94 1451.02 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M1221.66 1522.27 L1221.66 1532.4 L1233.72 1532.4 L1233.72 1536.95 L1221.66 1536.95 L1221.66 1556.3 Q1221.66 1560.66 1222.83 1561.9 Q1224.04 1563.14 1227.7 1563.14 L1233.72 1563.14 L1233.72 1568.04 L1227.7 1568.04 Q1220.92 1568.04 1218.35 1565.53 Q1215.77 1562.98 1215.77 1556.3 L1215.77 1536.95 L1211.47 1536.95 L1211.47 1532.4 L1215.77 1532.4 L1215.77 1522.27 L1221.66 1522.27 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M1261.6 1518.52 L1275.1 1518.52 L1275.1 1523.07 L1267.46 1523.07 L1267.46 1572.09 L1275.1 1572.09 L1275.1 1576.64 L1261.6 1576.64 L1261.6 1518.52 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M1310.3 1533.45 L1310.3 1538.98 Q1307.82 1537.71 1305.14 1537.07 Q1302.47 1536.44 1299.6 1536.44 Q1295.24 1536.44 1293.05 1537.77 Q1290.88 1539.11 1290.88 1541.79 Q1290.88 1543.82 1292.44 1545 Q1294 1546.15 1298.71 1547.2 L1300.72 1547.64 Q1306.96 1548.98 1309.57 1551.43 Q1312.21 1553.85 1312.21 1558.21 Q1312.21 1563.17 1308.26 1566.07 Q1304.35 1568.97 1297.47 1568.97 Q1294.61 1568.97 1291.49 1568.39 Q1288.4 1567.85 1284.96 1566.74 L1284.96 1560.69 Q1288.21 1562.38 1291.36 1563.24 Q1294.51 1564.07 1297.6 1564.07 Q1301.74 1564.07 1303.97 1562.66 Q1306.19 1561.23 1306.19 1558.65 Q1306.19 1556.27 1304.57 1554.99 Q1302.98 1553.72 1297.54 1552.54 L1295.5 1552.07 Q1290.06 1550.92 1287.64 1548.56 Q1285.22 1546.18 1285.22 1542.04 Q1285.22 1537.01 1288.78 1534.27 Q1292.35 1531.54 1298.9 1531.54 Q1302.15 1531.54 1305.02 1532.01 Q1307.88 1532.49 1310.3 1533.45 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M1335.22 1518.52 L1335.22 1576.64 L1321.73 1576.64 L1321.73 1572.09 L1329.33 1572.09 L1329.33 1523.07 L1321.73 1523.07 L1321.73 1518.52 L1335.22 1518.52 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip062)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  193.936,1276.04 2352.76,1276.04 
  "/>
<polyline clip-path="url(#clip062)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  193.936,1005.61 2352.76,1005.61 
  "/>
<polyline clip-path="url(#clip062)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  193.936,735.195 2352.76,735.195 
  "/>
<polyline clip-path="url(#clip062)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  193.936,464.774 2352.76,464.774 
  "/>
<polyline clip-path="url(#clip062)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  193.936,194.354 2352.76,194.354 
  "/>
<polyline clip-path="url(#clip060)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  193.936,1423.18 193.936,47.2441 
  "/>
<polyline clip-path="url(#clip060)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  193.936,1276.04 212.834,1276.04 
  "/>
<polyline clip-path="url(#clip060)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  193.936,1005.61 212.834,1005.61 
  "/>
<polyline clip-path="url(#clip060)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  193.936,735.195 212.834,735.195 
  "/>
<polyline clip-path="url(#clip060)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  193.936,464.774 212.834,464.774 
  "/>
<polyline clip-path="url(#clip060)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  193.936,194.354 212.834,194.354 
  "/>
<path clip-path="url(#clip060)" d="M50.9921 1276.49 L80.6679 1276.49 L80.6679 1280.42 L50.9921 1280.42 L50.9921 1276.49 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M90.8067 1258.76 L109.163 1258.76 L109.163 1262.69 L95.0891 1262.69 L95.0891 1271.16 Q96.1076 1270.82 97.1261 1270.65 Q98.1447 1270.47 99.1632 1270.47 Q104.95 1270.47 108.33 1273.64 Q111.709 1276.81 111.709 1282.23 Q111.709 1287.81 108.237 1290.91 Q104.765 1293.99 98.4456 1293.99 Q96.2697 1293.99 94.0012 1293.62 Q91.7558 1293.25 89.3484 1292.51 L89.3484 1287.81 Q91.4317 1288.94 93.6539 1289.5 Q95.8761 1290.05 98.353 1290.05 Q102.358 1290.05 104.696 1287.94 Q107.033 1285.84 107.033 1282.23 Q107.033 1278.62 104.696 1276.51 Q102.358 1274.4 98.353 1274.4 Q96.478 1274.4 94.603 1274.82 Q92.7512 1275.24 90.8067 1276.12 L90.8067 1258.76 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M120.922 1287.44 L125.807 1287.44 L125.807 1293.32 L120.922 1293.32 L120.922 1287.44 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M145.992 1261.83 Q142.381 1261.83 140.552 1265.4 Q138.746 1268.94 138.746 1276.07 Q138.746 1283.18 140.552 1286.74 Q142.381 1290.28 145.992 1290.28 Q149.626 1290.28 151.431 1286.74 Q153.26 1283.18 153.26 1276.07 Q153.26 1268.94 151.431 1265.4 Q149.626 1261.83 145.992 1261.83 M145.992 1258.13 Q151.802 1258.13 154.857 1262.74 Q157.936 1267.32 157.936 1276.07 Q157.936 1284.8 154.857 1289.4 Q151.802 1293.99 145.992 1293.99 Q140.181 1293.99 137.103 1289.4 Q134.047 1284.8 134.047 1276.07 Q134.047 1267.32 137.103 1262.74 Q140.181 1258.13 145.992 1258.13 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M51.9875 1006.07 L81.6633 1006.07 L81.6633 1010 L51.9875 1010 L51.9875 1006.07 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M95.7836 1018.96 L112.103 1018.96 L112.103 1022.89 L90.1586 1022.89 L90.1586 1018.96 Q92.8206 1016.21 97.4039 1011.58 Q102.01 1006.92 103.191 1005.58 Q105.436 1003.06 106.316 1001.32 Q107.219 999.562 107.219 997.872 Q107.219 995.117 105.274 993.381 Q103.353 991.645 100.251 991.645 Q98.0521 991.645 95.5984 992.409 Q93.1678 993.173 90.3901 994.724 L90.3901 990.002 Q93.2141 988.867 95.6678 988.289 Q98.1215 987.71 100.159 987.71 Q105.529 987.71 108.723 990.395 Q111.918 993.08 111.918 997.571 Q111.918 999.701 111.108 1001.62 Q110.321 1003.52 108.214 1006.11 Q107.635 1006.78 104.534 1010 Q101.432 1013.2 95.7836 1018.96 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M121.918 1017.02 L126.802 1017.02 L126.802 1022.89 L121.918 1022.89 L121.918 1017.02 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M137.033 988.335 L155.39 988.335 L155.39 992.27 L141.316 992.27 L141.316 1000.74 Q142.334 1000.4 143.353 1000.23 Q144.371 1000.05 145.39 1000.05 Q151.177 1000.05 154.556 1003.22 Q157.936 1006.39 157.936 1011.81 Q157.936 1017.39 154.464 1020.49 Q150.992 1023.57 144.672 1023.57 Q142.496 1023.57 140.228 1023.2 Q137.982 1022.83 135.575 1022.08 L135.575 1017.39 Q137.658 1018.52 139.881 1019.08 Q142.103 1019.63 144.58 1019.63 Q148.584 1019.63 150.922 1017.52 Q153.26 1015.42 153.26 1011.81 Q153.26 1008.2 150.922 1006.09 Q148.584 1003.98 144.58 1003.98 Q142.705 1003.98 140.83 1004.4 Q138.978 1004.82 137.033 1005.7 L137.033 988.335 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M100.76 720.993 Q97.1493 720.993 95.3206 724.558 Q93.515 728.1 93.515 735.229 Q93.515 742.336 95.3206 745.9 Q97.1493 749.442 100.76 749.442 Q104.395 749.442 106.2 745.9 Q108.029 742.336 108.029 735.229 Q108.029 728.1 106.2 724.558 Q104.395 720.993 100.76 720.993 M100.76 717.29 Q106.571 717.29 109.626 721.896 Q112.705 726.479 112.705 735.229 Q112.705 743.956 109.626 748.563 Q106.571 753.146 100.76 753.146 Q94.9502 753.146 91.8715 748.563 Q88.816 743.956 88.816 735.229 Q88.816 726.479 91.8715 721.896 Q94.9502 717.29 100.76 717.29 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M120.922 746.595 L125.807 746.595 L125.807 752.475 L120.922 752.475 L120.922 746.595 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M145.992 720.993 Q142.381 720.993 140.552 724.558 Q138.746 728.1 138.746 735.229 Q138.746 742.336 140.552 745.9 Q142.381 749.442 145.992 749.442 Q149.626 749.442 151.431 745.9 Q153.26 742.336 153.26 735.229 Q153.26 728.1 151.431 724.558 Q149.626 720.993 145.992 720.993 M145.992 717.29 Q151.802 717.29 154.857 721.896 Q157.936 726.479 157.936 735.229 Q157.936 743.956 154.857 748.563 Q151.802 753.146 145.992 753.146 Q140.181 753.146 137.103 748.563 Q134.047 743.956 134.047 735.229 Q134.047 726.479 137.103 721.896 Q140.181 717.29 145.992 717.29 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M95.7836 478.119 L112.103 478.119 L112.103 482.054 L90.1586 482.054 L90.1586 478.119 Q92.8206 475.364 97.4039 470.735 Q102.01 466.082 103.191 464.739 Q105.436 462.216 106.316 460.48 Q107.219 458.721 107.219 457.031 Q107.219 454.277 105.274 452.54 Q103.353 450.804 100.251 450.804 Q98.0521 450.804 95.5984 451.568 Q93.1678 452.332 90.3901 453.883 L90.3901 449.161 Q93.2141 448.027 95.6678 447.448 Q98.1215 446.869 100.159 446.869 Q105.529 446.869 108.723 449.554 Q111.918 452.24 111.918 456.73 Q111.918 458.86 111.108 460.781 Q110.321 462.679 108.214 465.272 Q107.635 465.943 104.534 469.161 Q101.432 472.355 95.7836 478.119 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M121.918 476.175 L126.802 476.175 L126.802 482.054 L121.918 482.054 L121.918 476.175 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M137.033 447.494 L155.39 447.494 L155.39 451.429 L141.316 451.429 L141.316 459.902 Q142.334 459.554 143.353 459.392 Q144.371 459.207 145.39 459.207 Q151.177 459.207 154.556 462.378 Q157.936 465.55 157.936 470.966 Q157.936 476.545 154.464 479.647 Q150.992 482.725 144.672 482.725 Q142.496 482.725 140.228 482.355 Q137.982 481.985 135.575 481.244 L135.575 476.545 Q137.658 477.679 139.881 478.235 Q142.103 478.79 144.58 478.79 Q148.584 478.79 150.922 476.684 Q153.26 474.577 153.26 470.966 Q153.26 467.355 150.922 465.249 Q148.584 463.142 144.58 463.142 Q142.705 463.142 140.83 463.559 Q138.978 463.976 137.033 464.855 L137.033 447.494 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M90.8067 177.074 L109.163 177.074 L109.163 181.009 L95.0891 181.009 L95.0891 189.481 Q96.1076 189.134 97.1261 188.972 Q98.1447 188.787 99.1632 188.787 Q104.95 188.787 108.33 191.958 Q111.709 195.129 111.709 200.546 Q111.709 206.125 108.237 209.226 Q104.765 212.305 98.4456 212.305 Q96.2697 212.305 94.0012 211.935 Q91.7558 211.564 89.3484 210.824 L89.3484 206.125 Q91.4317 207.259 93.6539 207.814 Q95.8761 208.37 98.353 208.37 Q102.358 208.37 104.696 206.263 Q107.033 204.157 107.033 200.546 Q107.033 196.935 104.696 194.828 Q102.358 192.722 98.353 192.722 Q96.478 192.722 94.603 193.139 Q92.7512 193.555 90.8067 194.435 L90.8067 177.074 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M120.922 205.754 L125.807 205.754 L125.807 211.634 L120.922 211.634 L120.922 205.754 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M145.992 180.153 Q142.381 180.153 140.552 183.717 Q138.746 187.259 138.746 194.389 Q138.746 201.495 140.552 205.06 Q142.381 208.601 145.992 208.601 Q149.626 208.601 151.431 205.06 Q153.26 201.495 153.26 194.389 Q153.26 187.259 151.431 183.717 Q149.626 180.153 145.992 180.153 M145.992 176.449 Q151.802 176.449 154.857 181.055 Q157.936 185.639 157.936 194.389 Q157.936 203.115 154.857 207.722 Q151.802 212.305 145.992 212.305 Q140.181 212.305 137.103 207.722 Q134.047 203.115 134.047 194.389 Q134.047 185.639 137.103 181.055 Q140.181 176.449 145.992 176.449 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip062)" style="stroke:#009af9; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  255.035,681.11 259.108,681.078 263.181,680.981 267.254,680.818 271.328,680.592 275.401,680.301 279.474,679.946 283.547,679.527 287.621,679.045 291.694,678.5 
  295.767,677.893 299.84,677.224 303.914,676.494 307.987,675.704 312.06,674.855 316.133,673.947 320.207,672.981 324.28,671.958 328.353,670.88 332.426,669.748 
  336.5,668.561 340.573,667.322 344.646,666.032 348.719,664.692 352.793,663.304 356.866,661.868 360.939,660.387 365.012,658.861 369.086,657.293 373.159,655.683 
  377.232,654.034 381.305,652.348 385.379,650.625 389.452,648.868 393.525,647.078 397.598,645.257 401.671,643.407 405.745,641.53 409.818,639.628 413.891,637.702 
  417.964,635.755 422.038,633.788 426.111,631.803 430.184,629.803 434.257,627.79 438.331,625.765 442.404,623.73 446.477,621.688 450.55,619.64 454.624,617.589 
  458.697,615.536 462.77,613.485 466.843,611.435 470.917,609.391 474.99,607.354 479.063,605.325 483.136,603.307 487.21,601.302 491.283,599.312 495.356,597.339 
  499.429,595.385 503.503,593.451 507.576,591.541 511.649,589.655 515.722,587.795 519.796,585.964 523.869,584.164 527.942,582.395 532.015,580.66 536.089,578.961 
  540.162,577.298 544.235,575.675 548.308,574.093 552.382,572.552 556.455,571.055 560.528,569.603 564.601,568.198 568.675,566.841 572.748,565.533 576.821,564.276 
  580.894,563.071 584.968,561.919 589.041,560.821 593.114,559.779 597.187,558.793 601.261,557.865 605.334,556.995 609.407,556.184 613.48,555.433 617.553,554.743 
  621.627,554.114 625.7,553.547 629.773,553.044 633.846,552.603 637.92,552.226 641.993,551.913 646.066,551.664 650.139,551.479 654.213,551.36 658.286,551.305 
  662.359,551.315 666.432,551.39 670.506,551.53 674.579,551.735 678.652,552.004 682.725,552.337 686.799,552.734 690.872,553.195 694.945,553.719 699.018,554.305 
  703.092,554.953 707.165,555.662 711.238,556.432 715.311,557.262 719.385,558.15 723.458,559.097 727.531,560.101 731.604,561.161 735.678,562.276 739.751,563.445 
  743.824,564.666 747.897,565.94 751.971,567.263 756.044,568.635 760.117,570.055 764.19,571.522 768.264,573.032 772.337,574.586 776.41,576.182 780.483,577.817 
  784.557,579.491 788.63,581.202 792.703,582.948 796.776,584.727 800.85,586.537 804.923,588.377 808.996,590.245 813.069,592.139 817.142,594.057 821.216,595.997 
  825.289,597.957 829.362,599.936 833.435,601.931 837.509,603.94 841.582,605.962 845.655,607.993 849.728,610.033 853.802,612.079 857.875,614.13 861.948,616.182 
  866.021,618.234 870.095,620.285 874.168,622.331 878.241,624.371 882.314,626.403 886.388,628.425 890.461,630.435 894.534,632.43 898.607,634.409 902.681,636.37 
  906.754,638.311 910.827,640.23 914.9,642.124 918.974,643.993 923.047,645.834 927.12,647.645 931.193,649.425 935.267,651.171 939.34,652.883 943.413,654.558 
  947.486,656.194 951.56,657.791 955.633,659.345 959.706,660.857 963.779,662.324 967.853,663.745 971.926,665.118 975.999,666.443 980.072,667.717 984.146,668.939 
  988.219,670.109 992.292,671.225 996.365,672.286 1000.44,673.291 1004.51,674.239 1008.59,675.128 1012.66,675.959 1016.73,676.73 1020.8,677.441 1024.88,678.09 
  1028.95,678.677 1033.02,679.202 1037.1,679.664 1041.17,680.062 1045.24,680.397 1049.32,680.668 1053.39,680.874 1057.46,681.015 1061.54,681.091 1065.61,681.103 
  1069.68,681.049 1073.76,680.931 1077.83,680.748 1081.9,680.501 1085.98,680.19 1090.05,679.814 1094.12,679.374 1098.2,678.872 1102.27,678.307 1106.34,677.679 
  1110.42,676.991 1114.49,676.242 1118.56,675.432 1122.64,674.563 1126.71,673.636 1130.78,672.651 1134.86,671.611 1138.93,670.514 1143,669.364 1147.08,668.16 
  1151.15,666.904 1155.22,665.598 1159.3,664.242 1163.37,662.838 1167.44,661.387 1171.51,659.892 1175.59,658.353 1179.66,656.771 1183.73,655.149 1187.81,653.488 
  1191.88,651.79 1195.95,650.056 1200.03,648.289 1204.1,646.489 1208.17,644.659 1212.25,642.801 1216.32,640.916 1220.39,639.007 1224.47,637.074 1228.54,635.121 
  1232.61,633.149 1236.69,631.16 1240.76,629.156 1244.83,627.139 1248.91,625.111 1252.98,623.074 1257.05,621.031 1261.13,618.982 1265.2,616.931 1269.27,614.879 
  1273.35,612.829 1277.42,610.782 1281.49,608.74 1285.57,606.706 1289.64,604.681 1293.71,602.668 1297.79,600.668 1301.86,598.684 1305.93,596.718 1310.01,594.771 
  1314.08,592.845 1318.15,590.943 1322.22,589.067 1326.3,587.217 1330.37,585.396 1334.44,583.606 1338.52,581.849 1342.59,580.126 1346.66,578.44 1350.74,576.791 
  1354.81,575.181 1358.88,573.613 1362.96,572.087 1367.03,570.605 1371.1,569.169 1375.18,567.78 1379.25,566.439 1383.32,565.148 1387.4,563.909 1391.47,562.721 
  1395.54,561.587 1399.62,560.508 1403.69,559.484 1407.76,558.518 1411.84,557.608 1415.91,556.758 1419.98,555.967 1424.06,555.236 1428.13,554.566 1432.2,553.958 
  1436.28,553.412 1440.35,552.929 1444.42,552.509 1448.5,552.153 1452.57,551.861 1456.64,551.633 1460.72,551.47 1464.79,551.371 1468.86,551.338 1472.93,551.369 
  1477.01,551.466 1481.08,551.626 1485.15,551.852 1489.23,552.142 1493.3,552.495 1497.37,552.913 1501.45,553.394 1505.52,553.938 1509.59,554.544 1513.67,555.212 
  1517.74,555.94 1521.81,556.729 1525.89,557.578 1529.96,558.485 1534.03,559.45 1538.11,560.471 1542.18,561.549 1546.25,562.68 1550.33,563.866 1554.4,565.103 
  1558.47,566.392 1562.55,567.731 1566.62,569.118 1570.69,570.553 1574.77,572.033 1578.84,573.557 1582.91,575.123 1586.99,576.731 1591.06,578.378 1595.13,580.063 
  1599.21,581.785 1603.28,583.54 1607.35,585.329 1611.43,587.148 1615.5,588.996 1619.57,590.871 1623.65,592.772 1627.72,594.696 1631.79,596.642 1635.86,598.607 
  1639.94,600.59 1644.01,602.588 1648.08,604.6 1652.16,606.624 1656.23,608.657 1660.3,610.698 1664.38,612.744 1668.45,614.794 1672.52,616.845 1676.6,618.895 
  1680.67,620.943 1684.74,622.986 1688.82,625.022 1692.89,627.049 1696.96,629.066 1701.04,631.069 1705.11,633.058 1709.18,635.03 1713.26,636.983 1717.33,638.915 
  1721.4,640.824 1725.48,642.709 1729.55,644.568 1733.62,646.398 1737.7,648.198 1741.77,649.965 1745.84,651.699 1749.92,653.398 1753.99,655.059 1758.06,656.682 
  1762.14,658.263 1766.21,659.803 1770.28,661.3 1774.36,662.751 1778.43,664.155 1782.5,665.512 1786.57,666.819 1790.65,668.076 1794.72,669.281 1798.79,670.433 
  1802.87,671.53 1806.94,672.573 1811.01,673.558 1815.09,674.487 1819.16,675.357 1823.23,676.168 1827.31,676.919 1831.38,677.609 1835.45,678.238 1839.53,678.805 
  1843.6,679.309 1847.67,679.75 1851.75,680.128 1855.82,680.442 1859.89,680.691 1863.97,680.876 1868.04,680.997 1872.11,681.052 1876.19,681.043 1880.26,680.969 
  1884.33,680.83 1888.41,680.627 1892.48,680.359 1896.55,680.027 1900.63,679.632 1904.7,679.173 1908.77,678.65 1912.85,678.066 1916.92,677.42 1920.99,676.712 
  1925.07,675.944 1929.14,675.116 1933.21,674.23 1937.28,673.285 1941.36,672.283 1945.43,671.225 1949.5,670.112 1953.58,668.946 1957.65,667.726 1961.72,666.455 
  1965.8,665.134 1969.87,663.764 1973.94,662.346 1978.02,660.882 1982.09,659.373 1986.16,657.822 1990.24,656.228 1994.31,654.595 1998.38,652.923 2002.46,651.215 
  2006.53,649.472 2010.6,647.695 2014.68,645.887 2018.75,644.049 2022.82,642.184 2026.9,640.292 2030.97,638.376 2035.04,636.438 2039.12,634.48 2043.19,632.504 
  2047.26,630.511 2051.34,628.505 2055.41,626.485 2059.48,624.456 2063.56,622.418 2067.63,620.374 2071.7,618.326 2075.78,616.276 2079.85,614.226 2083.92,612.177 
  2087.99,610.133 2092.07,608.095 2096.14,606.065 2100.21,604.045 2104.29,602.037 2108.36,600.044 2112.43,598.066 2116.51,596.107 2120.58,594.168 2124.65,592.251 
  2128.73,590.358 2132.8,588.491 2136.87,586.651 2140.95,584.841 2145.02,583.062 2149.09,581.317 2153.17,579.606 2157.24,577.932 2161.31,576.297 2165.39,574.701 
  2169.46,573.147 2173.53,571.635 2177.61,570.169 2181.68,568.748 2185.75,567.375 2189.83,566.051 2193.9,564.777 2197.97,563.554 2202.05,562.384 2206.12,561.268 
  2210.19,560.206 2214.27,559.201 2218.34,558.253 2222.41,557.362 2226.49,556.531 2230.56,555.759 2234.63,555.048 2238.71,554.398 2242.78,553.809 2246.85,553.283 
  2250.92,552.82 2255,552.421 2259.07,552.085 2263.14,551.813 2267.22,551.605 2271.29,551.463 2275.36,551.385 2279.44,551.371 2283.51,551.423 2287.58,551.539 
  2291.66,551.721 
  "/>
<polyline clip-path="url(#clip062)" style="stroke:#e26f46; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  255.035,735.195 259.108,728.706 263.181,722.224 267.254,715.754 271.328,709.304 275.401,702.88 279.474,696.488 283.547,690.135 287.621,683.826 291.694,677.569 
  295.767,671.37 299.84,665.234 303.914,659.169 307.987,653.179 312.06,647.271 316.133,641.451 320.207,635.725 324.28,630.098 328.353,624.576 332.426,619.164 
  336.5,613.868 340.573,608.694 344.646,603.646 348.719,598.73 352.793,593.95 356.866,589.312 360.939,584.819 365.012,580.477 369.086,576.29 373.159,572.261 
  377.232,568.396 381.305,564.697 385.379,561.169 389.452,557.815 393.525,554.638 397.598,551.642 401.671,548.829 405.745,546.203 409.818,543.766 413.891,541.52 
  417.964,539.468 422.038,537.611 426.111,535.952 430.184,534.493 434.257,533.234 438.331,532.177 442.404,531.323 446.477,530.673 450.55,530.227 454.624,529.986 
  458.697,529.951 462.77,530.121 466.843,530.496 470.917,531.076 474.99,531.859 479.063,532.846 483.136,534.036 487.21,535.426 491.283,537.017 495.356,538.805 
  499.429,540.79 503.503,542.97 507.576,545.341 511.649,547.903 515.722,550.652 519.796,553.585 523.869,556.7 527.942,559.993 532.015,563.462 536.089,567.102 
  540.162,570.911 544.235,574.883 548.308,579.016 552.382,583.306 556.455,587.747 560.528,592.335 564.601,597.067 568.675,601.936 572.748,606.939 576.821,612.07 
  580.894,617.325 584.968,622.697 589.041,628.181 593.114,633.773 597.187,639.466 601.261,645.255 605.334,651.134 609.407,657.096 613.48,663.137 617.553,669.25 
  621.627,675.429 625.7,681.668 629.773,687.96 633.846,694.299 637.92,700.68 641.993,707.095 646.066,713.537 650.139,720.002 654.213,726.482 658.286,732.971 
  662.359,739.461 666.432,745.948 670.506,752.424 674.579,758.882 678.652,765.317 682.725,771.722 686.799,778.09 690.872,784.415 694.945,790.691 699.018,796.912 
  703.092,803.071 707.165,809.162 711.238,815.179 715.311,821.116 719.385,826.968 723.458,832.728 727.531,838.391 731.604,843.95 735.678,849.4 739.751,854.736 
  743.824,859.953 747.897,865.045 751.971,870.007 756.044,874.833 760.117,879.52 764.19,884.062 768.264,888.456 772.337,892.696 776.41,896.779 780.483,900.7 
  784.557,904.455 788.63,908.042 792.703,911.455 796.776,914.693 800.85,917.75 804.923,920.626 808.996,923.316 813.069,925.818 817.142,928.129 821.216,930.247 
  825.289,932.171 829.362,933.897 833.435,935.425 837.509,936.753 841.582,937.879 845.655,938.802 849.728,939.522 853.802,940.037 857.875,940.348 861.948,940.453 
  866.021,940.353 870.095,940.047 874.168,939.537 878.241,938.822 882.314,937.904 886.388,936.783 890.461,935.46 894.534,933.938 898.607,932.216 902.681,930.296 
  906.754,928.182 910.827,925.875 914.9,923.376 918.974,920.69 923.047,917.819 927.12,914.765 931.193,911.531 935.267,908.121 939.34,904.537 943.413,900.785 
  947.486,896.866 951.56,892.787 955.633,888.55 959.706,884.16 963.779,879.62 967.853,874.936 971.926,870.112 975.999,865.154 980.072,860.065 984.146,854.852 
  988.219,849.52 992.292,844.073 996.365,838.517 1000.44,832.858 1004.51,827.101 1008.59,821.252 1012.66,815.318 1016.73,809.303 1020.8,803.215 1024.88,797.058 
  1028.95,790.84 1033.02,784.565 1037.1,778.242 1041.17,771.875 1045.24,765.472 1049.32,759.039 1053.39,752.582 1057.46,746.107 1061.54,739.622 1065.61,733.132 
  1069.68,726.644 1073.76,720.165 1077.83,713.701 1081.9,707.259 1085.98,700.845 1090.05,694.465 1094.12,688.125 1098.2,681.833 1102.27,675.594 1106.34,669.415 
  1110.42,663.302 1114.49,657.261 1118.56,651.297 1122.64,645.418 1126.71,639.628 1130.78,633.934 1134.86,628.342 1138.93,622.856 1143,617.483 1147.08,612.227 
  1151.15,607.095 1155.22,602.09 1159.3,597.219 1163.37,592.486 1167.44,587.895 1171.51,583.452 1175.59,579.161 1179.66,575.025 1183.73,571.05 1187.81,567.24 
  1191.88,563.597 1195.95,560.126 1200.03,556.83 1204.1,553.712 1208.17,550.775 1212.25,548.024 1216.32,545.459 1220.39,543.085 1224.47,540.902 1228.54,538.914 
  1232.61,537.122 1236.69,535.528 1240.76,534.133 1244.83,532.94 1248.91,531.949 1252.98,531.162 1257.05,530.578 1261.13,530.199 1265.2,530.025 1269.27,530.056 
  1273.35,530.292 1277.42,530.734 1281.49,531.38 1285.57,532.23 1289.64,533.282 1293.71,534.537 1297.79,535.992 1301.86,537.646 1305.93,539.497 1310.01,541.545 
  1314.08,543.787 1318.15,546.22 1322.22,548.841 1326.3,551.649 1330.37,554.64 1334.44,557.812 1338.52,561.161 1342.59,564.685 1346.66,568.379 1350.74,572.24 
  1354.81,576.264 1358.88,580.446 1362.96,584.783 1367.03,589.27 1371.1,593.904 1375.18,598.679 1379.25,603.59 1383.32,608.634 1387.4,613.803 1391.47,619.094 
  1395.54,624.5 1399.62,630.017 1403.69,635.64 1407.76,641.362 1411.84,647.178 1415.91,653.082 1419.98,659.068 1424.06,665.13 1428.13,671.262 1432.2,677.458 
  1436.28,683.712 1440.35,690.017 1444.42,696.367 1448.5,702.756 1452.57,709.178 1456.64,715.625 1460.72,722.091 1464.79,728.571 1468.86,735.058 1472.93,741.545 
  1477.01,748.025 1481.08,754.492 1485.15,760.94 1489.23,767.362 1493.3,773.752 1497.37,780.104 1501.45,786.41 1505.52,792.665 1509.59,798.863 1513.67,804.997 
  1517.74,811.061 1521.81,817.049 1525.89,822.955 1529.96,828.773 1534.03,834.498 1538.11,840.123 1542.18,845.644 1546.25,851.054 1550.33,856.348 1554.4,861.521 
  1558.47,866.567 1562.55,871.482 1566.62,876.26 1570.69,880.898 1574.77,885.39 1578.84,889.731 1582.91,893.919 1586.99,897.947 1591.06,901.812 1595.13,905.511 
  1599.21,909.039 1603.28,912.394 1607.35,915.571 1611.43,918.568 1615.5,921.382 1619.57,924.009 1623.65,926.448 1627.72,928.695 1631.79,930.748 1635.86,932.606 
  1639.94,934.267 1644.01,935.728 1648.08,936.99 1652.16,938.049 1656.23,938.905 1660.3,939.557 1664.38,940.005 1668.45,940.248 1672.52,940.287 1676.6,940.12 
  1680.67,939.749 1684.74,939.172 1688.82,938.392 1692.89,937.408 1696.96,936.222 1701.04,934.835 1705.11,933.249 1709.18,931.465 1713.26,929.485 1717.33,927.309 
  1721.4,924.942 1725.48,922.385 1729.55,919.641 1733.62,916.712 1737.7,913.602 1741.77,910.314 1745.84,906.851 1749.92,903.216 1753.99,899.412 1758.06,895.445 
  1762.14,891.317 1766.21,887.034 1770.28,882.598 1774.36,878.016 1778.43,873.29 1782.5,868.427 1786.57,863.429 1790.65,858.304 1794.72,853.056 1798.79,847.69 
  1802.87,842.212 1806.94,836.627 1811.01,830.94 1815.09,825.157 1819.16,819.284 1823.23,813.328 1827.31,807.293 1831.38,801.186 1835.45,795.014 1839.53,788.782 
  1843.6,782.496 1847.67,776.163 1851.75,769.788 1855.82,763.38 1859.89,756.943 1863.97,750.484 1868.04,744.011 1872.11,737.529 1876.19,731.044 1880.26,724.564 
  1884.33,718.094 1888.41,711.642 1892.48,705.213 1896.55,698.814 1900.63,692.451 1904.7,686.132 1908.77,679.861 1912.85,673.646 1916.92,667.492 1920.99,661.406 
  1925.07,655.394 1929.14,649.461 1933.21,643.614 1937.28,637.859 1941.36,632.201 1945.43,626.646 1949.5,621.2 1953.58,615.868 1957.65,610.655 1961.72,605.566 
  1965.8,600.607 1969.87,595.783 1973.94,591.098 1978.02,586.557 1982.09,582.165 1986.16,577.926 1990.24,573.844 1994.31,569.924 1998.38,566.169 2002.46,562.582 
  2006.53,559.169 2010.6,555.931 2014.68,552.873 2018.75,549.997 2022.82,547.307 2026.9,544.804 2030.97,542.491 2035.04,540.371 2039.12,538.446 2043.19,536.718 
  2047.26,535.188 2051.34,533.859 2055.41,532.73 2059.48,531.804 2063.56,531.082 2067.63,530.563 2071.7,530.249 2075.78,530.14 2079.85,530.236 2083.92,530.537 
  2087.99,531.043 2092.07,531.753 2096.14,532.666 2100.21,533.781 2104.29,535.099 2108.36,536.616 2112.43,538.332 2116.51,540.244 2120.58,542.352 2124.65,544.652 
  2128.73,547.143 2132.8,549.822 2136.87,552.687 2140.95,555.733 2145.02,558.96 2149.09,562.362 2153.17,565.937 2157.24,569.682 2161.31,573.592 2165.39,577.664 
  2169.46,581.893 2173.53,586.275 2177.61,590.806 2181.68,595.482 2185.75,600.298 2189.83,605.248 2193.9,610.328 2197.97,615.533 2202.05,620.858 2206.12,626.297 
  2210.19,631.845 2214.27,637.496 2218.34,643.245 2222.41,649.085 2226.49,655.012 2230.56,661.019 2234.63,667.101 2238.71,673.25 2242.78,679.461 2246.85,685.728 
  2250.92,692.045 2255,698.404 2259.07,704.801 2263.14,711.227 2267.22,717.678 2271.29,724.146 2275.36,730.625 2279.44,737.109 2283.51,743.591 2287.58,750.064 
  2291.66,756.523 
  "/>
<polyline clip-path="url(#clip062)" style="stroke:#3da44d; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  255.035,86.1857 259.108,86.5121 263.181,87.4852 267.254,89.1057 271.328,91.3722 275.401,94.2819 279.474,97.832 283.547,102.019 287.621,106.839 291.694,112.288 
  295.767,118.359 299.84,125.048 303.914,132.345 307.987,140.247 312.06,148.741 316.133,157.821 320.207,167.478 324.28,177.705 328.353,188.486 332.426,199.814 
  336.5,211.679 340.573,224.07 344.646,236.97 348.719,250.369 352.793,264.253 356.866,278.61 360.939,293.424 365.012,308.68 369.086,324.363 373.159,340.457 
  377.232,356.946 381.305,373.813 385.379,391.042 389.452,408.614 393.525,426.514 397.598,444.722 401.671,463.221 405.745,481.991 409.818,501.015 413.891,520.272 
  417.964,539.745 422.038,559.413 426.111,579.256 430.184,599.256 434.257,619.392 438.331,639.643 442.404,659.99 446.477,680.413 450.55,700.89 454.624,721.401 
  458.697,741.926 462.77,762.445 466.843,782.936 470.917,803.38 474.99,823.755 479.063,844.042 483.136,864.22 487.21,884.269 491.283,904.169 495.356,923.9 
  499.429,943.442 503.503,962.776 507.576,981.883 511.649,1000.74 515.722,1019.34 519.796,1037.65 523.869,1055.65 527.942,1073.34 532.015,1090.69 536.089,1107.68 
  540.162,1124.31 544.235,1140.54 548.308,1156.36 552.382,1171.77 556.455,1186.74 560.528,1201.26 564.601,1215.31 568.675,1228.88 572.748,1241.96 576.821,1254.53 
  580.894,1266.58 584.968,1278.1 589.041,1289.08 593.114,1299.5 597.187,1309.36 601.261,1318.64 605.334,1327.34 609.407,1335.45 613.48,1342.96 617.553,1349.86 
  621.627,1356.15 625.7,1361.82 629.773,1366.85 633.846,1371.26 637.92,1375.03 641.993,1378.16 646.066,1380.65 650.139,1382.5 654.213,1383.69 658.286,1384.24 
  662.359,1384.14 666.432,1383.39 670.506,1381.99 674.579,1379.94 678.652,1377.25 682.725,1373.92 686.799,1369.95 690.872,1365.34 694.945,1360.1 699.018,1354.24 
  703.092,1347.76 707.165,1340.67 711.238,1332.97 715.311,1324.67 719.385,1315.79 723.458,1306.32 727.531,1296.28 731.604,1285.68 735.678,1274.53 739.751,1262.84 
  743.824,1250.63 747.897,1237.89 751.971,1224.66 756.044,1210.94 760.117,1196.74 764.19,1182.08 768.264,1166.97 772.337,1151.43 776.41,1135.47 780.483,1119.12 
  784.557,1102.38 788.63,1085.27 792.703,1067.81 796.776,1050.02 800.85,1031.92 804.923,1013.52 808.996,994.838 813.069,975.9 817.142,956.72 821.216,937.319 
  825.289,917.715 829.362,897.93 833.435,877.981 837.509,857.89 841.582,837.675 845.655,817.358 849.728,796.958 853.802,776.497 857.875,755.995 861.948,735.472 
  866.021,714.949 870.095,694.445 874.168,673.981 878.241,653.579 882.314,633.258 886.388,613.04 890.461,592.944 894.534,572.991 898.607,553.199 902.681,533.589 
  906.754,514.181 910.827,494.994 914.9,476.048 918.974,457.361 923.047,438.952 927.12,420.841 931.193,403.043 935.267,385.578 939.34,368.462 943.413,351.714 
  947.486,335.349 951.56,319.385 955.633,303.836 959.706,288.719 963.779,274.049 967.853,259.841 971.926,246.108 975.999,232.865 980.072,220.124 984.146,207.897 
  988.219,196.198 992.292,185.038 996.365,174.429 1000.44,164.381 1004.51,154.905 1008.59,146.008 1012.66,137.7 1016.73,129.989 1020.8,122.883 1024.88,116.39 
  1028.95,110.517 1033.02,105.269 1037.1,100.65 1041.17,96.6656 1045.24,93.3193 1049.32,90.6143 1053.39,88.5535 1057.46,87.1413 1061.54,86.3772 1065.61,86.2619 
  1069.68,86.7954 1073.76,87.9771 1077.83,89.8054 1081.9,92.2782 1085.98,95.3945 1090.05,99.1522 1094.12,103.546 1098.2,108.571 1102.27,114.223 1106.34,120.495 
  1110.42,127.382 1114.49,134.875 1118.56,142.97 1122.64,151.658 1126.71,160.93 1130.78,170.775 1134.86,181.185 1138.93,192.149 1143,203.654 1147.08,215.69 
  1151.15,228.248 1155.22,241.313 1159.3,254.872 1163.37,268.911 1167.44,283.416 1171.51,298.371 1175.59,313.763 1179.66,329.576 1183.73,345.796 1187.81,362.406 
  1191.88,379.388 1195.95,396.726 1200.03,414.402 1204.1,432.398 1208.17,450.696 1212.25,469.279 1216.32,488.128 1220.39,507.225 1224.47,526.55 1228.54,546.082 
  1232.61,565.804 1236.69,585.694 1240.76,605.733 1244.83,625.902 1248.91,646.181 1252.98,666.549 1257.05,686.985 1261.13,707.469 1265.2,727.98 1269.27,748.498 
  1273.35,769.003 1277.42,789.474 1281.49,809.891 1285.57,830.233 1289.64,850.48 1293.71,870.611 1297.79,890.607 1301.86,910.446 1305.93,930.111 1310.01,949.58 
  1314.08,968.835 1318.15,987.857 1322.22,1006.62 1326.3,1025.12 1330.37,1043.33 1334.44,1061.23 1338.52,1078.8 1342.59,1096.03 1346.66,1112.89 1350.74,1129.38 
  1354.81,1145.48 1358.88,1161.16 1362.96,1176.42 1367.03,1191.24 1371.1,1205.6 1375.18,1219.49 1379.25,1232.9 1383.32,1245.81 1387.4,1258.2 1391.47,1270.08 
  1395.54,1281.42 1399.62,1292.21 1403.69,1302.45 1407.76,1312.12 1411.84,1321.21 1415.91,1329.71 1419.98,1337.62 1424.06,1344.93 1428.13,1351.63 1432.2,1357.71 
  1436.28,1363.17 1440.35,1368 1444.42,1372.2 1448.5,1375.76 1452.57,1378.68 1456.64,1380.96 1460.72,1382.59 1464.79,1383.58 1468.86,1383.91 1472.93,1383.6 
  1477.01,1382.64 1481.08,1381.03 1485.15,1378.77 1489.23,1375.87 1493.3,1372.34 1497.37,1368.16 1501.45,1363.35 1505.52,1357.91 1509.59,1351.85 1513.67,1345.17 
  1517.74,1337.89 1521.81,1330 1525.89,1321.51 1529.96,1312.44 1534.03,1302.79 1538.11,1292.58 1542.18,1281.8 1546.25,1270.49 1550.33,1258.63 1554.4,1246.26 
  1558.47,1233.37 1562.55,1219.98 1566.62,1206.11 1570.69,1191.76 1574.77,1176.96 1578.84,1161.72 1582.91,1146.06 1586.99,1129.98 1591.06,1113.51 1595.13,1096.66 
  1599.21,1079.44 1603.28,1061.89 1607.35,1044 1611.43,1025.81 1615.5,1007.33 1619.57,988.579 1623.65,969.571 1627.72,950.329 1631.79,930.872 1635.86,911.22 
  1639.94,891.392 1644.01,871.408 1648.08,851.29 1652.16,831.054 1656.23,810.722 1660.3,790.314 1664.38,769.852 1668.45,749.355 1672.52,728.845 1676.6,708.341 
  1680.67,687.865 1684.74,667.435 1688.82,647.073 1692.89,626.8 1696.96,606.635 1701.04,586.599 1705.11,566.712 1709.18,546.993 1713.26,527.463 1717.33,508.141 
  1721.4,489.046 1725.48,470.197 1729.55,451.613 1733.62,433.313 1737.7,415.315 1741.77,397.637 1745.84,380.297 1749.92,363.312 1753.99,346.699 1758.06,330.475 
  1762.14,314.656 1766.21,299.257 1770.28,284.294 1774.36,269.782 1778.43,255.736 1782.5,242.169 1786.57,229.096 1790.65,216.529 1794.72,204.481 1798.79,192.963 
  1802.87,181.988 1806.94,171.565 1811.01,161.707 1815.09,152.423 1819.16,143.722 1823.23,135.612 1827.31,128.102 1831.38,121.198 1835.45,114.908 1839.53,109.239 
  1843.6,104.196 1847.67,99.7854 1851.75,96.0098 1855.82,92.8732 1859.89,90.3787 1863.97,88.5284 1868.04,87.3239 1872.11,86.7676 1876.19,86.8599 1880.26,87.6003 
  1884.33,88.9879 1888.41,91.0213 1892.48,93.6985 1896.55,97.0174 1900.63,100.974 1904.7,105.565 1908.77,110.786 1912.85,116.63 1916.92,123.094 1920.99,130.169 
  1925.07,137.85 1929.14,146.128 1933.21,154.994 1937.28,164.441 1941.36,174.459 1945.43,185.038 1949.5,196.167 1953.58,207.835 1957.65,220.03 1961.72,232.739 
  1965.8,245.952 1969.87,259.654 1973.94,273.832 1978.02,288.471 1982.09,303.556 1986.16,319.072 1990.24,335.005 1994.31,351.339 1998.38,368.055 2002.46,385.139 
  2006.53,402.573 2010.6,420.339 2014.68,438.421 2018.75,456.799 2022.82,475.455 2026.9,494.371 2030.97,513.527 2035.04,532.906 2039.12,552.486 2043.19,572.249 
  2047.26,592.176 2051.34,612.245 2055.41,632.437 2059.48,652.731 2063.56,673.109 2067.63,693.548 2071.7,714.029 2075.78,734.531 2079.85,755.033 2083.92,775.516 
  2087.99,795.958 2092.07,816.34 2096.14,836.64 2100.21,856.839 2104.29,876.917 2108.36,896.852 2112.43,916.626 2116.51,936.218 2120.58,955.61 2124.65,974.78 
  2128.73,993.712 2132.8,1012.38 2136.87,1030.78 2140.95,1048.88 2145.02,1066.67 2149.09,1084.12 2153.17,1101.23 2157.24,1117.97 2161.31,1134.32 2165.39,1150.28 
  2169.46,1165.82 2173.53,1180.94 2177.61,1195.6 2181.68,1209.81 2185.75,1223.54 2189.83,1236.78 2193.9,1249.52 2197.97,1261.75 2202.05,1273.45 2206.12,1284.61 
  2210.19,1295.23 2214.27,1305.28 2218.34,1314.76 2222.41,1323.67 2226.49,1331.98 2230.56,1339.7 2234.63,1346.81 2238.71,1353.31 2242.78,1359.2 2246.85,1364.46 
  2250.92,1369.09 2255,1373.08 2259.07,1376.44 2263.14,1379.16 2267.22,1381.24 2271.29,1382.66 2275.36,1383.44 2279.44,1383.58 2283.51,1383.06 2287.58,1381.9 
  2291.66,1380.09 
  "/>
<path clip-path="url(#clip060)" d="
M1557.04 300.469 L2280.8 300.469 L2280.8 93.1086 L1557.04 93.1086  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<polyline clip-path="url(#clip060)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1557.04,300.469 2280.8,300.469 2280.8,93.1086 1557.04,93.1086 1557.04,300.469 
  "/>
<polyline clip-path="url(#clip060)" style="stroke:#009af9; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1581.03,144.949 1724.95,144.949 
  "/>
<path clip-path="url(#clip060)" d="M1769.28 141.28 Q1770.88 138.409 1773.1 137.044 Q1775.32 135.678 1778.33 135.678 Q1782.38 135.678 1784.58 138.525 Q1786.78 141.349 1786.78 146.581 L1786.78 162.229 L1782.5 162.229 L1782.5 146.719 Q1782.5 142.993 1781.18 141.187 Q1779.86 139.382 1777.15 139.382 Q1773.84 139.382 1771.92 141.581 Q1770 143.78 1770 147.576 L1770 162.229 L1765.72 162.229 L1765.72 146.719 Q1765.72 142.969 1764.4 141.187 Q1763.08 139.382 1760.32 139.382 Q1757.06 139.382 1755.14 141.604 Q1753.22 143.803 1753.22 147.576 L1753.22 162.229 L1748.94 162.229 L1748.94 136.303 L1753.22 136.303 L1753.22 140.331 Q1754.68 137.946 1756.71 136.812 Q1758.75 135.678 1761.55 135.678 Q1764.37 135.678 1766.34 137.113 Q1768.33 138.548 1769.28 141.28 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M1807.06 149.196 Q1801.9 149.196 1799.91 150.377 Q1797.92 151.557 1797.92 154.405 Q1797.92 156.673 1799.4 158.016 Q1800.9 159.335 1803.47 159.335 Q1807.01 159.335 1809.14 156.835 Q1811.3 154.312 1811.3 150.145 L1811.3 149.196 L1807.06 149.196 M1815.56 147.437 L1815.56 162.229 L1811.3 162.229 L1811.3 158.293 Q1809.84 160.655 1807.66 161.789 Q1805.49 162.9 1802.34 162.9 Q1798.36 162.9 1796 160.678 Q1793.66 158.432 1793.66 154.682 Q1793.66 150.307 1796.57 148.085 Q1799.51 145.863 1805.32 145.863 L1811.3 145.863 L1811.3 145.446 Q1811.3 142.507 1809.35 140.909 Q1807.43 139.289 1803.93 139.289 Q1801.71 139.289 1799.61 139.821 Q1797.5 140.354 1795.56 141.419 L1795.56 137.483 Q1797.89 136.581 1800.09 136.141 Q1802.29 135.678 1804.37 135.678 Q1810 135.678 1812.78 138.594 Q1815.56 141.511 1815.56 147.437 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M1840.86 137.067 L1840.86 141.094 Q1839.05 140.169 1837.11 139.706 Q1835.16 139.243 1833.08 139.243 Q1829.91 139.243 1828.31 140.215 Q1826.74 141.187 1826.74 143.131 Q1826.74 144.613 1827.87 145.469 Q1829 146.303 1832.43 147.067 L1833.89 147.391 Q1838.43 148.363 1840.32 150.145 Q1842.24 151.905 1842.24 155.076 Q1842.24 158.687 1839.37 160.793 Q1836.53 162.9 1831.53 162.9 Q1829.44 162.9 1827.18 162.483 Q1824.93 162.09 1822.43 161.28 L1822.43 156.881 Q1824.79 158.108 1827.08 158.733 Q1829.37 159.335 1831.62 159.335 Q1834.63 159.335 1836.25 158.317 Q1837.87 157.275 1837.87 155.4 Q1837.87 153.664 1836.69 152.738 Q1835.53 151.812 1831.57 150.956 L1830.09 150.608 Q1826.13 149.775 1824.37 148.062 Q1822.62 146.326 1822.62 143.317 Q1822.62 139.659 1825.21 137.669 Q1827.8 135.678 1832.57 135.678 Q1834.93 135.678 1837.01 136.025 Q1839.1 136.372 1840.86 137.067 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M1865.55 137.067 L1865.55 141.094 Q1863.75 140.169 1861.8 139.706 Q1859.86 139.243 1857.78 139.243 Q1854.61 139.243 1853.01 140.215 Q1851.43 141.187 1851.43 143.131 Q1851.43 144.613 1852.57 145.469 Q1853.7 146.303 1857.13 147.067 L1858.59 147.391 Q1863.12 148.363 1865.02 150.145 Q1866.94 151.905 1866.94 155.076 Q1866.94 158.687 1864.07 160.793 Q1861.23 162.9 1856.23 162.9 Q1854.14 162.9 1851.87 162.483 Q1849.63 162.09 1847.13 161.28 L1847.13 156.881 Q1849.49 158.108 1851.78 158.733 Q1854.07 159.335 1856.32 159.335 Q1859.33 159.335 1860.95 158.317 Q1862.57 157.275 1862.57 155.4 Q1862.57 153.664 1861.39 152.738 Q1860.23 151.812 1856.27 150.956 L1854.79 150.608 Q1850.83 149.775 1849.07 148.062 Q1847.31 146.326 1847.31 143.317 Q1847.31 139.659 1849.91 137.669 Q1852.5 135.678 1857.27 135.678 Q1859.63 135.678 1861.71 136.025 Q1863.8 136.372 1865.55 137.067 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M1874.33 156.349 L1879.21 156.349 L1879.21 162.229 L1874.33 162.229 L1874.33 156.349 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M1905.32 137.067 L1905.32 141.094 Q1903.52 140.169 1901.57 139.706 Q1899.63 139.243 1897.55 139.243 Q1894.37 139.243 1892.78 140.215 Q1891.2 141.187 1891.2 143.131 Q1891.2 144.613 1892.34 145.469 Q1893.47 146.303 1896.9 147.067 L1898.36 147.391 Q1902.89 148.363 1904.79 150.145 Q1906.71 151.905 1906.71 155.076 Q1906.71 158.687 1903.84 160.793 Q1900.99 162.9 1895.99 162.9 Q1893.91 162.9 1891.64 162.483 Q1889.4 162.09 1886.9 161.28 L1886.9 156.881 Q1889.26 158.108 1891.55 158.733 Q1893.84 159.335 1896.09 159.335 Q1899.1 159.335 1900.72 158.317 Q1902.34 157.275 1902.34 155.4 Q1902.34 153.664 1901.16 152.738 Q1900 151.812 1896.04 150.956 L1894.56 150.608 Q1890.6 149.775 1888.84 148.062 Q1887.08 146.326 1887.08 143.317 Q1887.08 139.659 1889.68 137.669 Q1892.27 135.678 1897.04 135.678 Q1899.4 135.678 1901.48 136.025 Q1903.56 136.372 1905.32 137.067 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M1938.8 126.257 Q1935.69 131.581 1934.19 136.789 Q1932.68 141.997 1932.68 147.344 Q1932.68 152.692 1934.19 157.946 Q1935.72 163.178 1938.8 168.479 L1935.09 168.479 Q1931.62 163.039 1929.88 157.784 Q1928.17 152.53 1928.17 147.344 Q1928.17 142.182 1929.88 136.951 Q1931.6 131.72 1935.09 126.257 L1938.8 126.257 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M1961.83 143.594 Q1965.18 144.312 1967.06 146.581 Q1968.96 148.849 1968.96 152.182 Q1968.96 157.298 1965.44 160.099 Q1961.92 162.9 1955.44 162.9 Q1953.26 162.9 1950.95 162.46 Q1948.66 162.043 1946.2 161.187 L1946.2 156.673 Q1948.15 157.807 1950.46 158.386 Q1952.78 158.965 1955.3 158.965 Q1959.7 158.965 1961.99 157.229 Q1964.3 155.493 1964.3 152.182 Q1964.3 149.127 1962.15 147.414 Q1960.02 145.678 1956.2 145.678 L1952.17 145.678 L1952.17 141.835 L1956.39 141.835 Q1959.84 141.835 1961.67 140.469 Q1963.49 139.081 1963.49 136.488 Q1963.49 133.826 1961.6 132.414 Q1959.72 130.979 1956.2 130.979 Q1954.28 130.979 1952.08 131.395 Q1949.88 131.812 1947.24 132.692 L1947.24 128.525 Q1949.91 127.784 1952.22 127.414 Q1954.56 127.044 1956.62 127.044 Q1961.94 127.044 1965.05 129.474 Q1968.15 131.882 1968.15 136.002 Q1968.15 138.872 1966.5 140.863 Q1964.86 142.831 1961.83 143.594 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M1991.99 143.594 Q1995.35 144.312 1997.22 146.581 Q1999.12 148.849 1999.12 152.182 Q1999.12 157.298 1995.6 160.099 Q1992.08 162.9 1985.6 162.9 Q1983.42 162.9 1981.11 162.46 Q1978.82 162.043 1976.36 161.187 L1976.36 156.673 Q1978.31 157.807 1980.62 158.386 Q1982.94 158.965 1985.46 158.965 Q1989.86 158.965 1992.15 157.229 Q1994.47 155.493 1994.47 152.182 Q1994.47 149.127 1992.31 147.414 Q1990.18 145.678 1986.36 145.678 L1982.34 145.678 L1982.34 141.835 L1986.55 141.835 Q1990 141.835 1991.83 140.469 Q1993.66 139.081 1993.66 136.488 Q1993.66 133.826 1991.76 132.414 Q1989.88 130.979 1986.36 130.979 Q1984.44 130.979 1982.24 131.395 Q1980.05 131.812 1977.41 132.692 L1977.41 128.525 Q1980.07 127.784 1982.38 127.414 Q1984.72 127.044 1986.78 127.044 Q1992.11 127.044 1995.21 129.474 Q1998.31 131.882 1998.31 136.002 Q1998.31 138.872 1996.67 140.863 Q1995.02 142.831 1991.99 143.594 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M2008.03 127.669 L2026.39 127.669 L2026.39 131.604 L2012.31 131.604 L2012.31 140.076 Q2013.33 139.729 2014.35 139.567 Q2015.37 139.382 2016.39 139.382 Q2022.17 139.382 2025.55 142.553 Q2028.93 145.724 2028.93 151.141 Q2028.93 156.719 2025.46 159.821 Q2021.99 162.9 2015.67 162.9 Q2013.49 162.9 2011.23 162.53 Q2008.98 162.159 2006.57 161.418 L2006.57 156.719 Q2008.66 157.854 2010.88 158.409 Q2013.1 158.965 2015.58 158.965 Q2019.58 158.965 2021.92 156.858 Q2024.26 154.752 2024.26 151.141 Q2024.26 147.53 2021.92 145.423 Q2019.58 143.317 2015.58 143.317 Q2013.7 143.317 2011.83 143.733 Q2009.98 144.15 2008.03 145.03 L2008.03 127.669 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M2038.19 127.669 L2056.55 127.669 L2056.55 131.604 L2042.48 131.604 L2042.48 140.076 Q2043.49 139.729 2044.51 139.567 Q2045.53 139.382 2046.55 139.382 Q2052.34 139.382 2055.72 142.553 Q2059.1 145.724 2059.1 151.141 Q2059.1 156.719 2055.62 159.821 Q2052.15 162.9 2045.83 162.9 Q2043.66 162.9 2041.39 162.53 Q2039.14 162.159 2036.73 161.418 L2036.73 156.719 Q2038.82 157.854 2041.04 158.409 Q2043.26 158.965 2045.74 158.965 Q2049.74 158.965 2052.08 156.858 Q2054.42 154.752 2054.42 151.141 Q2054.42 147.53 2052.08 145.423 Q2049.74 143.317 2045.74 143.317 Q2043.86 143.317 2041.99 143.733 Q2040.14 144.15 2038.19 145.03 L2038.19 127.669 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M2081.16 131.743 L2069.35 150.192 L2081.16 150.192 L2081.16 131.743 M2079.93 127.669 L2085.81 127.669 L2085.81 150.192 L2090.74 150.192 L2090.74 154.081 L2085.81 154.081 L2085.81 162.229 L2081.16 162.229 L2081.16 154.081 L2065.55 154.081 L2065.55 149.567 L2079.93 127.669 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M2111.32 131.743 L2099.51 150.192 L2111.32 150.192 L2111.32 131.743 M2110.09 127.669 L2115.97 127.669 L2115.97 150.192 L2120.9 150.192 L2120.9 154.081 L2115.97 154.081 L2115.97 162.229 L2111.32 162.229 L2111.32 154.081 L2095.72 154.081 L2095.72 149.567 L2110.09 127.669 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M2142.8 143.594 Q2146.16 144.312 2148.03 146.581 Q2149.93 148.849 2149.93 152.182 Q2149.93 157.298 2146.41 160.099 Q2142.89 162.9 2136.41 162.9 Q2134.23 162.9 2131.92 162.46 Q2129.63 162.043 2127.17 161.187 L2127.17 156.673 Q2129.12 157.807 2131.43 158.386 Q2133.75 158.965 2136.27 158.965 Q2140.67 158.965 2142.96 157.229 Q2145.28 155.493 2145.28 152.182 Q2145.28 149.127 2143.12 147.414 Q2140.99 145.678 2137.17 145.678 L2133.15 145.678 L2133.15 141.835 L2137.36 141.835 Q2140.81 141.835 2142.64 140.469 Q2144.47 139.081 2144.47 136.488 Q2144.47 133.826 2142.57 132.414 Q2140.69 130.979 2137.17 130.979 Q2135.25 130.979 2133.05 131.395 Q2130.85 131.812 2128.22 132.692 L2128.22 128.525 Q2130.88 127.784 2133.19 127.414 Q2135.53 127.044 2137.59 127.044 Q2142.91 127.044 2146.02 129.474 Q2149.12 131.882 2149.12 136.002 Q2149.12 138.872 2147.47 140.863 Q2145.83 142.831 2142.8 143.594 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M2162.82 158.293 L2179.14 158.293 L2179.14 162.229 L2157.2 162.229 L2157.2 158.293 Q2159.86 155.539 2164.44 150.909 Q2169.05 146.256 2170.23 144.914 Q2172.47 142.391 2173.35 140.655 Q2174.26 138.895 2174.26 137.206 Q2174.26 134.451 2172.31 132.715 Q2170.39 130.979 2167.29 130.979 Q2165.09 130.979 2162.64 131.743 Q2160.21 132.507 2157.43 134.057 L2157.43 129.335 Q2160.25 128.201 2162.71 127.622 Q2165.16 127.044 2167.2 127.044 Q2172.57 127.044 2175.76 129.729 Q2178.96 132.414 2178.96 136.905 Q2178.96 139.034 2178.15 140.956 Q2177.36 142.854 2175.25 145.446 Q2174.67 146.118 2171.57 149.335 Q2168.47 152.53 2162.82 158.293 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M2187.68 126.257 L2191.39 126.257 Q2194.86 131.72 2196.57 136.951 Q2198.31 142.182 2198.31 147.344 Q2198.31 152.53 2196.57 157.784 Q2194.86 163.039 2191.39 168.479 L2187.68 168.479 Q2190.76 163.178 2192.27 157.946 Q2193.79 152.692 2193.79 147.344 Q2193.79 141.997 2192.27 136.789 Q2190.76 131.581 2187.68 126.257 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip060)" style="stroke:#e26f46; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1581.03,196.789 1724.95,196.789 
  "/>
<path clip-path="url(#clip060)" d="M1769.28 193.12 Q1770.88 190.249 1773.1 188.884 Q1775.32 187.518 1778.33 187.518 Q1782.38 187.518 1784.58 190.365 Q1786.78 193.189 1786.78 198.421 L1786.78 214.069 L1782.5 214.069 L1782.5 198.559 Q1782.5 194.833 1781.18 193.027 Q1779.86 191.222 1777.15 191.222 Q1773.84 191.222 1771.92 193.421 Q1770 195.62 1770 199.416 L1770 214.069 L1765.72 214.069 L1765.72 198.559 Q1765.72 194.809 1764.4 193.027 Q1763.08 191.222 1760.32 191.222 Q1757.06 191.222 1755.14 193.444 Q1753.22 195.643 1753.22 199.416 L1753.22 214.069 L1748.94 214.069 L1748.94 188.143 L1753.22 188.143 L1753.22 192.171 Q1754.68 189.786 1756.71 188.652 Q1758.75 187.518 1761.55 187.518 Q1764.37 187.518 1766.34 188.953 Q1768.33 190.388 1769.28 193.12 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M1807.06 201.036 Q1801.9 201.036 1799.91 202.217 Q1797.92 203.397 1797.92 206.245 Q1797.92 208.513 1799.4 209.856 Q1800.9 211.175 1803.47 211.175 Q1807.01 211.175 1809.14 208.675 Q1811.3 206.152 1811.3 201.985 L1811.3 201.036 L1807.06 201.036 M1815.56 199.277 L1815.56 214.069 L1811.3 214.069 L1811.3 210.133 Q1809.84 212.495 1807.66 213.629 Q1805.49 214.74 1802.34 214.74 Q1798.36 214.74 1796 212.518 Q1793.66 210.272 1793.66 206.522 Q1793.66 202.147 1796.57 199.925 Q1799.51 197.703 1805.32 197.703 L1811.3 197.703 L1811.3 197.286 Q1811.3 194.347 1809.35 192.749 Q1807.43 191.129 1803.93 191.129 Q1801.71 191.129 1799.61 191.661 Q1797.5 192.194 1795.56 193.259 L1795.56 189.323 Q1797.89 188.421 1800.09 187.981 Q1802.29 187.518 1804.37 187.518 Q1810 187.518 1812.78 190.434 Q1815.56 193.351 1815.56 199.277 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M1840.86 188.907 L1840.86 192.934 Q1839.05 192.009 1837.11 191.546 Q1835.16 191.083 1833.08 191.083 Q1829.91 191.083 1828.31 192.055 Q1826.74 193.027 1826.74 194.971 Q1826.74 196.453 1827.87 197.309 Q1829 198.143 1832.43 198.907 L1833.89 199.231 Q1838.43 200.203 1840.32 201.985 Q1842.24 203.745 1842.24 206.916 Q1842.24 210.527 1839.37 212.633 Q1836.53 214.74 1831.53 214.74 Q1829.44 214.74 1827.18 214.323 Q1824.93 213.93 1822.43 213.12 L1822.43 208.721 Q1824.79 209.948 1827.08 210.573 Q1829.37 211.175 1831.62 211.175 Q1834.63 211.175 1836.25 210.157 Q1837.87 209.115 1837.87 207.24 Q1837.87 205.504 1836.69 204.578 Q1835.53 203.652 1831.57 202.796 L1830.09 202.448 Q1826.13 201.615 1824.37 199.902 Q1822.62 198.166 1822.62 195.157 Q1822.62 191.499 1825.21 189.509 Q1827.8 187.518 1832.57 187.518 Q1834.93 187.518 1837.01 187.865 Q1839.1 188.212 1840.86 188.907 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M1865.55 188.907 L1865.55 192.934 Q1863.75 192.009 1861.8 191.546 Q1859.86 191.083 1857.78 191.083 Q1854.61 191.083 1853.01 192.055 Q1851.43 193.027 1851.43 194.971 Q1851.43 196.453 1852.57 197.309 Q1853.7 198.143 1857.13 198.907 L1858.59 199.231 Q1863.12 200.203 1865.02 201.985 Q1866.94 203.745 1866.94 206.916 Q1866.94 210.527 1864.07 212.633 Q1861.23 214.74 1856.23 214.74 Q1854.14 214.74 1851.87 214.323 Q1849.63 213.93 1847.13 213.12 L1847.13 208.721 Q1849.49 209.948 1851.78 210.573 Q1854.07 211.175 1856.32 211.175 Q1859.33 211.175 1860.95 210.157 Q1862.57 209.115 1862.57 207.24 Q1862.57 205.504 1861.39 204.578 Q1860.23 203.652 1856.27 202.796 L1854.79 202.448 Q1850.83 201.615 1849.07 199.902 Q1847.31 198.166 1847.31 195.157 Q1847.31 191.499 1849.91 189.509 Q1852.5 187.518 1857.27 187.518 Q1859.63 187.518 1861.71 187.865 Q1863.8 188.212 1865.55 188.907 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M1874.33 208.189 L1879.21 208.189 L1879.21 214.069 L1874.33 214.069 L1874.33 208.189 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M1885.74 188.143 L1890.25 188.143 L1898.36 209.902 L1906.46 188.143 L1910.97 188.143 L1901.25 214.069 L1895.46 214.069 L1885.74 188.143 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M1942.15 178.097 Q1939.05 183.421 1937.55 188.629 Q1936.04 193.837 1936.04 199.184 Q1936.04 204.532 1937.55 209.786 Q1939.07 215.018 1942.15 220.319 L1938.45 220.319 Q1934.98 214.879 1933.24 209.624 Q1931.53 204.37 1931.53 199.184 Q1931.53 194.022 1933.24 188.791 Q1934.95 183.56 1938.45 178.097 L1942.15 178.097 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M1965.18 195.434 Q1968.54 196.152 1970.42 198.421 Q1972.31 200.689 1972.31 204.022 Q1972.31 209.138 1968.8 211.939 Q1965.28 214.74 1958.8 214.74 Q1956.62 214.74 1954.3 214.3 Q1952.01 213.883 1949.56 213.027 L1949.56 208.513 Q1951.5 209.647 1953.82 210.226 Q1956.13 210.805 1958.66 210.805 Q1963.05 210.805 1965.35 209.069 Q1967.66 207.333 1967.66 204.022 Q1967.66 200.967 1965.51 199.254 Q1963.38 197.518 1959.56 197.518 L1955.53 197.518 L1955.53 193.675 L1959.74 193.675 Q1963.19 193.675 1965.02 192.309 Q1966.85 190.921 1966.85 188.328 Q1966.85 185.666 1964.95 184.254 Q1963.08 182.819 1959.56 182.819 Q1957.64 182.819 1955.44 183.235 Q1953.24 183.652 1950.6 184.532 L1950.6 180.365 Q1953.26 179.624 1955.58 179.254 Q1957.92 178.884 1959.98 178.884 Q1965.3 178.884 1968.4 181.314 Q1971.5 183.722 1971.5 187.842 Q1971.5 190.712 1969.86 192.703 Q1968.22 194.671 1965.18 195.434 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M1995.35 195.434 Q1998.7 196.152 2000.58 198.421 Q2002.48 200.689 2002.48 204.022 Q2002.48 209.138 1998.96 211.939 Q1995.44 214.74 1988.96 214.74 Q1986.78 214.74 1984.47 214.3 Q1982.17 213.883 1979.72 213.027 L1979.72 208.513 Q1981.67 209.647 1983.98 210.226 Q1986.3 210.805 1988.82 210.805 Q1993.22 210.805 1995.51 209.069 Q1997.82 207.333 1997.82 204.022 Q1997.82 200.967 1995.67 199.254 Q1993.54 197.518 1989.72 197.518 L1985.69 197.518 L1985.69 193.675 L1989.91 193.675 Q1993.36 193.675 1995.18 192.309 Q1997.01 190.921 1997.01 188.328 Q1997.01 185.666 1995.11 184.254 Q1993.24 182.819 1989.72 182.819 Q1987.8 182.819 1985.6 183.235 Q1983.4 183.652 1980.76 184.532 L1980.76 180.365 Q1983.42 179.624 1985.74 179.254 Q1988.08 178.884 1990.14 178.884 Q1995.46 178.884 1998.56 181.314 Q2001.67 183.722 2001.67 187.842 Q2001.67 190.712 2000.02 192.703 Q1998.38 194.671 1995.35 195.434 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M2011.39 179.509 L2029.74 179.509 L2029.74 183.444 L2015.67 183.444 L2015.67 191.916 Q2016.69 191.569 2017.71 191.407 Q2018.73 191.222 2019.74 191.222 Q2025.53 191.222 2028.91 194.393 Q2032.29 197.564 2032.29 202.981 Q2032.29 208.559 2028.82 211.661 Q2025.35 214.74 2019.03 214.74 Q2016.85 214.74 2014.58 214.37 Q2012.34 213.999 2009.93 213.258 L2009.93 208.559 Q2012.01 209.694 2014.23 210.249 Q2016.46 210.805 2018.93 210.805 Q2022.94 210.805 2025.28 208.698 Q2027.61 206.592 2027.61 202.981 Q2027.61 199.37 2025.28 197.263 Q2022.94 195.157 2018.93 195.157 Q2017.06 195.157 2015.18 195.573 Q2013.33 195.99 2011.39 196.87 L2011.39 179.509 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M2041.55 179.509 L2059.91 179.509 L2059.91 183.444 L2045.83 183.444 L2045.83 191.916 Q2046.85 191.569 2047.87 191.407 Q2048.89 191.222 2049.91 191.222 Q2055.69 191.222 2059.07 194.393 Q2062.45 197.564 2062.45 202.981 Q2062.45 208.559 2058.98 211.661 Q2055.51 214.74 2049.19 214.74 Q2047.01 214.74 2044.74 214.37 Q2042.5 213.999 2040.09 213.258 L2040.09 208.559 Q2042.17 209.694 2044.4 210.249 Q2046.62 210.805 2049.1 210.805 Q2053.1 210.805 2055.44 208.698 Q2057.78 206.592 2057.78 202.981 Q2057.78 199.37 2055.44 197.263 Q2053.1 195.157 2049.1 195.157 Q2047.22 195.157 2045.35 195.573 Q2043.49 195.99 2041.55 196.87 L2041.55 179.509 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M2084.51 183.583 L2072.71 202.032 L2084.51 202.032 L2084.51 183.583 M2083.29 179.509 L2089.16 179.509 L2089.16 202.032 L2094.1 202.032 L2094.1 205.921 L2089.16 205.921 L2089.16 214.069 L2084.51 214.069 L2084.51 205.921 L2068.91 205.921 L2068.91 201.407 L2083.29 179.509 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M2114.67 183.583 L2102.87 202.032 L2114.67 202.032 L2114.67 183.583 M2113.45 179.509 L2119.33 179.509 L2119.33 202.032 L2124.26 202.032 L2124.26 205.921 L2119.33 205.921 L2119.33 214.069 L2114.67 214.069 L2114.67 205.921 L2099.07 205.921 L2099.07 201.407 L2113.45 179.509 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M2146.16 195.434 Q2149.51 196.152 2151.39 198.421 Q2153.28 200.689 2153.28 204.022 Q2153.28 209.138 2149.77 211.939 Q2146.25 214.74 2139.77 214.74 Q2137.59 214.74 2135.28 214.3 Q2132.98 213.883 2130.53 213.027 L2130.53 208.513 Q2132.47 209.647 2134.79 210.226 Q2137.1 210.805 2139.63 210.805 Q2144.03 210.805 2146.32 209.069 Q2148.63 207.333 2148.63 204.022 Q2148.63 200.967 2146.48 199.254 Q2144.35 197.518 2140.53 197.518 L2136.5 197.518 L2136.5 193.675 L2140.72 193.675 Q2144.16 193.675 2145.99 192.309 Q2147.82 190.921 2147.82 188.328 Q2147.82 185.666 2145.92 184.254 Q2144.05 182.819 2140.53 182.819 Q2138.61 182.819 2136.41 183.235 Q2134.21 183.652 2131.57 184.532 L2131.57 180.365 Q2134.23 179.624 2136.55 179.254 Q2138.89 178.884 2140.95 178.884 Q2146.27 178.884 2149.37 181.314 Q2152.47 183.722 2152.47 187.842 Q2152.47 190.712 2150.83 192.703 Q2149.19 194.671 2146.16 195.434 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M2176.32 195.434 Q2179.67 196.152 2181.55 198.421 Q2183.45 200.689 2183.45 204.022 Q2183.45 209.138 2179.93 211.939 Q2176.41 214.74 2169.93 214.74 Q2167.75 214.74 2165.44 214.3 Q2163.15 213.883 2160.69 213.027 L2160.69 208.513 Q2162.64 209.647 2164.95 210.226 Q2167.27 210.805 2169.79 210.805 Q2174.19 210.805 2176.48 209.069 Q2178.79 207.333 2178.79 204.022 Q2178.79 200.967 2176.64 199.254 Q2174.51 197.518 2170.69 197.518 L2166.66 197.518 L2166.66 193.675 L2170.88 193.675 Q2174.33 193.675 2176.16 192.309 Q2177.98 190.921 2177.98 188.328 Q2177.98 185.666 2176.09 184.254 Q2174.21 182.819 2170.69 182.819 Q2168.77 182.819 2166.57 183.235 Q2164.37 183.652 2161.73 184.532 L2161.73 180.365 Q2164.4 179.624 2166.71 179.254 Q2169.05 178.884 2171.11 178.884 Q2176.43 178.884 2179.53 181.314 Q2182.64 183.722 2182.64 187.842 Q2182.64 190.712 2180.99 192.703 Q2179.35 194.671 2176.32 195.434 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M2191.04 178.097 L2194.74 178.097 Q2198.22 183.56 2199.93 188.791 Q2201.66 194.022 2201.66 199.184 Q2201.66 204.37 2199.93 209.624 Q2198.22 214.879 2194.74 220.319 L2191.04 220.319 Q2194.12 215.018 2195.62 209.786 Q2197.15 204.532 2197.15 199.184 Q2197.15 193.837 2195.62 188.629 Q2194.12 183.421 2191.04 178.097 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip060)" style="stroke:#3da44d; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1581.03,248.629 1724.95,248.629 
  "/>
<path clip-path="url(#clip060)" d="M1769.28 244.96 Q1770.88 242.089 1773.1 240.724 Q1775.32 239.358 1778.33 239.358 Q1782.38 239.358 1784.58 242.205 Q1786.78 245.029 1786.78 250.261 L1786.78 265.909 L1782.5 265.909 L1782.5 250.399 Q1782.5 246.673 1781.18 244.867 Q1779.86 243.062 1777.15 243.062 Q1773.84 243.062 1771.92 245.261 Q1770 247.46 1770 251.256 L1770 265.909 L1765.72 265.909 L1765.72 250.399 Q1765.72 246.649 1764.4 244.867 Q1763.08 243.062 1760.32 243.062 Q1757.06 243.062 1755.14 245.284 Q1753.22 247.483 1753.22 251.256 L1753.22 265.909 L1748.94 265.909 L1748.94 239.983 L1753.22 239.983 L1753.22 244.011 Q1754.68 241.626 1756.71 240.492 Q1758.75 239.358 1761.55 239.358 Q1764.37 239.358 1766.34 240.793 Q1768.33 242.228 1769.28 244.96 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M1807.06 252.876 Q1801.9 252.876 1799.91 254.057 Q1797.92 255.237 1797.92 258.085 Q1797.92 260.353 1799.4 261.696 Q1800.9 263.015 1803.47 263.015 Q1807.01 263.015 1809.14 260.515 Q1811.3 257.992 1811.3 253.825 L1811.3 252.876 L1807.06 252.876 M1815.56 251.117 L1815.56 265.909 L1811.3 265.909 L1811.3 261.973 Q1809.84 264.335 1807.66 265.469 Q1805.49 266.58 1802.34 266.58 Q1798.36 266.58 1796 264.358 Q1793.66 262.112 1793.66 258.362 Q1793.66 253.987 1796.57 251.765 Q1799.51 249.543 1805.32 249.543 L1811.3 249.543 L1811.3 249.126 Q1811.3 246.187 1809.35 244.589 Q1807.43 242.969 1803.93 242.969 Q1801.71 242.969 1799.61 243.501 Q1797.5 244.034 1795.56 245.099 L1795.56 241.163 Q1797.89 240.261 1800.09 239.821 Q1802.29 239.358 1804.37 239.358 Q1810 239.358 1812.78 242.274 Q1815.56 245.191 1815.56 251.117 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M1840.86 240.747 L1840.86 244.774 Q1839.05 243.849 1837.11 243.386 Q1835.16 242.923 1833.08 242.923 Q1829.91 242.923 1828.31 243.895 Q1826.74 244.867 1826.74 246.811 Q1826.74 248.293 1827.87 249.149 Q1829 249.983 1832.43 250.747 L1833.89 251.071 Q1838.43 252.043 1840.32 253.825 Q1842.24 255.585 1842.24 258.756 Q1842.24 262.367 1839.37 264.473 Q1836.53 266.58 1831.53 266.58 Q1829.44 266.58 1827.18 266.163 Q1824.93 265.77 1822.43 264.96 L1822.43 260.561 Q1824.79 261.788 1827.08 262.413 Q1829.37 263.015 1831.62 263.015 Q1834.63 263.015 1836.25 261.997 Q1837.87 260.955 1837.87 259.08 Q1837.87 257.344 1836.69 256.418 Q1835.53 255.492 1831.57 254.636 L1830.09 254.288 Q1826.13 253.455 1824.37 251.742 Q1822.62 250.006 1822.62 246.997 Q1822.62 243.339 1825.21 241.349 Q1827.8 239.358 1832.57 239.358 Q1834.93 239.358 1837.01 239.705 Q1839.1 240.052 1840.86 240.747 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M1865.55 240.747 L1865.55 244.774 Q1863.75 243.849 1861.8 243.386 Q1859.86 242.923 1857.78 242.923 Q1854.61 242.923 1853.01 243.895 Q1851.43 244.867 1851.43 246.811 Q1851.43 248.293 1852.57 249.149 Q1853.7 249.983 1857.13 250.747 L1858.59 251.071 Q1863.12 252.043 1865.02 253.825 Q1866.94 255.585 1866.94 258.756 Q1866.94 262.367 1864.07 264.473 Q1861.23 266.58 1856.23 266.58 Q1854.14 266.58 1851.87 266.163 Q1849.63 265.77 1847.13 264.96 L1847.13 260.561 Q1849.49 261.788 1851.78 262.413 Q1854.07 263.015 1856.32 263.015 Q1859.33 263.015 1860.95 261.997 Q1862.57 260.955 1862.57 259.08 Q1862.57 257.344 1861.39 256.418 Q1860.23 255.492 1856.27 254.636 L1854.79 254.288 Q1850.83 253.455 1849.07 251.742 Q1847.31 250.006 1847.31 246.997 Q1847.31 243.339 1849.91 241.349 Q1852.5 239.358 1857.27 239.358 Q1859.63 239.358 1861.71 239.705 Q1863.8 240.052 1865.55 240.747 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M1874.33 260.029 L1879.21 260.029 L1879.21 265.909 L1874.33 265.909 L1874.33 260.029 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M1900.58 252.876 Q1895.42 252.876 1893.43 254.057 Q1891.43 255.237 1891.43 258.085 Q1891.43 260.353 1892.92 261.696 Q1894.42 263.015 1896.99 263.015 Q1900.53 263.015 1902.66 260.515 Q1904.81 257.992 1904.81 253.825 L1904.81 252.876 L1900.58 252.876 M1909.07 251.117 L1909.07 265.909 L1904.81 265.909 L1904.81 261.973 Q1903.36 264.335 1901.18 265.469 Q1899 266.58 1895.86 266.58 Q1891.87 266.58 1889.51 264.358 Q1887.18 262.112 1887.18 258.362 Q1887.18 253.987 1890.09 251.765 Q1893.03 249.543 1898.84 249.543 L1904.81 249.543 L1904.81 249.126 Q1904.81 246.187 1902.87 244.589 Q1900.95 242.969 1897.45 242.969 Q1895.23 242.969 1893.12 243.501 Q1891.02 244.034 1889.07 245.099 L1889.07 241.163 Q1891.41 240.261 1893.61 239.821 Q1895.81 239.358 1897.89 239.358 Q1903.52 239.358 1906.3 242.274 Q1909.07 245.191 1909.07 251.117 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M1943.15 229.937 Q1940.05 235.261 1938.54 240.469 Q1937.04 245.677 1937.04 251.024 Q1937.04 256.372 1938.54 261.626 Q1940.07 266.858 1943.15 272.159 L1939.44 272.159 Q1935.97 266.719 1934.24 261.464 Q1932.52 256.21 1932.52 251.024 Q1932.52 245.862 1934.24 240.631 Q1935.95 235.4 1939.44 229.937 L1943.15 229.937 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M1952.06 231.349 L1970.42 231.349 L1970.42 235.284 L1956.34 235.284 L1956.34 243.756 Q1957.36 243.409 1958.38 243.247 Q1959.4 243.062 1960.42 243.062 Q1966.2 243.062 1969.58 246.233 Q1972.96 249.404 1972.96 254.821 Q1972.96 260.399 1969.49 263.501 Q1966.02 266.58 1959.7 266.58 Q1957.52 266.58 1955.25 266.21 Q1953.01 265.839 1950.6 265.098 L1950.6 260.399 Q1952.68 261.534 1954.91 262.089 Q1957.13 262.645 1959.61 262.645 Q1963.61 262.645 1965.95 260.538 Q1968.29 258.432 1968.29 254.821 Q1968.29 251.21 1965.95 249.103 Q1963.61 246.997 1959.61 246.997 Q1957.73 246.997 1955.86 247.413 Q1954 247.83 1952.06 248.71 L1952.06 231.349 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M1992.17 249.497 Q1988.84 249.497 1986.92 251.279 Q1985.02 253.061 1985.02 256.186 Q1985.02 259.311 1986.92 261.094 Q1988.84 262.876 1992.17 262.876 Q1995.51 262.876 1997.43 261.094 Q1999.35 259.288 1999.35 256.186 Q1999.35 253.061 1997.43 251.279 Q1995.53 249.497 1992.17 249.497 M1987.5 247.506 Q1984.49 246.765 1982.8 244.705 Q1981.13 242.645 1981.13 239.682 Q1981.13 235.538 1984.07 233.131 Q1987.04 230.724 1992.17 230.724 Q1997.34 230.724 2000.28 233.131 Q2003.22 235.538 2003.22 239.682 Q2003.22 242.645 2001.53 244.705 Q1999.86 246.765 1996.87 247.506 Q2000.25 248.293 2002.13 250.585 Q2004.03 252.876 2004.03 256.186 Q2004.03 261.21 2000.95 263.895 Q1997.89 266.58 1992.17 266.58 Q1986.46 266.58 1983.38 263.895 Q1980.32 261.21 1980.32 256.186 Q1980.32 252.876 1982.22 250.585 Q1984.12 248.293 1987.5 247.506 M1985.79 240.122 Q1985.79 242.807 1987.45 244.312 Q1989.14 245.816 1992.17 245.816 Q1995.18 245.816 1996.87 244.312 Q1998.59 242.807 1998.59 240.122 Q1998.59 237.437 1996.87 235.932 Q1995.18 234.427 1992.17 234.427 Q1989.14 234.427 1987.45 235.932 Q1985.79 237.437 1985.79 240.122 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M2011.16 231.349 L2033.38 231.349 L2033.38 233.339 L2020.83 265.909 L2015.95 265.909 L2027.75 235.284 L2011.16 235.284 L2011.16 231.349 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M2046.53 261.973 L2062.85 261.973 L2062.85 265.909 L2040.9 265.909 L2040.9 261.973 Q2043.56 259.219 2048.15 254.589 Q2052.75 249.936 2053.93 248.594 Q2056.18 246.071 2057.06 244.335 Q2057.96 242.575 2057.96 240.886 Q2057.96 238.131 2056.02 236.395 Q2054.1 234.659 2050.99 234.659 Q2048.79 234.659 2046.34 235.423 Q2043.91 236.187 2041.13 237.737 L2041.13 233.015 Q2043.96 231.881 2046.41 231.302 Q2048.86 230.724 2050.9 230.724 Q2056.27 230.724 2059.47 233.409 Q2062.66 236.094 2062.66 240.585 Q2062.66 242.714 2061.85 244.636 Q2061.06 246.534 2058.96 249.126 Q2058.38 249.798 2055.28 253.015 Q2052.17 256.21 2046.53 261.973 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M2082.66 234.427 Q2079.05 234.427 2077.22 237.992 Q2075.41 241.534 2075.41 248.663 Q2075.41 255.77 2077.22 259.335 Q2079.05 262.876 2082.66 262.876 Q2086.29 262.876 2088.1 259.335 Q2089.93 255.77 2089.93 248.663 Q2089.93 241.534 2088.1 237.992 Q2086.29 234.427 2082.66 234.427 M2082.66 230.724 Q2088.47 230.724 2091.53 235.33 Q2094.6 239.913 2094.6 248.663 Q2094.6 257.39 2091.53 261.997 Q2088.47 266.58 2082.66 266.58 Q2076.85 266.58 2073.77 261.997 Q2070.72 257.39 2070.72 248.663 Q2070.72 239.913 2073.77 235.33 Q2076.85 230.724 2082.66 230.724 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M2106.85 261.973 L2123.17 261.973 L2123.17 265.909 L2101.22 265.909 L2101.22 261.973 Q2103.89 259.219 2108.47 254.589 Q2113.08 249.936 2114.26 248.594 Q2116.5 246.071 2117.38 244.335 Q2118.29 242.575 2118.29 240.886 Q2118.29 238.131 2116.34 236.395 Q2114.42 234.659 2111.32 234.659 Q2109.12 234.659 2106.66 235.423 Q2104.23 236.187 2101.46 237.737 L2101.46 233.015 Q2104.28 231.881 2106.73 231.302 Q2109.19 230.724 2111.22 230.724 Q2116.6 230.724 2119.79 233.409 Q2122.98 236.094 2122.98 240.585 Q2122.98 242.714 2122.17 244.636 Q2121.39 246.534 2119.28 249.126 Q2118.7 249.798 2115.6 253.015 Q2112.5 256.21 2106.85 261.973 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M2133.03 231.349 L2151.39 231.349 L2151.39 235.284 L2137.31 235.284 L2137.31 243.756 Q2138.33 243.409 2139.35 243.247 Q2140.37 243.062 2141.39 243.062 Q2147.17 243.062 2150.55 246.233 Q2153.93 249.404 2153.93 254.821 Q2153.93 260.399 2150.46 263.501 Q2146.99 266.58 2140.67 266.58 Q2138.49 266.58 2136.22 266.21 Q2133.98 265.839 2131.57 265.098 L2131.57 260.399 Q2133.66 261.534 2135.88 262.089 Q2138.1 262.645 2140.58 262.645 Q2144.58 262.645 2146.92 260.538 Q2149.26 258.432 2149.26 254.821 Q2149.26 251.21 2146.92 249.103 Q2144.58 246.997 2140.58 246.997 Q2138.7 246.997 2136.83 247.413 Q2134.97 247.83 2133.03 248.71 L2133.03 231.349 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M2173.72 246.765 Q2170.58 246.765 2168.72 248.918 Q2166.9 251.071 2166.9 254.821 Q2166.9 258.548 2168.72 260.723 Q2170.58 262.876 2173.72 262.876 Q2176.87 262.876 2178.7 260.723 Q2180.55 258.548 2180.55 254.821 Q2180.55 251.071 2178.7 248.918 Q2176.87 246.765 2173.72 246.765 M2183.01 232.113 L2183.01 236.372 Q2181.25 235.538 2179.44 235.099 Q2177.66 234.659 2175.9 234.659 Q2171.27 234.659 2168.82 237.784 Q2166.39 240.909 2166.04 247.228 Q2167.41 245.214 2169.47 244.149 Q2171.53 243.062 2174 243.062 Q2179.21 243.062 2182.22 246.233 Q2185.25 249.381 2185.25 254.821 Q2185.25 260.145 2182.1 263.362 Q2178.96 266.58 2173.72 266.58 Q2167.73 266.58 2164.56 261.997 Q2161.39 257.39 2161.39 248.663 Q2161.39 240.469 2165.28 235.608 Q2169.16 230.724 2175.72 230.724 Q2177.47 230.724 2179.26 231.071 Q2181.06 231.418 2183.01 232.113 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M2194.12 261.973 L2201.76 261.973 L2201.76 235.608 L2193.45 237.275 L2193.45 233.015 L2201.71 231.349 L2206.39 231.349 L2206.39 261.973 L2214.03 261.973 L2214.03 265.909 L2194.12 265.909 L2194.12 261.973 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip060)" d="M2222.2 229.937 L2225.9 229.937 Q2229.37 235.4 2231.09 240.631 Q2232.82 245.862 2232.82 251.024 Q2232.82 256.21 2231.09 261.464 Q2229.37 266.719 2225.9 272.159 L2222.2 272.159 Q2225.28 266.858 2226.78 261.626 Q2228.31 256.372 2228.31 251.024 Q2228.31 245.677 2226.78 240.469 Q2225.28 235.261 2222.2 229.937 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /></svg>



The data from the simualtion of the *defaultFMU*, are divided into position, velocity and acceleration data. The data for the acceleration will be needed later.


```julia
posDefault = collect(data[1] for data in defaultSimData.saveval)
velDefault = collect(data[2] for data in defaultSimData.saveval)
accDefault = collect(data[3] for data in defaultSimData.saveval)
```


    501-element Vector{Float64}:
      6.0
      5.996982513180007
      5.987986261034271
      5.9730046030442665
      5.95205107717745
      5.925151171646224
      5.892330901036602
      5.853619029884401
      5.809060133574773
      5.758686483925127
      ⋮
     -5.928264987575014
     -5.953392775960465
     -5.972564609104964
     -5.985762763815119
     -5.99297561097946
     -5.994196580640214
     -5.989425410415006
     -5.9786675103892755
     -5.961926527257058


## CS-NeuralFMU

In this section, the *defaultFMU* is inserted into a CS-NeuralFMU architecture. It has the goal to learn the initial state of the *referenceFMU*.


For the external force, a simple function is implemented that always returns a force of $0N$ at each time point. Also, all other functions and implementations would be possible here. Only for simplification reasons the function was chosen so simply.


```julia
function extForce(t)
    return [0.0]
end 
```


    extForce (generic function with 1 method)


#### Loss function

In order to train our model, a loss function must be implemented. The solver of the NeuralFMU can calculate the gradient of the loss function. The gradient descent is needed to adjust the weights in the neural network so that the sum of the error is reduced and the model becomes more accurate.

The loss function in this implmentation consists of the mean squared error (mse) from the acceleration data of the *referenceFMU* simulation (`accReference`) and the acceleration data of the network (`accNet`).
$$ mse = \frac{1}{n} \sum\limits_{i=0}^n (accReference[i] - accNet[i])^2 $$


```julia
# loss function for training
function lossSum()
    solution = csNeuralFMU(extForce, tStep)

    accNet = collect(data[1] for data in solution)
    
    Flux.Losses.mse(accReference, accNet)
end
```


    lossSum (generic function with 1 method)


#### Callback

To output the loss in certain time intervals, a callback is implemented as a function in the following. Here a counter is incremented, every twentieth pass the loss function is called and the average error is printed out.


```julia
# callback function for training
global counter = 0
function callb()
    global counter += 1

    if counter % 20 == 1
        avgLoss = lossSum()
        @info "Loss [$counter]: $(round(avgLoss, digits=5))"
    end
end
```


    callb (generic function with 1 method)


#### Structure of the CS-NeuralFMU

In the following, the topology of the CS-NeuralFMU is constructed. It consists of an input layer, which then leads into the *defaultFMU* model. The CS-FMU computes the outputs for the given system state and time step. After the *defaultFMU* follows a dense layer, which has exactly as many inputs as the model has outputs. The output of this layer consists of 16 output nodes and a *tanh* activation function. The next layer has 16 input and output nodes with the same activation function. The last layer is again a dense layer with 16 input nodes and the number of model outputs as output nodes. Here, it is important that no *tanh*-activation function follows, because otherwise the pendulums state values would be limited to the interval $[-1;1]$.


```julia
# NeuralFMU setup
numInputs = length(defaultFMU.modelDescription.inputValueReferences)
numOutputs = length(defaultFMU.modelDescription.outputValueReferences)

net = Chain(inputs -> fmiInputDoStepCSOutput(defaultFMU, tStep, inputs),
            Dense(numOutputs, 16, tanh),
            Dense(16, 16, tanh),
            Dense(16, numOutputs))
```


    Chain(
      var"#17#18"(),
      Dense(2, 16, tanh),                   [90m# 48 parameters[39m
      Dense(16, 16, tanh),                  [90m# 272 parameters[39m
      Dense(16, 2),                         [90m# 34 parameters[39m
    )[90m                   # Total: 6 arrays, [39m354 parameters, 1.758 KiB.


#### Definition of the CS-NeuralFMU

The instantiation of the CS-NeuralFMU is done as a one-liner. The FMU `defaultFMU`, the structure of the network `net`, start `tStart` and end time `tStop`, and the time steps `tSave` for saving are specified.


```julia
csNeuralFMU = CS_NeuralFMU(defaultFMU, net, (tStart, tStop); saveat=tSave);
```

#### Plot before training

Here the state trajactory of the *extForceFMU* is recorded. Doesn't really look like a pendulum yet, but the system is random initialized by default. In the later plots, the effect of learning can be seen.


```julia
solutionBefore = csNeuralFMU(extForce, tStep)
Plots.plot(tSave, collect(data[1] for data in solutionBefore), label="acc CS-NeuralFMU", linewidth=2)
```


<?xml version="1.0" encoding="utf-8"?>
<svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" width="600" height="400" viewBox="0 0 2400 1600">
<defs>
  <clipPath id="clip150">
    <rect x="0" y="0" width="2400" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip150)" d="
M0 1600 L2400 1600 L2400 0 L0 0  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip151">
    <rect x="480" y="0" width="1681" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip150)" d="
M224.098 1486.45 L2352.76 1486.45 L2352.76 47.2441 L224.098 47.2441  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip152">
    <rect x="224" y="47" width="2130" height="1440"/>
  </clipPath>
</defs>
<polyline clip-path="url(#clip152)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  284.343,1486.45 284.343,47.2441 
  "/>
<polyline clip-path="url(#clip152)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  685.977,1486.45 685.977,47.2441 
  "/>
<polyline clip-path="url(#clip152)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1087.61,1486.45 1087.61,47.2441 
  "/>
<polyline clip-path="url(#clip152)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1489.24,1486.45 1489.24,47.2441 
  "/>
<polyline clip-path="url(#clip152)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1890.88,1486.45 1890.88,47.2441 
  "/>
<polyline clip-path="url(#clip152)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  2292.51,1486.45 2292.51,47.2441 
  "/>
<polyline clip-path="url(#clip150)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  224.098,1486.45 2352.76,1486.45 
  "/>
<polyline clip-path="url(#clip150)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  284.343,1486.45 284.343,1467.55 
  "/>
<polyline clip-path="url(#clip150)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  685.977,1486.45 685.977,1467.55 
  "/>
<polyline clip-path="url(#clip150)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1087.61,1486.45 1087.61,1467.55 
  "/>
<polyline clip-path="url(#clip150)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1489.24,1486.45 1489.24,1467.55 
  "/>
<polyline clip-path="url(#clip150)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1890.88,1486.45 1890.88,1467.55 
  "/>
<polyline clip-path="url(#clip150)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  2292.51,1486.45 2292.51,1467.55 
  "/>
<path clip-path="url(#clip150)" d="M284.343 1517.37 Q280.732 1517.37 278.903 1520.93 Q277.098 1524.47 277.098 1531.6 Q277.098 1538.71 278.903 1542.27 Q280.732 1545.82 284.343 1545.82 Q287.977 1545.82 289.783 1542.27 Q291.611 1538.71 291.611 1531.6 Q291.611 1524.47 289.783 1520.93 Q287.977 1517.37 284.343 1517.37 M284.343 1513.66 Q290.153 1513.66 293.209 1518.27 Q296.287 1522.85 296.287 1531.6 Q296.287 1540.33 293.209 1544.94 Q290.153 1549.52 284.343 1549.52 Q278.533 1549.52 275.454 1544.94 Q272.399 1540.33 272.399 1531.6 Q272.399 1522.85 275.454 1518.27 Q278.533 1513.66 284.343 1513.66 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip150)" d="M676.358 1544.91 L683.997 1544.91 L683.997 1518.55 L675.687 1520.21 L675.687 1515.95 L683.951 1514.29 L688.627 1514.29 L688.627 1544.91 L696.266 1544.91 L696.266 1548.85 L676.358 1548.85 L676.358 1544.91 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip150)" d="M1082.26 1544.91 L1098.58 1544.91 L1098.58 1548.85 L1076.64 1548.85 L1076.64 1544.91 Q1079.3 1542.16 1083.88 1537.53 Q1088.49 1532.88 1089.67 1531.53 Q1091.92 1529.01 1092.8 1527.27 Q1093.7 1525.51 1093.7 1523.82 Q1093.7 1521.07 1091.75 1519.33 Q1089.83 1517.6 1086.73 1517.6 Q1084.53 1517.6 1082.08 1518.36 Q1079.65 1519.13 1076.87 1520.68 L1076.87 1515.95 Q1079.69 1514.82 1082.15 1514.24 Q1084.6 1513.66 1086.64 1513.66 Q1092.01 1513.66 1095.2 1516.35 Q1098.4 1519.03 1098.4 1523.52 Q1098.4 1525.65 1097.59 1527.57 Q1096.8 1529.47 1094.69 1532.07 Q1094.11 1532.74 1091.01 1535.95 Q1087.91 1539.15 1082.26 1544.91 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip150)" d="M1493.49 1530.21 Q1496.85 1530.93 1498.72 1533.2 Q1500.62 1535.47 1500.62 1538.8 Q1500.62 1543.92 1497.1 1546.72 Q1493.58 1549.52 1487.1 1549.52 Q1484.93 1549.52 1482.61 1549.08 Q1480.32 1548.66 1477.87 1547.81 L1477.87 1543.29 Q1479.81 1544.43 1482.13 1545.01 Q1484.44 1545.58 1486.96 1545.58 Q1491.36 1545.58 1493.65 1543.85 Q1495.97 1542.11 1495.97 1538.8 Q1495.97 1535.75 1493.82 1534.03 Q1491.69 1532.3 1487.87 1532.3 L1483.84 1532.3 L1483.84 1528.45 L1488.05 1528.45 Q1491.5 1528.45 1493.33 1527.09 Q1495.16 1525.7 1495.16 1523.11 Q1495.16 1520.45 1493.26 1519.03 Q1491.38 1517.6 1487.87 1517.6 Q1485.95 1517.6 1483.75 1518.01 Q1481.55 1518.43 1478.91 1519.31 L1478.91 1515.14 Q1481.57 1514.4 1483.88 1514.03 Q1486.22 1513.66 1488.28 1513.66 Q1493.61 1513.66 1496.71 1516.09 Q1499.81 1518.5 1499.81 1522.62 Q1499.81 1525.49 1498.17 1527.48 Q1496.52 1529.45 1493.49 1530.21 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip150)" d="M1893.89 1518.36 L1882.08 1536.81 L1893.89 1536.81 L1893.89 1518.36 M1892.66 1514.29 L1898.54 1514.29 L1898.54 1536.81 L1903.47 1536.81 L1903.47 1540.7 L1898.54 1540.7 L1898.54 1548.85 L1893.89 1548.85 L1893.89 1540.7 L1878.28 1540.7 L1878.28 1536.19 L1892.66 1514.29 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip150)" d="M2282.79 1514.29 L2301.15 1514.29 L2301.15 1518.22 L2287.07 1518.22 L2287.07 1526.7 Q2288.09 1526.35 2289.11 1526.19 Q2290.13 1526 2291.15 1526 Q2296.93 1526 2300.31 1529.17 Q2303.69 1532.34 2303.69 1537.76 Q2303.69 1543.34 2300.22 1546.44 Q2296.75 1549.52 2290.43 1549.52 Q2288.25 1549.52 2285.98 1549.15 Q2283.74 1548.78 2281.33 1548.04 L2281.33 1543.34 Q2283.41 1544.47 2285.64 1545.03 Q2287.86 1545.58 2290.33 1545.58 Q2294.34 1545.58 2296.68 1543.48 Q2299.02 1541.37 2299.02 1537.76 Q2299.02 1534.15 2296.68 1532.04 Q2294.34 1529.94 2290.33 1529.94 Q2288.46 1529.94 2286.58 1530.35 Q2284.73 1530.77 2282.79 1531.65 L2282.79 1514.29 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip152)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  224.098,1355.49 2352.76,1355.49 
  "/>
<polyline clip-path="url(#clip152)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  224.098,1061.26 2352.76,1061.26 
  "/>
<polyline clip-path="url(#clip152)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  224.098,767.02 2352.76,767.02 
  "/>
<polyline clip-path="url(#clip152)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  224.098,472.784 2352.76,472.784 
  "/>
<polyline clip-path="url(#clip152)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  224.098,178.548 2352.76,178.548 
  "/>
<polyline clip-path="url(#clip150)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  224.098,1486.45 224.098,47.2441 
  "/>
<polyline clip-path="url(#clip150)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  224.098,1355.49 242.996,1355.49 
  "/>
<polyline clip-path="url(#clip150)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  224.098,1061.26 242.996,1061.26 
  "/>
<polyline clip-path="url(#clip150)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  224.098,767.02 242.996,767.02 
  "/>
<polyline clip-path="url(#clip150)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  224.098,472.784 242.996,472.784 
  "/>
<polyline clip-path="url(#clip150)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  224.098,178.548 242.996,178.548 
  "/>
<path clip-path="url(#clip150)" d="M50.9921 1355.94 L80.6679 1355.94 L80.6679 1359.88 L50.9921 1359.88 L50.9921 1355.94 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip150)" d="M100.76 1341.29 Q97.1493 1341.29 95.3206 1344.86 Q93.515 1348.4 93.515 1355.53 Q93.515 1362.63 95.3206 1366.2 Q97.1493 1369.74 100.76 1369.74 Q104.395 1369.74 106.2 1366.2 Q108.029 1362.63 108.029 1355.53 Q108.029 1348.4 106.2 1344.86 Q104.395 1341.29 100.76 1341.29 M100.76 1337.59 Q106.571 1337.59 109.626 1342.19 Q112.705 1346.78 112.705 1355.53 Q112.705 1364.25 109.626 1368.86 Q106.571 1373.44 100.76 1373.44 Q94.9502 1373.44 91.8715 1368.86 Q88.816 1364.25 88.816 1355.53 Q88.816 1346.78 91.8715 1342.19 Q94.9502 1337.59 100.76 1337.59 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip150)" d="M120.922 1366.89 L125.807 1366.89 L125.807 1372.77 L120.922 1372.77 L120.922 1366.89 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip150)" d="M136.038 1338.21 L154.394 1338.21 L154.394 1342.15 L140.32 1342.15 L140.32 1350.62 Q141.339 1350.27 142.357 1350.11 Q143.376 1349.93 144.394 1349.93 Q150.181 1349.93 153.561 1353.1 Q156.941 1356.27 156.941 1361.68 Q156.941 1367.26 153.468 1370.37 Q149.996 1373.44 143.677 1373.44 Q141.501 1373.44 139.232 1373.07 Q136.987 1372.7 134.58 1371.96 L134.58 1367.26 Q136.663 1368.4 138.885 1368.95 Q141.107 1369.51 143.584 1369.51 Q147.589 1369.51 149.927 1367.4 Q152.265 1365.3 152.265 1361.68 Q152.265 1358.07 149.927 1355.97 Q147.589 1353.86 143.584 1353.86 Q141.709 1353.86 139.834 1354.28 Q137.982 1354.69 136.038 1355.57 L136.038 1338.21 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip150)" d="M176.153 1341.29 Q172.542 1341.29 170.714 1344.86 Q168.908 1348.4 168.908 1355.53 Q168.908 1362.63 170.714 1366.2 Q172.542 1369.74 176.153 1369.74 Q179.788 1369.74 181.593 1366.2 Q183.422 1362.63 183.422 1355.53 Q183.422 1348.4 181.593 1344.86 Q179.788 1341.29 176.153 1341.29 M176.153 1337.59 Q181.964 1337.59 185.019 1342.19 Q188.098 1346.78 188.098 1355.53 Q188.098 1364.25 185.019 1368.86 Q181.964 1373.44 176.153 1373.44 Q170.343 1373.44 167.265 1368.86 Q164.209 1364.25 164.209 1355.53 Q164.209 1346.78 167.265 1342.19 Q170.343 1337.59 176.153 1337.59 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip150)" d="M51.9875 1061.71 L81.6633 1061.71 L81.6633 1065.64 L51.9875 1065.64 L51.9875 1061.71 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip150)" d="M101.756 1047.06 Q98.1447 1047.06 96.316 1050.62 Q94.5104 1054.16 94.5104 1061.29 Q94.5104 1068.4 96.316 1071.96 Q98.1447 1075.5 101.756 1075.5 Q105.39 1075.5 107.196 1071.96 Q109.024 1068.4 109.024 1061.29 Q109.024 1054.16 107.196 1050.62 Q105.39 1047.06 101.756 1047.06 M101.756 1043.35 Q107.566 1043.35 110.621 1047.96 Q113.7 1052.54 113.7 1061.29 Q113.7 1070.02 110.621 1074.62 Q107.566 1079.21 101.756 1079.21 Q95.9456 1079.21 92.8669 1074.62 Q89.8114 1070.02 89.8114 1061.29 Q89.8114 1052.54 92.8669 1047.96 Q95.9456 1043.35 101.756 1043.35 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip150)" d="M121.918 1072.66 L126.802 1072.66 L126.802 1078.54 L121.918 1078.54 L121.918 1072.66 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip150)" d="M141.015 1074.6 L157.334 1074.6 L157.334 1078.54 L135.39 1078.54 L135.39 1074.6 Q138.052 1071.85 142.635 1067.22 Q147.242 1062.56 148.422 1061.22 Q150.668 1058.7 151.547 1056.96 Q152.45 1055.2 152.45 1053.51 Q152.45 1050.76 150.505 1049.02 Q148.584 1047.29 145.482 1047.29 Q143.283 1047.29 140.83 1048.05 Q138.399 1048.81 135.621 1050.37 L135.621 1045.64 Q138.445 1044.51 140.899 1043.93 Q143.353 1043.35 145.39 1043.35 Q150.76 1043.35 153.955 1046.04 Q157.149 1048.72 157.149 1053.21 Q157.149 1055.34 156.339 1057.26 Q155.552 1059.16 153.445 1061.75 Q152.867 1062.43 149.765 1065.64 Q146.663 1068.84 141.015 1074.6 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip150)" d="M167.195 1043.98 L185.552 1043.98 L185.552 1047.91 L171.478 1047.91 L171.478 1056.38 Q172.496 1056.04 173.515 1055.87 Q174.533 1055.69 175.552 1055.69 Q181.339 1055.69 184.718 1058.86 Q188.098 1062.03 188.098 1067.45 Q188.098 1073.03 184.626 1076.13 Q181.153 1079.21 174.834 1079.21 Q172.658 1079.21 170.39 1078.84 Q168.144 1078.47 165.737 1077.73 L165.737 1073.03 Q167.82 1074.16 170.042 1074.72 Q172.265 1075.27 174.741 1075.27 Q178.746 1075.27 181.084 1073.17 Q183.422 1071.06 183.422 1067.45 Q183.422 1063.84 181.084 1061.73 Q178.746 1059.62 174.741 1059.62 Q172.866 1059.62 170.991 1060.04 Q169.14 1060.46 167.195 1061.34 L167.195 1043.98 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip150)" d="M100.76 752.819 Q97.1493 752.819 95.3206 756.384 Q93.515 759.926 93.515 767.055 Q93.515 774.162 95.3206 777.726 Q97.1493 781.268 100.76 781.268 Q104.395 781.268 106.2 777.726 Q108.029 774.162 108.029 767.055 Q108.029 759.926 106.2 756.384 Q104.395 752.819 100.76 752.819 M100.76 749.115 Q106.571 749.115 109.626 753.722 Q112.705 758.305 112.705 767.055 Q112.705 775.782 109.626 780.388 Q106.571 784.972 100.76 784.972 Q94.9502 784.972 91.8715 780.388 Q88.816 775.782 88.816 767.055 Q88.816 758.305 91.8715 753.722 Q94.9502 749.115 100.76 749.115 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip150)" d="M120.922 778.421 L125.807 778.421 L125.807 784.3 L120.922 784.3 L120.922 778.421 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip150)" d="M145.992 752.819 Q142.381 752.819 140.552 756.384 Q138.746 759.926 138.746 767.055 Q138.746 774.162 140.552 777.726 Q142.381 781.268 145.992 781.268 Q149.626 781.268 151.431 777.726 Q153.26 774.162 153.26 767.055 Q153.26 759.926 151.431 756.384 Q149.626 752.819 145.992 752.819 M145.992 749.115 Q151.802 749.115 154.857 753.722 Q157.936 758.305 157.936 767.055 Q157.936 775.782 154.857 780.388 Q151.802 784.972 145.992 784.972 Q140.181 784.972 137.103 780.388 Q134.047 775.782 134.047 767.055 Q134.047 758.305 137.103 753.722 Q140.181 749.115 145.992 749.115 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip150)" d="M176.153 752.819 Q172.542 752.819 170.714 756.384 Q168.908 759.926 168.908 767.055 Q168.908 774.162 170.714 777.726 Q172.542 781.268 176.153 781.268 Q179.788 781.268 181.593 777.726 Q183.422 774.162 183.422 767.055 Q183.422 759.926 181.593 756.384 Q179.788 752.819 176.153 752.819 M176.153 749.115 Q181.964 749.115 185.019 753.722 Q188.098 758.305 188.098 767.055 Q188.098 775.782 185.019 780.388 Q181.964 784.972 176.153 784.972 Q170.343 784.972 167.265 780.388 Q164.209 775.782 164.209 767.055 Q164.209 758.305 167.265 753.722 Q170.343 749.115 176.153 749.115 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip150)" d="M101.756 458.583 Q98.1447 458.583 96.316 462.148 Q94.5104 465.689 94.5104 472.819 Q94.5104 479.925 96.316 483.49 Q98.1447 487.032 101.756 487.032 Q105.39 487.032 107.196 483.49 Q109.024 479.925 109.024 472.819 Q109.024 465.689 107.196 462.148 Q105.39 458.583 101.756 458.583 M101.756 454.879 Q107.566 454.879 110.621 459.486 Q113.7 464.069 113.7 472.819 Q113.7 481.546 110.621 486.152 Q107.566 490.736 101.756 490.736 Q95.9456 490.736 92.8669 486.152 Q89.8114 481.546 89.8114 472.819 Q89.8114 464.069 92.8669 459.486 Q95.9456 454.879 101.756 454.879 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip150)" d="M121.918 484.185 L126.802 484.185 L126.802 490.064 L121.918 490.064 L121.918 484.185 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip150)" d="M141.015 486.129 L157.334 486.129 L157.334 490.064 L135.39 490.064 L135.39 486.129 Q138.052 483.375 142.635 478.745 Q147.242 474.092 148.422 472.75 Q150.668 470.226 151.547 468.49 Q152.45 466.731 152.45 465.041 Q152.45 462.287 150.505 460.551 Q148.584 458.814 145.482 458.814 Q143.283 458.814 140.83 459.578 Q138.399 460.342 135.621 461.893 L135.621 457.171 Q138.445 456.037 140.899 455.458 Q143.353 454.879 145.39 454.879 Q150.76 454.879 153.955 457.564 Q157.149 460.25 157.149 464.74 Q157.149 466.87 156.339 468.791 Q155.552 470.689 153.445 473.282 Q152.867 473.953 149.765 477.171 Q146.663 480.365 141.015 486.129 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip150)" d="M167.195 455.504 L185.552 455.504 L185.552 459.439 L171.478 459.439 L171.478 467.912 Q172.496 467.564 173.515 467.402 Q174.533 467.217 175.552 467.217 Q181.339 467.217 184.718 470.388 Q188.098 473.56 188.098 478.976 Q188.098 484.555 184.626 487.657 Q181.153 490.736 174.834 490.736 Q172.658 490.736 170.39 490.365 Q168.144 489.995 165.737 489.254 L165.737 484.555 Q167.82 485.689 170.042 486.245 Q172.265 486.8 174.741 486.8 Q178.746 486.8 181.084 484.694 Q183.422 482.588 183.422 478.976 Q183.422 475.365 181.084 473.259 Q178.746 471.152 174.741 471.152 Q172.866 471.152 170.991 471.569 Q169.14 471.986 167.195 472.865 L167.195 455.504 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip150)" d="M100.76 164.347 Q97.1493 164.347 95.3206 167.912 Q93.515 171.453 93.515 178.583 Q93.515 185.689 95.3206 189.254 Q97.1493 192.796 100.76 192.796 Q104.395 192.796 106.2 189.254 Q108.029 185.689 108.029 178.583 Q108.029 171.453 106.2 167.912 Q104.395 164.347 100.76 164.347 M100.76 160.643 Q106.571 160.643 109.626 165.25 Q112.705 169.833 112.705 178.583 Q112.705 187.31 109.626 191.916 Q106.571 196.499 100.76 196.499 Q94.9502 196.499 91.8715 191.916 Q88.816 187.31 88.816 178.583 Q88.816 169.833 91.8715 165.25 Q94.9502 160.643 100.76 160.643 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip150)" d="M120.922 189.949 L125.807 189.949 L125.807 195.828 L120.922 195.828 L120.922 189.949 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip150)" d="M136.038 161.268 L154.394 161.268 L154.394 165.203 L140.32 165.203 L140.32 173.676 Q141.339 173.328 142.357 173.166 Q143.376 172.981 144.394 172.981 Q150.181 172.981 153.561 176.152 Q156.941 179.324 156.941 184.74 Q156.941 190.319 153.468 193.421 Q149.996 196.499 143.677 196.499 Q141.501 196.499 139.232 196.129 Q136.987 195.759 134.58 195.018 L134.58 190.319 Q136.663 191.453 138.885 192.009 Q141.107 192.564 143.584 192.564 Q147.589 192.564 149.927 190.458 Q152.265 188.351 152.265 184.74 Q152.265 181.129 149.927 179.023 Q147.589 176.916 143.584 176.916 Q141.709 176.916 139.834 177.333 Q137.982 177.75 136.038 178.629 L136.038 161.268 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip150)" d="M176.153 164.347 Q172.542 164.347 170.714 167.912 Q168.908 171.453 168.908 178.583 Q168.908 185.689 170.714 189.254 Q172.542 192.796 176.153 192.796 Q179.788 192.796 181.593 189.254 Q183.422 185.689 183.422 178.583 Q183.422 171.453 181.593 167.912 Q179.788 164.347 176.153 164.347 M176.153 160.643 Q181.964 160.643 185.019 165.25 Q188.098 169.833 188.098 178.583 Q188.098 187.31 185.019 191.916 Q181.964 196.499 176.153 196.499 Q170.343 196.499 167.265 191.916 Q164.209 187.31 164.209 178.583 Q164.209 169.833 167.265 165.25 Q170.343 160.643 176.153 160.643 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip152)" style="stroke:#009af9; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" points="
  284.343,889.522 288.359,891.842 292.376,894.863 296.392,898.577 300.408,902.973 304.425,908.038 308.441,913.761 312.457,920.128 316.474,927.128 320.49,934.748 
  324.506,942.978 328.523,951.806 332.539,961.223 336.555,971.224 340.572,981.801 344.588,992.953 348.604,1004.68 352.621,1016.99 356.637,1029.9 360.653,1043.42 
  364.67,1057.57 368.686,1072.38 372.702,1087.9 376.719,1104.18 380.735,1121.26 384.751,1139.22 388.768,1158.13 392.784,1178.05 396.8,1199.04 400.817,1221.17 
  404.833,1244.43 408.849,1268.79 412.866,1294.12 416.882,1320.15 420.898,1346.43 424.915,1372.25 428.931,1396.59 432.947,1418.04 436.964,1434.75 440.98,1444.55 
  444.996,1445.01 449.013,1433.75 453.029,1408.9 457.045,1369.49 461.062,1315.88 465.078,1249.82 469.094,1174.24 473.111,1092.63 477.127,1008.46 481.143,924.624 
  485.16,843.263 489.176,765.77 493.192,693.012 497.209,625.561 501.225,563.882 505.241,508.419 509.258,459.59 513.274,417.729 517.29,383.033 521.307,355.506 
  525.323,334.953 529.339,320.99 533.356,313.085 537.372,310.598 541.388,312.831 545.405,319.063 549.421,328.584 553.437,340.72 557.454,354.85 561.47,370.417 
  565.486,386.937 569.503,403.996 573.519,421.251 577.535,438.427 581.552,455.308 585.568,471.73 589.584,487.57 593.601,502.748 597.617,517.207 601.633,530.912 
  605.65,543.847 609.666,556.003 613.682,567.381 617.699,577.985 621.715,587.819 625.731,596.89 629.748,605.202 633.764,612.76 637.78,619.565 641.797,625.621 
  645.813,630.929 649.829,635.487 653.846,639.297 657.862,642.36 661.878,644.677 665.895,646.25 669.911,647.081 673.927,647.177 677.944,646.541 681.96,645.181 
  685.977,643.106 689.993,640.323 694.009,636.844 698.026,632.679 702.042,627.841 706.058,622.341 710.075,616.191 714.091,609.405 718.107,601.994 722.124,593.969 
  726.14,585.343 730.156,576.124 734.173,566.32 738.189,555.937 742.205,544.978 746.222,533.442 750.238,521.326 754.254,508.62 758.271,495.312 762.287,481.377 
  766.303,466.789 770.32,451.51 774.336,435.495 778.352,418.692 782.369,401.036 786.385,382.462 790.401,362.897 794.418,342.274 798.434,320.539 802.45,297.663 
  806.467,273.668 810.483,248.648 814.499,222.819 818.516,196.561 822.532,170.484 826.548,145.501 830.565,122.886 834.581,104.325 838.597,91.8976 842.614,87.9763 
  846.63,94.9858 850.646,115.032 854.663,149.439 858.679,198.332 862.695,260.418 866.712,333.129 870.728,413.086 874.744,496.747 878.761,580.977 882.777,663.371 
  886.793,742.304 890.81,816.758 894.826,886.081 898.842,949.785 902.859,1007.42 906.875,1058.56 910.891,1102.83 914.908,1139.98 918.924,1169.94 922.94,1192.84 
  926.957,1209 930.973,1218.9 934.989,1223.16 939.006,1222.45 943.022,1217.5 947.038,1209.02 951.055,1197.7 955.071,1184.19 959.087,1169.05 963.104,1152.81 
  967.12,1135.89 971.136,1118.67 975.153,1101.44 979.169,1084.44 983.185,1067.85 987.202,1051.8 991.218,1036.39 995.234,1021.68 999.251,1007.72 1003.27,994.525 
  1007.28,982.105 1011.3,970.464 1015.32,959.6 1019.33,949.507 1023.35,940.179 1027.37,931.612 1031.38,923.8 1035.4,916.74 1039.41,910.43 1043.43,904.871 
  1047.45,900.059 1051.46,895.997 1055.48,892.681 1059.5,890.113 1063.51,888.289 1067.53,887.206 1071.54,886.862 1075.56,887.251 1079.58,888.367 1083.59,890.201 
  1087.61,892.745 1091.63,895.99 1095.64,899.923 1099.66,904.533 1103.68,909.81 1107.69,915.74 1111.71,922.312 1115.72,929.512 1119.74,937.329 1123.76,945.752 
  1127.77,954.769 1131.79,964.373 1135.81,974.559 1139.82,985.321 1143.84,996.659 1147.86,1008.57 1151.87,1021.08 1155.89,1034.18 1159.9,1047.89 1163.92,1062.25 
  1167.94,1077.29 1171.95,1093.05 1175.97,1109.57 1179.99,1126.93 1184,1145.18 1188.02,1164.4 1192.03,1184.66 1196.05,1206.01 1200.07,1228.49 1204.08,1252.11 
  1208.1,1276.8 1212.12,1302.38 1216.13,1328.54 1220.15,1354.76 1224.17,1380.22 1228.18,1403.79 1232.2,1423.92 1236.21,1438.66 1240.23,1445.72 1244.25,1442.65 
  1248.26,1427.2 1252.28,1397.72 1256.3,1353.66 1260.31,1295.82 1264.33,1226.37 1268.35,1148.48 1272.36,1065.71 1276.38,981.388 1280.39,898.175 1284.41,817.953 
  1288.43,741.922 1292.44,670.825 1296.46,605.192 1300.48,545.474 1304.49,492.11 1308.51,445.496 1312.52,405.929 1316.54,373.547 1320.56,348.289 1324.57,329.896 
  1328.59,317.933 1332.61,311.829 1336.62,310.922 1340.64,314.502 1344.66,321.848 1348.67,332.262 1352.69,345.086 1356.7,359.718 1360.72,375.625 1364.74,392.345 
  1368.75,409.488 1372.77,426.733 1376.79,443.824 1380.8,460.563 1384.82,476.8 1388.84,492.429 1392.85,507.373 1396.87,521.587 1400.88,535.041 1404.9,547.721 
  1408.92,559.623 1412.93,570.748 1416.95,581.1 1420.97,590.684 1424.98,599.507 1429,607.573 1433.01,614.886 1437.03,621.448 1441.05,627.261 1445.06,632.325 
  1449.08,636.641 1453.1,640.208 1457.11,643.029 1461.13,645.105 1465.15,646.438 1469.16,647.032 1473.18,646.892 1477.19,646.023 1481.21,644.432 1485.23,642.129 
  1489.24,639.122 1493.26,635.423 1497.28,631.042 1501.29,625.992 1505.31,620.285 1509.33,613.932 1513.34,606.947 1517.36,599.341 1521.37,591.125 1525.39,582.312 
  1529.41,572.909 1533.42,562.924 1537.44,552.361 1541.46,541.222 1545.47,529.507 1549.49,517.209 1553.51,504.318 1557.52,490.817 1561.54,476.682 1565.55,461.884 
  1569.57,446.38 1573.59,430.123 1577.6,413.058 1581.62,395.12 1585.64,376.24 1589.65,356.347 1593.67,335.378 1597.68,313.286 1601.7,290.053 1605.72,265.723 
  1609.73,240.427 1613.75,214.425 1617.77,188.168 1621.78,162.356 1625.8,138.012 1629.82,116.538 1633.83,99.7519 1637.85,89.8432 1641.86,89.2293 1645.88,100.274 
  1649.9,124.879 1653.91,164.014 1657.93,217.352 1661.95,283.165 1665.96,358.561 1669.98,440.042 1674,524.149 1678.01,607.965 1682.03,689.343 1686.04,766.875 
  1690.06,839.686 1694.08,907.203 1698.09,968.957 1702.11,1024.5 1706.13,1073.43 1710.14,1115.4 1714.16,1150.21 1718.17,1177.86 1722.19,1198.54 1726.21,1212.63 
  1730.22,1220.66 1734.24,1223.26 1738.26,1221.14 1742.27,1215 1746.29,1205.57 1750.31,1193.51 1754.32,1179.44 1758.34,1163.92 1762.35,1147.45 1766.37,1130.42 
  1770.39,1113.19 1774.4,1096.03 1778.42,1079.16 1782.44,1062.75 1786.45,1046.91 1790.47,1031.73 1794.49,1017.27 1798.5,1003.56 1802.52,990.62 1806.53,978.457 
  1810.55,967.072 1814.57,956.461 1818.58,946.62 1822.6,937.542 1826.62,929.221 1830.63,921.656 1834.65,914.841 1838.66,908.776 1842.68,903.459 1846.7,898.89 
  1850.71,895.07 1854.73,891.996 1858.75,889.669 1862.76,888.084 1866.78,887.24 1870.8,887.132 1874.81,887.753 1878.83,889.099 1882.84,891.159 1886.86,893.926 
  1890.88,897.389 1894.89,901.537 1898.91,906.358 1902.93,911.84 1906.94,917.972 1910.96,924.739 1914.98,932.132 1918.99,940.136 1923.01,948.743 1927.02,957.942 
  1931.04,967.726 1935.06,978.089 1939.07,989.027 1943.09,1000.54 1947.11,1012.63 1951.12,1025.32 1955.14,1038.6 1959.15,1052.52 1963.17,1067.08 1967.19,1082.34 
  1971.2,1098.33 1975.22,1115.11 1979.24,1132.73 1983.25,1151.28 1987.27,1170.82 1991.29,1191.41 1995.3,1213.11 1999.32,1235.95 2003.33,1259.91 2007.35,1284.89 
  2011.37,1310.68 2015.38,1336.9 2019.4,1362.95 2023.42,1387.92 2027.43,1410.54 2031.45,1429.14 2035.47,1441.65 2039.48,1445.7 2043.5,1438.87 2047.51,1419.05 
  2051.53,1384.92 2055.55,1336.31 2059.56,1274.5 2063.58,1202.02 2067.6,1122.24 2071.61,1038.69 2075.63,954.521 2079.65,872.137 2083.66,793.186 2087.68,718.698 
  2091.69,649.33 2095.71,585.569 2099.73,527.863 2103.74,476.638 2107.76,432.266 2111.78,394.998 2115.79,364.91 2119.81,341.876 2123.82,325.579 2127.84,315.537 
  2131.86,311.145 2135.87,311.724 2139.89,316.558 2143.91,324.931 2147.92,336.157 2151.94,349.594 2155.96,364.664 2159.97,380.855 2163.99,397.729 2168,414.918 
  2172.02,432.124 2176.04,449.11 2180.05,465.692 2184.07,481.736 2188.09,497.145 2192.1,511.855 2196.12,525.824 2200.14,539.028 2204.15,551.457 2208.17,563.108 
  2212.18,573.983 2216.2,584.088 2220.22,593.428 2224.23,602.008 2228.25,609.833 2232.27,616.906 2236.28,623.229 2240.3,628.803 2244.31,633.629 2248.33,637.708 
  2252.35,641.039 2256.36,643.624 2260.38,645.466 2264.4,646.565 2268.41,646.928 2272.43,646.557 2276.45,645.461 2280.46,643.647 2284.48,641.123 2288.49,637.9 
  2292.51,633.987 
  "/>
<path clip-path="url(#clip150)" d="
M1616.44 198.898 L2281.8 198.898 L2281.8 95.2176 L1616.44 95.2176  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<polyline clip-path="url(#clip150)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1616.44,198.898 2281.8,198.898 2281.8,95.2176 1616.44,95.2176 1616.44,198.898 
  "/>
<polyline clip-path="url(#clip150)" style="stroke:#009af9; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1640.09,147.058 1782.01,147.058 
  "/>
<path clip-path="url(#clip150)" d="M1819.06 151.305 Q1813.9 151.305 1811.91 152.486 Q1809.92 153.666 1809.92 156.514 Q1809.92 158.782 1811.4 160.125 Q1812.9 161.444 1815.47 161.444 Q1819.01 161.444 1821.14 158.944 Q1823.3 156.421 1823.3 152.254 L1823.3 151.305 L1819.06 151.305 M1827.55 149.546 L1827.55 164.338 L1823.3 164.338 L1823.3 160.402 Q1821.84 162.763 1819.66 163.898 Q1817.49 165.009 1814.34 165.009 Q1810.36 165.009 1807.99 162.787 Q1805.66 160.541 1805.66 156.791 Q1805.66 152.416 1808.57 150.194 Q1811.51 147.972 1817.32 147.972 L1823.3 147.972 L1823.3 147.555 Q1823.3 144.615 1821.35 143.018 Q1819.43 141.398 1815.93 141.398 Q1813.71 141.398 1811.61 141.93 Q1809.5 142.463 1807.55 143.527 L1807.55 139.592 Q1809.89 138.69 1812.09 138.25 Q1814.29 137.787 1816.37 137.787 Q1822 137.787 1824.78 140.703 Q1827.55 143.62 1827.55 149.546 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip150)" d="M1854.99 139.407 L1854.99 143.389 Q1853.18 142.393 1851.35 141.907 Q1849.55 141.398 1847.69 141.398 Q1843.55 141.398 1841.26 144.037 Q1838.97 146.652 1838.97 151.398 Q1838.97 156.143 1841.26 158.782 Q1843.55 161.398 1847.69 161.398 Q1849.55 161.398 1851.35 160.912 Q1853.18 160.402 1854.99 159.407 L1854.99 163.342 Q1853.2 164.176 1851.28 164.592 Q1849.38 165.009 1847.23 165.009 Q1841.37 165.009 1837.93 161.328 Q1834.48 157.648 1834.48 151.398 Q1834.48 145.055 1837.95 141.421 Q1841.44 137.787 1847.51 137.787 Q1849.48 137.787 1851.35 138.203 Q1853.23 138.597 1854.99 139.407 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip150)" d="M1881.05 139.407 L1881.05 143.389 Q1879.24 142.393 1877.42 141.907 Q1875.61 141.398 1873.76 141.398 Q1869.61 141.398 1867.32 144.037 Q1865.03 146.652 1865.03 151.398 Q1865.03 156.143 1867.32 158.782 Q1869.61 161.398 1873.76 161.398 Q1875.61 161.398 1877.42 160.912 Q1879.24 160.402 1881.05 159.407 L1881.05 163.342 Q1879.27 164.176 1877.35 164.592 Q1875.45 165.009 1873.3 165.009 Q1867.44 165.009 1863.99 161.328 Q1860.54 157.648 1860.54 151.398 Q1860.54 145.055 1864.01 141.421 Q1867.51 137.787 1873.57 137.787 Q1875.54 137.787 1877.42 138.203 Q1879.29 138.597 1881.05 139.407 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip150)" d="M1929.59 132.44 L1929.59 137.37 Q1927.23 135.171 1924.55 134.083 Q1921.88 132.995 1918.87 132.995 Q1912.95 132.995 1909.8 136.629 Q1906.65 140.24 1906.65 147.092 Q1906.65 153.921 1909.8 157.555 Q1912.95 161.166 1918.87 161.166 Q1921.88 161.166 1924.55 160.078 Q1927.23 158.99 1929.59 156.791 L1929.59 161.676 Q1927.14 163.342 1924.38 164.176 Q1921.65 165.009 1918.6 165.009 Q1910.75 165.009 1906.23 160.217 Q1901.72 155.402 1901.72 147.092 Q1901.72 138.759 1906.23 133.967 Q1910.75 129.153 1918.6 129.153 Q1921.7 129.153 1924.43 129.986 Q1927.18 130.796 1929.59 132.44 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip150)" d="M1957.53 130.912 L1957.53 135.472 Q1954.87 134.199 1952.51 133.574 Q1950.15 132.949 1947.95 132.949 Q1944.13 132.949 1942.04 134.43 Q1939.98 135.912 1939.98 138.643 Q1939.98 140.935 1941.35 142.115 Q1942.74 143.273 1946.58 143.99 L1949.41 144.569 Q1954.64 145.565 1957.11 148.088 Q1959.61 150.588 1959.61 154.801 Q1959.61 159.824 1956.23 162.416 Q1952.88 165.009 1946.37 165.009 Q1943.92 165.009 1941.14 164.453 Q1938.39 163.898 1935.42 162.81 L1935.42 157.995 Q1938.27 159.592 1941 160.402 Q1943.73 161.213 1946.37 161.213 Q1950.38 161.213 1952.55 159.639 Q1954.73 158.064 1954.73 155.148 Q1954.73 152.602 1953.16 151.166 Q1951.61 149.731 1948.04 149.014 L1945.19 148.458 Q1939.96 147.416 1937.62 145.194 Q1935.29 142.972 1935.29 139.014 Q1935.29 134.43 1938.5 131.791 Q1941.74 129.153 1947.42 129.153 Q1949.85 129.153 1952.37 129.592 Q1954.89 130.032 1957.53 130.912 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip150)" d="M1964.57 149.453 L1977.04 149.453 L1977.04 153.25 L1964.57 153.25 L1964.57 149.453 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip150)" d="M1984.01 129.778 L1990.31 129.778 L2005.63 158.689 L2005.63 129.778 L2010.17 129.778 L2010.17 164.338 L2003.87 164.338 L1988.55 135.426 L1988.55 164.338 L1984.01 164.338 L1984.01 129.778 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip150)" d="M2041.47 150.31 L2041.47 152.393 L2021.88 152.393 Q2022.16 156.791 2024.52 159.106 Q2026.91 161.398 2031.14 161.398 Q2033.6 161.398 2035.89 160.796 Q2038.2 160.194 2040.47 158.99 L2040.47 163.018 Q2038.18 163.99 2035.77 164.5 Q2033.36 165.009 2030.89 165.009 Q2024.68 165.009 2021.05 161.398 Q2017.44 157.787 2017.44 151.629 Q2017.44 145.264 2020.86 141.537 Q2024.31 137.787 2030.15 137.787 Q2035.38 137.787 2038.41 141.166 Q2041.47 144.523 2041.47 150.31 M2037.21 149.06 Q2037.16 145.565 2035.24 143.481 Q2033.34 141.398 2030.19 141.398 Q2026.63 141.398 2024.48 143.412 Q2022.35 145.426 2022.02 149.083 L2037.21 149.06 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip150)" d="M2048.02 154.106 L2048.02 138.412 L2052.28 138.412 L2052.28 153.944 Q2052.28 157.625 2053.71 159.476 Q2055.15 161.305 2058.02 161.305 Q2061.47 161.305 2063.46 159.106 Q2065.47 156.907 2065.47 153.111 L2065.47 138.412 L2069.73 138.412 L2069.73 164.338 L2065.47 164.338 L2065.47 160.356 Q2063.92 162.717 2061.86 163.875 Q2059.82 165.009 2057.11 165.009 Q2052.65 165.009 2050.33 162.231 Q2048.02 159.453 2048.02 154.106 M2058.73 137.787 L2058.73 137.787 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip150)" d="M2093.53 142.393 Q2092.81 141.977 2091.95 141.791 Q2091.12 141.583 2090.1 141.583 Q2086.49 141.583 2084.54 143.944 Q2082.62 146.282 2082.62 150.68 L2082.62 164.338 L2078.34 164.338 L2078.34 138.412 L2082.62 138.412 L2082.62 142.44 Q2083.97 140.078 2086.12 138.944 Q2088.27 137.787 2091.35 137.787 Q2091.79 137.787 2092.32 137.856 Q2092.85 137.903 2093.5 138.018 L2093.53 142.393 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip150)" d="M2109.78 151.305 Q2104.61 151.305 2102.62 152.486 Q2100.63 153.666 2100.63 156.514 Q2100.63 158.782 2102.11 160.125 Q2103.62 161.444 2106.19 161.444 Q2109.73 161.444 2111.86 158.944 Q2114.01 156.421 2114.01 152.254 L2114.01 151.305 L2109.78 151.305 M2118.27 149.546 L2118.27 164.338 L2114.01 164.338 L2114.01 160.402 Q2112.55 162.763 2110.38 163.898 Q2108.2 165.009 2105.05 165.009 Q2101.07 165.009 2098.71 162.787 Q2096.37 160.541 2096.37 156.791 Q2096.37 152.416 2099.29 150.194 Q2102.23 147.972 2108.04 147.972 L2114.01 147.972 L2114.01 147.555 Q2114.01 144.615 2112.07 143.018 Q2110.15 141.398 2106.65 141.398 Q2104.43 141.398 2102.32 141.93 Q2100.22 142.463 2098.27 143.527 L2098.27 139.592 Q2100.61 138.69 2102.81 138.25 Q2105.01 137.787 2107.09 137.787 Q2112.72 137.787 2115.49 140.703 Q2118.27 143.62 2118.27 149.546 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip150)" d="M2127.04 128.319 L2131.3 128.319 L2131.3 164.338 L2127.04 164.338 L2127.04 128.319 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip150)" d="M2140.4 129.778 L2160.26 129.778 L2160.26 133.713 L2145.08 133.713 L2145.08 143.898 L2158.78 143.898 L2158.78 147.833 L2145.08 147.833 L2145.08 164.338 L2140.4 164.338 L2140.4 129.778 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip150)" d="M2167.67 129.778 L2174.64 129.778 L2183.46 153.296 L2192.32 129.778 L2199.29 129.778 L2199.29 164.338 L2194.73 164.338 L2194.73 133.99 L2185.82 157.694 L2181.12 157.694 L2172.21 133.99 L2172.21 164.338 L2167.67 164.338 L2167.67 129.778 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip150)" d="M2208.04 129.778 L2212.74 129.778 L2212.74 150.773 Q2212.74 156.328 2214.75 158.782 Q2216.77 161.213 2221.28 161.213 Q2225.77 161.213 2227.78 158.782 Q2229.8 156.328 2229.8 150.773 L2229.8 129.778 L2234.5 129.778 L2234.5 151.352 Q2234.5 158.111 2231.14 161.56 Q2227.81 165.009 2221.28 165.009 Q2214.73 165.009 2211.37 161.56 Q2208.04 158.111 2208.04 151.352 L2208.04 129.778 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /></svg>



#### Training of the CS-NeuralFMU

For the training of the CS-NeuralFMU the parameters are extracted. The known ADAM optimizer for minimizing the gradient descent is used as further passing parameters. In addition, the previously defined loss and callback function, as well as the number of epochs are passed.


```julia
# train
paramsNet = Flux.params(csNeuralFMU)

optim = ADAM()
Flux.train!(lossSum, paramsNet, Iterators.repeated((), 300), optim; cb=callb)
```

    ┌ Info: Loss [1]: 1.45099
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_CS.ipynb:8
    ┌ Info: Loss [21]: 0.39826
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_CS.ipynb:8
    ┌ Info: Loss [41]: 0.19409
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_CS.ipynb:8
    ┌ Info: Loss [61]: 0.1406
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_CS.ipynb:8
    ┌ Info: Loss [81]: 0.09556
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_CS.ipynb:8
    ┌ Info: Loss [101]: 0.06242
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_CS.ipynb:8
    ┌ Info: Loss [121]: 0.03803
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_CS.ipynb:8
    ┌ Info: Loss [141]: 0.02162
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_CS.ipynb:8
    ┌ Info: Loss [161]: 0.01175
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_CS.ipynb:8
    ┌ Info: Loss [181]: 0.00649
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_CS.ipynb:8
    ┌ Info: Loss [201]: 0.00393
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_CS.ipynb:8
    ┌ Info: Loss [221]: 0.00273
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_CS.ipynb:8
    ┌ Info: Loss [241]: 0.00214
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_CS.ipynb:8
    ┌ Info: Loss [261]: 0.00181
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_CS.ipynb:8
    ┌ Info: Loss [281]: 0.00157
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_CS.ipynb:8


#### Comparison of the plots

Here three plots are compared with each other and only the acceleration of the mass is considered. The first plot represents the *defaultFMU*, the second represents the *referenceFMU* and the third plot represents the result after training the CS-NeuralFMU. 


```julia
# plot results mass.a
solutionAfter = csNeuralFMU(extForce, tStep)

fig = Plots.plot(xlabel="t [s]", ylabel="mass acceleration [m/s^2]", linewidth=2,
                 xtickfontsize=12, ytickfontsize=12,
                 xguidefontsize=12, yguidefontsize=12,
                 legendfontsize=8, legend=:topright)

accNeuralFMU = collect(data[1] for data in solutionAfter)

Plots.plot!(fig, tSave, accDefault, label="defaultFMU", linewidth=2)
Plots.plot!(fig, tSave, accReference, label="referenceFMU", linewidth=2)
Plots.plot!(fig, tSave, accNeuralFMU, label="CS-NeuralFMU (300 eps.)", linewidth=2)
fig 
```


<?xml version="1.0" encoding="utf-8"?>
<svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" width="600" height="400" viewBox="0 0 2400 1600">
<defs>
  <clipPath id="clip240">
    <rect x="0" y="0" width="2400" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip240)" d="
M0 1600 L2400 1600 L2400 0 L0 0  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip241">
    <rect x="480" y="0" width="1681" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip240)" d="
M314.996 1401.58 L2352.76 1401.58 L2352.76 47.2441 L314.996 47.2441  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip242">
    <rect x="314" y="47" width="2039" height="1355"/>
  </clipPath>
</defs>
<polyline clip-path="url(#clip242)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  372.668,1401.58 372.668,47.2441 
  "/>
<polyline clip-path="url(#clip242)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  757.151,1401.58 757.151,47.2441 
  "/>
<polyline clip-path="url(#clip242)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1141.63,1401.58 1141.63,47.2441 
  "/>
<polyline clip-path="url(#clip242)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1526.12,1401.58 1526.12,47.2441 
  "/>
<polyline clip-path="url(#clip242)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1910.6,1401.58 1910.6,47.2441 
  "/>
<polyline clip-path="url(#clip242)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  2295.08,1401.58 2295.08,47.2441 
  "/>
<polyline clip-path="url(#clip240)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  314.996,1401.58 2352.76,1401.58 
  "/>
<polyline clip-path="url(#clip240)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  372.668,1401.58 372.668,1382.68 
  "/>
<polyline clip-path="url(#clip240)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  757.151,1401.58 757.151,1382.68 
  "/>
<polyline clip-path="url(#clip240)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1141.63,1401.58 1141.63,1382.68 
  "/>
<polyline clip-path="url(#clip240)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1526.12,1401.58 1526.12,1382.68 
  "/>
<polyline clip-path="url(#clip240)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1910.6,1401.58 1910.6,1382.68 
  "/>
<polyline clip-path="url(#clip240)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  2295.08,1401.58 2295.08,1382.68 
  "/>
<path clip-path="url(#clip240)" d="M372.668 1438.36 Q367.252 1438.36 364.509 1443.71 Q361.8 1449.02 361.8 1459.71 Q361.8 1470.37 364.509 1475.72 Q367.252 1481.03 372.668 1481.03 Q378.12 1481.03 380.828 1475.72 Q383.571 1470.37 383.571 1459.71 Q383.571 1449.02 380.828 1443.71 Q378.12 1438.36 372.668 1438.36 M372.668 1432.8 Q381.384 1432.8 385.967 1439.71 Q390.585 1446.59 390.585 1459.71 Q390.585 1472.8 385.967 1479.71 Q381.384 1486.59 372.668 1486.59 Q363.953 1486.59 359.335 1479.71 Q354.752 1472.8 354.752 1459.71 Q354.752 1446.59 359.335 1439.71 Q363.953 1432.8 372.668 1432.8 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M742.724 1479.68 L754.183 1479.68 L754.183 1440.13 L741.717 1442.63 L741.717 1436.24 L754.113 1433.74 L761.127 1433.74 L761.127 1479.68 L772.585 1479.68 L772.585 1485.58 L742.724 1485.58 L742.724 1479.68 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M1133.61 1479.68 L1158.09 1479.68 L1158.09 1485.58 L1125.18 1485.58 L1125.18 1479.68 Q1129.17 1475.55 1136.04 1468.6 Q1142.95 1461.62 1144.72 1459.61 Q1148.09 1455.82 1149.41 1453.22 Q1150.77 1450.58 1150.77 1448.05 Q1150.77 1443.91 1147.85 1441.31 Q1144.97 1438.71 1140.31 1438.71 Q1137.02 1438.71 1133.34 1439.85 Q1129.69 1441 1125.52 1443.32 L1125.52 1436.24 Q1129.76 1434.54 1133.44 1433.67 Q1137.12 1432.8 1140.18 1432.8 Q1148.23 1432.8 1153.02 1436.83 Q1157.81 1440.86 1157.81 1447.59 Q1157.81 1450.79 1156.6 1453.67 Q1155.42 1456.52 1152.26 1460.41 Q1151.39 1461.41 1146.74 1466.24 Q1142.09 1471.03 1133.61 1479.68 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M1532.49 1457.63 Q1537.52 1458.7 1540.34 1462.11 Q1543.18 1465.51 1543.18 1470.51 Q1543.18 1478.18 1537.91 1482.39 Q1532.63 1486.59 1522.91 1486.59 Q1519.64 1486.59 1516.17 1485.93 Q1512.73 1485.3 1509.05 1484.02 L1509.05 1477.25 Q1511.97 1478.95 1515.44 1479.82 Q1518.91 1480.68 1522.7 1480.68 Q1529.29 1480.68 1532.73 1478.08 Q1536.2 1475.48 1536.2 1470.51 Q1536.2 1465.93 1532.98 1463.36 Q1529.78 1460.75 1524.05 1460.75 L1518.01 1460.75 L1518.01 1454.99 L1524.33 1454.99 Q1529.5 1454.99 1532.25 1452.94 Q1534.99 1450.86 1534.99 1446.97 Q1534.99 1442.98 1532.14 1440.86 Q1529.33 1438.71 1524.05 1438.71 Q1521.17 1438.71 1517.87 1439.33 Q1514.57 1439.96 1510.61 1441.27 L1510.61 1435.02 Q1514.61 1433.91 1518.08 1433.36 Q1521.59 1432.8 1524.68 1432.8 Q1532.66 1432.8 1537.32 1436.45 Q1541.97 1440.06 1541.97 1446.24 Q1541.97 1450.55 1539.5 1453.53 Q1537.04 1456.48 1532.49 1457.63 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M1915.11 1439.85 L1897.41 1467.52 L1915.11 1467.52 L1915.11 1439.85 M1913.27 1433.74 L1922.09 1433.74 L1922.09 1467.52 L1929.49 1467.52 L1929.49 1473.36 L1922.09 1473.36 L1922.09 1485.58 L1915.11 1485.58 L1915.11 1473.36 L1891.71 1473.36 L1891.71 1466.59 L1913.27 1433.74 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M2280.5 1433.74 L2308.03 1433.74 L2308.03 1439.64 L2286.92 1439.64 L2286.92 1452.35 Q2288.45 1451.83 2289.98 1451.59 Q2291.51 1451.31 2293.03 1451.31 Q2301.72 1451.31 2306.78 1456.07 Q2311.85 1460.82 2311.85 1468.95 Q2311.85 1477.32 2306.65 1481.97 Q2301.44 1486.59 2291.96 1486.59 Q2288.69 1486.59 2285.29 1486.03 Q2281.92 1485.48 2278.31 1484.36 L2278.31 1477.32 Q2281.44 1479.02 2284.77 1479.85 Q2288.1 1480.68 2291.82 1480.68 Q2297.83 1480.68 2301.33 1477.52 Q2304.84 1474.36 2304.84 1468.95 Q2304.84 1463.53 2301.33 1460.37 Q2297.83 1457.21 2291.82 1457.21 Q2289.01 1457.21 2286.19 1457.84 Q2283.42 1458.46 2280.5 1459.78 L2280.5 1433.74 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M1277.49 1527.62 L1277.49 1538.66 L1290.65 1538.66 L1290.65 1543.62 L1277.49 1543.62 L1277.49 1564.74 Q1277.49 1569.49 1278.77 1570.85 Q1280.09 1572.2 1284.08 1572.2 L1290.65 1572.2 L1290.65 1577.55 L1284.08 1577.55 Q1276.69 1577.55 1273.88 1574.8 Q1271.06 1572.03 1271.06 1564.74 L1271.06 1543.62 L1266.38 1543.62 L1266.38 1538.66 L1271.06 1538.66 L1271.06 1527.62 L1277.49 1527.62 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M1321.06 1523.52 L1335.79 1523.52 L1335.79 1528.49 L1327.45 1528.49 L1327.45 1581.96 L1335.79 1581.96 L1335.79 1586.92 L1321.06 1586.92 L1321.06 1523.52 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M1374.19 1539.8 L1374.19 1545.85 Q1371.48 1544.46 1368.56 1543.76 Q1365.65 1543.07 1362.52 1543.07 Q1357.76 1543.07 1355.37 1544.53 Q1353.01 1545.99 1353.01 1548.9 Q1353.01 1551.12 1354.71 1552.41 Q1356.41 1553.66 1361.55 1554.8 L1363.74 1555.29 Q1370.54 1556.75 1373.39 1559.42 Q1376.27 1562.06 1376.27 1566.82 Q1376.27 1572.24 1371.97 1575.4 Q1367.7 1578.55 1360.2 1578.55 Q1357.07 1578.55 1353.67 1577.93 Q1350.3 1577.34 1346.55 1576.12 L1346.55 1569.53 Q1350.09 1571.37 1353.53 1572.3 Q1356.97 1573.21 1360.33 1573.21 Q1364.85 1573.21 1367.28 1571.68 Q1369.71 1570.12 1369.71 1567.3 Q1369.71 1564.7 1367.94 1563.31 Q1366.2 1561.92 1360.26 1560.64 L1358.04 1560.12 Q1352.11 1558.87 1349.47 1556.3 Q1346.83 1553.69 1346.83 1549.18 Q1346.83 1543.69 1350.72 1540.71 Q1354.6 1537.72 1361.76 1537.72 Q1365.3 1537.72 1368.42 1538.24 Q1371.55 1538.76 1374.19 1539.8 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M1401.38 1523.52 L1401.38 1586.92 L1386.65 1586.92 L1386.65 1581.96 L1394.95 1581.96 L1394.95 1528.49 L1386.65 1528.49 L1386.65 1523.52 L1401.38 1523.52 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip242)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  314.996,1256.75 2352.76,1256.75 
  "/>
<polyline clip-path="url(#clip242)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  314.996,990.57 2352.76,990.57 
  "/>
<polyline clip-path="url(#clip242)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  314.996,724.395 2352.76,724.395 
  "/>
<polyline clip-path="url(#clip242)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  314.996,458.22 2352.76,458.22 
  "/>
<polyline clip-path="url(#clip242)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  314.996,192.044 2352.76,192.044 
  "/>
<polyline clip-path="url(#clip240)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  314.996,1401.58 314.996,47.2441 
  "/>
<polyline clip-path="url(#clip240)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  314.996,1256.75 333.894,1256.75 
  "/>
<polyline clip-path="url(#clip240)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  314.996,990.57 333.894,990.57 
  "/>
<polyline clip-path="url(#clip240)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  314.996,724.395 333.894,724.395 
  "/>
<polyline clip-path="url(#clip240)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  314.996,458.22 333.894,458.22 
  "/>
<polyline clip-path="url(#clip240)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  314.996,192.044 333.894,192.044 
  "/>
<path clip-path="url(#clip240)" d="M118.58 1257.42 L163.094 1257.42 L163.094 1263.33 L118.58 1263.33 L118.58 1257.42 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M178.302 1230.83 L205.837 1230.83 L205.837 1236.73 L184.726 1236.73 L184.726 1249.44 Q186.253 1248.92 187.781 1248.67 Q189.309 1248.39 190.837 1248.39 Q199.517 1248.39 204.587 1253.15 Q209.656 1257.91 209.656 1266.03 Q209.656 1274.4 204.448 1279.05 Q199.239 1283.67 189.76 1283.67 Q186.496 1283.67 183.094 1283.12 Q179.726 1282.56 176.115 1281.45 L176.115 1274.4 Q179.24 1276.1 182.573 1276.94 Q185.906 1277.77 189.621 1277.77 Q195.628 1277.77 199.135 1274.61 Q202.642 1271.45 202.642 1266.03 Q202.642 1260.62 199.135 1257.46 Q195.628 1254.3 189.621 1254.3 Q186.809 1254.3 183.996 1254.92 Q181.219 1255.55 178.302 1256.87 L178.302 1230.83 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M223.475 1273.85 L230.802 1273.85 L230.802 1282.67 L223.475 1282.67 L223.475 1273.85 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M261.079 1235.44 Q255.663 1235.44 252.92 1240.79 Q250.211 1246.1 250.211 1256.8 Q250.211 1267.46 252.92 1272.8 Q255.663 1278.12 261.079 1278.12 Q266.531 1278.12 269.239 1272.8 Q271.982 1267.46 271.982 1256.8 Q271.982 1246.1 269.239 1240.79 Q266.531 1235.44 261.079 1235.44 M261.079 1229.89 Q269.795 1229.89 274.378 1236.8 Q278.996 1243.67 278.996 1256.8 Q278.996 1269.89 274.378 1276.8 Q269.795 1283.67 261.079 1283.67 Q252.364 1283.67 247.746 1276.8 Q243.163 1269.89 243.163 1256.8 Q243.163 1243.67 247.746 1236.8 Q252.364 1229.89 261.079 1229.89 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M120.073 991.247 L164.587 991.247 L164.587 997.15 L120.073 997.15 L120.073 991.247 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M185.767 1010.59 L210.246 1010.59 L210.246 1016.49 L177.33 1016.49 L177.33 1010.59 Q181.323 1006.46 188.198 999.511 Q195.108 992.532 196.878 990.518 Q200.246 986.733 201.566 984.129 Q202.92 981.49 202.92 978.955 Q202.92 974.824 200.003 972.219 Q197.121 969.615 192.469 969.615 Q189.17 969.615 185.49 970.761 Q181.844 971.907 177.677 974.233 L177.677 967.15 Q181.913 965.449 185.594 964.581 Q189.274 963.713 192.33 963.713 Q200.385 963.713 205.177 967.74 Q209.969 971.768 209.969 978.504 Q209.969 981.699 208.753 984.58 Q207.573 987.428 204.413 991.317 Q203.545 992.323 198.892 997.15 Q194.239 1001.94 185.767 1010.59 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M224.968 1007.67 L232.295 1007.67 L232.295 1016.49 L224.968 1016.49 L224.968 1007.67 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M247.642 964.65 L275.177 964.65 L275.177 970.553 L254.066 970.553 L254.066 983.261 Q255.593 982.74 257.121 982.497 Q258.649 982.219 260.177 982.219 Q268.857 982.219 273.927 986.976 Q278.996 991.733 278.996 999.858 Q278.996 1008.23 273.788 1012.88 Q268.579 1017.5 259.1 1017.5 Q255.836 1017.5 252.434 1016.94 Q249.066 1016.39 245.454 1015.27 L245.454 1008.23 Q248.579 1009.93 251.913 1010.76 Q255.246 1011.59 258.961 1011.59 Q264.968 1011.59 268.475 1008.43 Q271.982 1005.27 271.982 999.858 Q271.982 994.442 268.475 991.282 Q264.968 988.122 258.961 988.122 Q256.149 988.122 253.336 988.747 Q250.559 989.372 247.642 990.692 L247.642 964.65 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M193.233 703.093 Q187.816 703.093 185.073 708.44 Q182.365 713.753 182.365 724.447 Q182.365 735.107 185.073 740.454 Q187.816 745.766 193.233 745.766 Q198.684 745.766 201.392 740.454 Q204.135 735.107 204.135 724.447 Q204.135 713.753 201.392 708.44 Q198.684 703.093 193.233 703.093 M193.233 697.537 Q201.948 697.537 206.531 704.447 Q211.149 711.322 211.149 724.447 Q211.149 737.537 206.531 744.447 Q201.948 751.322 193.233 751.322 Q184.517 751.322 179.899 744.447 Q175.316 737.537 175.316 724.447 Q175.316 711.322 179.899 704.447 Q184.517 697.537 193.233 697.537 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M223.475 741.495 L230.802 741.495 L230.802 750.315 L223.475 750.315 L223.475 741.495 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M261.079 703.093 Q255.663 703.093 252.92 708.44 Q250.211 713.753 250.211 724.447 Q250.211 735.107 252.92 740.454 Q255.663 745.766 261.079 745.766 Q266.531 745.766 269.239 740.454 Q271.982 735.107 271.982 724.447 Q271.982 713.753 269.239 708.44 Q266.531 703.093 261.079 703.093 M261.079 697.537 Q269.795 697.537 274.378 704.447 Q278.996 711.322 278.996 724.447 Q278.996 737.537 274.378 744.447 Q269.795 751.322 261.079 751.322 Q252.364 751.322 247.746 744.447 Q243.163 737.537 243.163 724.447 Q243.163 711.322 247.746 704.447 Q252.364 697.537 261.079 697.537 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M185.767 478.237 L210.246 478.237 L210.246 484.14 L177.33 484.14 L177.33 478.237 Q181.323 474.105 188.198 467.161 Q195.108 460.181 196.878 458.168 Q200.246 454.383 201.566 451.779 Q202.92 449.14 202.92 446.605 Q202.92 442.473 200.003 439.869 Q197.121 437.265 192.469 437.265 Q189.17 437.265 185.49 438.411 Q181.844 439.557 177.677 441.883 L177.677 434.8 Q181.913 433.098 185.594 432.23 Q189.274 431.362 192.33 431.362 Q200.385 431.362 205.177 435.39 Q209.969 439.418 209.969 446.154 Q209.969 449.348 208.753 452.23 Q207.573 455.077 204.413 458.966 Q203.545 459.973 198.892 464.799 Q194.239 469.591 185.767 478.237 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M224.968 475.32 L232.295 475.32 L232.295 484.14 L224.968 484.14 L224.968 475.32 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M247.642 432.3 L275.177 432.3 L275.177 438.202 L254.066 438.202 L254.066 450.911 Q255.593 450.39 257.121 450.147 Q258.649 449.869 260.177 449.869 Q268.857 449.869 273.927 454.626 Q278.996 459.383 278.996 467.508 Q278.996 475.876 273.788 480.529 Q268.579 485.147 259.1 485.147 Q255.836 485.147 252.434 484.591 Q249.066 484.035 245.454 482.924 L245.454 475.876 Q248.579 477.577 251.913 478.41 Q255.246 479.244 258.961 479.244 Q264.968 479.244 268.475 476.084 Q271.982 472.924 271.982 467.508 Q271.982 462.091 268.475 458.931 Q264.968 455.772 258.961 455.772 Q256.149 455.772 253.336 456.397 Q250.559 457.022 247.642 458.341 L247.642 432.3 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M178.302 166.124 L205.837 166.124 L205.837 172.027 L184.726 172.027 L184.726 184.735 Q186.253 184.215 187.781 183.972 Q189.309 183.694 190.837 183.694 Q199.517 183.694 204.587 188.451 Q209.656 193.208 209.656 201.333 Q209.656 209.701 204.448 214.353 Q199.239 218.971 189.76 218.971 Q186.496 218.971 183.094 218.416 Q179.726 217.86 176.115 216.749 L176.115 209.701 Q179.24 211.402 182.573 212.235 Q185.906 213.069 189.621 213.069 Q195.628 213.069 199.135 209.909 Q202.642 206.749 202.642 201.333 Q202.642 195.916 199.135 192.756 Q195.628 189.597 189.621 189.597 Q186.809 189.597 183.996 190.222 Q181.219 190.847 178.302 192.166 L178.302 166.124 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M223.475 209.145 L230.802 209.145 L230.802 217.964 L223.475 217.964 L223.475 209.145 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M261.079 170.742 Q255.663 170.742 252.92 176.09 Q250.211 181.402 250.211 192.097 Q250.211 202.756 252.92 208.103 Q255.663 213.416 261.079 213.416 Q266.531 213.416 269.239 208.103 Q271.982 202.756 271.982 192.097 Q271.982 181.402 269.239 176.09 Q266.531 170.742 261.079 170.742 M261.079 165.187 Q269.795 165.187 274.378 172.097 Q278.996 178.972 278.996 192.097 Q278.996 205.187 274.378 212.096 Q269.795 218.971 261.079 218.971 Q252.364 218.971 247.746 212.096 Q243.163 205.187 243.163 192.097 Q243.163 178.972 247.746 172.097 Q252.364 165.187 261.079 165.187 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M35.6047 1168 Q31.2992 1165.61 29.2506 1162.27 Q27.202 1158.94 27.202 1154.43 Q27.202 1148.35 31.4728 1145.05 Q35.7089 1141.75 43.5561 1141.75 L67.0282 1141.75 L67.0282 1148.18 L43.7644 1148.18 Q38.1741 1148.18 35.4658 1150.16 Q32.7575 1152.14 32.7575 1156.2 Q32.7575 1161.16 36.0561 1164.05 Q39.3547 1166.93 45.0491 1166.93 L67.0282 1166.93 L67.0282 1173.35 L43.7644 1173.35 Q38.1394 1173.35 35.4658 1175.33 Q32.7575 1177.31 32.7575 1181.44 Q32.7575 1186.34 36.0908 1189.22 Q39.3894 1192.1 45.0491 1192.1 L67.0282 1192.1 L67.0282 1198.52 L28.1395 1198.52 L28.1395 1192.1 L34.1811 1192.1 Q30.6047 1189.91 28.9034 1186.86 Q27.202 1183.8 27.202 1179.6 Q27.202 1175.36 29.3547 1172.41 Q31.5075 1169.43 35.6047 1168 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M47.4797 1111.34 Q47.4797 1119.08 49.2505 1122.07 Q51.0213 1125.05 55.2921 1125.05 Q58.6949 1125.05 60.7087 1122.83 Q62.6879 1120.57 62.6879 1116.72 Q62.6879 1111.41 58.9379 1108.21 Q55.1532 1104.98 48.9033 1104.98 L47.4797 1104.98 L47.4797 1111.34 M44.8408 1098.59 L67.0282 1098.59 L67.0282 1104.98 L61.1254 1104.98 Q64.6671 1107.17 66.3684 1110.43 Q68.0351 1113.7 68.0351 1118.42 Q68.0351 1124.39 64.7018 1127.93 Q61.3337 1131.44 55.7088 1131.44 Q49.1463 1131.44 45.813 1127.07 Q42.4797 1122.66 42.4797 1113.94 L42.4797 1104.98 L41.8547 1104.98 Q37.445 1104.98 35.0492 1107.9 Q32.6186 1110.78 32.6186 1116.02 Q32.6186 1119.36 33.4172 1122.52 Q34.2158 1125.68 35.813 1128.59 L29.9103 1128.59 Q28.5561 1125.09 27.8964 1121.79 Q27.202 1118.49 27.202 1115.36 Q27.202 1106.93 31.577 1102.76 Q35.9519 1098.59 44.8408 1098.59 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M29.2853 1060.64 L35.3269 1060.64 Q33.9381 1063.35 33.2436 1066.27 Q32.5492 1069.18 32.5492 1072.31 Q32.5492 1077.07 34.0075 1079.46 Q35.4658 1081.82 38.3825 1081.82 Q40.6047 1081.82 41.8894 1080.12 Q43.1394 1078.42 44.2852 1073.28 L44.7713 1071.09 Q46.2297 1064.29 48.9033 1061.44 Q51.5421 1058.56 56.299 1058.56 Q61.7157 1058.56 64.8754 1062.87 Q68.0351 1067.14 68.0351 1074.64 Q68.0351 1077.76 67.4101 1081.16 Q66.8198 1084.53 65.6046 1088.28 L59.0074 1088.28 Q60.8476 1084.74 61.7851 1081.3 Q62.6879 1077.86 62.6879 1074.5 Q62.6879 1069.98 61.1601 1067.55 Q59.5976 1065.12 56.7852 1065.12 Q54.181 1065.12 52.7921 1066.89 Q51.4032 1068.63 50.1185 1074.57 L49.5977 1076.79 Q48.3477 1082.73 45.7783 1085.36 Q43.1741 1088 38.6603 1088 Q33.1742 1088 30.1881 1084.11 Q27.202 1080.23 27.202 1073.07 Q27.202 1069.53 27.7228 1066.41 Q28.2436 1063.28 29.2853 1060.64 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M29.2853 1023.59 L35.3269 1023.59 Q33.9381 1026.3 33.2436 1029.22 Q32.5492 1032.14 32.5492 1035.26 Q32.5492 1040.02 34.0075 1042.41 Q35.4658 1044.77 38.3825 1044.77 Q40.6047 1044.77 41.8894 1043.07 Q43.1394 1041.37 44.2852 1036.23 L44.7713 1034.05 Q46.2297 1027.24 48.9033 1024.39 Q51.5421 1021.51 56.299 1021.51 Q61.7157 1021.51 64.8754 1025.82 Q68.0351 1030.09 68.0351 1037.59 Q68.0351 1040.71 67.4101 1044.12 Q66.8198 1047.48 65.6046 1051.23 L59.0074 1051.23 Q60.8476 1047.69 61.7851 1044.25 Q62.6879 1040.82 62.6879 1037.45 Q62.6879 1032.93 61.1601 1030.5 Q59.5976 1028.07 56.7852 1028.07 Q54.181 1028.07 52.7921 1029.84 Q51.4032 1031.58 50.1185 1037.52 L49.5977 1039.74 Q48.3477 1045.68 45.7783 1048.32 Q43.1741 1050.96 38.6603 1050.96 Q33.1742 1050.96 30.1881 1047.07 Q27.202 1043.18 27.202 1036.02 Q27.202 1032.48 27.7228 1029.36 Q28.2436 1026.23 29.2853 1023.59 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M47.4797 971.06 Q47.4797 978.803 49.2505 981.789 Q51.0213 984.775 55.2921 984.775 Q58.6949 984.775 60.7087 982.553 Q62.6879 980.296 62.6879 976.442 Q62.6879 971.129 58.9379 967.935 Q55.1532 964.706 48.9033 964.706 L47.4797 964.706 L47.4797 971.06 M44.8408 958.317 L67.0282 958.317 L67.0282 964.706 L61.1254 964.706 Q64.6671 966.893 66.3684 970.157 Q68.0351 973.421 68.0351 978.143 Q68.0351 984.115 64.7018 987.657 Q61.3337 991.164 55.7088 991.164 Q49.1463 991.164 45.813 986.789 Q42.4797 982.379 42.4797 973.664 L42.4797 964.706 L41.8547 964.706 Q37.445 964.706 35.0492 967.622 Q32.6186 970.504 32.6186 975.747 Q32.6186 979.081 33.4172 982.24 Q34.2158 985.4 35.813 988.317 L29.9103 988.317 Q28.5561 984.81 27.8964 981.511 Q27.202 978.213 27.202 975.088 Q27.202 966.65 31.577 962.484 Q35.9519 958.317 44.8408 958.317 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M29.6325 917.171 L35.6047 917.171 Q34.1117 919.88 33.3825 922.623 Q32.6186 925.331 32.6186 928.109 Q32.6186 934.324 36.5769 937.762 Q40.5005 941.199 47.6185 941.199 Q54.7366 941.199 58.6949 937.762 Q62.6185 934.324 62.6185 928.109 Q62.6185 925.331 61.8893 922.623 Q61.1254 919.88 59.6324 917.171 L65.5351 917.171 Q66.7851 919.845 67.4101 922.727 Q68.0351 925.574 68.0351 928.803 Q68.0351 937.588 62.5143 942.761 Q56.9935 947.935 47.6185 947.935 Q38.1047 947.935 32.6533 942.727 Q27.202 937.484 27.202 928.387 Q27.202 925.435 27.827 922.623 Q28.4173 919.81 29.6325 917.171 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M29.6325 878.074 L35.6047 878.074 Q34.1117 880.783 33.3825 883.526 Q32.6186 886.234 32.6186 889.012 Q32.6186 895.227 36.5769 898.665 Q40.5005 902.102 47.6185 902.102 Q54.7366 902.102 58.6949 898.665 Q62.6185 895.227 62.6185 889.012 Q62.6185 886.234 61.8893 883.526 Q61.1254 880.783 59.6324 878.074 L65.5351 878.074 Q66.7851 880.748 67.4101 883.63 Q68.0351 886.477 68.0351 889.706 Q68.0351 898.491 62.5143 903.664 Q56.9935 908.838 47.6185 908.838 Q38.1047 908.838 32.6533 903.63 Q27.202 898.387 27.202 889.29 Q27.202 886.338 27.827 883.526 Q28.4173 880.713 29.6325 878.074 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M45.9866 833.7 L49.1116 833.7 L49.1116 863.074 Q55.7088 862.658 59.181 859.116 Q62.6185 855.54 62.6185 849.186 Q62.6185 845.505 61.7157 842.068 Q60.8129 838.595 59.0074 835.193 L65.049 835.193 Q66.5073 838.63 67.2712 842.241 Q68.0351 845.852 68.0351 849.568 Q68.0351 858.873 62.6185 864.324 Q57.2018 869.741 47.9658 869.741 Q38.4172 869.741 32.827 864.602 Q27.202 859.429 27.202 850.679 Q27.202 842.831 32.2714 838.283 Q37.3061 833.7 45.9866 833.7 M44.1116 840.088 Q38.8686 840.158 35.7436 843.04 Q32.6186 845.887 32.6186 850.609 Q32.6186 855.956 35.6394 859.186 Q38.6603 862.38 44.1463 862.866 L44.1116 840.088 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M13.0007 823.214 L13.0007 816.825 L67.0282 816.825 L67.0282 823.214 L13.0007 823.214 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M45.9866 770.193 L49.1116 770.193 L49.1116 799.568 Q55.7088 799.151 59.181 795.61 Q62.6185 792.033 62.6185 785.679 Q62.6185 781.998 61.7157 778.561 Q60.8129 775.089 59.0074 771.686 L65.049 771.686 Q66.5073 775.124 67.2712 778.735 Q68.0351 782.346 68.0351 786.061 Q68.0351 795.366 62.6185 800.818 Q57.2018 806.234 47.9658 806.234 Q38.4172 806.234 32.827 801.096 Q27.202 795.922 27.202 787.172 Q27.202 779.325 32.2714 774.776 Q37.3061 770.193 45.9866 770.193 M44.1116 776.582 Q38.8686 776.651 35.7436 779.533 Q32.6186 782.38 32.6186 787.103 Q32.6186 792.45 35.6394 795.679 Q38.6603 798.873 44.1463 799.359 L44.1116 776.582 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M34.1117 737.172 Q33.4867 738.249 33.2089 739.533 Q32.8964 740.783 32.8964 742.311 Q32.8964 747.728 36.438 750.644 Q39.945 753.526 46.5422 753.526 L67.0282 753.526 L67.0282 759.95 L28.1395 759.95 L28.1395 753.526 L34.1811 753.526 Q30.6395 751.513 28.9381 748.283 Q27.202 745.054 27.202 740.436 Q27.202 739.776 27.3061 738.978 Q27.3756 738.179 27.5492 737.207 L34.1117 737.172 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M47.4797 712.797 Q47.4797 720.54 49.2505 723.527 Q51.0213 726.513 55.2921 726.513 Q58.6949 726.513 60.7087 724.29 Q62.6879 722.034 62.6879 718.179 Q62.6879 712.867 58.9379 709.672 Q55.1532 706.443 48.9033 706.443 L47.4797 706.443 L47.4797 712.797 M44.8408 700.054 L67.0282 700.054 L67.0282 706.443 L61.1254 706.443 Q64.6671 708.631 66.3684 711.895 Q68.0351 715.159 68.0351 719.881 Q68.0351 725.853 64.7018 729.395 Q61.3337 732.902 55.7088 732.902 Q49.1463 732.902 45.813 728.527 Q42.4797 724.117 42.4797 715.402 L42.4797 706.443 L41.8547 706.443 Q37.445 706.443 35.0492 709.36 Q32.6186 712.242 32.6186 717.485 Q32.6186 720.818 33.4172 723.978 Q34.2158 727.138 35.813 730.054 L29.9103 730.054 Q28.5561 726.547 27.8964 723.249 Q27.202 719.95 27.202 716.825 Q27.202 708.388 31.577 704.221 Q35.9519 700.054 44.8408 700.054 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M17.0979 680.575 L28.1395 680.575 L28.1395 667.416 L33.1047 667.416 L33.1047 680.575 L54.2157 680.575 Q58.9726 680.575 60.3268 679.291 Q61.681 677.971 61.681 673.978 L61.681 667.416 L67.0282 667.416 L67.0282 673.978 Q67.0282 681.374 64.2851 684.186 Q61.5074 686.999 54.2157 686.999 L33.1047 686.999 L33.1047 691.686 L28.1395 691.686 L28.1395 686.999 L17.0979 686.999 L17.0979 680.575 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M28.1395 659.013 L28.1395 652.624 L67.0282 652.624 L67.0282 659.013 L28.1395 659.013 M13.0007 659.013 L13.0007 652.624 L21.0909 652.624 L21.0909 659.013 L13.0007 659.013 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M32.6186 624.187 Q32.6186 629.326 36.6464 632.312 Q40.6394 635.298 47.6185 635.298 Q54.5977 635.298 58.6254 632.346 Q62.6185 629.36 62.6185 624.187 Q62.6185 619.083 58.5907 616.097 Q54.5629 613.11 47.6185 613.11 Q40.7089 613.11 36.6811 616.097 Q32.6186 619.083 32.6186 624.187 M27.202 624.187 Q27.202 615.854 32.6186 611.097 Q38.0353 606.34 47.6185 606.34 Q57.1671 606.34 62.6185 611.097 Q68.0351 615.854 68.0351 624.187 Q68.0351 632.555 62.6185 637.312 Q57.1671 642.034 47.6185 642.034 Q38.0353 642.034 32.6186 637.312 Q27.202 632.555 27.202 624.187 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M43.5561 563.423 L67.0282 563.423 L67.0282 569.812 L43.7644 569.812 Q38.2436 569.812 35.5005 571.965 Q32.7575 574.118 32.7575 578.423 Q32.7575 583.597 36.0561 586.583 Q39.3547 589.569 45.0491 589.569 L67.0282 589.569 L67.0282 595.993 L28.1395 595.993 L28.1395 589.569 L34.1811 589.569 Q30.6742 587.277 28.9381 584.187 Q27.202 581.062 27.202 577 Q27.202 570.298 31.3686 566.861 Q35.5005 563.423 43.5561 563.423 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M13.0007 528.666 L13.0007 513.944 L17.9659 513.944 L17.9659 522.278 L71.4379 522.278 L71.4379 513.944 L76.4031 513.944 L76.4031 528.666 L13.0007 528.666 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M35.6047 470.056 Q31.2992 467.66 29.2506 464.327 Q27.202 460.993 27.202 456.479 Q27.202 450.403 31.4728 447.104 Q35.7089 443.806 43.5561 443.806 L67.0282 443.806 L67.0282 450.229 L43.7644 450.229 Q38.1741 450.229 35.4658 452.209 Q32.7575 454.188 32.7575 458.25 Q32.7575 463.215 36.0561 466.097 Q39.3547 468.979 45.0491 468.979 L67.0282 468.979 L67.0282 475.403 L43.7644 475.403 Q38.1394 475.403 35.4658 477.382 Q32.7575 479.361 32.7575 483.493 Q32.7575 488.389 36.0908 491.271 Q39.3894 494.153 45.0491 494.153 L67.0282 494.153 L67.0282 500.576 L28.1395 500.576 L28.1395 494.153 L34.1811 494.153 Q30.6047 491.965 28.9034 488.91 Q27.202 485.854 27.202 481.653 Q27.202 477.417 29.3547 474.465 Q31.5075 471.479 35.6047 470.056 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M15.1882 419.709 L15.1882 413.806 L73.6253 431.861 L73.6253 437.764 L15.1882 419.709 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M29.2853 382.313 L35.3269 382.313 Q33.9381 385.021 33.2436 387.938 Q32.5492 390.855 32.5492 393.98 Q32.5492 398.737 34.0075 401.132 Q35.4658 403.494 38.3825 403.494 Q40.6047 403.494 41.8894 401.792 Q43.1394 400.091 44.2852 394.952 L44.7713 392.764 Q46.2297 385.959 48.9033 383.112 Q51.5421 380.23 56.299 380.23 Q61.7157 380.23 64.8754 384.535 Q68.0351 388.806 68.0351 396.306 Q68.0351 399.431 67.4101 402.834 Q66.8198 406.202 65.6046 409.952 L59.0074 409.952 Q60.8476 406.41 61.7851 402.973 Q62.6879 399.535 62.6879 396.167 Q62.6879 391.653 61.1601 389.223 Q59.5976 386.792 56.7852 386.792 Q54.181 386.792 52.7921 388.563 Q51.4032 390.299 50.1185 396.237 L49.5977 398.459 Q48.3477 404.396 45.7783 407.035 Q43.1741 409.674 38.6603 409.674 Q33.1742 409.674 30.1881 405.785 Q27.202 401.896 27.202 394.744 Q27.202 391.202 27.7228 388.077 Q28.2436 384.952 29.2853 382.313 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M15.1882 343.563 L34.5283 324.709 L34.5283 331.688 L20.8131 346.966 L34.5283 362.244 L34.5283 369.223 L15.1882 350.369 L15.1882 343.563 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M61.1254 303.529 L61.1254 279.05 L67.0282 279.05 L67.0282 311.966 L61.1254 311.966 Q56.9935 307.973 50.0491 301.098 Q43.07 294.189 41.0561 292.418 Q37.2714 289.05 34.6672 287.73 Q32.0283 286.376 29.4936 286.376 Q25.3617 286.376 22.7576 289.293 Q20.1534 292.175 20.1534 296.827 Q20.1534 300.126 21.2992 303.807 Q22.4451 307.452 24.7714 311.619 L17.6881 311.619 Q15.9868 307.383 15.1187 303.702 Q14.2507 300.022 14.2507 296.966 Q14.2507 288.911 18.2784 284.119 Q22.3062 279.328 29.0422 279.328 Q32.2367 279.328 35.1186 280.543 Q37.9658 281.723 41.8547 284.883 Q42.8616 285.751 47.688 290.404 Q52.4796 295.057 61.1254 303.529 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M13.0007 250.3 L76.4031 250.3 L76.4031 265.022 L71.4379 265.022 L71.4379 256.724 L17.9659 256.724 L17.9659 265.022 L13.0007 265.022 L13.0007 250.3 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip242)" style="stroke:#009af9; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" points="
  372.668,85.5744 376.513,85.8956 380.358,86.8535 384.203,88.4486 388.048,90.6795 391.893,93.5435 395.737,97.0379 399.582,101.16 403.427,105.904 407.272,111.267 
  411.117,117.243 414.962,123.826 418.806,131.009 422.651,138.787 426.496,147.148 430.341,156.085 434.186,165.591 438.031,175.657 441.875,186.269 445.72,197.419 
  449.565,209.098 453.41,221.294 457.255,233.991 461.1,247.18 464.944,260.847 468.789,274.978 472.634,289.56 476.479,304.576 480.324,320.013 484.168,335.854 
  488.013,352.084 491.858,368.687 495.703,385.645 499.548,402.942 503.393,420.56 507.237,438.483 511.082,456.691 514.927,475.166 518.772,493.891 522.617,512.847 
  526.462,532.013 530.306,551.372 534.151,570.905 537.996,590.591 541.841,610.41 545.686,630.343 549.531,650.371 553.375,670.473 557.22,690.628 561.065,710.818 
  564.91,731.021 568.755,751.218 572.6,771.387 576.444,791.509 580.289,811.565 584.134,831.534 587.979,851.395 591.824,871.129 595.669,890.717 599.513,910.138 
  603.358,929.373 607.203,948.404 611.048,967.211 614.893,985.774 618.738,1004.08 622.582,1022.1 626.427,1039.82 630.272,1057.23 634.117,1074.31 637.962,1091.04 
  641.807,1107.4 645.651,1123.37 649.496,1138.95 653.341,1154.12 657.186,1168.85 661.031,1183.14 664.876,1196.97 668.72,1210.33 672.565,1223.2 676.41,1235.57 
  680.255,1247.44 684.1,1258.78 687.944,1269.58 691.789,1279.84 695.634,1289.54 699.479,1298.68 703.324,1307.25 707.169,1315.23 711.013,1322.62 714.858,1329.41 
  718.703,1335.6 722.548,1341.18 726.393,1346.14 730.238,1350.48 734.082,1354.19 737.927,1357.27 741.772,1359.72 745.617,1361.53 749.462,1362.71 753.307,1363.25 
  757.151,1363.15 760.996,1362.41 764.841,1361.03 768.686,1359.02 772.531,1356.37 776.376,1353.09 780.22,1349.18 784.065,1344.65 787.91,1339.49 791.755,1333.72 
  795.6,1327.35 799.445,1320.36 803.289,1312.79 807.134,1304.62 810.979,1295.87 814.824,1286.55 818.669,1276.67 822.514,1266.24 826.358,1255.27 830.203,1243.76 
  834.048,1231.74 837.893,1219.2 841.738,1206.18 845.583,1192.67 849.427,1178.69 853.272,1164.26 857.117,1149.39 860.962,1134.09 864.807,1118.39 868.651,1102.29 
  872.496,1085.81 876.341,1068.97 880.186,1051.79 884.031,1034.28 887.876,1016.46 891.72,998.349 895.565,979.963 899.41,961.321 903.255,942.443 907.1,923.346 
  910.945,904.05 914.789,884.575 918.634,864.94 922.479,845.164 926.324,825.266 930.169,805.268 934.014,785.189 937.858,765.049 941.703,744.869 945.548,724.668 
  949.393,704.467 953.238,684.285 957.083,664.142 960.927,644.06 964.772,624.059 968.617,604.157 972.462,584.377 976.307,564.738 980.152,545.257 983.996,525.955 
  987.841,506.851 991.686,487.965 995.531,469.316 999.376,450.923 1003.22,432.803 1007.07,414.976 1010.91,397.457 1014.76,380.266 1018.6,363.42 1022.44,346.934 
  1026.29,330.826 1030.13,315.112 1033.98,299.808 1037.82,284.928 1041.67,270.489 1045.51,256.504 1049.36,242.986 1053.2,229.951 1057.05,217.41 1060.89,205.375 
  1064.74,193.859 1068.58,182.875 1072.43,172.433 1076.27,162.543 1080.12,153.214 1083.96,144.457 1087.81,136.28 1091.65,128.69 1095.5,121.695 1099.34,115.305 
  1103.19,109.524 1107.03,104.358 1110.88,99.8115 1114.72,95.8897 1118.57,92.596 1122.41,89.9334 1126.26,87.905 1130.1,86.5149 1133.94,85.7628 1137.79,85.6494 
  1141.63,86.1745 1145.48,87.3376 1149.32,89.1372 1153.17,91.5712 1157.01,94.6387 1160.86,98.3373 1164.7,102.662 1168.55,107.609 1172.39,113.172 1176.24,119.345 
  1180.08,126.124 1183.93,133.499 1187.77,141.467 1191.62,150.019 1195.46,159.145 1199.31,168.836 1203.15,179.083 1207,189.874 1210.84,201.199 1214.69,213.046 
  1218.53,225.407 1222.38,238.267 1226.22,251.612 1230.07,265.431 1233.91,279.708 1237.76,294.429 1241.6,309.579 1245.44,325.144 1249.29,341.11 1253.13,357.459 
  1256.98,374.174 1260.82,391.24 1264.67,408.638 1268.51,426.352 1272.36,444.362 1276.2,462.653 1280.05,481.207 1283.89,500.004 1287.74,519.025 1291.58,538.251 
  1295.43,557.663 1299.27,577.241 1303.12,596.965 1306.96,616.818 1310.81,636.779 1314.65,656.827 1318.5,676.942 1322.34,697.105 1326.19,717.294 1330.03,737.49 
  1333.88,757.672 1337.72,777.822 1341.57,797.919 1345.41,817.941 1349.26,837.87 1353.1,857.686 1356.94,877.367 1360.79,896.895 1364.63,916.251 1368.48,935.415 
  1372.32,954.368 1376.17,973.09 1380.01,991.564 1383.86,1009.77 1387.7,1027.69 1391.55,1045.31 1395.39,1062.61 1399.24,1079.56 1403.08,1096.17 1406.93,1112.4 
  1410.77,1128.24 1414.62,1143.68 1418.46,1158.7 1422.31,1173.28 1426.15,1187.42 1430,1201.09 1433.84,1214.28 1437.69,1226.99 1441.53,1239.19 1445.38,1250.88 
  1449.22,1262.04 1453.07,1272.67 1456.91,1282.74 1460.76,1292.26 1464.6,1301.21 1468.44,1309.58 1472.29,1317.37 1476.13,1324.56 1479.98,1331.15 1483.82,1337.14 
  1487.67,1342.51 1491.51,1347.27 1495.36,1351.4 1499.2,1354.91 1503.05,1357.78 1506.89,1360.02 1510.74,1361.63 1514.58,1362.6 1518.43,1362.93 1522.27,1362.62 
  1526.12,1361.67 1529.96,1360.09 1533.81,1357.87 1537.65,1355.02 1541.5,1351.53 1545.34,1347.42 1549.19,1342.69 1553.03,1337.33 1556.88,1331.37 1560.72,1324.8 
  1564.57,1317.62 1568.41,1309.86 1572.26,1301.51 1576.1,1292.58 1579.95,1283.08 1583.79,1273.03 1587.63,1262.42 1591.48,1251.28 1595.32,1239.61 1599.17,1227.43 
  1603.01,1214.75 1606.86,1201.57 1610.7,1187.91 1614.55,1173.8 1618.39,1159.23 1622.24,1144.23 1626.08,1128.81 1629.93,1112.98 1633.77,1096.77 1637.62,1080.18 
  1641.46,1063.24 1645.31,1045.96 1649.15,1028.36 1653,1010.45 1656.84,992.26 1660.69,973.802 1664.53,955.092 1668.38,936.152 1672.22,917.001 1676.07,897.657 
  1679.91,878.14 1683.76,858.47 1687.6,838.667 1691.45,818.749 1695.29,798.736 1699.13,778.649 1702.98,758.508 1706.82,738.333 1710.67,718.145 1714.51,697.963 
  1718.36,677.808 1722.2,657.699 1726.05,637.657 1729.89,617.702 1733.74,597.853 1737.58,578.132 1741.43,558.557 1745.27,539.148 1749.12,519.925 1752.96,500.906 
  1756.81,482.11 1760.65,463.557 1764.5,445.265 1768.34,427.253 1772.19,409.537 1776.03,392.137 1779.88,375.069 1783.72,358.351 1787.57,341.998 1791.41,326.029 
  1795.26,310.458 1799.1,295.301 1802.95,280.573 1806.79,266.288 1810.63,252.463 1814.48,239.109 1818.32,226.241 1822.17,213.872 1826.01,202.013 1829.86,190.676 
  1833.7,179.872 1837.55,169.613 1841.39,159.91 1845.24,150.772 1849.08,142.207 1852.93,134.225 1856.77,126.832 1860.62,120.037 1864.46,113.846 1868.31,108.265 
  1872.15,103.302 1876,98.9606 1879.84,95.2442 1883.69,92.1569 1887.53,89.7016 1891.38,87.8803 1895.22,86.6947 1899.07,86.1472 1902.91,86.238 1906.76,86.9667 
  1910.6,88.3326 1914.45,90.3341 1918.29,92.9693 1922.13,96.236 1925.98,100.131 1929.82,104.65 1933.67,109.788 1937.51,115.541 1941.36,121.903 1945.2,128.867 
  1949.05,136.427 1952.89,144.575 1956.74,153.303 1960.58,162.601 1964.43,172.462 1968.27,182.875 1972.12,193.829 1975.96,205.314 1979.81,217.317 1983.65,229.827 
  1987.5,242.833 1991.34,256.32 1995.19,270.275 1999.03,284.684 2002.88,299.532 2006.72,314.805 2010.57,330.488 2014.41,346.565 2018.26,363.019 2022.1,379.835 
  2025.95,396.995 2029.79,414.482 2033.64,432.28 2037.48,450.369 2041.32,468.733 2045.17,487.352 2049.01,506.207 2052.86,525.281 2056.7,544.555 2060.55,564.008 
  2064.39,583.621 2068.24,603.375 2072.08,623.25 2075.93,643.226 2079.77,663.284 2083.62,683.402 2087.46,703.561 2091.31,723.741 2095.15,743.922 2099,764.083 
  2102.84,784.205 2106.69,804.266 2110.53,824.248 2114.38,844.13 2118.22,863.892 2122.07,883.515 2125.91,902.978 2129.76,922.263 2133.6,941.35 2137.45,960.22 
  2141.29,978.854 2145.14,997.233 2148.98,1015.34 2152.82,1033.16 2156.67,1050.66 2160.51,1067.84 2164.36,1084.68 2168.2,1101.16 2172.05,1117.26 2175.89,1132.96 
  2179.74,1148.26 2183.58,1163.14 2187.43,1177.57 2191.27,1191.56 2195.12,1205.07 2198.96,1218.11 2202.81,1230.65 2206.65,1242.69 2210.5,1254.2 2214.34,1265.19 
  2218.19,1275.64 2222.03,1285.53 2225.88,1294.86 2229.72,1303.63 2233.57,1311.81 2237.41,1319.41 2241.26,1326.41 2245.1,1332.81 2248.95,1338.6 2252.79,1343.78 
  2256.64,1348.34 2260.48,1352.27 2264.32,1355.58 2268.17,1358.25 2272.01,1360.29 2275.86,1361.7 2279.7,1362.47 2283.55,1362.6 2287.39,1362.09 2291.24,1360.94 
  2295.08,1359.16 
  "/>
<polyline clip-path="url(#clip242)" style="stroke:#e26f46; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" points="
  372.668,937.335 376.513,937.226 380.358,936.907 384.203,936.375 388.048,935.631 391.893,934.677 395.737,933.511 399.582,932.137 403.427,930.555 407.272,928.768 
  411.117,926.775 414.962,924.581 418.806,922.187 422.651,919.595 426.496,916.808 430.341,913.828 434.186,910.659 438.031,907.303 441.875,903.766 445.72,900.05 
  449.565,896.159 453.41,892.097 457.255,887.866 461.1,883.47 464.944,878.916 468.789,874.207 472.634,869.348 476.479,864.343 480.324,859.198 484.168,853.917 
  488.013,848.507 491.858,842.973 495.703,837.319 499.548,831.553 503.393,825.68 507.237,819.706 511.082,813.636 514.927,807.477 518.772,801.235 522.617,794.917 
  526.462,788.527 530.306,782.074 534.151,775.563 537.996,769.001 541.841,762.394 545.686,755.75 549.531,749.074 553.375,742.373 557.22,735.654 561.065,728.924 
  564.91,722.19 568.755,715.457 572.6,708.734 576.444,702.026 580.289,695.34 584.134,688.683 587.979,682.062 591.824,675.482 595.669,668.953 599.513,662.478 
  603.358,656.064 607.203,649.719 611.048,643.448 614.893,637.259 618.738,631.157 622.582,625.148 626.427,619.238 630.272,613.432 634.117,607.738 637.962,602.161 
  641.807,596.707 645.651,591.38 649.496,586.186 653.341,581.129 657.186,576.217 661.031,571.453 664.876,566.842 668.72,562.388 672.565,558.097 676.41,553.971 
  680.255,550.016 684.1,546.236 687.944,542.635 691.789,539.214 695.634,535.979 699.479,532.932 703.324,530.077 707.169,527.417 711.013,524.953 714.858,522.689 
  718.703,520.626 722.548,518.768 726.393,517.115 730.238,515.67 734.082,514.433 737.927,513.406 741.772,512.59 745.617,511.986 749.462,511.595 753.307,511.417 
  757.151,511.451 760.996,511.699 764.841,512.159 768.686,512.831 772.531,513.716 776.376,514.811 780.22,516.116 784.065,517.629 787.91,519.348 791.755,521.273 
  795.6,523.402 799.445,525.731 803.289,528.259 807.134,530.983 810.979,533.901 814.824,537.009 818.669,540.305 822.514,543.785 826.358,547.445 830.203,551.283 
  834.048,555.294 837.893,559.474 841.738,563.819 845.583,568.324 849.427,572.986 853.272,577.8 857.117,582.76 860.962,587.861 864.807,593.099 868.651,598.469 
  872.496,603.965 876.341,609.581 880.186,615.312 884.031,621.151 887.876,627.094 891.72,633.135 895.565,639.268 899.41,645.485 903.255,651.781 907.1,658.149 
  910.945,664.585 914.789,671.08 918.634,677.629 922.479,684.224 926.324,690.859 930.169,697.527 934.014,704.223 937.858,710.94 941.703,717.67 945.548,724.406 
  949.393,731.142 953.238,737.871 957.083,744.588 960.927,751.284 964.772,757.953 968.617,764.589 972.462,771.184 976.307,777.732 980.152,784.228 983.996,790.663 
  987.841,797.033 991.686,803.329 995.531,809.546 999.376,815.679 1003.22,821.72 1007.07,827.664 1010.91,833.505 1014.76,839.236 1018.6,844.852 1022.44,850.348 
  1026.29,855.719 1030.13,860.958 1033.98,866.06 1037.82,871.02 1041.67,875.834 1045.51,880.496 1049.36,885.003 1053.2,889.349 1057.05,893.529 1060.89,897.54 
  1064.74,901.378 1068.58,905.04 1072.43,908.521 1076.27,911.817 1080.12,914.926 1083.96,917.844 1087.81,920.569 1091.65,923.097 1095.5,925.428 1099.34,927.556 
  1103.19,929.482 1107.03,931.202 1110.88,932.715 1114.72,934.021 1118.57,935.116 1122.41,936.001 1126.26,936.674 1130.1,937.134 1133.94,937.382 1137.79,937.417 
  1141.63,937.239 1145.48,936.847 1149.32,936.243 1153.17,935.428 1157.01,934.401 1160.86,933.164 1164.7,931.718 1168.55,930.065 1172.39,928.206 1176.24,926.144 
  1180.08,923.879 1183.93,921.415 1187.77,918.753 1191.62,915.898 1195.46,912.851 1199.31,909.615 1203.15,906.194 1207,902.591 1210.84,898.809 1214.69,894.854 
  1218.53,890.728 1222.38,886.435 1226.22,881.98 1230.07,877.368 1233.91,872.602 1237.76,867.689 1241.6,862.632 1245.44,857.436 1249.29,852.107 1253.13,846.651 
  1256.98,841.072 1260.82,835.378 1264.67,829.571 1268.51,823.659 1272.36,817.648 1276.2,811.544 1280.05,805.352 1283.89,799.081 1287.74,792.733 1291.58,786.317 
  1295.43,779.839 1299.27,773.306 1303.12,766.725 1306.96,760.101 1310.81,753.44 1314.65,746.751 1318.5,740.039 1322.34,733.312 1326.19,726.576 1330.03,719.838 
  1333.88,713.104 1337.72,706.381 1341.57,699.676 1345.41,692.996 1349.26,686.348 1353.1,679.738 1356.94,673.171 1360.79,666.656 1364.63,660.199 1368.48,653.807 
  1372.32,647.486 1376.17,641.24 1380.01,635.077 1383.86,629.004 1387.7,623.027 1391.55,617.151 1395.39,611.383 1399.24,605.727 1403.08,600.19 1406.93,594.777 
  1410.77,589.493 1414.62,584.346 1418.46,579.338 1422.31,574.474 1426.15,569.761 1430,565.202 1433.84,560.803 1437.69,556.568 1441.53,552.5 1445.38,548.604 
  1449.22,544.884 1453.07,541.343 1456.91,537.986 1460.76,534.816 1464.6,531.835 1468.44,529.046 1472.29,526.452 1476.13,524.057 1479.98,521.862 1483.82,519.871 
  1487.67,518.083 1491.51,516.501 1495.36,515.128 1499.2,513.964 1503.05,513.01 1506.89,512.269 1510.74,511.739 1514.58,511.421 1518.43,511.317 1522.27,511.426 
  1526.12,511.749 1529.96,512.283 1533.81,513.03 1537.65,513.989 1541.5,515.158 1545.34,516.536 1549.19,518.122 1553.03,519.915 1556.88,521.912 1560.72,524.111 
  1564.57,526.511 1568.41,529.109 1572.26,531.902 1576.1,534.888 1579.95,538.064 1583.79,541.426 1587.63,544.971 1591.48,548.695 1595.32,552.595 1599.17,556.667 
  1603.01,560.907 1606.86,565.311 1610.7,569.873 1614.55,574.59 1618.39,579.457 1622.24,584.469 1626.08,589.622 1629.93,594.908 1633.77,600.325 1637.62,605.865 
  1641.46,611.524 1645.31,617.297 1649.15,623.176 1653,629.156 1656.84,635.232 1660.69,641.397 1664.53,647.645 1668.38,653.97 1672.22,660.366 1676.07,666.825 
  1679.91,673.342 1683.76,679.909 1687.6,686.523 1691.45,693.174 1695.29,699.856 1699.13,706.562 1702.98,713.285 1706.82,720.021 1710.67,726.761 1714.51,733.499 
  1718.36,740.227 1722.2,746.94 1726.05,753.629 1729.89,760.29 1733.74,766.916 1737.58,773.498 1741.43,780.032 1745.27,786.509 1749.12,792.924 1752.96,799.272 
  1756.81,805.544 1760.65,811.736 1764.5,817.839 1768.34,823.849 1772.19,829.76 1776.03,835.565 1779.88,841.26 1783.72,846.838 1787.57,852.292 1791.41,857.619 
  1795.26,862.812 1799.1,867.868 1802.95,872.78 1806.79,877.544 1810.63,882.154 1814.48,886.605 1818.32,890.896 1822.17,895.02 1826.01,898.973 1829.86,902.752 
  1833.7,906.351 1837.55,909.769 1841.39,913.002 1845.24,916.046 1849.08,918.898 1852.93,921.556 1856.77,924.016 1860.62,926.276 1864.46,928.335 1868.31,930.19 
  1872.15,931.839 1876,933.281 1879.84,934.513 1883.69,935.535 1887.53,936.346 1891.38,936.946 1895.22,937.332 1899.07,937.505 1902.91,937.465 1906.76,937.213 
  1910.6,936.747 1914.45,936.069 1918.29,935.179 1922.13,934.078 1925.98,932.767 1929.82,931.248 1933.67,929.523 1937.51,927.591 1941.36,925.457 1945.2,923.122 
  1949.05,920.587 1952.89,917.856 1956.74,914.932 1960.58,911.817 1964.43,908.515 1968.27,905.029 1972.12,901.362 1975.96,897.517 1979.81,893.5 1983.65,889.313 
  1987.5,884.962 1991.34,880.45 1995.19,875.781 1999.03,870.961 2002.88,865.995 2006.72,860.887 2010.57,855.642 2014.41,850.267 2018.26,844.765 2022.1,839.142 
  2025.95,833.405 2029.79,827.559 2033.64,821.61 2037.48,815.563 2041.32,809.425 2045.17,803.201 2049.01,796.899 2052.86,790.525 2056.7,784.085 2060.55,777.585 
  2064.39,771.031 2068.24,764.43 2072.08,757.79 2075.93,751.116 2079.77,744.417 2083.62,737.696 2087.46,730.961 2091.31,724.22 2095.15,717.48 2099,710.747 
  2102.84,704.028 2106.69,697.328 2110.53,690.655 2114.38,684.016 2118.22,677.418 2122.07,670.867 2125.91,664.369 2129.76,657.931 2133.6,651.559 2137.45,645.261 
  2141.29,639.041 2145.14,632.908 2148.98,626.865 2152.82,620.92 2156.67,615.078 2160.51,609.346 2164.36,603.729 2168.2,598.233 2172.05,592.862 2175.89,587.623 
  2179.74,582.521 2183.58,577.561 2187.43,572.748 2191.27,568.087 2195.12,563.582 2198.96,559.237 2202.81,555.058 2206.65,551.048 2210.5,547.212 2214.34,543.553 
  2218.19,540.074 2222.03,536.78 2225.88,533.674 2229.72,530.759 2233.57,528.037 2237.41,525.512 2241.26,523.185 2245.1,521.06 2248.95,519.138 2252.79,517.422 
  2256.64,515.913 2260.48,514.612 2264.32,513.521 2268.17,512.64 2272.01,511.972 2275.86,511.517 2279.7,511.274 2283.55,511.244 2287.39,511.428 2291.24,511.824 
  2295.08,512.434 
  "/>
<polyline clip-path="url(#clip242)" style="stroke:#3da44d; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" points="
  372.668,933.728 376.513,933.617 380.358,933.35 384.203,932.931 388.048,932.361 391.893,931.644 395.737,930.78 399.582,929.769 403.427,928.609 407.272,927.298 
  411.117,925.83 414.962,924.199 418.806,922.398 422.651,920.417 426.496,918.245 430.341,915.868 434.186,913.273 438.031,910.444 441.875,907.367 445.72,904.026 
  449.565,900.407 453.41,896.497 457.255,892.288 461.1,887.775 464.944,882.957 468.789,877.84 472.634,872.436 476.479,866.762 480.324,860.845 484.168,854.714 
  488.013,848.402 491.858,841.948 495.703,835.39 499.548,828.764 503.393,822.104 507.237,815.444 511.082,808.812 514.927,802.231 518.772,795.723 522.617,789.308 
  526.462,783 530.306,776.811 534.151,770.749 537.996,764.816 541.841,759.009 545.686,753.315 549.531,747.715 553.375,742.182 557.22,736.682 561.065,731.175 
  564.91,725.616 568.755,719.957 572.6,714.154 576.444,708.163 580.289,701.949 584.134,695.489 587.979,688.775 591.824,681.816 595.669,674.638 599.513,667.284 
  603.358,659.813 607.203,652.29 611.048,644.784 614.893,637.363 618.738,630.087 622.582,623.005 626.427,616.155 630.272,609.564 634.117,603.247 637.962,597.209 
  641.807,591.449 645.651,585.961 649.496,580.736 653.341,575.761 657.186,571.026 661.031,566.518 664.876,562.229 668.72,558.149 672.565,554.271 676.41,550.59 
  680.255,547.104 684.1,543.808 687.944,540.703 691.789,537.789 695.634,535.064 699.479,532.53 703.324,530.188 707.169,528.037 711.013,526.078 714.858,524.309 
  718.703,522.73 722.548,521.339 726.393,520.133 730.238,519.109 734.082,518.264 737.927,517.593 741.772,517.092 745.617,516.756 749.462,516.579 753.307,516.556 
  757.151,516.683 760.996,516.954 764.841,517.363 768.686,517.908 772.531,518.584 776.376,519.388 780.22,520.319 784.065,521.374 787.91,522.556 791.755,523.865 
  795.6,525.304 799.445,526.879 803.289,528.596 807.134,530.465 810.979,532.495 814.824,534.699 818.669,537.093 822.514,539.692 826.358,542.514 830.203,545.579 
  834.048,548.907 837.893,552.519 841.738,556.436 845.583,560.675 849.427,565.252 853.272,570.18 857.117,575.465 860.962,581.107 864.807,587.097 868.651,593.419 
  872.496,600.046 876.341,606.942 880.186,614.06 884.031,621.349 887.876,628.75 891.72,636.201 895.565,643.642 899.41,651.013 903.255,658.262 907.1,665.344 
  910.945,672.222 914.789,678.871 918.634,685.277 922.479,691.436 926.324,697.358 930.169,703.064 934.014,708.584 937.858,713.961 941.703,719.244 945.548,724.491 
  949.393,729.764 953.238,735.126 957.083,740.639 960.927,746.356 964.772,752.323 968.617,758.566 972.462,765.094 976.307,771.892 980.152,778.926 983.996,786.14 
  987.841,793.465 991.686,800.825 995.531,808.143 999.376,815.35 1003.22,822.387 1007.07,829.21 1010.91,835.789 1014.76,842.106 1018.6,848.154 1022.44,853.935 
  1026.29,859.457 1030.13,864.732 1033.98,869.771 1037.82,874.587 1041.67,879.194 1045.51,883.601 1049.36,887.816 1053.2,891.845 1057.05,895.692 1060.89,899.36 
  1064.74,902.849 1068.58,906.158 1072.43,909.286 1076.27,912.23 1080.12,914.987 1083.96,917.556 1087.81,919.934 1091.65,922.119 1095.5,924.112 1099.34,925.91 
  1103.19,927.516 1107.03,928.931 1110.88,930.156 1114.72,931.194 1118.57,932.048 1122.41,932.723 1126.26,933.222 1130.1,933.549 1133.94,933.709 1137.79,933.707 
  1141.63,933.546 1145.48,933.23 1149.32,932.763 1153.17,932.146 1157.01,931.382 1160.86,930.471 1164.7,929.413 1168.55,928.204 1172.39,926.843 1176.24,925.324 
  1180.08,923.64 1183.93,921.782 1187.77,919.741 1191.62,917.504 1195.46,915.059 1199.31,912.39 1203.15,909.483 1207,906.322 1210.84,902.893 1214.69,899.181 
  1218.53,895.176 1222.38,890.87 1226.22,886.258 1230.07,881.344 1233.91,876.133 1237.76,870.64 1241.6,864.886 1245.44,858.898 1249.29,852.705 1253.13,846.345 
  1256.98,839.854 1260.82,833.27 1264.67,826.631 1268.51,819.969 1272.36,813.316 1276.2,806.698 1280.05,800.139 1283.89,793.659 1287.74,787.277 1291.58,781.006 
  1295.43,774.858 1299.27,768.838 1303.12,762.946 1306.96,757.176 1310.81,751.514 1314.65,745.939 1318.5,740.42 1322.34,734.921 1326.19,729.402 1330.03,723.815 
  1333.88,718.115 1337.72,712.254 1341.57,706.194 1345.41,699.903 1349.26,693.362 1353.1,686.568 1356.94,679.535 1360.79,672.296 1364.63,664.899 1368.48,657.404 
  1372.32,649.88 1376.17,642.396 1380.01,635.016 1383.86,627.798 1387.7,620.788 1391.55,614.02 1395.39,607.517 1399.24,601.289 1403.08,595.341 1406.93,589.67 
  1410.77,584.268 1414.62,579.125 1418.46,574.228 1422.31,569.568 1426.15,565.132 1430,560.911 1433.84,556.896 1437.69,553.083 1441.53,549.466 1445.38,546.041 
  1449.22,542.808 1453.07,539.764 1456.91,536.911 1460.76,534.248 1464.6,531.777 1468.44,529.496 1472.29,527.408 1476.13,525.51 1479.98,523.803 1483.82,522.285 
  1487.67,520.954 1491.51,519.807 1495.36,518.841 1499.2,518.053 1503.05,517.437 1506.89,516.989 1510.74,516.705 1514.58,516.578 1518.43,516.604 1522.27,516.777 
  1526.12,517.093 1529.96,517.546 1533.81,518.133 1537.65,518.851 1541.5,519.695 1545.34,520.666 1549.19,521.762 1553.03,522.984 1556.88,524.335 1560.72,525.817 
  1564.57,527.438 1568.41,529.202 1572.26,531.122 1576.1,533.207 1579.95,535.47 1583.79,537.928 1587.63,540.597 1591.48,543.495 1595.32,546.643 1599.17,550.06 
  1603.01,553.768 1606.86,557.786 1610.7,562.132 1614.55,566.82 1618.39,571.862 1622.24,577.261 1626.08,583.015 1629.93,589.113 1633.77,595.535 1637.62,602.251 
  1641.46,609.222 1645.31,616.399 1649.15,623.728 1653,631.149 1656.84,638.601 1660.69,646.023 1664.53,653.359 1668.38,660.557 1672.22,667.575 1676.07,674.381 
  1679.91,680.952 1683.76,687.278 1687.6,693.359 1691.45,699.21 1695.29,704.853 1699.13,710.324 1702.98,715.667 1706.82,720.935 1710.67,726.186 1714.51,731.483 
  1718.36,736.889 1722.2,742.464 1726.05,748.26 1729.89,754.314 1733.74,760.65 1737.58,767.268 1741.43,774.147 1745.27,781.244 1749.12,788.5 1752.96,795.842 
  1756.81,803.194 1760.65,810.48 1764.5,817.636 1768.34,824.607 1772.19,831.352 1776.03,837.846 1779.88,844.076 1783.72,850.036 1787.57,855.732 1791.41,861.173 
  1795.26,866.369 1799.1,871.334 1802.95,876.081 1806.79,880.622 1810.63,884.965 1814.48,889.118 1818.32,893.088 1822.17,896.876 1826.01,900.486 1829.86,903.916 
  1833.7,907.165 1837.55,910.233 1841.39,913.115 1845.24,915.811 1849.08,918.318 1852.93,920.633 1856.77,922.755 1860.62,924.685 1864.46,926.421 1868.31,927.964 
  1872.15,929.317 1876,930.481 1879.84,931.459 1883.69,932.255 1887.53,932.872 1891.38,933.316 1895.22,933.589 1899.07,933.697 1902.91,933.643 1906.76,933.432 
  1910.6,933.067 1914.45,932.552 1918.29,931.888 1922.13,931.077 1925.98,930.119 1929.82,929.013 1933.67,927.756 1937.51,926.345 1941.36,924.774 1945.2,923.035 
  1949.05,921.119 1952.89,919.017 1956.74,916.715 1960.58,914.199 1964.43,911.456 1968.27,908.469 1972.12,905.223 1975.96,901.705 1979.81,897.901 1983.65,893.801 
  1987.5,889.398 1991.34,884.69 1995.19,879.68 1999.03,874.379 2002.88,868.802 2006.72,862.971 2010.57,856.916 2014.41,850.668 2018.26,844.264 2022.1,837.742 
  2025.95,831.139 2029.79,824.492 2033.64,817.832 2037.48,811.189 2041.32,804.59 2045.17,798.057 2049.01,791.609 2052.86,785.262 2056.7,779.031 2060.55,772.924 
  2064.39,766.946 2068.24,761.094 2072.08,755.361 2075.93,749.729 2079.77,744.175 2083.62,738.666 2087.46,733.165 2091.31,727.628 2095.15,722.009 2099,716.261 
  2102.84,710.341 2106.69,704.21 2110.53,697.841 2114.38,691.219 2118.22,684.347 2122.07,677.246 2125.91,669.952 2129.76,662.519 2133.6,655.009 2137.45,647.493 
  2141.29,640.038 2145.14,632.706 2148.98,625.552 2152.82,618.618 2156.67,611.933 2160.51,605.518 2164.36,599.38 2168.2,593.522 2172.05,587.938 2175.89,582.62 
  2179.74,577.557 2183.58,572.738 2187.43,568.15 2191.27,563.784 2195.12,559.63 2198.96,555.681 2202.81,551.931 2206.65,548.375 2210.5,545.012 2214.34,541.84 
  2218.19,538.857 2222.03,536.065 2225.88,533.464 2229.72,531.053 2233.57,528.834 2237.41,526.806 2241.26,524.969 2245.1,523.323 2248.95,521.864 2252.79,520.592 
  2256.64,519.503 2260.48,518.593 2264.32,517.86 2268.17,517.298 2272.01,516.902 2275.86,516.667 2279.7,516.589 2283.55,516.662 2287.39,516.88 2291.24,517.24 
  2295.08,517.735 
  "/>
<path clip-path="url(#clip240)" d="
M1468.95 299.749 L2284.83 299.749 L2284.83 92.3886 L1468.95 92.3886  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<polyline clip-path="url(#clip240)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1468.95,299.749 2284.83,299.749 2284.83,92.3886 1468.95,92.3886 1468.95,299.749 
  "/>
<polyline clip-path="url(#clip240)" style="stroke:#009af9; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1491.59,144.229 1627.44,144.229 
  "/>
<path clip-path="url(#clip240)" d="M1668.99 139.518 L1668.99 125.49 L1673.25 125.49 L1673.25 161.509 L1668.99 161.509 L1668.99 157.62 Q1667.65 159.935 1665.59 161.069 Q1663.55 162.18 1660.68 162.18 Q1655.99 162.18 1653.02 158.43 Q1650.08 154.68 1650.08 148.569 Q1650.08 142.458 1653.02 138.708 Q1655.99 134.958 1660.68 134.958 Q1663.55 134.958 1665.59 136.092 Q1667.65 137.203 1668.99 139.518 M1654.48 148.569 Q1654.48 153.268 1656.4 155.953 Q1658.35 158.615 1661.73 158.615 Q1665.11 158.615 1667.05 155.953 Q1668.99 153.268 1668.99 148.569 Q1668.99 143.87 1667.05 141.208 Q1665.11 138.523 1661.73 138.523 Q1658.35 138.523 1656.4 141.208 Q1654.48 143.87 1654.48 148.569 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M1704.2 147.481 L1704.2 149.564 L1684.62 149.564 Q1684.9 153.962 1687.26 156.277 Q1689.64 158.569 1693.88 158.569 Q1696.33 158.569 1698.62 157.967 Q1700.94 157.365 1703.21 156.161 L1703.21 160.189 Q1700.92 161.161 1698.51 161.671 Q1696.1 162.18 1693.62 162.18 Q1687.42 162.18 1683.79 158.569 Q1680.18 154.958 1680.18 148.8 Q1680.18 142.435 1683.6 138.708 Q1687.05 134.958 1692.88 134.958 Q1698.11 134.958 1701.15 138.337 Q1704.2 141.694 1704.2 147.481 M1699.94 146.231 Q1699.9 142.736 1697.98 140.652 Q1696.08 138.569 1692.93 138.569 Q1689.36 138.569 1687.21 140.583 Q1685.08 142.597 1684.76 146.254 L1699.94 146.231 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M1724.32 125.49 L1724.32 129.032 L1720.24 129.032 Q1717.95 129.032 1717.05 129.958 Q1716.17 130.884 1716.17 133.291 L1716.17 135.583 L1723.18 135.583 L1723.18 138.893 L1716.17 138.893 L1716.17 161.509 L1711.89 161.509 L1711.89 138.893 L1707.81 138.893 L1707.81 135.583 L1711.89 135.583 L1711.89 133.777 Q1711.89 129.449 1713.9 127.481 Q1715.92 125.49 1720.29 125.49 L1724.32 125.49 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M1739.67 148.476 Q1734.5 148.476 1732.51 149.657 Q1730.52 150.837 1730.52 153.685 Q1730.52 155.953 1732 157.296 Q1733.51 158.615 1736.08 158.615 Q1739.62 158.615 1741.75 156.115 Q1743.9 153.592 1743.9 149.425 L1743.9 148.476 L1739.67 148.476 M1748.16 146.717 L1748.16 161.509 L1743.9 161.509 L1743.9 157.573 Q1742.44 159.935 1740.27 161.069 Q1738.09 162.18 1734.94 162.18 Q1730.96 162.18 1728.6 159.958 Q1726.26 157.712 1726.26 153.962 Q1726.26 149.587 1729.18 147.365 Q1732.12 145.143 1737.93 145.143 L1743.9 145.143 L1743.9 144.726 Q1743.9 141.787 1741.96 140.189 Q1740.04 138.569 1736.54 138.569 Q1734.32 138.569 1732.21 139.101 Q1730.11 139.634 1728.16 140.699 L1728.16 136.763 Q1730.5 135.861 1732.7 135.421 Q1734.9 134.958 1736.98 134.958 Q1742.61 134.958 1745.38 137.874 Q1748.16 140.791 1748.16 146.717 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M1756.49 151.277 L1756.49 135.583 L1760.75 135.583 L1760.75 151.115 Q1760.75 154.796 1762.19 156.648 Q1763.62 158.476 1766.49 158.476 Q1769.94 158.476 1771.93 156.277 Q1773.95 154.078 1773.95 150.282 L1773.95 135.583 L1778.21 135.583 L1778.21 161.509 L1773.95 161.509 L1773.95 157.527 Q1772.4 159.888 1770.34 161.046 Q1768.3 162.18 1765.59 162.18 Q1761.12 162.18 1758.81 159.402 Q1756.49 156.624 1756.49 151.277 M1767.21 134.958 L1767.21 134.958 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M1786.98 125.49 L1791.24 125.49 L1791.24 161.509 L1786.98 161.509 L1786.98 125.49 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M1804.36 128.222 L1804.36 135.583 L1813.14 135.583 L1813.14 138.893 L1804.36 138.893 L1804.36 152.967 Q1804.36 156.138 1805.22 157.041 Q1806.1 157.944 1808.76 157.944 L1813.14 157.944 L1813.14 161.509 L1808.76 161.509 Q1803.83 161.509 1801.96 159.68 Q1800.08 157.828 1800.08 152.967 L1800.08 138.893 L1796.96 138.893 L1796.96 135.583 L1800.08 135.583 L1800.08 128.222 L1804.36 128.222 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M1818.92 126.949 L1838.79 126.949 L1838.79 130.884 L1823.6 130.884 L1823.6 141.069 L1837.3 141.069 L1837.3 145.004 L1823.6 145.004 L1823.6 161.509 L1818.92 161.509 L1818.92 126.949 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M1846.19 126.949 L1853.16 126.949 L1861.98 150.467 L1870.85 126.949 L1877.81 126.949 L1877.81 161.509 L1873.25 161.509 L1873.25 131.162 L1864.34 154.865 L1859.64 154.865 L1850.73 131.162 L1850.73 161.509 L1846.19 161.509 L1846.19 126.949 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M1886.56 126.949 L1891.26 126.949 L1891.26 147.944 Q1891.26 153.499 1893.28 155.953 Q1895.29 158.384 1899.8 158.384 Q1904.29 158.384 1906.31 155.953 Q1908.32 153.499 1908.32 147.944 L1908.32 126.949 L1913.02 126.949 L1913.02 148.523 Q1913.02 155.282 1909.66 158.731 Q1906.33 162.18 1899.8 162.18 Q1893.25 162.18 1889.9 158.731 Q1886.56 155.282 1886.56 148.523 L1886.56 126.949 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip240)" style="stroke:#e26f46; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1491.59,196.069 1627.44,196.069 
  "/>
<path clip-path="url(#clip240)" d="M1665.27 191.404 Q1664.55 190.988 1663.69 190.802 Q1662.86 190.594 1661.84 190.594 Q1658.23 190.594 1656.29 192.955 Q1654.36 195.293 1654.36 199.691 L1654.36 213.349 L1650.08 213.349 L1650.08 187.423 L1654.36 187.423 L1654.36 191.451 Q1655.71 189.089 1657.86 187.955 Q1660.01 186.798 1663.09 186.798 Q1663.53 186.798 1664.06 186.867 Q1664.6 186.914 1665.24 187.029 L1665.27 191.404 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M1690.87 199.321 L1690.87 201.404 L1671.29 201.404 Q1671.56 205.802 1673.93 208.117 Q1676.31 210.409 1680.55 210.409 Q1683 210.409 1685.29 209.807 Q1687.61 209.205 1689.87 208.001 L1689.87 212.029 Q1687.58 213.001 1685.17 213.511 Q1682.77 214.02 1680.29 214.02 Q1674.09 214.02 1670.45 210.409 Q1666.84 206.798 1666.84 200.64 Q1666.84 194.275 1670.27 190.548 Q1673.72 186.798 1679.55 186.798 Q1684.78 186.798 1687.81 190.177 Q1690.87 193.534 1690.87 199.321 M1686.61 198.071 Q1686.56 194.576 1684.64 192.492 Q1682.74 190.409 1679.6 190.409 Q1676.03 190.409 1673.88 192.423 Q1671.75 194.437 1671.43 198.094 L1686.61 198.071 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M1710.99 177.33 L1710.99 180.872 L1706.91 180.872 Q1704.62 180.872 1703.72 181.798 Q1702.84 182.724 1702.84 185.131 L1702.84 187.423 L1709.85 187.423 L1709.85 190.733 L1702.84 190.733 L1702.84 213.349 L1698.55 213.349 L1698.55 190.733 L1694.48 190.733 L1694.48 187.423 L1698.55 187.423 L1698.55 185.617 Q1698.55 181.289 1700.57 179.321 Q1702.58 177.33 1706.96 177.33 L1710.99 177.33 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M1736.73 199.321 L1736.73 201.404 L1717.14 201.404 Q1717.42 205.802 1719.78 208.117 Q1722.17 210.409 1726.4 210.409 Q1728.86 210.409 1731.15 209.807 Q1733.46 209.205 1735.73 208.001 L1735.73 212.029 Q1733.44 213.001 1731.03 213.511 Q1728.62 214.02 1726.15 214.02 Q1719.94 214.02 1716.31 210.409 Q1712.7 206.798 1712.7 200.64 Q1712.7 194.275 1716.12 190.548 Q1719.57 186.798 1725.41 186.798 Q1730.64 186.798 1733.67 190.177 Q1736.73 193.534 1736.73 199.321 M1732.47 198.071 Q1732.42 194.576 1730.5 192.492 Q1728.6 190.409 1725.45 190.409 Q1721.89 190.409 1719.73 192.423 Q1717.61 194.437 1717.28 198.094 L1732.47 198.071 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M1758.74 191.404 Q1758.02 190.988 1757.17 190.802 Q1756.33 190.594 1755.31 190.594 Q1751.7 190.594 1749.76 192.955 Q1747.84 195.293 1747.84 199.691 L1747.84 213.349 L1743.55 213.349 L1743.55 187.423 L1747.84 187.423 L1747.84 191.451 Q1749.18 189.089 1751.33 187.955 Q1753.48 186.798 1756.56 186.798 Q1757 186.798 1757.54 186.867 Q1758.07 186.914 1758.72 187.029 L1758.74 191.404 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M1784.34 199.321 L1784.34 201.404 L1764.76 201.404 Q1765.04 205.802 1767.4 208.117 Q1769.78 210.409 1774.02 210.409 Q1776.47 210.409 1778.76 209.807 Q1781.08 209.205 1783.35 208.001 L1783.35 212.029 Q1781.05 213.001 1778.65 213.511 Q1776.24 214.02 1773.76 214.02 Q1767.56 214.02 1763.92 210.409 Q1760.31 206.798 1760.31 200.64 Q1760.31 194.275 1763.74 190.548 Q1767.19 186.798 1773.02 186.798 Q1778.25 186.798 1781.29 190.177 Q1784.34 193.534 1784.34 199.321 M1780.08 198.071 Q1780.04 194.576 1778.11 192.492 Q1776.22 190.409 1773.07 190.409 Q1769.5 190.409 1767.35 192.423 Q1765.22 194.437 1764.9 198.094 L1780.08 198.071 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M1812.88 197.701 L1812.88 213.349 L1808.62 213.349 L1808.62 197.839 Q1808.62 194.159 1807.19 192.33 Q1805.75 190.502 1802.88 190.502 Q1799.43 190.502 1797.44 192.701 Q1795.45 194.9 1795.45 198.696 L1795.45 213.349 L1791.17 213.349 L1791.17 187.423 L1795.45 187.423 L1795.45 191.451 Q1796.98 189.113 1799.04 187.955 Q1801.12 186.798 1803.83 186.798 Q1808.3 186.798 1810.59 189.576 Q1812.88 192.33 1812.88 197.701 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M1840.04 188.418 L1840.04 192.4 Q1838.23 191.404 1836.4 190.918 Q1834.6 190.409 1832.74 190.409 Q1828.6 190.409 1826.31 193.048 Q1824.02 195.664 1824.02 200.409 Q1824.02 205.154 1826.31 207.793 Q1828.6 210.409 1832.74 210.409 Q1834.6 210.409 1836.4 209.923 Q1838.23 209.413 1840.04 208.418 L1840.04 212.353 Q1838.25 213.187 1836.33 213.603 Q1834.43 214.02 1832.28 214.02 Q1826.42 214.02 1822.98 210.339 Q1819.53 206.659 1819.53 200.409 Q1819.53 194.066 1823 190.432 Q1826.49 186.798 1832.56 186.798 Q1834.53 186.798 1836.4 187.215 Q1838.28 187.608 1840.04 188.418 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M1869.62 199.321 L1869.62 201.404 L1850.04 201.404 Q1850.31 205.802 1852.67 208.117 Q1855.06 210.409 1859.29 210.409 Q1861.75 210.409 1864.04 209.807 Q1866.35 209.205 1868.62 208.001 L1868.62 212.029 Q1866.33 213.001 1863.92 213.511 Q1861.52 214.02 1859.04 214.02 Q1852.84 214.02 1849.2 210.409 Q1845.59 206.798 1845.59 200.64 Q1845.59 194.275 1849.02 190.548 Q1852.47 186.798 1858.3 186.798 Q1863.53 186.798 1866.56 190.177 Q1869.62 193.534 1869.62 199.321 M1865.36 198.071 Q1865.31 194.576 1863.39 192.492 Q1861.49 190.409 1858.35 190.409 Q1854.78 190.409 1852.63 192.423 Q1850.5 194.437 1850.17 198.094 L1865.36 198.071 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M1876.79 178.789 L1896.66 178.789 L1896.66 182.724 L1881.47 182.724 L1881.47 192.909 L1895.17 192.909 L1895.17 196.844 L1881.47 196.844 L1881.47 213.349 L1876.79 213.349 L1876.79 178.789 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M1904.06 178.789 L1911.03 178.789 L1919.85 202.307 L1928.72 178.789 L1935.68 178.789 L1935.68 213.349 L1931.12 213.349 L1931.12 183.002 L1922.21 206.705 L1917.51 206.705 L1908.6 183.002 L1908.6 213.349 L1904.06 213.349 L1904.06 178.789 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M1944.43 178.789 L1949.13 178.789 L1949.13 199.784 Q1949.13 205.339 1951.15 207.793 Q1953.16 210.224 1957.67 210.224 Q1962.16 210.224 1964.18 207.793 Q1966.19 205.339 1966.19 199.784 L1966.19 178.789 L1970.89 178.789 L1970.89 200.363 Q1970.89 207.122 1967.53 210.571 Q1964.2 214.02 1957.67 214.02 Q1951.12 214.02 1947.77 210.571 Q1944.43 207.122 1944.43 200.363 L1944.43 178.789 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip240)" style="stroke:#3da44d; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1491.59,247.909 1627.44,247.909 
  "/>
<path clip-path="url(#clip240)" d="M1677.95 233.291 L1677.95 238.221 Q1675.59 236.022 1672.91 234.934 Q1670.24 233.846 1667.24 233.846 Q1661.31 233.846 1658.16 237.48 Q1655.01 241.092 1655.01 247.943 Q1655.01 254.772 1658.16 258.406 Q1661.31 262.017 1667.24 262.017 Q1670.24 262.017 1672.91 260.929 Q1675.59 259.841 1677.95 257.642 L1677.95 262.527 Q1675.5 264.193 1672.74 265.027 Q1670.01 265.86 1666.96 265.86 Q1659.11 265.86 1654.6 261.068 Q1650.08 256.253 1650.08 247.943 Q1650.08 239.61 1654.6 234.818 Q1659.11 230.004 1666.96 230.004 Q1670.06 230.004 1672.79 230.837 Q1675.55 231.647 1677.95 233.291 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M1705.89 231.763 L1705.89 236.323 Q1703.23 235.05 1700.87 234.425 Q1698.51 233.8 1696.31 233.8 Q1692.49 233.8 1690.41 235.281 Q1688.35 236.763 1688.35 239.494 Q1688.35 241.786 1689.71 242.967 Q1691.1 244.124 1694.94 244.842 L1697.77 245.42 Q1703 246.416 1705.48 248.939 Q1707.98 251.439 1707.98 255.652 Q1707.98 260.675 1704.6 263.267 Q1701.24 265.86 1694.74 265.86 Q1692.28 265.86 1689.5 265.304 Q1686.75 264.749 1683.79 263.661 L1683.79 258.846 Q1686.63 260.443 1689.36 261.253 Q1692.1 262.064 1694.74 262.064 Q1698.74 262.064 1700.92 260.49 Q1703.09 258.916 1703.09 255.999 Q1703.09 253.453 1701.52 252.017 Q1699.97 250.582 1696.4 249.865 L1693.55 249.309 Q1688.32 248.267 1685.99 246.045 Q1683.65 243.823 1683.65 239.865 Q1683.65 235.281 1686.86 232.642 Q1690.11 230.004 1695.78 230.004 Q1698.21 230.004 1700.73 230.443 Q1703.25 230.883 1705.89 231.763 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M1712.93 250.304 L1725.41 250.304 L1725.41 254.101 L1712.93 254.101 L1712.93 250.304 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M1732.37 230.629 L1738.67 230.629 L1753.99 259.541 L1753.99 230.629 L1758.53 230.629 L1758.53 265.189 L1752.23 265.189 L1736.91 236.277 L1736.91 265.189 L1732.37 265.189 L1732.37 230.629 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M1789.83 251.161 L1789.83 253.244 L1770.24 253.244 Q1770.52 257.642 1772.88 259.957 Q1775.27 262.249 1779.5 262.249 Q1781.96 262.249 1784.25 261.647 Q1786.56 261.045 1788.83 259.841 L1788.83 263.869 Q1786.54 264.841 1784.13 265.351 Q1781.73 265.86 1779.25 265.86 Q1773.04 265.86 1769.41 262.249 Q1765.8 258.638 1765.8 252.48 Q1765.8 246.115 1769.23 242.388 Q1772.67 238.638 1778.51 238.638 Q1783.74 238.638 1786.77 242.017 Q1789.83 245.374 1789.83 251.161 M1785.57 249.911 Q1785.52 246.416 1783.6 244.332 Q1781.7 242.249 1778.55 242.249 Q1774.99 242.249 1772.84 244.263 Q1770.71 246.277 1770.38 249.934 L1785.57 249.911 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M1796.38 254.957 L1796.38 239.263 L1800.64 239.263 L1800.64 254.795 Q1800.64 258.476 1802.07 260.328 Q1803.51 262.156 1806.38 262.156 Q1809.83 262.156 1811.82 259.957 Q1813.83 257.758 1813.83 253.962 L1813.83 239.263 L1818.09 239.263 L1818.09 265.189 L1813.83 265.189 L1813.83 261.207 Q1812.28 263.568 1810.22 264.726 Q1808.18 265.86 1805.48 265.86 Q1801.01 265.86 1798.69 263.082 Q1796.38 260.304 1796.38 254.957 M1807.1 238.638 L1807.1 238.638 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M1841.89 243.244 Q1841.17 242.828 1840.31 242.642 Q1839.48 242.434 1838.46 242.434 Q1834.85 242.434 1832.91 244.795 Q1830.98 247.133 1830.98 251.531 L1830.98 265.189 L1826.7 265.189 L1826.7 239.263 L1830.98 239.263 L1830.98 243.291 Q1832.33 240.929 1834.48 239.795 Q1836.63 238.638 1839.71 238.638 Q1840.15 238.638 1840.68 238.707 Q1841.22 238.754 1841.86 238.869 L1841.89 243.244 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M1858.14 252.156 Q1852.98 252.156 1850.98 253.337 Q1848.99 254.517 1848.99 257.365 Q1848.99 259.633 1850.48 260.976 Q1851.98 262.295 1854.55 262.295 Q1858.09 262.295 1860.22 259.795 Q1862.37 257.272 1862.37 253.105 L1862.37 252.156 L1858.14 252.156 M1866.63 250.397 L1866.63 265.189 L1862.37 265.189 L1862.37 261.253 Q1860.91 263.615 1858.74 264.749 Q1856.56 265.86 1853.41 265.86 Q1849.43 265.86 1847.07 263.638 Q1844.73 261.392 1844.73 257.642 Q1844.73 253.267 1847.65 251.045 Q1850.59 248.823 1856.4 248.823 L1862.37 248.823 L1862.37 248.406 Q1862.37 245.467 1860.43 243.869 Q1858.51 242.249 1855.01 242.249 Q1852.79 242.249 1850.68 242.781 Q1848.58 243.314 1846.63 244.379 L1846.63 240.443 Q1848.97 239.541 1851.17 239.101 Q1853.37 238.638 1855.45 238.638 Q1861.08 238.638 1863.85 241.554 Q1866.63 244.471 1866.63 250.397 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M1875.41 229.17 L1879.66 229.17 L1879.66 265.189 L1875.41 265.189 L1875.41 229.17 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M1888.76 230.629 L1908.62 230.629 L1908.62 234.564 L1893.44 234.564 L1893.44 244.749 L1907.14 244.749 L1907.14 248.684 L1893.44 248.684 L1893.44 265.189 L1888.76 265.189 L1888.76 230.629 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M1916.03 230.629 L1923 230.629 L1931.82 254.147 L1940.68 230.629 L1947.65 230.629 L1947.65 265.189 L1943.09 265.189 L1943.09 234.842 L1934.18 258.545 L1929.48 258.545 L1920.57 234.842 L1920.57 265.189 L1916.03 265.189 L1916.03 230.629 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M1956.4 230.629 L1961.1 230.629 L1961.1 251.624 Q1961.1 257.179 1963.11 259.633 Q1965.13 262.064 1969.64 262.064 Q1974.13 262.064 1976.15 259.633 Q1978.16 257.179 1978.16 251.624 L1978.16 230.629 L1982.86 230.629 L1982.86 252.203 Q1982.86 258.962 1979.5 262.411 Q1976.17 265.86 1969.64 265.86 Q1963.09 265.86 1959.73 262.411 Q1956.4 258.962 1956.4 252.203 L1956.4 230.629 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M2016.75 229.217 Q2013.65 234.541 2012.14 239.749 Q2010.64 244.957 2010.64 250.304 Q2010.64 255.652 2012.14 260.906 Q2013.67 266.138 2016.75 271.439 L2013.04 271.439 Q2009.57 265.999 2007.84 260.744 Q2006.12 255.49 2006.12 250.304 Q2006.12 245.142 2007.84 239.911 Q2009.55 234.68 2013.04 229.217 L2016.75 229.217 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M2039.78 246.554 Q2043.14 247.272 2045.01 249.541 Q2046.91 251.809 2046.91 255.142 Q2046.91 260.258 2043.39 263.059 Q2039.87 265.86 2033.39 265.86 Q2031.21 265.86 2028.9 265.42 Q2026.61 265.003 2024.15 264.147 L2024.15 259.633 Q2026.1 260.767 2028.41 261.346 Q2030.73 261.925 2033.25 261.925 Q2037.65 261.925 2039.94 260.189 Q2042.26 258.453 2042.26 255.142 Q2042.26 252.087 2040.1 250.374 Q2037.97 248.638 2034.15 248.638 L2030.13 248.638 L2030.13 244.795 L2034.34 244.795 Q2037.79 244.795 2039.62 243.429 Q2041.45 242.041 2041.45 239.448 Q2041.45 236.786 2039.55 235.374 Q2037.67 233.939 2034.15 233.939 Q2032.23 233.939 2030.03 234.355 Q2027.84 234.772 2025.2 235.652 L2025.2 231.485 Q2027.86 230.744 2030.17 230.374 Q2032.51 230.004 2034.57 230.004 Q2039.9 230.004 2043 232.434 Q2046.1 234.842 2046.1 238.962 Q2046.1 241.832 2044.46 243.823 Q2042.81 245.791 2039.78 246.554 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M2065.77 233.707 Q2062.16 233.707 2060.34 237.272 Q2058.53 240.814 2058.53 247.943 Q2058.53 255.05 2060.34 258.615 Q2062.16 262.156 2065.77 262.156 Q2069.41 262.156 2071.21 258.615 Q2073.04 255.05 2073.04 247.943 Q2073.04 240.814 2071.21 237.272 Q2069.41 233.707 2065.77 233.707 M2065.77 230.004 Q2071.58 230.004 2074.64 234.61 Q2077.72 239.193 2077.72 247.943 Q2077.72 256.67 2074.64 261.277 Q2071.58 265.86 2065.77 265.86 Q2059.96 265.86 2056.89 261.277 Q2053.83 256.67 2053.83 247.943 Q2053.83 239.193 2056.89 234.61 Q2059.96 230.004 2065.77 230.004 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M2095.94 233.707 Q2092.33 233.707 2090.5 237.272 Q2088.69 240.814 2088.69 247.943 Q2088.69 255.05 2090.5 258.615 Q2092.33 262.156 2095.94 262.156 Q2099.57 262.156 2101.38 258.615 Q2103.21 255.05 2103.21 247.943 Q2103.21 240.814 2101.38 237.272 Q2099.57 233.707 2095.94 233.707 M2095.94 230.004 Q2101.75 230.004 2104.8 234.61 Q2107.88 239.193 2107.88 247.943 Q2107.88 256.67 2104.8 261.277 Q2101.75 265.86 2095.94 265.86 Q2090.13 265.86 2087.05 261.277 Q2083.99 256.67 2083.99 247.943 Q2083.99 239.193 2087.05 234.61 Q2090.13 230.004 2095.94 230.004 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M2152.74 251.161 L2152.74 253.244 L2133.16 253.244 Q2133.44 257.642 2135.8 259.957 Q2138.18 262.249 2142.42 262.249 Q2144.87 262.249 2147.16 261.647 Q2149.48 261.045 2151.75 259.841 L2151.75 263.869 Q2149.45 264.841 2147.05 265.351 Q2144.64 265.86 2142.16 265.86 Q2135.96 265.86 2132.33 262.249 Q2128.71 258.638 2128.71 252.48 Q2128.71 246.115 2132.14 242.388 Q2135.59 238.638 2141.42 238.638 Q2146.65 238.638 2149.69 242.017 Q2152.74 245.374 2152.74 251.161 M2148.48 249.911 Q2148.44 246.416 2146.52 244.332 Q2144.62 242.249 2141.47 242.249 Q2137.9 242.249 2135.75 244.263 Q2133.62 246.277 2133.3 249.934 L2148.48 249.911 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M2163.85 261.3 L2163.85 275.05 L2159.57 275.05 L2159.57 239.263 L2163.85 239.263 L2163.85 243.198 Q2165.2 240.883 2167.23 239.772 Q2169.29 238.638 2172.14 238.638 Q2176.86 238.638 2179.8 242.388 Q2182.76 246.138 2182.76 252.249 Q2182.76 258.36 2179.8 262.11 Q2176.86 265.86 2172.14 265.86 Q2169.29 265.86 2167.23 264.749 Q2165.2 263.615 2163.85 261.3 M2178.34 252.249 Q2178.34 247.55 2176.4 244.888 Q2174.48 242.203 2171.1 242.203 Q2167.72 242.203 2165.77 244.888 Q2163.85 247.55 2163.85 252.249 Q2163.85 256.948 2165.77 259.633 Q2167.72 262.295 2171.1 262.295 Q2174.48 262.295 2176.4 259.633 Q2178.34 256.948 2178.34 252.249 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M2206.35 240.027 L2206.35 244.054 Q2204.55 243.129 2202.6 242.666 Q2200.66 242.203 2198.58 242.203 Q2195.4 242.203 2193.81 243.175 Q2192.23 244.147 2192.23 246.091 Q2192.23 247.573 2193.37 248.429 Q2194.5 249.263 2197.93 250.027 L2199.39 250.351 Q2203.92 251.323 2205.82 253.105 Q2207.74 254.865 2207.74 258.036 Q2207.74 261.647 2204.87 263.753 Q2202.02 265.86 2197.02 265.86 Q2194.94 265.86 2192.67 265.443 Q2190.43 265.05 2187.93 264.24 L2187.93 259.841 Q2190.29 261.068 2192.58 261.693 Q2194.87 262.295 2197.12 262.295 Q2200.13 262.295 2201.75 261.277 Q2203.37 260.235 2203.37 258.36 Q2203.37 256.624 2202.19 255.698 Q2201.03 254.772 2197.07 253.916 L2195.59 253.568 Q2191.63 252.735 2189.87 251.022 Q2188.11 249.286 2188.11 246.277 Q2188.11 242.619 2190.7 240.629 Q2193.3 238.638 2198.07 238.638 Q2200.43 238.638 2202.51 238.985 Q2204.59 239.332 2206.35 240.027 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M2215.13 259.309 L2220.01 259.309 L2220.01 265.189 L2215.13 265.189 L2215.13 259.309 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip240)" d="M2228.92 229.217 L2232.63 229.217 Q2236.1 234.68 2237.81 239.911 Q2239.55 245.142 2239.55 250.304 Q2239.55 255.49 2237.81 260.744 Q2236.1 265.999 2232.63 271.439 L2228.92 271.439 Q2232 266.138 2233.51 260.906 Q2235.03 255.652 2235.03 250.304 Q2235.03 244.957 2233.51 239.749 Q2232 234.541 2228.92 229.217 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /></svg>



Finally, the FMU is cleaned-up.


```julia
fmiUnload(defaultFMU)
```

### Summary

Based on the plots, it can be clearly seen that the CS-NeuralFMU model is able to learn the shift of the initial position. Even after only 300 runs, the curves overlap very much, so no further training with more runs is needed.

### Source

[1] Tobias Thummerer, Lars Mikelsons and Josef Kircher. 2021. **NeuralFMU: towards structural integration of FMUs into neural networks.** Martin Sjölund, Lena Buffoni, Adrian Pop and Lennart Ochel (Ed.). Proceedings of 14th Modelica Conference 2021, Linköping, Sweden, September 20-24, 2021. Linköping University Electronic Press, Linköping (Linköping Electronic Conference Proceedings ; 181), 297-306. [DOI: 10.3384/ecp21181297](https://doi.org/10.3384/ecp21181297)

