# Creation and training of ME-NeuralFMUs
Tutorial by Johannes Stoljar, Tobias Thummerer

## License
Copyright (c) 2021 Tobias Thummerer, Lars Mikelsons, Johannes Stoljar

Licensed under the MIT license. See [LICENSE](https://github.com/thummeto/FMIFlux.jl/blob/main/LICENSE) file in the project root for details.

## Motivation
The Julia Package *FMIFlux.jl* is motivated by the application of hybrid modeling. This package enables the user to integrate his simulation model between neural networks (NeuralFMU). For this, the simulation model must be exported as FMU (functional mock-up unit), which corresponds to a widely used standard. The big advantage of hybrid modeling with artificial neural networks is, that effects that are difficult to model (because they might be unknown) can be easily learned by the neural networks. For this purpose, the NeuralFMU is trained with measurement data containing the unmodeled physical effect. The final product is a simulation model including the orignially unmodeled effects. Another big advantage of the NeuralFMU is that it works with little data, because the FMU already contains the characterisitic functionality of the simulation and only the missing effects are added.

NeuralFMUs need not to be as easy as in this example. Basically a NeuralFMU can combine different ANN topologies that manipulate any FMU-input (system state, system inputs, time) and any FMU-output (system state derivative, system outputs, other system variables). However, for this example a NeuralFMU topology as shown in the following picture is used.

![NeuralFMU.svg](https://github.com/thummeto/FMIFlux.jl/blob/main/docs/src/examples/pics/NeuralFMU.svg?raw=true)

*NeuralFMU (ME) from* [[1]](#Source).

## Introduction to the example
In this example, simplified modeling of a one-dimensional spring pendulum (without friction) is compared to a model of the same system that includes a nonlinear friction model. The FMU with the simplified model will be named *simpleFMU* in the following and the model with the friction will be named *realFMU*. At the beginning, the actual state of both simulations is shown, whereby clear deviations can be seen in the graphs. The *realFMU* serves as a reference graph. The *simpleFMU* is then integrated into a NeuralFMU architecture and a training of the entire network is performed. After the training the final state is compared again to the *realFMU*. It can be clearly seen that by using the NeuralFMU, learning of the friction process has taken place.  


## Target group
The example is primarily intended for users who work in the field of first principle and/or hybrid modeling and are further interested in hybrid model building. The example wants to show how simple it is to combine FMUs with machine learning and to illustrate the advantages of this approach.


## Other formats
Besides this [Jupyter Notebook](https://github.com/thummeto/FMIFlux.jl/blob/main/example/simple_hybrid_ME.ipynb) there is also a [Julia file](https://github.com/thummeto/FMIFlux.jl/blob/main/example/simple_hybrid_ME.jl) with the same name, which contains only the code cells and for the documentation there is a [Markdown file](https://github.com/thummeto/FMIFlux.jl/blob/main/docs/src/examples/simple_hybrid_ME.md) corresponding to the notebook.  


## Getting started

### Installation prerequisites
|    | Description                       | Command     |  Alternative  |   
|:--- |:---                               |:---        |:---|
|1.  | Enter Package Manager via         |     ]       |     |
|2.  | Install FMI via                   |   add FMI   | add " https://github.com/ThummeTo/FMI.jl "   |
|3.  | Install FMIFlux via               | add FMIFlux | add " https://github.com/ThummeTo/FMIFlux.jl " |
|4.  | Install FMIZoo via                | add FMIZoo  | add " https://github.com/ThummeTo/FMIZoo.jl " |
|5.  | Install Flux via                  |  add Flux   |     |
|6.  | Install DifferentialEquations via | add DifferentialEquations |  |
|7.  | Install Plots via                 | add Plots   |     |

## Code section

To run the example, the previously installed packages must be included. 


```julia
# imports
using FMI
using FMIFlux
using FMIZoo
using Flux
using DifferentialEquations: Tsit5
import Plots
```

    WARNING: Method definition (::Type{DiffEqFlux.DeterministicCNF{M, P, RE, D, T, A, K} where K where A where T where D where RE where P where M})(M, P, RE, D, T, A, K) where {M, P, RE, D, T, A, K} in module DiffEqFlux at deprecated.jl:70 overwritten at C:\Users\thummeto\.julia\packages\DiffEqFlux\A5156\src\ffjord.jl:41.
      ** incremental compilation may be fatally broken for this module **
    


After importing the packages, the path to the *Functional Mock-up Units* (FMUs) is set. The FMU is a model exported meeting the *Functional Mock-up Interface* (FMI) Standard. The FMI is a free standard ([fmi-standard.org](http://fmi-standard.org/)) that defines a container and an interface to exchange dynamic models using a combination of XML files, binaries and C code zipped into a single file. 

The objec-orientated structure of the *SpringPendulum1D* (*simpleFMU*) can be seen in the following graphic and corresponds to a simple modeling.

![svg](https://github.com/thummeto/FMIFlux.jl/blob/main/docs/src/examples/pics/SpringPendulum1D.svg?raw=true)

In contrast, the model *SpringFrictionPendulum1D* (*realFMU*) is somewhat more accurate, because it includes a friction component. 

![svg](https://github.com/thummeto/FMIFlux.jl/blob/main/docs/src/examples/pics/SpringFrictionPendulum1D.svg?raw=true)

Next, the start time and end time of the simulation are set. Finally, a step size is specified to store the results of the simulation at these time steps.


```julia
tStart = 0.0
tStep = 0.01
tStop = 5.0
tSave = collect(tStart:tStep:tStop)
```


    501-element Vector{Float64}:
     0.0
     0.01
     0.02
     0.03
     0.04
     0.05
     0.06
     0.07
     0.08
     0.09
     ⋮
     4.92
     4.93
     4.94
     4.95
     4.96
     4.97
     4.98
     4.99
     5.0


### RealFMU

In the next lines of code the FMU of the *realFMU* model from *FMIZoo.jl* is loaded and instantiated.  


```julia
realFMU = fmiLoad("SpringFrictionPendulum1D", "Dymola", "2022x")
fmiInstantiate!(realFMU; loggingOn=false)
fmiInfo(realFMU)
```

    ┌ Info: fmi2Unzip(...): Successfully unzipped 29 files at `C:\Users\thummeto\AppData\Local\Temp\fmijl_XJkm8P\SpringFrictionPendulum1D`.
    └ @ FMIImport C:\Users\thummeto\Documents\FMIImport.jl\src\FMI2_ext.jl:77
    ┌ Info: fmi2Load(...): FMU resources location is `file:///C:/Users/thummeto/AppData/Local/Temp/fmijl_XJkm8P/SpringFrictionPendulum1D/resources`
    └ @ FMIImport C:\Users\thummeto\Documents\FMIImport.jl\src\FMI2_ext.jl:192
    ┌ Info: fmi2Load(...): FMU supports both CS and ME, using CS as default if nothing specified.
    └ @ FMIImport C:\Users\thummeto\Documents\FMIImport.jl\src\FMI2_ext.jl:195


    #################### Begin information for FMU ####################
    	Model name:			SpringFrictionPendulum1D
    	FMI-Version:			2.0
    	GUID:				{df491d8d-0598-4495-913e-5b025e54d7f2}
    	Generation tool:		Dymola Version 2022x (64-bit), 2021-10-08
    	Generation time:		2022-03-03T15:09:18Z
    	Var. naming conv.:		structured
    	Event indicators:		24
    	Inputs:				0
    	Outputs:			0
    	States:				2
    		33554432 ["mass.s"]
    		33554433 ["mass.v", "mass.v_relfric"]
    	Supports Co-Simulation:		true
    		Model identifier:	SpringFrictionPendulum1D
    		Get/Set State:		true
    		Serialize State:	true
    		Dir. Derivatives:	true
    		Var. com. steps:	true
    		Input interpol.:	true
    		Max order out. der.:	1
    	Supports Model-Exchange:	true
    		Model identifier:	SpringFrictionPendulum1D
    		Get/Set State:		true
    		Serialize State:	true
    		Dir. Derivatives:	true
    ##################### End information for FMU #####################


Both the start and end time are set via the *fmiSetupExperiment()* function. The experiment is initialized to get the information of the continuous states. You can get all continuous states of a FMU by the function *fmiGetContinuousStates()* and this is also done for the *realFMU*. It has two states: The first state is the position of the mass, which is initilized with $0.5m$, the second state is the velocity, which is initialized with $0\frac{m}{s}$.   


```julia
fmiSetupExperiment(realFMU, tStart, tStop)

fmiEnterInitializationMode(realFMU)
fmiExitInitializationMode(realFMU)

x₀ = fmiGetContinuousStates(realFMU)
```


    2-element Vector{Float64}:
     0.5
     0.0


In the following code block the *realFMU* is simulated, still specifying which variables are included. After the simulation is finished the result of the *realFMU* can be plotted. This plot also serves as a reference for the other model (*simpleFMU*).


```julia
vrs = ["mass.s", "mass.v", "mass.a", "mass.f"]
_, realSimData = fmiSimulate(realFMU, tStart, tStop; recordValues=vrs, saveat=tSave, setup=false, reset=false)
fmiPlot(realFMU, vrs, realSimData)
```


<?xml version="1.0" encoding="utf-8"?>
<svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" width="600" height="400" viewBox="0 0 2400 1600">
<defs>
  <clipPath id="clip940">
    <rect x="0" y="0" width="2400" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip940)" d="
M0 1600 L2400 1600 L2400 0 L0 0  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip941">
    <rect x="480" y="0" width="1681" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip940)" d="
M149.191 1423.18 L2352.76 1423.18 L2352.76 47.2441 L149.191 47.2441  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip942">
    <rect x="149" y="47" width="2205" height="1377"/>
  </clipPath>
</defs>
<polyline clip-path="url(#clip942)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  211.556,1423.18 211.556,47.2441 
  "/>
<polyline clip-path="url(#clip942)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  627.323,1423.18 627.323,47.2441 
  "/>
<polyline clip-path="url(#clip942)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1043.09,1423.18 1043.09,47.2441 
  "/>
<polyline clip-path="url(#clip942)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1458.86,1423.18 1458.86,47.2441 
  "/>
<polyline clip-path="url(#clip942)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1874.62,1423.18 1874.62,47.2441 
  "/>
<polyline clip-path="url(#clip942)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  2290.39,1423.18 2290.39,47.2441 
  "/>
<polyline clip-path="url(#clip940)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  149.191,1423.18 2352.76,1423.18 
  "/>
<polyline clip-path="url(#clip940)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  211.556,1423.18 211.556,1404.28 
  "/>
<polyline clip-path="url(#clip940)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  627.323,1423.18 627.323,1404.28 
  "/>
<polyline clip-path="url(#clip940)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1043.09,1423.18 1043.09,1404.28 
  "/>
<polyline clip-path="url(#clip940)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1458.86,1423.18 1458.86,1404.28 
  "/>
<polyline clip-path="url(#clip940)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1874.62,1423.18 1874.62,1404.28 
  "/>
<polyline clip-path="url(#clip940)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  2290.39,1423.18 2290.39,1404.28 
  "/>
<path clip-path="url(#clip940)" d="M211.556 1454.1 Q207.945 1454.1 206.116 1457.66 Q204.311 1461.2 204.311 1468.33 Q204.311 1475.44 206.116 1479.01 Q207.945 1482.55 211.556 1482.55 Q215.19 1482.55 216.996 1479.01 Q218.824 1475.44 218.824 1468.33 Q218.824 1461.2 216.996 1457.66 Q215.19 1454.1 211.556 1454.1 M211.556 1450.39 Q217.366 1450.39 220.422 1455 Q223.5 1459.58 223.5 1468.33 Q223.5 1477.06 220.422 1481.67 Q217.366 1486.25 211.556 1486.25 Q205.746 1486.25 202.667 1481.67 Q199.612 1477.06 199.612 1468.33 Q199.612 1459.58 202.667 1455 Q205.746 1450.39 211.556 1450.39 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M617.705 1481.64 L625.344 1481.64 L625.344 1455.28 L617.034 1456.95 L617.034 1452.69 L625.297 1451.02 L629.973 1451.02 L629.973 1481.64 L637.612 1481.64 L637.612 1485.58 L617.705 1485.58 L617.705 1481.64 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1037.74 1481.64 L1054.06 1481.64 L1054.06 1485.58 L1032.12 1485.58 L1032.12 1481.64 Q1034.78 1478.89 1039.36 1474.26 Q1043.97 1469.61 1045.15 1468.27 Q1047.4 1465.74 1048.28 1464.01 Q1049.18 1462.25 1049.18 1460.56 Q1049.18 1457.8 1047.23 1456.07 Q1045.31 1454.33 1042.21 1454.33 Q1040.01 1454.33 1037.56 1455.09 Q1035.13 1455.86 1032.35 1457.41 L1032.35 1452.69 Q1035.17 1451.55 1037.63 1450.97 Q1040.08 1450.39 1042.12 1450.39 Q1047.49 1450.39 1050.68 1453.08 Q1053.88 1455.77 1053.88 1460.26 Q1053.88 1462.39 1053.07 1464.31 Q1052.28 1466.2 1050.17 1468.8 Q1049.59 1469.47 1046.49 1472.69 Q1043.39 1475.88 1037.74 1481.64 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1463.1 1466.95 Q1466.46 1467.66 1468.34 1469.93 Q1470.23 1472.2 1470.23 1475.53 Q1470.23 1480.65 1466.72 1483.45 Q1463.2 1486.25 1456.72 1486.25 Q1454.54 1486.25 1452.22 1485.81 Q1449.93 1485.39 1447.48 1484.54 L1447.48 1480.02 Q1449.42 1481.16 1451.74 1481.74 Q1454.05 1482.32 1456.58 1482.32 Q1460.97 1482.32 1463.27 1480.58 Q1465.58 1478.84 1465.58 1475.53 Q1465.58 1472.48 1463.43 1470.77 Q1461.3 1469.03 1457.48 1469.03 L1453.45 1469.03 L1453.45 1465.19 L1457.66 1465.19 Q1461.11 1465.19 1462.94 1463.82 Q1464.77 1462.43 1464.77 1459.84 Q1464.77 1457.18 1462.87 1455.77 Q1461 1454.33 1457.48 1454.33 Q1455.56 1454.33 1453.36 1454.75 Q1451.16 1455.16 1448.52 1456.04 L1448.52 1451.88 Q1451.18 1451.14 1453.5 1450.77 Q1455.84 1450.39 1457.9 1450.39 Q1463.22 1450.39 1466.32 1452.83 Q1469.42 1455.23 1469.42 1459.35 Q1469.42 1462.22 1467.78 1464.21 Q1466.14 1466.18 1463.1 1466.95 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1877.63 1455.09 L1865.83 1473.54 L1877.63 1473.54 L1877.63 1455.09 M1876.41 1451.02 L1882.29 1451.02 L1882.29 1473.54 L1887.22 1473.54 L1887.22 1477.43 L1882.29 1477.43 L1882.29 1485.58 L1877.63 1485.58 L1877.63 1477.43 L1862.03 1477.43 L1862.03 1472.92 L1876.41 1451.02 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M2280.67 1451.02 L2299.03 1451.02 L2299.03 1454.96 L2284.95 1454.96 L2284.95 1463.43 Q2285.97 1463.08 2286.99 1462.92 Q2288.01 1462.73 2289.03 1462.73 Q2294.81 1462.73 2298.19 1465.9 Q2301.57 1469.08 2301.57 1474.49 Q2301.57 1480.07 2298.1 1483.17 Q2294.63 1486.25 2288.31 1486.25 Q2286.13 1486.25 2283.86 1485.88 Q2281.62 1485.51 2279.21 1484.77 L2279.21 1480.07 Q2281.29 1481.2 2283.52 1481.76 Q2285.74 1482.32 2288.21 1482.32 Q2292.22 1482.32 2294.56 1480.21 Q2296.9 1478.1 2296.9 1474.49 Q2296.9 1470.88 2294.56 1468.77 Q2292.22 1466.67 2288.21 1466.67 Q2286.34 1466.67 2284.46 1467.08 Q2282.61 1467.5 2280.67 1468.38 L2280.67 1451.02 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1199.28 1522.27 L1199.28 1532.4 L1211.35 1532.4 L1211.35 1536.95 L1199.28 1536.95 L1199.28 1556.3 Q1199.28 1560.66 1200.46 1561.9 Q1201.67 1563.14 1205.33 1563.14 L1211.35 1563.14 L1211.35 1568.04 L1205.33 1568.04 Q1198.55 1568.04 1195.97 1565.53 Q1193.4 1562.98 1193.4 1556.3 L1193.4 1536.95 L1189.1 1536.95 L1189.1 1532.4 L1193.4 1532.4 L1193.4 1522.27 L1199.28 1522.27 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1239.23 1518.52 L1252.72 1518.52 L1252.72 1523.07 L1245.09 1523.07 L1245.09 1572.09 L1252.72 1572.09 L1252.72 1576.64 L1239.23 1576.64 L1239.23 1518.52 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1287.93 1533.45 L1287.93 1538.98 Q1285.44 1537.71 1282.77 1537.07 Q1280.1 1536.44 1277.23 1536.44 Q1272.87 1536.44 1270.68 1537.77 Q1268.51 1539.11 1268.51 1541.79 Q1268.51 1543.82 1270.07 1545 Q1271.63 1546.15 1276.34 1547.2 L1278.35 1547.64 Q1284.58 1548.98 1287.19 1551.43 Q1289.84 1553.85 1289.84 1558.21 Q1289.84 1563.17 1285.89 1566.07 Q1281.97 1568.97 1275.1 1568.97 Q1272.23 1568.97 1269.12 1568.39 Q1266.03 1567.85 1262.59 1566.74 L1262.59 1560.69 Q1265.84 1562.38 1268.99 1563.24 Q1272.14 1564.07 1275.23 1564.07 Q1279.36 1564.07 1281.59 1562.66 Q1283.82 1561.23 1283.82 1558.65 Q1283.82 1556.27 1282.2 1554.99 Q1280.61 1553.72 1275.16 1552.54 L1273.13 1552.07 Q1267.68 1550.92 1265.26 1548.56 Q1262.85 1546.18 1262.85 1542.04 Q1262.85 1537.01 1266.41 1534.27 Q1269.98 1531.54 1276.53 1531.54 Q1279.78 1531.54 1282.64 1532.01 Q1285.51 1532.49 1287.93 1533.45 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1312.85 1518.52 L1312.85 1576.64 L1299.35 1576.64 L1299.35 1572.09 L1306.96 1572.09 L1306.96 1523.07 L1299.35 1523.07 L1299.35 1518.52 L1312.85 1518.52 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip942)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  149.191,1247.37 2352.76,1247.37 
  "/>
<polyline clip-path="url(#clip942)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  149.191,969.637 2352.76,969.637 
  "/>
<polyline clip-path="url(#clip942)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  149.191,691.907 2352.76,691.907 
  "/>
<polyline clip-path="url(#clip942)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  149.191,414.176 2352.76,414.176 
  "/>
<polyline clip-path="url(#clip942)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  149.191,136.446 2352.76,136.446 
  "/>
<polyline clip-path="url(#clip940)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  149.191,1423.18 149.191,47.2441 
  "/>
<polyline clip-path="url(#clip940)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  149.191,1247.37 168.089,1247.37 
  "/>
<polyline clip-path="url(#clip940)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  149.191,969.637 168.089,969.637 
  "/>
<polyline clip-path="url(#clip940)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  149.191,691.907 168.089,691.907 
  "/>
<polyline clip-path="url(#clip940)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  149.191,414.176 168.089,414.176 
  "/>
<polyline clip-path="url(#clip940)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  149.191,136.446 168.089,136.446 
  "/>
<path clip-path="url(#clip940)" d="M50.9921 1247.82 L80.6679 1247.82 L80.6679 1251.75 L50.9921 1251.75 L50.9921 1247.82 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M103.608 1234.16 L91.8021 1252.61 L103.608 1252.61 L103.608 1234.16 M102.381 1230.09 L108.26 1230.09 L108.26 1252.61 L113.191 1252.61 L113.191 1256.5 L108.26 1256.5 L108.26 1264.65 L103.608 1264.65 L103.608 1256.5 L88.0058 1256.5 L88.0058 1251.99 L102.381 1230.09 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M53.0754 970.088 L82.7512 970.088 L82.7512 974.023 L53.0754 974.023 L53.0754 970.088 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M96.8715 982.982 L113.191 982.982 L113.191 986.917 L91.2465 986.917 L91.2465 982.982 Q93.9086 980.227 98.4919 975.597 Q103.098 970.945 104.279 969.602 Q106.524 967.079 107.404 965.343 Q108.307 963.584 108.307 961.894 Q108.307 959.139 106.362 957.403 Q104.441 955.667 101.339 955.667 Q99.14 955.667 96.6863 956.431 Q94.2558 957.195 91.478 958.746 L91.478 954.024 Q94.3021 952.889 96.7558 952.311 Q99.2095 951.732 101.246 951.732 Q106.617 951.732 109.811 954.417 Q113.006 957.102 113.006 961.593 Q113.006 963.723 112.196 965.644 Q111.408 967.542 109.302 970.135 Q108.723 970.806 105.621 974.023 Q102.52 977.218 96.8715 982.982 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M101.246 677.705 Q97.6354 677.705 95.8067 681.27 Q94.0012 684.812 94.0012 691.941 Q94.0012 699.048 95.8067 702.613 Q97.6354 706.154 101.246 706.154 Q104.881 706.154 106.686 702.613 Q108.515 699.048 108.515 691.941 Q108.515 684.812 106.686 681.27 Q104.881 677.705 101.246 677.705 M101.246 674.002 Q107.057 674.002 110.112 678.608 Q113.191 683.191 113.191 691.941 Q113.191 700.668 110.112 705.275 Q107.057 709.858 101.246 709.858 Q95.4363 709.858 92.3576 705.275 Q89.3021 700.668 89.3021 691.941 Q89.3021 683.191 92.3576 678.608 Q95.4363 674.002 101.246 674.002 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M96.8715 427.521 L113.191 427.521 L113.191 431.456 L91.2465 431.456 L91.2465 427.521 Q93.9086 424.766 98.4919 420.137 Q103.098 415.484 104.279 414.142 Q106.524 411.618 107.404 409.882 Q108.307 408.123 108.307 406.433 Q108.307 403.679 106.362 401.943 Q104.441 400.206 101.339 400.206 Q99.14 400.206 96.6863 400.97 Q94.2558 401.734 91.478 403.285 L91.478 398.563 Q94.3021 397.429 96.7558 396.85 Q99.2095 396.271 101.246 396.271 Q106.617 396.271 109.811 398.956 Q113.006 401.642 113.006 406.132 Q113.006 408.262 112.196 410.183 Q111.408 412.081 109.302 414.674 Q108.723 415.345 105.621 418.563 Q102.52 421.757 96.8715 427.521 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M103.608 123.24 L91.8021 141.689 L103.608 141.689 L103.608 123.24 M102.381 119.166 L108.26 119.166 L108.26 141.689 L113.191 141.689 L113.191 145.578 L108.26 145.578 L108.26 153.726 L103.608 153.726 L103.608 145.578 L88.0058 145.578 L88.0058 141.064 L102.381 119.166 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip942)" style="stroke:#009af9; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  211.556,622.474 215.714,622.444 219.871,622.355 224.029,622.206 228.187,621.997 232.344,621.727 236.502,621.397 240.66,621.006 244.817,620.555 248.975,620.043 
  253.133,619.472 257.29,618.841 261.448,618.15 265.606,617.401 269.763,616.593 273.921,615.728 278.079,614.806 282.236,613.828 286.394,612.795 290.552,611.708 
  294.709,610.567 298.867,609.375 303.025,608.131 307.182,606.837 311.34,605.495 315.498,604.105 319.655,602.669 323.813,601.189 327.971,599.665 332.128,598.1 
  336.286,596.494 340.444,594.85 344.601,593.17 348.759,591.453 352.917,589.703 357.074,587.921 361.232,586.109 365.39,584.269 369.547,582.403 373.705,580.512 
  377.863,578.598 382.02,576.663 386.178,574.71 390.336,572.739 394.493,570.754 398.651,568.756 402.809,566.748 406.966,564.731 411.124,562.707 415.282,560.678 
  419.439,558.647 423.597,556.615 427.755,554.585 431.912,552.558 436.07,550.536 440.228,548.522 444.385,546.517 448.543,544.523 452.701,542.543 456.858,540.578 
  461.016,538.63 465.174,536.702 469.331,534.794 473.489,532.91 477.647,531.051 481.804,529.219 485.962,527.416 490.12,525.643 494.277,523.902 498.435,522.195 
  502.593,520.525 506.75,518.891 510.908,517.297 515.066,515.743 519.223,514.232 523.381,512.765 527.539,511.343 531.696,509.968 535.854,508.641 540.012,507.363 
  544.17,506.137 548.327,504.963 552.485,503.842 556.643,502.776 560.8,501.766 564.958,500.812 569.116,499.917 573.273,499.081 577.431,498.305 581.589,497.59 
  585.746,496.936 589.904,496.346 594.062,495.819 598.219,495.356 602.377,494.959 606.535,494.627 610.692,494.362 614.85,494.164 619.008,494.034 623.165,493.973 
  627.323,493.977 631.481,494.032 635.638,494.136 639.796,494.289 643.954,494.492 648.111,494.745 652.269,495.048 656.427,495.401 660.584,495.804 664.742,496.257 
  668.9,496.759 673.057,497.31 677.215,497.911 681.373,498.56 685.53,499.257 689.688,500.001 693.846,500.792 698.003,501.63 702.161,502.513 706.319,503.44 
  710.476,504.412 714.634,505.427 718.792,506.484 722.949,507.581 727.107,508.719 731.265,509.896 735.422,511.111 739.58,512.362 743.738,513.648 747.895,514.969 
  752.053,516.323 756.211,517.709 760.368,519.125 764.526,520.57 768.684,522.042 772.841,523.54 776.999,525.063 781.157,526.609 785.314,528.176 789.472,529.763 
  793.63,531.369 797.787,532.991 801.945,534.628 806.103,536.279 810.26,537.941 814.418,539.614 818.576,541.295 822.733,542.983 826.891,544.676 831.049,546.373 
  835.206,548.071 839.364,549.77 843.522,551.467 847.679,553.161 851.837,554.849 855.995,556.531 860.152,558.204 864.31,559.868 868.468,561.519 872.625,563.157 
  876.783,564.78 880.941,566.387 885.098,567.975 889.256,569.543 893.414,571.09 897.571,572.613 901.729,574.112 905.887,575.585 910.044,577.03 914.202,578.447 
  918.36,579.832 922.517,581.186 926.675,582.507 930.833,583.792 934.99,585.042 939.148,586.255 943.306,587.429 947.463,588.563 951.621,589.656 955.779,590.707 
  959.936,591.715 964.094,592.679 968.252,593.597 972.41,594.469 976.567,595.293 980.725,596.069 984.883,596.796 989.04,597.473 993.198,598.098 997.356,598.672 
  1001.51,599.193 1005.67,599.661 1009.83,600.075 1013.99,600.435 1018.14,600.739 1022.3,600.987 1026.46,601.178 1030.62,601.313 1034.77,601.389 1038.93,601.407 
  1043.09,601.381 1047.25,601.316 1051.41,601.213 1055.56,601.071 1059.72,600.89 1063.88,600.67 1068.04,600.411 1072.19,600.113 1076.35,599.776 1080.51,599.4 
  1084.67,598.985 1088.82,598.532 1092.98,598.04 1097.14,597.511 1101.3,596.944 1105.45,596.339 1109.61,595.699 1113.77,595.022 1117.93,594.31 1122.09,593.562 
  1126.24,592.781 1130.4,591.965 1134.56,591.116 1138.72,590.236 1142.87,589.323 1147.03,588.38 1151.19,587.408 1155.35,586.406 1159.5,585.377 1163.66,584.321 
  1167.82,583.24 1171.98,582.133 1176.14,581.003 1180.29,579.85 1184.45,578.676 1188.61,577.482 1192.77,576.268 1196.92,575.036 1201.08,573.788 1205.24,572.524 
  1209.4,571.246 1213.55,569.955 1217.71,568.652 1221.87,567.338 1226.03,566.015 1230.19,564.684 1234.34,563.347 1238.5,562.005 1242.66,560.659 1246.82,559.311 
  1250.97,557.961 1255.13,556.611 1259.29,555.263 1263.45,553.918 1267.6,552.578 1271.76,551.243 1275.92,549.915 1280.08,548.595 1284.23,547.285 1288.39,545.986 
  1292.55,544.699 1296.71,543.426 1300.87,542.168 1305.02,540.926 1309.18,539.701 1313.34,538.495 1317.5,537.309 1321.65,536.144 1325.81,535.002 1329.97,533.883 
  1334.13,532.788 1338.28,531.719 1342.44,530.677 1346.6,529.663 1350.76,528.678 1354.92,527.723 1359.07,526.799 1363.23,525.907 1367.39,525.048 1371.55,524.223 
  1375.7,523.433 1379.86,522.678 1384.02,521.96 1388.18,521.28 1392.33,520.638 1396.49,520.035 1400.65,519.471 1404.81,518.948 1408.96,518.467 1413.12,518.027 
  1417.28,517.63 1421.44,517.276 1425.6,516.965 1429.75,516.699 1433.91,516.478 1438.07,516.302 1442.23,516.172 1446.38,516.088 1450.54,516.051 1454.7,516.056 
  1458.86,516.089 1463.01,516.148 1467.17,516.235 1471.33,516.349 1475.49,516.49 1479.65,516.658 1483.8,516.855 1487.96,517.078 1492.12,517.329 1496.28,517.607 
  1500.43,517.913 1504.59,518.246 1508.75,518.605 1512.91,518.991 1517.06,519.404 1521.22,519.843 1525.38,520.307 1529.54,520.798 1533.69,521.313 1537.85,521.853 
  1542.01,522.418 1546.17,523.006 1550.33,523.617 1554.48,524.251 1558.64,524.908 1562.8,525.586 1566.96,526.286 1571.11,527.005 1575.27,527.745 1579.43,528.504 
  1583.59,529.282 1587.74,530.077 1591.9,530.889 1596.06,531.717 1600.22,532.56 1604.38,533.418 1608.53,534.29 1612.69,535.175 1616.85,536.071 1621.01,536.979 
  1625.16,537.897 1629.32,538.825 1633.48,539.761 1637.64,540.704 1641.79,541.654 1645.95,542.61 1650.11,543.571 1654.27,544.535 1658.43,545.502 1662.58,546.471 
  1666.74,547.441 1670.9,548.411 1675.06,549.38 1679.21,550.346 1683.37,551.31 1687.53,552.27 1691.69,553.225 1695.84,554.174 1700,555.116 1704.16,556.05 
  1708.32,556.976 1712.47,557.892 1716.63,558.797 1720.79,559.69 1724.95,560.571 1729.11,561.439 1733.26,562.292 1737.42,563.13 1741.58,563.952 1745.74,564.757 
  1749.89,565.545 1754.05,566.313 1758.21,567.062 1762.37,567.791 1766.52,568.499 1770.68,569.185 1774.84,569.849 1779,570.489 1783.16,571.105 1787.31,571.696 
  1791.47,572.262 1795.63,572.802 1799.79,573.315 1803.94,573.801 1808.1,574.259 1812.26,574.688 1816.42,575.088 1820.57,575.458 1824.73,575.798 1828.89,576.108 
  1833.05,576.386 1837.2,576.632 1841.36,576.846 1845.52,577.028 1849.68,577.177 1853.84,577.292 1857.99,577.373 1862.15,577.42 1866.31,577.433 1870.47,577.424 
  1874.62,577.401 1878.78,577.363 1882.94,577.312 1887.1,577.246 1891.25,577.165 1895.41,577.071 1899.57,576.961 1903.73,576.838 1907.89,576.699 1912.04,576.546 
  1916.2,576.379 1920.36,576.198 1924.52,576.002 1928.67,575.792 1932.83,575.568 1936.99,575.33 1941.15,575.078 1945.3,574.812 1949.46,574.533 1953.62,574.241 
  1957.78,573.936 1961.93,573.618 1966.09,573.287 1970.25,572.944 1974.41,572.589 1978.57,572.222 1982.72,571.844 1986.88,571.455 1991.04,571.054 1995.2,570.643 
  1999.35,570.223 2003.51,569.792 2007.67,569.352 2011.83,568.903 2015.98,568.446 2020.14,567.98 2024.3,567.507 2028.46,567.026 2032.62,566.538 2036.77,566.045 
  2040.93,565.545 2045.09,565.04 2049.25,564.53 2053.4,564.015 2057.56,563.496 2061.72,562.975 2065.88,562.45 2070.03,561.922 2074.19,561.393 2078.35,560.862 
  2082.51,560.331 2086.67,559.799 2090.82,559.267 2094.98,558.736 2099.14,558.207 2103.3,557.679 2107.45,557.153 2111.61,556.63 2115.77,556.111 2119.93,555.596 
  2124.08,555.085 2128.24,554.579 2132.4,554.079 2136.56,553.585 2140.71,553.097 2144.87,552.616 2149.03,552.143 2153.19,551.679 2157.35,551.223 2161.5,550.776 
  2165.66,550.338 2169.82,549.911 2173.98,549.494 2178.13,549.089 2182.29,548.695 2186.45,548.313 2190.61,547.943 2194.76,547.586 2198.92,547.243 2203.08,546.913 
  2207.24,546.598 2211.4,546.296 2215.55,546.01 2219.71,545.739 2223.87,545.484 2228.03,545.245 2232.18,545.022 2236.34,544.816 2240.5,544.627 2244.66,544.455 
  2248.81,544.301 2252.97,544.164 2257.13,544.046 2261.29,543.947 2265.44,543.866 2269.6,543.804 2273.76,543.761 2277.92,543.737 2282.08,543.733 2286.23,543.733 
  2290.39,543.733 
  "/>
<polyline clip-path="url(#clip942)" style="stroke:#e26f46; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  211.556,691.907 215.714,685.979 219.871,680.007 224.029,673.999 228.187,667.966 232.344,661.916 236.502,655.86 240.66,649.806 244.817,643.764 248.975,637.743 
  253.133,631.751 257.29,625.797 261.448,619.889 265.606,614.035 269.763,608.242 273.921,602.519 278.079,596.872 282.236,591.309 286.394,585.836 290.552,580.46 
  294.709,575.187 298.867,570.025 303.025,564.977 307.182,560.05 311.34,555.251 315.498,550.585 319.655,546.056 323.813,541.669 327.971,537.43 332.128,533.344 
  336.286,529.413 340.444,525.644 344.601,522.04 348.759,518.604 352.917,515.34 357.074,512.253 361.232,509.344 365.39,506.617 369.547,504.075 373.705,501.72 
  377.863,499.556 382.02,497.583 386.178,495.804 390.336,494.222 394.493,492.837 398.651,491.652 402.809,490.667 406.966,489.883 411.124,489.301 415.282,488.921 
  419.439,488.745 423.597,488.772 427.755,489.002 431.912,489.435 436.07,490.071 440.228,490.908 444.385,491.947 448.543,493.185 452.701,494.623 456.858,496.258 
  461.016,498.088 465.174,500.113 469.331,502.33 473.489,504.737 477.647,507.331 481.804,510.111 485.962,513.073 490.12,516.215 494.277,519.533 498.435,523.025 
  502.593,526.687 506.75,530.517 510.908,534.509 515.066,538.662 519.223,542.969 523.381,547.429 527.539,552.038 531.696,556.79 535.854,561.681 540.012,566.708 
  544.17,571.866 548.327,577.151 552.485,582.558 556.643,588.082 560.8,593.721 564.958,599.469 569.116,605.322 573.273,611.277 577.431,617.327 581.589,623.469 
  585.746,629.7 589.904,636.015 594.062,642.411 598.219,648.885 602.377,655.433 606.535,662.053 610.692,668.744 614.85,675.503 619.008,682.328 623.165,689.219 
  627.323,694.895 631.481,699.806 635.638,704.751 639.796,709.72 643.954,714.709 648.111,719.707 652.269,724.71 656.427,729.709 660.584,734.698 664.742,739.668 
  668.9,744.613 673.057,749.526 677.215,754.4 681.373,759.23 685.53,764.008 689.688,768.728 693.846,773.385 698.003,777.972 702.161,782.485 706.319,786.916 
  710.476,791.262 714.634,795.516 718.792,799.674 722.949,803.731 727.107,807.683 731.265,811.524 735.422,815.251 739.58,818.86 743.738,822.345 747.895,825.702 
  752.053,828.93 756.211,832.022 760.368,834.977 764.526,837.791 768.684,840.461 772.841,842.983 776.999,845.357 781.157,847.578 785.314,849.645 789.472,851.555 
  793.63,853.306 797.787,854.897 801.945,856.325 806.103,857.59 810.26,858.69 814.418,859.623 818.576,860.389 822.733,860.987 826.891,861.416 831.049,861.675 
  835.206,861.765 839.364,861.685 843.522,861.435 847.679,861.016 851.837,860.427 855.995,859.669 860.152,858.743 864.31,857.65 868.468,856.391 872.625,854.966 
  876.783,853.378 880.941,851.627 885.098,849.716 889.256,847.645 893.414,845.418 897.571,843.036 901.729,840.501 905.887,837.815 910.044,834.982 914.202,832.003 
  918.36,828.881 922.517,825.62 926.675,822.222 930.833,818.689 934.99,815.026 939.148,811.236 943.306,807.321 947.463,803.285 951.621,799.131 955.779,794.862 
  959.936,790.484 964.094,785.998 968.252,781.409 972.41,776.719 976.567,771.932 980.725,767.053 984.883,762.085 989.04,757.032 993.198,751.895 997.356,746.68 
  1001.51,741.389 1005.67,736.025 1009.83,730.591 1013.99,725.09 1018.14,719.524 1022.3,713.895 1026.46,708.206 1030.62,702.458 1034.77,696.654 1038.93,691.191 
  1043.09,687.373 1047.25,683.525 1051.41,679.654 1055.56,675.765 1059.72,671.864 1063.88,667.954 1068.04,664.042 1072.19,660.134 1076.35,656.234 1080.51,652.349 
  1084.67,648.484 1088.82,644.643 1092.98,640.831 1097.14,637.055 1101.3,633.318 1105.45,629.625 1109.61,625.982 1113.77,622.391 1117.93,618.859 1122.09,615.389 
  1126.24,611.985 1130.4,608.654 1134.56,605.397 1138.72,602.219 1142.87,599.124 1147.03,596.115 1151.19,593.196 1155.35,590.371 1159.5,587.642 1163.66,585.013 
  1167.82,582.486 1171.98,580.065 1176.14,577.753 1180.29,575.551 1184.45,573.462 1188.61,571.489 1192.77,569.635 1196.92,567.9 1201.08,566.287 1205.24,564.798 
  1209.4,563.435 1213.55,562.198 1217.71,561.09 1221.87,560.112 1226.03,559.265 1230.19,558.55 1234.34,557.967 1238.5,557.518 1242.66,557.203 1246.82,557.023 
  1250.97,556.977 1255.13,557.067 1259.29,557.291 1263.45,557.651 1267.6,558.145 1271.76,558.774 1275.92,559.536 1280.08,560.432 1284.23,561.46 1288.39,562.62 
  1292.55,563.911 1296.71,565.331 1300.87,566.879 1305.02,568.555 1309.18,570.356 1313.34,572.281 1317.5,574.327 1321.65,576.495 1325.81,578.78 1329.97,581.182 
  1334.13,583.699 1338.28,586.329 1342.44,589.068 1346.6,591.915 1350.76,594.867 1354.92,597.922 1359.07,601.078 1363.23,604.332 1367.39,607.681 1371.55,611.122 
  1375.7,614.653 1379.86,618.272 1384.02,621.975 1388.18,625.76 1392.33,629.624 1396.49,633.564 1400.65,637.578 1404.81,641.662 1408.96,645.814 1413.12,650.033 
  1417.28,654.314 1421.44,658.657 1425.6,663.057 1429.75,667.513 1433.91,672.022 1438.07,676.583 1442.23,681.193 1446.38,685.852 1450.54,690.557 1454.7,693.813 
  1458.86,696.498 1463.01,699.203 1467.17,701.923 1471.33,704.655 1475.49,707.395 1479.65,710.141 1483.8,712.888 1487.96,715.634 1492.12,718.375 1496.28,721.106 
  1500.43,723.825 1504.59,726.528 1508.75,729.211 1512.91,731.871 1517.06,734.505 1521.22,737.109 1525.38,739.681 1529.54,742.216 1533.69,744.712 1537.85,747.165 
  1542.01,749.573 1546.17,751.933 1550.33,754.242 1554.48,756.497 1558.64,758.695 1562.8,760.832 1566.96,762.908 1571.11,764.919 1575.27,766.863 1579.43,768.737 
  1583.59,770.539 1587.74,772.268 1591.9,773.92 1596.06,775.495 1600.22,776.99 1604.38,778.404 1608.53,779.734 1612.69,780.979 1616.85,782.138 1621.01,783.209 
  1625.16,784.192 1629.32,785.084 1633.48,785.885 1637.64,786.593 1641.79,787.208 1645.95,787.729 1650.11,788.155 1654.27,788.485 1658.43,788.72 1662.58,788.858 
  1666.74,788.899 1670.9,788.843 1675.06,788.69 1679.21,788.44 1683.37,788.092 1687.53,787.648 1691.69,787.107 1695.84,786.469 1700,785.736 1704.16,784.907 
  1708.32,783.983 1712.47,782.965 1716.63,781.853 1720.79,780.649 1724.95,779.353 1729.11,777.967 1733.26,776.492 1737.42,774.928 1741.58,773.277 1745.74,771.54 
  1749.89,769.718 1754.05,767.814 1758.21,765.829 1762.37,763.763 1766.52,761.619 1770.68,759.399 1774.84,757.104 1779,754.735 1783.16,752.296 1787.31,749.786 
  1791.47,747.209 1795.63,744.567 1799.79,741.86 1803.94,739.092 1808.1,736.263 1812.26,733.377 1816.42,730.434 1820.57,727.438 1824.73,724.389 1828.89,721.29 
  1833.05,718.143 1837.2,714.949 1841.36,711.711 1845.52,708.43 1849.68,705.108 1853.84,701.747 1857.99,698.349 1862.15,694.915 1866.31,691.722 1870.47,690.316 
  1874.62,688.899 1878.78,687.472 1882.94,686.037 1887.1,684.596 1891.25,683.15 1895.41,681.701 1899.57,680.25 1903.73,678.799 1907.89,677.349 1912.04,675.903 
  1916.2,674.462 1920.36,673.028 1924.52,671.603 1928.67,670.188 1932.83,668.785 1936.99,667.395 1941.15,666.022 1945.3,664.665 1949.46,663.327 1953.62,662.01 
  1957.78,660.714 1961.93,659.443 1966.09,658.196 1970.25,656.976 1974.41,655.784 1978.57,654.622 1982.72,653.491 1986.88,652.392 1991.04,651.328 1995.2,650.299 
  1999.35,649.306 2003.51,648.35 2007.67,647.434 2011.83,646.559 2015.98,645.724 2020.14,644.931 2024.3,644.182 2028.46,643.478 2032.62,642.819 2036.77,642.206 
  2040.93,641.64 2045.09,641.122 2049.25,640.652 2053.4,640.232 2057.56,639.862 2061.72,639.543 2065.88,639.274 2070.03,639.057 2074.19,638.893 2078.35,638.78 
  2082.51,638.72 2086.67,638.714 2090.82,638.761 2094.98,638.861 2099.14,639.014 2103.3,639.222 2107.45,639.483 2111.61,639.798 2115.77,640.166 2119.93,640.588 
  2124.08,641.064 2128.24,641.592 2132.4,642.173 2136.56,642.807 2140.71,643.493 2144.87,644.23 2149.03,645.018 2153.19,645.858 2157.35,646.747 2161.5,647.685 
  2165.66,648.673 2169.82,649.709 2173.98,650.792 2178.13,651.921 2182.29,653.097 2186.45,654.318 2190.61,655.583 2194.76,656.891 2198.92,658.241 2203.08,659.633 
  2207.24,661.066 2211.4,662.538 2215.55,664.048 2219.71,665.596 2223.87,667.179 2228.03,668.799 2232.18,670.452 2236.34,672.138 2240.5,673.856 2244.66,675.604 
  2248.81,677.382 2252.97,679.188 2257.13,681.021 2261.29,682.88 2265.44,684.763 2269.6,686.67 2273.76,688.599 2277.92,690.548 2282.08,691.907 2286.23,691.907 
  2290.39,691.907 
  "/>
<polyline clip-path="url(#clip942)" style="stroke:#3da44d; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  211.556,101.73 215.714,96.6426 219.871,92.579 224.029,89.5214 228.187,87.451 232.344,86.3465 236.502,86.1857 240.66,86.946 244.817,88.6044 248.975,91.1371 
  253.133,94.5199 257.29,98.7307 261.448,103.744 265.606,109.541 269.763,116.094 273.921,123.386 278.079,131.391 282.236,140.089 286.394,149.458 290.552,159.476 
  294.709,170.125 298.867,181.386 303.025,193.234 307.182,205.652 311.34,218.621 315.498,232.123 319.655,246.134 323.813,260.637 327.971,275.613 332.128,291.041 
  336.286,306.903 340.444,323.178 344.601,339.846 348.759,356.891 352.917,374.292 357.074,392.029 361.232,410.079 365.39,428.424 369.547,447.041 373.705,465.909 
  377.863,485.021 382.02,504.347 386.178,523.867 390.336,543.558 394.493,563.4 398.651,583.369 402.809,603.443 406.966,623.612 411.124,643.849 415.282,664.134 
  419.439,684.445 423.597,704.763 427.755,725.069 431.912,745.343 436.07,765.564 440.228,785.712 444.385,805.768 448.543,825.713 452.701,845.527 456.858,865.191 
  461.016,884.685 465.174,903.994 469.331,923.098 473.489,941.978 477.647,960.617 481.804,978.998 485.962,997.106 490.12,1014.92 494.277,1032.44 498.435,1049.63 
  502.593,1066.48 506.75,1082.99 510.908,1099.13 515.066,1114.9 519.223,1130.29 523.381,1145.28 527.539,1159.86 531.696,1174.03 535.854,1187.78 540.012,1201.1 
  544.17,1213.99 548.327,1226.45 552.485,1238.46 556.643,1250.03 560.8,1261.17 564.958,1271.87 569.116,1282.15 573.273,1292 577.431,1301.43 581.589,1310.47 
  585.746,1319.13 589.904,1327.42 594.062,1335.36 598.219,1342.99 602.377,1350.33 606.535,1357.43 610.692,1364.31 614.85,1371.04 619.008,1377.66 623.165,1384.24 
  627.323,1181.18 631.481,1184.93 635.638,1187.9 639.796,1190.1 643.954,1191.55 648.111,1192.24 652.269,1192.19 656.427,1191.42 660.584,1189.94 664.742,1187.75 
  668.9,1184.88 673.057,1181.34 677.215,1177.14 681.373,1172.31 685.53,1166.85 689.688,1160.79 693.846,1154.13 698.003,1146.9 702.161,1139.11 706.319,1130.78 
  710.476,1121.92 714.634,1112.54 718.792,1102.67 722.949,1092.33 727.107,1081.52 731.265,1070.26 735.422,1058.58 739.58,1046.49 743.738,1033.99 747.895,1021.11 
  752.053,1007.87 756.211,994.282 760.368,980.359 764.526,966.12 768.684,951.584 772.841,936.766 776.999,921.684 781.157,906.354 785.314,890.795 789.472,875.022 
  793.63,859.054 797.787,842.907 801.945,826.598 806.103,810.146 810.26,793.566 814.418,776.877 818.576,760.094 822.733,743.236 826.891,726.32 831.049,709.364 
  835.206,692.383 839.364,675.395 843.522,658.416 847.679,641.464 851.837,624.557 855.995,607.711 860.152,590.94 864.31,574.26 868.468,557.69 872.625,541.245 
  876.783,524.942 880.941,508.793 885.098,492.813 889.256,477.017 893.414,461.422 897.571,446.04 901.729,430.885 905.887,415.968 910.044,401.302 914.202,386.901 
  918.36,372.776 922.517,358.938 926.675,345.396 930.833,332.159 934.99,319.238 939.148,306.641 943.306,294.376 947.463,282.446 951.621,270.86 955.779,259.622 
  959.936,248.736 964.094,238.205 968.252,228.029 972.41,218.209 976.567,208.743 980.725,199.631 984.883,190.867 989.04,182.445 993.198,174.359 997.356,166.598 
  1001.51,159.151 1005.67,152.005 1009.83,145.142 1013.99,138.544 1018.14,132.188 1022.3,126.051 1026.46,120.101 1030.62,114.308 1034.77,108.632 1038.93,311.718 
  1043.09,308.497 1047.25,305.829 1051.41,303.718 1055.56,302.162 1059.72,301.159 1063.88,300.706 1068.04,300.798 1072.19,301.431 1076.35,302.599 1080.51,304.293 
  1084.67,306.507 1088.82,309.234 1092.98,312.462 1097.14,316.184 1101.3,320.39 1105.45,325.07 1109.61,330.214 1113.77,335.808 1117.93,341.841 1122.09,348.305 
  1126.24,355.189 1130.4,362.483 1134.56,370.172 1138.72,378.245 1142.87,386.691 1147.03,395.497 1151.19,404.65 1155.35,414.136 1159.5,423.944 1163.66,434.061 
  1167.82,444.473 1171.98,455.166 1176.14,466.129 1180.29,477.347 1184.45,488.808 1188.61,500.498 1192.77,512.405 1196.92,524.513 1201.08,536.81 1205.24,549.283 
  1209.4,561.915 1213.55,574.697 1217.71,587.614 1221.87,600.652 1226.03,613.796 1230.19,627.033 1234.34,640.347 1238.5,653.725 1242.66,667.155 1246.82,680.626 
  1250.97,694.121 1255.13,707.626 1259.29,721.126 1263.45,734.609 1267.6,748.061 1271.76,761.471 1275.92,774.827 1280.08,788.113 1284.23,801.318 1288.39,814.429 
  1292.55,827.434 1296.71,840.32 1300.87,853.078 1305.02,865.694 1309.18,878.158 1313.34,890.46 1317.5,902.59 1321.65,914.536 1325.81,926.29 1329.97,937.841 
  1334.13,949.181 1338.28,960.302 1342.44,971.195 1346.6,981.855 1350.76,992.273 1354.92,1002.44 1359.07,1012.36 1363.23,1022.02 1367.39,1031.41 1371.55,1040.54 
  1375.7,1049.4 1379.86,1057.98 1384.02,1066.29 1388.18,1074.33 1392.33,1082.09 1396.49,1089.57 1400.65,1096.79 1404.81,1103.74 1408.96,1110.42 1413.12,1116.85 
  1417.28,1123.02 1421.44,1128.95 1425.6,1134.65 1429.75,1140.13 1433.91,1145.39 1438.07,1150.47 1442.23,1155.36 1446.38,1160.1 1450.54,1164.69 1454.7,959.396 
  1458.86,961.498 1463.01,963.251 1467.17,964.65 1471.33,965.694 1475.49,966.381 1479.65,966.712 1483.8,966.685 1487.96,966.303 1492.12,965.566 1496.28,964.475 
  1500.43,963.035 1504.59,961.247 1508.75,959.117 1512.91,956.647 1517.06,953.842 1521.22,950.708 1525.38,947.249 1529.54,943.471 1533.69,939.381 1537.85,934.984 
  1542.01,930.286 1546.17,925.299 1550.33,920.027 1554.48,914.476 1558.64,908.654 1562.8,902.567 1566.96,896.227 1571.11,889.64 1575.27,882.814 1579.43,875.759 
  1583.59,868.483 1587.74,860.996 1591.9,853.307 1596.06,845.424 1600.22,837.358 1604.38,829.118 1608.53,820.713 1612.69,812.153 1616.85,803.447 1621.01,794.606 
  1625.16,785.638 1629.32,776.553 1633.48,767.362 1637.64,758.074 1641.79,748.699 1645.95,739.247 1650.11,729.727 1654.27,720.149 1658.43,710.524 1662.58,700.861 
  1666.74,691.17 1670.9,681.461 1675.06,671.742 1679.21,662.024 1683.37,652.317 1687.53,642.629 1691.69,632.971 1695.84,623.35 1700,613.777 1704.16,604.26 
  1708.32,594.807 1712.47,585.428 1716.63,576.13 1720.79,566.923 1724.95,557.813 1729.11,548.809 1733.26,539.919 1737.42,531.148 1741.58,522.505 1745.74,513.997 
  1749.89,505.629 1754.05,497.409 1758.21,489.341 1762.37,481.432 1766.52,473.687 1770.68,466.111 1774.84,458.708 1779,451.483 1783.16,444.439 1787.31,437.58 
  1791.47,430.908 1795.63,424.427 1799.79,418.139 1803.94,412.045 1808.1,406.146 1812.26,400.442 1816.42,394.935 1820.57,389.624 1824.73,384.507 1828.89,379.584 
  1833.05,374.852 1837.2,370.308 1841.36,365.949 1845.52,361.771 1849.68,357.769 1853.84,353.938 1857.99,350.271 1862.15,346.762 1866.31,551.962 1870.47,550.739 
  1874.62,549.67 1878.78,548.762 1882.94,548.017 1887.1,547.438 1891.25,547.025 1895.41,546.782 1899.57,546.708 1903.73,546.805 1907.89,547.072 1912.04,547.512 
  1916.2,548.124 1920.36,548.907 1924.52,549.863 1928.67,550.989 1932.83,552.284 1936.99,553.748 1941.15,555.379 1945.3,557.177 1949.46,559.138 1953.62,561.26 
  1957.78,563.543 1961.93,565.984 1966.09,568.578 1970.25,571.325 1974.41,574.22 1978.57,577.263 1982.72,580.446 1986.88,583.768 1991.04,587.226 1995.2,590.817 
  1999.35,594.533 2003.51,598.373 2007.67,602.333 2011.83,606.408 2015.98,610.592 2020.14,614.883 2024.3,619.276 2028.46,623.766 2032.62,628.346 2036.77,633.014 
  2040.93,637.765 2045.09,642.592 2049.25,647.491 2053.4,652.457 2057.56,657.487 2061.72,662.571 2065.88,667.708 2070.03,672.891 2074.19,678.115 2078.35,683.375 
  2082.51,688.665 2086.67,693.981 2090.82,699.318 2094.98,704.669 2099.14,710.029 2103.3,715.395 2107.45,720.76 2111.61,726.12 2115.77,731.469 2119.93,736.802 
  2124.08,742.114 2128.24,747.402 2132.4,752.658 2136.56,757.88 2140.71,763.062 2144.87,768.199 2149.03,773.288 2153.19,778.323 2157.35,783.301 2161.5,788.217 
  2165.66,793.067 2169.82,797.847 2173.98,802.554 2178.13,807.183 2182.29,811.731 2186.45,816.195 2190.61,820.571 2194.76,824.855 2198.92,829.046 2203.08,833.14 
  2207.24,837.133 2211.4,841.025 2215.55,844.812 2219.71,848.491 2223.87,852.061 2228.03,855.52 2232.18,858.865 2236.34,862.096 2240.5,865.21 2244.66,868.206 
  2248.81,871.083 2252.97,873.84 2257.13,876.476 2261.29,878.991 2265.44,881.383 2269.6,883.651 2273.76,885.797 2277.92,887.818 2282.08,691.907 2286.23,691.907 
  2290.39,691.907 
  "/>
<polyline clip-path="url(#clip942)" style="stroke:#c271d2; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  211.556,587.758 215.714,593.143 219.871,598.099 224.029,602.647 228.187,606.81 232.344,610.611 236.502,614.074 240.66,617.221 244.817,620.075 248.975,622.657 
  253.133,624.988 257.29,627.09 261.448,628.98 265.606,630.677 269.763,632.198 273.921,633.558 278.079,634.773 282.236,635.855 286.394,636.817 290.552,637.671 
  294.709,638.427 298.867,639.096 303.025,639.685 307.182,640.203 311.34,640.657 315.498,641.054 319.655,641.4 323.813,641.701 327.971,641.962 332.128,642.186 
  336.286,642.379 340.444,642.544 344.601,642.684 348.759,642.803 352.917,642.902 357.074,642.985 361.232,643.054 365.39,643.111 369.547,643.157 373.705,643.194 
  377.863,643.223 382.02,643.247 386.178,643.265 390.336,643.278 394.493,643.289 398.651,643.296 402.809,643.302 406.966,643.305 411.124,643.308 415.282,643.309 
  419.439,643.31 423.597,643.31 427.755,643.309 431.912,643.307 436.07,643.304 440.228,643.3 444.385,643.294 448.543,643.286 452.701,643.275 456.858,643.26 
  461.016,643.241 465.174,643.216 469.331,643.185 473.489,643.145 477.647,643.097 481.804,643.037 485.962,642.964 490.12,642.877 494.277,642.772 498.435,642.648 
  502.593,642.5 506.75,642.327 510.908,642.125 515.066,641.889 519.223,641.616 523.381,641.299 527.539,640.935 531.696,640.517 535.854,640.037 540.012,639.49 
  544.17,638.865 548.327,638.155 552.485,637.348 556.643,636.434 560.8,635.399 564.958,634.229 569.116,632.909 573.273,631.42 577.431,629.744 581.589,627.857 
  585.746,625.736 589.904,623.353 594.062,620.677 598.219,617.674 602.377,614.306 606.535,610.529 610.692,606.295 614.85,601.547 619.008,596.226 623.165,590.26 
  627.323,793.28 631.481,788.984 635.638,784.972 639.796,781.234 643.954,777.759 648.111,774.536 652.269,771.552 656.427,768.795 660.584,766.252 664.742,763.911 
  668.9,761.759 673.057,759.784 677.215,757.975 681.373,756.32 685.53,754.808 689.688,753.428 693.846,752.171 698.003,751.028 702.161,749.988 706.319,749.045 
  710.476,748.191 714.634,747.417 718.792,746.717 722.949,746.085 727.107,745.515 731.265,745.002 735.422,744.541 739.58,744.126 743.738,743.754 747.895,743.421 
  752.053,743.123 756.211,742.858 760.368,742.621 764.526,742.41 768.684,742.223 772.841,742.058 776.999,741.912 781.157,741.783 785.314,741.67 789.472,741.571 
  793.63,741.485 797.787,741.411 801.945,741.347 806.103,741.293 810.26,741.248 814.418,741.21 818.576,741.181 822.733,741.158 826.891,741.142 831.049,741.132 
  835.206,741.129 839.364,741.132 843.522,741.141 847.679,741.157 851.837,741.179 855.995,741.209 860.152,741.245 864.31,741.29 868.468,741.344 872.625,741.408 
  876.783,741.482 880.941,741.568 885.098,741.666 889.256,741.779 893.414,741.908 897.571,742.055 901.729,742.221 905.887,742.408 910.044,742.62 914.202,742.859 
  918.36,743.128 922.517,743.429 926.675,743.767 930.833,744.145 934.99,744.567 939.148,745.039 943.306,745.566 947.463,746.152 951.621,746.805 955.779,747.532 
  959.936,748.339 964.094,749.235 968.252,750.229 972.41,751.331 976.567,752.553 980.725,753.905 984.883,755.402 989.04,757.057 993.198,758.887 997.356,760.909 
  1001.51,763.142 1005.67,765.608 1009.83,768.331 1013.99,771.335 1018.14,774.65 1022.3,778.308 1026.46,782.343 1030.62,786.794 1034.77,791.704 1038.93,588.434 
  1043.09,591.92 1047.25,595.234 1051.41,598.377 1055.56,601.353 1059.72,604.165 1063.88,606.818 1068.04,609.316 1072.19,611.665 1076.35,613.869 1080.51,615.936 
  1084.67,617.869 1088.82,619.677 1092.98,621.364 1097.14,622.936 1101.3,624.4 1105.45,625.762 1109.61,627.028 1113.77,628.202 1117.93,629.291 1122.09,630.3 
  1126.24,631.233 1130.4,632.095 1134.56,632.891 1138.72,633.626 1142.87,634.303 1147.03,634.927 1151.19,635.5 1155.35,636.027 1159.5,636.51 1163.66,636.954 
  1167.82,637.359 1171.98,637.73 1176.14,638.069 1180.29,638.378 1184.45,638.658 1188.61,638.913 1192.77,639.143 1196.92,639.351 1201.08,639.538 1205.24,639.704 
  1209.4,639.853 1213.55,639.983 1217.71,640.098 1221.87,640.197 1226.03,640.28 1230.19,640.35 1234.34,640.406 1238.5,640.448 1242.66,640.478 1246.82,640.495 
  1250.97,640.499 1255.13,640.491 1259.29,640.47 1263.45,640.436 1267.6,640.389 1271.76,640.328 1275.92,640.254 1280.08,640.164 1284.23,640.06 1288.39,639.939 
  1292.55,639.801 1296.71,639.645 1300.87,639.47 1305.02,639.273 1309.18,639.055 1313.34,638.812 1317.5,638.544 1321.65,638.247 1325.81,637.921 1329.97,637.561 
  1334.13,637.167 1338.28,636.734 1342.44,636.261 1346.6,635.742 1350.76,635.175 1354.92,634.556 1359.07,633.88 1363.23,633.142 1367.39,632.338 1371.55,631.461 
  1375.7,630.506 1379.86,629.466 1384.02,628.334 1388.18,627.102 1392.33,625.763 1396.49,624.307 1400.65,622.724 1404.81,621.004 1408.96,619.137 1413.12,617.109 
  1417.28,614.906 1421.44,612.516 1425.6,609.922 1429.75,607.107 1433.91,604.054 1438.07,600.742 1442.23,597.15 1446.38,593.253 1450.54,589.027 1454.7,794.27 
  1458.86,791.842 1463.01,789.494 1467.17,787.229 1471.33,785.047 1475.49,782.947 1479.65,780.931 1483.8,778.996 1487.96,777.144 1492.12,775.371 1496.28,773.678 
  1500.43,772.063 1504.59,770.525 1508.75,769.06 1512.91,767.668 1517.06,766.346 1521.22,765.093 1525.38,763.905 1529.54,762.78 1533.69,761.717 1537.85,760.713 
  1542.01,759.766 1546.17,758.873 1550.33,758.031 1554.48,757.24 1558.64,756.497 1562.8,755.799 1566.96,755.145 1571.11,754.533 1575.27,753.96 1579.43,753.426 
  1583.59,752.927 1587.74,752.464 1591.9,752.033 1596.06,751.634 1600.22,751.265 1604.38,750.925 1608.53,750.612 1612.69,750.326 1616.85,750.065 1621.01,749.829 
  1625.16,749.616 1629.32,749.426 1633.48,749.258 1637.64,749.112 1641.79,748.986 1645.95,748.88 1650.11,748.795 1654.27,748.729 1658.43,748.682 1662.58,748.655 
  1666.74,748.647 1670.9,748.658 1675.06,748.688 1679.21,748.738 1683.37,748.807 1687.53,748.897 1691.69,749.006 1695.84,749.137 1700,749.289 1704.16,749.464 
  1708.32,749.661 1712.47,749.883 1716.63,750.129 1720.79,750.401 1724.95,750.701 1729.11,751.029 1733.26,751.387 1737.42,751.777 1741.58,752.199 1745.74,752.657 
  1749.89,753.152 1754.05,753.687 1758.21,754.262 1762.37,754.882 1766.52,755.548 1770.68,756.264 1774.84,757.032 1779,757.856 1783.16,758.738 1787.31,759.684 
  1791.47,760.696 1795.63,761.778 1799.79,762.935 1803.94,764.173 1808.1,765.494 1812.26,766.906 1816.42,768.413 1820.57,770.021 1824.73,771.737 1828.89,773.567 
  1833.05,775.518 1837.2,777.598 1841.36,779.815 1845.52,782.177 1849.68,784.693 1853.84,787.372 1857.99,790.225 1862.15,793.262 1866.31,587.933 1870.47,589.251 
  1874.62,590.551 1878.78,591.832 1882.94,593.092 1887.1,594.331 1891.25,595.547 1895.41,596.739 1899.57,597.906 1903.73,599.047 1907.89,600.163 1912.04,601.251 
  1916.2,602.311 1920.36,603.343 1924.52,604.347 1928.67,605.322 1932.83,606.268 1936.99,607.184 1941.15,608.071 1945.3,608.928 1949.46,609.757 1953.62,610.555 
  1957.78,611.325 1961.93,612.065 1966.09,612.777 1970.25,613.461 1974.41,614.115 1978.57,614.742 1982.72,615.341 1986.88,615.913 1991.04,616.458 1995.2,616.976 
  1999.35,617.468 2003.51,617.934 2007.67,618.374 2011.83,618.789 2015.98,619.179 2020.14,619.545 2024.3,619.886 2028.46,620.203 2032.62,620.497 2036.77,620.767 
  2040.93,621.014 2045.09,621.239 2049.25,621.44 2053.4,621.619 2057.56,621.776 2061.72,621.911 2065.88,622.024 2070.03,622.114 2074.19,622.183 2078.35,622.229 
  2082.51,622.254 2086.67,622.257 2090.82,622.238 2094.98,622.196 2099.14,622.132 2103.3,622.045 2107.45,621.936 2111.61,621.803 2115.77,621.647 2119.93,621.468 
  2124.08,621.264 2128.24,621.035 2132.4,620.781 2136.56,620.502 2140.71,620.196 2144.87,619.864 2149.03,619.505 2153.19,619.117 2157.35,618.7 2161.5,618.254 
  2165.66,617.777 2169.82,617.269 2173.98,616.729 2178.13,616.156 2182.29,615.548 2186.45,614.905 2190.61,614.225 2194.76,613.508 2198.92,612.752 2203.08,611.955 
  2207.24,611.118 2211.4,610.237 2215.55,609.313 2219.71,608.342 2223.87,607.325 2228.03,606.258 2232.18,605.142 2236.34,603.973 2240.5,602.75 2244.66,601.473 
  2248.81,600.138 2252.97,598.744 2257.13,597.289 2261.29,595.771 2265.44,594.189 2269.6,592.54 2273.76,590.823 2277.92,589.035 2282.08,784.993 2286.23,784.993 
  2290.39,784.993 
  "/>
<path clip-path="url(#clip940)" d="
M1390.67 352.309 L2279.3 352.309 L2279.3 93.1086 L1390.67 93.1086  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<polyline clip-path="url(#clip940)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1390.67,352.309 2279.3,352.309 2279.3,93.1086 1390.67,93.1086 1390.67,352.309 
  "/>
<polyline clip-path="url(#clip940)" style="stroke:#009af9; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1415.15,144.949 1562.06,144.949 
  "/>
<path clip-path="url(#clip940)" d="M1606.89 141.28 Q1608.49 138.409 1610.71 137.044 Q1612.93 135.678 1615.94 135.678 Q1619.99 135.678 1622.19 138.525 Q1624.39 141.349 1624.39 146.581 L1624.39 162.229 L1620.11 162.229 L1620.11 146.719 Q1620.11 142.993 1618.79 141.187 Q1617.47 139.382 1614.76 139.382 Q1611.45 139.382 1609.53 141.581 Q1607.61 143.78 1607.61 147.576 L1607.61 162.229 L1603.33 162.229 L1603.33 146.719 Q1603.33 142.969 1602.01 141.187 Q1600.69 139.382 1597.93 139.382 Q1594.67 139.382 1592.75 141.604 Q1590.83 143.803 1590.83 147.576 L1590.83 162.229 L1586.54 162.229 L1586.54 136.303 L1590.83 136.303 L1590.83 140.331 Q1592.28 137.946 1594.32 136.812 Q1596.36 135.678 1599.16 135.678 Q1601.98 135.678 1603.95 137.113 Q1605.94 138.548 1606.89 141.28 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1644.67 149.196 Q1639.51 149.196 1637.51 150.377 Q1635.52 151.557 1635.52 154.405 Q1635.52 156.673 1637.01 158.016 Q1638.51 159.335 1641.08 159.335 Q1644.62 159.335 1646.75 156.835 Q1648.9 154.312 1648.9 150.145 L1648.9 149.196 L1644.67 149.196 M1653.16 147.437 L1653.16 162.229 L1648.9 162.229 L1648.9 158.293 Q1647.45 160.655 1645.27 161.789 Q1643.09 162.9 1639.95 162.9 Q1635.96 162.9 1633.6 160.678 Q1631.26 158.432 1631.26 154.682 Q1631.26 150.307 1634.18 148.085 Q1637.12 145.863 1642.93 145.863 L1648.9 145.863 L1648.9 145.446 Q1648.9 142.507 1646.96 140.909 Q1645.04 139.289 1641.54 139.289 Q1639.32 139.289 1637.21 139.821 Q1635.11 140.354 1633.16 141.419 L1633.16 137.483 Q1635.5 136.581 1637.7 136.141 Q1639.9 135.678 1641.98 135.678 Q1647.61 135.678 1650.39 138.594 Q1653.16 141.511 1653.16 147.437 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1678.46 137.067 L1678.46 141.094 Q1676.66 140.169 1674.71 139.706 Q1672.77 139.243 1670.69 139.243 Q1667.51 139.243 1665.92 140.215 Q1664.34 141.187 1664.34 143.131 Q1664.34 144.613 1665.48 145.469 Q1666.61 146.303 1670.04 147.067 L1671.5 147.391 Q1676.03 148.363 1677.93 150.145 Q1679.85 151.905 1679.85 155.076 Q1679.85 158.687 1676.98 160.793 Q1674.13 162.9 1669.13 162.9 Q1667.05 162.9 1664.78 162.483 Q1662.54 162.09 1660.04 161.28 L1660.04 156.881 Q1662.4 158.108 1664.69 158.733 Q1666.98 159.335 1669.23 159.335 Q1672.24 159.335 1673.86 158.317 Q1675.48 157.275 1675.48 155.4 Q1675.48 153.664 1674.3 152.738 Q1673.14 151.812 1669.18 150.956 L1667.7 150.608 Q1663.74 149.775 1661.98 148.062 Q1660.22 146.326 1660.22 143.317 Q1660.22 139.659 1662.82 137.669 Q1665.41 135.678 1670.18 135.678 Q1672.54 135.678 1674.62 136.025 Q1676.7 136.372 1678.46 137.067 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1703.16 137.067 L1703.16 141.094 Q1701.36 140.169 1699.41 139.706 Q1697.47 139.243 1695.38 139.243 Q1692.21 139.243 1690.62 140.215 Q1689.04 141.187 1689.04 143.131 Q1689.04 144.613 1690.18 145.469 Q1691.31 146.303 1694.74 147.067 L1696.19 147.391 Q1700.73 148.363 1702.63 150.145 Q1704.55 151.905 1704.55 155.076 Q1704.55 158.687 1701.68 160.793 Q1698.83 162.9 1693.83 162.9 Q1691.75 162.9 1689.48 162.483 Q1687.24 162.09 1684.74 161.28 L1684.74 156.881 Q1687.1 158.108 1689.39 158.733 Q1691.68 159.335 1693.93 159.335 Q1696.94 159.335 1698.56 158.317 Q1700.18 157.275 1700.18 155.4 Q1700.18 153.664 1699 152.738 Q1697.84 151.812 1693.88 150.956 L1692.4 150.608 Q1688.44 149.775 1686.68 148.062 Q1684.92 146.326 1684.92 143.317 Q1684.92 139.659 1687.51 137.669 Q1690.11 135.678 1694.88 135.678 Q1697.24 135.678 1699.32 136.025 Q1701.4 136.372 1703.16 137.067 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1711.94 156.349 L1716.82 156.349 L1716.82 162.229 L1711.94 162.229 L1711.94 156.349 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1742.93 137.067 L1742.93 141.094 Q1741.13 140.169 1739.18 139.706 Q1737.24 139.243 1735.15 139.243 Q1731.98 139.243 1730.38 140.215 Q1728.81 141.187 1728.81 143.131 Q1728.81 144.613 1729.94 145.469 Q1731.08 146.303 1734.5 147.067 L1735.96 147.391 Q1740.5 148.363 1742.4 150.145 Q1744.32 151.905 1744.32 155.076 Q1744.32 158.687 1741.45 160.793 Q1738.6 162.9 1733.6 162.9 Q1731.52 162.9 1729.25 162.483 Q1727 162.09 1724.51 161.28 L1724.51 156.881 Q1726.87 158.108 1729.16 158.733 Q1731.45 159.335 1733.69 159.335 Q1736.7 159.335 1738.32 158.317 Q1739.94 157.275 1739.94 155.4 Q1739.94 153.664 1738.76 152.738 Q1737.61 151.812 1733.65 150.956 L1732.17 150.608 Q1728.21 149.775 1726.45 148.062 Q1724.69 146.326 1724.69 143.317 Q1724.69 139.659 1727.28 137.669 Q1729.88 135.678 1734.64 135.678 Q1737 135.678 1739.09 136.025 Q1741.17 136.372 1742.93 137.067 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1776.4 126.257 Q1773.3 131.581 1771.8 136.789 Q1770.29 141.997 1770.29 147.344 Q1770.29 152.692 1771.8 157.946 Q1773.32 163.178 1776.4 168.479 L1772.7 168.479 Q1769.23 163.039 1767.49 157.784 Q1765.78 152.53 1765.78 147.344 Q1765.78 142.182 1767.49 136.951 Q1769.2 131.72 1772.7 126.257 L1776.4 126.257 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1799.44 143.594 Q1802.79 144.312 1804.67 146.581 Q1806.56 148.849 1806.56 152.182 Q1806.56 157.298 1803.05 160.099 Q1799.53 162.9 1793.05 162.9 Q1790.87 162.9 1788.56 162.46 Q1786.26 162.043 1783.81 161.187 L1783.81 156.673 Q1785.75 157.807 1788.07 158.386 Q1790.38 158.965 1792.91 158.965 Q1797.31 158.965 1799.6 157.229 Q1801.91 155.493 1801.91 152.182 Q1801.91 149.127 1799.76 147.414 Q1797.63 145.678 1793.81 145.678 L1789.78 145.678 L1789.78 141.835 L1794 141.835 Q1797.44 141.835 1799.27 140.469 Q1801.1 139.081 1801.1 136.488 Q1801.1 133.826 1799.2 132.414 Q1797.33 130.979 1793.81 130.979 Q1791.89 130.979 1789.69 131.395 Q1787.49 131.812 1784.85 132.692 L1784.85 128.525 Q1787.51 127.784 1789.83 127.414 Q1792.17 127.044 1794.23 127.044 Q1799.55 127.044 1802.65 129.474 Q1805.75 131.882 1805.75 136.002 Q1805.75 138.872 1804.11 140.863 Q1802.47 142.831 1799.44 143.594 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1829.6 143.594 Q1832.95 144.312 1834.83 146.581 Q1836.73 148.849 1836.73 152.182 Q1836.73 157.298 1833.21 160.099 Q1829.69 162.9 1823.21 162.9 Q1821.03 162.9 1818.72 162.46 Q1816.43 162.043 1813.97 161.187 L1813.97 156.673 Q1815.92 157.807 1818.23 158.386 Q1820.55 158.965 1823.07 158.965 Q1827.47 158.965 1829.76 157.229 Q1832.07 155.493 1832.07 152.182 Q1832.07 149.127 1829.92 147.414 Q1827.79 145.678 1823.97 145.678 L1819.94 145.678 L1819.94 141.835 L1824.16 141.835 Q1827.61 141.835 1829.43 140.469 Q1831.26 139.081 1831.26 136.488 Q1831.26 133.826 1829.37 132.414 Q1827.49 130.979 1823.97 130.979 Q1822.05 130.979 1819.85 131.395 Q1817.65 131.812 1815.01 132.692 L1815.01 128.525 Q1817.68 127.784 1819.99 127.414 Q1822.33 127.044 1824.39 127.044 Q1829.71 127.044 1832.81 129.474 Q1835.92 131.882 1835.92 136.002 Q1835.92 138.872 1834.27 140.863 Q1832.63 142.831 1829.6 143.594 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1845.64 127.669 L1863.99 127.669 L1863.99 131.604 L1849.92 131.604 L1849.92 140.076 Q1850.94 139.729 1851.96 139.567 Q1852.98 139.382 1854 139.382 Q1859.78 139.382 1863.16 142.553 Q1866.54 145.724 1866.54 151.141 Q1866.54 156.719 1863.07 159.821 Q1859.6 162.9 1853.28 162.9 Q1851.1 162.9 1848.83 162.53 Q1846.59 162.159 1844.18 161.418 L1844.18 156.719 Q1846.26 157.854 1848.49 158.409 Q1850.71 158.965 1853.18 158.965 Q1857.19 158.965 1859.53 156.858 Q1861.87 154.752 1861.87 151.141 Q1861.87 147.53 1859.53 145.423 Q1857.19 143.317 1853.18 143.317 Q1851.31 143.317 1849.43 143.733 Q1847.58 144.15 1845.64 145.03 L1845.64 127.669 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1875.8 127.669 L1894.16 127.669 L1894.16 131.604 L1880.08 131.604 L1880.08 140.076 Q1881.1 139.729 1882.12 139.567 Q1883.14 139.382 1884.16 139.382 Q1889.94 139.382 1893.32 142.553 Q1896.7 145.724 1896.7 151.141 Q1896.7 156.719 1893.23 159.821 Q1889.76 162.9 1883.44 162.9 Q1881.26 162.9 1878.99 162.53 Q1876.75 162.159 1874.34 161.418 L1874.34 156.719 Q1876.43 157.854 1878.65 158.409 Q1880.87 158.965 1883.35 158.965 Q1887.35 158.965 1889.69 156.858 Q1892.03 154.752 1892.03 151.141 Q1892.03 147.53 1889.69 145.423 Q1887.35 143.317 1883.35 143.317 Q1881.47 143.317 1879.6 143.733 Q1877.74 144.15 1875.8 145.03 L1875.8 127.669 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1918.76 131.743 L1906.96 150.192 L1918.76 150.192 L1918.76 131.743 M1917.54 127.669 L1923.42 127.669 L1923.42 150.192 L1928.35 150.192 L1928.35 154.081 L1923.42 154.081 L1923.42 162.229 L1918.76 162.229 L1918.76 154.081 L1903.16 154.081 L1903.16 149.567 L1917.54 127.669 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1948.93 131.743 L1937.12 150.192 L1948.93 150.192 L1948.93 131.743 M1947.7 127.669 L1953.58 127.669 L1953.58 150.192 L1958.51 150.192 L1958.51 154.081 L1953.58 154.081 L1953.58 162.229 L1948.93 162.229 L1948.93 154.081 L1933.32 154.081 L1933.32 149.567 L1947.7 127.669 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1980.41 143.594 Q1983.76 144.312 1985.64 146.581 Q1987.54 148.849 1987.54 152.182 Q1987.54 157.298 1984.02 160.099 Q1980.5 162.9 1974.02 162.9 Q1971.84 162.9 1969.53 162.46 Q1967.24 162.043 1964.78 161.187 L1964.78 156.673 Q1966.73 157.807 1969.04 158.386 Q1971.36 158.965 1973.88 158.965 Q1978.28 158.965 1980.57 157.229 Q1982.88 155.493 1982.88 152.182 Q1982.88 149.127 1980.73 147.414 Q1978.6 145.678 1974.78 145.678 L1970.75 145.678 L1970.75 141.835 L1974.97 141.835 Q1978.42 141.835 1980.24 140.469 Q1982.07 139.081 1982.07 136.488 Q1982.07 133.826 1980.17 132.414 Q1978.3 130.979 1974.78 130.979 Q1972.86 130.979 1970.66 131.395 Q1968.46 131.812 1965.82 132.692 L1965.82 128.525 Q1968.49 127.784 1970.8 127.414 Q1973.14 127.044 1975.2 127.044 Q1980.52 127.044 1983.62 129.474 Q1986.73 131.882 1986.73 136.002 Q1986.73 138.872 1985.08 140.863 Q1983.44 142.831 1980.41 143.594 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M2000.43 158.293 L2016.75 158.293 L2016.75 162.229 L1994.8 162.229 L1994.8 158.293 Q1997.47 155.539 2002.05 150.909 Q2006.66 146.256 2007.84 144.914 Q2010.08 142.391 2010.96 140.655 Q2011.86 138.895 2011.86 137.206 Q2011.86 134.451 2009.92 132.715 Q2008 130.979 2004.9 130.979 Q2002.7 130.979 2000.24 131.743 Q1997.81 132.507 1995.04 134.057 L1995.04 129.335 Q1997.86 128.201 2000.31 127.622 Q2002.77 127.044 2004.8 127.044 Q2010.17 127.044 2013.37 129.729 Q2016.56 132.414 2016.56 136.905 Q2016.56 139.034 2015.75 140.956 Q2014.97 142.854 2012.86 145.446 Q2012.28 146.118 2009.18 149.335 Q2006.08 152.53 2000.43 158.293 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M2025.29 126.257 L2028.99 126.257 Q2032.47 131.72 2034.18 136.951 Q2035.92 142.182 2035.92 147.344 Q2035.92 152.53 2034.18 157.784 Q2032.47 163.039 2028.99 168.479 L2025.29 168.479 Q2028.37 163.178 2029.87 157.946 Q2031.4 152.692 2031.4 147.344 Q2031.4 141.997 2029.87 136.789 Q2028.37 131.581 2025.29 126.257 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip940)" style="stroke:#e26f46; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1415.15,196.789 1562.06,196.789 
  "/>
<path clip-path="url(#clip940)" d="M1606.89 193.12 Q1608.49 190.249 1610.71 188.884 Q1612.93 187.518 1615.94 187.518 Q1619.99 187.518 1622.19 190.365 Q1624.39 193.189 1624.39 198.421 L1624.39 214.069 L1620.11 214.069 L1620.11 198.559 Q1620.11 194.833 1618.79 193.027 Q1617.47 191.222 1614.76 191.222 Q1611.45 191.222 1609.53 193.421 Q1607.61 195.62 1607.61 199.416 L1607.61 214.069 L1603.33 214.069 L1603.33 198.559 Q1603.33 194.809 1602.01 193.027 Q1600.69 191.222 1597.93 191.222 Q1594.67 191.222 1592.75 193.444 Q1590.83 195.643 1590.83 199.416 L1590.83 214.069 L1586.54 214.069 L1586.54 188.143 L1590.83 188.143 L1590.83 192.171 Q1592.28 189.786 1594.32 188.652 Q1596.36 187.518 1599.16 187.518 Q1601.98 187.518 1603.95 188.953 Q1605.94 190.388 1606.89 193.12 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1644.67 201.036 Q1639.51 201.036 1637.51 202.217 Q1635.52 203.397 1635.52 206.245 Q1635.52 208.513 1637.01 209.856 Q1638.51 211.175 1641.08 211.175 Q1644.62 211.175 1646.75 208.675 Q1648.9 206.152 1648.9 201.985 L1648.9 201.036 L1644.67 201.036 M1653.16 199.277 L1653.16 214.069 L1648.9 214.069 L1648.9 210.133 Q1647.45 212.495 1645.27 213.629 Q1643.09 214.74 1639.95 214.74 Q1635.96 214.74 1633.6 212.518 Q1631.26 210.272 1631.26 206.522 Q1631.26 202.147 1634.18 199.925 Q1637.12 197.703 1642.93 197.703 L1648.9 197.703 L1648.9 197.286 Q1648.9 194.347 1646.96 192.749 Q1645.04 191.129 1641.54 191.129 Q1639.32 191.129 1637.21 191.661 Q1635.11 192.194 1633.16 193.259 L1633.16 189.323 Q1635.5 188.421 1637.7 187.981 Q1639.9 187.518 1641.98 187.518 Q1647.61 187.518 1650.39 190.434 Q1653.16 193.351 1653.16 199.277 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1678.46 188.907 L1678.46 192.934 Q1676.66 192.009 1674.71 191.546 Q1672.77 191.083 1670.69 191.083 Q1667.51 191.083 1665.92 192.055 Q1664.34 193.027 1664.34 194.971 Q1664.34 196.453 1665.48 197.309 Q1666.61 198.143 1670.04 198.907 L1671.5 199.231 Q1676.03 200.203 1677.93 201.985 Q1679.85 203.745 1679.85 206.916 Q1679.85 210.527 1676.98 212.633 Q1674.13 214.74 1669.13 214.74 Q1667.05 214.74 1664.78 214.323 Q1662.54 213.93 1660.04 213.12 L1660.04 208.721 Q1662.4 209.948 1664.69 210.573 Q1666.98 211.175 1669.23 211.175 Q1672.24 211.175 1673.86 210.157 Q1675.48 209.115 1675.48 207.24 Q1675.48 205.504 1674.3 204.578 Q1673.14 203.652 1669.18 202.796 L1667.7 202.448 Q1663.74 201.615 1661.98 199.902 Q1660.22 198.166 1660.22 195.157 Q1660.22 191.499 1662.82 189.509 Q1665.41 187.518 1670.18 187.518 Q1672.54 187.518 1674.62 187.865 Q1676.7 188.212 1678.46 188.907 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1703.16 188.907 L1703.16 192.934 Q1701.36 192.009 1699.41 191.546 Q1697.47 191.083 1695.38 191.083 Q1692.21 191.083 1690.62 192.055 Q1689.04 193.027 1689.04 194.971 Q1689.04 196.453 1690.18 197.309 Q1691.31 198.143 1694.74 198.907 L1696.19 199.231 Q1700.73 200.203 1702.63 201.985 Q1704.55 203.745 1704.55 206.916 Q1704.55 210.527 1701.68 212.633 Q1698.83 214.74 1693.83 214.74 Q1691.75 214.74 1689.48 214.323 Q1687.24 213.93 1684.74 213.12 L1684.74 208.721 Q1687.1 209.948 1689.39 210.573 Q1691.68 211.175 1693.93 211.175 Q1696.94 211.175 1698.56 210.157 Q1700.18 209.115 1700.18 207.24 Q1700.18 205.504 1699 204.578 Q1697.84 203.652 1693.88 202.796 L1692.4 202.448 Q1688.44 201.615 1686.68 199.902 Q1684.92 198.166 1684.92 195.157 Q1684.92 191.499 1687.51 189.509 Q1690.11 187.518 1694.88 187.518 Q1697.24 187.518 1699.32 187.865 Q1701.4 188.212 1703.16 188.907 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1711.94 208.189 L1716.82 208.189 L1716.82 214.069 L1711.94 214.069 L1711.94 208.189 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1723.35 188.143 L1727.86 188.143 L1735.96 209.902 L1744.07 188.143 L1748.58 188.143 L1738.86 214.069 L1733.07 214.069 L1723.35 188.143 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1779.76 178.097 Q1776.66 183.421 1775.15 188.629 Q1773.65 193.837 1773.65 199.184 Q1773.65 204.532 1775.15 209.786 Q1776.68 215.018 1779.76 220.319 L1776.06 220.319 Q1772.58 214.879 1770.85 209.624 Q1769.13 204.37 1769.13 199.184 Q1769.13 194.022 1770.85 188.791 Q1772.56 183.56 1776.06 178.097 L1779.76 178.097 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1802.79 195.434 Q1806.15 196.152 1808.02 198.421 Q1809.92 200.689 1809.92 204.022 Q1809.92 209.138 1806.4 211.939 Q1802.88 214.74 1796.4 214.74 Q1794.23 214.74 1791.91 214.3 Q1789.62 213.883 1787.17 213.027 L1787.17 208.513 Q1789.11 209.647 1791.43 210.226 Q1793.74 210.805 1796.26 210.805 Q1800.66 210.805 1802.95 209.069 Q1805.27 207.333 1805.27 204.022 Q1805.27 200.967 1803.12 199.254 Q1800.99 197.518 1797.17 197.518 L1793.14 197.518 L1793.14 193.675 L1797.35 193.675 Q1800.8 193.675 1802.63 192.309 Q1804.46 190.921 1804.46 188.328 Q1804.46 185.666 1802.56 184.254 Q1800.69 182.819 1797.17 182.819 Q1795.25 182.819 1793.05 183.235 Q1790.85 183.652 1788.21 184.532 L1788.21 180.365 Q1790.87 179.624 1793.19 179.254 Q1795.52 178.884 1797.58 178.884 Q1802.91 178.884 1806.01 181.314 Q1809.11 183.722 1809.11 187.842 Q1809.11 190.712 1807.47 192.703 Q1805.82 194.671 1802.79 195.434 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1832.95 195.434 Q1836.31 196.152 1838.18 198.421 Q1840.08 200.689 1840.08 204.022 Q1840.08 209.138 1836.56 211.939 Q1833.05 214.74 1826.56 214.74 Q1824.39 214.74 1822.07 214.3 Q1819.78 213.883 1817.33 213.027 L1817.33 208.513 Q1819.27 209.647 1821.59 210.226 Q1823.9 210.805 1826.43 210.805 Q1830.82 210.805 1833.12 209.069 Q1835.43 207.333 1835.43 204.022 Q1835.43 200.967 1833.28 199.254 Q1831.15 197.518 1827.33 197.518 L1823.3 197.518 L1823.3 193.675 L1827.51 193.675 Q1830.96 193.675 1832.79 192.309 Q1834.62 190.921 1834.62 188.328 Q1834.62 185.666 1832.72 184.254 Q1830.85 182.819 1827.33 182.819 Q1825.41 182.819 1823.21 183.235 Q1821.01 183.652 1818.37 184.532 L1818.37 180.365 Q1821.03 179.624 1823.35 179.254 Q1825.69 178.884 1827.75 178.884 Q1833.07 178.884 1836.17 181.314 Q1839.27 183.722 1839.27 187.842 Q1839.27 190.712 1837.63 192.703 Q1835.99 194.671 1832.95 195.434 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1849 179.509 L1867.35 179.509 L1867.35 183.444 L1853.28 183.444 L1853.28 191.916 Q1854.3 191.569 1855.31 191.407 Q1856.33 191.222 1857.35 191.222 Q1863.14 191.222 1866.52 194.393 Q1869.9 197.564 1869.9 202.981 Q1869.9 208.559 1866.43 211.661 Q1862.95 214.74 1856.63 214.74 Q1854.46 214.74 1852.19 214.37 Q1849.94 213.999 1847.54 213.258 L1847.54 208.559 Q1849.62 209.694 1851.84 210.249 Q1854.06 210.805 1856.54 210.805 Q1860.55 210.805 1862.88 208.698 Q1865.22 206.592 1865.22 202.981 Q1865.22 199.37 1862.88 197.263 Q1860.55 195.157 1856.54 195.157 Q1854.67 195.157 1852.79 195.573 Q1850.94 195.99 1849 196.87 L1849 179.509 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1879.16 179.509 L1897.51 179.509 L1897.51 183.444 L1883.44 183.444 L1883.44 191.916 Q1884.46 191.569 1885.48 191.407 Q1886.49 191.222 1887.51 191.222 Q1893.3 191.222 1896.68 194.393 Q1900.06 197.564 1900.06 202.981 Q1900.06 208.559 1896.59 211.661 Q1893.12 214.74 1886.8 214.74 Q1884.62 214.74 1882.35 214.37 Q1880.11 213.999 1877.7 213.258 L1877.7 208.559 Q1879.78 209.694 1882 210.249 Q1884.23 210.805 1886.7 210.805 Q1890.71 210.805 1893.05 208.698 Q1895.38 206.592 1895.38 202.981 Q1895.38 199.37 1893.05 197.263 Q1890.71 195.157 1886.7 195.157 Q1884.83 195.157 1882.95 195.573 Q1881.1 195.99 1879.16 196.87 L1879.16 179.509 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1922.12 183.583 L1910.31 202.032 L1922.12 202.032 L1922.12 183.583 M1920.89 179.509 L1926.77 179.509 L1926.77 202.032 L1931.7 202.032 L1931.7 205.921 L1926.77 205.921 L1926.77 214.069 L1922.12 214.069 L1922.12 205.921 L1906.52 205.921 L1906.52 201.407 L1920.89 179.509 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1952.28 183.583 L1940.48 202.032 L1952.28 202.032 L1952.28 183.583 M1951.05 179.509 L1956.93 179.509 L1956.93 202.032 L1961.86 202.032 L1961.86 205.921 L1956.93 205.921 L1956.93 214.069 L1952.28 214.069 L1952.28 205.921 L1936.68 205.921 L1936.68 201.407 L1951.05 179.509 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1983.76 195.434 Q1987.12 196.152 1988.99 198.421 Q1990.89 200.689 1990.89 204.022 Q1990.89 209.138 1987.37 211.939 Q1983.86 214.74 1977.37 214.74 Q1975.2 214.74 1972.88 214.3 Q1970.59 213.883 1968.14 213.027 L1968.14 208.513 Q1970.08 209.647 1972.4 210.226 Q1974.71 210.805 1977.24 210.805 Q1981.63 210.805 1983.92 209.069 Q1986.24 207.333 1986.24 204.022 Q1986.24 200.967 1984.09 199.254 Q1981.96 197.518 1978.14 197.518 L1974.11 197.518 L1974.11 193.675 L1978.32 193.675 Q1981.77 193.675 1983.6 192.309 Q1985.43 190.921 1985.43 188.328 Q1985.43 185.666 1983.53 184.254 Q1981.66 182.819 1978.14 182.819 Q1976.22 182.819 1974.02 183.235 Q1971.82 183.652 1969.18 184.532 L1969.18 180.365 Q1971.84 179.624 1974.16 179.254 Q1976.49 178.884 1978.55 178.884 Q1983.88 178.884 1986.98 181.314 Q1990.08 183.722 1990.08 187.842 Q1990.08 190.712 1988.44 192.703 Q1986.8 194.671 1983.76 195.434 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M2013.92 195.434 Q2017.28 196.152 2019.16 198.421 Q2021.05 200.689 2021.05 204.022 Q2021.05 209.138 2017.54 211.939 Q2014.02 214.74 2007.54 214.74 Q2005.36 214.74 2003.05 214.3 Q2000.75 213.883 1998.3 213.027 L1998.3 208.513 Q2000.24 209.647 2002.56 210.226 Q2004.87 210.805 2007.4 210.805 Q2011.8 210.805 2014.09 209.069 Q2016.4 207.333 2016.4 204.022 Q2016.4 200.967 2014.25 199.254 Q2012.12 197.518 2008.3 197.518 L2004.27 197.518 L2004.27 193.675 L2008.48 193.675 Q2011.93 193.675 2013.76 192.309 Q2015.59 190.921 2015.59 188.328 Q2015.59 185.666 2013.69 184.254 Q2011.82 182.819 2008.3 182.819 Q2006.38 182.819 2004.18 183.235 Q2001.98 183.652 1999.34 184.532 L1999.34 180.365 Q2002 179.624 2004.32 179.254 Q2006.66 178.884 2008.72 178.884 Q2014.04 178.884 2017.14 181.314 Q2020.24 183.722 2020.24 187.842 Q2020.24 190.712 2018.6 192.703 Q2016.96 194.671 2013.92 195.434 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M2028.65 178.097 L2032.35 178.097 Q2035.82 183.56 2037.54 188.791 Q2039.27 194.022 2039.27 199.184 Q2039.27 204.37 2037.54 209.624 Q2035.82 214.879 2032.35 220.319 L2028.65 220.319 Q2031.73 215.018 2033.23 209.786 Q2034.76 204.532 2034.76 199.184 Q2034.76 193.837 2033.23 188.629 Q2031.73 183.421 2028.65 178.097 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip940)" style="stroke:#3da44d; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1415.15,248.629 1562.06,248.629 
  "/>
<path clip-path="url(#clip940)" d="M1606.89 244.96 Q1608.49 242.089 1610.71 240.724 Q1612.93 239.358 1615.94 239.358 Q1619.99 239.358 1622.19 242.205 Q1624.39 245.029 1624.39 250.261 L1624.39 265.909 L1620.11 265.909 L1620.11 250.399 Q1620.11 246.673 1618.79 244.867 Q1617.47 243.062 1614.76 243.062 Q1611.45 243.062 1609.53 245.261 Q1607.61 247.46 1607.61 251.256 L1607.61 265.909 L1603.33 265.909 L1603.33 250.399 Q1603.33 246.649 1602.01 244.867 Q1600.69 243.062 1597.93 243.062 Q1594.67 243.062 1592.75 245.284 Q1590.83 247.483 1590.83 251.256 L1590.83 265.909 L1586.54 265.909 L1586.54 239.983 L1590.83 239.983 L1590.83 244.011 Q1592.28 241.626 1594.32 240.492 Q1596.36 239.358 1599.16 239.358 Q1601.98 239.358 1603.95 240.793 Q1605.94 242.228 1606.89 244.96 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1644.67 252.876 Q1639.51 252.876 1637.51 254.057 Q1635.52 255.237 1635.52 258.085 Q1635.52 260.353 1637.01 261.696 Q1638.51 263.015 1641.08 263.015 Q1644.62 263.015 1646.75 260.515 Q1648.9 257.992 1648.9 253.825 L1648.9 252.876 L1644.67 252.876 M1653.16 251.117 L1653.16 265.909 L1648.9 265.909 L1648.9 261.973 Q1647.45 264.335 1645.27 265.469 Q1643.09 266.58 1639.95 266.58 Q1635.96 266.58 1633.6 264.358 Q1631.26 262.112 1631.26 258.362 Q1631.26 253.987 1634.18 251.765 Q1637.12 249.543 1642.93 249.543 L1648.9 249.543 L1648.9 249.126 Q1648.9 246.187 1646.96 244.589 Q1645.04 242.969 1641.54 242.969 Q1639.32 242.969 1637.21 243.501 Q1635.11 244.034 1633.16 245.099 L1633.16 241.163 Q1635.5 240.261 1637.7 239.821 Q1639.9 239.358 1641.98 239.358 Q1647.61 239.358 1650.39 242.274 Q1653.16 245.191 1653.16 251.117 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1678.46 240.747 L1678.46 244.774 Q1676.66 243.849 1674.71 243.386 Q1672.77 242.923 1670.69 242.923 Q1667.51 242.923 1665.92 243.895 Q1664.34 244.867 1664.34 246.811 Q1664.34 248.293 1665.48 249.149 Q1666.61 249.983 1670.04 250.747 L1671.5 251.071 Q1676.03 252.043 1677.93 253.825 Q1679.85 255.585 1679.85 258.756 Q1679.85 262.367 1676.98 264.473 Q1674.13 266.58 1669.13 266.58 Q1667.05 266.58 1664.78 266.163 Q1662.54 265.77 1660.04 264.96 L1660.04 260.561 Q1662.4 261.788 1664.69 262.413 Q1666.98 263.015 1669.23 263.015 Q1672.24 263.015 1673.86 261.997 Q1675.48 260.955 1675.48 259.08 Q1675.48 257.344 1674.3 256.418 Q1673.14 255.492 1669.18 254.636 L1667.7 254.288 Q1663.74 253.455 1661.98 251.742 Q1660.22 250.006 1660.22 246.997 Q1660.22 243.339 1662.82 241.349 Q1665.41 239.358 1670.18 239.358 Q1672.54 239.358 1674.62 239.705 Q1676.7 240.052 1678.46 240.747 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1703.16 240.747 L1703.16 244.774 Q1701.36 243.849 1699.41 243.386 Q1697.47 242.923 1695.38 242.923 Q1692.21 242.923 1690.62 243.895 Q1689.04 244.867 1689.04 246.811 Q1689.04 248.293 1690.18 249.149 Q1691.31 249.983 1694.74 250.747 L1696.19 251.071 Q1700.73 252.043 1702.63 253.825 Q1704.55 255.585 1704.55 258.756 Q1704.55 262.367 1701.68 264.473 Q1698.83 266.58 1693.83 266.58 Q1691.75 266.58 1689.48 266.163 Q1687.24 265.77 1684.74 264.96 L1684.74 260.561 Q1687.1 261.788 1689.39 262.413 Q1691.68 263.015 1693.93 263.015 Q1696.94 263.015 1698.56 261.997 Q1700.18 260.955 1700.18 259.08 Q1700.18 257.344 1699 256.418 Q1697.84 255.492 1693.88 254.636 L1692.4 254.288 Q1688.44 253.455 1686.68 251.742 Q1684.92 250.006 1684.92 246.997 Q1684.92 243.339 1687.51 241.349 Q1690.11 239.358 1694.88 239.358 Q1697.24 239.358 1699.32 239.705 Q1701.4 240.052 1703.16 240.747 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1711.94 260.029 L1716.82 260.029 L1716.82 265.909 L1711.94 265.909 L1711.94 260.029 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1738.19 252.876 Q1733.02 252.876 1731.03 254.057 Q1729.04 255.237 1729.04 258.085 Q1729.04 260.353 1730.52 261.696 Q1732.03 263.015 1734.6 263.015 Q1738.14 263.015 1740.27 260.515 Q1742.42 257.992 1742.42 253.825 L1742.42 252.876 L1738.19 252.876 M1746.68 251.117 L1746.68 265.909 L1742.42 265.909 L1742.42 261.973 Q1740.96 264.335 1738.79 265.469 Q1736.61 266.58 1733.46 266.58 Q1729.48 266.58 1727.12 264.358 Q1724.78 262.112 1724.78 258.362 Q1724.78 253.987 1727.7 251.765 Q1730.64 249.543 1736.45 249.543 L1742.42 249.543 L1742.42 249.126 Q1742.42 246.187 1740.48 244.589 Q1738.56 242.969 1735.06 242.969 Q1732.84 242.969 1730.73 243.501 Q1728.63 244.034 1726.68 245.099 L1726.68 241.163 Q1729.02 240.261 1731.22 239.821 Q1733.42 239.358 1735.5 239.358 Q1741.13 239.358 1743.9 242.274 Q1746.68 245.191 1746.68 251.117 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1775.15 273.779 L1775.15 277.089 L1750.52 277.089 L1750.52 273.779 L1775.15 273.779 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1794.18 243.964 Q1793.46 243.548 1792.61 243.362 Q1791.77 243.154 1790.75 243.154 Q1787.14 243.154 1785.2 245.515 Q1783.28 247.853 1783.28 252.251 L1783.28 265.909 L1779 265.909 L1779 239.983 L1783.28 239.983 L1783.28 244.011 Q1784.62 241.649 1786.77 240.515 Q1788.93 239.358 1792 239.358 Q1792.44 239.358 1792.98 239.427 Q1793.51 239.474 1794.16 239.589 L1794.18 243.964 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1819.78 251.881 L1819.78 253.964 L1800.2 253.964 Q1800.48 258.362 1802.84 260.677 Q1805.22 262.969 1809.46 262.969 Q1811.91 262.969 1814.2 262.367 Q1816.52 261.765 1818.79 260.561 L1818.79 264.589 Q1816.5 265.561 1814.09 266.071 Q1811.68 266.58 1809.2 266.58 Q1803 266.58 1799.37 262.969 Q1795.75 259.358 1795.75 253.2 Q1795.75 246.835 1799.18 243.108 Q1802.63 239.358 1808.46 239.358 Q1813.69 239.358 1816.73 242.737 Q1819.78 246.094 1819.78 251.881 M1815.52 250.631 Q1815.48 247.136 1813.56 245.052 Q1811.66 242.969 1808.51 242.969 Q1804.94 242.969 1802.79 244.983 Q1800.66 246.997 1800.34 250.654 L1815.52 250.631 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1826.77 229.89 L1831.03 229.89 L1831.03 265.909 L1826.77 265.909 L1826.77 229.89 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1853.07 229.89 L1853.07 233.432 L1849 233.432 Q1846.7 233.432 1845.8 234.358 Q1844.92 235.284 1844.92 237.691 L1844.92 239.983 L1851.93 239.983 L1851.93 243.293 L1844.92 243.293 L1844.92 265.909 L1840.64 265.909 L1840.64 243.293 L1836.56 243.293 L1836.56 239.983 L1840.64 239.983 L1840.64 238.177 Q1840.64 233.849 1842.65 231.881 Q1844.67 229.89 1849.04 229.89 L1853.07 229.89 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1871.66 243.964 Q1870.94 243.548 1870.08 243.362 Q1869.25 243.154 1868.23 243.154 Q1864.62 243.154 1862.68 245.515 Q1860.75 247.853 1860.75 252.251 L1860.75 265.909 L1856.47 265.909 L1856.47 239.983 L1860.75 239.983 L1860.75 244.011 Q1862.1 241.649 1864.25 240.515 Q1866.4 239.358 1869.48 239.358 Q1869.92 239.358 1870.45 239.427 Q1870.99 239.474 1871.63 239.589 L1871.66 243.964 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1876.12 239.983 L1880.38 239.983 L1880.38 265.909 L1876.12 265.909 L1876.12 239.983 M1876.12 229.89 L1880.38 229.89 L1880.38 235.284 L1876.12 235.284 L1876.12 229.89 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1907.95 240.978 L1907.95 244.96 Q1906.15 243.964 1904.32 243.478 Q1902.51 242.969 1900.66 242.969 Q1896.52 242.969 1894.23 245.608 Q1891.93 248.224 1891.93 252.969 Q1891.93 257.714 1894.23 260.353 Q1896.52 262.969 1900.66 262.969 Q1902.51 262.969 1904.32 262.483 Q1906.15 261.973 1907.95 260.978 L1907.95 264.913 Q1906.17 265.747 1904.25 266.163 Q1902.35 266.58 1900.2 266.58 Q1894.34 266.58 1890.89 262.899 Q1887.44 259.219 1887.44 252.969 Q1887.44 246.626 1890.92 242.992 Q1894.41 239.358 1900.48 239.358 Q1902.44 239.358 1904.32 239.775 Q1906.19 240.168 1907.95 240.978 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1940.66 229.937 Q1937.56 235.261 1936.05 240.469 Q1934.55 245.677 1934.55 251.024 Q1934.55 256.372 1936.05 261.626 Q1937.58 266.858 1940.66 272.159 L1936.96 272.159 Q1933.49 266.719 1931.75 261.464 Q1930.04 256.21 1930.04 251.024 Q1930.04 245.862 1931.75 240.631 Q1933.46 235.4 1936.96 229.937 L1940.66 229.937 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1949.57 231.349 L1967.93 231.349 L1967.93 235.284 L1953.86 235.284 L1953.86 243.756 Q1954.87 243.409 1955.89 243.247 Q1956.91 243.062 1957.93 243.062 Q1963.72 243.062 1967.1 246.233 Q1970.48 249.404 1970.48 254.821 Q1970.48 260.399 1967 263.501 Q1963.53 266.58 1957.21 266.58 Q1955.04 266.58 1952.77 266.21 Q1950.52 265.839 1948.11 265.098 L1948.11 260.399 Q1950.2 261.534 1952.42 262.089 Q1954.64 262.645 1957.12 262.645 Q1961.12 262.645 1963.46 260.538 Q1965.8 258.432 1965.8 254.821 Q1965.8 251.21 1963.46 249.103 Q1961.12 246.997 1957.12 246.997 Q1955.24 246.997 1953.37 247.413 Q1951.52 247.83 1949.57 248.71 L1949.57 231.349 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1989.69 249.497 Q1986.36 249.497 1984.43 251.279 Q1982.54 253.061 1982.54 256.186 Q1982.54 259.311 1984.43 261.094 Q1986.36 262.876 1989.69 262.876 Q1993.02 262.876 1994.94 261.094 Q1996.86 259.288 1996.86 256.186 Q1996.86 253.061 1994.94 251.279 Q1993.05 249.497 1989.69 249.497 M1985.01 247.506 Q1982 246.765 1980.31 244.705 Q1978.65 242.645 1978.65 239.682 Q1978.65 235.538 1981.59 233.131 Q1984.55 230.724 1989.69 230.724 Q1994.85 230.724 1997.79 233.131 Q2000.73 235.538 2000.73 239.682 Q2000.73 242.645 1999.04 244.705 Q1997.37 246.765 1994.39 247.506 Q1997.77 248.293 1999.64 250.585 Q2001.54 252.876 2001.54 256.186 Q2001.54 261.21 1998.46 263.895 Q1995.41 266.58 1989.69 266.58 Q1983.97 266.58 1980.89 263.895 Q1977.84 261.21 1977.84 256.186 Q1977.84 252.876 1979.74 250.585 Q1981.63 248.293 1985.01 247.506 M1983.3 240.122 Q1983.3 242.807 1984.97 244.312 Q1986.66 245.816 1989.69 245.816 Q1992.7 245.816 1994.39 244.312 Q1996.1 242.807 1996.1 240.122 Q1996.1 237.437 1994.39 235.932 Q1992.7 234.427 1989.69 234.427 Q1986.66 234.427 1984.97 235.932 Q1983.3 237.437 1983.3 240.122 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M2008.67 231.349 L2030.89 231.349 L2030.89 233.339 L2018.35 265.909 L2013.46 265.909 L2025.27 235.284 L2008.67 235.284 L2008.67 231.349 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M2044.04 261.973 L2060.36 261.973 L2060.36 265.909 L2038.42 265.909 L2038.42 261.973 Q2041.08 259.219 2045.66 254.589 Q2050.27 249.936 2051.45 248.594 Q2053.69 246.071 2054.57 244.335 Q2055.48 242.575 2055.48 240.886 Q2055.48 238.131 2053.53 236.395 Q2051.61 234.659 2048.51 234.659 Q2046.31 234.659 2043.86 235.423 Q2041.42 236.187 2038.65 237.737 L2038.65 233.015 Q2041.47 231.881 2043.92 231.302 Q2046.38 230.724 2048.42 230.724 Q2053.79 230.724 2056.98 233.409 Q2060.17 236.094 2060.17 240.585 Q2060.17 242.714 2059.36 244.636 Q2058.58 246.534 2056.47 249.126 Q2055.89 249.798 2052.79 253.015 Q2049.69 256.21 2044.04 261.973 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M2080.17 234.427 Q2076.56 234.427 2074.73 237.992 Q2072.93 241.534 2072.93 248.663 Q2072.93 255.77 2074.73 259.335 Q2076.56 262.876 2080.17 262.876 Q2083.81 262.876 2085.61 259.335 Q2087.44 255.77 2087.44 248.663 Q2087.44 241.534 2085.61 237.992 Q2083.81 234.427 2080.17 234.427 M2080.17 230.724 Q2085.98 230.724 2089.04 235.33 Q2092.12 239.913 2092.12 248.663 Q2092.12 257.39 2089.04 261.997 Q2085.98 266.58 2080.17 266.58 Q2074.36 266.58 2071.29 261.997 Q2068.23 257.39 2068.23 248.663 Q2068.23 239.913 2071.29 235.33 Q2074.36 230.724 2080.17 230.724 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M2104.36 261.973 L2120.68 261.973 L2120.68 265.909 L2098.74 265.909 L2098.74 261.973 Q2101.4 259.219 2105.98 254.589 Q2110.59 249.936 2111.77 248.594 Q2114.02 246.071 2114.9 244.335 Q2115.8 242.575 2115.8 240.886 Q2115.8 238.131 2113.85 236.395 Q2111.93 234.659 2108.83 234.659 Q2106.63 234.659 2104.18 235.423 Q2101.75 236.187 2098.97 237.737 L2098.97 233.015 Q2101.79 231.881 2104.25 231.302 Q2106.7 230.724 2108.74 230.724 Q2114.11 230.724 2117.3 233.409 Q2120.5 236.094 2120.5 240.585 Q2120.5 242.714 2119.69 244.636 Q2118.9 246.534 2116.79 249.126 Q2116.22 249.798 2113.11 253.015 Q2110.01 256.21 2104.36 261.973 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M2130.54 231.349 L2148.9 231.349 L2148.9 235.284 L2134.83 235.284 L2134.83 243.756 Q2135.85 243.409 2136.86 243.247 Q2137.88 243.062 2138.9 243.062 Q2144.69 243.062 2148.07 246.233 Q2151.45 249.404 2151.45 254.821 Q2151.45 260.399 2147.97 263.501 Q2144.5 266.58 2138.18 266.58 Q2136.01 266.58 2133.74 266.21 Q2131.49 265.839 2129.09 265.098 L2129.09 260.399 Q2131.17 261.534 2133.39 262.089 Q2135.61 262.645 2138.09 262.645 Q2142.1 262.645 2144.43 260.538 Q2146.77 258.432 2146.77 254.821 Q2146.77 251.21 2144.43 249.103 Q2142.1 246.997 2138.09 246.997 Q2136.22 246.997 2134.34 247.413 Q2132.49 247.83 2130.54 248.71 L2130.54 231.349 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M2171.24 246.765 Q2168.09 246.765 2166.24 248.918 Q2164.41 251.071 2164.41 254.821 Q2164.41 258.548 2166.24 260.723 Q2168.09 262.876 2171.24 262.876 Q2174.39 262.876 2176.22 260.723 Q2178.07 258.548 2178.07 254.821 Q2178.07 251.071 2176.22 248.918 Q2174.39 246.765 2171.24 246.765 M2180.52 232.113 L2180.52 236.372 Q2178.76 235.538 2176.96 235.099 Q2175.17 234.659 2173.41 234.659 Q2168.79 234.659 2166.33 237.784 Q2163.9 240.909 2163.55 247.228 Q2164.92 245.214 2166.98 244.149 Q2169.04 243.062 2171.52 243.062 Q2176.72 243.062 2179.73 246.233 Q2182.77 249.381 2182.77 254.821 Q2182.77 260.145 2179.62 263.362 Q2176.47 266.58 2171.24 266.58 Q2165.24 266.58 2162.07 261.997 Q2158.9 257.39 2158.9 248.663 Q2158.9 240.469 2162.79 235.608 Q2166.68 230.724 2173.23 230.724 Q2174.99 230.724 2176.77 231.071 Q2178.58 231.418 2180.52 232.113 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M2191.63 261.973 L2199.27 261.973 L2199.27 235.608 L2190.96 237.275 L2190.96 233.015 L2199.22 231.349 L2203.9 231.349 L2203.9 261.973 L2211.54 261.973 L2211.54 265.909 L2191.63 265.909 L2191.63 261.973 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M2219.71 229.937 L2223.41 229.937 Q2226.89 235.4 2228.6 240.631 Q2230.34 245.862 2230.34 251.024 Q2230.34 256.21 2228.6 261.464 Q2226.89 266.719 2223.41 272.159 L2219.71 272.159 Q2222.79 266.858 2224.29 261.626 Q2225.82 256.372 2225.82 251.024 Q2225.82 245.677 2224.29 240.469 Q2222.79 235.261 2219.71 229.937 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip940)" style="stroke:#c271d2; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1415.15,300.469 1562.06,300.469 
  "/>
<path clip-path="url(#clip940)" d="M1606.89 296.8 Q1608.49 293.929 1610.71 292.564 Q1612.93 291.198 1615.94 291.198 Q1619.99 291.198 1622.19 294.045 Q1624.39 296.869 1624.39 302.101 L1624.39 317.749 L1620.11 317.749 L1620.11 302.239 Q1620.11 298.513 1618.79 296.707 Q1617.47 294.902 1614.76 294.902 Q1611.45 294.902 1609.53 297.101 Q1607.61 299.3 1607.61 303.096 L1607.61 317.749 L1603.33 317.749 L1603.33 302.239 Q1603.33 298.489 1602.01 296.707 Q1600.69 294.902 1597.93 294.902 Q1594.67 294.902 1592.75 297.124 Q1590.83 299.323 1590.83 303.096 L1590.83 317.749 L1586.54 317.749 L1586.54 291.823 L1590.83 291.823 L1590.83 295.851 Q1592.28 293.466 1594.32 292.332 Q1596.36 291.198 1599.16 291.198 Q1601.98 291.198 1603.95 292.633 Q1605.94 294.068 1606.89 296.8 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1644.67 304.716 Q1639.51 304.716 1637.51 305.897 Q1635.52 307.077 1635.52 309.925 Q1635.52 312.193 1637.01 313.536 Q1638.51 314.855 1641.08 314.855 Q1644.62 314.855 1646.75 312.355 Q1648.9 309.832 1648.9 305.665 L1648.9 304.716 L1644.67 304.716 M1653.16 302.957 L1653.16 317.749 L1648.9 317.749 L1648.9 313.813 Q1647.45 316.175 1645.27 317.309 Q1643.09 318.42 1639.95 318.42 Q1635.96 318.42 1633.6 316.198 Q1631.26 313.952 1631.26 310.202 Q1631.26 305.827 1634.18 303.605 Q1637.12 301.383 1642.93 301.383 L1648.9 301.383 L1648.9 300.966 Q1648.9 298.027 1646.96 296.429 Q1645.04 294.809 1641.54 294.809 Q1639.32 294.809 1637.21 295.341 Q1635.11 295.874 1633.16 296.939 L1633.16 293.003 Q1635.5 292.101 1637.7 291.661 Q1639.9 291.198 1641.98 291.198 Q1647.61 291.198 1650.39 294.114 Q1653.16 297.031 1653.16 302.957 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1678.46 292.587 L1678.46 296.614 Q1676.66 295.689 1674.71 295.226 Q1672.77 294.763 1670.69 294.763 Q1667.51 294.763 1665.92 295.735 Q1664.34 296.707 1664.34 298.651 Q1664.34 300.133 1665.48 300.989 Q1666.61 301.823 1670.04 302.587 L1671.5 302.911 Q1676.03 303.883 1677.93 305.665 Q1679.85 307.425 1679.85 310.596 Q1679.85 314.207 1676.98 316.313 Q1674.13 318.42 1669.13 318.42 Q1667.05 318.42 1664.78 318.003 Q1662.54 317.61 1660.04 316.8 L1660.04 312.401 Q1662.4 313.628 1664.69 314.253 Q1666.98 314.855 1669.23 314.855 Q1672.24 314.855 1673.86 313.837 Q1675.48 312.795 1675.48 310.92 Q1675.48 309.184 1674.3 308.258 Q1673.14 307.332 1669.18 306.476 L1667.7 306.128 Q1663.74 305.295 1661.98 303.582 Q1660.22 301.846 1660.22 298.837 Q1660.22 295.179 1662.82 293.189 Q1665.41 291.198 1670.18 291.198 Q1672.54 291.198 1674.62 291.545 Q1676.7 291.892 1678.46 292.587 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1703.16 292.587 L1703.16 296.614 Q1701.36 295.689 1699.41 295.226 Q1697.47 294.763 1695.38 294.763 Q1692.21 294.763 1690.62 295.735 Q1689.04 296.707 1689.04 298.651 Q1689.04 300.133 1690.18 300.989 Q1691.31 301.823 1694.74 302.587 L1696.19 302.911 Q1700.73 303.883 1702.63 305.665 Q1704.55 307.425 1704.55 310.596 Q1704.55 314.207 1701.68 316.313 Q1698.83 318.42 1693.83 318.42 Q1691.75 318.42 1689.48 318.003 Q1687.24 317.61 1684.74 316.8 L1684.74 312.401 Q1687.1 313.628 1689.39 314.253 Q1691.68 314.855 1693.93 314.855 Q1696.94 314.855 1698.56 313.837 Q1700.18 312.795 1700.18 310.92 Q1700.18 309.184 1699 308.258 Q1697.84 307.332 1693.88 306.476 L1692.4 306.128 Q1688.44 305.295 1686.68 303.582 Q1684.92 301.846 1684.92 298.837 Q1684.92 295.179 1687.51 293.189 Q1690.11 291.198 1694.88 291.198 Q1697.24 291.198 1699.32 291.545 Q1701.4 291.892 1703.16 292.587 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1711.94 311.869 L1716.82 311.869 L1716.82 317.749 L1711.94 317.749 L1711.94 311.869 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1739.53 281.73 L1739.53 285.272 L1735.45 285.272 Q1733.16 285.272 1732.26 286.198 Q1731.38 287.124 1731.38 289.531 L1731.38 291.823 L1738.39 291.823 L1738.39 295.133 L1731.38 295.133 L1731.38 317.749 L1727.1 317.749 L1727.1 295.133 L1723.02 295.133 L1723.02 291.823 L1727.1 291.823 L1727.1 290.017 Q1727.1 285.689 1729.11 283.721 Q1731.13 281.73 1735.5 281.73 L1739.53 281.73 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1768.39 281.777 Q1765.29 287.101 1763.79 292.309 Q1762.28 297.517 1762.28 302.864 Q1762.28 308.212 1763.79 313.466 Q1765.31 318.698 1768.39 323.999 L1764.69 323.999 Q1761.22 318.559 1759.48 313.304 Q1757.77 308.05 1757.77 302.864 Q1757.77 297.702 1759.48 292.471 Q1761.19 287.24 1764.69 281.777 L1768.39 281.777 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1787.84 298.605 Q1784.69 298.605 1782.84 300.758 Q1781.01 302.911 1781.01 306.661 Q1781.01 310.388 1782.84 312.563 Q1784.69 314.716 1787.84 314.716 Q1790.99 314.716 1792.81 312.563 Q1794.67 310.388 1794.67 306.661 Q1794.67 302.911 1792.81 300.758 Q1790.99 298.605 1787.84 298.605 M1797.12 283.953 L1797.12 288.212 Q1795.36 287.378 1793.56 286.939 Q1791.77 286.499 1790.01 286.499 Q1785.38 286.499 1782.93 289.624 Q1780.5 292.749 1780.15 299.068 Q1781.52 297.054 1783.58 295.989 Q1785.64 294.902 1788.12 294.902 Q1793.32 294.902 1796.33 298.073 Q1799.37 301.221 1799.37 306.661 Q1799.37 311.985 1796.22 315.202 Q1793.07 318.42 1787.84 318.42 Q1781.84 318.42 1778.67 313.837 Q1775.5 309.23 1775.5 300.503 Q1775.5 292.309 1779.39 287.448 Q1783.28 282.564 1789.83 282.564 Q1791.59 282.564 1793.37 282.911 Q1795.18 283.258 1797.12 283.953 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1821.59 299.114 Q1824.94 299.832 1826.82 302.101 Q1828.72 304.369 1828.72 307.702 Q1828.72 312.818 1825.2 315.619 Q1821.68 318.42 1815.2 318.42 Q1813.02 318.42 1810.71 317.98 Q1808.42 317.563 1805.96 316.707 L1805.96 312.193 Q1807.91 313.327 1810.22 313.906 Q1812.54 314.485 1815.06 314.485 Q1819.46 314.485 1821.75 312.749 Q1824.06 311.013 1824.06 307.702 Q1824.06 304.647 1821.91 302.934 Q1819.78 301.198 1815.96 301.198 L1811.94 301.198 L1811.94 297.355 L1816.15 297.355 Q1819.6 297.355 1821.43 295.989 Q1823.25 294.601 1823.25 292.008 Q1823.25 289.346 1821.36 287.934 Q1819.48 286.499 1815.96 286.499 Q1814.04 286.499 1811.84 286.915 Q1809.64 287.332 1807 288.212 L1807 284.045 Q1809.67 283.304 1811.98 282.934 Q1814.32 282.564 1816.38 282.564 Q1821.7 282.564 1824.81 284.994 Q1827.91 287.402 1827.91 291.522 Q1827.91 294.392 1826.26 296.383 Q1824.62 298.351 1821.59 299.114 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1836.4 283.189 L1858.62 283.189 L1858.62 285.179 L1846.08 317.749 L1841.19 317.749 L1853 287.124 L1836.4 287.124 L1836.4 283.189 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1867.79 283.189 L1886.15 283.189 L1886.15 287.124 L1872.07 287.124 L1872.07 295.596 Q1873.09 295.249 1874.11 295.087 Q1875.13 294.902 1876.15 294.902 Q1881.93 294.902 1885.31 298.073 Q1888.69 301.244 1888.69 306.661 Q1888.69 312.239 1885.22 315.341 Q1881.75 318.42 1875.43 318.42 Q1873.25 318.42 1870.99 318.05 Q1868.74 317.679 1866.33 316.938 L1866.33 312.239 Q1868.42 313.374 1870.64 313.929 Q1872.86 314.485 1875.34 314.485 Q1879.34 314.485 1881.68 312.378 Q1884.02 310.272 1884.02 306.661 Q1884.02 303.05 1881.68 300.943 Q1879.34 298.837 1875.34 298.837 Q1873.46 298.837 1871.59 299.253 Q1869.74 299.67 1867.79 300.55 L1867.79 283.189 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1912.07 299.114 Q1915.43 299.832 1917.3 302.101 Q1919.2 304.369 1919.2 307.702 Q1919.2 312.818 1915.68 315.619 Q1912.17 318.42 1905.68 318.42 Q1903.51 318.42 1901.19 317.98 Q1898.9 317.563 1896.45 316.707 L1896.45 312.193 Q1898.39 313.327 1900.71 313.906 Q1903.02 314.485 1905.55 314.485 Q1909.94 314.485 1912.24 312.749 Q1914.55 311.013 1914.55 307.702 Q1914.55 304.647 1912.4 302.934 Q1910.27 301.198 1906.45 301.198 L1902.42 301.198 L1902.42 297.355 L1906.63 297.355 Q1910.08 297.355 1911.91 295.989 Q1913.74 294.601 1913.74 292.008 Q1913.74 289.346 1911.84 287.934 Q1909.97 286.499 1906.45 286.499 Q1904.53 286.499 1902.33 286.915 Q1900.13 287.332 1897.49 288.212 L1897.49 284.045 Q1900.15 283.304 1902.47 282.934 Q1904.8 282.564 1906.87 282.564 Q1912.19 282.564 1915.29 284.994 Q1918.39 287.402 1918.39 291.522 Q1918.39 294.392 1916.75 296.383 Q1915.11 298.351 1912.07 299.114 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1940.92 287.263 L1929.11 305.712 L1940.92 305.712 L1940.92 287.263 M1939.69 283.189 L1945.57 283.189 L1945.57 305.712 L1950.5 305.712 L1950.5 309.601 L1945.57 309.601 L1945.57 317.749 L1940.92 317.749 L1940.92 309.601 L1925.31 309.601 L1925.31 305.087 L1939.69 283.189 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1962.26 313.813 L1978.58 313.813 L1978.58 317.749 L1956.63 317.749 L1956.63 313.813 Q1959.3 311.059 1963.88 306.429 Q1968.49 301.776 1969.67 300.434 Q1971.91 297.911 1972.79 296.175 Q1973.69 294.415 1973.69 292.726 Q1973.69 289.971 1971.75 288.235 Q1969.83 286.499 1966.73 286.499 Q1964.53 286.499 1962.07 287.263 Q1959.64 288.027 1956.86 289.577 L1956.86 284.855 Q1959.69 283.721 1962.14 283.142 Q1964.6 282.564 1966.63 282.564 Q1972 282.564 1975.2 285.249 Q1978.39 287.934 1978.39 292.425 Q1978.39 294.554 1977.58 296.476 Q1976.8 298.374 1974.69 300.966 Q1974.11 301.638 1971.01 304.855 Q1967.91 308.05 1962.26 313.813 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M1989.2 313.813 L1996.84 313.813 L1996.84 287.448 L1988.53 289.115 L1988.53 284.855 L1996.8 283.189 L2001.47 283.189 L2001.47 313.813 L2009.11 313.813 L2009.11 317.749 L1989.2 317.749 L1989.2 313.813 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M2019.36 313.813 L2027 313.813 L2027 287.448 L2018.69 289.115 L2018.69 284.855 L2026.96 283.189 L2031.63 283.189 L2031.63 313.813 L2039.27 313.813 L2039.27 317.749 L2019.36 317.749 L2019.36 313.813 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip940)" d="M2047.44 281.777 L2051.15 281.777 Q2054.62 287.24 2056.33 292.471 Q2058.07 297.702 2058.07 302.864 Q2058.07 308.05 2056.33 313.304 Q2054.62 318.559 2051.15 323.999 L2047.44 323.999 Q2050.52 318.698 2052.03 313.466 Q2053.55 308.212 2053.55 302.864 Q2053.55 297.517 2052.03 292.309 Q2050.52 287.101 2047.44 281.777 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /></svg>



 After the plots are created, the FMU is unloaded.


```julia
fmiUnload(realFMU)
```

The data from the simualtion of the *realFMU*, are divided into position and velocity data. These data will be needed later. 


```julia
velReal = collect(data[2] for data in realSimData.saveval)
posReal = collect(data[1] for data in realSimData.saveval)
```


    501-element Vector{Float64}:
     0.5
     0.5002147350958549
     0.5008572068150906
     0.5019307343013393
     0.5034377791759037
     0.5053797361931734
     0.507757421354132
     0.5105710760625108
     0.5138202661486302
     0.5175035429290001
     ⋮
     1.0647763670106132
     1.06549412855799
     1.0660771282511987
     1.066523587371857
     1.066831801521633
     1.0670001400125004
     1.0670339828286313
     1.0670339828276312
     1.067033982826631


### SimpleFMU

The following lines load, instantiate, simulate and plot the *simpleFMU* just like the *realFMU*. The differences between both systems can be clearly seen from the plots. In the plot for the *realFMU* it can be seen that the oscillation continues to decrease due to the effect of the friction. If you would simulate long enough, the oscillation would come to a standstill in a certain time. The oscillation in the *simpleFMU* behaves differently, since the friction was not taken into account here. The oscillation in this model would continue to infinity with the same oscillation amplitude. From this observation the desire of an improvement of this model arises.     


```julia
simpleFMU = fmiLoad("SpringPendulum1D", "Dymola", "2022x")

fmiInstantiate!(simpleFMU; loggingOn=false)
fmiInfo(simpleFMU)

vrs = ["mass.s", "mass.v", "mass.a"]
_, simpleSimData = fmiSimulate(simpleFMU, tStart, tStop; recordValues=vrs, saveat=tSave, reset=false)
fmiPlot(simpleFMU, vrs, simpleSimData)
```

    #################### Begin information for FMU ####################
    	Model name:			SpringPendulum1D
    	FMI-Version:			2.0
    	GUID:				{a3e886fa-675c-4308-8e91-3490e598ba11}
    	Generation tool:		Dymola Version 2022x (64-bit), 2021-10-08
    	Generation time:		2022-03-03T15:08:57Z
    	Var. naming conv.:		structured
    	Event indicators:		0
    	Inputs:				0
    	Outputs:			0
    	States:				2
    		33554432 ["mass.s"]
    		33554433 ["mass.v"]
    	Supports Co-Simulation:		true
    		Model identifier:	SpringPendulum1D
    		Get/Set State:		true
    		Serialize State:	true
    		Dir. Derivatives:	true
    		Var. com. steps:	true
    		Input interpol.:	true
    		Max order out. der.:	1
    	Supports Model-Exchange:	true
    		Model identifier:	SpringPendulum1D
    		Get/Set State:		true
    		Serialize State:	true
    		Dir. Derivatives:	true
    ##################### End information for FMU #####################

    ┌ Info: fmi2Unzip(...): Successfully unzipped 29 files at `C:\Users\thummeto\AppData\Local\Temp\fmijl_WrKVi8\SpringPendulum1D`.
    └ @ FMIImport C:\Users\thummeto\Documents\FMIImport.jl\src\FMI2_ext.jl:77
    ┌ Info: fmi2Load(...): FMU resources location is `file:///C:/Users/thummeto/AppData/Local/Temp/fmijl_WrKVi8/SpringPendulum1D/resources`
    └ @ FMIImport C:\Users\thummeto\Documents\FMIImport.jl\src\FMI2_ext.jl:192
    ┌ Info: fmi2Load(...): FMU supports both CS and ME, using CS as default if nothing specified.
    └ @ FMIImport C:\Users\thummeto\Documents\FMIImport.jl\src\FMI2_ext.jl:195


    



<?xml version="1.0" encoding="utf-8"?>
<svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" width="600" height="400" viewBox="0 0 2400 1600">
<defs>
  <clipPath id="clip030">
    <rect x="0" y="0" width="2400" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip030)" d="
M0 1600 L2400 1600 L2400 0 L0 0  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip031">
    <rect x="480" y="0" width="1681" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip030)" d="
M193.936 1423.18 L2352.76 1423.18 L2352.76 47.2441 L193.936 47.2441  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip032">
    <rect x="193" y="47" width="2160" height="1377"/>
  </clipPath>
</defs>
<polyline clip-path="url(#clip032)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  255.035,1423.18 255.035,47.2441 
  "/>
<polyline clip-path="url(#clip032)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  662.359,1423.18 662.359,47.2441 
  "/>
<polyline clip-path="url(#clip032)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1069.68,1423.18 1069.68,47.2441 
  "/>
<polyline clip-path="url(#clip032)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1477.01,1423.18 1477.01,47.2441 
  "/>
<polyline clip-path="url(#clip032)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1884.33,1423.18 1884.33,47.2441 
  "/>
<polyline clip-path="url(#clip032)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  2291.66,1423.18 2291.66,47.2441 
  "/>
<polyline clip-path="url(#clip030)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  193.936,1423.18 2352.76,1423.18 
  "/>
<polyline clip-path="url(#clip030)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  255.035,1423.18 255.035,1404.28 
  "/>
<polyline clip-path="url(#clip030)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  662.359,1423.18 662.359,1404.28 
  "/>
<polyline clip-path="url(#clip030)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1069.68,1423.18 1069.68,1404.28 
  "/>
<polyline clip-path="url(#clip030)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1477.01,1423.18 1477.01,1404.28 
  "/>
<polyline clip-path="url(#clip030)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1884.33,1423.18 1884.33,1404.28 
  "/>
<polyline clip-path="url(#clip030)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  2291.66,1423.18 2291.66,1404.28 
  "/>
<path clip-path="url(#clip030)" d="M255.035 1454.1 Q251.424 1454.1 249.595 1457.66 Q247.789 1461.2 247.789 1468.33 Q247.789 1475.44 249.595 1479.01 Q251.424 1482.55 255.035 1482.55 Q258.669 1482.55 260.474 1479.01 Q262.303 1475.44 262.303 1468.33 Q262.303 1461.2 260.474 1457.66 Q258.669 1454.1 255.035 1454.1 M255.035 1450.39 Q260.845 1450.39 263.9 1455 Q266.979 1459.58 266.979 1468.33 Q266.979 1477.06 263.9 1481.67 Q260.845 1486.25 255.035 1486.25 Q249.225 1486.25 246.146 1481.67 Q243.09 1477.06 243.09 1468.33 Q243.09 1459.58 246.146 1455 Q249.225 1450.39 255.035 1450.39 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M652.741 1481.64 L660.38 1481.64 L660.38 1455.28 L652.07 1456.95 L652.07 1452.69 L660.334 1451.02 L665.01 1451.02 L665.01 1481.64 L672.648 1481.64 L672.648 1485.58 L652.741 1485.58 L652.741 1481.64 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1064.34 1481.64 L1080.66 1481.64 L1080.66 1485.58 L1058.71 1485.58 L1058.71 1481.64 Q1061.37 1478.89 1065.96 1474.26 Q1070.56 1469.61 1071.74 1468.27 Q1073.99 1465.74 1074.87 1464.01 Q1075.77 1462.25 1075.77 1460.56 Q1075.77 1457.8 1073.83 1456.07 Q1071.91 1454.33 1068.8 1454.33 Q1066.61 1454.33 1064.15 1455.09 Q1061.72 1455.86 1058.94 1457.41 L1058.94 1452.69 Q1061.77 1451.55 1064.22 1450.97 Q1066.67 1450.39 1068.71 1450.39 Q1074.08 1450.39 1077.28 1453.08 Q1080.47 1455.77 1080.47 1460.26 Q1080.47 1462.39 1079.66 1464.31 Q1078.87 1466.2 1076.77 1468.8 Q1076.19 1469.47 1073.09 1472.69 Q1069.98 1475.88 1064.34 1481.64 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1481.26 1466.95 Q1484.61 1467.66 1486.49 1469.93 Q1488.39 1472.2 1488.39 1475.53 Q1488.39 1480.65 1484.87 1483.45 Q1481.35 1486.25 1474.87 1486.25 Q1472.69 1486.25 1470.38 1485.81 Q1468.08 1485.39 1465.63 1484.54 L1465.63 1480.02 Q1467.58 1481.16 1469.89 1481.74 Q1472.2 1482.32 1474.73 1482.32 Q1479.13 1482.32 1481.42 1480.58 Q1483.73 1478.84 1483.73 1475.53 Q1483.73 1472.48 1481.58 1470.77 Q1479.45 1469.03 1475.63 1469.03 L1471.6 1469.03 L1471.6 1465.19 L1475.82 1465.19 Q1479.27 1465.19 1481.09 1463.82 Q1482.92 1462.43 1482.92 1459.84 Q1482.92 1457.18 1481.02 1455.77 Q1479.15 1454.33 1475.63 1454.33 Q1473.71 1454.33 1471.51 1454.75 Q1469.31 1455.16 1466.67 1456.04 L1466.67 1451.88 Q1469.33 1451.14 1471.65 1450.77 Q1473.99 1450.39 1476.05 1450.39 Q1481.37 1450.39 1484.47 1452.83 Q1487.58 1455.23 1487.58 1459.35 Q1487.58 1462.22 1485.93 1464.21 Q1484.29 1466.18 1481.26 1466.95 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1887.34 1455.09 L1875.54 1473.54 L1887.34 1473.54 L1887.34 1455.09 M1886.12 1451.02 L1891.99 1451.02 L1891.99 1473.54 L1896.93 1473.54 L1896.93 1477.43 L1891.99 1477.43 L1891.99 1485.58 L1887.34 1485.58 L1887.34 1477.43 L1871.74 1477.43 L1871.74 1472.92 L1886.12 1451.02 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M2281.94 1451.02 L2300.29 1451.02 L2300.29 1454.96 L2286.22 1454.96 L2286.22 1463.43 Q2287.24 1463.08 2288.25 1462.92 Q2289.27 1462.73 2290.29 1462.73 Q2296.08 1462.73 2299.46 1465.9 Q2302.84 1469.08 2302.84 1474.49 Q2302.84 1480.07 2299.37 1483.17 Q2295.89 1486.25 2289.57 1486.25 Q2287.4 1486.25 2285.13 1485.88 Q2282.88 1485.51 2280.48 1484.77 L2280.48 1480.07 Q2282.56 1481.2 2284.78 1481.76 Q2287 1482.32 2289.48 1482.32 Q2293.49 1482.32 2295.82 1480.21 Q2298.16 1478.1 2298.16 1474.49 Q2298.16 1470.88 2295.82 1468.77 Q2293.49 1466.67 2289.48 1466.67 Q2287.61 1466.67 2285.73 1467.08 Q2283.88 1467.5 2281.94 1468.38 L2281.94 1451.02 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1221.66 1522.27 L1221.66 1532.4 L1233.72 1532.4 L1233.72 1536.95 L1221.66 1536.95 L1221.66 1556.3 Q1221.66 1560.66 1222.83 1561.9 Q1224.04 1563.14 1227.7 1563.14 L1233.72 1563.14 L1233.72 1568.04 L1227.7 1568.04 Q1220.92 1568.04 1218.35 1565.53 Q1215.77 1562.98 1215.77 1556.3 L1215.77 1536.95 L1211.47 1536.95 L1211.47 1532.4 L1215.77 1532.4 L1215.77 1522.27 L1221.66 1522.27 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1261.6 1518.52 L1275.1 1518.52 L1275.1 1523.07 L1267.46 1523.07 L1267.46 1572.09 L1275.1 1572.09 L1275.1 1576.64 L1261.6 1576.64 L1261.6 1518.52 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1310.3 1533.45 L1310.3 1538.98 Q1307.82 1537.71 1305.14 1537.07 Q1302.47 1536.44 1299.6 1536.44 Q1295.24 1536.44 1293.05 1537.77 Q1290.88 1539.11 1290.88 1541.79 Q1290.88 1543.82 1292.44 1545 Q1294 1546.15 1298.71 1547.2 L1300.72 1547.64 Q1306.96 1548.98 1309.57 1551.43 Q1312.21 1553.85 1312.21 1558.21 Q1312.21 1563.17 1308.26 1566.07 Q1304.35 1568.97 1297.47 1568.97 Q1294.61 1568.97 1291.49 1568.39 Q1288.4 1567.85 1284.96 1566.74 L1284.96 1560.69 Q1288.21 1562.38 1291.36 1563.24 Q1294.51 1564.07 1297.6 1564.07 Q1301.74 1564.07 1303.97 1562.66 Q1306.19 1561.23 1306.19 1558.65 Q1306.19 1556.27 1304.57 1554.99 Q1302.98 1553.72 1297.54 1552.54 L1295.5 1552.07 Q1290.06 1550.92 1287.64 1548.56 Q1285.22 1546.18 1285.22 1542.04 Q1285.22 1537.01 1288.78 1534.27 Q1292.35 1531.54 1298.9 1531.54 Q1302.15 1531.54 1305.02 1532.01 Q1307.88 1532.49 1310.3 1533.45 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1335.22 1518.52 L1335.22 1576.64 L1321.73 1576.64 L1321.73 1572.09 L1329.33 1572.09 L1329.33 1523.07 L1321.73 1523.07 L1321.73 1518.52 L1335.22 1518.52 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip032)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  193.936,1274.78 2352.76,1274.78 
  "/>
<polyline clip-path="url(#clip032)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  193.936,1004.94 2352.76,1004.94 
  "/>
<polyline clip-path="url(#clip032)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  193.936,735.102 2352.76,735.102 
  "/>
<polyline clip-path="url(#clip032)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  193.936,465.265 2352.76,465.265 
  "/>
<polyline clip-path="url(#clip032)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  193.936,195.428 2352.76,195.428 
  "/>
<polyline clip-path="url(#clip030)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  193.936,1423.18 193.936,47.2441 
  "/>
<polyline clip-path="url(#clip030)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  193.936,1274.78 212.834,1274.78 
  "/>
<polyline clip-path="url(#clip030)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  193.936,1004.94 212.834,1004.94 
  "/>
<polyline clip-path="url(#clip030)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  193.936,735.102 212.834,735.102 
  "/>
<polyline clip-path="url(#clip030)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  193.936,465.265 212.834,465.265 
  "/>
<polyline clip-path="url(#clip030)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  193.936,195.428 212.834,195.428 
  "/>
<path clip-path="url(#clip030)" d="M50.9921 1275.23 L80.6679 1275.23 L80.6679 1279.16 L50.9921 1279.16 L50.9921 1275.23 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M90.8067 1257.5 L109.163 1257.5 L109.163 1261.43 L95.0891 1261.43 L95.0891 1269.9 Q96.1076 1269.56 97.1261 1269.39 Q98.1447 1269.21 99.1632 1269.21 Q104.95 1269.21 108.33 1272.38 Q111.709 1275.55 111.709 1280.97 Q111.709 1286.55 108.237 1289.65 Q104.765 1292.73 98.4456 1292.73 Q96.2697 1292.73 94.0012 1292.36 Q91.7558 1291.99 89.3484 1291.25 L89.3484 1286.55 Q91.4317 1287.68 93.6539 1288.24 Q95.8761 1288.79 98.353 1288.79 Q102.358 1288.79 104.696 1286.69 Q107.033 1284.58 107.033 1280.97 Q107.033 1277.36 104.696 1275.25 Q102.358 1273.14 98.353 1273.14 Q96.478 1273.14 94.603 1273.56 Q92.7512 1273.98 90.8067 1274.86 L90.8067 1257.5 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M120.922 1286.18 L125.807 1286.18 L125.807 1292.06 L120.922 1292.06 L120.922 1286.18 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M145.992 1260.57 Q142.381 1260.57 140.552 1264.14 Q138.746 1267.68 138.746 1274.81 Q138.746 1281.92 140.552 1285.48 Q142.381 1289.02 145.992 1289.02 Q149.626 1289.02 151.431 1285.48 Q153.26 1281.92 153.26 1274.81 Q153.26 1267.68 151.431 1264.14 Q149.626 1260.57 145.992 1260.57 M145.992 1256.87 Q151.802 1256.87 154.857 1261.48 Q157.936 1266.06 157.936 1274.81 Q157.936 1283.54 154.857 1288.14 Q151.802 1292.73 145.992 1292.73 Q140.181 1292.73 137.103 1288.14 Q134.047 1283.54 134.047 1274.81 Q134.047 1266.06 137.103 1261.48 Q140.181 1256.87 145.992 1256.87 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M51.9875 1005.39 L81.6633 1005.39 L81.6633 1009.33 L51.9875 1009.33 L51.9875 1005.39 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M95.7836 1018.28 L112.103 1018.28 L112.103 1022.22 L90.1586 1022.22 L90.1586 1018.28 Q92.8206 1015.53 97.4039 1010.9 Q102.01 1006.25 103.191 1004.9 Q105.436 1002.38 106.316 1000.64 Q107.219 998.886 107.219 997.196 Q107.219 994.441 105.274 992.705 Q103.353 990.969 100.251 990.969 Q98.0521 990.969 95.5984 991.733 Q93.1678 992.497 90.3901 994.048 L90.3901 989.325 Q93.2141 988.191 95.6678 987.613 Q98.1215 987.034 100.159 987.034 Q105.529 987.034 108.723 989.719 Q111.918 992.404 111.918 996.895 Q111.918 999.024 111.108 1000.95 Q110.321 1002.84 108.214 1005.44 Q107.635 1006.11 104.534 1009.33 Q101.432 1012.52 95.7836 1018.28 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M121.918 1016.34 L126.802 1016.34 L126.802 1022.22 L121.918 1022.22 L121.918 1016.34 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M137.033 987.659 L155.39 987.659 L155.39 991.594 L141.316 991.594 L141.316 1000.07 Q142.334 999.719 143.353 999.557 Q144.371 999.372 145.39 999.372 Q151.177 999.372 154.556 1002.54 Q157.936 1005.71 157.936 1011.13 Q157.936 1016.71 154.464 1019.81 Q150.992 1022.89 144.672 1022.89 Q142.496 1022.89 140.228 1022.52 Q137.982 1022.15 135.575 1021.41 L135.575 1016.71 Q137.658 1017.84 139.881 1018.4 Q142.103 1018.95 144.58 1018.95 Q148.584 1018.95 150.922 1016.85 Q153.26 1014.74 153.26 1011.13 Q153.26 1007.52 150.922 1005.41 Q148.584 1003.31 144.58 1003.31 Q142.705 1003.31 140.83 1003.72 Q138.978 1004.14 137.033 1005.02 L137.033 987.659 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M100.76 720.901 Q97.1493 720.901 95.3206 724.465 Q93.515 728.007 93.515 735.137 Q93.515 742.243 95.3206 745.808 Q97.1493 749.349 100.76 749.349 Q104.395 749.349 106.2 745.808 Q108.029 742.243 108.029 735.137 Q108.029 728.007 106.2 724.465 Q104.395 720.901 100.76 720.901 M100.76 717.197 Q106.571 717.197 109.626 721.803 Q112.705 726.387 112.705 735.137 Q112.705 743.863 109.626 748.47 Q106.571 753.053 100.76 753.053 Q94.9502 753.053 91.8715 748.47 Q88.816 743.863 88.816 735.137 Q88.816 726.387 91.8715 721.803 Q94.9502 717.197 100.76 717.197 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M120.922 746.502 L125.807 746.502 L125.807 752.382 L120.922 752.382 L120.922 746.502 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M145.992 720.901 Q142.381 720.901 140.552 724.465 Q138.746 728.007 138.746 735.137 Q138.746 742.243 140.552 745.808 Q142.381 749.349 145.992 749.349 Q149.626 749.349 151.431 745.808 Q153.26 742.243 153.26 735.137 Q153.26 728.007 151.431 724.465 Q149.626 720.901 145.992 720.901 M145.992 717.197 Q151.802 717.197 154.857 721.803 Q157.936 726.387 157.936 735.137 Q157.936 743.863 154.857 748.47 Q151.802 753.053 145.992 753.053 Q140.181 753.053 137.103 748.47 Q134.047 743.863 134.047 735.137 Q134.047 726.387 137.103 721.803 Q140.181 717.197 145.992 717.197 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M95.7836 478.61 L112.103 478.61 L112.103 482.545 L90.1586 482.545 L90.1586 478.61 Q92.8206 475.855 97.4039 471.226 Q102.01 466.573 103.191 465.23 Q105.436 462.707 106.316 460.971 Q107.219 459.212 107.219 457.522 Q107.219 454.767 105.274 453.031 Q103.353 451.295 100.251 451.295 Q98.0521 451.295 95.5984 452.059 Q93.1678 452.823 90.3901 454.374 L90.3901 449.652 Q93.2141 448.517 95.6678 447.939 Q98.1215 447.36 100.159 447.36 Q105.529 447.36 108.723 450.045 Q111.918 452.73 111.918 457.221 Q111.918 459.351 111.108 461.272 Q110.321 463.17 108.214 465.763 Q107.635 466.434 104.534 469.651 Q101.432 472.846 95.7836 478.61 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M121.918 476.665 L126.802 476.665 L126.802 482.545 L121.918 482.545 L121.918 476.665 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M137.033 447.985 L155.39 447.985 L155.39 451.92 L141.316 451.92 L141.316 460.392 Q142.334 460.045 143.353 459.883 Q144.371 459.698 145.39 459.698 Q151.177 459.698 154.556 462.869 Q157.936 466.04 157.936 471.457 Q157.936 477.036 154.464 480.138 Q150.992 483.216 144.672 483.216 Q142.496 483.216 140.228 482.846 Q137.982 482.475 135.575 481.735 L135.575 477.036 Q137.658 478.17 139.881 478.726 Q142.103 479.281 144.58 479.281 Q148.584 479.281 150.922 477.175 Q153.26 475.068 153.26 471.457 Q153.26 467.846 150.922 465.739 Q148.584 463.633 144.58 463.633 Q142.705 463.633 140.83 464.05 Q138.978 464.466 137.033 465.346 L137.033 447.985 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M90.8067 178.148 L109.163 178.148 L109.163 182.083 L95.0891 182.083 L95.0891 190.555 Q96.1076 190.208 97.1261 190.046 Q98.1447 189.861 99.1632 189.861 Q104.95 189.861 108.33 193.032 Q111.709 196.203 111.709 201.62 Q111.709 207.199 108.237 210.301 Q104.765 213.379 98.4456 213.379 Q96.2697 213.379 94.0012 213.009 Q91.7558 212.639 89.3484 211.898 L89.3484 207.199 Q91.4317 208.333 93.6539 208.889 Q95.8761 209.444 98.353 209.444 Q102.358 209.444 104.696 207.338 Q107.033 205.231 107.033 201.62 Q107.033 198.009 104.696 195.903 Q102.358 193.796 98.353 193.796 Q96.478 193.796 94.603 194.213 Q92.7512 194.629 90.8067 195.509 L90.8067 178.148 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M120.922 206.828 L125.807 206.828 L125.807 212.708 L120.922 212.708 L120.922 206.828 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M145.992 181.227 Q142.381 181.227 140.552 184.791 Q138.746 188.333 138.746 195.463 Q138.746 202.569 140.552 206.134 Q142.381 209.676 145.992 209.676 Q149.626 209.676 151.431 206.134 Q153.26 202.569 153.26 195.463 Q153.26 188.333 151.431 184.791 Q149.626 181.227 145.992 181.227 M145.992 177.523 Q151.802 177.523 154.857 182.129 Q157.936 186.713 157.936 195.463 Q157.936 204.19 154.857 208.796 Q151.802 213.379 145.992 213.379 Q140.181 213.379 137.103 208.796 Q134.047 204.19 134.047 195.463 Q134.047 186.713 137.103 182.129 Q140.181 177.523 145.992 177.523 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip032)" style="stroke:#009af9; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  255.035,681.134 259.108,681.101 263.181,681.003 267.254,680.841 271.328,680.615 275.401,680.325 279.474,679.971 283.547,679.553 287.621,679.072 291.694,678.528 
  295.767,677.922 299.84,677.255 303.914,676.526 307.987,675.738 312.06,674.891 316.133,673.985 320.207,673.021 324.28,672.001 328.353,670.926 332.426,669.795 
  336.5,668.612 340.573,667.376 344.646,666.089 348.719,664.752 352.793,663.368 356.866,661.936 360.939,660.458 365.012,658.936 369.086,657.373 373.159,655.768 
  377.232,654.124 381.305,652.443 385.379,650.724 389.452,648.971 393.525,647.187 397.598,645.371 401.671,643.526 405.745,641.654 409.818,639.756 413.891,637.835 
  417.964,635.892 422.038,633.929 426.111,631.949 430.184,629.952 434.257,627.942 438.331,625.921 442.404,623.889 446.477,621.85 450.55,619.805 454.624,617.757 
  458.697,615.707 462.77,613.657 466.843,611.611 470.917,609.57 474.99,607.534 479.063,605.507 483.136,603.492 487.21,601.489 491.283,599.502 495.356,597.531 
  499.429,595.579 503.503,593.648 507.576,591.739 511.649,589.855 515.722,587.998 519.796,586.169 523.869,584.371 527.942,582.604 532.015,580.871 536.089,579.174 
  540.162,577.513 544.235,575.892 548.308,574.312 552.382,572.773 556.455,571.278 560.528,569.828 564.601,568.425 568.675,567.069 572.748,565.764 576.821,564.508 
  580.894,563.305 584.968,562.154 589.041,561.058 593.114,560.017 597.187,559.033 601.261,558.106 605.334,557.237 609.407,556.427 613.48,555.677 617.553,554.988 
  621.627,554.361 625.7,553.795 629.773,553.293 633.846,552.853 637.92,552.476 641.993,552.164 646.066,551.916 650.139,551.732 654.213,551.614 658.286,551.56 
  662.359,551.57 666.432,551.646 670.506,551.786 674.579,551.991 678.652,552.261 682.725,552.594 686.799,552.992 690.872,553.453 694.945,553.976 699.018,554.563 
  703.092,555.211 707.165,555.92 711.238,556.69 715.311,557.519 719.385,558.408 723.458,559.354 727.531,560.358 731.604,561.417 735.678,562.532 739.751,563.7 
  743.824,564.922 747.897,566.194 751.971,567.517 756.044,568.889 760.117,570.308 764.19,571.773 768.264,573.283 772.337,574.837 776.41,576.432 780.483,578.066 
  784.557,579.739 788.63,581.449 792.703,583.193 796.776,584.971 800.85,586.781 804.923,588.62 808.996,590.487 813.069,592.379 817.142,594.296 821.216,596.235 
  825.289,598.194 829.362,600.171 833.435,602.164 837.509,604.172 841.582,606.191 845.655,608.221 849.728,610.26 853.802,612.304 857.875,614.353 861.948,616.404 
  866.021,618.454 870.095,620.502 874.168,622.546 878.241,624.585 882.314,626.615 886.388,628.635 890.461,630.642 894.534,632.635 898.607,634.611 902.681,636.57 
  906.754,638.509 910.827,640.426 914.9,642.318 918.974,644.184 923.047,646.022 927.12,647.831 931.193,649.609 935.267,651.353 939.34,653.063 943.413,654.735 
  947.486,656.369 951.56,657.963 955.633,659.516 959.706,661.025 963.779,662.49 967.853,663.909 971.926,665.28 975.999,666.601 980.072,667.873 984.146,669.094 
  988.219,670.262 992.292,671.375 996.365,672.434 1000.44,673.436 1004.51,674.382 1008.59,675.269 1012.66,676.098 1016.73,676.867 1020.8,677.575 1024.88,678.222 
  1028.95,678.807 1033.02,679.329 1037.1,679.789 1041.17,680.186 1045.24,680.518 1049.32,680.786 1053.39,680.99 1057.46,681.129 1061.54,681.203 1065.61,681.213 
  1069.68,681.157 1073.76,681.037 1077.83,680.852 1081.9,680.602 1085.98,680.289 1090.05,679.911 1094.12,679.47 1098.2,678.965 1102.27,678.398 1106.34,677.769 
  1110.42,677.078 1114.49,676.327 1118.56,675.516 1122.64,674.645 1126.71,673.716 1130.78,672.73 1134.86,671.688 1138.93,670.59 1143,669.437 1147.08,668.232 
  1151.15,666.975 1155.22,665.667 1159.3,664.31 1163.37,662.905 1167.44,661.453 1171.51,659.956 1175.59,658.415 1179.66,656.833 1183.73,655.21 1187.81,653.548 
  1191.88,651.849 1195.95,650.114 1200.03,648.346 1204.1,646.545 1208.17,644.714 1212.25,642.855 1216.32,640.97 1220.39,639.06 1224.47,637.127 1228.54,635.173 
  1232.61,633.201 1236.69,631.211 1240.76,629.207 1244.83,627.191 1248.91,625.163 1252.98,623.126 1257.05,621.082 1261.13,619.034 1265.2,616.983 1269.27,614.932 
  1273.35,612.882 1277.42,610.835 1281.49,608.794 1285.57,606.76 1289.64,604.736 1293.71,602.724 1297.79,600.726 1301.86,598.743 1305.93,596.777 1310.01,594.831 
  1314.08,592.907 1318.15,591.006 1322.22,589.131 1326.3,587.283 1330.37,585.464 1334.44,583.676 1338.52,581.92 1342.59,580.199 1346.66,578.514 1350.74,576.868 
  1354.81,575.26 1358.88,573.694 1362.96,572.17 1367.03,570.691 1371.1,569.257 1375.18,567.871 1379.25,566.533 1383.32,565.245 1387.4,564.007 1391.47,562.823 
  1395.54,561.692 1399.62,560.615 1403.69,559.595 1407.76,558.631 1411.84,557.725 1415.91,556.878 1419.98,556.09 1424.06,555.362 1428.13,554.696 1432.2,554.092 
  1436.28,553.549 1440.35,553.07 1444.42,552.653 1448.5,552.301 1452.57,552.013 1456.64,551.789 1460.72,551.629 1464.79,551.535 1468.86,551.505 1472.93,551.54 
  1477.01,551.641 1481.08,551.805 1485.15,552.035 1489.23,552.329 1493.3,552.687 1497.37,553.109 1501.45,553.594 1505.52,554.141 1509.59,554.752 1513.67,555.423 
  1517.74,556.156 1521.81,556.949 1525.89,557.802 1529.96,558.713 1534.03,559.682 1538.11,560.707 1542.18,561.788 1546.25,562.924 1550.33,564.114 1554.4,565.355 
  1558.47,566.648 1562.55,567.99 1566.62,569.381 1570.69,570.819 1574.77,572.303 1578.84,573.831 1582.91,575.401 1586.99,577.012 1591.06,578.662 1595.13,580.351 
  1599.21,582.076 1603.28,583.835 1607.35,585.626 1611.43,587.448 1615.5,589.299 1619.57,591.177 1623.65,593.081 1627.72,595.007 1631.79,596.956 1635.86,598.923 
  1639.94,600.908 1644.01,602.908 1648.08,604.923 1652.16,606.948 1656.23,608.983 1660.3,611.026 1664.38,613.074 1668.45,615.124 1672.52,617.176 1676.6,619.228 
  1680.67,621.277 1684.74,623.321 1688.82,625.358 1692.89,627.386 1696.96,629.402 1701.04,631.406 1705.11,633.395 1709.18,635.367 1713.26,637.32 1717.33,639.252 
  1721.4,641.161 1725.48,643.045 1729.55,644.903 1733.62,646.732 1737.7,648.531 1741.77,650.297 1745.84,652.03 1749.92,653.726 1753.99,655.386 1758.06,657.006 
  1762.14,658.587 1766.21,660.124 1770.28,661.618 1774.36,663.067 1778.43,664.468 1782.5,665.822 1786.57,667.127 1790.65,668.38 1794.72,669.582 1798.79,670.73 
  1802.87,671.824 1806.94,672.862 1811.01,673.844 1815.09,674.769 1819.16,675.635 1823.23,676.442 1827.31,677.188 1831.38,677.874 1835.45,678.498 1839.53,679.06 
  1843.6,679.56 1847.67,679.996 1851.75,680.368 1855.82,680.677 1859.89,680.921 1863.97,681.1 1868.04,681.215 1872.11,681.265 1876.19,681.25 1880.26,681.17 
  1884.33,681.025 1888.41,680.816 1892.48,680.542 1896.55,680.204 1900.63,679.802 1904.7,679.336 1908.77,678.808 1912.85,678.217 1916.92,677.564 1920.99,676.85 
  1925.07,676.075 1929.14,675.241 1933.21,674.347 1937.28,673.396 1941.36,672.388 1945.43,671.323 1949.5,670.204 1953.58,669.03 1957.65,667.804 1961.72,666.527 
  1965.8,665.2 1969.87,663.823 1973.94,662.399 1978.02,660.929 1982.09,659.414 1986.16,657.856 1990.24,656.257 1994.31,654.618 1998.38,652.94 2002.46,651.226 
  2006.53,649.477 2010.6,647.695 2014.68,645.882 2018.75,644.039 2022.82,642.168 2026.9,640.272 2030.97,638.351 2035.04,636.408 2039.12,634.446 2043.19,632.466 
  2047.26,630.469 2051.34,628.458 2055.41,626.435 2059.48,624.402 2063.56,622.361 2067.63,620.314 2071.7,618.263 2075.78,616.21 2079.85,614.157 2083.92,612.106 
  2087.99,610.06 2092.07,608.02 2096.14,605.988 2100.21,603.967 2104.29,601.958 2108.36,599.963 2112.43,597.985 2116.51,596.025 2120.58,594.086 2124.65,592.169 
  2128.73,590.276 2132.8,588.409 2136.87,586.57 2140.95,584.761 2145.02,582.984 2149.09,581.24 2153.17,579.531 2157.24,577.859 2161.31,576.225 2165.39,574.632 
  2169.46,573.08 2173.53,571.572 2177.61,570.109 2181.68,568.692 2185.75,567.322 2189.83,566.002 2193.9,564.732 2197.97,563.513 2202.05,562.348 2206.12,561.237 
  2210.19,560.18 2214.27,559.18 2218.34,558.237 2222.41,557.353 2226.49,556.527 2230.56,555.762 2234.63,555.056 2238.71,554.413 2242.78,553.831 2246.85,553.311 
  2250.92,552.855 2255,552.463 2259.07,552.135 2263.14,551.87 2267.22,551.67 2271.29,551.535 2275.36,551.465 2279.44,551.46 2283.51,551.52 2287.58,551.645 
  2291.66,551.834 
  "/>
<polyline clip-path="url(#clip032)" style="stroke:#e26f46; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  255.035,735.102 259.108,728.627 263.181,722.162 267.254,715.709 271.328,709.276 275.401,702.868 279.474,696.49 283.547,690.15 287.621,683.853 291.694,677.608 
  295.767,671.421 299.84,665.299 303.914,659.246 307.987,653.268 312.06,647.372 316.133,641.565 320.207,635.85 324.28,630.234 328.353,624.724 332.426,619.324 
  336.5,614.039 340.573,608.875 344.646,603.837 348.719,598.931 352.793,594.16 356.866,589.53 360.939,585.047 365.012,580.711 369.086,576.529 373.159,572.504 
  377.232,568.641 381.305,564.946 385.379,561.421 389.452,558.072 393.525,554.894 397.598,551.896 401.671,549.081 405.745,546.453 409.818,544.015 413.891,541.77 
  417.964,539.717 422.038,537.858 426.111,536.196 430.184,534.733 434.257,533.472 438.331,532.414 442.404,531.559 446.477,530.907 450.55,530.459 454.624,530.216 
  458.697,530.179 462.77,530.347 466.843,530.72 470.917,531.298 474.99,532.08 479.063,533.065 483.136,534.252 487.21,535.641 491.283,537.229 495.356,539.014 
  499.429,540.997 503.503,543.173 507.576,545.542 511.649,548.101 515.722,550.846 519.796,553.775 523.869,556.886 527.942,560.176 532.015,563.641 536.089,567.277 
  540.162,571.081 544.235,575.049 548.308,579.177 552.382,583.461 556.455,587.898 560.528,592.481 564.601,597.208 568.675,602.072 572.748,607.069 576.821,612.193 
  580.894,617.442 584.968,622.809 589.041,628.288 593.114,633.873 597.187,639.56 601.261,645.342 605.334,651.214 609.407,657.171 613.48,663.206 617.553,669.313 
  621.627,675.485 625.7,681.716 629.773,688 633.846,694.334 637.92,700.708 641.993,707.116 646.066,713.553 650.139,720.01 654.213,726.482 658.286,732.963 
  662.359,739.447 666.432,745.927 670.506,752.396 674.579,758.848 678.652,765.274 682.725,771.67 686.799,778.032 690.872,784.351 694.945,790.62 699.018,796.834 
  703.092,802.985 707.165,809.068 711.238,815.077 715.311,821.007 719.385,826.852 723.458,832.604 727.531,838.259 731.604,843.81 735.678,849.251 739.751,854.58 
  743.824,859.79 747.897,864.874 751.971,869.829 756.044,874.649 760.117,879.328 764.19,883.863 768.264,888.25 772.337,892.485 776.41,896.562 780.483,900.477 
  784.557,904.226 788.63,907.805 792.703,911.212 796.776,914.444 800.85,917.496 804.923,920.366 808.996,923.049 813.069,925.545 817.142,927.849 821.216,929.962 
  825.289,931.88 829.362,933.601 833.435,935.123 837.509,936.445 841.582,937.565 845.655,938.483 849.728,939.197 853.802,939.708 857.875,940.013 861.948,940.114 
  866.021,940.009 870.095,939.699 874.168,939.184 878.241,938.466 882.314,937.544 886.388,936.419 890.461,935.093 894.534,933.566 898.607,931.841 902.681,929.92 
  906.754,927.803 910.827,925.493 914.9,922.993 918.974,920.305 923.047,917.432 927.12,914.376 931.193,911.14 935.267,907.728 939.34,904.144 943.413,900.391 
  947.486,896.472 951.56,892.392 955.633,888.154 959.706,883.763 963.779,879.223 967.853,874.539 971.926,869.715 975.999,864.758 980.072,859.67 984.146,854.456 
  988.219,849.124 992.292,843.677 996.365,838.122 1000.44,832.465 1004.51,826.71 1008.59,820.862 1012.66,814.928 1016.73,808.915 1020.8,802.827 1024.88,796.673 
  1028.95,790.458 1033.02,784.186 1037.1,777.864 1041.17,771.499 1045.24,765.099 1049.32,758.668 1053.39,752.215 1057.46,745.745 1061.54,739.262 1065.61,732.775 
  1069.68,726.291 1073.76,719.815 1077.83,713.356 1081.9,706.919 1085.98,700.509 1090.05,694.133 1094.12,687.797 1098.2,681.51 1102.27,675.276 1106.34,669.103 
  1110.42,662.996 1114.49,656.96 1118.56,651.002 1122.64,645.128 1126.71,639.344 1130.78,633.656 1134.86,628.071 1138.93,622.592 1143,617.225 1147.08,611.975 
  1151.15,606.849 1155.22,601.852 1159.3,596.988 1163.37,592.264 1167.44,587.68 1171.51,583.244 1175.59,578.96 1179.66,574.832 1183.73,570.865 1187.81,567.063 
  1191.88,563.429 1195.95,559.966 1200.03,556.677 1204.1,553.568 1208.17,550.64 1212.25,547.897 1216.32,545.343 1220.39,542.977 1224.47,540.802 1228.54,538.823 
  1232.61,537.04 1236.69,535.455 1240.76,534.071 1244.83,532.887 1248.91,531.905 1252.98,531.127 1257.05,530.553 1261.13,530.183 1265.2,530.019 1269.27,530.06 
  1273.35,530.306 1277.42,530.757 1281.49,531.412 1285.57,532.271 1289.64,533.334 1293.71,534.598 1297.79,536.063 1301.86,537.726 1305.93,539.588 1310.01,541.645 
  1314.08,543.896 1318.15,546.337 1322.22,548.968 1326.3,551.785 1330.37,554.786 1334.44,557.967 1338.52,561.326 1342.59,564.858 1346.66,568.56 1350.74,572.429 
  1354.81,576.461 1358.88,580.653 1362.96,584.998 1367.03,589.494 1371.1,594.135 1375.18,598.917 1379.25,603.836 1383.32,608.886 1387.4,614.063 1391.47,619.361 
  1395.54,624.775 1399.62,630.298 1403.69,635.926 1407.76,641.654 1411.84,647.477 1415.91,653.387 1419.98,659.378 1424.06,665.445 1428.13,671.581 1432.2,677.781 
  1436.28,684.039 1440.35,690.348 1444.42,696.703 1448.5,703.095 1452.57,709.519 1456.64,715.967 1460.72,722.437 1464.79,728.92 1468.86,735.408 1472.93,741.896 
  1477.01,748.377 1481.08,754.844 1485.15,761.292 1489.23,767.714 1493.3,774.105 1497.37,780.456 1501.45,786.761 1505.52,793.014 1509.59,799.208 1513.67,805.34 
  1517.74,811.402 1521.81,817.387 1525.89,823.29 1529.96,829.105 1534.03,834.824 1538.11,840.444 1542.18,845.96 1546.25,851.365 1550.33,856.654 1554.4,861.821 
  1558.47,866.86 1562.55,871.767 1566.62,876.538 1570.69,881.169 1574.77,885.653 1578.84,889.987 1582.91,894.165 1586.99,898.183 1591.06,902.038 1595.13,905.728 
  1599.21,909.246 1603.28,912.591 1607.35,915.757 1611.43,918.743 1615.5,921.544 1619.57,924.159 1623.65,926.586 1627.72,928.821 1631.79,930.862 1635.86,932.707 
  1639.94,934.354 1644.01,935.801 1648.08,937.049 1652.16,938.094 1656.23,938.936 1660.3,939.574 1664.38,940.008 1668.45,940.236 1672.52,940.258 1676.6,940.076 
  1680.67,939.689 1684.74,939.097 1688.82,938.301 1692.89,937.302 1696.96,936.1 1701.04,934.697 1705.11,933.094 1709.18,931.294 1713.26,929.297 1717.33,927.105 
  1721.4,924.722 1725.48,922.149 1729.55,919.388 1733.62,916.443 1737.7,913.316 1741.77,910.012 1745.84,906.532 1749.92,902.881 1753.99,899.062 1758.06,895.078 
  1762.14,890.934 1766.21,886.635 1770.28,882.184 1774.36,877.586 1778.43,872.846 1782.5,867.967 1786.57,862.955 1790.65,857.814 1794.72,852.552 1798.79,847.172 
  1802.87,841.681 1806.94,836.082 1811.01,830.381 1815.09,824.585 1819.16,818.7 1823.23,812.731 1827.31,806.685 1831.38,800.568 1835.45,794.384 1839.53,788.141 
  1843.6,781.844 1847.67,775.501 1851.75,769.118 1855.82,762.701 1859.89,756.256 1863.97,749.789 1868.04,743.307 1872.11,736.818 1876.19,730.326 1880.26,723.841 
  1884.33,717.367 1888.41,710.909 1892.48,704.475 1896.55,698.071 1900.63,691.705 1904.7,685.383 1908.77,679.111 1912.85,672.895 1916.92,666.74 1920.99,660.653 
  1925.07,654.641 1929.14,648.709 1933.21,642.865 1937.28,637.114 1941.36,631.459 1945.43,625.907 1949.5,620.465 1953.58,615.137 1957.65,609.93 1961.72,604.849 
  1965.8,599.899 1969.87,595.082 1973.94,590.405 1978.02,585.873 1982.09,581.49 1986.16,577.262 1990.24,573.193 1994.31,569.284 1998.38,565.541 2002.46,561.967 
  2006.53,558.567 2010.6,555.343 2014.68,552.3 2018.75,549.441 2022.82,546.765 2026.9,544.278 2030.97,541.983 2035.04,539.88 2039.12,537.973 2043.19,536.264 
  2047.26,534.754 2051.34,533.443 2055.41,532.334 2059.48,531.428 2063.56,530.726 2067.63,530.229 2071.7,529.937 2075.78,529.85 2079.85,529.968 2083.92,530.291 
  2087.99,530.82 2092.07,531.553 2096.14,532.489 2100.21,533.629 2104.29,534.97 2108.36,536.511 2112.43,538.251 2116.51,540.188 2120.58,542.32 2124.65,544.644 
  2128.73,547.16 2132.8,549.863 2136.87,552.752 2140.95,555.824 2145.02,559.075 2149.09,562.501 2153.17,566.101 2157.24,569.87 2161.31,573.804 2165.39,577.9 
  2169.46,582.153 2173.53,586.559 2177.61,591.113 2181.68,595.812 2185.75,600.65 2189.83,605.623 2193.9,610.726 2197.97,615.953 2202.05,621.298 2206.12,626.758 
  2210.19,632.327 2214.27,637.998 2218.34,643.767 2222.41,649.628 2226.49,655.573 2230.56,661.597 2234.63,667.696 2238.71,673.863 2242.78,680.091 2246.85,686.374 
  2250.92,692.705 2255,699.079 2259.07,705.488 2263.14,711.928 2267.22,718.392 2271.29,724.872 2275.36,731.362 2279.44,737.856 2283.51,744.346 2287.58,750.828 
  2291.66,757.295 
  "/>
<polyline clip-path="url(#clip032)" style="stroke:#3da44d; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  255.035,87.4932 259.108,87.8307 263.181,88.8074 267.254,90.4261 271.328,92.6875 275.401,95.5897 279.474,99.1318 283.547,103.309 287.621,108.12 291.694,113.558 
  295.767,119.618 299.84,126.293 303.914,133.576 307.987,141.457 312.06,149.932 316.133,158.993 320.207,168.627 324.28,178.825 328.353,189.582 332.426,200.887 
  336.5,212.72 340.573,225.077 344.646,237.947 348.719,251.315 352.793,265.162 356.866,279.482 360.939,294.26 365.012,309.475 369.086,325.11 373.159,341.155 
  377.232,357.593 381.305,374.412 385.379,391.595 389.452,409.126 393.525,426.973 397.598,445.128 401.671,463.574 405.745,482.295 409.818,501.273 413.891,520.49 
  417.964,539.922 422.038,559.548 426.111,579.352 430.184,599.314 434.257,619.414 438.331,639.63 442.404,659.944 446.477,680.337 450.55,700.787 454.624,721.272 
  458.697,741.771 462.77,762.263 466.843,782.725 470.917,803.142 474.99,823.497 479.063,843.763 483.136,863.92 487.21,883.947 491.283,903.823 495.356,923.526 
  499.429,943.048 503.503,962.362 507.576,981.449 511.649,1000.29 515.722,1018.86 519.796,1037.15 523.869,1055.13 527.942,1072.8 532.015,1090.13 536.089,1107.1 
  540.162,1123.7 544.235,1139.91 548.308,1155.72 552.382,1171.1 556.455,1186.06 560.528,1200.56 564.601,1214.59 568.675,1228.14 572.748,1241.2 576.821,1253.75 
  580.894,1265.79 584.968,1277.3 589.041,1288.26 593.114,1298.67 597.187,1308.51 601.261,1317.78 605.334,1326.47 609.407,1334.57 613.48,1342.07 617.553,1348.96 
  621.627,1355.23 625.7,1360.89 629.773,1365.91 633.846,1370.31 637.92,1374.07 641.993,1377.2 646.066,1379.68 650.139,1381.51 654.213,1382.7 658.286,1383.24 
  662.359,1383.13 666.432,1382.38 670.506,1380.98 674.579,1378.93 678.652,1376.23 682.725,1372.89 686.799,1368.92 690.872,1364.31 694.945,1359.07 699.018,1353.21 
  703.092,1346.73 707.165,1339.64 711.238,1331.94 715.311,1323.64 719.385,1314.76 723.458,1305.29 727.531,1295.26 731.604,1284.66 735.678,1273.52 739.751,1261.83 
  743.824,1249.62 747.897,1236.9 751.971,1223.67 756.044,1209.95 760.117,1195.76 764.19,1181.1 768.264,1166 772.337,1150.47 776.41,1134.52 780.483,1118.17 
  784.557,1101.45 788.63,1084.35 792.703,1066.9 796.776,1049.12 800.85,1031.03 804.923,1012.64 808.996,993.971 813.069,975.046 817.142,955.882 821.216,936.493 
  825.289,916.901 829.362,897.128 833.435,877.193 837.509,857.118 841.582,836.924 845.655,816.623 849.728,796.239 853.802,775.793 857.875,755.307 861.948,734.803 
  866.021,714.3 870.095,693.82 874.168,673.375 878.241,652.992 882.314,632.691 886.388,612.493 890.461,592.419 894.534,572.491 898.607,552.723 902.681,533.134 
  906.754,513.747 910.827,494.581 914.9,475.658 918.974,456.996 923.047,438.616 927.12,420.524 931.193,402.748 935.267,385.304 939.34,368.21 943.413,351.486 
  947.486,335.148 951.56,319.208 955.633,303.681 959.706,288.585 963.779,273.936 967.853,259.75 971.926,246.041 975.999,232.825 980.072,220.105 984.146,207.9 
  988.219,196.222 992.292,185.084 996.365,174.498 1000.44,164.475 1004.51,155.022 1008.59,146.146 1012.66,137.86 1016.73,130.171 1020.8,123.088 1024.88,116.619 
  1028.95,110.77 1033.02,105.543 1037.1,100.946 1041.17,96.9824 1045.24,93.6581 1049.32,90.9761 1053.39,88.9397 1057.46,87.5503 1061.54,86.8068 1065.61,86.7119 
  1069.68,87.266 1073.76,88.4685 1077.83,90.3184 1081.9,92.8142 1085.98,95.9521 1090.05,99.7292 1094.12,104.142 1098.2,109.186 1102.27,114.857 1106.34,121.148 
  1110.42,128.054 1114.49,135.568 1118.56,143.681 1122.64,152.387 1126.71,161.675 1130.78,171.537 1134.86,181.962 1138.93,192.941 1143,204.464 1147.08,216.518 
  1151.15,229.091 1155.22,242.17 1159.3,255.741 1163.37,269.791 1167.44,284.309 1171.51,299.279 1175.59,314.685 1179.66,330.511 1183.73,346.741 1187.81,363.358 
  1191.88,380.349 1195.95,397.697 1200.03,415.383 1204.1,433.388 1208.17,451.695 1212.25,470.283 1216.32,489.135 1220.39,508.238 1224.47,527.568 1228.54,547.106 
  1232.61,566.832 1236.69,586.725 1240.76,606.764 1244.83,626.932 1248.91,647.213 1252.98,667.582 1257.05,688.018 1261.13,708.501 1265.2,729.008 1269.27,749.519 
  1273.35,770.02 1277.42,790.488 1281.49,810.901 1285.57,831.237 1289.64,851.476 1293.71,871.596 1297.79,891.579 1301.86,911.411 1305.93,931.066 1310.01,950.526 
  1314.08,969.769 1318.15,988.775 1322.22,1007.53 1326.3,1026.01 1330.37,1044.2 1334.44,1062.08 1338.52,1079.64 1342.59,1096.85 1346.66,1113.69 1350.74,1130.16 
  1354.81,1146.23 1358.88,1161.9 1362.96,1177.14 1367.03,1191.93 1371.1,1206.27 1375.18,1220.13 1379.25,1233.51 1383.32,1246.39 1387.4,1258.76 1391.47,1270.61 
  1395.54,1281.92 1399.62,1292.68 1403.69,1302.89 1407.76,1312.52 1411.84,1321.59 1415.91,1330.06 1419.98,1337.94 1424.06,1345.21 1428.13,1351.88 1432.2,1357.92 
  1436.28,1363.35 1440.35,1368.14 1444.42,1372.31 1448.5,1375.83 1452.57,1378.71 1456.64,1380.95 1460.72,1382.54 1464.79,1383.49 1468.86,1383.79 1472.93,1383.43 
  1477.01,1382.43 1481.08,1380.78 1485.15,1378.49 1489.23,1375.55 1493.3,1371.97 1497.37,1367.75 1501.45,1362.9 1505.52,1357.42 1509.59,1351.32 1513.67,1344.6 
  1517.74,1337.28 1521.81,1329.35 1525.89,1320.82 1529.96,1311.71 1534.03,1302.02 1538.11,1291.77 1542.18,1280.95 1546.25,1269.6 1550.33,1257.7 1554.4,1245.28 
  1558.47,1232.36 1562.55,1218.93 1566.62,1205.02 1570.69,1190.64 1574.77,1175.81 1578.84,1160.53 1582.91,1144.83 1586.99,1128.72 1591.06,1112.21 1595.13,1095.33 
  1599.21,1078.08 1603.28,1060.49 1607.35,1042.58 1611.43,1024.36 1615.5,1005.85 1619.57,987.069 1623.65,968.032 1627.72,948.763 1631.79,929.28 1635.86,909.604 
  1639.94,889.756 1644.01,869.754 1648.08,849.612 1652.16,829.355 1656.23,809.003 1660.3,788.578 1664.38,768.102 1668.45,747.594 1672.52,727.073 1676.6,706.556 
  1680.67,686.066 1684.74,665.627 1688.82,645.257 1692.89,624.98 1696.96,604.816 1701.04,584.776 1705.11,564.885 1709.18,545.165 1713.26,525.635 1717.33,506.316 
  1721.4,487.229 1725.48,468.39 1729.55,449.812 1733.62,431.52 1737.7,413.531 1741.77,395.865 1745.84,378.541 1749.92,361.576 1753.99,344.98 1758.06,328.773 
  1762.14,312.972 1766.21,297.594 1770.28,282.656 1774.36,268.172 1778.43,254.156 1782.5,240.616 1786.57,227.571 1790.65,215.033 1794.72,203.017 1798.79,191.535 
  1802.87,180.599 1806.94,170.215 1811.01,160.394 1815.09,151.148 1819.16,142.487 1823.23,134.42 1827.31,126.955 1831.38,120.1 1835.45,113.856 1839.53,108.234 
  1843.6,103.24 1847.67,98.8776 1851.75,95.153 1855.82,92.0701 1859.89,89.6296 1863.97,87.8335 1868.04,86.6851 1872.11,86.1857 1876.19,86.3359 1880.26,87.136 
  1884.33,88.5851 1888.41,90.6797 1892.48,93.4191 1896.55,96.8006 1900.63,100.821 1904.7,105.476 1908.77,110.762 1912.85,116.672 1916.92,123.201 1920.99,130.342 
  1925.07,138.089 1929.14,146.432 1933.21,155.365 1937.28,164.877 1941.36,174.961 1945.43,185.606 1949.5,196.801 1953.58,208.534 1957.65,220.794 1961.72,233.567 
  1965.8,246.843 1969.87,260.609 1973.94,274.85 1978.02,289.552 1982.09,304.699 1986.16,320.276 1990.24,336.266 1994.31,352.658 1998.38,369.434 2002.46,386.576 
  2006.53,404.067 2010.6,421.887 2014.68,440.02 2018.75,458.447 2022.82,477.156 2026.9,496.123 2030.97,515.329 2035.04,534.754 2039.12,554.378 2043.19,574.18 
  2047.26,594.147 2051.34,614.258 2055.41,634.489 2059.48,654.821 2063.56,675.232 2067.63,695.7 2071.7,716.207 2075.78,736.738 2079.85,757.269 2083.92,777.777 
  2087.99,798.241 2092.07,818.641 2096.14,838.954 2100.21,859.166 2104.29,879.258 2108.36,899.205 2112.43,918.988 2116.51,938.586 2120.58,957.978 2124.65,977.144 
  2128.73,996.075 2132.8,1014.75 2136.87,1033.14 2140.95,1051.23 2145.02,1069 2149.09,1086.44 2153.17,1103.52 2157.24,1120.25 2161.31,1136.58 2165.39,1152.52 
  2169.46,1168.03 2173.53,1183.12 2177.61,1197.75 2181.68,1211.92 2185.75,1225.62 2189.83,1238.82 2193.9,1251.52 2197.97,1263.71 2202.05,1275.36 2206.12,1286.47 
  2210.19,1297.03 2214.27,1307.04 2218.34,1316.47 2222.41,1325.31 2226.49,1333.57 2230.56,1341.22 2234.63,1348.27 2238.71,1354.71 2242.78,1360.53 2246.85,1365.72 
  2250.92,1370.28 2255,1374.21 2259.07,1377.49 2263.14,1380.14 2267.22,1382.13 2271.29,1383.48 2275.36,1384.19 2279.44,1384.24 2283.51,1383.64 2287.58,1382.39 
  2291.66,1380.5 
  "/>
<path clip-path="url(#clip030)" d="
M1557.04 300.469 L2280.8 300.469 L2280.8 93.1086 L1557.04 93.1086  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<polyline clip-path="url(#clip030)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1557.04,300.469 2280.8,300.469 2280.8,93.1086 1557.04,93.1086 1557.04,300.469 
  "/>
<polyline clip-path="url(#clip030)" style="stroke:#009af9; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1581.03,144.949 1724.95,144.949 
  "/>
<path clip-path="url(#clip030)" d="M1769.28 141.28 Q1770.88 138.409 1773.1 137.044 Q1775.32 135.678 1778.33 135.678 Q1782.38 135.678 1784.58 138.525 Q1786.78 141.349 1786.78 146.581 L1786.78 162.229 L1782.5 162.229 L1782.5 146.719 Q1782.5 142.993 1781.18 141.187 Q1779.86 139.382 1777.15 139.382 Q1773.84 139.382 1771.92 141.581 Q1770 143.78 1770 147.576 L1770 162.229 L1765.72 162.229 L1765.72 146.719 Q1765.72 142.969 1764.4 141.187 Q1763.08 139.382 1760.32 139.382 Q1757.06 139.382 1755.14 141.604 Q1753.22 143.803 1753.22 147.576 L1753.22 162.229 L1748.94 162.229 L1748.94 136.303 L1753.22 136.303 L1753.22 140.331 Q1754.68 137.946 1756.71 136.812 Q1758.75 135.678 1761.55 135.678 Q1764.37 135.678 1766.34 137.113 Q1768.33 138.548 1769.28 141.28 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1807.06 149.196 Q1801.9 149.196 1799.91 150.377 Q1797.92 151.557 1797.92 154.405 Q1797.92 156.673 1799.4 158.016 Q1800.9 159.335 1803.47 159.335 Q1807.01 159.335 1809.14 156.835 Q1811.3 154.312 1811.3 150.145 L1811.3 149.196 L1807.06 149.196 M1815.56 147.437 L1815.56 162.229 L1811.3 162.229 L1811.3 158.293 Q1809.84 160.655 1807.66 161.789 Q1805.49 162.9 1802.34 162.9 Q1798.36 162.9 1796 160.678 Q1793.66 158.432 1793.66 154.682 Q1793.66 150.307 1796.57 148.085 Q1799.51 145.863 1805.32 145.863 L1811.3 145.863 L1811.3 145.446 Q1811.3 142.507 1809.35 140.909 Q1807.43 139.289 1803.93 139.289 Q1801.71 139.289 1799.61 139.821 Q1797.5 140.354 1795.56 141.419 L1795.56 137.483 Q1797.89 136.581 1800.09 136.141 Q1802.29 135.678 1804.37 135.678 Q1810 135.678 1812.78 138.594 Q1815.56 141.511 1815.56 147.437 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1840.86 137.067 L1840.86 141.094 Q1839.05 140.169 1837.11 139.706 Q1835.16 139.243 1833.08 139.243 Q1829.91 139.243 1828.31 140.215 Q1826.74 141.187 1826.74 143.131 Q1826.74 144.613 1827.87 145.469 Q1829 146.303 1832.43 147.067 L1833.89 147.391 Q1838.43 148.363 1840.32 150.145 Q1842.24 151.905 1842.24 155.076 Q1842.24 158.687 1839.37 160.793 Q1836.53 162.9 1831.53 162.9 Q1829.44 162.9 1827.18 162.483 Q1824.93 162.09 1822.43 161.28 L1822.43 156.881 Q1824.79 158.108 1827.08 158.733 Q1829.37 159.335 1831.62 159.335 Q1834.63 159.335 1836.25 158.317 Q1837.87 157.275 1837.87 155.4 Q1837.87 153.664 1836.69 152.738 Q1835.53 151.812 1831.57 150.956 L1830.09 150.608 Q1826.13 149.775 1824.37 148.062 Q1822.62 146.326 1822.62 143.317 Q1822.62 139.659 1825.21 137.669 Q1827.8 135.678 1832.57 135.678 Q1834.93 135.678 1837.01 136.025 Q1839.1 136.372 1840.86 137.067 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1865.55 137.067 L1865.55 141.094 Q1863.75 140.169 1861.8 139.706 Q1859.86 139.243 1857.78 139.243 Q1854.61 139.243 1853.01 140.215 Q1851.43 141.187 1851.43 143.131 Q1851.43 144.613 1852.57 145.469 Q1853.7 146.303 1857.13 147.067 L1858.59 147.391 Q1863.12 148.363 1865.02 150.145 Q1866.94 151.905 1866.94 155.076 Q1866.94 158.687 1864.07 160.793 Q1861.23 162.9 1856.23 162.9 Q1854.14 162.9 1851.87 162.483 Q1849.63 162.09 1847.13 161.28 L1847.13 156.881 Q1849.49 158.108 1851.78 158.733 Q1854.07 159.335 1856.32 159.335 Q1859.33 159.335 1860.95 158.317 Q1862.57 157.275 1862.57 155.4 Q1862.57 153.664 1861.39 152.738 Q1860.23 151.812 1856.27 150.956 L1854.79 150.608 Q1850.83 149.775 1849.07 148.062 Q1847.31 146.326 1847.31 143.317 Q1847.31 139.659 1849.91 137.669 Q1852.5 135.678 1857.27 135.678 Q1859.63 135.678 1861.71 136.025 Q1863.8 136.372 1865.55 137.067 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1874.33 156.349 L1879.21 156.349 L1879.21 162.229 L1874.33 162.229 L1874.33 156.349 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1905.32 137.067 L1905.32 141.094 Q1903.52 140.169 1901.57 139.706 Q1899.63 139.243 1897.55 139.243 Q1894.37 139.243 1892.78 140.215 Q1891.2 141.187 1891.2 143.131 Q1891.2 144.613 1892.34 145.469 Q1893.47 146.303 1896.9 147.067 L1898.36 147.391 Q1902.89 148.363 1904.79 150.145 Q1906.71 151.905 1906.71 155.076 Q1906.71 158.687 1903.84 160.793 Q1900.99 162.9 1895.99 162.9 Q1893.91 162.9 1891.64 162.483 Q1889.4 162.09 1886.9 161.28 L1886.9 156.881 Q1889.26 158.108 1891.55 158.733 Q1893.84 159.335 1896.09 159.335 Q1899.1 159.335 1900.72 158.317 Q1902.34 157.275 1902.34 155.4 Q1902.34 153.664 1901.16 152.738 Q1900 151.812 1896.04 150.956 L1894.56 150.608 Q1890.6 149.775 1888.84 148.062 Q1887.08 146.326 1887.08 143.317 Q1887.08 139.659 1889.68 137.669 Q1892.27 135.678 1897.04 135.678 Q1899.4 135.678 1901.48 136.025 Q1903.56 136.372 1905.32 137.067 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1938.8 126.257 Q1935.69 131.581 1934.19 136.789 Q1932.68 141.997 1932.68 147.344 Q1932.68 152.692 1934.19 157.946 Q1935.72 163.178 1938.8 168.479 L1935.09 168.479 Q1931.62 163.039 1929.88 157.784 Q1928.17 152.53 1928.17 147.344 Q1928.17 142.182 1929.88 136.951 Q1931.6 131.72 1935.09 126.257 L1938.8 126.257 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1961.83 143.594 Q1965.18 144.312 1967.06 146.581 Q1968.96 148.849 1968.96 152.182 Q1968.96 157.298 1965.44 160.099 Q1961.92 162.9 1955.44 162.9 Q1953.26 162.9 1950.95 162.46 Q1948.66 162.043 1946.2 161.187 L1946.2 156.673 Q1948.15 157.807 1950.46 158.386 Q1952.78 158.965 1955.3 158.965 Q1959.7 158.965 1961.99 157.229 Q1964.3 155.493 1964.3 152.182 Q1964.3 149.127 1962.15 147.414 Q1960.02 145.678 1956.2 145.678 L1952.17 145.678 L1952.17 141.835 L1956.39 141.835 Q1959.84 141.835 1961.67 140.469 Q1963.49 139.081 1963.49 136.488 Q1963.49 133.826 1961.6 132.414 Q1959.72 130.979 1956.2 130.979 Q1954.28 130.979 1952.08 131.395 Q1949.88 131.812 1947.24 132.692 L1947.24 128.525 Q1949.91 127.784 1952.22 127.414 Q1954.56 127.044 1956.62 127.044 Q1961.94 127.044 1965.05 129.474 Q1968.15 131.882 1968.15 136.002 Q1968.15 138.872 1966.5 140.863 Q1964.86 142.831 1961.83 143.594 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1991.99 143.594 Q1995.35 144.312 1997.22 146.581 Q1999.12 148.849 1999.12 152.182 Q1999.12 157.298 1995.6 160.099 Q1992.08 162.9 1985.6 162.9 Q1983.42 162.9 1981.11 162.46 Q1978.82 162.043 1976.36 161.187 L1976.36 156.673 Q1978.31 157.807 1980.62 158.386 Q1982.94 158.965 1985.46 158.965 Q1989.86 158.965 1992.15 157.229 Q1994.47 155.493 1994.47 152.182 Q1994.47 149.127 1992.31 147.414 Q1990.18 145.678 1986.36 145.678 L1982.34 145.678 L1982.34 141.835 L1986.55 141.835 Q1990 141.835 1991.83 140.469 Q1993.66 139.081 1993.66 136.488 Q1993.66 133.826 1991.76 132.414 Q1989.88 130.979 1986.36 130.979 Q1984.44 130.979 1982.24 131.395 Q1980.05 131.812 1977.41 132.692 L1977.41 128.525 Q1980.07 127.784 1982.38 127.414 Q1984.72 127.044 1986.78 127.044 Q1992.11 127.044 1995.21 129.474 Q1998.31 131.882 1998.31 136.002 Q1998.31 138.872 1996.67 140.863 Q1995.02 142.831 1991.99 143.594 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M2008.03 127.669 L2026.39 127.669 L2026.39 131.604 L2012.31 131.604 L2012.31 140.076 Q2013.33 139.729 2014.35 139.567 Q2015.37 139.382 2016.39 139.382 Q2022.17 139.382 2025.55 142.553 Q2028.93 145.724 2028.93 151.141 Q2028.93 156.719 2025.46 159.821 Q2021.99 162.9 2015.67 162.9 Q2013.49 162.9 2011.23 162.53 Q2008.98 162.159 2006.57 161.418 L2006.57 156.719 Q2008.66 157.854 2010.88 158.409 Q2013.1 158.965 2015.58 158.965 Q2019.58 158.965 2021.92 156.858 Q2024.26 154.752 2024.26 151.141 Q2024.26 147.53 2021.92 145.423 Q2019.58 143.317 2015.58 143.317 Q2013.7 143.317 2011.83 143.733 Q2009.98 144.15 2008.03 145.03 L2008.03 127.669 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M2038.19 127.669 L2056.55 127.669 L2056.55 131.604 L2042.48 131.604 L2042.48 140.076 Q2043.49 139.729 2044.51 139.567 Q2045.53 139.382 2046.55 139.382 Q2052.34 139.382 2055.72 142.553 Q2059.1 145.724 2059.1 151.141 Q2059.1 156.719 2055.62 159.821 Q2052.15 162.9 2045.83 162.9 Q2043.66 162.9 2041.39 162.53 Q2039.14 162.159 2036.73 161.418 L2036.73 156.719 Q2038.82 157.854 2041.04 158.409 Q2043.26 158.965 2045.74 158.965 Q2049.74 158.965 2052.08 156.858 Q2054.42 154.752 2054.42 151.141 Q2054.42 147.53 2052.08 145.423 Q2049.74 143.317 2045.74 143.317 Q2043.86 143.317 2041.99 143.733 Q2040.14 144.15 2038.19 145.03 L2038.19 127.669 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M2081.16 131.743 L2069.35 150.192 L2081.16 150.192 L2081.16 131.743 M2079.93 127.669 L2085.81 127.669 L2085.81 150.192 L2090.74 150.192 L2090.74 154.081 L2085.81 154.081 L2085.81 162.229 L2081.16 162.229 L2081.16 154.081 L2065.55 154.081 L2065.55 149.567 L2079.93 127.669 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M2111.32 131.743 L2099.51 150.192 L2111.32 150.192 L2111.32 131.743 M2110.09 127.669 L2115.97 127.669 L2115.97 150.192 L2120.9 150.192 L2120.9 154.081 L2115.97 154.081 L2115.97 162.229 L2111.32 162.229 L2111.32 154.081 L2095.72 154.081 L2095.72 149.567 L2110.09 127.669 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M2142.8 143.594 Q2146.16 144.312 2148.03 146.581 Q2149.93 148.849 2149.93 152.182 Q2149.93 157.298 2146.41 160.099 Q2142.89 162.9 2136.41 162.9 Q2134.23 162.9 2131.92 162.46 Q2129.63 162.043 2127.17 161.187 L2127.17 156.673 Q2129.12 157.807 2131.43 158.386 Q2133.75 158.965 2136.27 158.965 Q2140.67 158.965 2142.96 157.229 Q2145.28 155.493 2145.28 152.182 Q2145.28 149.127 2143.12 147.414 Q2140.99 145.678 2137.17 145.678 L2133.15 145.678 L2133.15 141.835 L2137.36 141.835 Q2140.81 141.835 2142.64 140.469 Q2144.47 139.081 2144.47 136.488 Q2144.47 133.826 2142.57 132.414 Q2140.69 130.979 2137.17 130.979 Q2135.25 130.979 2133.05 131.395 Q2130.85 131.812 2128.22 132.692 L2128.22 128.525 Q2130.88 127.784 2133.19 127.414 Q2135.53 127.044 2137.59 127.044 Q2142.91 127.044 2146.02 129.474 Q2149.12 131.882 2149.12 136.002 Q2149.12 138.872 2147.47 140.863 Q2145.83 142.831 2142.8 143.594 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M2162.82 158.293 L2179.14 158.293 L2179.14 162.229 L2157.2 162.229 L2157.2 158.293 Q2159.86 155.539 2164.44 150.909 Q2169.05 146.256 2170.23 144.914 Q2172.47 142.391 2173.35 140.655 Q2174.26 138.895 2174.26 137.206 Q2174.26 134.451 2172.31 132.715 Q2170.39 130.979 2167.29 130.979 Q2165.09 130.979 2162.64 131.743 Q2160.21 132.507 2157.43 134.057 L2157.43 129.335 Q2160.25 128.201 2162.71 127.622 Q2165.16 127.044 2167.2 127.044 Q2172.57 127.044 2175.76 129.729 Q2178.96 132.414 2178.96 136.905 Q2178.96 139.034 2178.15 140.956 Q2177.36 142.854 2175.25 145.446 Q2174.67 146.118 2171.57 149.335 Q2168.47 152.53 2162.82 158.293 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M2187.68 126.257 L2191.39 126.257 Q2194.86 131.72 2196.57 136.951 Q2198.31 142.182 2198.31 147.344 Q2198.31 152.53 2196.57 157.784 Q2194.86 163.039 2191.39 168.479 L2187.68 168.479 Q2190.76 163.178 2192.27 157.946 Q2193.79 152.692 2193.79 147.344 Q2193.79 141.997 2192.27 136.789 Q2190.76 131.581 2187.68 126.257 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip030)" style="stroke:#e26f46; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1581.03,196.789 1724.95,196.789 
  "/>
<path clip-path="url(#clip030)" d="M1769.28 193.12 Q1770.88 190.249 1773.1 188.884 Q1775.32 187.518 1778.33 187.518 Q1782.38 187.518 1784.58 190.365 Q1786.78 193.189 1786.78 198.421 L1786.78 214.069 L1782.5 214.069 L1782.5 198.559 Q1782.5 194.833 1781.18 193.027 Q1779.86 191.222 1777.15 191.222 Q1773.84 191.222 1771.92 193.421 Q1770 195.62 1770 199.416 L1770 214.069 L1765.72 214.069 L1765.72 198.559 Q1765.72 194.809 1764.4 193.027 Q1763.08 191.222 1760.32 191.222 Q1757.06 191.222 1755.14 193.444 Q1753.22 195.643 1753.22 199.416 L1753.22 214.069 L1748.94 214.069 L1748.94 188.143 L1753.22 188.143 L1753.22 192.171 Q1754.68 189.786 1756.71 188.652 Q1758.75 187.518 1761.55 187.518 Q1764.37 187.518 1766.34 188.953 Q1768.33 190.388 1769.28 193.12 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1807.06 201.036 Q1801.9 201.036 1799.91 202.217 Q1797.92 203.397 1797.92 206.245 Q1797.92 208.513 1799.4 209.856 Q1800.9 211.175 1803.47 211.175 Q1807.01 211.175 1809.14 208.675 Q1811.3 206.152 1811.3 201.985 L1811.3 201.036 L1807.06 201.036 M1815.56 199.277 L1815.56 214.069 L1811.3 214.069 L1811.3 210.133 Q1809.84 212.495 1807.66 213.629 Q1805.49 214.74 1802.34 214.74 Q1798.36 214.74 1796 212.518 Q1793.66 210.272 1793.66 206.522 Q1793.66 202.147 1796.57 199.925 Q1799.51 197.703 1805.32 197.703 L1811.3 197.703 L1811.3 197.286 Q1811.3 194.347 1809.35 192.749 Q1807.43 191.129 1803.93 191.129 Q1801.71 191.129 1799.61 191.661 Q1797.5 192.194 1795.56 193.259 L1795.56 189.323 Q1797.89 188.421 1800.09 187.981 Q1802.29 187.518 1804.37 187.518 Q1810 187.518 1812.78 190.434 Q1815.56 193.351 1815.56 199.277 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1840.86 188.907 L1840.86 192.934 Q1839.05 192.009 1837.11 191.546 Q1835.16 191.083 1833.08 191.083 Q1829.91 191.083 1828.31 192.055 Q1826.74 193.027 1826.74 194.971 Q1826.74 196.453 1827.87 197.309 Q1829 198.143 1832.43 198.907 L1833.89 199.231 Q1838.43 200.203 1840.32 201.985 Q1842.24 203.745 1842.24 206.916 Q1842.24 210.527 1839.37 212.633 Q1836.53 214.74 1831.53 214.74 Q1829.44 214.74 1827.18 214.323 Q1824.93 213.93 1822.43 213.12 L1822.43 208.721 Q1824.79 209.948 1827.08 210.573 Q1829.37 211.175 1831.62 211.175 Q1834.63 211.175 1836.25 210.157 Q1837.87 209.115 1837.87 207.24 Q1837.87 205.504 1836.69 204.578 Q1835.53 203.652 1831.57 202.796 L1830.09 202.448 Q1826.13 201.615 1824.37 199.902 Q1822.62 198.166 1822.62 195.157 Q1822.62 191.499 1825.21 189.509 Q1827.8 187.518 1832.57 187.518 Q1834.93 187.518 1837.01 187.865 Q1839.1 188.212 1840.86 188.907 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1865.55 188.907 L1865.55 192.934 Q1863.75 192.009 1861.8 191.546 Q1859.86 191.083 1857.78 191.083 Q1854.61 191.083 1853.01 192.055 Q1851.43 193.027 1851.43 194.971 Q1851.43 196.453 1852.57 197.309 Q1853.7 198.143 1857.13 198.907 L1858.59 199.231 Q1863.12 200.203 1865.02 201.985 Q1866.94 203.745 1866.94 206.916 Q1866.94 210.527 1864.07 212.633 Q1861.23 214.74 1856.23 214.74 Q1854.14 214.74 1851.87 214.323 Q1849.63 213.93 1847.13 213.12 L1847.13 208.721 Q1849.49 209.948 1851.78 210.573 Q1854.07 211.175 1856.32 211.175 Q1859.33 211.175 1860.95 210.157 Q1862.57 209.115 1862.57 207.24 Q1862.57 205.504 1861.39 204.578 Q1860.23 203.652 1856.27 202.796 L1854.79 202.448 Q1850.83 201.615 1849.07 199.902 Q1847.31 198.166 1847.31 195.157 Q1847.31 191.499 1849.91 189.509 Q1852.5 187.518 1857.27 187.518 Q1859.63 187.518 1861.71 187.865 Q1863.8 188.212 1865.55 188.907 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1874.33 208.189 L1879.21 208.189 L1879.21 214.069 L1874.33 214.069 L1874.33 208.189 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1885.74 188.143 L1890.25 188.143 L1898.36 209.902 L1906.46 188.143 L1910.97 188.143 L1901.25 214.069 L1895.46 214.069 L1885.74 188.143 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1942.15 178.097 Q1939.05 183.421 1937.55 188.629 Q1936.04 193.837 1936.04 199.184 Q1936.04 204.532 1937.55 209.786 Q1939.07 215.018 1942.15 220.319 L1938.45 220.319 Q1934.98 214.879 1933.24 209.624 Q1931.53 204.37 1931.53 199.184 Q1931.53 194.022 1933.24 188.791 Q1934.95 183.56 1938.45 178.097 L1942.15 178.097 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1965.18 195.434 Q1968.54 196.152 1970.42 198.421 Q1972.31 200.689 1972.31 204.022 Q1972.31 209.138 1968.8 211.939 Q1965.28 214.74 1958.8 214.74 Q1956.62 214.74 1954.3 214.3 Q1952.01 213.883 1949.56 213.027 L1949.56 208.513 Q1951.5 209.647 1953.82 210.226 Q1956.13 210.805 1958.66 210.805 Q1963.05 210.805 1965.35 209.069 Q1967.66 207.333 1967.66 204.022 Q1967.66 200.967 1965.51 199.254 Q1963.38 197.518 1959.56 197.518 L1955.53 197.518 L1955.53 193.675 L1959.74 193.675 Q1963.19 193.675 1965.02 192.309 Q1966.85 190.921 1966.85 188.328 Q1966.85 185.666 1964.95 184.254 Q1963.08 182.819 1959.56 182.819 Q1957.64 182.819 1955.44 183.235 Q1953.24 183.652 1950.6 184.532 L1950.6 180.365 Q1953.26 179.624 1955.58 179.254 Q1957.92 178.884 1959.98 178.884 Q1965.3 178.884 1968.4 181.314 Q1971.5 183.722 1971.5 187.842 Q1971.5 190.712 1969.86 192.703 Q1968.22 194.671 1965.18 195.434 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1995.35 195.434 Q1998.7 196.152 2000.58 198.421 Q2002.48 200.689 2002.48 204.022 Q2002.48 209.138 1998.96 211.939 Q1995.44 214.74 1988.96 214.74 Q1986.78 214.74 1984.47 214.3 Q1982.17 213.883 1979.72 213.027 L1979.72 208.513 Q1981.67 209.647 1983.98 210.226 Q1986.3 210.805 1988.82 210.805 Q1993.22 210.805 1995.51 209.069 Q1997.82 207.333 1997.82 204.022 Q1997.82 200.967 1995.67 199.254 Q1993.54 197.518 1989.72 197.518 L1985.69 197.518 L1985.69 193.675 L1989.91 193.675 Q1993.36 193.675 1995.18 192.309 Q1997.01 190.921 1997.01 188.328 Q1997.01 185.666 1995.11 184.254 Q1993.24 182.819 1989.72 182.819 Q1987.8 182.819 1985.6 183.235 Q1983.4 183.652 1980.76 184.532 L1980.76 180.365 Q1983.42 179.624 1985.74 179.254 Q1988.08 178.884 1990.14 178.884 Q1995.46 178.884 1998.56 181.314 Q2001.67 183.722 2001.67 187.842 Q2001.67 190.712 2000.02 192.703 Q1998.38 194.671 1995.35 195.434 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M2011.39 179.509 L2029.74 179.509 L2029.74 183.444 L2015.67 183.444 L2015.67 191.916 Q2016.69 191.569 2017.71 191.407 Q2018.73 191.222 2019.74 191.222 Q2025.53 191.222 2028.91 194.393 Q2032.29 197.564 2032.29 202.981 Q2032.29 208.559 2028.82 211.661 Q2025.35 214.74 2019.03 214.74 Q2016.85 214.74 2014.58 214.37 Q2012.34 213.999 2009.93 213.258 L2009.93 208.559 Q2012.01 209.694 2014.23 210.249 Q2016.46 210.805 2018.93 210.805 Q2022.94 210.805 2025.28 208.698 Q2027.61 206.592 2027.61 202.981 Q2027.61 199.37 2025.28 197.263 Q2022.94 195.157 2018.93 195.157 Q2017.06 195.157 2015.18 195.573 Q2013.33 195.99 2011.39 196.87 L2011.39 179.509 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M2041.55 179.509 L2059.91 179.509 L2059.91 183.444 L2045.83 183.444 L2045.83 191.916 Q2046.85 191.569 2047.87 191.407 Q2048.89 191.222 2049.91 191.222 Q2055.69 191.222 2059.07 194.393 Q2062.45 197.564 2062.45 202.981 Q2062.45 208.559 2058.98 211.661 Q2055.51 214.74 2049.19 214.74 Q2047.01 214.74 2044.74 214.37 Q2042.5 213.999 2040.09 213.258 L2040.09 208.559 Q2042.17 209.694 2044.4 210.249 Q2046.62 210.805 2049.1 210.805 Q2053.1 210.805 2055.44 208.698 Q2057.78 206.592 2057.78 202.981 Q2057.78 199.37 2055.44 197.263 Q2053.1 195.157 2049.1 195.157 Q2047.22 195.157 2045.35 195.573 Q2043.49 195.99 2041.55 196.87 L2041.55 179.509 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M2084.51 183.583 L2072.71 202.032 L2084.51 202.032 L2084.51 183.583 M2083.29 179.509 L2089.16 179.509 L2089.16 202.032 L2094.1 202.032 L2094.1 205.921 L2089.16 205.921 L2089.16 214.069 L2084.51 214.069 L2084.51 205.921 L2068.91 205.921 L2068.91 201.407 L2083.29 179.509 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M2114.67 183.583 L2102.87 202.032 L2114.67 202.032 L2114.67 183.583 M2113.45 179.509 L2119.33 179.509 L2119.33 202.032 L2124.26 202.032 L2124.26 205.921 L2119.33 205.921 L2119.33 214.069 L2114.67 214.069 L2114.67 205.921 L2099.07 205.921 L2099.07 201.407 L2113.45 179.509 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M2146.16 195.434 Q2149.51 196.152 2151.39 198.421 Q2153.28 200.689 2153.28 204.022 Q2153.28 209.138 2149.77 211.939 Q2146.25 214.74 2139.77 214.74 Q2137.59 214.74 2135.28 214.3 Q2132.98 213.883 2130.53 213.027 L2130.53 208.513 Q2132.47 209.647 2134.79 210.226 Q2137.1 210.805 2139.63 210.805 Q2144.03 210.805 2146.32 209.069 Q2148.63 207.333 2148.63 204.022 Q2148.63 200.967 2146.48 199.254 Q2144.35 197.518 2140.53 197.518 L2136.5 197.518 L2136.5 193.675 L2140.72 193.675 Q2144.16 193.675 2145.99 192.309 Q2147.82 190.921 2147.82 188.328 Q2147.82 185.666 2145.92 184.254 Q2144.05 182.819 2140.53 182.819 Q2138.61 182.819 2136.41 183.235 Q2134.21 183.652 2131.57 184.532 L2131.57 180.365 Q2134.23 179.624 2136.55 179.254 Q2138.89 178.884 2140.95 178.884 Q2146.27 178.884 2149.37 181.314 Q2152.47 183.722 2152.47 187.842 Q2152.47 190.712 2150.83 192.703 Q2149.19 194.671 2146.16 195.434 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M2176.32 195.434 Q2179.67 196.152 2181.55 198.421 Q2183.45 200.689 2183.45 204.022 Q2183.45 209.138 2179.93 211.939 Q2176.41 214.74 2169.93 214.74 Q2167.75 214.74 2165.44 214.3 Q2163.15 213.883 2160.69 213.027 L2160.69 208.513 Q2162.64 209.647 2164.95 210.226 Q2167.27 210.805 2169.79 210.805 Q2174.19 210.805 2176.48 209.069 Q2178.79 207.333 2178.79 204.022 Q2178.79 200.967 2176.64 199.254 Q2174.51 197.518 2170.69 197.518 L2166.66 197.518 L2166.66 193.675 L2170.88 193.675 Q2174.33 193.675 2176.16 192.309 Q2177.98 190.921 2177.98 188.328 Q2177.98 185.666 2176.09 184.254 Q2174.21 182.819 2170.69 182.819 Q2168.77 182.819 2166.57 183.235 Q2164.37 183.652 2161.73 184.532 L2161.73 180.365 Q2164.4 179.624 2166.71 179.254 Q2169.05 178.884 2171.11 178.884 Q2176.43 178.884 2179.53 181.314 Q2182.64 183.722 2182.64 187.842 Q2182.64 190.712 2180.99 192.703 Q2179.35 194.671 2176.32 195.434 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M2191.04 178.097 L2194.74 178.097 Q2198.22 183.56 2199.93 188.791 Q2201.66 194.022 2201.66 199.184 Q2201.66 204.37 2199.93 209.624 Q2198.22 214.879 2194.74 220.319 L2191.04 220.319 Q2194.12 215.018 2195.62 209.786 Q2197.15 204.532 2197.15 199.184 Q2197.15 193.837 2195.62 188.629 Q2194.12 183.421 2191.04 178.097 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip030)" style="stroke:#3da44d; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1581.03,248.629 1724.95,248.629 
  "/>
<path clip-path="url(#clip030)" d="M1769.28 244.96 Q1770.88 242.089 1773.1 240.724 Q1775.32 239.358 1778.33 239.358 Q1782.38 239.358 1784.58 242.205 Q1786.78 245.029 1786.78 250.261 L1786.78 265.909 L1782.5 265.909 L1782.5 250.399 Q1782.5 246.673 1781.18 244.867 Q1779.86 243.062 1777.15 243.062 Q1773.84 243.062 1771.92 245.261 Q1770 247.46 1770 251.256 L1770 265.909 L1765.72 265.909 L1765.72 250.399 Q1765.72 246.649 1764.4 244.867 Q1763.08 243.062 1760.32 243.062 Q1757.06 243.062 1755.14 245.284 Q1753.22 247.483 1753.22 251.256 L1753.22 265.909 L1748.94 265.909 L1748.94 239.983 L1753.22 239.983 L1753.22 244.011 Q1754.68 241.626 1756.71 240.492 Q1758.75 239.358 1761.55 239.358 Q1764.37 239.358 1766.34 240.793 Q1768.33 242.228 1769.28 244.96 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1807.06 252.876 Q1801.9 252.876 1799.91 254.057 Q1797.92 255.237 1797.92 258.085 Q1797.92 260.353 1799.4 261.696 Q1800.9 263.015 1803.47 263.015 Q1807.01 263.015 1809.14 260.515 Q1811.3 257.992 1811.3 253.825 L1811.3 252.876 L1807.06 252.876 M1815.56 251.117 L1815.56 265.909 L1811.3 265.909 L1811.3 261.973 Q1809.84 264.335 1807.66 265.469 Q1805.49 266.58 1802.34 266.58 Q1798.36 266.58 1796 264.358 Q1793.66 262.112 1793.66 258.362 Q1793.66 253.987 1796.57 251.765 Q1799.51 249.543 1805.32 249.543 L1811.3 249.543 L1811.3 249.126 Q1811.3 246.187 1809.35 244.589 Q1807.43 242.969 1803.93 242.969 Q1801.71 242.969 1799.61 243.501 Q1797.5 244.034 1795.56 245.099 L1795.56 241.163 Q1797.89 240.261 1800.09 239.821 Q1802.29 239.358 1804.37 239.358 Q1810 239.358 1812.78 242.274 Q1815.56 245.191 1815.56 251.117 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1840.86 240.747 L1840.86 244.774 Q1839.05 243.849 1837.11 243.386 Q1835.16 242.923 1833.08 242.923 Q1829.91 242.923 1828.31 243.895 Q1826.74 244.867 1826.74 246.811 Q1826.74 248.293 1827.87 249.149 Q1829 249.983 1832.43 250.747 L1833.89 251.071 Q1838.43 252.043 1840.32 253.825 Q1842.24 255.585 1842.24 258.756 Q1842.24 262.367 1839.37 264.473 Q1836.53 266.58 1831.53 266.58 Q1829.44 266.58 1827.18 266.163 Q1824.93 265.77 1822.43 264.96 L1822.43 260.561 Q1824.79 261.788 1827.08 262.413 Q1829.37 263.015 1831.62 263.015 Q1834.63 263.015 1836.25 261.997 Q1837.87 260.955 1837.87 259.08 Q1837.87 257.344 1836.69 256.418 Q1835.53 255.492 1831.57 254.636 L1830.09 254.288 Q1826.13 253.455 1824.37 251.742 Q1822.62 250.006 1822.62 246.997 Q1822.62 243.339 1825.21 241.349 Q1827.8 239.358 1832.57 239.358 Q1834.93 239.358 1837.01 239.705 Q1839.1 240.052 1840.86 240.747 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1865.55 240.747 L1865.55 244.774 Q1863.75 243.849 1861.8 243.386 Q1859.86 242.923 1857.78 242.923 Q1854.61 242.923 1853.01 243.895 Q1851.43 244.867 1851.43 246.811 Q1851.43 248.293 1852.57 249.149 Q1853.7 249.983 1857.13 250.747 L1858.59 251.071 Q1863.12 252.043 1865.02 253.825 Q1866.94 255.585 1866.94 258.756 Q1866.94 262.367 1864.07 264.473 Q1861.23 266.58 1856.23 266.58 Q1854.14 266.58 1851.87 266.163 Q1849.63 265.77 1847.13 264.96 L1847.13 260.561 Q1849.49 261.788 1851.78 262.413 Q1854.07 263.015 1856.32 263.015 Q1859.33 263.015 1860.95 261.997 Q1862.57 260.955 1862.57 259.08 Q1862.57 257.344 1861.39 256.418 Q1860.23 255.492 1856.27 254.636 L1854.79 254.288 Q1850.83 253.455 1849.07 251.742 Q1847.31 250.006 1847.31 246.997 Q1847.31 243.339 1849.91 241.349 Q1852.5 239.358 1857.27 239.358 Q1859.63 239.358 1861.71 239.705 Q1863.8 240.052 1865.55 240.747 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1874.33 260.029 L1879.21 260.029 L1879.21 265.909 L1874.33 265.909 L1874.33 260.029 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1900.58 252.876 Q1895.42 252.876 1893.43 254.057 Q1891.43 255.237 1891.43 258.085 Q1891.43 260.353 1892.92 261.696 Q1894.42 263.015 1896.99 263.015 Q1900.53 263.015 1902.66 260.515 Q1904.81 257.992 1904.81 253.825 L1904.81 252.876 L1900.58 252.876 M1909.07 251.117 L1909.07 265.909 L1904.81 265.909 L1904.81 261.973 Q1903.36 264.335 1901.18 265.469 Q1899 266.58 1895.86 266.58 Q1891.87 266.58 1889.51 264.358 Q1887.18 262.112 1887.18 258.362 Q1887.18 253.987 1890.09 251.765 Q1893.03 249.543 1898.84 249.543 L1904.81 249.543 L1904.81 249.126 Q1904.81 246.187 1902.87 244.589 Q1900.95 242.969 1897.45 242.969 Q1895.23 242.969 1893.12 243.501 Q1891.02 244.034 1889.07 245.099 L1889.07 241.163 Q1891.41 240.261 1893.61 239.821 Q1895.81 239.358 1897.89 239.358 Q1903.52 239.358 1906.3 242.274 Q1909.07 245.191 1909.07 251.117 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1943.15 229.937 Q1940.05 235.261 1938.54 240.469 Q1937.04 245.677 1937.04 251.024 Q1937.04 256.372 1938.54 261.626 Q1940.07 266.858 1943.15 272.159 L1939.44 272.159 Q1935.97 266.719 1934.24 261.464 Q1932.52 256.21 1932.52 251.024 Q1932.52 245.862 1934.24 240.631 Q1935.95 235.4 1939.44 229.937 L1943.15 229.937 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1952.06 231.349 L1970.42 231.349 L1970.42 235.284 L1956.34 235.284 L1956.34 243.756 Q1957.36 243.409 1958.38 243.247 Q1959.4 243.062 1960.42 243.062 Q1966.2 243.062 1969.58 246.233 Q1972.96 249.404 1972.96 254.821 Q1972.96 260.399 1969.49 263.501 Q1966.02 266.58 1959.7 266.58 Q1957.52 266.58 1955.25 266.21 Q1953.01 265.839 1950.6 265.098 L1950.6 260.399 Q1952.68 261.534 1954.91 262.089 Q1957.13 262.645 1959.61 262.645 Q1963.61 262.645 1965.95 260.538 Q1968.29 258.432 1968.29 254.821 Q1968.29 251.21 1965.95 249.103 Q1963.61 246.997 1959.61 246.997 Q1957.73 246.997 1955.86 247.413 Q1954 247.83 1952.06 248.71 L1952.06 231.349 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M1992.17 249.497 Q1988.84 249.497 1986.92 251.279 Q1985.02 253.061 1985.02 256.186 Q1985.02 259.311 1986.92 261.094 Q1988.84 262.876 1992.17 262.876 Q1995.51 262.876 1997.43 261.094 Q1999.35 259.288 1999.35 256.186 Q1999.35 253.061 1997.43 251.279 Q1995.53 249.497 1992.17 249.497 M1987.5 247.506 Q1984.49 246.765 1982.8 244.705 Q1981.13 242.645 1981.13 239.682 Q1981.13 235.538 1984.07 233.131 Q1987.04 230.724 1992.17 230.724 Q1997.34 230.724 2000.28 233.131 Q2003.22 235.538 2003.22 239.682 Q2003.22 242.645 2001.53 244.705 Q1999.86 246.765 1996.87 247.506 Q2000.25 248.293 2002.13 250.585 Q2004.03 252.876 2004.03 256.186 Q2004.03 261.21 2000.95 263.895 Q1997.89 266.58 1992.17 266.58 Q1986.46 266.58 1983.38 263.895 Q1980.32 261.21 1980.32 256.186 Q1980.32 252.876 1982.22 250.585 Q1984.12 248.293 1987.5 247.506 M1985.79 240.122 Q1985.79 242.807 1987.45 244.312 Q1989.14 245.816 1992.17 245.816 Q1995.18 245.816 1996.87 244.312 Q1998.59 242.807 1998.59 240.122 Q1998.59 237.437 1996.87 235.932 Q1995.18 234.427 1992.17 234.427 Q1989.14 234.427 1987.45 235.932 Q1985.79 237.437 1985.79 240.122 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M2011.16 231.349 L2033.38 231.349 L2033.38 233.339 L2020.83 265.909 L2015.95 265.909 L2027.75 235.284 L2011.16 235.284 L2011.16 231.349 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M2046.53 261.973 L2062.85 261.973 L2062.85 265.909 L2040.9 265.909 L2040.9 261.973 Q2043.56 259.219 2048.15 254.589 Q2052.75 249.936 2053.93 248.594 Q2056.18 246.071 2057.06 244.335 Q2057.96 242.575 2057.96 240.886 Q2057.96 238.131 2056.02 236.395 Q2054.1 234.659 2050.99 234.659 Q2048.79 234.659 2046.34 235.423 Q2043.91 236.187 2041.13 237.737 L2041.13 233.015 Q2043.96 231.881 2046.41 231.302 Q2048.86 230.724 2050.9 230.724 Q2056.27 230.724 2059.47 233.409 Q2062.66 236.094 2062.66 240.585 Q2062.66 242.714 2061.85 244.636 Q2061.06 246.534 2058.96 249.126 Q2058.38 249.798 2055.28 253.015 Q2052.17 256.21 2046.53 261.973 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M2082.66 234.427 Q2079.05 234.427 2077.22 237.992 Q2075.41 241.534 2075.41 248.663 Q2075.41 255.77 2077.22 259.335 Q2079.05 262.876 2082.66 262.876 Q2086.29 262.876 2088.1 259.335 Q2089.93 255.77 2089.93 248.663 Q2089.93 241.534 2088.1 237.992 Q2086.29 234.427 2082.66 234.427 M2082.66 230.724 Q2088.47 230.724 2091.53 235.33 Q2094.6 239.913 2094.6 248.663 Q2094.6 257.39 2091.53 261.997 Q2088.47 266.58 2082.66 266.58 Q2076.85 266.58 2073.77 261.997 Q2070.72 257.39 2070.72 248.663 Q2070.72 239.913 2073.77 235.33 Q2076.85 230.724 2082.66 230.724 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M2106.85 261.973 L2123.17 261.973 L2123.17 265.909 L2101.22 265.909 L2101.22 261.973 Q2103.89 259.219 2108.47 254.589 Q2113.08 249.936 2114.26 248.594 Q2116.5 246.071 2117.38 244.335 Q2118.29 242.575 2118.29 240.886 Q2118.29 238.131 2116.34 236.395 Q2114.42 234.659 2111.32 234.659 Q2109.12 234.659 2106.66 235.423 Q2104.23 236.187 2101.46 237.737 L2101.46 233.015 Q2104.28 231.881 2106.73 231.302 Q2109.19 230.724 2111.22 230.724 Q2116.6 230.724 2119.79 233.409 Q2122.98 236.094 2122.98 240.585 Q2122.98 242.714 2122.17 244.636 Q2121.39 246.534 2119.28 249.126 Q2118.7 249.798 2115.6 253.015 Q2112.5 256.21 2106.85 261.973 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M2133.03 231.349 L2151.39 231.349 L2151.39 235.284 L2137.31 235.284 L2137.31 243.756 Q2138.33 243.409 2139.35 243.247 Q2140.37 243.062 2141.39 243.062 Q2147.17 243.062 2150.55 246.233 Q2153.93 249.404 2153.93 254.821 Q2153.93 260.399 2150.46 263.501 Q2146.99 266.58 2140.67 266.58 Q2138.49 266.58 2136.22 266.21 Q2133.98 265.839 2131.57 265.098 L2131.57 260.399 Q2133.66 261.534 2135.88 262.089 Q2138.1 262.645 2140.58 262.645 Q2144.58 262.645 2146.92 260.538 Q2149.26 258.432 2149.26 254.821 Q2149.26 251.21 2146.92 249.103 Q2144.58 246.997 2140.58 246.997 Q2138.7 246.997 2136.83 247.413 Q2134.97 247.83 2133.03 248.71 L2133.03 231.349 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M2173.72 246.765 Q2170.58 246.765 2168.72 248.918 Q2166.9 251.071 2166.9 254.821 Q2166.9 258.548 2168.72 260.723 Q2170.58 262.876 2173.72 262.876 Q2176.87 262.876 2178.7 260.723 Q2180.55 258.548 2180.55 254.821 Q2180.55 251.071 2178.7 248.918 Q2176.87 246.765 2173.72 246.765 M2183.01 232.113 L2183.01 236.372 Q2181.25 235.538 2179.44 235.099 Q2177.66 234.659 2175.9 234.659 Q2171.27 234.659 2168.82 237.784 Q2166.39 240.909 2166.04 247.228 Q2167.41 245.214 2169.47 244.149 Q2171.53 243.062 2174 243.062 Q2179.21 243.062 2182.22 246.233 Q2185.25 249.381 2185.25 254.821 Q2185.25 260.145 2182.1 263.362 Q2178.96 266.58 2173.72 266.58 Q2167.73 266.58 2164.56 261.997 Q2161.39 257.39 2161.39 248.663 Q2161.39 240.469 2165.28 235.608 Q2169.16 230.724 2175.72 230.724 Q2177.47 230.724 2179.26 231.071 Q2181.06 231.418 2183.01 232.113 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M2194.12 261.973 L2201.76 261.973 L2201.76 235.608 L2193.45 237.275 L2193.45 233.015 L2201.71 231.349 L2206.39 231.349 L2206.39 261.973 L2214.03 261.973 L2214.03 265.909 L2194.12 265.909 L2194.12 261.973 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip030)" d="M2222.2 229.937 L2225.9 229.937 Q2229.37 235.4 2231.09 240.631 Q2232.82 245.862 2232.82 251.024 Q2232.82 256.21 2231.09 261.464 Q2229.37 266.719 2225.9 272.159 L2222.2 272.159 Q2225.28 266.858 2226.78 261.626 Q2228.31 256.372 2228.31 251.024 Q2228.31 245.677 2226.78 240.469 Q2225.28 235.261 2222.2 229.937 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /></svg>



The data from the simualtion of the *simpleFMU*, are divided into position and velocity data. These data will be needed later to plot the results. 


```julia
velSimple = collect(data[2] for data in simpleSimData.saveval)
posSimple = collect(data[1] for data in simpleSimData.saveval)
```


    501-element Vector{Float64}:
     0.5
     0.5003127019074967
     0.5012175433745238
     0.5027172504687035
     0.504812416566759
     0.5075012719497328
     0.5107830165354977
     0.5146534880772458
     0.5191107030735219
     0.5241484264969329
     ⋮
     1.69516502108285
     1.6976144547672483
     1.6994659284032172
     1.7007174453690572
     1.7013675684067706
     1.7014154196220592
     1.7008606804843265
     1.69970552855305
     1.6979508813706


## NeuralFMU

#### Loss function

In order to train our model, a loss function must be implemented. The solver of the NeuralFMU can calculate the gradient of the loss function. The gradient descent is needed to adjust the weights in the neural network so that the sum of the error is reduced and the model becomes more accurate.

The loss function in this implmentation consists of the mean squared error (mse) from the real position of the *realFMU* simulation (posReal) and the position data of the network (posNet).
$$ mse = \frac{1}{n} \sum\limits_{i=0}^n (posReal[i] - posNet[i])^2 $$

As it is indicated with the comments, one could also additionally consider the mse from the real velocity (velReal) and the velocity from the network (velNet). The error in this case would be calculated from the sum of both errors.


```julia
# loss function for training
function lossSum()
    solution = neuralFMU(x₀, tStart)

    posNet = collect(data[1] for data in solution.u)
    #velNet = collect(data[2] for data in solution.u)

    Flux.Losses.mse(posReal, posNet) #+ Flux.Losses.mse(velReal, velNet)
end
```


    lossSum (generic function with 1 method)


#### Callback

To output the loss in certain time intervals, a callback is implemented as a function in the following. Here a counter is incremented, every twentieth pass the loss function is called and the average error is printed out.


```julia
# callback function for training
global counter = 0
function callb()
    global counter += 1

    if counter % 20 == 1
        avgLoss = lossSum()
        @info "Loss [$counter]: $(round(avgLoss, digits=5))   Avg displacement in data: $(round(sqrt(avgLoss), digits=5))"
    end
end
```


    callb (generic function with 1 method)


#### Structure of the NeuralFMU

In the following, the topology of the NeuralFMU is constructed. It consists of an input layer, which then leads into the *simpleFMU* model. The ME-FMU computes the state derivatives for a given system state. Following the *simpleFMU* is a dense layer that has exactly as many inputs as the model has states (and therefore state derivatives). The output of this layer consists of 16 output nodes and a *tanh* activation function. The next layer has 16 input and output nodes with the same activation function. The last layer is again a dense layer with 16 input nodes and the number of states as outputs. Here, it is important that no *tanh*-activation function follows, because otherwise the pendulums state values would be limited to the interval $[-1;1]$.


```julia
# NeuralFMU setup
numStates = fmiGetNumberOfStates(simpleFMU)

net = Chain(inputs -> fmiEvaluateME(simpleFMU, inputs),
            Dense(numStates, 16, tanh),
            Dense(16, 16, tanh),
            Dense(16, numStates))
```


    Chain(
      var"#13#14"(),
      Dense(2, 16, tanh),                   [90m# 48 parameters[39m
      Dense(16, 16, tanh),                  [90m# 272 parameters[39m
      Dense(16, 2),                         [90m# 34 parameters[39m
    )[90m                   # Total: 6 arrays, [39m354 parameters, 1.758 KiB.


#### Definition of the NeuralFMU

The instantiation of the ME-NeuralFMU is done as a one-liner. The FMU (*simpleFMU*), the structure of the network `net`, start `tStart` and end time `tStop`, the numerical solver `Tsit5()` and the time steps `tSave` for saving are specified.


```julia
neuralFMU = ME_NeuralFMU(simpleFMU, net, (tStart, tStop), Tsit5(); saveat=tSave);
```

#### Plot before training

Here the state trajactory of the *simpleFMU* is recorded. Doesn't really look like a pendulum yet, but the system is random initialized by default. In the later plots, the effect of learning can be seen.


```julia
solutionBefore = neuralFMU(x₀, tStart)
fmiPlot(simpleFMU, solutionBefore)
```


<?xml version="1.0" encoding="utf-8"?>
<svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" width="600" height="400" viewBox="0 0 2400 1600">
<defs>
  <clipPath id="clip120">
    <rect x="0" y="0" width="2400" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip120)" d="
M0 1600 L2400 1600 L2400 0 L0 0  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip121">
    <rect x="480" y="0" width="1681" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip120)" d="
M186.274 1423.18 L2352.76 1423.18 L2352.76 47.2441 L186.274 47.2441  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip122">
    <rect x="186" y="47" width="2167" height="1377"/>
  </clipPath>
</defs>
<polyline clip-path="url(#clip122)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  247.59,1423.18 247.59,47.2441 
  "/>
<polyline clip-path="url(#clip122)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  656.36,1423.18 656.36,47.2441 
  "/>
<polyline clip-path="url(#clip122)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1065.13,1423.18 1065.13,47.2441 
  "/>
<polyline clip-path="url(#clip122)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1473.9,1423.18 1473.9,47.2441 
  "/>
<polyline clip-path="url(#clip122)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1882.67,1423.18 1882.67,47.2441 
  "/>
<polyline clip-path="url(#clip122)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  2291.44,1423.18 2291.44,47.2441 
  "/>
<polyline clip-path="url(#clip120)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  186.274,1423.18 2352.76,1423.18 
  "/>
<polyline clip-path="url(#clip120)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  247.59,1423.18 247.59,1404.28 
  "/>
<polyline clip-path="url(#clip120)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  656.36,1423.18 656.36,1404.28 
  "/>
<polyline clip-path="url(#clip120)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1065.13,1423.18 1065.13,1404.28 
  "/>
<polyline clip-path="url(#clip120)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1473.9,1423.18 1473.9,1404.28 
  "/>
<polyline clip-path="url(#clip120)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1882.67,1423.18 1882.67,1404.28 
  "/>
<polyline clip-path="url(#clip120)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  2291.44,1423.18 2291.44,1404.28 
  "/>
<path clip-path="url(#clip120)" d="M247.59 1454.1 Q243.978 1454.1 242.15 1457.66 Q240.344 1461.2 240.344 1468.33 Q240.344 1475.44 242.15 1479.01 Q243.978 1482.55 247.59 1482.55 Q251.224 1482.55 253.029 1479.01 Q254.858 1475.44 254.858 1468.33 Q254.858 1461.2 253.029 1457.66 Q251.224 1454.1 247.59 1454.1 M247.59 1450.39 Q253.4 1450.39 256.455 1455 Q259.534 1459.58 259.534 1468.33 Q259.534 1477.06 256.455 1481.67 Q253.4 1486.25 247.59 1486.25 Q241.779 1486.25 238.701 1481.67 Q235.645 1477.06 235.645 1468.33 Q235.645 1459.58 238.701 1455 Q241.779 1450.39 247.59 1450.39 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M646.742 1481.64 L654.381 1481.64 L654.381 1455.28 L646.07 1456.95 L646.07 1452.69 L654.334 1451.02 L659.01 1451.02 L659.01 1481.64 L666.649 1481.64 L666.649 1485.58 L646.742 1485.58 L646.742 1481.64 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M1059.78 1481.64 L1076.1 1481.64 L1076.1 1485.58 L1054.16 1485.58 L1054.16 1481.64 Q1056.82 1478.89 1061.4 1474.26 Q1066.01 1469.61 1067.19 1468.27 Q1069.44 1465.74 1070.32 1464.01 Q1071.22 1462.25 1071.22 1460.56 Q1071.22 1457.8 1069.27 1456.07 Q1067.35 1454.33 1064.25 1454.33 Q1062.05 1454.33 1059.6 1455.09 Q1057.17 1455.86 1054.39 1457.41 L1054.39 1452.69 Q1057.21 1451.55 1059.67 1450.97 Q1062.12 1450.39 1064.16 1450.39 Q1069.53 1450.39 1072.72 1453.08 Q1075.92 1455.77 1075.92 1460.26 Q1075.92 1462.39 1075.11 1464.31 Q1074.32 1466.2 1072.21 1468.8 Q1071.63 1469.47 1068.53 1472.69 Q1065.43 1475.88 1059.78 1481.64 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M1478.15 1466.95 Q1481.5 1467.66 1483.38 1469.93 Q1485.28 1472.2 1485.28 1475.53 Q1485.28 1480.65 1481.76 1483.45 Q1478.24 1486.25 1471.76 1486.25 Q1469.58 1486.25 1467.27 1485.81 Q1464.98 1485.39 1462.52 1484.54 L1462.52 1480.02 Q1464.47 1481.16 1466.78 1481.74 Q1469.1 1482.32 1471.62 1482.32 Q1476.02 1482.32 1478.31 1480.58 Q1480.62 1478.84 1480.62 1475.53 Q1480.62 1472.48 1478.47 1470.77 Q1476.34 1469.03 1472.52 1469.03 L1468.49 1469.03 L1468.49 1465.19 L1472.71 1465.19 Q1476.16 1465.19 1477.99 1463.82 Q1479.81 1462.43 1479.81 1459.84 Q1479.81 1457.18 1477.92 1455.77 Q1476.04 1454.33 1472.52 1454.33 Q1470.6 1454.33 1468.4 1454.75 Q1466.2 1455.16 1463.56 1456.04 L1463.56 1451.88 Q1466.23 1451.14 1468.54 1450.77 Q1470.88 1450.39 1472.94 1450.39 Q1478.26 1450.39 1481.37 1452.83 Q1484.47 1455.23 1484.47 1459.35 Q1484.47 1462.22 1482.82 1464.21 Q1481.18 1466.18 1478.15 1466.95 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M1885.68 1455.09 L1873.87 1473.54 L1885.68 1473.54 L1885.68 1455.09 M1884.45 1451.02 L1890.33 1451.02 L1890.33 1473.54 L1895.26 1473.54 L1895.26 1477.43 L1890.33 1477.43 L1890.33 1485.58 L1885.68 1485.58 L1885.68 1477.43 L1870.08 1477.43 L1870.08 1472.92 L1884.45 1451.02 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M2281.72 1451.02 L2300.07 1451.02 L2300.07 1454.96 L2286 1454.96 L2286 1463.43 Q2287.02 1463.08 2288.04 1462.92 Q2289.06 1462.73 2290.07 1462.73 Q2295.86 1462.73 2299.24 1465.9 Q2302.62 1469.08 2302.62 1474.49 Q2302.62 1480.07 2299.15 1483.17 Q2295.68 1486.25 2289.36 1486.25 Q2287.18 1486.25 2284.91 1485.88 Q2282.67 1485.51 2280.26 1484.77 L2280.26 1480.07 Q2282.34 1481.2 2284.57 1481.76 Q2286.79 1482.32 2289.26 1482.32 Q2293.27 1482.32 2295.61 1480.21 Q2297.94 1478.1 2297.94 1474.49 Q2297.94 1470.88 2295.61 1468.77 Q2293.27 1466.67 2289.26 1466.67 Q2287.39 1466.67 2285.51 1467.08 Q2283.66 1467.5 2281.72 1468.38 L2281.72 1451.02 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M1217.83 1522.27 L1217.83 1532.4 L1229.89 1532.4 L1229.89 1536.95 L1217.83 1536.95 L1217.83 1556.3 Q1217.83 1560.66 1219 1561.9 Q1220.21 1563.14 1223.87 1563.14 L1229.89 1563.14 L1229.89 1568.04 L1223.87 1568.04 Q1217.09 1568.04 1214.52 1565.53 Q1211.94 1562.98 1211.94 1556.3 L1211.94 1536.95 L1207.64 1536.95 L1207.64 1532.4 L1211.94 1532.4 L1211.94 1522.27 L1217.83 1522.27 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M1257.77 1518.52 L1271.27 1518.52 L1271.27 1523.07 L1263.63 1523.07 L1263.63 1572.09 L1271.27 1572.09 L1271.27 1576.64 L1257.77 1576.64 L1257.77 1518.52 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M1306.47 1533.45 L1306.47 1538.98 Q1303.99 1537.71 1301.31 1537.07 Q1298.64 1536.44 1295.77 1536.44 Q1291.41 1536.44 1289.22 1537.77 Q1287.05 1539.11 1287.05 1541.79 Q1287.05 1543.82 1288.61 1545 Q1290.17 1546.15 1294.88 1547.2 L1296.89 1547.64 Q1303.13 1548.98 1305.74 1551.43 Q1308.38 1553.85 1308.38 1558.21 Q1308.38 1563.17 1304.43 1566.07 Q1300.52 1568.97 1293.64 1568.97 Q1290.78 1568.97 1287.66 1568.39 Q1284.57 1567.85 1281.13 1566.74 L1281.13 1560.69 Q1284.38 1562.38 1287.53 1563.24 Q1290.68 1564.07 1293.77 1564.07 Q1297.91 1564.07 1300.13 1562.66 Q1302.36 1561.23 1302.36 1558.65 Q1302.36 1556.27 1300.74 1554.99 Q1299.15 1553.72 1293.7 1552.54 L1291.67 1552.07 Q1286.22 1550.92 1283.81 1548.56 Q1281.39 1546.18 1281.39 1542.04 Q1281.39 1537.01 1284.95 1534.27 Q1288.52 1531.54 1295.07 1531.54 Q1298.32 1531.54 1301.18 1532.01 Q1304.05 1532.49 1306.47 1533.45 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M1331.39 1518.52 L1331.39 1576.64 L1317.89 1576.64 L1317.89 1572.09 L1325.5 1572.09 L1325.5 1523.07 L1317.89 1523.07 L1317.89 1518.52 L1331.39 1518.52 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip122)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  186.274,1384.24 2352.76,1384.24 
  "/>
<polyline clip-path="url(#clip122)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  186.274,1095.64 2352.76,1095.64 
  "/>
<polyline clip-path="url(#clip122)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  186.274,807.038 2352.76,807.038 
  "/>
<polyline clip-path="url(#clip122)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  186.274,518.438 2352.76,518.438 
  "/>
<polyline clip-path="url(#clip122)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  186.274,229.838 2352.76,229.838 
  "/>
<polyline clip-path="url(#clip120)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  186.274,1423.18 186.274,47.2441 
  "/>
<polyline clip-path="url(#clip120)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  186.274,1384.24 205.172,1384.24 
  "/>
<polyline clip-path="url(#clip120)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  186.274,1095.64 205.172,1095.64 
  "/>
<polyline clip-path="url(#clip120)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  186.274,807.038 205.172,807.038 
  "/>
<polyline clip-path="url(#clip120)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  186.274,518.438 205.172,518.438 
  "/>
<polyline clip-path="url(#clip120)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  186.274,229.838 205.172,229.838 
  "/>
<path clip-path="url(#clip120)" d="M62.9365 1370.04 Q59.3254 1370.04 57.4967 1373.6 Q55.6912 1377.14 55.6912 1384.27 Q55.6912 1391.38 57.4967 1394.94 Q59.3254 1398.49 62.9365 1398.49 Q66.5707 1398.49 68.3763 1394.94 Q70.205 1391.38 70.205 1384.27 Q70.205 1377.14 68.3763 1373.6 Q66.5707 1370.04 62.9365 1370.04 M62.9365 1366.33 Q68.7467 1366.33 71.8022 1370.94 Q74.8809 1375.52 74.8809 1384.27 Q74.8809 1393 71.8022 1397.61 Q68.7467 1402.19 62.9365 1402.19 Q57.1264 1402.19 54.0477 1397.61 Q50.9921 1393 50.9921 1384.27 Q50.9921 1375.52 54.0477 1370.94 Q57.1264 1366.33 62.9365 1366.33 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M83.0984 1395.64 L87.9827 1395.64 L87.9827 1401.52 L83.0984 1401.52 L83.0984 1395.64 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M108.168 1370.04 Q104.557 1370.04 102.728 1373.6 Q100.922 1377.14 100.922 1384.27 Q100.922 1391.38 102.728 1394.94 Q104.557 1398.49 108.168 1398.49 Q111.802 1398.49 113.608 1394.94 Q115.436 1391.38 115.436 1384.27 Q115.436 1377.14 113.608 1373.6 Q111.802 1370.04 108.168 1370.04 M108.168 1366.33 Q113.978 1366.33 117.033 1370.94 Q120.112 1375.52 120.112 1384.27 Q120.112 1393 117.033 1397.61 Q113.978 1402.19 108.168 1402.19 Q102.358 1402.19 99.2789 1397.61 Q96.2234 1393 96.2234 1384.27 Q96.2234 1375.52 99.2789 1370.94 Q102.358 1366.33 108.168 1366.33 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M138.33 1370.04 Q134.719 1370.04 132.89 1373.6 Q131.084 1377.14 131.084 1384.27 Q131.084 1391.38 132.89 1394.94 Q134.719 1398.49 138.33 1398.49 Q141.964 1398.49 143.769 1394.94 Q145.598 1391.38 145.598 1384.27 Q145.598 1377.14 143.769 1373.6 Q141.964 1370.04 138.33 1370.04 M138.33 1366.33 Q144.14 1366.33 147.195 1370.94 Q150.274 1375.52 150.274 1384.27 Q150.274 1393 147.195 1397.61 Q144.14 1402.19 138.33 1402.19 Q132.519 1402.19 129.441 1397.61 Q126.385 1393 126.385 1384.27 Q126.385 1375.52 129.441 1370.94 Q132.519 1366.33 138.33 1366.33 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M63.9319 1081.44 Q60.3208 1081.44 58.4921 1085 Q56.6865 1088.54 56.6865 1095.67 Q56.6865 1102.78 58.4921 1106.34 Q60.3208 1109.89 63.9319 1109.89 Q67.5661 1109.89 69.3717 1106.34 Q71.2004 1102.78 71.2004 1095.67 Q71.2004 1088.54 69.3717 1085 Q67.5661 1081.44 63.9319 1081.44 M63.9319 1077.73 Q69.742 1077.73 72.7976 1082.34 Q75.8763 1086.92 75.8763 1095.67 Q75.8763 1104.4 72.7976 1109.01 Q69.742 1113.59 63.9319 1113.59 Q58.1217 1113.59 55.043 1109.01 Q51.9875 1104.4 51.9875 1095.67 Q51.9875 1086.92 55.043 1082.34 Q58.1217 1077.73 63.9319 1077.73 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M84.0938 1107.04 L88.978 1107.04 L88.978 1112.92 L84.0938 1112.92 L84.0938 1107.04 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M103.191 1108.98 L119.51 1108.98 L119.51 1112.92 L97.566 1112.92 L97.566 1108.98 Q100.228 1106.23 104.811 1101.6 Q109.418 1096.95 110.598 1095.6 Q112.844 1093.08 113.723 1091.34 Q114.626 1089.59 114.626 1087.9 Q114.626 1085.14 112.682 1083.4 Q110.76 1081.67 107.658 1081.67 Q105.459 1081.67 103.006 1082.43 Q100.575 1083.2 97.7974 1084.75 L97.7974 1080.02 Q100.621 1078.89 103.075 1078.31 Q105.529 1077.73 107.566 1077.73 Q112.936 1077.73 116.131 1080.42 Q119.325 1083.1 119.325 1087.59 Q119.325 1089.72 118.515 1091.65 Q117.728 1093.54 115.621 1096.14 Q115.043 1096.81 111.941 1100.02 Q108.839 1103.22 103.191 1108.98 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M129.371 1078.36 L147.728 1078.36 L147.728 1082.29 L133.654 1082.29 L133.654 1090.77 Q134.672 1090.42 135.691 1090.26 Q136.709 1090.07 137.728 1090.07 Q143.515 1090.07 146.894 1093.24 Q150.274 1096.41 150.274 1101.83 Q150.274 1107.41 146.802 1110.51 Q143.33 1113.59 137.01 1113.59 Q134.834 1113.59 132.566 1113.22 Q130.32 1112.85 127.913 1112.11 L127.913 1107.41 Q129.996 1108.54 132.219 1109.1 Q134.441 1109.65 136.918 1109.65 Q140.922 1109.65 143.26 1107.55 Q145.598 1105.44 145.598 1101.83 Q145.598 1098.22 143.26 1096.11 Q140.922 1094.01 136.918 1094.01 Q135.043 1094.01 133.168 1094.42 Q131.316 1094.84 129.371 1095.72 L129.371 1078.36 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M62.9365 792.837 Q59.3254 792.837 57.4967 796.402 Q55.6912 799.943 55.6912 807.073 Q55.6912 814.179 57.4967 817.744 Q59.3254 821.286 62.9365 821.286 Q66.5707 821.286 68.3763 817.744 Q70.205 814.179 70.205 807.073 Q70.205 799.943 68.3763 796.402 Q66.5707 792.837 62.9365 792.837 M62.9365 789.133 Q68.7467 789.133 71.8022 793.74 Q74.8809 798.323 74.8809 807.073 Q74.8809 815.8 71.8022 820.406 Q68.7467 824.99 62.9365 824.99 Q57.1264 824.99 54.0477 820.406 Q50.9921 815.8 50.9921 807.073 Q50.9921 798.323 54.0477 793.74 Q57.1264 789.133 62.9365 789.133 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M83.0984 818.439 L87.9827 818.439 L87.9827 824.318 L83.0984 824.318 L83.0984 818.439 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M98.2141 789.758 L116.57 789.758 L116.57 793.693 L102.496 793.693 L102.496 802.166 Q103.515 801.818 104.534 801.656 Q105.552 801.471 106.571 801.471 Q112.358 801.471 115.737 804.642 Q119.117 807.814 119.117 813.23 Q119.117 818.809 115.645 821.911 Q112.172 824.99 105.853 824.99 Q103.677 824.99 101.409 824.619 Q99.1632 824.249 96.7558 823.508 L96.7558 818.809 Q98.8391 819.943 101.061 820.499 Q103.284 821.054 105.76 821.054 Q109.765 821.054 112.103 818.948 Q114.441 816.841 114.441 813.23 Q114.441 809.619 112.103 807.513 Q109.765 805.406 105.76 805.406 Q103.885 805.406 102.01 805.823 Q100.159 806.24 98.2141 807.119 L98.2141 789.758 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M138.33 792.837 Q134.719 792.837 132.89 796.402 Q131.084 799.943 131.084 807.073 Q131.084 814.179 132.89 817.744 Q134.719 821.286 138.33 821.286 Q141.964 821.286 143.769 817.744 Q145.598 814.179 145.598 807.073 Q145.598 799.943 143.769 796.402 Q141.964 792.837 138.33 792.837 M138.33 789.133 Q144.14 789.133 147.195 793.74 Q150.274 798.323 150.274 807.073 Q150.274 815.8 147.195 820.406 Q144.14 824.99 138.33 824.99 Q132.519 824.99 129.441 820.406 Q126.385 815.8 126.385 807.073 Q126.385 798.323 129.441 793.74 Q132.519 789.133 138.33 789.133 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M63.9319 504.237 Q60.3208 504.237 58.4921 507.802 Q56.6865 511.343 56.6865 518.473 Q56.6865 525.579 58.4921 529.144 Q60.3208 532.686 63.9319 532.686 Q67.5661 532.686 69.3717 529.144 Q71.2004 525.579 71.2004 518.473 Q71.2004 511.343 69.3717 507.802 Q67.5661 504.237 63.9319 504.237 M63.9319 500.533 Q69.742 500.533 72.7976 505.14 Q75.8763 509.723 75.8763 518.473 Q75.8763 527.2 72.7976 531.806 Q69.742 536.389 63.9319 536.389 Q58.1217 536.389 55.043 531.806 Q51.9875 527.2 51.9875 518.473 Q51.9875 509.723 55.043 505.14 Q58.1217 500.533 63.9319 500.533 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M84.0938 529.839 L88.978 529.839 L88.978 535.718 L84.0938 535.718 L84.0938 529.839 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M97.9826 501.158 L120.205 501.158 L120.205 503.149 L107.658 535.718 L102.774 535.718 L114.58 505.093 L97.9826 505.093 L97.9826 501.158 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M129.371 501.158 L147.728 501.158 L147.728 505.093 L133.654 505.093 L133.654 513.566 Q134.672 513.218 135.691 513.056 Q136.709 512.871 137.728 512.871 Q143.515 512.871 146.894 516.042 Q150.274 519.214 150.274 524.63 Q150.274 530.209 146.802 533.311 Q143.33 536.389 137.01 536.389 Q134.834 536.389 132.566 536.019 Q130.32 535.649 127.913 534.908 L127.913 530.209 Q129.996 531.343 132.219 531.899 Q134.441 532.454 136.918 532.454 Q140.922 532.454 143.26 530.348 Q145.598 528.241 145.598 524.63 Q145.598 521.019 143.26 518.913 Q140.922 516.806 136.918 516.806 Q135.043 516.806 133.168 517.223 Q131.316 517.64 129.371 518.519 L129.371 501.158 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M53.7467 243.183 L61.3856 243.183 L61.3856 216.817 L53.0754 218.484 L53.0754 214.225 L61.3393 212.558 L66.0152 212.558 L66.0152 243.183 L73.654 243.183 L73.654 247.118 L53.7467 247.118 L53.7467 243.183 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M83.0984 241.239 L87.9827 241.239 L87.9827 247.118 L83.0984 247.118 L83.0984 241.239 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M108.168 215.637 Q104.557 215.637 102.728 219.202 Q100.922 222.743 100.922 229.873 Q100.922 236.979 102.728 240.544 Q104.557 244.086 108.168 244.086 Q111.802 244.086 113.608 240.544 Q115.436 236.979 115.436 229.873 Q115.436 222.743 113.608 219.202 Q111.802 215.637 108.168 215.637 M108.168 211.933 Q113.978 211.933 117.033 216.54 Q120.112 221.123 120.112 229.873 Q120.112 238.6 117.033 243.206 Q113.978 247.789 108.168 247.789 Q102.358 247.789 99.2789 243.206 Q96.2234 238.6 96.2234 229.873 Q96.2234 221.123 99.2789 216.54 Q102.358 211.933 108.168 211.933 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M138.33 215.637 Q134.719 215.637 132.89 219.202 Q131.084 222.743 131.084 229.873 Q131.084 236.979 132.89 240.544 Q134.719 244.086 138.33 244.086 Q141.964 244.086 143.769 240.544 Q145.598 236.979 145.598 229.873 Q145.598 222.743 143.769 219.202 Q141.964 215.637 138.33 215.637 M138.33 211.933 Q144.14 211.933 147.195 216.54 Q150.274 221.123 150.274 229.873 Q150.274 238.6 147.195 243.206 Q144.14 247.789 138.33 247.789 Q132.519 247.789 129.441 243.206 Q126.385 238.6 126.385 229.873 Q126.385 221.123 129.441 216.54 Q132.519 211.933 138.33 211.933 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip122)" style="stroke:#009af9; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  247.59,807.038 251.677,796.396 255.765,785.758 259.853,775.127 263.94,764.501 268.028,753.881 272.116,743.266 276.203,732.657 280.291,722.053 284.379,711.456 
  288.467,700.864 292.554,690.278 296.642,679.698 300.73,669.124 304.817,658.557 308.905,647.996 312.993,637.442 317.08,626.895 321.168,616.355 325.256,605.824 
  329.344,595.301 333.431,584.787 337.519,574.283 341.607,563.79 345.694,553.308 349.782,542.839 353.87,532.384 357.957,521.944 362.045,511.52 366.133,501.113 
  370.221,490.727 374.308,480.362 378.396,470.023 382.484,459.711 386.571,449.429 390.659,439.183 394.747,428.974 398.834,418.809 402.922,408.69 407.01,398.623 
  411.098,388.613 415.185,378.665 419.273,368.785 423.361,358.983 427.448,349.268 431.536,339.648 435.624,330.132 439.712,320.731 443.799,311.452 447.887,302.306 
  451.975,293.302 456.062,284.45 460.15,275.76 464.238,267.24 468.325,258.902 472.413,250.755 476.501,242.808 480.589,235.074 484.676,227.56 488.764,220.278 
  492.852,213.236 496.939,206.434 501.027,199.876 505.115,193.563 509.202,187.496 513.29,181.676 517.378,176.102 521.466,170.772 525.553,165.685 529.641,160.836 
  533.729,156.222 537.816,151.838 541.904,147.678 545.992,143.736 550.079,140.004 554.167,136.475 558.255,133.139 562.343,129.987 566.43,127.009 570.518,124.199 
  574.606,121.55 578.693,119.056 582.781,116.708 586.869,114.501 590.956,112.427 595.044,110.481 599.132,108.655 603.22,106.944 607.307,105.343 611.395,103.844 
  615.483,102.443 619.57,101.134 623.658,99.9114 627.746,98.771 631.833,97.7073 635.921,96.7157 640.009,95.7916 644.097,94.9305 648.184,94.1282 652.272,93.3814 
  656.36,92.6873 660.447,92.0435 664.535,91.4475 668.623,90.8968 672.711,90.3891 676.798,89.9222 680.886,89.4937 684.974,89.1015 689.061,88.7434 693.149,88.4175 
  697.237,88.1216 701.324,87.8539 705.412,87.6125 709.5,87.3955 713.588,87.2011 717.675,87.0277 721.763,86.8736 725.851,86.7372 729.938,86.6176 734.026,86.5141 
  738.114,86.426 742.201,86.3527 746.289,86.2936 750.377,86.248 754.465,86.2152 758.552,86.1947 762.64,86.1857 766.728,86.1876 770.815,86.1999 774.903,86.2217 
  778.991,86.2527 783.078,86.292 787.166,86.339 791.254,86.3932 795.342,86.4539 799.429,86.5205 803.517,86.5923 807.605,86.6687 811.692,86.7492 815.78,86.833 
  819.868,86.9198 823.955,87.01 828.043,87.1035 832.131,87.2002 836.219,87.2999 840.306,87.4026 844.394,87.508 848.482,87.6161 852.569,87.7266 856.657,87.8396 
  860.745,87.9547 864.832,88.0719 868.92,88.191 873.008,88.3118 877.096,88.4343 881.183,88.5581 885.271,88.6833 889.359,88.8096 893.446,88.9368 897.534,89.0648 
  901.622,89.1934 905.71,89.3224 909.797,89.4517 913.885,89.5811 917.973,89.7104 922.06,89.8395 926.148,89.9685 930.236,90.0975 934.323,90.2265 938.411,90.3555 
  942.499,90.4844 946.587,90.6133 950.674,90.7422 954.762,90.8711 958.85,90.9999 962.937,91.1286 967.025,91.2573 971.113,91.3859 975.2,91.5143 979.288,91.6426 
  983.376,91.7707 987.464,91.8986 991.551,92.0263 995.639,92.1537 999.727,92.2809 1003.81,92.4076 1007.9,92.534 1011.99,92.66 1016.08,92.7855 1020.17,92.9106 
  1024.25,93.035 1028.34,93.1588 1032.43,93.282 1036.52,93.4044 1040.6,93.526 1044.69,93.6468 1048.78,93.7668 1052.87,93.8862 1056.95,94.0049 1061.04,94.123 
  1065.13,94.2405 1069.22,94.3575 1073.31,94.4738 1077.39,94.5896 1081.48,94.7048 1085.57,94.8195 1089.66,94.9336 1093.74,95.0471 1097.83,95.1601 1101.92,95.2725 
  1106.01,95.3844 1110.09,95.4957 1114.18,95.6065 1118.27,95.7168 1122.36,95.8265 1126.45,95.9357 1130.53,96.0443 1134.62,96.1524 1138.71,96.2599 1142.8,96.3668 
  1146.88,96.4732 1150.97,96.579 1155.06,96.6842 1159.15,96.7888 1163.23,96.8929 1167.32,96.9963 1171.41,97.0991 1175.5,97.2013 1179.59,97.3029 1183.67,97.4037 
  1187.76,97.504 1191.85,97.6035 1195.94,97.7024 1200.02,97.8007 1204.11,97.8983 1208.2,97.9954 1212.29,98.092 1216.37,98.1879 1220.46,98.2833 1224.55,98.3782 
  1228.64,98.4725 1232.73,98.5662 1236.81,98.6594 1240.9,98.7521 1244.99,98.8442 1249.08,98.9358 1253.16,99.0269 1257.25,99.1175 1261.34,99.2075 1265.43,99.297 
  1269.51,99.3861 1273.6,99.4746 1277.69,99.5626 1281.78,99.6501 1285.87,99.7371 1289.95,99.8236 1294.04,99.9096 1298.13,99.9951 1302.22,100.08 1306.3,100.165 
  1310.39,100.249 1314.48,100.332 1318.57,100.415 1322.66,100.498 1326.74,100.58 1330.83,100.662 1334.92,100.743 1339.01,100.824 1343.09,100.904 1347.18,100.984 
  1351.27,101.063 1355.36,101.142 1359.44,101.22 1363.53,101.298 1367.62,101.375 1371.71,101.452 1375.8,101.528 1379.88,101.604 1383.97,101.68 1388.06,101.755 
  1392.15,101.829 1396.23,101.903 1400.32,101.977 1404.41,102.05 1408.5,102.123 1412.58,102.196 1416.67,102.268 1420.76,102.339 1424.85,102.41 1428.94,102.481 
  1433.02,102.551 1437.11,102.621 1441.2,102.691 1445.29,102.76 1449.37,102.829 1453.46,102.897 1457.55,102.965 1461.64,103.032 1465.72,103.099 1469.81,103.166 
  1473.9,103.232 1477.99,103.298 1482.08,103.363 1486.16,103.428 1490.25,103.493 1494.34,103.557 1498.43,103.621 1502.51,103.685 1506.6,103.748 1510.69,103.811 
  1514.78,103.873 1518.86,103.935 1522.95,103.997 1527.04,104.058 1531.13,104.119 1535.22,104.179 1539.3,104.24 1543.39,104.299 1547.48,104.359 1551.57,104.418 
  1555.65,104.477 1559.74,104.535 1563.83,104.593 1567.92,104.651 1572,104.708 1576.09,104.765 1580.18,104.821 1584.27,104.878 1588.36,104.934 1592.44,104.989 
  1596.53,105.045 1600.62,105.1 1604.71,105.154 1608.79,105.209 1612.88,105.263 1616.97,105.316 1621.06,105.37 1625.15,105.423 1629.23,105.476 1633.32,105.528 
  1637.41,105.58 1641.5,105.632 1645.58,105.684 1649.67,105.735 1653.76,105.786 1657.85,105.837 1661.93,105.888 1666.02,105.938 1670.11,105.988 1674.2,106.038 
  1678.29,106.088 1682.37,106.137 1686.46,106.186 1690.55,106.235 1694.64,106.284 1698.72,106.332 1702.81,106.38 1706.9,106.428 1710.99,106.476 1715.07,106.523 
  1719.16,106.57 1723.25,106.617 1727.34,106.663 1731.43,106.709 1735.51,106.755 1739.6,106.8 1743.69,106.846 1747.78,106.891 1751.86,106.935 1755.95,106.979 
  1760.04,107.023 1764.13,107.067 1768.21,107.111 1772.3,107.154 1776.39,107.196 1780.48,107.239 1784.57,107.281 1788.65,107.323 1792.74,107.365 1796.83,107.406 
  1800.92,107.447 1805,107.488 1809.09,107.528 1813.18,107.568 1817.27,107.608 1821.35,107.647 1825.44,107.687 1829.53,107.726 1833.62,107.764 1837.71,107.803 
  1841.79,107.841 1845.88,107.879 1849.97,107.916 1854.06,107.954 1858.14,107.991 1862.23,108.028 1866.32,108.064 1870.41,108.101 1874.49,108.137 1878.58,108.173 
  1882.67,108.209 1886.76,108.244 1890.85,108.28 1894.93,108.315 1899.02,108.35 1903.11,108.385 1907.2,108.42 1911.28,108.454 1915.37,108.488 1919.46,108.523 
  1923.55,108.557 1927.63,108.591 1931.72,108.624 1935.81,108.658 1939.9,108.692 1943.99,108.725 1948.07,108.759 1952.16,108.792 1956.25,108.826 1960.34,108.859 
  1964.42,108.893 1968.51,108.926 1972.6,108.96 1976.69,108.993 1980.78,109.027 1984.86,109.06 1988.95,109.094 1993.04,109.127 1997.13,109.16 2001.21,109.194 
  2005.3,109.227 2009.39,109.259 2013.48,109.292 2017.56,109.324 2021.65,109.356 2025.74,109.388 2029.83,109.42 2033.92,109.451 2038,109.482 2042.09,109.513 
  2046.18,109.543 2050.27,109.573 2054.35,109.603 2058.44,109.632 2062.53,109.661 2066.62,109.69 2070.7,109.718 2074.79,109.746 2078.88,109.774 2082.97,109.801 
  2087.06,109.828 2091.14,109.854 2095.23,109.88 2099.32,109.906 2103.41,109.931 2107.49,109.956 2111.58,109.981 2115.67,110.005 2119.76,110.029 2123.84,110.053 
  2127.93,110.076 2132.02,110.099 2136.11,110.122 2140.2,110.145 2144.28,110.167 2148.37,110.189 2152.46,110.211 2156.55,110.232 2160.63,110.254 2164.72,110.275 
  2168.81,110.296 2172.9,110.317 2176.98,110.338 2181.07,110.359 2185.16,110.38 2189.25,110.401 2193.34,110.422 2197.42,110.443 2201.51,110.464 2205.6,110.485 
  2209.69,110.506 2213.77,110.528 2217.86,110.549 2221.95,110.571 2226.04,110.593 2230.12,110.616 2234.21,110.639 2238.3,110.662 2242.39,110.686 2246.48,110.709 
  2250.56,110.733 2254.65,110.756 2258.74,110.779 2262.83,110.801 2266.91,110.824 2271,110.846 2275.09,110.868 2279.18,110.89 2283.26,110.911 2287.35,110.933 
  2291.44,110.954 
  "/>
<polyline clip-path="url(#clip122)" style="stroke:#e26f46; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  247.59,1384.24 251.677,1376.5 255.765,1368.75 259.853,1360.97 263.94,1353.18 268.028,1345.37 272.116,1337.54 276.203,1329.68 280.291,1321.81 284.379,1313.92 
  288.467,1306.01 292.554,1298.07 296.642,1290.12 300.73,1282.14 304.817,1274.14 308.905,1266.12 312.993,1258.08 317.08,1250.02 321.168,1241.94 325.256,1233.84 
  329.344,1225.72 333.431,1217.58 337.519,1209.43 341.607,1201.26 345.694,1193.07 349.782,1184.87 353.87,1176.66 357.957,1168.43 362.045,1160.2 366.133,1151.96 
  370.221,1143.72 374.308,1135.48 378.396,1127.25 382.484,1119.02 386.571,1110.81 390.659,1102.61 394.747,1094.44 398.834,1086.29 402.922,1078.19 407.01,1070.12 
  411.098,1062.1 415.185,1054.14 419.273,1046.25 423.361,1038.43 427.448,1030.69 431.536,1023.06 435.624,1015.53 439.712,1008.13 443.799,1000.86 447.887,993.728 
  451.975,986.754 456.062,979.945 460.15,973.314 464.238,966.871 468.325,960.627 472.413,954.594 476.501,948.783 480.589,943.204 484.676,937.869 488.764,932.79 
  492.852,927.973 496.939,923.421 501.027,919.136 505.115,915.12 509.202,911.373 513.29,907.895 517.378,904.684 521.466,901.739 525.553,899.057 529.641,896.634 
  533.729,894.464 537.816,892.542 541.904,890.862 545.992,889.415 550.079,888.193 554.167,887.187 558.255,886.385 562.343,885.778 566.43,885.353 570.518,885.106 
  574.606,885.026 578.693,885.107 582.781,885.339 586.869,885.716 590.956,886.23 595.044,886.872 599.132,887.637 603.22,888.518 607.307,889.507 611.395,890.597 
  615.483,891.784 619.57,893.059 623.658,894.419 627.746,895.856 631.833,897.365 635.921,898.941 640.009,900.579 644.097,902.274 648.184,904.021 652.272,905.817 
  656.36,907.657 660.447,909.54 664.535,911.463 668.623,913.423 672.711,915.417 676.798,917.443 680.886,919.498 684.974,921.581 689.061,923.688 693.149,925.817 
  697.237,927.967 701.324,930.134 705.412,932.318 709.5,934.515 713.588,936.725 717.675,938.944 721.763,941.172 725.851,943.406 729.938,945.646 734.026,947.891 
  738.114,950.14 742.201,952.392 746.289,954.647 750.377,956.905 754.465,959.164 758.552,961.424 762.64,963.684 766.728,965.944 770.815,968.202 774.903,970.459 
  778.991,972.713 783.078,974.964 787.166,977.212 791.254,979.455 795.342,981.693 799.429,983.925 803.517,986.151 807.605,988.371 811.692,990.582 815.78,992.785 
  819.868,994.98 823.955,997.167 828.043,999.345 832.131,1001.52 836.219,1003.68 840.306,1005.83 844.394,1007.98 848.482,1010.11 852.569,1012.24 856.657,1014.36 
  860.745,1016.47 864.832,1018.57 868.92,1020.66 873.008,1022.75 877.096,1024.82 881.183,1026.88 885.271,1028.94 889.359,1030.98 893.446,1033.02 897.534,1035.04 
  901.622,1037.06 905.71,1039.06 909.797,1041.06 913.885,1043.04 917.973,1045.01 922.06,1046.98 926.148,1048.93 930.236,1050.87 934.323,1052.8 938.411,1054.72 
  942.499,1056.63 946.587,1058.53 950.674,1060.42 954.762,1062.3 958.85,1064.17 962.937,1066.03 967.025,1067.88 971.113,1069.73 975.2,1071.56 979.288,1073.38 
  983.376,1075.19 987.464,1076.99 991.551,1078.78 995.639,1080.56 999.727,1082.33 1003.81,1084.09 1007.9,1085.85 1011.99,1087.59 1016.08,1089.32 1020.17,1091.04 
  1024.25,1092.76 1028.34,1094.46 1032.43,1096.15 1036.52,1097.84 1040.6,1099.51 1044.69,1101.17 1048.78,1102.83 1052.87,1104.47 1056.95,1106.11 1061.04,1107.73 
  1065.13,1109.35 1069.22,1110.96 1073.31,1112.56 1077.39,1114.14 1081.48,1115.72 1085.57,1117.29 1089.66,1118.86 1093.74,1120.41 1097.83,1121.95 1101.92,1123.49 
  1106.01,1125.01 1110.09,1126.53 1114.18,1128.04 1118.27,1129.54 1122.36,1131.03 1126.45,1132.52 1130.53,1133.99 1134.62,1135.46 1138.71,1136.91 1142.8,1138.36 
  1146.88,1139.81 1150.97,1141.24 1155.06,1142.66 1159.15,1144.08 1163.23,1145.49 1167.32,1146.89 1171.41,1148.28 1175.5,1149.66 1179.59,1151.04 1183.67,1152.4 
  1187.76,1153.76 1191.85,1155.11 1195.94,1156.46 1200.02,1157.79 1204.11,1159.12 1208.2,1160.44 1212.29,1161.75 1216.37,1163.06 1220.46,1164.35 1224.55,1165.64 
  1228.64,1166.92 1232.73,1168.2 1236.81,1169.47 1240.9,1170.73 1244.99,1171.98 1249.08,1173.22 1253.16,1174.46 1257.25,1175.69 1261.34,1176.91 1265.43,1178.13 
  1269.51,1179.34 1273.6,1180.54 1277.69,1181.74 1281.78,1182.93 1285.87,1184.11 1289.95,1185.28 1294.04,1186.45 1298.13,1187.61 1302.22,1188.76 1306.3,1189.91 
  1310.39,1191.05 1314.48,1192.19 1318.57,1193.31 1322.66,1194.43 1326.74,1195.55 1330.83,1196.66 1334.92,1197.76 1339.01,1198.85 1343.09,1199.94 1347.18,1201.02 
  1351.27,1202.1 1355.36,1203.17 1359.44,1204.23 1363.53,1205.29 1367.62,1206.34 1371.71,1207.38 1375.8,1208.42 1379.88,1209.45 1383.97,1210.48 1388.06,1211.5 
  1392.15,1212.51 1396.23,1213.52 1400.32,1214.52 1404.41,1215.52 1408.5,1216.51 1412.58,1217.5 1416.67,1218.48 1420.76,1219.45 1424.85,1220.42 1428.94,1221.38 
  1433.02,1222.34 1437.11,1223.29 1441.2,1224.23 1445.29,1225.17 1449.37,1226.11 1453.46,1227.04 1457.55,1227.96 1461.64,1228.88 1465.72,1229.79 1469.81,1230.7 
  1473.9,1231.6 1477.99,1232.5 1482.08,1233.39 1486.16,1234.27 1490.25,1235.15 1494.34,1236.03 1498.43,1236.9 1502.51,1237.77 1506.6,1238.63 1510.69,1239.48 
  1514.78,1240.33 1518.86,1241.18 1522.95,1242.02 1527.04,1242.85 1531.13,1243.68 1535.22,1244.51 1539.3,1245.33 1543.39,1246.15 1547.48,1246.96 1551.57,1247.76 
  1555.65,1248.57 1559.74,1249.36 1563.83,1250.16 1567.92,1250.94 1572,1251.73 1576.09,1252.5 1580.18,1253.28 1584.27,1254.05 1588.36,1254.81 1592.44,1255.57 
  1596.53,1256.33 1600.62,1257.08 1604.71,1257.83 1608.79,1258.57 1612.88,1259.31 1616.97,1260.04 1621.06,1260.77 1625.15,1261.5 1629.23,1262.22 1633.32,1262.94 
  1637.41,1263.65 1641.5,1264.36 1645.58,1265.06 1649.67,1265.76 1653.76,1266.46 1657.85,1267.15 1661.93,1267.84 1666.02,1268.52 1670.11,1269.2 1674.2,1269.88 
  1678.29,1270.55 1682.37,1271.22 1686.46,1271.89 1690.55,1272.55 1694.64,1273.2 1698.72,1273.86 1702.81,1274.51 1706.9,1275.15 1710.99,1275.79 1715.07,1276.43 
  1719.16,1277.07 1723.25,1277.7 1727.34,1278.32 1731.43,1278.95 1735.51,1279.57 1739.6,1280.18 1743.69,1280.8 1747.78,1281.4 1751.86,1282.01 1755.95,1282.61 
  1760.04,1283.21 1764.13,1283.8 1768.21,1284.39 1772.3,1284.98 1776.39,1285.56 1780.48,1286.14 1784.57,1286.72 1788.65,1287.29 1792.74,1287.86 1796.83,1288.43 
  1800.92,1288.99 1805,1289.55 1809.09,1290.11 1813.18,1290.66 1817.27,1291.21 1821.35,1291.76 1825.44,1292.3 1829.53,1292.84 1833.62,1293.38 1837.71,1293.91 
  1841.79,1294.44 1845.88,1294.97 1849.97,1295.49 1854.06,1296.01 1858.14,1296.53 1862.23,1297.05 1866.32,1297.56 1870.41,1298.07 1874.49,1298.57 1878.58,1299.07 
  1882.67,1299.57 1886.76,1300.07 1890.85,1300.56 1894.93,1301.06 1899.02,1301.54 1903.11,1302.03 1907.2,1302.51 1911.28,1302.99 1915.37,1303.47 1919.46,1303.94 
  1923.55,1304.41 1927.63,1304.88 1931.72,1305.35 1935.81,1305.81 1939.9,1306.27 1943.99,1306.73 1948.07,1307.19 1952.16,1307.64 1956.25,1308.09 1960.34,1308.54 
  1964.42,1308.99 1968.51,1309.43 1972.6,1309.87 1976.69,1310.31 1980.78,1310.75 1984.86,1311.18 1988.95,1311.61 1993.04,1312.04 1997.13,1312.47 2001.21,1312.89 
  2005.3,1313.32 2009.39,1313.74 2013.48,1314.15 2017.56,1314.57 2021.65,1314.98 2025.74,1315.39 2029.83,1315.8 2033.92,1316.2 2038,1316.6 2042.09,1317 
  2046.18,1317.4 2050.27,1317.8 2054.35,1318.19 2058.44,1318.58 2062.53,1318.96 2066.62,1319.35 2070.7,1319.73 2074.79,1320.11 2078.88,1320.49 2082.97,1320.86 
  2087.06,1321.24 2091.14,1321.61 2095.23,1321.97 2099.32,1322.34 2103.41,1322.7 2107.49,1323.06 2111.58,1323.42 2115.67,1323.78 2119.76,1324.13 2123.84,1324.48 
  2127.93,1324.83 2132.02,1325.18 2136.11,1325.52 2140.2,1325.86 2144.28,1326.2 2148.37,1326.54 2152.46,1326.88 2156.55,1327.21 2160.63,1327.54 2164.72,1327.87 
  2168.81,1328.2 2172.9,1328.53 2176.98,1328.85 2181.07,1329.17 2185.16,1329.49 2189.25,1329.81 2193.34,1330.13 2197.42,1330.44 2201.51,1330.76 2205.6,1331.07 
  2209.69,1331.38 2213.77,1331.69 2217.86,1332 2221.95,1332.3 2226.04,1332.61 2230.12,1332.91 2234.21,1333.22 2238.3,1333.52 2242.39,1333.82 2246.48,1334.12 
  2250.56,1334.41 2254.65,1334.71 2258.74,1335 2262.83,1335.29 2266.91,1335.58 2271,1335.87 2275.09,1336.15 2279.18,1336.44 2283.26,1336.72 2287.35,1337 
  2291.44,1337.28 
  "/>
<path clip-path="url(#clip120)" d="
M1587.09 248.629 L2280.54 248.629 L2280.54 93.1086 L1587.09 93.1086  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<polyline clip-path="url(#clip120)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1587.09,248.629 2280.54,248.629 2280.54,93.1086 1587.09,93.1086 1587.09,248.629 
  "/>
<polyline clip-path="url(#clip120)" style="stroke:#009af9; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1611.16,144.949 1755.59,144.949 
  "/>
<path clip-path="url(#clip120)" d="M1800.01 141.28 Q1801.61 138.409 1803.83 137.044 Q1806.06 135.678 1809.06 135.678 Q1813.12 135.678 1815.31 138.525 Q1817.51 141.349 1817.51 146.581 L1817.51 162.229 L1813.23 162.229 L1813.23 146.719 Q1813.23 142.993 1811.91 141.187 Q1810.59 139.382 1807.88 139.382 Q1804.57 139.382 1802.65 141.581 Q1800.73 143.78 1800.73 147.576 L1800.73 162.229 L1796.45 162.229 L1796.45 146.719 Q1796.45 142.969 1795.13 141.187 Q1793.81 139.382 1791.06 139.382 Q1787.79 139.382 1785.87 141.604 Q1783.95 143.803 1783.95 147.576 L1783.95 162.229 L1779.67 162.229 L1779.67 136.303 L1783.95 136.303 L1783.95 140.331 Q1785.41 137.946 1787.44 136.812 Q1789.48 135.678 1792.28 135.678 Q1795.11 135.678 1797.07 137.113 Q1799.06 138.548 1800.01 141.28 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M1837.79 149.196 Q1832.63 149.196 1830.64 150.377 Q1828.65 151.557 1828.65 154.405 Q1828.65 156.673 1830.13 158.016 Q1831.63 159.335 1834.2 159.335 Q1837.75 159.335 1839.87 156.835 Q1842.03 154.312 1842.03 150.145 L1842.03 149.196 L1837.79 149.196 M1846.29 147.437 L1846.29 162.229 L1842.03 162.229 L1842.03 158.293 Q1840.57 160.655 1838.39 161.789 Q1836.22 162.9 1833.07 162.9 Q1829.09 162.9 1826.73 160.678 Q1824.39 158.432 1824.39 154.682 Q1824.39 150.307 1827.31 148.085 Q1830.25 145.863 1836.06 145.863 L1842.03 145.863 L1842.03 145.446 Q1842.03 142.507 1840.08 140.909 Q1838.16 139.289 1834.67 139.289 Q1832.44 139.289 1830.34 139.821 Q1828.23 140.354 1826.29 141.419 L1826.29 137.483 Q1828.62 136.581 1830.82 136.141 Q1833.02 135.678 1835.11 135.678 Q1840.73 135.678 1843.51 138.594 Q1846.29 141.511 1846.29 147.437 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M1871.59 137.067 L1871.59 141.094 Q1869.78 140.169 1867.84 139.706 Q1865.89 139.243 1863.81 139.243 Q1860.64 139.243 1859.04 140.215 Q1857.47 141.187 1857.47 143.131 Q1857.47 144.613 1858.6 145.469 Q1859.74 146.303 1863.16 147.067 L1864.62 147.391 Q1869.16 148.363 1871.06 150.145 Q1872.98 151.905 1872.98 155.076 Q1872.98 158.687 1870.11 160.793 Q1867.26 162.9 1862.26 162.9 Q1860.18 162.9 1857.91 162.483 Q1855.66 162.09 1853.16 161.28 L1853.16 156.881 Q1855.52 158.108 1857.81 158.733 Q1860.11 159.335 1862.35 159.335 Q1865.36 159.335 1866.98 158.317 Q1868.6 157.275 1868.6 155.4 Q1868.6 153.664 1867.42 152.738 Q1866.26 151.812 1862.31 150.956 L1860.82 150.608 Q1856.87 149.775 1855.11 148.062 Q1853.35 146.326 1853.35 143.317 Q1853.35 139.659 1855.94 137.669 Q1858.53 135.678 1863.3 135.678 Q1865.66 135.678 1867.74 136.025 Q1869.83 136.372 1871.59 137.067 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M1896.29 137.067 L1896.29 141.094 Q1894.48 140.169 1892.54 139.706 Q1890.59 139.243 1888.51 139.243 Q1885.34 139.243 1883.74 140.215 Q1882.17 141.187 1882.17 143.131 Q1882.17 144.613 1883.3 145.469 Q1884.43 146.303 1887.86 147.067 L1889.32 147.391 Q1893.86 148.363 1895.75 150.145 Q1897.68 151.905 1897.68 155.076 Q1897.68 158.687 1894.81 160.793 Q1891.96 162.9 1886.96 162.9 Q1884.87 162.9 1882.61 162.483 Q1880.36 162.09 1877.86 161.28 L1877.86 156.881 Q1880.22 158.108 1882.51 158.733 Q1884.81 159.335 1887.05 159.335 Q1890.06 159.335 1891.68 158.317 Q1893.3 157.275 1893.3 155.4 Q1893.3 153.664 1892.12 152.738 Q1890.96 151.812 1887 150.956 L1885.52 150.608 Q1881.56 149.775 1879.81 148.062 Q1878.05 146.326 1878.05 143.317 Q1878.05 139.659 1880.64 137.669 Q1883.23 135.678 1888 135.678 Q1890.36 135.678 1892.44 136.025 Q1894.53 136.372 1896.29 137.067 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M1905.06 156.349 L1909.94 156.349 L1909.94 162.229 L1905.06 162.229 L1905.06 156.349 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M1936.05 137.067 L1936.05 141.094 Q1934.25 140.169 1932.3 139.706 Q1930.36 139.243 1928.28 139.243 Q1925.11 139.243 1923.51 140.215 Q1921.93 141.187 1921.93 143.131 Q1921.93 144.613 1923.07 145.469 Q1924.2 146.303 1927.63 147.067 L1929.09 147.391 Q1933.62 148.363 1935.52 150.145 Q1937.44 151.905 1937.44 155.076 Q1937.44 158.687 1934.57 160.793 Q1931.73 162.9 1926.73 162.9 Q1924.64 162.9 1922.37 162.483 Q1920.13 162.09 1917.63 161.28 L1917.63 156.881 Q1919.99 158.108 1922.28 158.733 Q1924.57 159.335 1926.82 159.335 Q1929.83 159.335 1931.45 158.317 Q1933.07 157.275 1933.07 155.4 Q1933.07 153.664 1931.89 152.738 Q1930.73 151.812 1926.77 150.956 L1925.29 150.608 Q1921.33 149.775 1919.57 148.062 Q1917.81 146.326 1917.81 143.317 Q1917.81 139.659 1920.41 137.669 Q1923 135.678 1927.77 135.678 Q1930.13 135.678 1932.21 136.025 Q1934.3 136.372 1936.05 137.067 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M1969.53 126.257 Q1966.43 131.581 1964.92 136.789 Q1963.42 141.997 1963.42 147.344 Q1963.42 152.692 1964.92 157.946 Q1966.45 163.178 1969.53 168.479 L1965.82 168.479 Q1962.35 163.039 1960.61 157.784 Q1958.9 152.53 1958.9 147.344 Q1958.9 142.182 1960.61 136.951 Q1962.33 131.72 1965.82 126.257 L1969.53 126.257 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M1992.56 143.594 Q1995.92 144.312 1997.79 146.581 Q1999.69 148.849 1999.69 152.182 Q1999.69 157.298 1996.17 160.099 Q1992.65 162.9 1986.17 162.9 Q1983.99 162.9 1981.68 162.46 Q1979.39 162.043 1976.93 161.187 L1976.93 156.673 Q1978.88 157.807 1981.19 158.386 Q1983.51 158.965 1986.03 158.965 Q1990.43 158.965 1992.72 157.229 Q1995.04 155.493 1995.04 152.182 Q1995.04 149.127 1992.88 147.414 Q1990.75 145.678 1986.93 145.678 L1982.91 145.678 L1982.91 141.835 L1987.12 141.835 Q1990.57 141.835 1992.4 140.469 Q1994.23 139.081 1994.23 136.488 Q1994.23 133.826 1992.33 132.414 Q1990.45 130.979 1986.93 130.979 Q1985.01 130.979 1982.81 131.395 Q1980.61 131.812 1977.98 132.692 L1977.98 128.525 Q1980.64 127.784 1982.95 127.414 Q1985.29 127.044 1987.35 127.044 Q1992.67 127.044 1995.78 129.474 Q1998.88 131.882 1998.88 136.002 Q1998.88 138.872 1997.24 140.863 Q1995.59 142.831 1992.56 143.594 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M2022.72 143.594 Q2026.08 144.312 2027.95 146.581 Q2029.85 148.849 2029.85 152.182 Q2029.85 157.298 2026.33 160.099 Q2022.81 162.9 2016.33 162.9 Q2014.16 162.9 2011.84 162.46 Q2009.55 162.043 2007.1 161.187 L2007.1 156.673 Q2009.04 157.807 2011.36 158.386 Q2013.67 158.965 2016.19 158.965 Q2020.59 158.965 2022.88 157.229 Q2025.2 155.493 2025.2 152.182 Q2025.2 149.127 2023.05 147.414 Q2020.92 145.678 2017.1 145.678 L2013.07 145.678 L2013.07 141.835 L2017.28 141.835 Q2020.73 141.835 2022.56 140.469 Q2024.39 139.081 2024.39 136.488 Q2024.39 133.826 2022.49 132.414 Q2020.61 130.979 2017.1 130.979 Q2015.17 130.979 2012.98 131.395 Q2010.78 131.812 2008.14 132.692 L2008.14 128.525 Q2010.8 127.784 2013.11 127.414 Q2015.45 127.044 2017.51 127.044 Q2022.84 127.044 2025.94 129.474 Q2029.04 131.882 2029.04 136.002 Q2029.04 138.872 2027.4 140.863 Q2025.75 142.831 2022.72 143.594 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M2038.76 127.669 L2057.12 127.669 L2057.12 131.604 L2043.04 131.604 L2043.04 140.076 Q2044.06 139.729 2045.08 139.567 Q2046.1 139.382 2047.12 139.382 Q2052.91 139.382 2056.29 142.553 Q2059.67 145.724 2059.67 151.141 Q2059.67 156.719 2056.19 159.821 Q2052.72 162.9 2046.4 162.9 Q2044.23 162.9 2041.96 162.53 Q2039.71 162.159 2037.3 161.418 L2037.3 156.719 Q2039.39 157.854 2041.61 158.409 Q2043.83 158.965 2046.31 158.965 Q2050.31 158.965 2052.65 156.858 Q2054.99 154.752 2054.99 151.141 Q2054.99 147.53 2052.65 145.423 Q2050.31 143.317 2046.31 143.317 Q2044.43 143.317 2042.56 143.733 Q2040.71 144.15 2038.76 145.03 L2038.76 127.669 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M2068.92 127.669 L2087.28 127.669 L2087.28 131.604 L2073.21 131.604 L2073.21 140.076 Q2074.23 139.729 2075.24 139.567 Q2076.26 139.382 2077.28 139.382 Q2083.07 139.382 2086.45 142.553 Q2089.83 145.724 2089.83 151.141 Q2089.83 156.719 2086.35 159.821 Q2082.88 162.9 2076.56 162.9 Q2074.39 162.9 2072.12 162.53 Q2069.87 162.159 2067.47 161.418 L2067.47 156.719 Q2069.55 157.854 2071.77 158.409 Q2073.99 158.965 2076.47 158.965 Q2080.48 158.965 2082.81 156.858 Q2085.15 154.752 2085.15 151.141 Q2085.15 147.53 2082.81 145.423 Q2080.48 143.317 2076.47 143.317 Q2074.6 143.317 2072.72 143.733 Q2070.87 144.15 2068.92 145.03 L2068.92 127.669 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M2111.89 131.743 L2100.08 150.192 L2111.89 150.192 L2111.89 131.743 M2110.66 127.669 L2116.54 127.669 L2116.54 150.192 L2121.47 150.192 L2121.47 154.081 L2116.54 154.081 L2116.54 162.229 L2111.89 162.229 L2111.89 154.081 L2096.29 154.081 L2096.29 149.567 L2110.66 127.669 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M2142.05 131.743 L2130.24 150.192 L2142.05 150.192 L2142.05 131.743 M2140.82 127.669 L2146.7 127.669 L2146.7 150.192 L2151.63 150.192 L2151.63 154.081 L2146.7 154.081 L2146.7 162.229 L2142.05 162.229 L2142.05 154.081 L2126.45 154.081 L2126.45 149.567 L2140.82 127.669 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M2173.53 143.594 Q2176.89 144.312 2178.76 146.581 Q2180.66 148.849 2180.66 152.182 Q2180.66 157.298 2177.14 160.099 Q2173.62 162.9 2167.14 162.9 Q2164.97 162.9 2162.65 162.46 Q2160.36 162.043 2157.91 161.187 L2157.91 156.673 Q2159.85 157.807 2162.16 158.386 Q2164.48 158.965 2167 158.965 Q2171.4 158.965 2173.69 157.229 Q2176.01 155.493 2176.01 152.182 Q2176.01 149.127 2173.85 147.414 Q2171.72 145.678 2167.91 145.678 L2163.88 145.678 L2163.88 141.835 L2168.09 141.835 Q2171.54 141.835 2173.37 140.469 Q2175.2 139.081 2175.2 136.488 Q2175.2 133.826 2173.3 132.414 Q2171.42 130.979 2167.91 130.979 Q2165.98 130.979 2163.79 131.395 Q2161.59 131.812 2158.95 132.692 L2158.95 128.525 Q2161.61 127.784 2163.92 127.414 Q2166.26 127.044 2168.32 127.044 Q2173.65 127.044 2176.75 129.474 Q2179.85 131.882 2179.85 136.002 Q2179.85 138.872 2178.21 140.863 Q2176.56 142.831 2173.53 143.594 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M2193.55 158.293 L2209.87 158.293 L2209.87 162.229 L2187.93 162.229 L2187.93 158.293 Q2190.59 155.539 2195.17 150.909 Q2199.78 146.256 2200.96 144.914 Q2203.21 142.391 2204.09 140.655 Q2204.99 138.895 2204.99 137.206 Q2204.99 134.451 2203.04 132.715 Q2201.12 130.979 2198.02 130.979 Q2195.82 130.979 2193.37 131.743 Q2190.94 132.507 2188.16 134.057 L2188.16 129.335 Q2190.98 128.201 2193.44 127.622 Q2195.89 127.044 2197.93 127.044 Q2203.3 127.044 2206.49 129.729 Q2209.69 132.414 2209.69 136.905 Q2209.69 139.034 2208.88 140.956 Q2208.09 142.854 2205.98 145.446 Q2205.41 146.118 2202.3 149.335 Q2199.2 152.53 2193.55 158.293 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M2218.41 126.257 L2222.12 126.257 Q2225.59 131.72 2227.3 136.951 Q2229.04 142.182 2229.04 147.344 Q2229.04 152.53 2227.3 157.784 Q2225.59 163.039 2222.12 168.479 L2218.41 168.479 Q2221.49 163.178 2223 157.946 Q2224.53 152.692 2224.53 147.344 Q2224.53 141.997 2223 136.789 Q2221.49 131.581 2218.41 126.257 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip120)" style="stroke:#e26f46; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1611.16,196.789 1755.59,196.789 
  "/>
<path clip-path="url(#clip120)" d="M1800.01 193.12 Q1801.61 190.249 1803.83 188.884 Q1806.06 187.518 1809.06 187.518 Q1813.12 187.518 1815.31 190.365 Q1817.51 193.189 1817.51 198.421 L1817.51 214.069 L1813.23 214.069 L1813.23 198.559 Q1813.23 194.833 1811.91 193.027 Q1810.59 191.222 1807.88 191.222 Q1804.57 191.222 1802.65 193.421 Q1800.73 195.62 1800.73 199.416 L1800.73 214.069 L1796.45 214.069 L1796.45 198.559 Q1796.45 194.809 1795.13 193.027 Q1793.81 191.222 1791.06 191.222 Q1787.79 191.222 1785.87 193.444 Q1783.95 195.643 1783.95 199.416 L1783.95 214.069 L1779.67 214.069 L1779.67 188.143 L1783.95 188.143 L1783.95 192.171 Q1785.41 189.786 1787.44 188.652 Q1789.48 187.518 1792.28 187.518 Q1795.11 187.518 1797.07 188.953 Q1799.06 190.388 1800.01 193.12 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M1837.79 201.036 Q1832.63 201.036 1830.64 202.217 Q1828.65 203.397 1828.65 206.245 Q1828.65 208.513 1830.13 209.856 Q1831.63 211.175 1834.2 211.175 Q1837.75 211.175 1839.87 208.675 Q1842.03 206.152 1842.03 201.985 L1842.03 201.036 L1837.79 201.036 M1846.29 199.277 L1846.29 214.069 L1842.03 214.069 L1842.03 210.133 Q1840.57 212.495 1838.39 213.629 Q1836.22 214.74 1833.07 214.74 Q1829.09 214.74 1826.73 212.518 Q1824.39 210.272 1824.39 206.522 Q1824.39 202.147 1827.31 199.925 Q1830.25 197.703 1836.06 197.703 L1842.03 197.703 L1842.03 197.286 Q1842.03 194.347 1840.08 192.749 Q1838.16 191.129 1834.67 191.129 Q1832.44 191.129 1830.34 191.661 Q1828.23 192.194 1826.29 193.259 L1826.29 189.323 Q1828.62 188.421 1830.82 187.981 Q1833.02 187.518 1835.11 187.518 Q1840.73 187.518 1843.51 190.434 Q1846.29 193.351 1846.29 199.277 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M1871.59 188.907 L1871.59 192.934 Q1869.78 192.009 1867.84 191.546 Q1865.89 191.083 1863.81 191.083 Q1860.64 191.083 1859.04 192.055 Q1857.47 193.027 1857.47 194.971 Q1857.47 196.453 1858.6 197.309 Q1859.74 198.143 1863.16 198.907 L1864.62 199.231 Q1869.16 200.203 1871.06 201.985 Q1872.98 203.745 1872.98 206.916 Q1872.98 210.527 1870.11 212.633 Q1867.26 214.74 1862.26 214.74 Q1860.18 214.74 1857.91 214.323 Q1855.66 213.93 1853.16 213.12 L1853.16 208.721 Q1855.52 209.948 1857.81 210.573 Q1860.11 211.175 1862.35 211.175 Q1865.36 211.175 1866.98 210.157 Q1868.6 209.115 1868.6 207.24 Q1868.6 205.504 1867.42 204.578 Q1866.26 203.652 1862.31 202.796 L1860.82 202.448 Q1856.87 201.615 1855.11 199.902 Q1853.35 198.166 1853.35 195.157 Q1853.35 191.499 1855.94 189.509 Q1858.53 187.518 1863.3 187.518 Q1865.66 187.518 1867.74 187.865 Q1869.83 188.212 1871.59 188.907 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M1896.29 188.907 L1896.29 192.934 Q1894.48 192.009 1892.54 191.546 Q1890.59 191.083 1888.51 191.083 Q1885.34 191.083 1883.74 192.055 Q1882.17 193.027 1882.17 194.971 Q1882.17 196.453 1883.3 197.309 Q1884.43 198.143 1887.86 198.907 L1889.32 199.231 Q1893.86 200.203 1895.75 201.985 Q1897.68 203.745 1897.68 206.916 Q1897.68 210.527 1894.81 212.633 Q1891.96 214.74 1886.96 214.74 Q1884.87 214.74 1882.61 214.323 Q1880.36 213.93 1877.86 213.12 L1877.86 208.721 Q1880.22 209.948 1882.51 210.573 Q1884.81 211.175 1887.05 211.175 Q1890.06 211.175 1891.68 210.157 Q1893.3 209.115 1893.3 207.24 Q1893.3 205.504 1892.12 204.578 Q1890.96 203.652 1887 202.796 L1885.52 202.448 Q1881.56 201.615 1879.81 199.902 Q1878.05 198.166 1878.05 195.157 Q1878.05 191.499 1880.64 189.509 Q1883.23 187.518 1888 187.518 Q1890.36 187.518 1892.44 187.865 Q1894.53 188.212 1896.29 188.907 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M1905.06 208.189 L1909.94 208.189 L1909.94 214.069 L1905.06 214.069 L1905.06 208.189 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M1916.47 188.143 L1920.99 188.143 L1929.09 209.902 L1937.19 188.143 L1941.7 188.143 L1931.98 214.069 L1926.19 214.069 L1916.47 188.143 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M1972.88 178.097 Q1969.78 183.421 1968.28 188.629 Q1966.77 193.837 1966.77 199.184 Q1966.77 204.532 1968.28 209.786 Q1969.8 215.018 1972.88 220.319 L1969.18 220.319 Q1965.71 214.879 1963.97 209.624 Q1962.26 204.37 1962.26 199.184 Q1962.26 194.022 1963.97 188.791 Q1965.68 183.56 1969.18 178.097 L1972.88 178.097 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M1995.92 195.434 Q1999.27 196.152 2001.15 198.421 Q2003.05 200.689 2003.05 204.022 Q2003.05 209.138 1999.53 211.939 Q1996.01 214.74 1989.53 214.74 Q1987.35 214.74 1985.04 214.3 Q1982.74 213.883 1980.29 213.027 L1980.29 208.513 Q1982.24 209.647 1984.55 210.226 Q1986.86 210.805 1989.39 210.805 Q1993.79 210.805 1996.08 209.069 Q1998.39 207.333 1998.39 204.022 Q1998.39 200.967 1996.24 199.254 Q1994.11 197.518 1990.29 197.518 L1986.26 197.518 L1986.26 193.675 L1990.48 193.675 Q1993.92 193.675 1995.75 192.309 Q1997.58 190.921 1997.58 188.328 Q1997.58 185.666 1995.68 184.254 Q1993.81 182.819 1990.29 182.819 Q1988.37 182.819 1986.17 183.235 Q1983.97 183.652 1981.33 184.532 L1981.33 180.365 Q1983.99 179.624 1986.31 179.254 Q1988.65 178.884 1990.71 178.884 Q1996.03 178.884 1999.13 181.314 Q2002.23 183.722 2002.23 187.842 Q2002.23 190.712 2000.59 192.703 Q1998.95 194.671 1995.92 195.434 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M2026.08 195.434 Q2029.43 196.152 2031.31 198.421 Q2033.21 200.689 2033.21 204.022 Q2033.21 209.138 2029.69 211.939 Q2026.17 214.74 2019.69 214.74 Q2017.51 214.74 2015.2 214.3 Q2012.91 213.883 2010.45 213.027 L2010.45 208.513 Q2012.4 209.647 2014.71 210.226 Q2017.03 210.805 2019.55 210.805 Q2023.95 210.805 2026.24 209.069 Q2028.55 207.333 2028.55 204.022 Q2028.55 200.967 2026.4 199.254 Q2024.27 197.518 2020.45 197.518 L2016.42 197.518 L2016.42 193.675 L2020.64 193.675 Q2024.09 193.675 2025.92 192.309 Q2027.74 190.921 2027.74 188.328 Q2027.74 185.666 2025.85 184.254 Q2023.97 182.819 2020.45 182.819 Q2018.53 182.819 2016.33 183.235 Q2014.13 183.652 2011.49 184.532 L2011.49 180.365 Q2014.16 179.624 2016.47 179.254 Q2018.81 178.884 2020.87 178.884 Q2026.19 178.884 2029.3 181.314 Q2032.4 183.722 2032.4 187.842 Q2032.4 190.712 2030.75 192.703 Q2029.11 194.671 2026.08 195.434 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M2042.12 179.509 L2060.48 179.509 L2060.48 183.444 L2046.4 183.444 L2046.4 191.916 Q2047.42 191.569 2048.44 191.407 Q2049.46 191.222 2050.48 191.222 Q2056.26 191.222 2059.64 194.393 Q2063.02 197.564 2063.02 202.981 Q2063.02 208.559 2059.55 211.661 Q2056.08 214.74 2049.76 214.74 Q2047.58 214.74 2045.31 214.37 Q2043.07 213.999 2040.66 213.258 L2040.66 208.559 Q2042.74 209.694 2044.97 210.249 Q2047.19 210.805 2049.67 210.805 Q2053.67 210.805 2056.01 208.698 Q2058.35 206.592 2058.35 202.981 Q2058.35 199.37 2056.01 197.263 Q2053.67 195.157 2049.67 195.157 Q2047.79 195.157 2045.92 195.573 Q2044.06 195.99 2042.12 196.87 L2042.12 179.509 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M2072.28 179.509 L2090.64 179.509 L2090.64 183.444 L2076.56 183.444 L2076.56 191.916 Q2077.58 191.569 2078.6 191.407 Q2079.62 191.222 2080.64 191.222 Q2086.42 191.222 2089.8 194.393 Q2093.18 197.564 2093.18 202.981 Q2093.18 208.559 2089.71 211.661 Q2086.24 214.74 2079.92 214.74 Q2077.74 214.74 2075.48 214.37 Q2073.23 213.999 2070.82 213.258 L2070.82 208.559 Q2072.91 209.694 2075.13 210.249 Q2077.35 210.805 2079.83 210.805 Q2083.83 210.805 2086.17 208.698 Q2088.51 206.592 2088.51 202.981 Q2088.51 199.37 2086.17 197.263 Q2083.83 195.157 2079.83 195.157 Q2077.95 195.157 2076.08 195.573 Q2074.23 195.99 2072.28 196.87 L2072.28 179.509 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M2115.24 183.583 L2103.44 202.032 L2115.24 202.032 L2115.24 183.583 M2114.02 179.509 L2119.9 179.509 L2119.9 202.032 L2124.83 202.032 L2124.83 205.921 L2119.9 205.921 L2119.9 214.069 L2115.24 214.069 L2115.24 205.921 L2099.64 205.921 L2099.64 201.407 L2114.02 179.509 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M2145.41 183.583 L2133.6 202.032 L2145.41 202.032 L2145.41 183.583 M2144.18 179.509 L2150.06 179.509 L2150.06 202.032 L2154.99 202.032 L2154.99 205.921 L2150.06 205.921 L2150.06 214.069 L2145.41 214.069 L2145.41 205.921 L2129.8 205.921 L2129.8 201.407 L2144.18 179.509 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M2176.89 195.434 Q2180.24 196.152 2182.12 198.421 Q2184.02 200.689 2184.02 204.022 Q2184.02 209.138 2180.5 211.939 Q2176.98 214.74 2170.5 214.74 Q2168.32 214.74 2166.01 214.3 Q2163.72 213.883 2161.26 213.027 L2161.26 208.513 Q2163.21 209.647 2165.52 210.226 Q2167.84 210.805 2170.36 210.805 Q2174.76 210.805 2177.05 209.069 Q2179.36 207.333 2179.36 204.022 Q2179.36 200.967 2177.21 199.254 Q2175.08 197.518 2171.26 197.518 L2167.23 197.518 L2167.23 193.675 L2171.45 193.675 Q2174.9 193.675 2176.72 192.309 Q2178.55 190.921 2178.55 188.328 Q2178.55 185.666 2176.66 184.254 Q2174.78 182.819 2171.26 182.819 Q2169.34 182.819 2167.14 183.235 Q2164.94 183.652 2162.3 184.532 L2162.3 180.365 Q2164.97 179.624 2167.28 179.254 Q2169.62 178.884 2171.68 178.884 Q2177 178.884 2180.1 181.314 Q2183.21 183.722 2183.21 187.842 Q2183.21 190.712 2181.56 192.703 Q2179.92 194.671 2176.89 195.434 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M2207.05 195.434 Q2210.41 196.152 2212.28 198.421 Q2214.18 200.689 2214.18 204.022 Q2214.18 209.138 2210.66 211.939 Q2207.14 214.74 2200.66 214.74 Q2198.48 214.74 2196.17 214.3 Q2193.88 213.883 2191.42 213.027 L2191.42 208.513 Q2193.37 209.647 2195.68 210.226 Q2198 210.805 2200.52 210.805 Q2204.92 210.805 2207.21 209.069 Q2209.53 207.333 2209.53 204.022 Q2209.53 200.967 2207.37 199.254 Q2205.24 197.518 2201.42 197.518 L2197.4 197.518 L2197.4 193.675 L2201.61 193.675 Q2205.06 193.675 2206.89 192.309 Q2208.72 190.921 2208.72 188.328 Q2208.72 185.666 2206.82 184.254 Q2204.94 182.819 2201.42 182.819 Q2199.5 182.819 2197.3 183.235 Q2195.1 183.652 2192.47 184.532 L2192.47 180.365 Q2195.13 179.624 2197.44 179.254 Q2199.78 178.884 2201.84 178.884 Q2207.16 178.884 2210.27 181.314 Q2213.37 183.722 2213.37 187.842 Q2213.37 190.712 2211.72 192.703 Q2210.08 194.671 2207.05 195.434 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip120)" d="M2221.77 178.097 L2225.47 178.097 Q2228.95 183.56 2230.66 188.791 Q2232.4 194.022 2232.4 199.184 Q2232.4 204.37 2230.66 209.624 Q2228.95 214.879 2225.47 220.319 L2221.77 220.319 Q2224.85 215.018 2226.35 209.786 Q2227.88 204.532 2227.88 199.184 Q2227.88 193.837 2226.35 188.629 Q2224.85 183.421 2221.77 178.097 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /></svg>



#### Training of the NeuralFMU

For the training of the NeuralFMU the parameters are extracted. The known ADAM optimizer for minimizing the gradient descent is used as further passing parameters. In addition, the previously defined loss and callback function, as well as the number of epochs are passed.


```julia
# train
paramsNet = Flux.params(neuralFMU)

optim = ADAM()
Flux.train!(lossSum, paramsNet, Iterators.repeated((), 300), optim; cb=callb) 
```

    ┌ Info: Loss [1]: 0.04719   Avg displacement in data: 0.21724
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_ME.ipynb:8
    ┌ Info: Loss [21]: 0.04325   Avg displacement in data: 0.20796
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_ME.ipynb:8
    ┌ Info: Loss [41]: 0.03975   Avg displacement in data: 0.19937
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_ME.ipynb:8
    ┌ Info: Loss [61]: 0.03358   Avg displacement in data: 0.18324
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_ME.ipynb:8
    ┌ Info: Loss [81]: 0.00654   Avg displacement in data: 0.08089
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_ME.ipynb:8
    ┌ Info: Loss [101]: 0.0033   Avg displacement in data: 0.05749
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_ME.ipynb:8
    ┌ Info: Loss [121]: 0.00263   Avg displacement in data: 0.05124
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_ME.ipynb:8
    ┌ Info: Loss [141]: 0.00241   Avg displacement in data: 0.04908
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_ME.ipynb:8
    ┌ Info: Loss [161]: 0.00221   Avg displacement in data: 0.04705
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_ME.ipynb:8
    ┌ Info: Loss [181]: 0.00205   Avg displacement in data: 0.04533
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_ME.ipynb:8
    ┌ Info: Loss [201]: 0.00191   Avg displacement in data: 0.04373
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_ME.ipynb:8
    ┌ Info: Loss [221]: 0.00179   Avg displacement in data: 0.04231
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_ME.ipynb:8
    ┌ Info: Loss [241]: 0.00168   Avg displacement in data: 0.04096
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_ME.ipynb:8
    ┌ Info: Loss [261]: 0.00158   Avg displacement in data: 0.03969
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_ME.ipynb:8
    ┌ Info: Loss [281]: 0.00148   Avg displacement in data: 0.03843
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_ME.ipynb:8


#### Comparison of the plots

Here three plots are compared with each other and only the position of the mass is considered. The first plot represents the *simpleFMU*, the second represents the *realFMU* (reference) and the third plot represents the result after training the NeuralFMU. 


```julia
# plot results mass.s
solutionAfter = neuralFMU(x₀, tStart)

fig = Plots.plot(xlabel="t [s]", ylabel="mass position [m]", linewidth=2,
                 xtickfontsize=12, ytickfontsize=12,
                 xguidefontsize=12, yguidefontsize=12,
                 legendfontsize=8, legend=:topright)

posNeuralFMU = collect(data[1] for data in solutionAfter.u)

Plots.plot!(fig, tSave, posSimple, label="SimpleFMU", linewidth=2)
Plots.plot!(fig, tSave, posReal, label="RealFMU", linewidth=2)
Plots.plot!(fig, tSave, posNeuralFMU, label="NeuralFMU (300 epochs)", linewidth=2)
fig 
```


<?xml version="1.0" encoding="utf-8"?>
<svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" width="600" height="400" viewBox="0 0 2400 1600">
<defs>
  <clipPath id="clip210">
    <rect x="0" y="0" width="2400" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip210)" d="
M0 1600 L2400 1600 L2400 0 L0 0  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip211">
    <rect x="480" y="0" width="1681" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip210)" d="
M303.503 1401.58 L2352.76 1401.58 L2352.76 47.2441 L303.503 47.2441  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip212">
    <rect x="303" y="47" width="2050" height="1355"/>
  </clipPath>
</defs>
<polyline clip-path="url(#clip212)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  361.501,1401.58 361.501,47.2441 
  "/>
<polyline clip-path="url(#clip212)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  748.152,1401.58 748.152,47.2441 
  "/>
<polyline clip-path="url(#clip212)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1134.8,1401.58 1134.8,47.2441 
  "/>
<polyline clip-path="url(#clip212)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1521.46,1401.58 1521.46,47.2441 
  "/>
<polyline clip-path="url(#clip212)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1908.11,1401.58 1908.11,47.2441 
  "/>
<polyline clip-path="url(#clip212)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  2294.76,1401.58 2294.76,47.2441 
  "/>
<polyline clip-path="url(#clip210)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  303.503,1401.58 2352.76,1401.58 
  "/>
<polyline clip-path="url(#clip210)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  361.501,1401.58 361.501,1382.68 
  "/>
<polyline clip-path="url(#clip210)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  748.152,1401.58 748.152,1382.68 
  "/>
<polyline clip-path="url(#clip210)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1134.8,1401.58 1134.8,1382.68 
  "/>
<polyline clip-path="url(#clip210)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1521.46,1401.58 1521.46,1382.68 
  "/>
<polyline clip-path="url(#clip210)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1908.11,1401.58 1908.11,1382.68 
  "/>
<polyline clip-path="url(#clip210)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  2294.76,1401.58 2294.76,1382.68 
  "/>
<path clip-path="url(#clip210)" d="M361.501 1438.36 Q356.084 1438.36 353.341 1443.71 Q350.633 1449.02 350.633 1459.71 Q350.633 1470.37 353.341 1475.72 Q356.084 1481.03 361.501 1481.03 Q366.952 1481.03 369.66 1475.72 Q372.403 1470.37 372.403 1459.71 Q372.403 1449.02 369.66 1443.71 Q366.952 1438.36 361.501 1438.36 M361.501 1432.8 Q370.216 1432.8 374.799 1439.71 Q379.417 1446.59 379.417 1459.71 Q379.417 1472.8 374.799 1479.71 Q370.216 1486.59 361.501 1486.59 Q352.785 1486.59 348.167 1479.71 Q343.584 1472.8 343.584 1459.71 Q343.584 1446.59 348.167 1439.71 Q352.785 1432.8 361.501 1432.8 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M733.725 1479.68 L745.183 1479.68 L745.183 1440.13 L732.718 1442.63 L732.718 1436.24 L745.114 1433.74 L752.128 1433.74 L752.128 1479.68 L763.586 1479.68 L763.586 1485.58 L733.725 1485.58 L733.725 1479.68 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1126.78 1479.68 L1151.26 1479.68 L1151.26 1485.58 L1118.35 1485.58 L1118.35 1479.68 Q1122.34 1475.55 1129.21 1468.6 Q1136.12 1461.62 1137.89 1459.61 Q1141.26 1455.82 1142.58 1453.22 Q1143.94 1450.58 1143.94 1448.05 Q1143.94 1443.91 1141.02 1441.31 Q1138.14 1438.71 1133.48 1438.71 Q1130.19 1438.71 1126.51 1439.85 Q1122.86 1441 1118.69 1443.32 L1118.69 1436.24 Q1122.93 1434.54 1126.61 1433.67 Q1130.29 1432.8 1133.35 1432.8 Q1141.4 1432.8 1146.19 1436.83 Q1150.98 1440.86 1150.98 1447.59 Q1150.98 1450.79 1149.77 1453.67 Q1148.59 1456.52 1145.43 1460.41 Q1144.56 1461.41 1139.91 1466.24 Q1135.26 1471.03 1126.78 1479.68 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1527.83 1457.63 Q1532.86 1458.7 1535.67 1462.11 Q1538.52 1465.51 1538.52 1470.51 Q1538.52 1478.18 1533.24 1482.39 Q1527.97 1486.59 1518.24 1486.59 Q1514.98 1486.59 1511.51 1485.93 Q1508.07 1485.3 1504.39 1484.02 L1504.39 1477.25 Q1507.31 1478.95 1510.78 1479.82 Q1514.25 1480.68 1518.04 1480.68 Q1524.63 1480.68 1528.07 1478.08 Q1531.54 1475.48 1531.54 1470.51 Q1531.54 1465.93 1528.31 1463.36 Q1525.12 1460.75 1519.39 1460.75 L1513.35 1460.75 L1513.35 1454.99 L1519.67 1454.99 Q1524.84 1454.99 1527.58 1452.94 Q1530.33 1450.86 1530.33 1446.97 Q1530.33 1442.98 1527.48 1440.86 Q1524.67 1438.71 1519.39 1438.71 Q1516.51 1438.71 1513.21 1439.33 Q1509.91 1439.96 1505.95 1441.27 L1505.95 1435.02 Q1509.94 1433.91 1513.42 1433.36 Q1516.92 1432.8 1520.01 1432.8 Q1528 1432.8 1532.65 1436.45 Q1537.31 1440.06 1537.31 1446.24 Q1537.31 1450.55 1534.84 1453.53 Q1532.38 1456.48 1527.83 1457.63 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1912.62 1439.85 L1894.91 1467.52 L1912.62 1467.52 L1912.62 1439.85 M1910.78 1433.74 L1919.6 1433.74 L1919.6 1467.52 L1927 1467.52 L1927 1473.36 L1919.6 1473.36 L1919.6 1485.58 L1912.62 1485.58 L1912.62 1473.36 L1889.22 1473.36 L1889.22 1466.59 L1910.78 1433.74 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M2280.17 1433.74 L2307.71 1433.74 L2307.71 1439.64 L2286.6 1439.64 L2286.6 1452.35 Q2288.13 1451.83 2289.65 1451.59 Q2291.18 1451.31 2292.71 1451.31 Q2301.39 1451.31 2306.46 1456.07 Q2311.53 1460.82 2311.53 1468.95 Q2311.53 1477.32 2306.32 1481.97 Q2301.11 1486.59 2291.63 1486.59 Q2288.37 1486.59 2284.97 1486.03 Q2281.6 1485.48 2277.99 1484.36 L2277.99 1477.32 Q2281.11 1479.02 2284.45 1479.85 Q2287.78 1480.68 2291.49 1480.68 Q2297.5 1480.68 2301.01 1477.52 Q2304.52 1474.36 2304.52 1468.95 Q2304.52 1463.53 2301.01 1460.37 Q2297.5 1457.21 2291.49 1457.21 Q2288.68 1457.21 2285.87 1457.84 Q2283.09 1458.46 2280.17 1459.78 L2280.17 1433.74 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1271.74 1527.62 L1271.74 1538.66 L1284.9 1538.66 L1284.9 1543.62 L1271.74 1543.62 L1271.74 1564.74 Q1271.74 1569.49 1273.03 1570.85 Q1274.35 1572.2 1278.34 1572.2 L1284.9 1572.2 L1284.9 1577.55 L1278.34 1577.55 Q1270.94 1577.55 1268.13 1574.8 Q1265.32 1572.03 1265.32 1564.74 L1265.32 1543.62 L1260.63 1543.62 L1260.63 1538.66 L1265.32 1538.66 L1265.32 1527.62 L1271.74 1527.62 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1315.32 1523.52 L1330.04 1523.52 L1330.04 1528.49 L1321.71 1528.49 L1321.71 1581.96 L1330.04 1581.96 L1330.04 1586.92 L1315.32 1586.92 L1315.32 1523.52 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1368.44 1539.8 L1368.44 1545.85 Q1365.73 1544.46 1362.82 1543.76 Q1359.9 1543.07 1356.78 1543.07 Q1352.02 1543.07 1349.62 1544.53 Q1347.26 1545.99 1347.26 1548.9 Q1347.26 1551.12 1348.96 1552.41 Q1350.66 1553.66 1355.8 1554.8 L1357.99 1555.29 Q1364.8 1556.75 1367.64 1559.42 Q1370.53 1562.06 1370.53 1566.82 Q1370.53 1572.24 1366.22 1575.4 Q1361.95 1578.55 1354.45 1578.55 Q1351.32 1578.55 1347.92 1577.93 Q1344.55 1577.34 1340.8 1576.12 L1340.8 1569.53 Q1344.34 1571.37 1347.78 1572.3 Q1351.22 1573.21 1354.59 1573.21 Q1359.1 1573.21 1361.53 1571.68 Q1363.96 1570.12 1363.96 1567.3 Q1363.96 1564.7 1362.19 1563.31 Q1360.46 1561.92 1354.52 1560.64 L1352.3 1560.12 Q1346.36 1558.87 1343.72 1556.3 Q1341.08 1553.69 1341.08 1549.18 Q1341.08 1543.69 1344.97 1540.71 Q1348.86 1537.72 1356.01 1537.72 Q1359.55 1537.72 1362.68 1538.24 Q1365.8 1538.76 1368.44 1539.8 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1395.63 1523.52 L1395.63 1586.92 L1380.91 1586.92 L1380.91 1581.96 L1389.21 1581.96 L1389.21 1528.49 L1380.91 1528.49 L1380.91 1523.52 L1395.63 1523.52 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip212)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  303.503,1361.96 2352.76,1361.96 
  "/>
<polyline clip-path="url(#clip212)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  303.503,1096.36 2352.76,1096.36 
  "/>
<polyline clip-path="url(#clip212)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  303.503,830.761 2352.76,830.761 
  "/>
<polyline clip-path="url(#clip212)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  303.503,565.16 2352.76,565.16 
  "/>
<polyline clip-path="url(#clip212)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  303.503,299.559 2352.76,299.559 
  "/>
<polyline clip-path="url(#clip210)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  303.503,1401.58 303.503,47.2441 
  "/>
<polyline clip-path="url(#clip210)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  303.503,1361.96 322.401,1361.96 
  "/>
<polyline clip-path="url(#clip210)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  303.503,1096.36 322.401,1096.36 
  "/>
<polyline clip-path="url(#clip210)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  303.503,830.761 322.401,830.761 
  "/>
<polyline clip-path="url(#clip210)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  303.503,565.16 322.401,565.16 
  "/>
<polyline clip-path="url(#clip210)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  303.503,299.559 322.401,299.559 
  "/>
<path clip-path="url(#clip210)" d="M136.497 1340.66 Q131.08 1340.66 128.337 1346.01 Q125.629 1351.32 125.629 1362.01 Q125.629 1372.67 128.337 1378.02 Q131.08 1383.33 136.497 1383.33 Q141.948 1383.33 144.656 1378.02 Q147.399 1372.67 147.399 1362.01 Q147.399 1351.32 144.656 1346.01 Q141.948 1340.66 136.497 1340.66 M136.497 1335.11 Q145.212 1335.11 149.795 1342.01 Q154.413 1348.89 154.413 1362.01 Q154.413 1375.1 149.795 1382.01 Q145.212 1388.89 136.497 1388.89 Q127.781 1388.89 123.163 1382.01 Q118.58 1375.1 118.58 1362.01 Q118.58 1348.89 123.163 1342.01 Q127.781 1335.11 136.497 1335.11 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M166.74 1379.06 L174.066 1379.06 L174.066 1387.88 L166.74 1387.88 L166.74 1379.06 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M189.413 1336.04 L216.948 1336.04 L216.948 1341.95 L195.837 1341.95 L195.837 1354.65 Q197.364 1354.13 198.892 1353.89 Q200.42 1353.61 201.948 1353.61 Q210.628 1353.61 215.698 1358.37 Q220.767 1363.13 220.767 1371.25 Q220.767 1379.62 215.559 1384.27 Q210.35 1388.89 200.871 1388.89 Q197.608 1388.89 194.205 1388.33 Q190.837 1387.78 187.226 1386.67 L187.226 1379.62 Q190.351 1381.32 193.684 1382.15 Q197.017 1382.99 200.732 1382.99 Q206.739 1382.99 210.246 1379.83 Q213.753 1376.67 213.753 1371.25 Q213.753 1365.83 210.246 1362.67 Q206.739 1359.51 200.732 1359.51 Q197.92 1359.51 195.108 1360.14 Q192.33 1360.76 189.413 1362.08 L189.413 1336.04 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M249.586 1340.66 Q244.17 1340.66 241.427 1346.01 Q238.718 1351.32 238.718 1362.01 Q238.718 1372.67 241.427 1378.02 Q244.17 1383.33 249.586 1383.33 Q255.038 1383.33 257.746 1378.02 Q260.489 1372.67 260.489 1362.01 Q260.489 1351.32 257.746 1346.01 Q255.038 1340.66 249.586 1340.66 M249.586 1335.11 Q258.302 1335.11 262.885 1342.01 Q267.503 1348.89 267.503 1362.01 Q267.503 1375.1 262.885 1382.01 Q258.302 1388.89 249.586 1388.89 Q240.871 1388.89 236.253 1382.01 Q231.67 1375.1 231.67 1362.01 Q231.67 1348.89 236.253 1342.01 Q240.871 1335.11 249.586 1335.11 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M137.99 1075.06 Q132.573 1075.06 129.83 1080.41 Q127.122 1085.72 127.122 1096.41 Q127.122 1107.07 129.83 1112.42 Q132.573 1117.73 137.99 1117.73 Q143.441 1117.73 146.149 1112.42 Q148.892 1107.07 148.892 1096.41 Q148.892 1085.72 146.149 1080.41 Q143.441 1075.06 137.99 1075.06 M137.99 1069.5 Q146.705 1069.5 151.288 1076.41 Q155.906 1083.29 155.906 1096.41 Q155.906 1109.5 151.288 1116.41 Q146.705 1123.29 137.99 1123.29 Q129.275 1123.29 124.657 1116.41 Q120.073 1109.5 120.073 1096.41 Q120.073 1083.29 124.657 1076.41 Q129.275 1069.5 137.99 1069.5 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M168.233 1113.46 L175.559 1113.46 L175.559 1122.28 L168.233 1122.28 L168.233 1113.46 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M189.066 1070.44 L222.399 1070.44 L222.399 1073.43 L203.58 1122.28 L196.253 1122.28 L213.962 1076.34 L189.066 1076.34 L189.066 1070.44 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M236.149 1070.44 L263.684 1070.44 L263.684 1076.34 L242.573 1076.34 L242.573 1089.05 Q244.1 1088.53 245.628 1088.29 Q247.156 1088.01 248.684 1088.01 Q257.364 1088.01 262.434 1092.77 Q267.503 1097.52 267.503 1105.65 Q267.503 1114.02 262.295 1118.67 Q257.086 1123.29 247.607 1123.29 Q244.343 1123.29 240.941 1122.73 Q237.573 1122.18 233.961 1121.07 L233.961 1114.02 Q237.086 1115.72 240.42 1116.55 Q243.753 1117.39 247.468 1117.39 Q253.475 1117.39 256.982 1114.23 Q260.489 1111.07 260.489 1105.65 Q260.489 1100.23 256.982 1097.07 Q253.475 1093.91 247.468 1093.91 Q244.656 1093.91 241.843 1094.54 Q239.066 1095.16 236.149 1096.48 L236.149 1070.44 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M122.712 850.778 L134.17 850.778 L134.17 811.23 L121.705 813.73 L121.705 807.341 L134.101 804.841 L141.115 804.841 L141.115 850.778 L152.573 850.778 L152.573 856.681 L122.712 856.681 L122.712 850.778 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M166.74 847.861 L174.066 847.861 L174.066 856.681 L166.74 856.681 L166.74 847.861 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M204.344 809.459 Q198.927 809.459 196.184 814.806 Q193.476 820.118 193.476 830.813 Q193.476 841.472 196.184 846.82 Q198.927 852.132 204.344 852.132 Q209.795 852.132 212.503 846.82 Q215.246 841.472 215.246 830.813 Q215.246 820.118 212.503 814.806 Q209.795 809.459 204.344 809.459 M204.344 803.903 Q213.059 803.903 217.642 810.813 Q222.26 817.688 222.26 830.813 Q222.26 843.903 217.642 850.813 Q213.059 857.688 204.344 857.688 Q195.628 857.688 191.01 850.813 Q186.427 843.903 186.427 830.813 Q186.427 817.688 191.01 810.813 Q195.628 803.903 204.344 803.903 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M249.586 809.459 Q244.17 809.459 241.427 814.806 Q238.718 820.118 238.718 830.813 Q238.718 841.472 241.427 846.82 Q244.17 852.132 249.586 852.132 Q255.038 852.132 257.746 846.82 Q260.489 841.472 260.489 830.813 Q260.489 820.118 257.746 814.806 Q255.038 809.459 249.586 809.459 M249.586 803.903 Q258.302 803.903 262.885 810.813 Q267.503 817.688 267.503 830.813 Q267.503 843.903 262.885 850.813 Q258.302 857.688 249.586 857.688 Q240.871 857.688 236.253 850.813 Q231.67 843.903 231.67 830.813 Q231.67 817.688 236.253 810.813 Q240.871 803.903 249.586 803.903 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M124.205 585.177 L135.663 585.177 L135.663 545.629 L123.198 548.129 L123.198 541.74 L135.594 539.24 L142.608 539.24 L142.608 585.177 L154.066 585.177 L154.066 591.08 L124.205 591.08 L124.205 585.177 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M168.233 582.26 L175.559 582.26 L175.559 591.08 L168.233 591.08 L168.233 582.26 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M196.878 585.177 L221.357 585.177 L221.357 591.08 L188.441 591.08 L188.441 585.177 Q192.434 581.045 199.309 574.101 Q206.219 567.122 207.989 565.108 Q211.357 561.323 212.677 558.719 Q214.031 556.08 214.031 553.545 Q214.031 549.413 211.114 546.809 Q208.232 544.205 203.58 544.205 Q200.281 544.205 196.601 545.351 Q192.955 546.497 188.788 548.823 L188.788 541.74 Q193.024 540.038 196.705 539.17 Q200.385 538.302 203.441 538.302 Q211.496 538.302 216.288 542.33 Q221.08 546.358 221.08 553.094 Q221.08 556.288 219.864 559.17 Q218.684 562.017 215.524 565.906 Q214.656 566.913 210.003 571.74 Q205.351 576.531 196.878 585.177 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M236.149 539.24 L263.684 539.24 L263.684 545.143 L242.573 545.143 L242.573 557.851 Q244.1 557.33 245.628 557.087 Q247.156 556.809 248.684 556.809 Q257.364 556.809 262.434 561.566 Q267.503 566.323 267.503 574.448 Q267.503 582.816 262.295 587.469 Q257.086 592.087 247.607 592.087 Q244.343 592.087 240.941 591.531 Q237.573 590.976 233.961 589.865 L233.961 582.816 Q237.086 584.517 240.42 585.351 Q243.753 586.184 247.468 586.184 Q253.475 586.184 256.982 583.024 Q260.489 579.865 260.489 574.448 Q260.489 569.031 256.982 565.872 Q253.475 562.712 247.468 562.712 Q244.656 562.712 241.843 563.337 Q239.066 563.962 236.149 565.281 L236.149 539.24 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M122.712 319.576 L134.17 319.576 L134.17 280.028 L121.705 282.528 L121.705 276.139 L134.101 273.639 L141.115 273.639 L141.115 319.576 L152.573 319.576 L152.573 325.479 L122.712 325.479 L122.712 319.576 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M166.74 316.659 L174.066 316.659 L174.066 325.479 L166.74 325.479 L166.74 316.659 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M189.413 273.639 L216.948 273.639 L216.948 279.542 L195.837 279.542 L195.837 292.25 Q197.364 291.729 198.892 291.486 Q200.42 291.208 201.948 291.208 Q210.628 291.208 215.698 295.965 Q220.767 300.722 220.767 308.847 Q220.767 317.215 215.559 321.868 Q210.35 326.486 200.871 326.486 Q197.608 326.486 194.205 325.93 Q190.837 325.375 187.226 324.264 L187.226 317.215 Q190.351 318.916 193.684 319.75 Q197.017 320.583 200.732 320.583 Q206.739 320.583 210.246 317.423 Q213.753 314.264 213.753 308.847 Q213.753 303.43 210.246 300.271 Q206.739 297.111 200.732 297.111 Q197.92 297.111 195.108 297.736 Q192.33 298.361 189.413 299.68 L189.413 273.639 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M249.586 278.257 Q244.17 278.257 241.427 283.604 Q238.718 288.917 238.718 299.611 Q238.718 310.271 241.427 315.618 Q244.17 320.93 249.586 320.93 Q255.038 320.93 257.746 315.618 Q260.489 310.271 260.489 299.611 Q260.489 288.917 257.746 283.604 Q255.038 278.257 249.586 278.257 M249.586 272.701 Q258.302 272.701 262.885 279.611 Q267.503 286.486 267.503 299.611 Q267.503 312.701 262.885 319.611 Q258.302 326.486 249.586 326.486 Q240.871 326.486 236.253 319.611 Q231.67 312.701 231.67 299.611 Q231.67 286.486 236.253 279.611 Q240.871 272.701 249.586 272.701 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M35.6047 1006.89 Q31.2992 1004.5 29.2506 1001.16 Q27.202 997.831 27.202 993.317 Q27.202 987.24 31.4728 983.942 Q35.7089 980.643 43.5561 980.643 L67.0282 980.643 L67.0282 987.067 L43.7644 987.067 Q38.1741 987.067 35.4658 989.046 Q32.7575 991.025 32.7575 995.088 Q32.7575 1000.05 36.0561 1002.93 Q39.3547 1005.82 45.0491 1005.82 L67.0282 1005.82 L67.0282 1012.24 L43.7644 1012.24 Q38.1394 1012.24 35.4658 1014.22 Q32.7575 1016.2 32.7575 1020.33 Q32.7575 1025.23 36.0908 1028.11 Q39.3894 1030.99 45.0491 1030.99 L67.0282 1030.99 L67.0282 1037.41 L28.1395 1037.41 L28.1395 1030.99 L34.1811 1030.99 Q30.6047 1028.8 28.9034 1025.75 Q27.202 1022.69 27.202 1018.49 Q27.202 1014.25 29.3547 1011.3 Q31.5075 1008.32 35.6047 1006.89 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M47.4797 950.227 Q47.4797 957.97 49.2505 960.956 Q51.0213 963.942 55.2921 963.942 Q58.6949 963.942 60.7087 961.72 Q62.6879 959.463 62.6879 955.609 Q62.6879 950.296 58.9379 947.102 Q55.1532 943.873 48.9033 943.873 L47.4797 943.873 L47.4797 950.227 M44.8408 937.484 L67.0282 937.484 L67.0282 943.873 L61.1254 943.873 Q64.6671 946.06 66.3684 949.324 Q68.0351 952.588 68.0351 957.31 Q68.0351 963.282 64.7018 966.824 Q61.3337 970.331 55.7088 970.331 Q49.1463 970.331 45.813 965.956 Q42.4797 961.546 42.4797 952.831 L42.4797 943.873 L41.8547 943.873 Q37.445 943.873 35.0492 946.789 Q32.6186 949.671 32.6186 954.914 Q32.6186 958.248 33.4172 961.407 Q34.2158 964.567 35.813 967.484 L29.9103 967.484 Q28.5561 963.977 27.8964 960.678 Q27.202 957.379 27.202 954.254 Q27.202 945.817 31.577 941.65 Q35.9519 937.484 44.8408 937.484 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M29.2853 899.533 L35.3269 899.533 Q33.9381 902.241 33.2436 905.158 Q32.5492 908.074 32.5492 911.199 Q32.5492 915.956 34.0075 918.352 Q35.4658 920.713 38.3825 920.713 Q40.6047 920.713 41.8894 919.012 Q43.1394 917.31 44.2852 912.171 L44.7713 909.984 Q46.2297 903.178 48.9033 900.331 Q51.5421 897.449 56.299 897.449 Q61.7157 897.449 64.8754 901.755 Q68.0351 906.026 68.0351 913.526 Q68.0351 916.651 67.4101 920.053 Q66.8198 923.421 65.6046 927.171 L59.0074 927.171 Q60.8476 923.63 61.7851 920.192 Q62.6879 916.755 62.6879 913.387 Q62.6879 908.873 61.1601 906.442 Q59.5976 904.012 56.7852 904.012 Q54.181 904.012 52.7921 905.783 Q51.4032 907.519 50.1185 913.456 L49.5977 915.678 Q48.3477 921.616 45.7783 924.255 Q43.1741 926.894 38.6603 926.894 Q33.1742 926.894 30.1881 923.005 Q27.202 919.116 27.202 911.963 Q27.202 908.421 27.7228 905.296 Q28.2436 902.171 29.2853 899.533 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M29.2853 862.484 L35.3269 862.484 Q33.9381 865.192 33.2436 868.109 Q32.5492 871.026 32.5492 874.151 Q32.5492 878.908 34.0075 881.303 Q35.4658 883.665 38.3825 883.665 Q40.6047 883.665 41.8894 881.963 Q43.1394 880.262 44.2852 875.123 L44.7713 872.935 Q46.2297 866.13 48.9033 863.283 Q51.5421 860.401 56.299 860.401 Q61.7157 860.401 64.8754 864.706 Q68.0351 868.977 68.0351 876.477 Q68.0351 879.602 67.4101 883.005 Q66.8198 886.373 65.6046 890.123 L59.0074 890.123 Q60.8476 886.581 61.7851 883.144 Q62.6879 879.706 62.6879 876.338 Q62.6879 871.824 61.1601 869.394 Q59.5976 866.963 56.7852 866.963 Q54.181 866.963 52.7921 868.734 Q51.4032 870.47 50.1185 876.408 L49.5977 878.63 Q48.3477 884.567 45.7783 887.206 Q43.1741 889.845 38.6603 889.845 Q33.1742 889.845 30.1881 885.956 Q27.202 882.067 27.202 874.915 Q27.202 871.373 27.7228 868.248 Q28.2436 865.123 29.2853 862.484 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M61.1949 821.443 L81.8197 821.443 L81.8197 827.866 L28.1395 827.866 L28.1395 821.443 L34.0422 821.443 Q30.57 819.429 28.9034 816.373 Q27.202 813.283 27.202 809.012 Q27.202 801.929 32.827 797.519 Q38.4519 793.075 47.6185 793.075 Q56.7852 793.075 62.4101 797.519 Q68.0351 801.929 68.0351 809.012 Q68.0351 813.283 66.3684 816.373 Q64.6671 819.429 61.1949 821.443 M47.6185 799.707 Q40.57 799.707 36.5769 802.623 Q32.5492 805.505 32.5492 810.575 Q32.5492 815.644 36.5769 818.561 Q40.57 821.443 47.6185 821.443 Q54.6671 821.443 58.6949 818.561 Q62.6879 815.644 62.6879 810.575 Q62.6879 805.505 58.6949 802.623 Q54.6671 799.707 47.6185 799.707 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M32.6186 767.415 Q32.6186 772.554 36.6464 775.54 Q40.6394 778.526 47.6185 778.526 Q54.5977 778.526 58.6254 775.575 Q62.6185 772.589 62.6185 767.415 Q62.6185 762.311 58.5907 759.325 Q54.5629 756.339 47.6185 756.339 Q40.7089 756.339 36.6811 759.325 Q32.6186 762.311 32.6186 767.415 M27.202 767.415 Q27.202 759.082 32.6186 754.325 Q38.0353 749.568 47.6185 749.568 Q57.1671 749.568 62.6185 754.325 Q68.0351 759.082 68.0351 767.415 Q68.0351 775.783 62.6185 780.54 Q57.1671 785.262 47.6185 785.262 Q38.0353 785.262 32.6186 780.54 Q27.202 775.783 27.202 767.415 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M29.2853 714.186 L35.3269 714.186 Q33.9381 716.895 33.2436 719.811 Q32.5492 722.728 32.5492 725.853 Q32.5492 730.61 34.0075 733.006 Q35.4658 735.367 38.3825 735.367 Q40.6047 735.367 41.8894 733.665 Q43.1394 731.964 44.2852 726.825 L44.7713 724.638 Q46.2297 717.832 48.9033 714.985 Q51.5421 712.103 56.299 712.103 Q61.7157 712.103 64.8754 716.409 Q68.0351 720.679 68.0351 728.179 Q68.0351 731.304 67.4101 734.707 Q66.8198 738.075 65.6046 741.825 L59.0074 741.825 Q60.8476 738.283 61.7851 734.846 Q62.6879 731.408 62.6879 728.04 Q62.6879 723.527 61.1601 721.096 Q59.5976 718.665 56.7852 718.665 Q54.181 718.665 52.7921 720.436 Q51.4032 722.172 50.1185 728.11 L49.5977 730.332 Q48.3477 736.27 45.7783 738.908 Q43.1741 741.547 38.6603 741.547 Q33.1742 741.547 30.1881 737.658 Q27.202 733.77 27.202 726.617 Q27.202 723.075 27.7228 719.95 Q28.2436 716.825 29.2853 714.186 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M28.1395 701.929 L28.1395 695.541 L67.0282 695.541 L67.0282 701.929 L28.1395 701.929 M13.0007 701.929 L13.0007 695.541 L21.0909 695.541 L21.0909 701.929 L13.0007 701.929 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M17.0979 675.853 L28.1395 675.853 L28.1395 662.694 L33.1047 662.694 L33.1047 675.853 L54.2157 675.853 Q58.9726 675.853 60.3268 674.568 Q61.681 673.249 61.681 669.256 L61.681 662.694 L67.0282 662.694 L67.0282 669.256 Q67.0282 676.652 64.2851 679.464 Q61.5074 682.277 54.2157 682.277 L33.1047 682.277 L33.1047 686.964 L28.1395 686.964 L28.1395 682.277 L17.0979 682.277 L17.0979 675.853 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M28.1395 654.291 L28.1395 647.902 L67.0282 647.902 L67.0282 654.291 L28.1395 654.291 M13.0007 654.291 L13.0007 647.902 L21.0909 647.902 L21.0909 654.291 L13.0007 654.291 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M32.6186 619.465 Q32.6186 624.603 36.6464 627.59 Q40.6394 630.576 47.6185 630.576 Q54.5977 630.576 58.6254 627.624 Q62.6185 624.638 62.6185 619.465 Q62.6185 614.36 58.5907 611.374 Q54.5629 608.388 47.6185 608.388 Q40.7089 608.388 36.6811 611.374 Q32.6186 614.36 32.6186 619.465 M27.202 619.465 Q27.202 611.131 32.6186 606.374 Q38.0353 601.617 47.6185 601.617 Q57.1671 601.617 62.6185 606.374 Q68.0351 611.131 68.0351 619.465 Q68.0351 627.833 62.6185 632.59 Q57.1671 637.312 47.6185 637.312 Q38.0353 637.312 32.6186 632.59 Q27.202 627.833 27.202 619.465 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M43.5561 558.701 L67.0282 558.701 L67.0282 565.09 L43.7644 565.09 Q38.2436 565.09 35.5005 567.243 Q32.7575 569.395 32.7575 573.701 Q32.7575 578.875 36.0561 581.861 Q39.3547 584.847 45.0491 584.847 L67.0282 584.847 L67.0282 591.27 L28.1395 591.27 L28.1395 584.847 L34.1811 584.847 Q30.6742 582.555 28.9381 579.465 Q27.202 576.34 27.202 572.277 Q27.202 565.576 31.3686 562.139 Q35.5005 558.701 43.5561 558.701 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M13.0007 523.944 L13.0007 509.222 L17.9659 509.222 L17.9659 517.555 L71.4379 517.555 L71.4379 509.222 L76.4031 509.222 L76.4031 523.944 L13.0007 523.944 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M35.6047 465.333 Q31.2992 462.938 29.2506 459.604 Q27.202 456.271 27.202 451.757 Q27.202 445.681 31.4728 442.382 Q35.7089 439.084 43.5561 439.084 L67.0282 439.084 L67.0282 445.507 L43.7644 445.507 Q38.1741 445.507 35.4658 447.486 Q32.7575 449.466 32.7575 453.528 Q32.7575 458.493 36.0561 461.375 Q39.3547 464.257 45.0491 464.257 L67.0282 464.257 L67.0282 470.681 L43.7644 470.681 Q38.1394 470.681 35.4658 472.66 Q32.7575 474.639 32.7575 478.771 Q32.7575 483.667 36.0908 486.549 Q39.3894 489.431 45.0491 489.431 L67.0282 489.431 L67.0282 495.854 L28.1395 495.854 L28.1395 489.431 L34.1811 489.431 Q30.6047 487.243 28.9034 484.188 Q27.202 481.132 27.202 476.931 Q27.202 472.695 29.3547 469.743 Q31.5075 466.757 35.6047 465.333 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M13.0007 411.41 L76.4031 411.41 L76.4031 426.132 L71.4379 426.132 L71.4379 417.834 L17.9659 417.834 L17.9659 426.132 L13.0007 426.132 L13.0007 411.41 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip212)" style="stroke:#009af9; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" points="
  361.501,1361.96 365.367,1361.63 369.234,1360.67 373.1,1359.08 376.967,1356.85 380.833,1353.99 384.7,1350.51 388.566,1346.39 392.433,1341.66 396.299,1336.31 
  400.166,1330.34 404.032,1323.77 407.899,1316.6 411.765,1308.85 415.632,1300.5 419.498,1291.59 423.365,1282.1 427.231,1272.06 431.098,1261.48 434.964,1250.35 
  438.831,1238.7 442.698,1226.54 446.564,1213.87 450.431,1200.71 454.297,1187.08 458.164,1172.99 462.03,1158.44 465.897,1143.47 469.763,1128.08 473.63,1112.28 
  477.496,1096.1 481.363,1079.55 485.229,1062.63 489.096,1045.38 492.962,1027.81 496.829,1009.94 500.695,991.785 504.562,973.358 508.428,954.678 512.295,935.763 
  516.161,916.636 520.028,897.318 523.894,877.825 527.761,858.176 531.627,838.392 535.494,818.493 539.36,798.499 543.227,778.426 547.093,758.297 550.96,738.133 
  554.826,717.956 558.693,697.786 562.559,677.644 566.426,657.548 570.293,637.513 574.159,617.565 578.026,597.724 581.892,578.011 585.759,558.448 589.625,539.055 
  593.492,519.839 597.358,500.828 601.225,482.04 605.091,463.496 608.958,445.216 612.824,427.217 616.691,409.516 620.557,392.125 624.424,375.067 628.29,358.359 
  632.157,342.018 636.023,326.062 639.89,310.507 643.756,295.363 647.623,280.646 651.489,266.373 655.356,252.559 659.222,239.218 663.089,226.365 666.955,214.011 
  670.822,202.163 674.688,190.837 678.555,180.046 682.421,169.8 686.288,160.111 690.154,150.988 694.021,142.437 697.887,134.466 701.754,127.085 705.621,120.303 
  709.487,114.126 713.354,108.561 717.22,103.613 721.087,99.2832 724.953,95.579 728.82,92.5042 732.686,90.0622 736.553,88.2558 740.419,87.087 744.286,86.5554 
  748.152,86.6609 752.019,87.4047 755.885,88.7862 759.752,90.8041 763.618,93.4565 767.485,96.7409 771.351,100.653 775.218,105.189 779.084,110.344 782.951,116.115 
  786.817,122.494 790.684,129.476 794.55,137.053 798.417,145.218 802.283,153.963 806.15,163.279 810.016,173.157 813.883,183.586 817.749,194.555 821.616,206.056 
  825.482,218.077 829.349,230.604 833.216,243.626 837.082,257.128 840.949,271.096 844.815,285.519 848.682,300.383 852.548,315.672 856.415,331.37 860.281,347.461 
  864.148,363.928 868.014,380.754 871.881,397.927 875.747,415.429 879.614,433.239 883.48,451.341 887.347,469.715 891.213,488.343 895.08,507.206 898.946,526.291 
  902.813,545.575 906.679,565.038 910.546,584.659 914.412,604.419 918.279,624.297 922.145,644.279 926.012,664.343 929.878,684.468 933.745,704.632 937.611,724.815 
  941.478,744.995 945.344,765.154 949.211,785.278 953.077,805.341 956.944,825.324 960.811,845.205 964.677,864.963 968.544,884.578 972.41,904.036 976.277,923.318 
  980.143,942.401 984.01,961.265 987.876,979.892 991.743,998.26 995.609,1016.35 999.476,1034.16 1003.34,1051.66 1007.21,1068.83 1011.08,1085.65 1014.94,1102.11 
  1018.81,1118.2 1022.67,1133.88 1026.54,1149.17 1030.41,1164.03 1034.27,1178.45 1038.14,1192.41 1042.01,1205.9 1045.87,1218.91 1049.74,1231.43 1053.61,1243.45 
  1057.47,1254.94 1061.34,1265.9 1065.21,1276.32 1069.07,1286.19 1072.94,1295.49 1076.81,1304.23 1080.67,1312.39 1084.54,1319.95 1088.41,1326.93 1092.27,1333.29 
  1096.14,1339.05 1100.01,1344.2 1103.87,1348.72 1107.74,1352.62 1111.6,1355.89 1115.47,1358.53 1119.34,1360.54 1123.2,1361.91 1127.07,1362.64 1130.94,1362.73 
  1134.8,1362.19 1138.67,1361 1142.54,1359.18 1146.4,1356.73 1150.27,1353.64 1154.14,1349.92 1158,1345.58 1161.87,1340.61 1165.74,1335.03 1169.6,1328.84 
  1173.47,1322.04 1177.34,1314.64 1181.2,1306.66 1185.07,1298.09 1188.93,1288.95 1192.8,1279.24 1196.67,1268.98 1200.53,1258.17 1204.4,1246.83 1208.27,1234.96 
  1212.13,1222.59 1216,1209.71 1219.87,1196.36 1223.73,1182.53 1227.6,1168.24 1231.47,1153.5 1235.33,1138.34 1239.2,1122.76 1243.07,1106.78 1246.93,1090.43 
  1250.8,1073.7 1254.67,1056.63 1258.53,1039.22 1262.4,1021.5 1266.27,1003.48 1270.13,985.181 1274,966.626 1277.86,947.823 1281.73,928.796 1285.6,909.565 
  1289.46,890.149 1293.33,870.568 1297.2,850.843 1301.06,830.992 1304.93,811.03 1308.8,790.98 1312.66,770.865 1316.53,750.704 1320.4,730.518 1324.26,710.33 
  1328.13,690.15 1332,670.004 1335.86,649.911 1339.73,629.894 1343.6,609.973 1347.46,590.169 1351.33,570.5 1355.2,550.979 1359.06,531.632 1362.93,512.478 
  1366.79,493.537 1370.66,474.829 1374.53,456.374 1378.39,438.183 1382.26,420.276 1386.13,402.673 1389.99,385.393 1393.86,368.453 1397.73,351.872 1401.59,335.665 
  1405.46,319.841 1409.33,304.422 1413.19,289.424 1417.06,274.862 1420.93,260.751 1424.79,247.106 1428.66,233.938 1432.53,221.256 1436.39,209.079 1440.26,197.417 
  1444.12,186.284 1447.99,175.691 1451.86,165.65 1455.72,156.162 1459.59,147.243 1463.46,138.902 1467.32,131.147 1471.19,123.987 1475.06,117.429 1478.92,111.479 
  1482.79,106.139 1486.66,101.418 1490.52,97.3201 1494.39,93.8508 1498.26,91.0134 1502.12,88.8112 1505.99,87.2427 1509.86,86.3116 1513.72,86.0191 1517.59,86.3658 
  1521.46,87.3514 1525.32,88.9751 1529.19,91.2349 1533.05,94.1275 1536.92,97.6508 1540.79,101.802 1544.65,106.575 1548.52,111.968 1552.39,117.973 1556.25,124.585 
  1560.12,131.798 1563.99,139.604 1567.85,147.996 1571.72,156.964 1575.59,166.499 1579.45,176.593 1583.32,187.236 1587.19,198.417 1591.05,210.125 1594.92,222.347 
  1598.79,235.071 1602.65,248.284 1606.52,261.974 1610.39,276.129 1614.25,290.733 1618.12,305.77 1621.98,321.226 1625.85,337.084 1629.72,353.329 1633.58,369.949 
  1637.45,386.925 1641.32,404.238 1645.18,421.871 1649.05,439.805 1652.92,458.023 1656.78,476.509 1660.65,495.246 1664.52,514.213 1668.38,533.39 1672.25,552.757 
  1676.12,572.294 1679.98,591.982 1683.85,611.808 1687.72,631.747 1691.58,651.779 1695.45,671.883 1699.31,692.039 1703.18,712.224 1707.05,732.423 1710.91,752.619 
  1714.78,772.786 1718.65,792.905 1722.51,812.954 1726.38,832.913 1730.25,852.761 1734.11,872.486 1737.98,892.065 1741.85,911.476 1745.71,930.699 1749.58,949.714 
  1753.45,968.502 1757.31,987.045 1761.18,1005.33 1765.05,1023.34 1768.91,1041.04 1772.78,1058.43 1776.65,1075.48 1780.51,1092.18 1784.38,1108.52 1788.24,1124.47 
  1792.11,1140.02 1795.98,1155.16 1799.84,1169.86 1803.71,1184.12 1807.58,1197.92 1811.44,1211.24 1815.31,1224.08 1819.18,1236.42 1823.04,1248.25 1826.91,1259.55 
  1830.78,1270.32 1834.64,1280.54 1838.51,1290.21 1842.38,1299.31 1846.24,1307.83 1850.11,1315.77 1853.98,1323.12 1857.84,1329.87 1861.71,1336.01 1865.58,1341.55 
  1869.44,1346.46 1873.31,1350.76 1877.17,1354.42 1881.04,1357.46 1884.91,1359.86 1888.77,1361.63 1892.64,1362.76 1896.51,1363.25 1900.37,1363.1 1904.24,1362.31 
  1908.11,1360.89 1911.97,1358.83 1915.84,1356.13 1919.71,1352.8 1923.57,1348.84 1927.44,1344.26 1931.31,1339.06 1935.17,1333.24 1939.04,1326.82 1942.91,1319.79 
  1946.77,1312.16 1950.64,1303.95 1954.5,1295.16 1958.37,1285.79 1962.24,1275.87 1966.1,1265.39 1969.97,1254.37 1973.84,1242.82 1977.7,1230.75 1981.57,1218.18 
  1985.44,1205.11 1989.3,1191.56 1993.17,1177.55 1997.04,1163.08 2000.9,1148.17 2004.77,1132.83 2008.64,1117.1 2012.5,1100.96 2016.37,1084.45 2020.24,1067.57 
  2024.1,1050.36 2027.97,1032.82 2031.84,1014.97 2035.7,996.832 2039.57,978.417 2043.43,959.748 2047.3,940.843 2051.17,921.723 2055.03,902.407 2058.9,882.916 
  2062.77,863.262 2066.63,843.467 2070.5,823.554 2074.37,803.541 2078.23,783.451 2082.1,763.303 2085.97,743.119 2089.83,722.909 2093.7,702.701 2097.57,682.515 
  2101.43,662.372 2105.3,642.293 2109.17,622.298 2113.03,602.404 2116.9,582.628 2120.77,562.993 2124.63,543.521 2128.5,524.231 2132.36,505.144 2136.23,486.278 
  2140.1,467.644 2143.96,449.267 2147.83,431.165 2151.7,413.358 2155.56,395.864 2159.43,378.7 2163.3,361.882 2167.16,345.422 2171.03,329.342 2174.9,313.657 
  2178.76,298.384 2182.63,283.54 2186.5,269.139 2190.36,255.189 2194.23,241.707 2198.1,228.708 2201.96,216.206 2205.83,204.214 2209.69,192.745 2213.56,181.808 
  2217.43,171.409 2221.29,161.564 2225.16,152.283 2229.03,143.575 2232.89,135.449 2236.76,127.915 2240.63,120.974 2244.49,114.637 2248.36,108.91 2252.23,103.799 
  2256.09,99.3096 2259.96,95.4472 2263.83,92.2148 2267.69,89.6125 2271.56,87.6455 2275.43,86.3159 2279.29,85.6252 2283.16,85.5744 2287.03,86.1637 2290.89,87.3909 
  2294.76,89.2551 
  "/>
<polyline clip-path="url(#clip212)" style="stroke:#e26f46; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" points="
  361.501,1361.96 365.367,1361.73 369.234,1361.05 373.1,1359.91 376.967,1358.31 380.833,1356.25 384.7,1353.72 388.566,1350.73 392.433,1347.28 396.299,1343.37 
  400.166,1338.99 404.032,1334.17 407.899,1328.88 411.765,1323.15 415.632,1316.97 419.498,1310.35 423.365,1303.3 427.231,1295.82 431.098,1287.91 434.964,1279.6 
  438.831,1270.87 442.698,1261.74 446.564,1252.23 450.431,1242.33 454.297,1232.06 458.164,1221.43 462.03,1210.44 465.897,1199.12 469.763,1187.46 473.63,1175.49 
  477.496,1163.2 481.363,1150.63 485.229,1137.77 489.096,1124.64 492.962,1111.25 496.829,1097.61 500.695,1083.75 504.562,1069.67 508.428,1055.39 512.295,1040.93 
  516.161,1026.29 520.028,1011.48 523.894,996.536 527.761,981.46 531.627,966.272 535.494,950.989 539.36,935.627 543.227,920.193 547.093,904.709 550.96,889.189 
  554.826,873.649 558.693,858.104 562.559,842.569 566.426,827.06 570.293,811.592 574.159,796.181 578.026,780.841 581.892,765.588 585.759,750.438 589.625,735.405 
  593.492,720.505 597.358,705.752 601.225,691.161 605.091,676.746 608.958,662.523 612.824,648.507 616.691,634.708 620.557,621.144 624.424,607.826 628.29,594.77 
  632.157,581.987 636.023,569.491 639.89,557.293 643.756,545.408 647.623,533.847 651.489,522.621 655.356,511.741 659.222,501.22 663.089,491.068 666.955,481.296 
  670.822,471.913 674.688,462.929 678.555,454.355 682.421,446.199 686.288,438.47 690.154,431.177 694.021,424.328 697.887,417.931 701.754,411.993 705.621,406.521 
  709.487,401.523 713.354,397.005 717.22,392.973 721.087,389.434 724.953,386.393 728.82,383.856 732.686,381.828 736.553,380.314 740.419,379.32 744.286,378.851 
  748.152,378.883 752.019,379.301 755.885,380.094 759.752,381.268 763.618,382.822 767.485,384.757 771.351,387.076 775.218,389.776 779.084,392.859 782.951,396.323 
  786.817,400.165 790.684,404.384 794.55,408.978 798.417,413.942 802.283,419.274 806.15,424.969 810.016,431.022 813.883,437.43 817.749,444.185 821.616,451.283 
  825.482,458.717 829.349,466.48 833.216,474.565 837.082,482.963 840.949,491.669 844.815,500.672 848.682,509.964 852.548,519.535 856.415,529.379 860.281,539.485 
  864.148,549.844 868.014,560.445 871.881,571.278 875.747,582.332 879.614,593.596 883.48,605.059 887.347,616.71 891.213,628.537 895.08,640.527 898.946,652.669 
  902.813,664.952 906.679,677.362 910.546,689.888 914.412,702.517 918.279,715.236 922.145,728.033 926.012,740.895 929.878,753.81 933.745,766.764 937.611,779.744 
  941.478,792.738 945.344,805.733 949.211,818.715 953.077,831.673 956.944,844.591 960.811,857.456 964.677,870.259 968.544,882.985 972.41,895.621 976.277,908.154 
  980.143,920.571 984.01,932.86 987.876,945.011 991.743,957.008 995.609,968.841 999.476,980.497 1003.34,991.965 1007.21,1003.23 1011.08,1014.29 1014.94,1025.13 
  1018.81,1035.73 1022.67,1046.08 1026.54,1056.19 1030.41,1066.02 1034.27,1075.59 1038.14,1084.86 1042.01,1093.84 1045.87,1102.52 1049.74,1110.89 1053.61,1118.93 
  1057.47,1126.64 1061.34,1134.01 1065.21,1141.04 1069.07,1147.71 1072.94,1154.01 1076.81,1159.95 1080.67,1165.51 1084.54,1170.69 1088.41,1175.47 1092.27,1179.86 
  1096.14,1183.85 1100.01,1187.43 1103.87,1190.6 1107.74,1193.35 1111.6,1195.68 1115.47,1197.57 1119.34,1199.04 1123.2,1200.06 1127.07,1200.65 1130.94,1200.79 
  1134.8,1200.59 1138.67,1200.09 1142.54,1199.3 1146.4,1198.22 1150.27,1196.83 1154.14,1195.15 1158,1193.17 1161.87,1190.89 1165.74,1188.31 1169.6,1185.43 
  1173.47,1182.26 1177.34,1178.79 1181.2,1175.03 1185.07,1170.98 1188.93,1166.64 1192.8,1162.02 1196.67,1157.11 1200.53,1151.94 1204.4,1146.49 1208.27,1140.77 
  1212.13,1134.79 1216,1128.55 1219.87,1122.06 1223.73,1115.32 1227.6,1108.34 1231.47,1101.13 1235.33,1093.68 1239.2,1086.02 1243.07,1078.15 1246.93,1070.07 
  1250.8,1061.79 1254.67,1053.33 1258.53,1044.68 1262.4,1035.86 1266.27,1026.88 1270.13,1017.74 1274,1008.46 1277.86,999.035 1281.73,989.484 1285.6,979.815 
  1289.46,970.036 1293.33,960.158 1297.2,950.188 1301.06,940.138 1304.93,930.017 1308.8,919.837 1312.66,909.608 1316.53,899.341 1320.4,889.043 1324.26,878.724 
  1328.13,868.397 1332,858.071 1335.86,847.758 1339.73,837.469 1343.6,827.214 1347.46,817 1351.33,806.839 1355.2,796.743 1359.06,786.72 1362.93,776.782 
  1366.79,766.938 1370.66,757.198 1374.53,747.572 1378.39,738.07 1382.26,728.701 1386.13,719.475 1389.99,710.401 1393.86,701.488 1397.73,692.746 1401.59,684.183 
  1405.46,675.809 1409.33,667.632 1413.19,659.66 1417.06,651.902 1420.93,644.365 1424.79,637.057 1428.66,629.988 1432.53,623.164 1436.39,616.592 1440.26,610.28 
  1444.12,604.235 1447.99,598.463 1451.86,592.971 1455.72,587.765 1459.59,582.853 1463.46,578.239 1467.32,573.929 1471.19,569.929 1475.06,566.243 1478.92,562.879 
  1482.79,559.84 1486.66,557.131 1490.52,554.756 1494.39,552.72 1498.26,551.027 1502.12,549.68 1505.99,548.684 1509.86,548.042 1513.72,547.759 1517.59,547.8 
  1521.46,548.05 1525.32,548.505 1529.19,549.168 1533.05,550.039 1536.92,551.119 1540.79,552.409 1544.65,553.909 1548.52,555.619 1552.39,557.539 1556.25,559.668 
  1560.12,562.006 1563.99,564.551 1567.85,567.301 1571.72,570.256 1575.59,573.413 1579.45,576.77 1583.32,580.325 1587.19,584.076 1591.05,588.018 1594.92,592.15 
  1598.79,596.469 1602.65,600.968 1606.52,605.645 1610.39,610.497 1614.25,615.52 1618.12,620.711 1621.98,626.062 1625.85,631.57 1629.72,637.23 1633.58,643.036 
  1637.45,648.984 1641.32,655.067 1645.18,661.279 1649.05,667.615 1652.92,674.068 1656.78,680.633 1660.65,687.302 1664.52,694.07 1668.38,700.93 1672.25,707.875 
  1676.12,714.899 1679.98,721.995 1683.85,729.155 1687.72,736.373 1691.58,743.642 1695.45,750.954 1699.31,758.303 1703.18,765.681 1707.05,773.08 1710.91,780.494 
  1714.78,787.914 1718.65,795.334 1722.51,802.746 1726.38,810.143 1730.25,817.517 1734.11,824.86 1737.98,832.165 1741.85,839.426 1745.71,846.633 1749.58,853.781 
  1753.45,860.862 1757.31,867.868 1761.18,874.793 1765.05,881.629 1768.91,888.369 1772.78,895.007 1776.65,901.535 1780.51,907.947 1784.38,914.235 1788.24,920.394 
  1792.11,926.417 1795.98,932.298 1799.84,938.029 1803.71,943.606 1807.58,949.022 1811.44,954.271 1815.31,959.347 1819.18,964.244 1823.04,968.958 1826.91,973.483 
  1830.78,977.813 1834.64,981.943 1838.51,985.868 1842.38,989.584 1846.24,993.086 1850.11,996.37 1853.98,999.43 1857.84,1002.26 1861.71,1004.86 1865.58,1007.23 
  1869.44,1009.36 1873.31,1011.24 1877.17,1012.88 1881.04,1014.27 1884.91,1015.41 1888.77,1016.29 1892.64,1016.91 1896.51,1017.28 1900.37,1017.37 1904.24,1017.3 
  1908.11,1017.12 1911.97,1016.84 1915.84,1016.44 1919.71,1015.94 1923.57,1015.32 1927.44,1014.6 1931.31,1013.76 1935.17,1012.82 1939.04,1011.76 1942.91,1010.59 
  1946.77,1009.31 1950.64,1007.92 1954.5,1006.42 1958.37,1004.81 1962.24,1003.1 1966.1,1001.28 1969.97,999.352 1973.84,997.32 1977.7,995.187 1981.57,992.952 
  1985.44,990.617 1989.3,988.182 1993.17,985.653 1997.04,983.029 2000.9,980.313 2004.77,977.506 2008.64,974.612 2012.5,971.633 2016.37,968.57 2020.24,965.427 
  2024.1,962.208 2027.97,958.913 2031.84,955.547 2035.7,952.112 2039.57,948.612 2043.43,945.05 2047.3,941.428 2051.17,937.75 2055.03,934.021 2058.9,930.243 
  2062.77,926.419 2066.63,922.555 2070.5,918.652 2074.37,914.716 2078.23,910.748 2082.1,906.755 2085.97,902.739 2089.83,898.705 2093.7,894.655 2097.57,890.595 
  2101.43,886.529 2105.3,882.46 2109.17,878.392 2113.03,874.33 2116.9,870.277 2120.77,866.239 2124.63,862.218 2128.5,858.219 2132.36,854.246 2136.23,850.303 
  2140.1,846.395 2143.96,842.525 2147.83,838.697 2151.7,834.916 2155.56,831.185 2159.43,827.509 2163.3,823.891 2167.16,820.335 2171.03,816.846 2174.9,813.426 
  2178.76,810.08 2182.63,806.812 2186.5,803.624 2190.36,800.521 2194.23,797.507 2198.1,794.584 2201.96,791.756 2205.83,789.027 2209.69,786.399 2213.56,783.876 
  2217.43,781.462 2221.29,779.158 2225.16,776.968 2229.03,774.896 2232.89,772.943 2236.76,771.113 2240.63,769.408 2244.49,767.83 2248.36,766.383 2252.23,765.068 
  2256.09,763.889 2259.96,762.846 2263.83,761.942 2267.69,761.18 2271.56,760.56 2275.43,760.086 2279.29,759.758 2283.16,759.58 2287.03,759.544 2290.89,759.544 
  2294.76,759.544 
  "/>
<polyline clip-path="url(#clip212)" style="stroke:#3da44d; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" points="
  361.501,1361.96 365.367,1356.66 369.234,1351.16 373.1,1345.45 376.967,1339.54 380.833,1333.42 384.7,1327.1 388.566,1320.58 392.433,1313.86 396.299,1306.93 
  400.166,1299.8 404.032,1292.47 407.899,1284.94 411.765,1277.21 415.632,1269.28 419.498,1261.15 423.365,1252.81 427.231,1244.28 431.098,1235.55 434.964,1226.61 
  438.831,1217.48 442.698,1208.15 446.564,1198.62 450.431,1188.89 454.297,1178.97 458.164,1168.84 462.03,1158.53 465.897,1148.02 469.763,1137.31 473.63,1126.42 
  477.496,1115.33 481.363,1104.05 485.229,1092.59 489.096,1080.94 492.962,1069.1 496.829,1057.07 500.695,1044.86 504.562,1032.47 508.428,1019.89 512.295,1007.13 
  516.161,994.187 520.028,981.065 523.894,967.764 527.761,954.298 531.627,940.665 535.494,926.857 539.36,912.866 543.227,898.691 547.093,884.33 550.96,869.788 
  554.826,855.072 558.693,840.19 562.559,825.156 566.426,809.986 570.293,794.7 574.159,779.32 578.026,763.872 581.892,748.384 585.759,732.89 589.625,717.423 
  593.492,702.024 597.358,686.734 601.225,671.599 605.091,656.673 608.958,642.002 612.824,627.632 616.691,613.602 620.557,599.947 624.424,586.698 628.29,573.88 
  632.157,561.517 636.023,549.626 639.89,538.22 643.756,527.309 647.623,516.896 651.489,506.981 655.356,497.562 659.222,488.629 663.089,480.169 666.955,472.166 
  670.822,464.625 674.688,457.541 678.555,450.905 682.421,444.707 686.288,438.94 690.154,433.595 694.021,428.663 697.887,424.136 701.754,420.007 705.621,416.269 
  709.487,412.914 713.354,409.935 717.22,407.326 721.087,405.081 724.953,403.193 728.82,401.657 732.686,400.467 736.553,399.618 740.419,399.105 744.286,398.924 
  748.152,399.07 752.019,399.538 755.885,400.326 759.752,401.429 763.618,402.845 767.485,404.569 771.351,406.6 775.218,408.934 779.084,411.569 782.951,414.503 
  786.817,417.734 790.684,421.26 794.55,425.079 798.417,429.191 802.283,433.594 806.15,438.287 810.016,443.271 813.883,448.544 817.749,454.106 821.616,459.958 
  825.482,466.1 829.349,472.533 833.216,479.257 837.082,486.274 840.949,493.584 844.815,501.19 848.682,509.093 852.548,517.292 856.415,525.766 860.281,534.526 
  864.148,543.585 868.014,552.955 871.881,562.644 875.747,572.659 879.614,583.003 883.48,593.678 887.347,604.682 891.213,616.011 895.08,627.659 898.946,639.617 
  902.813,651.874 906.679,664.416 910.546,677.226 914.412,690.286 918.279,703.573 922.145,717.065 926.012,730.733 929.878,744.55 933.745,758.483 937.611,772.5 
  941.478,786.562 945.344,800.631 949.211,814.661 953.077,828.601 956.944,842.414 960.811,856.067 964.677,869.531 968.544,882.778 972.41,895.783 976.277,908.522 
  980.143,920.977 984.01,933.129 987.876,944.964 991.743,956.469 995.609,967.634 999.476,978.453 1003.34,988.92 1007.21,999.033 1011.08,1008.79 1014.94,1018.2 
  1018.81,1027.27 1022.67,1035.99 1026.54,1044.39 1030.41,1052.48 1034.27,1060.25 1038.14,1067.7 1042.01,1074.83 1045.87,1081.65 1049.74,1088.16 1053.61,1094.36 
  1057.47,1100.26 1061.34,1105.86 1065.21,1111.17 1069.07,1116.19 1072.94,1120.93 1076.81,1125.38 1080.67,1129.56 1084.54,1133.46 1088.41,1137.09 1092.27,1140.45 
  1096.14,1143.54 1100.01,1146.38 1103.87,1148.96 1107.74,1151.28 1111.6,1153.35 1115.47,1155.18 1119.34,1156.75 1123.2,1158.09 1127.07,1159.18 1130.94,1160.04 
  1134.8,1160.66 1138.67,1161.05 1142.54,1161.21 1146.4,1161.14 1150.27,1160.85 1154.14,1160.34 1158,1159.6 1161.87,1158.64 1165.74,1157.47 1169.6,1156.09 
  1173.47,1154.49 1177.34,1152.68 1181.2,1150.66 1185.07,1148.44 1188.93,1146.02 1192.8,1143.39 1196.67,1140.56 1200.53,1137.54 1204.4,1134.32 1208.27,1130.91 
  1212.13,1127.31 1216,1123.52 1219.87,1119.55 1223.73,1115.39 1227.6,1111.05 1231.47,1106.52 1235.33,1101.83 1239.2,1096.95 1243.07,1091.9 1246.93,1086.68 
  1250.8,1081.3 1254.67,1075.74 1258.53,1070.03 1262.4,1064.15 1266.27,1058.12 1270.13,1051.93 1274,1045.58 1277.86,1039.09 1281.73,1032.45 1285.6,1025.67 
  1289.46,1018.75 1293.33,1011.69 1297.2,1004.5 1301.06,997.184 1304.93,989.739 1308.8,982.173 1312.66,974.488 1316.53,966.69 1320.4,958.783 1324.26,950.771 
  1328.13,942.658 1332,934.45 1335.86,926.15 1339.73,917.764 1343.6,909.296 1347.46,900.751 1351.33,892.135 1355.2,883.453 1359.06,874.709 1362.93,865.907 
  1366.79,857.055 1370.66,848.162 1374.53,839.239 1378.39,830.294 1382.26,821.339 1386.13,812.383 1389.99,803.438 1393.86,794.514 1397.73,785.624 1401.59,776.777 
  1405.46,767.988 1409.33,759.267 1413.19,750.628 1417.06,742.083 1420.93,733.645 1424.79,725.329 1428.66,717.148 1432.53,709.115 1436.39,701.246 1440.26,693.555 
  1444.12,686.057 1447.99,678.767 1451.86,671.701 1455.72,664.875 1459.59,658.306 1463.46,652.005 1467.32,645.974 1471.19,640.216 1475.06,634.738 1478.92,629.541 
  1482.79,624.63 1486.66,620.009 1490.52,615.678 1494.39,611.642 1498.26,607.902 1502.12,604.46 1505.99,601.317 1509.86,598.475 1513.72,595.933 1517.59,593.691 
  1521.46,591.751 1525.32,590.111 1529.19,588.771 1533.05,587.728 1536.92,586.983 1540.79,586.532 1544.65,586.373 1548.52,586.504 1552.39,586.922 1556.25,587.622 
  1560.12,588.603 1563.99,589.858 1567.85,591.385 1571.72,593.177 1575.59,595.23 1579.45,597.538 1583.32,600.096 1587.19,602.907 1591.05,605.971 1594.92,609.291 
  1598.79,612.869 1602.65,616.703 1606.52,620.794 1610.39,625.141 1614.25,629.742 1618.12,634.595 1621.98,639.698 1625.85,645.047 1629.72,650.637 1633.58,656.465 
  1637.45,662.524 1641.32,668.809 1645.18,675.313 1649.05,682.028 1652.92,688.948 1656.78,696.063 1660.65,703.363 1664.52,710.841 1668.38,718.483 1672.25,726.281 
  1676.12,734.222 1679.98,742.293 1683.85,750.482 1687.72,758.774 1691.58,767.158 1695.45,775.62 1699.31,784.139 1703.18,792.692 1707.05,801.258 1710.91,809.816 
  1714.78,818.345 1718.65,826.828 1722.51,835.247 1726.38,843.586 1730.25,851.828 1734.11,859.96 1737.98,867.969 1741.85,875.842 1745.71,883.567 1749.58,891.136 
  1753.45,898.537 1757.31,905.764 1761.18,912.81 1765.05,919.668 1768.91,926.333 1772.78,932.802 1776.65,939.072 1780.51,945.14 1784.38,951.007 1788.24,956.672 
  1792.11,962.131 1795.98,967.38 1799.84,972.419 1803.71,977.248 1807.58,981.868 1811.44,986.278 1815.31,990.479 1819.18,994.472 1823.04,998.256 1826.91,1001.83 
  1830.78,1005.2 1834.64,1008.37 1838.51,1011.33 1842.38,1014.09 1846.24,1016.64 1850.11,1018.99 1853.98,1021.15 1857.84,1023.11 1861.71,1024.87 1865.58,1026.43 
  1869.44,1027.81 1873.31,1028.99 1877.17,1029.99 1881.04,1030.8 1884.91,1031.43 1888.77,1031.88 1892.64,1032.15 1896.51,1032.24 1900.37,1032.16 1904.24,1031.91 
  1908.11,1031.49 1911.97,1030.9 1915.84,1030.14 1919.71,1029.21 1923.57,1028.13 1927.44,1026.88 1931.31,1025.47 1935.17,1023.91 1939.04,1022.19 1942.91,1020.32 
  1946.77,1018.3 1950.64,1016.13 1954.5,1013.82 1958.37,1011.36 1962.24,1008.77 1966.1,1006.04 1969.97,1003.17 1973.84,1000.18 1977.7,997.051 1981.57,993.8 
  1985.44,990.424 1989.3,986.929 1993.17,983.317 1997.04,979.59 2000.9,975.752 2004.77,971.806 2008.64,967.756 2012.5,963.604 2016.37,959.354 2020.24,955.01 
  2024.1,950.574 2027.97,946.051 2031.84,941.443 2035.7,936.755 2039.57,931.99 2043.43,927.152 2047.3,922.244 2051.17,917.272 2055.03,912.239 2058.9,907.15 
  2062.77,902.008 2066.63,896.818 2070.5,891.586 2074.37,886.315 2078.23,881.01 2082.1,875.677 2085.97,870.32 2089.83,864.944 2093.7,859.554 2097.57,854.157 
  2101.43,848.756 2105.3,843.358 2109.17,837.969 2113.03,832.591 2116.9,827.23 2120.77,821.894 2124.63,816.589 2128.5,811.323 2132.36,806.101 2136.23,800.93 
  2140.1,795.817 2143.96,790.768 2147.83,785.79 2151.7,780.888 2155.56,776.069 2159.43,771.34 2163.3,766.705 2167.16,762.172 2171.03,757.745 2174.9,753.431 
  2178.76,749.236 2182.63,745.164 2186.5,741.223 2190.36,737.416 2194.23,733.75 2198.1,730.23 2201.96,726.862 2205.83,723.649 2209.69,720.598 2213.56,717.714 
  2217.43,714.998 2221.29,712.453 2225.16,710.082 2229.03,707.889 2232.89,705.875 2236.76,704.043 2240.63,702.395 2244.49,700.933 2248.36,699.659 2252.23,698.573 
  2256.09,697.677 2259.96,696.971 2263.83,696.457 2267.69,696.134 2271.56,696.002 2275.43,696.061 2279.29,696.311 2283.16,696.751 2287.03,697.379 2290.89,698.196 
  2294.76,699.198 
  "/>
<path clip-path="url(#clip210)" d="
M1479.53 299.749 L2284.45 299.749 L2284.45 92.3886 L1479.53 92.3886  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<polyline clip-path="url(#clip210)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1479.53,299.749 2284.45,299.749 2284.45,92.3886 1479.53,92.3886 1479.53,299.749 
  "/>
<polyline clip-path="url(#clip210)" style="stroke:#009af9; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1502.3,144.229 1638.92,144.229 
  "/>
<path clip-path="url(#clip210)" d="M1683.93 128.083 L1683.93 132.643 Q1681.27 131.37 1678.91 130.745 Q1676.55 130.12 1674.35 130.12 Q1670.53 130.12 1668.45 131.601 Q1666.39 133.083 1666.39 135.814 Q1666.39 138.106 1667.75 139.287 Q1669.14 140.444 1672.99 141.162 L1675.81 141.74 Q1681.04 142.736 1683.52 145.259 Q1686.02 147.759 1686.02 151.972 Q1686.02 156.995 1682.64 159.587 Q1679.28 162.18 1672.78 162.18 Q1670.32 162.18 1667.55 161.624 Q1664.79 161.069 1661.83 159.981 L1661.83 155.166 Q1664.68 156.763 1667.41 157.573 Q1670.14 158.384 1672.78 158.384 Q1676.78 158.384 1678.96 156.81 Q1681.13 155.236 1681.13 152.319 Q1681.13 149.773 1679.56 148.337 Q1678.01 146.902 1674.44 146.185 L1671.6 145.629 Q1666.37 144.587 1664.03 142.365 Q1661.69 140.143 1661.69 136.185 Q1661.69 131.601 1664.91 128.962 Q1668.15 126.324 1673.82 126.324 Q1676.25 126.324 1678.77 126.763 Q1681.3 127.203 1683.93 128.083 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1693.12 135.583 L1697.38 135.583 L1697.38 161.509 L1693.12 161.509 L1693.12 135.583 M1693.12 125.49 L1697.38 125.49 L1697.38 130.884 L1693.12 130.884 L1693.12 125.49 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1726.48 140.56 Q1728.08 137.689 1730.3 136.324 Q1732.52 134.958 1735.53 134.958 Q1739.58 134.958 1741.78 137.805 Q1743.98 140.629 1743.98 145.861 L1743.98 161.509 L1739.7 161.509 L1739.7 145.999 Q1739.7 142.273 1738.38 140.467 Q1737.06 138.662 1734.35 138.662 Q1731.04 138.662 1729.12 140.861 Q1727.2 143.06 1727.2 146.856 L1727.2 161.509 L1722.92 161.509 L1722.92 145.999 Q1722.92 142.249 1721.6 140.467 Q1720.28 138.662 1717.52 138.662 Q1714.26 138.662 1712.34 140.884 Q1710.42 143.083 1710.42 146.856 L1710.42 161.509 L1706.13 161.509 L1706.13 135.583 L1710.42 135.583 L1710.42 139.611 Q1711.87 137.226 1713.91 136.092 Q1715.95 134.958 1718.75 134.958 Q1721.57 134.958 1723.54 136.393 Q1725.53 137.828 1726.48 140.56 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1756.6 157.62 L1756.6 171.37 L1752.31 171.37 L1752.31 135.583 L1756.6 135.583 L1756.6 139.518 Q1757.94 137.203 1759.98 136.092 Q1762.04 134.958 1764.88 134.958 Q1769.61 134.958 1772.55 138.708 Q1775.51 142.458 1775.51 148.569 Q1775.51 154.68 1772.55 158.43 Q1769.61 162.18 1764.88 162.18 Q1762.04 162.18 1759.98 161.069 Q1757.94 159.935 1756.6 157.62 M1771.09 148.569 Q1771.09 143.87 1769.14 141.208 Q1767.22 138.523 1763.84 138.523 Q1760.46 138.523 1758.52 141.208 Q1756.6 143.87 1756.6 148.569 Q1756.6 153.268 1758.52 155.953 Q1760.46 158.615 1763.84 158.615 Q1767.22 158.615 1769.14 155.953 Q1771.09 153.268 1771.09 148.569 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1782.57 125.49 L1786.83 125.49 L1786.83 161.509 L1782.57 161.509 L1782.57 125.49 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1817.92 147.481 L1817.92 149.564 L1798.33 149.564 Q1798.61 153.962 1800.97 156.277 Q1803.36 158.569 1807.59 158.569 Q1810.05 158.569 1812.34 157.967 Q1814.65 157.365 1816.92 156.161 L1816.92 160.189 Q1814.63 161.161 1812.22 161.671 Q1809.81 162.18 1807.34 162.18 Q1801.13 162.18 1797.5 158.569 Q1793.89 154.958 1793.89 148.8 Q1793.89 142.435 1797.31 138.708 Q1800.76 134.958 1806.6 134.958 Q1811.83 134.958 1814.86 138.337 Q1817.92 141.694 1817.92 147.481 M1813.66 146.231 Q1813.61 142.736 1811.69 140.652 Q1809.79 138.569 1806.64 138.569 Q1803.08 138.569 1800.92 140.583 Q1798.8 142.597 1798.47 146.254 L1813.66 146.231 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1825.09 126.949 L1844.95 126.949 L1844.95 130.884 L1829.77 130.884 L1829.77 141.069 L1843.47 141.069 L1843.47 145.004 L1829.77 145.004 L1829.77 161.509 L1825.09 161.509 L1825.09 126.949 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1852.36 126.949 L1859.33 126.949 L1868.15 150.467 L1877.01 126.949 L1883.98 126.949 L1883.98 161.509 L1879.42 161.509 L1879.42 131.162 L1870.51 154.865 L1865.81 154.865 L1856.9 131.162 L1856.9 161.509 L1852.36 161.509 L1852.36 126.949 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1892.73 126.949 L1897.43 126.949 L1897.43 147.944 Q1897.43 153.499 1899.44 155.953 Q1901.46 158.384 1905.97 158.384 Q1910.46 158.384 1912.48 155.953 Q1914.49 153.499 1914.49 147.944 L1914.49 126.949 L1919.19 126.949 L1919.19 148.523 Q1919.19 155.282 1915.83 158.731 Q1912.5 162.18 1905.97 162.18 Q1899.42 162.18 1896.06 158.731 Q1892.73 155.282 1892.73 148.523 L1892.73 126.949 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip210)" style="stroke:#e26f46; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1502.3,196.069 1638.92,196.069 
  "/>
<path clip-path="url(#clip210)" d="M1678.08 197.145 Q1679.58 197.654 1680.99 199.321 Q1682.43 200.988 1683.87 203.904 L1688.61 213.349 L1683.59 213.349 L1679.17 204.483 Q1677.45 201.011 1675.83 199.876 Q1674.24 198.742 1671.46 198.742 L1666.37 198.742 L1666.37 213.349 L1661.69 213.349 L1661.69 178.789 L1672.24 178.789 Q1678.17 178.789 1681.09 181.265 Q1684 183.742 1684 188.742 Q1684 192.006 1682.48 194.159 Q1680.97 196.312 1678.08 197.145 M1666.37 182.631 L1666.37 194.9 L1672.24 194.9 Q1675.62 194.9 1677.34 193.349 Q1679.07 191.775 1679.07 188.742 Q1679.07 185.71 1677.34 184.182 Q1675.62 182.631 1672.24 182.631 L1666.37 182.631 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1714.49 199.321 L1714.49 201.404 L1694.91 201.404 Q1695.18 205.802 1697.55 208.117 Q1699.93 210.409 1704.17 210.409 Q1706.62 210.409 1708.91 209.807 Q1711.23 209.205 1713.49 208.001 L1713.49 212.029 Q1711.2 213.001 1708.8 213.511 Q1706.39 214.02 1703.91 214.02 Q1697.71 214.02 1694.07 210.409 Q1690.46 206.798 1690.46 200.64 Q1690.46 194.275 1693.89 190.548 Q1697.34 186.798 1703.17 186.798 Q1708.4 186.798 1711.43 190.177 Q1714.49 193.534 1714.49 199.321 M1710.23 198.071 Q1710.18 194.576 1708.26 192.492 Q1706.37 190.409 1703.22 190.409 Q1699.65 190.409 1697.5 192.423 Q1695.37 194.437 1695.05 198.094 L1710.23 198.071 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1733.26 200.316 Q1728.1 200.316 1726.11 201.497 Q1724.12 202.677 1724.12 205.525 Q1724.12 207.793 1725.6 209.136 Q1727.11 210.455 1729.68 210.455 Q1733.22 210.455 1735.35 207.955 Q1737.5 205.432 1737.5 201.265 L1737.5 200.316 L1733.26 200.316 M1741.76 198.557 L1741.76 213.349 L1737.5 213.349 L1737.5 209.413 Q1736.04 211.775 1733.86 212.909 Q1731.69 214.02 1728.54 214.02 Q1724.56 214.02 1722.2 211.798 Q1719.86 209.552 1719.86 205.802 Q1719.86 201.427 1722.78 199.205 Q1725.72 196.983 1731.53 196.983 L1737.5 196.983 L1737.5 196.566 Q1737.5 193.627 1735.55 192.029 Q1733.63 190.409 1730.14 190.409 Q1727.92 190.409 1725.81 190.941 Q1723.7 191.474 1721.76 192.539 L1721.76 188.603 Q1724.1 187.701 1726.3 187.261 Q1728.49 186.798 1730.58 186.798 Q1736.2 186.798 1738.98 189.714 Q1741.76 192.631 1741.76 198.557 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1750.53 177.33 L1754.79 177.33 L1754.79 213.349 L1750.53 213.349 L1750.53 177.33 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1763.89 178.789 L1783.75 178.789 L1783.75 182.724 L1768.56 182.724 L1768.56 192.909 L1782.27 192.909 L1782.27 196.844 L1768.56 196.844 L1768.56 213.349 L1763.89 213.349 L1763.89 178.789 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1791.16 178.789 L1798.12 178.789 L1806.94 202.307 L1815.81 178.789 L1822.78 178.789 L1822.78 213.349 L1818.22 213.349 L1818.22 183.002 L1809.3 206.705 L1804.61 206.705 L1795.69 183.002 L1795.69 213.349 L1791.16 213.349 L1791.16 178.789 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1831.53 178.789 L1836.23 178.789 L1836.23 199.784 Q1836.23 205.339 1838.24 207.793 Q1840.25 210.224 1844.77 210.224 Q1849.26 210.224 1851.27 207.793 Q1853.29 205.339 1853.29 199.784 L1853.29 178.789 L1857.98 178.789 L1857.98 200.363 Q1857.98 207.122 1854.63 210.571 Q1851.29 214.02 1844.77 214.02 Q1838.22 214.02 1834.86 210.571 Q1831.53 207.122 1831.53 200.363 L1831.53 178.789 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip210)" style="stroke:#3da44d; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1502.3,247.909 1638.92,247.909 
  "/>
<path clip-path="url(#clip210)" d="M1661.69 230.629 L1667.99 230.629 L1683.31 259.541 L1683.31 230.629 L1687.85 230.629 L1687.85 265.189 L1681.55 265.189 L1666.23 236.277 L1666.23 265.189 L1661.69 265.189 L1661.69 230.629 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1719.14 251.161 L1719.14 253.244 L1699.56 253.244 Q1699.84 257.642 1702.2 259.957 Q1704.58 262.249 1708.82 262.249 Q1711.27 262.249 1713.56 261.647 Q1715.88 261.045 1718.15 259.841 L1718.15 263.869 Q1715.86 264.841 1713.45 265.351 Q1711.04 265.86 1708.56 265.86 Q1702.36 265.86 1698.73 262.249 Q1695.12 258.638 1695.12 252.48 Q1695.12 246.115 1698.54 242.388 Q1701.99 238.638 1707.82 238.638 Q1713.05 238.638 1716.09 242.017 Q1719.14 245.374 1719.14 251.161 M1714.88 249.911 Q1714.84 246.416 1712.92 244.332 Q1711.02 242.249 1707.87 242.249 Q1704.3 242.249 1702.15 244.263 Q1700.02 246.277 1699.7 249.934 L1714.88 249.911 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1725.69 254.957 L1725.69 239.263 L1729.95 239.263 L1729.95 254.795 Q1729.95 258.476 1731.39 260.328 Q1732.82 262.156 1735.69 262.156 Q1739.14 262.156 1741.13 259.957 Q1743.15 257.758 1743.15 253.962 L1743.15 239.263 L1747.41 239.263 L1747.41 265.189 L1743.15 265.189 L1743.15 261.207 Q1741.6 263.568 1739.54 264.726 Q1737.5 265.86 1734.79 265.86 Q1730.32 265.86 1728.01 263.082 Q1725.69 260.304 1725.69 254.957 M1736.41 238.638 L1736.41 238.638 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1771.2 243.244 Q1770.49 242.828 1769.63 242.642 Q1768.8 242.434 1767.78 242.434 Q1764.17 242.434 1762.22 244.795 Q1760.3 247.133 1760.3 251.531 L1760.3 265.189 L1756.02 265.189 L1756.02 239.263 L1760.3 239.263 L1760.3 243.291 Q1761.64 240.929 1763.8 239.795 Q1765.95 238.638 1769.03 238.638 Q1769.47 238.638 1770 238.707 Q1770.53 238.754 1771.18 238.869 L1771.2 243.244 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1787.45 252.156 Q1782.29 252.156 1780.3 253.337 Q1778.31 254.517 1778.31 257.365 Q1778.31 259.633 1779.79 260.976 Q1781.3 262.295 1783.86 262.295 Q1787.41 262.295 1789.54 259.795 Q1791.69 257.272 1791.69 253.105 L1791.69 252.156 L1787.45 252.156 M1795.95 250.397 L1795.95 265.189 L1791.69 265.189 L1791.69 261.253 Q1790.23 263.615 1788.05 264.749 Q1785.88 265.86 1782.73 265.86 Q1778.75 265.86 1776.39 263.638 Q1774.05 261.392 1774.05 257.642 Q1774.05 253.267 1776.97 251.045 Q1779.91 248.823 1785.72 248.823 L1791.69 248.823 L1791.69 248.406 Q1791.69 245.467 1789.74 243.869 Q1787.82 242.249 1784.33 242.249 Q1782.11 242.249 1780 242.781 Q1777.89 243.314 1775.95 244.379 L1775.95 240.443 Q1778.29 239.541 1780.49 239.101 Q1782.68 238.638 1784.77 238.638 Q1790.39 238.638 1793.17 241.554 Q1795.95 244.471 1795.95 250.397 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1804.72 229.17 L1808.98 229.17 L1808.98 265.189 L1804.72 265.189 L1804.72 229.17 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1818.08 230.629 L1837.94 230.629 L1837.94 234.564 L1822.75 234.564 L1822.75 244.749 L1836.46 244.749 L1836.46 248.684 L1822.75 248.684 L1822.75 265.189 L1818.08 265.189 L1818.08 230.629 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1845.35 230.629 L1852.31 230.629 L1861.13 254.147 L1870 230.629 L1876.97 230.629 L1876.97 265.189 L1872.41 265.189 L1872.41 234.842 L1863.49 258.545 L1858.79 258.545 L1849.88 234.842 L1849.88 265.189 L1845.35 265.189 L1845.35 230.629 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1885.72 230.629 L1890.42 230.629 L1890.42 251.624 Q1890.42 257.179 1892.43 259.633 Q1894.44 262.064 1898.96 262.064 Q1903.45 262.064 1905.46 259.633 Q1907.48 257.179 1907.48 251.624 L1907.48 230.629 L1912.17 230.629 L1912.17 252.203 Q1912.17 258.962 1908.82 262.411 Q1905.48 265.86 1898.96 265.86 Q1892.41 265.86 1889.05 262.411 Q1885.72 258.962 1885.72 252.203 L1885.72 230.629 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1946.06 229.217 Q1942.96 234.541 1941.46 239.749 Q1939.95 244.957 1939.95 250.304 Q1939.95 255.652 1941.46 260.906 Q1942.98 266.138 1946.06 271.439 L1942.36 271.439 Q1938.89 265.999 1937.15 260.744 Q1935.44 255.49 1935.44 250.304 Q1935.44 245.142 1937.15 239.911 Q1938.86 234.68 1942.36 229.217 L1946.06 229.217 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1969.1 246.554 Q1972.45 247.272 1974.33 249.541 Q1976.22 251.809 1976.22 255.142 Q1976.22 260.258 1972.71 263.059 Q1969.19 265.86 1962.71 265.86 Q1960.53 265.86 1958.22 265.42 Q1955.92 265.003 1953.47 264.147 L1953.47 259.633 Q1955.41 260.767 1957.73 261.346 Q1960.04 261.925 1962.57 261.925 Q1966.97 261.925 1969.26 260.189 Q1971.57 258.453 1971.57 255.142 Q1971.57 252.087 1969.42 250.374 Q1967.29 248.638 1963.47 248.638 L1959.44 248.638 L1959.44 244.795 L1963.66 244.795 Q1967.1 244.795 1968.93 243.429 Q1970.76 242.041 1970.76 239.448 Q1970.76 236.786 1968.86 235.374 Q1966.99 233.939 1963.47 233.939 Q1961.55 233.939 1959.35 234.355 Q1957.15 234.772 1954.51 235.652 L1954.51 231.485 Q1957.17 230.744 1959.49 230.374 Q1961.83 230.004 1963.89 230.004 Q1969.21 230.004 1972.31 232.434 Q1975.41 234.842 1975.41 238.962 Q1975.41 241.832 1973.77 243.823 Q1972.13 245.791 1969.1 246.554 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M1995.09 233.707 Q1991.48 233.707 1989.65 237.272 Q1987.85 240.814 1987.85 247.943 Q1987.85 255.05 1989.65 258.615 Q1991.48 262.156 1995.09 262.156 Q1998.72 262.156 2000.53 258.615 Q2002.36 255.05 2002.36 247.943 Q2002.36 240.814 2000.53 237.272 Q1998.72 233.707 1995.09 233.707 M1995.09 230.004 Q2000.9 230.004 2003.96 234.61 Q2007.03 239.193 2007.03 247.943 Q2007.03 256.67 2003.96 261.277 Q2000.9 265.86 1995.09 265.86 Q1989.28 265.86 1986.2 261.277 Q1983.15 256.67 1983.15 247.943 Q1983.15 239.193 1986.2 234.61 Q1989.28 230.004 1995.09 230.004 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M2025.25 233.707 Q2021.64 233.707 2019.81 237.272 Q2018.01 240.814 2018.01 247.943 Q2018.01 255.05 2019.81 258.615 Q2021.64 262.156 2025.25 262.156 Q2028.89 262.156 2030.69 258.615 Q2032.52 255.05 2032.52 247.943 Q2032.52 240.814 2030.69 237.272 Q2028.89 233.707 2025.25 233.707 M2025.25 230.004 Q2031.06 230.004 2034.12 234.61 Q2037.2 239.193 2037.2 247.943 Q2037.2 256.67 2034.12 261.277 Q2031.06 265.86 2025.25 265.86 Q2019.44 265.86 2016.36 261.277 Q2013.31 256.67 2013.31 247.943 Q2013.31 239.193 2016.36 234.61 Q2019.44 230.004 2025.25 230.004 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M2082.06 251.161 L2082.06 253.244 L2062.47 253.244 Q2062.75 257.642 2065.11 259.957 Q2067.5 262.249 2071.73 262.249 Q2074.19 262.249 2076.48 261.647 Q2078.79 261.045 2081.06 259.841 L2081.06 263.869 Q2078.77 264.841 2076.36 265.351 Q2073.96 265.86 2071.48 265.86 Q2065.28 265.86 2061.64 262.249 Q2058.03 258.638 2058.03 252.48 Q2058.03 246.115 2061.46 242.388 Q2064.9 238.638 2070.74 238.638 Q2075.97 238.638 2079 242.017 Q2082.06 245.374 2082.06 251.161 M2077.8 249.911 Q2077.75 246.416 2075.83 244.332 Q2073.93 242.249 2070.78 242.249 Q2067.22 242.249 2065.07 244.263 Q2062.94 246.277 2062.61 249.934 L2077.8 249.911 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M2093.17 261.3 L2093.17 275.05 L2088.89 275.05 L2088.89 239.263 L2093.17 239.263 L2093.17 243.198 Q2094.51 240.883 2096.55 239.772 Q2098.61 238.638 2101.46 238.638 Q2106.18 238.638 2109.12 242.388 Q2112.08 246.138 2112.08 252.249 Q2112.08 258.36 2109.12 262.11 Q2106.18 265.86 2101.46 265.86 Q2098.61 265.86 2096.55 264.749 Q2094.51 263.615 2093.17 261.3 M2107.66 252.249 Q2107.66 247.55 2105.71 244.888 Q2103.79 242.203 2100.41 242.203 Q2097.03 242.203 2095.09 244.888 Q2093.17 247.55 2093.17 252.249 Q2093.17 256.948 2095.09 259.633 Q2097.03 262.295 2100.41 262.295 Q2103.79 262.295 2105.71 259.633 Q2107.66 256.948 2107.66 252.249 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M2129.19 242.249 Q2125.76 242.249 2123.77 244.934 Q2121.78 247.596 2121.78 252.249 Q2121.78 256.902 2123.75 259.587 Q2125.74 262.249 2129.19 262.249 Q2132.59 262.249 2134.58 259.564 Q2136.57 256.878 2136.57 252.249 Q2136.57 247.642 2134.58 244.957 Q2132.59 242.249 2129.19 242.249 M2129.19 238.638 Q2134.74 238.638 2137.91 242.249 Q2141.08 245.86 2141.08 252.249 Q2141.08 258.615 2137.91 262.249 Q2134.74 265.86 2129.19 265.86 Q2123.61 265.86 2120.44 262.249 Q2117.29 258.615 2117.29 252.249 Q2117.29 245.86 2120.44 242.249 Q2123.61 238.638 2129.19 238.638 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M2166.8 240.258 L2166.8 244.24 Q2165 243.244 2163.17 242.758 Q2161.36 242.249 2159.51 242.249 Q2155.37 242.249 2153.08 244.888 Q2150.78 247.504 2150.78 252.249 Q2150.78 256.994 2153.08 259.633 Q2155.37 262.249 2159.51 262.249 Q2161.36 262.249 2163.17 261.763 Q2165 261.253 2166.8 260.258 L2166.8 264.193 Q2165.02 265.027 2163.1 265.443 Q2161.2 265.86 2159.05 265.86 Q2153.19 265.86 2149.74 262.179 Q2146.29 258.499 2146.29 252.249 Q2146.29 245.906 2149.77 242.272 Q2153.26 238.638 2159.33 238.638 Q2161.29 238.638 2163.17 239.055 Q2165.04 239.448 2166.8 240.258 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M2195.76 249.541 L2195.76 265.189 L2191.5 265.189 L2191.5 249.679 Q2191.5 245.999 2190.07 244.17 Q2188.63 242.342 2185.76 242.342 Q2182.31 242.342 2180.32 244.541 Q2178.33 246.74 2178.33 250.536 L2178.33 265.189 L2174.05 265.189 L2174.05 229.17 L2178.33 229.17 L2178.33 243.291 Q2179.86 240.953 2181.92 239.795 Q2184 238.638 2186.71 238.638 Q2191.18 238.638 2193.47 241.416 Q2195.76 244.17 2195.76 249.541 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M2220.78 240.027 L2220.78 244.054 Q2218.98 243.129 2217.03 242.666 Q2215.09 242.203 2213.01 242.203 Q2209.83 242.203 2208.24 243.175 Q2206.66 244.147 2206.66 246.091 Q2206.66 247.573 2207.8 248.429 Q2208.93 249.263 2212.36 250.027 L2213.82 250.351 Q2218.35 251.323 2220.25 253.105 Q2222.17 254.865 2222.17 258.036 Q2222.17 261.647 2219.3 263.753 Q2216.45 265.86 2211.45 265.86 Q2209.37 265.86 2207.1 265.443 Q2204.86 265.05 2202.36 264.24 L2202.36 259.841 Q2204.72 261.068 2207.01 261.693 Q2209.3 262.295 2211.55 262.295 Q2214.56 262.295 2216.18 261.277 Q2217.8 260.235 2217.8 258.36 Q2217.8 256.624 2216.62 255.698 Q2215.46 254.772 2211.5 253.916 L2210.02 253.568 Q2206.06 252.735 2204.3 251.022 Q2202.54 249.286 2202.54 246.277 Q2202.54 242.619 2205.14 240.629 Q2207.73 238.638 2212.5 238.638 Q2214.86 238.638 2216.94 238.985 Q2219.02 239.332 2220.78 240.027 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip210)" d="M2228.28 229.217 L2231.99 229.217 Q2235.46 234.68 2237.17 239.911 Q2238.91 245.142 2238.91 250.304 Q2238.91 255.49 2237.17 260.744 Q2235.46 265.999 2231.99 271.439 L2228.28 271.439 Q2231.36 266.138 2232.87 260.906 Q2234.39 255.652 2234.39 250.304 Q2234.39 244.957 2232.87 239.749 Q2231.36 234.541 2228.28 229.217 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /></svg>



#### Continue training and plotting

As can be seen from the previous figure, the plot of the NeuralFMU has not yet fully converged against the *realFMU*, so the training of the NeuralFMU is continued. After further training, the plot of *NeuralFMU* is added to the figure again. The effect of the longer training is well recognizable, since the plot of the NeuralFMU had further converged. 


```julia
Flux.train!(lossSum, paramsNet, Iterators.repeated((), 700), optim; cb=callb) 
# plot results mass.s
solutionAfter = neuralFMU(x₀, tStart)
posNeuralFMU = collect(data[1] for data in solutionAfter.u)
Plots.plot!(fig, tSave, posNeuralFMU, label="NeuralFMU (1000 epochs)", linewidth=2)
fig 
```

    ┌ Info: Loss [301]: 0.00138   Avg displacement in data: 0.03717
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_ME.ipynb:8
    ┌ Info: Loss [321]: 0.00129   Avg displacement in data: 0.0359
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_ME.ipynb:8
    ┌ Info: Loss [341]: 0.0012   Avg displacement in data: 0.0346
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_ME.ipynb:8
    ┌ Info: Loss [361]: 0.0011   Avg displacement in data: 0.03324
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_ME.ipynb:8
    ┌ Info: Loss [381]: 0.00101   Avg displacement in data: 0.03182
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_ME.ipynb:8
    ┌ Info: Loss [401]: 0.00092   Avg displacement in data: 0.03035
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_ME.ipynb:8
    ┌ Info: Loss [421]: 0.00083   Avg displacement in data: 0.02889
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_ME.ipynb:8
    ┌ Info: Loss [441]: 0.00074   Avg displacement in data: 0.02726
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_ME.ipynb:8
    ┌ Info: Loss [461]: 0.00065   Avg displacement in data: 0.0255
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_ME.ipynb:8
    ┌ Info: Loss [481]: 0.00056   Avg displacement in data: 0.02367
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_ME.ipynb:8
    ┌ Info: Loss [501]: 0.00048   Avg displacement in data: 0.02186
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_ME.ipynb:8
    ┌ Info: Loss [521]: 0.0004   Avg displacement in data: 0.02004
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_ME.ipynb:8
    ┌ Info: Loss [541]: 0.00033   Avg displacement in data: 0.01825
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_ME.ipynb:8
    ┌ Info: Loss [561]: 0.00027   Avg displacement in data: 0.01651
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_ME.ipynb:8
    ┌ Info: Loss [581]: 0.00023   Avg displacement in data: 0.015
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_ME.ipynb:8
    ┌ Info: Loss [601]: 0.00019   Avg displacement in data: 0.01372
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_ME.ipynb:8
    ┌ Info: Loss [621]: 0.00016   Avg displacement in data: 0.01267
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_ME.ipynb:8
    ┌ Info: Loss [641]: 0.00014   Avg displacement in data: 0.01184
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_ME.ipynb:8
    ┌ Info: Loss [661]: 0.00012   Avg displacement in data: 0.01117
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_ME.ipynb:8
    ┌ Info: Loss [681]: 0.00011   Avg displacement in data: 0.01062
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_ME.ipynb:8
    ┌ Info: Loss [701]: 0.0001   Avg displacement in data: 0.01014
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_ME.ipynb:8
    ┌ Info: Loss [721]: 9.0e-5   Avg displacement in data: 0.00971
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_ME.ipynb:8
    ┌ Info: Loss [741]: 9.0e-5   Avg displacement in data: 0.00933
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_ME.ipynb:8
    ┌ Info: Loss [761]: 8.0e-5   Avg displacement in data: 0.00899
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_ME.ipynb:8
    ┌ Info: Loss [781]: 0.00011   Avg displacement in data: 0.01052
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_ME.ipynb:8
    ┌ Info: Loss [801]: 7.0e-5   Avg displacement in data: 0.00847
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_ME.ipynb:8
    ┌ Info: Loss [821]: 7.0e-5   Avg displacement in data: 0.00825
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_ME.ipynb:8
    ┌ Info: Loss [841]: 6.0e-5   Avg displacement in data: 0.00802
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_ME.ipynb:8
    ┌ Info: Loss [861]: 6.0e-5   Avg displacement in data: 0.00783
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_ME.ipynb:8
    ┌ Info: Loss [881]: 6.0e-5   Avg displacement in data: 0.00766
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_ME.ipynb:8
    ┌ Info: Loss [901]: 6.0e-5   Avg displacement in data: 0.00751
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_ME.ipynb:8
    ┌ Info: Loss [921]: 5.0e-5   Avg displacement in data: 0.00738
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_ME.ipynb:8
    ┌ Info: Loss [941]: 5.0e-5   Avg displacement in data: 0.00726
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_ME.ipynb:8
    ┌ Info: Loss [961]: 6.0e-5   Avg displacement in data: 0.00744
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_ME.ipynb:8
    ┌ Info: Loss [981]: 5.0e-5   Avg displacement in data: 0.00712
    └ @ Main c:\Users\thummeto\Documents\FMIFlux.jl\example\simple_hybrid_ME.ipynb:8



<?xml version="1.0" encoding="utf-8"?>
<svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" width="600" height="400" viewBox="0 0 2400 1600">
<defs>
  <clipPath id="clip300">
    <rect x="0" y="0" width="2400" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip300)" d="
M0 1600 L2400 1600 L2400 0 L0 0  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip301">
    <rect x="480" y="0" width="1681" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip300)" d="
M303.503 1401.58 L2352.76 1401.58 L2352.76 47.2441 L303.503 47.2441  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip302">
    <rect x="303" y="47" width="2050" height="1355"/>
  </clipPath>
</defs>
<polyline clip-path="url(#clip302)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  361.501,1401.58 361.501,47.2441 
  "/>
<polyline clip-path="url(#clip302)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  748.152,1401.58 748.152,47.2441 
  "/>
<polyline clip-path="url(#clip302)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1134.8,1401.58 1134.8,47.2441 
  "/>
<polyline clip-path="url(#clip302)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1521.46,1401.58 1521.46,47.2441 
  "/>
<polyline clip-path="url(#clip302)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  1908.11,1401.58 1908.11,47.2441 
  "/>
<polyline clip-path="url(#clip302)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  2294.76,1401.58 2294.76,47.2441 
  "/>
<polyline clip-path="url(#clip300)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  303.503,1401.58 2352.76,1401.58 
  "/>
<polyline clip-path="url(#clip300)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  361.501,1401.58 361.501,1382.68 
  "/>
<polyline clip-path="url(#clip300)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  748.152,1401.58 748.152,1382.68 
  "/>
<polyline clip-path="url(#clip300)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1134.8,1401.58 1134.8,1382.68 
  "/>
<polyline clip-path="url(#clip300)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1521.46,1401.58 1521.46,1382.68 
  "/>
<polyline clip-path="url(#clip300)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1908.11,1401.58 1908.11,1382.68 
  "/>
<polyline clip-path="url(#clip300)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  2294.76,1401.58 2294.76,1382.68 
  "/>
<path clip-path="url(#clip300)" d="M361.501 1438.36 Q356.084 1438.36 353.341 1443.71 Q350.633 1449.02 350.633 1459.71 Q350.633 1470.37 353.341 1475.72 Q356.084 1481.03 361.501 1481.03 Q366.952 1481.03 369.66 1475.72 Q372.403 1470.37 372.403 1459.71 Q372.403 1449.02 369.66 1443.71 Q366.952 1438.36 361.501 1438.36 M361.501 1432.8 Q370.216 1432.8 374.799 1439.71 Q379.417 1446.59 379.417 1459.71 Q379.417 1472.8 374.799 1479.71 Q370.216 1486.59 361.501 1486.59 Q352.785 1486.59 348.167 1479.71 Q343.584 1472.8 343.584 1459.71 Q343.584 1446.59 348.167 1439.71 Q352.785 1432.8 361.501 1432.8 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M733.725 1479.68 L745.183 1479.68 L745.183 1440.13 L732.718 1442.63 L732.718 1436.24 L745.114 1433.74 L752.128 1433.74 L752.128 1479.68 L763.586 1479.68 L763.586 1485.58 L733.725 1485.58 L733.725 1479.68 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1126.78 1479.68 L1151.26 1479.68 L1151.26 1485.58 L1118.35 1485.58 L1118.35 1479.68 Q1122.34 1475.55 1129.21 1468.6 Q1136.12 1461.62 1137.89 1459.61 Q1141.26 1455.82 1142.58 1453.22 Q1143.94 1450.58 1143.94 1448.05 Q1143.94 1443.91 1141.02 1441.31 Q1138.14 1438.71 1133.48 1438.71 Q1130.19 1438.71 1126.51 1439.85 Q1122.86 1441 1118.69 1443.32 L1118.69 1436.24 Q1122.93 1434.54 1126.61 1433.67 Q1130.29 1432.8 1133.35 1432.8 Q1141.4 1432.8 1146.19 1436.83 Q1150.98 1440.86 1150.98 1447.59 Q1150.98 1450.79 1149.77 1453.67 Q1148.59 1456.52 1145.43 1460.41 Q1144.56 1461.41 1139.91 1466.24 Q1135.26 1471.03 1126.78 1479.68 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1527.83 1457.63 Q1532.86 1458.7 1535.67 1462.11 Q1538.52 1465.51 1538.52 1470.51 Q1538.52 1478.18 1533.24 1482.39 Q1527.97 1486.59 1518.24 1486.59 Q1514.98 1486.59 1511.51 1485.93 Q1508.07 1485.3 1504.39 1484.02 L1504.39 1477.25 Q1507.31 1478.95 1510.78 1479.82 Q1514.25 1480.68 1518.04 1480.68 Q1524.63 1480.68 1528.07 1478.08 Q1531.54 1475.48 1531.54 1470.51 Q1531.54 1465.93 1528.31 1463.36 Q1525.12 1460.75 1519.39 1460.75 L1513.35 1460.75 L1513.35 1454.99 L1519.67 1454.99 Q1524.84 1454.99 1527.58 1452.94 Q1530.33 1450.86 1530.33 1446.97 Q1530.33 1442.98 1527.48 1440.86 Q1524.67 1438.71 1519.39 1438.71 Q1516.51 1438.71 1513.21 1439.33 Q1509.91 1439.96 1505.95 1441.27 L1505.95 1435.02 Q1509.94 1433.91 1513.42 1433.36 Q1516.92 1432.8 1520.01 1432.8 Q1528 1432.8 1532.65 1436.45 Q1537.31 1440.06 1537.31 1446.24 Q1537.31 1450.55 1534.84 1453.53 Q1532.38 1456.48 1527.83 1457.63 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1912.62 1439.85 L1894.91 1467.52 L1912.62 1467.52 L1912.62 1439.85 M1910.78 1433.74 L1919.6 1433.74 L1919.6 1467.52 L1927 1467.52 L1927 1473.36 L1919.6 1473.36 L1919.6 1485.58 L1912.62 1485.58 L1912.62 1473.36 L1889.22 1473.36 L1889.22 1466.59 L1910.78 1433.74 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M2280.17 1433.74 L2307.71 1433.74 L2307.71 1439.64 L2286.6 1439.64 L2286.6 1452.35 Q2288.13 1451.83 2289.65 1451.59 Q2291.18 1451.31 2292.71 1451.31 Q2301.39 1451.31 2306.46 1456.07 Q2311.53 1460.82 2311.53 1468.95 Q2311.53 1477.32 2306.32 1481.97 Q2301.11 1486.59 2291.63 1486.59 Q2288.37 1486.59 2284.97 1486.03 Q2281.6 1485.48 2277.99 1484.36 L2277.99 1477.32 Q2281.11 1479.02 2284.45 1479.85 Q2287.78 1480.68 2291.49 1480.68 Q2297.5 1480.68 2301.01 1477.52 Q2304.52 1474.36 2304.52 1468.95 Q2304.52 1463.53 2301.01 1460.37 Q2297.5 1457.21 2291.49 1457.21 Q2288.68 1457.21 2285.87 1457.84 Q2283.09 1458.46 2280.17 1459.78 L2280.17 1433.74 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1271.74 1527.62 L1271.74 1538.66 L1284.9 1538.66 L1284.9 1543.62 L1271.74 1543.62 L1271.74 1564.74 Q1271.74 1569.49 1273.03 1570.85 Q1274.35 1572.2 1278.34 1572.2 L1284.9 1572.2 L1284.9 1577.55 L1278.34 1577.55 Q1270.94 1577.55 1268.13 1574.8 Q1265.32 1572.03 1265.32 1564.74 L1265.32 1543.62 L1260.63 1543.62 L1260.63 1538.66 L1265.32 1538.66 L1265.32 1527.62 L1271.74 1527.62 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1315.32 1523.52 L1330.04 1523.52 L1330.04 1528.49 L1321.71 1528.49 L1321.71 1581.96 L1330.04 1581.96 L1330.04 1586.92 L1315.32 1586.92 L1315.32 1523.52 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1368.44 1539.8 L1368.44 1545.85 Q1365.73 1544.46 1362.82 1543.76 Q1359.9 1543.07 1356.78 1543.07 Q1352.02 1543.07 1349.62 1544.53 Q1347.26 1545.99 1347.26 1548.9 Q1347.26 1551.12 1348.96 1552.41 Q1350.66 1553.66 1355.8 1554.8 L1357.99 1555.29 Q1364.8 1556.75 1367.64 1559.42 Q1370.53 1562.06 1370.53 1566.82 Q1370.53 1572.24 1366.22 1575.4 Q1361.95 1578.55 1354.45 1578.55 Q1351.32 1578.55 1347.92 1577.93 Q1344.55 1577.34 1340.8 1576.12 L1340.8 1569.53 Q1344.34 1571.37 1347.78 1572.3 Q1351.22 1573.21 1354.59 1573.21 Q1359.1 1573.21 1361.53 1571.68 Q1363.96 1570.12 1363.96 1567.3 Q1363.96 1564.7 1362.19 1563.31 Q1360.46 1561.92 1354.52 1560.64 L1352.3 1560.12 Q1346.36 1558.87 1343.72 1556.3 Q1341.08 1553.69 1341.08 1549.18 Q1341.08 1543.69 1344.97 1540.71 Q1348.86 1537.72 1356.01 1537.72 Q1359.55 1537.72 1362.68 1538.24 Q1365.8 1538.76 1368.44 1539.8 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1395.63 1523.52 L1395.63 1586.92 L1380.91 1586.92 L1380.91 1581.96 L1389.21 1581.96 L1389.21 1528.49 L1380.91 1528.49 L1380.91 1523.52 L1395.63 1523.52 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip302)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  303.503,1361.96 2352.76,1361.96 
  "/>
<polyline clip-path="url(#clip302)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  303.503,1096.36 2352.76,1096.36 
  "/>
<polyline clip-path="url(#clip302)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  303.503,830.761 2352.76,830.761 
  "/>
<polyline clip-path="url(#clip302)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  303.503,565.16 2352.76,565.16 
  "/>
<polyline clip-path="url(#clip302)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="
  303.503,299.559 2352.76,299.559 
  "/>
<polyline clip-path="url(#clip300)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  303.503,1401.58 303.503,47.2441 
  "/>
<polyline clip-path="url(#clip300)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  303.503,1361.96 322.401,1361.96 
  "/>
<polyline clip-path="url(#clip300)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  303.503,1096.36 322.401,1096.36 
  "/>
<polyline clip-path="url(#clip300)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  303.503,830.761 322.401,830.761 
  "/>
<polyline clip-path="url(#clip300)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  303.503,565.16 322.401,565.16 
  "/>
<polyline clip-path="url(#clip300)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  303.503,299.559 322.401,299.559 
  "/>
<path clip-path="url(#clip300)" d="M136.497 1340.66 Q131.08 1340.66 128.337 1346.01 Q125.629 1351.32 125.629 1362.01 Q125.629 1372.67 128.337 1378.02 Q131.08 1383.33 136.497 1383.33 Q141.948 1383.33 144.656 1378.02 Q147.399 1372.67 147.399 1362.01 Q147.399 1351.32 144.656 1346.01 Q141.948 1340.66 136.497 1340.66 M136.497 1335.11 Q145.212 1335.11 149.795 1342.01 Q154.413 1348.89 154.413 1362.01 Q154.413 1375.1 149.795 1382.01 Q145.212 1388.89 136.497 1388.89 Q127.781 1388.89 123.163 1382.01 Q118.58 1375.1 118.58 1362.01 Q118.58 1348.89 123.163 1342.01 Q127.781 1335.11 136.497 1335.11 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M166.74 1379.06 L174.066 1379.06 L174.066 1387.88 L166.74 1387.88 L166.74 1379.06 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M189.413 1336.04 L216.948 1336.04 L216.948 1341.95 L195.837 1341.95 L195.837 1354.65 Q197.364 1354.13 198.892 1353.89 Q200.42 1353.61 201.948 1353.61 Q210.628 1353.61 215.698 1358.37 Q220.767 1363.13 220.767 1371.25 Q220.767 1379.62 215.559 1384.27 Q210.35 1388.89 200.871 1388.89 Q197.608 1388.89 194.205 1388.33 Q190.837 1387.78 187.226 1386.67 L187.226 1379.62 Q190.351 1381.32 193.684 1382.15 Q197.017 1382.99 200.732 1382.99 Q206.739 1382.99 210.246 1379.83 Q213.753 1376.67 213.753 1371.25 Q213.753 1365.83 210.246 1362.67 Q206.739 1359.51 200.732 1359.51 Q197.92 1359.51 195.108 1360.14 Q192.33 1360.76 189.413 1362.08 L189.413 1336.04 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M249.586 1340.66 Q244.17 1340.66 241.427 1346.01 Q238.718 1351.32 238.718 1362.01 Q238.718 1372.67 241.427 1378.02 Q244.17 1383.33 249.586 1383.33 Q255.038 1383.33 257.746 1378.02 Q260.489 1372.67 260.489 1362.01 Q260.489 1351.32 257.746 1346.01 Q255.038 1340.66 249.586 1340.66 M249.586 1335.11 Q258.302 1335.11 262.885 1342.01 Q267.503 1348.89 267.503 1362.01 Q267.503 1375.1 262.885 1382.01 Q258.302 1388.89 249.586 1388.89 Q240.871 1388.89 236.253 1382.01 Q231.67 1375.1 231.67 1362.01 Q231.67 1348.89 236.253 1342.01 Q240.871 1335.11 249.586 1335.11 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M137.99 1075.06 Q132.573 1075.06 129.83 1080.41 Q127.122 1085.72 127.122 1096.41 Q127.122 1107.07 129.83 1112.42 Q132.573 1117.73 137.99 1117.73 Q143.441 1117.73 146.149 1112.42 Q148.892 1107.07 148.892 1096.41 Q148.892 1085.72 146.149 1080.41 Q143.441 1075.06 137.99 1075.06 M137.99 1069.5 Q146.705 1069.5 151.288 1076.41 Q155.906 1083.29 155.906 1096.41 Q155.906 1109.5 151.288 1116.41 Q146.705 1123.29 137.99 1123.29 Q129.275 1123.29 124.657 1116.41 Q120.073 1109.5 120.073 1096.41 Q120.073 1083.29 124.657 1076.41 Q129.275 1069.5 137.99 1069.5 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M168.233 1113.46 L175.559 1113.46 L175.559 1122.28 L168.233 1122.28 L168.233 1113.46 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M189.066 1070.44 L222.399 1070.44 L222.399 1073.43 L203.58 1122.28 L196.253 1122.28 L213.962 1076.34 L189.066 1076.34 L189.066 1070.44 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M236.149 1070.44 L263.684 1070.44 L263.684 1076.34 L242.573 1076.34 L242.573 1089.05 Q244.1 1088.53 245.628 1088.29 Q247.156 1088.01 248.684 1088.01 Q257.364 1088.01 262.434 1092.77 Q267.503 1097.52 267.503 1105.65 Q267.503 1114.02 262.295 1118.67 Q257.086 1123.29 247.607 1123.29 Q244.343 1123.29 240.941 1122.73 Q237.573 1122.18 233.961 1121.07 L233.961 1114.02 Q237.086 1115.72 240.42 1116.55 Q243.753 1117.39 247.468 1117.39 Q253.475 1117.39 256.982 1114.23 Q260.489 1111.07 260.489 1105.65 Q260.489 1100.23 256.982 1097.07 Q253.475 1093.91 247.468 1093.91 Q244.656 1093.91 241.843 1094.54 Q239.066 1095.16 236.149 1096.48 L236.149 1070.44 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M122.712 850.778 L134.17 850.778 L134.17 811.23 L121.705 813.73 L121.705 807.341 L134.101 804.841 L141.115 804.841 L141.115 850.778 L152.573 850.778 L152.573 856.681 L122.712 856.681 L122.712 850.778 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M166.74 847.861 L174.066 847.861 L174.066 856.681 L166.74 856.681 L166.74 847.861 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M204.344 809.459 Q198.927 809.459 196.184 814.806 Q193.476 820.118 193.476 830.813 Q193.476 841.472 196.184 846.82 Q198.927 852.132 204.344 852.132 Q209.795 852.132 212.503 846.82 Q215.246 841.472 215.246 830.813 Q215.246 820.118 212.503 814.806 Q209.795 809.459 204.344 809.459 M204.344 803.903 Q213.059 803.903 217.642 810.813 Q222.26 817.688 222.26 830.813 Q222.26 843.903 217.642 850.813 Q213.059 857.688 204.344 857.688 Q195.628 857.688 191.01 850.813 Q186.427 843.903 186.427 830.813 Q186.427 817.688 191.01 810.813 Q195.628 803.903 204.344 803.903 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M249.586 809.459 Q244.17 809.459 241.427 814.806 Q238.718 820.118 238.718 830.813 Q238.718 841.472 241.427 846.82 Q244.17 852.132 249.586 852.132 Q255.038 852.132 257.746 846.82 Q260.489 841.472 260.489 830.813 Q260.489 820.118 257.746 814.806 Q255.038 809.459 249.586 809.459 M249.586 803.903 Q258.302 803.903 262.885 810.813 Q267.503 817.688 267.503 830.813 Q267.503 843.903 262.885 850.813 Q258.302 857.688 249.586 857.688 Q240.871 857.688 236.253 850.813 Q231.67 843.903 231.67 830.813 Q231.67 817.688 236.253 810.813 Q240.871 803.903 249.586 803.903 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M124.205 585.177 L135.663 585.177 L135.663 545.629 L123.198 548.129 L123.198 541.74 L135.594 539.24 L142.608 539.24 L142.608 585.177 L154.066 585.177 L154.066 591.08 L124.205 591.08 L124.205 585.177 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M168.233 582.26 L175.559 582.26 L175.559 591.08 L168.233 591.08 L168.233 582.26 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M196.878 585.177 L221.357 585.177 L221.357 591.08 L188.441 591.08 L188.441 585.177 Q192.434 581.045 199.309 574.101 Q206.219 567.122 207.989 565.108 Q211.357 561.323 212.677 558.719 Q214.031 556.08 214.031 553.545 Q214.031 549.413 211.114 546.809 Q208.232 544.205 203.58 544.205 Q200.281 544.205 196.601 545.351 Q192.955 546.497 188.788 548.823 L188.788 541.74 Q193.024 540.038 196.705 539.17 Q200.385 538.302 203.441 538.302 Q211.496 538.302 216.288 542.33 Q221.08 546.358 221.08 553.094 Q221.08 556.288 219.864 559.17 Q218.684 562.017 215.524 565.906 Q214.656 566.913 210.003 571.74 Q205.351 576.531 196.878 585.177 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M236.149 539.24 L263.684 539.24 L263.684 545.143 L242.573 545.143 L242.573 557.851 Q244.1 557.33 245.628 557.087 Q247.156 556.809 248.684 556.809 Q257.364 556.809 262.434 561.566 Q267.503 566.323 267.503 574.448 Q267.503 582.816 262.295 587.469 Q257.086 592.087 247.607 592.087 Q244.343 592.087 240.941 591.531 Q237.573 590.976 233.961 589.865 L233.961 582.816 Q237.086 584.517 240.42 585.351 Q243.753 586.184 247.468 586.184 Q253.475 586.184 256.982 583.024 Q260.489 579.865 260.489 574.448 Q260.489 569.031 256.982 565.872 Q253.475 562.712 247.468 562.712 Q244.656 562.712 241.843 563.337 Q239.066 563.962 236.149 565.281 L236.149 539.24 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M122.712 319.576 L134.17 319.576 L134.17 280.028 L121.705 282.528 L121.705 276.139 L134.101 273.639 L141.115 273.639 L141.115 319.576 L152.573 319.576 L152.573 325.479 L122.712 325.479 L122.712 319.576 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M166.74 316.659 L174.066 316.659 L174.066 325.479 L166.74 325.479 L166.74 316.659 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M189.413 273.639 L216.948 273.639 L216.948 279.542 L195.837 279.542 L195.837 292.25 Q197.364 291.729 198.892 291.486 Q200.42 291.208 201.948 291.208 Q210.628 291.208 215.698 295.965 Q220.767 300.722 220.767 308.847 Q220.767 317.215 215.559 321.868 Q210.35 326.486 200.871 326.486 Q197.608 326.486 194.205 325.93 Q190.837 325.375 187.226 324.264 L187.226 317.215 Q190.351 318.916 193.684 319.75 Q197.017 320.583 200.732 320.583 Q206.739 320.583 210.246 317.423 Q213.753 314.264 213.753 308.847 Q213.753 303.43 210.246 300.271 Q206.739 297.111 200.732 297.111 Q197.92 297.111 195.108 297.736 Q192.33 298.361 189.413 299.68 L189.413 273.639 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M249.586 278.257 Q244.17 278.257 241.427 283.604 Q238.718 288.917 238.718 299.611 Q238.718 310.271 241.427 315.618 Q244.17 320.93 249.586 320.93 Q255.038 320.93 257.746 315.618 Q260.489 310.271 260.489 299.611 Q260.489 288.917 257.746 283.604 Q255.038 278.257 249.586 278.257 M249.586 272.701 Q258.302 272.701 262.885 279.611 Q267.503 286.486 267.503 299.611 Q267.503 312.701 262.885 319.611 Q258.302 326.486 249.586 326.486 Q240.871 326.486 236.253 319.611 Q231.67 312.701 231.67 299.611 Q231.67 286.486 236.253 279.611 Q240.871 272.701 249.586 272.701 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M35.6047 1006.89 Q31.2992 1004.5 29.2506 1001.16 Q27.202 997.831 27.202 993.317 Q27.202 987.24 31.4728 983.942 Q35.7089 980.643 43.5561 980.643 L67.0282 980.643 L67.0282 987.067 L43.7644 987.067 Q38.1741 987.067 35.4658 989.046 Q32.7575 991.025 32.7575 995.088 Q32.7575 1000.05 36.0561 1002.93 Q39.3547 1005.82 45.0491 1005.82 L67.0282 1005.82 L67.0282 1012.24 L43.7644 1012.24 Q38.1394 1012.24 35.4658 1014.22 Q32.7575 1016.2 32.7575 1020.33 Q32.7575 1025.23 36.0908 1028.11 Q39.3894 1030.99 45.0491 1030.99 L67.0282 1030.99 L67.0282 1037.41 L28.1395 1037.41 L28.1395 1030.99 L34.1811 1030.99 Q30.6047 1028.8 28.9034 1025.75 Q27.202 1022.69 27.202 1018.49 Q27.202 1014.25 29.3547 1011.3 Q31.5075 1008.32 35.6047 1006.89 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M47.4797 950.227 Q47.4797 957.97 49.2505 960.956 Q51.0213 963.942 55.2921 963.942 Q58.6949 963.942 60.7087 961.72 Q62.6879 959.463 62.6879 955.609 Q62.6879 950.296 58.9379 947.102 Q55.1532 943.873 48.9033 943.873 L47.4797 943.873 L47.4797 950.227 M44.8408 937.484 L67.0282 937.484 L67.0282 943.873 L61.1254 943.873 Q64.6671 946.06 66.3684 949.324 Q68.0351 952.588 68.0351 957.31 Q68.0351 963.282 64.7018 966.824 Q61.3337 970.331 55.7088 970.331 Q49.1463 970.331 45.813 965.956 Q42.4797 961.546 42.4797 952.831 L42.4797 943.873 L41.8547 943.873 Q37.445 943.873 35.0492 946.789 Q32.6186 949.671 32.6186 954.914 Q32.6186 958.248 33.4172 961.407 Q34.2158 964.567 35.813 967.484 L29.9103 967.484 Q28.5561 963.977 27.8964 960.678 Q27.202 957.379 27.202 954.254 Q27.202 945.817 31.577 941.65 Q35.9519 937.484 44.8408 937.484 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M29.2853 899.533 L35.3269 899.533 Q33.9381 902.241 33.2436 905.158 Q32.5492 908.074 32.5492 911.199 Q32.5492 915.956 34.0075 918.352 Q35.4658 920.713 38.3825 920.713 Q40.6047 920.713 41.8894 919.012 Q43.1394 917.31 44.2852 912.171 L44.7713 909.984 Q46.2297 903.178 48.9033 900.331 Q51.5421 897.449 56.299 897.449 Q61.7157 897.449 64.8754 901.755 Q68.0351 906.026 68.0351 913.526 Q68.0351 916.651 67.4101 920.053 Q66.8198 923.421 65.6046 927.171 L59.0074 927.171 Q60.8476 923.63 61.7851 920.192 Q62.6879 916.755 62.6879 913.387 Q62.6879 908.873 61.1601 906.442 Q59.5976 904.012 56.7852 904.012 Q54.181 904.012 52.7921 905.783 Q51.4032 907.519 50.1185 913.456 L49.5977 915.678 Q48.3477 921.616 45.7783 924.255 Q43.1741 926.894 38.6603 926.894 Q33.1742 926.894 30.1881 923.005 Q27.202 919.116 27.202 911.963 Q27.202 908.421 27.7228 905.296 Q28.2436 902.171 29.2853 899.533 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M29.2853 862.484 L35.3269 862.484 Q33.9381 865.192 33.2436 868.109 Q32.5492 871.026 32.5492 874.151 Q32.5492 878.908 34.0075 881.303 Q35.4658 883.665 38.3825 883.665 Q40.6047 883.665 41.8894 881.963 Q43.1394 880.262 44.2852 875.123 L44.7713 872.935 Q46.2297 866.13 48.9033 863.283 Q51.5421 860.401 56.299 860.401 Q61.7157 860.401 64.8754 864.706 Q68.0351 868.977 68.0351 876.477 Q68.0351 879.602 67.4101 883.005 Q66.8198 886.373 65.6046 890.123 L59.0074 890.123 Q60.8476 886.581 61.7851 883.144 Q62.6879 879.706 62.6879 876.338 Q62.6879 871.824 61.1601 869.394 Q59.5976 866.963 56.7852 866.963 Q54.181 866.963 52.7921 868.734 Q51.4032 870.47 50.1185 876.408 L49.5977 878.63 Q48.3477 884.567 45.7783 887.206 Q43.1741 889.845 38.6603 889.845 Q33.1742 889.845 30.1881 885.956 Q27.202 882.067 27.202 874.915 Q27.202 871.373 27.7228 868.248 Q28.2436 865.123 29.2853 862.484 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M61.1949 821.443 L81.8197 821.443 L81.8197 827.866 L28.1395 827.866 L28.1395 821.443 L34.0422 821.443 Q30.57 819.429 28.9034 816.373 Q27.202 813.283 27.202 809.012 Q27.202 801.929 32.827 797.519 Q38.4519 793.075 47.6185 793.075 Q56.7852 793.075 62.4101 797.519 Q68.0351 801.929 68.0351 809.012 Q68.0351 813.283 66.3684 816.373 Q64.6671 819.429 61.1949 821.443 M47.6185 799.707 Q40.57 799.707 36.5769 802.623 Q32.5492 805.505 32.5492 810.575 Q32.5492 815.644 36.5769 818.561 Q40.57 821.443 47.6185 821.443 Q54.6671 821.443 58.6949 818.561 Q62.6879 815.644 62.6879 810.575 Q62.6879 805.505 58.6949 802.623 Q54.6671 799.707 47.6185 799.707 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M32.6186 767.415 Q32.6186 772.554 36.6464 775.54 Q40.6394 778.526 47.6185 778.526 Q54.5977 778.526 58.6254 775.575 Q62.6185 772.589 62.6185 767.415 Q62.6185 762.311 58.5907 759.325 Q54.5629 756.339 47.6185 756.339 Q40.7089 756.339 36.6811 759.325 Q32.6186 762.311 32.6186 767.415 M27.202 767.415 Q27.202 759.082 32.6186 754.325 Q38.0353 749.568 47.6185 749.568 Q57.1671 749.568 62.6185 754.325 Q68.0351 759.082 68.0351 767.415 Q68.0351 775.783 62.6185 780.54 Q57.1671 785.262 47.6185 785.262 Q38.0353 785.262 32.6186 780.54 Q27.202 775.783 27.202 767.415 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M29.2853 714.186 L35.3269 714.186 Q33.9381 716.895 33.2436 719.811 Q32.5492 722.728 32.5492 725.853 Q32.5492 730.61 34.0075 733.006 Q35.4658 735.367 38.3825 735.367 Q40.6047 735.367 41.8894 733.665 Q43.1394 731.964 44.2852 726.825 L44.7713 724.638 Q46.2297 717.832 48.9033 714.985 Q51.5421 712.103 56.299 712.103 Q61.7157 712.103 64.8754 716.409 Q68.0351 720.679 68.0351 728.179 Q68.0351 731.304 67.4101 734.707 Q66.8198 738.075 65.6046 741.825 L59.0074 741.825 Q60.8476 738.283 61.7851 734.846 Q62.6879 731.408 62.6879 728.04 Q62.6879 723.527 61.1601 721.096 Q59.5976 718.665 56.7852 718.665 Q54.181 718.665 52.7921 720.436 Q51.4032 722.172 50.1185 728.11 L49.5977 730.332 Q48.3477 736.27 45.7783 738.908 Q43.1741 741.547 38.6603 741.547 Q33.1742 741.547 30.1881 737.658 Q27.202 733.77 27.202 726.617 Q27.202 723.075 27.7228 719.95 Q28.2436 716.825 29.2853 714.186 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M28.1395 701.929 L28.1395 695.541 L67.0282 695.541 L67.0282 701.929 L28.1395 701.929 M13.0007 701.929 L13.0007 695.541 L21.0909 695.541 L21.0909 701.929 L13.0007 701.929 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M17.0979 675.853 L28.1395 675.853 L28.1395 662.694 L33.1047 662.694 L33.1047 675.853 L54.2157 675.853 Q58.9726 675.853 60.3268 674.568 Q61.681 673.249 61.681 669.256 L61.681 662.694 L67.0282 662.694 L67.0282 669.256 Q67.0282 676.652 64.2851 679.464 Q61.5074 682.277 54.2157 682.277 L33.1047 682.277 L33.1047 686.964 L28.1395 686.964 L28.1395 682.277 L17.0979 682.277 L17.0979 675.853 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M28.1395 654.291 L28.1395 647.902 L67.0282 647.902 L67.0282 654.291 L28.1395 654.291 M13.0007 654.291 L13.0007 647.902 L21.0909 647.902 L21.0909 654.291 L13.0007 654.291 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M32.6186 619.465 Q32.6186 624.603 36.6464 627.59 Q40.6394 630.576 47.6185 630.576 Q54.5977 630.576 58.6254 627.624 Q62.6185 624.638 62.6185 619.465 Q62.6185 614.36 58.5907 611.374 Q54.5629 608.388 47.6185 608.388 Q40.7089 608.388 36.6811 611.374 Q32.6186 614.36 32.6186 619.465 M27.202 619.465 Q27.202 611.131 32.6186 606.374 Q38.0353 601.617 47.6185 601.617 Q57.1671 601.617 62.6185 606.374 Q68.0351 611.131 68.0351 619.465 Q68.0351 627.833 62.6185 632.59 Q57.1671 637.312 47.6185 637.312 Q38.0353 637.312 32.6186 632.59 Q27.202 627.833 27.202 619.465 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M43.5561 558.701 L67.0282 558.701 L67.0282 565.09 L43.7644 565.09 Q38.2436 565.09 35.5005 567.243 Q32.7575 569.395 32.7575 573.701 Q32.7575 578.875 36.0561 581.861 Q39.3547 584.847 45.0491 584.847 L67.0282 584.847 L67.0282 591.27 L28.1395 591.27 L28.1395 584.847 L34.1811 584.847 Q30.6742 582.555 28.9381 579.465 Q27.202 576.34 27.202 572.277 Q27.202 565.576 31.3686 562.139 Q35.5005 558.701 43.5561 558.701 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M13.0007 523.944 L13.0007 509.222 L17.9659 509.222 L17.9659 517.555 L71.4379 517.555 L71.4379 509.222 L76.4031 509.222 L76.4031 523.944 L13.0007 523.944 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M35.6047 465.333 Q31.2992 462.938 29.2506 459.604 Q27.202 456.271 27.202 451.757 Q27.202 445.681 31.4728 442.382 Q35.7089 439.084 43.5561 439.084 L67.0282 439.084 L67.0282 445.507 L43.7644 445.507 Q38.1741 445.507 35.4658 447.486 Q32.7575 449.466 32.7575 453.528 Q32.7575 458.493 36.0561 461.375 Q39.3547 464.257 45.0491 464.257 L67.0282 464.257 L67.0282 470.681 L43.7644 470.681 Q38.1394 470.681 35.4658 472.66 Q32.7575 474.639 32.7575 478.771 Q32.7575 483.667 36.0908 486.549 Q39.3894 489.431 45.0491 489.431 L67.0282 489.431 L67.0282 495.854 L28.1395 495.854 L28.1395 489.431 L34.1811 489.431 Q30.6047 487.243 28.9034 484.188 Q27.202 481.132 27.202 476.931 Q27.202 472.695 29.3547 469.743 Q31.5075 466.757 35.6047 465.333 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M13.0007 411.41 L76.4031 411.41 L76.4031 426.132 L71.4379 426.132 L71.4379 417.834 L17.9659 417.834 L17.9659 426.132 L13.0007 426.132 L13.0007 411.41 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip302)" style="stroke:#009af9; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" points="
  361.501,1361.96 365.367,1361.63 369.234,1360.67 373.1,1359.08 376.967,1356.85 380.833,1353.99 384.7,1350.51 388.566,1346.39 392.433,1341.66 396.299,1336.31 
  400.166,1330.34 404.032,1323.77 407.899,1316.6 411.765,1308.85 415.632,1300.5 419.498,1291.59 423.365,1282.1 427.231,1272.06 431.098,1261.48 434.964,1250.35 
  438.831,1238.7 442.698,1226.54 446.564,1213.87 450.431,1200.71 454.297,1187.08 458.164,1172.99 462.03,1158.44 465.897,1143.47 469.763,1128.08 473.63,1112.28 
  477.496,1096.1 481.363,1079.55 485.229,1062.63 489.096,1045.38 492.962,1027.81 496.829,1009.94 500.695,991.785 504.562,973.358 508.428,954.678 512.295,935.763 
  516.161,916.636 520.028,897.318 523.894,877.825 527.761,858.176 531.627,838.392 535.494,818.493 539.36,798.499 543.227,778.426 547.093,758.297 550.96,738.133 
  554.826,717.956 558.693,697.786 562.559,677.644 566.426,657.548 570.293,637.513 574.159,617.565 578.026,597.724 581.892,578.011 585.759,558.448 589.625,539.055 
  593.492,519.839 597.358,500.828 601.225,482.04 605.091,463.496 608.958,445.216 612.824,427.217 616.691,409.516 620.557,392.125 624.424,375.067 628.29,358.359 
  632.157,342.018 636.023,326.062 639.89,310.507 643.756,295.363 647.623,280.646 651.489,266.373 655.356,252.559 659.222,239.218 663.089,226.365 666.955,214.011 
  670.822,202.163 674.688,190.837 678.555,180.046 682.421,169.8 686.288,160.111 690.154,150.988 694.021,142.437 697.887,134.466 701.754,127.085 705.621,120.303 
  709.487,114.126 713.354,108.561 717.22,103.613 721.087,99.2832 724.953,95.579 728.82,92.5042 732.686,90.0622 736.553,88.2558 740.419,87.087 744.286,86.5554 
  748.152,86.6609 752.019,87.4047 755.885,88.7862 759.752,90.8041 763.618,93.4565 767.485,96.7409 771.351,100.653 775.218,105.189 779.084,110.344 782.951,116.115 
  786.817,122.494 790.684,129.476 794.55,137.053 798.417,145.218 802.283,153.963 806.15,163.279 810.016,173.157 813.883,183.586 817.749,194.555 821.616,206.056 
  825.482,218.077 829.349,230.604 833.216,243.626 837.082,257.128 840.949,271.096 844.815,285.519 848.682,300.383 852.548,315.672 856.415,331.37 860.281,347.461 
  864.148,363.928 868.014,380.754 871.881,397.927 875.747,415.429 879.614,433.239 883.48,451.341 887.347,469.715 891.213,488.343 895.08,507.206 898.946,526.291 
  902.813,545.575 906.679,565.038 910.546,584.659 914.412,604.419 918.279,624.297 922.145,644.279 926.012,664.343 929.878,684.468 933.745,704.632 937.611,724.815 
  941.478,744.995 945.344,765.154 949.211,785.278 953.077,805.341 956.944,825.324 960.811,845.205 964.677,864.963 968.544,884.578 972.41,904.036 976.277,923.318 
  980.143,942.401 984.01,961.265 987.876,979.892 991.743,998.26 995.609,1016.35 999.476,1034.16 1003.34,1051.66 1007.21,1068.83 1011.08,1085.65 1014.94,1102.11 
  1018.81,1118.2 1022.67,1133.88 1026.54,1149.17 1030.41,1164.03 1034.27,1178.45 1038.14,1192.41 1042.01,1205.9 1045.87,1218.91 1049.74,1231.43 1053.61,1243.45 
  1057.47,1254.94 1061.34,1265.9 1065.21,1276.32 1069.07,1286.19 1072.94,1295.49 1076.81,1304.23 1080.67,1312.39 1084.54,1319.95 1088.41,1326.93 1092.27,1333.29 
  1096.14,1339.05 1100.01,1344.2 1103.87,1348.72 1107.74,1352.62 1111.6,1355.89 1115.47,1358.53 1119.34,1360.54 1123.2,1361.91 1127.07,1362.64 1130.94,1362.73 
  1134.8,1362.19 1138.67,1361 1142.54,1359.18 1146.4,1356.73 1150.27,1353.64 1154.14,1349.92 1158,1345.58 1161.87,1340.61 1165.74,1335.03 1169.6,1328.84 
  1173.47,1322.04 1177.34,1314.64 1181.2,1306.66 1185.07,1298.09 1188.93,1288.95 1192.8,1279.24 1196.67,1268.98 1200.53,1258.17 1204.4,1246.83 1208.27,1234.96 
  1212.13,1222.59 1216,1209.71 1219.87,1196.36 1223.73,1182.53 1227.6,1168.24 1231.47,1153.5 1235.33,1138.34 1239.2,1122.76 1243.07,1106.78 1246.93,1090.43 
  1250.8,1073.7 1254.67,1056.63 1258.53,1039.22 1262.4,1021.5 1266.27,1003.48 1270.13,985.181 1274,966.626 1277.86,947.823 1281.73,928.796 1285.6,909.565 
  1289.46,890.149 1293.33,870.568 1297.2,850.843 1301.06,830.992 1304.93,811.03 1308.8,790.98 1312.66,770.865 1316.53,750.704 1320.4,730.518 1324.26,710.33 
  1328.13,690.15 1332,670.004 1335.86,649.911 1339.73,629.894 1343.6,609.973 1347.46,590.169 1351.33,570.5 1355.2,550.979 1359.06,531.632 1362.93,512.478 
  1366.79,493.537 1370.66,474.829 1374.53,456.374 1378.39,438.183 1382.26,420.276 1386.13,402.673 1389.99,385.393 1393.86,368.453 1397.73,351.872 1401.59,335.665 
  1405.46,319.841 1409.33,304.422 1413.19,289.424 1417.06,274.862 1420.93,260.751 1424.79,247.106 1428.66,233.938 1432.53,221.256 1436.39,209.079 1440.26,197.417 
  1444.12,186.284 1447.99,175.691 1451.86,165.65 1455.72,156.162 1459.59,147.243 1463.46,138.902 1467.32,131.147 1471.19,123.987 1475.06,117.429 1478.92,111.479 
  1482.79,106.139 1486.66,101.418 1490.52,97.3201 1494.39,93.8508 1498.26,91.0134 1502.12,88.8112 1505.99,87.2427 1509.86,86.3116 1513.72,86.0191 1517.59,86.3658 
  1521.46,87.3514 1525.32,88.9751 1529.19,91.2349 1533.05,94.1275 1536.92,97.6508 1540.79,101.802 1544.65,106.575 1548.52,111.968 1552.39,117.973 1556.25,124.585 
  1560.12,131.798 1563.99,139.604 1567.85,147.996 1571.72,156.964 1575.59,166.499 1579.45,176.593 1583.32,187.236 1587.19,198.417 1591.05,210.125 1594.92,222.347 
  1598.79,235.071 1602.65,248.284 1606.52,261.974 1610.39,276.129 1614.25,290.733 1618.12,305.77 1621.98,321.226 1625.85,337.084 1629.72,353.329 1633.58,369.949 
  1637.45,386.925 1641.32,404.238 1645.18,421.871 1649.05,439.805 1652.92,458.023 1656.78,476.509 1660.65,495.246 1664.52,514.213 1668.38,533.39 1672.25,552.757 
  1676.12,572.294 1679.98,591.982 1683.85,611.808 1687.72,631.747 1691.58,651.779 1695.45,671.883 1699.31,692.039 1703.18,712.224 1707.05,732.423 1710.91,752.619 
  1714.78,772.786 1718.65,792.905 1722.51,812.954 1726.38,832.913 1730.25,852.761 1734.11,872.486 1737.98,892.065 1741.85,911.476 1745.71,930.699 1749.58,949.714 
  1753.45,968.502 1757.31,987.045 1761.18,1005.33 1765.05,1023.34 1768.91,1041.04 1772.78,1058.43 1776.65,1075.48 1780.51,1092.18 1784.38,1108.52 1788.24,1124.47 
  1792.11,1140.02 1795.98,1155.16 1799.84,1169.86 1803.71,1184.12 1807.58,1197.92 1811.44,1211.24 1815.31,1224.08 1819.18,1236.42 1823.04,1248.25 1826.91,1259.55 
  1830.78,1270.32 1834.64,1280.54 1838.51,1290.21 1842.38,1299.31 1846.24,1307.83 1850.11,1315.77 1853.98,1323.12 1857.84,1329.87 1861.71,1336.01 1865.58,1341.55 
  1869.44,1346.46 1873.31,1350.76 1877.17,1354.42 1881.04,1357.46 1884.91,1359.86 1888.77,1361.63 1892.64,1362.76 1896.51,1363.25 1900.37,1363.1 1904.24,1362.31 
  1908.11,1360.89 1911.97,1358.83 1915.84,1356.13 1919.71,1352.8 1923.57,1348.84 1927.44,1344.26 1931.31,1339.06 1935.17,1333.24 1939.04,1326.82 1942.91,1319.79 
  1946.77,1312.16 1950.64,1303.95 1954.5,1295.16 1958.37,1285.79 1962.24,1275.87 1966.1,1265.39 1969.97,1254.37 1973.84,1242.82 1977.7,1230.75 1981.57,1218.18 
  1985.44,1205.11 1989.3,1191.56 1993.17,1177.55 1997.04,1163.08 2000.9,1148.17 2004.77,1132.83 2008.64,1117.1 2012.5,1100.96 2016.37,1084.45 2020.24,1067.57 
  2024.1,1050.36 2027.97,1032.82 2031.84,1014.97 2035.7,996.832 2039.57,978.417 2043.43,959.748 2047.3,940.843 2051.17,921.723 2055.03,902.407 2058.9,882.916 
  2062.77,863.262 2066.63,843.467 2070.5,823.554 2074.37,803.541 2078.23,783.451 2082.1,763.303 2085.97,743.119 2089.83,722.909 2093.7,702.701 2097.57,682.515 
  2101.43,662.372 2105.3,642.293 2109.17,622.298 2113.03,602.404 2116.9,582.628 2120.77,562.993 2124.63,543.521 2128.5,524.231 2132.36,505.144 2136.23,486.278 
  2140.1,467.644 2143.96,449.267 2147.83,431.165 2151.7,413.358 2155.56,395.864 2159.43,378.7 2163.3,361.882 2167.16,345.422 2171.03,329.342 2174.9,313.657 
  2178.76,298.384 2182.63,283.54 2186.5,269.139 2190.36,255.189 2194.23,241.707 2198.1,228.708 2201.96,216.206 2205.83,204.214 2209.69,192.745 2213.56,181.808 
  2217.43,171.409 2221.29,161.564 2225.16,152.283 2229.03,143.575 2232.89,135.449 2236.76,127.915 2240.63,120.974 2244.49,114.637 2248.36,108.91 2252.23,103.799 
  2256.09,99.3096 2259.96,95.4472 2263.83,92.2148 2267.69,89.6125 2271.56,87.6455 2275.43,86.3159 2279.29,85.6252 2283.16,85.5744 2287.03,86.1637 2290.89,87.3909 
  2294.76,89.2551 
  "/>
<polyline clip-path="url(#clip302)" style="stroke:#e26f46; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" points="
  361.501,1361.96 365.367,1361.73 369.234,1361.05 373.1,1359.91 376.967,1358.31 380.833,1356.25 384.7,1353.72 388.566,1350.73 392.433,1347.28 396.299,1343.37 
  400.166,1338.99 404.032,1334.17 407.899,1328.88 411.765,1323.15 415.632,1316.97 419.498,1310.35 423.365,1303.3 427.231,1295.82 431.098,1287.91 434.964,1279.6 
  438.831,1270.87 442.698,1261.74 446.564,1252.23 450.431,1242.33 454.297,1232.06 458.164,1221.43 462.03,1210.44 465.897,1199.12 469.763,1187.46 473.63,1175.49 
  477.496,1163.2 481.363,1150.63 485.229,1137.77 489.096,1124.64 492.962,1111.25 496.829,1097.61 500.695,1083.75 504.562,1069.67 508.428,1055.39 512.295,1040.93 
  516.161,1026.29 520.028,1011.48 523.894,996.536 527.761,981.46 531.627,966.272 535.494,950.989 539.36,935.627 543.227,920.193 547.093,904.709 550.96,889.189 
  554.826,873.649 558.693,858.104 562.559,842.569 566.426,827.06 570.293,811.592 574.159,796.181 578.026,780.841 581.892,765.588 585.759,750.438 589.625,735.405 
  593.492,720.505 597.358,705.752 601.225,691.161 605.091,676.746 608.958,662.523 612.824,648.507 616.691,634.708 620.557,621.144 624.424,607.826 628.29,594.77 
  632.157,581.987 636.023,569.491 639.89,557.293 643.756,545.408 647.623,533.847 651.489,522.621 655.356,511.741 659.222,501.22 663.089,491.068 666.955,481.296 
  670.822,471.913 674.688,462.929 678.555,454.355 682.421,446.199 686.288,438.47 690.154,431.177 694.021,424.328 697.887,417.931 701.754,411.993 705.621,406.521 
  709.487,401.523 713.354,397.005 717.22,392.973 721.087,389.434 724.953,386.393 728.82,383.856 732.686,381.828 736.553,380.314 740.419,379.32 744.286,378.851 
  748.152,378.883 752.019,379.301 755.885,380.094 759.752,381.268 763.618,382.822 767.485,384.757 771.351,387.076 775.218,389.776 779.084,392.859 782.951,396.323 
  786.817,400.165 790.684,404.384 794.55,408.978 798.417,413.942 802.283,419.274 806.15,424.969 810.016,431.022 813.883,437.43 817.749,444.185 821.616,451.283 
  825.482,458.717 829.349,466.48 833.216,474.565 837.082,482.963 840.949,491.669 844.815,500.672 848.682,509.964 852.548,519.535 856.415,529.379 860.281,539.485 
  864.148,549.844 868.014,560.445 871.881,571.278 875.747,582.332 879.614,593.596 883.48,605.059 887.347,616.71 891.213,628.537 895.08,640.527 898.946,652.669 
  902.813,664.952 906.679,677.362 910.546,689.888 914.412,702.517 918.279,715.236 922.145,728.033 926.012,740.895 929.878,753.81 933.745,766.764 937.611,779.744 
  941.478,792.738 945.344,805.733 949.211,818.715 953.077,831.673 956.944,844.591 960.811,857.456 964.677,870.259 968.544,882.985 972.41,895.621 976.277,908.154 
  980.143,920.571 984.01,932.86 987.876,945.011 991.743,957.008 995.609,968.841 999.476,980.497 1003.34,991.965 1007.21,1003.23 1011.08,1014.29 1014.94,1025.13 
  1018.81,1035.73 1022.67,1046.08 1026.54,1056.19 1030.41,1066.02 1034.27,1075.59 1038.14,1084.86 1042.01,1093.84 1045.87,1102.52 1049.74,1110.89 1053.61,1118.93 
  1057.47,1126.64 1061.34,1134.01 1065.21,1141.04 1069.07,1147.71 1072.94,1154.01 1076.81,1159.95 1080.67,1165.51 1084.54,1170.69 1088.41,1175.47 1092.27,1179.86 
  1096.14,1183.85 1100.01,1187.43 1103.87,1190.6 1107.74,1193.35 1111.6,1195.68 1115.47,1197.57 1119.34,1199.04 1123.2,1200.06 1127.07,1200.65 1130.94,1200.79 
  1134.8,1200.59 1138.67,1200.09 1142.54,1199.3 1146.4,1198.22 1150.27,1196.83 1154.14,1195.15 1158,1193.17 1161.87,1190.89 1165.74,1188.31 1169.6,1185.43 
  1173.47,1182.26 1177.34,1178.79 1181.2,1175.03 1185.07,1170.98 1188.93,1166.64 1192.8,1162.02 1196.67,1157.11 1200.53,1151.94 1204.4,1146.49 1208.27,1140.77 
  1212.13,1134.79 1216,1128.55 1219.87,1122.06 1223.73,1115.32 1227.6,1108.34 1231.47,1101.13 1235.33,1093.68 1239.2,1086.02 1243.07,1078.15 1246.93,1070.07 
  1250.8,1061.79 1254.67,1053.33 1258.53,1044.68 1262.4,1035.86 1266.27,1026.88 1270.13,1017.74 1274,1008.46 1277.86,999.035 1281.73,989.484 1285.6,979.815 
  1289.46,970.036 1293.33,960.158 1297.2,950.188 1301.06,940.138 1304.93,930.017 1308.8,919.837 1312.66,909.608 1316.53,899.341 1320.4,889.043 1324.26,878.724 
  1328.13,868.397 1332,858.071 1335.86,847.758 1339.73,837.469 1343.6,827.214 1347.46,817 1351.33,806.839 1355.2,796.743 1359.06,786.72 1362.93,776.782 
  1366.79,766.938 1370.66,757.198 1374.53,747.572 1378.39,738.07 1382.26,728.701 1386.13,719.475 1389.99,710.401 1393.86,701.488 1397.73,692.746 1401.59,684.183 
  1405.46,675.809 1409.33,667.632 1413.19,659.66 1417.06,651.902 1420.93,644.365 1424.79,637.057 1428.66,629.988 1432.53,623.164 1436.39,616.592 1440.26,610.28 
  1444.12,604.235 1447.99,598.463 1451.86,592.971 1455.72,587.765 1459.59,582.853 1463.46,578.239 1467.32,573.929 1471.19,569.929 1475.06,566.243 1478.92,562.879 
  1482.79,559.84 1486.66,557.131 1490.52,554.756 1494.39,552.72 1498.26,551.027 1502.12,549.68 1505.99,548.684 1509.86,548.042 1513.72,547.759 1517.59,547.8 
  1521.46,548.05 1525.32,548.505 1529.19,549.168 1533.05,550.039 1536.92,551.119 1540.79,552.409 1544.65,553.909 1548.52,555.619 1552.39,557.539 1556.25,559.668 
  1560.12,562.006 1563.99,564.551 1567.85,567.301 1571.72,570.256 1575.59,573.413 1579.45,576.77 1583.32,580.325 1587.19,584.076 1591.05,588.018 1594.92,592.15 
  1598.79,596.469 1602.65,600.968 1606.52,605.645 1610.39,610.497 1614.25,615.52 1618.12,620.711 1621.98,626.062 1625.85,631.57 1629.72,637.23 1633.58,643.036 
  1637.45,648.984 1641.32,655.067 1645.18,661.279 1649.05,667.615 1652.92,674.068 1656.78,680.633 1660.65,687.302 1664.52,694.07 1668.38,700.93 1672.25,707.875 
  1676.12,714.899 1679.98,721.995 1683.85,729.155 1687.72,736.373 1691.58,743.642 1695.45,750.954 1699.31,758.303 1703.18,765.681 1707.05,773.08 1710.91,780.494 
  1714.78,787.914 1718.65,795.334 1722.51,802.746 1726.38,810.143 1730.25,817.517 1734.11,824.86 1737.98,832.165 1741.85,839.426 1745.71,846.633 1749.58,853.781 
  1753.45,860.862 1757.31,867.868 1761.18,874.793 1765.05,881.629 1768.91,888.369 1772.78,895.007 1776.65,901.535 1780.51,907.947 1784.38,914.235 1788.24,920.394 
  1792.11,926.417 1795.98,932.298 1799.84,938.029 1803.71,943.606 1807.58,949.022 1811.44,954.271 1815.31,959.347 1819.18,964.244 1823.04,968.958 1826.91,973.483 
  1830.78,977.813 1834.64,981.943 1838.51,985.868 1842.38,989.584 1846.24,993.086 1850.11,996.37 1853.98,999.43 1857.84,1002.26 1861.71,1004.86 1865.58,1007.23 
  1869.44,1009.36 1873.31,1011.24 1877.17,1012.88 1881.04,1014.27 1884.91,1015.41 1888.77,1016.29 1892.64,1016.91 1896.51,1017.28 1900.37,1017.37 1904.24,1017.3 
  1908.11,1017.12 1911.97,1016.84 1915.84,1016.44 1919.71,1015.94 1923.57,1015.32 1927.44,1014.6 1931.31,1013.76 1935.17,1012.82 1939.04,1011.76 1942.91,1010.59 
  1946.77,1009.31 1950.64,1007.92 1954.5,1006.42 1958.37,1004.81 1962.24,1003.1 1966.1,1001.28 1969.97,999.352 1973.84,997.32 1977.7,995.187 1981.57,992.952 
  1985.44,990.617 1989.3,988.182 1993.17,985.653 1997.04,983.029 2000.9,980.313 2004.77,977.506 2008.64,974.612 2012.5,971.633 2016.37,968.57 2020.24,965.427 
  2024.1,962.208 2027.97,958.913 2031.84,955.547 2035.7,952.112 2039.57,948.612 2043.43,945.05 2047.3,941.428 2051.17,937.75 2055.03,934.021 2058.9,930.243 
  2062.77,926.419 2066.63,922.555 2070.5,918.652 2074.37,914.716 2078.23,910.748 2082.1,906.755 2085.97,902.739 2089.83,898.705 2093.7,894.655 2097.57,890.595 
  2101.43,886.529 2105.3,882.46 2109.17,878.392 2113.03,874.33 2116.9,870.277 2120.77,866.239 2124.63,862.218 2128.5,858.219 2132.36,854.246 2136.23,850.303 
  2140.1,846.395 2143.96,842.525 2147.83,838.697 2151.7,834.916 2155.56,831.185 2159.43,827.509 2163.3,823.891 2167.16,820.335 2171.03,816.846 2174.9,813.426 
  2178.76,810.08 2182.63,806.812 2186.5,803.624 2190.36,800.521 2194.23,797.507 2198.1,794.584 2201.96,791.756 2205.83,789.027 2209.69,786.399 2213.56,783.876 
  2217.43,781.462 2221.29,779.158 2225.16,776.968 2229.03,774.896 2232.89,772.943 2236.76,771.113 2240.63,769.408 2244.49,767.83 2248.36,766.383 2252.23,765.068 
  2256.09,763.889 2259.96,762.846 2263.83,761.942 2267.69,761.18 2271.56,760.56 2275.43,760.086 2279.29,759.758 2283.16,759.58 2287.03,759.544 2290.89,759.544 
  2294.76,759.544 
  "/>
<polyline clip-path="url(#clip302)" style="stroke:#3da44d; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" points="
  361.501,1361.96 365.367,1356.66 369.234,1351.16 373.1,1345.45 376.967,1339.54 380.833,1333.42 384.7,1327.1 388.566,1320.58 392.433,1313.86 396.299,1306.93 
  400.166,1299.8 404.032,1292.47 407.899,1284.94 411.765,1277.21 415.632,1269.28 419.498,1261.15 423.365,1252.81 427.231,1244.28 431.098,1235.55 434.964,1226.61 
  438.831,1217.48 442.698,1208.15 446.564,1198.62 450.431,1188.89 454.297,1178.97 458.164,1168.84 462.03,1158.53 465.897,1148.02 469.763,1137.31 473.63,1126.42 
  477.496,1115.33 481.363,1104.05 485.229,1092.59 489.096,1080.94 492.962,1069.1 496.829,1057.07 500.695,1044.86 504.562,1032.47 508.428,1019.89 512.295,1007.13 
  516.161,994.187 520.028,981.065 523.894,967.764 527.761,954.298 531.627,940.665 535.494,926.857 539.36,912.866 543.227,898.691 547.093,884.33 550.96,869.788 
  554.826,855.072 558.693,840.19 562.559,825.156 566.426,809.986 570.293,794.7 574.159,779.32 578.026,763.872 581.892,748.384 585.759,732.89 589.625,717.423 
  593.492,702.024 597.358,686.734 601.225,671.599 605.091,656.673 608.958,642.002 612.824,627.632 616.691,613.602 620.557,599.947 624.424,586.698 628.29,573.88 
  632.157,561.517 636.023,549.626 639.89,538.22 643.756,527.309 647.623,516.896 651.489,506.981 655.356,497.562 659.222,488.629 663.089,480.169 666.955,472.166 
  670.822,464.625 674.688,457.541 678.555,450.905 682.421,444.707 686.288,438.94 690.154,433.595 694.021,428.663 697.887,424.136 701.754,420.007 705.621,416.269 
  709.487,412.914 713.354,409.935 717.22,407.326 721.087,405.081 724.953,403.193 728.82,401.657 732.686,400.467 736.553,399.618 740.419,399.105 744.286,398.924 
  748.152,399.07 752.019,399.538 755.885,400.326 759.752,401.429 763.618,402.845 767.485,404.569 771.351,406.6 775.218,408.934 779.084,411.569 782.951,414.503 
  786.817,417.734 790.684,421.26 794.55,425.079 798.417,429.191 802.283,433.594 806.15,438.287 810.016,443.271 813.883,448.544 817.749,454.106 821.616,459.958 
  825.482,466.1 829.349,472.533 833.216,479.257 837.082,486.274 840.949,493.584 844.815,501.19 848.682,509.093 852.548,517.292 856.415,525.766 860.281,534.526 
  864.148,543.585 868.014,552.955 871.881,562.644 875.747,572.659 879.614,583.003 883.48,593.678 887.347,604.682 891.213,616.011 895.08,627.659 898.946,639.617 
  902.813,651.874 906.679,664.416 910.546,677.226 914.412,690.286 918.279,703.573 922.145,717.065 926.012,730.733 929.878,744.55 933.745,758.483 937.611,772.5 
  941.478,786.562 945.344,800.631 949.211,814.661 953.077,828.601 956.944,842.414 960.811,856.067 964.677,869.531 968.544,882.778 972.41,895.783 976.277,908.522 
  980.143,920.977 984.01,933.129 987.876,944.964 991.743,956.469 995.609,967.634 999.476,978.453 1003.34,988.92 1007.21,999.033 1011.08,1008.79 1014.94,1018.2 
  1018.81,1027.27 1022.67,1035.99 1026.54,1044.39 1030.41,1052.48 1034.27,1060.25 1038.14,1067.7 1042.01,1074.83 1045.87,1081.65 1049.74,1088.16 1053.61,1094.36 
  1057.47,1100.26 1061.34,1105.86 1065.21,1111.17 1069.07,1116.19 1072.94,1120.93 1076.81,1125.38 1080.67,1129.56 1084.54,1133.46 1088.41,1137.09 1092.27,1140.45 
  1096.14,1143.54 1100.01,1146.38 1103.87,1148.96 1107.74,1151.28 1111.6,1153.35 1115.47,1155.18 1119.34,1156.75 1123.2,1158.09 1127.07,1159.18 1130.94,1160.04 
  1134.8,1160.66 1138.67,1161.05 1142.54,1161.21 1146.4,1161.14 1150.27,1160.85 1154.14,1160.34 1158,1159.6 1161.87,1158.64 1165.74,1157.47 1169.6,1156.09 
  1173.47,1154.49 1177.34,1152.68 1181.2,1150.66 1185.07,1148.44 1188.93,1146.02 1192.8,1143.39 1196.67,1140.56 1200.53,1137.54 1204.4,1134.32 1208.27,1130.91 
  1212.13,1127.31 1216,1123.52 1219.87,1119.55 1223.73,1115.39 1227.6,1111.05 1231.47,1106.52 1235.33,1101.83 1239.2,1096.95 1243.07,1091.9 1246.93,1086.68 
  1250.8,1081.3 1254.67,1075.74 1258.53,1070.03 1262.4,1064.15 1266.27,1058.12 1270.13,1051.93 1274,1045.58 1277.86,1039.09 1281.73,1032.45 1285.6,1025.67 
  1289.46,1018.75 1293.33,1011.69 1297.2,1004.5 1301.06,997.184 1304.93,989.739 1308.8,982.173 1312.66,974.488 1316.53,966.69 1320.4,958.783 1324.26,950.771 
  1328.13,942.658 1332,934.45 1335.86,926.15 1339.73,917.764 1343.6,909.296 1347.46,900.751 1351.33,892.135 1355.2,883.453 1359.06,874.709 1362.93,865.907 
  1366.79,857.055 1370.66,848.162 1374.53,839.239 1378.39,830.294 1382.26,821.339 1386.13,812.383 1389.99,803.438 1393.86,794.514 1397.73,785.624 1401.59,776.777 
  1405.46,767.988 1409.33,759.267 1413.19,750.628 1417.06,742.083 1420.93,733.645 1424.79,725.329 1428.66,717.148 1432.53,709.115 1436.39,701.246 1440.26,693.555 
  1444.12,686.057 1447.99,678.767 1451.86,671.701 1455.72,664.875 1459.59,658.306 1463.46,652.005 1467.32,645.974 1471.19,640.216 1475.06,634.738 1478.92,629.541 
  1482.79,624.63 1486.66,620.009 1490.52,615.678 1494.39,611.642 1498.26,607.902 1502.12,604.46 1505.99,601.317 1509.86,598.475 1513.72,595.933 1517.59,593.691 
  1521.46,591.751 1525.32,590.111 1529.19,588.771 1533.05,587.728 1536.92,586.983 1540.79,586.532 1544.65,586.373 1548.52,586.504 1552.39,586.922 1556.25,587.622 
  1560.12,588.603 1563.99,589.858 1567.85,591.385 1571.72,593.177 1575.59,595.23 1579.45,597.538 1583.32,600.096 1587.19,602.907 1591.05,605.971 1594.92,609.291 
  1598.79,612.869 1602.65,616.703 1606.52,620.794 1610.39,625.141 1614.25,629.742 1618.12,634.595 1621.98,639.698 1625.85,645.047 1629.72,650.637 1633.58,656.465 
  1637.45,662.524 1641.32,668.809 1645.18,675.313 1649.05,682.028 1652.92,688.948 1656.78,696.063 1660.65,703.363 1664.52,710.841 1668.38,718.483 1672.25,726.281 
  1676.12,734.222 1679.98,742.293 1683.85,750.482 1687.72,758.774 1691.58,767.158 1695.45,775.62 1699.31,784.139 1703.18,792.692 1707.05,801.258 1710.91,809.816 
  1714.78,818.345 1718.65,826.828 1722.51,835.247 1726.38,843.586 1730.25,851.828 1734.11,859.96 1737.98,867.969 1741.85,875.842 1745.71,883.567 1749.58,891.136 
  1753.45,898.537 1757.31,905.764 1761.18,912.81 1765.05,919.668 1768.91,926.333 1772.78,932.802 1776.65,939.072 1780.51,945.14 1784.38,951.007 1788.24,956.672 
  1792.11,962.131 1795.98,967.38 1799.84,972.419 1803.71,977.248 1807.58,981.868 1811.44,986.278 1815.31,990.479 1819.18,994.472 1823.04,998.256 1826.91,1001.83 
  1830.78,1005.2 1834.64,1008.37 1838.51,1011.33 1842.38,1014.09 1846.24,1016.64 1850.11,1018.99 1853.98,1021.15 1857.84,1023.11 1861.71,1024.87 1865.58,1026.43 
  1869.44,1027.81 1873.31,1028.99 1877.17,1029.99 1881.04,1030.8 1884.91,1031.43 1888.77,1031.88 1892.64,1032.15 1896.51,1032.24 1900.37,1032.16 1904.24,1031.91 
  1908.11,1031.49 1911.97,1030.9 1915.84,1030.14 1919.71,1029.21 1923.57,1028.13 1927.44,1026.88 1931.31,1025.47 1935.17,1023.91 1939.04,1022.19 1942.91,1020.32 
  1946.77,1018.3 1950.64,1016.13 1954.5,1013.82 1958.37,1011.36 1962.24,1008.77 1966.1,1006.04 1969.97,1003.17 1973.84,1000.18 1977.7,997.051 1981.57,993.8 
  1985.44,990.424 1989.3,986.929 1993.17,983.317 1997.04,979.59 2000.9,975.752 2004.77,971.806 2008.64,967.756 2012.5,963.604 2016.37,959.354 2020.24,955.01 
  2024.1,950.574 2027.97,946.051 2031.84,941.443 2035.7,936.755 2039.57,931.99 2043.43,927.152 2047.3,922.244 2051.17,917.272 2055.03,912.239 2058.9,907.15 
  2062.77,902.008 2066.63,896.818 2070.5,891.586 2074.37,886.315 2078.23,881.01 2082.1,875.677 2085.97,870.32 2089.83,864.944 2093.7,859.554 2097.57,854.157 
  2101.43,848.756 2105.3,843.358 2109.17,837.969 2113.03,832.591 2116.9,827.23 2120.77,821.894 2124.63,816.589 2128.5,811.323 2132.36,806.101 2136.23,800.93 
  2140.1,795.817 2143.96,790.768 2147.83,785.79 2151.7,780.888 2155.56,776.069 2159.43,771.34 2163.3,766.705 2167.16,762.172 2171.03,757.745 2174.9,753.431 
  2178.76,749.236 2182.63,745.164 2186.5,741.223 2190.36,737.416 2194.23,733.75 2198.1,730.23 2201.96,726.862 2205.83,723.649 2209.69,720.598 2213.56,717.714 
  2217.43,714.998 2221.29,712.453 2225.16,710.082 2229.03,707.889 2232.89,705.875 2236.76,704.043 2240.63,702.395 2244.49,700.933 2248.36,699.659 2252.23,698.573 
  2256.09,697.677 2259.96,696.971 2263.83,696.457 2267.69,696.134 2271.56,696.002 2275.43,696.061 2279.29,696.311 2283.16,696.751 2287.03,697.379 2290.89,698.196 
  2294.76,699.198 
  "/>
<polyline clip-path="url(#clip302)" style="stroke:#c271d2; stroke-linecap:butt; stroke-linejoin:round; stroke-width:8; stroke-opacity:1; fill:none" points="
  361.501,1361.96 365.367,1361.01 369.234,1359.68 373.1,1357.95 376.967,1355.84 380.833,1353.34 384.7,1350.45 388.566,1347.17 392.433,1343.51 396.299,1339.45 
  400.166,1335.01 404.032,1330.18 407.899,1324.97 411.765,1319.36 415.632,1313.38 419.498,1307.01 423.365,1300.27 427.231,1293.14 431.098,1285.64 434.964,1277.76 
  438.831,1269.51 442.698,1260.89 446.564,1251.9 450.431,1242.55 454.297,1232.85 458.164,1222.79 462.03,1212.38 465.897,1201.62 469.763,1190.53 473.63,1179.1 
  477.496,1167.34 481.363,1155.27 485.229,1142.88 489.096,1130.19 492.962,1117.2 496.829,1103.92 500.695,1090.36 504.562,1076.53 508.428,1062.44 512.295,1048.1 
  516.161,1033.53 520.028,1018.73 523.894,1003.72 527.761,988.511 531.627,973.117 535.494,957.549 539.36,941.82 543.227,925.946 547.093,909.942 550.96,893.828 
  554.826,877.621 558.693,861.342 562.559,845.013 566.426,828.655 570.293,812.293 574.159,795.951 578.026,779.657 581.892,763.438 585.759,747.318 589.625,731.328 
  593.492,715.511 597.358,699.905 601.225,684.545 605.091,669.461 608.958,654.679 612.824,640.222 616.691,626.107 620.557,612.35 624.424,598.96 628.29,585.944 
  632.157,573.304 636.023,561.037 639.89,549.139 643.756,537.615 647.623,526.468 651.489,515.698 655.356,505.304 659.222,495.287 663.089,485.646 666.955,476.382 
  670.822,467.497 674.688,458.99 678.555,450.863 682.421,443.118 686.288,435.756 690.154,428.78 694.021,422.191 697.887,415.992 701.754,410.185 705.621,404.774 
  709.487,399.759 713.354,395.142 717.22,390.925 721.087,387.113 724.953,383.707 728.82,380.71 732.686,378.123 736.553,375.949 740.419,374.19 744.286,372.847 
  748.152,371.922 752.019,371.416 755.885,371.33 759.752,371.664 763.618,372.421 767.485,373.599 771.351,375.201 775.218,377.224 779.084,379.671 782.951,382.54 
  786.817,385.832 790.684,389.545 794.55,393.68 798.417,398.236 802.283,403.212 806.15,408.606 810.016,414.417 813.883,420.641 817.749,427.275 821.616,434.317 
  825.482,441.761 829.349,449.604 833.216,457.84 837.082,466.463 840.949,475.468 844.815,484.848 848.682,494.596 852.548,504.705 856.415,515.166 860.281,525.972 
  864.148,537.112 868.014,548.578 871.881,560.357 875.747,572.428 879.614,584.781 883.48,597.402 887.347,610.276 891.213,623.388 895.08,636.722 898.946,650.259 
  902.813,663.98 906.679,677.864 910.546,691.888 914.412,706.029 918.279,720.263 922.145,734.562 926.012,748.9 929.878,763.247 933.745,777.574 937.611,791.848 
  941.478,806.038 945.344,820.108 949.211,834.024 953.077,847.747 956.944,861.249 960.811,874.519 964.677,887.552 968.544,900.342 972.41,912.884 976.277,925.172 
  980.143,937.203 984.01,948.972 987.876,960.476 991.743,971.712 995.609,982.677 999.476,993.369 1003.34,1003.79 1007.21,1013.93 1011.08,1023.79 1014.94,1033.38 
  1018.81,1042.69 1022.67,1051.73 1026.54,1060.49 1030.41,1068.97 1034.27,1077.18 1038.14,1085.12 1042.01,1092.79 1045.87,1100.17 1049.74,1107.28 1053.61,1114.11 
  1057.47,1120.65 1061.34,1126.91 1065.21,1132.88 1069.07,1138.55 1072.94,1143.94 1076.81,1149.03 1080.67,1153.82 1084.54,1158.32 1088.41,1162.52 1092.27,1166.42 
  1096.14,1170.01 1100.01,1173.31 1103.87,1176.29 1107.74,1178.97 1111.6,1181.35 1115.47,1183.41 1119.34,1185.16 1123.2,1186.61 1127.07,1187.74 1130.94,1188.55 
  1134.8,1189.06 1138.67,1189.25 1142.54,1189.13 1146.4,1188.69 1150.27,1187.93 1154.14,1186.87 1158,1185.48 1161.87,1183.78 1165.74,1181.77 1169.6,1179.44 
  1173.47,1176.81 1177.34,1173.86 1181.2,1170.6 1185.07,1167.04 1188.93,1163.18 1192.8,1159.01 1196.67,1154.55 1200.53,1149.79 1204.4,1144.74 1208.27,1139.4 
  1212.13,1133.78 1216,1127.88 1219.87,1121.71 1223.73,1115.26 1227.6,1108.55 1231.47,1101.58 1235.33,1094.35 1239.2,1086.87 1243.07,1079.16 1246.93,1071.21 
  1250.8,1063.04 1254.67,1054.65 1258.53,1046.05 1262.4,1037.25 1266.27,1028.26 1270.13,1019.1 1274,1009.76 1277.86,1000.27 1281.73,990.626 1285.6,980.848 
  1289.46,970.946 1293.33,960.931 1297.2,950.817 1301.06,940.617 1304.93,930.344 1308.8,920.011 1312.66,909.633 1316.53,899.224 1320.4,888.799 1324.26,878.365 
  1328.13,867.936 1332,857.526 1335.86,847.149 1339.73,836.82 1343.6,826.552 1347.46,816.358 1351.33,806.25 1355.2,796.24 1359.06,786.341 1362.93,776.562 
  1366.79,766.914 1370.66,757.408 1374.53,748.052 1378.39,738.856 1382.26,729.829 1386.13,720.978 1389.99,712.311 1393.86,703.834 1397.73,695.556 1401.59,687.48 
  1405.46,679.614 1409.33,671.962 1413.19,664.529 1417.06,657.324 1420.93,650.353 1424.79,643.617 1428.66,637.12 1432.53,630.863 1436.39,624.849 1440.26,619.079 
  1444.12,613.555 1447.99,608.279 1451.86,603.252 1455.72,598.476 1459.59,593.952 1463.46,589.681 1467.32,585.664 1471.19,581.903 1475.06,578.397 1478.92,575.149 
  1482.79,572.158 1486.66,569.424 1490.52,566.949 1494.39,564.733 1498.26,562.775 1502.12,561.077 1505.99,559.64 1509.86,558.465 1513.72,557.549 1517.59,556.894 
  1521.46,556.497 1525.32,556.359 1529.19,556.476 1533.05,556.849 1536.92,557.476 1540.79,558.354 1544.65,559.482 1548.52,560.858 1552.39,562.48 1556.25,564.344 
  1560.12,566.448 1563.99,568.79 1567.85,571.366 1571.72,574.174 1575.59,577.209 1579.45,580.469 1583.32,583.949 1587.19,587.646 1591.05,591.556 1594.92,595.674 
  1598.79,599.996 1602.65,604.517 1606.52,609.233 1610.39,614.138 1614.25,619.228 1618.12,624.497 1621.98,629.937 1625.85,635.539 1629.72,641.297 1633.58,647.203 
  1637.45,653.251 1641.32,659.434 1645.18,665.745 1649.05,672.177 1652.92,678.722 1656.78,685.374 1660.65,692.125 1664.52,698.968 1668.38,705.894 1672.25,712.897 
  1676.12,719.969 1679.98,727.101 1683.85,734.286 1687.72,741.516 1691.58,748.782 1695.45,756.077 1699.31,763.393 1703.18,770.719 1707.05,778.05 1710.91,785.374 
  1714.78,792.685 1718.65,799.973 1722.51,807.229 1726.38,814.445 1730.25,821.61 1734.11,828.717 1737.98,835.756 1741.85,842.721 1745.71,849.606 1749.58,856.405 
  1753.45,863.114 1757.31,869.727 1761.18,876.239 1765.05,882.646 1768.91,888.943 1772.78,895.126 1776.65,901.191 1780.51,907.133 1784.38,912.95 1788.24,918.638 
  1792.11,924.194 1795.98,929.615 1799.84,934.898 1803.71,940.041 1807.58,945.04 1811.44,949.895 1815.31,954.603 1819.18,959.163 1823.04,963.573 1826.91,967.832 
  1830.78,971.938 1834.64,975.891 1838.51,979.691 1842.38,983.335 1846.24,986.821 1850.11,990.147 1853.98,993.312 1857.84,996.314 1861.71,999.152 1865.58,1001.82 
  1869.44,1004.33 1873.31,1006.67 1877.17,1008.83 1881.04,1010.83 1884.91,1012.66 1888.77,1014.32 1892.64,1015.81 1896.51,1017.13 1900.37,1018.27 1904.24,1019.25 
  1908.11,1020.05 1911.97,1020.68 1915.84,1021.15 1919.71,1021.44 1923.57,1021.56 1927.44,1021.52 1931.31,1021.31 1935.17,1020.93 1939.04,1020.39 1942.91,1019.68 
  1946.77,1018.81 1950.64,1017.79 1954.5,1016.6 1958.37,1015.26 1962.24,1013.76 1966.1,1012.11 1969.97,1010.3 1973.84,1008.35 1977.7,1006.26 1981.57,1004.02 
  1985.44,1001.65 1989.3,999.135 1993.17,996.491 1997.04,993.718 2000.9,990.819 2004.77,987.8 2008.64,984.663 2012.5,981.413 2016.37,978.053 2020.24,974.59 
  2024.1,971.026 2027.97,967.367 2031.84,963.617 2035.7,959.783 2039.57,955.868 2043.43,951.879 2047.3,947.82 2051.17,943.697 2055.03,939.517 2058.9,935.285 
  2062.77,931.007 2066.63,926.688 2070.5,922.332 2074.37,917.947 2078.23,913.537 2082.1,909.108 2085.97,904.667 2089.83,900.218 2093.7,895.768 2097.57,891.321 
  2101.43,886.882 2105.3,882.458 2109.17,878.052 2113.03,873.669 2116.9,869.315 2120.77,864.992 2124.63,860.707 2128.5,856.463 2132.36,852.264 2136.23,848.114 
  2140.1,844.017 2143.96,839.976 2147.83,835.996 2151.7,832.08 2155.56,828.23 2159.43,824.45 2163.3,820.744 2167.16,817.112 2171.03,813.557 2174.9,810.081 
  2178.76,806.687 2182.63,803.375 2186.5,800.148 2190.36,797.007 2194.23,793.955 2198.1,790.993 2201.96,788.121 2205.83,785.343 2209.69,782.658 2213.56,780.067 
  2217.43,777.573 2221.29,775.176 2225.16,772.876 2229.03,770.674 2232.89,768.571 2236.76,766.567 2240.63,764.664 2244.49,762.86 2248.36,761.156 2252.23,759.552 
  2256.09,758.048 2259.96,756.644 2263.83,755.34 2267.69,754.135 2271.56,753.029 2275.43,752.022 2279.29,751.111 2283.16,750.299 2287.03,749.585 2290.89,748.968 
  2294.76,748.448 
  "/>
<path clip-path="url(#clip300)" d="
M1449.37 351.589 L2284.45 351.589 L2284.45 92.3886 L1449.37 92.3886  Z
  " fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<polyline clip-path="url(#clip300)" style="stroke:#000000; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1449.37,351.589 2284.45,351.589 2284.45,92.3886 1449.37,92.3886 1449.37,351.589 
  "/>
<polyline clip-path="url(#clip300)" style="stroke:#009af9; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1472.14,144.229 1608.76,144.229 
  "/>
<path clip-path="url(#clip300)" d="M1653.77 128.083 L1653.77 132.643 Q1651.11 131.37 1648.75 130.745 Q1646.39 130.12 1644.19 130.12 Q1640.37 130.12 1638.29 131.601 Q1636.23 133.083 1636.23 135.814 Q1636.23 138.106 1637.59 139.287 Q1638.98 140.444 1642.82 141.162 L1645.65 141.74 Q1650.88 142.736 1653.36 145.259 Q1655.86 147.759 1655.86 151.972 Q1655.86 156.995 1652.48 159.587 Q1649.12 162.18 1642.62 162.18 Q1640.16 162.18 1637.38 161.624 Q1634.63 161.069 1631.67 159.981 L1631.67 155.166 Q1634.51 156.763 1637.25 157.573 Q1639.98 158.384 1642.62 158.384 Q1646.62 158.384 1648.8 156.81 Q1650.97 155.236 1650.97 152.319 Q1650.97 149.773 1649.4 148.337 Q1647.85 146.902 1644.28 146.185 L1641.43 145.629 Q1636.2 144.587 1633.87 142.365 Q1631.53 140.143 1631.53 136.185 Q1631.53 131.601 1634.75 128.962 Q1637.99 126.324 1643.66 126.324 Q1646.09 126.324 1648.61 126.763 Q1651.13 127.203 1653.77 128.083 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1662.96 135.583 L1667.22 135.583 L1667.22 161.509 L1662.96 161.509 L1662.96 135.583 M1662.96 125.49 L1667.22 125.49 L1667.22 130.884 L1662.96 130.884 L1662.96 125.49 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1696.32 140.56 Q1697.92 137.689 1700.14 136.324 Q1702.36 134.958 1705.37 134.958 Q1709.42 134.958 1711.62 137.805 Q1713.82 140.629 1713.82 145.861 L1713.82 161.509 L1709.54 161.509 L1709.54 145.999 Q1709.54 142.273 1708.22 140.467 Q1706.9 138.662 1704.19 138.662 Q1700.88 138.662 1698.96 140.861 Q1697.04 143.06 1697.04 146.856 L1697.04 161.509 L1692.75 161.509 L1692.75 145.999 Q1692.75 142.249 1691.43 140.467 Q1690.12 138.662 1687.36 138.662 Q1684.1 138.662 1682.18 140.884 Q1680.25 143.083 1680.25 146.856 L1680.25 161.509 L1675.97 161.509 L1675.97 135.583 L1680.25 135.583 L1680.25 139.611 Q1681.71 137.226 1683.75 136.092 Q1685.79 134.958 1688.59 134.958 Q1691.41 134.958 1693.38 136.393 Q1695.37 137.828 1696.32 140.56 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1726.43 157.62 L1726.43 171.37 L1722.15 171.37 L1722.15 135.583 L1726.43 135.583 L1726.43 139.518 Q1727.78 137.203 1729.81 136.092 Q1731.87 134.958 1734.72 134.958 Q1739.44 134.958 1742.38 138.708 Q1745.35 142.458 1745.35 148.569 Q1745.35 154.68 1742.38 158.43 Q1739.44 162.18 1734.72 162.18 Q1731.87 162.18 1729.81 161.069 Q1727.78 159.935 1726.43 157.62 M1740.93 148.569 Q1740.93 143.87 1738.98 141.208 Q1737.06 138.523 1733.68 138.523 Q1730.3 138.523 1728.36 141.208 Q1726.43 143.87 1726.43 148.569 Q1726.43 153.268 1728.36 155.953 Q1730.3 158.615 1733.68 158.615 Q1737.06 158.615 1738.98 155.953 Q1740.93 153.268 1740.93 148.569 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1752.41 125.49 L1756.67 125.49 L1756.67 161.509 L1752.41 161.509 L1752.41 125.49 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1787.75 147.481 L1787.75 149.564 L1768.17 149.564 Q1768.45 153.962 1770.81 156.277 Q1773.19 158.569 1777.43 158.569 Q1779.88 158.569 1782.17 157.967 Q1784.49 157.365 1786.76 156.161 L1786.76 160.189 Q1784.47 161.161 1782.06 161.671 Q1779.65 162.18 1777.17 162.18 Q1770.97 162.18 1767.34 158.569 Q1763.73 154.958 1763.73 148.8 Q1763.73 142.435 1767.15 138.708 Q1770.6 134.958 1776.43 134.958 Q1781.67 134.958 1784.7 138.337 Q1787.75 141.694 1787.75 147.481 M1783.49 146.231 Q1783.45 142.736 1781.53 140.652 Q1779.63 138.569 1776.48 138.569 Q1772.92 138.569 1770.76 140.583 Q1768.63 142.597 1768.31 146.254 L1783.49 146.231 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1794.93 126.949 L1814.79 126.949 L1814.79 130.884 L1799.61 130.884 L1799.61 141.069 L1813.31 141.069 L1813.31 145.004 L1799.61 145.004 L1799.61 161.509 L1794.93 161.509 L1794.93 126.949 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1822.2 126.949 L1829.17 126.949 L1837.98 150.467 L1846.85 126.949 L1853.82 126.949 L1853.82 161.509 L1849.26 161.509 L1849.26 131.162 L1840.35 154.865 L1835.65 154.865 L1826.73 131.162 L1826.73 161.509 L1822.2 161.509 L1822.2 126.949 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1862.57 126.949 L1867.27 126.949 L1867.27 147.944 Q1867.27 153.499 1869.28 155.953 Q1871.29 158.384 1875.81 158.384 Q1880.3 158.384 1882.31 155.953 Q1884.33 153.499 1884.33 147.944 L1884.33 126.949 L1889.03 126.949 L1889.03 148.523 Q1889.03 155.282 1885.67 158.731 Q1882.34 162.18 1875.81 162.18 Q1869.26 162.18 1865.9 158.731 Q1862.57 155.282 1862.57 148.523 L1862.57 126.949 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip300)" style="stroke:#e26f46; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1472.14,196.069 1608.76,196.069 
  "/>
<path clip-path="url(#clip300)" d="M1647.92 197.145 Q1649.42 197.654 1650.83 199.321 Q1652.27 200.988 1653.7 203.904 L1658.45 213.349 L1653.43 213.349 L1649 204.483 Q1647.29 201.011 1645.67 199.876 Q1644.07 198.742 1641.3 198.742 L1636.2 198.742 L1636.2 213.349 L1631.53 213.349 L1631.53 178.789 L1642.08 178.789 Q1648.01 178.789 1650.93 181.265 Q1653.84 183.742 1653.84 188.742 Q1653.84 192.006 1652.31 194.159 Q1650.81 196.312 1647.92 197.145 M1636.2 182.631 L1636.2 194.9 L1642.08 194.9 Q1645.46 194.9 1647.18 193.349 Q1648.91 191.775 1648.91 188.742 Q1648.91 185.71 1647.18 184.182 Q1645.46 182.631 1642.08 182.631 L1636.2 182.631 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1684.33 199.321 L1684.33 201.404 L1664.74 201.404 Q1665.02 205.802 1667.38 208.117 Q1669.77 210.409 1674 210.409 Q1676.46 210.409 1678.75 209.807 Q1681.06 209.205 1683.33 208.001 L1683.33 212.029 Q1681.04 213.001 1678.63 213.511 Q1676.23 214.02 1673.75 214.02 Q1667.55 214.02 1663.91 210.409 Q1660.3 206.798 1660.3 200.64 Q1660.3 194.275 1663.73 190.548 Q1667.18 186.798 1673.01 186.798 Q1678.24 186.798 1681.27 190.177 Q1684.33 193.534 1684.33 199.321 M1680.07 198.071 Q1680.02 194.576 1678.1 192.492 Q1676.2 190.409 1673.06 190.409 Q1669.49 190.409 1667.34 192.423 Q1665.21 194.437 1664.88 198.094 L1680.07 198.071 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1703.1 200.316 Q1697.94 200.316 1695.95 201.497 Q1693.96 202.677 1693.96 205.525 Q1693.96 207.793 1695.44 209.136 Q1696.94 210.455 1699.51 210.455 Q1703.05 210.455 1705.18 207.955 Q1707.34 205.432 1707.34 201.265 L1707.34 200.316 L1703.1 200.316 M1711.6 198.557 L1711.6 213.349 L1707.34 213.349 L1707.34 209.413 Q1705.88 211.775 1703.7 212.909 Q1701.53 214.02 1698.38 214.02 Q1694.4 214.02 1692.04 211.798 Q1689.7 209.552 1689.7 205.802 Q1689.7 201.427 1692.62 199.205 Q1695.55 196.983 1701.37 196.983 L1707.34 196.983 L1707.34 196.566 Q1707.34 193.627 1705.39 192.029 Q1703.47 190.409 1699.98 190.409 Q1697.75 190.409 1695.65 190.941 Q1693.54 191.474 1691.6 192.539 L1691.6 188.603 Q1693.93 187.701 1696.13 187.261 Q1698.33 186.798 1700.42 186.798 Q1706.04 186.798 1708.82 189.714 Q1711.6 192.631 1711.6 198.557 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1720.37 177.33 L1724.63 177.33 L1724.63 213.349 L1720.37 213.349 L1720.37 177.33 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1733.73 178.789 L1753.59 178.789 L1753.59 182.724 L1738.4 182.724 L1738.4 192.909 L1752.11 192.909 L1752.11 196.844 L1738.4 196.844 L1738.4 213.349 L1733.73 213.349 L1733.73 178.789 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1760.99 178.789 L1767.96 178.789 L1776.78 202.307 L1785.65 178.789 L1792.61 178.789 L1792.61 213.349 L1788.05 213.349 L1788.05 183.002 L1779.14 206.705 L1774.44 206.705 L1765.53 183.002 L1765.53 213.349 L1760.99 213.349 L1760.99 178.789 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1801.36 178.789 L1806.06 178.789 L1806.06 199.784 Q1806.06 205.339 1808.08 207.793 Q1810.09 210.224 1814.61 210.224 Q1819.1 210.224 1821.11 207.793 Q1823.12 205.339 1823.12 199.784 L1823.12 178.789 L1827.82 178.789 L1827.82 200.363 Q1827.82 207.122 1824.47 210.571 Q1821.13 214.02 1814.61 214.02 Q1808.05 214.02 1804.7 210.571 Q1801.36 207.122 1801.36 200.363 L1801.36 178.789 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip300)" style="stroke:#3da44d; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1472.14,247.909 1608.76,247.909 
  "/>
<path clip-path="url(#clip300)" d="M1631.53 230.629 L1637.82 230.629 L1653.15 259.541 L1653.15 230.629 L1657.68 230.629 L1657.68 265.189 L1651.39 265.189 L1636.06 236.277 L1636.06 265.189 L1631.53 265.189 L1631.53 230.629 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1688.98 251.161 L1688.98 253.244 L1669.4 253.244 Q1669.68 257.642 1672.04 259.957 Q1674.42 262.249 1678.66 262.249 Q1681.11 262.249 1683.4 261.647 Q1685.72 261.045 1687.99 259.841 L1687.99 263.869 Q1685.69 264.841 1683.29 265.351 Q1680.88 265.86 1678.4 265.86 Q1672.2 265.86 1668.56 262.249 Q1664.95 258.638 1664.95 252.48 Q1664.95 246.115 1668.38 242.388 Q1671.83 238.638 1677.66 238.638 Q1682.89 238.638 1685.93 242.017 Q1688.98 245.374 1688.98 251.161 M1684.72 249.911 Q1684.68 246.416 1682.75 244.332 Q1680.86 242.249 1677.71 242.249 Q1674.14 242.249 1671.99 244.263 Q1669.86 246.277 1669.54 249.934 L1684.72 249.911 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1695.53 254.957 L1695.53 239.263 L1699.79 239.263 L1699.79 254.795 Q1699.79 258.476 1701.23 260.328 Q1702.66 262.156 1705.53 262.156 Q1708.98 262.156 1710.97 259.957 Q1712.99 257.758 1712.99 253.962 L1712.99 239.263 L1717.24 239.263 L1717.24 265.189 L1712.99 265.189 L1712.99 261.207 Q1711.43 263.568 1709.37 264.726 Q1707.34 265.86 1704.63 265.86 Q1700.16 265.86 1697.85 263.082 Q1695.53 260.304 1695.53 254.957 M1706.25 238.638 L1706.25 238.638 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1741.04 243.244 Q1740.32 242.828 1739.47 242.642 Q1738.63 242.434 1737.61 242.434 Q1734 242.434 1732.06 244.795 Q1730.14 247.133 1730.14 251.531 L1730.14 265.189 L1725.86 265.189 L1725.86 239.263 L1730.14 239.263 L1730.14 243.291 Q1731.48 240.929 1733.63 239.795 Q1735.79 238.638 1738.86 238.638 Q1739.3 238.638 1739.84 238.707 Q1740.37 238.754 1741.02 238.869 L1741.04 243.244 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1757.29 252.156 Q1752.13 252.156 1750.14 253.337 Q1748.15 254.517 1748.15 257.365 Q1748.15 259.633 1749.63 260.976 Q1751.13 262.295 1753.7 262.295 Q1757.24 262.295 1759.37 259.795 Q1761.53 257.272 1761.53 253.105 L1761.53 252.156 L1757.29 252.156 M1765.79 250.397 L1765.79 265.189 L1761.53 265.189 L1761.53 261.253 Q1760.07 263.615 1757.89 264.749 Q1755.72 265.86 1752.57 265.86 Q1748.59 265.86 1746.23 263.638 Q1743.89 261.392 1743.89 257.642 Q1743.89 253.267 1746.8 251.045 Q1749.74 248.823 1755.55 248.823 L1761.53 248.823 L1761.53 248.406 Q1761.53 245.467 1759.58 243.869 Q1757.66 242.249 1754.17 242.249 Q1751.94 242.249 1749.84 242.781 Q1747.73 243.314 1745.79 244.379 L1745.79 240.443 Q1748.12 239.541 1750.32 239.101 Q1752.52 238.638 1754.61 238.638 Q1760.23 238.638 1763.01 241.554 Q1765.79 244.471 1765.79 250.397 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1774.56 229.17 L1778.82 229.17 L1778.82 265.189 L1774.56 265.189 L1774.56 229.17 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1787.92 230.629 L1807.78 230.629 L1807.78 234.564 L1792.59 234.564 L1792.59 244.749 L1806.3 244.749 L1806.3 248.684 L1792.59 248.684 L1792.59 265.189 L1787.92 265.189 L1787.92 230.629 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1815.18 230.629 L1822.15 230.629 L1830.97 254.147 L1839.84 230.629 L1846.8 230.629 L1846.8 265.189 L1842.24 265.189 L1842.24 234.842 L1833.33 258.545 L1828.63 258.545 L1819.72 234.842 L1819.72 265.189 L1815.18 265.189 L1815.18 230.629 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1855.55 230.629 L1860.25 230.629 L1860.25 251.624 Q1860.25 257.179 1862.27 259.633 Q1864.28 262.064 1868.79 262.064 Q1873.29 262.064 1875.3 259.633 Q1877.31 257.179 1877.31 251.624 L1877.31 230.629 L1882.01 230.629 L1882.01 252.203 Q1882.01 258.962 1878.66 262.411 Q1875.32 265.86 1868.79 265.86 Q1862.24 265.86 1858.89 262.411 Q1855.55 258.962 1855.55 252.203 L1855.55 230.629 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1915.9 229.217 Q1912.8 234.541 1911.29 239.749 Q1909.79 244.957 1909.79 250.304 Q1909.79 255.652 1911.29 260.906 Q1912.82 266.138 1915.9 271.439 L1912.2 271.439 Q1908.73 265.999 1906.99 260.744 Q1905.28 255.49 1905.28 250.304 Q1905.28 245.142 1906.99 239.911 Q1908.7 234.68 1912.2 229.217 L1915.9 229.217 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1938.93 246.554 Q1942.29 247.272 1944.16 249.541 Q1946.06 251.809 1946.06 255.142 Q1946.06 260.258 1942.54 263.059 Q1939.03 265.86 1932.54 265.86 Q1930.37 265.86 1928.05 265.42 Q1925.76 265.003 1923.31 264.147 L1923.31 259.633 Q1925.25 260.767 1927.57 261.346 Q1929.88 261.925 1932.41 261.925 Q1936.8 261.925 1939.1 260.189 Q1941.41 258.453 1941.41 255.142 Q1941.41 252.087 1939.26 250.374 Q1937.13 248.638 1933.31 248.638 L1929.28 248.638 L1929.28 244.795 L1933.49 244.795 Q1936.94 244.795 1938.77 243.429 Q1940.6 242.041 1940.6 239.448 Q1940.6 236.786 1938.7 235.374 Q1936.83 233.939 1933.31 233.939 Q1931.39 233.939 1929.19 234.355 Q1926.99 234.772 1924.35 235.652 L1924.35 231.485 Q1927.01 230.744 1929.33 230.374 Q1931.66 230.004 1933.72 230.004 Q1939.05 230.004 1942.15 232.434 Q1945.25 234.842 1945.25 238.962 Q1945.25 241.832 1943.61 243.823 Q1941.97 245.791 1938.93 246.554 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1964.93 233.707 Q1961.32 233.707 1959.49 237.272 Q1957.68 240.814 1957.68 247.943 Q1957.68 255.05 1959.49 258.615 Q1961.32 262.156 1964.93 262.156 Q1968.56 262.156 1970.37 258.615 Q1972.2 255.05 1972.2 247.943 Q1972.2 240.814 1970.37 237.272 Q1968.56 233.707 1964.93 233.707 M1964.93 230.004 Q1970.74 230.004 1973.79 234.61 Q1976.87 239.193 1976.87 247.943 Q1976.87 256.67 1973.79 261.277 Q1970.74 265.86 1964.93 265.86 Q1959.12 265.86 1956.04 261.277 Q1952.98 256.67 1952.98 247.943 Q1952.98 239.193 1956.04 234.61 Q1959.12 230.004 1964.93 230.004 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1995.09 233.707 Q1991.48 233.707 1989.65 237.272 Q1987.85 240.814 1987.85 247.943 Q1987.85 255.05 1989.65 258.615 Q1991.48 262.156 1995.09 262.156 Q1998.72 262.156 2000.53 258.615 Q2002.36 255.05 2002.36 247.943 Q2002.36 240.814 2000.53 237.272 Q1998.72 233.707 1995.09 233.707 M1995.09 230.004 Q2000.9 230.004 2003.96 234.61 Q2007.03 239.193 2007.03 247.943 Q2007.03 256.67 2003.96 261.277 Q2000.9 265.86 1995.09 265.86 Q1989.28 265.86 1986.2 261.277 Q1983.15 256.67 1983.15 247.943 Q1983.15 239.193 1986.2 234.61 Q1989.28 230.004 1995.09 230.004 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M2051.9 251.161 L2051.9 253.244 L2032.31 253.244 Q2032.59 257.642 2034.95 259.957 Q2037.34 262.249 2041.57 262.249 Q2044.03 262.249 2046.32 261.647 Q2048.63 261.045 2050.9 259.841 L2050.9 263.869 Q2048.61 264.841 2046.2 265.351 Q2043.79 265.86 2041.32 265.86 Q2035.11 265.86 2031.48 262.249 Q2027.87 258.638 2027.87 252.48 Q2027.87 246.115 2031.29 242.388 Q2034.74 238.638 2040.58 238.638 Q2045.81 238.638 2048.84 242.017 Q2051.9 245.374 2051.9 251.161 M2047.64 249.911 Q2047.59 246.416 2045.67 244.332 Q2043.77 242.249 2040.62 242.249 Q2037.06 242.249 2034.9 244.263 Q2032.78 246.277 2032.45 249.934 L2047.64 249.911 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M2063.01 261.3 L2063.01 275.05 L2058.72 275.05 L2058.72 239.263 L2063.01 239.263 L2063.01 243.198 Q2064.35 240.883 2066.39 239.772 Q2068.45 238.638 2071.29 238.638 Q2076.02 238.638 2078.96 242.388 Q2081.92 246.138 2081.92 252.249 Q2081.92 258.36 2078.96 262.11 Q2076.02 265.86 2071.29 265.86 Q2068.45 265.86 2066.39 264.749 Q2064.35 263.615 2063.01 261.3 M2077.5 252.249 Q2077.5 247.55 2075.55 244.888 Q2073.63 242.203 2070.25 242.203 Q2066.87 242.203 2064.93 244.888 Q2063.01 247.55 2063.01 252.249 Q2063.01 256.948 2064.93 259.633 Q2066.87 262.295 2070.25 262.295 Q2073.63 262.295 2075.55 259.633 Q2077.5 256.948 2077.5 252.249 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M2099.03 242.249 Q2095.6 242.249 2093.61 244.934 Q2091.62 247.596 2091.62 252.249 Q2091.62 256.902 2093.59 259.587 Q2095.58 262.249 2099.03 262.249 Q2102.43 262.249 2104.42 259.564 Q2106.41 256.878 2106.41 252.249 Q2106.41 247.642 2104.42 244.957 Q2102.43 242.249 2099.03 242.249 M2099.03 238.638 Q2104.58 238.638 2107.75 242.249 Q2110.92 245.86 2110.92 252.249 Q2110.92 258.615 2107.75 262.249 Q2104.58 265.86 2099.03 265.86 Q2093.45 265.86 2090.28 262.249 Q2087.13 258.615 2087.13 252.249 Q2087.13 245.86 2090.28 242.249 Q2093.45 238.638 2099.03 238.638 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M2136.64 240.258 L2136.64 244.24 Q2134.84 243.244 2133.01 242.758 Q2131.2 242.249 2129.35 242.249 Q2125.21 242.249 2122.91 244.888 Q2120.62 247.504 2120.62 252.249 Q2120.62 256.994 2122.91 259.633 Q2125.21 262.249 2129.35 262.249 Q2131.2 262.249 2133.01 261.763 Q2134.84 261.253 2136.64 260.258 L2136.64 264.193 Q2134.86 265.027 2132.94 265.443 Q2131.04 265.86 2128.89 265.86 Q2123.03 265.86 2119.58 262.179 Q2116.13 258.499 2116.13 252.249 Q2116.13 245.906 2119.6 242.272 Q2123.1 238.638 2129.16 238.638 Q2131.13 238.638 2133.01 239.055 Q2134.88 239.448 2136.64 240.258 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M2165.6 249.541 L2165.6 265.189 L2161.34 265.189 L2161.34 249.679 Q2161.34 245.999 2159.9 244.17 Q2158.47 242.342 2155.6 242.342 Q2152.15 242.342 2150.16 244.541 Q2148.17 246.74 2148.17 250.536 L2148.17 265.189 L2143.89 265.189 L2143.89 229.17 L2148.17 229.17 L2148.17 243.291 Q2149.7 240.953 2151.76 239.795 Q2153.84 238.638 2156.55 238.638 Q2161.02 238.638 2163.31 241.416 Q2165.6 244.17 2165.6 249.541 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M2190.62 240.027 L2190.62 244.054 Q2188.82 243.129 2186.87 242.666 Q2184.93 242.203 2182.84 242.203 Q2179.67 242.203 2178.08 243.175 Q2176.5 244.147 2176.5 246.091 Q2176.5 247.573 2177.64 248.429 Q2178.77 249.263 2182.2 250.027 L2183.65 250.351 Q2188.19 251.323 2190.09 253.105 Q2192.01 254.865 2192.01 258.036 Q2192.01 261.647 2189.14 263.753 Q2186.29 265.86 2181.29 265.86 Q2179.21 265.86 2176.94 265.443 Q2174.7 265.05 2172.2 264.24 L2172.2 259.841 Q2174.56 261.068 2176.85 261.693 Q2179.14 262.295 2181.39 262.295 Q2184.39 262.295 2186.02 261.277 Q2187.64 260.235 2187.64 258.36 Q2187.64 256.624 2186.46 255.698 Q2185.3 254.772 2181.34 253.916 L2179.86 253.568 Q2175.9 252.735 2174.14 251.022 Q2172.38 249.286 2172.38 246.277 Q2172.38 242.619 2174.97 240.629 Q2177.57 238.638 2182.33 238.638 Q2184.7 238.638 2186.78 238.985 Q2188.86 239.332 2190.62 240.027 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M2198.12 229.217 L2201.83 229.217 Q2205.3 234.68 2207.01 239.911 Q2208.75 245.142 2208.75 250.304 Q2208.75 255.49 2207.01 260.744 Q2205.3 265.999 2201.83 271.439 L2198.12 271.439 Q2201.2 266.138 2202.71 260.906 Q2204.23 255.652 2204.23 250.304 Q2204.23 244.957 2202.71 239.749 Q2201.2 234.541 2198.12 229.217 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><polyline clip-path="url(#clip300)" style="stroke:#c271d2; stroke-linecap:butt; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="
  1472.14,299.749 1608.76,299.749 
  "/>
<path clip-path="url(#clip300)" d="M1631.53 282.469 L1637.82 282.469 L1653.15 311.381 L1653.15 282.469 L1657.68 282.469 L1657.68 317.029 L1651.39 317.029 L1636.06 288.117 L1636.06 317.029 L1631.53 317.029 L1631.53 282.469 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1688.98 303.001 L1688.98 305.084 L1669.4 305.084 Q1669.68 309.482 1672.04 311.797 Q1674.42 314.089 1678.66 314.089 Q1681.11 314.089 1683.4 313.487 Q1685.72 312.885 1687.99 311.681 L1687.99 315.709 Q1685.69 316.681 1683.29 317.191 Q1680.88 317.7 1678.4 317.7 Q1672.2 317.7 1668.56 314.089 Q1664.95 310.478 1664.95 304.32 Q1664.95 297.955 1668.38 294.228 Q1671.83 290.478 1677.66 290.478 Q1682.89 290.478 1685.93 293.857 Q1688.98 297.214 1688.98 303.001 M1684.72 301.751 Q1684.68 298.256 1682.75 296.172 Q1680.86 294.089 1677.71 294.089 Q1674.14 294.089 1671.99 296.103 Q1669.86 298.117 1669.54 301.774 L1684.72 301.751 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1695.53 306.797 L1695.53 291.103 L1699.79 291.103 L1699.79 306.635 Q1699.79 310.316 1701.23 312.168 Q1702.66 313.996 1705.53 313.996 Q1708.98 313.996 1710.97 311.797 Q1712.99 309.598 1712.99 305.802 L1712.99 291.103 L1717.24 291.103 L1717.24 317.029 L1712.99 317.029 L1712.99 313.047 Q1711.43 315.408 1709.37 316.566 Q1707.34 317.7 1704.63 317.7 Q1700.16 317.7 1697.85 314.922 Q1695.53 312.144 1695.53 306.797 M1706.25 290.478 L1706.25 290.478 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1741.04 295.084 Q1740.32 294.668 1739.47 294.482 Q1738.63 294.274 1737.61 294.274 Q1734 294.274 1732.06 296.635 Q1730.14 298.973 1730.14 303.371 L1730.14 317.029 L1725.86 317.029 L1725.86 291.103 L1730.14 291.103 L1730.14 295.131 Q1731.48 292.769 1733.63 291.635 Q1735.79 290.478 1738.86 290.478 Q1739.3 290.478 1739.84 290.547 Q1740.37 290.594 1741.02 290.709 L1741.04 295.084 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1757.29 303.996 Q1752.13 303.996 1750.14 305.177 Q1748.15 306.357 1748.15 309.205 Q1748.15 311.473 1749.63 312.816 Q1751.13 314.135 1753.7 314.135 Q1757.24 314.135 1759.37 311.635 Q1761.53 309.112 1761.53 304.945 L1761.53 303.996 L1757.29 303.996 M1765.79 302.237 L1765.79 317.029 L1761.53 317.029 L1761.53 313.093 Q1760.07 315.455 1757.89 316.589 Q1755.72 317.7 1752.57 317.7 Q1748.59 317.7 1746.23 315.478 Q1743.89 313.232 1743.89 309.482 Q1743.89 305.107 1746.8 302.885 Q1749.74 300.663 1755.55 300.663 L1761.53 300.663 L1761.53 300.246 Q1761.53 297.307 1759.58 295.709 Q1757.66 294.089 1754.17 294.089 Q1751.94 294.089 1749.84 294.621 Q1747.73 295.154 1745.79 296.219 L1745.79 292.283 Q1748.12 291.381 1750.32 290.941 Q1752.52 290.478 1754.61 290.478 Q1760.23 290.478 1763.01 293.394 Q1765.79 296.311 1765.79 302.237 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1774.56 281.01 L1778.82 281.01 L1778.82 317.029 L1774.56 317.029 L1774.56 281.01 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1787.92 282.469 L1807.78 282.469 L1807.78 286.404 L1792.59 286.404 L1792.59 296.589 L1806.3 296.589 L1806.3 300.524 L1792.59 300.524 L1792.59 317.029 L1787.92 317.029 L1787.92 282.469 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1815.18 282.469 L1822.15 282.469 L1830.97 305.987 L1839.84 282.469 L1846.8 282.469 L1846.8 317.029 L1842.24 317.029 L1842.24 286.682 L1833.33 310.385 L1828.63 310.385 L1819.72 286.682 L1819.72 317.029 L1815.18 317.029 L1815.18 282.469 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1855.55 282.469 L1860.25 282.469 L1860.25 303.464 Q1860.25 309.019 1862.27 311.473 Q1864.28 313.904 1868.79 313.904 Q1873.29 313.904 1875.3 311.473 Q1877.31 309.019 1877.31 303.464 L1877.31 282.469 L1882.01 282.469 L1882.01 304.043 Q1882.01 310.802 1878.66 314.251 Q1875.32 317.7 1868.79 317.7 Q1862.24 317.7 1858.89 314.251 Q1855.55 310.802 1855.55 304.043 L1855.55 282.469 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1915.9 281.057 Q1912.8 286.381 1911.29 291.589 Q1909.79 296.797 1909.79 302.144 Q1909.79 307.492 1911.29 312.746 Q1912.82 317.978 1915.9 323.279 L1912.2 323.279 Q1908.73 317.839 1906.99 312.584 Q1905.28 307.33 1905.28 302.144 Q1905.28 296.982 1906.99 291.751 Q1908.7 286.52 1912.2 281.057 L1915.9 281.057 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1925.58 313.093 L1933.22 313.093 L1933.22 286.728 L1924.91 288.395 L1924.91 284.135 L1933.17 282.469 L1937.85 282.469 L1937.85 313.093 L1945.48 313.093 L1945.48 317.029 L1925.58 317.029 L1925.58 313.093 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1964.93 285.547 Q1961.32 285.547 1959.49 289.112 Q1957.68 292.654 1957.68 299.783 Q1957.68 306.89 1959.49 310.455 Q1961.32 313.996 1964.93 313.996 Q1968.56 313.996 1970.37 310.455 Q1972.2 306.89 1972.2 299.783 Q1972.2 292.654 1970.37 289.112 Q1968.56 285.547 1964.93 285.547 M1964.93 281.844 Q1970.74 281.844 1973.79 286.45 Q1976.87 291.033 1976.87 299.783 Q1976.87 308.51 1973.79 313.117 Q1970.74 317.7 1964.93 317.7 Q1959.12 317.7 1956.04 313.117 Q1952.98 308.51 1952.98 299.783 Q1952.98 291.033 1956.04 286.45 Q1959.12 281.844 1964.93 281.844 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M1995.09 285.547 Q1991.48 285.547 1989.65 289.112 Q1987.85 292.654 1987.85 299.783 Q1987.85 306.89 1989.65 310.455 Q1991.48 313.996 1995.09 313.996 Q1998.72 313.996 2000.53 310.455 Q2002.36 306.89 2002.36 299.783 Q2002.36 292.654 2000.53 289.112 Q1998.72 285.547 1995.09 285.547 M1995.09 281.844 Q2000.9 281.844 2003.96 286.45 Q2007.03 291.033 2007.03 299.783 Q2007.03 308.51 2003.96 313.117 Q2000.9 317.7 1995.09 317.7 Q1989.28 317.7 1986.2 313.117 Q1983.15 308.51 1983.15 299.783 Q1983.15 291.033 1986.2 286.45 Q1989.28 281.844 1995.09 281.844 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M2025.25 285.547 Q2021.64 285.547 2019.81 289.112 Q2018.01 292.654 2018.01 299.783 Q2018.01 306.89 2019.81 310.455 Q2021.64 313.996 2025.25 313.996 Q2028.89 313.996 2030.69 310.455 Q2032.52 306.89 2032.52 299.783 Q2032.52 292.654 2030.69 289.112 Q2028.89 285.547 2025.25 285.547 M2025.25 281.844 Q2031.06 281.844 2034.12 286.45 Q2037.2 291.033 2037.2 299.783 Q2037.2 308.51 2034.12 313.117 Q2031.06 317.7 2025.25 317.7 Q2019.44 317.7 2016.36 313.117 Q2013.31 308.51 2013.31 299.783 Q2013.31 291.033 2016.36 286.45 Q2019.44 281.844 2025.25 281.844 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M2082.06 303.001 L2082.06 305.084 L2062.47 305.084 Q2062.75 309.482 2065.11 311.797 Q2067.5 314.089 2071.73 314.089 Q2074.19 314.089 2076.48 313.487 Q2078.79 312.885 2081.06 311.681 L2081.06 315.709 Q2078.77 316.681 2076.36 317.191 Q2073.96 317.7 2071.48 317.7 Q2065.28 317.7 2061.64 314.089 Q2058.03 310.478 2058.03 304.32 Q2058.03 297.955 2061.46 294.228 Q2064.9 290.478 2070.74 290.478 Q2075.97 290.478 2079 293.857 Q2082.06 297.214 2082.06 303.001 M2077.8 301.751 Q2077.75 298.256 2075.83 296.172 Q2073.93 294.089 2070.78 294.089 Q2067.22 294.089 2065.07 296.103 Q2062.94 298.117 2062.61 301.774 L2077.8 301.751 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M2093.17 313.14 L2093.17 326.89 L2088.89 326.89 L2088.89 291.103 L2093.17 291.103 L2093.17 295.038 Q2094.51 292.723 2096.55 291.612 Q2098.61 290.478 2101.46 290.478 Q2106.18 290.478 2109.12 294.228 Q2112.08 297.978 2112.08 304.089 Q2112.08 310.2 2109.12 313.95 Q2106.18 317.7 2101.46 317.7 Q2098.61 317.7 2096.55 316.589 Q2094.51 315.455 2093.17 313.14 M2107.66 304.089 Q2107.66 299.39 2105.71 296.728 Q2103.79 294.043 2100.41 294.043 Q2097.03 294.043 2095.09 296.728 Q2093.17 299.39 2093.17 304.089 Q2093.17 308.788 2095.09 311.473 Q2097.03 314.135 2100.41 314.135 Q2103.79 314.135 2105.71 311.473 Q2107.66 308.788 2107.66 304.089 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M2129.19 294.089 Q2125.76 294.089 2123.77 296.774 Q2121.78 299.436 2121.78 304.089 Q2121.78 308.742 2123.75 311.427 Q2125.74 314.089 2129.19 314.089 Q2132.59 314.089 2134.58 311.404 Q2136.57 308.718 2136.57 304.089 Q2136.57 299.482 2134.58 296.797 Q2132.59 294.089 2129.19 294.089 M2129.19 290.478 Q2134.74 290.478 2137.91 294.089 Q2141.08 297.7 2141.08 304.089 Q2141.08 310.455 2137.91 314.089 Q2134.74 317.7 2129.19 317.7 Q2123.61 317.7 2120.44 314.089 Q2117.29 310.455 2117.29 304.089 Q2117.29 297.7 2120.44 294.089 Q2123.61 290.478 2129.19 290.478 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M2166.8 292.098 L2166.8 296.08 Q2165 295.084 2163.17 294.598 Q2161.36 294.089 2159.51 294.089 Q2155.37 294.089 2153.08 296.728 Q2150.78 299.344 2150.78 304.089 Q2150.78 308.834 2153.08 311.473 Q2155.37 314.089 2159.51 314.089 Q2161.36 314.089 2163.17 313.603 Q2165 313.093 2166.8 312.098 L2166.8 316.033 Q2165.02 316.867 2163.1 317.283 Q2161.2 317.7 2159.05 317.7 Q2153.19 317.7 2149.74 314.019 Q2146.29 310.339 2146.29 304.089 Q2146.29 297.746 2149.77 294.112 Q2153.26 290.478 2159.33 290.478 Q2161.29 290.478 2163.17 290.895 Q2165.04 291.288 2166.8 292.098 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M2195.76 301.381 L2195.76 317.029 L2191.5 317.029 L2191.5 301.519 Q2191.5 297.839 2190.07 296.01 Q2188.63 294.182 2185.76 294.182 Q2182.31 294.182 2180.32 296.381 Q2178.33 298.58 2178.33 302.376 L2178.33 317.029 L2174.05 317.029 L2174.05 281.01 L2178.33 281.01 L2178.33 295.131 Q2179.86 292.793 2181.92 291.635 Q2184 290.478 2186.71 290.478 Q2191.18 290.478 2193.47 293.256 Q2195.76 296.01 2195.76 301.381 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M2220.78 291.867 L2220.78 295.894 Q2218.98 294.969 2217.03 294.506 Q2215.09 294.043 2213.01 294.043 Q2209.83 294.043 2208.24 295.015 Q2206.66 295.987 2206.66 297.931 Q2206.66 299.413 2207.8 300.269 Q2208.93 301.103 2212.36 301.867 L2213.82 302.191 Q2218.35 303.163 2220.25 304.945 Q2222.17 306.705 2222.17 309.876 Q2222.17 313.487 2219.3 315.593 Q2216.45 317.7 2211.45 317.7 Q2209.37 317.7 2207.1 317.283 Q2204.86 316.89 2202.36 316.08 L2202.36 311.681 Q2204.72 312.908 2207.01 313.533 Q2209.3 314.135 2211.55 314.135 Q2214.56 314.135 2216.18 313.117 Q2217.8 312.075 2217.8 310.2 Q2217.8 308.464 2216.62 307.538 Q2215.46 306.612 2211.5 305.756 L2210.02 305.408 Q2206.06 304.575 2204.3 302.862 Q2202.54 301.126 2202.54 298.117 Q2202.54 294.459 2205.14 292.469 Q2207.73 290.478 2212.5 290.478 Q2214.86 290.478 2216.94 290.825 Q2219.02 291.172 2220.78 291.867 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /><path clip-path="url(#clip300)" d="M2228.28 281.057 L2231.99 281.057 Q2235.46 286.52 2237.17 291.751 Q2238.91 296.982 2238.91 302.144 Q2238.91 307.33 2237.17 312.584 Q2235.46 317.839 2231.99 323.279 L2228.28 323.279 Q2231.36 317.978 2232.87 312.746 Q2234.39 307.492 2234.39 302.144 Q2234.39 296.797 2232.87 291.589 Q2231.36 286.381 2228.28 281.057 Z" fill="#000000" fill-rule="evenodd" fill-opacity="1" /></svg>



Finally, the FMU is cleaned-up.


```julia
fmiUnload(simpleFMU)
```

### Summary

Based on the plots, it can be seen that the NeuralFMU is able to adapt the friction model of the *realFMU*. After 300 runs, the curves do not overlap very well, but this can be achieved by longer training (1000 runs) or a better initialization.

### Source

[1] Tobias Thummerer, Lars Mikelsons and Josef Kircher. 2021. **NeuralFMU: towards structural integration of FMUs into neural networks.** Martin Sjölund, Lena Buffoni, Adrian Pop and Lennart Ochel (Ed.). Proceedings of 14th Modelica Conference 2021, Linköping, Sweden, September 20-24, 2021. Linköping University Electronic Press, Linköping (Linköping Electronic Conference Proceedings ; 181), 297-306. [DOI: 10.3384/ecp21181297](https://doi.org/10.3384/ecp21181297)

